Package: GGIR
Type: Package
Title: Raw Accelerometer Data Analysis
Version: 3.3-6
Date: 2026-04-21
Authors@R: c(person("Vincent T","van Hees",role=c("aut","cre"),
                    email="v.vanhees@accelting.com"),
             person("Jairo H","Migueles",role="aut",
                    email="jairo@jhmigueles.com",
                    comment = c(ORCID = "0000-0003-0366-6935")),
             person("Severine","Sabia",role="ctb"),
             person("Matthew R","Patterson",role="ctb"),
             person("Zhou","Fang",role="ctb"),
             person("Joe","Heywood",role="ctb"),
             person("Joan","Capdevila Pujol",role="ctb"),
             person("Lena","Kushleyeva",role="ctb"),
             person("Mathilde","Chen",role="ctb"),
             person("Manasa","Yerramalla",role="ctb"),
             person(given = "Patrick",family = "Bos",
                    role = "ctb", email = "egpbos@gmail.com",
                    comment = c(ORCID = "0000-0002-6033-960X")),
             person(given = "Taren",family = "Sanders", role = "ctb"),
             person("Chenxuan","Zhao",role="ctb"),
             person("Ian","Meneghel Danilevicz", role="ctb",
                    email="ian.meneghel-danilevicz @ inserm.fr",
                    comment = c(ORCID = "0000-0003-4541-0524")),
             person("Victor","Barreto Mesquita", role="ctb",
                    email="victormesquita40@hotmail.com"),
             person("Gaia","Segantin",role="ctb"),
             person("Xiaoyu","Zong",role="ctb", comment = c(ORCID = "0000-0001-5646-710X")),
             person("Medical Research Council UK",  role = c("cph", "fnd")),
             person("Accelting",  role = c("cph", "fnd")),
             person("French National Research Agency",  role = c("cph", "fnd"))
             )
Maintainer: Vincent T van Hees <v.vanhees@accelting.com>
Description: A tool to process and analyse data collected with wearable raw acceleration sensors as described in Migueles and colleagues (JMPB 2019), and van Hees and colleagues (JApplPhysiol 2014; PLoSONE 2015). The package has been developed and tested for binary data from 'GENEActiv' <https://activinsights.com/>, binary (.gt3x) and .csv-export data from  'Actigraph' <https://ametris.com/> devices, and binary (.cwa) and .csv-export data from 'Axivity' <https://axivity.com>. These devices are currently widely used in research on human daily physical activity. Further, the package can handle accelerometer data file from any other sensor brand providing that the data is stored in csv format. Also the package allows for external function embedding.
URL: https://github.com/wadpac/GGIR/, https://wadpac.github.io/GGIR/
BugReports: https://github.com/wadpac/GGIR/issues
License: Apache License (== 2.0) | file LICENSE
Suggests: testthat, covr, knitr, rmarkdown, actilifecounts, readxl
Imports: data.table, foreach, doParallel, signal, zoo, unisensR, ineq,
        methods, psych, irr, lubridate, GGIRread, ActCR, read.gt3x
Depends: stats, utils, R (>= 3.5)
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-04-21 07:21:43 UTC; vvanh
Author: Vincent T van Hees [aut, cre],
  Jairo H Migueles [aut] (ORCID: <https://orcid.org/0000-0003-0366-6935>),
  Severine Sabia [ctb],
  Matthew R Patterson [ctb],
  Zhou Fang [ctb],
  Joe Heywood [ctb],
  Joan Capdevila Pujol [ctb],
  Lena Kushleyeva [ctb],
  Mathilde Chen [ctb],
  Manasa Yerramalla [ctb],
  Patrick Bos [ctb] (ORCID: <https://orcid.org/0000-0002-6033-960X>),
  Taren Sanders [ctb],
  Chenxuan Zhao [ctb],
  Ian Meneghel Danilevicz [ctb] (ORCID:
    <https://orcid.org/0000-0003-4541-0524>),
  Victor Barreto Mesquita [ctb],
  Gaia Segantin [ctb],
  Xiaoyu Zong [ctb] (ORCID: <https://orcid.org/0000-0001-5646-710X>),
  Medical Research Council UK [cph, fnd],
  Accelting [cph, fnd],
  French National Research Agency [cph, fnd]
Repository: CRAN
Date/Publication: 2026-04-21 10:22:10 UTC
Built: R 4.5.2; ; 2026-04-21 19:21:04 UTC; unix

export(g.analyse, g.calibrate,
       g.getmeta, g.impute,
       g.inspectfile,
       g.plot, g.part1,
       g.part2, g.part3,
       g.part4, g.part5,
       GGIR, g.shell.GGIR, g.getbout,
       g.abr.day.names,
       g.applymetrics, g.create.sp.mat, g.detecmidnight,
       g.dotorcomma, g.extractheadervars,
       g.getM5L5,
       g.getstarttime, g.loadlog,
       g.plot5, g.readaccfile,
       g.report.part2, g.report.part4,
       g.report.part5, g.sib.det,
       g.sib.plot, 
       g.weardec, isfilelist,
       iso8601chartime2POSIX, is_this_a_dst_night,
       POSIXtime2iso8601, is.ISO8601,
       updateBlocksize, g.IVIS,
       createConfigFile, identify_levels,
       g.sib.sum, datadir2fnames,
       read.myacc.csv, g.part5.addfirstwake,
       g.part5.addsib, g.part5.definedays,
       g.part5.fixmissingnight, g.part5.onsetwaketiming,
       g.part5.wakesleepwindows, g.part5.savetimeseries,
       get_nw_clip_block_params, get_starttime_weekday_truncdata,
       ismovisens, g.readtemp_movisens,
       g.conv.actlog, g.fragmentation,
       g.intensitygradient, g.part5.handle_lux_extremes,
       g.part5.lux_persegment,
       g.sibreport, HASPT,
       HASIB, g.part4_extractid,
       CalcSleepRegularityIndex, load_params,
       check_params, extract_params,
       g.imputeTimegaps,
       tidyup_df, cosinor_IS_IV_Analyses,
       detect_nonwear_clipping, apply_cosinor_IS_IV_Analyses,
       ShellDoc2Vignette, parametersVignette,
       correctOlderMilestoneData,
       convertEpochData, appendRecords, extractID,
       g.part5_analyseSegment, g.part5_initialise_ts,
       g.part5.analyseRest, part6AlignIndividuals,
       part6PairwiseAggregation, g.part6, g.report.part6, 
       check_log, g.report.part5_dictionary, DFA, ABI, SSP, visualReport,
       splitRecords, getSplitNames, addSplitNames, getPart1BasicInfo,
       markerButtonForRest, aggregateEvent, inspect_binFile_brand,
       g.part3_correct_guider, g.part3_alignIndexVectors, filterNonwearNight)

importFrom("grDevices", "colors", "dev.off", "pdf",
           "rgb", "rainbow", "palette", "adjustcolor", "gray.colors", "dev.list", "png")
importFrom("graphics", "abline", "axis", "par", "plot", "plot.new",
           "rect","axis.POSIXct", "barplot", "box", "legend",
           "lines", "mtext", "text", "arrows", "segments", "layout")
importFrom("utils", "read.csv", "sessionInfo", "write.csv",
           "installed.packages", "setTxtProgressBar",
           "txtProgressBar", "write.table", "head",
           "memory.limit", "memory.size", "tail", "unzip", "lsf.str")
importFrom("stats", "aggregate.data.frame", "weighted.mean",
           "rnorm","median","aggregate","C", "lm.wfit",
           "quantile", "sd","coef", "lm", "embed", "na.pass", "na.omit",
            "residuals", "fitted", "reshape", "cor", "arima", "lm.fit")
import(data.table)
importFrom("methods", "is")
importFrom("utils", "available.packages", "packageVersion", "help", "read.table")

![](vignettes/GGIR-MASTERLOGO-RGB.png)

![GitHub Actions R-CMD-check](https://github.com/wadpac/GGIR/workflows/R-CMD-check/badge.svg)
[![codecov](https://codecov.io/gh/wadpac/GGIR/graph/badge.svg?token=p5Po3OpEPG)](https://app.codecov.io/gh/wadpac/GGIR)
[![](https://cranlogs.r-pkg.org/badges/last-month/GGIR)](https://cran.r-project.org/package=GGIR)

## Getting started:
The package [documentation](https://wadpac.github.io/GGIR/) provides extensive documentation on GGIR and its functionalities.

## Getting help:
For guidance on how to seek help see [this page](https://wadpac.github.io/GGIR/articles/chapter0_Support.html).

## Contribution guidelines:
We always welcome contributions to the package.
If you want to contribute to the development of GGIR, have a look at the [contribution guidelines](https://github.com/wadpac/GGIR/blob/master/CONTRIBUTING.md).

### Images usaged
The copyright of the GGIR logo as contained in the file vignettes/GGIR-MASTERLOGO-RGB.png lies with Accelting (Almere, The Netherlands), please contact v.vanhees@accelting.com to ask for permission to use this logo.

All other images in this repository are released under the Creative Commons Attribution 4.0 International (CC BY 4.0) license.

## ----echo=FALSE, out.width = "60%", out.extra='style="border: 0; padding:20px"', fig.alt="GGIR logo"----
knitr::include_graphics("GGIR-MASTERLOGO-RGB.png")

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(echo = TRUE)

## ----eval=FALSE---------------------------------------------------------------
# library(GGIR)
# GGIR(datadir="C:/mystudy/mydata",
#  outputdir="D:/myresults")

## ----out.width = "700px",echo=FALSE, fig.alt="Example visualreport"-----------
knitr::include_graphics("reportexample.jpg")

## ----echo=FALSE, out.width = "60%", out.extra='style="border: 0; padding:20px"', fig.alt="GGIR logo"----
knitr::include_graphics("GGIR-MASTERLOGO-RGB.png")

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(echo = TRUE)

## ----eval=FALSE---------------------------------------------------------------
# library(GGIR)
# data = read.myacc.csv(rmc.file = "C:/mystudy/mydata/datafile.csv",
#                rmc.nrow = Inf,
#                rmc.skip = 0,
#                rmc.dec = ".",
#                rmc.firstrow.acc = 2,
#                rmc.col.acc = 2:4,
#                rmc.col.temp = 5,
#                rmc.col.time=1,
#                rmc.unit.acc = "g",
#                rmc.unit.temp = "C",
#                rmc.unit.time = "POSIX",
#                rmc.format.time = "%d/%m/%Y %H:%M:%OS",
#                desiredtz = "Europe/London",
#                rmc.sf = 100)

## ----eval=FALSE---------------------------------------------------------------
# library(GGIR)
# GGIR(
#              mode=c(1,2,3,4,5),
#              datadir="C:/mystudy/mydata/datafile.csv",
#              outputdir="D:/myresults",
#              do.report=c(2,4,5),
#              #=====================
#              # read.myacc.csv arguments
#              #=====================
#              rmc.nrow = Inf,
#              rmc.dec = ".",
#              rmc.firstrow.acc = 2,
#              rmc.col.acc = 2:4,
#              rmc.col.temp = 5,
#              rmc.col.time=1,
#              rmc.unit.acc = "g",
#              rmc.unit.temp = "C",
#              rmc.unit.time = "POSIX",
#              rmc.format.time = "%d/%m/%Y %H:%M:%OS",
#              desiredtz = "Europe/London",
#              rmc.sf = 100,
#              rmc.noise = 0.013
# )

## ----echo=FALSE, out.width = "60%", out.extra='style="border: 0; padding:20px"', fig.alt="GGIR logo"----
knitr::include_graphics("GGIR-MASTERLOGO-RGB.png")

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  echo = TRUE,
  python.reticulate = FALSE
)

## ----eval=FALSE---------------------------------------------------------------
# calculateCounts = function(data=c(), parameters=c()) {
#   # data: 3 column matrix with acc data
#   # parameters: the sample rate of data
#   library("activityCounts")
#   if (ncol(data) == 4) data = data[,2:4]
#   mycounts = counts(data = data, hertz = parameters,
#                     x_axis = 1, y_axis = 2, z_axis = 3,
#                     start_time = Sys.time())
#   mycounts = mycounts[,2:4] #Note: do not provide timestamps to GGIR
#   return(mycounts)
# }

## ----eval=FALSE---------------------------------------------------------------
# source("~/calculateCounts.R")
# myfun =  list(FUN = calculateCounts,
#               parameters = 30,
#               expected_sample_rate = 30,
#               expected_unit = "g",
#               colnames = c("countsX","countsY","countsZ"),
#               outputres = 1,
#               minlength = 1,
#               outputtype = "numeric",
#               aggfunction = sum,
#               timestamp = F,
#               reporttype = c("scalar", "scalar", "scalar"))

## ----eval=FALSE---------------------------------------------------------------
# library(GGIR)
# GGIR(datadir = "~/myaccelerometerdata",
#              outputdir = "~/myresults",
#              mode = 1:2,
#              epochvalues2csv = TRUE,
#              do.report = 2,
#              myfun = myfun) #<= this is where object myfun is provided to GGIR

## ----eval=FALSE---------------------------------------------------------------
# dominant_frequency = function(data=c(), parameters=c()) {
#   # data: 3 column matrix with acc data
#   # parameters: the sample rate of data
#   source_python("dominant_frequency.py")
#   sf=parameters
#   N = nrow(data)
#   ws = 5 # windowsize
#   if (ncol(data) == 4) data= data[,2:4]
#   data = data.frame(t = floor(seq(0,(N - 1)/sf, by = 1/sf)/ws),
#                     x = data[,1], y = data[,2], z = data[,3])
#   df = aggregate(data, by = list(data$t),
#                  FUN=function(x) {return(dominant_frequency(x, sf))})
#   df = df[, -c(1:2)]
#   return(df)
# }
# }

## ----eval=FALSE---------------------------------------------------------------
#   library("reticulate")
#   use_virtualenv("~/myvenv", required = TRUE) # Local Python environment
#   py_install("numpy", pip = TRUE)
# 

## ----eval=FALSE---------------------------------------------------------------
# source("~/dominant_frequency.R")
# myfun =  list(FUN = dominant_frequency,
#               parameters = 30,
#               expected_sample_rate = 30,
#               expected_unit = "g",
#               colnames = c("domfreqX", "domfreqY", "domfreqZ"),
#               minlength = 5,
#               outputres = 5,
#               outputtype = "numeric",
#               aggfunction = median
#               timestamp = F,
#               reporttype = c("scalar", "scalar", "scalar"))

## ----eval=FALSE---------------------------------------------------------------
# library(GGIR)
# GGIR(datadir = "~/myaccelerometerdata",
#              outputdir = "~/myresults",
#              mode = 1:2,
#              epochvalues2csv = TRUE,
#              do.report = 2,
#              myfun = myfun,
#              do.parallel = FALSE)

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(echo = TRUE)

## ----eval=FALSE---------------------------------------------------------------
# library("GGIR")
# GGIR(datadir = "/your/data/directory",
#              outputdir = "/your/output/directory",
#              mode = 1:5, # <= run GGIR parts 1 to 5
#              do.report = c(2, 5), # <= generate csv-report for GGIR part 2 and part 5
#              qwindow = c(0, 6, 12, 18, 24),
#              timewindow = "MM")

## ----eval=FALSE---------------------------------------------------------------
# library("GGIR")
# GGIR(datadir = "/your/data/directory",
#              outputdir = "/your/output/directory",
#              mode = 1:5, # <= run GGIR parts 1 to 5
#              do.report = c(2, 5), # <= generate csv-report for GGIR part 2 and part 5
#              qwindow = "/path/to/your/activity/log.csv",
#              timewindow = "MM")

## ----echo=FALSE, out.width = "60%", out.extra='style="border: 0; padding:20px"', fig.alt="GGIR logo"----
knitr::include_graphics("GGIR-MASTERLOGO-RGB.png")

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(echo = TRUE)

## ----echo=FALSE, out.width = "60%", out.extra='style="border: 0; padding:20px"', fig.alt="GGIR logo"----
knitr::include_graphics("GGIR-MASTERLOGO-RGB.png")

load_params = function(topic = c("sleep", "metrics", "rawdata",
                                 "247", "phyact", "cleaning",
                                 "output", "general")) {

  # initialize params objects
  params_sleep = params_metrics = params_rawdata = c()
  params_247 = params_phyact = params_cleaning = c()
  params_output = params_general = c()

  # if (length(jsonfile) > 0) {
  #   # PLACEHOLDER FOR WHEN JSON FILE CONFIG FILE WILL BE FACILITATE
  # } else {
  if ("sleep" %in% topic) {
    params_sleep = list(anglethreshold = 5, timethreshold = 5,
                        ignorenonwear = TRUE,
                        HASPT.algo = "HDCZA",
                        HASIB.algo = "vanHees2015", Sadeh_axis = "Y",
                        longitudinal_axis = c(),
                        HASPT.ignore.invalid = FALSE,
                        loglocation = c(), colid = 1, coln1 = 2,
                        nnights = c(),
                        relyonguider = FALSE,
                        def.noc.sleep = 1,
                        sleeplogsep = NULL, sleepwindowType = "SPT",
                        possible_nap_window = NULL,
                        possible_nap_dur = NULL,
                        possible_nap_gap = 0,
                        possible_nap_edge_acc = Inf,
                        nap_model = c(), sleepefficiency.metric = 1,
                        HDCZA_threshold = c(),
                        oakley_threshold = 20,
                        consider_marker_button = FALSE,
                        impute_marker_button = FALSE,
                        sib_must_fully_overlap_with_TimeInBed = c(TRUE, TRUE),
                        nap_markerbutton_method = 0,
                        nap_markerbutton_max_distance = 30,
                        SRI1_smoothing_wsize_hrs = NULL,
                        SRI1_smoothing_frac = NULL,
                        spt_min_block_dur =  30,
                        spt_max_gap_dur =  60,
                        spt_max_gap_ratio = 1,
                        HorAngle_threshold = 60,
                        guider_cor_maxgap_hrs = 2,
                        guider_cor_min_frac_sib = 0.5,
                        guider_cor_min_hrs = 2,
                        guider_cor_meme_frac_out = 0.9,
                        guider_cor_meme_frac_in = 0.4,
                        guider_cor_meme_min_hrs = 1,
                        guider_cor_do = FALSE,
                        guider_cor_meme_min_dys = 3)
  }
  if ("metrics" %in% topic) {
    params_metrics = list(do.anglex = FALSE, do.angley = FALSE, do.anglez = TRUE,
                          do.zcx = FALSE, do.zcy = FALSE, do.zcz = FALSE,
                          do.enmo = TRUE, do.lfenmo = FALSE, do.en = FALSE,
                          do.mad = FALSE, do.enmoa = FALSE,
                          do.roll_med_acc_x = FALSE, do.roll_med_acc_y = FALSE,
                          do.roll_med_acc_z = FALSE, do.dev_roll_med_acc_x = FALSE,
                          do.dev_roll_med_acc_y = FALSE, do.dev_roll_med_acc_z = FALSE,
                          do.bfen = FALSE, do.hfen = FALSE, do.hfenplus = FALSE, do.lfen = FALSE,
                          do.lfx = FALSE, do.lfy = FALSE, do.lfz = FALSE,
                          do.hfx = FALSE, do.hfy = FALSE, do.hfz = FALSE,
                          do.bfx = FALSE, do.bfy = FALSE, do.bfz = FALSE,
                          do.brondcounts = FALSE,
                          do.neishabouricounts = FALSE,
                          hb = 15, lb = 0.2, n = 4,
                          zc.lb = 0.25, zc.hb = 3, zc.sb = 0.01, zc.order = 2, zc.scale = 1,
                          actilife_LFE = FALSE)
  }
  if ("rawdata" %in% topic) {
    params_rawdata = list(
      chunksize = 1, spherecrit = 0.3, minloadcrit = 168, printsummary = FALSE,
      do.cal = TRUE, backup.cal.coef = "retrieve", dynrange = c(),
      minimumFileSizeMB = 2, rmc.dec = ".",
      rmc.firstrow.acc = c(), rmc.firstrow.header = c(),
      rmc.header.length = c(), rmc.col.acc = 1:3,
      rmc.col.temp = c(), rmc.col.time = c(),
      rmc.unit.acc = "g",  rmc.unit.temp = "C",
      rmc.unit.time = "POSIX", rmc.format.time = "%Y-%m-%d %H:%M:%OS",
      rmc.bitrate = c(),  rmc.dynamic_range = c(),
      rmc.unsignedbit = TRUE, rmc.origin = "1970-01-01",
      rmc.desiredtz = NULL, rmc.configtz = NULL,  rmc.sf = c(),
      rmc.headername.sf = c(), rmc.headername.sn = c(),
      rmc.headername.recordingid = c(), rmc.header.structure = c(),
      rmc.check4timegaps = FALSE,  rmc.noise = 13, nonwear_range_threshold = 150,
      rmc.col.wear = c(), rmc.doresample = FALSE,
      interpolationType = 1,
      imputeTimegaps = TRUE, frequency_tol = 0.1, rmc.scalefactor.acc = 1)
  }
  if ("247" %in% topic) {
    params_247 = list(qwindow = c(0,24), qlevels = c(),
                      qwindow_dateformat = "%d-%m-%Y", ilevels = c(),
                      IVIS_windowsize_minutes = 60, IVIS_epochsize_seconds = c(),
                      IVIS.activity.metric = 2, IVIS_acc_threshold = 20, qM5L5 = c(),
                      MX.ig.min.dur = 10, M5L5res = 10, winhr = 5, iglevels = c(),
                      LUXthresholds = c(0, 100, 500, 1000, 3000, 5000, 10000),
                      LUX_cal_constant = c(), LUX_cal_exponent = c(), LUX_day_segments = c(),
                      L5M5window = c(0, 24), cosinor = FALSE,
                      part6CR = FALSE, part6HCA = FALSE,
                      part6Window = c("start", "end"),
                      part6DFA = FALSE, clevels = c(30, 150),
                      SRI2_WASOmin = 30, part2CR = FALSE)

  }
  if ("phyact" %in% topic) {
    params_phyact = list(mvpathreshold = NULL, boutcriter = NULL,
                         mvpadur = c(1,5,10),
                         boutcriter.in = 0.9, boutcriter.lig = 0.8,
                         boutcriter.mvpa = 0.8, threshold.lig = 40,
                         threshold.mod = 100, threshold.vig = 400,
                         boutdur.mvpa = c(1,5,10), boutdur.in = c(10,20,30),
                         boutdur.lig = c(1,5,10), frag.metrics = c(),
                         part6_threshold_combi = NULL)
  }
  if ("cleaning" %in% topic) {
    params_cleaning = list(includedaycrit = 16, ndayswindow = 7,
                           strategy = 1, data_masking_strategy = 1,
                           maxdur = 0,
                           hrs.del.start = 0, hrs.del.end = 0,
                           includedaycrit.part5 = 2/3, excludefirstlast.part5 = FALSE,
                           TimeSegments2ZeroFile = c(), do.imp = TRUE,
                           data_cleaning_file = c(), minimum_MM_length.part5 = 23,
                           excludefirstlast = FALSE, #<= to cleaning
                           includenightcrit = 16, #<= to cleaning
                           excludefirst.part4 = FALSE, # => to cleaning
                           excludelast.part4 = FALSE, max_calendar_days = 0,
                           nonWearEdgeCorrection = TRUE, nonwear_approach = "2023",
                           segmentWEARcrit.part5 = 0.5,
                           segmentDAYSPTcrit.part5 = c(0.9, 0),
                           study_dates_file = c(), study_dates_dateformat = "%d-%m-%Y",
                           includecrit.part6 = c(2/3, 2/3),
                           includenightcrit.part5 = 0,
                           nonwearFiltermaxHours = NULL,
                           nonwearFilterWindow = NULL)
  }
  if ("output" %in% topic) {
    params_output = list(epochvalues2csv = FALSE, save_ms5rawlevels = TRUE,
                         save_ms5raw_format = "RData", save_ms5raw_without_invalid = TRUE,
                         storefolderstructure = FALSE, timewindow = c("MM","WW"),
                         viewingwindow = 1, dofirstpage = TRUE, visualreport = TRUE,
                         week_weekend_aggregate.part5 = FALSE, do.part3.pdf = FALSE,
                         outliers.only = FALSE, criterror = 3, do.visual = TRUE,
                         do.sibreport = TRUE, do.part2.pdf = NULL, do.part2.png = NULL,
                         sep_reports = ",", sep_config = ",", 
                         dec_reports = ".", dec_config = ".", 
                         visualreport_without_invalid = TRUE,
                         old_visualreport = FALSE, visualreport_hrsPerRow = 36,
                         visualreport_focus = "day",
                         visualreport_validcrit = 0, require_complete_lastnight_part5 = FALSE,
                         method_research_vars = NULL)

  }
  if ("general" %in% topic) {
    params_general = list(overwrite = FALSE, acc.metric = "ENMO",
                          maxNcores = c(), print.filename = FALSE,
                          do.parallel = TRUE, windowsizes = c(5,900,3600),
                          desiredtz = "", configtz = c(), idloc = 1, dayborder = 0,
                          part5_agg2_60seconds = FALSE,
                          sensor.location = "wrist",
                          expand_tail_max_hours = NULL, recordingEndSleepHour = NULL,
                          dataFormat = "raw", maxRecordingInterval = NULL,
                          extEpochData_timeformat = "%d-%m-%Y %H:%M:%S",
                          recording_split_times = NULL,
                          recording_split_timeformat = "%d/%m/%Y %H:%M",
                          recording_split_overlap = 0,
                          recording_split_ignore_edges = FALSE,
                          use_trycatch_serial = FALSE)
  }
  invisible(list(params_sleep = params_sleep,
                 params_metrics = params_metrics,
                 params_rawdata = params_rawdata,
                 params_247 = params_247,
                 params_phyact = params_phyact,
                 params_cleaning = params_cleaning,
                 params_output = params_output,
                 params_general = params_general))
}

library(GGIR)
context("markerButtonForRest")

tz = "Pacific/Auckland"
startday = as.POSIXct(x = "2022-06-02 08:00:00", tz = tz)
ts = data.frame(time = seq(startday, startday + (16 * 3600), by = 60))
ts$marker = 0
ts$sibdetection = 0

test_that("Marker buttons influence sib adjustment as expected", {
  sib_start_original = as.POSIXct(c("2022-06-02 14:00:00", "2022-06-02 15:05:00"), tz = tz)
  sib_end_original = as.POSIXct(c("2022-06-02 14:20:00", "2022-06-02 15:25:00"), tz = tz)
  marker_times = as.POSIXct(c("2022-06-02 13:50:00", "2022-06-02 14:30:00"), tz = tz)
  
  sibreport = data.frame(ID = rep("test123", 2), type = c("sib", "sib"),
                         start = sib_start_original,
                         end = sib_end_original)
  ts$marker[which(ts$time %in% marker_times)] = 1

  sibreport$duration = as.numeric(difftime(time1 = sibreport$end,
                                           time2 = sibreport$start, units = "mins", tz = tz))
  params_sleep = load_params()$params_sleep
  params_sleep[["possible_nap_dur"]] =  c(15, 240)
  params_sleep[["possible_nap_window"]] =  c(9, 18)
  params_sleep[["nap_markerbutton_max_distance"]] = 30
  
  #==========================================================
  # ( 1 ) Method 1
  params_sleep[["nap_markerbutton_method"]] = 1
  test1 = markerButtonForRest(sibreport = sibreport, ts = ts,
                                  params_sleep = params_sleep)
  
  # updated timing for first sib because it has marker buttons nearby
  # but not for second sib
  expect_equal(test1$start, as.POSIXct(c("2022-06-02 13:50:00", "2022-06-02 15:05:00"), tz =  tz))
  expect_equal(test1$end, as.POSIXct(c("2022-06-02 14:30:00", "2022-06-02 15:25:00"), tz =  tz))
  # no sib is ignored in method 1
  expect_equal(test1$ignore, c(FALSE, FALSE))
  
  #==========================================================
  # ( 2 ) Method 2
  params_sleep[["nap_markerbutton_method"]] = 2
  test2 = markerButtonForRest(sibreport = sibreport, ts = ts,
                             params_sleep = params_sleep)
  
  # timing for sibs are not updated
  expect_equal(test2$start, sib_start_original)
  expect_equal(test2$end, sib_end_original)
  # second sib is ignored because it does not have marker button nearby
  expect_equal(test2$ignore, c(FALSE, TRUE))
  
  #==========================================================
  # ( 3 ) Method 3
  params_sleep[["nap_markerbutton_method"]] = 3
  test3 = markerButtonForRest(sibreport = sibreport, ts = ts,
                             params_sleep = params_sleep)
  
  # updated timing for first sib because it has marker buttons nearby
  # but not for second sib
  expect_equal(test3$start, test1$start)
  expect_equal(test3$end, test1$end)
  # second sib is ignored because it does not have marker button nearby
  expect_equal(test3$ignore, c(FALSE, TRUE))
  
  
  #==========================================================
  # ( 4 )  Method 3 - Distance 1
  params_sleep[["nap_markerbutton_max_distance"]] = 1
  params_sleep[["nap_markerbutton_method"]] = 3
  test4 = markerButtonForRest(sibreport = sibreport, ts = ts,
                             params_sleep = params_sleep)
  
  # timings are not updated because both sibs do not have marker button within less than 1 minute
  expect_equal(test4$start, sib_start_original)
  expect_equal(test4$end, sib_end_original)
  # both sibs are ignored because they do not have marker button within less than 1 minute
  expect_equal(test4$ignore, c(TRUE, TRUE))
  
  #==========================================================
  # ( 5 )  Method 3 - Distance 1 - Add 1 marker buttons inside the first sib
  marker_times = c(marker_times, as.POSIXct("2022-06-02 14:05:00", tz = tz))
  ts$marker[which(ts$time %in% marker_times)] = 1
  params_sleep[["nap_markerbutton_max_distance"]] = 1
  params_sleep[["nap_markerbutton_method"]] = 3
  test5 = markerButtonForRest(sibreport = sibreport, ts = ts,
                             params_sleep = params_sleep)
  
  # timings are not updated because both sibs do not have marker button within less than 1 minute
  # on both sides of the middle
  expect_equal(test5$start, sib_start_original)
  expect_equal(test5$end, sib_end_original)
  # both sibs are ignored because they do not have marker button within less than 1 minute
  expect_equal(test5$ignore, c(TRUE, TRUE))
  
  #==========================================================
  # ( 6 )  Method 3 - Distance 1 - Add 2nd marker button inside the first sib
  marker_times = c(marker_times, as.POSIXct("2022-06-02 14:15:00", tz = tz))
  ts$marker[which(ts$time %in% marker_times)] = 1
  params_sleep[["nap_markerbutton_max_distance"]] = 1
  params_sleep[["nap_markerbutton_method"]] = 3
  test6 = markerButtonForRest(sibreport = sibreport, ts = ts,
                             params_sleep = params_sleep)
  
  # updated timing for first sib because it has marker buttons nearby
  # but not for second sib
  expect_equal(test6$start, as.POSIXct(c("2022-06-02 14:05:00", "2022-06-02 15:05:00"), tz =  tz))
  expect_equal(test6$end, as.POSIXct(c("2022-06-02 14:15:00", "2022-06-02 15:25:00"), tz =  tz))
  # both sibs are ignored because they do not have marker button within less than 1 minute
  expect_equal(test6$ignore, c(FALSE, TRUE))
})

library(GGIR)
context("Integration")
test_that("All GGIR parts work together (integration test)", {
  skip_on_cran()
  Ndays = 2
  create_test_acc_csv(Nmin = Ndays*1440)
  create_test_sleeplog_csv(advanced = FALSE)
  fn = "123A_testaccfile.csv"
  sleeplog_fn = "testsleeplogfile.csv"
  metadatadir = paste0(getwd(), "/output_test")
  desiredtz = "Europe/London"
  dn = "output_test"
  do.parallel = FALSE
  if (file.exists(dn))  unlink(dn, recursive = TRUE)
  minimumFileSizeMB = 0
  #--------------------------------------------
  # isfilelist
  expect_true(isfilelist("file1.bin"))
  expect_true(isfilelist("file1.csv"))
  expect_true(isfilelist("file1.wav"))
  expect_true(isfilelist("file1.cwa"))
  expect_true(isfilelist("file1.gt3x"))
  expect_true(isfilelist(c("file1.bin", "file2.bin")))
  #--------------------------------------------
  # part 1
  g.part1(datadir = fn, metadatadir = metadatadir, f0 = 1, f1 = 1,
          overwrite = TRUE, desiredtz = desiredtz,
          do.enmo = TRUE, do.anglez = TRUE ,do.cal = TRUE,
          windowsizes = c(15,3600,3600), do.parallel = do.parallel,
          minimumFileSizeMB = minimumFileSizeMB, verbose = FALSE)
  expect_true(dir.exists(dn))
  rn = dir("output_test/meta/basic/",full.names = TRUE)
  load(rn[1])
  expect_equal(round(C$scale, digits = 5), c(0.98473, 0.98425, 0.98462), tolerance = 3)
  expect_that(nrow(M$metalong), equals(47))
  # expect_that(M$metalong[2,1],equals("2016-06-23T09:15:00+0100")) # turned off because not consistent across machines, to investigate
  expect_that(nrow(M$metashort),equals(11280))
  expect_equal(round(mean(M$metashort$ENMO), digits = 5), 0.02911, tolerance = 3)
  expect_that(I$monc, equals(3))
  expect_that(I$sf, equals(3))
  expect_that(I$dformc, equals(2))
  expect_equal(C$npoints, 7494, tolerance = 5)
  #-------------------------
  # part 2 with data_masking_strategy = 3
  g.part2(datadir = fn, metadatadir = metadatadir, f0 = 1, f1 = 1,
          idloc = 2, desiredtz = desiredtz, ndayswindow = 1,
          data_masking_strategy = 3, overwrite = TRUE, hrs.del.start = 0, hrs.del.end = 0,
          maxdur = Ndays, includedaycrit = 0, do.parallel = do.parallel, myfun = c(),
          verbose = FALSE)
  dirname = "output_test/meta/ms2.out/"
  rn = dir(dirname,full.names = TRUE)
  load(rn[1])
  load("output_test/meta/basic/meta_123A_testaccfile.csv.RData")
  expect_equal(nrow(IMP$metashort), 11280)
  expect_equal(rle(IMP$rout$r4)$lengths[1], 10)
  expect_equal(rle(IMP$rout$r4)$lengths[3], 13)
  expect_equal(round(mean(IMP$metashort$ENMO), digits = 5), 0.00802, tolerance = 3)
  expect_equal(round(as.numeric(SUM$summary$meas_dur_def_proto_day), digits = 3), 1)
  expect_equal(SUM$summary$`N valid weekend days (WE)`, 1)
  expect_equal(SUM$summary$`N valid weekdays (WD)`, 2)
  # check the ndayswindow included is 24 hours exactly 
  # ndayswindow = 1 with windowsizes = c(15, 3600, 3600)
  first_epoch_in_protocol = rle(IMP$rout$r4)$lengths[1] + 1
  last_epoch_in_protocol = max(which(IMP$rout$r4 == 0))
  expect_equal(last_epoch_in_protocol - first_epoch_in_protocol + 1,  24)
  
  # part 2 with data_masking_strategy = 3 and hrs.del.start = 6 and hrs.del.end = 6
  g.part2(datadir = fn, metadatadir = metadatadir, f0 = 1, f1 = 1,
          idloc = 2, desiredtz = desiredtz, ndayswindow = 1,
          strategy = 3, overwrite = TRUE, hrs.del.start = 6, hrs.del.end = 6,
          maxdur = Ndays, includedaycrit = 0, do.parallel = do.parallel, myfun = c(),
          verbose = FALSE)
  dirname = "output_test/meta/ms2.out/"
  rn = dir(dirname,full.names = TRUE)
  load(rn[1])
  expect_equal(nrow(IMP$metashort), 11280)
  expect_equal(rle(IMP$rout$r4)$lengths[1], 16)
  expect_equal(rle(IMP$rout$r4)$lengths[3], 19)
  # check the ndayswindow included is 12 hours exactly (24 minus hrs.del.start/end)
  # ndayswindow = 1 with windowsizes = c(15, 3600, 3600)
  first_epoch_in_protocol = rle(IMP$rout$r4)$lengths[1] + 1
  last_epoch_in_protocol = max(which(IMP$rout$r4 == 0))
  expect_equal(last_epoch_in_protocol - first_epoch_in_protocol + 1,  12)
  
  # part 2 with data_masking_strategy = 5
  g.part2(datadir = fn, metadatadir = metadatadir, f0 = 1, f1 = 1,
          idloc = 2, desiredtz = desiredtz, ndayswindow = 1,
          strategy = 5, overwrite = TRUE, hrs.del.start = 0, hrs.del.end = 0,
          maxdur = Ndays, includedaycrit = 0, do.parallel = do.parallel, myfun = c(),
          verbose = FALSE)
  dirname = "output_test/meta/ms2.out/"
  rn = dir(dirname,full.names = TRUE)
  load(rn[1])
  expect_equal(nrow(IMP$metashort), 11280)
  expect_equal(rle(IMP$rout$r4)$lengths[1], 15)
  expect_equal(rle(IMP$rout$r4)$lengths[3], 8)
  expect_equal(round(mean(IMP$metashort$ENMO), digits = 5), 0.03398, tolerance = 3)
  expect_equal(round(as.numeric(SUM$summary$meas_dur_def_proto_day), digits = 3), 1)
  expect_equal(SUM$summary$`N valid weekend days (WE)`, 1)
  expect_equal(SUM$summary$`N valid weekdays (WD)`, 2)
  # check the ndayswindow included is 24 hours exactly
  # ndayswindow = 1 with windowsizes = c(15, 3600, 3600)
  first_epoch_in_protocol = rle(IMP$rout$r4)$lengths[1] + 1
  last_epoch_in_protocol = max(which(IMP$rout$r4 == 0))
  expect_equal(last_epoch_in_protocol - first_epoch_in_protocol + 1,  24)
  
  # part 2 with data_masking_strategy = 5 and hrs.del.start = 6 and hrs.del.end = 6
  g.part2(datadir = fn, metadatadir = metadatadir, f0 = 1, f1 = 1,
          idloc = 2, desiredtz = desiredtz, ndayswindow = 1,
          data_masking_strategy = 5, overwrite = TRUE, hrs.del.start = 6, hrs.del.end = 6,
          maxdur = Ndays, includedaycrit = 0, do.parallel = do.parallel, myfun = c(),
          verbose = FALSE)
  dirname = "output_test/meta/ms2.out/"
  rn = dir(dirname,full.names = TRUE)
  load(rn[1])
  expect_equal(nrow(IMP$metashort), 11280)
  expect_equal(rle(IMP$rout$r4)$lengths[1], 21)
  expect_equal(rle(IMP$rout$r4)$lengths[3], 14)
  # check the ndayswindow included is 12 hours exactly (24 minus hrs.del.start/end)
  # ndayswindow = 1 with windowsizes = c(15, 3600, 3600)
  first_epoch_in_protocol = rle(IMP$rout$r4)$lengths[1] + 1
  last_epoch_in_protocol = max(which(IMP$rout$r4 == 0))
  expect_equal(last_epoch_in_protocol - first_epoch_in_protocol + 1,  12)
  
  # part 2 with data_masking_strategy = 2 and iglevels = 1
  g.part2(datadir = fn, metadatadir = metadatadir, f0 = 1, f1 = 1,
          idloc = 2, desiredtz = desiredtz,
          data_masking_strategy = 2,overwrite = TRUE, hrs.del.start = 0,hrs.del.end = 0,
          maxdur = Ndays, includedaycrit = 0, do.imp = FALSE, epochvalues2csv = TRUE, iglevels = TRUE,
          do.parallel = do.parallel, myfun = c(), winhr = 16, MX.ig.min.dur = 14, verbose = FALSE)
  dirname = "output_test/meta/ms2.out/"
  rn = dir(dirname,full.names = TRUE)
  load(rn[1])
  expect_that(nrow(IMP$metashort),equals(11280))
  expect_equal(mean(IMP$metashort$ENMO), 0.029, tolerance = 3)
  expect_equal(as.numeric(SUM$summary$meas_dur_def_proto_day), 1, tolerance = 2)
  expect_equal(as.numeric(SUM$daysummary$`L16_ig_gradient_ENMO_mg_0-24hr`[1]), -1.12, tolerance = 2)
  expect_equal(as.numeric(SUM$daysummary$`L16hr_ENMO_mg_0-24hr`), c(33.5, 19.8333, 13), tolerance = 0.001)
  # part 2
  
  # part 2 with data_masking_strategy = 1
  g.part2(datadir = fn, metadatadir = metadatadir, f0 = 1, f1 = 1,
          idloc = 2, desiredtz = desiredtz,
          data_masking_strategy = 1,overwrite = TRUE, hrs.del.start = 0,hrs.del.end = 0,
          maxdur = Ndays, includedaycrit = 0, qM5L5 = c(0.2,0.4), winhr = c(3,10),
          do.parallel = do.parallel, myfun = c(), qlevels = c(0.5, 0.9),
          cosinor = TRUE, verbose = FALSE)
  params_output = load_params()$params_output
  g.report.part2(metadatadir = metadatadir, f0 = 1, f1 = 1, params_output = params_output)
  dirname = "output_test/meta/ms2.out/"
  rn = dir(dirname,full.names = TRUE)
  load(rn[1])
  summarycsv = "output_test/results/part2_summary.csv"
  daysummarycsv = "output_test/results/part2_daysummary.csv"
  expect_true(dir.exists(dirname))
  expect_true(file.exists(summarycsv))
  expect_true(file.exists(daysummarycsv))
  expect_that(nrow(IMP$metashort), equals(11280))
  expect_equal(round(mean(IMP$metashort$ENMO), digits = 5), 0.02911, tolerance = 0.01)
  expect_that(round(as.numeric(SUM$summary$meas_dur_dys), digits = 5), equals(1.95833))
  expect_that(ncol(SUM$daysummary), equals(34))
  expect_equal(SUM$daysummary$`p50_ENMO_mg_0-24hr`, c(17.15, 33,  0), tolerance = 0.5)
  expect_equal(round(as.numeric(SUM$daysummary$`p90_ENMO_mg_0-24hr`)), c(44, 54, 41), tolerance = 1)
  expect_equal(mean(as.numeric(SUM$daysummary$`M3_ENMO_mg_0-24hr`)), 89.26, tolerance = 1)
  expect_equal(mean(as.numeric(SUM$daysummary$`M3_q40_ENMO_mg_0-24hr`)), 37.383, tolerance = 1)
  expect_equal(as.numeric(SUM$summary$cosinor_acrophase), 3.260855, tolerance = 1)
  expect_equal(as.numeric(SUM$summary$cosinorExt_beta), 109.618, tolerance = 2)
  #--------------------------------------------
  # part 3
  g.part3(metadatadir = metadatadir, f0 = 1, f1 = 1, anglethreshold = 5, desiredtz = desiredtz,
          timethreshold = 5,ignorenonwear = FALSE, overwrite = TRUE, do.part3.pdf = TRUE,
          do.parallel = do.parallel)
  dirname = "output_test/meta/ms3.out/"
  rn3 = dir(dirname,full.names = TRUE)
  load(rn3[1])
  
  expect_true(dir.exists(dirname))
  expect_that(round(sum(sib.cla.sum[,4:7]), digits = 0), equals(2957))
  
  #--------------------------------------------
  # part 4
  g.part4(datadir = fn, metadatadir = metadatadir, f0 = 1, f1 = 1,
          idloc = 2, loglocation = sleeplog_fn, do.visual = TRUE, outliers.only = FALSE,
          excludefirstlast = FALSE, criterror = 1, includenightcrit = 0, #nnights = 7,
          colid = 1, coln1 = 2, relyonguider = FALSE, desiredtz = desiredtz,
          storefolderstructure = FALSE, overwrite = TRUE)
  dirname = "output_test/meta/ms4.out/"
  rn = dir(dirname,full.names = TRUE)
  load(rn[1])
  vis_sleep_file = "output_test/results/visualisation_sleep.pdf"
  params_sleep = load_params()$params_sleep
  params_sleep[["loglocation"]] = sleeplog_fn
  g.report.part4(datadir = fn, metadatadir = metadatadir, 
                 f0 = 1, f1 = 1, params_sleep = params_sleep,
                 params_output = params_output, verbose = FALSE)
  expect_true(dir.exists(dirname))
  expect_true(file.exists(vis_sleep_file))
  expect_that(round(nightsummary$number_sib_wakinghours[1], digits = 4), equals(6))
  expect_true(as.logical(nightsummary$acc_available[1]))
  expect_true(as.logical(nightsummary$sleeplog_used[1]))

  #--------------------------------------------
  #part 5
  g.part5(datadir = fn, metadatadir = metadatadir, f0 = 1, f1 = 1, desiredtz = desiredtz,
          data_masking_strategy = 1, maxdur = Ndays, hrs.del.start = 0, hrs.del.end = 0,
          loglocation = sleeplog_fn,
          overwrite = TRUE, excludefirstlast = FALSE, do.parallel = do.parallel,
          frag.metrics = "all", save_ms5rawlevels = TRUE, save_ms5raw_format = "csv",
          part5_agg2_60seconds = TRUE, do.sibreport = TRUE, 
          iglevels = 1, timewindow = c("MM", "WW", "OO"),
          possible_nap_window = c(0, 24),
          possible_nap_dur = c(0, 240),
          possible_nap_edge_acc = Inf,
          possible_nap_gap = 0)
  sibreport_dirname = "output_test/meta/ms5.outraw/sib.reports"
  expect_true(dir.exists(sibreport_dirname))
  expect_true(file.exists(paste0(sibreport_dirname, "/sib_report_123A_testaccfile_T5A5.csv")))
  
  dirname = "output_test/meta/ms5.out/"
  rn = dir(dirname,full.names = TRUE)
  load(rn[1])
  
  dirname = "output_test/meta/ms5.out/"
  
  expect_true(dir.exists(dirname))
  expect_true(file.exists(rn[1]))
  expect_that(nrow(output),equals(5))
  expect_that(ncol(output),equals(158))
  expect_that(round(as.numeric(output$wakeup[2]), digits = 4), equals(36))
  expect_that(as.numeric(output$dur_day_spt_min[4]), equals(1150)) # WW window duration
  expect_that(as.numeric(output$dur_day_spt_min[5]), equals(1680)) # OO window duration
  dirname_raw = "output_test/meta/ms5.outraw/40_100_400"
  rn2 = dir(dirname_raw,full.names = TRUE, recursive = T)
  rn2_index = grep(pattern = "[.]csv", x = rn2, value = FALSE)
  expect_true(file.exists(rn2[rn2_index]))
  TSFILE = read.csv(rn2[rn2_index])
  expect_that(nrow(TSFILE),equals(2820))
  expect_equal(ncol(TSFILE), 13)
  expect_equal(length(unique(TSFILE$class_id)), 11)
  #GGIR
  suppressWarnings(GGIR(mode = c(2,3,4,5), datadir = fn, outputdir = getwd(),
                        studyname = "test", f0 = 1, f1 = 1,
                        do.report = c(2,4,5), overwrite = FALSE, visualreport = FALSE, viewingwindow = 1,
                        do.parallel = do.parallel, minimumFileSizeMB = minimumFileSizeMB,
                        verbose = FALSE, storefolderstructure = TRUE))
  suppressWarnings(GGIR(mode = c(), datadir = fn, outputdir = getwd(), studyname = "test",
                        f0 = 1, f1 = 1,
                        do.report = c(), overwrite = FALSE, visualreport = TRUE, viewingwindow = 1,
                        do.parallel = do.parallel, minimumFileSizeMB = minimumFileSizeMB, 
                        verbose = FALSE, old_visualreport = TRUE))
  expect_true(file.exists("output_test/results/part2_daysummary.csv"))
  expect_true(file.exists("output_test/results/part2_summary.csv"))
  expect_true(file.exists("output_test/results/part4_nightsummary_sleep_cleaned.csv"))
  expect_true(file.exists("output_test/results/part4_summary_sleep_cleaned.csv"))
  expect_true(file.exists("output_test/results/file summary reports/old_report_123A_testaccfile.csv.pdf"))
  expect_true(file.exists("output_test/results/part5_daysummary_MM_L40M100V400_T5A5.csv"))
  expect_true(file.exists("output_test/results/part5_daysummary_WW_L40M100V400_T5A5.csv"))
  expect_true(file.exists("output_test/results/part5_daysummary_OO_L40M100V400_T5A5.csv"))
  dn = "output_test"
  
  # Expect warning when unknown parameters are provided
  expect_error(GGIR(mode = NULL, datadir = fn, outputdir = getwd(),
                    studyname = "test", f0 = 1, f1 = 1,
                    do.report = NULL, overwrite = FALSE,
                    verbose = FALSE,
                    iamnewhere = 0),
               paste0("\nParameter iamnewhere is unknown to GGIR and will not be",
                      " used, please check for typos or remove."))
  
  expect_error(GGIR(mode = NULL, datadir = fn, outputdir = getwd(),
                    studyname = "test", f0 = 1, f1 = 1,
                    do.report = NULL, overwrite = FALSE,
                    verbose = FALSE,
                    iamnewhere = 0,
                    iamnewtoo = 1),
               paste0("\nParameters iamnewhere and iamnewtoo are unknown",
                      " to GGIR and will not be used, please check for ",
                      "typos or remove these."))
  

  #=======================
  # Different variations on part 4:
  #--------------------------------------------
  #--------------------------------------------
  # part 4 with sleepwindowType = TimeInBed and sleepefficiency.metric = 2
  g.part4(datadir = fn, metadatadir = metadatadir, f0 = 1, f1 = 1,
          idloc = 2, loglocation = sleeplog_fn, do.visual = TRUE, outliers.only = FALSE,
          excludefirstlast = FALSE, criterror = 1, includenightcrit = 0, #nnights = 7,
          colid = 1, coln1 = 2, relyonguider = FALSE, desiredtz = desiredtz,
          storefolderstructure = FALSE, overwrite = TRUE,
          sleepwindowType = "TimeInBed", sleepefficiency.metric = 2)
  dirname = "output_test/meta/ms4.out/"
  rn = dir(dirname,full.names = TRUE)
  load(rn[1])
  expect_true("sleeplatency" %in% colnames(nightsummary))
  expect_true("sleepefficiency" %in% colnames(nightsummary))
  expect_equal(round(nightsummary$sleeplatency[1], 3), 0.171)
  expect_equal(round(nightsummary$sleepefficiency[1], 3), 0.951)
  
  #--------------------------------------------
  # part 4 with sleepwindowType = TimeInBed and sleepefficiency.metric = 1
  g.part4(datadir = fn, metadatadir = metadatadir, f0 = 1, f1 = 1,
          idloc = 2, loglocation = sleeplog_fn, do.visual = TRUE, outliers.only = FALSE,
          excludefirstlast = FALSE, criterror = 1, includenightcrit = 0, #nnights = 7,
          colid = 1, coln1 = 2, relyonguider = FALSE, desiredtz = desiredtz,
          storefolderstructure = FALSE, overwrite = TRUE,
          sleepwindowType = "TimeInBed", sleepefficiency.metric = 1)
  dirname = "output_test/meta/ms4.out/"
  rn = dir(dirname,full.names = TRUE)
  load(rn[1])
  expect_true("sleeplatency" %in% colnames(nightsummary))
  expect_true("sleepefficiency" %in% colnames(nightsummary))
  expect_equal(round(nightsummary$sleepefficiency[1], 3), 0.851)
  
  #--------------------------------------------
  # part 4 without sleeplog
  g.part4(datadir = fn, metadatadir = metadatadir, f0 = 1, f1 = 1,
          idloc = 2, loglocation = c(), do.visual = TRUE, outliers.only = FALSE,
          excludefirstlast = FALSE, criterror = 1, includenightcrit = 0,
          nnights = 7, colid = 1, coln1 = 2,
          relyonguider = FALSE, desiredtz = desiredtz,
          storefolderstructure = TRUE, overwrite = TRUE, verbose = FALSE,
          sleepwindowType = "SPT")
  
  dirname = "output_test/meta/ms4.out/"
  rn = dir(dirname,full.names = TRUE)
  load(rn[1])
  expect_true(file.exists(vis_sleep_file))
  expect_equal(round(nightsummary$SptDuration[1], digits = 4), 18.0792, tolerance = 0.0005)
  expect_true(as.logical(nightsummary$acc_available[1]))
  expect_false(as.logical(nightsummary$sleeplog_used[1]))
  
  #----------------------------------------------------------------------
  # Part 4 - daysleeper scenario by modifying the part 3 output & criterror = 0, and relyonguider=TRUE
  SPTE_end = c(37, 37) # turn into midday
  row2copy = which(sib.cla.sum$sib.onset.time == "2016-06-24T05:10:15+0100")
  newrow = sib.cla.sum[row2copy,]
  newrow$sib.onset.time = "2016-06-24T12:30:15+0100"
  newrow$sib.end.time = "2016-06-24T12:45:15+0100"
  sib.cla.sum = rbind(sib.cla.sum, newrow)
  sib.cla.sum = sib.cla.sum[order(sib.cla.sum$sib.onset.time),]
  sib.cla.sum$sib.period[1:11] = 1:11
  part3_guider = "HDCZA"
  save(L5list, SPTE_end, SPTE_start, sib.cla.sum, tib.threshold, part3_guider,
       file = rn3[1])
  g.part4(datadir = fn, metadatadir = metadatadir, f0 = 1, f1 = 1,
          idloc = 2, loglocation = c(), do.visual = TRUE,
          outliers.only = FALSE, excludefirstlast = FALSE,
          criterror = 0, includenightcrit = 0, nnights = 7,
          colid = 1, coln1 = 2, relyonguider = TRUE,
          desiredtz = desiredtz, storefolderstructure = TRUE,
          overwrite = TRUE, verbose = FALSE)
  dirname = "output_test/meta/ms4.out/"
  rn = dir(dirname,full.names = TRUE)
  load(rn[1])
  vis_sleep_file = "output_test/results/visualisation_sleep.pdf"
  
  expect_true(dir.exists(dirname))
  expect_true(file.exists(vis_sleep_file))
  expect_that(round(nightsummary$number_sib_wakinghours[1], digits = 4), equals(0))
  expect_equal(nightsummary$SptDuration[1], 16.15, tolerance = 0.01)
  expect_true(as.logical(nightsummary$acc_available[1]))
  expect_false(as.logical(nightsummary$sleeplog_used[1]))
  #----------------------------------------------------------------------
  # Part 4 - DST+1
  load(rn3[1])
  changetime = function(x) {
    x = gsub("-06-", "-03-", x)
    x = gsub("-23T", "-26T", x)
    x = gsub("-24T", "-27T", x)
    x = gsub("-25T", "-28T", x)
    return(x)
  }
  sib.cla.sum$start.time.day = changetime(sib.cla.sum$start.time.day)
  sib.cla.sum$sib.onset.time = changetime(sib.cla.sum$sib.onset.time)
  sib.cla.sum$sib.end.time = changetime(sib.cla.sum$sib.end.time)
  save(L5list, SPTE_end, SPTE_start, sib.cla.sum, tib.threshold, file = rn3[1])
  g.part4(datadir = fn, metadatadir = metadatadir, f0 = 1, f1 = 1,
          idloc = 2, loglocation = sleeplog_fn, do.visual = TRUE,
          outliers.only = FALSE, excludefirstlast = FALSE, criterror = 0,
          includenightcrit = 0, nnights = 7, colid = 1, coln1 = 2,
          relyonguider = TRUE, desiredtz = desiredtz,
          storefolderstructure = TRUE, overwrite = TRUE, verbose = FALSE)
  dirname = "output_test/meta/ms4.out/"
  rn = dir(dirname,full.names = TRUE)
  load(rn[1])
  expect_true(dir.exists(dirname))
  expect_that(round(nightsummary$number_sib_wakinghours[1], digits = 4), equals(10))
  expect_that(round(nightsummary$SptDuration[1], digits = 4), equals(7))
  
  #----------------------------------------------------------------------
  # Part 4 - DST-1
  load(rn3[1])
  changetime = function(x) {
    x = gsub("-03-", "-10-", x)
    x = gsub("-26T", "-29T", x)
    x = gsub("-27T", "-30T", x)
    x = gsub("-28T", "-31T", x)
    return(x)
  }
  sib.cla.sum$start.time.day = changetime(sib.cla.sum$start.time.day)
  sib.cla.sum$sib.onset.time  = changetime(sib.cla.sum$sib.onset.time)
  sib.cla.sum$sib.end.time = changetime(sib.cla.sum$sib.end.time)
  save(L5list, SPTE_end, SPTE_start, sib.cla.sum, tib.threshold, file = rn3[1])
  g.part4(datadir = fn, metadatadir = metadatadir, f0 = 1, f1 = 1,
          idloc = 2, loglocation = sleeplog_fn, do.visual = TRUE,
          outliers.only = FALSE, excludefirstlast = FALSE,
          criterror = 0, includenightcrit = 0, nnights = 7,
          colid = 1, coln1 = 2, relyonguider = FALSE,
          desiredtz = desiredtz, storefolderstructure = TRUE,
          overwrite = TRUE, verbose = FALSE)
  dirname = "output_test/meta/ms4.out/"
  rn = dir(dirname,full.names = TRUE)
  load(rn[1])
  expect_true(dir.exists(dirname))
  expect_that(round(nightsummary$number_sib_wakinghours[1], digits = 4), equals(6))
  expect_equal(round(nightsummary$SptDuration[1], digits = 4), 13.0792)
  #---------------------
  # Part 1 with external function:
  exampleExtFunction = function(data=c(), parameters=c()) {
    data = data.frame(data, agglevel = round((1:nrow(data)) / (30 * 60 * 15)))
    output = aggregate(data, by = list(data$agglevel), FUN = mean)
    output = output[, -c(1, 2, ncol(output))]
    return(output)
  }
  myfun =  list(FUN = exampleExtFunction,
                parameters = 1.1,
                expected_sample_rate = 3, # resample data to 30 Hertz before applying function
                expected_unit = "mg",
                minlength = 15,
                outputres = 15,
                colnames = c("A", "B", "C"),
                outputtype = "numeric", #"numeric" (averaging is possible), "category" (majority vote)
                aggfunction = mean,
                timestamp = as.numeric(Sys.time())) # for unit test only
  
  g.part1(datadir = fn, metadatadir = metadatadir, f0 = 1, f1 = 1,
          overwrite = TRUE, desiredtz = desiredtz,
          do.parallel = do.parallel, myfun = myfun,
          do.enmo = TRUE, do.anglez = TRUE, do.cal = FALSE,
          windowsizes = c(15, 300, 3600),
          chunksize = 2, do.en = TRUE, do.anglex = TRUE, do.angley = TRUE,
          do.roll_med_acc_x = TRUE, do.roll_med_acc_y = TRUE, do.roll_med_acc_z = TRUE,
          do.dev_roll_med_acc_x = TRUE, do.dev_roll_med_acc_y = TRUE, do.dev_roll_med_acc_z = TRUE,
          do.bfx = TRUE, do.bfy = TRUE, do.bfz = TRUE, do.hfen = TRUE,
          do.hfx = TRUE, do.hfy = TRUE, do.hfz = TRUE, do.lfen = TRUE,
          do.enmoa = TRUE, verbose = FALSE)
  
  rn = dir("output_test/meta/basic/", full.names = TRUE)
  load(rn[1])
  expect_equal(ncol(M$metashort), 24)
  expect_equal(mean(M$metashort$B, na.rm = T), 24.673, tolerance = 3)
  expect_equal(mean(M$metashort$C, na.rm = T), -6.642, tolerance = 3)
  expect_equal(mean(M$metashort$EN, na.rm = T), 1.029, tolerance = 3)
  expect_equal(mean(M$metashort$angley, na.rm = T), 0.765, tolerance = 3)
  expect_equal(mean(M$metashort$roll_med_acc_x, na.rm = T), 0.729, tolerance = 3)
  expect_equal(mean(M$metashort$roll_med_acc_z, na.rm = T), 0.007, tolerance = 3)
  expect_equal(mean(M$metashort$dev_roll_med_acc_x, na.rm = T), 0.007, tolerance = 3)
  expect_equal(mean(M$metashort$ENMOa, na.rm = T), 0.03, tolerance = 3)
  
  if (dir.exists(dn))  unlink(dn, recursive = TRUE)
  if (file.exists(fn)) unlink(fn)
  if (file.exists(sleeplog_fn)) file.remove(sleeplog_fn)
})

library(GGIR)
context("loading and checking params")
test_that("load_params can load parameters", {
  skip_on_cran()
  
  params = load_params()
  expect_equal(params$params_sleep[[1]], 5)
  expect_equal(params$params_sleep[[3]], TRUE)
  expect_equal(params$params_cleaning[[5]], 0)
  expect_equal(params$params_sleep[[6]], "Y")
  
  # Test length of objects
  expect_equal(length(params), 8)
  expect_equal(length(params$params_sleep), 43)
  expect_equal(length(params$params_metrics), 41)
  expect_equal(length(params$params_rawdata), 39)
  expect_equal(length(params$params_247), 26)
  expect_equal(length(params$params_cleaning), 28)
  expect_equal(length(params$params_phyact), 14)
  expect_equal(length(params$params_output), 28)
  expect_equal(length(params$params_general), 22)

  params_sleep = params$params_sleep
  params_metrics = params$params_metrics
  params_rawdata = params$params_rawdata
  params_247 = params$params_247
  params_phyact = params$params_phyact
  params_cleaning = params$params_cleaning
  params_output = params$params_output
  params_general = params$params_general
  
  # Test that parameter check does not generate warnings:
  expect_warning(check_params(params_sleep), regexp = NA)
  # Test that parameter check does not generate errors:
  expect_error(check_params(params_sleep), regexp = NA)
  # Test that parameter check produces error when numeric is given a character value
  params_sleep$timethreshold = "ABC"
  expect_error(check_params(params_sleep), regexp = "Sleep parameter timethreshold is not numeric")
  params_sleep = params$params_sleep
  # Test that parameter check produces error when character is given a numeric value
  params_sleep$Sadeh_axis = 123
  expect_error(check_params(params_sleep), regexp = "Sleep parameter Sadeh_axis is not character")
  params_sleep = params$params_sleep
  # Test that parameter check produces error when boolean is given a numeric value
  params_general$overwrite = 123
  expect_error(check_params(params_general = params_general), regexp = "general parameter overwrite is not boolean")
  params_general = params$params_general
  
  # Error if brondcounts calculated
  params_metrics$do.brondcounts = TRUE
  expect_error(check_params(params_metrics = params_metrics))
  params_metrics = params$params_metrics
  
  # change HASPT algo to "notused" if length(def.noc.sleep) == 2
  params_sleep$def.noc.sleep = c(21, 8)
  check = check_params(params_sleep = params_sleep)
  expect_equal(check$params_sleep$HASPT.algo, "notused")
  params_sleep = params$params_sleep
  
  # if sensor.location = "hip", then HASPT should turn to "HorAngle"
  # also, three angles metrics should be turned to TRUE
  # also, sleepwindowType should be turned to TimeInBed
  params_general$sensor.location = "hip"
  check = expect_warning(check_params(params_general = params_general, 
                                      params_metrics = params_metrics, 
                                      params_sleep = params_sleep))
  expect_equal(check$params_sleep$HASPT.algo, "HorAngle")
  expect_true(check$params_metrics$do.anglex)
  expect_true(check$params_metrics$do.angley)
  expect_true(check$params_metrics$do.anglez)
  expect_equal(check$params_sleep$sleepwindowType, "TimeInBed")
  params_general = params$params_general
  
  # if HASIB.algo is "Sadeh1994", "Galland2012", "ColeKripke1992"
  # then Sadeh_axis should be one of "X", "Y", "Z"
  # and zero-crossing counts should be calculated
  # otherwise, Sadeh_axis should be empty
  params_sleep$HASIB.algo = "Sadeh1994"
  params_sleep$Sadeh_axis = "y"
  expect_warning(check_params(params_metrics = params_metrics, 
                              params_sleep = params_sleep),
                 regexp = "Parameter Sadeh_axis does not have meaningful value")
  params_sleep$Sadeh_axis = "Y"
  check = check_params(params_metrics = params_metrics, 
                       params_sleep = params_sleep)
  expect_true(check$params_metrics$do.zcy)
  params_sleep = params$params_sleep
  check = check_params(params_metrics = params_metrics, 
                       params_sleep = params_sleep)
  expect_equal(check$params_sleep$Sadeh_axis, "")
  params_sleep = params$params_sleep
  params_metrics = params$params_metrics
  
  # loglocation = "" should be turned to c()
  # loglocation = "C:\\mystudy\\mylog.csv" should turn the \ into /
  params_sleep$loglocation = ""
  check = check_params(params_sleep = params_sleep)
  expect_null(check$params_sleep$loglocation)
  params_sleep$loglocation = "C:\\mystudy\\mylog.csv"
  check = check_params(params_sleep = params_sleep)
  expect_true(grepl("/", check$params_sleep$loglocation))
  params_sleep = params$params_sleep
  
  # loglocation = "C:/mystudy/mylog.csv" and def.noc.sleep of length != 1 are not compatible
  params_sleep$loglocation = "C:/mystudy/mylog.csv"
  params_sleep$def.noc.sleep = c()
  expect_warning(check_params(params_sleep = params_sleep),
                 regexp = "loglocation was specified and def.noc.sleep")
  params_sleep = params$params_sleep
  
  # if no sleeplog and HASPT != "HorAngle", then sleepwindowType should be "SPT"
  params_sleep$sleepwindowType = "TimeInBed"
  expect_warning(check_params(params_sleep = params_sleep),
                 regexp = "Auto-updating sleepwindowType to SPT")
  params_sleep = params$params_sleep
  
  # if data_masking_strategy != 1, hrs.del.start and hrs.del.end != 0 do not make sense
  params_cleaning$data_masking_strategy = 2
  params_cleaning$hrs.del.start = 1
  expect_warning(check_params(params_cleaning = params_cleaning),
                 regexp = "Setting parameter hrs.del.start")
  params_cleaning = params$params_cleaning
  params_cleaning$data_masking_strategy = 2
  params_cleaning$hrs.del.end = 1
  expect_warning(check_params(params_cleaning = params_cleaning),
                 regexp = "Setting parameter hrs.del.end")
  params_cleaning = params$params_cleaning
  
  # if data_masking_strategy != c(3,5), ndayswindow != 7 does not make sense
  params_cleaning$ndayswindow = 14
  expect_warning(check_params(params_cleaning = params_cleaning),
                 regexp = "Setting parameter ndayswindow in combination")
  params_cleaning = params$params_cleaning
  
  # if data_masking_strategy == 5, ndayswindow should be integer
  params_cleaning$data_masking_strategy = 5
  params_cleaning$ndayswindow = 7.5
  check = expect_warning(check_params(params_cleaning = params_cleaning),
                         regexp = "Parameter ndayswindow has been rounded")
  expect_equal(check$params_cleaning$ndayswindow, 8)
  params_cleaning = params$params_cleaning
  
  # data_cleaning_file = "C:\\mystudy\\mycleaningfile.csv" should turn the \ into /
  params_cleaning$data_cleaning_file = "C:\\mystudy\\mycleaningfile.csv"
  check = check_params(params_cleaning = params_cleaning)
  expect_true(grepl("/", check$params_cleaning$data_cleaning_file))
  params_cleaning = params$params_cleaning
  
  # bout.metric should not be used anymore
  params_phyact$bout.metric = 6
  expect_warning(check_params(params_phyact = params_phyact),
                 regexp = "Parameters bout.metric and closedbout are no longer")
  params_phyact = params$params_phyact
  
  # length of mvpadur should be 3
  params_phyact$mvpadur = c(1, 2, 3, 5, 10, 15)
  check =  expect_warning(check_params(params_phyact = params_phyact),
                          regexp = "mvpadur needs to be a vector")
  expect_equal(length(check$params_phyact$mvpadur), 3)
  params_phyact = params$params_phyact
  
  # last value in LUX_day_segments should be 24
  params_247$LUX_day_segments = c(9, 18)
  check = check_params(params_247 = params_247)
  expect_equal(check$params_247$LUX_day_segments[length(check$params_247$LUX_day_segments)], 24)
  params_247 = params$params_247
  
  # if dataFormat = "actiwatch_awd"
  params_general$dataFormat = "actiwatch_awd"
  check = check_params(params_general = params_general,
                       params_metrics = params_metrics)
  expect_false(check$params_metrics$do.zcy)
  expect_equal(check$params_general$acc.metric, "ExtAct")
  # also, no other metrics should be extracted
  params_metrics$do.anglex = TRUE
  params_general$acc.metric = "ExtAct"
  check = check_params(params_general = params_general,
                       params_metrics = params_metrics)
  expect_false(check$params_metrics$do.zcy)
  expect_false(check$params_metrics$do.anglex)
  expect_equal(check$params_general$acc.metric, "ExtAct")
  params_general = params$params_general
  params_metrics = params$params_metrics
  
  # if dataFormat = "ukbiobank", then acc.metric = "LFENMO"
  params_general$dataFormat = "ukbiobank"
  check = check_params(params_general = params_general,
                       params_metrics = params_metrics)
  expect_equal(check$params_general$acc.metric, "LFENMO")
  # also, no other metrics should be extracted
  params_metrics$do.anglex = TRUE
  params_general$acc.metric = "LFENMO"
  check = check_params(params_general = params_general,
                       params_metrics = params_metrics)
  expect_false(check$params_metrics$do.anglex)
  params_general = params$params_general
  params_metrics = params$params_metrics
  
  # maxRecordingInterval cannot be higher than 21 days (21*24)
  params_general$maxRecordingInterval = 30*24
  expect_error(check_params(params_general = params_general),
               regexp = "A maxRecordingInterval value higher than 21 days")
  
  # segmentWEARcrit should be a number between 0 and 1
  params_cleaning$segmentWEARcrit.part5 = NULL
  expect_warning(check_params(params_cleaning = params_cleaning),
                 regexp = "segmentWEARcrit.part5 is expected to be")
  params_cleaning$segmentWEARcrit.part5 = 16
  expect_error(check_params(params_cleaning = params_cleaning),
               regexp = "Incorrect value of segmentWEARcrit.part5")
  params_cleaning$segmentWEARcrit.part5 = 0.5
  
  # segmentDAYSPTcrit.part5 should be two numbers between 0 and 1
  params_cleaning$segmentDAYSPTcrit.part5 = NULL
  expect_error(check_params(params_cleaning = params_cleaning),
               regexp = "Parameter segmentDAYSPTcrit.part5 is expected to be a numeric vector of length 2")
  params_cleaning$segmentDAYSPTcrit.part5 = c(-1, 0)
  expect_error(check_params(params_cleaning = params_cleaning),
               regexp = paste0("Incorrect values of segmentDAYSPTcrit.part5,",
                               " these should be a fractions between zero",
                               " and one, please change."))
  
})

library(GGIR)
context("aggregateEvent")
test_that("Events from external function are correctly aggregated", {
  skip_on_cran()
  
  # create dummy test data
  # 5 minutes, where minute 2-4 are the minutes of our window
  # minute 2 has no steps and ENMO = 0
  # minute 3 has 60 steps and ENMO = 80
  # minute 4 has 60 steps and ENMO = 120
  metashort = data.frame(ENMO = c(rep(0, 24), rep(0.08, 12), rep(0.12, 12), rep(1, 12)),
                         step_count = c(rep(0, 24), rep(5, 36)))
  cn_metashort = colnames(metashort)
  anwi_index = 1
  # varnum = metashort$step_count[anwindices]
  ws3 = 5
  anwi_nameindices = "_1234hrs"
  daysummary = matrix("", 1, 23)
  fi = 1
  di = 1
  ds_names = ""
  myfun = list(ilevels = c(0, 50, 100),
               clevels = c(0, 30, 50),
               qlevels = c(0.25, 0.5, 0.75),
               ebout.dur = c(1, 5, 10),
               ebout_th.cad = 30,
               ebout.th.acc = 50,
               ebout.criter = 0.8,
               ebout.condition = "AND")
  # run function  
  segmentInfo = list(anwi_nameindices = anwi_nameindices,
                     anwi_index = anwi_index)
  params_247 = load_params("247")$params_247
  eventAgg = aggregateEvent(metric_name = "step_count",
                            epochsize = ws3,
                            daysummary = daysummary,
                            ds_names = ds_names,
                            fi = fi, di = di, vari = metashort,
                            segmentInfo = segmentInfo, myfun, params_247)
  
  daysummary = as.data.frame(eventAgg$daysummary)
  names(daysummary)[1:length(eventAgg$ds_names)] = eventAgg$ds_names
  
  # Dimensions
  expect_equal(ncol(daysummary), 23)
  
  # total steps
  expect_equal(daysummary$ExtFunEvent_tot_step_count_1234hrs, "180")
  
  #mean cadence
  expect_equal(daysummary$ExtFunEvent_mn_cad_1234hrs, "36")
  
  # Percentiles of cadence
  expect_equal(daysummary$ExtFunEvent_cad_p25_1234hrs, "0")
  expect_equal(daysummary$ExtFunEvent_cad_p50_1234hrs, "60")
  expect_equal(daysummary$ExtFunEvent_cad_p75_1234hrs, "60")
  
  #total steps per acc range
  expect_equal(daysummary$`ExtFunEvent_tot_step_count_acc0-50mg_ENMO_1234hrs`, "0")
  expect_equal(daysummary$`ExtFunEvent_tot_step_count_acc50-100mg_ENMO_1234hrs`, "60")
  expect_equal(daysummary$ExtFunEvent_tot_step_count_accatleast100mg_ENMO_1234hrs, "120")
  
  #mean cadence per acc range
  expect_equal(daysummary$`ExtFunEvent_mn_cad_acc0-50mg_ENMO_1234hrs`, "0")
  expect_equal(daysummary$`ExtFunEvent_mn_cad_acc50-100mg_ENMO_1234hrs`, "60")
  expect_equal(daysummary$ExtFunEvent_mn_cad_accatleast100mg_ENMO_1234hrs, "60")
  
  #mean cadence per cadence range
  expect_equal(daysummary$`ExtFunEvent_mn_cad_cad0-30spm_1234hrs`, "0")
  expect_equal(daysummary$`ExtFunEvent_mn_cad_cad30-50spm_1234hrs`, "0")
  expect_equal(daysummary$ExtFunEvent_mn_cad_cadatleast50spm_1234hrs, "60")
  
  #total steps per cadence range
  expect_equal(daysummary$`ExtFunEvent_tot_step_count_cad0-30spm_1234hrs`, "0")
  expect_equal(daysummary$`ExtFunEvent_tot_step_count_cad30-50spm_1234hrs`, "0")
  expect_equal(daysummary$ExtFunEvent_tot_step_count_cadatleast50spm_1234hrs, "180")
  
  # mean acc per cadence range
  expect_equal(daysummary$`ExtFunEvent_mn_ENMO_cad0-30spm_1234hrs`, "0")
  expect_equal(daysummary$`ExtFunEvent_mn_ENMO_cad30-50spm_1234hrs`, "0")
  expect_equal(daysummary$ExtFunEvent_mn_ENMO_cadatleast50spm_1234hrs, "400")
  
  # time per cadence range
  expect_equal(daysummary$`ExtFunEvent_dur_cad0-30spm_1234hrs`, "2")
  expect_equal(daysummary$`ExtFunEvent_dur_cad30-50spm_1234hrs`, "0")
  expect_equal(daysummary$ExtFunEvent_dur_cadatleast50spm_1234hrs, "3")
  
  # MX
  # First create longer time series
  ENMO = seq(1/12, 1440, by = 1/12)
  step_count = rep(5, 1440 * 12)
  
  set.seed(123)
  steps = rnorm(n = length(which(ENMO > 40)), mean = 2, sd = 1)
  step_count[which(ENMO > 40)] = abs(steps)
  metashort = data.frame(ENMO = ENMO,
                         step_count = step_count)
  params_247[["winhr"]] = 2
  segmentInfo = list(anwi_nameindices = anwi_nameindices,
                     anwi_index = anwi_index,
                     anwi_t0 = 1,
                     anwi_t1 = 12 * 60 * 5)
  eventAgg = aggregateEvent(metric_name = "step_count",
                            epochsize = ws3,
                            daysummary = daysummary,
                            ds_names = ds_names,
                            fi = fi, di = di, vari = metashort,
                            segmentInfo = segmentInfo, myfun, params_247)
  daysummary = as.data.frame(eventAgg$daysummary)
  names(daysummary)[1:length(eventAgg$ds_names)] = eventAgg$ds_names
  expect_equal(daysummary$ExtFunEvent_L2hr_cad_1234hrs, 2.5, tolerance = 0.01)
  expect_equal(daysummary$ExtFunEvent_L2_cad_meancad_1234hrs, 24.17291, tolerance = 0.001)
  expect_equal(daysummary$ExtFunEvent_L2_cad_mean_ENMO_mg_1234hrs, 210041.7, tolerance = 0.05)
  expect_equal(daysummary$ExtFunEvent_M2hr_cad_1234hrs, 0)
  expect_equal(daysummary$ExtFunEvent_M2_cad_meancad_1234hrs, 36.26251, tolerance = 0.05)
  expect_equal(daysummary$ExtFunEvent_M2_cad_mean_ENMO_mg_1234hrs, 60041.67, tolerance = 0.01)
})

library(GGIR)
context("HASPT")
test_that("HASPT generate correct output", {
  skip_on_cran()

  # prepare dummy data
  N = 17280
  epochSize = 5
  nomovement = c(5000:6000, 6100:8000)
  
  invalid = rep(0, N)
  invalid[5400:7800] = 1
  
  set.seed(1234)
  activity = abs(rnorm(n = N, mean = 0, sd = 60))
  activity[nomovement] = 0
  
  set.seed(1234)
  angle = rnorm(n = N, mean = 0, sd = 20)
  angle[which(angle > 90)] = 90
  angle[which(angle < -90)] = -90
  angle[nomovement] = angle[5000]
  
  sibs = rep(0, N)
  sibs[nomovement] = 1
  sibs[c(5000:5050, 6000:6050, 7950:8000)] = 0
  
  marker = rep(0, N)
  marker[c(4900, 7900)] = 1
  
  # Test all HASPT.algo setting
  
  # HDCZA with HASPT.ignore.invalid = FALSE
  params_sleep = load_params()$params_sleep
  params_sleep[["spt_min_block_dur"]] = 30
  params_sleep[["spt_max_gap_dur"]] = 60
  params_sleep[["HDCZA_threshold"]] = NULL
  params_sleep[["HASPT.ignore.invalid"]] = FALSE
  params_sleep[["try_marker_button"]] = FALSE
  test = HASPT(angle = angle, params_sleep = params_sleep,
               ws3 = epochSize, HASPT.algo = "HDCZA", invalid = invalid)
  expect_equal(test$SPTE_start, 5000)
  expect_equal(test$SPTE_end, 8000)
  
  # HDCZA with HASPT.ignore.invalid = FALSE, spt_max_gap_dur = 60 and spt_max_gap_ratio = 0.05
  params_sleep = load_params()$params_sleep
  params_sleep[["spt_min_block_dur"]] = 30
  params_sleep[["spt_max_gap_dur"]] = 60
  params_sleep[["spt_max_gap_ratio"]] = 0.05
  params_sleep[["HDCZA_threshold"]] = NULL
  params_sleep[["HASPT.ignore.invalid"]] = FALSE
  params_sleep[["try_marker_button"]] = FALSE
  test = HASPT(angle = angle, params_sleep = params_sleep,
               HASPT.algo = "HDCZA", 
               ws3 = epochSize, invalid = invalid)
  expect_equal(test$SPTE_start, 6099)
  expect_equal(test$SPTE_end, 8000)
  
  # HDCZA with HASPT.ignore.invalid = TRUE
  params_sleep = load_params()$params_sleep
  params_sleep[["spt_min_block_dur"]] = 30
  params_sleep[["spt_max_gap_dur"]] = 60
  params_sleep[["HDCZA_threshold"]] = NULL
  params_sleep[["HASPT.ignore.invalid"]] = TRUE
  params_sleep[["try_marker_button"]] = FALSE
  test = HASPT(angle = angle, params_sleep = params_sleep,
               ws3 = epochSize, HASPT.algo = "HDCZA", invalid = invalid)
  expect_equal(test$SPTE_start, 5000)
  expect_equal(test$SPTE_end, 5400)
  
  # notused
  params_sleep = load_params()$params_sleep
  params_sleep[["spt_min_block_dur"]] = 30
  params_sleep[["spt_max_gap_dur"]] = 60
  params_sleep[["HDCZA_threshold"]] = NULL
  params_sleep[["HASPT.ignore.invalid"]] = FALSE
  params_sleep[["try_marker_button"]] = FALSE
  test = HASPT(angle = angle, params_sleep = params_sleep,
               ws3 = epochSize, HASPT.algo = "notused", invalid = invalid)
  expect_null(test$SPTE_start)
  expect_null(test$SPTE_end)
  
  # HorAngle
  params_sleep = load_params()$params_sleep
  params_sleep[["spt_min_block_dur"]] = 30
  params_sleep[["spt_max_gap_dur"]] = 60
  params_sleep[["HDCZA_threshold"]] = NULL
  params_sleep[["HASPT.ignore.invalid"]] = FALSE
  params_sleep[["try_marker_button"]] = FALSE
  params_sleep[["HorAngle_threshold"]] = 60
  test = HASPT(angle = angle, params_sleep = params_sleep,
               ws3 = epochSize, HASPT.algo = "HorAngle", invalid = invalid)
  expect_equal(test$SPTE_start, 2415)
  expect_equal(test$SPTE_end, 17097)
  
  # NotWorn
  params_sleep = load_params()$params_sleep
  params_sleep[["spt_min_block_dur"]] = 30
  params_sleep[["spt_max_gap_dur"]] = 60
  params_sleep[["HDCZA_threshold"]] = NULL
  params_sleep[["HASPT.ignore.invalid"]] = FALSE
  params_sleep[["try_marker_button"]] = FALSE
  params_sleep[["HorAngle_threshold"]] = 60
  test = HASPT(angle = angle, params_sleep = params_sleep,
               ws3 = epochSize, HASPT.algo = "NotWorn",
               invalid = invalid, activity = activity,
               marker = marker)
  expect_equal(test$SPTE_start, 5028)
  expect_equal(test$SPTE_end, 7971)
  expect_equal(test$tib.threshold, 0.2835165, tolerance = 0.001)
  
  # HLRB
  params_sleep = load_params()$params_sleep
  params_sleep[["spt_min_block_dur"]] = 30
  params_sleep[["spt_max_gap_dur"]] = 60
  params_sleep[["HDCZA_threshold"]] = NULL
  params_sleep[["HASPT.ignore.invalid"]] = FALSE
  params_sleep[["try_marker_button"]] = FALSE
  test = HASPT(angle = angle, params_sleep = params_sleep,
               ws3 = epochSize, HASPT.algo = "HLRB", invalid = invalid,
               activity = activity, sibs = sibs)
  expect_equal(test$SPTE_start, 5052)
  expect_equal(test$SPTE_end, 7950)
  expect_equal(test$tib.threshold, 0)
  
  # marker button
  params_sleep = load_params()$params_sleep
  params_sleep[["spt_min_block_dur"]] = 30
  params_sleep[["spt_max_gap_dur"]] = 60
  params_sleep[["HDCZA_threshold"]] = NULL
  params_sleep[["HASPT.ignore.invalid"]] = FALSE
  params_sleep[["consider_marker_button"]] = TRUE
  marker[c(4900, 4901, 4902, 7900, 7901, 7902)] = 1
  downsample = seq(1, N, by = 3)
  test = HASPT(angle = angle, params_sleep = params_sleep,
               ws3 = epochSize,
               HASPT.algo = "markerbutton", invalid = invalid,
               marker = marker, sibs = sibs)
  expect_equal(test$SPTE_start, 4902) # note: times 3 because these are indices in the downsampled data
  expect_equal(test$SPTE_end, 7902)
  expect_equal(test$part3_guider, "markerbutton")
  
  # MotionWare
  # As we do not know whether implementation is correct
  # this unit-test also has limited value. I am keeping it to at least
  # make an attempt to monitor changes in functionality
  params_sleep = load_params()$params_sleep
  params_sleep[["spt_min_block_dur"]] = 30
  params_sleep[["spt_max_gap_dur"]] = 60
  params_sleep[["HDCZA_threshold"]] = NULL
  params_sleep[["HASPT.ignore.invalid"]] = FALSE
  params_sleep[["consider_marker_button"]] = FALSE
  marker[c(4900, 4901, 4902, 7900, 7901, 7902)] = 1
  downsample = seq(1, N, by = 3)
  test = HASPT(angle = NULL, params_sleep = params_sleep,
               ws3 = 15,
               HASPT.algo = "MotionWare", invalid = invalid[downsample],
               activity = activity[downsample] / 1000,
               marker = marker[downsample], sibs = sibs[downsample])
  expect_equal(test$SPTE_start, 0) # note: times 3 because these are indices in the downsampled data
  expect_equal(test$SPTE_end, 5761)
  expect_equal(test$part3_guider, "MotionWare")
})

library(GGIR)
context("g.weardec detection")
test_that("g.weardec relabels data correctly", {
  skip_on_cran()
  ws2 = 900
  wearthreshold = 2
  Ntotal = 300
  createNW = function(Ntotal = 300, Nnonwear = 0, invert = FALSE) {
    wearhalf = ((Ntotal - Nnonwear) / 2)
    N1 = floor(wearhalf)
    N2 = ceiling(wearhalf)
    N3 = Ntotal - (N1 + N2)
    if (invert == FALSE) {
      ts = c(rep(0, N1), rep(3, N3), rep(0, N2))
    } else {
      ts = c(rep(3, N1), rep(0, N3), rep(3, N2))
    }
    return(ts)
  }
  params_cleaning = load_params()$params_cleaning
  # No non-wear, no clipping
  metalong = data.frame(nonwearscore = createNW(),
                        clippingscore = rep(0, Ntotal))
  out = g.weardec(metalong = metalong, wearthreshold, ws2, params_cleaning = params_cleaning)
  expect_equal(sum(out$r1), 0)
  expect_equal(sum(out$r2), 0)
  expect_equal(sum(out$r3), 0)
  expect_equal(out$LC, 0)
  expect_equal(out$LC2, 0)
  
  # Non-wear a few epoch in the first 3 hours, no clipping
  metalong = data.frame(nonwearscore = c(rep(0, 3), rep(3, 2), rep(0, Ntotal - 5)),
                        clippingscore = rep(0, Ntotal))
  out = g.weardec(metalong, wearthreshold, ws2, params_cleaning = params_cleaning)
  expect_equal(sum(out$r1), 2)
  expect_equal(sum(out$r3), 3)
  expect_equal(sum(out$r1 + out$r3), 5)
  
  # Non-wear a few epoch in the last 3 hours, no clipping
  metalong = data.frame(nonwearscore = c(rep(0, Ntotal - 5), rep(3, 2), rep(0, 3)),
                        clippingscore = rep(0, Ntotal))
  out = g.weardec(metalong, wearthreshold, ws2, params_cleaning = params_cleaning)
  expect_equal(sum(out$r1), 2)
  expect_equal(sum(out$r3), 3)
  expect_equal(sum(out$r1 + out$r3), 5)
  
  # Non-wear 20 in middle + clipping
  metalong = data.frame(nonwearscore = createNW(Nnonwear = 20),
                        clippingscore = c(rep(0, Ntotal - 100), rep(1, 100)))
  out = g.weardec(metalong, wearthreshold, ws2, params_cleaning = params_cleaning)
  expect_equal(sum(out$r1), 20)
  expect_equal(sum(out$r2), 100)
  expect_equal(sum(out$r3), 0)
  
  # Wear 20 in the middle + clipping
  metalong = data.frame(nonwearscore = createNW(Nnonwear = 20, invert = TRUE),
                        clippingscore = c(rep(0, Ntotal - 200), rep(1, 200)))
  out = g.weardec(metalong, wearthreshold, ws2, params_cleaning = params_cleaning)
  expect_equal(sum(out$r1), 280)
  expect_equal(sum(out$r2), 200)
  expect_equal(sum(out$r3), 20)
  
  
  # Non-wear 100 in middle + clipping
  metalong = data.frame(nonwearscore = createNW(Nnonwear = 100),
                        clippingscore = c(rep(0, Ntotal - 15), rep(1, 15)))
  out = g.weardec(metalong, wearthreshold, ws2, params_cleaning = params_cleaning)
  expect_equal(sum(out$r1), 100)
  expect_equal(sum(out$r2), 15)
  expect_equal(sum(out$r3), 0)
  
  # Wear 100 in the middle + clipping
  metalong = data.frame(nonwearscore = createNW(Nnonwear = 100, invert = TRUE),
                        clippingscore = c(rep(0, Ntotal - 280), rep(1, 280)))
    out = g.weardec(metalong, wearthreshold, ws2, params_cleaning = params_cleaning)
  expect_equal(sum(out$r1), 200)
  expect_equal(sum(out$r2), 280)
  expect_equal(sum(out$r3), 0)
  
  # Filter short nonwear during the night: defined with a set window
  params_cleaning$nonwearFiltermaxHours = 12
  params_cleaning$nonwearFilterWindow = c(19, 9)
  alternatingNonwear = rep(c(0, 0, 0, 0, 3, 3, 3, 3), ceiling(Ntotal / 8))
  metalong = data.frame(nonwearscore = alternatingNonwear[1:Ntotal],
                        clippingscore = c(rep(0, Ntotal - 280), rep(1, 280)))
  tz = "Europe/Amsterdam"
  timestamp = POSIXtime2iso8601(seq(as.POSIXct("2024-10-29 00:00:00", tz = tz),
                                    by = ws2, length.out = nrow(metalong)), tz = tz)
  metalong$timestamp = timestamp
  out = g.weardec(metalong, wearthreshold, ws2, params_cleaning = params_cleaning)
  expect_equal(sum(out$r1), 60)
  expect_equal(sum(out$r2), 280)
  expect_equal(sum(out$r3), 48)
})

library(GGIR)
context("convertEpochData")
test_that("External epoch data is correctly converted", {
  skip_on_cran()
  params_general = load_params()$params_general
  params_general[["overwrite"]] = TRUE
  params_general[["extEpochData_timeformat"]] = "%d/%m/%Y %H:%M:%S"
  # Set smaller than usual windowsizes, because this test recordings are clipped short
  params_general[["windowsizes"]][2] = 60
  params_general[["windowsizes"]][3] = 120
  
  # Function to create a new directory (dn) and copies the test file there
  dn = "./tmp_testdata"
  move2folder = function(fn, dn) {
    if (!dir.exists(dn)) {
      dir.create(dn)
    }
    file.copy(from = fn, to = dn)
  }
  # Create output directory
  QCout = "./output_tmp_testdata/results/QC"
  QCbasis = "./output_tmp_testdata/meta/basic"
  if (!dir.exists(QCout)) dir.create(QCout, recursive = TRUE)
  if (!dir.exists(QCbasis)) dir.create(QCbasis, recursive = TRUE)
  outputdir = "./output_tmp_testdata"
  
  # Tidy up by deleting output folder
  if (file.exists(outputdir)) unlink(outputdir, recursive = TRUE)
  
  # AWD
  cat("\nActiwatch AWD")
  move2folder(system.file("testfiles/Actiwatch.AWD", package = "GGIR")[1], dn)
  params_general[["windowsizes"]][1] = 60
  params_general[["dataFormat"]] = "actiwatch_awd"
  params_general[["extEpochData_timeformat"]] = "%d-%b-%Y %H:%M:%S"
  convertEpochData(datadir = dn, metadatadir = "./output_tmp_testdata",
                   params_general = params_general)
  
  if (dir.exists(dn))  unlink(dn, recursive = TRUE)
  load(paste0(QCbasis, "/meta_Actiwatch.AWD.RData"))
  expect_equal(sum(M$metalong$nonwearscore), 63)
  expect_equal(nrow(M$metashort), 329)
  expect_equal(ncol(M$metashort), 3)
  expect_equal(colnames(M$metashort), c("timestamp", "ExtAct", "marker"))

  # Tidy up by deleting output folder
  if (file.exists(outputdir)) unlink(outputdir, recursive = TRUE)
  
  # Actiwatch CSV
  cat("\nActiwatch CSV")
  move2folder(system.file("testfiles/Actiwatch.csv", package = "GGIR")[1], dn)
  params_general[["windowsizes"]][1] = 15
  params_general[["dataFormat"]] = "actiwatch_csv"
  params_general[["extEpochData_timeformat"]] = "%d-%m-%Y %H:%M:%S"
  expect_error(convertEpochData(datadir = dn, metadatadir = "./output_tmp_testdata",
               params_general = params_general))
  params_general[["extEpochData_timeformat"]] = "%d/%m/%Y %H:%M:%S"
  convertEpochData(datadir = dn, metadatadir = "./output_tmp_testdata",
                   params_general = params_general)
  if (dir.exists(dn))  unlink(dn, recursive = TRUE)
  load(paste0(QCbasis, "/meta_Actiwatch.csv.RData"))
  expect_equal(sum(M$metalong$nonwearscore), 600)
  expect_equal(nrow(M$metashort), 860)
  expect_equal(ncol(M$metashort), 4)
  # Note: It says markering in next file because test file was Dutch
  # this is an open issue https://github.com/wadpac/GGIRread/issues/75
  expect_equal(colnames(M$metashort)[1:4], c("timestamp", "ExtAct", "markering", "ExtSleep"))

  # Tidy up by deleting output folder
  if (file.exists(outputdir)) unlink(outputdir, recursive = TRUE)
  
  # ukbiobank CSV
  cat("\nukbiobank CSV")
  move2folder(system.file("testfiles/ukbiobank.csv", package = "GGIR")[1], dn)
  params_general[["windowsizes"]][1] = 5
  params_general[["dataFormat"]] = "ukbiobank_csv"
  convertEpochData(datadir = dn, metadatadir = "./output_tmp_testdata",
                   params_general = params_general)
  if (dir.exists(dn))  unlink(dn, recursive = TRUE)
  load(paste0(QCbasis, "/meta_ukbiobank.csv.RData"))
  expect_equal(sum(M$metalong$nonwearscore), 0)
  expect_equal(nrow(M$metashort), 492)
  expect_equal(ncol(M$metashort), 2)
  expect_equal(colnames(M$metashort), c("timestamp", "LFENMO"))

  # Tidy up by deleting output folder
  if (file.exists(outputdir)) unlink(outputdir, recursive = TRUE)
  
  # ActiGraph mode 61
  cat("\nActiGraph mode 61")
  move2folder(system.file("testfiles/ActiGraph61.csv", package = "GGIR")[1], dn)
  params_general[["windowsizes"]][1] = 5
  params_general[["dataFormat"]] = "actigraph_csv"
  params_general[["extEpochData_timeformat"]] = "%m/%d/%Y %H:%M:%S"
  convertEpochData(datadir = dn, metadatadir = "./output_tmp_testdata",
                   params_general = params_general)
  if (dir.exists(dn))  unlink(dn, recursive = TRUE)
  load(paste0(QCbasis, "/meta_ActiGraph61.csv.RData"))
  expect_equal(sum(M$metalong$nonwearscore), 165)
  expect_equal(nrow(M$metashort), 984)
  expect_equal(ncol(M$metashort), 6)
  expect_true(all(c("timestamp", "NeishabouriCount_x", "NeishabouriCount_y", 
                    "NeishabouriCount_z", "ExtStep") %in% colnames(M$metashort)))
  
  # Tidy up by deleting output folder
  if (file.exists(outputdir)) unlink(outputdir, recursive = TRUE)
  
  # ActiGraph mode 13
  cat("\nActiGraph mode 13")
  move2folder(system.file("testfiles/ActiGraph13.csv", package = "GGIR")[1], dn)
  params_general[["windowsizes"]][1] = 15
  params_general[["dataFormat"]] = "actigraph_csv"
  params_general[["extEpochData_timeformat"]] = "%m/%d/%Y %H:%M:%S"
  convertEpochData(datadir = dn, metadatadir = "./output_tmp_testdata",
                   params_general = params_general)
  if (dir.exists(dn))  unlink(dn, recursive = TRUE)
  load(paste0(QCbasis, "/meta_ActiGraph13.csv.RData"))
  expect_equal(sum(M$metalong$nonwearscore), 291)
  expect_equal(nrow(M$metashort), 988)
  expect_equal(ncol(M$metashort), 6)
  expect_true(all(c("timestamp", "NeishabouriCount_x", "NeishabouriCount_y",  "ExtStep",
                    "NeishabouriCount_z") %in% colnames(M$metashort)))
  
  # Tidy up by deleting output folder
  if (file.exists(outputdir)) unlink(outputdir, recursive = TRUE)

  # ActiGraph with headers
  cat("\nActiGraph with headers")
  move2folder(system.file("testfiles/ActiGraph13_timestamps_headers.csv", package = "GGIR")[1], dn)
  params_general[["windowsizes"]][1] = 1
  params_general[["dataFormat"]] = "actigraph_csv"
  params_general[["extEpochData_timeformat"]] = "%d-%m-%Y %H:%M:%S"
  convertEpochData(datadir = dn, metadatadir = "./output_tmp_testdata",
                   params_general = params_general)
  if (dir.exists(dn))  unlink(dn, recursive = TRUE)
  load(paste0(QCbasis, "/meta_ActiGraph13_timestamps_headers.csv.RData"))
  expect_equal(sum(M$metalong$nonwearscore), 0)
  expect_equal(nrow(M$metashort), 960)
  expect_equal(ncol(M$metashort), 6)
  expect_true(all(c("timestamp", "NeishabouriCount_x", "NeishabouriCount_y", 
                    "NeishabouriCount_z", "ExtStep", "NeishabouriCount_vm") %in% colnames(M$metashort)))
  
  # Tidy up by deleting output folder
  if (file.exists(outputdir)) unlink(outputdir, recursive = TRUE)
  

  # Sensewear
  cat("\nSensewear")
  move2folder(system.file("testfiles/sensewear.xls", package = "GGIR")[1], dn)
  params_general[["windowsizes"]][1] = 60
  params_general[["windowsizes"]][2] = 60
  params_general[["windowsizes"]][3] = 120
  params_general[["dataFormat"]] = "sensewear_xls"
  params_general[["extEpochData_timeformat"]] = "%d-%b-%Y %H:%M:%S"
  convertEpochData(datadir = dn, metadatadir = "./output_tmp_testdata",
                   params_general = params_general)
  if (dir.exists(dn))  unlink(dn, recursive = TRUE)
  load(paste0(QCbasis, "/meta_sensewear.xls.RData"))
  expect_equal(sum(M$metalong$nonwearscore), 0)
  expect_equal(nrow(M$metashort), 60)
  expect_equal(ncol(M$metashort), 4)
  expect_true(all(c("timestamp", "ExtAct", "ExtStep",
                    "ExtSleep") %in% colnames(M$metashort)))
  
  # Tidy up by deleting output folder
  if (file.exists(outputdir)) unlink(outputdir, recursive = TRUE)
  
  # Philips Health Band
  cat("\nPhilips Health Band")
  move2folder(system.file("testfiles/DataList_AH1234567890_PhilipsHealthBand.xlsx",
                          package = "GGIRread")[1], dn)
  move2folder(system.file("testfiles/Sleep_Wake_AH1234567890_PhilipsHealthBand.xlsx",
                          package = "GGIRread")[1], dn)
  params_general = load_params()$params_general
  params_general[["windowsizes"]][1] = 60
  params_general[["dataFormat"]] = "phb_xlsx"
  params_general[["extEpochData_timeformat"]] = "%m-%d-%Y %H:%M:%S"
  convertEpochData(datadir = dn, metadatadir = "./output_tmp_testdata",
                   params_general = params_general)
  if (dir.exists(dn))  unlink(dn, recursive = TRUE)
  
  load(paste0(QCbasis, "/meta_tmp_testdata.RData"))
  expect_equal(sum(M$metalong$nonwearscore), 0)
  expect_equal(nrow(M$metashort), 240)
  expect_equal(ncol(M$metashort), 5)
  expect_true(all(c("timestamp", "ExtAct", "ExtStep", "marker",
                    "ExtSleep") %in% colnames(M$metashort)))
  
  # Tidy up by deleting output folder
  if (file.exists(outputdir)) unlink(outputdir, recursive = TRUE)
  
  cat("\nFitbit")
  dir.create(paste0(dn, "/participant1"), recursive = TRUE)
  move2folder(system.file("testfiles/calories-1995-06-23_Fitbit.json",
                          package = "GGIRread")[1], paste0(dn, "/participant1"))
  move2folder(system.file("testfiles/sleep-1995-06-23_Fitbit.json",
                          package = "GGIRread")[1], paste0(dn, "/participant1"))
  move2folder(system.file("testfiles/steps-1995-06-23_Fitbit.json",
                          package = "GGIRread")[1], paste0(dn, "/participant1"))
  params_general = load_params()$params_general
  params_general[["windowsizes"]][1] = 60
  params_general[["dataFormat"]] = "fitbit_json"
  convertEpochData(datadir = dn, metadatadir = "./output_tmp_testdata",
                   params_general = params_general)
  if (dir.exists(dn))  unlink(dn, recursive = TRUE)
  load(paste0(QCbasis, "/meta_participant1.RData"))
  expect_equal(sum(M$metalong$nonwearscore), 4782)
  expect_equal(nrow(M$metashort), 23925)
  expect_equal(ncol(M$metashort), 4)
  expect_equal(ncol(M$QClog), 10)
  expect_equal(M$QClog$filehealth_2vars_missing, 99.81)
  expect_true(all(c("timestamp", "ExtAct", "ExtStep", "ExtSleep") %in% colnames(M$metashort)))
  
  # Tidy up by deleting output folder
  if (file.exists(outputdir)) unlink(outputdir, recursive = TRUE)
  
  
})

library(GGIR)
context("g.part5.analyseRest")

tz = "Pacific/Auckland"
ds_names = rep("", 40)
dsummary = matrix("", 1, 40)
startday = as.POSIXct(x = "2022-06-02 08:00:00", tz = tz)
ts = data.frame(time = seq(startday, startday + (16 * 3600), by = 60))
ts$sibdetection = 0

test_that("Overlap 1 nap and 1 sib", {
  fi = 1
  di = 1
  sibreport = data.frame(ID = rep("test123", 2), type = c("nap", "sib"),
                         start = c("2022-06-02 14:00:00", "2022-06-02 14:05:00"),
                         end = c("2022-06-02 14:20:00", "2022-06-02 14:20:00"))
  sibreport$duration = as.numeric(difftime(time1 = sibreport$end,
                                           time2 = sibreport$start, units = "mins", tz = tz))
  params_sleep = load_params()$params_sleep
  params_sleep[["possible_nap_dur"]] =  c(0, 240)
  params_sleep[["possible_nap_window"]] =  c(9, 18)
  restAnalyses = g.part5.analyseRest(sibreport = sibreport, dsummary = dsummary,
                                     ds_names = ds_names, fi = fi, di = di,
                                     ts = ts, tz = tz,
                                     params_sleep = params_sleep)
  fi = restAnalyses$fi
  di = restAnalyses$di
  dsummary = restAnalyses$dsummary
  ds_names = restAnalyses$ds_names
  dsummary = as.numeric(dsummary[, which(ds_names != "")])
  ds_names = ds_names[ds_names != ""]
  names(dsummary) = ds_names
  dsummary = as.data.frame(t(dsummary))
  
  expect_equal(dsummary$nbouts_day_denap, 1)
  expect_equal(dsummary$nbouts_day_srnap, 1)
  expect_equal(dsummary$frag_mean_dur_denap_day, 15)
  expect_equal(dsummary$frag_mean_dur_srnap_day, 20)
  expect_equal(dsummary$perc_denap_overl_srnap, 100)
  expect_equal(dsummary$perc_srnap_overl_denap, 75)
  expect_equal(sum(dsummary), 323)
})

test_that("Overlap 1 nonwear and 1 sib", {
  fi = 1
  di = 1
  sibreport = data.frame(ID = rep("test123", 2), type = c("nonwear", "sib"),
                         start = c("2022-06-02 14:00:00", "2022-06-02 14:05:00"),
                         end = c("2022-06-02 14:20:00", "2022-06-02 14:20:00"))
  sibreport$duration = as.numeric(difftime(time1 = sibreport$end, time2 = sibreport$start, units = "mins", tz = tz))
  params_sleep = load_params()$params_sleep
  params_sleep[["possible_nap_dur"]] =  c(0, 240)
  params_sleep[["possible_nap_window"]] =  c(9, 18)
  restAnalyses = g.part5.analyseRest(sibreport = sibreport, dsummary = dsummary,
                                     ds_names = ds_names, fi = fi, di = di,
                                     ts = ts, tz = tz,params_sleep = params_sleep)
  fi = restAnalyses$fi
  di = restAnalyses$di
  dsummary = restAnalyses$dsummary
  ds_names = restAnalyses$ds_names
  dsummary = as.numeric(dsummary[, which(ds_names != "")])
  ds_names = ds_names[ds_names != ""]
  names(dsummary) = ds_names
  dsummary = as.data.frame(t(dsummary))
  
  expect_equal(dsummary$nbouts_day_denap, 1)
  expect_equal(dsummary$nbouts_day_srnonw, 1)
  expect_equal(dsummary$frag_mean_dur_denap_day, 15)
  expect_equal(dsummary$frag_mean_dur_srnonw_day, 20)
  expect_equal(dsummary$perc_denap_overl_srnonw, 100)
  expect_equal(dsummary$perc_srnonw_overl_denap, 75)
  expect_equal(sum(dsummary), 323)
})


test_that("No overlap 1 nonwear, 1 nap, and 1 sib", {
  fi = 1
  di = 1
  sibreport = data.frame(ID = rep("test123", 3), type = c("nonwear", "nap", "sib"),
                         start = c("2022-06-02 12:00:00", "2022-06-02 13:00:00", "2022-06-02 15:00:00"),
                         end = c("2022-06-02 12:20:00", "2022-06-02 13:20:00", "2022-06-02 15:20:00"))
  sibreport$duration = as.numeric(difftime(time1 = sibreport$end, time2 = sibreport$start, units = "mins", tz = tz))
  params_sleep = load_params()$params_sleep
  params_sleep[["possible_nap_dur"]] =  c(0, 240)
  params_sleep[["possible_nap_window"]] =  c(9, 18)
  restAnalyses = g.part5.analyseRest(sibreport = sibreport, dsummary = dsummary,
                                     ds_names = ds_names, fi = fi, di = di,
                                     ts = ts, tz = tz, params_sleep = params_sleep)
  fi = restAnalyses$fi
  di = restAnalyses$di
  dsummary = restAnalyses$dsummary
  expect_equal(sum(as.numeric(restAnalyses$dsummary), na.rm = TRUE), 129)
  
  ds_names = restAnalyses$ds_names
  dsummary = as.numeric(dsummary[, which(ds_names != "")])
  ds_names = ds_names[ds_names != ""]
  names(dsummary) = ds_names
  dsummary = as.data.frame(t(dsummary))
  
  expect_equal(ncol(dsummary), 27)
  expect_equal(nrow(dsummary), 1)
  expect_equal(dsummary$nbouts_day_denap, 1)
  expect_equal(dsummary$nbouts_day_srnap, 1)
  expect_equal(dsummary$nbouts_day_srnonw, 1)
  expect_equal(dsummary$frag_mean_dur_denap_day, 20)
  expect_equal(dsummary$frag_mean_dur_srnap_day, 20)
  expect_equal(dsummary$frag_mean_dur_srnonw_day, 20)
  expect_equal(dsummary$perc_denap_overl_srnap, 0)
  expect_equal(dsummary$perc_denap_overl_srnonw, 0)
  expect_equal(dsummary$perc_srnonw_overl_denap, 0)
  expect_equal(sum(dsummary, na.rm = TRUE), 129)
  
})

library(GGIR)
context("g.part3_correct_guider")
test_that("Guider that misses half of the night can be corrected", {
  skip_on_cran()
  # create dummy test data
  desiredtz = "Europe/London"
  epochSize = 5
  time = seq(as.POSIXct("2025-05-21 15:42:55", tz = desiredtz),
             as.POSIXct("2025-05-23 15:42:55", tz = desiredtz), by = epochSize)
  spt_crude_estimate = rep(0, length(time))

  # define classified sleep windows
  SPTE_start = c(22, 26)
  SPTE_end = c(28, 32)
  spt_crude_estimate[which(time >= as.POSIXct("2025-05-21 22:00:00", tz = desiredtz) &
                             time < as.POSIXct("2025-05-22 04:00:00", tz = desiredtz))] = 2
  spt_crude_estimate[which(time >= as.POSIXct("2025-05-23 02:00:00", tz = desiredtz) &
                             time < as.POSIXct("2025-05-23 08:00:00", tz = desiredtz))] = 2

  # Define additional sleep windows that are missed by guider algorithm
  spt_crude_estimate[which(time >= as.POSIXct("2025-05-22 05:00:00", tz = desiredtz) &
                             time < as.POSIXct("2025-05-22 06:00:00", tz = desiredtz))] = 1
  spt_crude_estimate[which(time >= as.POSIXct("2025-05-23 00:00:00", tz = desiredtz) &
                             time < as.POSIXct("2025-05-23 01:00:00", tz = desiredtz))] = 1
  output = data.frame(time = POSIXtime2iso8601(time, desiredtz),
                      spt_crude_estimate = spt_crude_estimate,
                      invalid = rep(0, length(time)),
                      night = rep(1, length(time)),
                      T5A5 = rep(1, length(time)))
  output$night[which(time >= as.POSIXct("2025-05-22 12:00:00", tz = desiredtz) &
                       time < as.POSIXct("2025-05-23 12:00:00", tz = desiredtz))] = 2
  output$night[which(time >= as.POSIXct("2025-05-23 12:00:00", tz = desiredtz))] = 3

  SLE = list(output = output,
             SPTE_end = SPTE_end,
             SPTE_start = SPTE_start)
  # The test whether missed sleep windows are correctly identified
  params_sleep = GGIR::load_params()$params_sleep
  params_sleep[["guider_cor_do"]] = TRUE # to run correction
  params_sleep[["guider_cor_maxgap_hrs "]] = Inf # to use max gap size
  params_sleep[["guider_cor_meme_min_dys"]] = Inf # to not do meme
  params_sleep[["guider_cor_min_frac_sib"]]  = 0.8
  params_sleep[["guider_cor_min_hrs"]]  = 1
  params_sleep[["guider_cor_meme_min_hrs"]]  = 2
  SLE = g.part3_correct_guider(SLE, desiredtz, epochSize,
                               params_sleep = params_sleep)
  expect_equal(SLE$SPTE_start, c(22, 24))
  expect_equal(SLE$SPTE_end, c(29.999, 31.999))
  expect_equal(nrow(SLE$output), 34561)
})


test_that("Guider that focusses on afternoon nap can be corrected", {
  skip_on_cran()
  #=============================================================
  # One night with sleep in the afternoon
  #=============================================================
  # create dummy test data
  desiredtz = "Europe/London"
  epochSize = 5
  time = seq(as.POSIXct("2025-05-21 15:42:55", tz = desiredtz),
             as.POSIXct("2025-05-25 15:42:55", tz = desiredtz), by = epochSize)
  spt_crude_estimate = rep(0, length(time))
  # define classified sleep windows
  SPTE_start = c(13, 22, 26, 22)
  SPTE_end = c(16, 28, 32, 30)
  spt_crude_estimate[which(time >= as.POSIXct("2025-05-21 13:00:00", tz = desiredtz) &
                             time < as.POSIXct("2025-05-21 16:00:00", tz = desiredtz))] = 2
  spt_crude_estimate[which(time >= as.POSIXct("2025-05-22 22:00:00", tz = desiredtz) &
                             time < as.POSIXct("2025-05-23 04:00:00", tz = desiredtz))] = 2
  spt_crude_estimate[which(time >= as.POSIXct("2025-05-24 02:00:00", tz = desiredtz) &
                             time < as.POSIXct("2025-05-24 08:00:00", tz = desiredtz))] = 2
  spt_crude_estimate[which(time >= as.POSIXct("2025-05-24 22:00:00", tz = desiredtz) &
                             time < as.POSIXct("2025-05-25 06:00:00", tz = desiredtz))] = 2
  
  # Define additional sleep windows that are missed by guider algorithm
  
  spt_crude_estimate[which(time >= as.POSIXct("2025-05-22 01:00:00", tz = desiredtz) &
                             time < as.POSIXct("2025-05-22 03:30:00", tz = desiredtz))] = 1
  spt_crude_estimate[which(time >= as.POSIXct("2025-05-22 05:00:00", tz = desiredtz) &
                             time < as.POSIXct("2025-05-22 06:00:00", tz = desiredtz))] = 1
  output = data.frame(time = POSIXtime2iso8601(time, desiredtz),
                      spt_crude_estimate = spt_crude_estimate,
                      invalid = rep(0, length(time)),
                      night = rep(1, length(time)),
                      T5A5 = rep(1, length(time)))
  
  output$night[which(time < as.POSIXct("2025-05-22 12:00:00", tz = desiredtz))] = 1
  output$night[which(time >= as.POSIXct("2025-05-22 12:00:00", tz = desiredtz) &
                       time < as.POSIXct("2025-05-23 12:00:00", tz = desiredtz))] = 2
  output$night[which(time >= as.POSIXct("2025-05-23 12:00:00", tz = desiredtz) &
                       time < as.POSIXct("2025-05-24 12:00:00", tz = desiredtz))] = 3
  output$night[which(time >= as.POSIXct("2025-05-24 12:00:00", tz = desiredtz))] = 4
  
  SLE = list(output = output,
             SPTE_end = SPTE_end,
             SPTE_start = SPTE_start)
  # The test whether missed sleep windows are correctly identified
  params_sleep = GGIR::load_params()$params_sleep
  params_sleep[["guider_cor_do"]] = TRUE # to run correction
  params_sleep[["guider_cor_maxgap_hrs "]] = 3 # to use max gap size
  params_sleep[["guider_cor_meme_min_dys"]] = 3 # to not do meme
  params_sleep[["guider_cor_min_frac_sib"]]  = 0.8
  params_sleep[["guider_cor_min_hrs"]]  = 1
  params_sleep[["guider_cor_meme_min_hrs"]]  = 2
  SLE = g.part3_correct_guider(SLE, desiredtz, epochSize,
                               params_sleep = params_sleep)
  expect_equal(SLE$SPTE_start, c(25, 22, 26, 22))
  expect_equal(SLE$SPTE_end, c(29.999, 28, 32, 30))
  expect_equal(nrow(SLE$output), 69121)
  
})

library(GGIR)
context("splitRecords")

test_that("Recording can be split", {
  skip_on_cran()
  #==========================================
  # Create dummy file
  data(data.getmeta)
  data(data.calibrate)
  data(data.inspectfile)
  M = data.getmeta
  tz = "Europe/London"
  
  # duplicate data
  M$metashort =  rbind(M$metashort, M$metashort)
  M$metalong =  rbind(M$metalong, M$metalong)
  starttime = as.POSIXct(M$metashort$timestamp[1], tz = tz)
  M$metashort$timestamp = seq(from = starttime, by = 5, length.out = nrow(M$metashort))
  starttime = as.POSIXct(M$metalong$timestamp[1], tz = tz)
  M$metalong$timestamp = seq(from = starttime, by = 900, length.out = nrow(M$metalong))
  
  # change timestamp format to what it has been since 2017
  M$metashort$timestamp = POSIXtime2iso8601(x = M$metashort$timestamp, tz = tz)
  M$metalong$timestamp = POSIXtime2iso8601(x = M$metalong$timestamp, tz = tz)
  colnames(M$metashort)[which(colnames(M$metashort) == "angle")] = "anglez"
  I = data.inspectfile
  I$filename = "1_1-1-1900.bin"
  C = data.calibrate
  
  if (dir.exists("./output_test"))  unlink("./output_test", recursive = TRUE)
  dn = "./output_test/meta/basic"
  if (!dir.exists(dn)) {
    dir.create(dn, recursive = TRUE)
  }
  result_folder = "./output_test/results/QC"
  if (!dir.exists(result_folder)) {
    dir.create(result_folder, recursive = TRUE)
  }
  
  filefoldername = filename_dir = I$filename
  tail_expansion_log = NULL
  save(M, C, I,
       filefoldername = filefoldername,
       filename_dir = filename_dir,
       tail_expansion_log = tail_expansion_log,
       file = paste0(dn, "/meta_1_1-1-1900.bin.RData"))
  
  times_to_split = data.frame(ID = "1", time1 = "2013-11-15 00:05", time2 = "2013-11-18")
  csv_file = "time_chunks_per_participant.csv"
  write.csv(x = times_to_split, file = csv_file, row.names = FALSE)
  
  # Prepare params object
  params_general = load_params()$params_general
  params_general[["desiredtz"]] = tz
  params_general[["idloc"]] = 2
  params_general[["recording_split_times"]] = "time_chunks_per_participant.csv"
  params_general[["recording_split_overlap"]] = 0
  params_general[["recording_split_timeformat"]] = "%Y-%m-%d %H:%M"
  params_general[["recording_split_ignore_edges"]] == FALSE
  
  splitRecords(metadatadir = "./output_test",
               params_general = params_general)
  
  # File created?
  expect_true(dir.exists("./output_test/meta/basic"))
  expect_equal(length(dir("./output_test/meta/basic")), 2)
  expect_true(file.exists("./output_test/meta/basic/meta_1_1-1-1900_split1_startrecTOtime1.bin.RData"))
  expect_true(file.exists("./output_test/meta/basic/meta_1_1-1-1900_split2_time1TOendrec.bin.RData"))

  # File content correct?
  load("./output_test/meta/basic/meta_1_1-1-1900_split1_startrecTOtime1.bin.RData")
  expect_equal(nrow(M$metashort), 8820)
  expect_equal(nrow(M$metalong), 49)
  expect_equal(M$metashort$timestamp[1], "2013-11-14T11:45:00+0000")
  expect_equal(tail(M$metashort$timestamp, n = 1), "2013-11-14T23:59:55+0000")
  expect_equal(M$metalong$timestamp[1],  "2013-11-14T11:45:00+0000")
  expect_equal(tail(M$metalong$timestamp, n = 1), "2013-11-14T23:45:00+0000")
  load("./output_test/meta/basic/meta_1_1-1-1900_split2_time1TOendrec.bin.RData")
  expect_equal(nrow(M$metashort), 28619)
  expect_equal(nrow(M$metalong), 158)
  expect_equal(M$metashort$timestamp[1], "2013-11-15T00:00:00+0000")
  expect_equal(tail(M$metashort$timestamp, n = 1), "2013-11-16T15:44:50+0000")
  expect_equal(M$metalong$timestamp[1], "2013-11-15T00:00:00+0000")
  expect_equal(tail(M$metalong$timestamp, n = 1), "2013-11-16T15:15:00+0000")
  
  # Test in context of GGIR as a whole:
  GGIR(datadir = "D:/test", outputdir = ".", desiredtz = tz,
       mode = 2:5, do.report = c(2, 4, 5), verbose = FALSE, idloc = 2,
       recording_split_times = "time_chunks_per_participant.csv",
       visualreport = FALSE, recording_split_overlap = 0,
       recording_split_timeformat = "%Y-%m-%d %H:%M")
  
  # We expect split names to be logged in part 2 csv reports
  P2 = read.csv("./output_test/results/part2_summary.csv")
  expect_equal(P2$split1_name, c("startrec", "time1"))
  expect_equal(P2$split2_name, c("time1", "endrec"))
  P2day = read.csv("./output_test/results/part2_daysummary.csv")
  expect_equal(P2day$split1_name, c("startrec", "time1", "time1"))
  expect_equal(P2day$split2_name, c("time1", "endrec", "endrec"))
  QC = read.csv("./output_test/results/QC/data_quality_report.csv")
  expect_equal(QC$split1_name, c("startrec", "time1"))
  expect_equal(QC$split2_name, c("time1", "endrec"))
  # We expect split names to be logged in part 4 csv reports
  P4night = read.csv("./output_test/results/part4_nightsummary_sleep_cleaned.csv")
  expect_equal(P4night$split1_name, "time1")
  expect_equal(P4night$split2_name, "endrec")
  P4 = read.csv("./output_test/results/part4_summary_sleep_cleaned.csv")
  expect_equal(P4$split1_name, "time1")
  expect_equal(P4$split2_name, "endrec")
  # We expect split names to be logged in part 5 csv reports
  P5day = read.csv("./output_test/results/part5_daysummary_MM_L40M100V400_T5A5.csv")
  expect_equal(P5day$split1_name, "time1")
  expect_equal(P5day$split2_name, "endrec")
  P5 = read.csv("./output_test/results/part5_personsummary_MM_L40M100V400_T5A5.csv")
  expect_equal(P5$split1_name, "time1")
  expect_equal(P5$split2_name, "endrec")
  
  if (dir.exists("./output_test"))  unlink("./output_test", recursive = TRUE)
  if (file.exists(csv_file)) unlink(csv_file, recursive = TRUE)
})


test_that("Recording can be split and ignore edges", {
  skip_on_cran()
  
  #==========================================
  # Create dummy file
  data(data.getmeta)
  data(data.calibrate)
  data(data.inspectfile)
  M = data.getmeta
  
  # change timestamp format to what it has been since 2017
  M$metashort$timestamp = POSIXtime2iso8601(x = M$metashort$timestamp, tz = "Europe/London")
  M$metalong$timestamp = POSIXtime2iso8601(x = M$metalong$timestamp, tz = "Europe/London")
  I = data.inspectfile
  I$filename = "1_1-1-1900.bin"
  C = data.calibrate
  
  dn = "./testfolder/meta/basic"
  if (!dir.exists(dn)) {
    dir.create(dn, recursive = TRUE)
  }
  filefoldername = filename_dir = I$filename
  tail_expansion_log = NULL
  save(M, C, I,
       filefoldername = filefoldername,
       filename_dir = filename_dir,
       tail_expansion_log = tail_expansion_log,
       file = paste0(dn, "/meta_1_1-1-1900.bin.RData"))
  
  times_to_split = data.frame(ID = "1", time1 = "2013-11-15 00:05", time2 = "2013-11-16")
  csv_file = "time_chunks_per_participant.csv"
  write.csv(x = times_to_split, file = csv_file, row.names = FALSE)
  
  # Prepare params object
  params_general = load_params()$params_general
  params_general[["desiredtz"]] = "Europe/Amsterdam"
  params_general[["idloc"]] = 2
  params_general[["recording_split_times"]] = "time_chunks_per_participant.csv"
  params_general[["recording_split_overlap"]] = 0
  params_general[["recording_split_timeformat"]] = "%Y-%m-%d %H:%M"
  params_general[["recording_split_ignore_edges"]] = TRUE
  
  splitRecords(metadatadir = "./testfolder",
               params_general = params_general)
  
  # File created?
  expect_true(dir.exists("./testfolder/meta/basic"))
  expect_equal(length(dir("./testfolder/meta/basic")), 1)
  expect_true(file.exists("./testfolder/meta/basic/meta_1_1-1-1900.bin.RData"))
  
  # File content correct?
  load("./testfolder/meta/basic/meta_1_1-1-1900.bin.RData")
  expect_equal(nrow(M$metashort), 18720)
  expect_equal(nrow(M$metalong), 104)
  expect_equal(M$metashort$timestamp[1], "2013-11-14T11:45:00+0000")
  expect_equal(tail(M$metashort$timestamp, n = 1), "2013-11-15T13:44:55+0000")
  expect_equal(M$metalong$timestamp[1],  "2013-11-14T11:45:00+0000")
  expect_equal(tail(M$metalong$timestamp, n = 1), "2013-11-15T13:30:00+0000")
  
  if (dir.exists("./testfolder"))  unlink("./testfolder", recursive = TRUE)
  if (file.exists(csv_file)) unlink(csv_file, recursive = TRUE)
})

library(GGIR)
context("g.part6")

test_that("Part 6 with household co-analysis", {

  # Create test files for household co-analysis
  metadatadir = "./output_testpart6"
  part6_threshold_combi = "40_120_400"
  dn = paste0(metadatadir, "/meta/ms5.outraw/", part6_threshold_combi)
  if (!dir.exists(dn)) dir.create(path = dn, recursive = TRUE)
  dn2 = paste0(metadatadir, "/meta/basic")
  if (!dir.exists(dn2)) dir.create(path = dn2, recursive = TRUE)
  data(data.metalong)
  M = data.metalong
  save(M, file = paste0(dn2, "/meta_800-900-001_left wrist.bin.RData"))
  save(M, file = paste0(dn2, "/meta_800-900-002_left wrist.bin.RData"))
  save(M, file = paste0(dn2, "/meta_800-900-003_left wrist.bin.RData"))
  
  
  # Update parameters to align with datset
  params_general = load_params(topic = "general")$params_general
  params_general[["desiredtz"]] = "America/Curacao"
  params_phyact = load_params(topic = "phyact")$params_phyact
  params_phyact[["part6_threshold_combi"]] = part6_threshold_combi
  thresholds = as.numeric(unlist(strsplit(part6_threshold_combi, "_")))
  params_phyact[["threshold.lig"]] = thresholds[1]
  params_phyact[["threshold.mod"]] = thresholds[2] 
  params_phyact[["threshold.vig"]] = thresholds[3]
  params_247 = load_params(topic = "247")$params_247
  params_247[["part6HCA"]] = TRUE
  
  # Use time shifts to simulate three household members
  data(data.ts)
  mdat = data.ts
  
  Lnames = c("spt_sleep", "spt_wake_IN", "spt_wake_LIG", "spt_wake_MOD",
             "spt_wake_VIG", "day_IN_unbt", "day_LIG_unbt", "day_MOD_unbt",
             "day_VIG_unbt", "day_MVPA_bts_10", "day_IN_bts_30",
             "day_IN_bts_10_30", "day_LIG_bts_10")
  
  mdat$timenum = mdat$timenum - (5 * 60) 
  mdat$timestamp = as.POSIXct(mdat$timenum, tz =  params_general[["desiredtz"]], origin = "1970-01-01")
  filename = "800-900-001_left wrist.bin"
  save(mdat, filename, Lnames, file = paste0(dn, "/800-900-001_left wrist.RData"))
  mdat$timenum = mdat$timenum + (7 * 60)
  mdat$timestamp = as.POSIXct(mdat$timenum, tz =  params_general[["desiredtz"]], origin = "1970-01-01")
  filename = "800-900-002_left wrist.bin"
  save(mdat, filename, Lnames, file = paste0(dn, "/800-900-002_left wrist.RData"))
  mdat$timenum = mdat$timenum + (14 * 60)
  mdat$timestamp = as.POSIXct(mdat$timenum, tz =  params_general[["desiredtz"]], origin = "1970-01-01")
  filename = "800-900-003_left wrist.bin"
  save(mdat, filename, Lnames, file = paste0(dn, "/800-900-003_left wrist.RData"))
  mdat_file3 = mdat
  
  # Run household co-analysis  
  g.part6(metadatadir = metadatadir,
          params_general = params_general,
          params_phyact = params_phyact,
          params_247 = params_247, verbose = FALSE)
  
  # Check aligned time series output file
  path_to_ts = paste0(metadatadir, "/results/part6HouseholdCoAnalysis/alignedTimeseries/timeseries_HID_900.RData")
  expect_true(file.exists(path_to_ts))
  load(path_to_ts)
  TS = alignedTimeseries
  expect_equal(nrow(TS), 10032)
  expect_equal(ncol(TS), 43)
  expect_equal(sum(TS$ACC.001[1000:1200]), 840.91)
  expect_equal(sum(TS$lightmean.001[1000:1200]), 1.166)
  expect_equal(sum(TS$ACC.002[2000:2300]), 7731.019)
  expect_equal(sum(TS$lightmean.002[2000:2300]), 2223.294, tolerance = 0.002)
  
  # Check pairwise comparisons
  path_to_pairwisecomp = paste0(metadatadir, "/results/part6HouseholdCoAnalysis/pairwise_summary_all_housholds.csv")
  expect_true(file.exists(path_to_pairwisecomp))
  PC = read.csv(path_to_pairwisecomp)
  expect_equal(nrow(PC), 18)
  expect_equal(ncol(PC), 41)
  expect_equal(PC$Npairs, rep(3, 18))
  expect_equal(sum(PC$wakeup_acc_1_before_2_mg), 171.863)
  expect_equal(PC$wakeup_time1[c(1, 8, 16)], c("01:35:00", "02:58:00", "03:39:00"))
  expect_equal(sum(PC$day_Kappa_active), 5.091)
  
  #===================================================================
  # For circadian rhythm analysis we set one window to invalid
  mdat_file3$invalid_wakinghours[which(mdat_file3$window == 4)] = 100 # set one window to invalid
  mdat_file3$invalid_sleepperiod[which(mdat_file3$window == 4)] = 100 # set one window to invalid
  mdat_file3$invalid_fullwindow[which(mdat_file3$window == 4)] = 100 # set one window to invalid
  mdat_file3$ACC[which(mdat$window == 3)] = NA # set one window to invalid
  mdat = mdat_file3
  filename = "800-900-003_left wrist.bin"
  save(mdat, filename, Lnames, file = paste0(dn, "/800-900-003_left wrist.RData"))
  
  # Run Circadian rhythm analysis with default window
  params_247[["part6HCA"]] = FALSE
  params_general[["do.parallel"]] = FALSE
  params_general[["overwrite"]] = TRUE
  params_247[["part2CR"]] = TRUE
  params_247[["part6CR"]] = TRUE
  params_247[["part6Window"]] = c("start", "end")
  params_cleaning = load_params(topic = "cleaning")$params_cleaning
  g.part6(metadatadir = metadatadir,
          params_general = params_general,
          params_phyact = params_phyact,
          params_247 = params_247,
          params_cleaning = params_cleaning,
          f0 = 1, f1 = 3,
          verbose = FALSE)
  path_to_ms6 = paste0(metadatadir, "/meta/ms6.out/800-900-003_left wrist.RData")
  expect_true(file.exists(path_to_ms6))
  
  load(path_to_ms6)
  expect_equal(ncol(output_part6), 50)
  expect_equal(output_part6$starttime, "2022-06-02 03:16:00")
  expect_equal(output_part6$cosinor_mes, 2.471082, tolerance = 0.00001)
  expect_equal(output_part6$cosinorExt_minimum, 1.302779, tolerance = 0.00001)
  expect_equal(output_part6$cosinorExt_MESOR, 2.171, tolerance = 0.00001)
  expect_equal(sum(output_part6[5:27]), 620.6144, tolerance = 0.0001)
  expect_equal(output_part6$SleepRegularityIndex2, 95.87223, tolerance = 0.0001)
  expect_equal(output_part6$SleepRegularityIndex2_Ndaypairs, 4)
  
  # Run Circadian rhythm analysis with non-default window
  params_247[["part6Window"]] = c("W2", "W-1") # second wake till last wake
  g.part6(metadatadir = metadatadir,
          params_general = params_general,
          params_phyact = params_phyact,
          params_247 = params_247,
          params_cleaning = params_cleaning,
          f0 = 1, f1 = 3,
          verbose = FALSE)
  path_to_ms6 = paste0(metadatadir, "/meta/ms6.out/800-900-003_left wrist.RData")
  
  expect_true(file.exists(path_to_ms6))
  
  load(path_to_ms6)
  expect_equal(ncol(output_part6), 50)
  expect_equal(output_part6$starttime, "2022-06-03 01:57:00")
  expect_equal(output_part6$cosinor_mes, 2.431978, tolerance = 0.00001)
  expect_equal(output_part6$cosinorExt_minimum, 1.286542, tolerance = 0.00001)
  expect_equal(output_part6$cosinorExt_MESOR, 2.140128, tolerance = 0.00001)
  expect_equal(sum(output_part6[5:27]), 680.9232, tolerance = 0.0001)
  expect_equal(output_part6$SleepRegularityIndex2, 96.72433, tolerance = 0.0001)
  expect_equal(output_part6$SleepRegularityIndex2_Ndaypairs, 3)
  output_part6_with_invalid = output_part6
  rm(output_part6)
  
  # Assess impact of storing part 5 output without invalid
  mdat = mdat_file3[-which(mdat$window == 4),] # delete the invalid window
  filename = "800-900-003_left wrist.bin"
  save(mdat, filename, Lnames, file = paste0(dn, "/800-900-003_left wrist.RData"))
  
  g.part6(metadatadir = metadatadir,
          params_general = params_general,
          params_phyact = params_phyact,
          params_247 = params_247,
          params_cleaning = params_cleaning,
          f0 = 1, f1 = 3,
          verbose = FALSE)
  path_to_ms6 = paste0(metadatadir, "/meta/ms6.out/800-900-003_left wrist.RData")
  
  expect_true(file.exists(path_to_ms6))
  
  load(path_to_ms6)
  output_part6_without_invalid = output_part6
  
  # Compare output_part6_with_invalid and output_part6_without_invalid
  expect_equal(ncol(output_part6_with_invalid), ncol(output_part6_without_invalid))
  expect_equal(output_part6_with_invalid$starttime, output_part6_without_invalid$starttime)
  expect_equal(output_part6_with_invalid$cosinor_mes, output_part6_without_invalid$cosinor_mes, tolerance = 0.00001)
  expect_equal(output_part6_with_invalid$cosinorExt_minimum, output_part6_without_invalid$cosinorExt_minimum, tolerance = 0.00001)
  expect_equal(output_part6_with_invalid$cosinorExt_MESOR, output_part6_without_invalid$cosinorExt_MESOR, tolerance = 0.00001)
  expect_equal(sum(output_part6_with_invalid[5:27]), sum(output_part6_without_invalid[5:27]), tolerance = 0.0001)

  # Remove test files
  if (file.exists(metadatadir))  unlink(metadatadir, recursive = TRUE)
}
)

create_test_acc_csv = function(sf = 3, Nmin = 2000, storagelocation = c(),
                               start_time = NULL, starts_at_midnight = FALSE) {
  # function to create a test Actigraph csv file needed for testing GGIR
  # adds variation in angle to enable autocalibration
  # adds some activity periods
  # rm(list=ls())
  # sf = 3
  # Nmin = 2000
  # storagelocation = c()
  #===============================================
  gensequence = function(t0,t1) {
    x = round(t0 * 3600 * sf):(round(t1 * 3600 * sf) - 1)
    return(x)
  }
  genrotationmatrix = function(phi) {   # create rotation matrix
    Rx = Ry = matrix(0,3,3) # no z rotation needed, all we need is to populate sphere on all sides
    Ry[1,] = c(cos(phi),0,sin(phi))
    Ry[2,] = c(0,1,0)
    Ry[3,] = c(-sin(phi),0,cos(phi))
    Rx[1,] = c(cos(phi),-sin(phi),0)
    Rx[2,] = c(sin(phi),cos(phi),0)
    Rx[3,] = c(0,0,1)
    invisible(list(Rx = Rx, Ry = Ry))
  }
  
  generate_repeated_timestamp_blocks = function(first_onset, block_duration, period_length, sf, max_timestamp) {
    # Generate a series of timestamp blocks: the first one at first_onset, then one more each period. 
    if (max_timestamp <= first_onset) {
      return(c())
    }
    block_onsets = seq(first_onset + 1, max_timestamp, by = period_length) # one block onset for each period
    block_timestamps = rep(block_onsets, each = block_duration)
    block_timestamps = block_timestamps + rep(0:(block_duration - 1), times = length(block_onsets))
    
    # Make sure the entire duration of the last block fits within the time period for which we have data.
    # Trim it if necessary.
    block_timestamps = block_timestamps[which(block_timestamps <= max_timestamp)]
    
    return(block_timestamps)
  }
  
  generate_daily_timestamp_blocks = function(first_onset, block_duration, sf, max_timestamp) {
    # Generate a series of timestamp blocks: the first one at first_onset, then one more every 24 hours. 
    day_length = 24*3600*sf
    block_timestamps = generate_repeated_timestamp_blocks(first_onset, block_duration, day_length, sf, max_timestamp)
    return(block_timestamps)
  }
  
  #==================================
  
  if (length(storagelocation) == 0) storagelocation = getwd()
  if (Nmin < 720) Nmin = 720 # only make this file for tests with at least 12 hours of data
  if (length(start_time) == 0 || start_time == "") {
    if (starts_at_midnight) {
      start_time = "00:00:00"
    } else {
      start_time = "08:55:30"
    }
  } else {
    if (is.na(as.POSIXct(start_time, format = "%H:%M:%S"))) {
      stop(paste0("start_time \"", start_time, "\" not in the right format, should be \"%H:%M:%S\", ex. \"08:55:30\""))
    }
  }
  header = c(paste0("------------ Data File Created By ActiGraph GT3X+ ActiLife v6.13.3 Firmware v1.8.0 date format M/d/yyyy at ",sf," Hz  Filter Normal -----------"),
             "Serial Number: MOS2D12345678",
             paste0("Start Time ", start_time),
             "Start Date 6/23/2016",
             "Epoch Period (hh:mm:ss) 00:00:00",
             "Download Time 10:28:51",
             "Download Date 6/25/2016",
             "Current Memory Address: 0",
             "Current Battery Voltage: 3.97     Mode = 12",
             "--------------------------------------------------")
  
  variablenames = c("Accelerometer X","Accelerometer Y","Accelerometer Z")
  Nrows = Nmin*sf*60
  # default values, including some calibration error such that auto-calibration can do its work
  accx = 1.015
  accy = 0.005
  accz = -0.005
  testdata = matrix(0,Nrows,3)
  # add enough noise to not trigger the non-wear detection
  set.seed(300)
  testdata[,1] = accx + round(rnorm(n = Nrows, mean = 0.03, sd = 0.02), digits = 4) #gravity
  set.seed(400)
  testdata[,2] = accy + round(rnorm(n = Nrows, mean = 0.03, sd = 0.1), digits = 4) #no gravity
  Nsamplesinday = 24*3600*sf
  set.seed(500)
  testdata[,3] = accz + round(rnorm(n = Nrows, mean = 0.03, sd = 0.02), digits = 4) + #no gravity
    sin(((1:Nrows)/((Nrows)/(Nsamplesinday/0.1)))) * 0.05 +
    sin(((1:Nrows)/((Nrows)/(Nsamplesinday/2)))) * 0.1
  
  #======================================
  # generate 2.5-hour sleep blocks: at 12 hours from the start of the recording, and then every 24 hours
  sleep1 = generate_daily_timestamp_blocks(first_onset = 12 * 3600 * sf,
                                           block_duration = 2.5 * 3600 * sf, sf = sf,
                                           max_timestamp = Nrows)
  
  # generate 3-hour sleep blocks: at 15 hours from the start of the recording, and then every 24 hours
  sleep2 = generate_daily_timestamp_blocks(first_onset = 15 * 3600 * sf,
                                           block_duration = 3 * 3600 * sf,
                                           sf = sf, max_timestamp = Nrows)
  
  sleep_periods = sort(unique(c(sleep1,sleep2)))
  
  set.seed(300)
  testdata[sleep_periods, 1] = accx + round(rnorm(n = length(sleep_periods),mean = 0, sd = 0.015), digits = 4)
  set.seed(400)
  testdata[sleep_periods, 2] = accy + round(rnorm(n = length(sleep_periods),mean = 0, sd = 0.015), digits = 4)
  set.seed(500)
  testdata[sleep_periods, 3] = accz + round(rnorm(n = length(sleep_periods),mean = 0, sd = 0.015), digits = 4)
  
  # add some rotations to the acceleormeter, to avoid non-wear detection
  for (j in 1:(Nrows/(sf*3600))) { # loop to add angle changes to every night
    t0v = seq(12 + ((j - 1) * 24), 22 + ((j - 1) * 24), by = 0.25) # start times of angle positions
    phiv = 1:length(t0v)
    phiv = scale(x = phiv,center = 0,scale = diff(range(phiv))/(2*pi))  # rescale such that angle goes 180 degrees
    gammav = scale(x = phiv,center = 0,scale = diff(range(phiv))/(2*pi))  # rescale such that angle goes 180 degrees
    for (i in 1:length(t0v)) {
      B = gensequence(t0 = t0v[i], t1 = t0v[i] + 0.245) #long activity starts at the 3rd hour of every 24 hours
      Rm_phi = genrotationmatrix(phiv[i])
      Rm_gamma = genrotationmatrix(gammav[i])
      Ry = Rm_phi$Ry; Rx = Rm_gamma$Rx
      B = B[which(B < nrow(testdata))]
      if (length(B) > 0) {
        testdata[B,1] = accx # set constant such that it is detected as a period without movement
        testdata[B,2] = accy
        testdata[B,3] = accz
        testdata[B,] = t(Ry %*% t(testdata[B,]))
        testdata[B,] = t(Rx %*% t(testdata[B,]))
      }
    }
  }
  #============================================================
  # insert 2-minute blocks of activity at the beginning of every 15 minutes
  act2_15 = generate_repeated_timestamp_blocks(first_onset = 0, block_duration = 2 * 60 * sf,
                                               period_length = 15 * 60 * sf,
                                               sf = sf,
                                               max_timestamp = Nrows)
  
  ACC = rep(0,Nrows) #empty ACC pattern
  set.seed(300)
  ACC[act2_15] = 0.08 + round(rnorm(n = length(act2_15),
                                    mean = 0, sd = 0.01) , digits = 4) # add 0.08g + 10 mg noise
  testdata[,1] = testdata[,1] + ACC
  #======================================
  # generate one 2-hour block of non-wear starting at the 5th minute of every 24 hours
  nw = generate_daily_timestamp_blocks(first_onset = 5 * 60 * sf,
                                       block_duration = 2 * 3600 * sf,
                                       sf = sf, max_timestamp = Nrows)
  
  testdata[nw,1] = accx # constant values to trigger non-wear detection
  testdata[nw,2] = accy
  testdata[nw,3] = accz
  #======================================
  # insert 20 and 4 minute blocks of activity
  
  # generate one 20-minute block of activity starting at the 3rd hour of every 24 hours
  act1 = generate_daily_timestamp_blocks(first_onset = 3 * 3600 * sf,
                                         block_duration = 20 * 60 * sf,
                                         sf = sf, max_timestamp = Nrows)
  
  # generate one 4-minute block of activity starting at the 3.5th hour of every 24 hours
  act2 = generate_daily_timestamp_blocks(first_onset = 3.5 * 3600 * sf,
                                         block_duration = 4 * 60 * sf,
                                         sf = sf, max_timestamp = Nrows)
  
  act_periods = sort(unique(c(act1,act2)))
  
  ACC = rep(0,Nrows) #empty ACC pattern
  set.seed(300)
  ACC[act_periods] = 0.5 + round(rnorm(n = length(act_periods),
                                       mean = 0, sd = 0.04), digits = 4) # add 0.5g + 20 mg noise
  testdata[,1] = testdata[,1] + ACC
  #======================================
  # insert variations in angle to facilitate autocalibration
  t0v = seq(4.5, 9, by = 1/120) # start times of angle positions, and 30 seconds each
  phiv = 1:length(t0v)
  phiv = scale(x = phiv, center = 0, scale = diff(range(phiv)) / (20 * pi))  # rescale such that angle go 4 times 360 degrees
  gammav = scale(x = phiv, center = 0, scale = diff(range(phiv)) / (2 * pi))  # rescale such that angle goes 180 degrees
  for (i in 1:length(t0v)) {
    B = gensequence(t0 = t0v[i], t1 = t0v[i] + 0.24) #long activity starts at the 3rd hour of every 24 hours
    Rm_phi = genrotationmatrix(phiv[i])
    Rm_gamma = genrotationmatrix(gammav[i])
    Ry = Rm_phi$Ry; Rx = Rm_gamma$Rx
    testdata[B,1] = accx # set constant such that it is detected as a period without movement
    testdata[B,2] = accy
    testdata[B,3] = accz
    testdata[B,] = t(Ry %*% t(testdata[B,]))
    testdata[B,] = t(Rx %*% t(testdata[B,]))
  }
  testfile = matrix(" ", Nrows + 11, 3)
  testfile[1:10,1] = header
  testfile[11,1:3] = variablenames
  testfile[12:(Nrows + 11), 1:3] = testdata
  data.table::fwrite(x = testfile, file = paste0(storagelocation,"/123A_testaccfile.csv"),
              row.names = FALSE, col.names = FALSE)
}

library(GGIR)
context("HASIB with count data and HorAngle algorithm")
test_that("Checks that HASIB with count data is calculated and HorAngle algorithm applied", {
  skip_on_cran()
  
  # test data -----
  Ndays = 2
  create_test_acc_csv(Nmin = Ndays*1440, sf = 10)
  fn = "123A_testaccfile.csv"
  desiredtz = "Europe/London"
  dn = "output_test"
  if (file.exists(dn)) unlink(dn, recursive = TRUE)

  #----------------------------------------------------------------------
  # Parts 3:4 with HorAngle and SIB with Sadeh1994
  # generate file with sf = 30 to calculate zcy for using Sadeh1994
  GGIR(mode = c(1:4), datadir = fn, outputdir = getwd(),
       studyname = "test", f0 = 1, f1 = 1,
       do.report = c(), do.visual = FALSE,
       overwrite = FALSE, visualreport = FALSE, 
       verbose = FALSE, storefolderstructure = TRUE,
       do.zcy = TRUE,
       do.anglex = TRUE,
       do.angley = TRUE,
       do.anglez = TRUE,
       HASIB.algo = "Sadeh1994",
       Sadeh_axis = "Y",
       HASPT.algo = "HorAngle",
       longitudinal_axis = NULL,
       sleepwindowType = "TimeInBed",
       sensor.location = "hip") # so that it also covers the identification of the longitudinal axis
  dirname = "output_test/meta/ms4.out/"
  rn = dir(dirname,full.names = TRUE)
  load(rn[1])
  expect_equal(unique(nightsummary$sleepparam), "Sadeh1994_ZC")
  expect_equal(unique(nightsummary$guider), "HorAngle")
  
})

library(GGIR)
context("Very short recordings")
test_that("Test number of days in very short recordings", {
  skip_on_cran()
  create_test_acc_csv(Nmin = 720, start_time = "09:00:00")
  fn = "123A_testaccfile.csv"
  desiredtz = "Europe/London"
  dn = "output_test"
  if (file.exists(dn))  unlink(dn, recursive = TRUE)
  minimumFileSizeMB = 0
  #--------------------------------------------
  # run GGIR
  GGIR(datadir = fn, outputdir = getwd(), studyname = "test", 
       mode = 1:2, do.report = 2, idloc = 2,
       verbose = FALSE, desiredtz = desiredtz,
       do.cal = FALSE, do.imp = FALSE, windowsizes = c(5, 60, 60),
       ilevels = seq(0, 800, by = 100), includedaycrit = 0)
  #--------------------------------------------
  # part 2 data contains 2 complete days
  rn = "output_test/results/part2_daysummary.csv"
  out2 = read.csv(rn)
  expect_equal(nrow(out2), 1)
  expect_equal(round(out2$N.hours), 12)
  
  if (dir.exists(dn))  unlink(dn, recursive = TRUE)
  if (file.exists(fn)) unlink(fn)
})

library(GGIR)
context("g.part1")
test_that("Part 1 can run with all metrics", {
  skip_on_cran()
  #=======================
  # Part 1 with all metrics
  Ndays = 2
  # Using sf = 30 Hertz  because metric zc filters at 4 Hertz and brondcounts requires 30 Hertz
  create_test_acc_csv(Nmin = Ndays * 1440, sf = 30)
  fn = "123A_testaccfile.csv"
  metadatadir = paste0(getwd(), "/output_test")
  desiredtz = "Europe/London"
  dn = "output_test"
  
  # external function,  which adds a index per second
  exampleExtFunction = function(data, parameters) {
    output = data.frame(secondInDay = seq(from = 0, to = floor(nrow(data) / 30), by = 1))
    output$copy1 = output$copy2 = output$secondInDay
    return(output)
  }
  myfun =  list(FUN = exampleExtFunction,
                parameters = 1,
                expected_sample_rate =  30, # resample data to 30 Hertz before applying function
                expected_unit = "mg",
                minlength = 1,
                outputres = 1,
                colnames = c("secondInDay", "copy1", "copy2"),
                outputtype = "numeric", #"numeric" (averaging is possible), "category" (majority vote)
                aggfunction = mean,
                timestamp = F,
                reporttype = rep("scalar", 3)) # for unit test only
  
  if (file.exists(dn))  unlink(dn, recursive = TRUE)
  g.part1(datadir = fn, metadatadir = metadatadir, f0 = 1, f1 = 1, overwrite = TRUE, desiredtz = desiredtz,
          do.cal = FALSE, do.anglex = TRUE,
          # We are not doing all the metrics, because Travis-CI cannot allocate enough memory
          do.enmo = TRUE,do.lfenmo = TRUE,
          do.bfen = TRUE, do.hfenplus = TRUE,
          do.mad = TRUE, do.zcx = TRUE,
          windowsizes = c(15, 3600, 3600), do.parallel = FALSE,
          minimumFileSizeMB = 0, verbose = FALSE, myfun = myfun)
  
  rn = dir("output_test/meta/basic/",full.names = TRUE)
  load(rn[1])
  expect_equal(ncol(M$metashort), 12)
  expect_true(nrow(M$metashort) == 11280)
  expect_equal(mean(M$metashort$BFEN),  0.0458, tolerance = 4)
  expect_equal(mean(M$metashort$LFENMO),  0.0447, tolerance = 4)
  expect_equal(mean(M$metashort$HFENplus),  0.0914, tolerance = 4)
  expect_equal(mean(M$metashort$MAD),  0.0073, tolerance = 4)
  expect_equal(mean(M$metashort$anglex),  57.4683, tolerance = 4)
  expect_equal(mean(M$metashort$anglez),  0.3522, tolerance = 4)
  expect_equal(mean(M$metashort$ZCX),  14.94, tolerance = 2)
  expect_equal(mean(M$metashort$secondInDay),  42318.65, tolerance = 2)
  expect_equal(M$metashort$copy1, M$metashort$copy2)
  if (file.exists(fn)) file.remove(fn)
})

library(GGIR)
context("qwindow")
test_that("qwindow implementation in part 2", {
  skip_on_cran()
  
  Ndays = 2
  create_test_acc_csv(Nmin = Ndays*1440)
  fn = "123A_testaccfile.csv"
  actlog = data.frame(id = c("123A", "124A"),
                      date = c("23/06/2016", "23/06/2016"),
                      work = c("7:45:30", "7:45:00"),
                      travelhome = c("17:00:00", "17:00:00"),
                      home = c("17:30:00", "18:00:00"),
                      date = c("24/06/2016", "24/06/2016"),
                      work = c("7:45:15", "7:45:00"),
                      travelhome = c("17:00:00", "17:00:00"),
                      home = c("17:30:00", "18:00:00"),
                      sport = c("19:00:00", "19:00:00"),
                      home = c("20:00:00", "20:00:00"))
  actlog_fn = "testactlog.csv"
  write.csv(x = actlog, file = actlog_fn, row.names = FALSE)
  
  #--------------------------------------------
  # tests
  
  # part 2
  GGIR(mode = 1:2, datadir = fn, outputdir = getwd(), studyname = "test",
       idloc = 2,
       qwindow = actlog_fn, qwindow_dateformat = "%d/%m/%Y",
       do.report = 2, visualreport = FALSE, verbose = FALSE)
  
  # long format generated
  long_fn = paste(getwd(), "output_test", "results",
                  "part2_daysummary_longformat.csv", sep = .Platform$file.sep)
  expect_true(file.exists(long_fn))
  out = read.csv(long_fn)
  expect_true("qwindow_name" %in% colnames(out))
  expect_true("daystart-dayend" %in% out$qwindow_name)
  expect_true("sport-home" %in% out$qwindow_name)
  expect_true("work-travelhome" %in% out$qwindow_name)

  # remove generated files
  if (file.exists(fn)) file.remove(fn)
  if (file.exists(actlog_fn)) file.remove(actlog_fn)
  if (dir.exists(file.path(getwd(), "output_test"))) unlink(file.path(getwd(), "output_test"), recursive = TRUE)
})

library(GGIR)
context("create_test_acc_csv")
test_that("create_test_acc_csv produces a file", {
  create_test_acc_csv()
  fn = "123A_testaccfile.csv"
  expect_true(file.exists(fn))
  if (file.exists(fn)) file.remove(fn)

  # generate a test file with a non-wear block straddling the end of the recording 
  create_test_acc_csv(Nmin = 3000)
  expect_true(file.exists(fn))
  if (file.exists(fn)) file.remove(fn)

  # generate a test file with a sleep block straddling the end of the recording 
  create_test_acc_csv(Nmin = 2309)
  expect_true(file.exists(fn))
  if (file.exists(fn)) file.remove(fn)

  # generate a test file with an activity block straddling the end of the recording 
  create_test_acc_csv(Nmin = 2311)
  expect_true(file.exists(fn))
  if (file.exists(fn)) file.remove(fn)
})

library(GGIR)
context("g.readaccfile")
test_that("g.readaccfile and g.inspectfile can read movisens, gt3x, cwa, Axivity csv, actigraph csv, ad-hoc csv, and Parmay Matrix files correctly", {
  skip_on_cran()
  
  desiredtz = "Pacific/Auckland"
  configtz = "Europe/Berlin"
  params = extract_params(input = list(frequency_tol = 0.1, interpolationType = 1,
                                       desiredtz = desiredtz, configtz = configtz))
  params_rawdata = params$params_rawdata
  params_general = params$params_general
  
  filequality = list(filecorrupt = FALSE, filetooshort = FALSE)
  dayborder = 0
  
  # For Axivity csv files, we'll be able to read files with both unix and formatted (Y-M-D h:m:s) timestamps
  Ax3CsvFile  = system.file("testfiles/ax3_testfile_unix_timestamps.csv", package = "GGIR")[1]
  Ax6CsvFile  = system.file("testfiles/ax6_testfile_formatted_timestamps.csv", package = "GGIR")[1]
  
  Ax3CwaFile  = system.file("testfiles/ax3_testfile.cwa", package = "GGIRread")[1]
  Ax6CwaFile  = system.file("testfiles/ax6_testfile.cwa", package = "GGIRread")[1]
  
  GAfile  = system.file("testfiles/GENEActiv_testfile.bin", package = "GGIRread")[1]
  gt3xfile  = system.file("testfiles/actigraph_testfile.gt3x", package = "GGIR")[1]
  mtxfile  = system.file("testfiles/mtx_25Hz_acc_HR.BIN", package = "GGIRread")[1]
  
  cat("\nActigraph .csv")
  
  create_test_acc_csv()
  filename = "123A_testaccfile.csv"
  on.exit({if (file.exists(filename)) file.remove(filename)}, add = TRUE)
  
  Icsv = g.inspectfile(filename, desiredtz = desiredtz)
  expect_equal(Icsv$monc, MONITOR$ACTIGRAPH)
  expect_equal(Icsv$dformc, FORMAT$CSV)
  
  csv_read = g.readaccfile(filename, blocksize = 10, blocknumber = 1, filequality = filequality,
                           dayborder = dayborder, ws = 3,
                           PreviousEndPage = 1, inspectfileobject = Icsv,
                           params_rawdata = params_rawdata, params_general = params_general)
  expect_equal(nrow(csv_read$P$data), 3000)
  expect_false(csv_read$filequality$filecorrupt)
  expect_false(csv_read$filequality$filetooshort)
  expect_equal(sum(csv_read$P$data), 3151.11, tolerance = .01, scale = 1)
  
  cat("\nActigraph .gt3x")
  # actigraph .gt3x
  Igt3x = g.inspectfile(gt3xfile, desiredtz = desiredtz)
  expect_equal(Igt3x$monc, MONITOR$ACTIGRAPH)
  expect_equal(Igt3x$dformc, FORMAT$GT3X)
  expect_equal(Igt3x$sf, 30)
  EHV = g.extractheadervars(Igt3x)
  expect_equal(EHV$deviceSerialNumber, "MOS2E39180594_firmware_1.9.2")
  
  gt3x_read = g.readaccfile(gt3xfile, blocksize = 3000, blocknumber = 1, filequality = filequality,
                            dayborder = dayborder, ws = 3,
                            PreviousEndPage = 1, inspectfileobject = Igt3x,
                            params_rawdata = params_rawdata, params_general = params_general)
  expect_equal(nrow(gt3x_read$P$data), 17640)
  expect_false(gt3x_read$filequality$filecorrupt)
  expect_false(gt3x_read$filequality$filetooshort)
  expect_equal(sum(gt3x_read$P$data[c("x","y","z")]), 2732.35, tolerance = .01, scale = 1)
  
  Mgt3x = g.getmeta(datafile = gt3xfile, desiredtz = desiredtz, windowsize = c(1,300,300),
                    inspectfileobject = Igt3x)
  expect_true(Mgt3x$filetooshort)
  expect_false(Mgt3x$filecorrupt)
  
  cat("\nAxivity .cwa")
  
  Icwa = g.inspectfile(Ax3CwaFile, params_rawdata = params_rawdata, params_general = params_general)
  expect_equal(Icwa$monc, MONITOR$AXIVITY)
  expect_equal(Icwa$dformc, FORMAT$CWA)
  expect_equal(Icwa$sf, 100)
  EHV = g.extractheadervars(Icwa)
  expect_equal(EHV$deviceSerialNumber,"39434")
  
  cwa_read = g.readaccfile(Ax3CwaFile, blocksize = 10, blocknumber = 1, filequality = filequality,
                           dayborder = dayborder, ws = 2, 
                           PreviousEndPage = 1, inspectfileobject = Icwa,
                           params_rawdata = params_rawdata, params_general = params_general)
  expect_equal(cwa_read$P$header$blocks, 145)
  expect_equal(sum(cwa_read$P$data[c("x","y","z")]), 280.53, tolerance = .01, scale = 1)
  
  Mcwa = g.getmeta(Ax3CwaFile, desiredtz = desiredtz, windowsize = c(1,300,300),
                   inspectfileobject = Icwa)
  expect_true(Mcwa$filetooshort)
  expect_false(Mcwa$filecorrupt)
  
  ax3_start_timestamp = cwa_read$P$data$time[1]
  
  Icwa = g.inspectfile(Ax6CwaFile, params_rawdata = params_rawdata, params_general = params_general)
  cwa_read = g.readaccfile(Ax6CwaFile, blocksize = 10, blocknumber = 1, filequality = filequality,
                           dayborder = dayborder, ws = 2, 
                           PreviousEndPage = 1, inspectfileobject = Icwa,
                           params_rawdata = params_rawdata, params_general = params_general)
  ax6_start_timestamp = cwa_read$P$data$time[1]
  
  cat("\nAxivity .csv")
  
  for (csvData in list(list(Ax3CsvFile, 200, -11.80, ax3_start_timestamp),
                       list(Ax6CsvFile, 200, 14.84, ax6_start_timestamp))) {
    IAxivityCsv = g.inspectfile(csvData[[1]], params_rawdata = params_rawdata, params_general = params_general)
    expect_equal(IAxivityCsv$monc, MONITOR$AXIVITY)
    expect_equal(IAxivityCsv$dformc, FORMAT$CSV)
    
    csv_read = g.readaccfile(csvData[[1]], blocksize = 1, blocknumber = 1, filequality = filequality,
                             dayborder = dayborder, ws = 1, 
                             PreviousEndPage = 1, inspectfileobject = IAxivityCsv,
                             params_rawdata = params_rawdata, params_general = params_general)
    
    # For both ax3 and ax6 files, we expect 4 columns: timestamp and XYZ.
    # All gyro data in ax6 files gets ignored.
    expect_equal(ncol(csv_read$P$data), 4)
    
    expect_equal(nrow(csv_read$P$data), csvData[[2]])
    expect_false(csv_read$filequality$filecorrupt)
    expect_false(csv_read$filequality$filetooshort)
    expect_equal(sum(csv_read$P$data[c("x","y","z")]), csvData[[3]], tolerance = .01, scale = 1)
    
    # check that the timestamps for the Axivity csv look the same as they did for
    # the original cwa version of the same file (this verifies that timestamp conversion
    # worked the same for both formats)
    expect_equal(csv_read$P$data$time[1], csvData[[4]], tolerance = .01, scale = 1)
    
    MAxCsv = g.getmeta(datafile = Ax3CsvFile, desiredtz = desiredtz, windowsize = c(1,300,300),
                       inspectfileobject = IAxivityCsv)
    expect_true(MAxCsv$filetooshort)
    expect_false(MAxCsv$filecorrupt)   
  }
  
  cat("\nGENEActiv .bin")
  # GENEActiv .bin
  IGA = g.inspectfile(GAfile, desiredtz = desiredtz)
  expect_equal(IGA$monc, MONITOR$GENEACTIV)
  expect_equal(IGA$dformc, FORMAT$BIN)
  expect_equal(IGA$sf, 86)
  
  EHV = g.extractheadervars(IGA)
  expect_equal(EHV$deviceSerialNumber,"012967")
  
  GA_num_blocks = 2
  GA_read = g.readaccfile(GAfile, blocksize = GA_num_blocks, blocknumber = 1, filequality = filequality,
                          dayborder = dayborder, ws = 3,
                          desiredtz = desiredtz, PreviousEndPage = 1, inspectfileobject = IGA)
  
  # As of R 4.0, an extra header row is extracted, which affects the positioning of the values.
  # expect_equal(as.numeric(as.character(wav_read$P$header$hvalues[7])),17) 
  
  expect_equal(round(sum(GA_read$P$data[, 2:4]), digits = 2), -271.35)
  expect_equal(GA_read$endpage, GA_num_blocks)
  
  # print(GA_read$P$header)
  # expect_equal(as.character(unlist(GA_read$P$header[3, 1])), "216 Hours")
  
  MGA = g.getmeta(GAfile, desiredtz = desiredtz, windowsize = c(1,300,300), verbose = FALSE,
                  inspectfileobject = IGA)
  expect_true(MGA$filetooshort)
  
  cat("\n Movisens")
  
  output_dir = "output_unisensExample"
  on.exit({if (file.exists(output_dir)) unlink(output_dir, recursive = TRUE)}, add = TRUE)
  if (file.exists(output_dir)) unlink(output_dir, recursive = TRUE)
  
  zip_file = "0.3.4.zip"
  on.exit({if (file.exists(zip_file)) unlink(zip_file)}, add = TRUE)
  if (!file.exists(zip_file)) {
    # link to a tagged release of Unisens/unisensR github repo
    movisens_url = "https://github.com/Unisens/unisensR/archive/refs/tags/0.3.4.zip"
    download.file(url = movisens_url, destfile = zip_file, quiet = TRUE)
  }
  
  movisens_dir = "unisensR-0.3.4"
  on.exit({if (file.exists(movisens_dir)) unlink(movisens_dir, recursive = TRUE)}, add = TRUE)
  if (file.exists(movisens_dir)) {
    unlink(movisens_dir, recursive = TRUE)
  }
  unzip(zipfile = zip_file, exdir = ".")
  movisensFile = file.path(getwd(), "unisensR-0.3.4/tests/unisensExample/acc.bin")
  
  Mcsv = g.inspectfile(movisensFile, desiredtz = desiredtz)
  expect_equal(Mcsv$monc, MONITOR$MOVISENS)
  expect_equal(Mcsv$dformc, FORMAT$BIN)
  expect_equal(Mcsv$sf, 64)
  
  movisens_blocksize = 3000
  movisens_read = g.readaccfile(movisensFile, blocksize = movisens_blocksize, blocknumber = 1, filequality = filequality,
                                dayborder = dayborder, ws = 3,
                                PreviousEndPage = 1, inspectfileobject = Mcsv,
                                params_rawdata = params_rawdata, params_general = params_general)
  expect_equal(nrow(movisens_read$P$data), movisens_blocksize)
  expect_false(movisens_read$filequality$filecorrupt)
  expect_false(movisens_read$filequality$filetooshort)
  expect_equal(sum(movisens_read$P$data[c("x","y","z")]), 4383.67, tolerance = .01, scale = 1)
  expect_equal(movisens_read$endpage, movisens_blocksize)
  
  # read the next block (set PreviousEndPage to movisens_read$endpage)
  movisens_read2 = g.readaccfile(movisensFile, blocksize = movisens_blocksize, blocknumber = 2, filequality = filequality,
                                 dayborder = dayborder, ws = 3,
                                 PreviousEndPage = movisens_read$endpage, inspectfileobject = Mcsv,
                                 params_rawdata = params_rawdata, params_general = params_general)
  expect_equal(nrow(movisens_read2$P$data), movisens_blocksize)
  expect_equal(movisens_read2$endpage, movisens_blocksize * 2)
  
  # if the 1st sample of 2nd block is identical to the last sample of the 1st block,
  # this means that we calculated the startpage of the 2nd block incorrectly.
  expect_false(any(movisens_read2$P$data[1,] == movisens_read$P$data[nrow(movisens_read$P$data),]))
  cat("\n ad-hoc csv file")
  # ad-hoc csv file
  
  # Create test file: No header, with temperature, with time.
  # The first row of the file will contain column names.
  # Have this file contain 6000 samples. We'll read it in 2 sets of 3000 lines,
  # and if there are any lines unnecessarily skipped, then the second attempt to 
  # read a block of 3000 will return fewer than 3000 lines.
  N = 6000
  sf = 30
  x = Sys.time() + ((0:(N - 1)) / sf)
  timestamps = as.POSIXlt(x, origin = "1970-1-1", tz = configtz)
  mydata = data.frame(Xcol = rnorm(N), timecol = timestamps, Ycol = rnorm(N), Zcol = rnorm(N),
                      tempcol = rnorm(N) + 20)
  testfile = "testcsv.csv"
  on.exit({if (file.exists(testfile)) file.remove(testfile)}, add = TRUE)
  
  write.csv(mydata, file = testfile, row.names = FALSE)
  
  # check that for files with no header, g.inspectfile() errors out if sampling rate 
  # is not specified as rmc.sf, or if rmc.sf == 0
  expect_error(g.inspectfile(testfile, 
                             rmc.dec = ".", rmc.unit.time = "POSIX", 
                             rmc.firstrow.acc = 1, rmc.firstrow.header = c(), 
                             rmc.col.acc = c(1, 3, 4), rmc.col.temp = 5, rmc.col.time = 2, 
                             rmc.unit.acc = "g", rmc.unit.temp = "C", rmc.origin = "1970-01-01"),
               regexp = "File header doesn't specify sample rate. Please provide rmc.sf value to process")
  expect_error(g.inspectfile(testfile, 
                             rmc.dec = ".", rmc.sf = 0, rmc.unit.time = "POSIX", 
                             rmc.firstrow.acc = 1, rmc.firstrow.header = c(), 
                             rmc.col.acc = c(1, 3, 4), rmc.col.temp = 5, rmc.col.time = 2, 
                             rmc.unit.acc = "g", rmc.unit.temp = "C", rmc.origin = "1970-01-01"),
               regexp = "File header doesn't specify sample rate. Please provide a non-zero rmc.sf value to process")
  
  AHcsv = g.inspectfile(testfile, 
                        rmc.dec = ".", rmc.sf = 30, rmc.unit.time = "POSIX", 
                        rmc.firstrow.acc = 1, rmc.firstrow.header = c(), 
                        rmc.col.acc = c(1, 3, 4), rmc.col.temp = 5, rmc.col.time = 2, 
                        rmc.unit.acc = "g", rmc.unit.temp = "C", rmc.origin = "1970-01-01")
  
  expect_equal(AHcsv$monc, MONITOR$AD_HOC)
  expect_equal(AHcsv$dformc, FORMAT$AD_HOC_CSV)
  expect_equal(AHcsv$sf, 30)
  
  # Read the file starting with row 1 (rmc.firstrow.acc = 1); this row contains column names.
  # Verify that full 3000 rows are still read.
  csv_read = g.readaccfile(testfile, blocksize = 10, blocknumber = 1, filequality = filequality, # blocksize is # of pages of 300 samples
                           dayborder = dayborder, ws = 3, desiredtz = desiredtz, configtz = configtz,
                           PreviousEndPage = c(), inspectfileobject = AHcsv,
                           rmc.dec = ".", rmc.sf = 30, rmc.unit.time = "POSIX", 
                           rmc.firstrow.acc = 1, rmc.firstrow.header = c(), 
                           rmc.col.acc = c(1, 3, 4), rmc.col.temp = 5, rmc.col.time = 2, 
                           rmc.unit.acc = "g", rmc.unit.temp = "C", rmc.origin = "1970-01-01")
  
  expect_equal(nrow(csv_read$P$data), 3000)
  expect_false(csv_read$filequality$filecorrupt)
  expect_false(csv_read$filequality$filetooshort)
  expect_equal(sum(csv_read$P$data[c("x","y","z")]), -38.90, tolerance = .01, scale = 1)
  # endpage should be (# of rows read + 1) because for the next block we'll need to skip
  # not just the rows read but also the row containing column names.
  expect_equal(csv_read$endpage, 3001)
  
  # since the 1st row of the file contains column names, pointing rmc.firstrow.acc 2
  # should lead to the same eaxt 3000 lines being read (the lines after the column names).
  csv_read2 = g.readaccfile(testfile, blocksize = 10, blocknumber = 1, filequality = filequality,
                            dayborder = dayborder, ws = 3, desiredtz = desiredtz, configtz = configtz,
                            PreviousEndPage = c(), inspectfileobject = AHcsv,
                            rmc.dec = ".", rmc.sf = 30, rmc.unit.time = "POSIX", 
                            rmc.firstrow.acc = 2, rmc.firstrow.header = c(), 
                            rmc.col.acc = c(1, 3, 4), rmc.col.temp = 5, rmc.col.time = 2, 
                            rmc.unit.acc = "g", rmc.unit.temp = "C", rmc.origin = "1970-01-01")
  
  expect_equal(nrow(csv_read2$P$data), 3000)
  expect_false(csv_read2$filequality$filecorrupt)
  expect_false(csv_read2$filequality$filetooshort)
  expect_equal(sum(csv_read2$P$data[c("x","y","z")]), sum(csv_read$P$data[c("x","y","z")]), tolerance = .01, scale = 1)
  expect_equal(csv_read2$endpage, 3000)
  
  # reading the next 3000 lines should also give the same result for rmc.firstrow.acc == 1 or 2.
  csv_read3 = g.readaccfile(testfile, blocksize = 10, blocknumber = 2, filequality = filequality,
                            dayborder = dayborder, ws = 3, desiredtz = desiredtz, configtz = configtz,
                            PreviousEndPage = csv_read$endpage, inspectfileobject = AHcsv,
                            rmc.dec = ".", rmc.sf = 30, rmc.unit.time = "POSIX", 
                            rmc.firstrow.acc = 1, rmc.firstrow.header = c(), 
                            rmc.col.acc = c(1, 3, 4), rmc.col.temp = 5, rmc.col.time = 2, 
                            rmc.unit.acc = "g", rmc.unit.temp = "C", rmc.origin = "1970-01-01")
  
  expect_equal(nrow(csv_read3$P$data), 3000)
  
  csv_read4 = g.readaccfile(testfile, blocksize = 10, blocknumber = 2, filequality = filequality,
                            dayborder = dayborder, ws = 3, desiredtz = desiredtz, configtz = configtz,
                            PreviousEndPage = csv_read2$endpage, inspectfileobject = AHcsv,
                            rmc.dec = ".", rmc.sf = 30, rmc.unit.time = "POSIX", 
                            rmc.firstrow.acc = 2, rmc.firstrow.header = c(), 
                            rmc.col.acc = c(1, 3, 4), rmc.col.temp = 5, rmc.col.time = 2, 
                            rmc.unit.acc = "g", rmc.unit.temp = "C", rmc.origin = "1970-01-01")
  
  expect_equal(nrow(csv_read4$P$data), 3000)
  expect_equal(sum(csv_read3$P$data[c("x","y","z")]), sum(csv_read4$P$data[c("x","y","z")]), tolerance = .01, scale = 1)
  
  # Create test file: 2-column header, with time,
  # but sample rate not specified in the header
  
  N = 6000
  sf = 30
  x = Sys.time() + ((0:(N - 1)) / sf)
  timestamps = as.POSIXlt(x, origin = "1970-1-1", tz = configtz)
  mydata = data.frame(Xcol = rnorm(N), timecol = timestamps, Ycol = rnorm(N), Zcol = rnorm(N))
  S1 = as.matrix(mydata)
  
  hd_NR = 10
  hd = matrix("", hd_NR + 1, ncol(S1))
  hd[1, 1:2] = c("ID","12345")
  hd[2, 1:2] = c("serial_number","30")
  hd[3, 1:2] = c("bit","8")
  hd[4, 1:2] = c("dynamic_range","6")
  
  S1 = rbind(hd, S1)
  S1[hd_NR + 1,] = colnames(S1)
  colnames(S1) = NULL
  
  testfile_two_col = "testcsv2col.csv"
  on.exit({if (file.exists(testfile_two_col)) file.remove(testfile_two_col)}, add = TRUE)
  write.table(S1, file = testfile_two_col, col.names = FALSE, row.names = FALSE)
  
  # Create test file: 1-column header, with time,
  # but sample rate not specified in the header
  S1 = as.matrix(mydata)
  hd = matrix("", hd_NR + 1, ncol(S1))
  hd[1, 1:2] = c("ID: 12345", "")
  hd[2, 1:2] = c("serial_number: 4321", "")
  hd[3, 1:2] = c("bit: 8", "")
  hd[4, 1:2] = c("dynamic_range: 6", "")
  
  S1 = rbind(hd, S1)
  S1[hd_NR + 1,] = colnames(S1)
  colnames(S1) = NULL
  
  testfile_one_col = "testcsv1col.csv"
  on.exit({if (file.exists(testfile_one_col)) file.remove(testfile_one_col)}, add = TRUE)
  write.table(S1, file = testfile_one_col, col.names = FALSE, row.names = FALSE)
  
  for (csvData in list(list(testfile_one_col, ": "),
                       list(testfile_two_col, c()))) {
    # check that for a file whose header doesn't specify sampling rate,
    # g.inspectfile() errors out if sampling rate is not specified as rmc.sf, or if rmc.sf==0
    expect_error(g.inspectfile(csvData[[1]],
                               rmc.dec = ".", rmc.unit.time = "POSIX", 
                               rmc.firstrow.acc = 11, rmc.firstrow.header = 1,
                               rmc.col.acc = c(1, 3, 4), rmc.col.time = 2, 
                               rmc.unit.acc = "g", rmc.origin = "1970-01-01",
                               rmc.header.structure = csvData[[2]]),
                 regexp = "File header doesn't specify sample rate. Please provide rmc.sf value to process")
    expect_error(g.inspectfile(csvData[[1]],
                               rmc.dec = ".", rmc.sf = 0, rmc.unit.time = "POSIX", 
                               rmc.firstrow.acc = 11, rmc.firstrow.header = 1,
                               rmc.col.acc = c(1, 3, 4), rmc.col.time = 2, 
                               rmc.unit.acc = "g", rmc.origin = "1970-01-01",
                               rmc.header.structure = csvData[[2]]),
                 regexp = "File header doesn't specify sample rate. Please provide a non-zero rmc.sf value to process")
    
    # check that for a file whose header doesn't specify sampling rate,
    # g.inspectfile() returns sf == rmc.sf if the latter was specified
    I = g.inspectfile(csvData[[1]],
                      rmc.dec = ".", rmc.sf = 80, rmc.unit.time = "POSIX", 
                      rmc.firstrow.acc = 11, rmc.firstrow.header = 1,
                      rmc.col.acc = c(1, 3, 4), rmc.col.time = 2, 
                      rmc.unit.acc = "g", rmc.origin = "1970-01-01",
                      rmc.header.structure = csvData[[2]])
    expect_equal(I$sf, 80)
  }
  
  # Create test file: 2-column header, with temperature, with time,
  # and sample rate correctly specified in the header
  S1 = as.matrix(mydata)
  hd_NR = 10
  hd = matrix("", hd_NR + 1, ncol(S1))
  hd[1, 1:2] = c("ID","12345")
  hd[2, 1:2] = c("sample_freq","40")
  hd[3, 1:2] = c("serial_number","9876")
  hd[4, 1:2] = c("bit","8")
  hd[5, 1:2] = c("dynamic_range","6")
  S1 = rbind(hd, S1)
  S1[hd_NR + 1,] = colnames(S1)
  colnames(S1) = NULL
  
  testfile_two_col = "testcsv2col.csv"
  on.exit({if (file.exists(testfile_two_col)) file.remove(testfile_two_col)}, add = TRUE)
  write.table(S1, file = testfile_two_col, col.names = FALSE, row.names = FALSE)
  
  # Create test file: 1-column header, with time,
  # and sample rate not specified in the header
  S1 = as.matrix(mydata)
  hd = matrix("", hd_NR + 1, ncol(S1))
  hd[1, 1:2] = c("ID: 12345", "")
  hd[2, 1:2] = c("sample_freq: 40", "")
  hd[3, 1:2] = c("serial_number: 4321", "")
  hd[4, 1:2] = c("bit: 8", "")
  hd[5, 1:2] = c("dynamic_range: 6", "")
  S1 = rbind(hd, S1)
  S1[hd_NR + 1,] = colnames(S1)
  colnames(S1) = NULL
  
  testfile_one_col = "testcsv1col.csv"
  on.exit({if (file.exists(testfile_one_col)) file.remove(testfile_one_col)}, add = TRUE)
  write.table(S1, file = testfile_one_col, col.names = FALSE, row.names = FALSE)
  
  for (csvData in list(list(testfile_one_col, ": "),
                       list(testfile_two_col, c()))) {
    # check that g.inspectfile() returns sf value that was specified in the header, even if rmc.sf was also specified
    I = g.inspectfile(csvData[[1]],
                      rmc.dec = ".", rmc.sf = 80, rmc.headername.sf = "sample_freq", 
                      rmc.unit.time = "POSIX", 
                      rmc.firstrow.acc = 11, rmc.firstrow.header = 1, 
                      rmc.col.acc = c(1, 3, 4), rmc.col.time = 2, 
                      rmc.unit.acc = "g", rmc.origin = "1970-01-01",
                      rmc.headername.sn = "serial_number",
                      rmc.headername.recordingid = "ID",
                      rmc.bitrate = "bit", rmc.dynamic_range = "dynamic_range",
                      rmc.header.structure = csvData[[2]])
    
    expect_equal(I$sf, 40)
    
    # check that g.inspectfile() correctly reads the sf value from the header
    I = g.inspectfile(csvData[[1]],
                      rmc.dec = ".", rmc.headername.sf = "sample_freq", 
                      rmc.unit.time = "POSIX", 
                      rmc.firstrow.acc = 11, rmc.firstrow.header = 1, 
                      rmc.col.acc = c(1, 3, 4), rmc.col.time = 2, 
                      rmc.unit.acc = "g", rmc.origin = "1970-01-01", 
                      rmc.header.structure = csvData[[2]])
    expect_equal(I$sf, 40)
  }
  cat("\nParmay Matrix .BIN")
  
  Imtx = g.inspectfile(mtxfile, params_rawdata = params_rawdata, params_general = params_general)
  expect_equal(Imtx$monc, MONITOR$PARMAY_MTX)
  expect_equal(Imtx$dformc, FORMAT$BIN)
  expect_equal(Imtx$sf, 25)
  EHV = g.extractheadervars(Imtx)
  expect_equal(EHV$deviceSerialNumber,"not extracted") # not stored in file, but leave this here to facilitate implementation in case that in the future it is available
  
  mtx_read = g.readaccfile(mtxfile, blocksize = 10, blocknumber = 1, filequality = filequality,
                           dayborder = dayborder, ws = 2, 
                           PreviousEndPage = 1, inspectfileobject = Imtx,
                           params_rawdata = params_rawdata, params_general = params_general)
  expect_equal(nrow(mtx_read$P$QClog), 4) # 4 blocks of data in this file

  expect_equal(mean(mtx_read$P$data[, "x"]), 0.3484907, tolerance = .001, scale = 1)
  expect_equal(mean(mtx_read$P$data[, "y"]), 0.445011, tolerance = .001, scale = 1)
  expect_equal(mean(mtx_read$P$data[, "z"]), -0.5303114, tolerance = .001, scale = 1)

  Mmtx = g.getmeta(mtxfile, desiredtz = desiredtz, windowsize = c(1,300,300),
                   inspectfileobject = Imtx)
  expect_true(Mmtx$filetooshort)
  expect_false(Mmtx$filecorrupt)
  
  # test decimal separator recognition extraction
  decn =  g.dotorcomma(Ax3CwaFile,dformat = FORMAT$CWA, mon = MONITOR$AXIVITY, desiredtz = desiredtz)
  expect_equal(decn,".")
  decn =  g.dotorcomma(GAfile, dformat = FORMAT$BIN, mon = MONITOR$GENEACTIV, desiredtz = desiredtz)
  expect_equal(decn,".")
  
  #also test one small other function:
  datadir  = system.file("testfiles", package = "GGIR")[1]
  fnames = datadir2fnames(datadir = datadir, filelist = FALSE)
  expect_equal(length(fnames$fnames), 8)
  expect_equal(length(fnames$fnamesfull), 8)
  
  if (dir.exists("unisensR-0.3.4/")) unlink("unisensR-0.3.4/", recursive = TRUE)
  if (file.exists(testfile_one_col)) file.remove(testfile_one_col)
  if (file.exists(testfile_two_col)) file.remove(testfile_two_col)
  if (file.exists(testfile)) file.remove(testfile)
  if (file.exists(filename)) file.remove(filename)
  if (file.exists(zip_file)) file.remove(zip_file)
})

library(GGIR)
context("Lux handling in part 5")
test_that("lux_per_segment is correctly calculated", {
  skip_on_cran()
  
  # This test also covers: dayborder != 0 and part5_agg2_60seconds = TRUE
  
  # test data -----
  create_test_acc_csv(Nmin=1440*2)
  fn = "123A_testaccfile.csv"
  dn = "output_test"
  if (file.exists(dn)) unlink(dn, recursive = TRUE)
  
  # run part 1
  GGIR(mode = 1, datadir = fn, outputdir = getwd(), studyname = "test",
       do.report = c(), dayborder = 23, verbose = FALSE, nonwear_approach = "2013")
  
  # add lightmean and lightpeak to metalong
  meta_fn = paste(getwd(), "output_test", "meta", "basic", 
                  "meta_123A_testaccfile.csv.RData", sep = .Platform$file.sep)
  load(meta_fn)
  set.seed(400)
  M$metalong$lightmean = rnorm(n = nrow(M$metalong),mean=1000,sd=2000)
  M$metalong$lightmean[which(M$metalong$lightmean < 0)] = 0
  M$metalong$lightmean = M$metalong$lightmean / 1000
  M$metalong$lightpeak = M$metalong$lightmean + 0.2
  
  save(C, I, M, filefoldername, filename_dir, tail_expansion_log,
       file = meta_fn)
  
  # now run parts 2:5
  GGIR(mode = 2:5, datadir = fn, outputdir = getwd(), studyname = "test",
       overwrite = FALSE, do.report = c(2, 4, 5), verbose = FALSE,
       LUXthresholds = c(0, 500, 1000), 
       LUX_cal_constant = 1, 
       LUX_cal_exponent = 0.2, 
       LUX_day_segments = c(9, 15, 24), 
       dayborder = 23, part5_agg2_60seconds = TRUE,
       save_ms5rawlevels = TRUE, save_ms5raw_without_invalid = FALSE,
       save_ms5raw_format = "RData", nonwear_approach = "2013",
       timewindow = "MM")
  
  # Only segment 9 to 15hr is calculated because it is the only segment
  # containing awake data in the file generated by create_test_acc_csv
  
  # tests -----
  p5_fn = paste(getwd(), "output_test", "results", 
                "part5_daysummary_MM_L40M100V400_T5A5.csv",
                sep = .Platform$file.sep)
  df = read.csv(p5_fn)
  
  expect_equal(length(which(grepl("^LUX", colnames(df)))), 17) # max and mean day, mean spt, mean mvpa, and ranges ()
  expect_equal(df$LUX_max_day, 3.3)
  expect_equal(df$LUX_mean_day, 2)
  expect_equal(df$LUX_mean_day_mvpa, 2.7)
  expect_equal(df$LUX_min_0_500_day, df$dur_day_min) #in this artificial data, all LUX values in the day are expected to be below 500
  
  # dayborder and ws3 = 60
  p5raw_fn = paste(getwd(), "output_test", "meta", 
                   "ms5.outraw", "40_100_400",
                   "123A_testaccfile_T5A5.RData",
                   sep = .Platform$file.sep)
  expect_true(file.exists(p5raw_fn))
  mdat = c()
  load(p5raw_fn)
  expect_equal(diff(mdat[1:2, "timenum"]), 60) #epoch = 60
  pm11 = grep("23:00:00", as.character(mdat$timestamp))[1]
  expect_equal(diff(mdat[(pm11 - 1):pm11, "window"]), 1) #dayborder = 23 (change in window at 23:00)
  
  outfolder = paste(getwd(), "output_test", sep = .Platform$file.sep)
  if (file.exists(outfolder))  unlink(outfolder, recursive = TRUE)
  if (file.exists(dn)) unlink(dn, recursive = TRUE)
  if (file.exists(fn)) unlink(fn, recursive = TRUE)
})

library(GGIR)
context("Chainof5parts")
test_that("chainof5parts", {
  skip_on_cran()
  
  Ndays = 3
  create_test_acc_csv(Nmin = Ndays*1440, start_time = "09:00:00")
  fn = "123A_testaccfile.csv"
  desiredtz = "Europe/London"
  dn = "output_test"
  do.parallel = FALSE
  if (file.exists(dn))  unlink(dn, recursive = TRUE)
  minimumFileSizeMB = 0
  #--------------------------------------------
  # create study dates file
  studydates = data.frame(ID = "123A",
                          start = "24-06-2016", # intentional wrong format (hyphen) to force tests on check_log
                          end = "25/06/2016")
  
  studydates = expect_warning(check_log(studydates, dateformat = "%d/%m/%Y", colid = 1, datecols = 2:3,
                                        logPath = "study_dates_file.csv", logtype = "study dates log"),
                              regexp = "You may be mixing slash and hyphen")
  
  write.csv(studydates, "study_dates_file.csv", row.names = FALSE)
  
  
  # PART 1 ------------------------------------------------------------------
  GGIR(datadir = fn, outputdir = getwd(), studyname = "test", mode = 1, verbose = FALSE,
       overwrite = TRUE,
       idloc = 2, desiredtz = desiredtz)
  
  # PART 2 (tests) ----------------------------------------------------------
  
  # Strategy 1 ----
  GGIR(datadir = fn, outputdir = getwd(), studyname = "test", mode = 2, verbose = FALSE,
       # study dates file
       study_dates_file = "study_dates_file.csv",
       study_dates_dateformat = "%d/%m/%Y",
       # strategy
       data_masking_strategy = 1, hrs.del.start = 0,hrs.del.end = 0)
  # load data
  load(dir("output_test/meta/ms2.out/",full.names = TRUE)[1])
  # removed time before 24/06/2016 and after  25/06/2016
  expect_equal(rle(IMP$rout$r4)$lengths[1], 60)
  expect_equal(rle(IMP$rout$r4)$lengths[3], 36)
  # check the total time included is exactly 2 days (= 192 long epochs)
  first_epoch_in_protocol = rle(IMP$rout$r4)$lengths[1] + 1
  last_epoch_in_protocol = max(which(IMP$rout$r4 == 0))
  expect_equal(last_epoch_in_protocol - first_epoch_in_protocol + 1,  192)
  
  GGIR(datadir = fn, outputdir = getwd(), studyname = "test", mode = 2, verbose = FALSE,
       # study dates file
       study_dates_file = "study_dates_file.csv",
       study_dates_dateformat = "%d/%m/%Y",
       # strategy
       data_masking_strategy = 1, hrs.del.start = 4, hrs.del.end = 4)
  # load data
  load(dir("output_test/meta/basic/",full.names = TRUE)[1])
  load(dir("output_test/meta/ms2.out/",full.names = TRUE)[1])
  # first epoch in protocol 24/06/2016 04:00:00
  first_epoch_in_protocol = rle(IMP$rout$r4)$lengths[1] + 1
  expect_equal(M$metalong$timestamp[first_epoch_in_protocol], "2016-06-24T04:00:00+0100")
  # last epoch in protocol 25/06/2016 19:45:00
  last_epoch_in_protocol = max(which(IMP$rout$r4 == 0))
  expect_equal(M$metalong$timestamp[last_epoch_in_protocol], "2016-06-25T19:45:00+0100")
  
  # Strategy 2 ----
  GGIR(datadir = fn, outputdir = getwd(), studyname = "test", mode = 2, verbose = FALSE,
       # study dates file
       study_dates_file = "study_dates_file.csv",
       study_dates_dateformat = "%d/%m/%Y",
       # strategy
       data_masking_strategy = 2, hrs.del.start = 0,hrs.del.end = 0)
  # load data
  load(dir("output_test/meta/basic/",full.names = TRUE)[1])
  load(dir("output_test/meta/ms2.out/",full.names = TRUE)[1])
  # first epoch in protocol 24/06/2016 00:00:00
  first_epoch_in_protocol = rle(IMP$rout$r4)$lengths[1] + 1
  expect_equal(M$metalong$timestamp[first_epoch_in_protocol], "2016-06-24T00:00:00+0100")
  # last epoch in protocol 25/06/2016 23:45:00
  last_epoch_in_protocol = max(which(IMP$rout$r4 == 0))
  expect_equal(M$metalong$timestamp[last_epoch_in_protocol], "2016-06-25T23:45:00+0100")
  
  # Strategy 3 ----
  GGIR(datadir = fn, outputdir = getwd(), studyname = "test", mode = 2, verbose = FALSE,
       # study dates file
       study_dates_file = "study_dates_file.csv",
       study_dates_dateformat = "%d/%m/%Y",
       # strategy
       data_masking_strategy = 3, ndayswindow = 1,
       hrs.del.start = 0,hrs.del.end = 0)
  # load data
  load(dir("output_test/meta/basic/",full.names = TRUE)[1])
  load(dir("output_test/meta/ms2.out/",full.names = TRUE)[1])
  # first and last epochs are within 24/06/2016 and 25/06/2016
  first_epoch_in_protocol = rle(IMP$rout$r4)$lengths[1] + 1
  t0 = iso8601chartime2POSIX(M$metalong$timestamp[first_epoch_in_protocol], tz = "Europe/London")
  last_epoch_in_protocol = max(which(IMP$rout$r4 == 0))
  t1 = iso8601chartime2POSIX(M$metalong$timestamp[last_epoch_in_protocol], tz = "Europe/London")
  expect_true(t0 == as.POSIXct("2016-06-24 14:45:00", tz = "Europe/London"))
  expect_true(t1 == as.POSIXct("2016-06-25 14:30:00", tz = "Europe/London"))
  # check the total time included is exactly 1 day (= 96 long epochs)
  expect_equal(last_epoch_in_protocol - first_epoch_in_protocol + 1,  96)
  
  GGIR(datadir = fn, outputdir = getwd(), studyname = "test", mode = 2, verbose = FALSE,
       # study dates file
       study_dates_file = "study_dates_file.csv",
       study_dates_dateformat = "%d/%m/%Y",
       # strategy
       data_masking_strategy = 3, ndayswindow = 1,
       hrs.del.start = 6,hrs.del.end = 6)
  # load data
  load(dir("output_test/meta/basic/",full.names = TRUE)[1])
  load(dir("output_test/meta/ms2.out/",full.names = TRUE)[1])
  # first and last epochs are within 24/06/2016 and 25/06/2016
  first_epoch_in_protocol = rle(IMP$rout$r4)$lengths[1] + 1
  t0 = iso8601chartime2POSIX(M$metalong$timestamp[first_epoch_in_protocol], tz = "Europe/London")
  last_epoch_in_protocol = max(which(IMP$rout$r4 == 0))
  t1 = iso8601chartime2POSIX(M$metalong$timestamp[last_epoch_in_protocol], tz = "Europe/London")
  expect_true(t0 >= as.POSIXct("2016-06-24 00:00:00", tz = "Europe/London"))
  expect_true(t1 < as.POSIXct("2016-06-26 00:00:00", tz = "Europe/London"))
  # check the total time included is exactly 0.5 days (= 48 long epochs)
  expect_equal(last_epoch_in_protocol - first_epoch_in_protocol + 1,  48)
  
  # Strategy 4 ----
  GGIR(datadir = fn, outputdir = getwd(), studyname = "test", mode = 2, verbose = FALSE,
       # study dates file
       study_dates_file = "study_dates_file.csv",
       study_dates_dateformat = "%d/%m/%Y",
       # strategy
       data_masking_strategy = 4, ndayswindow = 7,
       hrs.del.start = 0,hrs.del.end = 0)
  # load data
  load(dir("output_test/meta/basic/",full.names = TRUE)[1])
  load(dir("output_test/meta/ms2.out/",full.names = TRUE)[1])
  # first epoch in protocol 24/06/2016 00:00:00
  first_epoch_in_protocol = rle(IMP$rout$r4)$lengths[1] + 1
  expect_equal(M$metalong$timestamp[first_epoch_in_protocol], "2016-06-24T00:00:00+0100")
  # last epoch in protocol 25/06/2016 23:45:00
  last_epoch_in_protocol = max(which(IMP$rout$r4 == 0))
  expect_equal(M$metalong$timestamp[last_epoch_in_protocol], "2016-06-25T23:45:00+0100")
  
  # Strategy 5 -----
  GGIR(datadir = fn, outputdir = getwd(), studyname = "test", mode = 2, verbose = FALSE,
       # study dates file
       study_dates_file = "study_dates_file.csv",
       study_dates_dateformat = "%d/%m/%Y",
       # strategy
       data_masking_strategy = 5, ndayswindow = 1,
       hrs.del.start = 0,hrs.del.end = 0)
  # load data
  load(dir("output_test/meta/basic/",full.names = TRUE)[1])
  load(dir("output_test/meta/ms2.out/",full.names = TRUE)[1])
  # the day most active in this recording is 25/06/2016
  first_epoch_in_protocol = rle(IMP$rout$r4)$lengths[1] + 1
  t0 = iso8601chartime2POSIX(M$metalong$timestamp[first_epoch_in_protocol], tz = "Europe/London")
  last_epoch_in_protocol = max(which(IMP$rout$r4 == 0))
  t1 = iso8601chartime2POSIX(M$metalong$timestamp[last_epoch_in_protocol], tz = "Europe/London")
  expect_true(t0 == as.POSIXct("2016-06-25 00:00:00", tz = "Europe/London"))
  expect_true(t1 == as.POSIXct("2016-06-25 23:45:00", tz = "Europe/London"))
  # check the total time included is exactly 1 day (= 96 long epochs)
  expect_equal(last_epoch_in_protocol - first_epoch_in_protocol + 1,  96)
  
  GGIR(datadir = fn, outputdir = getwd(), studyname = "test", mode = 2, verbose = FALSE,
       # study dates file
       study_dates_file = "study_dates_file.csv",
       study_dates_dateformat = "%d/%m/%Y",
       # strategy
       data_masking_strategy = 5, ndayswindow = 1,
       hrs.del.start = 5,hrs.del.end = 5)
  # load data
  load(dir("output_test/meta/basic/",full.names = TRUE)[1])
  load(dir("output_test/meta/ms2.out/",full.names = TRUE)[1])
  # the day most active in this recording is 25/06/2016
  first_epoch_in_protocol = rle(IMP$rout$r4)$lengths[1] + 1
  t0 = iso8601chartime2POSIX(M$metalong$timestamp[first_epoch_in_protocol], tz = "Europe/London")
  last_epoch_in_protocol = max(which(IMP$rout$r4 == 0))
  t1 = iso8601chartime2POSIX(M$metalong$timestamp[last_epoch_in_protocol], tz = "Europe/London")
  expect_true(t0 == as.POSIXct("2016-06-25 05:00:00", tz = "Europe/London"))
  expect_true(t1 == as.POSIXct("2016-06-25 18:45:00", tz = "Europe/London"))
  
  # start date out of recorded dates (warning and do not trim) -----
  # studydates = data.frame(ID = "123A",
  #                         start = "18/06/2016",
  #                         end = "25/06/2016")
  # write.csv(studydates, "study_dates_file.csv", row.names = FALSE)
  # expect_warning(
  #   GGIR(datadir = fn, outputdir = getwd(), studyname = "test", mode = 2, verbose = FALSE,
  #        # study dates file
  #        study_dates_file = "study_dates_file.csv", 
  #        study_dates_dateformat = "%d/%m/%Y",
  #        # strategy
  #        data_masking_strategy = 1, ndayswindow = 7,
  #        hrs.del.start = 0,hrs.del.end = 0),
  #   regexp = "The start date")
  
  # end date out of recorded dates (warning and do not trim) -----
  # studydates = data.frame(ID = "123A",
  #                         start = "24/06/2016",
  #                         end = "26/06/2016")
  # write.csv(studydates, "study_dates_file.csv", row.names = FALSE)
  # expect_warning(
  #   GGIR(datadir = fn, outputdir = getwd(), studyname = "test", mode = 2, verbose = FALSE,
  #        # study dates file
  #        study_dates_file = "study_dates_file.csv", 
  #        study_dates_dateformat = "%d/%m/%Y",
  #        # strategy
  #        data_masking_strategy = 1, ndayswindow = 7,
  #        hrs.del.start = 0,hrs.del.end = 0),
  #   regexp = "The end date")
  
  # ID not in study dates file (warning and do not trim) -----
  studydates = data.frame(ID = "dummyID",
                          start = "24/06/2016",
                          end = "25/06/2016")
  write.csv(studydates, "study_dates_file.csv", row.names = FALSE)
  expect_warning(
    GGIR(datadir = fn, outputdir = getwd(), studyname = "test", mode = 2, verbose = FALSE,
         # study dates file
         study_dates_file = "study_dates_file.csv", 
         study_dates_dateformat = "%d/%m/%Y",
         # strategy
         data_masking_strategy = 1, ndayswindow = 7,
         hrs.del.start = 0,hrs.del.end = 0),
    regexp = "The ID 123A does not appear")
  
  
  # Missing start date and data_masking_strategy 1
  studydates = data.frame(ID = "123A",
                          start = "",
                          end = "25/06/2016")
  write.csv(studydates, "study_dates_file.csv", row.names = FALSE)
  GGIR(datadir = fn, outputdir = getwd(), studyname = "test", mode = 2, verbose = FALSE,
       # study dates file
       study_dates_file = "study_dates_file.csv", 
       study_dates_dateformat = "%d/%m/%Y",
       # strategy
       data_masking_strategy = 1, hrs.del.start = 0,hrs.del.end = 0)
  # load data
  load(dir("output_test/meta/basic/",full.names = TRUE)[1])
  load(dir("output_test/meta/ms2.out/",full.names = TRUE)[1])
  first_epoch_in_protocol = which(IMP$rout$r4 == 0)[1]
  t0 = iso8601chartime2POSIX(M$metalong$timestamp[first_epoch_in_protocol], tz = "Europe/London")
  last_epoch_in_protocol = max(which(IMP$rout$r4 == 0))
  t1 = iso8601chartime2POSIX(M$metalong$timestamp[last_epoch_in_protocol], tz = "Europe/London")
  expect_true(t0 == as.POSIXct("2016-06-23 09:00:00", tz = "Europe/London"))
  expect_true(t1 == as.POSIXct("2016-06-25 23:45:00", tz = "Europe/London"))
  
  # Missing end date and data_masking_strategy 1
  studydates = data.frame(ID = "123A",
                          start = "24/06/2016",
                          end = "")
  write.csv(studydates, "study_dates_file.csv", row.names = FALSE)
  GGIR(datadir = fn, outputdir = getwd(), studyname = "test", mode = 2, verbose = FALSE,
       # study dates file
       study_dates_file = "study_dates_file.csv", 
       study_dates_dateformat = "%d/%m/%Y",
       # strategy
       data_masking_strategy = 1, hrs.del.start = 0,hrs.del.end = 0)
  # load data
  load(dir("output_test/meta/basic/",full.names = TRUE)[1])
  load(dir("output_test/meta/ms2.out/",full.names = TRUE)[1])
  first_epoch_in_protocol = which(IMP$rout$r4 == 0)[1]
  t0 = iso8601chartime2POSIX(M$metalong$timestamp[first_epoch_in_protocol], tz = "Europe/London")
  last_epoch_in_protocol = max(which(IMP$rout$r4 == 0))
  t1 = iso8601chartime2POSIX(M$metalong$timestamp[last_epoch_in_protocol], tz = "Europe/London")
  expect_true(t0 == as.POSIXct("2016-06-24 00:00:00", tz = "Europe/London"))
  expect_true(t1 == as.POSIXct("2016-06-26 08:45:00", tz = "Europe/London"))
  
  if (file.exists("study_dates_file.csv")) file.remove("study_dates_file.csv")
  if (dir.exists(dn))  unlink(dn, recursive = TRUE)
  if (file.exists(fn)) unlink(fn)
})

library(GGIR)
context("g.part3")
test_that("Part 3 runs fine on a short file not containing a full night", {
  skip_on_cran()
  #=======================
  fn = "123A_testaccfile.csv"
  dn = "output_test"
  if (file.exists(dn)) unlink(dn, recursive = TRUE)

  # Create a test file that doesn't contain a 4am but does contain a noon and a midnight
  # (4am, noon and midnight all carry significance in the internal logic of g.sib.det)
  create_test_acc_csv(Nmin = 1000, starts_at_midnight = FALSE)
  GGIR(mode = c(1:3), datadir = fn, outputdir = getwd(), studyname = "test", f0 = 1, f1 = 1,
       do.report = c(), visualreport = FALSE, viewingwindow = 1,
       do.parallel = FALSE, minimumFileSizeMB = 0, verbose = FALSE)
  
  out.file.path = paste0(dn, "/meta/ms3.out/", fn, ".RData")
  expect_true(file.exists(out.file.path))

  if (file.exists(dn)) unlink(dn, recursive = TRUE)

  # Create a test file that doesn't contain a 4am or a noon, but does have a midnight
  # (4am, noon and midnight all carry significance in the internal logic of g.sib.det)
  create_test_acc_csv(Nmin = 720, start_time = "13:00:00")
  GGIR(mode = c(1:3), datadir = fn, outputdir = getwd(), studyname = "test", f0 = 1, f1 = 1,
       do.report = c(), visualreport = FALSE, viewingwindow = 1,
       do.parallel = FALSE, minimumFileSizeMB = 0, verbose = FALSE)
  
  out.file.path = paste0(dn, "/meta/ms3.out/", fn, ".RData")
  expect_true(file.exists(out.file.path))

  if (file.exists(fn)) file.remove(fn)
  if (file.exists(dn)) unlink(dn, recursive = TRUE)
})

library(GGIR)
context("TryCatch errors in sequential mode")
test_that("Test that potential errors found in sequential mode are skipped and logged correctly", {
  skip_on_cran()
  create_test_acc_csv(Nmin = 2*1440)
  corrupt_files = dir(system.file("testfiles", package = "GGIR"), full.names = T)
  fn = c(corrupt_files, "123A_testaccfile.csv")
  desiredtz = "Europe/London"
  dn = "output_test"
  if (file.exists(dn))  unlink(dn, recursive = TRUE)
  minimumFileSizeMB = 0
  #--------------------------------------------
  # run GGIR (this should not trigger error, but log it and inform in console)
  console_log = capture.output(
    GGIR(datadir = fn, outputdir = getwd(), studyname = "test", 
         do.cal = FALSE, do.imp = FALSE,
         do.parallel = FALSE,
         minimumFileSizeMB = minimumFileSizeMB,
         use_trycatch_serial = TRUE)
  )
  #--------------------------------------------
  # Were errors skipped and logged?
  #   - did part 5 run?
  expect_true(any(grepl("^ Part 5", console_log)))
  #   - were errors logged in part 1?
  expect_true(any(grepl("Errors in part 1... for:", console_log)))
  
  if (dir.exists(dn))  unlink(dn, recursive = TRUE)
  if (file.exists("123A_testaccfile.csv")) unlink("123A_testaccfile.csv")
})

library(GGIR)
context("recordingEndSleepHour")
test_that("recordingEndSleepHour works as expected", {
  skip_on_cran()
  #=======================
  create_test_acc_csv(Nmin = 3.5 * 1440) # ends at 20:45
  fn = "123A_testaccfile.csv"
  tz = "Europe/Amsterdam"
  # this should NOT trigger data expansion
  GGIR(mode = 1, datadir = fn, outputdir = getwd(), 
       studyname = "test", overwrite = TRUE, 
       recordingEndSleepHour = 21, desiredtz = tz,
       visualreport = FALSE, do.report = c(), verbose = FALSE)
  rn = dir("output_test/meta/basic/", full.names = TRUE)
  load(rn[1])
  expect_true(nrow(M$metashort) == 60300)
  # expect_true(M$metashort$timestamp[nrow(M$metashort)] == "2016-06-25T20:44:55+0200")
  
  # errors and warnings work properly
  expect_error( # error from using expand_tail_max_hour instead of new argument
    GGIR(mode = 1, datadir = fn, outputdir = getwd(), 
         studyname = "test", overwrite = TRUE,  desiredtz = tz,
         expand_tail_max_hours = 5, recordingEndSleepHour = NULL,
         visualreport = FALSE, do.report = c(), verbose = FALSE)
  )
  expect_warning( # warning from using both expand_tail_max_hour and recordingEndSleepHour
    GGIR(mode = 1, datadir = fn, outputdir = getwd(), 
         studyname = "test", overwrite = TRUE, desiredtz = tz,
         expand_tail_max_hours = 5, recordingEndSleepHour = 20,
         visualreport = FALSE, do.report = c(), verbose = FALSE)
  )
  
  expect_error( # error from recordingEndSleepHour being too early,
    # this should not produce any output
    GGIR(datadir = fn, outputdir = getwd(),
         studyname = "test", overwrite = TRUE, desiredtz = tz,
         recordingEndSleepHour = 6,
         minimum_MM_length.part5 = 15, verbose = FALSE)
  )
  
  
  # No warning, this should work
  GGIR(datadir = fn, outputdir = getwd(),
       studyname = "test", overwrite = TRUE, desiredtz = tz,
       recordingEndSleepHour = 20, do.anglex = TRUE, do.angley = TRUE,
       minimum_MM_length.part5 = 6, verbose = FALSE)
  rn = dir("output_test/meta/basic/", full.names = TRUE)
  load(rn[1])
  expect_true(nrow(M$metashort) > 60300) # metashort is expanded
  
  # expanded time is not reports
  p2 = read.csv("output_test/results/part2_daysummary.csv")
  expect_equal(p2$N.hours[nrow(p2)], 20.75)  # N hours in last day does not include the expanded time
  
  p4 = read.csv("output_test/results/part4_nightsummary_sleep_cleaned.csv")
  p4full = read.csv("output_test/results/QC/part4_nightsummary_sleep_full.csv")
  expect_equal(nrow(p4), 3) # Night 3 is not in the part 4 reports
  expect_equal(nrow(p4full), 3) # Night 3 is not in the part 4 reports
  expect_equal(sum(p4$sleeponset), 62.77)
  expect_equal(sum(p4$wakeup), 93.505)
  expect_equal(sum(p4$guider_onset), 62.778)
  expect_equal(sum(p4$guider_wakeup), 93.506)
  expect_equal(sum(p4$number_sib_sleepperiod), 114)
  expect_true(all(is.na(p4$longitudinal_axis)))
  
  p5 = read.csv("output_test/results/part5_daysummary_MM_L40M100V400_T5A5.csv")
  expect_equal(nrow(p5), 4) # expanded day appears in MM report
  expect_true(p5$dur_day_spt_min[nrow(p5)] < 23*60) # but expanded time is not accounted for in estimates
  
  #================================================================
  # Redo but now with sensor.location="hip", HASPT.algo = "HorAngle", and all three angles
  # we do not expect sleep to end in last hour here. Further, this allows for improving
  # coverage for the hip-attachment scenario for which we do not have specific unit
  # tests yet. By re-using the test file from this unit test we do not have to create it 
  # twice.

  # No warning, this should work
  GGIR(datadir = fn, outputdir = getwd(),
       studyname = "test", overwrite = TRUE, desiredtz = tz,
       recordingEndSleepHour = 20, do.anglex = TRUE, do.angley = TRUE,
       save_ms5rawlevels = TRUE, sensor.location = "hip", HASPT.algo = "HorAngle",
       save_ms5raw_format = "RData", sleepwindowType = "TimeInBed",
       minimum_MM_length.part5 = 6, verbose = FALSE)
  rn = dir("output_test/meta/basic/", full.names = TRUE)
  load(rn[1])
  expect_true(nrow(M$metashort) > 60300) # metashort is expanded
  
  # expanded time is not reports
  p2 = read.csv("output_test/results/part2_daysummary.csv")
  expect_equal(p2$N.hours[nrow(p2)], 20.75)  # N hours in last day does not include the expanded time
  
  p4 = read.csv("output_test/results/part4_nightsummary_sleep_cleaned.csv")
  p4full = read.csv("output_test/results/QC/part4_nightsummary_sleep_full.csv")
  expect_equal(nrow(p4), 3) # Night 3 is not in the part 4 reports
  expect_equal(nrow(p4full), 3) # Night 3 is not in the part 4 reports
  expect_equal(sum(p4$sleeponset), 62.1)
  expect_equal(sum(p4$wakeup), 81.096)
  expect_equal(sum(p4$guider_inbedStart), 53.17, tolerance = 0.005)
  expect_equal(sum(p4$guider_inbedEnd), 86.85, tolerance = 0.005)
  expect_equal(sum(p4$number_sib_sleepperiod), 24)
  expect_equal(sum(p4$sleepefficiency), 0.501)
  expect_equal(sum(p4$longitudinal_axis), 9)
  
  p5 = read.csv("output_test/results/part5_daysummary_MM_L40M100V400_T5A5.csv")
  expect_equal(nrow(p5), 3) # expanded day appears in MM report
  expect_true(p5$dur_day_spt_min[nrow(p5)] < 23*60) # but expanded time is not accounted for in estimates
  
  # test that time series output has all expected columns including the multiple angle columns
  # this does not align with the focus on sleep in this unit test, but if we have to 
  # write a separate unit test for it then the functions needs to be run twice.
  ts_file = "output_test/meta/ms5.outraw/40_100_400/123A_testaccfile_T5A5.RData"
  expect_true(file.exists(ts_file))
  load(ts_file)
  expect_true(all(c("timenum", "ACC", "SleepPeriodTime", "invalidepoch",
                    "guider", "window", "angle", "anglex", "angley",
                    "class_id", "invalid_fullwindow", "invalid_sleepperiod",
                    "invalid_wakinghours", "timestamp") %in% names(mdat)))
  
  #========================================================
  # Test require_complete_lastnight_part5 = TRUE
  create_test_acc_csv(Nmin = 3 * 1440) # ends at 8:45
  # delete config.csv such that GGIR uses its own default parameter values
  if (file.exists("output_test/config.csv"))  unlink("output_test/config.csv", recursive = TRUE)
  GGIR(datadir = fn, outputdir = getwd(), mode = 1:5,
       studyname = "test", overwrite = TRUE, desiredtz = tz, 
       verbose = FALSE, require_complete_lastnight_part5 = TRUE,
       do.report = 5)
  p5 = read.csv("output_test/results/part5_daysummary_MM_L40M100V400_T5A5.csv")
  expect_equal(nrow(p5), 2) # last window is ignored from because the night ends at 8:45
  
  
  if (file.exists("output_test"))  unlink("output_test", recursive = TRUE)
  if (file.exists(fn)) file.remove(fn)
})

library(GGIR)
context("g.part3")
test_that("Part 3 identifies 1-night scenario", {
  skip_on_cran()
  #=======================
  # Create a test file for a bit over a day worth of data.
  # Nmin=2000 is currently the default, but hardcoding it here in case the default changes.
  create_test_acc_csv(Nmin = 2000)
  fn = "123A_testaccfile.csv"
  dn = "output_test"
  if (file.exists(dn)) unlink(dn, recursive = TRUE)

  GGIR(mode = c(1:3), datadir = fn, outputdir = getwd(), studyname = "test", f0 = 1, f1 = 1,
       do.report = c(), visualreport = FALSE, viewingwindow = 1,
       overwrite = TRUE, do.parallel = FALSE, minimumFileSizeMB = 0,
       verbose = FALSE)
  
  out.file.path = paste0(dn, "/meta/ms3.out/", fn, ".RData")
  expect_true(file.exists(out.file.path))
  load(out.file.path)

  # check that the data file had only one night. The 'night' column should contain only 1s
  for (i in 1:nrow(sib.cla.sum)) {
    expect_that(sib.cla.sum$night[i],equals(1))
  }
   
  if (file.exists(fn)) file.remove(fn)
  if (file.exists(dn)) unlink(dn, recursive = TRUE)
})

library(GGIR)
context("detect_nonwear_clipping")
test_that("detects non wear time", {
  skip_on_cran()
  
  
  # This will produce a 2-day long acc file with a 2-hour block of nonwear
  # starting at the 5th minute of every day
  Ndays = 2
  sf = 3
  create_test_acc_csv(Nmin = Ndays * 1440, sf = sf)
  
  data = as.matrix(read.csv("123A_testaccfile.csv", skip = 10))
  colnames(data) = c("x", "y", "z")

  # 2013 algorithm ------
  # clipthres to 1.7 to test the clipping detection
  NWCW = detect_nonwear_clipping(data = data, nonwear_approach = "2013", 
                                 sf = sf, clipthres = 1.7)
  CW = NWCW$CWav; NW = NWCW$NWav
  NW_rle_2013 = rle(NW)
  CW = sum(NWCW$CWav > 0)

  # 2023 algorithm ------
  NWCW = detect_nonwear_clipping(data = data, nonwear_approach = "2023", sf = sf)
  NW = NWCW$NWav
  NW_rle_2023 = rle(NW)
  # tests ----------------
  # Does it find the 2 periods of nonwear?
  expect_equal(sum(NW_rle_2013$values == 3), 2)
  expect_equal(sum(NW_rle_2023$values == 3), 2)
  
  # Expect the 2023 algorithm finds more nonwear than the 2013
  total_nonwear_2013 = sum(NW_rle_2013$lengths[which(NW_rle_2013$values == 3)])
  total_nonwear_2023 = sum(NW_rle_2023$lengths[which(NW_rle_2023$values == 3)])
  expect_true(total_nonwear_2023 > total_nonwear_2013)
  
  # Expect six ws2 windows with some clipping (values over 1.7 in this test)
  expect_equal(CW, 6)
  
  # remove generated file ------
  if (file.exists("123A_testaccfile.csv")) file.remove("123A_testaccfile.csv")
})

test_that("the wear column is processed correctly", {
  skip_on_cran()

  N = 3000
  sf = 10

  op = options(digits.secs = 4)
  on.exit(options(op))
  
  timeseq = ((0:(N - 1))/sf)

  set.seed(100)
  accx = rnorm(N)
  set.seed(200)
  accy = rnorm(N)
  set.seed(300)
  accz = rnorm(N)
  wear = c(rep(TRUE,N/3),rep(FALSE,N/4),rep(TRUE,N/3),rep(FALSE,N/12))

  # create a csv file with a wear channel
  S = data.frame(x = accx, y = accy, z = accz, time = timeseq, wear = wear, stringsAsFactors = TRUE)
  testfile = "testcsv.csv"
  write.csv(S, file = testfile, row.names = FALSE)
  on.exit({if (file.exists(testfile)) file.remove(testfile)}, add = TRUE)


  I_obj = g.inspectfile(testfile,
                        rmc.dec = ".",
                        rmc.firstrow.acc = 1, rmc.firstrow.header = c(), rmc.unit.time = "UNIXsec",
                        rmc.col.acc = 1:3, rmc.col.temp = c(), rmc.col.time = 4,
                        rmc.sf = sf, desiredtz = "",
                        rmc.col.wear = 5)

  filequality = data.frame(filetooshort = FALSE, filecorrupt = FALSE,
                           filedoesnotholdday = FALSE, NFilePagesSkipped = 0)
  
  M_obj = g.getmeta(datafile = testfile, desiredtz = "", windowsizes = c(1,60,120),
                    inspectfileobject = I_obj,
                    rmc.dec = ".",
                    rmc.firstrow.acc = 1, rmc.firstrow.header = c(), rmc.unit.time = "UNIXsec",
                    rmc.col.acc = 1:3, rmc.col.temp = c(), rmc.col.time = 4,
                    rmc.sf = sf, desiredtz = "",
                    rmc.col.wear = 5)
  expect_equal(M_obj$metalong$nonwearscore, c(3,0,3,3,3))
})

library(GGIR)
context("Recording starts at midnight")
test_that("Test recordings that start at midnight", {
  skip_on_cran()
  
  Ndays = 2
  create_test_acc_csv(Nmin = Ndays*1440, starts_at_midnight = TRUE)
  create_test_sleeplog_csv(advanced = TRUE, begin_date = "2016/06/21")
  fn = "123A_testaccfile.csv"
  sleeplog_fn = "testsleeplogfile.csv"
  desiredtz = "Europe/London"
  dn = "output_test"
  if (file.exists(dn))  unlink(dn, recursive = TRUE)
  minimumFileSizeMB = 0
  #--------------------------------------------
  # run GGIR
  GGIR(datadir = fn, outputdir = getwd(), studyname = "test", idloc = 2,
       verbose = FALSE, desiredtz = desiredtz,
       loglocation = sleeplog_fn, colid = 1, coln1 = 2, 
       nnights = 7, timewindow = "MM", visualreport = FALSE)
  #--------------------------------------------
  # part 1 milestone data starts at midnight
  expect_true(dir.exists(dn))
  rn = dir("output_test/meta/basic/",full.names = TRUE)
  load(rn[1])
  expect_true(grepl("T00:00:00", M$metashort$timestamp[1]))
  expect_true(grepl("T00:00:00", M$metalong$timestamp[1]))
  #-------------------------
  # part 2 data contains 2 complete days
  rn = "output_test/results/part2_daysummary.csv"
  out2 = read.csv(rn)
  expect_equal(nrow(out2), 2)
  expect_true(all(out2$N.hours == 24))
  
  #--------------------------------------------
  # part 4 data contains correct night numbers
  rn = "output_test/results/QC/part4_nightsummary_sleep_full.csv"
  out4 = read.csv(rn)
  expect_equal(nrow(out4), 2)
  expect_equal(out4$guider, rep("sleeplog", 2))
  expect_equal(out4$night, 1:2)
  expect_equal(out4$guider_onset_ts, c("23:00:02", "23:00:03"))
  
  #--------------------------------------------
  # part 5 data contains 2 complete days
  rn = "output_test/results/part5_daysummary_MM_L40M100V400_T5A5.csv"
  out5 = read.csv(rn)
  expect_equal(nrow(out5), 2)
  expect_true(all(out5$dur_day_spt_min == 1440))
  
  
  #===============================================
  # Repeat part 4 with Basic sleeplog
  if (file.exists(sleeplog_fn)) file.remove(sleeplog_fn)
  create_test_sleeplog_csv(advanced = FALSE)
  #--------------------------------------------
  # run GGIR
  GGIR(datadir = fn, outputdir = getwd(), studyname = "test", idloc = 2,
       verbose = FALSE, desiredtz = desiredtz,
       loglocation = sleeplog_fn, colid = 1, coln1 = 2, 
       nnights = 7, timewindow = "MM", mode = 4, do.report = 4,
       visualreport = FALSE, overwrite = TRUE)
  
  #--------------------------------------------
  # part 4 data contains correct night numbers
  rn = "output_test/results/QC/part4_nightsummary_sleep_full.csv"
  out4 = read.csv(rn)
  expect_equal(nrow(out4), 2)
  expect_equal(out4$guider, rep("sleeplog", 2))
  expect_equal(out4$night, 1:2)
  expect_equal(out4$guider_onset_ts, c("23:00:00", "03:00:00"))
  
  if (dir.exists(dn))  unlink(dn, recursive = TRUE)
  if (file.exists(fn)) unlink(fn)
  if (file.exists(sleeplog_fn)) file.remove(sleeplog_fn)
})

HASPT = function(angle, params_sleep = NULL, ws3 = 5,
                 HASPT.algo = "HDCZA", invalid,
                 activity = NULL, marker = NULL,
                 sibs = NULL) {
  tib.threshold = SPTE_start = SPTE_end = part3_guider = spt_crude_estimate = NULL
  
  #-------------------------------------
  # Use marker button as guider:
  # if available and required by user (only Actiwatch and Philips Health band at the moment)
  # See documentation for parameter consider_marker_button.
  if (length(marker) > 0 && params_sleep[["consider_marker_button"]] == TRUE) {
    button_pressed = which(marker != 0)
    N_markers = length(button_pressed)
    # Only consider when there is more than 1 marker and the range spans more than 1 hour
    fraction_invalid = length(which(invalid == 1)) / length(invalid)
    if (N_markers > 1 &&
        diff(range(button_pressed)) * (60/ws3) > 60 &&
        fraction_invalid < 0.5) {
      # Consider the 10 marker buttons closest to the day of interest,
      # or all marker buttons for that day if more than 10
      ranking = sort(marker[button_pressed], decreasing = TRUE, index.return = TRUE)$ix
      button_pressed = button_pressed[ranking[1:pmax(length(which(marker == 1)), 10)]]
      # find marker pair with lowest amount of activity in between and longest duration
      pairs = expand.grid(button_pressed, button_pressed)
      pairs = pairs[which(pairs$Var1 < pairs$Var2),]
      # define duration
      pairs$duration_hours = (pairs$Var2 - pairs$Var1) / (3600/ws3)
      
      # ignore pairs that reflect both short distances and involve imputed marker buttons
      pairs$mark1 = marker[pairs$Var1]
      pairs$mark2 = marker[pairs$Var2]
      pairs_to_ignore = which((pairs$duration_hours < 3 & (pairs$mark1 < 1 | pairs$mark2 < 1)) |
                                (pairs$duration_hours < 1))
      if (length(pairs_to_ignore) > 0 & length(pairs_to_ignore) < nrow(pairs)) {
        pairs = pairs[-pairs_to_ignore,]
      }
      # define activity
      pairs$activity = 0
      if (is.null(activity) && !is.null(sibs)) {
        activity_tmp = 1 - sibs
      } else {
        activity_tmp = activity
      }
      # ignore invalid timepoints
      invalid_index = which(invalid == 1)
      if (length(invalid_index) > 0) {
        activity_tmp[invalid_index] = NA
      }
      pairs$fraction_sibs_outside = NA
      pairs$fraction_sibs_inside = NA
      # create table of each marker button pair and their statistics
      for (i in 1:nrow(pairs)) {
        i0 = pairs$Var1[i]
        i1 = pairs$Var2[i]
        pairs$activity[i] = mean(activity_tmp[i0:i1], na.rm = TRUE) + 1
        if (i0 > 1 && i1 < length(sibs)) {
          outside = c(1:(i0 - 1), (i1 + 1):length(sibs))
          sum_outside = sum(sibs[outside])
          fraction_outside = sum_outside / length(outside)
          fraction_inside = sum(sibs[i0:i1]) / (i1 - i0 + 1)
          pairs$fraction_sibs_inside[i] = fraction_inside
          pairs$fraction_sibs_outside[i] = fraction_outside
        }
      }
      # Only consider when fraction sibs outside is less than inside and time spent outside is less than 6 hours
      pairs = pairs[which(pairs$fraction_sibs_outside < pairs$fraction_sibs_inside),]
      if (nrow(pairs) > 0) {
        # Define performance score
        pairs$dur_score = pmax(abs(pairs$duration_hours - 8), 1) - 1
        pairs$score = pairs$activity * pairs$dur_score
        # find winning marker button pair
        winner = which.min(pairs$score)[1]
        if (!is.na(winner)) {
          start = pairs$Var1[winner]
          end = pairs$Var2[winner]
          HASPT.algo = "markerbutton"
          return(invisible(list(SPTE_start = start, SPTE_end = end,
                                tib.threshold = 0,
                                part3_guider = HASPT.algo)))
        }
      }
    }
  }

  # internal functions ---------
  adjustlength = function(x, invalid) {
    if (length(invalid) > length(x)) {
      invalid = invalid[1:length(x)]
    } else if (length(invalid) < length(x)) {
      invalid = c(invalid, rep(0, length(x) - length(invalid)))
    }
    return(invalid)
  }
  rebuild_rle = function(rlex, N) {
    # after rle values have been changed
    # this function aids in rebuilding it
    y = rep(rlex$values, rlex$length)
    return(rle(y[1:N]))
  }
  # main code -----------------
  if (HASPT.algo != "notused") {
    if (HASPT.algo == "HDCZA") { # original, default
      # x = 5-min rolling median of abs differences
      # threshold = 10th percentile (constrained to 0.13-0.5 if required)
      medabsdi = function(angle) {
        #50th percentile, do not use mean because that will be outlier sensitive
        angvar = stats::median(abs(diff(angle)))
        return(angvar)
      }
      k1 = 5 * (60/ws3)
      
      x = zoo::rollapply(angle, width = k1, FUN = medabsdi, fill = 0) # 5 minute rolling median of the absolute difference
      if (is.null(params_sleep[["HDCZA_threshold"]])) {
        params_sleep[["HDCZA_threshold"]] = c(10, 15)
      }
      if (length(params_sleep[["HDCZA_threshold"]]) == 2) {
        threshold = quantile(x, probs = params_sleep[["HDCZA_threshold"]][1] / 100) * params_sleep[["HDCZA_threshold"]][2]
        if (threshold < 0.13) {
          threshold = 0.13
        } else if (threshold > 0.50) {
          threshold = 0.50
        }
      } else {
        threshold = params_sleep[["HDCZA_threshold"]]
      }
    } else if (HASPT.algo == "HorAngle") {  # if hip, then require horizontal angle
      # x = absolute angle
      # threshold = 45 degrees
      x = abs(angle)
      threshold = params_sleep[["HorAngle_threshold"]]
    } else if (HASPT.algo == "NotWorn") {
      # When protocol is to not wear sensor during the night,
      # and data is collected in count units we do not know angle
      # as needed for HorAngle and HDCZA.
      # Instead look for longest period of zero or very low intensity
      
      # First attempt:
      # However, we need to take into account that there may be some
      # noise in the data, so threshold needs to be above zero
      x = activity
      # smooth x to 5 minute rolling average to reduce sensitivity to sudden peaks
      ma <- function(x, n = 300 / ws3){stats::filter(x, rep(1 / n, n), sides = 2, circular = TRUE)}
      x = ma(x)
      nonzero = which(x != 0)
      if (length(nonzero) > 0) {
        activityThreshold = sd(x[nonzero], na.rm = TRUE) * 0.05
        # For sensewear external data this will not work as it mostly has values of 1 and up.
        if (activityThreshold < min(activity)) {
          activityThreshold = quantile(x, probs = 0.1)
        }
      } else {
        activityThreshold = 0
      }
      # this algorithm looked for x <= threshold, now a minimum quantity is added
      # to the threshold to allow for consistent definition of nomov below
      # i.e., x < threshold
      threshold = activityThreshold + 0.001
      # Always set HASPT.ignore.invalid to NA for HASPT.algo NotWorn
      # because NotWorn is by definition interested in invalid periods
      # and we definitely do not want to rely on imputed time series
      params_sleep[["HASPT.ignore.invalid"]] = NA
    } else if (HASPT.algo == "MotionWare") {
      
      # Auto-Sleep Detection / Marking based on description in
      # Information bulletin no.3 sleep algorithms by Cambridge Neurotechnologies
      # Section 1.6. The letters A, B, C, D, and E below refer to sections in 
      # this description.
      
      # parameters that GGIR user may want to change
      # keep hardcoded for now while this is still experimental
      MarkerButtonLimit = 3
      minimum_sleep_fraction = 0.6
      Prescan_Sleep_Fraction = 0.8
      # (A - classify sleep/wake)
      if (ws3 == 15) {
        score_threshold = 1.5
      } else if (ws3 == 30) {
        score_threshold = 3
      } else if (ws3 == 60) {
        score_threshold = 6
      }
      sleep_wake_score = rep(0, length(activity))
      sleep_wake_score[which(activity <= score_threshold)] = 1
      
      # (B - ignore nonwear)
      # Ignore all detected nonwear
      sleep_wake_score[which(invalid == 1)] = 0 
      
      # Only do analysis if at least 1 hour of sleep and wake in the data
      # and less than 33% of the data points is invalid
      Npoints = length(sleep_wake_score) 
      start_log = end_log = NULL
      # initialise output
      start = 0
      end = 0 #length(sibs)
      tib.threshold = 0
      if (length(which(invalid == 1)) / Npoints <= 0.33 &&
          length(which(sleep_wake_score == 1)) > 60 * (60/ws3) &&
          length(which(sleep_wake_score == 0)) > 60 * (60/ws3)) {
        # (F while loop to repeat steps C, D, E to find all sleep periods) 
        start_point = 1
        while (start_point < Npoints - 180 * (60/ws3)) { 
          # (C - find midpoint)
          find_midpoint = function(cnt, sleep_wake_score, ws3) {
            midpoint = NULL
            avscore = 0
            while (avscore < Prescan_Sleep_Fraction) { #0.8
              j0 = cnt
              j1 = cnt + ((60/ws3) * 180)
              if (j1 > Npoints) break
              avscore = mean(sleep_wake_score[j0:j1])
              cnt = cnt + 1
            }
            if (avscore >= Prescan_Sleep_Fraction) {
              while (avscore > Prescan_Sleep_Fraction - 0.02) {
                j1 = j1 + 1
                if (j1 > Npoints) break
                avscore = mean(sleep_wake_score[j0:j1])
              }
              midpoint = round((j1 - j0) / 2) + j0
            } else {
              j0 = j1 = midpoint = NULL
            }
            invisible(list(j0 = j0, j1 = j1, midpoint = midpoint))
          }
          # use j0 and j1 as estimate for start and end for now
          fmout = find_midpoint(cnt = start_point, sleep_wake_score, ws3)
          
          start = fmout$j0
          end = fmout$j1
          midpoint = fmout$midpoint
          if (is.null(midpoint)) break
          # (D - extend)
          find_edge = function(x, where, midpoint, minimum_sleep_fraction) {
            if (where == "before") {
              xrle = rle(rev(x[1:midpoint]))
            } else if (where == "after") {
              xrle = rle(x[midpoint:length(x)])
            }
            log_len = 0

            if (xrle$values[1] == 0) {
              # If midpoint equals wake, then always include that segment              
              log_len = log_len + xrle$lengths[1]
              xrle$values = xrle$values[-1]
              xrle$lengths = xrle$lengths[-1]
            }
            if (length(xrle$values) > 0) {
              if (xrle$values[1] == 1) {
                # Always include first sleep segment
                log_len = log_len + xrle$lengths[1]
                xrle$values = xrle$values[-1]
                xrle$lengths = xrle$lengths[-1]
              }
              Nsegments = length(xrle$lengths)
              Nout = floor(Nsegments / 2)
              if (Nsegments > 1) {
                # Explore which extension is possible
                wakesegs = cumsum(xrle$lengths[seq(1, by = 2, length.out = Nout)])
                sleepsegs = cumsum(xrle$lengths[seq(2, by = 2, length.out = Nout)])
                extension_duration = sleepsegs + wakesegs
                fractions = sleepsegs / (sleepsegs + wakesegs)
                # ignore (keep) extensions that do not meet sleep fraction
                too_little_sleep = which(fractions < minimum_sleep_fraction)
                if (length(too_little_sleep) > 0) {
                  last_segment = (too_little_sleep[1] - 1) * 2
                } else {
                  last_segment = length(xrle$lengths)
                }
                if (last_segment > 0 && !is.na(last_segment)) {
                  log_len = log_len + sum(xrle$lengths[1:last_segment])
                }
              }
            }
            if (where == "before") {
              edge = midpoint - log_len
            } else if (where == "after") {
              edge = midpoint + log_len
            }
            return(edge)
          }
          start = find_edge(sleep_wake_score, where = "before", midpoint, minimum_sleep_fraction)
          end = find_edge(sleep_wake_score, where = "after", midpoint, minimum_sleep_fraction)
          # (E)
          # ignore marker button if while loop is stuck
          do_no_use_marker = FALSE
          if (length(start_log) > 2 && 
              start_log[length(start_log)] == start_log[length(start_log) - 1]) {
            do_no_use_marker = TRUE
          }
          # is marker button data present?
          if (length(marker) > 0 && do_no_use_marker == FALSE) {
            marker_indices = which(marker == 1)
            if (length(marker_indices) > 1) {
              # check for nearby marker buttons and use those
              delta_start = abs(start - marker_indices)
              ds2 = which(delta_start < (60/ws3) * 60 * MarkerButtonLimit)
              if (length(ds2) > 0) {
                start = rev(marker_indices[ds2])[1] # use last
              }
              delta_end = abs(end - marker_indices)
              de2 = which(delta_end < (60/ws3) * 60 * MarkerButtonLimit)
              if (length(de2) > 0) {
                end = rev(marker_indices[de2])[1] # use last
              }
            }
          }
          start_log = c(start_log, start)
          end_log = c(end_log, end)
          start_point = end + (60/ws3) * 60
        }
        dur_log = end_log - start_log + 1
        # Keep longest sleep periods, because at the moment GGIR cannot handle multiple sleep periods
        start = start_log[which.max(dur_log)[1]]
        end = end_log[which.max(dur_log)[1]]
      }
      # exit function with the result
      return(invisible(list(SPTE_start = start, SPTE_end = end, tib.threshold = 0,
                            part3_guider = HASPT.algo)))
    } else if (HASPT.algo == "HLRB") {
      start = end = NULL
      # smooth classification
      sibs = round(zoo::rollmean(x = sibs, k = (60/ws3) * 60, fill = 0))
      MaxSleepGap = 1
      # ignore all X hour gaps
      srle = rle(c(0, as.numeric(sibs), 0))
      wakegaps = which(srle$values == 0 & srle$lengths < (60/ws3) * 60 * 1)
      wakegaps = wakegaps[which(wakegaps %in% c(1, length(srle$values)) == FALSE)]
      if (length(wakegaps) > 0) {
        srle$values[wakegaps] = 1
      }
      sibs = rep(srle$values, srle$lengths)
      sibs = sibs[-c(1, length(sibs))]
      # keep longest sleep period
      srle = rle(c(0, as.numeric(sibs), 0))
      if (length(srle$lengths) > 2) {
        size_longest_sleep = max(srle$lengths[which(srle$values == 1)])
        sleepgaps = which(srle$values == 1 & srle$lengths < size_longest_sleep)
        if (length(sleepgaps) > 0) {
          srle$values[sleepgaps] = 0
        }
      }
      sibs = c(0, rep(srle$values, srle$lengths), 0)
      start = which(diff(sibs) == 1)
      end = which(diff(sibs) == -1)
      # If there are multiple windows with equal length take last one
      if (length(start) > 1 & length(end) > 1) {
        start = start[length(end)]
        end = end[length(end)]
      }
      sibs = sibs[-c(1, length(sibs))]
      # If start and end are not found try again
      # with shorter filter window
      if (length(start) == 0 & length(end) == 0) {
        start = 1
        end = length(sibs)
      }
      # exit function with the result
      return(invisible(list(SPTE_start = start, SPTE_end = end, tib.threshold = 0,
                            part3_guider = HASPT.algo)))
    }

    # Now define nomov periods with the selected strategy for invalid time
    nomov = rep(0,length(x)) # no movement
    invalid = adjustlength(x, invalid)
    if (is.na(params_sleep[["HASPT.ignore.invalid"]])) { # all invalid = no movement
      nomov[which(x < threshold | invalid == 1)] = 1
    } else if (params_sleep[["HASPT.ignore.invalid"]] == FALSE) { # calculate no movement over the imputed angle
      nomov[which(x < threshold)] = 1
    } else if (params_sleep[["HASPT.ignore.invalid"]] == TRUE) {  # all invalid = movement
      nomov[which(x < threshold & invalid == 0)] = 1
    }
    
    # initialise output
    SPTE_end = c()
    SPTE_start = c()
    tib.threshold = c()
    part3_guider = "none"
    #------------------------------------------------------
    # apply final 3 steps to estimate the main SPT window
    # these steps are the same for HDCZA, HorAngle, and NotWorn
    N = length(x)
    spt_estimate = rep(NA, N)
    nomov = c(0, nomov, 0)
    rle_nomov = rle(nomov)
    # only keep the blocks that are long enough
    fraction_night_invalid = sum(invalid) / length(invalid)
    if (fraction_night_invalid < 1) {
      # Step -3: ignore blocks that are too short
      blocks_to_remove = which(rle_nomov$values == 1 & rle_nomov$lengths <= (60 / ws3) * params_sleep[["spt_min_block_dur"]])    
      blocks_to_remove = blocks_to_remove[which(blocks_to_remove %in% c(1, length(rle_nomov$values)) == FALSE)]
      if (length(blocks_to_remove) > 0) {
        rle_nomov$values[blocks_to_remove] = 0
        rle_nomov = rebuild_rle(rle_nomov, N)
      }
      # Step -2: fill gaps that are short
      Nsegments = length(rle_nomov$lengths)
      if (!is.null(params_sleep[["spt_max_gap_ratio"]]) && params_sleep[["spt_max_gap_ratio"]] < 1 && Nsegments > 3) {
        # Note: spt_max_gap_ratio is NULL by default
        gap_ratios = data.frame(values = rle_nomov$values, lengths = rle_nomov$lengths)
        gap_ratios$ratio = gap_ratios$length_after = gap_ratios$length_before = 0
        gap_ratios$length_after[1:(Nsegments - 1)] = gap_ratios$lengths[2:Nsegments]
        gap_ratios$length_before[2:Nsegments] = gap_ratios$lengths[1:(Nsegments - 1)]
        gaps_to_fill = which(gap_ratios$values == 0 &
                               gap_ratios$lengths < (60 / ws3) * params_sleep[["spt_max_gap_dur"]] &
                               gap_ratios$lengths / gap_ratios$length_after < params_sleep[["spt_max_gap_ratio"]] &
                               gap_ratios$lengths / gap_ratios$length_before < params_sleep[["spt_max_gap_ratio"]])
      } else {
        gaps_to_fill = which(rle_nomov$values == 0 & rle_nomov$lengths < (60 / ws3) * params_sleep[["spt_max_gap_dur"]])
      }
      if (length(gaps_to_fill) > 0) {
        gaps_to_fill = gaps_to_fill[which(gaps_to_fill %in% c(1, length(rle_nomov$values)) == FALSE)]
      }
      if (length(gaps_to_fill) > 0) {
        rle_nomov$values[gaps_to_fill] = 1
        rle_nomov = rebuild_rle(rle_nomov, N)
      }
      # Retain estimate before the selecting the longest block
      spt_crude_estimate = rep(rle_nomov$values, rle_nomov$length)
      # Step -1: keep indices for longest spt block
      if (1 %in% rle_nomov$values) {
        max_length =  max(rle_nomov$length[which(rle_nomov$values == 1)])
        rle_nomov$values[which(rle_nomov$values == 1 & rle_nomov$length == max_length)[1]] = 2
        rle_nomov$values[which(rle_nomov$values != 2)] = 0
        rle_nomov$values[which(rle_nomov$values == 2)] = 1
      }
      spt_estimate = rep(rle_nomov$values, rle_nomov$length)
      spt_estimate = spt_estimate[1:length(x)]
      # identify start and end of longest block
      SPTE_start = which(diff(c(0, spt_estimate, 0)) == 1) - 1
      SPTE_end = which(diff(c(0, spt_estimate, 0)) == -1) - 1
      if (length(SPTE_start) == 1 && length(SPTE_end) == 1 && SPTE_start == 0) SPTE_start = 1
      if (length(SPTE_start) > 0 & length(SPTE_end) > 0) {
        spt_crude_estimate[SPTE_start:SPTE_end] = 2 # clarify in crude estimate what final estimate is
      }
      spt_crude_estimate = spt_crude_estimate[1:max(c(length(angle), length(activity), length(sibs)))]
      part3_guider = HASPT.algo
      if (is.na(params_sleep[["HASPT.ignore.invalid"]])) {
        # investigate if invalid time was included in the SPT definition,
        # and if so, keep track of that in the guider. This is needed in the
        # case that sleeplog is used, to inform part 4 that it should
        # trust the sleeplog times for this specific night.
        spt_long = rep(0, length(invalid))
        if (length(SPTE_start) > 0 & length(SPTE_end) > 0) {
          spt_long[SPTE_start:SPTE_end] = 1
        }
        invalid_in_spt = which(invalid == 1 & spt_long == 1)
        if (length(invalid_in_spt)) {
          part3_guider = paste0(HASPT.algo, "+invalid")
        }
      }
    }
    tib.threshold = threshold
  }
  invisible(list(SPTE_start = SPTE_start, SPTE_end = SPTE_end, tib.threshold = tib.threshold,
                 part3_guider = part3_guider, spt_crude_estimate = spt_crude_estimate))
}

createConfigFile = function(config.parameters = c(), GGIRversion = "") {
  if (length(config.parameters) == 0) {
    stop("Configuration parameters not be found.")
  }
  out = matrix("", length(config.parameters) + 2, 3)
  possible_params_objectnames = c("params_247", "params_cleaning", "params_general",
                                  "params_metrics", "params_output",
                                  "params_phyact", "params_rawdata", "params_sleep")
  for (i in 1:length(config.parameters)) {
    NM = names(config.parameters)[i]
    if (NM %in% possible_params_objectnames) {
      # Replace NULL values before converting to data.frame
      config.parameters[[i]] <- lapply(config.parameters[[i]], lapply, function(x)ifelse(is.null(x), "c()", x))
      Value = as.data.frame(t(data.frame(t(sapply(config.parameters[[i]],c)), stringsAsFactors = FALSE)))
      Value$col1 = row.names(Value)
      Value$col3 = NM
      Value = Value[,c("col1", "V1", "col3")]
      colnames(out) = colnames(Value)
      # Replace empty lists by c(), replace vector lists by collapse vector
      # Replace non-empty lists by same value but without the list
      myfun = function(x) {
        x = ifelse(test = is.list(x) & length(unlist(x) == 1),
                   yes = ifelse(test = length(unlist(x)) > 1,
                                yes = paste0("c(", paste0(unlist(x), collapse = ","),")", collapse = ""),
                                no = unlist(x)),
                   no =  x)
        x = ifelse(test = is.list(x) & length(unlist(x)) == 0,
                   yes = "c()",
                   no =  x)
        return(x)
      }
      Value$V1 <- as.character(lapply(Value$V1, myfun))
      out = rbind(out, as.matrix(Value)) # append to the end
    } else {
      out[i,1] = NM
      Value = config.parameters[[i]]
      if (length(Value) == 0) {
        Value = 'c()'
      }
      if (length(Value) > 1) {
        Value = paste0("c(",paste(Value,collapse = ","),")")
      }
      if (is.function(Value) == FALSE | is.list(Value) == FALSE) {
        out[i,2] = as.character(Value)
      } else {
        out[i,2] = 'c()' # function or list objects are not stored in the config file, the user will have to provide these explicitely
      }
      if (NM %in% c("windowsizes", "chunksize",
                    "minloadcrit", "do.enmo", "do.lfenmo", "do.en", "do.bfen", "do.hfen",
                    "do.hfenplus", "do.mad", "do.anglex", "do.angley", "do.anglez",
                    "do.roll_med_acc_x", "do.roll_med_acc_y", "do.roll_med_acc_z",
                    "do.dev_roll_med_acc_x", "do.dev_roll_med_acc_y", "do.dev_roll_med_acc_z",
                    "do.enmoa", "printsummary", "do.cal", "print.filename",
                    "backup.cal.coef", "dayborder", "dynrange",
                    "configtz", "do.lfen", "hb", "lb", "n", "myfun") ==  TRUE) {
        out[i,3] = "Calibration, Feature extraction, Epoch size, Time zone"
      } else if (NM %in% c("strategy", "data_masking_strategy", "hrs.del.start", "hrs.del.end", "maxdur",
                           "includedaycrit", "M5L5res", "winhr", "qwindow", "qlevels",
                           "ilevels", "mvpathreshold", "boutcriter", "ndayswindow",
                           "do.imp", "epochvalues2csv", "mvpadur",
                           "dayborder", "closedbout",
                           "IVIS_windowsize_minutes", "IVIS_epochsize_seconds", "iglevels",
                           "IVIS.activity.metric", "TimeSegments2ZeroFile", "qM5L5", "do.part3.pdf") ==  TRUE) {
        out[i,3] = "Study design, Parameters descriptive analysis"
      } else if (NM %in% c("anglethreshold", "timethreshold",
                           "ignorenonwear", "constrain2range", "sensor.location") == TRUE) {
        out[i,3] = "Parameters sleep detection"
      } else if (NM %in% c("colid", "coln1", "nnights", "outliers.only",
                           "excludefirstlast", "criterror", "includenightcrit", "loglocation",
                           "relyonsleeplog", "sleeplogidnum", "def.noc.sleep", "do.visual") ==  TRUE) {
        out[i,3] = "Parameters sleep period time detection with or wihout sleeplog"
      } else if (NM %in% c("excludefirstlast.part5", "windowsizes", "boutcriter.in",
                           "boutcriter.lig", "boutcriter.mvpa", "threshold.lig", "threshold.mod",
                           "threshold.vig", "timewindow", "boutdur.mvpa",
                           "boutdur.in", "boutdur.lig", "save_ms5rawlevels") ==  TRUE) {
        out[i,3] = "Parameters time-use variables"
      } else if (NM %in% c("viewingwindow", "visualreport", "dofirstpage") ==  TRUE) {
        out[i,3] = "Visual report"
      }
    }
  }
  SI = sessionInfo()
  GGIRread_version = "could not extract version"
  if (is.element('GGIRread', installed.packages()[,1])) {
    GGIRread_version = as.character(utils::packageVersion("GGIRread"))
    if (length(GGIRread_version) != 1) GGIRread_version = sessionInfo()$otherPkgs$GGIRread$Version
  }
  out = rbind(out, matrix(c("GGIRread_version", GGIRread_version, "not applicable"), nrow = 1))
  out = rbind(out, matrix(c("GGIRversion", GGIRversion, "not applicable"), nrow = 1))
  out = rbind(out, matrix(c("R_version", SI$R.version$version.string, "not applicable"), nrow = 1)) 
  out = out[which(!is.na(out[,1])),]
  out = as.data.frame(out, stringsAsFactors = TRUE)
  row.names(out) <- NULL
  colnames(out) = c("argument","value","context")
  out$value = as.character(out$value)
  out$argument = as.character(out$argument)
  out$context = factor(as.character(out$context))
  out = out[which(out$argument %in% c("", possible_params_objectnames) == FALSE),]
  out = out[order(out$context),]
  return(out)
}

visualReport = function(metadatadir = c(), 
                        f0 = c(), f1 = c(), 
                        verbose = TRUE,
                        part6_threshold_combi = NULL,
                        GGIRversion = NULL,
                        params_sleep = NULL, params_output = NULL,
                        params_general = NULL) {
  if (!file.exists(paste0(metadatadir, "/results/file summary reports"))) {
    dir.create(file.path(paste0(metadatadir, "/results"), "file summary reports"))
  }
  desiredtz = params_general[["desiredtz"]]
  #------------------------------------------------------
  fnames = dir(paste0(metadatadir,"/meta/ms5.outraw/", part6_threshold_combi))
  if (f1 > length(fnames) | f1 == 0) f1 = length(fnames)
  if (f0 > length(fnames) | f0 == 0) f0 = 1
  
  main_visualreport = function(i, f0, f1, metadatadir, part6_threshold_combi,
                               fnames.ms5raw, ylabels_plot2, GGIRversion, desiredtz,
                               behavioral_code_names, behavioral_codes,
                               binary_vars) {
    # Declare local functions (internal, because this eases passing the function on to parallel process:
    correctRect = function(starti, endi, NR, epochSize) {
      if (endi[length(endi)] > NR) {
        if (starti[length(starti)] < endi[length(endi)] - 1) {
          endi[length(endi)] = endi[length(endi)] - 1
        } else {
          starti = starti[1:(length(starti) - 1)]
          endi = endi[1:(length(endi) - 1)]
        }
      }
      invisible(list(starti = starti, endi = endi))
    }
    
    changepoints = function(x) {
      Nvalues = length(x)
      if (Nvalues > 3) {
        changes = which(x[2:Nvalues] != x[1:(Nvalues - 1)])
        changes = sort(unique(c(1, changes, changes + 1, Nvalues)))
        if (length(changes) == 0) {
          changes = 1:Nvalues
        }
      } else {
        changes = 1:Nvalues
      } 
      return(changes)
    }  
    
    panelplot = function(mdat, ylabels_plot2, binary_vars,
                         behavioral_code_names, behavioral_codes, title = "", 
                         hrsPerRow = NULL, plotid = 0, 
                         legend_items = NULL, lux_available = FALSE,
                         step_count_available = FALSE, epochSize = 60, focus = "day",
                         temperature_available = FALSE) {
      window_duration = mdat$timenum[nrow(mdat)] - mdat$timenum[1]
      signalcolor = "black"
      
      # Prepare time tick points
      date = paste0(as.numeric(format(mdat$timestamp, "%d")), " ", format(mdat$timestamp, "%b"))
      hour = as.numeric(format(mdat$timestamp, "%H"))
      min = as.numeric(format(mdat$timestamp, "%M"))
      sec = as.numeric(format(mdat$timestamp, "%S"))
      ticks = which(min == 0 & sec == 0 & hour %in% seq(0, 24, by = 1))
      atTime = mdat$timestamp[ticks]
      datLIM = as.Date(min(mdat$timestamp, na.rm = TRUE), tz = desiredtz)
      if (focus == "day") {
        XLIM = as.POSIXct(paste0(datLIM[1], " 00:00:00"), tz = desiredtz)
      } else if (focus == "night") {
        XLIM = as.POSIXct(paste0(datLIM[1], " 12:00:00"), tz = desiredtz)
      }
      XLIM[2] = XLIM[1] + hrsPerRow * 3600
      
      if (step_count_available == TRUE) {
        steps_per_hour = round(zoo::rollsum(x = mdat$step_count, k = epochSize, fill = NA) / 100)
        steps_per_hour = as.character(steps_per_hour)
        steps_per_hour = gsub(pattern = "0", replacement = "", x = steps_per_hour)
      }
      if (lux_available == TRUE) {
        lux_per_hour = round(zoo::rollmax(x = mdat$lightpeak, k = epochSize, fill = NA) / 1000)
        lux_per_hour = as.character(lux_per_hour)
        lux_per_hour = gsub(pattern = "0", replacement = "", x = lux_per_hour)
      }
      
      par(mar = c(3, 0, 1.5, 2.5))
      #============================================
      # Acceleration and angle signal
      # scale 0 - 100, where bottom half is used for angle and top half for acceleration
      if (lux_available == TRUE) {
        Ymax = 110
      } else {
        Ymax = 100
      }
      if (temperature_available == TRUE) {
        Ymin = -10
      } else {
        Ymin = 0
      }
      accy = (mdat$ACC / 10) + 50
      
      # identify angle columns
      anglecols = sort(grep(pattern = "angle", x = colnames(mdat), value = TRUE))
      Nangles = length(anglecols)
      if (Nangles == 1) {
        ang1 = (((mdat[, anglecols] + 90) * 40) / 180)
      } else if (Nangles == 2) {
        ang1 = (((mdat[, anglecols[1]] + 90) * 40) / 180)
        ang2 = (((mdat[, anglecols[1]] + 90) * 40) / 180) #+ 20
      } else if (Nangles == 3) {
        ang1 = (((mdat[, anglecols[1]] + 90) * 40) / 180)
        ang2 = (((mdat[, anglecols[2]] + 90) * 40) / 180) #+ 13
        ang3 = (((mdat[, anglecols[3]] + 90) * 40) / 180) #+ 26
      }
      
      if (lux_available == TRUE) {
        # luxy = ceiling(pmin(mdat$lightpeak + 1, 20000) / 2000) + 100
        luxy = (pmin(mdat$lightpeak, 20000) / 2000) + 100
      }
      if (temperature_available == TRUE) {
        temperaturey = round(pmax(pmin(diff(c(0, mdat$temperature)), 1), -1))
        temperaturey[which(temperaturey < 0)] = -1
        temperaturey[which(temperaturey > 0)] = 1
      }
      plot(0:1, 0:1,
           ylim = c(Ymin, Ymax), xlim = XLIM, 
           xaxt = 'n', axes = FALSE,
           xlab = "", ylab = "")
      
      cex_axis = 0.5
      cex_mtext = ifelse(hrsPerRow <= 36, yes = 0.4, no = 0.35)
      line = 0
      line_delta = 0.5
      # assign timestamp axis:
      mtext(text = hour[ticks], side = 1, line = line, cex = cex_mtext, at = atTime)
      mtext(text = "hour", side = 1, line = line, cex = cex_mtext,
            at = atTime[length(atTime)] + 5000, adj = 0)
      
      line = line + line_delta
      # with window numbers
      windowNow = mdat$window[ticks]
      changes = changepoints(windowNow)
      mtext(text = windowNow[changes], side = 1, line = line, cex = cex_mtext,
            at = atTime[changes])
      mtext(text = "window", side = 1, line = line, cex = cex_mtext,
            at = atTime[length(atTime)] + 5000, adj = 0)
      
      line = line + line_delta
      # with dates
      dateNow = as.numeric(format(mdat$timestamp[ticks], "%d"))
      
      changes = changepoints(dateNow)
      mtext(text = dateNow[changes], side = 1, line = line, cex = cex_mtext,
            at = atTime[changes])
      mtext(text = "date", side = 1, line = line, cex = cex_mtext,
            at = atTime[length(atTime)] + 5000, adj = 0)
      line = line + line_delta
      
      # optional variables
      if (lux_available == TRUE) {
        mtext(text = lux_per_hour[ticks], side = 1, line = line, cex = cex_mtext,
              at = atTime)
        mtext(text = "kLux", side = 1, line = line, cex = cex_mtext,
              at = atTime[length(atTime)] + 5000, adj = 0)
        line = line + line_delta
      }
      if (step_count_available == TRUE) {
        mtext(text = steps_per_hour[ticks], side = 1, line = line, cex = cex_mtext,
              at = atTime)
        mtext(text = "steps x100", side = 1, line = line, cex = cex_mtext,
              at = atTime[length(atTime)] + 5000, adj = 0)
      }
      
      #============================================
      # add rectangular blocks to reflect classes
      Nlevels = max(legend_items$level, na.rm = TRUE) #sum(Nlevels) - 1 # -1 because invalidepoch will not be a rectangle
      for (labi in 1:length(legend_items$name)) {
        if (legend_items$name[labi] == "invalid") {
          # Will be plotted further down
        } else if (legend_items$name[labi] %in% c("nap", "sib")) {
          bin_ts = mdat$sib
          if (legend_items$name[labi] == "sib") {
            bin_ts[which(bin_ts != 1)] = 0
          } else if (legend_items$name[labi] == "nap") {
            bin_ts[which(bin_ts != 2)] = 0
            bin_ts[which(bin_ts == 2)] = 1
          }
          freqtab = table(bin_ts)
          if (length(freqtab) > 1 || names(freqtab)[1] == "1") {
            starti = which(diff(c(0, bin_ts)) == 1)
            endi = which(diff(c(bin_ts, 0)) == -1) + 1
            newi = correctRect(starti, endi, NR = nrow(mdat), epochSize)
            t0 = mdat$timestamp[starti = newi$starti]
            t1 = mdat$timestamp[endi = newi$endi]
            y0 = 25
            y1 = 50
            if (legend_items$name[labi] == "nap") {
              y1 = 100
            }
            col = legend_items$col[labi]
            rect(xleft = t0, xright = t1, ybottom = y0, ytop = y1, col = col, border = FALSE)
          }
        } else {
          diary_names = c("diary_nap", "diary_nonwear", "diary_sleepwindow", "diary_timeinbed")
          if (legend_items$name[labi] %in% diary_names) {
            # Diary-based classes
            if (legend_items$name[labi] == "diary_sleepwindow") {
              relevant_labels = c("sleeplog", "sleeplog+bedlog")
            } else if (legend_items$name[labi] == "diary_timeinbed") {
              relevant_labels = c("bedlog", "sleeplog+bedlog")
            } else {
              relevant_labels = legend_items$name[labi]
            }
            tempi = which(mdat$selfreported %in% relevant_labels)
            y0 = 0
            y1 = 25
            if (legend_items$name[labi] == "diary_sleepwindow") {
              y0 = 0
              y1 = 12.5
            } else if (legend_items$name[labi] == "diary_timeinbed") {
              y0 = 12.5
              y1 = 25
            }
          } else {
            # Accelerometer- based classes
            tempi = which(mdat$class_id == legend_items$code[labi])
            y0 = 50
            if (length(grep(pattern = "unbt", x = legend_items$name[labi])) > 0) {
              y1 = 90 # lower rectangle for unbouted behaviour
            } else {
              y1 = 100
            }
          }
          if (length(tempi) > 0) {
            A = rep(0, nrow(mdat))
            A[tempi] = 1
            starti = which(diff(c(0, A)) == 1)
            endi = which(diff(c(A, 0)) == -1) + 1
            newi = correctRect(starti, endi, NR = nrow(mdat), epochSize)
            starti = newi$starti
            endi = newi$endi
            t0 = mdat$timestamp[starti]
            t1 = mdat$timestamp[endi]
            col = legend_items$col[labi]
            rect(xleft = t0, xright = t1, ybottom = y0, ytop = y1 , col = col, border = FALSE)
          }
        }
      }
      
      # Add angle grid lines
      abline(h = c(10, 30), lty = 2, lwd = 0.3)
      if (Nangles > 1) {
        abline(h = c(0, 20, 40), lty = 3, lwd = 0.3)
      }
      # Add acc line
      lines(mdat$timestamp, accy, type = "l",
            col = signalcolor,
            lwd = 0.3)
      
      if (lux_available == TRUE) {
        # Add lux
        lines(mdat$timestamp, luxy, type = "l",
              col = signalcolor,
              lwd = 0.3)
      }
      if (temperature_available == TRUE) {
        # Add temperature
        text(x = mdat$timestamp[which(temperaturey == -1)], y = -9, labels = "-", cex = 0.5)
        text(x = mdat$timestamp[which(temperaturey == 1)], y = -9, labels = "+", cex = 0.5)
      }
      angleColor = ifelse(Nangles > 1, yes = "orange", no = signalcolor)
      # Add angle lines on top
      if (Nangles > 0) {
        lines(mdat$timestamp, ang1, type = "l", col = angleColor, lwd = 0.3)
      }
      if (Nangles > 1) {
        lines(mdat$timestamp, ang2, type = "l", col = "red", lwd = 0.3)
      }
      if (Nangles > 2) {
        lines(mdat$timestamp, ang3, type = "l", col = "green", lwd = 0.3)
      }
      text(x = mdat$timestamp[1], y = 60, labels = "Acceleration",
           pos = 4, cex = 0.7, col = signalcolor, font = 2)
      textAngle = ifelse(Nangles > 1, yes = "Angles", no = "Angle")
      text(x = mdat$timestamp[1], y = 40, labels = textAngle,
           pos = 4, cex = 0.7, col = signalcolor, font = 2)
      
      if (lux_available == TRUE) {
        text(x = mdat$timestamp[1], y = 105, labels = "Lux",
             pos = 4, cex = 0.7, col = signalcolor, font = 2)
      }
      if (temperature_available == TRUE) {
        text(x = mdat$timestamp[1], y = -2, labels = "Temperature change",
             pos = 4, cex = 0.7, col = signalcolor, font = 2)
      }
      # Highlight invalid epochs as hashed area on top of all rects
      if ("invalid" %in% legend_items$name) {
        freqtab = table(mdat$invalid)
        if (length(freqtab) > 1 || names(freqtab)[1] == "1") {
          starti = which(diff(c(0, mdat$invalid)) == 1)
          endi = which(diff(c(mdat$invalid, 0)) == -1) + 1
          newi = correctRect(starti, endi, NR = nrow(mdat), epochSize)
          t0 = mdat$timestamp[newi$starti]
          t1 = mdat$timestamp[newi$endi]
          col = legend_items$col[which(legend_items$name == "invalid")]
          if (temperature_available == TRUE) {
            y0 = -5
          } else {
            y0 = 5
          }
          if (lux_available == TRUE) {
            y1 = 105
          } else {
            y1 = 95
          }
          transparantWhite = grDevices::adjustcolor(col = "white", alpha.f = 0.8)
          rect(xleft = t0, xright = t1, ybottom = y0, ytop = y1,
               col = transparantWhite, lwd = 0.8, border = "grey")
        }
      }
      
      # marker button
      if ("marker" %in% colnames(mdat)) {
        marker_moments = which(mdat$marker == 1)
        if (length(marker_moments) > 0) {
          arrows(x0 = mdat$timestamp[marker_moments],
                 x1 = mdat$timestamp[marker_moments],
                 y0 = 30, y1 = 0, col = "purple", lwd = 1, length = 0.06)
          text(x = mdat$timestamp[marker_moments],y = 30, labels = "M", font = 2, col = "purple", cex = 1.2)
        }
      }
      
      # onset/wake lines:
      window_edges = which(diff(mdat$SleepPeriodTime) != 0)
      if (length(window_edges) > 0) {
        abline(v = mdat$timestamp[window_edges], col = "black", lwd = 1.5, lty = 2)
        for (wei in 1:length(window_edges)) {
          # add text labels for onset, wake, weekday and guider being used
          # the below guider names are the guider names as defined in 
          # function g.part5.savetimeseries
          guider_names = c('unknown', 'sleeplog', 'HDCZA', 'setwindow', 
                           'L512', 'HorAngle', 'NotWorn', 'markerbutton',
                           'HLRB', 'MotionWare')
          guider_name =  paste0("guided by: ", guider_names[mdat$guider[window_edges[wei]] + 1])        
          
          if (mdat$SleepPeriodTime[window_edges[wei]] == 1) {
            toptext = paste0("wake-up ", weekdays(mdat$timestamp[window_edges[wei]]))
            
            text(x = mdat$timestamp[window_edges[wei]], y = 92,
                 labels = guider_name, las = 2, srt = 90, pos = 2,
                 offset = 0.4, adj = 0, cex = 0.6)
          } else {
            toptext = "onset"  
            text(x = mdat$timestamp[window_edges[wei]], y = 2,
                 labels = guider_name, las = 2, srt = 90, pos = 4,
                 offset = 0.4, adj = 0, cex = 0.6)
          }
          mtext(text = toptext, side = 3, line = 0, cex = 0.5,
                at = mdat$timestamp[window_edges[wei]])
          
        }
      }
      # diaryImputationCode availabe
      if ("diaryImputationCode" %in% names(mdat)) {
        temp_index_0 = pmax(1, round(nrow(mdat) * (hrsPerRow - 24) / 24))
        diaryImputationCode = unique(mdat$diaryImputationCode[temp_index_0:nrow(mdat)])
        diaryImputationCode = diaryImputationCode[is.na(diaryImputationCode) == FALSE]
        if (length(diaryImputationCode) > 0) {
          diaryImputationCode = unlist(strsplit(as.character(diaryImputationCode), ""))
          missingZeros = 4 - length(diaryImputationCode)
          if (missingZeros > 0) {
            diaryImputationCode = c(rep("0", missingZeros), diaryImputationCode)
          }
          diaryImputationCode = paste0(diaryImputationCode, collapse = "")
          diaryImputationCode = paste0("diary imputation: ", diaryImputationCode)
          mtext(text = diaryImputationCode, side = 4, line = -2, cex = 0.3, las = 1)
        }
      }
    }
    
    gen_col_names = function(behavioral_code_names, behavioral_codes = NULL,
                             name, legend_items, colour = NULL,
                             level = NULL, reverse = TRUE) {
      # generate colour and names for legend
      vars = grep(pattern = name, x = behavioral_code_names, value = TRUE)
      Nitems = length(vars)
      if (Nitems > 0) {
        legend_items$name = c(legend_items$name, vars)
        legend_items$level = c(legend_items$level, rep(level, Nitems))
        if (!is.null(behavioral_codes)) {
          legend_items$code = c(legend_items$code, behavioral_codes[which(behavioral_code_names %in% vars)])
        } else {
          legend_items$code = c(legend_items$code, rep(-1, Nitems))
        }
        col = rep(colour, length.out = Nitems)
        if (length(colour) == 1) {
          for (ci in 1:Nitems) {
            col[ci] = grDevices::adjustcolor(col = col[ci],
                                             alpha.f = 0.2 + (ci / Nitems) * 0.8)
          }
        }
        if (reverse == TRUE) col = rev(col)
        legend_items$col = c(legend_items$col, col)
      }
      return(legend_items)
    }
    # end of internal funtions
    #-------------------
    mdat = NULL
    load(file = paste0(metadatadir, "/meta/ms5.outraw/",
                       part6_threshold_combi, "/", fnames.ms5raw[i]))
    if (length(mdat) != 0 && nrow(mdat) != 0) {
      names(mdat)[which(names(mdat) == "sibdetection")] = "sib"
      names(mdat)[which(names(mdat) == "invalidepoch")] = "invalid"
      mdat$sib[which(mdat$SleepPeriodTime == 1)] = 0
      if ("selfreported" %in% colnames(mdat)) {
        sr_levelnames = levels(mdat$selfreported)
        sr_levelnames = gsub("nap", "diary_nap", x = sr_levelnames)
        sr_levelnames = gsub("nonwear", "diary_nonwear", x = sr_levelnames)
        levels(mdat$selfreported) = sr_levelnames
      }
      # assess which information is available
      if ("lightpeak" %in% colnames(mdat)) {
        lux_available = TRUE
      } else {
        lux_available = FALSE
      }
      if ("temperature" %in% colnames(mdat)) {
        temperature_available = TRUE
      } else {
        temperature_available = FALSE
      }
      if ("step_count" %in% colnames(mdat)) {
        step_count_available = TRUE
      } else {
        step_count_available = FALSE
      }
      epochSize = diff(mdat$timenum[1:2])
      # Define legend
      legend_items = list(col = NULL, name = NULL, code = NULL, level = NULL)
      
      # Sleep diary
      legend_items = gen_col_names(ylabels_plot2, name = "diary",
                                   legend_items = legend_items,
                                   colour = c("#FF00FF", "#FFD700", "#7CFC00", "red"),
                                   level = 1, reverse = TRUE) #"#222255" #
      # SIB (day time)
      legend_items$col = c(legend_items$col, "steelblue4") #"#56B4E9") #"#D55E00""#E69F00") "#56B4E9"
      legend_items$name = c(legend_items$name, "sib")
      legend_items$code = c(legend_items$code, -1)
      legend_items$level = c(legend_items$level, 2)
      
      legend_items$col = c(legend_items$col, "steelblue1") #"blue3") # "#D55E00""#E69F00") "#56B4E9" "#0072B2"
      legend_items$name = c(legend_items$name, "nap")
      legend_items$code = c(legend_items$code, -1)
      legend_items$level = c(legend_items$level, 2)
      
      # Sleep in SPT
      legend_items = gen_col_names(behavioral_code_names,
                                   behavioral_codes = behavioral_codes,
                                   name = "spt_sleep", #sleep_in_spt
                                   legend_items = legend_items, colour = "white", #"#F0E442"
                                   level = 2)
      # Wake in SPT
      legend_items = gen_col_names(behavioral_code_names,
                                   behavioral_codes = behavioral_codes,
                                   name = "spt_wake", #sleep_in_spt
                                   legend_items = legend_items, colour = "yellow3",
                                   level = 2, reverse = FALSE)
      # Inactivity
      not_spt = grep(pattern = "spt", x = behavioral_code_names, invert = TRUE)
      legend_items = gen_col_names(behavioral_code_names = behavioral_code_names[not_spt],
                                   behavioral_codes = behavioral_codes[not_spt],
                                   name = "inactive",
                                   legend_items = legend_items, colour = "#CC79A7",
                                   level = 2, reverse = FALSE)
      # LIPA
      legend_items = gen_col_names(behavioral_code_names = behavioral_code_names[not_spt],
                                   behavioral_codes = behavioral_codes[not_spt],
                                   name = "lipa",
                                   legend_items = legend_items, colour = "#009E73",
                                   level = 2, reverse = FALSE)
      # MVPA
      legend_items = gen_col_names(behavioral_code_names = behavioral_code_names[not_spt],
                                   behavioral_codes = behavioral_codes[not_spt],
                                   name = "mod|vig|mvpa",
                                   legend_items = legend_items, colour = "#D55E00",
                                   level = 2, reverse = FALSE)
      # Invalid
      legend_items$col = c(legend_items$col, "grey") # ""
      legend_items$name = c(legend_items$name, "invalid")
      legend_items$code = c(legend_items$code, -1)
      legend_items$level = c(legend_items$level, 0)
      
      simple_filename = gsub(pattern = ".RData", "", x = fnames.ms5raw[i] )
      hrsPerRow = params_output[["visualreport_hrsPerRow"]]
      focus = params_output[["visualreport_focus"]]
      if (focus == "day") {
        dayedges = which(format(mdat$timestamp, "%H") == "00" &
                           format(mdat$timestamp, "%M") == "00" &
                           format(mdat$timestamp, "%S") == "00")
      } else if (focus == "night") {
        dayedges = which(format(mdat$timestamp, "%H") == "12" &
                           format(mdat$timestamp, "%M") == "00" &
                           format(mdat$timestamp, "%S") == "00")
      }
      # Identify indices for start and end of each window
      if (dayedges[1] == 1) {
        # recording starts at edge
        subploti = dayedges
        dayEnds = c(dayedges[2:length(dayedges)] + ((hrsPerRow - 24) * (3600/epochSize)) - 1, nrow(mdat))
      } else {
        # recording does not start at edge
        subploti = c(1, dayedges)
        dayEnds = c(dayedges + ((hrsPerRow - 24) * (3600/epochSize)) - 1, nrow(mdat))
        
      }
      subploti = cbind(subploti, dayEnds)
      invalid = which(mdat$invalidepoch == 1)
      subploti[which(subploti[,2] > nrow(mdat)), 2] = nrow(mdat)
      NdaysPerPage = 8
      skip_pdf_generation = FALSE
      if ( params_output[["visualreport_validcrit"]] > 0) {
        # Only keep windows that meet the fraction of valid data
        # as specified with parameter visualreport_validcrit
        for (si in 1:nrow(subploti)) {
          Nvalid = length(which(mdat$invalid[(subploti[si, 1] + 1):subploti[si, 2]] == 0))
          Ntotal =  length(mdat$invalid[(subploti[si, 1] + 1):subploti[si, 2]])
          frac_valid = Nvalid / Ntotal
          if (frac_valid < params_output[["visualreport_validcrit"]]) {
            subploti[si, ] = -1
          }
        }
        valid_rows = which(subploti[,1] > 0)
        if (length(valid_rows) > 0) {
          subploti = subploti[valid_rows, , drop = FALSE]
        }
        if (length(valid_rows) == 0 || inherits(x = subploti, what = "matrix") == FALSE || nrow(subploti) == 0) {
          message(paste0("Recording ", simple_filename, " skipped from visual",
                         " report generation because valid",
                         " data criteria (visualreport_validcrit) not met."))
          skip_pdf_generation = TRUE
        }
      }
      if (skip_pdf_generation == FALSE) {
        if (nrow(subploti) > 0) {
          # Generate pdf
          pdf(paste0(metadatadir, "/results/file summary reports/report_",
                     simple_filename, ".pdf"), paper = "a4",
              width = 0, height = 0)
          par(mfrow = c(NdaysPerPage, 1), mgp = c(2, 0.8, 0), omi = c(0, 0, 0, 0), bty = "n")
          cntplot = 0
          for (ani in 1:nrow(subploti)) { # loop across subplots
            if (cntplot >= 8) cntplot = 0
            if (cntplot == 0) {
              # pdf file header with file characteristics and legend
              par(mar = c(3, 0, 1.5, 2.5))
              plot.new()
              legendnames = legend_items$name
              legendnames[which(legendnames == "sib")] = "no movement (sib daytime)"
              legendnames[which(legendnames == "nap")] = "nap"
              legendnames[which(legendnames == "sleep_in_spt")] = "sleep"
              legendnames = gsub(pattern = "_", replacement = " ", x = legendnames)
              boutvars = grep(pattern = "bts", x = legendnames, value = FALSE)
              for (bvi in 1:length(boutvars)) {
                tmp_split = unlist(strsplit(legendnames[boutvars[bvi]], " "))
                if (length(tmp_split) == 3) {
                  legendnames[boutvars[bvi]] = paste0(tmp_split[1], " ", tmp_split[2],
                                                      " >=", tmp_split[3])
                } else {
                  legendnames[boutvars[bvi]] = paste0(tmp_split[1], " ",tmp_split[2],
                                                      " [", tmp_split[3], ",", tmp_split[4], ")")
                }
              }
              legendnames[boutvars] = paste0(legendnames[boutvars], " mins")
              legendcolors = legend_items$col
              
              not_invalid = which(legendnames != "invalid")
              legendpch = rep(15, length(legendnames))
              legendpch[which(legendnames == "invalid")] = 0
              legendcolors[which(legendnames == "invalid")] = "grey" #"white"
              legendnames[which(legendnames == "invalid")] = "ignored/imputed"
              legend("topright", legend = legendnames,
                     col = legendcolors,
                     ncol = length(legend_items$name) %/% 6 + 1, cex = 0.8,
                     pch = legendpch, pt.cex = 2, bty = "n", title = "Legend:",
                     title.font = 2, title.adj = 0)
              if (is.null(GGIRversion)) GGIRversion = "Not identified"
              
              RecDuration = round((nrow(mdat) * epochSize) / (24 * 3600), digits = 1)
              RecDurUnit = "days"
              if (RecDuration < 1) {
                RecDuration * 24
                RecDurUnit = "hours"
              }
              if (desiredtz == "") desiredtz = Sys.timezone()
              if (nchar(simple_filename) > 10) {
                added_text1 = "Start of filename: "
                added_text2 = "..."
              } else {
                added_text1 = "Filename: "
                added_text2 = ""
                
              }
              legend("topleft", legend = c(paste0(added_text1,
                                                  substr(x = simple_filename, start = 1, stop = 10),
                                                  added_text2),
                                           paste0("Start date: ", as.Date(mdat$timestamp[1])),
                                           paste0("Duration: ", RecDuration, " ", RecDurUnit),
                                           paste0("GGIR version: " , GGIRversion),
                                           paste0("Timezone: ", desiredtz)),
                     cex = 0.9,
                     bty = "n", title = "Info:",
                     title.font = 2, title.adj = 0)
              
              mtext(text = "Accelerometer time series report generated by R package GGIR",
                    side = 3, line = 0.2, cex = 1.2, font = 2)
              cntplot = cntplot + 1
            }
            
            if (subploti[ani, 2] - subploti[ani, 1] > 60 * (60/epochSize)) { # we need at least 1 hour for the ticks
              panelplot(mdat[(subploti[ani, 1] + 1):subploti[ani, 2], ],
                        ylabels_plot2, binary_vars,
                        behavioral_code_names, behavioral_codes, title = "",
                        hrsPerRow = hrsPerRow, plotid = ani,
                        legend_items = legend_items, lux_available = lux_available,
                        step_count_available = step_count_available, epochSize = epochSize,
                        focus = focus, temperature_available = temperature_available)
              cntplot = cntplot + 1
            }
          }
          while (!is.null(dev.list()))  dev.off()
        }
      }
    }
    
  } # end function
  
  # End of defining local functions
  #---------------------------------------
  
  # Attempt to load time series directly
  # Identify correct subfolder
  expected_ts_path = paste0(metadatadir, "/meta/ms5.outraw/", part6_threshold_combi)
  expected_ms5raw_path = paste0(metadatadir, "/meta/ms5.outraw")
  if (dir.exists(expected_ts_path)) {
    fnames.ms5raw = dir(expected_ts_path, pattern = "[.]RData")
    N_ts_files = length(fnames.ms5raw)
    if (f1 > N_ts_files) f1 = N_ts_files # this is intentionally ms3 and not ms4, do not change!
    if (f1 == 0) f1 = N_ts_files
    
    ts_exists = ifelse(length(fnames.ms5raw) > 0, yes = TRUE, no = FALSE) 
    mdat = NULL
    if (ts_exists) {
      # Extract behavioural class names and codes:
      legendfiles = list.files(path = paste0(metadatadir, "/meta/ms5.outraw"),
                               pattern = "codes", full.names = TRUE)
      df = file.info(legendfiles)
      legendfiles = rownames(df)[which.max(df$mtime)]
      legendF = data.table::fread(file = rownames(df)[which.max(df$mtime)], data.table = FALSE)
      behavioral_code_names = legendF$class_name # behavioural class names (characters)
      behavioral_codes = legendF$class_id # behavioural class codes (numeric)
      # reorder and rename behavioural class names and codes:
      neworder = c(grep("sleep", x = behavioral_code_names), grep("IN", x = behavioral_code_names),
                   grep("LIG", x = behavioral_code_names), grep("MOD", x = behavioral_code_names),
                   grep("VIG", x = behavioral_code_names), grep("MVPA", x = behavioral_code_names)) 
      behavioral_code_names = behavioral_code_names[neworder]
      behavioral_codes = behavioral_codes[neworder]
      behavioral_code_names = gsub("day_|spt_", "", x = behavioral_code_names)
      behavioral_code_names = gsub("sleep", "spt_sleep", x = behavioral_code_names)
      behavioral_code_names = gsub("wake_IN", "spt_wake_inactive", x = behavioral_code_names)
      behavioral_code_names = gsub("wake_LIG", "spt_wake_lipa", x = behavioral_code_names)
      behavioral_code_names = gsub("wake_MOD", "spt_wake_moderate", x = behavioral_code_names)
      behavioral_code_names = gsub("wake_VIG", "spt_wake_vigorous", x = behavioral_code_names)
      behavioral_code_names = tolower(behavioral_code_names)
      behavioral_code_names = gsub("lig_", "lipa_", x = behavioral_code_names)
      behavioral_code_names = gsub("in_bts", "inactive_bts", x = behavioral_code_names)
      behavioral_code_names = gsub("in_unbt", "inactive_unbt", x = behavioral_code_names)
      # move unbouted to the end for logical order
      neworder = c(grep(pattern = "unbt", x = behavioral_code_names, invert = TRUE),
                   grep(pattern = "unbt", x = behavioral_code_names)) 
      behavioral_code_names = behavioral_code_names[neworder]
      behavioral_codes = behavioral_codes[neworder]
      
      
      
      suitable_file_found = FALSE
      filei = f0
      while (suitable_file_found == FALSE && filei <= f1) {
        load(file = paste0(metadatadir, "/meta/ms5.outraw/",
                           part6_threshold_combi, "/", fnames.ms5raw[filei]))
        if (length(mdat) != 0 && nrow(mdat) != 0) {
          names(mdat)[which(names(mdat) == "sibdetection")] = "sib"
          names(mdat)[which(names(mdat) == "invalidepoch")] = "invalid"
          mdat$sib[which(mdat$SleepPeriodTime == 1)] = 0
          ylabels_plot2 = NULL
          # rename selfreported terminology
          if ("selfreported" %in% colnames(mdat) &&
              !is.null(params_sleep) &&
              !is.null(params_sleep[["loglocation"]])) {
            ylabels_plot2 = c("nap", "nonwear", "sleeplog", "bedlog")
          }
          binary_vars = c("SleepPeriodTime", "sibdetection", "invalidepoch")
          ylabels_plot2 = c(binary_vars, ylabels_plot2)
          ylabels_plot2 = gsub("invalidepoch", "invalid", x = ylabels_plot2)
          ylabels_plot2 = gsub("SleepPeriodTime", "spt", x = ylabels_plot2)
          ylabels_plot2 = gsub("sibdetection", "sib", x = ylabels_plot2)
          ylabels_plot2 = gsub("nap", "diary_nap", x = ylabels_plot2)
          ylabels_plot2 = gsub("nonwear", "diary_nonwear", x = ylabels_plot2)
          ylabels_plot2 = gsub("sleeplog", "diary_sleepwindow", x = ylabels_plot2)
          ylabels_plot2 = gsub("bedlog", "diary_timeinbed", x = ylabels_plot2)
          ylabels_plot2 = tolower(ylabels_plot2)
          suitable_file_found = TRUE
        }
        filei = filei + 1
      }
      
      # loop through files
      if (params_general[["do.parallel"]] == TRUE) {
        cores = parallel::detectCores()
        Ncores = cores[1]
        if (Ncores > 3) {
          if (length(params_general[["maxNcores"]]) == 0) params_general[["maxNcores"]] = Ncores
          Ncores2use = min(c(Ncores - 1, params_general[["maxNcores"]], (f1 - f0) + 1))
          if (Ncores2use > 1) {
            cl <- parallel::makeCluster(Ncores2use) # not to overload your computer
            parallel::clusterExport(cl = cl,
                                    varlist = c(unclass(lsf.str(envir = asNamespace("GGIR"), all = T)),
                                                "MONITOR", "FORMAT"),
                                    envir = as.environment(asNamespace("GGIR")))
            parallel::clusterEvalQ(cl, Sys.setlocale("LC_TIME", "C"))
            doParallel::registerDoParallel(cl)
          } else {
            # Don't process in parallel if only one core
            params_general[["do.parallel"]] = FALSE
          }
        } else {
          if (verbose == TRUE) cat(paste0("\nparallel processing not possible because number of available cores (",Ncores,") < 4"))
          params_general[["do.parallel"]] = FALSE
        }
      }
      if (params_general[["do.parallel"]] == TRUE) {
        if (verbose == TRUE) cat(paste0('\n Busy processing ... see ', metadatadir,'/results/file summary reports', ' for progress\n'))
        # check whether we are indevelopment mode:
        GGIRinstalled = is.element('GGIR', installed.packages()[,1])
        packages2passon = functions2passon = NULL
        GGIRloaded = "GGIR" %in% .packages()
        if (GGIRloaded) { #pass on package
          packages2passon = 'GGIR'
          errhand = 'pass'
        } else {
          # pass on functions
          functions2passon = c()
          errhand = 'stop'
        }
        i = 0 # declare i because foreach uses it, without declaring it
        `%myinfix%` = foreach::`%dopar%`
        output_list = foreach::foreach(i = f0:f1, .packages = packages2passon,
                                       .export = functions2passon, .errorhandling = errhand) %myinfix% {
                                         tryCatchResult = tryCatch({
                                           main_visualreport(i, f0, f1, metadatadir, 
                                                             part6_threshold_combi,
                                                             fnames.ms5raw, ylabels_plot2,
                                                             GGIRversion, desiredtz,
                                                             behavioral_code_names, behavioral_codes,
                                                             binary_vars)
                                         })
                                         return(tryCatchResult)
                                       }
        
        on.exit(parallel::stopCluster(cl))
        for (oli in 1:length(output_list)) { # logged error and warning messages
          if (is.null(unlist(output_list[oli])) == FALSE) {
            if (verbose == TRUE) cat(paste0("\nErrors and warnings for ", fnames[oli]))
            browser()
            print(unlist(output_list[oli])) # print any error and warnings observed
          }
        }
      } else {
        for (i in f0:f1) {
          if (verbose == TRUE) cat(paste0(i, " "))
          main_visualreport(i, f0, f1, metadatadir, part6_threshold_combi,
                            fnames.ms5raw, ylabels_plot2, GGIRversion, desiredtz,
                            behavioral_code_names, behavioral_codes,
                            binary_vars)
        }
      }
    }
  }
}

g.abr.day.names = function(daynames) {
  tmp =rep(0,length(daynames))
  daynamesnew =rep("",length(daynames))
  tmp[which(daynames=="Monday")] = 1
  tmp[which(daynames=="Tuesday")] = 2
  tmp[which(daynames=="Wednesday")] = 3
  tmp[which(daynames=="Thursday")] = 4
  tmp[which(daynames=="Friday")] = 5
  tmp[which(daynames=="Saturday")] = 6
  tmp[which(daynames=="Sunday")] = 7
  daynamesnew[which(tmp == 1)] = "MON"
  daynamesnew[which(tmp == 2)] = "TUE"
  daynamesnew[which(tmp == 3)] = "WED"
  daynamesnew[which(tmp == 4)] = "THU"
  daynamesnew[which(tmp == 5)] = "FRI"
  daynamesnew[which(tmp == 6)] = "SAT"
  daynamesnew[which(tmp == 7)] = "SUN"
  g.convert.day.names= daynamesnew
}

g.shell.GGIR = function(...) {
  message(paste0("Please note that instead of g.shell.GGIR() you can just write ",
                 "GGIR(), because all g.shell.GGIR() does is forward its input to ",
                 "function GGIR()."), call. = FALSE)
  .args <- as.list(match.call()[-1])
  do.call(GGIR, .args)
}

g.part5.onsetwaketiming = function(qqq, ts, min, sec, hour, timewindowi) {
  onset = wake = 0
  skiponset = TRUE; skipwake = TRUE
  # Onset index
  if (timewindowi == "OO") {
    # For OO onset is by definition the start and end of the window
    onseti = qqq[1] + 1
  } else {
    # For WW and MM onset needs to be search in the window
    onseti = c(qqq[1]:qqq[2])[which(diff(ts$diur[qqq[1]:(qqq[2] - 1)]) == 1)] + 1
    if (length(onseti) > 1) {
      onseti = onseti[length(onseti)] # in the case if MM use last onset
    }
  }
  # Wake index
  if (timewindowi == "WW") {
    # For WW wake is by definition the start and end of the window
    wakei = qqq[2] + 1
  } else {
    # For OO and MM wake needs to be search in the window
    wakei = c(qqq[1]:qqq[2])[which(diff(ts$diur[qqq[1]:(qqq[2] - 1)]) == -1) + 1]
    if (length(wakei) > 1) wakei = wakei[1] # in the case if MM use first wake-up time
  }
  # Onset time
  if (length(onseti) == 1) { # in MM window it is possible to not have an onset
    if (is.na(onseti) == FALSE) {
      onset = hour[onseti] + (min[onseti]/60) + (sec[onseti]/3600)
      skiponset = FALSE
    }
  }
  # Wake time
  if (length(wakei) == 1) { # in MM window it is possible to not have a wake
    if (is.na(wakei) == FALSE) {
      wake = hour[wakei] + (min[wakei]/60) + (sec[wakei]/3600)
      skipwake = FALSE
    }
  }
  if (wake >= 12 & wake <= 18) { # daysleeper and onset in the morning or afternoon
    if (onset <= 18 & skiponset == FALSE) onset = onset + 24
    if (wake <= 18 & skipwake == FALSE) {
      wake = wake + 24
    }
  } else if (wake <= 12) { # no daysleeper, but onset before noon
    if (wake <= 12 & skipwake == FALSE) wake = wake + 24
    if (onset <= 12 & skiponset == FALSE) onset = onset + 24
  }
  if (wake >= 12 & onset <= 12 & skiponset == FALSE) onset = onset + 24
  if (wake > 36 & onset > 36) {
    # both on the next afternoon is not possible,
    # so this was an overcorrection and reverse this:
    onset = onset - 24
    wake = wake - 24
  }
  return(invisible(
    list(
      wake = wake,
      onset = onset,
      wakei = wakei,
      onseti = onseti,
      skiponset = skiponset,
      skipwake = skipwake
    )
  ))
}

g.part5_initialise_ts = function(IMP, M, params_247, params_general, longitudinal_axis = c()) {
  # extract key variables from the mile-stone data: time, acceleration and elevation angle
  # note that this is imputed ACCELERATION because we use this for describing behaviour:
  scale = ifelse(test = grepl("^Brond|^Neishabouri|^ZC|^ExtAct|^ExtHeartRate", params_general[["acc.metric"]]), yes = 1, no = 1000)

  # Use anglez by default or longitudinal axis if specified when sensor is worn on hip
  
  if (params_general[["sensor.location"]] != "hip") {
    longitudinal_axis = NULL
  }
  if (is.null(longitudinal_axis) && "anglez" %in% names(IMP$metashort)) {
    angleName = "anglez"
  } else if (longitudinal_axis == 1 && "anglex" %in% names(IMP$metashort)) {
    angleName = "anglex"
  } else if (longitudinal_axis == 2 && "anglex" %in% names(IMP$metashort)) {
    angleName = "angley"
  } else if (longitudinal_axis == 3 && "anglex" %in% names(IMP$metashort)) {
    angleName = "anglez"
  } else {
    angleName = NULL
  }
  if (!is.null(angleName) && angleName %in% names(IMP$metashort)) {
    ts = data.frame(time = IMP$metashort[,1], ACC = IMP$metashort[,params_general[["acc.metric"]]] * scale,
                    guider = rep("unknown", nrow(IMP$metashort)),
                    angle = as.numeric(as.matrix(IMP$metashort[,which(names(IMP$metashort) == angleName)])))
    # also store other angles if available:
    for (otherAngle in c("anglex", "angley", "anglez")) {
      if (otherAngle %in% names(IMP$metashort) == TRUE && angleName != otherAngle) {
        if (otherAngle == "anglex") {
          ts$anglex = as.numeric(as.matrix(IMP$metashort[, otherAngle]))
        } else if (otherAngle == "angley") {
          ts$angley = as.numeric(as.matrix(IMP$metashort[, otherAngle]))
        } else if (otherAngle == "anglez") {
          ts$anglez = as.numeric(as.matrix(IMP$metashort[, otherAngle]))
        }
      }
    }
  } else {
    ts = data.frame(time = IMP$metashort[,1], ACC = IMP$metashort[,params_general[["acc.metric"]]] * scale,
                    guider = rep("unknown", nrow(IMP$metashort)))
  }
  if ("step_count" %in% colnames(IMP$metashort)) {
    ts$step_count = 0
    ts$step_count = IMP$metashort$step_count
  }
  
  Nts = nrow(ts)
  # add non-wear column
  nonwear = IMP$rout[,5]
  nonwear = rep(nonwear, each = (IMP$windowsizes[2]/IMP$windowsizes[1]))
  if (length(nonwear) > Nts) {
    nonwear = nonwear[1:Nts]
  } else if (length(nonwear) < Nts) {
    nonwear = c(nonwear, rep(0, (Nts - length(nonwear))))
  }
  ts$nonwear = 0 # initialise column
  ts$nonwear = nonwear
  
  # Add temperature and light, if present
  lightpeak_available = "lightpeak" %in% colnames(M$metalong)
  temperature_available = "temperaturemean" %in% colnames(M$metalong)
  repeatvalues = function(x, windowsizes, Nts) {
    x = rep(x, each = (windowsizes[2]/windowsizes[1]))
    if (length(x) > Nts) {
      x = x[1:Nts]
    } else if (length(x) < Nts) {
      x = c(x, rep(0, (Nts - length(x))))
    }
    return(x)
  }
  if (lightpeak_available == TRUE) {
    luz = M$metalong$lightpeak
    if (length(params_247[["LUX_cal_constant"]]) > 0 &
        length(params_247[["LUX_cal_exponent"]]) > 0) { # re-calibrate light
      luz = params_247[["LUX_cal_constant"]] * exp(params_247[["LUX_cal_exponent"]] * luz)
    }
    handle_luz_extremes = g.part5.handle_lux_extremes(luz)
    luz = handle_luz_extremes$lux
    correction_log = handle_luz_extremes$correction_log
    # repeate values to match resolution of other data
    luz = repeatvalues(x = luz, windowsizes = IMP$windowsizes, Nts)
    correction_log = repeatvalues(x = correction_log, windowsizes = IMP$windowsizes, Nts)
    ts$lightpeak_imputationcode = ts$lightpeak = 0 # initialise column
    ts$lightpeak = luz
    ts$lightpeak_imputationcode = correction_log
  }
  if (temperature_available == TRUE) {
    temperature = M$metalong$temperaturemean
    # repeate values to match resolution of other data
    ts$temperature = repeatvalues(x = temperature, windowsizes = IMP$windowsizes, Nts)
  }
  if ("marker" %in% colnames(M$metashort)) {
    ts$marker = NA
    ts$marker = IMP$metashort$marker
  }
  return(ts)
}

# NOTE ABOUT DOCUMENTATION:
# GGIR does not use Roxygen. The documentation below is not used. 
# All function documentation can be found in the man/*.Rd files.
# Please edit documentation there.
#
#- @title Activity balance index (ABI)
#- 
#- @description This function estimates the Activity balance index (ABI), which is a transformation of the self-similarity parameter (SSP), also known as scaling exponent or alpha.  
#- @param x the estimated self-similarity parameter (SSP)
#-
#- @details ABI = exp(-abs(SSP-1)/exp(-2))
#-
#- @return The estimated Activity balance index (ABI) is a real number between zero and one.
#-
#- @author Ian Meneghel Danilevicz  
#- 
#- @references C.-K. Peng, S.V. Buldyrev, S. Havlin, M. Simons, H.E. Stanley, A.L. Goldberger Phys. Rev. E, 49 (1994), p. 1685
#- Mesquita, Victor & Filho, Florencio & Rodrigues, Paulo. (2020). Detection of crossover points in detrended fluctuation analysis: An application to EEG signals of patients with epilepsy. Bioinformatics. 10.1093/bioinformatics/btaa955. 
#-
#- @examples 
#- # Estimate Activity balance index of a very known time series available on R base: the sunspot.year.
#- 
#- ssp = SSP(sunspot.year, scale = 1.2)
#- abi = ABI(ssp)
#-
ABI = function(x){
  if (is.na(x)) {
    y = NA
  } else {
    y = exp(-abs(x - 1) * exp(2))  
  }
  return(y)  
}

library(GGIR)
context("DFA, SSP and ABI")
test_that("DFA produces expected abi and ssp from sunspot.year dataset", {
  skip_on_cran()
  ssp = SSP(sunspot.year)
  abi = ABI(ssp)
  expect_equal(abi, 0.145756, tolerance = 0.0001)
  expect_equal(ssp, 0.7393684, tolerance = 0.0001)
})

test_that("DFA skips calculation if there are NA values", {
  skip_on_cran()
  sy_tmp = sunspot.year
  sy_tmp[5] = NA
  ssp = SSP(sy_tmp)
  abi = ABI(ssp)
  expect_equal(abi, NA)
  expect_equal(ssp, NA)
})

test_that("DFA can handle too short time series", {
  skip_on_cran()
  ssp = SSP(sunspot.year[1:2])
  abi = ABI(ssp)
  expect_equal(abi, NA)
  expect_equal(ssp, NA)
})

g.report.part5 = function(metadatadir = c(), f0 = c(), f1 = c(), loglocation = c(),
                          params_cleaning = NULL,
                          LUX_day_segments = c(), params_output,
                          verbose = TRUE) {
  # description: function to load milestone data generated by g.part5 and to merge these into a spreadsheet
  # here no additional information of analyses are added. This function therefore is 
  # primary to wrap up the output from parallel processed accelerometer files.
  
  if (params_cleaning[["includedaycrit.part5"]] > 1) {
    # I put this warning here, and not in check_params because that would generate
    # the warning every time the check_params is used across GGIR.
    warning(paste0("\nNote that the behaviour of parameter includedaycrit.part5 ",
                   "has changed for values above 1. These are now treated as the mimimum ",
                   "absolute number of valid hours during the waking hours of a day.",
                   "If you prefer to keep the old functionality then divide ",
                   "your current value (which is above 1) by 24."), call. = FALSE)
  }
  getValidDayIndices = function(x, window, params_cleaning) {
    if (window != "Segments") {
      if (params_cleaning[["includedaycrit.part5"]] >= 0 &
          params_cleaning[["includedaycrit.part5"]] <= 1) { # if includedaycrit.part5 is used as a ratio
        includeday_wearPercentage = params_cleaning[["includedaycrit.part5"]] * 100
        includeday_absolute = 0
      } else if (params_cleaning[["includedaycrit.part5"]] > 1 &
                 params_cleaning[["includedaycrit.part5"]] <= 25) { # if includedaycrit.part5 is used like params_cleaning[["includedaycrit"]] as a number of hours
        includeday_wearPercentage = 0
        includeday_absolute = params_cleaning[["includedaycrit.part5"]] * 60
      }
      if (params_cleaning[["includenightcrit.part5"]] >= 0 &
          params_cleaning[["includenightcrit.part5"]] <= 1) { # if includenightcrit.part5 is used as a ratio
        includenight_wearPercentage = params_cleaning[["includenightcrit.part5"]] * 100
        includenight_absolute = 0
      } else if (params_cleaning[["includenightcrit.part5"]] > 1 &
                 params_cleaning[["includenightcrit.part5"]] <= 25) { # if includenightcrit.part5 is used like params_cleaning[["includenightcrit"]] as a number of hours
        includenight_wearPercentage = 0
        includenight_absolute = params_cleaning[["includenightcrit.part5"]] * 60
      }
      include_window = rep(TRUE, nrow(x))
      if (length(params_cleaning[["data_cleaning_file"]]) > 0) { # allow for forced relying on guider based on external params_cleaning[["data_cleaning_file"]]
        DaCleanFile = data.table::fread(params_cleaning[["data_cleaning_file"]], data.table = FALSE)
        days2exclude = which(paste(x$ID, x$window_number) %in% paste(DaCleanFile$ID, DaCleanFile$day_part5))
        if (length(days2exclude) > 0) {
          include_window[days2exclude] = FALSE
        }
      } else {
        include_window = rep(TRUE,nrow(x))
      }
      # Below we intentionally only set a criteria on daytime, because for
      # the night time we only need start and end of the SPT window.
      
      # Add extra columns to ease the check
      x$nonwear_perc_day_spt = as.numeric(x$nonwear_perc_day_spt)
      x$nonwear_perc_day = as.numeric(x$nonwear_perc_day)
      x$nonwear_perc_spt = as.numeric(x$nonwear_perc_spt)
      x$wear_min_day_spt = (1 - (x$nonwear_perc_day_spt / 100)) * x$dur_day_spt_min #valid minute during waking hours
      
      x$wear_min_day = (1 - (x$nonwear_perc_day / 100)) * x$dur_day_min #valid minute during waking hours
      x$wear_perc_day = 100 - x$nonwear_perc_day #wear percentage during waking hours
      x$wear_min_spt = (1 - (x$nonwear_perc_spt / 100)) * x$dur_spt_min #valid minute during waking hours
      x$wear_perc_spt = 100 - x$nonwear_perc_spt #wear percentage during waking hours

      x$lastHour = as.numeric(x$lastHour)
      x$calendar_date = as.Date(x$calendar_date)

      minimumValidMinutesMM = 0 # default
      if (length(params_cleaning[["includedaycrit"]]) == 2) {
        minimumValidMinutesMM = params_cleaning[["includedaycrit"]][2] * 60
      }
      if (params_output[["require_complete_lastnight_part5"]] == FALSE) {
        x$lastWindow = FALSE
        x$lastDate = x$calendar_date
      } else {
        x$lastWindow = x$window_number == max(x$window_number)
        x$lastDate = as.Date(x$lastDate)
      }
      if (window == "WW" | window == "OO") {
        indices = which(x$wear_perc_day >= includeday_wearPercentage &
                          x$wear_min_day >= includeday_absolute &
                          x$wear_perc_spt >= includenight_wearPercentage &
                          x$wear_min_spt >= includenight_absolute &
                          x$dur_spt_min > 0 & x$dur_day_min > 0 &
                          ((x$lastWindow == TRUE & x$lastHour >= 15 & (x$lastDate - x$calendar_date) >= -1 & window == "WW") |
                             (x$lastWindow == TRUE & x$lastHour >= 9 & (x$lastDate - x$calendar_date) >= -1 & window == "OO") |
                             x$lastWindow == FALSE) &
                          include_window == TRUE &
                          x$wear_min_day_spt >= minimumValidMinutesMM)
      } else if (window == "MM") {
        indices = which(x$wear_perc_day >= includeday_wearPercentage &
                          x$wear_min_day >= includeday_absolute &
                          x$wear_perc_spt >= includenight_wearPercentage &
                          x$wear_min_spt >= includenight_absolute &
                          x$dur_spt_min > 0 & x$dur_day_min > 0 &
                          ((x$lastWindow == TRUE & x$lastHour > 9 & (x$lastDate - x$calendar_date) >= -1) |
                             x$lastWindow == FALSE) &
                          x$dur_day_spt_min >= (params_cleaning[["minimum_MM_length.part5"]] * 60) &
                          include_window == TRUE &
                          x$wear_min_day_spt >= minimumValidMinutesMM)
        # Note: By default for MM analysis only full days are interesting (23 hours for one day in the year)
      }
    } else if (window == "Segments") {
      # clean based on segments (even if a day is not valid, a certain segment of that
      # day could be valid if participant wore device enough in that part of the day)
      maxpernwday = 100 - (params_cleaning[["segmentWEARcrit.part5"]] * 100)
      include_window = rep(TRUE, nrow(x))
      indices = which(as.numeric(x$nonwear_perc_day_spt) <= maxpernwday &
                        as.numeric(x$dur_day_min) / as.numeric(x$dur_day_spt_min) >= params_cleaning[["segmentDAYSPTcrit.part5"]][1] &
                        as.numeric(x$dur_spt_min) / as.numeric(x$dur_day_spt_min) >= params_cleaning[["segmentDAYSPTcrit.part5"]][2] &
                        include_window == TRUE)
    }
    # exclude first and last window?
    if (params_cleaning[["excludefirstlast.part5"]] == TRUE) {
      x$window_number = as.numeric(x$window_number)
      # identify first and last day per file
      first_days = aggregate(window_number ~ filename, data = x, FUN = min, na.rm = TRUE)
      last_days = aggregate(window_number ~ filename, data = x, FUN = max, na.rm = TRUE)
      
      # match first and last days with the output dataframe
      exclude_firsts = which(paste(x$filename, x$window_number) %in% paste(first_days$filename, first_days$window_number))
      exclude_lasts = which(paste(x$filename, x$window_number) %in% paste(last_days$filename, last_days$window_number))
      
      # keep only indices that do not match with first and last days
      indices2exclude = which(indices %in% c(exclude_firsts, exclude_lasts))
      if (length(indices2exclude) > 0) indices = indices[-indices2exclude]
    }
    return(indices)
  }
  ms5.out = "/meta/ms5.out"
  if (file.exists(paste(metadatadir, ms5.out, sep = ""))) {
    if (length(dir(paste(metadatadir, ms5.out, sep = ""))) == 0) {
      try.generate.report = FALSE #do not run this function if there is no milestone data from g.part5
    } else {
      try.generate.report = TRUE
    }
  } else {
    try.generate.report = FALSE #do not run this function if there is no milestone data from g.part5
    warning(paste0('Cannot generate part5 report because no part 5 milestone data",
                   " is available. First run part 5 with argument mode=5.'))
  }
  if (try.generate.report == TRUE) {
    #======================================================================
    # loop through meta-files
    fnames.ms5 = list.files(paste0(metadatadir, ms5.out), full.names = TRUE)
    if (f1 > length(fnames.ms5)) {
      f1 = length(fnames.ms5)
    }
    if (verbose == TRUE) {
      cat(paste0(" loading all the milestone data from part 5 this can",
                 " take a few minutes\n"))
    }
    myfun = function(x, expectedCols = c()) {
      tail_expansion_log = last_timestamp = output = NULL
      load(file = x)
      cut = which(output[, 1] == "")
      if (length(cut) > 0 & length(cut) < nrow(output)) {
        output = output[-cut, which(colnames(output) != "")]
      }
      if (exists("last_timestamp") == TRUE) {
        output$lastHour = as.numeric(format(last_timestamp, "%H"))
        output$lastDate = as.Date(last_timestamp)
      } else {
        output$lastHour = Inf # use dummy value
        output$lastDate = Inf # use dummy value
      }
      out = as.matrix(output)
      if (length(expectedCols) > 0) {
        tmp = as.data.frame(matrix(0, 0, length(expectedCols)))
        colnames(tmp) = expectedCols
        out = base::merge(tmp, out, all = TRUE)
      }
      if (!is.null(expectedCols) && ncol(out) > length(expectedCols)) {
        warning(paste0("Columns dropped in output part5 for ",
                       basename(x),
                       " because these could not be matched to columns for earlier",
                       " recordings in this dataset. Please check whether these recordings ",
                       " were processed with a",
                       " different GGIR version or configuration. If yes, reprocess",
                       " consistently. If not, ",
                       " consider renaming this file such that it is",
                       " alphabetically first and by that processed first, which",
                       " should address the issue."), call. = FALSE)
      }
      if (length(tail_expansion_log) != 0) {
        col2na = grep(pattern = paste0("sleep_efficiency|N_atleast5minwakenight|daysleeper|",
                                       "daysleeper|sleeplog_used|_spt_sleep|_spt_wake"),
                      x = names(out), value = FALSE)
        window_number = as.numeric(out[,"window_number"])
        lastwindow_indices = which(window_number == max(window_number, na.rm = TRUE))
        if (length(col2na) > 0 & length(lastwindow_indices) > 0) {
          out[lastwindow_indices, col2na] = "" # set last row to NA for all sleep related variables
        }
      }
      return(out)
    }
    # It can be that first part 5 milestone file has less columns when it 
    # was processed with a different GGIR version or configuration.
    # This would then complicate
    # appending later milestone files to it. So, check first a random sample for highest number
    # of columns. If this fails to identify a file with the maximum number of
    # column then the above function will provide a warning to
    # inform the user and with instructions on how to address it.
    expectedCols = NULL
    set.seed(1234)
    # Create list of file indices to try
    testfiles = unique(sample(x = f0:f1, size = pmin(5, f1 - f0 + 1)))
    for (testf in testfiles) {
      out_try = myfun(fnames.ms5[testf])
      if (ncol(out_try) > length(expectedCols)) {
        # expectedCols should equal the column names of the milestone
        # file with the largest number of columns
        expectedCols = colnames(out_try) 
      }
    }
    # Now load and append all files
    outputfinal = as.data.frame(do.call(rbind,
                                        lapply(fnames.ms5[f0:f1], myfun, expectedCols)),
                                stringsAsFactors = FALSE)
    
    # Find columns filled with missing values
    cut = which(sapply(outputfinal, function(x) all(x == "")) == TRUE)
    if (length(cut) > 0) {
      outputfinal = outputfinal[,-cut]
    }
    
    # revise filename
    outputfinal$filename = gsub(".RData$", "", outputfinal$filename)

    # order data.frame
    outputfinal$window_number = as.numeric(gsub(" ", "", outputfinal$window_number))
    outputfinal = outputfinal[order(outputfinal$filename, outputfinal$window_number, outputfinal$window), ]

    # split results to different spreadsheets in order to minimize individual
    # filesize and to ease organising dataset
    uwi = as.character(unique(outputfinal$window))
    if (!all(uwi %in% c("MM", "WW", "OO"))) {
      uwi = c(uwi[uwi %in% c("MM", "WW", "OO")], "Segments")
    }
    uTRLi = as.character(unique(outputfinal$TRLi))
    uTRMi = as.character(unique(outputfinal$TRMi))
    uTRVi = as.character(unique(outputfinal$TRVi))
    usleepparam = as.character(unique(outputfinal$sleepparam))
    # replace NaN by empty cell value
    for (kra in 1:ncol(outputfinal)) {
      krad = which(is.nan(outputfinal[,kra]) == TRUE)
      if (length(krad) > 0) {
        outputfinal[krad,kra] = ""
      }
    }
    outputfinal$daytype = 0
    outputfinal$daytype[which(outputfinal$weekday == "Sunday" |
                                outputfinal$weekday == "Saturday")] = "WE"
    outputfinal$daytype[which(outputfinal$weekday == "Monday" |
                                outputfinal$weekday == "Tuesday" |
                                outputfinal$weekday == "Wednesday" |
                                outputfinal$weekday == "Thursday" |
                                outputfinal$weekday == "Friday")] = "WD"
    outputfinal$nonwear_perc_day = as.numeric(outputfinal$nonwear_perc_day)
    outputfinal$nonwear_perc_spt = as.numeric(outputfinal$nonwear_perc_spt)
    outputfinal$dur_spt_min = as.numeric(outputfinal$dur_spt_min)
    outputfinal$dur_day_min = as.numeric(outputfinal$dur_day_min)
    outputfinal$guider = as.character(outputfinal$guider)
    outputfinal$sleeplog_used = as.numeric(outputfinal$sleeplog_used)
    outputfinal$dur_spt_min = as.numeric(outputfinal$dur_spt_min)
    outputfinal$dur_day_min = as.numeric(outputfinal$dur_day_min)
    outputfinal$dur_day_spt_min = as.numeric(outputfinal$dur_day_spt_min)
    # loop to store various variants of the analysis separately
    outputfinal_bu = outputfinal
    if (verbose == TRUE) cat(" generating csv report for every parameter configurations...\n")
    for (j in 1:length(uwi)) {
      for (h1 in 1:length(uTRLi)) {
        for (h2 in 1:length(uTRMi)) {
          for (h3 in 1:length(uTRVi)) {
            for (h4 in 1:length(usleepparam)) {
              if (verbose == TRUE) {
                cat(paste0(" ", uwi[j], "-", uTRLi[h1], "-", uTRMi[h2],
                           "-", uTRVi[h3], "-", usleepparam[h4]))
              }
              select_window = as.character(outputfinal$window) == uwi[j]
              if (!(uwi[j] %in% c("MM", "WW", "OO"))) select_window = !(as.character(outputfinal$window) %in% c("MM", "WW", "OO"))
              seluwi = which(select_window &
                               as.character(outputfinal$TRLi) == uTRLi[h1] &
                               as.character(outputfinal$TRMi) == uTRMi[h2] &
                               as.character(outputfinal$TRVi) == uTRVi[h3] &
                               as.character(outputfinal$sleepparam) == usleepparam[h4])
              # store spreadsheet
              if (nrow(outputfinal[seluwi,]) == 0) {
                if (verbose == TRUE) cat("report not stored, because no results available")
              } else {
                CN = colnames(outputfinal)
                outputfinal2 = outputfinal
                colnames(outputfinal2) = CN
                delcol = grep(pattern = "TRLi|TRMi|TRVi|sleepparam",
                              x = colnames(outputfinal2))
                if (uwi[j] != "Segments") {
                  delcol = c(delcol, which(colnames(outputfinal2) == "window"))
                }
                outputfinal2 = outputfinal2[,-delcol]
                OF3 = outputfinal2[seluwi,]
                OF3 = as.data.frame(OF3, stringsAsFactors = TRUE)
                #-------------------------------------------------------------
                # store all summaries in csv files without cleaning criteria
                OF3_clean = tidyup_df(OF3)
                OF3_clean = addSplitNames(OF3_clean) # If recording was split
                data.table::fwrite(
                  OF3_clean,
                  paste(
                    metadatadir,
                    "/results/QC/part5_daysummary_full_",
                    uwi[j], "_L", uTRLi[h1], "M", uTRMi[h2], "V", uTRVi[h3],
                    "_", usleepparam[h4], ".csv", sep = ""), row.names = FALSE, na = "",
                  sep = params_output[["sep_reports"]],
                  dec = params_output[["dec_reports"]])
                # store all summaries in csv files with cleaning criteria
                validdaysi = getValidDayIndices(x = OF3_clean, window = uwi[j],
                                                params_cleaning = params_cleaning)
                if ("lastHour" %in% colnames(OF3_clean)) {
                  OF3_clean = OF3_clean[, -which(colnames(OF3_clean) %in% c("lastHour", "lastDate"))]
                }
                if (length(validdaysi) > 0) {
                  data.table::fwrite(
                    OF3_clean[validdaysi, ],
                    paste(metadatadir, "/results/part5_daysummary_",
                          uwi[j], "_L", uTRLi[h1], "M", uTRMi[h2], "V", uTRVi[h3],
                          "_", usleepparam[h4], ".csv", sep = ""), row.names = FALSE, na = "",
                    sep = params_output[["sep_reports"]],
                    dec = params_output[["dec_reports"]])
                }
                #------------------------------------------------------------------------------------
                #also compute summary per person
                agg_plainNweighted = function(df,
                                              filename = "filename",
                                              daytype = "daytype",
                                              window = "MM") {
                  # function to take both the weighted (by weekday/weekendday) and plain average of all numeric variables
                  # df: input data.frame (OF3 outside this function)
                  
                  ignorevar = c("daysleeper", "cleaningcode", "night_number", "sleeplog_used",
                                "ID", "acc_available", "window_number",
                                "boutcriter.mvpa", "boutcriter.lig", "boutcriter.in", "bout.metric") # skip cosinor variables
                  for (ee in 1:ncol(df)) { # make sure that numeric columns have class numeric
                    nr = nrow(df)
                    if (nr > 30) nr = 30
                    options(warn = -1)
                    trynum = as.numeric(as.character(df[1:nr, ee]))
                    options(warn = 0)
                    if (length(which(is.na(trynum) == TRUE)) != nr &
                        length(which(ignorevar == names(df)[ee])) == 0) {
                      options(warn = -1)
                      class(df[, ee]) = "numeric"
                      options(warn = 0)
                    }
                  }
                  plain_mean = function(x) {
                    options(warn = -1)
                    plain_mean = mean(x, na.rm = TRUE)
                    options(warn = 0)
                    if (is.na(plain_mean) == TRUE) {
                      plain_mean = x[1]
                    }
                    return(plain_mean)
                  }
                  # aggregate across all days
                  if (window == "Segments") {
                    by = list(df$filename, df$window)
                  } else {
                    by = list(df$filename)
                  }
                  PlainAggregate = aggregate.data.frame(df, by = by, FUN = plain_mean)
                  PlainAggregate = PlainAggregate[, -grep("^Group", colnames(PlainAggregate))]
                  # aggregate per day type (weekday or weekenddays)
                  if (window == "Segments") {
                    by = list(df$filename, df$window, df$daytype)
                  } else {
                    by = list(df$filename, df$daytype)
                  }
                  AggregateWDWE = aggregate.data.frame(df, by = by, plain_mean)
                  AggregateWDWE = AggregateWDWE[, -grep("^Group", colnames(AggregateWDWE))]
                  # Add counted number of days for Gini, Cov, alpha Fragmentation variables, because 
                  # days are dropped if there are not enough fragments:
                  vars_with_mininum_Nfrag = c("FRAG_Gini_dur_PA_day", "FRAG_CoV_dur_PA_day",
                                              "FRAG_alpha_dur_PA_day", "FRAG_Gini_dur_IN_day",
                                              "FRAG_CoV_dur_IN_day")
                  vars_with_mininum_Nfrag_i = which(vars_with_mininum_Nfrag %in% colnames(df) == TRUE)
                  if (length(vars_with_mininum_Nfrag_i) > 0) {
                    varname_minfrag = vars_with_mininum_Nfrag[vars_with_mininum_Nfrag_i[1]]
                    DAYCOUNT_Frag_Multiclass = aggregate.data.frame(
                      df[, varname_minfrag],
                      by = by,
                      FUN = function(x) length(which(is.na(x) == FALSE))
                    )
                    if (window != "Segments") {
                      colnames(DAYCOUNT_Frag_Multiclass)[1:2] = c("filename","daytype")
                      # AL10F, abbreviation for: at least 10 fragments
                      colnames(DAYCOUNT_Frag_Multiclass)[3] = "Nvaliddays_AL10F" 
                      by.x = c("filename", "daytype")
                    } else if (window == "Segments") {
                      colnames(DAYCOUNT_Frag_Multiclass)[1:3] = c("filename","window", "daytype")
                      # AL10F, abbreviation for: at least 10 fragments
                      colnames(DAYCOUNT_Frag_Multiclass)[4] = "Nvalidsegments_AL10F"
                      by.x = c("filename","window", "daytype")
                    }
                    AggregateWDWE = merge(AggregateWDWE, DAYCOUNT_Frag_Multiclass, by.x = by.x)
                  }
                  len = NULL
                  AggregateWDWE$len <- 0
                  AggregateWDWE$len[which(as.character(AggregateWDWE$daytype) == "WD")] = 5 #weighting of weekdays
                  AggregateWDWE$len[which(as.character(AggregateWDWE$daytype) == "WE")] = 2 #weighting of weekend days
                  if (window == "Segments") {
                    dt <- data.table::as.data.table(AggregateWDWE[,which(lapply(AggregateWDWE, class) == "numeric" | 
                                                                           names(AggregateWDWE) == filename |
                                                                           names(AggregateWDWE) == "window")])
                  } else {
                    dt <- data.table::as.data.table(AggregateWDWE[,which(lapply(AggregateWDWE, class) == "numeric" | 
                                                                           names(AggregateWDWE) == filename)])
                  }
                  
                  options(warn = -1)
                  .SD <- .N <- count <- a <- NULL
                  if (window == "Segments") {
                    WeightedAggregate <- dt[, lapply(.SD, weighted.mean, w = len, na.rm = TRUE), by = list(filename, window)]
                  } else {
                    WeightedAggregate <- dt[, lapply(.SD, weighted.mean, w = len, na.rm = TRUE), by = list(filename)]
                  }
                  options(warn = 0)
                  LUXmetrics = c("above1000", "timeawake", "mean", "imputed", "ignored")
                  add_missing_LUX = function(x, LUX_day_segments, weeksegment = c(), LUXmetrics) {
                    # missing columns, add these:
                    NLUXseg = length(LUX_day_segments)
                    if (length(weeksegment) > 0) {
                      LUX_segment_vars_expected = paste0("LUX_", LUXmetrics, "_",
                                                         LUX_day_segments[1:(NLUXseg - 1)],
                                                         "-", LUX_day_segments[2:(NLUXseg)],
                                                         "hr_day_", weeksegment)
                    } else {
                      luxvars = paste0("LUX_", LUXmetrics, "_")
                      segments = paste0(LUX_day_segments[1:(NLUXseg - 1)], 
                                        "_", LUX_day_segments[2:(NLUXseg)], "hr_day")
                      LUX_segment_vars_expected = c()
                      for (luxi in 1:length(luxvars)) {
                        LUX_segment_vars_expected = c(LUX_segment_vars_expected,
                                                      paste0(luxvars[luxi], segments))
                      }
                    }
                    dummy_df = as.data.frame(matrix(NaN, 1, length(LUX_segment_vars_expected)))
                    colnames(dummy_df) = LUX_segment_vars_expected
                    if (length(which(LUX_segment_vars_expected %in% colnames(x))) > 0) {
                      x = as.data.frame(merge(x, dummy_df, all.x = T))
                      # re-order
                      current_location = which(colnames(x) %in% LUX_segment_vars_expected == TRUE)
                      neworder = sort(colnames(x)[current_location])
                      x = cbind(x[,-current_location], x[,LUX_segment_vars_expected])
                    }
                    return(x)
                  }
                  LUX_segment_vars = c()
                  for (li in 1:length(LUXmetrics)) {
                    LUX_segment_vars = c(LUX_segment_vars, grep(
                      pattern = paste0("LUX_", LUXmetrics[li]),
                      x = colnames(WeightedAggregate),
                      value = TRUE
                    ))
                  }
                  if (length(LUX_segment_vars) > 0 &
                      length(LUX_segment_vars) < 24 &
                      length(LUX_day_segments) > 0) {
                    WeightedAggregate = add_missing_LUX(
                      WeightedAggregate,
                      LUX_day_segments,
                      weeksegment = c(),
                      LUXmetrics = LUXmetrics
                    )
                  }
                  # merge them into one output data.frame (G)
                  LUX_segment_vars = c()
                  for (li in 1:length(LUXmetrics)) {
                    LUX_segment_vars = colnames(PlainAggregate) %in% grep(
                      x = colnames(PlainAggregate),
                      pattern = paste0("LUX_", LUXmetrics[li]),
                      value = TRUE
                    )
                  }
                  charcol = which(
                    lapply(PlainAggregate, class) != "numeric" &
                      names(PlainAggregate) != filename & !(LUX_segment_vars)
                  )
                  numcol = which(lapply(PlainAggregate, class) == "numeric" | LUX_segment_vars)
                  WeightedAggregate = as.data.frame(WeightedAggregate, stringsAsFactors = TRUE)
                  
                  if (window == "Segments") {
                    by = c("filename", "window")
                  } else {
                    by = "filename"
                  }
                  G = base::merge(PlainAggregate,
                                  WeightedAggregate,
                                  by = by,
                                  all.x = TRUE)
                  p0b = paste0(names(PlainAggregate[, charcol]), ".x")
                  p1 = paste0(names(PlainAggregate[, numcol]), ".x")
                  p2 = paste0(names(PlainAggregate[, numcol]), ".y")
                  for (i in 1:length(p0b)) {
                    names(G)[which(names(G) == p0b[i])] = paste0(names(PlainAggregate[, charcol])[i])
                  }
                  for (i in 1:length(p1)) {
                    names(G)[which(names(G) == p1[i])] = paste0(names(PlainAggregate[, numcol])[i], "_pla")
                  }
                  for (i in 1:length(p2)) {
                    names(G)[which(names(G) == p2[i])] = paste0(names(PlainAggregate[, numcol])[i], "_wei")
                  }
                  # expand output with weekday (WD) and weekend (WE) day aggregates
                  if (params_output[["week_weekend_aggregate.part5"]] == TRUE) {
                    for (weeksegment in c("WD", "WE")) {
                      temp_aggregate = AggregateWDWE[which(AggregateWDWE$daytype == weeksegment), ]
                      charcol = which(lapply(temp_aggregate, class) != "numeric" &
                                        names(temp_aggregate) != filename)
                      numcol = which(lapply(temp_aggregate, class) %in% c("numeric", "integer") == TRUE)
                      names(temp_aggregate)[numcol] = paste0(names(temp_aggregate)[numcol], "_", weeksegment)
                      if (window == "Segments") {
                        temp_aggregate = temp_aggregate[, c(which(colnames(temp_aggregate) == "filename" |
                                                                    colnames(temp_aggregate) == "window"), numcol)]
                      } else {
                        temp_aggregate = temp_aggregate[, c(which(colnames(temp_aggregate) == "filename"), numcol)]
                      }
                      LUX_segment_vars = c()
                      for (li in 1:length(LUXmetrics)) {
                        LUX_segment_vars = grep(
                          pattern = paste0("LUX_", LUXmetrics[li]),
                          x = colnames(temp_aggregate),
                          value = TRUE
                        )
                      }
                      if (length(LUX_segment_vars) > 0 &
                          length(LUX_segment_vars) < 24 & length(LUX_day_segments) > 0) {
                        temp_aggregate = add_missing_LUX(temp_aggregate, LUX_day_segments, weeksegment, LUXmetrics)
                      }
                      G = base::merge(G, temp_aggregate,
                                      by = by, all.x = TRUE)
                    }
                  }
                  G = G[,-which(names(G) %in% c("len", "daytype", "len_WE", "len_WD"))]
                  return(G)
                }
                #---------------------------------------------
                # Calculate, weighted and plain mean of all variables
                # add column to define what are weekenddays and weekdays as needed for function agg_plainNweighted
                # before processing OF3, first identify which days have enough monitor wear time
                
                validdaysi = getValidDayIndices(x = OF3, window = uwi[j],
                                                params_cleaning = params_cleaning)
                if (length(validdaysi) > 0) { # do not attempt to aggregate if there are no valid days
                  # aggregate OF3 (days) to person summaries in OF4
                  OF4 = agg_plainNweighted(df = OF3[validdaysi,], filename = "filename", 
                                           daytype = "daytype", window = uwi[j])
                  # calculate additional variables
                  columns2keep = c("filename", "night_number", "daysleeper",
                                   "cleaningcode","sleeplog_used","guider",
                                   "acc_available", "nonwear_perc_day", "nonwear_perc_spt",
                                   "daytype", "dur_day_min",
                                   "dur_spt_min")
                  if (uwi[j] == "Segments") {
                    columns2keep = c(columns2keep, "window")
                  }
                  OF3tmp = OF3[, columns2keep]
                  foo34 = function(df,aggPerIndividual,nameold,namenew,cval,window) {
                    # function to help with calculating additional variables
                    # related to counting how many days of measurement there are
                    # that meet a certain criteria
                    # cval is a vector with 0 and 1, indicating whether the criteria is met
                    # aggPerIndividual is the aggregate data (per individual)
                    # df is the non-aggregated data (days across individuals
                    # we want to extra the number of days per individuals that meet the
                    # criteria in df, and make it allign with aggPerIndividual.
                    df2 = function(x) df2 = length(which(x == cval)) # check which values meets criterion
                    if (window == "Segments") by = list(df$filename, df$window)
                    if (window != "Segments") by = list(df$filename)
                    mmm = as.data.frame(aggregate.data.frame(df, by = by, FUN = df2),
                                        stringsAsFactors = TRUE)
                    mmm2 = data.frame(
                      filename = mmm$Group.1,
                      cc = mmm[, nameold],
                      stringsAsFactors = TRUE
                    )
                    if (window == "Segments") {
                      mmm2$window = mmm$Group.2
                      by = c("filename", "window")
                    } else if (window != "Segments") {
                      by = "filename"
                    }
                    aggPerIndividual = merge(aggPerIndividual, mmm2, by = by)
                    names(aggPerIndividual)[which(names(aggPerIndividual) == "cc")] = namenew
                    foo34 = aggPerIndividual
                  }
                  # # calculate number of valid days (both night and day criteria met)
                  OF3tmp$validdays = 0
                  # criteria is that nonwear percentage needs to be below threshold for both day and night:
                  OF3tmp$validdays[validdaysi] = 1
                  # now we have a label for the valid days, we can create a new variable
                  # in OF4 that is a count of the number of valid days:
                  namenew = "Nvaliddays"
                  if (uwi[j] == "Segments") namenew = "Nvalidsegments"
                  OF4 = foo34(
                    df = OF3tmp,
                    aggPerIndividual = OF4,
                    nameold = "validdays",
                    namenew = namenew,
                    cval = 1,
                    window = uwi[j]
                  )
                  # do the same for WE (weekend days):
                  OF3tmp$validdays = 0
                  OF3tmp$validdays[validdaysi[which(OF3tmp$daytype[validdaysi] == "WE")]] = 1
                  namenew = "Nvaliddays_WE"
                  if (uwi[j] == "Segments") namenew = "Nvalidsegments_WE"
                  OF4 = foo34(
                    df = OF3tmp,
                    aggPerIndividual = OF4,
                    nameold = "validdays",
                    namenew = namenew,
                    cval = 1,
                    window = uwi[j]
                  )
                  # do the same for WD (weekdays):
                  OF3tmp$validdays = 0
                  OF3tmp$validdays[validdaysi[which(OF3tmp$daytype[validdaysi] == "WD")]] = 1
                  namenew = "Nvaliddays_WD"
                  if (uwi[j] == "Segments") namenew = "Nvalidsegments_WD"
                  OF4 = foo34(df = OF3tmp, 
                              aggPerIndividual = OF4, 
                              nameold = "validdays", 
                              namenew = namenew, 
                              cval = 1, 
                              window = uwi[j]) # create variable from it
                  # do the same for daysleeper,cleaningcode, sleeplog_used, acc_available:
                  OF3tmp$validdays = 1
                  # redefine by considering only valid days
                  OF4 = foo34(
                    df = OF3tmp[validdaysi,],
                    aggPerIndividual = OF4,
                    nameold = "daysleeper",
                    namenew = "Ndaysleeper",
                    cval = 1,
                    window = uwi[j]
                  )
                  OF4 = foo34(
                    df = OF3tmp[validdaysi,],
                    aggPerIndividual = OF4,
                    nameold = "cleaningcode",
                    namenew = "Ncleaningcodezero",
                    cval = 0,
                    window = uwi[j]
                  )
                  for (ccode in 1:6) {
                    OF4 = foo34(
                      df = OF3tmp[validdaysi, ],
                      aggPerIndividual = OF4,
                      nameold = "cleaningcode",
                      namenew = paste0("Ncleaningcode", ccode),
                      cval = ccode,
                      window = uwi[j]
                    )
                  }
                  OF4 = foo34(
                    df = OF3tmp[validdaysi, ],
                    aggPerIndividual = OF4,
                    nameold = "sleeplog_used",
                    namenew = "Nsleeplog_used",
                    cval = TRUE,
                    window = uwi[j]
                  )
                  OF4 = foo34(
                    df = OF3tmp[validdaysi, ],
                    aggPerIndividual = OF4,
                    nameold = "acc_available",
                    namenew = "Nacc_available",
                    cval = 1,
                    window = uwi[j]
                  )
                  if ("markerbutton" %in% OF3$guider) {
                    OF3tmp$markerbutton_guider = 0
                    OF3tmp$markerbutton_guider[which(OF3tmp$guider[validdaysi] == "markerbutton")] = 1
                    OF4 = foo34(
                      df = OF3tmp[validdaysi, ],
                      aggPerIndividual = OF4,
                      nameold = "markerbutton_guider",
                      namenew = "Nmarkerbutton_used",
                      cval = 1,
                      window = uwi[j])
                  }
                  # Move valid day count variables to beginning of dataframe
                  OF4 = cbind(OF4[, 1:5],
                              OF4[, (ncol(OF4) - 10):ncol(OF4)],
                              OF4[, 6:(ncol(OF4) - 11)])
                  nom = names(OF4)
                  cut = which(nom == "sleeponset_ts" | nom == "wakeup_ts"
                              | nom == "night_number"  | nom == "window_number"
                              | nom == "daysleeper" | nom == "cleaningcode"
                              | nom == "acc_available"
                              | nom == "guider" | nom == "L5TIME" | nom == "M5TIME"
                              | nom == "L10TIME" | nom == "M10TIME"
                              | nom == "acc_available" | nom == "daytype"
                              | nom %in% paste0("sleeplog_used_", c("pla", "wei", "WD", "WE"))
                              | nom %in% paste0("window_number_", c("pla", "wei", "WD", "WE")))
                  names(OF4)[which(names(OF4) == "weekday")] = "startday"
                  OF4 = OF4[,-cut]
                  for (col4 in 1:ncol(OF4)) {
                    navalues = which(is.na(OF4[,col4]) == TRUE)
                    if (length(navalues) > 0) {
                      OF4[navalues, col4] = ""
                    }
                  }
                  #Move Nvaliddays variables to the front of the spreadsheet
                  Nvaliddays_variables = grep(x = colnames(OF4), pattern = "Nvaliddays", value = FALSE)
                  Nvaliddays_variables = unique(c(
                    which(colnames(OF4) == "Nvaliddays"),
                    which(colnames(OF4) == "Nvaliddays_WD"),
                    which(colnames(OF4) == "Nvaliddays_WE"),
                    Nvaliddays_variables
                  ))
                  OF4 = OF4[, unique(c(1:4, Nvaliddays_variables, 5:ncol(OF4)))]
                  #-------------------------------------------------------------
                  # store all summaries in csv files
                  OF4_clean = tidyup_df(OF4)
                  OF4_clean = addSplitNames(OF4_clean)  # If recording was split
                  # remove helper variables lastHour and lastdate, these are not relevant for end-user
                  lastHour_lastdate = grep(pattern = "lastHour|lastDate", x = colnames(OF4_clean))
                  if (length(lastHour_lastdate) > 0) {
                    OF4_clean = OF4_clean[, -lastHour_lastdate]
                  }
                  data.table::fwrite(OF4_clean,paste(metadatadir,"/results/part5_personsummary_",
                                                     uwi[j],"_L",uTRLi[h1],"M",
                                                     uTRMi[h2], "V", uTRVi[h3], "_",
                                                     usleepparam[h4], ".csv", sep = ""), 
                                     row.names = FALSE, na = "",
                                     sep = params_output[["sep_reports"]],
                                     dec = params_output[["dec_reports"]])
                }
              }
            }
          }
        }
      }
    }
    rm(outputfinal, outputfinal2)
  }
}

isfilelist = function(datadir) {
  filelist = FALSE
  #   verify whether datadir is a directory or a list of files
  if (length(datadir) == 1) { #could be a directory or one file
    if (length(unlist(strsplit(datadir,"[.]bi"))) > 1) filelist = TRUE
    if (length(unlist(strsplit(datadir,"[.]cs"))) > 1) filelist = TRUE
    if (length(unlist(strsplit(datadir,"[.]wa"))) > 1) filelist = TRUE
    if (length(unlist(strsplit(datadir,"[.]cw"))) > 1) filelist = TRUE #XD added this line to make it possible to process single cwa file
    if (length(unlist(strsplit(datadir,"[.]gt"))) > 1) filelist = TRUE #XD added this line to make it possible to process single cwa file
    if (length(unlist(strsplit(datadir,"[.]BI"))) > 1) filelist = TRUE #added to process BIN files (Parmay Matrix devices)
  } else { #multiple files
    filelist = TRUE    
  }
  return(filelist)
}

g.part5 = function(datadir = c(), metadatadir = c(), f0=c(), f1=c(),
                   params_sleep = c(), params_metrics = c(),
                   params_247 = c(), params_phyact = c(),
                   params_cleaning = c(), params_output = c(),
                   params_general = c(), verbose = TRUE, ...) {
  options(encoding = "UTF-8")
  filename_dir = NULL
  # This function called by function GGIR
  # and aims to combine all the milestone output from the previous parts
  # in order to facilitate a varierty of analysis on time-use, interactions
  # between day and night activity, and circadian rhythms
  #----------------------------------------------------------
  # Extract and check parameters
  input = list(...)
  params = extract_params(params_sleep = params_sleep,
                          params_metrics = params_metrics,
                          params_247 = params_247,
                          params_phyact = params_phyact,
                          params_cleaning = params_cleaning,
                          params_output = params_output,
                          params_general = params_general,
                          input = input, params2check = c("sleep", "metrics", "247",
                                                          "phyact", "cleaning",
                                                          "general", "output"))
  params_sleep = params$params_sleep
  params_metrics = params$params_metrics
  params_247 = params$params_247
  params_phyact = params$params_phyact
  params_cleaning = params$params_cleaning
  params_output = params$params_output
  params_general = params$params_general
  #======================================================================
  # create new folder (if not existent) for storing milestone data
  checkMilestoneFolders(metadatadir, partNumber = 5)
  ms5.out = "/meta/ms5.out"
  if (params_output[["save_ms5rawlevels"]] == TRUE | params_output[["do.sibreport"]] == TRUE) {
    ms5.outraw = "/meta/ms5.outraw"
    if (file.exists(paste(metadatadir, ms5.outraw, sep = ""))) {
    } else {
      dir.create(file.path(metadatadir, ms5.outraw))
    }
    if (params_output[["save_ms5rawlevels"]] == TRUE) {
      # Create on subfolder per configuration
      configurations = c()
      for (TRLi in params_phyact[["threshold.lig"]]) {
        for (TRMi in params_phyact[["threshold.mod"]]) {
          for (TRVi in params_phyact[["threshold.vig"]]) {
            configurations = c(configurations, paste0(TRLi, "_", TRMi, "_", TRVi))
          }
        }
      }
      for (hi in configurations) {
        folder2create = paste0(metadatadir, ms5.outraw, "/", hi)
        if (dir.exists(folder2create) == FALSE) {
          dir.create(path = folder2create)
        }
      }
    }
  }
  nightsummary = M = IMP = sib.cla.sum = c() # declaring variable as otherwise R is confused where they come from, while in fact they are loaded as part of the load operations
  #======================================================================
  # compile lists of milestone data filenames
  fnames.ms3 = dir(paste(metadatadir, "/meta/ms3.out", sep = ""))
  
  fnames.ms5 = dir(paste(metadatadir, "/meta/ms5.out", sep = ""))
  # path to sleeplog milestonedata, if it exists:
  
  if (length(params_sleep[["loglocation"]]) > 0) {
    sleeplogRDA = paste0(metadatadir,"/meta/sleeplog_", basename(params_sleep[["loglocation"]]), ".RData")
  } else {
    sleeplogRDA = paste(metadatadir, "/meta/sleeplog.RData", sep = "")
  }
  if (file.exists(sleeplogRDA) == TRUE) {
    sleeplog = logs_diaries = c()
    load(sleeplogRDA)
    if (length(logs_diaries) > 0) {# new format
      if (is.list(logs_diaries)) { # advanced format
        if (params_sleep[["sleepwindowType"]] == "TimeInBed" && length(logs_diaries$bedlog) > 0) {
          sleeplog = logs_diaries$bedlog
        } else {
          sleeplog = logs_diaries$sleeplog
        }
      } else {
        sleeplog = logs_diaries
      }
    }
  } else {
    sleeplog = logs_diaries = c()
  }
  # Extract activity diary if applicable
  if (is.character(params_247[["qwindow"]])) {
    if (length(grep(pattern = "onlyfilter|filteronly", x = params_247[["qwindow"]])) == 0) {
      epochSize_tmp = ifelse(params_general[["part5_agg2_60seconds"]],
                             yes = 60, no = params_general[["windowsizes"]][1])
      params_247[["qwindow"]] = g.conv.actlog(params_247[["qwindow"]],
                                              params_247[["qwindow_dateformat"]],
                                              epochSize = epochSize_tmp)
      # This will be an object with numeric qwindow values for all individuals and days
    } else {
      # ignore the diary specified by qwindow because user only want to use
      # it for filtering night time nonwear in part 2, but not as a way to
      # do day segment analysis.
      params_247[["qwindow"]] = c(0, 24)
    }
  }
  #------------------------------------------------
  # specify parameters
  ffdone = fnames.ms5 #ffdone is now a list of files that have already been processed by g.part5
  if (f1 > length(fnames.ms3)) f1 = length(fnames.ms3) # this is intentionally ms3 and not ms4, do not change!
  params_phyact[["boutdur.mvpa"]] = sort(params_phyact[["boutdur.mvpa"]],decreasing = TRUE)
  params_phyact[["boutdur.lig"]] = sort(params_phyact[["boutdur.lig"]],decreasing = TRUE)
  params_phyact[["boutdur.in"]] = sort(params_phyact[["boutdur.in"]],decreasing = TRUE)
  #--------------------------------
  # get full file path and folder name if requested by end-user and keep this for storage in output
  if (params_output[["storefolderstructure"]] == TRUE && dir.exists(datadir)) {
    extractfilenames = function(x) as.character(unlist(strsplit(x,".RDa"))[1])
    referencefnames = sapply(fnames.ms3,extractfilenames)
    folderstructure = getfolderstructure(datadir,referencefnames)
    fullfilenames = folderstructure$fullfilenames
    foldername = folderstructure$foldername
  } else {
    referencefnames = fullfilenames = gsub(pattern = "[.]RData", replacement = "", x = fnames.ms3)
    foldername = rep("", length(fnames.ms3))
  }
  if (f0 > length(fnames.ms3)) f0 = 1
  if (f1 == 0 | length(f1) == 0 | f1 > length(fnames.ms3))  f1 = length(fnames.ms3)
  #=========================================================
  # Declare core functionality, which at the end of this g.part5 is either
  # applied to the file in parallel with foreach or serially with a loop
  main_part5 = function(i, metadatadir = c(), f0=c(), f1=c(),
                        params_sleep = c(), params_metrics = c(),
                        params_247 = c(), params_phyact = c(),
                        params_cleaning = c(), params_output = c(),
                        params_general = c(), ms5.out, ms5.outraw,
                        fnames.ms3, sleeplog, logs_diaries,
                        referencefnames, folderstructure,
                        fullfilenames, foldername, ffdone, verbose) {
    tail_expansion_log =  desiredtz_part1 = NULL
    filename_dir = NULL # to be loaded
    fnames.ms1 = dir(paste(metadatadir, "/meta/basic", sep = ""))
    fnames.ms2 = dir(paste(metadatadir, "/meta/ms2.out", sep = ""))
    fnames.ms4 = dir(paste(metadatadir, "/meta/ms4.out", sep = ""))
    nfeatures = 500
    ws3 = params_general[["windowsizes"]][1]
    ds_names = rep("",nfeatures)
    di = 1
    DaCleanFile = c()
    if (length(params_cleaning[["data_cleaning_file"]]) > 0) {
      if (file.exists(params_cleaning[["data_cleaning_file"]])) DaCleanFile = data.table::fread(params_cleaning[["data_cleaning_file"]], data.table = FALSE)
    }
    if (length(ffdone) > 0) {
      if (length(which(ffdone == fnames.ms3[i])) > 0) {
        skip = 1 #skip this file because it was analysed before")
      } else {
        skip = 0 #do not skip this file
      }
    } else {
      skip = 0
    }
    if (params_general[["overwrite"]] == TRUE) skip = 0
    # skip files from ms3 if there is no equivalent in ms4
    selp = which(fnames.ms4 == fnames.ms3[i])
    if (length(selp) > 0) {
      if (file.exists(paste0(metadatadir, "/meta/ms4.out/", fnames.ms4[selp])) == FALSE) {
        skip = 1
      }
    } else {
      skip = 1
    }
    if (skip == 0) {
      # load output g.part2
      selp = which(fnames.ms2 == fnames.ms3[i]) # so, fnames.ms3[i] is the reference point for filenames
      load(file = paste0(metadatadir, "/meta/ms2.out/", fnames.ms2[selp]))
      # load output g.part4
      selp = which(fnames.ms4 == fnames.ms3[i])
      load(file = paste0(metadatadir,"/meta/ms4.out/", fnames.ms4[selp]))
      summarysleep = nightsummary
      rm(nightsummary)
      idindex = which(summarysleep$filename == fnames.ms3[i])
      ID = summarysleep$ID[idindex[1]]
      ndays = nrow(summarysleep) #/ length(unique(summarysleep$sleepparam))
      dsummary = matrix("", ((nrow(summarysleep) + 12) * length(unique(summarysleep$sleepparam))
                             * length(unique(params_phyact[["threshold.lig"]]))
                             * length(unique(params_phyact[["threshold.mod"]]))
                             * length(unique(params_phyact[["threshold.vig"]]))
                             * length(unique(params_output[["timewindow"]]))), nfeatures)
      di = 1
      fi = 1
      SPTE_end = c() # if it is not loaded from part3 milestone data then this will be the default
      # Only attempt to load file if it has at least 1 night of data
      if (length(idindex) > 0 & nrow(summarysleep) > 0) { 
        summarysleep_tmp = summarysleep
        #======================================================================
        # load output g.part1
        selp = which(fnames.ms1 == paste0("meta_", fnames.ms3[i]))
        if (length(selp) != 1) {
          if (verbose == TRUE) cat("Warning: Milestone data part 1 could not be retrieved")
        }
        load(paste0(metadatadir, "/meta/basic/", fnames.ms1[selp]))
        if (!is.null(desiredtz_part1)) {
          params_general[["desiredtz"]] = desiredtz_part1
        }
        # convert to character/numeric if stored as factor in metashort and metalong
        M$metashort = correctOlderMilestoneData(M$metashort)
        M$metalong = correctOlderMilestoneData(M$metalong)
        filename = filename_dir # object comes from load() call above
        # load output g.part3
        longitudinal_axis = NULL # initialise var that is part of ms3.out
        load(paste0(metadatadir, "/meta/ms3.out/", fnames.ms3[i]))
        # remove expanded time so that it is not used for behavioral classification
        if (length(tail_expansion_log) != 0) {
          expanded_short = which(IMP$r5long == -1)
          expanded_long = which(IMP$rout$r5 == -1)
          IMP$metashort = IMP$metashort[-expanded_short,]
          IMP$rout = IMP$rout[-expanded_long,]
          M$metashort = M$metashort[-expanded_short,]
          M$metalong = M$metalong[-expanded_long,]
        }
        
        #====================================
        # Initialise time series data.frame (ts) which will hold the time series
        # which forms the center of all part 5 activity
        # note longitudinal_axis comes from loaded part 3 milestone data and not from params_sleep
        
        ts = g.part5_initialise_ts(IMP, M, params_247, params_general,
                                   longitudinal_axis = longitudinal_axis)
        Nts = nrow(ts)
        lightpeak_available = "lightpeak" %in% names(ts)
        
        rm(IMP, M ,I)
        clock2numtime = function(x) { # function used for converting sleeplog times to hour times
          x2 = as.numeric(unlist(strsplit(x, ":"))) / c(1, 60, 3600)
          return(sum(x2))
        }
        Nepochsinhour = (60/ws3) * 60
        #=======================
        # extract epoch by epoch classification of the entire measurement (day specific summaries will follow further down)
        S = sib.cla.sum
        rm(sib.cla.sum)
        def = unique(S$definition)
        cut = which(S$fraction.night.invalid > 0.9 | S$nsib.periods == 0)
        if (length(cut) > 0) S = S[-cut,]
        if (params_general[["part5_agg2_60seconds"]] == TRUE) {
          ts_backup = ts
        }
        # Remove impossible entries:
        pko = which(summarysleep_tmp$sleeponset == 0 & summarysleep_tmp$wakeup == 0 & summarysleep_tmp$SptDuration == 0)
        if (length(pko) > 0) {
          summarysleep_tmp = summarysleep_tmp[-pko,]
        }
        # if expanded time with expand_tail_max_hours, then latest wakeup might not be in data
        # reset latest wakeup to latest observed timestamp
        if (length(tail_expansion_log) != 0) {
          last_night = S[which(S$night == max(S$night)),]
          last_wakeup = last_night$sib.end.time[which(last_night$sib.period == max(last_night$sib.period))]
          if (!(last_wakeup %in% ts$time)) {
            replaceLastWakeup = which(S$sib.end.time == last_wakeup)
            S$sib.end.time[replaceLastWakeup] = ts$time[nrow(ts)]
          }
        }
        for (sibDef in def) { # loop through sleep definitions (defined by angle and time threshold in g.part3)
          ws3new = ws3 # reset wse3new, because if part5_agg2_60seconds is TRUE then this will have been change in the previous iteration of the loop
          if (params_general[["part5_agg2_60seconds"]] == TRUE) {
            ts = ts_backup
          }
          # extract time and from that the indices for midnights
          time_POSIX = iso8601chartime2POSIX(ts$time,tz = params_general[["desiredtz"]])
          tempp = as.POSIXlt(time_POSIX) #unclass(time_POSIX)
          if (is.na(tempp$sec[1]) == TRUE) {
            time_POSIX = as.POSIXct(ts$time, tz = params_general[["desiredtz"]])
            tempp = as.POSIXlt(time_POSIX)
          }
          sec = tempp$sec
          min = tempp$min
          hour = tempp$hour
          if (params_general[["dayborder"]] == 0) {
            nightsi = which(sec == 0 & min == 0 & hour == 0)
            nightsi2 = nightsi # nightsi2 will be used in g.part5.wakesleepwindows
          } else {
            nightsi = which(sec == 0 & min == (params_general[["dayborder"]] - floor(params_general[["dayborder"]])) * 60 & hour == floor(params_general[["dayborder"]])) #shift the definition of midnight if required
            nightsi2 = which(sec == 0 & min == 0 & hour == 0)
          }
          # include last window if has been expanded and not present in ts
          if (length(tail_expansion_log) != 0 && nrow(ts) > max(nightsi)) nightsi[length(nightsi) + 1] = nrow(ts)
          # create copy of only relevant part of sleep summary dataframe
          summarysleep_tmp2 = summarysleep_tmp[which(summarysleep_tmp$sleepparam == sibDef),]
          # Add sustained inactivity bouts (sib) to the time series
          ts = g.part5.addsib(ts,
                              epochSize = ws3new,
                              part3_output = S[S$definition == sibDef,],
                              desiredtz = params_general[["desiredtz"]],
                              sibDefinition = sibDef,
                              nightsi)
          if (nrow(summarysleep_tmp2) > 0) {
            if (!all(is.na(summarysleep_tmp$sleepparam))) {
              # Fix missing nights in part 4 data:
              summarysleep_tmp2 = g.part5.fixmissingnight(summarysleep_tmp2, sleeplog = sleeplog, ID)
            } else {
              summarysleep_tmp2 = summarysleep_tmp2[-which(is.na(summarysleep_tmp$sleepparam)), ]
            }
          }
          #Initialise diur variable, which will  indicate the diurnal rhythm: 0 if wake/daytime, 1 if sleep/nighttime
          ts$diur = 0
          if (nrow(summarysleep_tmp2) > 0) {
            # Add defenition of wake and sleep windows in diur column of data.frame ts
            ts = g.part5.wakesleepwindows(ts,
                                          part4_output = summarysleep_tmp2,
                                          desiredtz = params_general[["desiredtz"]],
                                          nightsi = nightsi2,
                                          sleeplog,
                                          epochSize = ws3, ID, Nepochsinhour)
            # Add first waking up time, if it is missing:
            ts = g.part5.addfirstwake(ts, summarysleep = summarysleep_tmp2, nightsi, sleeplog, ID,
                                      Nepochsinhour, SPTE_end)
            # Convert time column from iso8601 to POSIX regardless of whether it is aggregated
            # to ensure the format is consistent
            ts$time = iso8601chartime2POSIX(ts$time,tz = params_general[["desiredtz"]])
            if (params_general[["part5_agg2_60seconds"]] == TRUE) { # Optionally aggregate to 1 minute epoch:
              ts$time_num = floor(as.numeric(ts$time) / 60) * 60
              
              # only include angle if angle is present
              angleColName = grep(pattern = "angle", x = names(ts), value = TRUE)
              if (lightpeak_available == TRUE) {
                light_columns = c("lightpeak", "lightpeak_imputationcode")
              } else {
                light_columns = NULL
              }
              temperature_col = grep(pattern = "temperature", x = names(ts), value = TRUE)
              
              # aggregate steps by taking the sum
              stepcount_available = ifelse("step_count" %in% names(ts), yes = TRUE, no = FALSE)
              if (stepcount_available) {
                step_count_tmp = aggregate(ts$step_count, by = list(ts$time_num), FUN = function(x) sum(x))
                colnames(step_count_tmp)[2] = "step_count"
              }
              # aggregate guider names as the first value per time segment
              agg_guider = aggregate(ts$guider,
                                     by = list(ts$time_num), FUN = function(x) x[1])
              colnames(agg_guider)[2] = "guider"
              # aggregate all other variables by taking the mean
              ts = aggregate(ts[,c("ACC","sibdetection", "diur", "nonwear", angleColName, light_columns, temperature_col)],
                             by = list(ts$time_num), FUN = function(x) mean(x))
              ts = merge(x = ts, y = agg_guider, by = "Group.1")
              if (stepcount_available) {
                ts = merge(x = ts, y = step_count_tmp, by = "Group.1")
              }
              ts$sibdetection = round(ts$sibdetection)
              ts$diur = round(ts$diur)
              ts$nonwear = round(ts$nonwear)
              names(ts)[1] = "time"
              # # convert back to iso8601 format
              ts$time = as.POSIXct(ts$time, origin = "1970-1-1", tz = params_general[["desiredtz"]])
              ws3new = 60 # change because below it is used to decide how many epochs are there in
              # extract nightsi again
              time_POSIX = ts$time
              tempp = as.POSIXlt(time_POSIX) #unclass(time_POSIX)
              if (is.na(tempp$sec[1]) == TRUE) {
                time_POSIX = as.POSIXct(ts$time, tz = params_general[["desiredtz"]])
                tempp = as.POSIXlt(time_POSIX)
              }
              sec = tempp$sec
              min = tempp$min
              hour = tempp$hour
              if (params_general[["dayborder"]] == 0) {
                nightsi = which(sec == 0 & min == 0 & hour == 0)
              } else {
                nightsi = which(sec == 0 & min == (params_general[["dayborder"]] - floor(params_general[["dayborder"]])) * 60 & hour == floor(params_general[["dayborder"]])) #shift the definition of midnight if required
              }
              if (length(tail_expansion_log) != 0 & nrow(ts) > max(nightsi)) nightsi[length(nightsi) + 1] = nrow(ts) # include last window
              Nts = nrow(ts)
            }
            #===============================================
            # Use sib.report to classify naps, non-wear and integrate these in time series
            # does not depend on bout detection criteria or window definitions.
            if (params_output[["do.sibreport"]]  == TRUE) {
              IDtmp = as.character(ID)
              sibreport = g.sibreport(ts, ID = IDtmp, epochlength = ws3new, logs_diaries,
                                      desiredtz = params_general[["desiredtz"]])
              
              # Add self-reported classes to ts object
              ts$selfreported = NA
              
              if ("imputecode" %in% colnames(sibreport)) {
                if ("logImputationCode" %in% colnames(ts) == FALSE) {
                  ts$diaryImputationCode = NA # initialise value
                }
                addImputationCode = TRUE
              } else {
                addImputationCode = FALSE
              }
              for (srType in c("sleeplog", "nap", "nonwear", "bedlog")) {
                sr_index = which(sibreport$type == srType)
                if (length(sr_index) > 0) {
                  for (sii in sr_index) {
                    ts_index = which(ts$time >= sibreport$start[sii] & ts$time < sibreport$end[sii])
                    if (addImputationCode == TRUE && srType %in% c("sleeplog", "bedlog")) {
                      ts$diaryImputationCode[ts_index] = as.numeric(sibreport$imputecode[sii])
                    }
                    ts_index1 = ts_index[which(is.na(ts$selfreported[ts_index]))]
                    ts_index2 = ts_index[which(!is.na(ts$selfreported[ts_index]))]
                    if (length(ts_index1) > 0) {
                      ts$selfreported[ts_index1] = srType
                    }
                    if (length(ts_index2) > 0) {
                      ts$selfreported[ts_index2] = paste0(ts$selfreported[ts_index2], "+", srType)
                    }
                  }
                }
              }
              ts$selfreported = as.factor(ts$selfreported)
              # store in csv file:
              ms5.sibreport = "/meta/ms5.outraw/sib.reports"
              if (!file.exists(paste(metadatadir, ms5.sibreport, sep = ""))) {
                dir.create(file.path(metadatadir, ms5.sibreport))
              }
              shortendFname = gsub(pattern = "[.]|RData|csv|cwa|bin", replacement = "", x = fnames.ms3[i], ignore.case = TRUE)
              sibreport_fname =  paste0(metadatadir,ms5.sibreport,"/sib_report_", shortendFname, "_",sibDef,".csv")
              if (length(sibreport) > 0) {
                data.table::fwrite(x = sibreport, file = sibreport_fname, row.names = FALSE,
                                   sep = params_output[["sep_reports"]],
                                   dec = params_output[["dec_reports"]],
                                   dateTimeAs = "write.csv")
              }
            } else {
              sibreport = NULL
            }
            # backup of nightsi outside threshold defintions to avoid
            # overwriting the backup after the first iteration
            nightsi_bu = nightsi
            for (TRLi in params_phyact[["threshold.lig"]]) {
              for (TRMi in params_phyact[["threshold.mod"]]) {
                for (TRVi in params_phyact[["threshold.vig"]]) {
                  # derive behavioral levels (class), e.g. MVPA, inactivity bouts, etc.
                  levelList = identify_levels(ts = ts, TRLi = TRLi, TRMi = TRMi, TRVi = TRVi,
                                              ws3 = ws3new, params_phyact = params_phyact)
                  LEVELS = levelList$LEVELS
                  OLEVELS = levelList$OLEVELS
                  Lnames = levelList$Lnames
                  ts = levelList$ts
                  
                  #=============================================
                  # NOW LOOP TROUGH DAYS AND GENERATE DAY SPECIFIC SUMMARY VARIABLES
                  # we want there to be one more nights in the accelerometer data than there are nights with sleep results
                  NNIGHTSSLEEP = length(unique(summarysleep_tmp2$calendar_date)) # nights with sleep results
                  NNIGHTSACC = length(nightsi) #acc
                  #-------------------------------
                  # ignore all nights in 'inights' before the first waking up and after the last waking up
                  FM = which(diff(ts$diur) == -1)
                  SO = which(diff(ts$diur) == 1)
                  # now 0.5+6+0.5 midnights and 7 days
                  for (timewindowi in params_output[["timewindow"]]) {
                    ts$window = 0
                    nightsi = nightsi_bu
                    # part3 estimate used for first night then
                    part3_estimates_firstnight = which(ts$guider == "part3_estimate")
                    if (length(part3_estimates_firstnight) > 0) {
                      ts$guider[part3_estimates_firstnight] = rev(params_sleep[["HASPT.algo"]])[1]
                    }
                    if (timewindowi == "WW") {
                      if (length(FM) > 0) {
                        # ignore first and last midnight because we did not do sleep detection on it
                        nightsi = nightsi[nightsi > FM[1] & nightsi < FM[length(FM)]]
                      }
                    } else if (timewindowi == "OO") {
                      if (length(SO) > 0) {
                        # ignore data before the first sleep onset and after the last sleep onset
                        nightsi = nightsi[nightsi > SO[1] & nightsi < SO[length(SO)]]
                      }
                    } else {
                      # if first night is missing then nights needs to align with diur
                      startend_sleep = which(abs(diff(ts$diur)) == 1)
                      Nepochsin12Hours =  (60/ws3new) * 60 * 12
                      nightsi = nightsi[nightsi >= (startend_sleep[1] - Nepochsin12Hours) &
                                          nightsi <= (startend_sleep[length(startend_sleep)] + Nepochsin12Hours)]  # newly added on 25-11-2019
                    }
                    if (timewindowi == "MM") {
                      Nwindows = length(nightsi) + 1  # +1 to include the data after last awakening
                    } else if (timewindowi == "WW") {
                      Nwindows = length(which(diff(ts$diur) == -1))
                    } else if (timewindowi == "OO") {
                      Nwindows = length(which(diff(ts$diur) == 1))
                    }
                    indjump = 1
                    qqq_backup = c()
                    add_one_day_to_next_date = FALSE
                    lastDay = ifelse(Nwindows > 0 && length(nightsi) > 0, yes = FALSE, no = TRUE) # skip while loop if there are no days to analyses
                    wi = 1
                    while (lastDay == FALSE) { #loop through windows
                      # Define indices of start and end of the day window (e.g. midnight-midnight, or waking-up or wakingup
                      defdays = g.part5.definedays(nightsi, wi, indjump,
                                                   epochSize = ws3new, qqq_backup, ts, 
                                                   timewindowi, Nwindows, qwindow = params_247[["qwindow"]],
                                                   ID = ID, dayborder = params_general[["dayborder"]])
                      qqq = defdays$qqq
                      qqq_backup = defdays$qqq_backup
                      segments = defdays$segments
                      segments_names = defdays$segments_names
                      lastDay = defdays$lastDay
                      if (length(which(is.na(qqq) == TRUE)) == 0) { #if it is a meaningful day then none of the values in qqq should be NA
                        if ((qqq[2] - qqq[1]) * ws3new > 900) {
                          ts$window[qqq[1]:qqq[2]] = wi
                          if (di == 1) next_si = 1 else next_si = sum(dsummary[,1] != "") + 1
                          for (si in next_si:(next_si + length(segments) - 1)) {
                            fi = 1
                            current_segment_i = si - next_si + 1
                            segStart = segments[[current_segment_i]][1]
                            segEnd = segments[[current_segment_i]][2]
                            extraRowsNeeded = max(c(si, di)) - nrow(dsummary)
                            if (extraRowsNeeded > 0) {
                              dsummary = rbind(dsummary, matrix(data = "", nrow = extraRowsNeeded, ncol = ncol(dsummary)))
                            }
                            if (timewindowi == "MM" & si > 1) { # because first segment is always full window
                              if (("segment" %in% colnames(ts)) == FALSE) ts$segment = NA
                              if (!is.na(segStart) && !is.na(segEnd)) {
                                ts$segment[segStart:segEnd] = si
                              }
                            }
                            # Already store basic information about the file
                            # in the output matrix: 
                            dsummary[si,fi:(fi + 1)] = c(ID, fnames.ms3[i])
                            ds_names[fi:(fi + 1)] = c("ID", "filename"); fi = fi + 2
                            
                            ##################################################
                            # Analysis per segment:
                            # Group categories of objects together
                            # to reduce number of individual objects that need to be
                            # passed on to analyseSegment
                            indexlog = list(fileIndex = i,
                                            winType = timewindowi,
                                            winIndex = wi,
                                            winStartEnd = qqq,
                                            segIndex1 = si,
                                            segIndex2 = current_segment_i,
                                            segStartEnd = c(segStart, segEnd),
                                            columnIndex = fi)
                            timeList = list(ts = ts,
                                            sec = sec,
                                            min = min,
                                            hour = hour,
                                            time_POSIX = time_POSIX,
                                            epochSize = ws3new)
                            gas = g.part5_analyseSegment(indexlog, timeList, levelList,
                                                         segments,
                                                         segments_names,
                                                         dsummary, ds_names,
                                                         # parameter objects
                                                         params_general, params_output,
                                                         params_sleep, params_247,
                                                         params_phyact,
                                                         # sleep related
                                                         sumSleep = summarysleep_tmp2, sibDef,
                                                         # other arguments
                                                         fullFilename = fullfilenames[i],
                                                         add_one_day_to_next_date,
                                                         lightpeak_available, tail_expansion_log,
                                                         foldernamei = foldername[i],
                                                         sibreport = sibreport)
                            # Extract essential object to be used as input for the next 
                            # segment
                            indexlog = gas$indexlog
                            ds_names = gas$ds_names
                            dsummary = gas$dsummary
                            timeList = gas$timeList
                            doNext = gas$doNext
                            add_one_day_to_next_date = gas$add_one_day_to_next_date
                            # indexlog
                            qqq = indexlog$winStartEnd
                            si = indexlog$segIndex1
                            current_segment_i = indexlog$segIndex2
                            segStart = indexlog$segStartEnd[1]
                            segEnd = indexlog$segStartEnd[2]
                            fi = indexlog$columnIndex
                            # timeList
                            ts = timeList$ts
                            ws3new = timeList$epochSize
                            Lnames = levelList$Lnames = timeList$Lnames
                            LEVELS = levelList$LEVELS = timeList$LEVELS
                            if (doNext == TRUE) next
                          }
                          #===============================================
                          # FOLDER STRUCTURE
                          if (params_output[["storefolderstructure"]] == TRUE) {
                            if ("filename_dir" %in% ds_names) fi = which( ds_names == "filename_dir")
                            dsummary[di,fi] = fullfilenames[i] #full filename structure
                            ds_names[fi] = "filename_dir"; fi = fi + 1
                            dsummary[di,fi] = foldername[i] #store the lowest foldername
                            if ("foldername" %in% ds_names) fi = which( ds_names == "foldername")
                            ds_names[fi] = "foldername"; fi = fi + 1
                          }
                          di = di + 1
                        }
                      }
                      di = di + 1
                      wi = wi + 1
                    }
                  }
                  if (params_output[["save_ms5rawlevels"]] == TRUE || params_247[["part6HCA"]] == TRUE || params_247[["part6CR"]] == TRUE) {
                    legendfile = paste0(metadatadir,ms5.outraw,"/behavioralcodes",as.Date(Sys.time()),".csv")
                    legendtable = data.frame(class_name = Lnames, class_id = 0:(length(Lnames) - 1), stringsAsFactors = FALSE)
                    if (!file.exists(legendfile)) {
                      data.table::fwrite(legendtable, file = legendfile, row.names = FALSE,
                                         sep = params_output[["sep_reports"]],
                                         dec = params_output[["dec_reports"]])
                    }
                    # I moved this bit of code to the end, because we want guider to be included (VvH April 2020)
                    rawlevels_fname =  paste0(metadatadir, ms5.outraw, "/", TRLi, "_", TRMi, "_", TRVi, "/",
                                              gsub(pattern = "[.]|rdata|csv|cwa|gt3x|bin",
                                                   replacement = "", x = fnames.ms3[i], ignore.case = TRUE),
                                              "_", sibDef, ".", params_output[["save_ms5raw_format"]])
                    # save time series to csv files
                    if (params_output[["do.sibreport"]] == TRUE & length(params_sleep[["nap_model"]]) > 0) {
                      napNonwear_col = "nap1_nonwear2"
                    } else {
                      napNonwear_col = NULL
                    }
                    if (params_output[["do.sibreport"]]  == TRUE) {
                      selfreported_col = "selfreported"
                    } else {
                      selfreported_col = NULL
                    }
                    
                    if (lightpeak_available == TRUE) {
                      lightpeak_col = "lightpeak"
                    } else {
                      lightpeak_col = NULL
                    }
                    angle_col = grep(pattern = "angle", x = names(ts), value = TRUE)
                    if (length(angle_col) == 0) {
                      angle_col = NULL
                    }
                    temperature_col = grep(pattern = "temperature", x = names(ts), value = TRUE)
                    if (length(temperature_col) == 0) {
                      temperature_col = NULL
                    }
                    step_count_col = grep(pattern = "step_count", x = names(ts), value = TRUE)
                    if (length(step_count_col) == 0) {
                      step_count_col = NULL
                    }
                    
                    diaryImputationCode_col = grep(pattern = "diaryImputationCode", 
                                                   x = names(ts), value = TRUE)
                    if (length(diaryImputationCode_col) == 0) {
                      diaryImputationCode_col = NULL
                    }
                    
                    marker_col = grep(pattern = "marker", x = names(ts), value = TRUE)
                    if (length(marker_col) == 0) {
                      marker_col = NULL
                    }
                    g.part5.savetimeseries(ts = ts[, c("time", "ACC", "diur", "nonwear",
                                                       "guider", "window", "sibdetection", napNonwear_col,
                                                       lightpeak_col, selfreported_col,
                                                       angle_col, temperature_col, step_count_col,
                                                       diaryImputationCode_col, marker_col)],
                                           LEVELS = LEVELS,
                                           desiredtz = params_general[["desiredtz"]],
                                           rawlevels_fname = rawlevels_fname,
                                           DaCleanFile = DaCleanFile,
                                           includedaycrit.part5 = params_cleaning[["includedaycrit.part5"]],
                                           includenightcrit.part5 = params_cleaning[["includenightcrit.part5"]],
                                           ID = ID,
                                           params_output = params_output,
                                           params_247 = params_247,
                                           Lnames = Lnames, timewindow = timewindowi,
                                           filename = filename)
                  }
                }
              }
            }
          }
        }
        if ("angle" %in% colnames(ts)) {
          ts = ts[, -which(colnames(ts) == "angle")]
        }
        #remove NA values
        for (kik in 1:ncol(dsummary)) {
          naval = which(is.na(dsummary[, kik]) == TRUE)
          if (length(naval) > 0) dsummary[naval, kik] = ""
        }
        output = data.frame(dsummary,stringsAsFactors = FALSE)
        names(output) = ds_names
        
        # correct definition of sleep log availability for window = WW, because now it
        # also relies on sleep log from previous night
        whoareWW = which(output$window == "WW") # look up WW
        if (length(params_sleep[["loglocation"]]) > 0) { #only do this if there is a sleep log
          if (length(whoareWW) > 0) {
            whoareNOSL = which(output$sleeplog_used[whoareWW] == "0") #look up nights with no Sleeplog
            if (length(whoareNOSL) > 0) {
              for (k23 in 1:length(whoareNOSL)) {
                k24 = whoareWW[(whoareNOSL[k23] - 1)]
                if (length(k24) > 0) {
                  if (k24 > 0) {
                    output$sleeplog_used[k24] = "0"
                  }
                }
              }
            }
          }
        }
        # tidy up output data frame, because it may have a lot of empty rows and columns
        emptyrows = which(output[,1] == "" & output[,2] == "")
        if (length(emptyrows) > 0) output = output[-emptyrows,]
        lastcolumn = which(colnames(output) == "boutdur.mvpa")
        if (length(lastcolumn) > 0) {
          if (ncol(output) > lastcolumn) {
            emptycols = sapply(output, function(x)all(x == ""))# Find columns filled with missing values which(output[1,] == "" & output[2,] == "")
            # ignore columns with the LUX hour
            emptycols = which(emptycols == TRUE &
                                colnames(output) %in%
                                grep(pattern = "LUX_|FRAG_|dur_|ACC_|Nbouts_|Nblocks_",
                                     x = colnames(output), value = TRUE) == FALSE)
            if (length(emptycols) > 0) emptycols = emptycols[which(emptycols > lastcolumn)]
            # While we explore the fragmentation variables, we want to make sure that all variables are kept in the output
            FRAG_variables_indices = grep(pattern = "FRAG_",x = names(output))
            emptycols = emptycols[which(emptycols %in% FRAG_variables_indices == FALSE)]
            if (length(emptycols) > 0) output = output[,-emptycols]
          }
          if (length(output) > 0) {
            if (nrow(output) > 0) {
              GGIRversion = "GGIR not used"
              if (is.element('GGIR', installed.packages()[,1])) {
                GGIRversion = as.character(utils::packageVersion("GGIR"))
                if (length(GGIRversion) != 1) GGIRversion = sessionInfo()$otherPkgs$GGIR$Version
              }
              output$GGIRversion = GGIRversion
              # Capture final timestamp to ease filtering last window in g.report.part5
              last_timestamp = time_POSIX[length(time_POSIX)] 
              # move experimental nap columns to the end of output
              if ("nap" %in% params_output[["method_research_vars"]]) {
                output = output[, c(grep(pattern = "denap|srnap|srnonw|sibreport_n_items", x = names(output), invert = TRUE, value = FALSE),
                                    grep(pattern = "denap|srnap|srnonw|sibreport_n_items", x = names(output), invert = FALSE, value = FALSE))]
              } else {
                output = output[, grep(pattern = "denap|srnap|srnonw|sibreport_n_items", x = names(output), invert = TRUE, value = FALSE)]
              }
              save(output, tail_expansion_log, GGIRversion, last_timestamp,
                   file = paste(metadatadir, ms5.out, "/", fnames.ms3[i], sep = ""))
            }
          }
        }
        rm(output, dsummary)
      }
    }
  }
  #======================================================================
  # loop through milestone data-files or filenames stored in output of g.part2 and g.part4
  # setup parallel backend to use many processors
  if (params_general[["do.parallel"]] == TRUE) {
    cores = parallel::detectCores()
    Ncores = cores[1]
    if (Ncores > 3) {
      if (length(params_general[["maxNcores"]]) == 0) params_general[["maxNcores"]] = Ncores
      Ncores2use = min(c(Ncores - 1, params_general[["maxNcores"]], (f1 - f0) + 1))
      if (Ncores2use > 1) {
        cl <- parallel::makeCluster(Ncores2use) # not to overload your computer
        parallel::clusterExport(cl = cl,
                                varlist = c(unclass(lsf.str(envir = asNamespace("GGIR"), all = T)),
                                            "MONITOR", "FORMAT"),
                                envir = as.environment(asNamespace("GGIR")))
        parallel::clusterEvalQ(cl, Sys.setlocale("LC_TIME", "C"))
        doParallel::registerDoParallel(cl)
      } else {
        # Don't process in parallel if only one core
        params_general[["do.parallel"]] = FALSE
      }
    } else {
      if (verbose == TRUE) cat(paste0("\nparallel processing not possible because number of available cores (",Ncores,") < 4"))
      params_general[["do.parallel"]] = FALSE
    }
  }
  if (params_general[["do.parallel"]] == TRUE) {
    if (verbose == TRUE) cat(paste0('\n Busy processing ... see ', metadatadir, ms5.out, ' for progress\n'))
    # check whether we are in development mode:
    GGIRinstalled = is.element('GGIR', installed.packages()[,1])
    packages2passon = functions2passon = NULL
    GGIRloaded = "GGIR" %in% .packages()
    if (GGIRloaded) { #pass on package
      packages2passon = 'GGIR'
      errhand = 'pass'
    } else {
      # pass on functions
      functions2passon = c("is.ISO8601", "iso8601chartime2POSIX", "identify_levels", "g.getbout",
                           "g.sibreport", "extract_params", "load_params", "check_params",
                           "correctOlderMilestoneData",
                           "g.part5.addfirstwake", "g.part5.addsib",
                           "g.part5.definedays", "g.part5.fixmissingnight",
                           "g.part5.handle_lux_extremes", "g.part5.lux_persegment",
                           "g.part5.savetimeseries", "g.part5.wakesleepwindows",
                           "g.part5.onsetwaketiming", "g.part5_analyseSegment",
                           "g.part5_initialise_ts", "g.part5.analyseRest",
                           "g.fragmentation", "g.intensitygradient")
      errhand = 'stop'
    }
    i = 0 # declare i because foreach uses it, without declaring it
    `%myinfix%` = foreach::`%dopar%`
    output_list = foreach::foreach(i = f0:f1,  .packages = packages2passon,
                                   .export = functions2passon, .errorhandling = errhand) %myinfix% {
                                     tryCatchResult = tryCatch({
                                       main_part5(i, metadatadir, f0, f1,
                                                  params_sleep, params_metrics,
                                                  params_247, params_phyact,
                                                  params_cleaning, params_output,
                                                  params_general, ms5.out, ms5.outraw,
                                                  fnames.ms3, sleeplog, logs_diaries,
                                                  referencefnames, folderstructure,
                                                  fullfilenames, foldername, ffdone, verbose)
                                     })
                                     return(tryCatchResult)
                                   }
    on.exit(parallel::stopCluster(cl))
    for (oli in 1:length(output_list)) { # logged error and warning messages
      if (is.null(unlist(output_list[oli])) == FALSE) {
        if (verbose == TRUE) cat(paste0("\nErrors and warnings for ", fnames.ms3[oli]))
        print(unlist(output_list[oli])) # print any error and warnings observed
      }
    }
  } else {
    errors = list()
    for (i in f0:f1) {
      if (verbose == TRUE) cat(paste0(i, " "))
      function_to_evaluate = expression(
        main_part5(i, metadatadir, f0, f1,
                   params_sleep, params_metrics,
                   params_247, params_phyact,
                   params_cleaning, params_output,
                   params_general, ms5.out, ms5.outraw,
                   fnames.ms3, sleeplog, logs_diaries,
                   referencefnames, folderstructure,
                   fullfilenames, foldername, ffdone, verbose)
      )
      if (params_general[["use_trycatch_serial"]] == TRUE) {
        tryCatch(
          eval(function_to_evaluate),
          error = function(e) {
            err_msg = conditionMessage(e)
            errors[[as.character(fnames.ms3[i])]] <<- err_msg
          }
        )
      } else {
        eval(function_to_evaluate)
      }
    }
    # show logged errors after the loop:
    if (params_general[["use_trycatch_serial"]] == TRUE && verbose == TRUE) {
      if (length(errors) > 0) {
        cat(paste0("\n\nErrors in part 5... for:"))
        cat(paste0("\n-", names(errors), ": ", unlist(errors), collapse = ""))
      }
    }
  }
}

check_params = function(params_sleep = c(), params_metrics = c(),
                        params_rawdata = c(), params_247 = c(),
                        params_phyact = c(), params_cleaning = c(),
                        params_output = c(), params_general = c()) {
  
  check_class = function(category, params, parnames, parclass) {
    for (parname in parnames) {
      if (length(params[[parname]]) > 0) {
        if (params[[parname]][1] %in% c("c()","NULL") == FALSE) { # because some variables are initialised empty
          x = params[[parname]]
          if (parclass == "numeric") {
            if (!is.numeric(x)) {
              stop(paste0("\n", category, " parameter ", parname, " is not ", parclass))
            }
          }
          if (parclass == "boolean") {
            if (!is.logical(x)) {
              stop(paste0("\n", category, " parameter ", parname, " is not ", parclass))
            }
          }
          if (parclass == "character") {
            if (!is.character(x)) {
              stop(paste0("\n", category, " parameter ", parname, " is not ", parclass))
            }
          }
        }
      }
    }
  }
  #-----------------------------------------------------------------------------------------
  if (length(params_sleep) > 0) { # Check class of sleep parameters
    numeric_params = c("anglethreshold", "timethreshold", "longitudinal_axis", 
                       "possible_nap_window", "possible_nap_dur",
                       "colid", "coln1", "def.noc.sleep", "nnights", 
                       "sleepefficiency.metric", "possible_nap_edge_acc", "HDCZA_threshold",
                       "possible_nap_gap", "oakley_threshold",
                       "nap_markerbutton_method",
                       "nap_markerbutton_max_distance",
                       "SRI1_smoothing_wsize_hrs",
                       "SRI1_smoothing_frac",
                       "spt_min_block_dur",
                       "spt_max_gap_dur", "spt_max_gap_ratio", "HorAngle_threshold",
                       "guider_cor_maxgap_hrs",
                       "guider_cor_min_frac_sib", "guider_cor_min_hrs",
                       "guider_cor_meme_frac_out",
                       "guider_cor_meme_frac_in", "guider_cor_meme_min_hrs",
                       "guider_cor_meme_min_dys")
    boolean_params = c("ignorenonwear", "HASPT.ignore.invalid",
                       "relyonguider", "sleeplogidnum",
                       "impute_marker_button", "consider_marker_button",
                       "sib_must_fully_overlap_with_TimeInBed", "guider_cor_do")
    character_params = c("HASPT.algo", "HASIB.algo", "Sadeh_axis", "nap_model",
                         "sleeplogsep", "sleepwindowType", "loglocation")
    check_class("Sleep", params = params_sleep, parnames = numeric_params, parclass = "numeric")
    check_class("Sleep", params = params_sleep, parnames = boolean_params, parclass = "boolean")
    check_class("Sleep", params = params_sleep, parnames = character_params, parclass = "character")
  }
  if (length(params_metrics) > 0) { # Check class of metrics parameters
    boolean_params = c("do.anglex", "do.angley", "do.anglez",
                       "do.zcx", "do.zcy", "do.zcz",
                       "do.enmo", "do.lfenmo", "do.en", "do.mad", "do.enmoa",
                       "do.roll_med_acc_x", "do.roll_med_acc_y", "do.roll_med_acc_z",
                       "do.dev_roll_med_acc_x", "do.dev_roll_med_acc_y", "do.dev_roll_med_acc_z",
                       "do.bfen", "do.hfen", "do.hfenplus", "do.lfen",
                       "do.lfx", "do.lfy", "do.lfz", "do.hfx", "do.hfy", "do.hfz",
                       "do.bfx", "do.bfy", "do.bfz", "do.brondcounts")
    check_class("Metrics", params = params_metrics, parnames = boolean_params, parclass = "boolean")
    check_class("Metrics", params = params_metrics, parnames = c("hb", "lb", "n", "zc.lb", "zc.hb",
                                                                 "zc.sb", "zc.order", "zc.scale"), parclass = "numeric")
  }
  if (length(params_rawdata) > 0) {
    numeric_params = c("chunksize", "spherecrit", "minloadcrit", "minimumFileSizeMB", "dynrange",
                       "rmc.col.acc", "interpolationType",
                       "rmc.firstrow.acc", "rmc.firstrow.header", "rmc.header.length",
                       "rmc.col.temp", "rmc.col.time",
                       "rmc.sf", "rmc.col.wear", "rmc.noise", "frequency_tol",
                       "rmc.scalefactor.acc", "nonwear_range_threshold")
    boolean_params = c("printsummary", "do.cal", "rmc.unsignedbit", "rmc.check4timegaps", "rmc.doresample",
                       "imputeTimegaps")
    character_params = c("backup.cal.coef", "rmc.dec", "rmc.unit.acc",
                         "rmc.unit.temp", "rmc.unit.time", "rmc.format.time",
                         "rmc.origin", "rmc.desiredtz", "rmc.configtz", "rmc.headername.sf",
                         "rmc.headername.sn", "rmc.headername.recordingid",
                         "rmc.header.structure")
    if (is.logical(params_rawdata[["rmc.noise"]])) {
      # Older config files used this, so overwrite with NULL value
      params_rawdata[["rmc.noise"]] = c()
    }
    check_class("Raw data", params = params_rawdata, parnames = numeric_params, parclass = "numeric")
    check_class("Raw data", params = params_rawdata, parnames = boolean_params, parclass = "boolean")
    check_class("Raw data", params = params_rawdata, parnames = character_params, parclass = "character")

    if (params_rawdata[["chunksize"]] < 0.1) params_rawdata[["chunksize"]] = 0.1
  }
  if (length(params_247) > 0) {
    # iglevels and qwindow can be numeric or character, so not tested
    numeric_params = c("qlevels", "ilevels", "IVIS_windowsize_minutes", "IVIS_epochsize_seconds",
                       "IVIS.activity.metric", "IVIS_acc_threshold",
                       "qM5L5", "MX.ig.min.dur", "M5L5res", "winhr", "LUXthresholds", "LUX_cal_constant",
                       "LUX_cal_exponent", "LUX_day_segments", "L5M5window", "clevels", "SRI2_WASOmin")
    boolean_params = c("cosinor", "part6CR", "part6HCA", "part6DFA", "part2CR")
    character_params = c("qwindow_dateformat", "part6Window")
    check_class("247", params = params_247, parnames = numeric_params, parclass = "numeric")
    check_class("247", params = params_247, parnames = boolean_params, parclass = "boolean")
    check_class("247", params = params_247, parnames = character_params, parclass = "character")
  }
  if (length(params_phyact) > 0) {
    numeric_params = c("mvpathreshold", "boutcriter", "mvpadur",
                       "boutcriter.in", "boutcriter.lig", "boutcriter.mvpa",
                       "threshold.lig", "threshold.mod", "threshold.vig", "boutdur.mvpa",
                       "boutdur.in", "boutdur.lig")
    character_params = c("frag.metrics", "part6_threshold_combi")
    check_class("phyact", params = params_phyact, parnames = numeric_params, parclass = "numeric")
    # check_class("phyact", params = params_phyact, parnames = boolean_params, parclass = "boolean")
    check_class("phyact", params = params_phyact, parnames = character_params, parclass = "character")
    
  }
  if (length(params_cleaning) > 0) {
    numeric_params = c("includedaycrit", "ndayswindow", "data_masking_strategy", "maxdur", "hrs.del.start",
                       "hrs.del.end", "includedaycrit.part5", "minimum_MM_length.part5",
                       "includenightcrit", "max_calendar_days", "includecrit.part6", "includenightcrit.part5",
                       "nonwearFiltermaxHours", "nonwearFilterWindow")

    boolean_params = c("excludefirstlast.part5", "do.imp", "excludefirstlast",
                       "excludefirst.part4", "excludelast.part4", "nonWearEdgeCorrection")
    character_params = c("data_cleaning_file", "TimeSegments2ZeroFile")
    check_class("cleaning", params = params_cleaning, parnames = numeric_params, parclass = "numeric")
    check_class("cleaning", params = params_cleaning, parnames = boolean_params, parclass = "boolean")
    check_class("cleaning", params = params_cleaning, parnames = character_params, parclass = "character")
  }
  if (length(params_output) > 0) {
    numeric_params = c("viewingwindow", "criterror", "visualreport_hrsPerRow",
                       "visualreport_validcrit")
    boolean_params = c("epochvalues2csv", "save_ms5rawlevels", "save_ms5raw_without_invalid",
                       "storefolderstructure", "dofirstpage", "visualreport", "week_weekend_aggregate.part5",
                       "do.part3.pdf", "outliers.only", "do.visual", "do.sibreport", "visualreport_without_invalid",
                       "do.part2.pdf", "do.part2.png", "old_visualreport", "require_complete_lastnight_part5")

    character_params = c("save_ms5raw_format", "timewindow", "sep_reports", "sep_config",
                         "dec_reports", "dec_config", "visualreport_focus", "method_research_vars")
    check_class("output", params = params_output, parnames = numeric_params, parclass = "numeric")
    check_class("output", params = params_output, parnames = boolean_params, parclass = "boolean")
    check_class("output", params = params_output, parnames = character_params, parclass = "character")
  }
  if (length(params_general) > 0) {
    numeric_params = c("maxNcores", "windowsizes", "idloc", "dayborder",
                       "expand_tail_max_hours", "maxRecordingInterval",
                       "recording_split_overlap")
    boolean_params = c("overwrite", "print.filename", "do.parallel", "part5_agg2_60seconds",
                       "recording_split_ignore_edges", "use_trycatch_serial")
    character_params = c("acc.metric", "desiredtz", "configtz", "sensor.location", 
                         "dataFormat", "extEpochData_timeformat", "recording_split_times",
                         "recording_split_timeformat")
    check_class("general", params = params_general, parnames = numeric_params, parclass = "numeric")
    check_class("general", params = params_general, parnames = boolean_params, parclass = "boolean")
    check_class("general", params = params_general, parnames = character_params, parclass = "character")

    ws3 = params_general[["windowsizes"]][1]; ws2 = params_general[["windowsizes"]][2]; ws = params_general[["windowsizes"]][3]
    if (ws2/60 != round(ws2/60)) {
      ws2 = as.numeric(60 * ceiling(ws2/60))
      warning(paste0("The long windowsize needs to be a multitude of 1 minute periods.\n",
                     "Long windowsize has now been automatically adjusted to ",
                     ws2, " seconds in order to meet this criteria."), call. = FALSE)
    }
    if (ws2/ws3 != round(ws2/ws3)) {
      def = c(1,5,10,15,20,30,60)
      def2 = abs(def - ws3)
      ws3 = as.numeric(def[which(def2 == min(def2))])
      warning(paste0("The long windowsize needs to be a multitude of short windowsize.\n",
                     "The short windowsize has now been automatically adjusted to ",
                     ws3, " seconds in order to meet this criteria.\n"), call. = FALSE)
    }
    if (ws/ws2 != round(ws/ws2)) {
      ws = ws2 * ceiling(ws/ws2)
      warning(paste0("The third value of parameter windowsizes needs to be a multitude of the second value.\n",
                     "The third value has been automatically adjusted to ",
                     ws, " seconds in order to meet this criteria.\n"), call. = FALSE)
    }    
    params_general[["windowsizes"]] = c(ws3, ws2, ws)
  }
  #-----------------------------------------------------------------------------------
  # Check value combinations and apply corrections if not logical
  if (length(params_metrics) > 0) {
    if (params_metrics[["do.brondcounts"]] == TRUE) {
      stop(paste0("\nThe brondcounts option has been deprecated following issues with the ",
                  "activityCounts package. We will reinsert brondcounts ",
                  "once the issues are resolved. Consider using parameter do.neishabouricounts, ",
                  "for more information see package documentation."), call. = FALSE)
    }
  }
  
  if (length(params_rawdata) > 0) {
    if (params_rawdata[["frequency_tol"]] < 0 | params_rawdata[["frequency_tol"]] > 1) {
      stop(paste0("\nParameter frequency_tol is ", params_rawdata[["frequency_tol"]],
                  " , please adjust such that it is a number between 0 and 1"))
    }
  }
  
  if (length(params_sleep) > 0) {
    if (length(params_sleep[["def.noc.sleep"]]) != 2) {
      if (params_sleep[["HASPT.algo"]][1] %in% c("HorAngle", "NotWorn", "MotionWare", "HLRB") == FALSE) {
        params_sleep[["HASPT.algo"]] = "HDCZA"
      }
      if (length(params_sleep[["HASPT.algo"]]) == 2 && params_sleep[["HASPT.algo"]][2] == "NotWorn") {
        params_sleep[["HASPT.algo"]] = params_sleep[["HASPT.algo"]][2:1] # NotWorn is expected to be first
      }
    } else if (length(params_sleep[["def.noc.sleep"]]) == 2) {
      params_sleep[["HASPT.algo"]] = "notused"
    }
    if (length(params_sleep[["possible_nap_gap"]]) != 1) {
      stop(paste0("Parameter possible_nap_gap has length ", length(params_sleep[["possible_nap_gap"]]), 
                  " while length 1 is expected"), call. = FALSE)
    }
    if (!is.null(params_sleep[["possible_nap_window"]]) &&
        length(params_sleep[["possible_nap_window"]]) != 2) {
      stop(paste0("Parameter possible_nap_window has length ", length(params_sleep[["possible_nap_window"]]), 
                  " while length 2 is expected"), call. = FALSE)
    }
    if (!is.null(params_sleep[["possible_nap_dur"]]) &&
        length(params_sleep[["possible_nap_dur"]]) != 2) {
      stop(paste0("Parameter possible_nap_dur has length ", length(params_sleep[["possible_nap_dur"]]), 
                  " while length 2 is expected"), call. = FALSE)
    }
    if (!is.null(params_sleep[["possible_nap_window"]]) &&
        !is.null(params_sleep[["possible_nap_dur"]])) {
      params_output[["do.sibreport"]] = TRUE
      params_output[["save_ms5raw_format"]] = unique(c(params_output[["save_ms5raw_format"]], "RData"))
      params_output[["save_ms5rawlevels"]] = TRUE
      params_output[["save_ms5raw_without_invalid"]] = FALSE
    }
  }
  
  if (length(params_metrics) > 0 & length(params_sleep) > 0) {
    
    sib_90s_algo_names = c("Sadeh1994", "Galland2012", "ColeKripke1992", "Oakley1997")
    if (any(params_sleep[["HASIB.algo"]] %in% sib_90s_algo_names == TRUE)) {
      if (params_sleep[["Sadeh_axis"]] %in% c("X","Y","Z") == FALSE) {
        warning("Parameter Sadeh_axis does not have meaningful value, it needs to be X, Y or Z (capital)", call. = FALSE)
      }
      if (params_sleep[["Sadeh_axis"]] == "X" & params_metrics[["do.zcx"]] == FALSE) params_metrics[["do.zcx"]] =  TRUE
      if (params_sleep[["Sadeh_axis"]] == "Y" & params_metrics[["do.zcy"]] == FALSE) params_metrics[["do.zcy"]] =  TRUE
      if (params_sleep[["Sadeh_axis"]] == "Z" & params_metrics[["do.zcz"]] == FALSE) params_metrics[["do.zcz"]] =  TRUE
    } else { # vanHees2015
      params_sleep[["Sadeh_axis"]] = "" # not used
    }
  }
  
  if (length(params_general) > 0 & length(params_metrics) > 0 & length(params_sleep) > 0) {
    if (params_general[["sensor.location"]] == "hip" &&
        params_sleep[["HASPT.algo"]][1] %in% c("notused", "NotWorn") == FALSE) {
      if (params_metrics[["do.anglex"]] == FALSE | params_metrics[["do.angley"]] == FALSE | params_metrics[["do.anglez"]] == FALSE) {
        warning(paste0("\nWhen working with hip data all three angle metrics are needed,",
                       "so GGIR now auto-sets parameters do.anglex, do.angley, and do.anglez to TRUE."), call. = FALSE)
        params_metrics[["do.anglex"]] = params_metrics[["do.angley"]] = params_metrics[["do.anglez"]] = TRUE
      }
      if (length(params_sleep[["HASPT.algo"]]) == 1 && params_sleep[["HASPT.algo"]][1] != "HorAngle") {
        warning("\nChanging HASPT.algo value to HorAngle, because sensor.location is set as hip", call. = FALSE)
        params_sleep[["HASPT.algo"]] = "HorAngle"; params_sleep[["def.noc.sleep"]] = 1
      }
    }
  }
  
  if (length(params_sleep) > 0) {
    if (length(params_sleep[["loglocation"]]) == 1) {
      if (params_sleep[["loglocation"]] == "") {
        params_sleep[["loglocation"]] = c() #inserted because some users mistakingly use this
      } else {
        # Convert paths from Windows specific slashed to generic slashes
        params_sleep[["loglocation"]] = gsub(pattern = "\\\\", replacement = "/", x = params_sleep[["loglocation"]])
      }
    }
    
    if (length(params_sleep[["loglocation"]]) > 0 & length(params_sleep[["def.noc.sleep"]]) != 1) {
      warning(paste0("\nloglocation was specified and def.noc.sleep does not have length of 1, this is not compatible. ",
                     " We assume you want to use the sleeplog and misunderstood",
                     " parameter def.noc.sleep. Therefore, we will reset def.noc.sleep to its default value of 1"), call. = FALSE)
      params_sleep[["def.noc.sleep"]] = 1
    }
    
    if (params_sleep[["HASPT.algo"]][1] == "HorAngle" & params_sleep[["sleepwindowType"]] != "TimeInBed") {
      warning("\nHASPT.algo is set to HorAngle, therefore auto-updating sleepwindowType to TimeInBed", call. = FALSE)
      params_sleep[["sleepwindowType"]] = "TimeInBed"
    }
    
    if (length(params_sleep[["loglocation"]]) == 0 &
        params_sleep[["HASPT.algo"]][1] != "HorAngle" &
        params_sleep[["HASPT.algo"]][1] != "NotWorn" &
        params_sleep[["sleepwindowType"]] != "SPT") {
      warning("\nAuto-updating sleepwindowType to SPT because no sleeplog used and neither HASPT.algo HorAngle or NotWorn used.", call. = FALSE)
      params_sleep[["sleepwindowType"]] = "SPT"
    }
  }
  
  if (length(params_cleaning) > 0) {
    # overwrite data_masking_strategy with strategy in case the latter is used
    if (params_cleaning[["strategy"]] != params_cleaning[["data_masking_strategy"]]
        & params_cleaning[["strategy"]] != 1) {
      params_cleaning[["data_masking_strategy"]] = params_cleaning[["strategy"]]
    }
    if (params_cleaning[["data_masking_strategy"]] %in% c(2, 4) & params_cleaning[["hrs.del.start"]] != 0) {
      warning(paste0("\nSetting parameter hrs.del.start in combination with data_masking_strategy = ",
                     params_cleaning[["data_masking_strategy"]]," is not meaningful, because this is only used when straytegy = 1"), call. = FALSE)
    }
    if (params_cleaning[["data_masking_strategy"]] %in% c(2, 4) & params_cleaning[["hrs.del.end"]] != 0) {
      warning(paste0("\nSetting parameter hrs.del.end in combination with data_masking_strategy = ",
                     params_cleaning[["data_masking_strategy"]]," is not meaningful, because this is only used when straytegy = 1"), call. = FALSE)
    }
    if (!(params_cleaning[["data_masking_strategy"]] %in% c(3, 5)) & params_cleaning[["ndayswindow"]] != 7) {
      warning(paste0("\nSetting parameter ndayswindow in combination with data_masking_strategy = ",
                     params_cleaning[["data_masking_strategy"]]," is not meaningful, because this is only used when data_masking_strategy = 3 or data_masking_strategy = 5"), call. = FALSE)
    }
    if (params_cleaning[["data_masking_strategy"]] == 5 &
        params_cleaning[["ndayswindow"]] != round(params_cleaning[["ndayswindow"]])) {
      newValue = round(params_cleaning[["ndayswindow"]])
      warning(paste0("\nParameter ndayswindow has been rounded from ",
                     params_cleaning[["ndayswindow"]], " to ", newValue, " days",
                     "because when data_masking_strategy == 5 we expect an integer value", call. = FALSE))
      params_cleaning[["ndayswindow"]] = newValue
    }
    if (length(params_cleaning[["data_cleaning_file"]]) > 0) {
      # Convert paths from Windows specific slashed to generic slashes
      params_cleaning[["data_cleaning_file"]] = gsub(pattern = "\\\\",
                                                     replacement = "/", x = params_cleaning[["data_cleaning_file"]])
    }
    if (params_cleaning[["includedaycrit.part5"]] < 0) {
      stop("\nNegative value of includedaycrit.part5 is not allowed, please change.")
    } else if (params_cleaning[["includedaycrit.part5"]]  > 24) {
      stop(paste0("\nIncorrect value of includedaycrit.part5, this should be",
                     " a fraction of the day between zero and one or the number ",
                     "of hours in a day."))
    }
    if (!is.null(params_cleaning[["nonwearFiltermaxHours"]])) {
      if (params_cleaning[["nonwearFiltermaxHours"]] < 0 ||
          params_cleaning[["nonwearFiltermaxHours"]] > 12) {
        stop("Parameters nonwearFiltermaxHours is expected to have a value > 0 and < 12")
      }
      if (!is.null(params_cleaning[["nonwearFilterWindow"]])) {
        if (length(params_cleaning[["nonwearFilterWindow"]]) != 2) {
          stop("Parameter nonwearFilterWindow does not have expected length of 2, please fix.", call. = FALSE)
        }
        if (params_cleaning[["nonwearFilterWindow"]][1] < params_cleaning[["nonwearFilterWindow"]][2] &&
            params_cleaning[["nonwearFilterWindow"]][2] > 18 &&
            params_cleaning[["nonwearFilterWindow"]][1] < 12) {
          warning(paste0("The NonwearFilter applied to window starting at ", 
                         params_cleaning[["nonwearFilterWindow"]][1], " and ending at ",
                         params_cleaning[["nonwearFilterWindow"]][2], 
                         " this is probably not the night, please check that order of",
                         " values in nonwearFilterWindow is correct"), call. = FALSE)
        }
      }
    }

    if (params_cleaning[["includenightcrit.part5"]] < 0) {
      stop("\nNegative value of includenightcrit.part5 is not allowed, please change.")
    } else if (params_cleaning[["includenightcrit.part5"]]  > 24) {
      stop(paste0("\nIncorrect value of includenightcrit.part5, this should be",
                     " a fraction of the day between zero and one or the number ",
                     "of hours in a day."))
    }
    if (any(params_cleaning[["includecrit.part6"]] < 0) | any(params_cleaning[["includecrit.part6"]] > 1)) {
      stop("Values of includecrit.part6 are not in the range [0, 1]. Please fix.")
    }
  }
  if (length(params_phyact) > 0) {
    if (length(params_phyact[["bout.metric"]]) > 0 |
        length(params_phyact[["closedbout"]]) > 0) {
      warning(paste0("\nParameters bout.metric and closedbout are no longer used",
                     " by GGIR and ignored, we now use one piece of code stored in",
                     " function g.getbout."), call. = FALSE)
    }
    if (length(params_phyact[["mvpadur"]]) != 3) {
      params_phyact[["mvpadur"]] = c(1,5,10)
      warning("\nmvpadur needs to be a vector with length three, value now reset to default c(1, 5, 10)", call. = FALSE)
    }
    if (is.null(params_phyact[["mvpathreshold"]]) == TRUE) {
      params_phyact[["mvpathreshold"]] = params_phyact[["threshold.mod"]]
    }
    if (is.null(params_phyact[["boutcriter"]]) == TRUE) {
      params_phyact[["boutcriter"]] = params_phyact[["boutcriter.mvpa"]]
    }
    if ((length(params_phyact[["threshold.lig"]]) == 1 &&
        length(params_phyact[["threshold.mod"]]) == 1 &&
        length(params_phyact[["threshold.vig"]]) == 1) | is.null(params_phyact[["part6_threshold_combi"]])) {
      params_phyact[["part6_threshold_combi"]] = paste(params_phyact[["threshold.lig"]][1],
                                                       params_phyact[["threshold.mod"]][1],
                                                       params_phyact[["threshold.vig"]][1], sep = "_")
    }
  }
  # params output 
  if (length(params_output) > 0) {
    if (!all(params_output[["save_ms5raw_format"]] %in% c("RData", "csv"))) {
      formats2keep = which(params_output[["save_ms5raw_format"]] %in% c("RData", "csv"))
      if (length(formats2keep) > 0) {
        params_output[["save_ms5raw_format"]] = params_output[["save_ms5raw_format"]][formats2keep]
      } else {
        stop("Parameter save_ms5raw_format incorrectly specified, please fix.", call. = FALSE)
      }
    }
    if (params_output[["sep_reports"]] == params_output[["dec_reports"]]) {
      stop(paste0("\nYou have set sep_reports and dec_reports both to ",
                  params_output[["sep_reports"]], " this is ambiguous. Please fix."))
    }
    if (params_output[["sep_config"]] == params_output[["dec_config"]]) {
      stop(paste0("\nYou have set sep_config and dec_config both to ",
                  params_output[["sep_config"]], " this is ambiguous. Please fix."))
    }
    if (params_output[["visualreport_hrsPerRow"]] < 24 ||
        params_output[["visualreport_hrsPerRow"]] > 48) {
      stop("Parameter visualreport_hrsPerRow is expected to be set in the range 24-48")
    }
    
    if (!is.null(params_output[["do.part2.pdf"]]) && is.null(params_output[["do.part2.png"]])) {
      params_output[["do.part2.png"]] = params_output[["do.part2.pdf"]]
    }
    if (is.null(params_output[["do.part2.png"]])) {
      params_output[["do.part2.png"]] = TRUE
    }
    
  }
  # params 247
  if (length(params_247) > 0) {
    if (length(params_247[["iglevels"]]) > 0) {
      if (length(params_247[["iglevels"]]) == 1) {
        params_247[["iglevels"]] = c(seq(0, 4000, by = 25), 8000) # to introduce option to just say TRUE
      }
    }
    if (length(params_247[["qwindow"]]) > 0) {
      if (is.character(params_247[["qwindow"]])) {
        # Convert paths from Windows specific slashed to generic slashes
        params_247[["qwindow"]] = gsub(pattern = "\\\\", replacement = "/", x = params_247[["qwindow"]])
      }
    }
    if (length(params_247[["LUX_day_segments"]]) > 0) {
      params_247[["LUX_day_segments"]] = sort(unique(round(params_247[["LUX_day_segments"]])))
      if (params_247[["LUX_day_segments"]][1] != 0) {
        params_247[["LUX_day_segments"]] = c(0, params_247[["LUX_day_segments"]])
      }
      if (params_247[["LUX_day_segments"]][length(params_247[["LUX_day_segments"]])] != 24) {
        params_247[["LUX_day_segments"]] = c(params_247[["LUX_day_segments"]], 24)
      }
    }
    if (params_247[["cosinor"]] == TRUE && params_247[["part2CR"]] == FALSE) {
      params_247[["part2CR"]] = TRUE
    }
    # params 247 & params output
    if (length(params_output[["save_ms5raw_format"]]) == 1 && 
        params_output[["save_ms5raw_format"]] == "csv") {
      # always add RData if only csv is specified, because otherwise visualreport cannot be generated
      params_output[["save_ms5raw_format"]] = c(params_output[["save_ms5raw_format"]], "RData")
    }

    if (length(params_247[["clevels"]]) == 1) {
      warning("\nParameter clevels expects a number vector of at least 2 values, current length is 1", call. = FALSE)
    }
  }
  
  if (length(params_output) > 0 && length(params_247) > 0) {
    if (params_247[["part6HCA"]] == TRUE || params_247[["part6CR"]] == TRUE || 
        params_output[["visualreport"]] == TRUE) {
      # Add RData because part 6 / visualreport will need it
      params_output[["save_ms5raw_format"]] = unique(c(params_output[["save_ms5raw_format"]], "RData"))
      params_output[["save_ms5rawlevels"]] = TRUE
      params_output[["save_ms5raw_without_invalid"]] = FALSE
    }
  }
  if (!is.null(params_general[["expand_tail_max_hours"]])) {
    if (is.null(params_general[["recordingEndSleepHour"]]) & params_general[["expand_tail_max_hours"]] != 0) {
      params_general[["recordingEndSleepHour"]] = 24 - params_general[["expand_tail_max_hours"]] # redefine the argument
      params_general[["expand_tail_max_hours"]] = NULL # set to null so that it keeps this configuration in the config file for the next run of the script.
      stop("\nThe parameter expand_tail_max_hours has been replaced by",
           " recordingEndSleepHour which has a different definition. Please",
           " see the documentation for further details and replace",
           " expand_tail_max_hour in your function call and config.csv file.", call. = FALSE)
    } else {
      # If both are defined, this is probably because expand_tail_max_hours is
      # in the config file from a previous run
      params_general[["expand_tail_max_hours"]] = NULL # set to null so that it keeps this configuration in the config file for the next run of the script.
      warning("\nBoth expand_tail_max_hours and recordingEndSleepHour",
              " are defined. GGIR will only use recordingEndSleepHour",
              " and expand_tail_max_hours will be set to NULL.", call. = FALSE)
    }
  }
  
  if (length(params_general) > 0 & length(params_sleep) > 0) {
    if (params_sleep[["HASPT.algo"]][1] == "HorAngle") {
      # Not everywhere in the code we check that when HASPT.algo is HorAngle, sensor.location is hip.
      # However, the underlying assumption is that they are linked. Even if a study would
      # use a waist or chest worn sensor we refer to it as hip as the orientation and need
      # for detecting longitudinal axis are the same. 
      # Therefore, sensor.location should be forced to hip if HASPT.algo is HorAngle.
      # On the other hand hip does not mean that HorAngle needs to be used, because
      # when using count data from the hip the user may prefer the HASPT.algo=NotWorn.
      params_general[["sensor.location"]] = "hip"
    }
    
    if (params_general[["dataFormat"]] %in% c("sensewear_xls", "fitbit_json")) {
      if (params_sleep[["HASIB.algo"]][1] != "data") {
        params_sleep[["HASIB.algo"]] = "data"
        warning(paste0("HASIB.algo has been set to \"data\", because dataFormat is",
                       " ", params_general[["dataFormat"]], " for which the sleep ",
                       " classification inside the data is the only route."), call. = FALSE)
      }
    }
    
  }
  
  if (length(params_metrics) > 0 & length(params_general) > 0) {
    if (params_general[["dataFormat"]] %in% c("actiwatch_awd", "actiwatch_csv", "phb_xlsx")) {
      params_metrics[["do.enmo"]] = params_metrics[["do.anglez"]] = FALSE
      params_metrics[["do.anglex"]] = params_metrics[["do.angley"]] = FALSE
      params_metrics[["do.anglez"]] = params_metrics[["do.enmoa"]] = FALSE
      params_metrics[["do.enmo"]] = params_metrics[["do.bfen"]] = FALSE
      params_metrics[["do.mad"]] = params_metrics[["do.lfenmo"]] = FALSE
      params_general[["acc.metric"]] = "ExtAct"
      if (length(params_sleep) > 0) {
        if (any(params_sleep[["HASIB.algo"]] == "vanHees2015")) {
          stop(paste0("\nSleep algorithm ", params_sleep[["HASIB.algo"]], " is not a valid",
                      " setting in combination with dataFormat set to ",
                      params_general[["dataFormat"]], " Please fix"), call. = FALSE)
        }
      }
    } else if (params_general[["dataFormat"]] == "ukbiobank") {
      # Turn all commonly used metrics to FALSE
      params_metrics[["do.anglex"]] = params_metrics[["do.angley"]] = FALSE
      params_metrics[["do.anglez"]] = params_metrics[["do.enmoa"]] = FALSE
      params_metrics[["do.enmo"]] = params_metrics[["do.bfen"]] = FALSE
      params_metrics[["do.mad"]] = FALSE
      params_general[["acc.metric"]] = "LFENMO"
    } else if (params_general[["dataFormat"]] == "sensewear_xls") {
      params_general[["acc.metric"]] = "ExtAct"
    }
  }
  if (!is.null(params_general[["recordingEndSleepHour"]])) {
    # stop if expand_tail_max_hours was defined before 7pm
    if (params_general[["recordingEndSleepHour"]] < 19) {
      stop(paste0("\nrecordingEndSleepHour expects the latest time at which",
                  " the participant is expected to fall asleep. recordingEndSleepHour",
                  " has been defined as ", params_general[["recordingEndSleepHour"]],
                  ", which does not look plausible, please specify time at or later than 19:00",
                  " . Please note that it is your responsibility as user to verify that the
                  assumption is credible."), call. = FALSE)
    }
  }
  
  if (length(params_general) > 0) {
    if (!is.null(params_general[["maxRecordingInterval"]])) {
      if (params_general[["maxRecordingInterval"]] > 24 * 21) {
        stop(paste0("A maxRecordingInterval value higher than 21 days (504 hours) is not permitted,",
                    " please specify a lower value."), call. = FALSE)
      }
    }
    if (!is.null(params_general[["recording_split_times"]])) {
      if (!file.exists(params_general[["recording_split_times"]])) {
        stop(paste0("File .../", basename(params_general[["recording_split_times"]]),
                    " as specified with parameter recording_split_times does not exist, ",
                    " please fix."), call. = FALSE)
        
      }
    }
  }
  
  # cleaning parameters for segments
  if (length(params_cleaning) > 0) {
    if (is.null(params_cleaning[["includedaycrit.part5"]]) == TRUE) {
      stop(paste0("\nSetting includedaycrit.part5 to an empty value is not allowed",
                  ", please change."), call. = FALSE)
    } else if (params_cleaning[["includedaycrit.part5"]] < 0) {
      stop(paste0("\nNegative value of includedaycrit.part5 is not allowed",
                     ", please change."), call. = FALSE)
    } else if (params_cleaning[["includedaycrit.part5"]] > 25) {
      stop(paste0("\nIncorrect value of includedaycrit.part5, this should ",
                     "be a fraction of the day between zero and one or the ",
                     "number of hours in a day."), call. = FALSE)
    }
    if (is.null(params_cleaning[["segmentWEARcrit.part5"]])) {
      # if null, then assign default value
      params_cleaning[["segmentWEARcrit.part5"]] = 0.5
      warning(paste0("\nsegmentWEARcrit.part5 is expected to be a number between 0 and 1",
                     ", the default value has been assigned (i.e., 0.5) "), call. = FALSE)
    } else if (params_cleaning[["segmentWEARcrit.part5"]] < 0 | 
               params_cleaning[["segmentWEARcrit.part5"]] > 1) {
      stop(paste0("Incorrect value of segmentWEARcrit.part5, this should be a ",
                  "fraction between zero and one, please change."), 
           call. = FALSE)
    }
    if (length(params_cleaning[["segmentDAYSPTcrit.part5"]]) != 2) {
      stop("\nParameter segmentDAYSPTcrit.part5 is expected to be a numeric vector of length 2", call. = FALSE)
    }
    if (any(params_cleaning[["segmentDAYSPTcrit.part5"]] < 0) | 
        any(params_cleaning[["segmentDAYSPTcrit.part5"]] > 1)) {
      stop(paste0("Incorrect values of segmentDAYSPTcrit.part5, these should be a ",
                  "fractions between zero and one, please change."), 
           call. = FALSE)
    }
  }
  
  invisible(list(params_sleep = params_sleep,
                 params_metrics = params_metrics,
                 params_rawdata = params_rawdata,
                 params_247 = params_247,
                 params_phyact = params_phyact,
                 params_cleaning = params_cleaning,
                 params_output = params_output,
                 params_general = params_general))
}

aggregateEvent = function(metric_name, epochsize,
                          daysummary,  ds_names, fi, di,
                          vari, segmentInfo, myfun = NULL, params_247) {
  anwi_nameindices = segmentInfo$anwi_nameindices
  anwi_index = segmentInfo$anwi_index
  anwi_t0 = segmentInfo$anwi_t0
  anwi_t1 = segmentInfo$anwi_t1
  if ("ilevels" %in% names(myfun) == FALSE) myfun$ilevels = c(0, 80)
  if ("clevels" %in% names(myfun) == FALSE) myfun$clevels = c(0, 30)
  if ("qlevels" %in% names(myfun) == FALSE) myfun$qlevels = c(0.25, 0.5, 0.75)
  if (length(myfun$ilevels) == 0) myfun$ilevels = 0
  if (length(myfun$clevels) == 0) myfun$clevels = 0
  if (length(myfun$qlevels) == 0) myfun$qlevels = 0.5
  acc.thresholds = myfun$ilevels
  cadence.thresholds = myfun$clevels
  qlevels = myfun$qlevels
  fi_start = fi
  cadence = vari[, metric_name] * (60/epochsize) # assumption now that cadence is also relevant if the event detected is not steps
  # aggregate per window total
  varnameevent = paste0("ExtFunEvent_tot_", metric_name, anwi_nameindices[anwi_index])
  daysummary[di, fi] = sum(vari[, metric_name])
  ds_names[fi] = varnameevent; fi = fi + 1
  
  # cadence
  varnameevent = paste0("ExtFunEvent_mn_cad", anwi_nameindices[anwi_index])
  daysummary[di, fi] = mean(cadence, na.rm = TRUE)
  ds_names[fi] = varnameevent; fi = fi + 1
  
  # cadence percentiles
  cadence_quantiles = quantile(cadence, probs = qlevels)
  varnameevent = paste0("ExtFunEvent_cad_p", qlevels * 100, anwi_nameindices[anwi_index])
  daysummary[di, fi:(fi + length(qlevels) - 1)] = as.numeric(cadence_quantiles)
  ds_names[fi:(fi + length(qlevels) - 1)] = varnameevent; fi = fi + length(qlevels)
  
  #========================================
  # per acceleration level
  cn_vari = colnames(vari)
  acc.metrics = cn_vari[cn_vari %in% c("timestamp", "anglex", "angley", "anglez", metric_name) == FALSE]
  
  for (ami in 1:length(acc.metrics)) {
    vari[, acc.metrics[ami]] = as.numeric(vari[, acc.metrics[ami]])
    for (ti in 1:length(acc.thresholds)) {
      # step_count per acceleration level
      if (ti < length(acc.thresholds)) {
        acc_level_name = paste0(acc.thresholds[ti], "-", acc.thresholds[ti + 1], "mg",  "_",  acc.metrics[ami])
        whereAccLevel = which(vari[, acc.metrics[ami]] >= (acc.thresholds[ti]/1000) &
                                vari[, acc.metrics[ami]] < (acc.thresholds[ti + 1]/1000))
      } else {
        acc_level_name = paste0("atleast", acc.thresholds[ti], "mg",  "_",  acc.metrics[ami])
        whereAccLevel = which(vari[,acc.metrics[ami]] >= (acc.thresholds[ti]/1000))
      }
      varnameevent = paste0("ExtFunEvent_tot_", metric_name, "_acc", acc_level_name, anwi_nameindices[anwi_index])
      if (length(whereAccLevel) > 0)  {
        daysummary[di, fi] = sum(vari[whereAccLevel, metric_name], na.rm = TRUE)
      } else {
        daysummary[di, fi] = 0
      }
      ds_names[fi] = varnameevent; fi = fi + 1
      
      # cadence per acceleration level
      varnameevent = paste0("ExtFunEvent_mn_cad_acc", 
                            acc_level_name, anwi_nameindices[anwi_index])
      if (length(whereAccLevel) > 0)  {
        daysummary[di, fi] = mean(cadence[whereAccLevel], na.rm = TRUE)
      } else {
        daysummary[di, fi] = 0
      }
      ds_names[fi] = varnameevent; fi = fi + 1
    }
  }
  
  #========================================
  # per cadence level
  for (ti in 1:length(cadence.thresholds)) {
    # define cadence level
    if (ti < length(cadence.thresholds)) {
      cadence_level_name = paste0(cadence.thresholds[ti], "-", cadence.thresholds[ti + 1], "spm")
      whereCadenceLevel = which(cadence >= cadence.thresholds[ti] &
                                  cadence < cadence.thresholds[ti + 1])
    } else {
      cadence_level_name = paste0("atleast", cadence.thresholds[ti], "spm")
      whereCadenceLevel = which(cadence >= cadence.thresholds[ti])
    }
    # cadence per cadence level
    varnameevent = paste0("ExtFunEvent_mn_cad_cad", cadence_level_name, anwi_nameindices[anwi_index])
    if (length(whereCadenceLevel) > 0)  {
      daysummary[di, fi] = mean(cadence[whereCadenceLevel], na.rm = TRUE)
    } else {
      daysummary[di, fi] = 0
    }
    ds_names[fi] = varnameevent; fi = fi + 1
    
    # step count per cadence level
    varnameevent = paste0("ExtFunEvent_tot_", metric_name, "_cad", 
                          cadence_level_name, anwi_nameindices[anwi_index])
    if (length(whereCadenceLevel) > 0)  {
      daysummary[di, fi] = sum(vari[whereCadenceLevel, metric_name], na.rm = TRUE)
    } else {
      daysummary[di, fi] = 0
    }
    ds_names[fi] = varnameevent; fi = fi + 1
    
    # time per cadence level
    varnameevent = paste0("ExtFunEvent_dur_cad", cadence_level_name, anwi_nameindices[anwi_index])
    daysummary[di, fi] = length(whereCadenceLevel) / (60/epochsize)
    ds_names[fi] = varnameevent; fi = fi + 1
    
    for (ami in 1:length(acc.metrics)) {
      # acceleration per cadence level
      varnameevent = paste0("ExtFunEvent_mn_",  acc.metrics[ami],"_cad",
                            cadence_level_name, anwi_nameindices[anwi_index])
      if (length(whereCadenceLevel) > 0)  {
        daysummary[di,fi] = mean(vari[whereCadenceLevel, acc.metrics[ami]], na.rm = TRUE) * 1000
      } else {
        daysummary[di, fi] = 0
      }
      ds_names[fi] = varnameevent; fi = fi + 1
    }
  }
  if (length(params_247[["winhr"]]) > 0) { 
    # Event based MXLX
    tseg = c((anwi_t0[anwi_index] - 1) / (3600 / epochsize), anwi_t1[anwi_index] / (3600 / epochsize))
    for (winhr_value in params_247[["winhr"]]) {
      MXLXout = MXLX(Y = cadence, X = winhr_value, epochSize = epochsize,
                     tseg = tseg,
                     resolutionMin = params_247[["M5L5res"]])
      # Describe for MX and LX:
      LXMXwindow_name = anwi_nameindices[anwi_index]
      for (WX in c("L", "M")) {
        if (!all(is.na(MXLXout[, grep("M", names(MXLXout))]))) {
          i0 = MXLXout[, grep(paste0("start_", WX, winhr_value), names(MXLXout))]
          i1 = MXLXout[, grep(paste0("end_", WX, winhr_value), names(MXLXout))]
          WXi = i0:i1
          # # Code used to help check that MXLX makes sense with real study data
          # x11()
          # plot(cadence, type = "l")
          # lines(WXi, cadence[WXi], type = "l", col = "red")
          
          # Timing
          daysummary[di, fi] = MXLXout[, which(names(MXLXout) == paste0(WX, winhr_value, "hr"))]
          ds_names[fi] = paste0("ExtFunEvent_", WX, winhr_value,  "hr_cad", LXMXwindow_name); fi = fi + 1
          # cadence in the form of mean, percentiles.
          daysummary[di, fi] = mean(cadence[WXi])
          ds_names[fi] = paste0("ExtFunEvent_", WX, winhr_value, "_cad_meancad", LXMXwindow_name); fi = fi + 1
          if (length(params_247[["qM5L5"]]) > 0) {
            for (qle in params_247[["qM5L5"]]) {
              daysummary[di, fi] = quantile(x = cadence[WXi], probs = qle)
              ds_names[fi] = paste0("ExtFunEvent_", WX, winhr_value, "_cad_q", qle*100, "cad", LXMXwindow_name); fi = fi + 1
            }
          }
          for (ami in 1:length(acc.metrics)) {
            # Acceleration in the form of mean, percentiles.
            daysummary[di, fi] = mean(vari[WXi, acc.metrics[ami]]) * 1000
            ds_names[fi] = paste0("ExtFunEvent_", WX, winhr_value, "_cad_mean_",  acc.metrics[ami], "_mg", LXMXwindow_name); fi = fi + 1
            if (length(params_247[["qM5L5"]]) > 0) {
              for (qle in params_247[["qM5L5"]]) {
                daysummary[di, fi] = quantile(x = vari[WXi, acc.metrics[ami]], probs = qle)
                ds_names[fi] = paste0("ExtFunEvent_", WX, winhr_value, "_cad_q", qle*100, "acc", LXMXwindow_name); fi = fi + 1
              }
            }
          }
        }
      }
    }
  }
  invisible(list(ds_names = ds_names, daysummary = daysummary, fi = fi))
}

getSplitNames = function(filename) {
  # extract file split names from filename
  segment_names = NULL
  if (length(grep(pattern = "_split", x =  filename)) > 0) {
    filename_tmp1 = unlist(strsplit(filename, "_split|[.]_split"))
    filename_tmp2 = unlist(strsplit(filename_tmp1[2], "[.]"))
    segment_names = unlist(strsplit(filename_tmp2[1], "_|TO"))
    filename = paste0(filename_tmp1[1], ".", filename_tmp2[2])
  }
  invisible(list(segment_names = segment_names, filename = filename))
}

g.fragmentation = function(frag.metrics = c("mean", "TP", "Gini", "power",
                                            "CoV", "NFragPM", "all"),
                           LEVELS = c(),
                           Lnames=c(), xmin=1,
                           mode = "day") { 
  # This function was originally loosely inspired by R package ActFrag by Junrui Di.
  # In contrast to R package ActFrag this function assumes
  # that non-wear and missing values have already been taken care of (imputed)
  # outside this function or set to NA. Further, the algorithms are not all exactly the same,
  # and there are some additional metrics. For further dicussion see
  # https://wadpac.github.io/GGIR/articles/chapter11_DescribingDataCutPoints.html
  # Further, I am avoiding the term sedentary because sedentary implies
  # that the activity type sitting  was classified, which is generally
  # difficult to justify.
  
  # LEVELS: vector with behavioural classes produced by GGIR
  # Lnames: Names of brehavioural classes.
  # frag.metrics: metric to define fragmentation
  # xmin is shortest recordable (not necessarily observed) boutlength
  
  TransProb = function(x, a = 1, b = c(2,3)) {
    TPab = TPba = NA
    Nab = Nba = 0
    totDur_ab = totDur_ba = 0
    if (length(x) > 0) {
      frag = rle(x)
      Nsegments = length(frag$value)
      if (Nsegments > 1) {
        # Scenario where we have more than 1 segment
        ab = which(frag$value[1:(Nsegments - 1)] %in% a &
                     frag$value[2:Nsegments] %in% b)
        ba = which(frag$value[1:(Nsegments - 1)] %in% b &
                     frag$value[2:Nsegments] %in% a)
        Nab = length(ab)
        Nba = length(ba)
        
        # In GGIR part 6 we set day or spt to NA, by which the series is split in parts
        # separate by strings of NA values. In GGIR part 5 this would not happen
        # and we only need the last item of the time series for that day. However,
        # if last value is NA then ignore it.
        lastitems = which(is.na(frag$value[1:(Nsegments - 1)]) == FALSE &
                                      is.na(frag$value[2:Nsegments] == TRUE))
        if (!is.na(frag$value[Nsegments])) {
          lastitems = unique(c(lastitems, Nsegments))
        }
        tmp = frag$value[-lastitems]
        # Next use this to detect how often last value of each string is reference class (a or b)
        if (length(lastitems) > 0 & length(which(is.na(tmp) == FALSE)) > 1) {
          count_a_ending = sum(ifelse(frag$value[lastitems] %in% a, 1, 0))
          count_b_ending = sum(ifelse(frag$value[lastitems] %in% b, 1, 0))
        } else {
          count_a_ending = 0
          count_b_ending = 0
        }
        # Total duration from a to b
        totDur_ab = sum(frag$length[which(frag$value %in% a)]) - count_a_ending
        totDur_ba = sum(frag$length[which(frag$value %in% b)]) - count_b_ending
        epsilon = 1e-6
        TPab = (Nab + epsilon) / (totDur_ab + epsilon)
        TPba = (Nba + epsilon) / (totDur_ba + epsilon)
        # Round to 6 digits because preceding step introduces bias at 7 decimal places
        TPab = round(TPab, digits = 6)
        TPba = round(TPba, digits = 6)
        durations_ab = frag$length[ab]
        durations_ba = frag$length[ba]
        totDur_ab = sum(durations_ab)
        totDur_ba = sum(durations_ba)
      } else if (Nsegments == 1) {
        # Scenario where we have only one segment of data
        if (frag$value[1] %in% a) { # Only a
          TPab = 0
          TPba = 1
          totDur_ab = frag$length[1]
          totDur_ba = 0
          Nab = 1
          Nba = 0
        } else { # Only b
          TPab = 1
          TPba = 0
          totDur_ab = 0
          totDur_ba = frag$length[1]
          Nab = 0
          Nba = 1
        }
      }
    }
    invisible(list(TPab = TPab,
                   TPba = TPba,
                   Nab = Nab, Nba = Nba,
                   totDur_ab = totDur_ab,
                   totDur_ba = totDur_ba))
  }
  
  if ("all" %in% frag.metrics) {
    frag.metrics = c("mean", "TP", "Gini", "power",
                     "CoV", "NFragPM", "all")
  }
  output = list()
  Nepochs = length(LEVELS)
  if (mode == "day") {
    # convert to class names to numeric class ids for inactive, LIPA and MVPA:
    classes.in = c("day_IN_unbt", Lnames[grep(pattern = "day_IN_bts", x = Lnames)])
    class.in.ids = which(Lnames %in%  classes.in) - 1 
    classes.lig = c("day_LIG_unbt", Lnames[grep(pattern = "day_LIG_bts", x = Lnames)])
    class.lig.ids = which(Lnames %in%  classes.lig) - 1
    classes.mvpa = c("day_MOD_unbt", "day_VIG_unbt", Lnames[grep(pattern = "day_MVPA_bts", x = Lnames)])
    class.mvpa.ids = which(Lnames %in% classes.mvpa) - 1
  } else {
    do.frag = FALSE
  }
  # define expected output to standardize length and output names
  if ("TP" %in% frag.metrics && mode == "day") {
    output[["TP_IN2PA"]] = output[["TP_PA2IN"]] = output[["Nfrag_IN2PA"]] = output[["Nfrag_PA2IN"]] = NA
    output[["TP_IN2LIPA"]] = output[["Nfrag_IN2LIPA"]] = NA
    output[["TP_IN2MVPA"]] = output[["Nfrag_IN2MVPA"]] = NA
    output[["Nfrag_LIPA"]] = output[["mean_dur_LIPA"]] = NA
    output[["Nfrag_MVPA"]] = output[["mean_dur_MVPA"]] = NA
  }
  if (Nepochs > 1 & mode == "day") { # metrics that require more than just binary
    #====================================================
    # Convert LEVELS in three classes: Inactivity (1), Light = LIPA (2), and MVPA (3)
    y = rep(0, Nepochs)
    is.na(y[is.na(LEVELS)]) = TRUE
    y[which(LEVELS %in% class.in.ids)] = 1
    y[which(LEVELS %in% class.lig.ids)] = 2
    y[which(LEVELS %in% class.mvpa.ids)] = 3
    #====================================================
    # TP (transition probability) metrics that depend on multiple classes
    if ("TP" %in% frag.metrics) {
      out = TransProb(y, a = 1, b = c(2, 3)) #IN <-> PA
      output[["TP_IN2PA"]] = out$TPab
      output[["TP_PA2IN"]] = out$TPba
      output[["Nfrag_IN2PA"]] = out$Nab
      output[["Nfrag_PA2IN"]] = out$Nba
      
      out = TransProb(y, a = 1, b = 2) #IN to LIPA
      output[["TP_IN2LIPA"]] = out$TPab
      output[["Nfrag_IN2LIPA"]] = out$Nab

      out = TransProb(y, a = 1, b = 3) #IN to MVPA
      output[["TP_IN2MVPA"]] = out$TPab
      output[["Nfrag_IN2MVPA"]] = out$Nab

      out = TransProb(y, a = 2, b = c(1, 3)) #LIPA <-> rest
      output[["Nfrag_LIPA"]] = out$Nab
      output[["mean_dur_LIPA"]] = out$totDur_ab / out$Nab

      out = TransProb(y, a = 3, b = c(1, 2)) #MVPA <-> rest
      output[["Nfrag_MVPA"]] = out$Nab
      output[["mean_dur_MVPA"]] = out$totDur_ab / out$Nab
    }
  }
  #====================================================
  # Binary fragmentation for the metrics that do not depend on multiple classes
  if (mode == "day") {
    x = rep(0,Nepochs)
    is.na(x[is.na(LEVELS)]) = TRUE
    x[which(LEVELS %in% class.in.ids)] = 1 # inactivity becomes 1 because this is behaviour of interest
    x = as.integer(x)
    frag2levels = rle(x)
    Nfrag2levels = length(which(is.na(frag2levels$values) == FALSE))
    out = TransProb(x, a = 1, b = 0) #IN <-> PA
    output[["Nfrag_PA"]] = out$Nba
    output[["Nfrag_IN"]] = out$Nab
    # Define default values
    if ("mean" %in% frag.metrics) {
      output[["mean_dur_PA"]] = output[["mean_dur_IN"]] = 0
    }
    if ("Gini" %in% frag.metrics) {
      output[["Gini_dur_PA"]] = output[["Gini_dur_IN"]] = NA
    }
    if ("CoV" %in% frag.metrics) { #coefficient of variation
      output[["CoV_dur_PA"]] = output[["CoV_dur_IN"]] = NA
    }
    if ("power" %in% frag.metrics) {
      output[["alpha_dur_PA"]] = output[["alpha_dur_IN"]] = NA
      output[["x0.5_dur_PA"]] = output[["x0.5_dur_IN"]] = NA
      output[["W0.5_dur_PA"]] = output[["W0.5_dur_IN"]] = NA
    }
    if ("power" %in% frag.metrics | "CoV" %in% frag.metrics | "Gini" %in% frag.metrics) {
      output[["SD_dur_PA"]] = output[["SD_dur_IN"]] = NA
    }
    if ("NFragPM" %in% frag.metrics) {
      output[["NFragPM_PA"]] = 0
      output[["NFragPM_IN"]] = 0
    }
    if (Nfrag2levels > 1) {
      if ("mean" %in% frag.metrics) {
        output[["mean_dur_PA"]] = out$totDur_ba / out$Nba
        output[["mean_dur_IN"]] = out$totDur_ab / out$Nab
      }
      if ("NFragPM" %in% frag.metrics) {
        # Identify to metric named fragmentation by Chastin,
        # but renamed into Number of Fragments Per Minutes to be 
        # a better reflection of the calculation
        output[["NFragPM_PA"]] = output[["Nfrag_PA"]] / out$totDur_ba
        output[["NFragPM_IN"]] = output[["Nfrag_IN"]] / out$totDur_ab
      }
      # minimum number of required fragments (10 here this is the sum of the number
      # of PA and IN fragments, so basically we allow for 5 PA and 5 IN fragments)
      # because the metrics below are less informative with a few metrics
      if (Nfrag2levels >= 10) {
        DurationIN = frag2levels$length[which(frag2levels$value == 1)]
        DurationPA = frag2levels$length[which(frag2levels$value == 0)]
        SD0 = sd(DurationPA, na.rm = TRUE)
        SD1 = sd(DurationIN, na.rm = TRUE)
        output[["SD_dur_PA"]] = SD0 # maybe not a fragmentation metric, but helpful to understand other metrics
        output[["SD_dur_IN"]] = SD1
        if ("Gini" %in% frag.metrics) {
          output[["Gini_dur_PA"]] = ineq::Gini(DurationPA,corr = T)
          output[["Gini_dur_IN"]] = ineq::Gini(DurationIN,corr = T)
        }
        if ("CoV" %in% frag.metrics) { #coefficient of variation as described by Blikman 2015
          output[["CoV_dur_PA"]] = sd(log(DurationPA)) / mean(log(DurationPA))
          output[["CoV_dur_IN"]] = sd(log(DurationIN)) / mean(log(DurationIN))
        }
        if ("power" %in% frag.metrics) {
          calc_alpha = function(x, xmin) {
            nr = length(x)
            alpha = 1 + nr/sum(log(x/(xmin))) # adapted to match Chastin 2010 instead of ActFrag
            return(alpha)
          }
          if (SD0 != 0) {
            output[["alpha_dur_PA"]] = calc_alpha(DurationPA, xmin)
            output[["x0.5_dur_PA"]] = 2^(1 / (output[["alpha_dur_PA"]] - 1) * xmin) # according to Chastin 2010
            output[["W0.5_dur_PA"]] = sum(DurationPA[which(DurationPA > output[["x0.5_dur_PA"]])]) / sum(DurationPA)
          }
          if (SD1 != 0) {
            output[["alpha_dur_IN"]] = calc_alpha(DurationIN, xmin)
            output[["x0.5_dur_IN"]] = 2^(1 / (output[["alpha_dur_IN"]] - 1) * xmin) # according to Chastin 2010
            output[["W0.5_dur_IN"]] = sum(DurationIN[which(DurationIN > output[["x0.5_dur_IN"]])]) / sum(DurationIN)
          }
        }
      }
    }
  } else if (mode == "spt") {
    # Binary fragmentation metrics for spt:
    # Active - Rest transitions during SPT:
    x = rep(0, Nepochs)
    is.na(x[is.na(LEVELS)]) = TRUE
    # convert to class names to numeric class ids for inactive, LIPA and MVPA:
    classes.pa = c("spt_wake_LIG", "spt_wake_MOD", "spt_wake_VIG")
    class.pa = which(Lnames %in% classes.pa) - 1
    PAi = which(LEVELS %in% class.pa)
    if (length(PAi) > 0) {
      x[PAi] = 1
    }
    out = TransProb(x = x, a = 0, b = 1)
    output[["Nfrag_IN"]] = out$Nab
    output[["Nfrag_PA"]] = out$Nba
    output[["TP_IN2PA"]] = out$TPab
    output[["TP_PA2IN"]] = out$TPba
    # Wake - Sleep transitions during SPT:
    x = rep(0, Nepochs)
    is.na(x[is.na(LEVELS)]) = TRUE
    classes.wake = c("spt_wake_IN", "spt_wake_LIG", "spt_wake_MOD", "spt_wake_VIG")
    class.wake = which(Lnames %in% classes.wake) - 1 
    wakei = which(LEVELS %in% class.wake)
    if (length(wakei) > 0) {
      x[wakei] = 1
    }
    out = TransProb(x = x, a = 0, b = 1)
    output[["Nfrag_sleep"]] = out$Nab
    output[["Nfrag_wake"]] = out$Nba
    output[["TP_sleep2wake"]] = out$TPab
    output[["TP_wake2sleep"]] = out$TPba
  }
  return(output)
}

library(GGIR)
context("fragmentation")
test_that("fragmentation calculates the expected fragmentation metric values", {
  skip_on_cran()
  
  Lnames = c("spt_sleep", "spt_wake_IN", "spt_wake_LIG", "spt_wake_MOD", "spt_wake_VIG",
             "day_IN_unbt", "day_LIG_unbt", "day_MOD_unbt", "day_VIG_unbt",
             "day_MVPA_bts_10", "day_IN_bts_30",
             "day_IN_bts_10_30", "day_LIG_bts_10")
  
  # Fragmentation of daytime behaviour
  x = c(6, 5, 6, 7, 6, 6, 7, 6, 6, 5, 6, 6, 6, 5, 7, 6, 6, 5, 5, 5, 6, 7, 6,
        6, 6, 6, 7, 6, 5, 5, 5, 5, 5, 6, 6, 6, 5, 6, 6, 6, 6, 6, 6, 6, 6, 6,
        7, 7, 6, 5, 6, 5, 6, 5, rep(12, 11), 5, 6, 6, 6, 5, 6, rep(9, 14), 6,
        5, 7, 7, 6, 7, 7, 7, 6, 6, 6, 5, 6, 5, 5, 5, 6, 5, 5, 5, 5, 5, 5, 5,
        5, 6, 7, 7, 7, 6, 7, 7, 6, 5, 5, 5, 6, 6, 5, 5, 5, 6, 7, 7, 7, 7, 7,
        6, 7, 6, 6, 6, 6, 6, 6, 6, 5, 5, 6, 5, 5, 5, 5, 5, 5, 5, 5, 5, 6, 6,
        5, 5, 5, 5, 5, 5, 6, 6, 6, 6, rep(11, 11), 7, 6, 7, 6, 6, 7, 6, 5, 5,
        6, rep(11, 12), 6, rep(11, 11), 6, 6, 7, 6, 6, 6, 6, rep(11, 12), 6,
        5, 5, 5, 5, 5, 6, 5, 6, 6, 6, 5, 5, 6, 6, 7, 6, 6, 6, 5, 5, 6, 5, 5, 6,
        6, rep(10, 98), 6, 6, 5, 5, 5, 5, 5, 5, 6, 6, 5, 5, 5, 5, 6, rep(11, 15),
        6, rep(11, 21), 6, 6, 6, 6, 6, 6, 5, 6, 5, 6, 5, 12, 12, 12, 12, 12, 12,
        12, 12, 12, 12, 12, 12, 7,  rep(10, 55), 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5,
        6, 6, 6,  rep(11, 22), 6, 6, rep(10, 100), 6, rep(11, 23), 6, 5, 5, 5, 5,
        5, 7, rep(10, 133), 6, rep(10, 119), 6, 6, 5, 6, 6, 6, 6, 6, 7, 7, 7, 6,
        5, 6, 6, 6, 5, 5, 5)

  out = g.fragmentation(frag.metrics = "all",
                        LEVELS = x,
                        Lnames = Lnames, mode = "day")
  
  expect_equal(out$mean_dur_PA, 4.020408, tolerance = 0.0001)
  expect_equal(out$mean_dur_IN, 15.10417, tolerance = 0.0001)
  expect_equal(out$Gini_dur_PA, 0.506352, tolerance = 0.0001)
  expect_equal(out$Gini_dur_IN, 0.768544, tolerance = 0.0001)
  expect_equal(out$alpha_dur_PA, 2.055102, tolerance = 0.0001)
  expect_equal(out$x0.5_dur_PA, 1.928897, tolerance = 0.0001)
  expect_equal(out$W0.5_dur_IN, 0.9629121, tolerance = 0.0001)
  
  expect_equal(out$CoV_dur_IN, 1.035708, tolerance = 0.0001)
  expect_equal(out$CoV_dur_PA, 0.9859804, tolerance = 0.0001)
  
  expect_equal(out$TP_IN2MVPA, 0.005502, tolerance = 0.0001)
  expect_equal(out$TP_IN2LIPA, 0.060523, tolerance = 0.0001)
  expect_equal(out$TP_IN2PA, 0.066025, tolerance = 0.0001)
  expect_equal(out$Nfrag_PA, 49)
  expect_equal(out$Nfrag_IN2LIPA, 44)
  expect_equal(out$Nfrag_IN2MVPA, 4)
  expect_equal(out$Nfrag_IN, 48)
  
  # Fragmentation of nighttime (SPT) behaviour
  # classes: wake 1, sleep 0
  x = c(rep(1, 15), rep(0, 10), rep(1, 30),
        rep(0, 15), rep(1, 20), rep(0, 25), rep(1, 10))
  
  out = g.fragmentation(frag.metrics = "all",
                        LEVELS = x,
                        Lnames = Lnames, mode = "spt")
  expect_equal(out$Nfrag_wake, 3)
  expect_equal(out$Nfrag_sleep, 3)
  expect_equal(out$TP_wake2sleep, 0.040541, tolerance = 0.0001) # 3+epsilon / 65+epsilon
  expect_equal(out$TP_sleep2wake, 0.06, tolerance = 0.0001) # 3+epsilon / 50+epsilon

  # SPT: Test ability to handle NA blocks as needed for g.part6
  # classes: wake 1, sleep 0
  x = c(rep(1, 5), rep(0, 5), rep(1, 5), rep(0, 5), rep(NA, 15),
        rep(0, 5), rep(1, 5), rep(0, 5), rep(1, 5), rep(NA, 15),
        rep(1, 5), rep(0, 5), rep(1, 5), rep(0, 5))

  out = g.fragmentation(frag.metrics = "all",
                        LEVELS = x,
                        Lnames = Lnames, mode = "spt")
  expect_equal(out$Nfrag_sleep, 4)
  expect_equal(out$Nfrag_wake, 5)
  expect_equal(out$TP_sleep2wake, 0.142857, tolerance = 0.0001)
  expect_equal(out$TP_wake2sleep, 0.172414, tolerance = 0.0001)
  
  
  # day: Test ability to handle NA blocks as needed for g.part6
  # 9 is MVPA, 10 is IN
  x = c(rep(10, 5), rep(9, 5), rep(10, 5), rep(9, 5), rep(NA, 15),
        rep(9, 5), rep(10, 5), rep(9, 5), rep(10, 5), rep(NA, 15),
        rep(10, 5), rep(9, 5), rep(10, 5), rep(9, 5))
  
  out = g.fragmentation(frag.metrics = "all",
                        LEVELS = x,
                        Lnames = Lnames, mode = "day")
  expect_equal(out$Nfrag_PA2IN, 4) # 4+epsilon/20+epsilon
  expect_equal(out$Nfrag_IN2PA, 5) # 5+epsilon/25+epsilon
  expect_equal(out$TP_PA2IN, 0.142857, tolerance = 0.0001)
  expect_equal(out$TP_IN2PA, 0.172414, tolerance = 0.0001)
  
  
  # same but now with more variation in segment length
  x = c(rep(10, 4), rep(9, 6), rep(10, 9), rep(9, 2), rep(NA, 12),
        rep(9, 8), rep(10, 5), rep(9, 3), rep(10, 20), rep(NA, 19),
        rep(10, 3), rep(9, 2), rep(10, 7), rep(9, 7))
  
  out = g.fragmentation(frag.metrics = "all",
                        LEVELS = x,
                        Lnames = Lnames, mode = "day")
  expect_equal(out$Nfrag_PA2IN, 4) # 4+epsilon/20+epsilon
  expect_equal(out$Nfrag_IN2PA, 5) # 5+epsilon/25+epsilon
  expect_equal(out$TP_PA2IN, 0.153846, tolerance = 0.0001)
  expect_equal(out$TP_IN2PA, 0.106383, tolerance = 0.0001)
  
})

detectEventBouts = function(myfun, varnum_event, varnum, 
                            UnitReScale, daysummary, ds_names,
                            di, fi, ws3, boutnameEnding) {
  if ("ebout.dur" %in% names(myfun) == FALSE) myfun$ebout.dur = c(1, 5, 10)
  if ("ebout.th.cad" %in% names(myfun) == FALSE) myfun$ebout.th.cad = 30
  if ("ebout.th.acc" %in% names(myfun) == FALSE) myfun$ebout.th.acc = 50
  if ("ebout.criter" %in% names(myfun) == FALSE) myfun$ebout.criter = 0.8
  if ("ebout.condition" %in% names(myfun) == FALSE) myfun$ebout.condition = "AND"
  
  # varnum = metashort[anwindices, mi]
  cadence = varnum_event * (60/ws3)
  # Event bouts
  # Loop over bout durations
  for (boutdur in myfun$ebout.dur) {
    boutduration = boutdur * (60/ws3) # per minute
    
    # Loop over cadence thresholds
    for (cad_th in myfun$ebout.th.cad) {
      
      # Loop over acceleration thresholds
      for (acc_th in myfun$ebout.th.acc) {
        rr1 = matrix(0, length(varnum), 1)
        if (myfun$ebout.condition == "AND") {
          p = which(varnum * UnitReScale >= acc_th &
                      cadence >= cad_th); rr1[p] = 1
        } else if (myfun$ebout.condition == "OR") {
          p = which(varnum * UnitReScale >= acc_th |
                      cadence >= cad_th); rr1[p] = 1
        }
        getboutout = g.getbout(x = rr1, boutduration = boutduration,
                               boutcriter = myfun$ebout.criter,
                               ws3 = ws3)
        # time spent in bouts in minutes
        eventbout = length(which(getboutout == 1)) / (60/ws3)
        eboutname = paste0("ExtFunEvent_totdur_B", boutdur,
                           "M", (myfun$ebout.criter  * 100),
                           "%_cad",cad_th, myfun$ebout.condition,
                           "acc", acc_th)
        ebout_varname = paste0(eboutname, "_", boutnameEnding)
        # fi = correct_fi(di, ds_names, fi, varname = ebout_varname)
        daysummary[di,fi] = eventbout # total time in ebouts
        ds_names[fi] = ebout_varname;
        fi = fi + 1
        # number of bouts
        rle_bout = rle(as.numeric(getboutout))
        rle_bout1 = which(rle_bout$values == 1)
        number_of_bouts = length(rle_bout1)
        
        eboutname = paste0("ExtFunEvent_number_B", boutdur,
                           "M", (myfun$ebout.criter  * 100),
                           "%_cad",cad_th, myfun$ebout.condition,
                           "acc", acc_th)
        ebout_varname = paste0(eboutname, "_", boutnameEnding)
        daysummary[di,fi] = number_of_bouts
        ds_names[fi] = ebout_varname;
        fi = fi + 1
        # average bout duration in minutes
        if (number_of_bouts > 0) {
          mn_dur_bouts = mean(rle_bout$lengths[which(rle_bout$values == 1)]) / (60/ws3)
        } else {
          mn_dur_bouts = 0
        }
        eboutname = paste0("ExtFunEvent_meandur_B", boutdur,
                           "M", (myfun$ebout.criter  * 100),
                           "%_cad",cad_th, myfun$ebout.condition,
                           "acc", acc_th)
        ebout_varname = paste0(eboutname, "_", boutnameEnding)
        daysummary[di,fi] = mn_dur_bouts
        ds_names[fi] = ebout_varname;
        fi = fi + 1
      }
    }
  }
  invisible(list(daysummary = daysummary, ds_names = ds_names,
                 fi = fi, di = di))
}

library(GGIR)
context("detectEventBouts")
test_that("Detect bouts in events", {
  skip_on_cran()
  
  # create dummy test data
  metashort = data.frame(ENMO = c(rep(0, 24), rep(0.12, 42), rep(0, 24)),
                         step_count = c(rep(0, 24), rep(5, 42), rep(0, 24)))
  cn_metashort = colnames(metashort)
  anwindices = 1:84
  anwi_index = 1
  varnum = metashort$step_count[anwindices]
  varnum_event = rep(0, length(varnum))
  varnum_event[which(metashort$ENMO[anwindices] > 0.05)] = 10
  ws3 = 5
  anwi_nameindices = "_1234hrs"
  daysummary = matrix("", 1, 50)
  fi = 1
  di = 1
  ds_names = ""
  myfun = list(ilevels = c(0, 50, 100),
               clevels = c(0, 30, 50),
               qlevels = c(0.25, 0.5, 0.75),
               ebout.dur = c(1, 5, 10),
               ebout.th.cad = c(30, 100),
               ebout.th.acc = c(50, 100),
               ebout.criter = 1,
               ebout.condition = "AND")

  # run function  
  eventBouts = detectEventBouts(myfun, varnum_event = varnum_event,
                                varnum = varnum, 
                                UnitReScale = 1000,
                                daysummary = daysummary,
                                ds_names = ds_names,
                                di = di, fi = fi,
                                ws3 = ws3,
                                boutnameEnding = "_anyRandomText")
  
  
  # all bouts of 1 minute have the same estimates as they all meet the cadence and acc criteria
  expect_equal(as.numeric(eventBouts$daysummary)[1:12], c(rep(c(3.5, 1.0, 3.5), 4)))
  
  # expect different estimates are calculated
  expect_true("ExtFunEvent_number_B5M100%_cad30ANDacc50__anyRandomText" %in% eventBouts$ds_names)
  expect_true("ExtFunEvent_totdur_B1M100%_cad30ANDacc100__anyRandomText" %in% eventBouts$ds_names)
  expect_true("ExtFunEvent_totdur_B1M100%_cad100ANDacc100__anyRandomText" %in% eventBouts$ds_names)
  expect_true("ExtFunEvent_totdur_B1M100%_cad100ANDacc50__anyRandomText" %in% eventBouts$ds_names)
  
  # Bouts of 5 or 10 minutes are undetected with any cadence and acc criteria
  expect_equal(sum(as.numeric(eventBouts$daysummary)[13:36]), 0)
  
})

g.readaccfile = function(filename, blocksize, blocknumber, filequality,
                         ws, PreviousEndPage = 1, inspectfileobject = c(),
                         PreviousLastValue = c(0, 0, 1), PreviousLastTime = NULL,
                         params_rawdata = c(), params_general = c(), header = NULL, ...) {
  #get input variables
  input = list(...)
  if (length(input) > 0 ||
      length(params_rawdata) == 0 || length(params_general) == 0) {
    # Extract and check parameters if user provides more arguments than just the parameter arguments,
    # or if params_[...] aren't specified (so need to be filled with defaults).
    # So, inside GGIR this will not be used, but it is used when g.getmeta is used on its own
    # as if it was still the old g.getmeta function
    params = extract_params(params_rawdata = params_rawdata,
                            params_general = params_general,
                            input = input) # load default parameters
    params_rawdata = params$params_rawdata
    params_general = params$params_general
  }
  
  desiredtz = params_general[["desiredtz"]]
  configtz = params_general[["configtz"]]
  if (length(configtz) == 0) configtz = desiredtz

  I = inspectfileobject
  mon = I$monc
  if (mon == MONITOR$VERISENSE) mon = MONITOR$ACTIGRAPH
  dformat = I$dformc
  sf = I$sf
  decn = I$decn

  if ((mon == MONITOR$ACTIGRAPH && dformat == FORMAT$CSV) ||
      (mon == MONITOR$AXIVITY && dformat == FORMAT$CSV) || 
      dformat == FORMAT$AD_HOC_CSV) {
    blocksize = blocksize * 300
  }

  if (blocknumber < 1) blocknumber = 1

  # startpage should only be specified for blocknumber 1.
  # The next time (blocknumber > 1) the startpage will be derived from the previous
  # endpage and the blocksize.

  if ((mon == MONITOR$GENEACTIV && dformat == FORMAT$BIN) || dformat == FORMAT$GT3X ||
      (mon == MONITOR$MOVISENS && dformat == FORMAT$BIN) || 
      (mon == MONITOR$PARMAY_MTX && dformat == FORMAT$BIN)) {
    # for GENEActiv binary data, gt3x format data, Movisens data, and Parmay Matrix data, 
    # page selection is defined from start to end **including end**
    if (blocknumber > 1 && length(PreviousEndPage) != 0) {
      startpage = PreviousEndPage + 1
    } else {
      startpage = blocksize * (blocknumber - 1) + 1 # pages are numbered starting with page 1
    }
    endpage = startpage + blocksize - 1 # -1 because both startpage and endpage will be read,
                                        # and we want to read blocksize # of samples
  } else {
    # for other monitor brands and data formats
    # page selection is defined from start to end **excluding end itself**,
    # so start page of one block equals the end page of previous block
    if (blocknumber > 1 && length(PreviousEndPage) != 0) {
      startpage = PreviousEndPage
    } else {
      startpage = blocksize * (blocknumber - 1) # pages are numbered starting with page 0

      if (mon == MONITOR$ACTIGRAPH && dformat == FORMAT$CSV) {
        headerlength = 10
        startpage = startpage + headerlength
      }
    }
    endpage = startpage + blocksize
  }

  P = c()
  isLastBlock = FALSE

  if (mon == MONITOR$GENEACTIV && dformat == FORMAT$BIN) {    
    try(expr = {P = GGIRread::readGENEActiv(filename = filename, start = startpage,
                                            end = endpage, desiredtz = desiredtz,
                                            configtz = configtz)}, silent = TRUE)
    if (length(P) > 0 && ("data.out" %in% names(P))) {
      names(P)[names(P) == "data.out"] = "data"
      
      # If sampling rate is fractional which is possible for GENEActiv, 
      # resample to integer sampling rate to ensure downstream GGIR code
      # can derive non-ambiguous epoch level aggregates.
      if (P$header$SampleRate != round(P$header$SampleRate)) {
        coln_original = colnames(P$data)
        rawData = P$data[, 2:ncol(P$data)]
        rawTime = P$data$time
        P$header$SampleRate = round(P$header$SampleRate)
        if (sf != P$header$SampleRate) stop("sampling rate inconsistency, please contact maintainer")
        rawAccel = as.matrix(rawData)
        step = 1/sf
        timeRes = seq(rawTime[1], rawTime[length(rawTime)], step)
        timeRes = timeRes[1 : (length(timeRes) - 1)]
        accelRes = GGIRread::resample(rawAccel, rawTime, timeRes, nrow(rawAccel), params_rawdata[["interpolationType"]]) # this is now the resampled acceleration data
        P$data = data.frame(timeRes, accelRes)
        colnames(P$data) = coln_original
      }
      if (nrow(P$data) < (blocksize*300)) {
        isLastBlock = TRUE
      }
    }
  } else if (mon == MONITOR$ACTIGRAPH && dformat == FORMAT$CSV) {    
    # load rows 11:13  to investigate whether the file has a header
    # invisible because R complains about poor Actigraph file format,
    # this is an an ActiGraph problem not a GGIR problem, so we ignore it
    quiet <- function(x) {
      # from https://stackoverflow.com/a/54136863/5311763
      sink(tempfile())
      on.exit(sink())
      invisible(force(x))
    }

    # skip 1 more row only if the file has a header. Only the first chunk of data can have a header.
    if (blocknumber == 1) {
      testheader =  quiet(data.table::fread(filename, nrows = 2, skip = 10,
                                            dec = decn, showProgress = FALSE,
                                            header = TRUE, data.table=FALSE, stringsAsFactors=FALSE))
      if (suppressWarnings(is.na(as.numeric(colnames(testheader)[1])))) { # first value is *not* a number, so file starts with a header
        startpage = startpage + 1
        endpage = endpage + 1
      }
    }
    
    try(expr = {
      P$data = quiet(data.table::fread(filename, nrows = blocksize, skip = startpage,
                                       dec = decn, showProgress = FALSE,
                                       header = FALSE, # header should always be FALSE to prevent acceleration values from being mistaken for header when reading chunks 2 and on
                                       data.table=FALSE, stringsAsFactors=FALSE))
    }, silent = TRUE)
    if (length(P$data) > 0) {
      if (ncol(P$data) < 3) {
        P$data = c()
      } else {
        if (ncol(P$data) > 3) {
          P$data = P$data[, 2:4] # remove timestamp column, keep only XYZ columns
        }
        colnames(P$data) = c("x", "y", "z")
      }
    }
  } else if (mon == MONITOR$AXIVITY && dformat == FORMAT$CWA) {
    if (utils::packageVersion("GGIRread") < "0.3.0") {
      # ignore frequency_tol parameter
      apply_readAxivity = function(bstart, bend) {
        try(expr = {P = GGIRread::readAxivity(filename = filename, start = bstart, end = bend,
                                              progressBar = FALSE,
                                              desiredtz = desiredtz,
                                              configtz = configtz,
                                              interpolationType = params_rawdata[["interpolationType"]],
                                              header = header)
            }, silent = TRUE)
        return(P)
      }
    } else {
      # pass on frequency_tol parameter to GGIRread::readAxivity function
      apply_readAxivity = function(bstart, bend) {
        try(expr = {P = GGIRread::readAxivity(filename = filename, start = bstart, end = bend, 
                                              progressBar = FALSE,
                                              desiredtz = desiredtz,
                                              configtz = configtz,
                                              interpolationType = params_rawdata[["interpolationType"]],
                                              frequency_tol = params_rawdata[["frequency_tol"]],
                                              header = header)
            }, silent = TRUE)
        return(P)
      }
    }

    P = apply_readAxivity(bstart = startpage, bend = endpage)
    if (length(P) == 0) { 
      # If data reading is not successful then try following steps to retrieve issue
      # I am not sure if this is still relevant after all the improvements to GGIRread
      # but leaving this in just in case it is still needed
      
      PtestLastPage = PtestStartPage = NULL
      # Try to read the last page of the block because if it exists then there might be
      # something wrong with the first page(s).
      PtestLastPage = apply_readAxivity(bstart = endpage, bend = endpage)
      if (length(PtestLastPage) > 1) { 
        # Last page exist, so there must be something wrong with the first page
        NFilePagesSkipped = 0
        while (length(PtestStartPage) == 0) { # Try loading the first page of the block by iteratively skipping a page
          NFilePagesSkipped = NFilePagesSkipped + 1
          startpage = startpage + NFilePagesSkipped
          PtestStartPage = apply_readAxivity(bstart = startpage, bend = startpage)
          if (NFilePagesSkipped == 10 & length(PtestStartPage) == 0) PtestStartPage = FALSE # stop after 10 attempts
        }
      }
      if (length(PtestStartPage) > 1) {
        # Now we know on which page we can start and end the block, we can try again to
        # read the entire block:
        P = apply_readAxivity(bstart = startpage, bend = endpage)
        if (length(P) > 1) { # data reading succesful
          filequality$NFilePagesSkipped = NFilePagesSkipped # store number of pages jumped

          # Add replications of Ptest to the beginning of P to achieve same data 
          # length as under normal conditions
          P$data = rbind(do.call("rbind",
                                 replicate(NFilePagesSkipped, PtestStartPage$data, simplify = FALSE)),
                         P$data)
        }
      }
    }
    if ("temp" %in% colnames(P$data)) {
      colnames(P$data)[colnames(P$data) == "temp"] = "temperature"
    }
  } else if (mon == MONITOR$AXIVITY && dformat == FORMAT$CSV) {
    try(expr = {
      rawData = data.table::fread(filename, nrows = blocksize,
                            skip = startpage,
                            dec = decn, showProgress = FALSE, header = FALSE,
                            data.table=FALSE, stringsAsFactors=FALSE)
    }, silent = TRUE)
    if (length(rawData) > 0) {
      if (nrow(rawData) < blocksize) {
        isLastBlock = TRUE
      }

      rawTime = rawData[,1]

      if(class(rawTime)[1] == "character") {
        # If timestamps in the csv file are formatted (Y-M-D h:m:s.f), but some of the dates are formatted poorly,
        # data.table::fread() might have trouble parsing them as timestamps, and will instead return them as strings. 
        # as.POSIXct() is less brittle and might still be able to parse such timestamps, but it will be a very slow process.
        # For instance, one omGUI export contained timestamps like "2023-11-11 15:22:60.000" (which should normally be "2023-11-11 15:23:00.000").
        # data.table::fread() couldn't parse this but as.POSIXct() could, albeit very slowly.

        rawTime = as.POSIXct(rawTime, tz=configtz, origin = "1970-01-01")

        # if as.POSIXct() also failed to parse this data as timestamps, there's nothing else we can do.
        if(class(rawTime)[1] != "POSIXct") {
          stop(paste0("Corrupt timestamp data in ", filename),  call. = FALSE)
        } else {
          warning(paste0("Corrupt timestamp data in ", filename, 
                         ". This will greatly slow down processing. To avoid this, use the original .cwa file, ",
                         "or export your data with Unix timestamps instead."),  call. = FALSE)
        }
      } else {
        # If timestamps in the csv file are formatted (Y-M-D h:m:s.f), data.table::fread assumes them to be
        # in the timezone of the current device (i.e. as if configtz == "").
        # If this is not the case, we need to force the correct timezone (force, as opposed to converting to that timezone).
        if (!is.numeric(rawTime) && configtz != "") {
          rawTime = lubridate::force_tz(rawTime, configtz)
        }

        # A similar thing needs to be done if the csv file contains Unix timestamps as well.
        # OmGui converts device timestamps to Unix timestamps as if the timestamps were originally in UTC.
        # So we need to convert the Unix timestamp into a hh:mm:ss format, then force that timestamp
        # from  UTC into configtz timzone.
        if (is.numeric(rawTime)) {
          rawTime = as.POSIXct(rawTime, tz="UTC", origin = "1970-01-01")
          rawTime = lubridate::force_tz(rawTime, configtz)
        }
      }

      rawTime = as.numeric(rawTime)

      # resample the acceleration data, because AX3 data is stored at irregular time points
      rawAccel = as.matrix(rawData[,2:4])
      step = 1/sf
      timeRes = seq(rawTime[1], rawTime[length(rawTime)], step)
      timeRes = timeRes[1 : (length(timeRes) - 1)]

      # at the moment the function is designed for reading the 3 acceleration channels only,
      # because that is the situation of the use-case we had.
      accelRes = GGIRread::resample(rawAccel, rawTime, timeRes, nrow(rawAccel), params_rawdata[["interpolationType"]]) # this is now the resampled acceleration data
      P$data = data.frame(timeRes, accelRes)
      colnames(P$data) = c("time", "x", "y", "z")
    }
  } else if (mon == MONITOR$MOVISENS && dformat == FORMAT$BIN) {
    file_length = unisensR::getUnisensSignalSampleCount(dirname(filename), "acc.bin")
    if (endpage > file_length) {
      endpage = file_length
      isLastBlock = TRUE
    }
    P$data = unisensR::readUnisensSignalEntry(dirname(filename), "acc.bin",
                                              startIndex = startpage,
                                              endIndex = endpage)
    if (length(P$data) > 0) {
      if (ncol(P$data) < 3) {
        P$data = c()
      } else {
        colnames(P$data) = c("x", "y", "z")
        # there may or may not be a temp.bin file containing temperature
        try(expr = {P$data$temperature = g.readtemp_movisens(filename,
                                                             from = startpage, to = endpage,
                                                             acc_sf = sf, acc_length = nrow(P$data),
                                                             interpolationType = params_rawdata[["interpolationType"]])
        }, silent = TRUE)
      }
    } 
  } else if (mon == MONITOR$ACTIGRAPH && dformat == FORMAT$GT3X) {
    P$data = try(expr = {read.gt3x::read.gt3x(path = filename, batch_begin = startpage,
                                              batch_end = endpage, asDataFrame = TRUE)}, silent = TRUE)
    if (length(P$data) == 0 || inherits(P$data, "try-error") == TRUE) { # too short or no data at all
      P$data = c()
    } else { # If data passes these checks then it is usefull
      colnames(P$data)[colnames(P$data) == "X"] = "x"
      colnames(P$data)[colnames(P$data) == "Y"] = "y"
      colnames(P$data)[colnames(P$data) == "Z"] = "z"

      # read.gt3x::read.gt3x returns timestamps as POSIXct with GMT timezone, but they are actally in local time of the device.
      # Don't just convert timezones, instead force the correct local timezone of the device (configtz)
      # while keeping the same hh:mm:ss time.
      P$data$time = lubridate::force_tz(P$data$time, configtz)
    }
  } else if (mon == MONITOR$AD_HOC && dformat == FORMAT$AD_HOC_CSV) { # user-specified csv format
    # skip 1 more row only if rmc.firstrow.acc points at a row containing column names.
    # This is only relevant for the first chunk of data.
    if (blocknumber == 1) {
      testheader =  data.table::fread(filename, nrows = 2, skip = params_rawdata[["rmc.firstrow.acc"]]-1,
                                      dec = decn, showProgress = FALSE,
                                      header = TRUE, data.table=FALSE, stringsAsFactors=FALSE)
      if (suppressWarnings(is.na(as.numeric(colnames(testheader)[1])))) { # first value is *not* a number, so file starts with a header
        startpage = startpage + 1
        endpage = endpage + 1
      }
    }

    try(expr = {P = read.myacc.csv(rmc.file = filename,
                                   rmc.nrow = blocksize, rmc.skip = startpage,
                                   rmc.dec = params_rawdata[["rmc.dec"]],
                                   rmc.firstrow.acc = params_rawdata[["rmc.firstrow.acc"]],
                                   rmc.firstrow.header = params_rawdata[["rmc.firstrow.header"]],
                                   rmc.header.length = params_rawdata[["rmc.header.length"]],
                                   rmc.col.acc = params_rawdata[["rmc.col.acc"]],
                                   rmc.col.temp = params_rawdata[["rmc.col.temp"]],
                                   rmc.col.time = params_rawdata[["rmc.col.time"]],
                                   rmc.unit.acc = params_rawdata[["rmc.unit.acc"]],
                                   rmc.unit.temp = params_rawdata[["rmc.unit.temp"]],
                                   rmc.unit.time = params_rawdata[["rmc.unit.time"]],
                                   rmc.format.time = params_rawdata[["rmc.format.time"]],
                                   rmc.bitrate = params_rawdata[["rmc.bitrate"]],
                                   rmc.dynamic_range = params_rawdata[["rmc.dynamic_range"]],
                                   rmc.unsignedbit = params_rawdata[["rmc.unsignedbit"]],
                                   rmc.origin = params_rawdata[["rmc.origin"]],
                                   rmc.desiredtz = params_rawdata[["rmc.desiredtz"]],
                                   rmc.configtz = params_rawdata[["rmc.configtz"]],
                                   rmc.sf = params_rawdata[["rmc.sf"]],
                                   rmc.headername.sf = params_rawdata[["rmc.headername.sf"]],
                                   rmc.headername.sn = params_rawdata[["rmc.headername.sn"]],
                                   rmc.headername.recordingid = params_rawdata[["rmc.headername.sn"]],
                                   rmc.header.structure = params_rawdata[["rmc.header.structure"]],
                                   rmc.check4timegaps = params_rawdata[["rmc.check4timegaps"]],
                                   rmc.col.wear = params_rawdata[["rmc.col.wear"]],
                                   rmc.doresample = params_rawdata[["rmc.doresample"]],
                                   rmc.scalefactor.acc = params_rawdata[["rmc.scalefactor.acc"]],
                                   interpolationType = params_rawdata[["interpolationType"]],
                                   PreviousLastValue = PreviousLastValue,
                                   PreviousLastTime = PreviousLastTime,
                                   desiredtz = desiredtz,
                                   configtz = configtz,
                                   header = header)
    }, silent = TRUE)
    if (length(sf) == 0) sf = params_rawdata[["rmc.sf"]]
  } else if (mon == MONITOR$PARMAY_MTX && dformat == FORMAT$BIN) {
    try(expr = {P = GGIRread::readParmayMatrix(filename = filename, output = "all",
                                               start = startpage, end = endpage, 
                                               desiredtz = desiredtz, configtz = configtz,
                                               interpolationType = params_rawdata[["interpolationType"]])}, silent = TRUE)
    # fix colnames to match expectations of GGIR
    colnames(P$data) = gsub("acc_", "", colnames(P$data))
    colnames(P$data) = gsub("ambient_temp", "temperature", colnames(P$data))
    if (P$lastchunk) {
      isLastBlock = TRUE
    }
  } 

  # if first block isn't read then the file is probably corrupt
  if (length(P$data) <= 1 || nrow(P$data) == 0) {
    P = c()
    isLastBlock = TRUE
    if (blocknumber == 1) {
      warning('\nFile empty, possibly corrupt.\n')
      filequality$filetooshort = TRUE
      filequality$filecorrupt = TRUE
    }
  } else if (nrow(P$data) < (sf * ws * 2 + 1)) {
    # a shorter chunk of data than expected was read
    isLastBlock = TRUE

    if (blocknumber == 1) {
      # not enough data for analysis
      P = c()
      filequality$filetooshort = TRUE
    }
  }

  # remove any columns we don't need/expect
  P$data = P$data[,which(colnames(P$data) %in% c("x", "y", "z", "time", "light", "temperature", "wear"))]

  # every column except for time and wear should be numeric.
  # If it isn't, some non-numeric input was present, and we don't know how to deal with it.
  for (col in c("x", "y", "z", "light", "temperature")) {
    if ((col %in% colnames(P$data)) && !is.numeric(P$data[, col])) {
      stop(paste0("Corrupt file. ", col, " column contains non-numeric data."))
    }
  }

  # the wear column should be logical, but we will coerse it to numeric right away,
  # so that later it could be combined with the other numeric columns into a numeric matrix

  if ("wear" %in% colnames(P$data)) {
    if (!is.logical(P$data$wear)) {
      stop("Corrupt file. The wear column should contail TRUE/FALSE values.")
    }
    P$data$wear = as.numeric(P$data$wear)
  }

  # the time column at this point will be either Unix timestamps or POSIXct objects.
  # If POSIXct, we'll convert them to Unix timestamps, so that later this column 
  # could be combined with the other numeric columns into a numeric matrix
  if (("time" %in% colnames(P$data)) && !is.numeric(P$data$time)) { 
    P$data$time = as.numeric(P$data$time)
  }

  invisible(list(P = P,
                 filequality = filequality,
                 isLastBlock = isLastBlock,
                 endpage = endpage,  startpage = startpage))
}

g.getmeta = function(datafile, params_metrics = c(), params_rawdata = c(),
                     params_general = c(), params_cleaning = c(), daylimit = FALSE,
                     offset = c(0, 0, 0), scale = c(1, 1, 1), tempoffset = c(0, 0, 0),
                     meantempcal = c(), myfun = c(),
                     inspectfileobject = c(),
                     verbose = TRUE, ...) {
  
  #get input variables
  input = list(...)
  if (length(input) > 0 ||
      length(params_metrics) == 0 || length(params_rawdata) == 0 ||
      length(params_general) == 0 || length(params_cleaning) == 0) {
    # Extract and check parameters if user provides more arguments than just the parameter arguments,
    # or if params_[...] aren't specified (so need to be filled with defaults).
    # So, inside GGIR this will not be used, but it is used when g.getmeta is used on its own
    # as if it was still the old g.getmeta function
    params = extract_params(params_metrics = params_metrics,
                            params_rawdata = params_rawdata,
                            params_general = params_general,
                            params_cleaning = params_cleaning,
                            input = input,
                            params2check = c("metrics", "rawdata", "general", "cleaning")) # load default parameters
    params_metrics = params$params_metrics
    params_rawdata = params$params_rawdata
    params_general = params$params_general
    params_cleaning = params$params_cleaning
  }
  
  metrics2do = data.frame(do.bfen = params_metrics[["do.bfen"]],
                          do.enmo = params_metrics[["do.enmo"]],
                          do.lfenmo = params_metrics[["do.lfenmo"]],
                          do.en = params_metrics[["do.en"]],
                          do.hfen = params_metrics[["do.hfen"]],
                          do.hfenplus = params_metrics[["do.hfenplus"]],
                          do.mad = params_metrics[["do.mad"]],
                          do.anglex = params_metrics[["do.anglex"]],
                          do.angley = params_metrics[["do.angley"]],
                          do.anglez = params_metrics[["do.anglez"]],
                          do.roll_med_acc_x = params_metrics[["do.roll_med_acc_x"]],
                          do.roll_med_acc_y = params_metrics[["do.roll_med_acc_y"]],
                          do.roll_med_acc_z = params_metrics[["do.roll_med_acc_z"]],
                          do.dev_roll_med_acc_x = params_metrics[["do.dev_roll_med_acc_x"]],
                          do.dev_roll_med_acc_y = params_metrics[["do.dev_roll_med_acc_y"]],
                          do.dev_roll_med_acc_z = params_metrics[["do.dev_roll_med_acc_z"]],
                          do.enmoa = params_metrics[["do.enmoa"]],
                          do.lfen = params_metrics[["do.lfen"]],
                          do.lfx = params_metrics[["do.lfx"]],
                          do.lfy = params_metrics[["do.lfy"]],
                          do.lfz = params_metrics[["do.lfz"]],
                          do.hfx = params_metrics[["do.hfx"]],
                          do.hfy = params_metrics[["do.hfy"]],
                          do.hfz = params_metrics[["do.hfz"]],
                          do.bfx = params_metrics[["do.bfx"]],
                          do.bfy = params_metrics[["do.bfy"]],
                          do.bfz = params_metrics[["do.bfz"]],
                          do.zcx = params_metrics[["do.zcx"]],
                          do.zcy = params_metrics[["do.zcy"]],
                          do.zcz = params_metrics[["do.zcz"]],
                          do.brondcounts = params_metrics[["do.brondcounts"]],
                          do.neishabouricounts = params_metrics[["do.neishabouricounts"]],
                          stringsAsFactors = FALSE)
  
  nmetrics = sum(c(params_metrics[["do.bfen"]], params_metrics[["do.enmo"]],
                   params_metrics[["do.lfenmo"]], params_metrics[["do.en"]],
                   params_metrics[["do.hfen"]], params_metrics[["do.hfenplus"]],
                   params_metrics[["do.mad"]], params_metrics[["do.anglex"]],
                   params_metrics[["do.angley"]], params_metrics[["do.anglez"]],
                   params_metrics[["do.roll_med_acc_x"]], params_metrics[["do.roll_med_acc_y"]],
                   params_metrics[["do.roll_med_acc_z"]],
                   params_metrics[["do.dev_roll_med_acc_x"]], params_metrics[["do.dev_roll_med_acc_y"]],
                   params_metrics[["do.dev_roll_med_acc_z"]],
                   params_metrics[["do.enmoa"]], params_metrics[["do.lfen"]],
                   params_metrics[["do.lfx"]], params_metrics[["do.lfy"]],
                   params_metrics[["do.lfz"]],  params_metrics[["do.hfx"]],
                   params_metrics[["do.hfy"]], params_metrics[["do.hfz"]],
                   params_metrics[["do.bfx"]], params_metrics[["do.bfy"]],
                   params_metrics[["do.bfz"]],
                   params_metrics[["do.zcx"]], params_metrics[["do.zcy"]],
                   params_metrics[["do.zcz"]], params_metrics[["do.brondcounts"]] * 3,
                   params_metrics[["do.neishabouricounts"]] * 4))
  if (length(myfun) != 0) {
    nmetrics = nmetrics + length(myfun$colnames)
    # check myfun object already, because we do not want to discover
    # bugs after waiting for the data to load
    check_myfun(myfun, params_general[["windowsizes"]])
  }
  
  if (length(nmetrics) == 0) {
    warning("No metrics selected.", call. = FALSE)
  }

  ws3 = params_general[["windowsizes"]][1]; ws2 = params_general[["windowsizes"]][2]; ws = params_general[["windowsizes"]][3]

  PreviousEndPage = c()
  
  filequality = data.frame(filetooshort = FALSE, filecorrupt = FALSE,
                           filedoesnotholdday = FALSE, NFilePagesSkipped = 0)
  filetooshort = FALSE
  filecorrupt = FALSE
  filedoesnotholdday = FALSE
  NFilePagesSkipped = 0
  
  i = 1 #counter to keep track of which binary block is being read
  count = 1 #counter to keep track of the number of seconds that have been read
  count2 = 1 #count number of blocks read with length "ws2" (long epoch, 15 minutes by default)
  LD = 2 #dummy variable used to identify end of file and to make the process stop
  bsc_qc = data.frame(time = c(), size = c(), stringsAsFactors = FALSE)
  
  if (length(inspectfileobject) > 0) {
    INFI = inspectfileobject
  } else {
    stop("argument inspectfileobject not specified")
  }
  
  mon = INFI$monc
  dformat = INFI$dformc
  sf = INFI$sf

  if (is.null(sf)) { # sf is NULL for corrupt files
    return(invisible(list(filecorrupt = TRUE, filetooshort = FALSE, NFilePagesSkipped = 0,
                     metalong = c(), metashort = c(), wday = c(), wdayname = c(),
                     windowsizes = c(), bsc_qc = bsc_qc, QClog = NULL)))
  }

  hvars = g.extractheadervars(INFI)
  deviceSerialNumber = hvars$deviceSerialNumber

  # get now-wear, clip, and blocksize parameters (thresholds)
  if (mon == MONITOR$PARMAY_MTX) params_rawdata[["dynrange"]] = GGIRread::readParmayMatrix(datafile, output = "dynrange")
  ncb_params = get_nw_clip_block_params(monc = mon, dformat = dformat,
                                        deviceSerialNumber = deviceSerialNumber,
                                        sf = sf,
                                        params_rawdata = params_rawdata)
  clipthres = ncb_params$clipthres
  blocksize = ncb_params$blocksize
  sdcriter = ncb_params$sdcriter
  racriter = ncb_params$racriter
  n_decimal_places = 4 # number of decimal places to which features should be rounded
  # creating matrices for storing output
  S = matrix(0,0,4) #dummy variable needed to cope with head-tailing succeeding blocks of data
  nev = 80*10^7 # number expected values
  NR = ceiling(nev / (sf*ws3)) + 1000 #NR = number of 'ws3' second rows (this is for 10 days at 80 Hz)
  metashort = matrix(" ",NR,(1 + nmetrics)) #generating output matrix for acceleration signal
  QClog = NULL

  #===============================================
  # Read file
  isLastBlock = FALSE # dummy variable part "end of loop mechanism"

  PreviousLastValue = c(0, 0, 1)
  PreviousLastTime = NULL
  header = NULL

  while (LD > 1) {
    if (verbose == TRUE) {
      if (i  == 1) {
        cat(paste0("\nLoading chunk: ", i))
      } else {
        cat(paste0(" ", i))
      }
    }
    
    accread = g.readaccfile(filename = datafile, blocksize = blocksize, blocknumber = i,
                            filequality = filequality,
                            ws = ws, PreviousEndPage = PreviousEndPage,
                            inspectfileobject = INFI,
                            PreviousLastValue = PreviousLastValue,
                            PreviousLastTime = PreviousLastTime,
                            params_rawdata = params_rawdata, params_general = params_general, 
                            header = header)
    header = accread$header

    if ("PreviousLastValue" %in% names(accread$P)) { # output when reading ad-hoc csv
      PreviousLastValue = accread$P$PreviousLastValue
      PreviousLastTime = accread$P$PreviousLastTime
    } 
        
    filequality = accread$filequality
    filetooshort = filequality$filetooshort
    filecorrupt = filequality$filecorrupt
    filedoesnotholdday = filequality$filedoesnotholdday
    NFilePagesSkipped = filequality$NFilePagesSkipped
    isLastBlock = accread$isLastBlock
    PreviousEndPage = accread$endpage
    
    #============
    #process data as read from binary file
    if (length(accread$P) > 0) { # would have been set to zero if file was corrupt or empty
      data = accread$P$data
      QClog = rbind(QClog, accread$P$QClog)
      rm(accread)

      if (i == 1) {
        light.available = ("light" %in% colnames(data))

        use.temp = ("temperature" %in% colnames(data))
        if (use.temp) {
          if (mean(data$temperature[1:10], na.rm = TRUE) > 50) {
            warning("temperature value is unreaslistically high (> 50 Celcius) and will not be used.", call. = FALSE)
            use.temp = FALSE
          }
        }

        # output matrix for 15 minutes summaries
        if (!use.temp && !light.available) {
          metalong = matrix(" ", ((nev/(sf*ws2)) + 100), 4)
          metricnames_long = c("timestamp","nonwearscore","clippingscore","EN")
        } else if (use.temp && !light.available) {
          metalong = matrix(" ", ((nev/(sf*ws2)) + 100), 5)
          metricnames_long = c("timestamp","nonwearscore","clippingscore","temperaturemean","EN")
        } else if (!use.temp && light.available) {
          metalong = matrix(" ", ((nev/(sf*ws2)) + 100), 6)
          metricnames_long = c("timestamp","nonwearscore","clippingscore","lightmean","lightpeak","EN")
        } else if (use.temp && light.available) {
          metalong = matrix(" ", ((nev/(sf*ws2)) + 100), 7)
          metricnames_long = c("timestamp","nonwearscore","clippingscore","lightmean","lightpeak","temperaturemean","EN")
        }
      }

      if (params_rawdata[["imputeTimegaps"]] && (dformat == FORMAT$CSV ||
                                                 dformat == FORMAT$AD_HOC_CSV ||
                                                 dformat == FORMAT$GT3X)) {
        P = g.imputeTimegaps(data, sf = sf, k = 0.25,
                             PreviousLastValue = PreviousLastValue,
                             PreviousLastTime = PreviousLastTime,
                             epochsize = c(ws3, ws2))
        data = P$x
        PreviousLastValue = data[nrow(data), c("x", "y", "z")]
        if ("time" %in% colnames(data)) {
          PreviousLastTime = as.POSIXct(data$time[nrow(data)], origin = "1970-1-1")
        } else {
          PreviousLastTime = NULL
        }
        QClog = rbind(QClog, P$QClog)
        rm(P)
      }

      gc()

      data = as.matrix(data, rownames.force = FALSE)
      #add leftover data from last time
      if (nrow(S) > 0) {
        if (params_rawdata[["imputeTimegaps"]]) {
          if ("remaining_epochs" %in% colnames(data)) {
            if (ncol(S) == (ncol(data) - 1)) {
              # this block has time gaps while the previous block did not
              S = cbind(S, 1)
              colnames(S)[ncol(S)] = "remaining_epochs"
            }
          } else if ("remaining_epochs" %in% colnames(S)) {
            if ((ncol(S) - 1) == ncol(data)) {
              # this block does not have time gaps while the previous blog did
              data = cbind(data, 1)
              colnames(data)[ncol(S)] = "remaining_epochs"
            }
          }
        }
        # Timegaps inside data chunks are dealt with in g.imputeTimegaps
        # but timegaps in between data chunks are not dealt with there,
        # which could in theory be an issue for csv format data with known time gaps,
        # such as ActivPAL and ActiGraph.
        # The time gaps needs to coincide exactly with
        # edges of a data chunk, by which probability of this occurring is low.
        # The code blow aims to detect and handle this scenario.
        if ("time" %in% colnames(data) && "time" %in% colnames(S)) {
          timegap_between_chunks = as.numeric(data[1, "time"] - S[nrow(S), "time"])
          if (timegap_between_chunks > 3600 * sf) {
            stop(paste0("Time gap observed of more than 1 hour between data ",
                        "chunks for ", basename(datafile), " . Please contact ",
                        "package maintainer."), call. = FALSE)
          } else if (timegap_between_chunks > 3 / sf && timegap_between_chunks <= 60 * sf) {
            # impute time gap of more than 3 samples and equal to or less than 1 hour
            # normalise last value
            S[nrow(S), c("x", "y", "z")] = S[nrow(S), c("x", "y", "z")] / sqrt(sum(S[nrow(S), c("x", "y", "z")]^2))
            # replicate last row
            newRows = do.call(rbind, replicate(round(timegap_between_chunks  * sf), S[nrow(S), ], simplify = FALSE))
            # append to end
            S = rbind(S, newRows)
          }
        }
        data = rbind(S, data)
      }
      if (i == 1) {
        SWMT = get_starttime_weekday_truncdata(mon, dformat,
                                               data, header, desiredtz = params_general[["desiredtz"]],
                                               sf, datafile,  ws2,
                                               configtz = params_general[["configtz"]])
        starttime = SWMT$starttime
        wday = SWMT$wday; wdayname = SWMT$wdayname
        data = SWMT$data
        
        rm(SWMT)
      }

      LD = nrow(data)
      if (LD < (ws*sf) && i == 1) {
        warning('\nWarning data too short for doing non-wear detection 3\n')
        isLastBlock = TRUE
        LD = 0 #ignore rest of the data and store what has been loaded so far.
      }
      
      #store data that could not be used for this block, but will be added to next block
      if (LD >= (ws*sf)) {
        
        use = (floor(LD / (ws2*sf))) * (ws2*sf) #number of datapoint to use # changes from ws to ws2 Vvh 23/4/2017
        if (length(myfun) != 0) { # if using external function, then check that use is a multitude of the expected windowlength
          Nminlength = use / myfun$minlength
          if (Nminlength != floor(Nminlength)) { # it is not a multitude
            use = floor(Nminlength) * myfun$minlength # correct use accordingly
          }
        }
        if ((LD - use) > 1) {
          S = data[(use + 1):LD,] #store leftover data
          if (ncol(S) == 1) {
            S = t(S)
          }
        } else { #use all data
          S = matrix(0, 0, ncol(data))
        }
        data = data[1:use,]
        LD = nrow(data) #redefine LD because there is less data
        if ("remaining_epochs" %in% colnames(data)) { #
          # remove remaining_epochs from data object and keep it seperately
          remaining_epochs = data[,"remaining_epochs"]
          data = data[, -which(colnames(data) == "remaining_epochs")]
        }
        ##==================================================
        # Feature calculation
        dur = nrow(data)	#duration of experiment in data points
        durexp = nrow(data) / (sf*ws)	#duration of experiment in hrs
        #--------------------------------------------
        temperature = light = c()
        if (light.available) {
          light = data[, "light"]
        }
        if (use.temp) {
          temperature = data[, "temperature"]
        }

        # rescale data
        data[, c("x", "y", "z")] = scale(data[, c("x", "y", "z")], center = -offset, scale = 1/scale)
        if (use.temp && length(meantempcal) > 0) {
          yy = cbind(temperature,
                     temperature,
                     temperature)
          data[, c("x", "y", "z")] = data[, c("x", "y", "z")] + scale(yy, center = rep(meantempcal,3), scale = 1/tempoffset)
        }

        EN = sqrt(data[, "x"]^2 + data[, "y"]^2 + data[, "z"]^2) # Do not delete Used for long epoch calculation
        accmetrics = g.applymetrics(data = data,
                                    sf = sf, ws3 = ws3,
                                    metrics2do = metrics2do,
                                    n = params_metrics[["n"]],
                                    lb = params_metrics[["lb"]],
                                    hb = params_metrics[["hb"]],
                                    zc.lb = params_metrics[["zc.lb"]],
                                    zc.hb = params_metrics[["zc.hb"]],
                                    zc.sb = params_metrics[["zc.sb"]],
                                    zc.order = params_metrics[["zc.order"]],
                                    actilife_LFE = params_metrics[["actilife_LFE"]])
        # round decimal places, because due to averaging we get a lot of information
        # that only slows down computation and increases storage size
        accmetrics = lapply(accmetrics, round, n_decimal_places)
        accmetrics = data.frame(sapply(accmetrics,c)) # collapse to data.frame
        # update LD in case data has been imputed at epoch level
        if (floor(LD / (ws3 * sf)) < nrow(accmetrics)) { # then, data has been imputed
          LD = nrow(accmetrics) * ws3 * sf
        }
        #--------------------------------------------------------------------
        if (length(myfun) != 0) { # apply external function to the data to extract extra features
          #starttime
          if (is.logical(myfun$timestamp) == T) {
            if (myfun$timestamp == TRUE) {
              myfun$timestamp = starttime
            } else {
              myfun$timestamp = c()
            }
          }
          OutputExternalFunction = applyExtFunction(data, myfun, sf, ws3, interpolationType = params_rawdata[["interpolationType"]])
        }
      }
      if (LD >= (ws*sf)) { #LD != 0
        #-----------------------------------------------------
        #extend metashort and metalong if it is expected to be too short
        if (exists("remaining_epochs")) {
          totalgap = sum(remaining_epochs[which(remaining_epochs != 1)])
        } else {
          totalgap = 0
        }
        if (count > (nrow(metashort) - ((2.5*(3600/ws3) * 24)) + totalgap)) {
          extension = matrix(" ", ((3600/ws3) * 24) + totalgap, ncol(metashort)) #add another day to metashort once you reach the end of it
          metashort = rbind(metashort,extension)
          extension2 = matrix(" ", ((3600/ws2) * 24)  + (totalgap * (ws2/ws3)), ncol(metalong)) #add another day to metashort once you reach the end of it
          metalong = rbind(metalong, extension2)
        }
        col_msi = 2
        # Add metric time series to metashort object
        metnames = grep(pattern = "BrondCount|NeishabouriCount", x = names(accmetrics), invert = TRUE, value = TRUE)
        for (metnam in metnames) {
          dovalue = paste0("do.",tolower(metnam))
          dovalue = gsub(pattern = "angle_", replacement = "angle", x = dovalue)
          if (params_metrics[[dovalue]] == TRUE) {
            metashort[count:(count - 1 + length(accmetrics[[metnam]])), col_msi] = accmetrics[[metnam]]
            col_msi = col_msi + 1
          }
        }
        if (params_metrics[["do.brondcounts"]] == TRUE) {
          metashort[count:(count - 1 + length(accmetrics$BrondCount_x)), col_msi] = accmetrics$BrondCount_x
          metashort[count:(count - 1 + length(accmetrics$BrondCount_y)), col_msi + 1] = accmetrics$BrondCount_y
          metashort[count:(count - 1 + length(accmetrics$BrondCount_z)), col_msi + 2] = accmetrics$BrondCount_z
          col_msi = col_msi + 3
          metnames = c(metnames, "BrondCount_x", "BrondCount_y", "BrondCount_z")
        }
        if (params_metrics[["do.neishabouricounts"]] == TRUE) {
          metashort[count:(count - 1 + length(accmetrics$NeishabouriCount_x)), col_msi] = accmetrics$NeishabouriCount_x
          metashort[count:(count - 1 + length(accmetrics$NeishabouriCount_y)), col_msi + 1] = accmetrics$NeishabouriCount_y
          metashort[count:(count - 1 + length(accmetrics$NeishabouriCount_z)), col_msi + 2] = accmetrics$NeishabouriCount_z
          metashort[count:(count - 1 + length(accmetrics$NeishabouriCount_vm)), col_msi + 3] = accmetrics$NeishabouriCount_vm
          col_msi = col_msi + 3
          metnames = c(metnames, "NeishabouriCount_x", "NeishabouriCount_y", "NeishabouriCount_z", "NeishabouriCount_vm")
        }
        metnames = gsub(pattern = "angle_", replacement = "angle", x = metnames)
        if (length(myfun) != 0) { # if an external function is applied.
          NcolEF = ncol(OutputExternalFunction) - 1 # number of extra columns needed
          metashort[count:(count - 1 + nrow(OutputExternalFunction)), col_msi:(col_msi + NcolEF)] = as.matrix(OutputExternalFunction); col_msi = col_msi + NcolEF + 1
        }
        
        length_acc_metrics =  length(accmetrics[[1]]) # changing indicator to whatever metric is calculated, EN produces incompatibility when deriving both ENMO and ENMOa
        rm(accmetrics)
        # update blocksize depending on available memory
        BlocksizeNew = updateBlocksize(blocksize = blocksize, bsc_qc = bsc_qc)
        bsc_qc = BlocksizeNew$bsc_qc
        blocksize = BlocksizeNew$blocksize
        ##==================================================
        # MODULE 2 - non-wear time & clipping
        NWCW = detect_nonwear_clipping(data = data, windowsizes = c(ws3, ws2, ws), sf = sf,
                                       clipthres = clipthres, sdcriter = sdcriter, racriter = racriter,
                                       nonwear_approach = params_cleaning[["nonwear_approach"]],
                                       params_rawdata = params_rawdata)
        NWav = NWCW$NWav; CWav = NWCW$CWav; nmin = NWCW$nmin
        # metalong
        col_mli = 2
        metalong[count2:((count2 - 1) + length(NWav)),col_mli] = NWav; col_mli = col_mli + 1
        metalong[count2:((count2 - 1) + length(NWav)),col_mli] = CWav; col_mli = col_mli + 1

        if(light.available) {
          #light (running mean)
          lightc = cumsum(c(0,light))
          select = seq(1, length(lightc), by = (ws2 * sf))
          lightmean = diff(lightc[round(select)]) / abs(diff(round(select)))
          rm(lightc)
          #light (running max)
          lightmax = matrix(0, length(lightmean), 1)
          for (li in 1:(length(light)/(ws2*sf))) {
            tempm = max(light[((li - 1) * (ws2 * sf)):(li * (ws2 * sf))])
            if (length(tempm) > 0) {
              lightmax[li] = tempm[1]
            } else {
              lightmax[li] = max(light[((li - 1) * (ws2 * sf)):(li * (ws2 * sf))])
            }
          }

          metalong[(count2):((count2 - 1) + length(NWav)), col_mli] = round(lightmean, digits = n_decimal_places)
          col_mli = col_mli + 1
          metalong[(count2):((count2 - 1) + length(NWav)), col_mli] = round(lightmax, digits = n_decimal_places)
          col_mli = col_mli + 1
        }

        if(use.temp) {
          #temperature (running mean)
          temperaturec = cumsum(c(0, temperature))
          select = seq(1, length(temperaturec), by = (ws2 * sf))
          temperatureb = diff(temperaturec[round(select)]) / abs(diff(round(select)))
          rm(temperaturec)

          metalong[(count2):((count2 - 1) + length(NWav)), col_mli] = round(temperatureb, digits = n_decimal_places)
          col_mli = col_mli + 1
        }

        #EN going from sample to ws2
        ENc = cumsum(c(0, EN))
        select = seq(1, length(ENc), by = (ws2 * sf)) #<= EN is derived from data, so it needs the new sf
        ENb = diff(ENc[round(select)]) / abs(diff(round(select)))
        rm(ENc, EN)
        metalong[(count2):((count2 - 1) + length(NWav)), col_mli] = round(ENb, digits = n_decimal_places)

        if (exists("remaining_epochs")) {
          # Impute long gaps at epoch levels, because imputing them at raw level would
          # be too memory hungry
          impute_at_epoch_level = function(gapsize, timeseries, gap_index, metnames) {
            # gap_index: where do gaps occur (epoch indexing)
            # gap_size: how long is gap (epoch numbers)
            if (any(duplicated(gap_index))) {
              # When 2 gap_index are within the same epoch (either short or long)
              # we would have a duplicated gap_index here, then combine information
              dup_index_tmp = which(duplicated(gap_index))
              dup_index = gap_index[dup_index_tmp]
              for (dup_index_i in dup_index) {
                to_combine = which(gap_index == dup_index_i)
                length_to_combine = length(to_combine) # In the unlikely event that a gap_index appears more than 2, this should be able to deal with it.
                delete = to_combine[-1] # leave only the first index and remove duplicates
                gap_index = gap_index[-delete] # remove from gap index
                gapsize[to_combine[1]] = sum(gapsize[to_combine]) - (length_to_combine - 1) # minus 1 because it was summed 1 to each gapsize (which is +2 when it is duplicated) in the function call
                gapsize = gapsize[-delete]
              }
            }
            if ("nonwearscore" %in% metnames) {
              timeseries[gap_index, which(metnames == "nonwearscore")]  = 3
            } else {
              # set all features to zero except time and angle feature
              timeseries[gap_index, grep(pattern = "time|angle",
                                         x = metnames, invert = TRUE, value = FALSE)] = 0
              # set EN to 1 if it is available
              if ("EN" %in% metnames) timeseries[gap_index, which(metnames == "EN")] = 1
            }
            N_time = nrow(timeseries)
            newindi = rep(1, N_time)
            newindi[gap_index] = as.numeric(gapsize)
            newindi = rep(1:N_time, newindi)
            timeseries = timeseries[newindi,]
            return(timeseries)
          }
          gaps_to_fill = which(remaining_epochs != 1)
          if (length(gaps_to_fill) > 0) {
            nr_before = c(nrow(metalong), nrow(metashort))
            # metalong
            metalong = impute_at_epoch_level(gapsize = floor(remaining_epochs[gaps_to_fill] * (ws3/ws2)) + 1, # plus 1 needed to count for current epoch
                                             timeseries = metalong,
                                             gap_index = floor(gaps_to_fill / (ws2 * sf)) + count2, # Using floor so that the gap is filled in the epoch in which it is occurring
                                             metnames = metricnames_long)
            # metashort
            # added epoch-level nonwear to metashort to get it imputed, then remove it
            metashort = impute_at_epoch_level(gapsize = remaining_epochs[gaps_to_fill], # gapsize in epochs
                                              timeseries = metashort,
                                              gap_index = floor(gaps_to_fill / (ws3 * sf)) + count, # Using floor so that the gap is filled in the epoch in which it is occurring
                                              metnames = c("timestamp", metnames)) # epoch level index of gap
            nr_after = c(nrow(metalong), nrow(metashort))
            count2 = count2 + (nr_after[1] - nr_before[1])
            count = count + (nr_after[2] - nr_before[2])
          }
        }
        col_mli = col_mli + 1
        count2 = count2 + nmin
        count = count + length_acc_metrics
        if (exists("data")) rm(data)
        if (exists("light")) rm(light)
        if (exists("temperature")) rm(temperature)
        gc()
      } #end of section which is skipped when switchoff == 1
    } else {
      LD = 0 #once LD < 1 the analysis stops, so this is a trick to stop it
      # stop reading because there is not enough data in this block
    }
    if (isLastBlock) LD = 0
    if (ceiling(daylimit) != FALSE) {
      if (i == ceiling(daylimit)) { #to speed up testing only read first 'i' blocks of data
        LD = 0 #once LD < 1 the analysis stops, so this is a trick to stop it
        if (verbose == TRUE) cat(paste0("\nstopped reading data because this analysis is limited to ", ceiling(daylimit), " days\n"))
      }
    }
    i = i + 1 #go to next block
  }
  # deriving timestamps
  if (!filecorrupt && !filetooshort && !filedoesnotholdday) {
    cut = count:nrow(metashort)
    if (length(cut) > 1) {
      metashort = metashort[-cut,]
      # for a very small file, there could be just one row in metashort[-cut,], so it gets coerced to a vector.
      # But what we actually need is a 1-row matrix. So we need to transpose it. 
      if(is.vector(metashort)) { 
        metashort = as.matrix(t(metashort))
      }
    }
    if (nrow(metashort) > 1) {
      starttime3 = round(as.numeric(starttime)) #numeric time but relative to the desiredtz
      time5 = seq(starttime3, (starttime3 + ((nrow(metashort) - 1) * ws3)), by = ws3)
      time6 = as.POSIXlt(time5,origin = "1970-01-01", tz = params_general[["desiredtz"]])
      time6 = strftime(time6, format = "%Y-%m-%dT%H:%M:%S%z")
      metashort[,1] = as.character(time6)
    }
    cut2 = count2:nrow(metalong)
    if (length(cut2) > 1) {
      metalong = metalong[-cut2,]
      # for a very small file, there could be just one row in metalong[-cut2,], so it gets coerced to a vector.
      # But what we actually need is a 1-row matrix. So we need to transpose it. 
      if(is.vector(metalong)) { 
        metalong = as.matrix(t(metalong))
      }
    }
    if (nrow(metalong) > 2) {
      starttime4 = round(as.numeric(starttime)) #numeric time but relative to the desiredtz
      time1 = seq(starttime4,(starttime4 + (nrow(metalong) * ws2) - 1), by = ws2)
      if (length(params_general[["desiredtz"]]) == 0) {
        warning("desiredtz not specified, system timezone used as default")
        params_general[["desiredtz"]] = ""
      }
      time2 = as.POSIXlt(time1, origin = "1970-01-01", tz = params_general[["desiredtz"]])
      time2 = strftime(time2, format = "%Y-%m-%dT%H:%M:%S%z")
      metalong[, 1] = as.character(time2)
    }
    metricnames_short = c("timestamp", metnames)
    
    # Following code is needed to make sure that algorithms that produce character value
    # output are not assumed to be numeric
    NbasicMetrics = length(metricnames_short)
    if (length(myfun) != 0) {
      metricnames_short = c(metricnames_short, myfun$colnames)
      if (myfun$outputtype == "numeric") NbasicMetrics = NbasicMetrics + length(myfun$colnames)
    }
    metashort = data.frame(A = metashort, stringsAsFactors = FALSE)
    names(metashort) = metricnames_short
    for (ncolms in 2:NbasicMetrics) {
      metashort[,ncolms] = as.numeric(metashort[,ncolms])
    }
    
    metalong = data.frame(A = metalong, stringsAsFactors = FALSE)
    names(metalong) = metricnames_long
    for (ncolml in 2:ncol(metalong)) {
      metalong[,ncolml] = as.numeric(metalong[,ncolml])
    }
  } else {
    metalong = metashort = wday = wdayname = params_general[["windowsizes"]] = c()
  }
  if (length(metashort) == 0 | filedoesnotholdday == TRUE) filetooshort = TRUE
  invisible(list(filecorrupt = filecorrupt, filetooshort = filetooshort, NFilePagesSkipped = NFilePagesSkipped,
                 metalong = metalong, metashort = metashort, wday = wday, wdayname = wdayname,
                 windowsizes = params_general[["windowsizes"]], bsc_qc = bsc_qc, QClog = QClog))
}

.onAttach <- function(...) {
  if (!interactive()) return()
  # stop interactive calling from `library(GGIR)`
  repos = getOption("repos")
  if ("@CRAN@" %in% repos) {
    repos = utils::getCRANmirrors()
    # choose cloud/first if option triggers `contrib.url`
    # to call `chooseCRANmirror`
    repos = repos$URL[1]
    packageStartupMessage(
      "No CRAN mirror set, so using ", repos,
      " to check GGIR package version")
  }
  pkgs <- available.packages(repos = repos)
  cran_version <- package_version(pkgs[which(pkgs[,1] == "GGIR"),"Version"])
  if (length(cran_version) == 0) return() # handle no internet connection
  local_version <- packageVersion("GGIR")
  behind_cran <- cran_version > local_version
  if (interactive()) {
    if (behind_cran) {
      msg <- paste0("A newer version of GGIR is available with bug fixes and new features. [", local_version," --> ", cran_version, "]")
      packageStartupMessage(msg)
    }
  }
}

identify_levels = function(ts, TRLi, TRMi, TRVi, ws3, params_phyact = c(), ...) {
  #get input variables
  input = list(...)
  if (length(input) > 0 || length(params_phyact) == 0) {
    # Extract and check parameters if user provides more arguments than just the parameter arguments,
    # or if params_[...] aren't specified (so need to be filled with defaults).
    # So, inside GGIR this will not be used, but it is used when identify_level is used on its own
    # as if it was still the old identify_level function
    params = extract_params(params_phyact = params_phyact,
                            input = input) # load default parameters
    params_phyact = params$params_phyact
  }
  
  #=======================================================
  # LABEL INSENTITY LEVELS
  LEVELS = rep(0, length(ts$time))
  OLEVELS = rep(0, length(ts$time)) #to capture moderate and vigorous seperately
  LEVELS[ts$sibdetection == 1 & ts$diur == 1] = 0 #Sleep during the Sleep Period Time Window
  LEVELS[ts$sibdetection == 0 & ts$diur == 1] = 1 #Wakefullness during Sleep Period Time Window
  # activity during the night
  LEVELS[ts$sibdetection == 0 & ts$diur == 1 & ts$ACC >= TRLi & ts$ACC < TRMi] = 2 #LIGHT
  LEVELS[ts$sibdetection == 0 & ts$diur == 1 & ts$ACC >= TRMi & ts$ACC < TRVi] = 3 #MODERATE
  LEVELS[ts$sibdetection == 0 & ts$diur == 1 & ts$ACC >= TRVi] = 4 #VIGOROUS
  Lnames = c("spt_sleep","spt_wake_IN", "spt_wake_LIG", "spt_wake_MOD", "spt_wake_VIG")
  # activity during the day
  #======================================
  LEVELS[ts$diur == 0 & ts$ACC < TRLi] = 5 # Inactivit
  LEVELS[ts$diur == 0 & ts$ACC >= TRLi & ts$ACC < TRMi] = 6 #LIGHT
  LEVELS[ts$diur == 0 & ts$ACC >= TRMi & ts$ACC < TRVi] = 7 #MODERATE
  LEVELS[ts$diur == 0 & ts$ACC >= TRVi] = 8 #VIGOROUS
  Lnames = c(Lnames,"day_IN_unbt", "day_LIG_unbt", "day_MOD_unbt","day_VIG_unbt")
  # store separate copy of moderate and vigorous levels
  # note that when OLEVELS equal zero this is sleep period time window
  OLEVELS[LEVELS == 5] = 1 #IN
  OLEVELS[LEVELS == 6] = 2 #LIGHT
  OLEVELS[LEVELS == 7] = 3 #MOD
  OLEVELS[LEVELS == 8] = 4 #VIG
  

  #-------------------------------------
  # MVPA BOUTS
  LN = length(ts$time)
  
  boutduration = params_phyact[["boutdur.mvpa"]] * (60 / ws3) 
  NBL = length(boutduration) #number of bout lengths
  CL = 9 #current level
  refe = rep(0, LN)
  bc.mvpa = c()
  for (BL in 1:NBL) { # needs to be flexible to varibale number of bout lengths
    rr1 = rep(0, LN)
    p = which(ts$ACC >= TRMi & refe == 0 & ts$diur == 0); rr1[p] = 1
    out1 = g.getbout(x = rr1, boutduration = boutduration[BL],
                     boutcriter = params_phyact[["boutcriter.mvpa"]], ws3 = ws3)
    LEVELS[ts$diur == 0 & out1 == 1] = CL
    bc.mvpa = rbind(bc.mvpa,out1)
    refe = refe + out1
    if (BL == 1) {
      Lnames = c(Lnames,paste0("day_MVPA_bts_", params_phyact[["boutdur.mvpa"]][BL]))
    } else {
      Lnames = c(Lnames,paste0("day_MVPA_bts_", params_phyact[["boutdur.mvpa"]][BL], "_",
                               params_phyact[["boutdur.mvpa"]][BL - 1]))
    }
    CL = CL + 1
  }
  #-------------------------------------
  # INACTIVITY BOUTS
  LN = length(ts$time)
  boutduration = params_phyact[["boutdur.in"]] * (60/ws3)
  NBL = length(boutduration) #number of bout lengths
  bc.in = c()
  for (BL in 1:NBL) {
    rr1 = rep(0, LN)
    p = which(ts$ACC < TRLi & refe == 0 & ts$diur == 0); rr1[p] = 1
    out1 = g.getbout(x = rr1, boutduration = boutduration[BL],
                     boutcriter = params_phyact[["boutcriter.in"]], ws3 = ws3)
    LEVELS[ts$diur == 0 & out1 == 1] = CL
    bc.in = rbind(bc.in, out1)
    refe = refe + out1
    if (BL == 1) {
      Lnames = c(Lnames,paste0("day_IN_bts_", params_phyact[["boutdur.in"]][BL]))
    } else {
      Lnames = c(Lnames,paste0("day_IN_bts_", params_phyact[["boutdur.in"]][BL], "_",
                               params_phyact[["boutdur.in"]][BL - 1]))
    }
    CL = CL + 1
  }
  #-------------------------------------
  # LIGHT BOUTS
  LN = length(ts$time)
  boutduration = params_phyact[["boutdur.lig"]] * (60 / ws3)
  NBL = length(boutduration) #number of bout lengths
  bc.lig = c()
  for (BL in 1:NBL) {
    rr1 = rep(0, LN)
    p = which(ts$ACC >= TRLi & refe == 0 & ts$ACC < TRMi & ts$diur == 0); rr1[p] = 1
    out1 = g.getbout(x = rr1, boutduration = boutduration[BL],
                     boutcriter = params_phyact[["boutcriter.lig"]], ws3 = ws3)                    
    LEVELS[ts$diur == 0 & out1 == 1] = CL
    bc.lig = rbind(bc.lig,out1)
    refe = refe + out1
    if (BL == 1) {
      Lnames = c(Lnames,paste0("day_LIG_bts_",params_phyact[["boutdur.lig"]][BL]))
    } else {
      Lnames = c(Lnames,paste0("day_LIG_bts_", params_phyact[["boutdur.lig"]][BL],
                               "_", params_phyact[["boutdur.lig"]][BL - 1]))
    }
    CL = CL + 1
  }
  invisible(list(LEVELS = LEVELS, OLEVELS = OLEVELS, Lnames = Lnames, 
                 bc.mvpa = bc.mvpa, bc.lig = bc.lig, bc.in = bc.in, ts = ts,
                 threshold = c(TRLi, TRMi, TRVi)))
}

library(GGIR)
context("identify_levels")
test_that("identify_levels", {
  time = rep(0, 10000) # the values of time are not used in the function, we only care about the length the object time
  diur = c(rep(0, 2000), rep(1, 4000), rep(0, 4000))
  detection = time
  set.seed(300)
  detection[which(runif(n = 10000, min = 0, max = 1) < 0.1)] = 1
  set.seed(1000)
  ACC = rnorm(n = 10000, mean = 0, sd = 40)
  ts = data.frame(ACC = ACC,time = time, diur = diur, sibdetection = detection)
  levels = identify_levels(ts, TRLi = 40, TRMi = 100, TRVi = 400,
                             boutdur.mvpa = 10, boutcriter.mvpa = 0.8,
                             boutdur.lig = 5, boutcriter.lig = 0.8,
                             boutdur.in = 10, boutcriter.in = 0.9,
                             ws3 = 5)
  expect_equal(as.numeric(table(levels$LEVELS)), c(403, 3012, 564, 21, 3792, 779, 34, 1395))
  expect_equal(levels$Lnames[12], "day_LIG_bts_5")
})

g.create.sp.mat = function(nsp,spo,sleepdet.t,daysleep=FALSE) {
  # function to convert data into sleep period matrix part of g.part4.R  
  if (daysleep == FALSE) {
    th2 = 12
  } else {
    th2 = 18
  }  
  weekdays = c("Sunday","Monday","Tuesday","Wednesday","Thursday","Friday","Saturday")
  for (sp in 1:nsp) {
    spo[sp,1] = sp
    tmp7 = format(sleepdet.t$sib.onset.time[which(sleepdet.t$sib.period == sp)])[1]
    tmp7w = format(sleepdet.t$sib.onset.time[which(sleepdet.t$sib.period == sp)])[1]
    if (length(tmp7) > 0) {
      if (tmp7 != "") {
        tmp8 = unlist(strsplit(tmp7," "))
        tmp8w = unlist(strsplit(tmp7w," "))
        tmp9 = unlist(strsplit(tmp8[2],":"))
        tmp9w = unlist(strsplit(tmp8w[2],":"))
        if (length(tmp8) == 1) {
          tmp10 = 0
        } else {
          tmp10 = as.numeric(tmp9[1]) + (as.numeric(tmp9[2])/60) + (as.numeric(tmp9[3])/3600)
        }
        if (length(tmp8w) == 1) {
          tmp10w = 0
        } else {
          tmp10w = as.numeric(tmp9w[1]) + (as.numeric(tmp9w[2])/60) + (as.numeric(tmp9w[3])/3600)
        }
        # get weekday
        if (sp == 1) { #first sleep period
          soso = unclass(as.POSIXlt(tmp7))
          wday = soso$wday #day of the week 0-6 and 0 is Sunday
          #remember day of the week and date
          wday_safe = wday + 1 #change to 1-7 number
          calendar_date_safe = paste(soso$mday,"/",(soso$mon+1),"/",(soso$year+1900),sep="")
        }
        if (sp == nsp)   { #waking up of last sleep period
          soso = unclass(as.POSIXlt(tmp7w))
          wday = soso$wday #day of the week 0-6 and 0 is Sunday
          wday = wday + 1 #change to 1-7 number
          # check whether both beginning and end of the night are on the same day
          if (wday == (wday_safe+1)) { #normal night starting on one day and ending at the ned
            wdayname = weekdays[wday_safe]
            calendar_date = calendar_date_safe
          } else { #start and end at the same day
            if (tmp10w > th2) { #end of last sleep period starts at first day (end of night is before midnight)
              wdayname = weekdays[wday_safe]
              calendar_date = paste(soso$mday,"/",(soso$mon+1),"/",(soso$year+1900),sep="")
            } else { #second day
              tmp7 = as.POSIXlt(tmp7) - (24*3600) #change date
              soso = unclass(as.POSIXlt(tmp7))
              wday = soso$wday #day of the week 0-6 and 0 is Sunday
              wday = wday + 1 #change to 1-7 number
              wdayname = weekdays[wday]
              calendar_date = paste(soso$mday,"/",(soso$mon+1),"/",(soso$year+1900),sep="")
            }
          }
        }

        # wake
        tmp11 = format(sleepdet.t$sib.end.time[which(sleepdet.t$sib.period == sp)])
        tmp12 = unlist(strsplit(tmp11," "))
        tmp13 = unlist(strsplit(tmp12[2],":"))
        if (length(tmp12) == 1) {
          tmp14 = 0
        } else {
          tmp14 = as.numeric(tmp13[1]) + (as.numeric(tmp13[2])/60) + (as.numeric(tmp13[3])/3600)
        }
        if (tmp10 != tmp14) {#added 15/12/2014
        if (tmp10 < 12) tmp10 = tmp10 + 24
        if (tmp14 <= 12) tmp14 = tmp14 + 24 #changed < into <= on 18-11-2014
        }
        spo[sp,2] = tmp10
        spo[sp,3] = tmp14
      }
    }
  }
  invisible(list(spo=spo,calendar_date=calendar_date,wdayname=wdayname))
}

g.part5.analyseRest = function(sibreport = NULL, dsummary = NULL,
                               ds_names = NULL, fi = NULL, di = NULL,
                               ts = NULL, tz = NULL,
                               params_sleep = NULL) {
  # define function to summarise overlap between selfreported behaviour and sibs
  summarise_overlap = function(srep_tmp, X, Y, xi, yi, name = "", sumobject = NULL) {
    # X: column name in srep_temp to reflect overlap SIB with Selfreport
    # Y: column name in srep_temp to reflect overlap Selfreport with SIB
    # xi: indices of overlapping sibs corresponding to X
    # yi: indices of overlapping sibs corresponding to Y
    # name: name of self-reported behaviour to be used inside the variable name
    
    dsummary = sumobject$dsummary
    ds_names = sumobject$ds_names
    fi = sumobject$fi
    di = sumobject$di
    
    calcOverlapPercentage = function(overlap, duration) {
      return(sum(overlap * duration) / sum(duration))
    }
    # Overlap sib with selfreport
    if (length(xi) > 0) {
      dsummary[di,fi:(fi + 2)] = c(mean(srep_tmp$duration[xi]),
                                   sum(srep_tmp$duration[xi]),
                                   calcOverlapPercentage(overlap = srep_tmp[xi, X],
                                                         duration = srep_tmp$duration[xi]))
    } else {
      dsummary[di,fi:(fi + 2)] = c(0, 0, 0)
    }
    ds_names[fi:(fi + 2)] = c(paste0("mdur_denap_overl_", name),
                              paste0("tdur_denap_overl_", name),
                              paste0("perc_denap_overl_", name))
    fi = fi + 3
    # Overlap selfreportwith sib
    if (length(xi) > 0) {
      dsummary[di,fi:(fi + 2)] = c(mean(srep_tmp$duration[yi]),
                                   sum(srep_tmp$duration[yi]),
                                   calcOverlapPercentage(overlap = srep_tmp[yi, Y],
                                                         duration = srep_tmp$duration[yi]))
    } else {
      dsummary[di,fi:(fi + 2)] = c(0, 0, 0)
    }
    ds_names[fi:(fi + 2)] = c(paste0("mdur_", name, "_overl_denap"),
                              paste0("tdur_", name, "_overl_denap"),
                              paste0("perc_", name ,"_overl_denap"))
    fi = fi + 3
    invisible(list(fi = fi, ds_names = ds_names, dsummary = dsummary, di = di))
  }
  
  
  if (!is.null(sibreport) &&
      length(sibreport[[1]]) > 1)  {
    #---------------------------------------
    # transform time to POSIX
    if (is.ISO8601(as.character(ts$time[1]))) {
      ts$time = iso8601chartime2POSIX(ts$time, tz = tz)
    }
    sibreport$end = as.POSIXct(sibreport$end, tz = tz)
    sibreport$start = as.POSIXct(sibreport$start, tz = tz)
    
    #---------------------------------------
    # merge sibs when gap is shorter than possible_nap_gap
    if (params_sleep[["possible_nap_gap"]] > 0) {
      sibreport$gap2next = NA
      Nrow = nrow(sibreport)
      sibreport$gap2next[1:(Nrow - 1)] = (as.numeric(sibreport$start[2:Nrow]) - as.numeric(sibreport$end[1:(Nrow - 1)])) / 60
      sibreport$gap2next[which(sibreport$type != "sib" | sibreport$gap2next < 0)] = NA
      iter = 1
      while (iter < nrow(sibreport)) {
        if (!is.na(sibreport$gap2next[iter]) &&
            sibreport$gap2next[iter] < params_sleep[["possible_nap_gap"]]) {
          sibreport$end[iter] = sibreport$end[iter + 1]
          sibreport$mean_acc_1min_after[iter] = sibreport$mean_acc_1min_after[iter + 1]
          sibreport = sibreport[-(iter + 1),]
          sibreport$gap2next[iter] = as.numeric(sibreport$start[iter + 1]) - as.numeric(sibreport$end[iter])
          # no need to increment iter, because by merging the sib blocks
          # the current iter now refers to the next gap
        } else {
          iter = iter + 1
        }
        if (iter > nrow(sibreport) - 1) {
          break()
        }
      }
      epochSize = as.numeric(difftime(ts$time[2], ts$time[1], units = "mins"))
      sibreport$duration = as.numeric(difftime(sibreport$end, sibreport$start, units = "mins")) + epochSize
    }
    
    # Only consider sib episodes with minimum duration
    if (length(grep(pattern = "mean_acc_1min", x = colnames(sibreport))) > 0) {
      sibreport$acc_edge = pmax(sibreport$mean_acc_1min_before, sibreport$mean_acc_1min_after)
    } else {
      sibreport$acc_edge = 0
    }
    #-----------------------------------------------------------
    # Consider using marker button data to aid nap detection
    sibreport = markerButtonForRest(sibreport, params_sleep, ts)
    
    sibreport$startHour = as.numeric(format(sibreport$start, "%H"))
    sibreport$endHour = as.numeric(format(sibreport$end, "%H"))
    
    overlapMidnight = which(sibreport$endHour < sibreport$startHour)
    if (length(overlapMidnight) > 0) {
      sibreport$endHour[overlapMidnight] = sibreport$endHour[overlapMidnight] + 24
    }
    
    longboutsi = which((sibreport$type == "sib" &
                          sibreport$duration >= params_sleep[["possible_nap_dur"]][1] &
                          sibreport$duration < params_sleep[["possible_nap_dur"]][2] &
                          sibreport$acc_edge <= params_sleep[["possible_nap_edge_acc"]] &
                          sibreport$startHour >= params_sleep[["possible_nap_window"]][1] &
                          sibreport$endHour < params_sleep[["possible_nap_window"]][2] &
                          sibreport$ignore == FALSE) |
                         (sibreport$type != "sib" & sibreport$duration >= 1))
    
    Nlongbouts  = length(longboutsi)
    # add segment of sleeplog to sibreport. Done here and not near 
    # function g.sibreport called in function g.part5 because the segment needs to
    # be part of the time range otherwise it is omitted further down
    sibreport = sibreport[which(sibreport$type != "sleeplog" & sibreport$type != "sleeplog+bedlog"),]
    sleeplogi = which(ts$selfreported == "sleeplog" | ts$selfreported == "sleeplog+bedlog")
    if (length(sleeplogi) > 0) {
      dsi = diff(sleeplogi)
      sl_starts = c(1, which(dsi != 1) + 1)
      sl_ends = c(which(dsi != 1), length(sleeplogi))
      if (length(sl_starts) > 0) {
        for (Nsi in 1:length(sl_starts)) {
          newline = nrow(sibreport) + 1
          sibreport[newline,] = NA
          sibreport$ID[newline] = sibreport$ID[newline - 1]
          sibreport$type[newline] = "sleeplog"
          sibreport$start[newline] = format(ts$time[sleeplogi[sl_starts[Nsi]]])
          sibreport$end[newline] = format(ts$time[sleeplogi[sl_ends[Nsi]]])
        }
      }
    }
  } else {
    Nlongbouts = 0
    longboutsi = NULL
  }
  ds_names[fi:(fi + 26)] = c("sibreport_n_items",
                             "sibreport_n_items_day", "nbouts_day_denap",
                             "nbouts_day_srnap", "nbouts_day_srnonw",
                             "noverl_denap_srnap", "noverl_denap_srnonw",
                             "noverl_srnap_denap", "noverl_srnonw_denap",
                             "frag_mean_dur_denap_day", "dur_day_denap_min",
                             "frag_mean_dur_srnap_day", "dur_day_srnap_min",
                             "frag_mean_dur_srnonw_day", "dur_day_srnonw_min",
                             "mdur_denap_overl_srnap", "tdur_denap_overl_srnap",
                             "perc_denap_overl_srnap", "mdur_srnap_overl_denap",
                             "tdur_srnap_overl_denap", "perc_srnap_overl_denap",
                             "mdur_denap_overl_srnonw", "tdur_denap_overl_srnonw",
                             "perc_denap_overl_srnonw", "mdur_srnonw_overl_denap",
                             "tdur_srnonw_overl_denap", "perc_srnonw_overl_denap")
  dsummary[di,fi] = Nlongbouts
  fi = fi + 1
  if (length(longboutsi) > 0) {
    sibreport = sibreport[longboutsi,]
    srep_tmp = sibreport[which(sibreport$start >= min(ts$time) &
                                 sibreport$end <= max(ts$time)),]
    
    # update ts time series with the classified naps
    if ("sib" %in% srep_tmp$type) {
      sibnaps = which(srep_tmp$type == "sib")
      srep_tmp_rowsdelete = NULL
      if (length(sibnaps) > 0) {
        for (sni in 1:length(sibnaps)) {
          sibnap = which(ts$time >= srep_tmp$start[sibnaps[sni]] & ts$time <= srep_tmp$end[sibnaps[sni]])
          if (length(sibnap) > 0) {
            # Only consider nap it does not overlap for more than 10% with known nonwear.
            fractionInvalid = length(which(ts$nonwear[sibnap] == 1)) / length(sibnap)
            if (fractionInvalid < 0.1) {
              ts$sibdetection[sibnap] = 2
            } else {
              srep_tmp_rowsdelete = c(srep_tmp_rowsdelete , sibnaps[sni])
            }
          }
        }
        if (!is.null(srep_tmp_rowsdelete )) {
          srep_tmp = srep_tmp[-srep_tmp_rowsdelete,]
        }
      }
    }
    #	identify overlapping and non-overlapping, (nap-sib, non-wear-sib, sib, nap, nonwear)
    #	calculate for all five categories number, total duration, mean duration
    # but also account for possibility that some of these categories do not exist
    
    # for qc purposes:
    dsummary[di,fi] = nrow(srep_tmp)
    # if (ds_names[fi] != "sibreport_n_items_day") stop("mismatch in columnnames 1")
    fi = fi + 1
    if (nrow(srep_tmp) > 0) {
      sibs = which(srep_tmp$type == "sib")
      srep_tmp$SIBoverlapNonwear = 0
      srep_tmp$SIBoverlapNap = 0
      srep_tmp$SIBoverlapSleeplog = 0
      srep_tmp$NonwearOverlapSIB = 0
      srep_tmp$NapOverlapSIB = 0
      srep_tmp$SleeplogOverlapSIB = 0
      srep_tmp$start = as.POSIXct(srep_tmp$start, tz = tz)
      srep_tmp$end = as.POSIXct(srep_tmp$end, tz = tz)
      if (length(sibs) > 0) {
        classes = unique(srep_tmp$type)
        selfreport = which(srep_tmp$type == "nonwear" | srep_tmp$type == "nap" | srep_tmp$type == "sleeplog") 
        # summarise overlap between selfreported and accelerometer-based SIB
        if (length(selfreport) > 0) {
          for (si in sibs) {
            for (sr in selfreport) {
              # SIB overlap with selfreported behaviour
              if (srep_tmp$start[si] <= srep_tmp$end[sr] &
                  srep_tmp$end[si] >= srep_tmp$start[sr]) {
                end_overlap = as.numeric(pmin(srep_tmp$end[si], srep_tmp$end[sr]))
                start_overlap = as.numeric(pmax(srep_tmp$start[si], srep_tmp$start[sr]))
                duration_overlap = end_overlap - start_overlap
                duration_sib = as.numeric(srep_tmp$end[si]) - as.numeric(srep_tmp$start[si])
                # percentage of overlap
                perc_overlap = round(100 * (duration_overlap / duration_sib), digits = 1)
                if (srep_tmp$type[sr] == "nonwear") {
                  srep_tmp$SIBoverlapNonwear[si] = perc_overlap
                } else if (srep_tmp$type[sr] == "nap") {
                  srep_tmp$SIBoverlapNap[si] = perc_overlap
                } else if (srep_tmp$type[sr] == "sleeplog") {
                  srep_tmp$SIBoverlapSleeplog[si] = perc_overlap
                }
              }
              # Selfreport behaviour overlap with SIB
              if (srep_tmp$start[sr] <= srep_tmp$end[si] &
                  srep_tmp$end[sr] >= srep_tmp$start[si]) {
                end_overlap = as.numeric(pmin(srep_tmp$end[si], srep_tmp$end[sr]))
                start_overlap = as.numeric(pmax(srep_tmp$start[si], srep_tmp$start[sr]))
                duration_overlap = end_overlap - start_overlap
                duration_sr = as.numeric(srep_tmp$end[sr]) - as.numeric(srep_tmp$start[sr])
                # percentage of overlap
                perc_overlap = round(100 * (duration_overlap / duration_sr), digits = 1)
                if (srep_tmp$type[sr] == "nonwear") {
                  srep_tmp$NonwearOverlapSIB[sr] = perc_overlap
                } else if (srep_tmp$type[sr] == "nap") {
                  srep_tmp$NapOverlapSIB[sr] = perc_overlap
                } else if (srep_tmp$type[sr] == "sleeplog") {
                  srep_tmp$SleeplogOverlapSIB[sr] = perc_overlap
                }
              }
            }
          }
        }
      }
      # Identify where segments overlap
      sibs_indices = which(srep_tmp$type == "sib")
      nap_indices = which(srep_tmp$type == "nap")
      nonwear_indices = which(srep_tmp$type == "nonwear")
      # sleeplog_indices = which(srep_tmp$type == "sleeplog")
      SIBoverlapNap_indices = which(srep_tmp$SIBoverlapNap != 0)
      SIBoverlapNonwear_indices = which(srep_tmp$SIBoverlapNonwear != 0)
      SIBoverlapSleeplog_indices = which(srep_tmp$SIBoverlapSleeplog != 0)
      NapOverlapSIB_indices = which(srep_tmp$NapOverlapSIB != 0)
      NonwearOverlapSIB_indices = which(srep_tmp$NonwearOverlapSIB != 0)
      SleeplogOverlapSIB_indices = which(srep_tmp$NonwearOverlapSIB != 0)
      
      # Count number of occurrences (do not count sleeplog because not informative)
      dsummary[di,fi:(fi + 6)] = c(length(sibs_indices),
                                   length(nap_indices),
                                   length(nonwear_indices),
                                   length(SIBoverlapNap_indices),
                                   length(SIBoverlapNonwear_indices),
                                   length(NapOverlapSIB_indices),
                                   length(NonwearOverlapSIB_indices))
      # if (any(ds_names[fi:(fi + 6)] != c("nbouts_day_denap", "nbouts_day_srnap", "nbouts_day_srnonw", 
      #                                    "noverl_denap_srnap", "noverl_denap_srnonw",
      #                                    "noverl_srnap_denap", "noverl_srnonw_denap"))) {
      #   stop("mismatch in columnnames 2")
      # }
      fi = fi + 7      
      
      # mean and total duration in sib per day
      if (length(sibs_indices) > 0) {
        dsummary[di,fi:(fi + 1)] = c(mean(srep_tmp$duration[sibs_indices]),
                                     sum(srep_tmp$duration[sibs_indices]))
      } else {
        dsummary[di,fi:(fi + 1)] = c(0, 0)
      }
      # if (any(ds_names[fi:(fi + 1)] != c("frag_mean_dur_denap_day", "dur_day_denap_min"))) {
      #   stop("mismatch in columnnames 3")
      # }
      fi = fi + 2
      # mean and total duration in self-reported naps per day
      if (length(nap_indices) > 0) {
        srep_tmp$duration[nap_indices] = (as.numeric(srep_tmp$end[nap_indices]) -
                                            as.numeric(srep_tmp$start[nap_indices])) / 60
        dsummary[di,fi:(fi + 1)] = c(mean(srep_tmp$duration[nap_indices]),
                                     sum(srep_tmp$duration[nap_indices]))
      } else {
        dsummary[di,fi:(fi + 1)] = c(0, 0)
      }
      # if (any(ds_names[fi:(fi + 1)] != c("frag_mean_dur_srnap_day", "dur_day_srnap_min"))) {
      #   stop("mismatch in columnnames 4")
      # }
      fi = fi + 2
      # mean and total duration in self-reported nonwear per day
      if (length(nonwear_indices) > 0) {
        dsummary[di,fi:(fi + 1)] = c(mean(srep_tmp$duration[nonwear_indices]),
                                     sum(srep_tmp$duration[nonwear_indices]))
      } else {
        dsummary[di,fi:(fi + 1)] = c(0, 0)
      }
      # if (any(ds_names[fi:(fi + 1)] != c("frag_mean_dur_srnonw_day", "dur_day_srnonw_min"))) {
      #   stop("mismatch in columnnames 5")
      # }
      fi = fi + 2
      so = list(fi = fi, ds_names = ds_names, dsummary = dsummary, di = di)
      # Self-reported naps
      so = summarise_overlap(
        srep_tmp,
        X = "SIBoverlapNap",
        Y = "NapOverlapSIB",
        xi = SIBoverlapNap_indices,
        yi = NapOverlapSIB_indices,
        name = "srnap",
        sumobject = so
      )
      # Self-reported nonwear
      so = summarise_overlap(
        srep_tmp,
        X = "SIBoverlapNonwear",
        Y = "NonwearOverlapSIB",
        xi = SIBoverlapNonwear_indices,
        yi = NonwearOverlapSIB_indices,
        name = "srnonw",
        sumobject = so
      )
      dsummary = so$dsummary; ds_names = so$ds_names; fi = so$fi
      rm(srep_tmp)
    } else {
      fi  =  fi + 25
    }
  } else {
    fi  =  fi + 26
  }
  invisible(list(fi = fi, di = di, ds_names = ds_names, dsummary = dsummary, ts = ts))
}

filterNonwearNight = function(r1, metalong, qwindowImp, desiredtz,
                              params_cleaning, ws2) {
  nonwearFiltermaxHours  = params_cleaning[["nonwearFiltermaxHours"]]
  nonwearFilterWindow = params_cleaning[["nonwearFilterWindow"]]
  nonwearEventsFiltered = nonwearHoursFiltered = 0
  # Identify method to use
  if (!is.null(nonwearFilterWindow)) {
    filter_method = 1 # set window as provided by user
  } else {
    # If nonwearFilterWindow is not provided then
    # look for qwindow is available
    if (is.null(qwindowImp)) {
      stop(paste0("Please specify parameter nonwearFilterWindow or ",
                  "qwindow as diary with columns to define the window for ",
                  "filtering short nonwear. See documentation for ",
                  "parameter nonwearFiltermaxHours"), call. = FALSE)
    }
    if (inherits(qwindowImp, "data.frame")) {
      filter_method = 2 # Use it as a dataframe
    } else {
      filter_method = 1 
      nonwearFilterWindow = qwindowImp
    }
  }
  # Get metalong timestamps
  metalong$time_POSIX = iso8601chartime2POSIX(metalong$timestamp, tz = desiredtz)
  metalong$hour = as.numeric(format(metalong$time_POSIX, "%H")) + as.numeric(format(metalong$time_POSIX, "%M")) / 60
  # Mask short nonwear during the night
  x = rle(as.numeric(r1))
  if (length(x$values) != 1) {
    r1B = as.data.frame(lapply(x, rep, times =  x$lengths))
    r1B$hr = metalong$hour
    r1B$filter_method = filter_method
    r1B$lengths = (r1B$lengths * ws2) / 3600 # convert unit from epoch to hours
    r1B$filterWindow = 0
    # Define night window
    if (filter_method == 1) { 
      # set window same for all recordings
      if (nonwearFilterWindow[1] > nonwearFilterWindow[2]) {
        r1B$filterWindow[which(r1B$hr >= nonwearFilterWindow[1] |
                                 r1B$hr < nonwearFilterWindow[2])] = 1
      } else {
        r1B$filterWindow[which(r1B$hr >= nonwearFilterWindow[1] &
                                 r1B$hr < nonwearFilterWindow[2])] = 1
      }
    } else if (filter_method == 2) { 
      # window defined by diary input from qwindow specific per participant
      r1B$date = as.Date(metalong$time_POSIX)
      for (qi in 1:nrow(qwindowImp)) {
        qwindow_temp = qwindowImp$qwindow_values[[qi]]
        # Only consider window when both start and end are reported
        available = c(length(grep(pattern = "inbed|sleeponset|lightsout", x = qwindowImp$qwindow_names[[qi]])) > 0,
                      length(grep(pattern = "outbed|wakeup|lightsoff", x = qwindowImp$qwindow_names[[qi]])) > 0)
        # filter only SPT and time in bed reports
        qwindow_temp = qwindow_temp[grep(pattern = "bed|wakeup|sleeponset|lights", x = qwindowImp$qwindow_names[[qi]])]
        qwindow_temp = sort(qwindow_temp)
        # Ignore if somehow only teh default window is available
        isDefaultWindow = length(qwindow_temp) == 2 && qwindow_temp[1] == 0 && qwindow_temp[2] == 24
        if (length(qwindow_temp) > 1 && isDefaultWindow == FALSE && all(available)) {
          # convert to continuous scale to ease finding start and end
          below18 = which(qwindow_temp < 18)
          if (length(below18) > 0) {
            qwindow_temp[below18] = qwindow_temp[below18] + 24
          }
          start = min(qwindow_temp)
          end = max(qwindow_temp)
          if (length(below18) > 0) {
            start = ifelse(start >= 24, yes = start - 24, no = start)
            end = ifelse(end >= 24, yes = end - 24, no = end)
          }
          if (start > end) {
            r1B$filterWindow[which(r1B$date == qwindowImp$date[qi] & 
                                     r1B$hr >= start |
                                     r1B$hr < end)] = 1
          } else {
            r1B$filterWindow[which(r1B$date == qwindowImp$date[qi] & 
                                     r1B$hr >= start &
                                     r1B$hr < end)] = 1
          }
        } else {
          # If diary has one date missing use default window
          r1B$filter_method[which(r1B$hr >= 0 &
                                    r1B$hr < 6)] = 4
          # use midnight - 6am as fall back option
          r1B$filterWindow[which(r1B$hr >= 0 &
                                   r1B$hr < 6)] = 1
        }
      }
    }
    # Identify short nonwear during the night
    short_nonwear_night = which(r1B$values == 1 &
                                  r1B$lengths < nonwearFiltermaxHours &
                                  r1B$filterWindow == 1)
    
    if (length(short_nonwear_night) > 0) {
      nonwearHoursFiltered = (length(which(diff(short_nonwear_night) == 1)) * ws2) / 3600
      nonwearEventsFiltered = length(which(diff(short_nonwear_night) != 1)) + 1
      r1[short_nonwear_night] = 0
    }
  }
  invisible(list(r1 = r1, nonwearHoursFiltered = nonwearHoursFiltered,
                 nonwearEventsFiltered = nonwearEventsFiltered))
}

is.ISO8601 = function(x) {
  is.ISO = FALSE
  if (is.character(x)) {
    NNeg = length(unlist(strsplit(x,"[-]")))
    NPos = length(unlist(strsplit(x,"[+]")))
    if (NPos == 2 | NNeg == 4 | NNeg == 2) {
      is.ISO = TRUE
    }
  }
  return(is.ISO)
}

library(GGIR)
context("is.ISO8601")
test_that("is.ISO8601 is able to detect ISO8601", {
  # west of greenwhich with negative timezone
  x = "2017-01-01 13:15:12"
  expect_equal(is.ISO8601(x),FALSE)
  tz = "America/Aruba"
  x_converted = as.character(POSIXtime2iso8601(x,tz))
  expect_equal(is.ISO8601(x_converted),TRUE)
  
  # east of greenwhich with positive timezone
  x = "2017-01-01 13:15:12"
  tz = "Europe/Amsterdam"
  x_converted = as.character(POSIXtime2iso8601(x,tz))
  expect_equal(is.ISO8601(x_converted),TRUE)
})

g.report.part2 = function(metadatadir = c(), f0 = c(), f1 = c(), 
                          store.long = FALSE, params_output, myfun = c(), verbose = TRUE,
                          desiredtz = "") {
  ms2.out = "/meta/ms2.out"
  if (file.exists(paste0(metadatadir,ms2.out))) {
    if (length(dir(paste0(metadatadir,ms2.out))) == 0) {
      try.generate.report = FALSE
    } else {
      try.generate.report = TRUE
    }
  } else {
    try.generate.report = FALSE
  }
  if (try.generate.report == TRUE) {
    #---------------------------------
    # Specifying directories with meta-data and extracting filenames
    path = paste0(metadatadir, "/meta/basic/")  #values stored per long epoch, e.g. 15 minutes
    fnames = dir(path) # part 1
    fnames.ms2 = dir(paste0(metadatadir,ms2.out))  #part 2
    #---------------------------------
    SUMMARY = daySUMMARY = c()
    
    if (length(f0) ==  0) f0 = 1
    if (length(f1) ==  0) f1 = length(fnames)
    if (f0 > length(fnames)) f0 = 1
    if (f1 > length(fnames)) f1 = length(fnames)
    
    #-----------------------------
    # Loop through all the files
    for (i in f0:f1) {
      if (verbose == TRUE) cat(paste0(" ",i))
      # First load part 1 data
      M = c()
      part1_fname2read = paste0(path, fnames[i])
      try(expr = {load(part1_fname2read)}, silent = TRUE) #reading RData-file
      # Checks for availability of file and data
      if (length(M) == 0) {
        tmp = unlist(strsplit(part1_fname2read, "/"))
        part1_fname2read_display = paste0(tmp[length(tmp) - 1], "/", tmp[length(tmp)])
        warning(paste0("Struggling to read: ", part1_fname2read_display))
        next
      }
      fname = as.character(unlist(strsplit(fnames[i],"eta_"))[2])
      selp = which(fnames.ms2 == fname)
      if (length(selp) == 0 ) {
        if (verbose == TRUE) cat(paste0("File ", fname, " not available in part 2"))
      }
      # If part 1 milestone data indicates that file was useful
      # try loading part 2
      if (M$filecorrupt == FALSE & M$filetooshort == FALSE & length(selp) > 0) { 
        # Load part 2 data
        SUM = IMP = c()
        part2_fname2read = paste0(metadatadir, ms2.out, "/", fnames.ms2[selp])
        try(expr = {load(file = part2_fname2read)}, silent = TRUE)
        # Checks for availability of file and data
        if (length(IMP) == 0) {
          warning(paste0("Struggling to read: ", part2_fname2read))
        }
        if (M$filecorrupt == FALSE & M$filetooshort == FALSE) {
          if (i == 1 | i == f0) {
            SUMMARY = SUM$summary
            daySUMMARY = SUM$daysummary
          } else {
            bind_with_prev_data = function(df1, df2) {
              df1 = data.table::rbindlist(list(df1, df2), fill = TRUE)
              df1 = as.data.frame(df1)
              return(df1)
            }
            SUMMARY = bind_with_prev_data(SUMMARY, SUM$summary)
            daySUMMARY = bind_with_prev_data(daySUMMARY, SUM$daysummary)
          }
        }
      }
      if (length(SUMMARY) == 0 | length(daySUMMARY) == 0) {
        warning("No summary data available to be stored in csv-reports", call. = FALSE)
      }
      #-----------------
      # create data quality report
      if (length(C$cal.error.end) == 0) C$cal.error.end = " "
      if (M$filetooshort == TRUE | M$filecorrupt == TRUE) {
        C$cal.error.start = 0
        C$npoints = 0
      }
      tm = which(colnames(M$metalong) == "temperaturemean")
      if (length(tm) > 0) {
        tmean = as.character(mean(as.numeric(as.matrix(M$metalong[1:(nrow(M$metalong) - 1), tm]))))
      } else {
        tmean = ""
      }
      #=========
      if (is.null(I$sf) == FALSE) {
        header = I$header
        hnames = rownames(header)
        hvalues = as.character(as.matrix(header))
        pp = which(hvalues == "")
        hvalues[pp] = c("not stored in header")
      }
      mon = I$monn
      
      if (mon == "genea") {
        deviceSerialNumber = hvalues[which(hnames == "Serial_Number")] #serial number
      } else if (mon == "geneactive") {
        if ("Device_Unique_Serial_Code" %in% hnames) {
          deviceSerialNumber = hvalues[which(hnames == "Device_Unique_Serial_Code")] #serial number
          if (I$dformn == "csv") { #if it was stored in csv-format then underscores were replaced by spaces (by company)
            deviceSerialNumber = hvalues[which(hnames == "Device Unique Serial Code")] #serial number
          }
        } else {
          deviceSerialNumber = hvalues[which(hnames == "serial_number")] #serial number
        }
      } else if (mon == "actigraph" | mon == "axivity" | mon == "verisense") {
        deviceSerialNumber = "not extracted"
      } else if (I$monc %in% c(5, 7, 94, 96, 97, 98, 99)) { #movisense 5, parmay matrix 7, Actiwatch 98, Sensewear 99
        deviceSerialNumber = "not extracted"
      } else if (I$monc == 0) {
        if (header != "no header") {
          deviceSerialNumber = hvalues[which(hnames == "device_serial_number")]
          if (length(deviceSerialNumber) == 0) {
            deviceSerialNumber = "not extracted"
          }
        } else {
          deviceSerialNumber = "not extracted"
        }
      }
      if (length(C$cal.error.start) == 0) {
        C$cal.error.start = " "
      }
      if (length(C$npoints) == 0) {
        C$npoints = " "
      }
      if (length(C$tempoffset) == 0) {
        C$tempoffset = c(0, 0, 0)
      }
      if (length(M$NFilePagesSkipped) == 0) M$NFilePagesSkipped = 0 # to make the code work for historical part1 output.
      fname2store = unlist(strsplit(fnames[i], "eta_|[.]RDat"))[2]
      QC = data.frame(filename = fname2store,
                      file.corrupt = M$filecorrupt,
                      file.too.short = M$filetooshort,
                      use.temperature = C$use.temp,
                      scale.x = C$scale[1], scale.y = C$scale[2], scale.z = C$scale[3],
                      offset.x = C$offset[1], offset.y = C$offset[2], offset.z = C$offset[3],
                      temperature.offset.x = C$tempoffset[1],  temperature.offset.y = C$tempoffset[2],
                      temperature.offset.z = C$tempoffset[3],
                      cal.error.start = C$cal.error.start,
                      cal.error.end = C$cal.error.end,
                      n.10sec.windows = C$npoints,
                      n.hours.considered = C$nhoursused, QCmessage = C$QCmessage, mean.temp = tmean,
                      device.serial.number = deviceSerialNumber,
                      NFilePagesSkipped = M$NFilePagesSkipped, stringsAsFactors = FALSE)
      if (is.null(I$sf) == TRUE) {
        # When ActiGraph is corrupt
        # Note that QC shape is consistent here
        if (i == 1 | i == f0) {
          QCout = QC
        } else {
          # fill up missing columns with NA
          QC[setdiff(names(QCout), names(QC))] <- NA
          QCout = rbind(QCout, QC)
        }
        next
      }
      
      filehealth_cols = grep(pattern = "filehealth", x = names(SUMMARY), value = FALSE)
      if (length(filehealth_cols) > 0) {
        # migrate filehealth columns to QC report, only applicable to Axivity data
        QC = merge(x = QC, y = SUMMARY[, c(which(names(SUMMARY) == "filename"), filehealth_cols)], by = "filename")
        SUMMARY = SUMMARY[, -filehealth_cols]
      }
      if (i == 1 | i == f0) {
        QCout = QC
      } else {
        n1 = ncol(QCout)
        n2 = ncol(QC)
        # n1 and n2 are expected to be consistent, but in case of Axivity possibly not
        if (n1 > n2) {
          QC = cbind(QC, matrix(" ", 1, (n1 - n2)))
          colnames(QC) = colnames(QCout)
          QC = QC[, colnames(QCout)] # reorder to match order of QCout
        } else if (n1 < n2) {
          newcolnames = colnames(QC)[which(colnames(QC) %in% colnames(QCout) == FALSE)]
          newcols = (n1 + 1):(n1 +  length(newcolnames))
          QCout = cbind(QCout, matrix(" ", 1, n2 - n1))
          colnames(QCout)[newcols] = newcolnames
          QCout = QCout[, colnames(QC)] # reorder to match order of QC
        }
        QCout = rbind(QCout, QC)
      }
    }
    # tidy up memory
    if (exists("IMP")) rm(IMP)
    rm(M); rm(I)

    #--------------------------------------
    # Store Event reports
    # split daySUMMARY in two files and reoder EventVariable names if they exist
    ds_names = names(daySUMMARY)
    EventVars = grep(pattern = "ExtFunEvent_", x = ds_names, value = FALSE)
    NotEventVars = grep(pattern = "ExtFunEvent_", x = ds_names, value = FALSE, invert = TRUE)
    eventName = "event"
    if (length(myfun) > 0) {
      if ("name" %in% names(myfun)) {
        eventName =  myfun$name
      }
    }
    if (length(EventVars) > 0) {
      dayEVENTSUMMARY = daySUMMARY[ , c("ID", "filename", "calendar_date",
                                        "N valid hours", "N hours", "weekday",
                                        sort(names(daySUMMARY[, EventVars])))]
      names(dayEVENTSUMMARY) = gsub(pattern = "ExtFunEvent_", replacement = "", x = names(dayEVENTSUMMARY))
      daySUMMARY = daySUMMARY[,NotEventVars]
      dayEVENTSUMMARY_clean = tidyup_df(dayEVENTSUMMARY)
      dayEVENTSUMMARY_clean = addSplitNames(dayEVENTSUMMARY_clean)  # If recording was split
      data.table::fwrite(x = dayEVENTSUMMARY_clean,
                         file = paste0(metadatadir, "/results/part2_day", eventName, "summary.csv"),
                         row.names = F, na = "", sep = params_output[["sep_reports"]],
                         dec = params_output[["dec_reports"]])
    }
    # split SUMMARY in two files and reoder EventVariable names if they exist
    s_names = names(SUMMARY)
    EventVars = grep(pattern = "ExtFunEvent_", x = s_names, value = FALSE)
    NotEventVars = grep(pattern = "ExtFunEvent_", x = s_names, value = FALSE, invert = TRUE)
    if (length(EventVars) > 0) {
      EVENTSUMMARY = SUMMARY[ , c("ID", "filename", "start_time",
                                  "wear_dur_def_proto_day",
                                  sort(names(SUMMARY[, EventVars])))]
      names(EVENTSUMMARY) = gsub(pattern = "ExtFunEvent_", replacement = "", x = names(EVENTSUMMARY))
      SUMMARY = SUMMARY[,NotEventVars]
      EVENTSUMMARY_clean = tidyup_df(EVENTSUMMARY)
      EVENTSUMMARY_clean = addSplitNames(EVENTSUMMARY_clean)  # If recording was split
      data.table::fwrite(x = EVENTSUMMARY_clean,
                         file =  paste0(metadatadir, "/results/part2_", eventName, "summary.csv"),
                         row.names = F, na = "", sep = params_output[["sep_reports"]],
                         dec = params_output[["dec_reports"]])
    }
    #-----------------------------
    # tidy up data.frames
    if (length(SUMMARY) > 0 & length(daySUMMARY) > 0) {
      SUMMARY_clean = tidyup_df(SUMMARY)
      daySUMMARY_clean = tidyup_df(daySUMMARY)
      daySUMMARY_clean$start_time = daySUMMARY_clean$calendar_date
      # reorder to have starttime next to calendar_date
      old_vars = which(colnames(daySUMMARY_clean) != "start_time")
      new_var = which(colnames(daySUMMARY_clean) == "start_time")
      daySUMMARY_clean = daySUMMARY_clean[, c(old_vars[1:2], new_var, old_vars[3:length(old_vars)])]
      # format calendar dates
      dd = iso8601chartime2POSIX(daySUMMARY_clean$calendar_date, tz = desiredtz) 
      daySUMMARY_clean$calendar_date = format(dd, format = "%Y-%m-%d")
      #===============================================================================
      # store final matrices again
      SUMMARY_clean = addSplitNames(SUMMARY_clean)  # If recording was split
      daySUMMARY_clean = addSplitNames(daySUMMARY_clean)  # If recording was split
      data.table::fwrite(x = SUMMARY_clean, file = paste0(metadatadir, "/results/part2_summary.csv"),
                         row.names = F, na = "", sep = params_output[["sep_reports"]],
                         dec = params_output[["dec_reports"]])
      data.table::fwrite(x = daySUMMARY_clean, paste0(metadatadir, "/results/part2_daysummary.csv"),
                         row.names = F, na = "", sep = params_output[["sep_reports"]],
                         dec = params_output[["dec_reports"]])
      if (store.long == TRUE) { # Convert daySUMMARY to long format if there are multiple segments per day
        df = g.convert.part2.long(daySUMMARY)
        df_clean = tidyup_df(df)
        df_clean = addSplitNames(df_clean)  # If recording was split
        data.table::fwrite(x = df_clean, file = paste0(metadatadir, "/results/part2_daysummary_longformat.csv"),
                           row.names = F, na = "", sep = params_output[["sep_reports"]],
                           dec = params_output[["dec_reports"]])
      }
      QCout = addSplitNames(QCout)  # If recording was split
      data.table::fwrite(x = QCout, file = paste0(metadatadir, "/results/QC/data_quality_report.csv"),
                         row.names = F, na = "",
                         sep = params_output[["sep_reports"]],
                         dec = params_output[["dec_reports"]])
    }
  }
}

g.part3 = function(metadatadir = c(), f0, f1, myfun = c(),
                   params_sleep = c(), params_metrics = c(),
                   params_output = c(),
                   params_general = c(), verbose = TRUE, ...) {
  #----------------------------------------------------------
  # Extract and check parameters
  input = list(...)
  params = extract_params(params_sleep = params_sleep,
                          params_metrics = params_metrics,
                          params_general = params_general,
                          params_output = params_output, input = input,
                          params2check = c("sleep", "metrics",
                                           "general", "output")) # load default parameters
  params_sleep = params$params_sleep
  params_metrics = params$params_metrics
  params_general = params$params_general
  params_output = params$params_output
  
  checkMilestoneFolders(metadatadir, partNumber = 3)
  #------------------------------------------------------
  fnames = dir(paste(metadatadir,"/meta/ms2.out", sep = ""))
  if (f1 > length(fnames) | f1 == 0) f1 = length(fnames)
  if (f0 > length(fnames) | f0 == 0) f0 = 1
  #========================================================================
  # check which files have already been processed, such that no double work is done
  # ffdone a matrix with all the binary filenames that have been processed
  ffdone = fdone = dir(paste(metadatadir,"/meta/ms3.out", sep = ""))
  if (length(fdone) > 0) {
    for (ij in 1:length(fdone)) {
      tmp = unlist(strsplit(fdone[ij], ".RData"))
      ffdone[ij] = tmp[1]
    }
  } else {
    ffdone = c()
  }
  #=========================================================
  # Declare core functionality, which at the end of this g.part3 is either
  # applied to the file in parallel with foreach or serially with a loop
  main_part3 = function(i, metadatadir = c(), f0, f1, myfun = c(),
                        params_sleep = c(), params_metrics = c(),
                        params_output = c(),
                        params_general = c(), fnames, ffdone, verbose) {
    tail_expansion_log = desiredtz_part1 = NULL
    nightsperpage = 7
    FI = file.info(paste(metadatadir, "/meta/ms2.out/", fnames[i], sep = ""))
    if (is.na(FI$size) == TRUE) FI$size = 0
    if (FI$size == 0 | is.na(FI$size) == TRUE | length(FI$size) == 0) {
      if (verbose == TRUE) {
        cat(paste("P3 file ", fnames[i], sep = ""))
        cat("Filename not recognised")
      }
    }
    fname = unlist(strsplit(fnames[i], ".RData"))[1]
    #=========================================================
    #check whether file has already been processed
    #by comparing filename to read with list of processed files
    if (length(ffdone) > 0) {
      # 1=skip this file because it was analysed before"), 0 =do not skip file
      skip = ifelse(test = length(which(ffdone == fname)) > 0, yes = 1, no = 0)
    } else {
      skip = 0
    }
    if (params_general[["overwrite"]] == TRUE) skip = 0
    if (skip == 0) {
      # Load previously stored meta-data from part1.R
      SUM = IMP = M = c()
      load(paste(metadatadir, "/meta/basic/meta_", fnames[i], sep = ""))
      load(paste(metadatadir, "/meta/ms2.out/", fnames[i], sep = ""))
      if (!is.null(desiredtz_part1)) {
        params_general[["desiredtz"]] = desiredtz_part1
      } else {
        desiredtz_part1 = params_general[["desiredtz"]]
      }
      if (M$filecorrupt == FALSE & M$filetooshort == FALSE) {
        if (params_general[["sensor.location"]] == "hip") {
          if (is.null(params_sleep[["longitudinal_axis"]])) {
            # If user does not specify longitudinal axis explicitly then re-use
            # estimated longitudinal axis from part 2
            if (is.na(SUM$summary$if_hip_long_axis_id) == FALSE) {
              params_sleep[["longitudinal_axis"]] = SUM$summary$if_hip_long_axis_id
            } else {
              # If not available default to 2
              params_sleep[["longitudinal_axis"]] = 2
            }
          }
        }
        SLE = g.sib.det(M, IMP, I, twd = c(-12,12),
                        acc.metric = params_general[["acc.metric"]],
                        desiredtz = params_general[["desiredtz"]],
                        myfun = myfun,
                        sensor.location = params_general[["sensor.location"]],
                        params_sleep = params_sleep, zc.scale = params_metrics[["zc.scale"]])
        # Optional correction
        if (params_sleep[["guider_cor_do"]] == TRUE &&
            length(SLE$SPTE_start) > 0 && any(!is.na(SLE$SPTE_start)) &&
            length(SLE$SPTE_end) > 0 && any(!is.na(SLE$SPTE_end))) {
          SLE = g.part3_correct_guider(SLE, desiredtz = params_general[["desiredtz"]],
                                       epochSize = M$windowsizes[1],
                                       params_sleep = params_sleep)
        }
        
        if ("spt_crude_estimate" %in% names(SLE$output)) {
          SLE$output = SLE$output[, -which(names(SLE$output) == "spt_crude_estimate")]
        }
        # SleepRegulartiyIndex calculation
        if (!is.null(SLE$output)) {
          if (nrow(SLE$output) > 2 * 24 * (3600/M$windowsizes[1])) { # only calculate SRI if there are at least two days of data
            SleepRegularityIndex = CalcSleepRegularityIndex(data = SLE$output,
                                                            epochsize = M$windowsizes[1],
                                                            desiredtz = params_general[["desiredtz"]],
                                                            SRI1_smoothing_wsize_hrs = params_sleep[["SRI1_smoothing_wsize_hrs"]],
                                                            SRI1_smoothing_frac = params_sleep[["SRI1_smoothing_frac"]])
          } else {
            SleepRegularityIndex = NA
          }
        } else {
          SleepRegularityIndex = NA
        }
        L5list = SLE$L5list
        SPTE_end = SLE$SPTE_end
        SPTE_start = SLE$SPTE_start
        SPTE_corrected = SLE$SPTE_corrected
        tib.threshold = SLE$tib.threshold
        longitudinal_axis = SLE$longitudinal_axis
        part3_guider = SLE$part3_guider
        if (length(SLE$output) > 0 & SLE$detection.failed == FALSE) {
          ID = SUM$summary$ID
          if (is.list(ID)) ID = unlist(ID)
          datename = as.character(unlist(strsplit(format(as.matrix(M$metashort[1]))," "))[1])
          plottitle = " "
          if (params_output[["do.part3.pdf"]] == TRUE) {
            pdf(paste(metadatadir, "/meta/sleep.qc/graphperday_id_", ID, "_",
                      I$filename, ".pdf", sep = ""), width = 8.2, height = 11.7)
            g.sib.plot(SLE, M, I, plottitle ,nightsperpage = nightsperpage, desiredtz = params_general[["desiredtz"]])
            dev.off()
          }
          sib.cla.sum = c()
          sib.cla.sum = g.sib.sum(SLE, M,
                                  ignorenonwear = params_sleep[["ignorenonwear"]],
                                  desiredtz = params_general[["desiredtz"]])
          rec_starttime = IMP$metashort[1,1] # this may be used in g.loadlog to align sleeplog with recording
          ms3out_filename = paste(metadatadir, "/meta/ms3.out/", fname, ".RData",sep = "")
          GGIRversion = utils::packageVersion("GGIR")
          save(sib.cla.sum, L5list, SPTE_end, SPTE_start, tib.threshold, rec_starttime, ID,
               longitudinal_axis, SleepRegularityIndex, tail_expansion_log, GGIRversion,
               part3_guider, desiredtz_part1, SPTE_corrected,
               file = ms3out_filename)
        }
      }
    }
  }
  
  if (params_general[["do.parallel"]] == TRUE) {
    cores = parallel::detectCores()
    Ncores = cores[1]
    if (Ncores > 3) {
      if (length(params_general[["maxNcores"]]) == 0) params_general[["maxNcores"]] = Ncores
      Ncores2use = min(c(Ncores - 1, params_general[["maxNcores"]], (f1 - f0) + 1))
      if (Ncores2use > 1) {
        cl <- parallel::makeCluster(Ncores2use) # not to overload your computer
        parallel::clusterExport(cl = cl,
                                varlist = c(unclass(lsf.str(envir = asNamespace("GGIR"), all = T)),
                                            "MONITOR", "FORMAT"),
                                envir = as.environment(asNamespace("GGIR")))
        parallel::clusterEvalQ(cl, Sys.setlocale("LC_TIME", "C"))
        doParallel::registerDoParallel(cl)
      } else {
        # Don't process in parallel if only one core
        params_general[["do.parallel"]] = FALSE
      }
    } else {
      if (verbose == TRUE) cat(paste0("\nparallel processing not possible because number of available cores (",Ncores,") < 4"))
      params_general[["do.parallel"]] = FALSE
    }
  }
  if (params_general[["do.parallel"]] == TRUE) {
    if (verbose == TRUE) cat(paste0('\n Busy processing ... see ', metadatadir,'/meta/ms3.out', ' for progress\n'))
    # check whether we are indevelopment mode:
    GGIRinstalled = is.element('GGIR', installed.packages()[,1])
    packages2passon = functions2passon = NULL
    GGIRloaded = "GGIR" %in% .packages()
    if (GGIRloaded) { #pass on package
      packages2passon = 'GGIR'
      errhand = 'pass'
    } else {
      # pass on functions
      functions2passon = c("g.sib.det", "g.detecmidnight", "iso8601chartime2POSIX",
                           "g.sib.plot", "g.sib.sum", "HASPT", "HASIB", "CalcSleepRegularityIndex",
                           "extract_params", "load_params", "check_params", 
                           "g.part3_correct_guider", "g.part3_alignIndexVectors")
      errhand = 'stop'
    }
    i = 0 # declare i because foreach uses it, without declaring it
    `%myinfix%` = foreach::`%dopar%`
    output_list = foreach::foreach(i = f0:f1, .packages = packages2passon,
                                   .export = functions2passon, .errorhandling = errhand) %myinfix% {
                                     tryCatchResult = tryCatch({
                                       main_part3(i, metadatadir, f0, f1, myfun,
                                                  params_sleep, params_metrics,
                                                  params_output,
                                                  params_general, fnames, ffdone, verbose)
                                     })
                                     return(tryCatchResult)
                                   }
    
    on.exit(parallel::stopCluster(cl))
    for (oli in 1:length(output_list)) { # logged error and warning messages
      if (is.null(unlist(output_list[oli])) == FALSE) {
        if (verbose == TRUE) cat(paste0("\nErrors and warnings for ", fnames[oli]))
        print(unlist(output_list[oli])) # print any error and warnings observed
      }
    }
  } else {
    errors = list()
    for (i in f0:f1) {
      if (verbose == TRUE) cat(paste0(i, " "))
      function_to_evaluate = expression(
        main_part3(i, metadatadir, f0, f1, myfun,
                   params_sleep, params_metrics,
                   params_output,
                   params_general, fnames, ffdone, verbose)
      )
      if (params_general[["use_trycatch_serial"]] == TRUE) {
        tryCatch(
          eval(function_to_evaluate),
          error = function(e) {
            err_msg = conditionMessage(e)
            errors[[as.character(fnames[i])]] <<- err_msg
          }
        )
      } else {
        eval(function_to_evaluate)
      }
    }
    # show logged errors after the loop:
    if (params_general[["use_trycatch_serial"]] == TRUE && verbose == TRUE) {
      if (length(errors) > 0) {
        cat(paste0("\n\nErrors in part 3... for:"))
        cat(paste0("\n-", names(errors), ": ", unlist(errors), collapse = ""))
      }
    }
  }
}

create_test_sleeplog_csv = function(Nnights = 7,
                                    storagelocation = c(),
                                    advanced = FALSE,
                                    sep = ",",
                                    begin_date = "2016/06/25",
                                    type = "sleeplog") {
  
  # function to create a test sleeplog file needed for testing GGIR
  # Nnights is the number of nights the sleeplog has
  if (length(storagelocation) == 0) storagelocation = getwd()
  if (Nnights == 0) Nnights = 1
  if (advanced == FALSE) {
    sleeplog = as.data.frame(t(c("123",rep(c("23:00:00","07:00:00"),Nnights))), stringsAsFactors = TRUE)
    if (ncol(sleeplog) >= 5) {
      # let's add one night where the person slept untill the afternoon
      sleeplog[, 4:5] = c("03:00:00", "17:00:00") 
    }
    colnames(sleeplog) = c("ID",rep(c("onset","wakeup"),Nnights))
    write.table(
      sleeplog,
      file = paste0(storagelocation, "/testsleeplogfile.csv"),
      row.names = FALSE,
      col.names = TRUE,
      sep = ",",
      fileEncoding = "UTF-8"
    )
  } else {
    times = c("07:00:00", "23:00:00", "13:00:00", "13:30:00", "15:00:00", "15:40:00",  "17:00:00", "17:15:00")
    dates = seq(as.Date(begin_date), as.Date(begin_date) + Nnights, by = 1)
    Cnames = c(
      "_wakeup",
      "_onset",
      "_nap1_start",
      "_nap1_end",
      "_nap2_start",
      "_nap2_end",
      "_nonwear1_off",
      "_nonwear1_on"
    )
    if (type == "bedlog") {
      Cnames[1:2] = c("_bedstart", "_bedend")
    }
    if (type == "both") {
      Cnames = c(Cnames[1:2], c("_bedend", "_bedstart"), Cnames[3:8])
      bedtimes = c("08:00:00", "22:00:00")
      times = c(times[1:2], bedtimes, times[3:8])
    }
    names = "ID"
    log = c()
    for (di in 1:length(dates)) {
      # include day number in the time, to ease testing sleeplog processing
      if (di < 10) {
        times2 = gsub(pattern = ":00:00", replacement = paste0(":00:0",di), x = times)
      } else {
        times2 = gsub(pattern = ":00:00", replacement = paste0(":00:", ifelse(di < 59, yes = di, no = "00")),
                                                               x = times)
      }
      log = c(log, as.character(dates[di]), times2)
      names = c(names, c(paste0("D",di,"_date"),paste0("D",di, Cnames)))
    }
    sleeplog = as.data.frame(t(c("123A",log)), stringsAsFactors = TRUE)
    colnames(sleeplog) = names
    
    data.table::fwrite(sleeplog, file = paste0(storagelocation,"/testsleeplogfile.csv"),
                row.names = FALSE, col.names = TRUE, sep = sep)
  }
}

library(GGIR)
context("gloadlog")
test_that("gloadlog is able to load different log formats", {
  tempdir = "mytestdir"
  if (dir.exists(tempdir) == FALSE) dir.create(tempdir)
  ID = "123A"

  # test that a sleep log with any separator from the set of [,\t |;] can be successfully parsed
  for (sep in c(",", "\t", " ", "|", ";")) {
    create_test_sleeplog_csv(Nnights = 7, advanced = TRUE, sep = sep, type = "sleeplog")
    fn = "testsleeplogfile.csv"
    expect_true(file.exists(fn))

    # rec_starttime = "2016-06-25T20:20:20+0200"
    rec_starttime = format(as.POSIXct("2016-06-25 20:20:20"), "%Y-%m-%dT%H:%M:%S%z")
    save(ID, rec_starttime, file = "mytestdir/dummyms3.RData")
    cat(paste0("\nLocal timezone: ", Sys.timezone()))
    logs1 = g.loadlog(loglocation = fn, coln1 = 2, colid = 1,
                      meta.sleep.folder = tempdir , desiredtz = "")
    # after deprecating nnights, the whole sleeplog is read (it contains 7 nights)
    # amend tests as needed
    expect_equal(nrow(logs1$sleeplog), 7) 
    expect_equal(ncol(logs1$sleeplog), 5)
    expect_equal(logs1$sleeplog$night, as.character(1:7))
    expect_equal(logs1$sleeplog$sleepwake, c("7:0:2", "7:0:3", "7:0:4", "7:0:5", "7:0:6", "7:0:7", "7:0:8"))
    expect_equal(nrow(logs1$nonwearlog), 8)
    expect_equal(ncol(logs1$nonwearlog), 4)
    expect_equal(nrow(logs1$naplog), 8)
    expect_equal(ncol(logs1$naplog), 6)
    
    # start of accelerometer recording 6 days earlier
    rec_starttime = format(as.POSIXct("2016-06-20 20:20:20"), "%Y-%m-%dT%H:%M:%S%z")
    save(ID, rec_starttime, file = "mytestdir/dummyms3.RData")
    logs2 = g.loadlog(loglocation = fn, coln1 = 2, colid = 1,
                     meta.sleep.folder = tempdir , desiredtz = "")
    expect_equal(nrow(logs2$sleeplog), 7)
    expect_equal(ncol(logs2$sleeplog), 5)
    expect_equal(logs2$sleeplog$night, as.character(6:12)) # these are nights in the acc recording
    expect_equal(logs2$sleeplog$sleepwake, c("7:0:2", "7:0:3", "7:0:4", "7:0:5", "7:0:6", "7:0:7", "7:0:8"))
    expect_equal(nrow(logs2$nonwearlog), 8)
    expect_equal(ncol(logs2$nonwearlog), 4)
    expect_equal(nrow(logs2$naplog), 8)
    expect_equal(ncol(logs2$naplog), 6)
    
    # start of accelerometer recording 3 days later
    rec_starttime = format(as.POSIXct("2016-06-29 20:20:20"), "%Y-%m-%dT%H:%M:%S%z")
    save(ID, rec_starttime, file = "mytestdir/dummyms3.RData")
    logs3 = g.loadlog(loglocation = fn, coln1 = 2, colid = 1,
                      meta.sleep.folder = tempdir , desiredtz = "")
    expect_equal(nrow(logs3$sleeplog), 3)
    expect_equal(ncol(logs3$sleeplog), 5)
    expect_equal(logs3$sleeplog$night, as.character(1:3)) # 1:3 because these are the first three nights of the acc recording
    expect_equal(logs3$sleeplog$sleeponset, c("23:0:5", "23:0:6", "23:0:7"))
    expect_equal(nrow(logs3$nonwearlog), 4)
    expect_equal(ncol(logs3$nonwearlog), 4)
    expect_equal(nrow(logs3$naplog), 4)
    expect_equal(ncol(logs3$naplog), 6)
  }
  
  # Test basic sleeplog with only 1 night:
  create_test_sleeplog_csv(Nnights = 1, advanced = FALSE, sep = ",")
  fn = "testsleeplogfile.csv"
  expect_true(file.exists(fn))
  
  # rec_starttime = "2016-06-25T20:20:20+0200"
  rec_starttime = format(as.POSIXct("2016-06-25 20:20:20"), "%Y-%m-%dT%H:%M:%S%z")
  save(ID, rec_starttime, file = "mytestdir/dummyms3.RData")
  logs4 = g.loadlog(loglocation = fn, coln1 = 2, colid = 1,
                    meta.sleep.folder = tempdir , desiredtz = "")
  expect_equal(nrow(logs4$sleeplog), 1)
  expect_equal(ncol(logs4$sleeplog), 5)
  expect_equal(logs4$sleeplog$night, "1")
  
  
  # Test advanced sleeplog with only 1 night:
  create_test_sleeplog_csv(Nnights = 1, advanced = TRUE, sep = ",")
  fn = "testsleeplogfile.csv"
  expect_true(file.exists(fn))
  rec_starttime = format(as.POSIXct("2016-06-25 20:20:20"), "%Y-%m-%dT%H:%M:%S%z")
  save(ID, rec_starttime, file = "mytestdir/dummyms3.RData")
  logs5 = g.loadlog(loglocation = fn, coln1 = 2, colid = 1,
                    meta.sleep.folder = tempdir , desiredtz = "")
  expect_equal(nrow(logs5$sleeplog), 1)
  expect_equal(ncol(logs5$sleeplog), 5)
  expect_equal(logs5$sleeplog$night, "1")
  
  
  # Test advanced bedlog:
  create_test_sleeplog_csv(Nnights = 3, advanced = TRUE, sep = ",", type = "bedlog")
  fn = "testsleeplogfile.csv"
  expect_true(file.exists(fn))
  rec_starttime = format(as.POSIXct("2016-06-25 20:20:20"), "%Y-%m-%dT%H:%M:%S%z")
  save(ID, rec_starttime, file = "mytestdir/dummyms3.RData")
  logs6 = g.loadlog(loglocation = fn, coln1 = 2, colid = 1,
                    meta.sleep.folder = tempdir , desiredtz = "")
  expect_equal(logs6$sleeplog, NULL)
  expect_equal(nrow(logs6$bedlog), 3)
  expect_equal(colnames(logs6$bedlog), c("ID", "night", "duration", "bedstart", "bedend"))
  expect_equal(logs6$bedlog$night, c("1", "2", "3"))
  
  
  # Test advanced bedlog + sleeplog:
  create_test_sleeplog_csv(Nnights = 3, advanced = TRUE, sep = ",", type = "both")
  fn = "testsleeplogfile.csv"
  expect_true(file.exists(fn))
  rec_starttime = format(as.POSIXct("2016-06-25 20:20:20"), "%Y-%m-%dT%H:%M:%S%z")
  save(ID, rec_starttime, file = "mytestdir/dummyms3.RData")
  logs7 = g.loadlog(loglocation = fn, coln1 = 2, colid = 1,
                    meta.sleep.folder = tempdir , desiredtz = "")
  expect_equal(nrow(logs7$sleeplog), 3)
  expect_equal(ncol(logs7$sleeplog), 5)
  expect_equal(nrow(logs7$bedlog), 3)
  expect_equal(ncol(logs7$bedlog), 5)
  expect_equal(logs7$bedlog$bedend[1], "8:0:2")
  expect_equal(logs7$bedlog$bedstart[1], "22:0:1")
  expect_equal(logs7$sleeplog$sleepwake[1], "7:0:2")
  expect_equal(logs7$sleeplog$sleeponset[1], "23:0:1")
  
  if (dir.exists(tempdir)) unlink(tempdir, recursive = TRUE)
  if (file.exists(fn)) unlink(fn, recursive = TRUE)
})

library(GGIR)
context("g.part5.fixmissingnight")
test_that("is able to identify missing night and retrieve sleeplog estimate", {
  nightsummary = data.frame(ID = c(123, 123), night = c(1,3),
                            sleeponset = c(22.92083, 26.92083),
                            wakeup = c(35.99583, 31.16667),
                            SptDuration = c(13.075000, 4.245833),
                            sleepparam = c("T5A5", "T5A5"),
                            guider_onset = c(23, 27),
                            guider_wakeup = c(31, 36),
                            guider_SptDuration = c(8, 9), 
                            error_onset = c(-0.07916667, -0.07916667),
                            error_wake = c(4.995833, -4.833333),
                            error_dur = c(5.075000, -4.754167),
                            fraction_night_invalid = c(0.04166667, 0.00000000),
                            SleepDurationInSpt = c(12.8750, 4.1375),
                            duration_sib_wakinghours = c(4.95000 , 14.77917),
                            number_sib_sleepperiod = c(25 , 14),
                            number_sib_wakinghours = c(6, 17),
                            duration_sib_wakinghours_atleast15min = c(3.741667, 11.154167),
                            sleeponset_ts = c("22:55:15", "02:55:15"), 
                            wakeup_ts = c("11:59:45", "07:10:00"),
                            guider_onset_ts = c("23:00:00", "03:00:00"), 
                            guider_wakeup_ts = c("07:00:00", "17:00:00"),
                            page = c(1, 1),
                            daysleeper = c(0, 0),
                            weekday = c("Thursday", "Saturday"),
                            calendar_date = c("23/6/2016", "25/6/2016"),
                            filename = c("123A_testaccfile.csv.RData", "123A_testaccfile.csv.RData"),
                            cleaningcode = c(0, 0),
                            sleeplog_used = c(1, 1), 
                            acc_available = c(1, 1),
                            guider = c("sleeplog", "sleeplog"))
  
  create_test_sleeplog_csv()
  fn = "testsleeplogfile.csv"
  expect_true(file.exists(fn))
  SLEEPLOG = g.loadlog(loglocation = fn, coln1 = 2, colid = 1)
  
  out = g.part5.fixmissingnight(summarysleep = nightsummary,
                                sleeplog = SLEEPLOG$sleeplog,
                                ID = 123)
  expect_equal(round(out$sleeponset[2], digits = 4), 3)
  expect_equal(round(out$wakeup[2], digits = 4), 17)
  
  # also test g.part5.addfirstwake
  N = 54 * 60 * 60 / 5
  t0 = as.POSIXlt("2016-06-22 22:00:00", tz = "Europe/Amsterdam",
                  origin = "1970-01-01")
  # the values of time are not used in the function, we only care about the length the object time
  time = seq(t0, t0 + ((N * 5) - 1), by = 5) 
  time = POSIXtime2iso8601(time, tz = "")
  diur = c(rep(0, 18720), #awake until second midnight (to trigger detection of first wake)
           rep(1, 5040),  #spt until 7 am next day
           rep(0, 12240),
           rep(1, N - 36000))
  detection = rep(0, N)
  set.seed(300)
  detection[which(runif(n = N, min = 0, max = 1) < 0.1)] = 1
  ACC = rnorm(n = N, mean = 0, sd = 40)
  
  ts = data.frame(time = time, ACC = ACC, diur = diur, sibdetection = detection, guider = "unknown")
  nightsi = grep("00:00:00", time)

  ts = g.part5.addfirstwake(ts, summarysleep = nightsummary, 
                            nightsi = nightsi, sleeplog = SLEEPLOG$sleeplog,
                            ID = 123, Nepochsinhour = 720, 
                            SPTE_end = NA)
  
  if (file.exists(fn)) file.remove(fn)
})

library(GGIR)
context("create_test_sleeplog_csv")
test_that("create_test_sleeplog_csv produces a file", {
  create_test_sleeplog_csv(advanced = FALSE)
  fn = "testsleeplogfile.csv"
  expect_true(file.exists(fn))
  if (file.exists(fn)) file.remove(fn)
  create_test_sleeplog_csv(advanced = TRUE)
  fn = "testsleeplogfile.csv"
  expect_true(file.exists(fn))
  if (file.exists(fn)) file.remove(fn)
})

inspect_binFile_brand = function(filename) {
  mon = "not_recognised"
  # Extract information from the fileheader
  suppressWarnings({fh = readLines(filename, 69)})
  # look for device type field
  suppressWarnings({deviceGeneactiv = grep("Device Type", fh, ignore.case = T)})
  if (length(deviceGeneactiv) > 0) {
    if (grepl("GENEActiv|GENEAsleep", fh[deviceGeneactiv])) mon = MONITOR$GENEACTIV
  } else {
    # check if it is a Matrix device
    raw = readBin(filename, "raw", file.info(filename)$size)
    # Read the header (bytes 513-516)
    header_raw = raw[513:516]
    header = rawToChar(header_raw[header_raw != 0], multiple = FALSE)
    # if matrix device, header is MDTC, otherwise it is a corrupt file
    if (header == "MDTC") mon = MONITOR$PARMAY_MTX
  }
  return(mon)
}

getPart1BasicInfo = function(fn, idloc, tz) {
  M = NULL # load from fn
  load(fn)
  if (is.null(M$metashort)) return()
  hvars = g.extractheadervars(I)
  if (exists("Clist")) {
    ID = NA # If Clist exists then ignore this file as it was previously appended
  } else {
    ID = extractID(hvars, idloc, fname = I$filename)
  }
  start = as.POSIXct(x = M$metashort$timestamp[1], format = "%Y-%m-%dT%H:%M:%S%z", tz = tz)
  end = as.POSIXct(x = M$metashort$timestamp[nrow(M$metashort)], format = "%Y-%m-%dT%H:%M:%S%z", tz = tz)
  info = data.frame(ID = ID, start = start, end = end, filename = fn, brand = I$monn)
  return(info)
}

g.part2 = function(datadir = c(), metadatadir = c(), f0 = c(), f1 = c(),
                   myfun=c(), params_cleaning = c(), params_247 = c(),
                   params_phyact = c(), params_output = c(), params_general = c(),
                   verbose = TRUE, ...) {
  
  #----------------------------------------------------------
  # Extract and check parameters
  input = list(...)
  params = extract_params(params_cleaning = params_cleaning,
                          params_247 = params_247,
                          params_phyact = params_phyact,
                          params_output = params_output,
                          params_general = params_general, input = input,
                          params2check = c("cleaning", "247", "phyact", "output", "general")) # load default parameters
  params_cleaning = params$params_cleaning
  params_247 = params$params_247
  params_phyact = params$params_phyact
  params_output = params$params_output
  params_general = params$params_general
  #-----------------------------
  use_qwindow_as_diary = TRUE # If there is a diary specified via qwindow use it as qwindow
  if (is.numeric(params_247[["qwindow"]])) {
    params_247[["qwindow"]] = params_247[["qwindow"]][order(params_247[["qwindow"]])]
  } else if (is.character(params_247[["qwindow"]])) {
    if (length(grep(pattern = "onlyfilter|filteronly", x = params_247[["qwindow"]])) > 0) {
      # Do not use diary specified for qwindow if it has the word
      # "onlyfilter" or "filteronly", but use it for filterning nighttime nonwear
      # note that this filtering is only use if parameter nonwearFiltermaxHours is specified.
      use_qwindow_as_diary = FALSE 
    }
    tmp_activityDiary_file = paste0(metadatadir, "/meta/activityDiary_", basename(params_247[["qwindow"]]), ".RData")
    
    if (!file.exists(tmp_activityDiary_file) || (file.exists(tmp_activityDiary_file) && 
                                                 file.info(params_247[["qwindow"]])$mtime >= file.info(tmp_activityDiary_file)$mtime)) {
      if (verbose == TRUE) cat("\nConverting activity diary...")
      # This will be an object with numeric qwindow values for all individuals and days
      params_247[["qwindow"]] = g.conv.actlog(params_247[["qwindow"]],
                                              params_247[["qwindow_dateformat"]],
                                              epochSize = params_general[["windowsizes"]][1])
      qwindow = params_247[["qwindow"]]
      save(qwindow, file = tmp_activityDiary_file)
    } else {
      load(tmp_activityDiary_file)
      params_247[["qwindow"]] = qwindow
    }
    rm(qwindow)
  }
  
  #---------------------------------
  # Specifying directories with meta-data and extracting filenames
  path = paste0(metadatadir,"/meta/basic/")  #values stored per long epoch, e.g. 15 minutes
  checkMilestoneFolders(metadatadir, partNumber = 2)
  fnames = dir(path)
  if (f1 > length(fnames)) f1 = length(fnames)
  # create output folders
  ffdone = c()
  ms2.out = "/meta/ms2.out"
  if (file.exists(paste0(metadatadir,ms2.out)) == FALSE) {
    dir.create(file.path(metadatadir,ms2.out))
  }
  csvfolder = "/meta/csv"
  if (params_output[["epochvalues2csv"]] == TRUE) {
    if (file.exists(paste0(metadatadir, csvfolder)) == FALSE) {
      dir.create(file.path(metadatadir,csvfolder))
    }
  }
  ffdone = dir(paste0(metadatadir,ms2.out))
  if (params_output[["do.part2.png"]] == TRUE) {
    QC_plotfolder = paste0(metadatadir, "/results/QC/plots_part2")
    if (!dir.exists(QC_plotfolder)) {
      dir.create(QC_plotfolder)
    }
  }
  #---------------------------------
  # house keeping variables
  daySUMMARY = c()
  if (length(f0) == 0) f0 = 1
  if (length(f1) == 0) f1 = length(fnames)
  #--------------------------------
  # get full file path and folder name if requested by end-user and keep this for storage in output
  foldername = fullfilenames = folderstructure = referencefnames = c()
  if (params_output[["storefolderstructure"]] == TRUE) {
    extractfilenames = function(x) {
      x2 = as.character(unlist(strsplit(x,".RDa"))[1])
      x3 = unlist(strsplit(x2,"eta_"))[2]
      return(x3)
    }
    referencefnames = as.character(sapply(fnames, extractfilenames))
    folderstructure = getfolderstructure(datadir, referencefnames)
    fullfilenames = folderstructure$fullfilenames
    foldername = folderstructure$foldername
  }
  #--------------------------------
  # Loop through all the files
  # fnames = sort(fnames)
  if (f1 > length(fnames)) f1 = length(fnames)
  if (f0 > f1) f0 = 1
  
  #---------------------------------------
  cnt78 = 1
  
  #=========================================================
  # Declare core functionality, which at the end of this g.part2 is either
  # applied to the file in parallel with foreach or serially with a loop
  main_part2 = function(i, ffdone, fnames,
                        metadatadir = c(),
                        myfun=c(), params_cleaning = c(), params_247 = c(),
                        params_phyact = c(), params_output = c(), params_general = c(),
                        path, ms2.out, foldername, fullfilenames, folderstructure, referencefnames,
                        daySUMMARY, csvfolder, cnt78, verbose, use_qwindow_as_diary) {
    Nappended = I_list = tail_expansion_log = desiredtz_part1 = NULL
    if (length(ffdone) > 0) {
      if (length(which(ffdone == as.character(unlist(strsplit(fnames[i], "eta_"))[2]))) > 0) {
        skip = 1 #skip this file because it was analysed before")
      } else {
        skip = 0 #do not skip this file
      }
    } else {
      skip = 0
    }
    if (params_general[["overwrite"]] == TRUE) skip = 0
    M = I = c()
    if (skip == 0) {
      file2read = paste0(path,fnames[i])
      try(expr = {load(file2read)}, silent = TRUE) #reading RData-file
    }
    if (length(M) == 0) skip = 1
    if (skip == 0) {
      filename_dir = c()
      filefoldername = c()
      if (!is.null(desiredtz_part1)) {
        params_general[["desiredtz"]] = desiredtz_part1
      }
      # convert to character/numeric if stored as factor in metashort and metalong
      M$metashort = correctOlderMilestoneData(M$metashort)
      M$metalong = correctOlderMilestoneData(M$metalong)
      # extract ID centrally to be used in GGIR
      if (is.null(I$sf)) {
        M$filecorrupt = TRUE
      } else {
        hvars = g.extractheadervars(I)
        ID = extractID(hvars = hvars, idloc = params_general[["idloc"]], fname = I$filename)
      }
      if (M$filecorrupt == FALSE & M$filetooshort == FALSE) {
        #-----------------------
        # If required by user, ignore specific timewindows for imputation and set them to zeroinstead:
        TimeSegments2Zero = c() # set defaul
        # Check whether csv file exists with start-end end times of timewindows to be ignored (where 0 movement will be assumed)
        if (length(params_cleaning[["TimeSegments2ZeroFile"]]) > 0) {
          TimeSegments2ZeroAll = data.table::fread(params_cleaning[["TimeSegments2ZeroFile"]], data.table = FALSE)
          # Check whether this individual is part of the file
          filei = which(TimeSegments2ZeroAll$filename == as.character(unlist(strsplit(fnames[i], "eta_"))[2]))
          if (length(filei) > 0) {
            # If yes, load the timestamps that indicate the windows to be ignored
            TimeSegments2Zero = TimeSegments2ZeroAll[filei,]
            # Check that they fall withint the measurement
            TimeSegments2Zero$windowstart = as.POSIXlt(TimeSegments2Zero$windowstart,tz = params_general[["desiredtz"]])
            TimeSegments2Zero$windowend = as.POSIXlt(TimeSegments2Zero$windowend,tz = params_general[["desiredtz"]])
            timespan0 = iso8601chartime2POSIX(M$metashort$timestamp[1], tz = params_general[["desiredtz"]])
            timespan1 = iso8601chartime2POSIX(M$metashort$timestamp[nrow(M$metashort)],
                                              tz = params_general[["desiredtz"]])
            validtimes = which(TimeSegments2Zero$windowstart > timespan0 &
                                 TimeSegments2Zero$windowend < timespan1)
            if (length(validtimes) > 0) {
              TimeSegments2Zero = TimeSegments2Zero[validtimes,c("windowstart","windowend")]
              
            } else {
              TimeSegments2Zero = c()
            }
          }
        }
        if (length(myfun) > 0) {
          for (mi in 1:length(myfun$outputtype)) {
            if (myfun$outputtype[mi] == "character") {
              # At the moment we do not have a strategy in place on how to impute categorical variables
              # produced by external functions. Therefore, for the moment ignore these variables until
              # there is a plan.
              M$metashort = M$metashort[,-which(names(M$metashort) %in% myfun$colnames == TRUE)]
            }
          }
        }
        qwindowImp = params_247[["qwindow"]]
        if (use_qwindow_as_diary == FALSE) {
          # reset qwindow to default, because it is only used
          # for filtering short nighttime nonwear 
          params_247[["qwindow"]] = c(0, 24)
        }
        if (inherits(qwindowImp, "data.frame")) {
          qwindowImp = qwindowImp[which(qwindowImp$ID == ID),]
          if (nrow(qwindowImp) == 0) {
            qwindowImp = c(0, 6) # If participant not present in diary
          }
        } else {
          qwindowImp = NULL
        }
        IMP = g.impute(M, I,
                       params_cleaning = params_cleaning,
                       dayborder = params_general[["dayborder"]],
                       desiredtz = params_general[["desiredtz"]],
                       TimeSegments2Zero = TimeSegments2Zero,
                       acc.metric = params_general[["acc.metric"]],
                       ID = ID, qwindowImp = qwindowImp)
        
        if (params_output[["do.part2.png"]] == TRUE & length(IMP) > 0) {
          Ndays = (nrow(M$metalong) * M$windowsizes[2]) / (3600 * 24)
          if (params_cleaning[["maxdur"]] != 0) {
            durplot = params_cleaning[["maxdur"]]
          } else {
            durplot = 7 #how many DAYS to plot? (only used if maxdur is not specified as a number above zero)
          }
          IDname = gsub(pattern = "meta_|[.]bin|[.]RData|[.]csv|[.]gt3x", replacement = "", x = fnames[i])
          png(paste0(QC_plotfolder, "/plot_", IDname, ".png"),
              width = 7, height = 7, units = "in", res = 300)
          g.plot(IMP, M, I,
                 durplot = ifelse(test = Ndays > durplot, yes = Ndays, no = durplot))
          dev.off()
        }
        if (params_cleaning[["do.imp"]] == FALSE) { #for those interested in sensitivity analysis
          IMP$metashort = M$metashort
          # IMP$metalong = M$metalong
        }
        # do not use expanded time with expand_tail_max_hours in summary reports
        if (length(tail_expansion_log) != 0) {
          M_bu = M; IMP_bu = IMP # back up to be reversed later on
          expanded_time_short = which(IMP$r5long == -1)
          expanded_time_long = which(IMP$rout$r5 == -1)
          M$metashort = M$metashort[-expanded_time_short,]
          IMP$metashort = IMP$metashort[-expanded_time_short,]
          M$metalong = M$metalong[-expanded_time_long,]
          IMP$rout = IMP$rout[-expanded_time_long,]
        }
        
        SUM = g.analyse(I, C, M, IMP,
                        params_247 = params_247,
                        params_phyact = params_phyact,
                        params_general = params_general,
                        params_cleaning = params_cleaning,
                        myfun = myfun, ID = ID)
        RDname = as.character(unlist(strsplit(fnames[i], "eta_"))[2])
        # reset M and IMP so that they include the expanded time (needed for sleep detection in parts 3 and 4)
        if (length(tail_expansion_log) != 0) {
          M = M_bu
          IMP = IMP_bu
        }
        if (params_output[["epochvalues2csv"]] == TRUE) {
          if (length(IMP$metashort) > 0) {
            data.table::fwrite(IMP$metashort, paste0(metadatadir, "/", csvfolder, "/", RDname, ".csv"),
                               row.names = FALSE, sep = params_output[["sep_reports"]],
                               dec = params_output[["dec_reports"]])
          }
        }
        if (M$filecorrupt == FALSE & M$filetooshort == FALSE) {
          if (cnt78 == 1) {
            SUMMARY = SUM$summary
            daySUMMARY = SUM$daysummary
            cnt78 = 2
          }
        }
        
        NumberRDinFilename = length(unlist(strsplit(RDname,"[.]RD")))
        if (NumberRDinFilename == 1) { # to avoid getting .RData.RData
          RDname = paste0(RDname,".RData")
        }
        if (params_output[["storefolderstructure"]] == TRUE) {
          SUM$daysummary$filename_dir = fullfilenames[i] #full filename structure
          SUM$daysummary$foldername = foldername[i] #lowest foldername
          SUM$summary$filename_dir = fullfilenames[i]
          SUM$summary$foldername = foldername[i]
        }
        if (!is.null(params_general[["maxRecordingInterval"]])) {
          getValue = function(x, name) {
            if (name %in% names(x)) {
              out = x[name]
            } else {
              out = NULL
            }
            return(out)
          }
          if (!is.null(I_list)) {
            # This recording is the result of appending 2 or more recordings
            NappendedRecordings = Nappended
            sf_appendedRecordings = paste0(unlist(lapply(X = I_list, FUN = getValue, name = "sf")), collapse = " ")
            names_appendedRecordings = paste0(unlist(lapply(X = I_list, FUN = getValue, name = "filename")), collapse = " ")
            # Note if overlap is positive it is overlap, if it is negative there was a gap
            overlap_appendedRecordings = paste0(unlist(lapply(X = I_list[[1]], FUN = getValue, name = "interval")), collapse = " ")
          } else {
            # This recording is has not been appended
            NappendedRecordings = 0
            sf_appendedRecordings = getValue(I, "sf")
            names_appendedRecordings = getValue(I, "filename")
            overlap_appendedRecordings = ""
          }
          SUM$summary$NappendedRecordings = NappendedRecordings
          SUM$summary$sf_appendedRecordings  = sf_appendedRecordings
          SUM$summary$names_appendedRecordings = names_appendedRecordings
          SUM$summary$overlap_hrs_appendedRecordings = overlap_appendedRecordings
        }
        # convert daysummary and summary variables to numeric
        char2num_df = function(df) {
          # save as numeric columns that can be coerced to numeric
          df = lapply(df, function(x) tryCatch(as.numeric(as.character(x)), 
                                               error = function(cond) return(x),
                                               warning = function(cond) return(x)))
          # list to data frame
          df = as.data.frame(df, check.names = FALSE)
          return(df)
        }
        # Only apply this to column that are not ID because ID can be number
        # combined with letter E, which would then resolve to numeric
        noID = which(colnames(SUM$summary) != "ID")
        SUM$summary[noID] = char2num_df(SUM$summary[noID])
        noIDday = which(colnames(SUM$daysummary) != "ID")
        SUM$daysummary[noIDday] = char2num_df(SUM$daysummary[noIDday])
        GGIRversion = utils::packageVersion("GGIR")
        save(SUM, IMP, tail_expansion_log, GGIRversion, file = paste0(metadatadir, ms2.out, "/", RDname)) #IMP is needed for g.plot in g.report.part2
      }
      if (M$filecorrupt == FALSE & M$filetooshort == FALSE) rm(IMP)
      rm(M); rm(I)
    }
  } # end of main_part2
  
  #--------------------------------------------------------------------------------
  # Run the code either parallel or in serial (file index starting with f0 and ending with f1)
  cores = parallel::detectCores()
  Ncores = cores[1]
  if (params_general[["do.parallel"]] == TRUE) {
    if (Ncores > 3) {
      if (length(params_general[["maxNcores"]]) == 0) params_general[["maxNcores"]] = Ncores
      Ncores2use = min(c(Ncores - 1, params_general[["maxNcores"]], (f1 - f0) + 1))
      if (Ncores2use > 1) {
        cl <- parallel::makeCluster(Ncores2use) # not to overload your computer
        parallel::clusterExport(cl = cl,
                                varlist = c(unclass(lsf.str(envir = asNamespace("GGIR"), all = T)),
                                            "MONITOR", "FORMAT"),
                                envir = as.environment(asNamespace("GGIR")))
        parallel::clusterEvalQ(cl, Sys.setlocale("LC_TIME", "C"))
        doParallel::registerDoParallel(cl)
      } else {
        # Don't process in parallel if only one core
        params_general[["do.parallel"]] = FALSE
      }
    } else {
      if (verbose == TRUE) cat(paste0("\nparallel processing not possible because number of available cores (",Ncores,") < 4"))
      params_general[["do.parallel"]] = FALSE
    }
  }
  if (params_general[["do.parallel"]] == TRUE) {
    if (verbose == TRUE) cat(paste0('\n Busy processing ... see ', metadatadir, ms2.out, ' for progress\n'))
    # check whether we are indevelopment mode:
    GGIRinstalled = is.element('GGIR', installed.packages()[,1])
    packages2passon = functions2passon = NULL
    GGIRloaded = "GGIR" %in% .packages()
    if (GGIRloaded) { #pass on package
      packages2passon = 'GGIR'
      errhand = 'pass'
    } else {
      # pass on functions
      functions2passon = c("g.analyse", "g.impute", "g.weardec", "g.detecmidnight",
                           "g.extractheadervars", "g.analyse.avday", "g.getM5L5", "g.IVIS",
                           "g.analyse.perday", "g.getbout", "g.analyse.perfile", "g.intensitygradient",
                           "iso8601chartime2POSIX", "extract_params", "load_params", "check_params",
                           "correctOlderMilestoneData", "cosinor_IS_IV_Analyses", "extractID")
      errhand = 'stop'
    }
    i = 0 # declare i because foreach uses it, without declaring it
    `%myinfix%` = foreach::`%dopar%`
    output_list = foreach::foreach(i = f0:f1, .packages = packages2passon,
                                   .export = functions2passon, .errorhandling = errhand, .verbose = F) %myinfix% {
                                     tryCatchResult = tryCatch({
                                       main_part2(i, ffdone, fnames, metadatadir,
                                                  myfun, params_cleaning, params_247,
                                                  params_phyact, params_output, params_general,
                                                  path, ms2.out, foldername, fullfilenames,
                                                  folderstructure, referencefnames,
                                                  daySUMMARY, csvfolder, cnt78,
                                                  verbose, use_qwindow_as_diary)
                                       
                                     })
                                     return(tryCatchResult)
                                   }
    on.exit(parallel::stopCluster(cl))
    for (oli in 1:length(output_list)) { # logged error and warning messages
      if (is.null(unlist(output_list[oli])) == FALSE) {
        if (verbose == TRUE) cat(paste0("\nErrors and warnings for ",fnames[oli]))
        print(unlist(output_list[oli])) # print any error and warnings observed
      }
    }
  } else {
    errors = list()
    for (i in f0:f1) {
      if (verbose == TRUE) cat(paste0(i, " "))
      function_to_evaluate = expression(
        main_part2(i, ffdone, fnames, metadatadir,
                   myfun, params_cleaning, params_247,
                   params_phyact, params_output, params_general,
                   path, ms2.out, foldername, fullfilenames,
                   folderstructure, referencefnames,
                   daySUMMARY, csvfolder, 
                   cnt78, verbose, use_qwindow_as_diary)
      )
      if (params_general[["use_trycatch_serial"]] == TRUE) {
        tryCatch(
          eval(function_to_evaluate),
          error = function(e) {
            err_msg = conditionMessage(e)
            errors[[as.character(fnames[i])]] <<- err_msg
          }
        )
      } else {
        eval(function_to_evaluate)
      }
    }
    # show logged errors after the loop:
    if (params_general[["use_trycatch_serial"]] == TRUE && verbose == TRUE) {
      if (length(errors) > 0) {
        cat(paste0("\n\nErrors in part 2... for:"))
        cat(paste0("\n-", names(errors), ": ", unlist(errors), collapse = ""))
      }
    }
  }
}

g.readtemp_movisens = function(datafile, from = c(), to = c(), acc_sf, acc_length, interpolationType=1) {
    # Acceleration data and temperature were sampled at different rates,
    # so we need to resample temperature to get the same sampling rate
    # as what we have for acceleration.

    temp_sf = 1 # temperature is most likely sampled at 1Hz, but we'll double-check later

    temp_from = ceiling(from / acc_sf * temp_sf)
    temp_to = ceiling(to / acc_sf * temp_sf)

    temperature = unisensR::readUnisensSignalEntry(dirname(datafile), "temp.bin", 
                                                   startIndex = temp_from, endIndex = temp_to)
    new_temp_sf = attr(temperature, "sampleRate")

    # We had guessed that temperature was sampled at 1Hz. Let's check, and if we were wrong, then re-do.
    if (temp_sf != new_temp_sf) {
        temp_from = ceiling(from / acc_sf * new_temp_sf)
        temp_to = ceiling(to / acc_sf * new_temp_sf)

        temperature = unisensR::readUnisensSignalEntry(dirname(datafile), "temp.bin", 
                                                       startIndex = temp_from, endIndex = temp_to)
    }

    temperature = temperature$temp

    # we don't care about the exact timestamp values because we'll throw the timestamps away anyway.
    rawTime = seq_len(length(temperature))
    timeRes = seq(from = 1, to = rawTime[length(rawTime)], length.out = acc_length)

    temperature = GGIRread::resample(as.matrix(temperature), rawTime, timeRes, length(temperature), type=interpolationType)

    invisible(temperature)
}

ShellDoc2Vignette <- function(argument = "mode") {  
  if (!is.character(argument)) argument <- deparse(substitute(argument))
  .utils <- asNamespace("utils")
  .tools <- asNamespace("tools")
  helpfile = .utils$.getHelpFile(help("GGIR"))
  
  
  hs <- utils::capture.output(switch("Rd", Rd = .tools$prepare_Rd(helpfile)))
  
  # substitute curly braces to avoid problems later on
  hs = gsub(pattern = "\\{|\\}", replacement = "__", x = hs)
  
  
  # argument to look for with Rd format
  lookfor = paste0("\\item__",argument, "____")
  
  # basic arguments
  start_args = grep("__\\arguments__", hs, fixed = TRUE)
  end_args = grep("__\\details__", x = hs, fixed = TRUE)
  basic_args = hs[start_args:(end_args - 1)]
  
  # Arguments in parameters
  start_details = grep("__\\details__", hs, fixed = TRUE)
  end_details = grep("__\\examples__", x = hs, fixed = TRUE)
  parameters = hs[start_details:(end_details - 1)]
  
  tmp = grep(lookfor, basic_args); if (length(tmp) > 1) tmp = tmp[1]
  tmp2 = grep(lookfor, parameters); if (length(tmp2) > 1) tmp2 = tmp2[1]
  
  # Find the argument definition --------------------------------------------
  if (length(tmp) > 0) { # the argument is within the basic arguments of GGIR()
    definitions = grep("\\item", basic_args, fixed = TRUE)
    def0 = tmp +  1
    def1 = (definitions[which(definitions == tmp) + 1]) - 1
    
    def = paste(basic_args[def0:def1], collapse = " ")
  } else { # argument is in details (parameter objects)
    definitions = grep("\\item_", parameters, fixed = TRUE)
    def0 = tmp2
    def1 = (definitions[which(definitions == tmp2) + 1]) - 1
    
    if (def0 == max(definitions)) { # last parameter in GGIR.Rd
      def1_tmp = grep("__\\\\", parameters[(def0 + 1):length(parameters)])
      def1 = def0 + def1_tmp[1] - 1
    }
    
    if (is.na(def1)) {
      def1_tmp = grep("__", parameters)
      def1_tmp = which(diff(def1_tmp) > 1)
      def1 = def1_tmp[length(def1_tmp)]
    }
    
    def = paste(parameters[def0:def1], collapse = " ")
  }
  
  
  # Cleaning the definition text --------------------------------------------
  # clean argument name from definition
  def = gsub(pattern = lookfor, replacement = "", x = def, fixed = TRUE)
  
  # clean links in def
  def = gsub(pattern = "\\\\link", replacement = "", x = def)
  
  # clean \\code in def (used to mention arguments in Rd)
  def = gsub(pattern = "\\\\code", replacement = "", x = def)
  
  # clean multiple spaces in def
  def = gsub(pattern = "^ *|(?<= ) | *$", replacement = "", x = def, perl = TRUE)
  
  # clean double quotes in def
  # def = gsub(pattern = "\"", replacement = "'", x = def)
  
  # clean double underscore at the end
  def = gsub(pattern = "__", replacement = "", x = def)
  
  # clean next subsection if it is there
  if (grepl("\\\\subsection", def)) {
    stop = gregexpr("\\\\subsection", def)[[1]][1]
    def = substr(x = def, start = 1, stop = stop - 1)
  }
  
  
  # return ------------------------------------------------------------------
  return(def)
}

library(GGIR)
context("ShellDoc2Vignette and parametersVignette")
test_that("ShellDoc2Vignette retrieves the parameters definition correctly", {
  
  # ShellDoc2Vignette
  x = ShellDoc2Vignette(mode)
  expect_true(grepl("^Numeric", x))
  
  x = ShellDoc2Vignette("desiredtz")
  expect_true(grepl("^Character", x))
  
  x = ShellDoc2Vignette("visualreport_without_invalid")
  expect_true(grepl("^Boolean", x))
  
  x = ShellDoc2Vignette("possible_nap_dur")
  expect_true(grepl("^Numeric", x))
  
  # ParametersVignette
  x = parametersVignette("sleep")
  expect_equal(length(x), 1)
  expect_equal(class(x), "character")
  
  x = parametersVignette("metrics")
  expect_equal(length(x), 1)
  expect_equal(class(x), "character")
  
  x = parametersVignette("247")
  expect_equal(length(x), 1)
  expect_equal(class(x), "character")
  
  x = parametersVignette("rawdata")
  expect_equal(length(x), 1)
  expect_equal(class(x), "character")
  
  x = parametersVignette("phyact")
  expect_equal(length(x), 1)
  expect_equal(class(x), "character")
  
  x = parametersVignette("cleaning")
  expect_equal(length(x), 1)
  expect_equal(class(x), "character")
  
  x = parametersVignette("output")
  expect_equal(length(x), 1)
  expect_equal(class(x), "character")
  
  x = parametersVignette("general")
  expect_equal(length(x), 1)
  expect_equal(class(x), "character")
  
})

g.plot = function(IMP, M, I, durplot) {
  
    # Extracting filename and monitor type 
  fname = I$filename
  mon = I$monc
  monn = I$monn
  dformat = I$dformc
  # cutting out long"___" trains in some of the filenames (especially for Genea)
  fname2 = unlist(strsplit(fname,"_______"))
  if (length(fname2) > 1) fname = paste0(fname2[1], fname2[2])
  
  # Time variable and duration of recording
  ws3 = M$windowsizes[1]
  ws2 = M$windowsizes[2]
  n_ws2_perday = (1440*60) / ws2
  timeline = 1:nrow(M$metalong)
  durplot = durplot * n_ws2_perday
  
  # Moved function here, because it was nowhere used inside GGIR
  createcoordinates = function(r,timeline) {
    if (length(which(abs(diff(r)) == 1) > 0)) {	
      if (r[1] == 1) {
        x0 = c(timeline[1],timeline[which(diff(r) == 1) + 1])
        x1 = timeline[which(diff(r) == -1) + 1]
        if (r[length(timeline)] == 1) { #file ends with non-wear
          x1 = c(x1,timeline[length(timeline)])
        }
      } else {
        x0 = timeline[which(diff(r) == 1) + 1]
        x1 = timeline[which(diff(r) == -1) + 1]
        if (r[length(timeline)] == 1) { #file ends with non-wear
          x1 = c(x1,timeline[length(timeline)])
        }
      }
    } else {
      x0 = c()
      x1 = c()
    }
    invisible(list(x0 = x0, x1 = x1))
  }
  
  #create coordinates for rectangles non-wear
  cd1 = createcoordinates(IMP$rout[,1], timeline) #nonwear
  cd2 = createcoordinates(IMP$rout[,2], timeline) #clipping
  cd3 = createcoordinates(IMP$rout[,3], timeline) #additional nonwear
  cd4 = createcoordinates(IMP$rout[,4], timeline) #protocol
  s0 = cd1$x0; 	s1 = cd1$x1
  b0 = cd2$x0; 	b1 = cd2$x1
  g0 = cd3$x0; 	g1 = cd3$x1
  w0 = cd4$x0; 	w1 = cd4$x1  
  dens = 30
  
  # start plot with empty canvas
  plot.new()	
  par(fig = c(0, 1, 0, 1), new = T, mar = c(5, 4, 3, 0.5))
  plot(seq(0, durplot), seq(0, durplot), col = "white", type = "l", axes = F, 
       xlab = "", ylab = "", main = paste0("device brand: ", monn, " | filename: ", fname), cex.main = 0.6)#dummy plot
  # draw coloured rectangles
  Y0 = -50
  Y1 = durplot * 0.98 # leave space around legend
  legend_names = c("not worn", "signal clipping", "also not worn", "study protocol masked")
  legend_lty = c(NA, NA, NA, NA)
  legend_density = c(100, 100, 100, dens)
  x.intersp = rep(0.5, 4)
  legend_colors = c()
  legend_index = c()
  if (length(s0) > 0) { #non-wear
    CL = colors()[148]
    for (ri in 1:length(s0)) {
      rect(s0[ri], Y0, s1[ri], Y1, border = colors()[148], col = CL, lwd = 0.6)
    }
    legend_index = 1
    legend_colors = CL
  }
  if (length(b0) > 0) { #clip
    CL = colors()[464]
    for (ri in 1:length(b0)) {
      rect(b0[ri], Y0, b1[ri], Y1, border = colors()[464], col = CL, lwd = 0.6)
    }
    legend_index = c(legend_index, 2)
    legend_colors = c(legend_colors, CL)
  }
  if (length(g0) > 0) {
    CL = colors()[150]
    for (ri in 1:length(g0)) { #additional non-wear
      rect(g0[ri], Y0, g1[ri], Y1, border = colors()[150], col = CL, lwd = 0.6)
    }
    legend_index = c(legend_index, 3)
    legend_colors = c(legend_colors, CL)
  }
  if (length(w0) > 0) {
    CL = colors()[24]
    for (ri in 1:length(w0)) { #protocol
      rect(w0[ri], Y0, w1[ri], Y1, border = colors()[24], col = CL, density = dens, lwd = 0.6)
    }
    legend_index = c(legend_index, 4)
    legend_colors = c(legend_colors, CL)
  }
  
  # Add legend for rectengles
  legend_names = legend_names[legend_index]
  legend_lty = legend_lty[legend_index]
  legend_density = legend_density[legend_index]
  x.intersp = x.intersp[legend_index]
  
  
  
  abline(v = timeline[length(timeline)], col = "blue", lwd = 1)	
  
  # derive acceleration signal
  # Index of first acceleration metric in metashort other than timestamp, or angle-related
  IndOtMetric = which(colnames(M$metashort) %in% c("timestamp","anglex","angley","anglez") == FALSE)[1]
  metricName = colnames(M$metashort)[IndOtMetric]
  accel = as.numeric(as.matrix(M$metashort[,IndOtMetric]))
  

  accel2 = cumsum(c(0, accel))
  select = seq(1,length(accel2), by = ws2 / ws3)
  Acceleration = diff(accel2[round(select)]) / abs(diff(round(select[1:(length(select))])))
  if (length(timeline) > length(Acceleration)) {
    Acceleration = c(Acceleration, rep(0, abs(length(Acceleration) - length(timeline))))
  } else if (length(timeline) < length(Acceleration)) {
    Acceleration = Acceleration[1:length(timeline)]
  }
  
  MEND = length(timeline)
  mnights = grep("00:00:00", M$metalong$timestamp)
  noons = grep("12:00:00", M$metalong$timestamp)
  abline(v = noons, lwd = 0.5, col = "grey", lty = 2)
  abline(v = mnights, lwd = 0.5, lty = 3)
  if (length(legend_index) > 0) {
    legend("top", legend = legend_names, col = legend_colors, density = legend_density, #lty = legend_lty,  lty = legend_lty, 
           fill = legend_colors, border = legend_colors,
           x.intersp = x.intersp, ncol = 4, cex = 0.6, lwd = 0.6,
           bg = "white", box.col = "black")
  }
  extranoons = c()
  if (length(mnights) > 0 & length(noons) > 0) {
    # axis 1: midnight, noon labels (including one extra day at the beginning and end)
    extramnights = c(mnights[1] - n_ws2_perday, mnights, max(mnights) + n_ws2_perday)
    extranoons = c(noons[1] - n_ws2_perday, noons, max(noons) + n_ws2_perday)
    if (extranoons[1] < extramnights[1]) extranoons = extranoons[-1]
    if (max(extranoons) > max(extramnights)) extranoons = extranoons[-length(extranoons)]
    ticks_12hours = sort(c(extramnights, extranoons))
    x_labels_12hours = rep("", length(ticks_12hours)) # no tick labels
    # axis 2: day counting (including one extra day at the beginning and end)
    if (length(extranoons) > 1) {
      extramnights = c(mnights[1] - n_ws2_perday, mnights, max(mnights) + n_ws2_perday)
      ticks_dayborders = extramnights
      x_labels_days = ceiling((extranoons / n_ws2_perday) - 0.5) + 1
    }
  } else {
    ticks_12hours = seq(0, nrow(M$metalong) + n_ws2_perday, by = n_ws2_perday)
    x_labels_12hours = 1:length(ticks_12hours)
  }
  # creating plot functions to avoid duplicated code
  plot_acc = function(timeline, Acceleration, durplot, ticks_12hours, metricName, x_labels_12hours) {
    if (metricName %in% c("ZCX", "ZCY", "ZCX", "ExtAct") == TRUE | 
        length(grep(pattern = "count", x = metricName, ignore.case = TRUE)) > 0) {
      # Metric is not on a G scale
      ylabel = paste0(metricName, ifelse(metricName == "ExtAct", yes = "", no = " (counts)"))
      YLIM = c(0, max(Acceleration, na.rm = TRUE) * 1.05)
      YTICKS = NULL
      if (YLIM[2] < 0.8) {
        YLIM[2] = 0.8
        YTICKS = c(0, 0.2, 0.4, 0.6, 0.8)
      } else if (YLIM[2] >= 0.8 & YLIM[2] < 1.5) {
        YLIM[2] = 1.5
        YTICKS = c(0, 0.5, 1, 1.5)
      } else if (YLIM[2] >= 1.5 & YLIM[2] < 5) {
        YLIM[2] = 5
        YTICKS = 0:5
      } else if (YLIM[2] >= 5 & YLIM[2] < 10) {
        YLIM[2] = 10
        YTICKS = seq(0, 10, by = 2)
      } else if (YLIM[2] >= 10 & YLIM[2] < 250) {
        YLIM[2] = 250
        YTICKS = seq(0, 250, by = 50)
      } else if (YLIM[2] >= 250 & YLIM[2] < 500) {
        YLIM[2] = 500
        YTICKS = seq(0, 500, by = 100)
      } else if (YLIM[2] >= 500 & YLIM[2] < 1000) {
        YLIM[2] = 1000
        YTICKS = seq(0, 1000, by = 200)
      }
      if (is.null(YTICKS)) {
        YTICKS = round(c(0, YLIM[2] * 1/4, YLIM[2] * 2/4, YLIM[2] * 3/4, YLIM[2]))
      }
    } else {
      ylabel = expression(paste("Acceleration (m", italic("g"), ")"))
      YLIM = c(0, 0.6)
      YTICKS = c(0, 0.2, 0.4, 0.6)
    }
    plot(timeline, Acceleration, type = "l", bty = "l",
         lwd = 0.1, axes = FALSE, cex.lab = 0.8,
         xlab = "Recording day", ylab = ylabel,  xlim = c(0, durplot), 
         ylim = YLIM, )
    axis(side = 2, at = YTICKS, las = 1, cex.axis = 0.8)
    # axis 2 (day counting)
    if (length(extranoons) > 1) { # only if more than 1 day
      axis(side = 1, at = extranoons, labels = x_labels_days, 
           cex.axis = ifelse(test = length(extranoons) > 10, yes = 0.6, no = 0.8),
           font = 1, tick = FALSE, lwd = 0.8, mgp = c(3,1,0)) #line = -0.4
      axis(side = 1, at = ticks_dayborders, labels = NA, tck = -0.05, mgp = c(3,1,0))# line = 0.5, 
    }
  }
  plot_nonwear = function(timeline, M, durplot, ticks_12hours) {
    plot(timeline, M$metalong$nonwearscore, type = "s",
         xlab = "", ylab = "Non-wear score", axes = F,
         lwd = 0.1, xlim = c(0,durplot), ylim = c(0, 3), cex.lab = 0.8)
    axis(side = 2,at = c(0, 1, 2, 3), cex.axis = 0.8, las = 1)
  }
  # plot data
  if (mon == MONITOR$GENEACTIV || (mon == MONITOR$AXIVITY && dformat == FORMAT$CWA)) {
    # Recordings with temperature
    par(fig = c(0,1, 0, 0.65), new = T)
    plot_acc(timeline, Acceleration, durplot, ticks_12hours, metricName, x_labels_12hours)
    
    par(fig = c(0, 1, 0.45, 0.80), new = T)
    plot_nonwear(timeline, M, durplot, ticks_12hours)
    
    par(fig = c(0, 1, 0.60, 0.95), new = T)
    plot(timeline, M$metalong$temperaturemean[1:MEND], type = "l",
         xlab = "", ylab = "Temp. (C)", axes = F, lwd = 0.1,
         xlim = c(0, durplot), ylim = c(20,35), cex.lab = 0.8)
    abline(h = 20, col = "black", lwd = 1, lty = 2)
    abline(h = 35, col = "black", lwd = 1, lty = 2)
    axis(side = 2,at = c(20,35), cex.axis = 0.8, las = 1)
  } else {
    # Recordings without temperature
    par(fig = c(0, 1, 0, 0.80), new = T)
    plot_acc(timeline, Acceleration, durplot, ticks_12hours, metricName, x_labels_12hours)

    par(fig = c(0, 1, 0.60, 0.95), new = T)
    plot_nonwear(timeline, M, durplot, ticks_12hours)
  }
}

MONITOR = setNames(0:7, c("AD_HOC", "GENEA", "GENEACTIV", "ACTIGRAPH", "AXIVITY",  "MOVISENS", "VERISENSE", "PARMAY_MTX")) # GENEA is deprecated
MONITOR = as.environment(as.list(MONITOR))
lockEnvironment(MONITOR, bindings = TRUE)

FORMAT = setNames(1:6, c("BIN", "CSV", "WAV", "CWA", "AD_HOC_CSV", "GT3X")) # WAV is deprecated
FORMAT = as.environment(as.list(FORMAT))
lockEnvironment(FORMAT, bindings = TRUE)

correctOlderMilestoneData = function(x) {
  # Function to convert older format milestone data
  # for which temperature and light data may have been
  # stored in factor format
  
  # timestamp is assumed to be always present and in character format
  if (is.factor(x$timestamp[1])) {
    x$timestamp = format(x$timestamp)
  }
  # temperature and light may not always be present
  # if present they are expected to be in numeric format
  metrics = c("ENMO","LFENMO", "BFEN", "EN", "HFEN", "HFENplus", "MAD", "ENMOa",
              "ZCX", "ZCY", "ZCZ", "BrondCount_x", "BrondCount_y",
              "BrondCount_z", "NeishabouriCount_x", "NeishabouriCount_y", 
              "NeishabouriCount_z", "NeishabouriCount_vm", 
              "anglex", "angley", "anglez", "temperature", "light")
  pattern = paste(metrics, collapse = "|")
  columns2check = grep(pattern = pattern, x = names(x), value = FALSE)
  if (length(columns2check) > 0) {
    for (column in columns2check) {
      if (is.factor(x[1, column])) {
        x[, column] = as.numeric(as.character(x[, column]))
      }
    }
  }
  return(x)
}

library(GGIR)
context("correctOlderMilestoneData")
test_that("Function changes the class of metashort and metalong columns when needed", {
  skip_on_cran()
  
  # data prep
  time0 = as.POSIXct("2023-01-24 10:00:00 CET")
  time1 = as.POSIXct("2023-01-24 10:59:55 CET")
  timestamp = POSIXtime2iso8601(seq(time0, time1, by = 5), tz = "")
  metashort = data.frame(timestamp = as.factor(timestamp),
                         ENMO = factor(1:(60*60/5)),
                         anglez = factor(1:(60*60/5)),
                         extFun_dummy = factor(1:2))
  
  # function test
  metashort2 = correctOlderMilestoneData(metashort)
  
  # tests
  expect_equal(class(metashort$timestamp), "factor")
  expect_equal(class(metashort$ENMO), "factor")
  expect_equal(class(metashort$anglez), "factor")
  expect_equal(class(metashort$extFun_dummy), "factor")
  
  expect_equal(class(metashort2$timestamp), "character")
  expect_equal(class(metashort2$ENMO), "numeric")
  expect_equal(class(metashort2$anglez), "numeric")
  expect_equal(class(metashort2$extFun_dummy), "factor")
})

g.part3_correct_guider = function(SLE, desiredtz, epochSize,
                                  params_sleep) {
  
  guider_cor_maxgap_hrs = params_sleep[["guider_cor_maxgap_hrs"]]
  guider_cor_min_frac_sib = params_sleep[["guider_cor_min_frac_sib"]]
  guider_cor_min_hrs = params_sleep[["guider_cor_min_hrs"]]
  guider_cor_meme_frac_out = params_sleep[["guider_cor_meme_frac_out"]]
  guider_cor_meme_frac_in = params_sleep[["guider_cor_meme_frac_in"]]
  guider_cor_meme_min_hrs = params_sleep[["guider_cor_meme_min_hrs"]]
  # guider_cor_do = params_sleep[["guider_cor_do"]]
  guider_cor_meme_min_dys = params_sleep[["guider_cor_meme_min_dys"]]
  
  # local functions:
  get_matching_indices = function(SLE, reference_window, tz) {
    correct_for_DST = function(x, epochSize) {
      # When clock moves forward a timepoint may be missed
      missingday = which(diff(x) > (25 * 3600/epochSize))
      if (length(missingday) > 0) {
        missingday = missingday[1]
        x = c(x[1:missingday],
              x[missingday] + 24 * 3600/epochSize,
              x[(missingday + 1):length(x)])
      }
      return(x)
    }
    ref_min = ref_max = NULL
    for (ki in 1:2) {
      if (reference_window[ki] >= 24) {
        reference_window[ki] = reference_window[ki] - 24
      }
      HR = floor(reference_window[ki])
      MIN = floor((reference_window[ki] - HR) * 60)
      SEC = floor((reference_window[ki] - HR - (MIN / 60)) * 3600)
      SEC = floor(SEC / epochSize) * epochSize
      HR = ifelse(HR < 10, yes = paste0("0", HR), no = as.character(HR))
      MIN = ifelse(MIN < 10, yes = paste0("0", MIN), no = as.character(MIN))
      SEC = ifelse(SEC < 10, yes = paste0("0", SEC), no = as.character(SEC))
      ref_time = paste(HR, MIN, SEC, sep = ":")
      if (ki == 1) {
        ref_min = which(SLE$output$clocktime == ref_time)
      } else if (ki == 2) {
        ref_max = which(SLE$output$clocktime == ref_time)
      }
    }
    if (ref_max[1] < ref_min[1]) {
      # recording started after the start of reference causing a shift
      # correct for this
      ref_min = c(1, ref_min)
    }
    if (length(ref_min) > length(ref_max)) {
      # recording end before end of reference causing a shift
      # correct for this
      ref_max = c(ref_max, length(SLE$output$clocktime))
    }
    ref_min = correct_for_DST(ref_min, epochSize)    
    ref_max = correct_for_DST(ref_max, epochSize)
    
    invisible(list(ref_min = ref_min, ref_max = ref_max))
  }
  
  get_crude_estimate = function(SLE, tSegment) {
    # get crude estimate values between min-max pair
    crude_est = SLE$output$spt_crude_estimate[tSegment]
    sib = SLE$output[tSegment, grep(pattern = "time|invalid|night|estimate|medmed", x = colnames(SLE$output), invert = TRUE)]
    temp_time = SLE$output$time_POSIX[tSegment]
    # At this point rle_rest$values can have the following values:
    # 2: guider based classification of main sleep window
    # 1: other resting windows that were discarded
    # 0: remaining time
    if (1 %in% crude_est) { 
      # omit 1- segments that have less than 80% sib
      bin_est = as.integer(crude_est == 1) # maps 2 to 0 to avoid conflicts in cases in which 1 is preceeding 2 or vice versa
      class_changes = diff(c(0, bin_est, 0)) 
      segment_start = which(class_changes == 1)
      segment_end = which(class_changes == -1) - 1
      if (length(segment_end) == 0) {
        crude_est[which(crude_est == 1)] = 0
      } else {
        for (gi in seq_along(segment_start)) {
          if (mean(sib[segment_start[gi]:segment_end[gi]]) < 0.8) {
            crude_est[segment_start[gi]:segment_end[gi]] = 0
          }
        }
      }
    }
    return(crude_est)
  }
  
  clean_SLE = function(SLE) {
    # remove temp columns on exit
    temp_columns = c("clocktime", "time_POSIX", "spt_crude_estimate", "medmed")
    SLE$output = SLE$output[, which(names(SLE$output) %in% temp_columns == FALSE)]
    return(SLE)
  }
  
  convert_ts_to_hours = function(SLE, crude_est, tSegment, ref_value = 1) {
    # convert timeseries to zeros to clock hour in the day
    # redefine window by range
    new_window = range(which(crude_est == ref_value ))
    new_SPTE = SLE$output$clocktime[tSegment[new_window]]
    convert_time = function(x) {
      spit_time = as.numeric(unlist(strsplit(x, ":")))
      time_hours = round(sum((spit_time * c(3600, 60, 1)) / 3600), digits = 3)
      if (time_hours <= 12) time_hours = time_hours + 24
      return(time_hours)
    }
    new_SPTE = unlist(lapply(X = new_SPTE, FUN = convert_time))
    if (new_SPTE[1] > new_SPTE[2] && new_SPTE[2] > 12 && new_SPTE[1] < 24) {
      new_SPTE[2] = new_SPTE[2] + 24
    } else if (new_SPTE[1] > new_SPTE[2] & new_SPTE[1] > 24) {
      new_SPTE[1] = new_SPTE[1] - 24
    }
    return(new_SPTE)
  }
  
  #-----------------------------------------
  # main code
  # HASPT includes a guider estimate for every night regardless of how much
  # invalid data there is, but here we should only the valid nights
  invalid_per_night = aggregate(x = SLE$output$invalid, by = list(SLE$output$night), FUN = sum)
  names(invalid_per_night) = c("night", "count")
  valid_nights = invalid_per_night$night[which(invalid_per_night$night > 0 & invalid_per_night$count < 24 * 3600 / epochSize * 0.333)]
  if (length(valid_nights) > 0) {
    valid_nights = valid_nights[which(is.na(SLE$SPTE_start[valid_nights]) == FALSE &
                                        is.na(SLE$SPTE_end[valid_nights]) == FALSE)]
  }
  SLE$SPTE_corrected = rep(0, length(SLE$SPTE_start))
  # require at least 2 nights as otherwise min-max becomes meaningless
  if (length(valid_nights) < 1) {
    SLE = clean_SLE(SLE)
    return(SLE)
  }
  
  # Step 1) Identify reference guider across all nights
  # from the earliest (min) start till the latest (max) end.
  reference_window = c(min(SLE$SPTE_start[valid_nights], na.rm = TRUE),
                       max(SLE$SPTE_end[valid_nights], na.rm = TRUE))
  SLE$output$time_POSIX = iso8601chartime2POSIX(SLE$output$time, tz = desiredtz)
  SLE$output$clocktime = format(SLE$output$time_POSIX, format = "%H:%M:%S")
  
  # Get indices of full time series where the reference guider matches the timestamps
  ref_indices = get_matching_indices(SLE, reference_window,  tz = desiredtz)
  ref_min = ref_indices$ref_min
  ref_max = ref_indices$ref_max
  if (length(ref_min) < 2 || length(ref_max) < 2) {
    SLE = clean_SLE(SLE)
    return(SLE)
  }
  if (length(valid_nights) >= guider_cor_meme_min_dys) {
    # Step 2) Identify median-median window and corresponding indices in the time series
    # this window will be used to assess whether HDCZA window can be replaced
    medmed_reference_window = c(median(SLE$SPTE_start[valid_nights], na.rm = TRUE),
                                median(SLE$SPTE_end[valid_nights], na.rm = TRUE))
    ref_indices = get_matching_indices(SLE, medmed_reference_window,  tz = desiredtz)
    ref_med1 = ref_indices$ref_min
    ref_med2 = ref_indices$ref_max
    if (length(ref_med1) < 2 || length(ref_med2) < 2) {
      SLE = clean_SLE(SLE)
      return(SLE)
    }
    # Deal with non-matching index vectors caused by incomplete first or last night
    Ntimepoints = length(SLE$output$clocktime)
    newVectors = g.part3_alignIndexVectors(x = ref_min, y = ref_max,
                                           a = ref_med1, b = ref_med2,
                                           N = Ntimepoints) 
    ref_min = newVectors$x
    ref_max = newVectors$y
    ref_med1 = newVectors$a
    ref_med2 = newVectors$b
    
    if (length(ref_min) != length(ref_med1) ||
        length(ref_max) != length(ref_med2)) {
      stop("index vectors do not match")
    }
    # Step 3): Select an alternative HDCZA window if the default falls outside med-med
    # and there is a secondary HDCZA in med-med.
    
    # - Check each night for main HDCZA outside med-med window and correct if needed
    for (ji in 1:length(ref_min)) {
      if (ji %in% valid_nights) {
        # get crude estimate values between min-max pair
        tSegment = ref_min[ji]:ref_max[ji]
        crude_est = get_crude_estimate(SLE, tSegment)
        sib = SLE$output[tSegment, grep(pattern = "time|invalid|night|estimate|medmed", x = colnames(SLE$output), invert = TRUE)]
        # derive time series across min-max with medmed indicated
        tSegment_med = ref_med1[ji]:ref_med2[ji]
        SLE$output$medmed = 0
        SLE$output$medmed[tSegment_med] = 1
        medmed = SLE$output$medmed[tSegment]
        # Is original HDCZA estimate largely outside median-median window?
        if (length(tSegment_med) > 3600 / epochSize &&
            length(which(crude_est == 2)) != 0  &&
            length(which(crude_est == 2 & medmed == 0)) / 
            length(which(crude_est == 2)) > guider_cor_meme_frac_out) {
          # To assess other criteria first create summary per segment
          # and give each non-zero segment a unique segment id
          temp_rle = rle(crude_est)
          nonzero = which(temp_rle$values != 0)
          temp_rle$values[nonzero] = 1:length(nonzero)
          seg_id = rep(temp_rle$values, temp_rle$lengths)
          
          df = data.frame(crude_est = crude_est, medmed = medmed, 
                          index = 1:length(crude_est), sib = sib,
                          seg_id = seg_id)
          # Crude estimate per segment (code by 0, 1 or 2 as explained above)
          segment_level = aggregate(df[, c("crude_est", "seg_id")], by = list(df$seg_id), FUN = mean)[, 1:2]
          names(segment_level) = c("seg_id", "crude_est")
          # sib
          segment_sib = aggregate(df[, c("sib", "seg_id")], by = list(df$seg_id), FUN = mean)[, 1:2]
          names(segment_sib) = c("seg_id", "sib")
          segment_summary = merge(segment_level, segment_sib, by = "seg_id")
          
          # Overlaps with median-median window
          perc_value_one = function(x) {
            return(length(which(x == 1)) / length(x))
          }
          segment_overlap_medmed = aggregate(df[, c("medmed", "seg_id")], by = list(df$seg_id), FUN = perc_value_one)[, 1:2]
          names(segment_overlap_medmed) = c("seg_id", "medmed")
          segment_summary = merge(segment_summary, segment_overlap_medmed, by = "seg_id")
          
          # duration
          segment_size_hours = aggregate(df$index, by = list(df$seg_id), FUN = length)[, 1:2]
          segment_size_hours[, 2] = segment_size_hours[, 2] / (3600 / epochSize)
          names(segment_size_hours) = c("seg_id", "segment_size_hours")
          segment_summary = merge(segment_summary, segment_size_hours, by = "seg_id")
          
          # Start
          segment_start = aggregate(df[, c("index", "seg_id")], by = list(df$seg_id), FUN = min)[, 1:2]
          names(segment_start) = c("seg_id", "start_index")
          segment_summary = merge(segment_summary, segment_start, by = "seg_id")
          
          # End
          segment_end = aggregate(df[, c("index", "seg_id")], by = list(df$seg_id), FUN = max)[, 1:2]
          names(segment_end) = c("seg_id", "end_index")
          segment_summary = merge(segment_summary, segment_end, by = "seg_id")
          
          rm(segment_sib, segment_overlap_medmed, segment_start, segment_size_hours, segment_end)
          
          SLE$output = SLE$output[, -which(colnames(SLE$output) == "medmed")]
          # Identify whether there is a secondary HDCZA window that meets
          # following criteria:
          # - (Mostly) Overlaps with the median-median window
          # - lasts at least 3 hours
          # - has at least 80% sustained inactivity (already removed above)
          new_main_HDCZA = which(segment_summary$crude_est == 1 &
                                   segment_summary$medmed > guider_cor_meme_frac_in &
                                   segment_summary$segment_size_hours > guider_cor_meme_min_hrs &
                                   segment_summary$sib > guider_cor_min_frac_sib)

          if (length(new_main_HDCZA) > 0) {
            # If yes, then select this other HDCZA window.
            # Update such that old window has value 1
            old_2 = which(segment_summary$crude_est == 2)
            segment_summary$crude_est[old_2] = 1
            crude_est[segment_summary$start_index[old_2]:segment_summary$end_index[old_2]] = 1
            # Update such that new window has 2
            segment_summary$crude_est[new_main_HDCZA] = 2
            crude_est[segment_summary$start_index[new_main_HDCZA]:segment_summary$end_index[new_main_HDCZA]] = 2
            
            # Update corresponding SLE$SPTE_start and SLE$SPTE_end values.
            new_SPTE = convert_ts_to_hours(SLE, crude_est, tSegment, ref_value = 2)
            SLE$output$spt_crude_estimate[tSegment] = crude_est
            SLE$SPTE_start[ji] = new_SPTE[1]
            SLE$SPTE_end[ji] = new_SPTE[2]
            SLE$SPTE_corrected[ji] = 2
          }
        }
      }
    }
    if (any(SLE$SPTE_corrected != 0)) {
      # If any night was corrected, repeat the initial steps of min-max and index extraction:
      # Identify reference guider across all nights
      reference_window = c(min(SLE$SPTE_start[valid_nights], na.rm = TRUE),
                           max(SLE$SPTE_end[valid_nights], na.rm = TRUE))
      
      # Get indices of full time series where the reference guider matches the timestamps
      ref_indices = get_matching_indices(SLE, reference_window,  tz = desiredtz)
      ref_min = ref_indices$ref_min
      ref_max = ref_indices$ref_max
      if (length(ref_min) < 2 || length(ref_max) < 2) {
        SLE = clean_SLE(SLE)
        return(SLE)
      }
    }
  }
  # Step 4): Loop again over nights but this time expand HDCZA with neighboring HDCZA
  # if criteria are met.
  
  
  # # Check each night and correct if needed
  for (ji in 1:length(ref_min)) {
    if (ji %in% valid_nights) {
      # get crude estimate values between min-max pair
      tSegment = ref_min[ji]:ref_max[ji]
      crude_est = get_crude_estimate(SLE, tSegment)
      # hours_in_class_1 = (length(which(crude_est == 1)) * epochSize) / 3600
      # if (hours_in_class_1 < 2) {
      #   # only consider nights with 2 or more hours outside main sleep window
      #   # we are only interested in fixing major night distortions
      #   next 
      # }
      
      # only consider window if there is rest outside guider
      # as indicated by a 1, because 2 is the current guider and zero is not resting
      if (1 %in% crude_est & 2 %in% crude_est) { 
        rle_rest = rle(crude_est)
        #--------------------------------------------
        # Identify long resting blocks
        long_rest = which(rle_rest$values == 1 & rle_rest$lengths * epochSize >= guider_cor_min_hrs * 3600)
        #--------------------------------------------
        # Remove any long rest that are separated from main sleep
        # by a too long wake period
        if (length(long_rest) > 0) {
          if (!is.null(guider_cor_maxgap_hrs) && 
              !is.infinite(guider_cor_maxgap_hrs)) {
            long_wake = which(rle_rest$values == 0 & rle_rest$lengths * epochSize >= guider_cor_maxgap_hrs * 3600)
            if (length(long_wake) > 0) {
              rle_rest$values[long_wake] = -1
              ind2remove = NULL
              original = which(rle_rest$values == 2) # only one segment is expected to be 2
              for (lri in 1:length(long_rest)) {
                # for each long rest (1)
                # check whether any of the long wake (-1) separates
                # it from the original HDCZA (2)
                this_long_rest = long_rest[lri]
                too_long_wake = which(rle_rest$values == -1)
                if (any(too_long_wake > original & too_long_wake < this_long_rest) |
                    any(too_long_wake < original & too_long_wake > this_long_rest)) {
                  ind2remove = c(ind2remove, lri)
                }
              }
              if (!is.null(ind2remove)) {
                long_rest = long_rest[-ind2remove]
              }
            }
          }
        }
        #--------------------------------------------
        # Any remaining long rests are considered
        if (length(long_rest) > 0) {
          rle_rest$values[long_rest] = 2
          rle_rest$values[which(rle_rest$values == 1)] = 0
          rle_rest$values[which(rle_rest$values == 2)] = 1
          N = length(crude_est)
          crude_est = rep(rle_rest$values, rle_rest$lengths)[1:N]
          
          new_SPTE = convert_ts_to_hours(SLE, crude_est, tSegment, ref_value = 1)
          SLE$SPTE_start[ji] = new_SPTE[1]
          SLE$SPTE_end[ji] = new_SPTE[2]
          if (SLE$SPTE_corrected[ji] == 0) {
            SLE$SPTE_corrected[ji] = 1
          } else {
            SLE$SPTE_corrected[ji] = 3
          }
        }
      }
    }
  }
  SLE = clean_SLE(SLE)
  return(SLE)
}

convertEpochData = function(datadir = c(), metadatadir = c(),
                            params_general = c(), f0 = c(), f1 = c(), verbose = TRUE) {
  
  # Function to convert externally derived epoch data
  # to a format that allows GGIR to process them as if they were part 1 output.
  tz = params_general[["desiredtz"]]
  epSizeShort = params_general[["windowsizes"]][1]
  epSizeLong = params_general[["windowsizes"]][2]
  epSizeNonWear = params_general[["windowsizes"]][3]
  
  # Identify input data files
  if (dir.exists(datadir) == FALSE) {
    stop("\nWhen working with external data, argument datadir is expected to be a directory")
  }
  
  if (params_general[["dataFormat"]] == "phb_xlsx") {
    # Philips Health Band data comes with two files per recording that need to be matched
    # Identify all xlsx files
    xlsx_files = dir(datadir, recursive = TRUE, full.names = TRUE, pattern = "[.]xlsx")
    xlsx_files = xlsx_files[grep(pattern = "sleep|datalist", x = xlsx_files, ignore.case = TRUE)]
    # Identify the pairs by looking for matching ID
    fileOverview = data.frame(filename = xlsx_files)
    extractID = function(x) {
      x = basename(x)
      # remove _ in sleep_wake to ease finding the ID
      x = gsub(pattern = "sleep_wake", replacement = "sleepwake", x = tolower(x))
      ID = unlist(strsplit(x, "_"))[2]
      return(ID)
    }
    fileOverview$ID = unlist(lapply(fileOverview$filename, FUN = extractID))
    # Put matching file pairs in a vector of lists
    uids = unique(fileOverview$ID)
    fnames = rep(NA, length(uids))
    for (uid in 1:length(uids)) {
      # Only keep recording where we have both files
      matchingFiles = which(fileOverview$ID == uids[uid])
      if (length(matchingFiles) == 2) {
        fnames[uid] = list(fileOverview$filename[matchingFiles])
      } else if (length(matchingFiles) == 1) {
        warning(paste0("No matching file found for ", fileOverview$filename[matchingFiles]), call. = FALSE)
      } else if (length(matchingFiles) > 2) {
        warning(paste0("There are more than 2 files for ID ", uids[uid], "."), call. = FALSE)
      }
    }
  } else if (params_general[["dataFormat"]] == "fitbit_json") {
    # Fitbit comes with one folder per participant that has multiple files
    # Identify folders
    participantFolders = dir(datadir, include.dirs = TRUE, full.names = TRUE)
    participantFolders = participantFolders[dir.exists(participantFolders)]
    IDs = fnames = rep(NA, length(participantFolders))
    for (uid in 1:length(participantFolders)) {
      IDs[uid] = basename(participantFolders[uid]) # Use folder name as ID because ID is not inside the files
      # Identify files in each folder
      jsonFiles = dir(participantFolders[uid], pattern = "json", recursive = TRUE, full.names = TRUE)
      jsonFiles = grep(pattern = "sleep|steps|calories|heart_rate", x = jsonFiles, value = TRUE)
      jsonFiles = grep(pattern = "resting_heart_rate|time_in_heart", invert = TRUE, x = jsonFiles, value = TRUE)
      if (length(jsonFiles) > 0) {
        fnames[uid] = list(jsonFiles = jsonFiles)
      }
    }
    IDs = IDs[!is.na(fnames)]
    fnames = fnames[!is.na(fnames)]
  } else {
    # Data with just one recording per file.
    # Here we do a check that no unexpected file types are found.
    fnames_csv = dir(datadir, full.names = TRUE,pattern = "[.]csv")
    fnames_awd = dir(datadir, full.names = TRUE,pattern = "[.]awd|[.]AWD")
    fnames_xls = dir(datadir, full.names = TRUE,pattern = "[.]xls")
    if (length(which(c(length(fnames_csv), length(fnames_awd), length(fnames_xls)) != 0)) > 1) {
      stop("Do not mix csv and awd files in the same data directory")
    } else {
      if (length(fnames_awd) > 0) {
        if (params_general[["dataFormat"]] != "actiwatch_awd") {
          stop("Specified dataFormat does not match the data")
        }
        fnames = fnames_awd
      } else if (length(fnames_csv) > 0) {
        if (params_general[["dataFormat"]] == "actiwatch_awd") {
          stop("Specified dataFormat does not match the data")
        }
        fnames = fnames_csv
      } else if (length(fnames_xls) > 0) {
        if (params_general[["dataFormat"]] != "sensewear_xls") {
          stop("Specified dataFormat does not match the data")
        }
        fnames = fnames_xls
      }
    }
  }
  #-------------
  # Create output folder, if we had raw data function g.part1 would normally do this:
  if (!file.exists(metadatadir)) {
    dir.create(metadatadir)
    dir.create(file.path(metadatadir, "meta"))
    dir.create(file.path(metadatadir, "meta", "basic"))
    dir.create(file.path(metadatadir, "results"))
    dir.create(file.path(metadatadir, "results", "QC"))
  }
  
  #============
  # Based on knowledge about data format
  # we can already assign monitor names and codes
  # data format names and codes
  # and sample rate
  actiwatchData = length(grep(pattern = "actiwatch", x = params_general[["dataFormat"]], ignore.case = TRUE)) > 0
  if (params_general[["dataFormat"]] == "ukbiobank_csv") {
    deviceName = "Axivity"
    monn = "axivity"
    monc = 4
    dformc = 4
    epSizeShort = 5
  } else if (params_general[["dataFormat"]] == "actigraph_csv") {
    deviceName = "ActiGraph"
    monn = "actigraph"
    monc = 3
    dformc = 95
    epSizeShort = 0 # extract from file header
  } else if (actiwatchData == TRUE) {
    deviceName = "Actiwatch"
    monn = "actiwatch"
    monc = 99
    dformc = 99
  } else if (params_general[["dataFormat"]] == "sensewear_xls") {
    deviceName = "Sensewear"
    monn = "sensewear"
    monc = 98
    dformc = 98
  } else if (params_general[["dataFormat"]] == "phb_xlsx") {
    deviceName = "PhilipsHealthBand"
    monn = "philipshealthband"
    monc = 97
    dformc = 97
  } else if (params_general[["dataFormat"]] == "fitbit_json") {
    deviceName = "Fitbit"
    monn = "Fitbit"
    monc = 96
    dformc = 96
  } else if (params_general[["dataFormat"]] == "actical_csv") {
    deviceName = "Actical"
    monn = "actical"
    monc = 94
    dformc = 94
    epSizeShort = 0 # extract from file header
  }
  sf = NA
  dformn = params_general[["dataFormat"]]
  
  # Create template output objects to be stored in the milestone data (RData)
  C = list(cal.error.end = 0, cal.error.start = 0)
  C$scale = c(1, 1, 1)
  C$offset = c(0, 0, 0)
  C$tempoffset =  c(0, 0, 0)
  C$QCmessage = "Autocalibration not done"
  C$npoints = 0
  C$nhoursused = 0
  C$use.temp = TRUE
  M = list(filecorrupt = FALSE, filetooshort = FALSE,
           metalong = data.frame(
             timestamp = rep(0, 3),
             nonwearscore = rep(0, 3),
             clippingscore = rep(0, 3),
             lightmean = rep(0, 3),
             lightpeak = rep(0, 3),
             temperaturemean = rep(0, 3),
             EN = rep(0, 3)
           ),
           metashort = data.frame(timestamp = rep(0, 3),
                                  ExtAct = rep(0, 3)), 
           wday = 0,
           wdayname = "Sunday",
           windowsizes = params_general[["windowsizes"]],
           bsc_qc = data.frame(A = 1:3,
                               B = 1:3))
  dummyheader = data.frame(uniqueSerialCode = 0, 
                           frequency = NA, # NA because we work with processed data
                           start = "", # if relevant, this will be filled later on in code
                           device = deviceName,
                           firmwareVersion = "unknown",
                           block = 0)
  dummyheader = t(dummyheader)
  colnames(dummyheader) = "value"
  dummyheader = as.data.frame(dummyheader)
  I = list(
    header = dummyheader,
    monc = monc,
    monn = monn,
    dformc = dformc,
    dformn = dformn,
    sf = sf,
    filename = "unknown",
    deviceSerialNumber = "unknown")
  
  if (length(f0) == 0) f0 = 1
  if (length(f1) == 0) f1 = length(fnames)
  
  
  I_bu = I # backup version of I (for now only used in actigraph_csv)
  main_convert = function(i, fnames, metadatadir, params_general, I_bu,
                          epSizeShort, epSizeLong, tz, verbose, M, C) {
    myfun = NULL
    if (verbose == TRUE) {
      if (i  == 1) {
        cat(paste0("\nP1 file: ", i))
      } else {
        cat(paste0(" ", i))
      }
    }
    
    # Declare local funcions:
    matAggregate = function(mat, step) {
      # Aggregate matrix mat by taking over step number of rows
      # as sum unless column names is sleep or nonwear in that case
      # we take the rounded mean.
      mat = rbind(rep(0, ncol(mat)), mat)
      cumsum2 = function(x) {
        x = cumsum(ifelse(is.na(x), 0, x)) + x*0
        return(x)
      }
      mat = apply(mat, 2, cumsum2)
      mat = mat[seq(1, nrow(mat), by = step), , drop = FALSE]
      mat = apply(mat, 2, diff)
      # Correct non incremental variables
      for (niv in c("sleep", "ExtSleep", "light", "nonwear", "marker", "light")) {
        if (niv %in% colnames(mat)) mat[, niv] = round(mat[, niv] / step)
      }
      # Incremental variables are counts, calories, ExtAct as so far
      # none of the data formats provide movement expressed as average acceleration
      # per epoch
      return(mat)
    }
    
    checkEpochMatch = function(desiredEpochSize, epSizeShort) {
      # Check whether desired and derived epoch size match
      if (!is.null(desiredEpochSize) && epSizeShort != desiredEpochSize) {
        stop(paste0("\nThe short epoch size as specified by the user (",
                    desiredEpochSize, " seconds) does NOT match the short",
                    " epoch size we see in the data (", epSizeShort,
                    " seconds). Please correct."), call. = FALSE)
      }
      return()
    }
    #-----------------------------------------------------------
    # main code
    
    # filename
    if (!all(is.na(unlist(fnames[i])))) {
      if (params_general[["dataFormat"]] %in% c("fitbit_json", "phb_xlsx")) {
        fname = basename(dirname(unlist(fnames[i])[1]))
      } else {
        fname = basename(unlist(fnames[i])[1])
      }
      
      outputFileName = paste0(metadatadir, "/meta/basic/meta_", fname, ".RData")
      skip = TRUE
      if (params_general[["overwrite"]] == TRUE) {
        skip = FALSE
      } else {
        if (!file.exists(outputFileName)) {
          skip = FALSE
        }
      }
    } else {
      skip = TRUE
    }
    if (skip == FALSE) {
      I = I_bu
      D_extraVars = QClog = NULL
      I$filename = filename_dir = fname
      if (params_general[["dataFormat"]] == "ukbiobank_csv") {
        # read data
        D = data.table::fread(input = fnames[i],
                              header = TRUE,
                              data.table = FALSE,
                              sep = ",")
        header = as.character(data.table::fread(input = fnames[i], header = FALSE,
                                                nrows = 1, data.table = FALSE, sep = ",")[1, 1])
        # extract date/timestamp from fileheader
        timestamp = unlist(strsplit(header," - "))[2]
        timestamp_POSIX = as.POSIXlt(timestamp, tz = tz)
        D = list(data = D, epochSize = epSizeShort, startTime = timestamp_POSIX)
      } else if (params_general[["dataFormat"]] == "sensewear_xls") {
        # read data
        D = as.data.frame(readxl::read_excel(path = fnames[i], col_types = "text"))
        # Convert timestamps which is read as Excel timestamp
        timestamp_POSIX = format(as.POSIXct(as.numeric(D[, grep(pattern = "Time", x = colnames(D))]) * (60*60*24),
                                            origin = "1899-12-30", tz = "GMT"))
        timestamp_POSIX = as.POSIXct(timestamp_POSIX, tz = tz)
        
        D = D[, c("METs", "Step Counter", "Sleep")]
        colnames(D) = c("ExtAct", "ExtStep", "ExtSleep")
        D$ExtAct = as.numeric(D$ExtAct)
        D$ExtStep = as.numeric(D$ExtStep)
        D$ExtSleep = as.numeric(D$ExtSleep)
        epochSize = difftime(timestamp_POSIX[2], timestamp_POSIX[1], 
                             units = "secs")
        epSizeShort = as.numeric(epochSize)
        timestamp_POSIX = timestamp_POSIX[1]
        D = list(data = D, epochSize = epSizeShort, startTime = timestamp_POSIX)
      } else if (params_general[["dataFormat"]] == "actigraph_csv") {
        D = GGIRread::readActiGraphCount(filename = fnames[i], 
                                         timeformat = params_general[["extEpochData_timeformat"]],
                                         desiredtz = params_general[["desiredtz"]],
                                         configtz = params_general[["configtz"]],
                                         timeformatName = "extEpochData_timeformat")
        
        # Rename to align with GGIR metric naming
        cnd = colnames(D$data)
        colnames(D$data)[which(cnd == "x")] = "NeishabouriCount_x"
        colnames(D$data)[which(cnd == "y")] = "NeishabouriCount_y"
        colnames(D$data)[which(cnd == "z")] = "NeishabouriCount_z"
        colnames(D$data)[which(cnd == "vm")] = "NeishabouriCount_vm"
      } else if (params_general[["dataFormat"]] == "actical_csv") {
        D = GGIRread::readActicalCount(filename = fnames[i],
                                       timeformat = params_general[["extEpochData_timeformat"]],
                                       desiredtz = params_general[["desiredtz"]],
                                       configtz = params_general[["configtz"]],
                                       timeformatName = "extEpochData_timeformat")
        # Rename to align with GGIR metric naming
        colnames(D$data)[which(colnames(D$data) == "counts")] = "ExtAct" 
      } else if (length(grep(pattern = "actiwatch", x = params_general[["dataFormat"]], ignore.case = TRUE)) > 0) {
        D = GGIRread::readActiwatchCount(filename = fnames[i], 
                                         timeformat = params_general[["extEpochData_timeformat"]],
                                         desiredtz = params_general[["desiredtz"]],
                                         configtz = params_general[["configtz"]],
                                         timeformatName = "extEpochData_timeformat")
        # Rename to align with GGIR metric naming
        colnames(D$data)[which(colnames(D$data) == "counts")] = "ExtAct"
        extraVars  = grep(pattern = "light|nonwear", x = colnames(D$data))
        if (length(extraVars) > 0) {
          # split the extraVars
          D_extraVars = D$data[, extraVars, drop = FALSE]
          D$data = D$data[, -extraVars, drop = FALSE]
        }
        if (D$epochSize == 120) {
          # Oddly data can be stored in 120 second epochs even though original publications
          # never proposed this. The rest of GGIR cannot handle epoch size of 120 seconds
          # and most practical solution seems to duplicate data to simulate 60 second epoch
          D$data = D$data[rep(1:nrow(D$data), each = 2),]
          D$data[, "ExtAct"] = D$data[, "ExtAct"] / 2
          D$epochSize = 60
        }
      } else if (params_general[["dataFormat"]] == "phb_xlsx") {
        # phb = Philips Health Band
        D = GGIRread::mergePHBdata(filenames = fnames[[i]], 
                                   timeformat = params_general[["extEpochData_timeformat"]],
                                   desiredtz = params_general[["desiredtz"]],
                                   configtz = params_general[["configtz"]],
                                   timeformatName = "extEpochData_timeformat")
        if (nrow(D$data) < 2) return()
        epochSize = difftime(D$data$timestamp[2], D$data$timestamp[1], 
                             units = "secs")
        epSizeShort = as.numeric(epochSize)
        timestamp_POSIX = D$data$timestamp[1]
        
        D$epochSize = epSizeShort
        D$startTime = timestamp_POSIX
        
        extraVars  = grep(pattern = "nonwear", x = colnames(D$data))
        if (length(extraVars) > 0) {
          # split the extraVars
          D_extraVars = D$data[, extraVars, drop = FALSE]
          D$data = D$data[, -extraVars, drop = FALSE]
        }
        colnames(D$data)[which(colnames(D$data) == "counts")] = "ExtAct"
        D$data = D$data[, grep(pattern = "cardio|heart|sleepevent|battery|duration|missing|activem|vo2|energy|respiration|timestamp",
                               x = colnames(D$data), ignore.case = TRUE, invert = TRUE)]
      } else if (params_general[["dataFormat"]] == "fitbit_json") {
        data = GGIRread::mergeFitbitData(filenames = fnames[[i]],
                                         desiredtz = params_general[["desiredtz"]],
                                         configtz = params_general[["configtz"]])
        ID = IDs[i]
        I$filename = filename_dir = fname = ID
        epochSizes = names(table(diff(data$dateTime)))
        if (length(epochSizes) > 1) {
          stop(paste0("multiple epoch sizes encountered in Fitbit data ",
                      basename(fnames[[i]])), call. = FALSE)
        } else if (length(epochSizes) == 1) {
          epSizeShort = as.numeric(epochSizes)
        } else {
          stop(paste0("No epoch size recognised in Fitbit data ",
                      basename(fnames[[i]])), call. = FALSE)
        }
        timestamp_POSIX = data$dateTime[1]
        D = list(data = data, epochSize = epSizeShort, startTime = timestamp_POSIX)
        fnames[[i]] = list(unlist(fnames[i]), ID) # add ID to fnames object
        # convert sleeplevel to sleep because GGIR does not deal with sleep stages
        if ("sleeplevel" %in% names(D$data)) {
          D$data$sleep = 0
          D$data$sleep[which(D$data$sleeplevel %in% c("asleep", "deep", "light", "rem", "restless"))] = 1
          D$data = D$data[, -which(names(D$data) == "sleeplevel")]
        }
        if ("calories" %in% names(D$data)) {
          names(D$data)[which(names(D$data) == "calories")] = "ExtAct"
        }
        if ("heart_rate" %in% names(D$data)) {
          names(D$data)[which(names(D$data) == "heart_rate")] = "ExtHeartRate"
        }
        D$data = D$data[, -which(names(D$data) %in% c("seconds", "dateTime"))]
      }
      if ("sleep" %in% colnames(D$data)) colnames(D$data)[which(colnames(D$data) == "sleep")] = "ExtSleep"
      if ("steps" %in% colnames(D$data)) colnames(D$data)[which(colnames(D$data) == "steps")] = "ExtStep"
      # Convert lists to objects as expected by code below
      timestamp_POSIX = D$startTime
      epSizeShort = D$epochSize
      D = D$data
      
      #-----------------------------------
      # Aggregate long epochs if desired epoch length is larger
      if (!is.null(D_extraVars)) {
        if ("light" %in% colnames(D_extraVars)) {
          step = epSizeLong %/% epSizeShort
          D_extraVars = matAggregate(D_extraVars, step)
        }
      }
      
      #-----------------------------------
      # Aggregate short epochs if desired epoch length is larger
      desiredEpochSize = params_general[["windowsizes"]][1]
      if (!is.null(desiredEpochSize)) {
        if (desiredEpochSize > epSizeShort) {
          step = desiredEpochSize %/% epSizeShort
          
          D = matAggregate(D, step)
          epSizeShort = epSizeShort * step
        }
        checkEpochMatch(desiredEpochSize, epSizeShort)
      }
      
      quartlystart = (ceiling(as.numeric(timestamp_POSIX) / epSizeLong)) * epSizeLong
      ts_longEp_POSIX = as.POSIXlt(quartlystart, tz = tz, origin = "1970-01-01") # POSIX time
      M$wdayname = weekdays(x = ts_longEp_POSIX, abbreviate = FALSE) # extract weekday as day name
      M$wday = ts_longEp_POSIX$wday # extract weekday as a daynumber
      if (length(ts_longEp_POSIX) > 2) {
        # calculate time difference relative to original data
        # to know how much epoch to ignore at the beginning of the file
        deltastart = as.numeric(difftime(ts_longEp_POSIX, timestamp_POSIX,units = "sec") / 5)
        # shorten the data (object D) accordingly
        D = D[(deltastart + 1):nrow(D), drop = FALSE]
      }
      # Define compatible metashort and metalong dimensions
      LMS = nrow(D)
      LML = LMS * epSizeShort / epSizeLong
      expected_LML = floor(LML)
      if (expected_LML != LML) {
        # this means LMS and LML are not compatible
        # keep only until last epSizeLong available
        # redefine metashort length
        LMS = expected_LML * epSizeLong / epSizeShort
        D = D[1:LMS, , drop = FALSE]
        LML = expected_LML
      }
      # Create timeseries for metashort
      time_shortEp_num = seq(quartlystart, quartlystart + ((nrow(D) - 1) * epSizeShort), by = epSizeShort)
      time_longEp_num = seq(quartlystart, quartlystart + ((nrow(D) - 1) * epSizeShort), by = epSizeLong)
      if (LML * 15 > LMS) {
        time_longEp_num = time_longEp_num[1:floor(length(time_shortEp_num) / (epSizeLong/epSizeShort))]
        time_shortEp_num = time_shortEp_num[1:(length(time_longEp_num) * (epSizeLong/epSizeShort))]
      }
      starttime = as.POSIXlt(time_longEp_num[1], origin = "1970-1-1", tz = tz)
      time_shortEp_8601 = POSIXtime2iso8601(x = as.POSIXlt(time_shortEp_num, tz = tz,
                                                           origin = "1970-01-01"),
                                            tz = tz)
      time_longEp_8601 = POSIXtime2iso8601(x = as.POSIXlt(time_longEp_num, tz = tz, origin = "1970-01-01"),
                                           tz = tz)
      # File formats with possibly more than 1 column
      morethan1 = c("actigraph_csv", "sensewear_xls", "actiwatch_awd",
                    "actiwatch_csv", "phb_xlsx", "actical_csv", "fitbit_json") #, "
      if (params_general[["dataFormat"]] %in% morethan1 == FALSE) {
        M$metashort = data.frame(timestamp = time_shortEp_8601,
                                 ExtAct = D[1:length(time_shortEp_8601),1],stringsAsFactors = FALSE)
      } else {
        M$metashort = as.data.frame(cbind(time_shortEp_8601,
                                          D[1:length(time_shortEp_8601), ]))
        colnames(M$metashort) = c("timestamp", colnames(D))
        for (ic in 2:ncol(M$metashort)) {
          M$metashort[, ic] <- as.numeric(M$metashort[, ic])
        }
      }
      LML = length(time_longEp_8601)
      if (params_general[["dataFormat"]] == "ukbiobank_csv") {
        names(M$metashort)[2] = "LFENMO"
        # Use UKBiobank's second column, which is a non-wear score
        imp = D[, 2]
        M$metashort$LFENMO = M$metashort$LFENMO/1000
        nonwear_in_data = TRUE
      } else if (params_general[["dataFormat"]] == "fitbit_json") {
        nonwear_in_data = TRUE
        imp = unlist(D[, "ExtAct"])
        missingValueRows = rowSums(is.na(D))
        
        # Create log of data missingnes
        varnames = colnames(D)
        Ntotal = nrow(D)
        
        perc_na = function(x, colname, Ntotal) {
          y = round((length(which(is.na(x[,colname]) == TRUE)) / Ntotal) * 100, digits = 2)
          return(y)
        }
        perc_value = function(x, value, Ntotal) {
          y = round((length(which(x == value)) / Ntotal) * 100, digits = 2)
          return(y)
        }
        n_ExtStep_missing = ifelse(test = "ExtStep" %in% varnames,
                                   yes = perc_na(D, "ExtStep", Ntotal),
                                   no = NA)
        n_ExtHeartRate_missing = ifelse(test = "ExtHeartRate" %in% varnames,
                                        yes = perc_na(D, "ExtHeartRate", Ntotal),
                                        no = NA)
        n_ExtAct_missing = ifelse(test = "ExtAct" %in% varnames,
                                  yes = perc_na(D, "ExtAct", Ntotal),
                                  no = NA)
        n_ExtSleep_missing = ifelse(test = "ExtSleep" %in% varnames,
                                    yes = perc_na(D, "ExtSleep", Ntotal),
                                    no = NA)
        n_ExtSleep_and_ExtAct_missing = ifelse(test = all(c("ExtSleep", "ExtAct") %in% varnames),
                                               yes = length(which(is.na(D[,"ExtSleep"]) == TRUE &
                                                                    is.na(D[,"ExtAct"]) == TRUE)) / Ntotal,
                                               no = NA)
        QClog = data.frame(filehealth_0vars_missing = perc_value(missingValueRows, 0, Ntotal),
                           filehealth_1vars_missing = perc_value(missingValueRows, 1, Ntotal),
                           filehealth_2vars_missing = perc_value(missingValueRows, 2, Ntotal),
                           filehealth_3vars_missing = perc_value(missingValueRows, 3, Ntotal),
                           filehealth_4vars_missing = perc_value(missingValueRows, 4, Ntotal),
                           filehealth_ExtAct_missing = n_ExtAct_missing,
                           filehealth_ExtStep_missing = n_ExtStep_missing,
                           filehealth_ExtHeartRate_missing = n_ExtHeartRate_missing,
                           filehealth_ExtSleep_missing = n_ExtSleep_missing,
                           filehealth_ExtSleep_y_ExtAct_missing = n_ExtSleep_and_ExtAct_missing)
      } else {
        imp = unlist(D[, 1])
        nonwear_in_data = FALSE
      }
      if (nonwear_in_data == TRUE) {
        if (params_general[["dataFormat"]] == "ukbiobank_csv") {
          # Take long epoch mean of UK Biobank based invalid data indicater per 5 seconds
          imp2 = cumsum(imp)
          imp3 = diff(imp2[seq(1, length(imp2),
                               by = ((60/epSizeShort) * (epSizeLong/60)))]) / ((60/epSizeShort) * (epSizeLong/60)) # rolling mean
          nonwearscore = round(imp3 * 3) # create three level nonwear score from it, not really necessary for GGIR, but helps to retain some of the information
        } else if (params_general[["dataFormat"]] == "fitbit_json") {
          if (any(c("ExtAct", "ExtSleep") %in% colnames(D) == FALSE)) {
            nonwearscore = rep(3, nrow(D))
          } else {
            
            # Step 1: Use missing data as first indication of nonwear
            missingValueRows = rowSums(is.na(D[, c("ExtAct", "ExtSleep")]))
            nonZero = which(missingValueRows > 0)
            if (length(nonZero) > 0) {
              missingValueRows[nonZero] = 1
            }
            imp2 = cumsum(missingValueRows)
            imp3 = diff(imp2[seq(1, length(imp2),
                                 by = ((60/epSizeShort) * (epSizeLong/60)))]) / ((60/epSizeShort) * (epSizeLong/60)) # rolling mean
            nonwearscore = round(imp3) * 3 # create three level nonwear score from it, not really necessary for GGIR, but helps to retain some of the information
            # Step 2: Use windows with no movement as seocnd indication of nonwear
            ni = 1
            for (g in seq(from = 1, to = length(imp), by = epSizeLong / epSizeShort)) {
              iend = g + (epSizeNonWear / epSizeShort) - 1
              indices = g:iend
              if (iend <= length(imp)) {
                if (any(is.na(imp[indices])) || sum(imp[indices], na.rm = TRUE) <= 0) {
                  nonwearscore[ni] = 3
                }
                if (length(which(!is.na(imp[indices]))) > 1) {
                  if (sd(imp[indices], na.rm = TRUE) < 0.0001 &&
                      mean(imp[indices], na.rm = TRUE) < 2) {
                    nonwearscore[ni] = 3
                  }
                } else {
                  nonwearscore[ni] = 3
                }
              }
              ni = ni + 1
            }
          }
        }
      } else {
        # Using rolling long window sum to indicate whether it is nonwear
        nonwearscore = rep(0, LML)
        ni = 1
        for (g in seq(from = 1, to = length(imp), by = epSizeLong / epSizeShort)) {
          iend = g + (epSizeNonWear / epSizeShort) - 1
          indices = g:iend
          if (iend <= length(imp)) {
            if (any(is.na(imp[indices])) || sum(imp[indices], na.rm = TRUE) <= 0) {
              nonwearscore[ni] = 3
            }
            if (length(which(!is.na(imp[indices]))) > 0) {
              # For Fitbit and Sensewear nonwear detection is based on calories/MET values
              if (params_general[["dataFormat"]] %in% c("fitbit_json", "sensewear_xls") &&
                  sd(imp[indices], na.rm = TRUE) < 0.0001 &&
                  mean(imp[indices], na.rm = TRUE) < 2) {
                nonwearscore[ni] = 3
              }
            } else {
              nonwearscore[ni] = 3
            }
          }
          ni = ni + 1
        }
        if (length(nonwearscore) > length(time_longEp_8601)) nonwearscore = nonwearscore[1:length(time_longEp_8601)]
        nonwearscore = round(nonwearscore)
        if (any(is.na(nonwearscore))) nonwearscore[which(is.na(nonwearscore))] = 3
      }
      if (length(nonwearscore) < LML) {
        nonwearscore = c(nonwearscore, rep(0, LML - length(nonwearscore)))
      }
      #--------------------------------------------
      # Create myfun object to trigger outcome type specific analysis
      remove_missing_vars_from_myfun = function(myfun, availableVars) {
        # function to correct the myfun object for variables
        # that are not in the data
        var_missing = which(myfun$colnames %in% availableVars == FALSE)
        if (length(var_missing) > 0) {
          myfun[["colnames"]] = myfun[["colnames"]][-var_missing]
          myfun[["outputtype"]] = myfun[["outputtype"]][-var_missing]
          myfun[["reporttype"]] = myfun[["reporttype"]][-var_missing]
        }
        return(myfun)
      }
      if (params_general[["dataFormat"]] == "sensewear_xls") {
        myfun = list(FUN = NA,
                     parameters = NA, 
                     expected_sample_rate = NA, 
                     expected_unit = "g", 
                     colnames = c("ExtAct", "ExtStep", "ExtSleep"),
                     outputres = epSizeShort,
                     minlength = NA,
                     outputtype = c("numeric", "numeric", "numeric"),
                     aggfunction = NA,
                     timestamp = F, 
                     reporttype = c("scalar", "event", "type"))
        
      } else if (params_general[["dataFormat"]] == "fitbit_json") {
        myfun = list(FUN = NA,
                     parameters = NA, 
                     expected_sample_rate = NA, 
                     expected_unit = "g", 
                     colnames = c("ExtAct", "ExtStep","ExtHeartRate", "ExtSleep"),
                     outputres = epSizeShort,
                     minlength = NA,
                     outputtype = c("numeric", "numeric", "numeric", "numeric"),
                     aggfunction = NA,
                     timestamp = F, 
                     reporttype = c("scalar", "event",  "scalar", "type"))
        myfun = remove_missing_vars_from_myfun(myfun, availableVars = colnames(D))
      } else if (params_general[["dataFormat"]] %in% c("phb_xlsx")) {
        myfun = list(FUN = NA,
                     parameters = NA, 
                     expected_sample_rate = NA, 
                     expected_unit = "g", 
                     colnames = c("ExtAct", "ExtStep", "ExtSleep"),
                     outputres = epSizeShort,
                     minlength = NA,
                     outputtype = c("numeric", "numeric", "numeric"),
                     aggfunction = NA,
                     timestamp = F, 
                     reporttype = c("scalar", "event", "type"))
        myfun = remove_missing_vars_from_myfun(myfun, availableVars = colnames(D))
      } else if (params_general[["dataFormat"]] == "actiwatch_csv" && "ExtSleep" %in% colnames(D)) {
        myfun = list(FUN = NA,
                     parameters = NA, 
                     expected_sample_rate = NA, 
                     expected_unit = "g", 
                     colnames = c("ExtAct", "ExtSleep"),
                     outputres = epSizeShort,
                     minlength = NA,
                     outputtype = c("numeric", "numeric"),
                     aggfunction = NA,
                     timestamp = F, 
                     reporttype = c("scalar", "type"))
        myfun = remove_missing_vars_from_myfun(myfun, availableVars = colnames(D))
        # flip sleep scoring 1 <-> 0 to be consistent with GGIR,
        # where 1 is sleep and 0 is wake:
        M$metashort$ExtSleep[which(M$metashort$ExtSleep == 0)] = -1
        M$metashort$ExtSleep[which(M$metashort$ExtSleep == 1)] = 0
        M$metashort$ExtSleep[which(M$metashort$ExtSleep == -1)] = 1
      }
      if (params_general[["dataFormat"]] == "phb_xlsx") {
        neg_indices = which(D$data$ExtAct < 0)
        D$data$ExtAct[neg_indices] = 0
      }
      # create data.frame for metalong, note that light and temperature are just set at zero by default
      M$metalong = data.frame(timestamp = time_longEp_8601, nonwearscore = nonwearscore, #rep(0,LML)
                              clippingscore = rep(0,LML), lightmean = rep(0, LML),
                              lightpeak = rep(0,LML), temperaturemean = rep(0, LML),
                              EN = rep(0,LML))
      # If light was loaded (Actiwatch/PHB) then store this in metalong
      if (!is.null(D_extraVars) && "light" %in% colnames(D_extraVars)) {
        if (nrow(D_extraVars) == nrow(M$metalong)) {
          M$metalong$lightpeak = D_extraVars[, "light"]
        }
      }
      # update weekday name and number based on actual data
      M$wdayname = weekdays(x = starttime, abbreviate = FALSE)
      M$wday = as.POSIXlt(starttime)$wday + 1
      # replace missing values by zero
      for (activityColumn in c("ExtAct", "ExtStep")) {
        if (activityColumn %in% colnames(M$metashort)) {
          M$metashort[is.na(M$metashort[, activityColumn]),  activityColumn] = 0
        }
      }
      M$QClog = QClog
      # Save these files as new meta-file
      filefoldername = NA
      save(M, C, I, myfun, filename_dir, filefoldername,
           file = outputFileName)
    }
  }
  
  #--------------------------------------------------------------------------------
  # Run the code either parallel or in serial (file index starting with f0 and ending with f1)
  cores = parallel::detectCores()
  Ncores = cores[1]
  if (params_general[["do.parallel"]] == TRUE) {
    if (Ncores > 3) {
      if (length(params_general[["maxNcores"]]) == 0) params_general[["maxNcores"]] = Ncores
      Ncores2use = min(c(Ncores - 1, params_general[["maxNcores"]], (f1 - f0) + 1))
      if (Ncores2use > 1) {
        cl <- parallel::makeCluster(Ncores2use) # not to overload your computer
        parallel::clusterExport(cl = cl,
                                varlist = c(unclass(lsf.str(envir = asNamespace("GGIR"), all = T)),
                                            "MONITOR", "FORMAT"),
                                envir = as.environment(asNamespace("GGIR")))
        parallel::clusterEvalQ(cl, Sys.setlocale("LC_TIME", "C"))
        doParallel::registerDoParallel(cl)
      } else {
        # Don't process in parallel if only one core
        params_general[["do.parallel"]] = FALSE
      }
    } else {
      if (verbose == TRUE) cat(paste0("\nparallel processing not possible because number of available cores (",Ncores,") < 4"))
      params_general[["do.parallel"]] = FALSE
    }
  }
  if (params_general[["do.parallel"]] == TRUE) {
    if (verbose == TRUE) cat(paste0('\n Busy processing ... see ',
                                    metadatadir, '/meta/basic',
                                    ' for progress\n'))
    # check whether we are indevelopment mode:
    GGIRinstalled = is.element('GGIR', installed.packages()[,1])
    packages2passon = functions2passon = NULL
    GGIRloaded = "GGIR" %in% .packages()
    if (GGIRloaded) { #pass on package
      packages2passon = 'GGIR'
      errhand = 'pass'
    } else {
      # pass on functions
      functions2passon = c("POSIXtime2iso8601", "aggregateEvent")
      errhand = 'stop'
    }
    i = 0 # declare i because foreach uses it, without declaring it
    `%myinfix%` = foreach::`%dopar%`
    output_list = foreach::foreach(i = f0:f1, .packages = packages2passon,
                                   .export = functions2passon, .errorhandling = errhand, .verbose = F) %myinfix% {
                                     tryCatchResult = tryCatch({
                                       main_convert(i, fnames, metadatadir, params_general,
                                                    I_bu, epSizeShort, epSizeLong, tz, 
                                                    verbose, M, C)
                                     })
                                     return(tryCatchResult)
                                   }
    on.exit(parallel::stopCluster(cl))
    for (oli in 1:length(output_list)) { # logged error and warning messages
      if (is.null(unlist(output_list[oli])) == FALSE) {
        if (verbose == TRUE) cat(paste0("\nErrors and warnings for ",fnames[oli]))
        print(unlist(output_list[oli])) # print any error and warnings observed
      }
    }
  } else {
    for (i in f0:f1) {
      main_convert(i, fnames, metadatadir, params_general, I_bu,
                   epSizeShort, epSizeLong, tz, verbose, M, C)
    }
  }
  
}

library(GGIR)
context("extractID")
test_that("extractID recognizes IDs correctly", {
  skip_on_cran()
  
  hvars = list(ID = "123A_testaccfile.csv",
               iID = "not extracted",
               IDd = "123A",
               HN = "not extracted",
               sensor.location = "not extracted",
               SX = "not available",
               deviceSerialNumber = "MOS2D12345678_firmware_NA")
  
  # idloc = 2
  ID = extractID(hvars, idloc = 2, fname = "123A_testaccfile.csv")
  expect_equal(ID, "123A")
  
  # idloc = 3
  ID = extractID(hvars, idloc = 3, fname = "123A_testaccfile.csv")
  expect_equal(ID, "123A_testaccfile.csv")
  
  hvars$ID = "123A-testaccfile.csv"
  ID = extractID(hvars, idloc = 3, fname = "123A_testaccfile.csv")
  expect_equal(ID, "123A")
  
  # idloc = 4
  ID = extractID(hvars, idloc = 4, fname = "123A_testaccfile.csv")
  expect_equal(ID, "123A")
  
  # idloc = 5
  ID = extractID(hvars, idloc = 5, fname = "123A testaccfile.csv")
  expect_equal(ID, "123A")
  
  # idloc = 6
  ID = extractID(hvars, idloc = 6, fname = "123A.testaccfile.csv")
  expect_equal(ID, "123A")
  
  # idloc = 7
  ID = extractID(hvars, idloc = 7, fname = "123A-testaccfile.csv")
  expect_equal(ID, "123A")
  
  # NAs
  hvars$ID = "NA"
  ID = extractID(hvars, idloc = 1, fname = "123A-testaccfile.csv")
  expect_equal(ID, "notextracted")
  
  hvars$iID = "NA"
  hvars$ID = "123A_testaccfile.csv"
  ID = extractID(hvars, idloc = 1, fname = "123A-testaccfile.csv")
  expect_equal(ID, "123A_testaccfile.csv")
  
  # no ID
  hvars$ID = hvars$IDd = NULL
  expect_warning(extractID(hvars, idloc = 1, 
                           fname = "123A-testaccfile.csv"),
                 regexp = "Unable to extract ID")
  
})

library("testthat")
library("GGIR")
LC_TIME_backup = Sys.getlocale("LC_TIME")
Sys.setlocale("LC_TIME", "C") # standardises all language in GGIR to English
test_check("GGIR")
Sys.setlocale("LC_TIME", LC_TIME_backup) # reverts language setting

g.analyse.perfile = function(I, C, metrics_nav,
                             AveAccAve24hr, doquan, doiglevels, tooshort, 
                             params_247, params_cleaning, params_general,
                             output_avday, output_perday,
                             dataqual_summary, file_summary) {
  
  # extract objects from lists in input:
  cosinor_coef = output_avday$cosinor_coef
  daysummary = output_perday$daysummary
  ds_names = output_perday$ds_names
  lookat = metrics_nav$lookat
  colnames_to_lookat = metrics_nav$colnames_to_lookat
  
  filesummary = matrix(" ", 1, 150) #matrix to be stored with summary per participant
  s_names = rep(" ", ncol(filesummary))
  vi = 1
  # tidy up daysummary
  cut = which(ds_names == " " | ds_names == "" | is.na(ds_names) == T)
  if (length(cut > 0)) {
    ds_names = ds_names[-cut]
    daysummary = daysummary[,-cut]
  }
  # for a very small file, there could be just one row in daysummary[-cut,], so it gets coerced to a vector.
  # But what we actually need is a 1-row matrix. So we need to transpose it. 
  if(is.vector(daysummary)) { 
    daysummary = t(daysummary)
  }

  # Person identification number
  filesummary[vi] = file_summary$ID
  # Identify which of the metrics are in g-units to aid deciding whether to multiply by 1000
  g_variables_lookat = lookat[grep(x = colnames_to_lookat, pattern = "BrondCount|ZCX|ZCY|ZCZ|NeishabouriCount|ExtAct|ExtHeartRate", invert = TRUE)]
  
  # Serial number
  filesummary[(vi + 1)] = file_summary$deviceSerialNumber
  s_names[vi:(vi + 1)] = c("ID","device_sn")
  vi = vi + 2
  # starttime of measurement, body location, filename
  filesummary[vi] = file_summary$sensor.location
  if (!is.null(params_general[["recording_split_times"]])) {
    filename = file_summary$fname
    splitnames = getSplitNames(filename)
    segment_names = splitnames$segment_names
    filename = splitnames$filename
  } else {
    segment_names = NULL
    filename = file_summary$fname
  }
  filesummary[(vi + 1)] = filename
  filesummary[(vi + 2)] = file_summary$startt # starttime of measurement
  s_names[vi:(vi + 2)] = c("bodylocation","filename","start_time")
  vi = vi + 3
  
  if (!is.null(params_general[["recording_split_times"]]) && !is.null(segment_names)) {
    filesummary[vi] = segment_names[2]
    filesummary[vi + 1] = segment_names[3]
    s_names[vi:(vi + 1)] = c("split1_name","split2_name")
    vi = vi + 2
  }
  # weekday on which measurement started, sample frequency and device
  filesummary[vi] = file_summary$wdayname
  filesummary[(vi + 1)] = I$sf
  filesummary[(vi + 2)] = I$monn
  s_names[vi:(vi + 2)] = c("startday", "samplefreq", "device")
  vi = vi + 3
  # clipsing score and measurement duration in days
  filesummary[vi] = dataqual_summary$clipping_score
  filesummary[vi + 1] = dataqual_summary$meas_dur_dys #measurement duration in days
  # completeness core and measurement duration with different definitions
  filesummary[vi + 2] = dataqual_summary$dcomplscore #completeness of the day
  filesummary[vi + 3] = dataqual_summary$meas_dur_def_proto_day #measurement duration according to protocol
  filesummary[vi + 4] = dataqual_summary$wear_dur_def_proto_day #wear duration in days (out of measurement protocol)
  s_names[vi:(vi + 4)] = c("clipping_score", "meas_dur_dys", "complete_24hcycle", 
                           "meas_dur_def_proto_day", "wear_dur_def_proto_day")
  vi = vi + 5
  if (!is.null(params_cleaning[["nonwearFiltermaxHours"]])) {
    filesummary[vi] = dataqual_summary$nonwearHoursFiltered
    filesummary[vi + 1] = dataqual_summary$nonwearEventsFiltered
    s_names[vi:(vi + 1)] = c("nonwear_hours_filtered", "nonwear_events_filtered")
    vi = vi + 2
  }
  # calibration error after auto-calibration
  if (length(C$cal.error.end) == 0)   C$cal.error.end = c(" ")
  filesummary[vi] = C$cal.error.end
  filesummary[vi + 1] = C$QCmessage
  for (la in 1:length(lookat)) {
    AveAccAve24hr[la] = 	AveAccAve24hr[la] * ifelse(test = lookat[la] %in% g_variables_lookat, yes = 1000, no = 1)
  }
  q0 = length(AveAccAve24hr) + 1
  filesummary[(vi + 2):(vi + q0)] = AveAccAve24hr
  colnames_to_lookat = paste0(colnames_to_lookat, "_fullRecordingMean")
  step_count_var = grep("step_count", x = colnames_to_lookat)
  if (length(step_count_var) > 0) {
    colnames_to_lookat[step_count_var] = paste0("ExtFunEvent_", colnames_to_lookat[step_count_var])
  }
  s_names[vi:(vi + q0)] = c("calib_err",
                            "calib_status", colnames_to_lookat)
  vi = vi + q0 + 2
  # QClog summary
  if ("Dur_chsum_failed" %in% names(file_summary)) {
    # readAxivity QClog
    # These are summaries of the file health check by the GGIRread::readAxivity
    # the function handles data blocks (1-3 seconds) with faulty data by imputing 
    # them and logging the information.
    # Normally we do not expect issue with cwa files, but by logging the information
    # we will facilitate better insight into when this happens.
    filesummary[vi:(vi + 6)] = c(ifelse(is.null(file_summary$Dur_imputed), " ", file_summary$Dur_imputed), # total imputed
                                 ifelse(is.null(file_summary$Dur_chsum_failed), " ", file_summary$Dur_chsum_failed), # checksum
                                 ifelse(is.null(file_summary$Dur_nonincremental), " ", file_summary$Dur_nonincremental), # nonincremental id between blocks
                                 ifelse(is.null(file_summary$Dur_freqissue_5_10), " ", file_summary$Dur_freqissue_5_10), # bias 5-10%
                                 ifelse(is.null(file_summary$Dur_freqissue_10_20), " ", file_summary$Dur_freqissue_10_20), # bias 10-20%
                                 ifelse(is.null(file_summary$Dur_freqissue_20_30), " ", file_summary$Dur_freqissue_20_30), # bias 20-30%
                                 ifelse(is.null(file_summary$Dur_freqissue_30), " ", file_summary$Dur_freqissue_30)) # bias >30%
    s_names[vi:(vi + 6)] = c("filehealth_totimp_min",
                             "filehealth_checksumfail_min",
                             "filehealth_niblockid_min", # non incremental block id
                             "filehealth_fbias0510_min", # frequency bias
                             "filehealth_fbias1020_min", 
                             "filehealth_fbias2030_min",
                             "filehealth_fbias30_min")
    vi = vi + 7
    filesummary[vi:(vi + 6)] = c( ifelse(is.null(file_summary$Nblocks_imputed), " ", file_summary$Nblocks_imputed),
                                  ifelse(is.null(file_summary$Nblocks_chsum_failed), " ", file_summary$Nblocks_chsum_failed),
                                  ifelse(is.null(file_summary$Nblocks_nonincremental), " ", file_summary$Nblocks_nonincremental),
                                  ifelse(is.null(file_summary$Nblock_freqissue_5_10), " ", file_summary$Nblock_freqissue_5_10),
                                  ifelse(is.null(file_summary$Nblock_freqissue_10_20), " ", file_summary$Nblock_freqissue_10_20),
                                  ifelse(is.null(file_summary$Nblock_freqissue_20_30), " ", file_summary$Nblock_freqissue_20_30),
                                  ifelse(is.null(file_summary$Nblock_freqissue_30), " ", file_summary$Nblock_freqissue_30))
    s_names[vi:(vi + 6)] = c("filehealth_totimp_N",
                             "filehealth_checksumfail_N",
                             "filehealth_niblockid_N",
                             "filehealth_fbias0510_N",
                             "filehealth_fbias1020_N",
                             "filehealth_fbias2030_N",
                             "filehealth_fbias30_N")
    vi = vi + 7
  } else if ("Dur_imputed" %in% names(file_summary)) {
    # ActiGraph QClog
    # This also logs time gaps in ActiGraph files, which correspond with periods
    # in which the idle sleep mode has been activated
    filesummary[vi:(vi + 1)] = c(file_summary$Dur_imputed, # total imputed
                                 file_summary$Nblocks_imputed)
    s_names[vi:(vi + 1)] = c("filehealth_totimp_min",
                    "filehealth_totimp_N")
    vi = vi + 2
  }
  if ("filehealth_ExtSleep_y_ExtAct_missing" %in% names(file_summary)) {
    Fitbit_filehealth_columns = c("filehealth_0vars_missing",
                                  "filehealth_1vars_missing",
                                  "filehealth_2vars_missing",
                                  "filehealth_3vars_missing",
                                  "filehealth_4vars_missing",
                                  "filehealth_ExtAct_missing",
                                  "filehealth_ExtStep_missing",
                                  "filehealth_ExtHeartRate_missing",
                                  "filehealth_ExtSleep_missing",
                                  "filehealth_ExtSleep_y_ExtAct_missing")
    filesummary[vi:(vi + 9)] = file_summary[, Fitbit_filehealth_columns]
    s_names[vi:(vi + 9)] = Fitbit_filehealth_columns
    vi = vi = 10
  }
  #quantile, ML5, and intensity gradient variables
  if (doquan == TRUE) {
    q1 = length(output_avday$QUAN)
    filesummary[vi:((vi - 1) + q1)] = output_avday$QUAN * ifelse(test = lookat[la] %in% g_variables_lookat, yes = 1000, no = 1)
    s_names[vi:((vi - 1) + q1)] = paste0(output_avday$qlevels_names, "_fullRecording")
    vi = vi + q1
    q1 = length(output_avday$ML5AD)
    filesummary[vi:((vi - 1) + q1)] = as.numeric(output_avday$ML5AD)
    s_names[vi:((vi - 1) + q1)] = paste0(output_avday$ML5AD_names, "_fullRecording")
    vi = vi + q1
  }
  if (doiglevels == TRUE) {
    # intensity gradient (as described by Alex Rowlands 2018)
    # applied to the averageday per metric (except from angle metrics)
    q1 = length(output_avday$igfullr)
    filesummary[vi:((vi - 1) + q1)] = output_avday$igfullr
    s_names[vi:((vi - 1) + q1)] = paste0(output_avday$output_avday$igfullr_names, "_fullRecording")
    vi = vi + q1
  }
  if (tooshort == 0) {
    #====================================================================
    # Summarise per recording (not per day) - additional variables if the recording is long enough
    #====================================================================
    # Recognise weekenddays with enough data
    wkend  = which(daysummary[, which(ds_names == "weekday")] == "Saturday" |
                     daysummary[, which(ds_names == "weekday")] == "Sunday")
    columnWithAlwaysData = which(ds_names == "N hours" | ds_names == "N_hours")
    NVHcolumn = which(ds_names == "N valid hours" | ds_names == "N_valid_hours" ) #only count in the days for which the inclusion criteria is met
    v1 = which(is.na(as.numeric(daysummary[wkend, columnWithAlwaysData])) == F &
                 as.numeric(daysummary[wkend, NVHcolumn]) >= params_cleaning[["includedaycrit"]][1])
    wkend = wkend[v1]
    wkday  = which(daysummary[,which(ds_names == "weekday")] != "Saturday" & daysummary[,which(ds_names == "weekday")] != "Sunday")
    v2 = which(is.na(as.numeric(daysummary[wkday, columnWithAlwaysData])) == F  &
                 as.numeric(daysummary[wkday, NVHcolumn]) >= params_cleaning[["includedaycrit"]][1])
    wkday = wkday[v2]
    validdays = which(is.na(as.numeric(daysummary[, columnWithAlwaysData])) == F  &
                      as.numeric(daysummary[, NVHcolumn]) >= params_cleaning[["includedaycrit"]][1])
    # Add number of weekend and weekdays to filesummary
    filesummary[vi:(vi + 1)] = c(length(wkend),  length(wkday)) # number of weekend days & weekdays
    iNA = which(is.na(filesummary[vi:(vi + 1)]) == TRUE)
    if (length(iNA) > 0) filesummary[(vi:(vi + 1))[iNA]] = 0
    s_names[vi:(vi + 1)] = c("N valid weekend days (WE)","N valid weekdays (WD)")
    vi = vi + 2
    # Cosinor analysis + IV + IS + phi
    if (length(cosinor_coef) > 0) {
      filesummary[vi]  = c(cosinor_coef$timeOffsetHours)
      s_names[vi] = c("cosinor_timeOffsetHours")
      vi = vi + 1
      try(expr = {filesummary[vi:(vi + 5)]  = as.numeric(c(cosinor_coef$coef$params$mes,
                                                           cosinor_coef$coef$params$amp,
                                                           cosinor_coef$coef$params$acr,
                                                           cosinor_coef$coef$params$acrotime,
                                                           cosinor_coef$coef$params$ndays,
                                                           cosinor_coef$coef$params$R2))}, silent = TRUE)
      s_names[vi:(vi + 5)] = c("cosinor_mes", "cosinor_amp", "cosinor_acrophase",
                               "cosinor_acrotime", "cosinor_ndays", "cosinor_R2")
      vi = vi + 6
      try(expr = {filesummary[vi:(vi + 10)]  = c(cosinor_coef$coefext$params$minimum,
                                                 cosinor_coef$coefext$params$amp,
                                                 cosinor_coef$coefext$params$alpha,
                                                 cosinor_coef$coefext$params$beta,
                                                 cosinor_coef$coefext$params$acrotime,
                                                 cosinor_coef$coefext$params$UpMesor,
                                                 cosinor_coef$coefext$params$DownMesor,
                                                 cosinor_coef$coefext$params$MESOR,
                                                 cosinor_coef$coefext$params$ndays,
                                                 cosinor_coef$coefext$params$F_pseudo,
                                                 cosinor_coef$coefext$params$R2)}, silent = TRUE)
      s_names[vi:(vi + 10)] = c("cosinorExt_minimum", "cosinorExt_amp", "cosinorExt_alpha",
                                "cosinorExt_beta", "cosinorExt_acrotime", "cosinorExt_UpMesor",
                                "cosinorExt_DownMesor", "cosinorExt_MESOR",
                                "cosinorExt_ndays", "cosinorExt_F_pseudo", "cosinorExt_R2")
      vi = vi + 11
      filesummary[vi:(vi + 2)]  = c(cosinor_coef$IVIS$InterdailyStability,
                                    cosinor_coef$IVIS$IntradailyVariability,
                                    cosinor_coef$IVIS$phi)
      s_names[vi:(vi + 2)] = c("IS", "IV", "phi")
      vi = vi + 3
    } else {
      vi = vi + 21
    }
    
    # Variables per metric - summarise with stratification to weekdays and weekend days
    daytoweekvar = c(5:length(ds_names))
    md = which(ds_names[daytoweekvar] %in% c("measurementday", "weekday", "qwindow_timestamps", "qwindow_names"))
    if (length(md) > 0) daytoweekvar = daytoweekvar[-md]
    
    dtwtel = 0
    if (length(daytoweekvar) >= 1) {
      sp = length(daytoweekvar) + 1
      for (dtwi in daytoweekvar) {
        #check whether columns is empty:
        uncona = unique(daysummary[,dtwi])
        storevalue = !(length(uncona) == 1 & length(params_247[["qwindow"]]) > 2 & uncona[1] == "")
        if (is.na(storevalue) == TRUE) storevalue = FALSE
        # Only do next 15-isch lines of code if:
        # - there is more than 1 day of data
        # - there are multiple daysegments (qwindow)
        # - first value is not empty
        if (storevalue == TRUE) {
          # Plain average of available days
          v4 = mean(suppressWarnings(as.numeric(daysummary[validdays, dtwi])), na.rm = TRUE)
          filesummary[(vi + 1 + (dtwtel*sp))] = v4 # #average all availabel days
          s_names[(vi + 1 + (dtwtel * sp))] = paste0("AD_", ds_names[dtwi])
          # Average of available days per weekenddays / weekdays:
          dtw_wkend = suppressWarnings(as.numeric(daysummary[wkend,dtwi]))
          dtw_wkday = suppressWarnings(as.numeric(daysummary[wkday,dtwi]))
          filesummary[(vi + 2 + (dtwtel * sp))] = suppressWarnings(mean(dtw_wkend, na.rm = TRUE))
          filesummary[(vi + 3 + (dtwtel * sp))] = suppressWarnings(mean(dtw_wkday, na.rm = TRUE))
          s_names[(vi + 2 + (dtwtel * sp))] = paste0("WE_", ds_names[dtwi]) # Weekend no weighting
          s_names[(vi + 3 + (dtwtel * sp))] = paste0("WD_", ds_names[dtwi]) # Weekdays no weighting
          # Weighted average of available days per weekenddays / weekdays:
          if (length(dtw_wkend) > 2) {
            dtw_wkend = c((dtw_wkend[1] + dtw_wkend[3]) / 2, dtw_wkend[2])
          }
          if (length(dtw_wkday) > 5) {
            dtw_wkday = c((dtw_wkday[1] + dtw_wkday[6]) / 2, dtw_wkday[2:5])
          }
          filesummary[(vi + 4 + (dtwtel * sp))] = suppressWarnings(mean(dtw_wkend, na.rm = TRUE))
          filesummary[(vi + 5 + (dtwtel * sp))] = suppressWarnings(mean(dtw_wkday, na.rm = TRUE))
          s_names[(vi + 4 + (dtwtel * sp))] = paste0("WWE_", ds_names[dtwi]) # Weekend with weighting
          s_names[(vi + 5 + (dtwtel * sp))] = paste0("WWD_", ds_names[dtwi]) # Weekdays with weighting
          dtwtel = dtwtel + 1
        }
        # Plain average of available days
        v4 = mean(suppressWarnings(as.numeric(daysummary[validdays, dtwi])), na.rm = TRUE)
        filesummary[(vi + 1 + (dtwtel * sp))] = v4
        s_names[(vi + 1 + (dtwtel * sp))] = paste0("AD_", ds_names[dtwi])
        # Average of available days per weekenddays / weekdays:
        dtw_wkend = suppressWarnings(as.numeric(daysummary[wkend,dtwi]))
        dtw_wkday = suppressWarnings(as.numeric(daysummary[wkday,dtwi]))
        if (storevalue == TRUE) {
          filesummary[(vi + 2 + (dtwtel * sp))] = mean(dtw_wkend, na.rm = TRUE)
          filesummary[(vi + 3 + (dtwtel * sp))] = mean(dtw_wkday, na.rm = TRUE)
        }
        s_names[(vi + 2 + (dtwtel * sp))] = paste0("WE_", ds_names[dtwi]) # Weekend no weighting
        s_names[(vi + 3 + (dtwtel * sp))] = paste0("WD_", ds_names[dtwi]) # Weekdays no weighting
        # Weighted average of available days per weekenddays / weekdays:
        if (length(dtw_wkend) > 2) {
          dtw_wkend = c((dtw_wkend[1] + dtw_wkend[3]) / 2, dtw_wkend[2])
        }
        if (length(dtw_wkday) > 5) {
          dtw_wkday = c((dtw_wkday[1] + dtw_wkday[6]) / 2, dtw_wkday[2:5])
        }
        if (storevalue == TRUE) {
          filesummary[(vi + 4 + (dtwtel * sp))] = mean(dtw_wkend, na.rm = TRUE) # Weekend with weighting
          filesummary[(vi + 5 + (dtwtel * sp))] = mean(dtw_wkday, na.rm = TRUE) # Weekdays with weighting
        }
        s_names[(vi + 4 + (dtwtel * sp))] = paste0("WWE_", ds_names[dtwi])
        s_names[(vi + 5 + (dtwtel * sp))] = paste0("WWD_", ds_names[dtwi])
        dtwtel = dtwtel + 1
      }
      vi = vi + 6 + ((dtwtel * sp) - 1)
    }
    filesummary[vi] = params_cleaning[["data_masking_strategy"]]
    filesummary[(vi + 1)] = params_cleaning[["hrs.del.start"]]
    filesummary[(vi + 2)] = params_cleaning[["hrs.del.end"]]
    filesummary[(vi + 3)] = params_cleaning[["maxdur"]]
    filesummary[(vi + 4)] = params_general[["windowsizes"]][1]
    filesummary[(vi + 5)] = metrics_nav$longitudinal_axis_id
    #get GGIR version
    GGIRversion = "GGIR not used"
    if (is.element('GGIR', installed.packages()[,1])) {
      GGIRversion = as.character(utils::packageVersion("GGIR"))
      if (length(GGIRversion) != 1) GGIRversion = sessionInfo()$otherPkgs$GGIR$Version
    }
    filesummary[(vi + 6)] = GGIRversion #"2014-03-14 12:14:00 GMT"
    s_names[vi:(vi + 6)] = as.character(c(paste0("data exclusion stategy (value=1, ignore specific hours;",
                                                 " value=2, ignore all data before the first midnight and",
                                                 " after the last midnight)"),
                                          "n hours ignored at start of meas (if data_masking_strategy=1)",
                                          "n hours ignored at end of meas (if data_masking_strategy=1)",
                                          "n days of measurement after which all data is ignored (if data_masking_strategy=1)",
                                          "epoch size to which acceleration was averaged (seconds)",
                                          "if_hip_long_axis_id", "GGIR version"))
    vi = vi + 7
  }
  # tidy up daysummary object
  mw = which(is.na(daysummary) == T)
  mw = c(mw, grep(pattern = "NaN", x = daysummary))
  if (length(mw) > 0) {
    daysummary[mw] = " "
  }
  
  if (min(dim(as.matrix(daysummary))) == 1) {
    if (nrow(as.matrix(daysummary)) != 1) {
      daysummary = t(daysummary) #if there is only one day of data
    }
  }
  daysummary = data.frame(value = daysummary, stringsAsFactors = FALSE)
  names(daysummary) = ds_names
  # remove double columns with 1-6am variables
  columnswith16am = grep("1-6am", x = colnames(daysummary))
  if (length(columnswith16am) > 1) {
    daysummary = daysummary[,-columnswith16am[2:length(columnswith16am)]]
  }
  # tidy up filesummary object
  mw = which(is.na(filesummary) == T)
  mw = c(mw, grep(pattern = "NaN", x = filesummary))
  if (length(mw) > 0) {
    filesummary[mw] = " "
  }
  cut = which(as.character(s_names) == " " | as.character(s_names) == "" | is.na(s_names) == T | duplicated(s_names) |
                s_names %in% c("AD_", "WE_", "WD_", "WWD_", "WWE_",
                               "AD_N hours", "WE_N hours", "WD_N hours", "WWD_N hours", "WWE_N hours",
                               "AD_N valid hours", "WE_N valid hours", "WD_N valid hours", "WWD_N valid hours", "WWE_N valid hours"))
  if (length(cut) > 0) {
    s_names = s_names[-cut]
    filesummary = filesummary[-cut]
  }
  filesummary = data.frame(value = t(filesummary), stringsAsFactors = FALSE) #needs to be t() because it will be a column otherwise
  names(filesummary) = s_names
  
  columns2order = c()
  if (ncol(filesummary) > 37) {
    columns2order = grep(pattern = "AD_|WE_|WD_|WWD_|WWE_", x = names(filesummary))
  }
  options(encoding = "UTF-8")
  if (length(columns2order) > 0) {
    selectcolumns = c(names(filesummary)[1:(columns2order[1] - 1)],
                      grep(pattern = "^AD_", x = names(filesummary), value = T),
                      grep(pattern = "^WD_", x = names(filesummary), value = T),
                      grep(pattern = "^WE_", x = names(filesummary), value = T),
                      grep(pattern = "^WWD_", x = names(filesummary), value = T),
                      grep(pattern = "^WWE_", x = names(filesummary), value = T))
    selectcolumns = c(selectcolumns, names(filesummary)[which(names(filesummary) %in% selectcolumns == FALSE)])
  } else {
    selectcolumns = names(filesummary)
  }
  selectcolumns = selectcolumns[which(selectcolumns %in% colnames(filesummary) == TRUE)]
  filesummary = filesummary[,selectcolumns]
  filesummary = filesummary[,!duplicated(filesummary)]
  invisible(list(filesummary = filesummary, daysummary = daysummary))
}

g.report.part4 = function(datadir = c(), metadatadir = c(), f0 = c(), f1 = c(), 
                          data_cleaning_file = c(), params_sleep, params_output,
                          verbose = TRUE) {
  # description: function to merge generate report from milestone data generated by g.part4 (if store.ms was
  # set to TRUE)
  output = NULL
  ms4.out = "/meta/ms4.out"
  if (file.exists(paste(metadatadir, ms4.out, sep = ""))) {
    if (length(dir(paste(metadatadir, ms4.out, sep = ""))) == 0) {
      try.generate.report = FALSE
    } else {
      try.generate.report = TRUE
    }
  } else {
    try.generate.report = FALSE
  }
  if (try.generate.report == TRUE) {
    resultfolder = metadatadir  #resultfolder = 'Q:/studies/sleep/output_pi_sleep_wrist'
    meta.sleep.folder = paste(metadatadir, "/meta/ms3.out", sep = "")
    if (length(params_sleep[["loglocation"]]) > 0) {
      only.use.sleeplog = TRUE
    } else {
      only.use.sleeplog = FALSE
    }
    # ============================================================================ Go through
    # accelerometer datafiles and merge with sleep log data
    fnames = dir(meta.sleep.folder)
    if (f1 > length(fnames)) {
      if (verbose == TRUE) cat(paste0("\nf1 changed from, ", f1, " to ", length(fnames)))
      f1 = length(fnames)
    }
    if (length(f1) == 0 | f1 > length(fnames))
      f1 = length(fnames)
    #-----------------------------------------------------
    colnames_nightsummary2 = c("ID", "night", "sleeponset", "wakeup", "SptDuration", "sleepparam", "guider_onset",
                               "guider_wakeup", "guider_SptDuration", "error_onset", "error_wake", "error_dur", "fraction_night_invalid",
                               "SleepDurationInSpt", "WASO", "duration_sib_wakinghours", "number_sib_sleepperiod", "number_of_awakenings",
                               "number_sib_wakinghours", "duration_sib_wakinghours_atleast15min", "sleeponset_ts", "wakeup_ts", "guider_onset_ts",
                               "guider_wakeup_ts", "sleeplatency", "sleepefficiency", "page", "daysleeper", "weekday", "calendar_date",
                               "filename", "cleaningcode", "sleeplog_used", "sleeplog_ID", "acc_available", "guider", "SleepRegularityIndex1", "SriFractionValid",
                               "longitudinal_axis", "guider_corrected")
    nightsummary2 = as.data.frame(matrix(0, 0, length(colnames_nightsummary2)))

    sumi = 1
    sleeplog_used = rep(" ", ((f1 - f0) + 1))
    fnames.ms4 = list.files(paste0(metadatadir, ms4.out), full.names = TRUE)
    if (length(fnames.ms4) < f1)
      f1 = length(fnames.ms4)
    if (verbose == TRUE) cat(" loading all the milestone data from part 4 this can take a few minutes\n")
    myfun = function(x) {
      tail_expansion_log = NULL
      load(file = x)
      cut = which(nightsummary[, 1] == "")
      if (length(cut) > 0 & length(cut) < nrow(nightsummary)) {
        nightsummary = nightsummary[-cut, ]
      }
      if (length(tail_expansion_log) != 0) {
        nightsummary = nightsummary[-which(nightsummary$night == max(nightsummary$night)),] # remove last row because it may not be trustworthy
      }
      if ("GGIRversion" %in% colnames(nightsummary) == FALSE) {
        if (nrow(nightsummary) > 0) {
          nightsummary$GGIRversion = "" #before 3.0-10 this column did not exist
        } else {
          nightsummary[1, ] = NA
          nightsummary$GGIRversion = NA
          nightsummary = nightsummary[0, ]
        }
      }
      out = as.matrix(nightsummary)
    }
    nightsummary2 = as.data.frame(do.call(rbind, lapply(fnames.ms4, myfun)), stringsAsFactors = FALSE)
    nightsummary2$night = as.numeric(gsub(" ", "", nightsummary2$night))
    nightsummary2$calendar_date = as.Date(nightsummary2$calendar_date, format = "%d/%m/%Y")
    nightsummary2$calendar_date = format(nightsummary2$calendar_date, format = "%Y-%m-%d")
    nightsummary2$filename = gsub(".RData$", "", nightsummary2$filename)
    
    if (params_sleep[["sleepwindowType"]] == "TimeInBed" || params_sleep[["consider_marker_button"]] == TRUE) {
      colnames(nightsummary2) = gsub(replacement = "guider_inbedStart", pattern = "guider_onset", x = colnames(nightsummary2))
      colnames(nightsummary2) = gsub(replacement = "guider_inbedEnd", pattern = "guider_wakeup", x = colnames(nightsummary2))
      colnames(nightsummary2) = gsub(replacement = "guider_inbedDuration", pattern = "guider_SptDuration",
                                     x = colnames(nightsummary2))
    }
    # ====================================== Add non-wearing during SPT from part 5, if it is availabe:
    ms5.out = "/meta/ms5.out"
    if (file.exists(paste(metadatadir, ms5.out, sep = ""))) {
      if (length(dir(paste(metadatadir, ms5.out, sep = ""))) == 0) {
        try.add.part5.variable = FALSE  #do not run this function if there is no milestone data from g.part5
      } else {
        # check WW windows are calculated in the first file
        load(dir(paste(metadatadir, ms5.out, sep = ""), full.names = T)[1]) # this loads object output
        if ("WW" %in% output$window) { 
          try.add.part5.variable = TRUE
        } else {
          try.add.part5.variable = FALSE #do not run this function if WW windows are not calculated
        }
      }
    } else {
      try.add.part5.variable = FALSE  #do not run this function if there is no milestone data from g.part5
    }
    if (try.add.part5.variable == TRUE) {
      # ====================================================================== loop through meta-files
      fnames.ms5 = list.files(paste0(metadatadir, ms5.out), full.names = TRUE)
      if (f1 > length(fnames.ms5))
        f1 = length(fnames.ms5)
      # cat(' loading all the milestone data from part 5 this can take a few minutes\n')
      myfun5 = function(x) {
        load(file = x) # this loads the content of a RData file that has object output
        cut = which(output[, 1] == "")
        if (length(cut) > 0 & length(cut) < nrow(output)) {
          output = output[-cut, which(colnames(output) != "")]
        }
        WW = which(output[, "window"] == "WW")
        out = as.matrix(output[WW, which(colnames(output) %in% c("ID", "nonwear_perc_spt", "ACC_spt_mg", "night_number", "window"))])
      }
      outputp5 = as.data.frame(do.call(rbind, lapply(fnames.ms5[f0:f1], myfun5)), stringsAsFactors = FALSE)
      dupl = which(duplicated(outputp5[, c("ID", "night_number")]) == TRUE)
      # Note: another approach to removing duplicates could be to take the average...
      if (length(dupl) > 0) {
        # some days in part 5 can have two SPTs, remove first and keep second SPT
        outputp5 = outputp5[-dupl, ]  #-dupl2[1]
      }
      colnames(outputp5)[which(colnames(outputp5) == "night_number")] = "night"
      # merge should now work, but if ID is numeric and stored as character with a leading zero then
      # part 5 ID will not have this leading zero, so, we need fix this now:
      remove_oldID = FALSE
      if (is.character(nightsummary2$ID) & is.character(outputp5$ID)) {
        options(warn = -1)
        # next line could generate warning about NA creation when content is not numeric this is why
        # we turn of warnings
        testnumeric = !is.na(as.numeric(nightsummary2$ID))
        options(warn = 0)
        if (length(which(testnumeric == TRUE)) > (nrow(nightsummary2) * (2/3))) {
          nightsummary2$ID_old = nightsummary2$ID
          nightsummary2$ID = as.character(as.numeric(nightsummary2$ID))
          remove_oldID = TRUE
        }
      }
      # merge in variable
      outputp5$night = as.numeric(outputp5$night)
      nightsummary2 = base::merge(nightsummary2, outputp5, by = c("ID", "night"), all.x = TRUE)
      if (remove_oldID == TRUE) {
        nightsummary2$ID = nightsummary2$ID_old
        nightsummary2 = nightsummary2[, -which(names(nightsummary2) == "ID_old")]
      }
      nightsummary2 = nightsummary2[order(nightsummary2$ID, nightsummary2$night), ]
      nightsummary2$nonwear_perc_spt = as.numeric(nightsummary2$nonwear_perc_spt)
      nightsummary2$ACC_spt_mg = as.numeric(nightsummary2$ACC_spt_mg)
    }
    # =============
    skip = FALSE
    if (length(nightsummary2) != 0) {
      NumberNotNA = length(which(is.na(nightsummary2[, 3:25]) == FALSE))
      if (NumberNotNA == 0) {
        skip = TRUE
        warning("\nCannot create report part 4 report, because no sleep estimates present in milestone data.", call. = FALSE)
      }
    } else {
      skip = TRUE
      warning("\nCannot create report part 4 report, because no milestone data found for part4.", call. = FALSE)
    }
    if (skip == FALSE) {
      # skip if no data was loaded or if it all rows were NA values
      #----------------
      nightsummary = nightsummary2
      pko = which(nightsummary$sleeponset == 0 & nightsummary$wakeup == 0 & nightsummary$SptDuration == 0)
      if (length(pko) > 0) {
        nightsummary = nightsummary[-pko, ]
      }
      ##################################################### COLLAPSING nightsummary TO A ONELINE
      ##################################################### personsummary PER PARTICIPANT
      if (nrow(nightsummary) == 0) {
        if (verbose == TRUE) cat("report not stored, because no results available")
      } else {
        nightsummary_clean = tidyup_df(nightsummary)
        data.table::fwrite(nightsummary_clean, file = paste(resultfolder, "/results/QC/part4_nightsummary_sleep_full.csv",
                                                            sep = ""), row.names = FALSE, na = "",
                           sep = params_output[["sep_reports"]],
                           dec = params_output[["dec_reports"]])
        nightsummary_bu = nightsummary
      }
      ####
      summarynames_backup = c()
      for (dotwice in 1:2) {
        # store data twice, once full and once cleaned
        if (dotwice == 2) {
          # ignore nights that were derived without sleep log?
          if (only.use.sleeplog == TRUE) {
            del = which(nightsummary$cleaningcode > 0 | nightsummary$sleeplog_used == "FALSE"  |
                          nightsummary$guider == "NotWorn" | nightsummary$guider == "NotWorn+invalid")
          } else {
            # only delete nights with no or no valid accelerometer data or when accelerometer not worn, but consider nights with
            # missing sleep log data
            del = which(nightsummary$cleaningcode > 1 | nightsummary$guider == "NotWorn" | 
                          nightsummary$guider == "NotWorn+invalid")
          }
          if (length(del) > 0) {
            nightsummary = nightsummary_bu[-del, ]
          }
          # Ignore nights based on data_cleaning_file if this is provided include_window = rep(TRUE,
          # nrow(x)) allow for forced relying on guider based on external data_cleaning_file
          if (length(data_cleaning_file) > 0) {
            DaCleanFile = data.table::fread(data_cleaning_file, data.table = FALSE)
            if ("night_part4" %in% colnames(DaCleanFile)) {
              days2exclude = which(paste(nightsummary$ID, nightsummary$night) %in% paste(DaCleanFile$ID, DaCleanFile$night_part4))
              if (length(days2exclude) > 0) {
                nightsummary = nightsummary[-days2exclude, ]
              }
            }
          }
          # ignore also all columns related to error (difference between guider and final estimate,
          # which were mainly used for methodological research
          coldel = which(colnames(nightsummary) %in% c("error_onset", "error_wake", "error_dur") == TRUE)
          if (length(coldel) > 0)
            nightsummary = nightsummary[, -coldel]
        }
        NIDS = max(c(length(unique(nightsummary$filename)), length(unique(nightsummary$ID))))
        NDEF = length(unique(nightsummary$sleepparam))
        uuu = unique(nightsummary$sleepparam)
        rem = which(uuu == 0 | uuu == "0" | is.na(uuu) == TRUE)
        if (length(rem) > 0) {
          uuu = uuu[-rem]
          NDEF = length(uuu)
        }
        if (params_output[["storefolderstructure"]] == TRUE) {
          personSummary = matrix(0, NIDS, ((NDEF * 3 * 22) + 15 + (6 * 3)))
        } else {
          personSummary = matrix(0, NIDS, ((NDEF * 3 * 22) + 13 + (6 * 3)))
        }
        # unique filenames, previously we used unique IDs, but that would not allow
        # for repeated measurements of the same ID to be summarised separately
        uniquefn = unique(nightsummary$filename) 
        if (nrow(nightsummary) > 0) {
          for (i in 1:length(uniquefn)) {
            personSummarynames = c()  #moved here on 3/12/2014
            # fully cleaned from nights that need to be deleted
            this_file = which(nightsummary$filename == uniquefn[i])
            nightsummary.tmp = nightsummary[this_file, ]  #back up
            udef = as.character(unique(nightsummary.tmp$sleepparam))
            if (length(which(as.character(udef) == "0") > 0))
              udef = udef[-c(which(as.character(udef) == "0"))]
            udefn = udef
            #-------------------------------------------
            # general info about file
            personSummary[i, 1] = nightsummary.tmp$ID[1]  #id
            personSummarynames = c(personSummarynames, "ID")
            personSummary[i, 2] = uniquefn[i] #as.character(nightsummary$filename[which(nightsummary$ID == uid[i])][1])  #filename
            if (length(unlist(strsplit(as.character(personSummary[i, 2]), ".RDa"))) > 1)
              personSummary[i, 2] = unlist(strsplit(personSummary[i, 2], ".RDa"))[1]
            personSummarynames = c(personSummarynames, "filename")
            cntt = 2
            personSummary[i, cntt + 1] = as.character(nightsummary$calendar_date[this_file[1]])  #date
            personSummarynames = c(personSummarynames, "calendar_date")
            personSummary[i, cntt + 2] = nightsummary$weekday[this_file[1]]  #date
            personSummarynames = c(personSummarynames, "weekday")
            # sleep log used
            personSummary[i, cntt + 3] = as.character(nightsummary.tmp$sleeplog_used[1])
            personSummarynames = c(personSummarynames, paste("sleeplog_used", sep = ""))
            this_sleepparam = which(nightsummary.tmp$sleepparam == udef[1])
            # sleep log ID
            personSummary[i, cntt + 4] = as.character(nightsummary.tmp$sleeplog_ID[1])
            personSummarynames = c(personSummarynames, paste("sleeplog_ID", sep = ""))
            # total number of nights with acceleration and accelerometer worn
            personSummary[i, cntt + 5] = length(which((nightsummary.tmp$acc_available[this_sleepparam] == "TRUE" |
                                                         nightsummary.tmp$acc_available[this_sleepparam] == "1") &
                                                        nightsummary.tmp$cleaningcode[this_sleepparam] != 2))
            personSummarynames = c(personSummarynames, paste("n_nights_acc", sep = ""))
            # total number of nights with sleep log
            n_nights_sleeplog = length(nightsummary.tmp$night[which(nightsummary.tmp$sleepparam == udef[1] &
                                                                      nightsummary.tmp$guider == "sleeplog")])  # number of nights with sleeplog
            personSummary[i, cntt + 6] = n_nights_sleeplog
            personSummarynames = c(personSummarynames, paste("n_nights_sleeplog", sep = ""))
            cntt = cntt + 6
            # total number of nights with marker button as guider
            if (params_sleep[["consider_marker_button"]] == TRUE) {
              personSummary[i, cntt + 1] = length(nightsummary.tmp$night[which(nightsummary.tmp$sleepparam == udef[1] &
                                                    nightsummary.tmp$guider == "markerbutton")])
              personSummarynames = c(personSummarynames, "n_nights_markerbutton")
              cntt = cntt + 1
            }
            # total number of complete weekend and week nights
            th3 = nightsummary.tmp$weekday[this_sleepparam]
            if (only.use.sleeplog == TRUE) {
              validcleaningcode = 0
            } else if (only.use.sleeplog == FALSE) {
              validcleaningcode = 1
            }
            
            personSummary[i, cntt + 1] = length(which(nightsummary.tmp$cleaningcode[this_sleepparam] <= validcleaningcode &
                                                        (th3 == "Friday" | th3 == "Saturday")))
            personSummary[i, cntt + 2] = length(which(nightsummary.tmp$cleaningcode[this_sleepparam] <= validcleaningcode &
                                                        (th3 == "Monday" | th3 == "Tuesday" | th3 == "Wednesday" |
                                                           th3 == "Thursday" | th3 == "Sunday")))
            personSummarynames = c(personSummarynames, paste("n_WE_nights_complete", sep = ""), paste("n_WD_nights_complete",
                                                                                                      sep = ""))
            # number of days with sleep during the day
            personSummary[i, cntt + 3] = length(which(nightsummary.tmp$daysleep[this_sleepparam] == 1 &
                                                        (th3 == "Friday" | th3 == "Saturday")))
            personSummary[i, cntt + 4] = length(which(nightsummary.tmp$daysleep[this_sleepparam] == 1 & 
                                                         (th3 == "Monday" | th3 == "Tuesday" | 
                                                            th3 == "Wednesday" | th3 == "Thursday" |
                                                            th3 == "Sunday")))
            personSummarynames = c(personSummarynames, paste("n_WEnights_daysleeper", sep = ""), paste("n_WDnights_daysleeper",
                                                                                                       sep = ""))
            cnt = cntt + 4
            #-------------------------------------------
            # sleep log summary
            turn_numeric = function(x, varnames) {
              cnx = colnames(x)
              for (i in 1:length(varnames)) {
                if (varnames[i] %in% cnx) {
                  x[, varnames[i]] = as.numeric(x[, varnames[i]])
                }
              }
              return(x)
            }
            if (params_sleep[["sleepwindowType"]] == "SPT" && params_sleep[["consider_marker_button"]] == FALSE) {
              gdn = c("guider_SptDuration", "guider_onset", "guider_wakeup")
            } else if (params_sleep[["sleepwindowType"]] == "TimeInBed" || 
                       params_sleep[["consider_marker_button"]] == TRUE) {
              gdn = c("guider_inbedDuration", "guider_inbedStart", "guider_inbedEnd")
            }
            if (dotwice == 1) {
              nightsummary.tmp = turn_numeric(x = nightsummary.tmp, varnames = gdn)
            }
            # names of numeric columns
            varnames_tmp = c("SptDuration", "sleeponset",
                             "wakeup", "WASO", "SleepDurationInSpt",
                             "number_sib_sleepperiod", "duration_sib_wakinghours",
                             "number_of_awakenings", "number_sib_wakinghours", 
                             "duration_sib_wakinghours_atleast15min",
                             "sleeplatency", "sleepefficiency", "number_of_awakenings",
                             "guider_inbedDuration", "guider_inbedStart",
                             "guider_inbedEnd", "guider_SptDuration", "guider_onset",
                             "guider_wakeup", "SleepRegularityIndex1",
                             "SriFractionValid", "N_nights_guider_corrected")
            nightsummary.tmp = turn_numeric(x = nightsummary.tmp, 
                                            varnames = varnames_tmp)
            weekday = nightsummary.tmp$weekday[this_sleepparam]
            if (dotwice == 1) {
              for (k in 1:3) {
                if (k == 1) {
                  TW = "AD"
                  Seli = 1:length(weekday)
                } else if (k == 2) {
                  TW = "WD"
                  Seli = which(weekday == "Monday" | weekday == "Tuesday" | weekday == "Wednesday" | weekday ==
                                 "Thursday" | weekday == "Sunday")
                } else if (k == 3) {
                  TW = "WE"
                  Seli = which(weekday == "Friday" | weekday == "Saturday")
                }
                relevant_rows = this_sleepparam[Seli]
                if (length(relevant_rows) > 0) {
                  for (gdni in 1:length(gdn)) {
                    personSummary[i, cnt + 1] = mean(nightsummary.tmp[relevant_rows, gdn[gdni]], na.rm = TRUE)
                    personSummary[i, cnt + 2] = sd(nightsummary.tmp[relevant_rows, gdn[gdni]], na.rm = TRUE)
                    personSummarynames = c(personSummarynames, paste(gdn[gdni], "_", TW, "_mn", sep = ""),
                                           paste(gdn[gdni], "_", TW, "_sd", sep = ""))
                    cnt = cnt + 2
                  }
                }
                if ("nonwear_perc_spt" %in% colnames(nightsummary.tmp)) {
                  personSummary[i, cnt + 1] = mean(nightsummary.tmp$nonwear_perc_spt[this_sleepparam[Seli]], na.rm = TRUE)
                  personSummarynames = c(personSummarynames, paste("nonwear_perc_spt_", TW, "_mn", sep = ""))
                  cnt = cnt + 1
                }
                if ("ACC_spt_mg" %in% colnames(nightsummary.tmp)) {
                  personSummary[i, cnt + 1] = mean(nightsummary.tmp$ACC_spt_mg[this_sleepparam[Seli]], na.rm = TRUE)
                  personSummarynames = c(personSummarynames, paste("ACC_spt_mg_", TW, "_mn", sep = ""))
                  cnt = cnt + 1
                }
              }
            }
            nightsummary$cleaningcode = as.numeric(nightsummary$cleaningcode)
            nightsummary$ID = as.character(nightsummary$ID)
            # uid = as.character(uid)
            #-------------------------------------------
            # accelerometer summary
            #----------------------------------------------
            if (only.use.sleeplog == FALSE) {
              # when sleep log is not available
              if (dotwice == 2) {
                CRIT = which(nightsummary$filename == uniquefn[i] &
                               (nightsummary$cleaningcode == 0 | nightsummary$cleaningcode == 1))
              } else {
                CRIT = which(nightsummary$filename == uniquefn[i])
              }
            } else {
              CRIT = which(nightsummary$filename == uniquefn[i] &
                             nightsummary$cleaningcode == 0)  #when sleep log is available
            }
            personSummarynames_backup = c()
            if (length(CRIT) > 0) {
              # summarise data if there is data
              #-----------------------------------------------
              for (j in 1:length(udef)) {
                weekday = nightsummary.tmp$weekday[which(nightsummary.tmp$sleepparam == udef[j])]
                for (k in 1:3) {
                  if (ncol(personSummary) < (cnt + 50)) {
                    # expand personSummary matrix if there is a change that is not big enough
                    expansion = matrix(NA, nrow(personSummary), 50)
                    if (nrow(expansion) != nrow(personSummary))
                      expansion = t(expansion)
                    personSummary = cbind(personSummary, expansion)
                  }
                  if (k == 1) {
                    TW = "AD"
                    Seli = 1:length(weekday)
                  } else if (k == 2) {
                    TW = "WD"
                    Seli = which(weekday == "Monday" | weekday == "Tuesday" | weekday == "Wednesday" | weekday ==
                                   "Thursday" | weekday == "Sunday")
                  } else if (k == 3) {
                    TW = "WE"
                    Seli = which(weekday == "Friday" | weekday == "Saturday")
                  }
                  indexUdef = which(nightsummary.tmp$sleepparam == udef[j])[Seli]
                  personSummary[i, (cnt + 1)] = mean(nightsummary.tmp$SptDuration[indexUdef], na.rm = TRUE)
                  personSummary[i, (cnt + 2)] = sd(nightsummary.tmp$SptDuration[indexUdef], na.rm = TRUE)
                  personSummarynames = c(personSummarynames, paste("SptDuration_", TW, "_", udefn[j], "_mn",
                                                                   sep = ""), paste("SptDuration_", TW, "_", udefn[j], "_sd", sep = ""))
                  personSummary[i, (cnt + 3)] = mean(nightsummary.tmp$SleepDurationInSpt[indexUdef], na.rm = TRUE)
                  personSummary[i, (cnt + 4)] = sd(nightsummary.tmp$SleepDurationInSpt[indexUdef], na.rm = TRUE)
                  personSummarynames = c(personSummarynames, paste("SleepDurationInSpt_", TW, "_", udefn[j],
                                                                   "_mn", sep = ""), paste("SleepDurationInSpt_", TW, "_", udefn[j], "_sd", sep = ""))
                  personSummary[i, (cnt + 5)] = mean(nightsummary.tmp$WASO[indexUdef], na.rm = TRUE)
                  personSummary[i, (cnt + 6)] = sd(nightsummary.tmp$WASO[indexUdef], na.rm = TRUE)
                  personSummarynames = c(personSummarynames, paste("WASO_", TW, "_", udefn[j], "_mn", sep = ""),
                                         paste("WASO_", TW, "_", udefn[j], "_sd", sep = ""))
                  personSummary[i, (cnt + 7)] = mean(nightsummary.tmp$duration_sib_wakinghours[indexUdef],
                                                     na.rm = TRUE)
                  personSummary[i, (cnt + 8)] = sd(nightsummary.tmp$duration_sib_wakinghours[indexUdef],
                                                   na.rm = TRUE)
                  personSummarynames = c(personSummarynames, paste("duration_sib_wakinghours_", TW, "_",
                                                                   udefn[j], "_mn", sep = ""), paste("duration_sib_wakinghours_", TW, "_", udefn[j], "_sd",
                                                                                                     sep = ""))
                  personSummary[i, (cnt + 9)] = mean(nightsummary.tmp$number_sib_sleepperiod[indexUdef],
                                                     na.rm = TRUE)
                  personSummary[i, (cnt + 10)] = sd(nightsummary.tmp$number_sib_sleepperiod[indexUdef],
                                                    na.rm = TRUE)
                  personSummarynames = c(personSummarynames, paste("number_sib_sleepperiod_", TW, "_", udefn[j],
                                                                   "_mn", sep = ""), paste("number_sib_sleepperiod_", TW, "_", udefn[j], "_sd", sep = ""))
                  personSummary[i, (cnt + 11)] = mean(nightsummary.tmp$number_of_awakenings[indexUdef],
                                                      na.rm = TRUE)
                  personSummary[i, (cnt + 12)] = sd(nightsummary.tmp$number_of_awakenings[indexUdef], na.rm = TRUE)
                  personSummarynames = c(personSummarynames, paste("number_of_awakenings_", TW, "_", udefn[j],
                                                                   "_mn", sep = ""), paste("number_of_awakenings_", TW, "_", udefn[j], "_sd", sep = ""))
                  personSummary[i, (cnt + 13)] = mean(nightsummary.tmp$number_sib_wakinghours[indexUdef],
                                                      na.rm = TRUE)
                  personSummary[i, (cnt + 14)] = sd(nightsummary.tmp$number_sib_wakinghours[indexUdef],
                                                    na.rm = TRUE)
                  personSummarynames = c(personSummarynames, paste("number_sib_wakinghours_", TW, "_", udefn[j],
                                                                   "_mn", sep = ""), paste("number_sib_wakinghours_", TW, "_", udefn[j], "_sd", sep = ""))
                  personSummary[i, (cnt + 15)] = mean(nightsummary.tmp$duration_sib_wakinghours_atleast15min[indexUdef],
                                                      na.rm = TRUE)
                  personSummary[i, (cnt + 16)] = sd(nightsummary.tmp$duration_sib_wakinghours_atleast15min[indexUdef],
                                                    na.rm = TRUE)
                  personSummarynames = c(personSummarynames, paste("duration_sib_wakinghours_atleast15min_",
                                                                   TW, "_", udefn[j], "_mn", sep = ""), paste("duration_sib_wakinghours_atleast15min_",
                                                                                                              TW, "_", udefn[j], "_sd", sep = ""))
                  # average sibd during the day
                  AVEsibdDUR = c(nightsummary.tmp$duration_sib_wakinghours[indexUdef]/nightsummary.tmp$number_sib_wakinghours[indexUdef])
                  if (length(which(nightsummary.tmp$number_sib_wakinghours[indexUdef] == 0))) {
                    AVEsibdDUR[which(nightsummary.tmp$number_sib_wakinghours[indexUdef] == 0)] = 0
                  }
                  personSummary[i, (cnt + 17)] = mean(AVEsibdDUR, na.rm = TRUE)
                  personSummary[i, (cnt + 18)] = sd(AVEsibdDUR, na.rm = TRUE)
                  personSummarynames = c(personSummarynames, paste("average_dur_sib_wakinghours_", TW, "_",
                                                                   udefn[j], "_mn", sep = ""), paste("average_dur_sib_wakinghours_", TW, "_", udefn[j],
                                                                                                     "_sd", sep = ""))
                  NDAYsibd = length(which(nightsummary.tmp$number_sib_wakinghours[indexUdef] > 0))
                  if (length(NDAYsibd) == 0) NDAYsibd = 0
                  personSummary[i, (cnt + 19)] = NDAYsibd
                  personSummarynames = c(personSummarynames, paste("n_days_w_sib_wakinghours_", TW, "_",
                                                                   udefn[j], sep = ""))
                  personSummary[i, (cnt + 20)] = mean(nightsummary.tmp$sleeponset[indexUdef], na.rm = TRUE)
                  personSummary[i, (cnt + 21)] = sd(nightsummary.tmp$sleeponset[indexUdef], na.rm = TRUE)
                  personSummarynames = c(personSummarynames, paste("sleeponset_", TW, "_", udefn[j], "_mn",
                                                                   sep = ""), paste("sleeponset_", TW, "_", udefn[j], "_sd", sep = ""))
                  personSummary[i, (cnt + 22)] = mean(nightsummary.tmp$wakeup[indexUdef], na.rm = TRUE)
                  personSummary[i, (cnt + 23)] = sd(nightsummary.tmp$wakeup[indexUdef], na.rm = TRUE)
                  personSummarynames = c(personSummarynames, paste("wakeup_", TW, "_", udefn[j], "_mn",
                                                                   sep = ""), paste("wakeup_", TW, "_", udefn[j], "_sd", sep = ""))
                  personSummary[i, (cnt + 24)] = mean(nightsummary.tmp$SleepRegularityIndex1[indexUdef],
                                                      na.rm = TRUE)
                  personSummary[i, (cnt + 25)] = sd(nightsummary.tmp$SleepRegularityIndex1[indexUdef], na.rm = TRUE)
                  personSummarynames = c(personSummarynames, paste("SleepRegularityIndex1_", TW, "_", udefn[j],
                                                                   "_mn", sep = ""), paste("SleepRegularityIndex1_", TW, "_", udefn[j], "_sd", sep = ""))
                  personSummary[i, (cnt + 26)] = mean(nightsummary.tmp$SriFractionValid[indexUdef], na.rm = TRUE)
                  personSummary[i, (cnt + 27)] = sd(nightsummary.tmp$SriFractionValid[indexUdef], na.rm = TRUE)
                  personSummarynames = c(personSummarynames, paste("SriFractionValid_", TW, "_", udefn[j],
                                                                   "_mn", sep = ""), paste("SriFractionValid_", TW, "_", udefn[j], "_sd", sep = ""))
                  cnt = cnt + 27
                  if ("N_nights_guider_corrected" %in% personSummarynames == FALSE) {
                    personSummary[i, (cnt + 1)] = sum(as.numeric(nightsummary.tmp$guider_corrected[indexUdef]),
                                                      na.rm = TRUE)
                    personSummarynames = c(personSummarynames, "N_nights_guider_corrected")
                    cnt = cnt + 1
                  }
                  if (params_sleep[["sleepwindowType"]] == "TimeInBed" || params_sleep[["consider_marker_button"]] == TRUE) {
                    sleepefficiency = nightsummary.tmp$sleepefficiency[indexUdef]
                    latency = nightsummary.tmp$sleeplatency[indexUdef]
                    if (params_sleep[["sib_must_fully_overlap_with_TimeInBed"]][1] == FALSE) {
                      negative_latency = which(latency < 0)
                      N_neg_lat = length(negative_latency)
                      if (N_neg_lat > 0) latency = latency[-negative_latency]
                      personSummary[i, (cnt + 1)] = N_neg_lat
                      personSummarynames = c(personSummarynames, "N_nights_negative_latency")
                      cnt = cnt + 1
                      Nlatency = length(latency)
                      if (Nlatency > 0) {
                        meanLatency = mean(latency, na.rm = TRUE)
                        meanSleepefficiency = mean(sleepefficiency, na.rm = TRUE)
                      } else {
                        meanLatency = NA
                        meanSleepefficiency = NA
                      }
                      if (Nlatency > 1) {
                        sdLatency = sd(latency, na.rm = TRUE)
                        sdSleepefficiency = sd(sleepefficiency, na.rm = TRUE)
                      } else {
                        sdLatency = NA
                        sdSleepefficiency = NA
                      }
                    } else {
                      meanLatency = mean(latency, na.rm = TRUE)
                      meanSleepefficiency = mean(sleepefficiency, na.rm = TRUE)
                      sdLatency = sd(latency, na.rm = TRUE)
                      sdSleepefficiency = sd(sleepefficiency, na.rm = TRUE)
                    }
                    personSummary[i, (cnt + 1)] = meanSleepefficiency
                    personSummary[i, (cnt + 2)] = sdSleepefficiency
                    personSummarynames = c(personSummarynames, paste("sleep_efficiency_", TW, "_", udefn[j],
                                                                     "_mn", sep = ""), paste("sleep_efficiency_", TW, "_", udefn[j], "_sd", sep = ""))
                    cnt = cnt + 2
                    personSummary[i, (cnt + 1)] = meanLatency
                    personSummary[i, (cnt + 2)] = sdLatency
                    personSummarynames = c(personSummarynames, paste("sleeplatency_", TW, "_", udefn[j],
                                                                     "_mn", sep = ""), paste("sleeplatency_", TW, "_", udefn[j], "_sd", sep = ""))
                    personSummary[i, (cnt + 3)] = mean(nightsummary.tmp$guider_inbedStart[indexUdef], na.rm = TRUE)
                    personSummary[i, (cnt + 4)] = sd(nightsummary.tmp$guider_inbedStart[indexUdef], na.rm = TRUE)
                    personSummarynames = c(personSummarynames, paste("guider_inbedStart_", TW, "_", udefn[j],
                                                                     "_mn", sep = ""), paste("guider_inbedStart_", TW, "_", udefn[j], "_sd", sep = ""))
                    cnt = cnt + 4
                    personSummary[i, (cnt + 1)] = mean(nightsummary.tmp$guider_inbedEnd[indexUdef], na.rm = TRUE)
                    personSummary[i, (cnt + 2)] = sd(nightsummary.tmp$guider_inbedEnd[indexUdef], na.rm = TRUE)
                    personSummarynames = c(personSummarynames, paste("guider_inbedEnd_", TW, "_", udefn[j],
                                                                     "_mn", sep = ""), paste("guider_inbedEnd_", TW, "_", udefn[j], "_sd", sep = ""))
                    personSummary[i, (cnt + 3)] = mean(nightsummary.tmp$guider_inbedDuration[indexUdef],
                                                       na.rm = TRUE)
                    personSummary[i, (cnt + 4)] = sd(nightsummary.tmp$guider_inbedDuration[indexUdef],
                                                     na.rm = TRUE)
                    personSummarynames = c(personSummarynames, paste("guider_inbedDuration_", TW, "_", udefn[j],
                                                                     "_mn", sep = ""), paste("guider_inbedDuration_", TW, "_", udefn[j], "_sd", sep = ""))
                    cnt = cnt + 4
                  }
                }
              }
              personSummary[i, cnt + 1] = as.character(nightsummary$GGIRversion[this_file[1]])
              cnt = cnt + 1
              personSummarynames = c(personSummarynames, "GGIRversion")
              personSummarynames_backup = personSummarynames
            }
          }
          # replace matrix values 'NA' and 'NaN' by empty cells
          for (colli in 1:ncol(personSummary)) {
            missingv = which(is.na(personSummary[, colli]) == TRUE |
                               personSummary[, colli] == "NA" |
                               personSummary[, colli] == "NaN")
            if (length(missingv) > 0) {
              personSummary[missingv, colli] = ""
            }
          }
          personSummary = as.data.frame(personSummary, stringsAsFactors = TRUE)
          if (length(personSummarynames) != ncol(personSummary)) {
            if (length(personSummarynames_backup) > 0) {
              names(personSummary) = personSummarynames_backup
            } else {
              if (length(personSummarynames) > ncol(personSummary)) {
                names(personSummary)[1:length(personSummarynames)] = personSummarynames
              } else {
                names(personSummary) = personSummarynames[1:ncol(personSummary)]
              }
            }
          } else {
            names(personSummary) = personSummarynames
          }
          # remove empty columns in personpersonSummary, if any
          emptycolumns = which(is.na(colnames(personSummary)) == TRUE)
          if (length(emptycolumns) > 0) {
            personSummary = personSummary[, -emptycolumns]
          }
        }
        #----------------------------------------------------
        # Identify whether recordings where split.
        # If yes, then extract names for each plot.
        if (nrow(nightsummary) != 0) {
          personSummary = addSplitNames(personSummary)
          if ("split1_name" %in% colnames(personSummary) == TRUE) {
            nightsummary = addSplitNames(nightsummary)
          }
        }
        #----------------------------------------------------
        if (nrow(nightsummary) == 0) {
          if (verbose == TRUE) {
            if (dotwice == 1) {
              cat("\npart 4 full report not stored, because no results available")
            } else {
              cat("\npart 4 cleaned report not stored, because no results available")
            }
          }
        } else {
          nightsummary_clean = tidyup_df(nightsummary)
          personSummary_clean = tidyup_df(personSummary)
          if (dotwice == 1) {
            csv_name = paste0(resultfolder, "/results/QC/part4_nightsummary_sleep_full.csv")
            data.table::fwrite(nightsummary_clean, file = csv_name, row.names = FALSE,
                               na = "", sep = params_output[["sep_reports"]],
                               dec = params_output[["dec_reports"]])
            csv_name = paste0(resultfolder, "/results/QC/part4_summary_sleep_full.csv")
            data.table::fwrite(personSummary_clean, file = csv_name, row.names = FALSE,
                               na = "", sep = params_output[["sep_reports"]],
                               dec = params_output[["dec_reports"]])
          } else {
            csv_name = paste0(resultfolder, "/results/part4_nightsummary_sleep_cleaned.csv")
            data.table::fwrite(nightsummary_clean, file = csv_name, row.names = FALSE,
                               na = "", sep = params_output[["sep_reports"]],
                               dec = params_output[["dec_reports"]])
            csv_name = paste0(resultfolder, "/results/part4_summary_sleep_cleaned.csv")
            data.table::fwrite(personSummary_clean, file = csv_name, row.names = FALSE,
                               na = "", sep = params_output[["sep_reports"]],
                               dec = params_output[["dec_reports"]])
          }
        }
      }
    }
  }
}

g.part4 = function(datadir = c(), metadatadir = c(), f0 = f0, f1 = f1,
                   params_sleep = c(), params_metrics = c(),
                   params_cleaning = c(), params_output = c(),
                   params_general = c(), verbose = TRUE, ...) {
  #----------------------------------------------------------
  # Extract and check parameters
  input = list(...)
  params = extract_params(params_sleep = params_sleep, params_metrics = params_metrics,
                          params_general = params_general, params_output = params_output,
                          params_cleaning = params_cleaning, input = input,
                          params2check = c("sleep", "metrics",
                                           "general", "output", "cleaning")) # load default parameters
  params_sleep = params$params_sleep
  params_metrics = params$params_metrics
  params_cleaning = params$params_cleaning
  params_output = params$params_output
  params_general = params$params_general
  if (exists("relyonsleeplog") == TRUE & exists("relyonguider") == FALSE) relyonguider = params_sleep[["relyonsleeplog"]]
  # description: function to load sleep detection from g.part3 and to convert it into night-specific summary measures of sleep,
  # possibly aided by sleep log/diary information (if available and provided by end-user)
  nnpp = 40 # number of nights to be displayed in the report (hard-coded not a critical parameter for most scenarios)
  #------------------------------------------------
  ms3.out = "/meta/ms3.out"
  meta.sleep.folder = paste0(metadatadir, ms3.out)
  ms4.out = "/meta/ms4.out"
  checkMilestoneFolders(metadatadir, partNumber = 4)
  #------------------------------------------------
  # Get sleeplog data
  if (length(params_sleep[["loglocation"]]) > 0) {
    dolog = TRUE
  } else {
    dolog = FALSE
  }
  if (dolog == TRUE) {
    sleeplogRDataFile = paste0(metadatadir,"/meta/sleeplog_", basename(params_sleep[["loglocation"]]), ".RData")
    # only re-process sleeplog if sleeplog.RData does not exist 
    # or if sleeplog is from a date equal to or after sleeplog.RData
    # or if sleeplog.Rdata file is older than the news files in milestone 3.
    if (!file.exists(sleeplogRDataFile) || 
        file.info(params_sleep[["loglocation"]])$mtime >= file.info(sleeplogRDataFile)$mtime ||
        max(file.info(dir(meta.sleep.folder, full.names = TRUE))$mtime) >= file.info(sleeplogRDataFile)$mtime) {
      logs_diaries = g.loadlog(params_sleep[["loglocation"]], 
                               coln1 = params_sleep[["coln1"]],
                               colid = params_sleep[["colid"]],
                               meta.sleep.folder = meta.sleep.folder,
                               desiredtz = params_general[["desiredtz"]],
                               sleepwindowType = params_sleep[["sleepwindowType"]])
      
      if (params_sleep[["sleepwindowType"]] == "SPT" && length(logs_diaries$bedlog) > 0 &&
          length(logs_diaries$sleeplog) == 0) {
        stop(paste0("The sleep diary as provided only appears to have time indicators",
                    " for time in bed and not for the Sleep Period Time window, while",
                    " parameter sleepwindowType is set to SPT (default). Either change sleepwindowType to",
                    " \"TimeInBed\" or change your sleep diary column names for sleep timing",
                    " to \"wakeup\" and \"sleeponset\"."), call. = FALSE)
      } else if (params_sleep[["sleepwindowType"]] == "TimeInBed" && length(logs_diaries$bedlog) == 0 &&
                 length(logs_diaries$sleeplog) > 0) {
        stop(paste0("The sleep diary as provided only appears to have time indicators",
                    " for SPT and not for the Time in Bed, while",
                    " parameter sleepwindowType is set to TimeInBed. Either change sleepwindowType to",
                    " \"SPT\" or change your sleep diary column names for sleep timing",
                    " to \"outbed\" and \"inbed\"."), call. = FALSE)
      }
      save(logs_diaries, file = sleeplogRDataFile)
    } else {
      load(file = sleeplogRDataFile)
    }
    if (params_sleep[["sleepwindowType"]] == "TimeInBed" && length(logs_diaries$bedlog) > 0) {
      sleeplog = logs_diaries$bedlog
    } else {
      sleeplog = logs_diaries$sleeplog
    }
    sleeplog$night = as.numeric(sleeplog$night)
    sleeplog$duration = as.numeric(sleeplog$duration)
    if (is.null(logs_diaries$sleeplog) && is.null(logs_diaries$bedlog)) {
      dolog = FALSE
      rm(sleeplog)
    }
  }
  #------------------------------------------------
  # get list of accelerometer milestone data files from sleep (produced by g.part3)
  fnames = dir(meta.sleep.folder)

  if (f1 > length(fnames)) f1 = length(fnames)
  if (f0 > length(fnames)) f0 = 1
  if (f1 == 0 | length(f1) == 0 | f1 > length(fnames))  f1 = length(fnames)
  #-----------------------------------------------------
  # related to plotting
  cnt = 1 #counter to keep track of how many nights have been plotted in the visual report
  idlabels = rep(0,nnpp) # initialize vaiable for plot labels
  pagei = 1
  cnt67 = 1  # counter used to decide whether to create a new pdf file for the plots
  #-----------------------------------------------------
  # initialize output variable names
  colnamesnightsummary = c("ID", "night", "sleeponset", "wakeup", "SptDuration", "sleepparam", "guider_onset",
                           "guider_wakeup", "guider_SptDuration", "error_onset", "error_wake", "error_dur", "fraction_night_invalid",
                           "SleepDurationInSpt", "WASO", "duration_sib_wakinghours",
                           "number_sib_sleepperiod", "number_of_awakenings",
                           "number_sib_wakinghours", "duration_sib_wakinghours_atleast15min",
                           "sleeponset_ts", "wakeup_ts", "guider_onset_ts",  "guider_wakeup_ts",
                           "sleeplatency", "sleepefficiency", "page", "daysleeper", "weekday", "calendar_date",
                           "filename", "cleaningcode", "sleeplog_used", "sleeplog_ID", "acc_available", "guider", "SleepRegularityIndex1", "SriFractionValid",
                           "longitudinal_axis", "guider_corrected")
  
  
  if (params_output[["storefolderstructure"]] == TRUE) {
    colnamesnightsummary = c(colnamesnightsummary, "filename_dir", "foldername")
  }
  # initialize variable to hold sleeplog derived sleep duration if not sleep log was used then the
  # estimates of the sleep period time window will be used instead
  logdur = rep(0, length(fnames))
  # ======================================================================== check which files have
  # already been processed, such that no double work is done
  ffdone = c()
  ms4.out = "/meta/ms4.out"
  fnames.ms4 = dir(paste0(metadatadir, ms4.out))
  # fnames.ms4 = sort(fnames.ms4)
  ffdone = fnames.ms4
  # ffdone a matrix with all the binary filenames that have been processed
  # fnames = sort(fnames)
  #--------------------------------
  # get original file path of the accelerometer (some studies may like to keep track of original folder
  # structure if their study structure is embodied in folder structure)
  if (params_output[["storefolderstructure"]] == TRUE) {
    filelist = FALSE
    if (length(datadir) == 1) {
      # could be a directory or one file
      if (length(unlist(strsplit(datadir, split = "[.](cs|bi|cw)"))) > 1)
        filelist = TRUE
    } else {
      # multiple files
      filelist = TRUE
    }
    if (filelist == FALSE) {
      fnamesfull = dir(datadir, recursive = TRUE, pattern = "[.](csv|bin|cwa)")
    } else {
      fnamesfull = datadir
    }
    f17 = function(X) {
      out = unlist(strsplit(X, "/"))
      f17 = out[(length(out) - 1)]
    }
    ffd = ffp = rep("", length(fnamesfull))
    if (length(fnamesfull) > 0) {
      fnamesshort = basename(fnamesfull)
      foldername = apply(X = as.matrix(fnamesfull), MARGIN = 1, FUN = f17)
      for (i in 1:length(fnames)) {
        #
        ff = as.character(unlist(strsplit(fnames[i], ".RDa"))[1])
        if (length(which(fnamesshort == ff)) > 0) {
          ffd[i] = fnamesfull[which(fnamesshort == ff)]
          ffp[i] = foldername[which(fnamesshort == ff)]
        }
      }
    }
  }
  convertHRsinceprevMN2Clocktime = function(x) {
    # x = hours Since Previous Midnight
    if (x > 24) x = x - 24
    HR = floor(x)
    MI = floor((x - HR) * 60)
    SE = round(((x - HR) - (MI/60)) * 3600)
    if (SE == 60) {
      MI = MI + 1
      SE = 0
    }
    if (MI == 60) {
      HR = HR + 1
      MI = 0
    }
    if (HR == 24) HR = 0
    if (HR < 10) HR = paste0("0", HR)
    if (MI < 10) MI = paste0("0", MI)
    if (SE < 10) SE = paste0("0", SE)
    return(paste0(HR, ":", MI, ":", SE))
  }
  if (length(params_cleaning[["data_cleaning_file"]]) > 0) {
    # allow for forced relying on guider based on external data_cleaning_file
    DaCleanFile = data.table::fread(params_cleaning[["data_cleaning_file"]], data.table = FALSE)
  }
  # =================================================================
  # start of loop through the
  # participants
  for (i in f0:f1) {
    tail_expansion_log = NULL
    # decide whether file was processed before
    if (params_general[["overwrite"]] == TRUE) {
      skip = 0  # this will make that analyses is done regardless of whether it was done before
    } else {
      skip = 0  #do not skip this file
      if (length(ffdone) > 0) {
        if (length(which(ffdone == fnames[i])) > 0)
          skip = 1  #skip this file because it was analysed before')
      }
    }
    if (skip == 0) {
      if (verbose == TRUE) cat(paste0(" ", i))
      addlegend = FALSE
      if (cnt67 == 1) {
        # only create new pdf if there is actually new plots to be generated keep pdf for QC
        # purposes
        if (params_output[["do.visual"]] == TRUE) {
          pdf(file = paste0(metadatadir, "/results/visualisation_sleep.pdf"), width = 8.27, height = 11.69)
          par(mar = c(4, 5, 1, 2) + 0.1)
          plot(c(0, 0), c(1, 1), xlim = c(12, 36), ylim = c(0, nnpp), col = "white", axes = FALSE, xlab = "time",
               ylab = "", main = paste0("Page ", pagei))
          axis(side = 1, at = 12:36, labels = c(12:24, 1:12), cex.axis = 0.7)
          abline(v = c(18, 24, 30), lwd = 0.2, lty = 2)
          abline(v = c(15, 21, 27, 33), lwd = 0.2, lty = 3, col = "grey")
          addlegend = TRUE
        }
        cnt67 = 2
      }
      nightsummary = as.data.frame(matrix(0, 0, length(colnamesnightsummary)))
      colnames(nightsummary) = colnamesnightsummary
      if (params_sleep[["sleepwindowType"]] == "TimeInBed") {
        colnames(nightsummary) = gsub(replacement = "guider_inbedStart", pattern = "guider_onset", x = colnames(nightsummary))
        colnames(nightsummary) = gsub(replacement = "guider_inbedEnd", pattern = "guider_wakeup", x = colnames(nightsummary))
        colnames(nightsummary) = gsub(replacement = "guider_inbedDuration", pattern = "guider_SptDuration",
                                      x = colnames(nightsummary))
      }
      sumi = 1  # counter to keep track of where we are in filling the output matrix 'nightsummary'
      ID  = SPTE_end = SPTE_start = L5list = sib.cla.sum = SPTE_corrected = NULL
      desiredtz_part1 = longitudinal_axis = part3_guider = NULL
      # load milestone 3 data (RData files), check whether there is data, identify id numbers...
      load(paste0(meta.sleep.folder, "/", fnames[i]))
      if (!is.null(desiredtz_part1)) {
        params_general[["desiredtz"]] = desiredtz_part1
      }
      accid = c()
      logid = NA  # keep track of what log id matched to accid
      if (length(ID) > 0) {
        if (!is.na(ID)) {
          # continue with same ID as extracted in GGIR parts 1-3:
          accid = ID
        }
        # if ID not available, function part4_extractid will attempt to extract it from the file name
      }
      if (exists("SRI") == FALSE) SRI = NA
      
      # Ignore night zero being the night before the recording or partially 
      # overlapping with the start of the recording
      # but for which some timestamps are in the sib-report
      sib.cla.sum = sib.cla.sum[which(sib.cla.sum$night != 0),]
      
      if (nrow(sib.cla.sum) != 0) {
        # there needs to be some information
        sib.cla.sum$sib.onset.time = iso8601chartime2POSIX(sib.cla.sum$sib.onset.time, tz = params_general[["desiredtz"]])
        sib.cla.sum$sib.end.time = iso8601chartime2POSIX(sib.cla.sum$sib.end.time, tz = params_general[["desiredtz"]])
        # extract the identifier from accelerometer data and matching indices of sleeplog:
        idwi = g.part4_extractid(params_general[["idloc"]], fname = fnames[i],
                                 dolog,
                                 sleeplog, accid = accid)
        accid = idwi$accid
        wi = idwi$matching_indices_sleeplog
        #-----------------------------------------------------------
        # create overview of night numbers in the data file: nnightlist
        if (dolog == TRUE) {
          logid = sleeplog$ID[wi][1]
          sleeplog_matching_ID = which(sleeplog$ID == accid)
          if (length(sleeplog_matching_ID) == 0) {
            # try without letters
            sleeplog_matching_ID = which(sleeplog$ID == gsub("[^0-9.-]", "", accid))
            if (length(sleeplog_matching_ID) == 0) {
              # try without leading zeros
              sleeplog_matching_ID = which(sleeplog$ID == gsub("^0+", "", accid))
              if (length(sleeplog_matching_ID) == 0) {
                # try interpret as integer that was accidentally stored as decimal 
                # number, e.g. 123.00
                split_by_dot = unlist(strsplit(accid, "[.]"))
                if (length(split_by_dot) == 2) {
                  sleeplog_matching_ID = which(sleeplog$ID == split_by_dot[1])
                }
                if (length(sleeplog_matching_ID) == 0) {
                  # Consider all sleeplog entries
                  sleeplog_matching_ID = 1:length(sleeplog$night)
                  warning(paste0("No matching sleeplog entry found for acceleromete recording ", accid))
                }
              }
            }
          }
          first_night = min(min(sib.cla.sum$night), 
                            min(as.numeric(sleeplog$night[sleeplog_matching_ID])))
          last_night = max(max(sib.cla.sum$night), 
                           max(as.numeric(sleeplog$night[sleeplog_matching_ID])))
        } else {
          first_night = min(sib.cla.sum$night)
          last_night = max(sib.cla.sum$night)
        }
        nnightlist = first_night:last_night
        
        if (length(nnightlist) < length(wi)) {
          nnightlist = nnightlist[1:length(wi)]
        }
        # create overview of which night numbers in the file that have a value and are not equal
        # to zero
        nnights.list = nnightlist
        nnights.list = nnights.list[which(is.na(nnights.list) == FALSE)]
        if (params_cleaning[["excludefirstlast"]] == TRUE & params_cleaning[["excludelast.part4"]] == FALSE & params_cleaning[["excludefirst.part4"]] == FALSE) {
          # exclude first and last night
          if (length(nnights.list) >= 3) {
            nnights.list = nnights.list[2:(length(nnights.list) - 1)]
          } else {
            nnights.list = c()
          }
        } else if (params_cleaning[["excludelast.part4"]] == FALSE & params_cleaning[["excludefirst.part4"]] == TRUE) {
          if (length(nnights.list) >= 2) {
            nnights.list = nnights.list[2:length(nnights.list)]
          } else {
            nnights.list = c()
          }
        } else if (params_cleaning[["excludelast.part4"]] == TRUE & params_cleaning[["excludefirst.part4"]] == FALSE) {
          if (length(nnights.list) >= 2) {
            nnights.list = nnights.list[1:(length(nnights.list) - 1)]
          } else {
            nnights.list = c()
          }
        }
        # initialize variables calendardate and daysleeper from the nnights.list variable
        calendar_date = wdayname = rep("", length(nnights.list))
        # daysleeper variable to keep track of whether the person woke up after noon or a certain
        # day
        daysleeper = rep(FALSE, length(nnights.list))
        ###########################################################
        nightj = 1
        # max(c(max(nnights.list),
        guider.df = data.frame(matrix(NA, length(nnights.list), 5), stringsAsFactors = FALSE)
        names(guider.df) = c("ID", "night", "duration", "sleeponset", "sleepwake")
        guider.df$night = nnights.list
        if (dolog == TRUE) {
          if (length(wi) > 0) {
            wi2 = wi[which(sleeplog$night[wi] %in% guider.df$night)]
            guider.df[which(guider.df$night %in% sleeplog$night[wi2]),] = sleeplog[wi2,]
          }
        }
        for (j in nnights.list) {
          # go through the nights get default onset and wake (based on sleeplog or on heuristic
          # algorithms) def.noc.sleep is an input argument the GGIR user can use to specify what
          # detection strategy is used in the absense of a sleep diary
          if ((length(params_sleep[["def.noc.sleep"]]) == 0 ||
               length(SPTE_start) == 0 ||
               length(SPTE_start[which(is.na(SPTE_start) == FALSE)]) == 0) &&
              length(params_sleep[["def.noc.sleep"]]) != 2) {
            # use L5+/-6hr algorithm if SPTE fails OR if the user explicitely asks for it (length
            # zero argument)
            guider = "notavailable"
            if (length(L5list) > 0 & length(L5list) >= j) {
              defaultGuiderOnset = L5list[j] - 6
              defaultGuiderWake = L5list[j] + 6
              guider = "L512"
            } else {
              # This should never happen, but just as a final backup
              defaultGuiderOnset = 21
              defaultGuiderWake = 31
            }
            defaultGuider = guider
            defaultGuiderCorrected = NULL
          } else if ((length(params_sleep[["def.noc.sleep"]]) == 1 ||
                      length(params_sleep[["loglocation"]]) != 0) &&
                     length(SPTE_start) != 0) {
            # use SPTE algorithm (inside the g.sib.det function) as backup for sleeplog OR if user
            # explicitely asks for it
            defaultGuiderOnset = SPTE_start[j]
            defaultGuiderWake = SPTE_end[j]
            defaultGuiderCorrected = SPTE_corrected[j]
            defaultGuider = part3_guider[j] # HDCZA, NotWorn, HorAngle (or plus invalid)
            if (is.null(defaultGuider)) {
              # this ensures compatibility with previous versions in which part3_guider was not stored
              # for newer version we expect defaultGuider to always not be NULL
              guider = params_sleep[["HASPT.algo"]][1]
              defaultGuider = guider
            } else {
              if (is.na(defaultGuider)) { 
                # No default guider available, for example when sleeplog is 
                # available but no accelerometer
                # In that case guider will be set to "sleeplog" later on
              } else {
                guider = defaultGuider
              }
            }
            if (is.na(defaultGuiderOnset) == TRUE) {
              # If SPTE was not derived for this night, use average estimate for other nights
              availableestimate = which(is.na(SPTE_start) == FALSE)
              cleaningcode = 6
              if (length(availableestimate) > 0) {
                defaultGuiderOnset = mean(SPTE_start[availableestimate])
                guider = defaultGuider = names(sort(table(part3_guider[availableestimate]), decreasing = TRUE)[1])
              } else {
                defaultGuiderOnset = L5list[j] - 6
                guider = "L512"
              }
            }
            if (is.na(defaultGuiderWake) == TRUE) {
              # If SPTE was not derived for this night, use average estimate for other nights
              availableestimate = which(is.na(SPTE_end) == FALSE)
              cleaningcode = 6
              if (length(availableestimate) > 0) {
                defaultGuiderWake = mean(SPTE_end[availableestimate])
                guider = defaultGuider = names(sort(table(part3_guider[availableestimate]), decreasing = TRUE)[1])
              } else {
                defaultGuiderWake = L5list[j] + 6
                guider = "L512"
              }
            }
          } else if (length(params_sleep[["def.noc.sleep"]]) == 2) {
            # use constant onset and waking time as specified with def.noc.sleep argument
            defaultGuiderOnset = params_sleep[["def.noc.sleep"]][1]  #onset
            defaultGuiderWake = params_sleep[["def.noc.sleep"]][2]  #wake
            defaultGuider = guider = "setwindow"
            defaultGuiderCorrected = NULL
          }
          if (defaultGuiderOnset >= 24) {
            defaultGuiderOnset = defaultGuiderOnset - 24
          }
          if (defaultGuiderWake >= 24) {
            defaultGuiderWake = defaultGuiderWake - 24
          }
          defaultdur = defaultGuiderWake - defaultGuiderOnset  #default sleep duration based on sleeplog, L5+/-6hr, or HDCZA algorithm
          sleeplog_used = FALSE
          if (dolog == TRUE) {
            #-----------------------------------------------------------
            # If sleep log is available and values are not missing
            if (all(!is.na(guider.df[which(guider.df$night == j), 4:5]))) {
              sleeplog_used = TRUE
            }
          }
          if (sleeplog_used == FALSE) {
            #-----------------------------------------------------------
            # If sleep log is not available available, use default values calculated above (with
            # the heuristic algorithm HDCZA or if that fails L5+/-6hr.
            guider.df[which(guider.df$night == j), 1:5] = c(accid, j, defaultdur, convertHRsinceprevMN2Clocktime(defaultGuiderOnset),
                                                            convertHRsinceprevMN2Clocktime(defaultGuiderWake))
            cleaningcode = 1
          }
          nightj = nightj + 1
          # keep track of whether enough accelerometer data is available for each night
          acc_available = TRUE  #default assumption
          # initialize dataframe to hold sleep period overview:
          spocum = data.frame(nb = numeric(0), start = numeric(0),  end = numeric(0),
                              overlapGuider = numeric(0), def = character(0))
          
          spocumi = 1  # counter for sleep periods
          # continue now with the specific data of the night
          
          guider.df2 = guider.df[which(guider.df$night == j), ]
          # ================================================================================ get
          # sleeplog (or HDCZA or L5+/-6hr algorithm) onset and waking time and assess whether it
          # is a nightworker onset
          tmp1 = format(guider.df2$sleeponset[1])
          tmp2 = unlist(strsplit(tmp1, ":"))
          GuiderOnset = sum(as.numeric(tmp2) / c(1, 60, 3600))
          # wake
          tmp4 = format(guider.df2$sleepwake[1])
          tmp5 = unlist(strsplit(tmp4, ":"))
          GuiderWake = sum(as.numeric(tmp5) / c(1, 60, 3600))
          # Assess whether it is a daysleeper or a nightsleeper
          daysleeper[j] = FALSE  # default
          if (is.na(GuiderOnset) == FALSE & is.na(GuiderWake) == FALSE & tmp1 != "" & tmp4 != "") {
            # is the sleep log valid?  transform possible Single Digit clock times to Double
            # Digits: hours, minutes and seconds
            doubleDigitClocktime = function(x) {
              x = unlist(strsplit(x, ":"))
              xHR = as.numeric(x[1])
              xMI = as.numeric(x[2])
              xSE = as.numeric(x[3])
              if (xHR < 10) xHR = paste0("0", xHR)
              if (xMI < 10) xMI = paste0("0", xMI)
              if (xSE < 10) xSE = paste0("0", xSE)
              x = paste0(xHR, ":", xMI, ":", xSE)
              return(x)
            }
            tmp1 = doubleDigitClocktime(tmp1)
            tmp4 = doubleDigitClocktime(tmp4)
            #------------------------------------------------------------------
            # does sleep period overlap with noon? If yes, then classify as daysleeper
            if (GuiderWake > 12 & GuiderOnset < 12) daysleeper[j] = TRUE
            if (GuiderWake > 12 & GuiderOnset > GuiderWake) daysleeper[j] = TRUE
            # change time stamps to be a continues time
            if (GuiderOnset < 12) GuiderOnset = GuiderOnset + 24  #shift 24 hours to create continues time
            if (GuiderWake <= 12) GuiderWake = GuiderWake + 24  #shift 24 hours to create continues time
            if (GuiderWake > 12 & GuiderWake < 18 & daysleeper[j] == TRUE) GuiderWake = GuiderWake + 24  # NEW 10/5/2018 by Vincent
            if (daysleeper[j] == TRUE) {
              logdur[i] = GuiderOnset - GuiderWake
            } else {
              logdur[i] = GuiderWake - GuiderOnset
            }
            if (sleeplog_used == TRUE) {
              cleaningcode = 0
              guider = "sleeplog"
            }
          } else {
            GuiderOnset = defaultGuiderOnset  #use default assumption about onset
            GuiderWake = defaultGuiderWake + 24  #use default assumption about wake
            logdur[i] = GuiderWake - GuiderOnset
            cleaningcode = 1  # no diary available for this night, so fall back on detaults
            sleeplog_used = FALSE
          }
          #-----------------------------------------
          # plan analysis according to knowledge about whether it is a daysleeper or not if you
          # are not excluding the last day and is a daysleeper and not the last night
          if (params_cleaning[["excludefirstlast"]] == FALSE) {
            if (daysleeper[j] == TRUE & j != max(nnights.list)) {
              loaddays = 2
            } else {
              loaddays = 1
            }
            if (daysleeper[j] == TRUE & j == max(nnights.list)) {
              # and is a daysleeper and the last night
              daysleeper[j] = FALSE  #treat it like a normal day
              loaddays = 1
              if (GuiderWake > 36)  GuiderWake = 36  #round diary to noon
              logdur[i] = GuiderWake - GuiderOnset
            }
          } else {
            # if you are excluding the last day dont worrry about whether it is the last day,
            # because it is not included
            if (daysleeper[j] == TRUE) {
              loaddays = 2
            } else {
              loaddays = 1
            }
          }
          # now generate empty overview for this night / person
          dummyspo = data.frame(nb = numeric(1), start = numeric(1),  end = numeric(1),
                                overlapGuider = numeric(1), def = character(1), duration = numeric(1))
          dummyspo$nb[1] = 1
          spo_day = c()
          spo_day_exists = FALSE
          # ============================================================================================
          # Loop through sleep definitions
          defs = unique(sib.cla.sum$definition)  # definition of sleep episode metric (see van Hees 2015 PLoSONE paper)
          for (k in defs) {
            # load twice if daysleeper because we also need data from the afternoon on the next day
            for (loaddaysi in 1:loaddays) {
              qq = sib.cla.sum
              sleepdet = qq[which(qq$night == (j + (loaddaysi - 1))), ]
              if (nrow(sleepdet) == 0) {
                if (spocumi == 1) {
                  spocum = dummyspo
                } else {
                  spocum = rbind(spocum, dummyspo)
                }
                spocumi = spocumi + 1
                cleaningcode = 3
                acc_available = FALSE
              } else if (length(tail_expansion_log) != 0 & sumi == max(nnightlist)) {
                # so that we skip the visualization of last night when data has been expanded
                # also, cleaningcode is more realistic as acc data is not available for this night
                cleaningcode = 3
                acc_available = FALSE
              } else {
                acc_available = TRUE
              }
              
              if (nrow(sleepdet) == 0) next
              ki = which(sleepdet$definition == k)
              if (length(ki) == 0) next
              sleepdet.t = sleepdet[ki, ]
              if (loaddaysi == 1) remember_fraction_invalid_day1 = sleepdet.t$fraction.night.invalid[1]
              # now get sleep periods
              nsp = length(unique(sleepdet.t$sib.period))  #number of sleep periods
              spo = data.frame(nb = numeric(nsp), start = numeric(nsp),  end = numeric(nsp),
                               overlapGuider = numeric(nsp), def = character(nsp))
              if (nsp <= 1 & unique(sleepdet.t$sib.period)[1] == 0) {
                # no sleep periods
                spo$nb[1] = 1
                spo[1, c("start", "end", "overlapGuider")] = 0
                spo$def[1] = k
                if (daysleeper[j] == TRUE) {
                  tmpCmd = paste0("spo_day", k, "= c()")
                  eval(parse(text = tmpCmd))  ##
                  spo_day_exists = TRUE
                }
              } else {
                DD = g.create.sp.mat(nsp, spo, sleepdet.t, daysleep = daysleeper[j])
                if (loaddaysi == 1) {
                  wdayname[j] = DD$wdayname
                  calendar_date[j] = DD$calendar_date
                }
                spo = DD$spo
                if (daysleeper[j] == TRUE) {
                  if (loaddaysi == 1) {
                    w1 = which(spo$end >= 18)  #only use periods ending after 6pm
                    if (length(w1) > 0) {
                      spo = spo[w1, ]
                      if (nrow(spo) == 1) {
                        if (spo$start[1] <= 18) spo$start[1] = 18  #turn start time on 1st day before 6pm to 6pm
                      } else {
                        spo$start[which(spo$start <= 18)] = 18  #turn start times on 1st day before 6pm to 6pm
                      }
                      tmpCmd = paste0("spo_day", k, "= spo")  #spo needs to be rememered specific to definition
                      eval(parse(text = tmpCmd))
                      spo_day_exists = TRUE
                    } else {
                      tmpCmd = paste0("spo_day", k, "= c()")
                      eval(parse(text = tmpCmd))
                      spo_day_exists = TRUE
                    }
                  } else if (loaddaysi == 2 & spo_day_exists == TRUE) {
                    # length check added because day may have been skipped
                    w2 = which(spo$start < 18)  #only use periods starting before 6pm
                    if (length(w2) > 0) {
                      spo = spo[w2, ]
                      if (ncol(spo) == 1) spo = t(spo)
                      if (nrow(spo) == 1) {
                        if (spo$end[1] > 18) spo$end[1] = 18  #turn end time on 2nd day after 6pm to 6pm
                      } else {
                        spo$end[which(spo$end > 18)] = 18  #turn end times on 2nd day after 6pm to 6pm
                      }
                      spo[, c("start", "end")] = spo[, c("start", "end")] + 24  # + 24 to create continues timelines for day 2 relative to day 1
                      tmpCmd = paste0("spo_day2", k, "= spo")  #spo needs to be rememered specific to definition
                      eval(parse(text = tmpCmd))
                    } else {
                      tmpCmd = paste0("spo_day2", k, "= c()")
                      eval(parse(text = tmpCmd))
                    }
                    # attach days together as being one day
                    name1 = paste0("spo_day", k)
                    name2 = paste0("spo_day2", k)
                    tmpCmd = paste0("spo = rbind(", name1, ",", name2, ")")
                    eval(parse(text = tmpCmd))
                  }
                }
                if (daysleeper[j] == TRUE) {
                  if (GuiderWake < 21 & GuiderWake > 12 & GuiderOnset > GuiderWake) {
                    # waking up in the afternoon should have value above 36
                    GuiderWake = GuiderWake + 24
                  }
                }
                # Detect whether none of the SIBs overlaps with the SPT, if this is the case then
                # probably the night is not going to be very useful for sleep analysis however,
                # for part 5 we may still want to have some estimate to define the waking-up to
                # waking windows.  So, we will now add two tiny artificial sibs (1-minute) at the
                # beginning and end of the guider-based SPT to make that the SPT is still detected
                # guided entirely by the guider.  and add cleaningcode = 5, this means that the
                # night will be omitted from the part 4 cleaned results, but is still available
                # for inspection in the full part 4 results and available for part 5, because in
                # part 5 we are only interested in the edges of the SPT and not what happens in
                # it.
                relyonguider_thisnight = FALSE
                # user specified to rely on guider in data cleaning file:
                # relyonguider_thisnight = TRUE and cleaningcode = 5
                if (length(params_cleaning[["data_cleaning_file"]]) > 0) {
                  if (length(which(DaCleanFile$relyonguider_part4 == j &
                                   DaCleanFile$ID == accid)) > 0) {
                    relyonguider_thisnight = TRUE
                    cleaningcode = 5 # user specified to rely on guider
                  }
                }
                # No SIBs overlap with SPT window
                # relyonguider_thisnight = TRUE and cleaningcode = 5
                if (length(spo) == 0) {
                  # add empty spo object, in case it was removed above
                  # we do this because code below assumes that spo is a matrix
                  spo = data.frame(nb = numeric(1), start = numeric(1),  end = numeric(1),
                                   overlapGuider = numeric(1), def = character(1))
                  spo$nb[1] = 1
                  spo[1, 2:4] = 0
                  spo$def[1] = k
                }
                if (length(which(spo$start < GuiderWake &
                                 spo$end > GuiderOnset)) == 0) {
                  relyonguider_thisnight = TRUE
                  cleaningcode = 5 
                }
                # if invalid time was used in part3 with HASPT.ignore.invalid,
                # then rely on guider, but cleaningcode should not be 5
                if (grepl("+invalid", guider) | grepl("+invalid", defaultGuider)) {
                  relyonguider_thisnight = TRUE # rely on guider because some nonwear was used to help the slep identification
                }
                # If no SIBs overlap with the guider window
                if (relyonguider_thisnight == TRUE) {
                  if (guider != "NotWorn") {
                    newlines = rbind(spo[1, ], spo[1, ])
                    newlines[1, 1:4] = c(nrow(spo) + 1, GuiderOnset, GuiderOnset + 1/60, 1)
                    newlines[2, 1:4] = c(nrow(spo) + 1, GuiderWake - 1/60, GuiderWake, 1)
                    spo = rbind(spo, newlines) # When NotWorn was used then fully trust on guider and ignore sibs detected
                  } else {
                    newlines = spo[1, ] # initialise object
                    newlines[1, 1:4] = c(nrow(spo) + 1, GuiderOnset, GuiderWake, 1)
                    spo = newlines
                  }
                  spo = spo[order(spo$start), ]
                  spo$nb = 1:nrow(spo)
                  relyonguider_thisnight = TRUE
                }
                # spo is now a data.frame of start and end for each sib (sustained inactivity bout)
                # Classify as being part of the guider window or not
                for (evi in 1:nrow(spo)) {
                  if (spo$start[evi] < GuiderWake && spo$end[evi] > GuiderOnset) {
                    if ((params_sleep[["sleepwindowType"]] == "TimeInBed" || guider == "markerbutton") &&
                        sum(params_sleep[["sib_must_fully_overlap_with_TimeInBed"]]) != 0) {
                      if (all(params_sleep[["sib_must_fully_overlap_with_TimeInBed"]]) &&
                          spo$start[evi] > GuiderOnset && spo$end[evi] < GuiderWake) {
                        spo$overlapGuider[evi] = 1  # Consider only full overlap
                      }
                      if (all(params_sleep[["sib_must_fully_overlap_with_TimeInBed"]] == c(FALSE, TRUE)) &&
                          spo$end[evi] < GuiderWake) {
                        spo$overlapGuider[evi] = 1 # Consider only full overlap
                      }
                      if (all(params_sleep[["sib_must_fully_overlap_with_TimeInBed"]] == c(TRUE, FALSE)) &&
                          spo$start[evi] > GuiderOnset) {
                        spo$overlapGuider[evi] = 1  # Consider only full overlap
                      }
                    } else {
                      spo$overlapGuider[evi] = 1  # Consider partial overlap
                    }
                    if (params_sleep[["relyonguider"]] == TRUE | relyonguider_thisnight == TRUE) {
                      # Redefine sib start and end if it overlaps with guider
                      # to match guider 
                      if ((spo$start[evi] < GuiderWake && spo$end[evi] > GuiderWake) |
                          (spo$start[evi] < GuiderWake && spo$end[evi] < spo$start[evi])) {
                        spo$end[evi] = GuiderWake
                      }
                      if ((spo$start[evi] < GuiderOnset && spo$end[evi] > GuiderOnset) |
                          (spo$end[evi] > GuiderOnset && spo$end[evi] < spo$start[evi])) {
                        spo$start[evi] = GuiderOnset
                      }
                    }
                  }
                }
                spo$duration = spo$end - spo$start
                if (daysleeper[j] == TRUE) {
                  # for the labelling above it was needed to have times > 36, but for the plotting
                  # time in the second day needs to be returned to a normal 24 hour scale.
                  reversetime2 = which(spo$start >= 36)
                  reversetime3 = which(spo$end >= 36)
                  if (length(reversetime2) > 0) spo$start[reversetime2] = spo$start[reversetime2] - 24
                  if (length(reversetime3) > 0) spo$end[reversetime3] = spo$end[reversetime3] - 24
                }
              }
            }
            #------------------------------------------------------------------------
            # Object spo contains all the sib for one sib defintion
            # Object spocum contains all the sib for multiple sib definitions
            if (exists("spo")) {
              spo$def = k
              if (spocumi == 1) {
                spocum = spo
              } else {
                spocum = rbind(spocum, spo)
              }
              spocumi = spocumi + 1
            }
          }
          #------------------------------------------------------------------------
          # Take variables 'spocum', GuiderOnset and GuiderWake to derive nightsummary measures and to
          # create a plot for the current night in the current participant
          #------------------------------------------------------------------------
          # PLOTTING related
          if (params_output[["do.visual"]] == TRUE) {
            if (cnt == (nnpp + 1)) {
              if (verbose == TRUE) cat(" NEW ")
              pagei = pagei + 1
              # add y-axis before starting new page
              if (length(idlabels) < nnpp) {
                idlabels = c(idlabels, rep(" ", length(idlabels) - nnpp))
              } else if (length(idlabels) > nnpp) {
                idlabels = idlabels[1:nnpp]
              }
              axis(side = 2, at = 1:nnpp, labels = idlabels, las = 1, cex.axis = 0.6)
              idlabels = rep(0, nnpp)
              plot(c(0, 0), c(1, 1), xlim = c(12, 36), ylim = c(0, nnpp), col = "white", axes = FALSE,
                   xlab = "time", ylab = "", main = paste0("Page ", pagei))
              axis(side = 1, at = 12:36, labels = c(12:24, 1:12), cex.axis = 0.7)
              abline(v = c(18, 24, 30), lwd = 0.2, lty = 2)
              abline(v = c(15, 21, 27, 33), lwd = 0.2, lty = 3, col = "grey")
              cnt = 1
              addlegend = TRUE
            }
          }
          if (length(spocum) > 0) {
            NAvalues = which(is.na(spocum$def) == TRUE)
            if (length(NAvalues) > 0) {
              spocum = spocum[-NAvalues, ]
            }
          }
          # Fill matrix 'nightsummary' with key sleep parameters
          if (length(spocum) > 0 & class(spocum)[1] == "data.frame" & length(calendar_date) >= j) {
            if (nrow(spocum) > 0 & ncol(spocum) >= 5 & calendar_date[j] != "") {
              undef = unique(spocum$def)
              undef = undef[undef != ""]
              for (defi in undef) {
                #------------------------------------------------------------------------
                # nightsummary
                rowswithdefi = which(spocum$def == defi)
                if (length(rowswithdefi) > 0) {
                  # only process day if there are at least 2 sustained inactivity bouts
                  spocum.t = spocum[rowswithdefi, ]
                  evin = 2
                  if (guider == "NotWorn") {
                    while (evin <= nrow(spocum.t)) {
                      if (spocum.t$start[evin] - spocum.t$start[evin - 1] < -2) {
                        spocum.t$start[evin] = spocum.t$start[evin] + 24
                      }
                      if (spocum.t$end[evin] - spocum.t$end[evin - 1] < -2) {
                        spocum.t$end[evin] = spocum.t$end[evin] + 24
                      }
                      evin = evin + 1
                    }
                  }
                  # in DST it can be that a double hour is not recognized as part of the SPT
                  correct01010pattern = function(x) {
                    x = as.numeric(x)
                    if (length(which(diff(x) == 1)) > 1) {
                      minone = which(diff(x) == -1) + 1
                      plusone = which(diff(x) == 1)
                      matchingvalue = which(minone %in% plusone == TRUE)
                      if (length(matchingvalue) > 0) x[minone[matchingvalue]] = 1
                    }
                    return(x)
                  }
                  delta_t1 = diff(as.numeric(spocum.t$end))
                  spocum.t$overlapGuider = correct01010pattern(spocum.t$overlapGuider)
                  #----------------------------
                  nightsummary[sumi, 1] = accid
                  nightsummary[sumi, 2] = j  #night
                  # remove double rows
                  spocum.t = spocum.t[!duplicated(spocum.t), ]
                  
                  #------------------------------------
                  # ACCELEROMETER
                  if (length(which(as.numeric(spocum.t$overlapGuider) == 1)) > 0) {
                    rtl = which(spocum.t$overlapGuider == 1)
                    nightsummary[sumi, 3] = spocum.t$start[rtl[1]]
                    nightsummary[sumi, 4] = spocum.t$end[rtl[length(rtl)]]
                  } else {
                    cleaningcode = 5  # only for first day, other cleaningcode is assigned to wrong day
                    nightsummary[sumi, 3] = GuiderOnset  #use default assumption about onset
                    nightsummary[sumi, 4] = GuiderWake  #use default assumption about wake
                  }
                  nightsummary[, 3] = as.numeric(nightsummary[, 3])  # onset
                  nightsummary[, 4] = as.numeric(nightsummary[, 4])  # wake
                  # onset after wake is impossible and even more impossible if wake occurs
                  # before none, while we previously labelled it as daysleep
                  if (nightsummary[sumi, 3] > nightsummary[sumi, 4] &
                      nightsummary[sumi, 4] < 36 & daysleeper[j] == TRUE) {
                    nightsummary[sumi, 4] = nightsummary[sumi, 4] + 24  # correction for overcorrection in waking time
                  }
                  if (nightsummary[sumi, 3] == nightsummary[sumi, 4] & nightsummary[sumi, 4] == 18) {
                    # sleeping from 6pm to 6pm (probably non-wear)
                    nightsummary[sumi, 4] = nightsummary[sumi, 4] + 24
                  }
                  nightsummary[sumi, 5] = nightsummary[sumi, 4] - nightsummary[sumi, 3]  #duration Spt
                  nightsummary[, 5] = as.numeric(nightsummary[, 5])
                  nightsummary[sumi, 6] = defi  #sleep definition
                  #------------------------------------
                  # Correct GuiderOnset and GuiderWake to fall within [12-36] window and
                  # store as guider_onset and guider_wake, except when it is a daysleeper
                  # because for daysleeper we expect the window to run beyond time 36 (noon next day)
                  if (GuiderOnset > 36) {
                    nightsummary[sumi, 7] = GuiderOnset - 24
                  } else {
                    nightsummary[sumi, 7] = GuiderOnset
                  }
                  if (GuiderWake > 36 & daysleeper[j] == FALSE) {
                    nightsummary[sumi, 8] = GuiderWake - 24
                  } else {
                    nightsummary[sumi, 8] = GuiderWake
                  }
                  if (nightsummary[sumi, 7] > nightsummary[sumi, 8]) {
                    nightsummary[sumi, 9] = abs((36 - nightsummary[sumi, 7]) + (nightsummary[sumi,
                                                                                             8] - 12))
                  } else {
                    nightsummary[sumi, 9] = abs(nightsummary[sumi, 8] - nightsummary[sumi, 7])
                  }
                  #------------------------------------
                  # Calculate differences between final estimate and guider
                  # In the output we call this error
                  nightsummary[sumi, 10] = nightsummary[sumi, 3] - nightsummary[sumi, 7]  #error onset
                  nightsummary[sumi, 11] = nightsummary[sumi, 4] - nightsummary[sumi, 8]  #error wake
                  # sometimes difference calculations (error) can be in the wrong direction,
                  # e.g. log = 11, acc is 35
                  if (nightsummary[sumi, 10] > 12) nightsummary[sumi, 10] = -(24 - nightsummary[sumi, 10])
                  if (nightsummary[sumi, 10] < -12) nightsummary[sumi, 10] = -(nightsummary[sumi, 10] + 24)
                  if (nightsummary[sumi, 11] > 12) nightsummary[sumi, 11] = -(24 - nightsummary[sumi, 11])
                  if (nightsummary[sumi, 11] < -12) nightsummary[sumi, 11] = -(nightsummary[sumi, 11] + 24)
                  nightsummary[sumi, 12] = nightsummary[sumi, 5] - nightsummary[sumi, 9]  #error duration
                  #------------------------------------
                  # Other variables
                  if (acc_available == TRUE) {
                    nightsummary[sumi, 13] = remember_fraction_invalid_day1  #sleepdet.t$fraction.night.invalid[1]
                    if (remember_fraction_invalid_day1 > ((24 - params_cleaning[["includenightcrit"]])/24)) {
                      cleaningcode = 2  # only for first day, other cleaningcode is assigned to wrong day
                    }
                  } else {
                    nightsummary[sumi, 13] = 1
                  }
                  # Accumulated sustained inactivity bouts during SPT (nocturnal) and dyatime
                  overlap = which(spocum.t$overlapGuider == 1)
                  nocs = spocum.t$duration[overlap]
                  no_overlap = which(spocum.t$overlapGuider == 0)
                  sibds = spocum.t$duration[no_overlap]
                  # it is possible that nocs is negative if when sleep episode starts before dst
                  # in the autumn and ends inside the dst hour
                  negval = which(nocs < 0)
                  if (length(negval) > 0) {
                    kk0 = as.numeric(spocum.t$start[which(spocum.t$overlapGuider == 1)])  # episode onsets
                    kk1 = as.numeric(spocum.t$end[which(spocum.t$overlapGuider == 1)])  # episode endings
                    kk1[negval] = kk1[negval] + 1
                    nocs = kk1 - kk0
                  }
                  if (length(nocs) > 0) {
                    spocum.t.dur.noc = sum(nocs)
                  } else {
                    spocum.t.dur.noc = 0
                  }
                  # ======================================================================================================
                  # check whether it is day saving time (DST) the next day (= the night
                  # connected to the present day)
                  is_this_a_dst_night_output = is_this_a_dst_night(calendar_date = calendar_date[j],
                                                                   tz = params_general[["desiredtz"]])
                  dst_night_or_not = is_this_a_dst_night_output$dst_night_or_not
                  dsthour = is_this_a_dst_night_output$dsthour
                  # if yes, then check whether any of the sleep episodes overlaps dst in spring,
                  # one hour skipped
                  if (dst_night_or_not == 1) {
                    checkoverlap = spocum.t[which(spocum.t$overlapGuider == 1), c("start", "end")]
                    if (nrow(checkoverlap) > 0) {
                      overlaps = which(checkoverlap[, 1] <= (dsthour + 24) & checkoverlap[, 2] >=
                                         (dsthour + 25))
                    } else {
                      overlaps = c()
                    }
                    if (length(overlaps) > 0) {
                      # if yes, then reduce the length of those sleep episodes
                      spocum.t.dur.noc = spocum.t.dur.noc - 1
                      nightsummary[sumi, 5] = nightsummary[sumi, 5] - 1
                      nightsummary[sumi, 9] = nightsummary[sumi, 9] - 1
                    }
                  } else if (dst_night_or_not == -1) {
                    # dst in autumn, one double hour spocum.t.dur.noc has been calculated
                    # correctly including the double hour However, time elapse between onset and
                    # wake needs to be expanded by 1 if onset was before dst and waking up was
                    # after dst.
                    if (nightsummary[sumi, 3] <= (dsthour + 24) &
                        nightsummary[sumi, 4] >= (dsthour + 25)) {
                      nightsummary[sumi, 5] = nightsummary[sumi, 5] + 1  # accelerometer derived sleep duration
                    }
                    if (nightsummary[sumi, 7] <= (dsthour + 24) &
                        nightsummary[sumi, 8] >= (dsthour + 25)) {
                      nightsummary[sumi, 9] = nightsummary[sumi, 9] + 1  # sleep log sleepduration
                    }
                    # does SPT end within double hour?
                    correctSptEdgingInDoubleHour = function(nightsummary, onsetcol, wakecol, durcol,
                                                            dsthour, delta_t1) {
                      wakeInDoubleHour = nightsummary[, wakecol] >= (dsthour + 24) & nightsummary[,
                                                                                                  wakecol] <= (dsthour + 25)
                      onsetInDoubleHour = nightsummary[, onsetcol] >= (dsthour + 24) & nightsummary[,
                                                                                                    onsetcol] <= (dsthour + 25)
                      onsetBeforeDoubleHour = nightsummary[, onsetcol] <= (dsthour + 24)
                      wakeAfterDoubleHour = nightsummary[, wakecol] >= (dsthour + 25)
                      timeWentBackward = length(which(delta_t1 < 0)) > 0
                      if (onsetBeforeDoubleHour == TRUE & wakeInDoubleHour == TRUE) {
                        if (timeWentBackward == TRUE) {
                          # if time went back then it ended in the second hour of the double
                          # hour and the + 1 is definitely justified
                          nightsummary[, durcol] = nightsummary[, durcol] + 1
                        } else if (timeWentBackward == FALSE) {
                          # if time did not go back then SPT may have ended in either the first
                          # or second of the double hour TO DO: distinguish these rare cases,
                          # for now assume it ends in the second hour to avoid sleep efficiency
                          # above 100%.
                          nightsummary[, durcol] = nightsummary[, durcol] + 1
                        }
                      }
                      if (wakeAfterDoubleHour == TRUE & onsetInDoubleHour == TRUE) {
                        if (timeWentBackward == TRUE) {
                          # if time went back then it ended in the second hour of the double
                          # hour and the + 1 is definitely justified
                          nightsummary[, durcol] = nightsummary[, durcol] + 1  # accelerometer derived sleep duration
                        } else if (timeWentBackward == FALSE) {
                          # if time did not go back then SPT may have ended in either the first
                          # or second of the double hour TO DO: distinguish these rare cases,
                          # for now assume it ends in the second hour to avoid sleep efficiency
                          # above 100%.
                          nightsummary[, durcol] = nightsummary[, durcol] + 1  # accelerometer derived sleep duration
                        }
                      }
                      return(nightsummary)
                    }
                    nightsummary[sumi, ] = correctSptEdgingInDoubleHour(nightsummary[sumi, ], onsetcol = 3,
                                                                        wakecol = 4, durcol = 5, dsthour = dsthour, delta_t1 = delta_t1)
                    nightsummary[sumi, ] = correctSptEdgingInDoubleHour(nightsummary[sumi, ], onsetcol = 7,
                                                                        wakecol = 8, durcol = 9, dsthour = dsthour, delta_t1 = delta_t1)
                  }
                  # ======================================================================================================
                  sibds_atleast15min = 0
                  if (length(sibds) > 0) {
                    spocum.t.dur_sibd = sum(sibds)
                    atleast15min = which(sibds >= 1/4)
                    if (length(atleast15min) > 0) {
                      sibds_atleast15min = sibds[atleast15min]
                      spocum.t.dur_sibd_atleast15min = sum(sibds_atleast15min)
                    } else {
                      spocum.t.dur_sibd_atleast15min = 0
                    }
                  } else {
                    spocum.t.dur_sibd = 0
                    spocum.t.dur_sibd_atleast15min = 0
                  }
                  
                  nightsummary[sumi, 14] = spocum.t.dur.noc  #SleepDurationInSpt
                  nightsummary[sumi, 15] = nightsummary[sumi, 5] - spocum.t.dur.noc  #WASO
                  nightsummary[sumi, 16] = spocum.t.dur_sibd  #total sib (sustained inactivty bout) duration during wakinghours
                  nightsummary[sumi, 17] = length(which(spocum.t$overlapGuider == 1))  #number of nocturnalsleep periods
                  nightsummary[sumi, 18] = nightsummary[sumi, 17] - 1  #number of awakenings
                  nightsummary[sumi, 19] = length(which(spocum.t$overlapGuider == 0))  #number of sib (sustained inactivty bout) during wakinghours
                  nightsummary[sumi, 20] = as.numeric(spocum.t.dur_sibd_atleast15min)  #total sib (sustained inactivty bout) duration during wakinghours of at least 5 minutes
                  #-------------------------------------------------------
                  # Also report timestamps in non-numeric format:
                  acc_onset = nightsummary[sumi, 3]
                  acc_wake = nightsummary[sumi, 4]
                  if (acc_onset > 24) acc_onset = acc_onset - 24
                  if (acc_wake > 24) acc_wake = acc_wake - 24
                  #--------------------------------------------
                  # convert into clocktime
                  acc_onsetTS = convertHRsinceprevMN2Clocktime(acc_onset)
                  acc_wakeTS = convertHRsinceprevMN2Clocktime(acc_wake)
                  nightsummary[sumi, 21] = acc_onsetTS
                  nightsummary[sumi, 22] = acc_wakeTS
                  #----------------------------------------------
                  nightsummary[sumi, 23] = tmp1  #guider_onset_ts
                  nightsummary[sumi, 24] = tmp4  #guider_wake_ts
                  if (params_sleep[["sleepwindowType"]] == "TimeInBed" || guider == "markerbutton") {
                    # If guider is a sleeplog and if the sleeplog recorded time in bed or a marker button
                    # then calculate:
                    # sleep latency:
                    nightsummary[sumi, 25] = round(nightsummary[sumi, 3] - nightsummary[sumi, 7],
                                                   digits = 7)  #sleeponset - guider_onset
                    # sleep efficiency:
                    if (params_sleep[["sleepefficiency.metric"]] == 1) {
                      nightsummary[sumi, 26] = round(nightsummary[sumi, 14]/nightsummary[sumi, 9], digits = 5)  #accumulated nocturnal sleep / guider
                    } else if (params_sleep[["sleepefficiency.metric"]] == 2) {
                      nightsummary[sumi, 26] = round(nightsummary[sumi, 14]/(nightsummary[sumi, 5] + nightsummary[sumi, 25]), digits = 5)  #accumulated nocturnal sleep / detected spt + latency
                    }
                  }
                  nightsummary[sumi, 27] = pagei
                  nightsummary[sumi, 28] = daysleeper[j]
                  nightsummary[sumi, 29] = wdayname[j]
                  nightsummary[sumi, 30] = calendar_date[j]
                  nightsummary[sumi, 31] = fnames[i]
                  # nightsummary
                  #------------------------------------------------------------------------
                  # PLOT
                  if (params_output[["do.visual"]] == TRUE) {
                    if (defi == undef[1]) {
                      # only decide whether to plot the first time
                      if (params_output[["outliers.only"]] == TRUE) {
                        if (abs(nightsummary$error_onset[sumi]) > params_output[["criterror"]] | abs(nightsummary$error_wake[sumi]) >
                            params_output[["criterror"]] | abs(nightsummary$error_dur[sumi]) > (params_output[["criterror"]] * 2)) {
                          doplot = TRUE
                          if (verbose == TRUE) cat(" PLOT ")
                        } else {
                          doplot = FALSE
                        }
                      } else {
                        doplot = TRUE
                      }
                    }
                    # upcoming 5 lines added to avoid ending up with meaningless visualisations
                    # of nights for which no sleep log entry was available, and for which L5
                    # method provided estimates
                    if (length(params_sleep[["loglocation"]]) > 0) {
                      cleaningcriterion = 1
                    } else {
                      cleaningcriterion = 2
                    }
                    if (doplot == TRUE & cleaningcode < cleaningcriterion) {
                      idlabels[cnt] = paste0("ID", accid, " night", j)
                      den = 20
                      defii = which(undef == defi)
                      qtop = ((defii/length(undef)) * 0.6) - 0.3
                      qbot = (((defii - 1)/length(undef)) * 0.6) - 0.3
                      # add bar for each sleep defintion of accelerometer
                      for (pli in 1:nrow(spocum.t)) {
                        if (spocum.t$start[pli] > spocum.t$end[pli]) {
                          if (pli > 1 & pli < nrow(spocum.t) &
                              abs(as.numeric(spocum.t$start[pli]) - as.numeric(spocum.t$end[pli])) < 2) {
                            spocum.t[pli, c("start", "end")] = spocum.t[pli, c("end", "start")]
                          }
                        }
                        if (spocum.t$overlapGuider[pli] == 1) {
                          colb = rainbow(length(undef), start = 0.7, end = 1)
                        } else {
                          colb = rainbow(length(undef), start = 0.2, end = 0.4)
                        }
                        if (spocum.t$start[pli] > spocum.t$end[pli]) {
                          # plot sib that starts on the right (morning) and ends on the left (afternoon)
                          rect(xleft = spocum.t$start[pli], ybottom = (cnt + qbot), xright = 36,
                               ytop = (cnt + qtop), col = colb[defii], border = NA)
                          rect(xleft = 12, ybottom = (cnt + qbot), xright = spocum.t$end[pli], ytop = (cnt + qtop),
                               col = colb[defii], border = NA)
                        } else {
                          rect(xleft = spocum.t$start[pli], ybottom = (cnt + qbot), xright = spocum.t$end[pli],
                               ytop = (cnt + qtop), col = colb[defii], border = NA)
                        }
                      }
                      GuiderWaken = GuiderWake
                      GuiderOnsetn = GuiderOnset
                      
                      if (GuiderWake > 36) GuiderWaken = GuiderWake - 24
                      if (GuiderOnset > 36) GuiderOnsetn = GuiderOnset - 24
                      if (defi == undef[length(undef)]) {
                        # only plot log for last definition night sleeper
                        
                        if (GuiderOnsetn > GuiderWaken) {
                          # day sleeper
                          rect(xleft = GuiderOnsetn, ybottom = (cnt - 0.3), xright = 36, ytop = (cnt + 0.3),
                               col = "black", border = TRUE, density = den)
                          rect(xleft = 12, ybottom = (cnt - 0.3), xright = GuiderWaken, ytop = (cnt + 0.3),
                               col = "black", border = TRUE, density = den)
                        } else {
                          rect(xleft = GuiderOnsetn, ybottom = (cnt - 0.3), xright = GuiderWaken, ytop = (cnt + 0.3),
                               col = "black", border = TRUE, density = den)
                        }
                      }
                    }
                  }
                  # PLOT
                  #------------------------------------------------------------------------
                  nightsummary[sumi, 32] = cleaningcode
                  nightsummary[sumi, 33] = sleeplog_used
                  nightsummary[sumi, 34] = logid
                  nightsummary[sumi, 35] = acc_available
                  nightsummary[sumi, 36] = guider
                  # Extract SRI for this night
                  nightsummary[sumi, 37:38] = NA
                  if (!exists("SleepRegularityIndex")) {
                    SleepRegularityIndex = NA
                  }
                  SRI = SleepRegularityIndex
                  if (is.data.frame(SRI) == TRUE) {
                    calendar_date_asDate = format(as.Date(calendar_date[j], format = "%d/%m/%Y"), format = ("%d/%m/%Y"))
                    calendar_date_reformat = format(x = calendar_date_asDate, format = "%d/%m/%Y")
                    SRIindex = which(SRI$date == calendar_date_reformat & SRI$frac_valid > (params_cleaning[["includenightcrit"]]/24))
                    if (length(SRIindex) > 0) {
                      nightsummary[sumi, 37] = SRI$SleepRegularityIndex[SRIindex[1]]
                      nightsummary[sumi, 38] = SRI$frac_valid[SRIindex[1]]
                    }
                  }
                  if (length(longitudinal_axis) == 0) {
                    nightsummary[sumi, 39] = NA
                  } else {
                    nightsummary[sumi, 39] = longitudinal_axis
                  }
                  if (length(defaultGuiderCorrected) == 0 | guider == "sleeplog") {
                    nightsummary[sumi, 40] = NA
                  } else {
                    nightsummary[sumi, 40] = defaultGuiderCorrected
                  }
                  if (params_output[["storefolderstructure"]] == TRUE) {
                    nightsummary[sumi, 41] = ffd[i]  #full filename structure
                    nightsummary[sumi, 42] = ffp[i]  #use the lowest foldername as foldername name
                  }
                  sumi = sumi + 1
                }  #run through definitions
                if (params_output[["do.visual"]] == TRUE) {
                  if (cleaningcode < cleaningcriterion & doplot == TRUE) {
                    lines(x = c(12, 36), y = c(cnt, cnt), lwd = 0.2, lty = 2)  #abline(h=cnt,lwd=0.2,lty=2)
                    if (daysleeper[j] == TRUE) {
                      lines(x = c(18, 18), y = c((cnt - 0.3), (cnt + 0.3)), lwd = 2, lty = 2, col = "black")
                    }
                    # only increase count if there was bar plotted and it is the last definition
                    if (defi == undef[length(undef)]) {
                      cnt = cnt + 1
                    }
                  }
                }
              }
              if (addlegend == TRUE) {
                colb_spt = rainbow(length(undef), start = 0.7, end = 1)
                colb_day = rainbow(length(undef), start = 0.2, end = 0.4)
                colb = c(colb_spt, colb_day)
                legnames = c(paste0("sib", undef, "_spt"), paste0("sib", undef, "_day"), "guider, e.g. diary")
                legend("top", legend = legnames, density = c(rep(NA, 2 * length(undef)), 40), fill = c(colb, "black"),
                       border = c(colb, "black"), ncol = min(c(3, length(legnames))), cex = 0.7)
                addlegend = FALSE
              }
            }
          }
          # END OF PLOT AND nightsummary MEASURES
        }  #nights
        if (length(nnights.list) == 0) {
          # if there were no nights to analyse
          nightsummary[sumi, 1:2] = c(accid, 0)
          nightsummary[sumi, c(3:30, 34, 36:39)] = NA
          nightsummary[sumi, 31] = fnames[i]
          nightsummary[sumi, 32] = 4  #cleaningcode = 4 (no nights of accelerometer available)
          nightsummary[sumi, c(33, 35)] = c(FALSE, TRUE)  #sleeplog_used acc_available
          if (params_output[["storefolderstructure"]] == TRUE) {
            nightsummary[sumi, 41:42] = c(ffd[i], ffp[i])  #full filename structure and use the lowest foldername as foldername name
          }
          sumi = sumi + 1
        }
        if (params_sleep[["sleepwindowType"]] != "TimeInBed" &&
            params_sleep[["consider_marker_button"]] == FALSE) {
          nightsummary = nightsummary[, which(colnames(nightsummary) %in% c("sleeplatency", "sleepefficiency") ==
                                                FALSE)]
        }
        GGIRversion = "GGIR not used"
        if (is.element('GGIR', installed.packages()[,1])) {
          GGIRversion = as.character(utils::packageVersion("GGIR"))
          if (length(GGIRversion) != 1) GGIRversion = sessionInfo()$otherPkgs$GGIR$Version
        }
        if (nrow(nightsummary) > 0) {
          nightsummary$GGIRversion = GGIRversion
        }
        save(nightsummary, tail_expansion_log, GGIRversion, file = paste0(metadatadir, ms4.out, "/", fnames[i]))
      }
    }
  }  #end of loop through acc files
  if (cnt67 == 2 & params_output[["do.visual"]] == TRUE) {
    if (cnt - 1 != (nnpp + 1)) {
      zerolabel = which(idlabels == 0)
      if (length(zerolabel) > 0) idlabels[zerolabel] = " "
      axis(side = 2, at = 1:nnpp, labels = idlabels, las = 1, cex.axis = 0.5)
    }
    dev.off()
    cnt67 = 1
  }
}

g.part5.lux_persegment = function(ts, sse, LUX_day_segments, epochSize, desiredtz = "") {
  # only look at waking hours of this day
  sse = sse[which(ts$diur[sse] == 0)]
  # create equal length vector representing per epoch the first hour of the corresponding day segment
  if ("POSIXt" %in% class(ts$time[sse[1]])) {
    posixTime = ts$time[sse]
  } else {
    posixTime = iso8601chartime2POSIX(ts$time[sse], tz = desiredtz)
  }
  first_hour_seg = as.numeric(format(posixTime,"%H"))
  for (ldi in 1:(length(LUX_day_segments) - 1)) {
    # round all hours to bottom of its class
    tmpl = which(first_hour_seg >= LUX_day_segments[ldi] &
                   first_hour_seg <  LUX_day_segments[ldi + 1])
    if (length(tmpl) > 0) {
      first_hour_seg[tmpl] = LUX_day_segments[ldi]
    }
  }
  Nepochperday = 24 * (3600 / epochSize)
  if (length(first_hour_seg) > Nepochperday) {
    first_hour_seg = first_hour_seg[1:Nepochperday]
  }
  if (length(sse) > Nepochperday) {
    sse = sse[1:Nepochperday]
  }
  # expanding first time segment
  expected_N_seg1 = (rep(LUX_day_segments,2)[which(LUX_day_segments == first_hour_seg[1])+1] - LUX_day_segments[which(LUX_day_segments == first_hour_seg[1])]) * 60 * (60/epochSize)
  actual_N_seg1 = length(which(first_hour_seg[1:round(Nepochperday*0.66)] == first_hour_seg[1]))
  missingN_seg1 = expected_N_seg1 - actual_N_seg1
  if (missingN_seg1 > 0) {
    extension = (sse[1]-missingN_seg1):(sse[1]-1)
    ts$lightpeak[extension] = 0 # ensure that light during SPT is treated as zeros
    sse = c(extension,sse)
    first_hour_seg = c(rep(first_hour_seg[1], missingN_seg1), first_hour_seg)
  }
  # removing again the data after 24 hours to avoid overlap  Nepochperday = 24 * (3600 / ws3)
  if (length(first_hour_seg) > Nepochperday) {
    first_hour_seg = first_hour_seg[1:Nepochperday]
  }
  if (length(sse) > Nepochperday) {
    sse = sse[1:Nepochperday]
  }
  # if less than 24 hours makes sure that last incomplete segment is removed
  if (length(sse) < Nepochperday) {
    last_seg = first_hour_seg[length(first_hour_seg)]
    N_last_seg = length(which(first_hour_seg[round(length(first_hour_seg)*0.5):length(first_hour_seg)] == last_seg))
    sse = sse[1:(length(sse) - N_last_seg)]
    first_hour_seg = first_hour_seg[1:(length(first_hour_seg) - N_last_seg)]
  }
  # Aggregation per segment of the window:
  # Function to help standardise metric per seg output across days and recordings
  standardise_luxperseg = function(x, LUX_day_segments, LUXmetricname = "") {
    colnames(x) = c("seg", "light")
    if (24 %in% LUX_day_segments) LUX_day_segments = LUX_day_segments[which(LUX_day_segments != 24)] # remove end of day
    Nsegs = length(LUX_day_segments)
    x = base::merge(x, data.frame(seg= LUX_day_segments, 
                                  light = rep(NA, Nsegs)),
                    by =c("seg"), all.y=TRUE)
    x = x[,c("seg","light.x")]
    colnames(x) = c("seg", "light")
    LUX_day_segments = c(LUX_day_segments, 24) # end of day back in
    end_of_segment = LUX_day_segments[which(x$seg %in% LUX_day_segments) + 1]
    invisible(list(values = x$light,
                   names = paste0("LUX_",LUXmetricname,"_",x$seg,"_",end_of_segment, "hr_day")))
  }
  # Fraction above 1000 LUX
  fraction_above_thousand = function(x, epochSize) {
    timeabove1000 = length(which(x > 1000)) / (60/epochSize)
    return(timeabove1000)
  }
  LUXabove1000 = aggregate(ts$lightpeak[sse], by =  list(first_hour_seg), fraction_above_thousand, epochSize)
  # Time awake
  countvalue = function(x, epochSize, value) {
    return(length(which(x == value)) / (60/epochSize))
  }
  LUXwaketime = aggregate(ts$diur[sse], by = list(first_hour_seg), FUN = countvalue, value = 0, epochSize = epochSize)
  # Mean light
  mymean = function(x) {
    return(round(mean(x, na.rm = T), digits = 1))
  }
  LUXmean = aggregate(ts$lightpeak[sse], by =  list(first_hour_seg), mymean)
  # Time light imputed
  LUXlightimputed = aggregate(ts$lightpeak_imputationcode[sse], by = list(first_hour_seg), FUN = countvalue, value = 1, epochSize = epochSize)
  LUXlightignored = aggregate(ts$lightpeak_imputationcode[sse], by = list(first_hour_seg), FUN = countvalue, value = 2, epochSize = epochSize)
  
  standardLPS1 = standardise_luxperseg(x = LUXabove1000, LUX_day_segments, LUXmetricname = "above1000")
  standardLPS2 = standardise_luxperseg(x = LUXwaketime, LUX_day_segments, LUXmetricname = "timeawake")
  standardLPS3 = standardise_luxperseg(x = LUXmean, LUX_day_segments, LUXmetricname = "mean")
  standardLPS4 = standardise_luxperseg(x = LUXlightimputed, LUX_day_segments, LUXmetricname = "imputed")
  standardLPS5 = standardise_luxperseg(x = LUXlightignored, LUX_day_segments, LUXmetricname = "ignored")
  
  values = c(standardLPS1$values, standardLPS2$values,
             standardLPS3$values, standardLPS4$values,
             standardLPS5$values)
  names = c(standardLPS1$names, standardLPS2$names,
            standardLPS3$names, standardLPS4$names,
            standardLPS5$names)
  invisible(list(values=values, names=names))
}

library(GGIR)
context("lux_per_segment")
test_that("lux_per_segment is correctly calculated", {
  skip_on_cran()
  N = 10000
  t0 = as.POSIXlt("2021-05-27 22:08:28", tz = "Europe/Amsterdam", origin = "1970-01-01")
  # the values of time are not used in the function, we only care about the length the object time
  time = seq(t0, t0 + ((N * 60) - 1), by = 60)
  diur = c(rep(1,1000),rep(0,8000),rep(1, 1000))
  detection = rep(0, N)
  set.seed(300)
  detection[which(runif(n = N, min = 0, max = 1) < 0.1)] = 1
  ACC = rnorm(n = N, mean = 0, sd = 40)
  lightpeak_imputationcode = lightpeak = ((1:N) / N) * 1000
  lightpeak[c(4:10)] = NA
  lightpeak_imputationcode[c(4:10)] = 2
  
  ts = data.frame(ACC = ACC, time = time, diur = diur, sibdetection = detection, 
                  lightpeak = lightpeak, lightpeak_imputationcode = lightpeak_imputationcode)
  sse = 1001:9000
  LUX_day_segments = c(9, 15, 24)
  ws3new = 60
  luxperseg = g.part5.lux_persegment(ts, sse, LUX_day_segments, epochSize = ws3new)
  expect_equal(round(luxperseg$values[3]), 12)
  expect_equal(round(luxperseg$values[4]), 540)
  expect_equal(round(luxperseg$values[5]), 3)
  expect_equal(round(luxperseg$values[10]), 0)
  expect_equal(luxperseg$names[10], "LUX_ignored_15_24hr_day")
  expect_equal(luxperseg$names[3], "LUX_timeawake_9_15hr_day" )
})

g.part4_extractid = function(idloc, fname, dolog, sleeplog, accid = c()) {
  if (length(accid) == 0) {
    #------------------------------------------------------
    # extract the identifier from accelerometer data if it was not found in the GGIR part 3 milestone data
    if (idloc %in% c(2, 5, 6, 7) == TRUE) { #idloc is an argument to specify where the participant identifier can be found
      if (idloc == 2) {
        getIDfromChar = function(x) {
          return(as.character(unlist(strsplit(x, "_")))[1])
        }
      } else if (idloc == 5) {
        getIDfromChar = function(x) {
          return(as.character(unlist(strsplit(x, " ")))[1])
        }
      } else if (idloc == 6) {
        getIDfromChar = function(x) {
          return(as.character(unlist(strsplit(x, "[.]")))[1])
        }
      } else if (idloc == 7) {
        getIDfromChar = function(x) {
          return(as.character(unlist(strsplit(x, "-")))[1])
        }
      }
      accid = apply(as.matrix(as.character(fname)), MARGIN = c(1), FUN = getIDfromChar)
    } else { # get id from filename
      newaccid = fname
      if (length(unlist(strsplit(newaccid, "_"))) > 1) newaccid = unlist(strsplit(newaccid, "_"))[1]
      if (length(unlist(strsplit(newaccid, " "))) > 1) newaccid = unlist(strsplit(newaccid, " "))[1]
      if (length(unlist(strsplit(newaccid, "[.]RDa"))) > 1) newaccid = unlist(strsplit(newaccid, "[.]RDa"))[1]
      if (length(unlist(strsplit(newaccid, "[.]cs"))) > 1) newaccid = unlist(strsplit(newaccid, "[.]cs"))[1]
      accid = newaccid[1]
    }
  }
  # get matching identifier from sleeplog
  if (dolog == TRUE) { 
    
    # convert to character
    logid = as.character(sleeplog$ID)
    accid2 = as.character(accid)
    
    # remove spaces in ID, to ease matching, because some accelerometer brands at several spaces behind ID
    logid = gsub(pattern = " ", replacement = "", x = as.character(logid))
    accid2 = gsub(pattern = " ", replacement = "", x = as.character(accid2))
    
    # attempt to match 1 - works if both IDs are identical
    matching_indices_sleeplog = which(logid == accid2)
    matched = length(matching_indices_sleeplog)
    matched_unique = unique(sleeplog$ID[matching_indices_sleeplog])
    
    # attempt to match 2 - ignore case
    if (matched == 0) {
      matching_indices_sleeplog = which(tolower(logid) == tolower(accid2))
      matched = length(matching_indices_sleeplog)
      matched_unique = unique(sleeplog$ID[matching_indices_sleeplog])
    } 
    
    # attempt to match 3 - get rid of letters
    if (matched == 0) {
      # remove all letters
      accid2 = gsub("[^0-9.-]", "", accid2)
      logid = gsub("[^0-9.-]", "", sleeplog$ID)
      matching_indices_sleeplog = which(logid == accid2)
      matched = length(matching_indices_sleeplog)
      matched_unique = unique(sleeplog$ID[matching_indices_sleeplog])
    }
    
    # attempt to match 4 - get rid of extra leading 0s
    if (matched == 0) {
      # remove leading zeros
      accid2 = gsub("^0+", "", accid2)
      logid = gsub("^0+", "", logid)
      matching_indices_sleeplog = which(logid == accid2)
      matched = length(matching_indices_sleeplog)
      matched_unique = unique(sleeplog$ID[matching_indices_sleeplog])
    }
    
    # if matched to more than one entrance in sleeplog, warn the user
    if (length(matched_unique) > 1) {
      warning(paste0("\n", as.character(accid), " matched to more than one entrance ",
                     "in the sleeplog (i.e., ", paste(as.character(matched_unique), collapse = ", "), 
                     ").\nPlease revise the IDs in your sleeplog. ", matched_unique[1], " used."))
      matching_indices_sleeplog = which(sleeplog$ID == matched_unique[1])
    }
  } else if (dolog == FALSE) { 
    matching_indices_sleeplog = 1
  }
  invisible(list(accid = accid, matching_indices_sleeplog = matching_indices_sleeplog))
}

g.part5.fixmissingnight = function(summarysleep, sleeplog = c(), ID) {
  #========================================================
  # Added 24 March 2020, for the rare situation when part 4 does misses a night.
  # This is possible when the accelerometer was not worn the entire day and
  # ignorenonwear was set to TRUE, part4 then gives up and does not try to store anything.
  # rather than fiddling with part 4 again it seems more logical to address this here,
  # because only part 5 needs this.
  clock2numtime = function(x) { # function used for converting sleeplog times to hour times
    x2 = as.numeric(unlist(strsplit(x,":"))) / c(1,60,3600)
    return(sum(x2))
  }
  hr_to_clocktime = function(x) {
    hrsNEW = floor(x)
    minsUnrounded = (x - hrsNEW) *60
    minsNEW =  floor(minsUnrounded)
    secsNEW =  floor( (minsUnrounded - minsNEW)*60)
    if (minsNEW < 10) minsNEW = paste0(0,minsNEW)
    if (secsNEW < 10) secsNEW = paste0(0,secsNEW)
    if (hrsNEW < 10) hrsNEW = paste0(0,hrsNEW)
    time = paste0(hrsNEW,":",minsNEW,":",secsNEW)
    return(time)
  }
  potentialnight = min(summarysleep$night):max(summarysleep$night)
  missingnight = which(as.numeric(potentialnight) %in% as.numeric(summarysleep$night) == FALSE)
  
  if ("guider_wakeup" %in% colnames(summarysleep) == TRUE) {
    guider_onset = "guider_onset"
    guider_wakeup = "guider_wakeup"
  } else {
    guider_onset = "guider_inbedStart"
    guider_wakeup = "guider_inbedEnd"
  }
  if (length(missingnight) > 0) {
    for (mi in missingnight) {
      missingNight = potentialnight[mi]
      newnight = summarysleep[1,]
      newnight[which(names(newnight) %in% c("ID","night", "sleepparam","filename","filename_dir","foldername") == FALSE)] = NA
      newnight$wakeup = newnight[,guider_wakeup] = newnight$sleeponset = newnight[,guider_onset] = NA
      newnight$night = missingNight
      newnight$calendar_date = format(as.Date(as.POSIXlt(summarysleep$calendar_date[mi-1],format="%d/%m/%Y") + (36*3600)), "%d/%m/%Y")
      timesplit = as.numeric(unlist(strsplit(as.character(newnight$calendar_date),"/"))) # remove leading zeros
      newnight$calendar_date = paste0(timesplit[1],"/",timesplit[2],"/",timesplit[3])
      newnight$daysleeper = 0
      newnight$acc_available = 0
      if (length(sleeplog) > 0) { # we impute with sleeplog (TO DO: Also implement catch for if sleeplog is not available)
        sleeplogonset = sleeplog$sleeponset[which(sleeplog$ID == ID & sleeplog$night == missingNight)]
        sleeplogwake = sleeplog$sleepwake[which(sleeplog$ID == ID & sleeplog$night == missingNight)]
        newnight$sleeplog_used = 0
        newnight$guider = "nosleeplog_accnotworn"
        if (length(sleeplogonset) != 0 & length( sleeplogwake) != 0) {
          if (is.na(sleeplogonset) == FALSE & is.na(sleeplogonset) == FALSE) {
            sleeplogonset_hr = clock2numtime(sleeplogonset)
            sleeplogwake_hr= clock2numtime(sleeplogwake)
            newnight$sleeponset = newnight[,guider_onset] = sleeplogonset_hr
            newnight$wakeup = newnight[,guider_wakeup] = sleeplogwake_hr
            if (sleeplogwake_hr > 36) {
              newnight$daysleeper = 1
              newnight$sleeponset_ts = hr_to_clocktime(sleeplogonset_hr)
              newnight$wakeup_ts = hr_to_clocktime(sleeplogwake_hr)
              newnight$sleeplog_used = 1
              newnight$guider = "sleeplog"
            }
          }
        }
      } else {
        newnight$sleeplog_used = 0
        newnight$guider = "nosleeplog_accnotworn"
      }
      newnight$cleaningcode = 5
      summarysleep = rbind(summarysleep[1:(mi-1),],
                                newnight,
                                summarysleep[mi:nrow(summarysleep),])
    }
  }
  return(summarysleep)
}

g.detecmidnight = function(time, desiredtz, dayborder) {
  # code in this function is able to deal with two types of timestamp format
  convert2clock = function(x) { #ISO format
    return(format(as.POSIXlt(x, format = "%Y-%m-%dT%H:%M:%S%z", tz = desiredtz), "%H:%M:%S"))
  }
  convert2clock_space = function(x) { #POSIX format
    return(unlist(strsplit(x, " "))[2])
  }
  checkmidnight = function(x){
    temp1 = as.numeric(unlist(strsplit(x, ":")))
    return(sum(temp1 / c(1, 60, 3600)))
  }
  space = ifelse(length(unlist(strsplit(time[1], " "))) > 1,TRUE,FALSE)
  if (space == TRUE) {
    time_clock = sapply(time, FUN = convert2clock_space)
    checkmidnight_out = lapply(time_clock, FUN = checkmidnight)
  } else {
    time_clock = convert2clock(time)
    checkmidnight_out = unlist(lapply(time_clock, FUN = checkmidnight))
  }
  midn = which(checkmidnight_out == dayborder) #replaced "== 0" for "== dayborder" to work in any scenario
  if (length(midn) == 0) { # measurement with no midnights, use last timestamp as dummy midnight for g.analyse() to work
    midnights = time[length(time)]
    midnightsi = length(time)
  } else {
    midnights = format(time[midn])
    midnightsi = midn
  }
  if (length(midn) == 0) { # measurement with no midnights, use last timestamp as dummy midnight for g.analyse() to work
    lastmidnight = time[length(time)]
    lastmidnighti = length(time)
    firstmidnight = time[1]
    firstmidnighti = 1
  } else {
    lastmidnight = midnights[length(midnights)]
    lastmidnighti = midnightsi[length(midnights)]
    firstmidnight = midnights[1]
    firstmidnighti = midnightsi[1]
  }
  invisible(list(firstmidnight = firstmidnight, firstmidnighti = firstmidnighti,
                 lastmidnight = lastmidnight, lastmidnighti = lastmidnighti,
                 midnights = midnights, midnightsi = midnightsi))
}

GGIR = function(mode = 1:5, datadir = c(), outputdir = c(),
                studyname = c(), f0 = 1, f1 = 0,
                do.report = c(2, 4, 5, 6), configfile = c(),
                myfun = c(), verbose = TRUE,
                ...) {
  
  # Set language to English while using GGIR
  LC_TIME_backup = Sys.getlocale("LC_TIME")
  Sys.setlocale("LC_TIME", "C")
  on.exit({Sys.setlocale("LC_TIME", LC_TIME_backup)}, add = TRUE)
  
  #get input variables
  input = list(...)
  # Check for duplicated arguments
  if (length(input) > 1) {
    argNames = names(input)
    dupArgNames = duplicated(argNames)
    if (any(dupArgNames)) { # Warn user about those duplicated arguments
      for (dupi in unique(argNames[dupArgNames])) {
        dupArgValues = input[which(argNames %in% dupi)]
        if (all(dupArgValues == dupArgValues[[1]])) { # duplicated arguments, but no confusion about what value should be
          warning(paste0("\nParameter ", dupi, " has been provided more than once. Try to avoid this."))
        } else {# duplicated arguments, and confusion about what value should be,
          warning(paste0("\nParameter ", dupi, " has been provided more than once and with inconsistent values. Please fix."))
        }
      }
    }
    default_params = load_params()
    known_param_names =  unlist(lapply(names(default_params), function(n) names(default_params[[n]])))
    if (any(argNames %in% known_param_names == FALSE)) {
      unknown_param_names = argNames[argNames %in% known_param_names == FALSE]
      if (length(unknown_param_names) == 1) {
        stop(paste0("\nParameter ", unknown_param_names,
                       " is unknown to GGIR and will not be used, ",
                    "please check for typos or remove."), call. = FALSE)
      } else {
        stop(paste0("\nParameters ", paste0(unknown_param_names, collapse = " and "),
                       " are unknown to GGIR and will not be used, ",
                    "please check for typos or remove these."), call. = FALSE)
      }
      rm(unknown_param_names)
    }
    rm(default_params, known_param_names)
  }
  
  if (length(datadir) == 0) {
    stop('\nVariable datadir is not specified')
  }
  
  if (length(outputdir) == 0) {
    stop('\nVariable outputdir is not specified')
  }
  
  # Convert paths from Windows specific slashed to generic slashes
  outputdir = gsub(pattern = "\\\\", replacement = "/", x = outputdir)
  datadir = gsub(pattern = "\\\\", replacement = "/", x = datadir)
  
  
  
  # Establish which parts need to be processed:
  dopart1 = dopart2 = dopart3 = dopart4 = dopart5 = dopart6 = FALSE
  if (length(which(mode == 0)) > 0) {
    dopart1 = dopart2 = dopart3 = dopart4 = dopart5 = TRUE # intentionally not including dopart6
  } else {
    if (length(which(mode == 1)) > 0) dopart1 = TRUE
    if (length(which(mode == 2)) > 0) dopart2 = TRUE
    if (length(which(mode == 3)) > 0) dopart3 = TRUE
    if (length(which(mode == 4)) > 0) dopart4 = TRUE
    if (length(which(mode == 5)) > 0) dopart5 = TRUE
    if (length(which(mode == 6)) > 0) dopart6 = TRUE
  }
  # identify output folder names
  filelist = isfilelist(datadir)
  if (filelist == TRUE) {
    metadatadir = paste0(outputdir, "/output_", studyname)
  } else {
    outputfoldername = unlist(strsplit(datadir, "/"))[length(unlist(strsplit(datadir, "/")))]
    metadatadir = paste0(outputdir, "/output_", outputfoldername)
  }
  
  # Configuration file - check whether it exists or auto-load
  configfile_csv = c()
  ex = "csv"
  if (length(configfile) > 0) { # Get extension of file
    ex = unlist(strsplit(basename(configfile), split = "\\."))
    ex = ex[length(ex)]
  }
  if (ex == "csv") { # at a later point there may also be other file extensions
    if (dir.exists(metadatadir) | length(configfile) > 0) {
      # so if outputdir was not created (first time) and if configfile is not
      # specified then we can safely assume that there is no configfile
      # and this section will be skipped, alternatively an attempt is
      # made to retrieve settings from the configfile
      if (length(configfile) > 0) {
        if (!file.exists(configfile)) {
          stop("\nDo not supply argument configfile if the configfile does not exist yet")
        } else {
          configfile_csv = configfile
        }
      }
      if (dir.exists(metadatadir) & length(configfile) == 0) {
        config_file_in_outputdir = paste0(metadatadir, "/config.csv")
        # So, if both exist, we prioritise configfile over the configfile
        # stored in outputdir and this is skipped
        if (file.exists(config_file_in_outputdir)) configfile_csv = config_file_in_outputdir
      }
    }
  }
  
  #----------------------------------------------------------
  # Extract parameters from user input, configfile and/or defaults.
  params = extract_params(input = input, configfile_csv = configfile_csv) # load default parameters here in g.shell.GGIR
  params_sleep = params$params_sleep
  params_metrics = params$params_metrics
  params_rawdata = params$params_rawdata
  params_247 = params$params_247
  params_phyact = params$params_phyact
  params_cleaning = params$params_cleaning
  params_output = params$params_output
  params_general = params$params_general
  
  if (params_general[["dataFormat"]] == "ukbiobank") {
    warning("\nRunnning part 3, 4, and 5 are disabled when dataFormat is ukbiobank epoch", call. = FALSE)
    dopart3 = dopart4 = dopart5 = FALSE
    mode = mode[which(mode <= 2)]
  }
  
  if (dopart3 == TRUE & params_metrics[["do.anglez"]] == FALSE & params_general[["dataFormat"]] == "raw") {
    params_metrics[["do.anglez"]] = TRUE
  }
  
  if (length(myfun) != 0) { # Run check on myfun object, if provided
    check_myfun(myfun, params_general[["windowsizes"]])
  }
  
  if (params_output[["visualreport"]] == TRUE &
      params_output[["old_visualreport"]] == TRUE &
      params_general[["dataFormat"]] != "raw") {
    params_output[["visualreport"]] == FALSE
  }
  
  # check package dependencies
  if (params_metrics$do.neishabouricounts == TRUE) {
    is_actilifecounts_installed = is.element('actilifecounts', installed.packages()[,1])
    if (is_actilifecounts_installed == FALSE) {
      stop("If you want to derive Neishabouricounts, please install package: actilifecounts.", call. = FALSE)
    } else {
      if (utils::packageVersion("actilifecounts") < "1.1.0") {
        stop("Please update R package actilifecounts to version 1.1.0 or higher", call. = FALSE)
      }
    }
  }
  if (1 %in% mode) {
    checkFormat = TRUE
    if (all(dir.exists(datadir)) == TRUE) {
      rawaccfiles = dir(datadir, full.names = TRUE)[f0:f1]
    } else if (all(file.exists(datadir))) {
      rawaccfiles = datadir[f0:f1]
    } else {
      checkFormat = FALSE
    }
    if (checkFormat == TRUE) {
      is_readxl_installed = is.element('readxl', installed.packages()[,1])
      if (is_readxl_installed == FALSE) {
        getExt = function(x) {
          tmp = unlist(strsplit(x, "[.]"))
          return(tmp[length(tmp)])
        }
        rawaccfiles_formats = unique(unlist(lapply(rawaccfiles, FUN = getExt)))
        if (any(grepl("xls", rawaccfiles_formats))) {
          stop("If you are working with xls data for Sensewear, please install package: readxl.", call. = FALSE)
        }
      }
    }
  }
  
  # This check cannot be part of check_params because it needs to know mode:
  if (6 %in% mode && params_247[["part6CR"]] == FALSE && params_247[["part6HCA"]]) {
    warning(paste0("Both part6CR and part6HCA are set to FALSE by which there is ",
                   "not analysis to be run in part 6."), call. = FALSE)
  }
  #===========================
  # Establish default start / end file index to process
  if (filelist == FALSE) {
    if (dir.exists(datadir) == FALSE && 1 %in% mode) {
      # Note: The check whether datadir exists is only relevant when running part 1
      # For other scenarios it can be convenient to keep specifying datadir
      # even if the actual path is no longer available, because GGIR uses the most
      # distal folder name to identify the output directory. For example,
      # if part 2-5 are processed on a different system then part 1.
      
      stop("\nDirectory specified by argument datadir does not exist")
    }
    if (datadir == outputdir || grepl(paste(datadir, '/', sep = ''), outputdir)) {
      stop(paste0('\nError: The file path specified by argument outputdir should ',
                  'NOT equal or be a subdirectory of the path specified by argument datadir'))
    }
  }
  if (dir.exists(outputdir) == FALSE) {
    stop("\nDirectory specified by argument outputdir does not exist")
  }
  if (file.access(outputdir, mode = 2) != 0) {
    stop("\nUser does not seem to have write access permissions for the directory specified by argument outputdir.\n")
  }
  if (filelist == TRUE) {
    if (length(studyname) == 0) {
      stop('\nError: studyname must be specified if datadir is a list of files')
    }
  }
  # Correct f1 
  if (is.null(f0) || f0 < 1) { # What file to start with?
    f0 = 1
  }
  f1_orinally_null = ifelse(!is.null(f1) && f1 == 0, yes = TRUE, no = FALSE)
  if ((is.null(f1) || f1 < 1) && 1 %in% mode) {  # What file to end with?
    # Do not modify f1 here when not attempting to process GGIR part 1
    # we only expect datadir to exist when running part 1
    if (filelist == FALSE && params_general[["dataFormat"]] %in% c("phb_xlsx", "fitbit_json") == FALSE) {
      f1 <- length(dir(datadir, recursive = TRUE, ignore.case = TRUE, pattern = "[.](csv|bin|Rda|wa|cw|gt3|AW)")) # modified by JH
    } else if (params_general[["dataFormat"]] %in% c("phb_xlsx", "fitbit_json")) {
      # use number of directories as indicator for number of recordings
      f1 = length(dir(datadir, recursive = FALSE))
    } else {
      f1 = length(datadir) #modified
    }
  }
  if (is.null(f1)) {
    f1 = 0
  }
  
  #-----------------------------------------------------------
  # Print GGIR header to console
  GGIRversion = "could not extract version"
  if (is.element('GGIR', installed.packages()[,1])) {
    GGIRversion = as.character(utils::packageVersion("GGIR"))
    if (length(GGIRversion) != 1) GGIRversion = sessionInfo()$otherPkgs$GGIR$Version
  }
  if (verbose == TRUE) {
    cat(paste0("\n   GGIR version: ",GGIRversion,"\n"))
    cat("\n   << Please cite GGIR in your publications with doi: 10.5281/zenodo.1051064 >>\n")
    random_message = sample(x = 1:3, size = 1)
    cat(paste0("\n   [Note #", random_message, "]"))
    if (random_message == 1) {
      cat("\n   To help us track where GGIR was in the literature, post a link to your publication ")
      cat("\n   in https://github.com/wadpac/GGIR/discussions/categories/show-and-tell")
      cat("\n   or email it to v.vanhees@accelting.com")
    } else if (random_message == 2) {
      cat("\n   To make your research reproducible and interpretable always report:")
      cat("\n     (1) GGIR version")
      cat("\n     (2) Accelerometer brand and product name")
      cat("\n     (3) How you configured the accelerometer")
      cat("\n     (4) Study protocol and wear instructions given to the participants")
      cat("\n     (5) How GGIR was used: Share the config.csv file or your R script.")
      cat("\n     (6) How you post-processed / cleaned GGIR output")
      cat("\n     (7) How reported outcomes relate to the specific variable names in GGIR")
    } else if (random_message == 3) {
      cat("\n   The sustainability of GGIR depends on YOU as user, please consider:")
      cat("\n     (1) Helping other users on https://github.com/wadpac/GGIR/discussions")
      cat("\n     (2) Providing feedback on or help improve GGIR documentation")
      cat("\n     (3) Providing feedback on or help improve GGIR source code")
      cat("\n     (4) Finding funding for work on GGIR")
      cat("\n   For questions about contributing contact v.vanhees@accelting.com.")
    }
    cat("\n")
  }
  #-----------------------------------------------------------
  # Now run GGIR parts 1-5
  print_console_header = function(headerTitle) {
    cat('\n')
    cat(paste0(rep('_', options()$width), collapse = ''))
    cat("\n",headerTitle,"\n")
  }
  if (dopart1 == TRUE) {
    if (verbose == TRUE) print_console_header("Part 1")
    
    if (!is.null(params_general[["maxRecordingInterval"]]) & params_general[["overwrite"]] == TRUE) {
      # When we want to overwrite previously processed data and append recordings
      # it is necessary to first empty folder meta/basic to avoid confusion with
      # previously appended data
      basic_folder = paste0(metadatadir, "/meta/basic")
      if (dir.exists(basic_folder)) {
        basic_ms_files = dir(basic_folder, full.names = TRUE)
        if (length(basic_ms_files) > 0) {
          for (fnr in basic_ms_files) unlink(fnr, recursive = TRUE)
          rm(fnr)
        }
        rm(basic_folder, basic_ms_files)
      }
    }
    if (params_general[["dataFormat"]] == "raw") {
      g.part1(datadir = datadir, metadatadir = metadatadir, f0 = f0, f1 = f1,
              studyname = studyname, myfun = myfun,
              params_rawdata = params_rawdata, params_metrics = params_metrics,
              params_cleaning = params_cleaning, params_general = params_general,
              verbose = verbose)
    } else {
      # Skip g.part1, but instead convert epoch data to a format that
      # looks as if it came out of g.part1
      if (utils::packageVersion("GGIRread") <= "1.0.1") {
        stop(paste0("To work with external epoch ",
                    "data GGIRread version 1.0.2 or above is now required. ",
                    "Please update R package GGIRread. "), call. = FALSE)
      }
      convertEpochData(datadir = datadir,
                       metadatadir = metadatadir,
                       params_general = params_general,
                       f0 = f0,
                       f1 = f1,
                       verbose = verbose)
    }
    if (!is.null(params_general[["maxRecordingInterval"]])) {
      # Append recordings when ID and brand match and gap between
      # recordings does not exceed maxRecordingInterval,
      # where GGIR prohibits user from use a value larger
      # than 504 hours (21 days)
      appendRecords(metadatadir = metadatadir,
                    desiredtz = params_general[["desiredtz"]],
                    idloc = params_general[["idloc"]],
                    maxRecordingInterval = params_general[["maxRecordingInterval"]])
      if (f1_orinally_null == TRUE) {
        f1 = length(dir(paste0(metadatadir, "/meta/basic/")))
      }
    }
    if (!is.null(params_general[["recording_split_times"]])) {
      # Split recordings based on user specified time points
      splitRecords(metadatadir = metadatadir,
                   params_general = params_general)
      if (f1_orinally_null == TRUE) {
        f1 = length(dir(paste0(metadatadir, "/meta/basic/")))
      }
    }
  }
  if (dopart2 == TRUE) {
    if (verbose == TRUE) print_console_header("Part 2")
    if (f1 == 0) f1 = length(dir(paste0(metadatadir, "/meta/basic")))
    g.part2(datadir = datadir, metadatadir = metadatadir, f0 = f0, f1 = f1,
            myfun = myfun, params_cleaning = params_cleaning,
            params_247 = params_247, params_phyact = params_phyact,
            params_output = params_output,
            params_general = params_general,
            verbose = verbose)
  }
  if (dopart3 == TRUE) {
    if (verbose == TRUE) print_console_header("Part 3")
    if (f1 == 0) f1 = length(dir(paste0(metadatadir,"/meta/basic")))
    g.part3(metadatadir = metadatadir, f0 = f0, f1 = f1, myfun = myfun,
            params_sleep = params_sleep, params_output = params_output,
            params_metrics = params_metrics,
            params_general = params_general,
            verbose = verbose)
  }
  if (dopart4 == TRUE) {
    if (verbose == TRUE) print_console_header("Part 4")
    if (f1 == 0) f1 = length(dir(paste0(metadatadir,"/meta/ms3.out")))
    g.part4(datadir = datadir, metadatadir = metadatadir, f0 = f0, f1 = f1,
            params_sleep = params_sleep, params_metrics = params_metrics,
            params_general = params_general, params_output = params_output,
            params_cleaning = params_cleaning,
            verbose = verbose)
  }
  if (dopart5 == TRUE) {
    if (verbose == TRUE) print_console_header("Part 5")
    if (f1 == 0) f1 = length(dir(paste0(metadatadir, "/meta/ms3.out"))) # this is intentionally ms3 and not ms4, do not change!
    g.part5(datadir = datadir, metadatadir = metadatadir, f0 = f0, f1 = f1,
            params_sleep = params_sleep, params_metrics = params_metrics,
            params_general = params_general, params_output = params_output,
            params_cleaning = params_cleaning, params_247 = params_247,
            params_phyact = params_phyact,
            verbose = verbose)
  }
  if (dopart6 == TRUE) {
    if (verbose == TRUE) print_console_header("Part 6")
    if (f1 == 0) {
      f1 = length(dir(paste0(metadatadir, "/meta/ms5.outraw/", 
                             params_phyact[["part6_threshold_combi"]])))
    }
    g.part6(datadir = datadir, metadatadir = metadatadir, f0 = f0, f1 = f1,
            params_general = params_general, params_phyact = params_phyact,
            params_247 = params_247, params_cleaning = params_cleaning,
            verbose = verbose)
  }
  
  #--------------------------------------------------
  # Store configuration parameters in config file
  LS = ls()
  LS = LS[which(LS %in% c("input", "txt", "dopart1", "dopart2", "dopart3", "LS",
                          "dopart4", "dopart5", "dopart6", "metadatadir", "ci", "config",
                          "configfile", "filelist", "outputfoldername", "numi", "logi",
                          "conv2logical", "conv2num", "SI", "params", "argNames", "dupArgNames",
                          "print_console_header", "configfile_csv", "myfun", "ex",
                          "GGIRversion",  "dupArgValues", "verbose", "is_GGIRread_installed", 
                          "is_read.gt3x_installed", "is_ActCR_installed", 
                          "is_actilifecounts_installed", "rawaccfiles", "is_readxl_installed", 
                          "checkFormat", "getExt", "rawaccfiles_formats",
                          "f1_orinally_null", "random_message") == FALSE)]
  
  config.parameters = mget(LS)
  config.matrix = as.data.frame(createConfigFile(config.parameters, GGIRversion))
  config.matrix$context[which(config.matrix$context == "")] = "not applicable"
  if (dir.exists(metadatadir)) {
    data.table::fwrite(config.matrix, file = paste0(metadatadir, "/config.csv"),
                       row.names = FALSE, sep = params_output[["sep_config"]],
                       dec = params_output[["dec_config"]])
  } else {
    warning("\nCould not write config file.")
  }
  #==========================
  # Report generation:
  # -----
  # check a few basic assumptions before continuing
  if (length(which(do.report == 2)) > 0) {
    N.files.basic = length(dir(paste0(metadatadir, "/meta/basic")))
    if (N.files.basic > 0) {
      if (verbose == TRUE) print_console_header("Report part 2")
      if (length(f0) == 0) f0 = 1
      if (f1 == 0) f1 = N.files.basic
      if (length(params_247[["qwindow"]]) > 2 |
          (is.character(params_247[["qwindow"]]) && length(grep(pattern = "onlyfilter|filteronly", x = params_247[["qwindow"]])) == 0) |
          (length(params_247[["qwindow"]]) == 2 & !all(c(0, 24) %in% params_247[["qwindow"]]))) {
        store.long = TRUE
      } else {
        store.long = FALSE
      }
      g.report.part2(metadatadir = metadatadir, f0 = f0, f1 = f1,
                     store.long = store.long, params_output,
                     verbose = verbose, desiredtz = params_general[["desiredtz"]])
    }
  }
  if (length(which(do.report == 4)) > 0) {
    N.files.ms4.out = length(dir(paste0(metadatadir, "/meta/ms4.out")))
    if (N.files.ms4.out > 0) {
      if (verbose == TRUE) print_console_header("Report part 4")
      if (N.files.ms4.out < f0) f0 = 1
      if (N.files.ms4.out < f1) f1 = N.files.ms4.out
      if (f1 == 0) f1 = N.files.ms4.out
      g.report.part4(datadir = datadir, metadatadir = metadatadir, f0 = f0, f1 = f1,
                     data_cleaning_file = params_cleaning[["data_cleaning_file"]],
                     params_sleep = params_sleep, params_output = params_output,
                     verbose = verbose)
    }
  }
  if (length(which(do.report == 5)) > 0) {
    N.files.ms5.out = length(dir(paste0(metadatadir, "/meta/ms5.out")))
    if (N.files.ms5.out > 0) {
      if (verbose == TRUE) print_console_header("Report part 5")
      if (N.files.ms5.out < f0) f0 = 1
      if (N.files.ms5.out < f1) f1 = N.files.ms5.out
      if (f1 == 0) f1 = N.files.ms5.out
      g.report.part5(metadatadir = metadatadir, f0 = f0, f1 = f1,
                     loglocation = params_sleep[["loglocation"]],
                     params_cleaning = params_cleaning,
                     LUX_day_segments = params_247[["LUX_day_segments"]], params_output = params_output,
                     verbose = verbose)
      g.report.part5_dictionary(metadatadir = metadatadir, params_output = params_output)
    }
  }
  if (length(which(do.report == 6)) > 0) {
    N.files.ms6.out = length(dir(paste0(metadatadir, "/meta/ms6.out")))
    if (N.files.ms6.out > 0) {
      if (verbose == TRUE) print_console_header("Report part 6")
      if (N.files.ms6.out < f0) f0 = 1
      if (N.files.ms6.out < f1) f1 = N.files.ms6.out
      if (f1 == 0) f1 = N.files.ms6.out
      g.report.part6(metadatadir = metadatadir, f0 = f0, f1 = f1,
                     params_cleaning = params_cleaning, params_output = params_output,
                     verbose = verbose)
    }
  }
  if (params_output[["visualreport"]] == TRUE) {
    # g.plot uses metadata from part 1, 3, and 4
    files.basic = gsub("^meta_", "", dir(paste0(metadatadir, "/meta/basic"))) # without "meta_" to ease matching to data in ms3.out and ms4.out
    files.ms3.out = dir(paste0(metadatadir, "/meta/ms3.out"))
    files.ms4.out = dir(paste0(metadatadir, "/meta/ms4.out"))
    # at least one metafile from the same recording in each folder
    files.available = Reduce(intersect, list(files.basic, files.ms3.out, files.ms4.out))
    
    if (length(files.available) > 0) {
      if (verbose == TRUE) print_console_header("Generate visual reports")
      # The new visual report
      visualReport(metadatadir = metadatadir,
                   f0 = f0, f1 = f1,
                   verbose = verbose,
                   part6_threshold_combi = params_phyact[["part6_threshold_combi"]],
                   GGIRversion = GGIRversion,
                   params_sleep = params_sleep,
                   params_output = params_output,
                   params_general = params_general)
      if (params_output[["old_visualreport"]] == TRUE) {
        g.plot5(metadatadir = metadatadir,
                dofirstpage = params_output[["dofirstpage"]],
                viewingwindow = params_output[["viewingwindow"]],
                f0 = f0, f1 = f1,
                overwrite = params_general[["overwrite"]],
                metric = params_general[["acc.metric"]],
                desiredtz = params_general[["desiredtz"]],
                threshold.lig = params_phyact[["threshold.lig"]],
                threshold.mod = params_phyact[["threshold.mod"]],
                threshold.vig = params_phyact[["threshold.vig"]],
                visualreport_without_invalid = params_output[["visualreport_without_invalid"]],
                includedaycrit = params_cleaning[["includedaycrit"]][1],
                includenightcrit = params_cleaning[["includenightcrit"]],
                verbose = TRUE)
      }
    }
  }
}

updateBlocksize = function(blocksize = c(),
                           bsc_qc = data.frame(time = c(), size = c())) {
  if (length(blocksize) == 0) {
    warning("Blocksize is zero, please contact maintainers")
  }
  gco = gc()
  memuse = gco[2, 2] # memuse in mb
  bsc_qc_new_row = data.frame(
    time = format(Sys.time()),
    size = memuse,
    stringsAsFactors = FALSE
  )
  if (nrow(bsc_qc) == 0) {
    bsc_qc = bsc_qc_new_row
  } else {
    bsc_qc = rbind(bsc_qc,bsc_qc_new_row)
  }
  if (memuse > 4000) {
    if (nrow(bsc_qc) < 5) {
      blocksize = round(blocksize * 0.8)
    }
  }
  blocksize = round(blocksize)
  return(list(blocksize = blocksize, bsc_qc = bsc_qc))
}

library(GGIR)
context("updateBlocksize")
test_that("updateBlocksize works is able to recognize available memory and update blocksize", {
  currenttime = Sys.time()
  out = updateBlocksize(blocksize = 1000, 
                        bsc_qc=data.frame(time=as.character(currenttime), 
                                          size=2, stringsAsFactors = FALSE))
  ra = round(out$blocksize/100)
  eval = (ra == 10 | ra == 8)
  expect_equal(eval,TRUE)
})

g.weardec = function(metalong, wearthreshold, ws2, params_cleaning = NULL,
                     desiredtz = "", qwindowImp = c()) {
  nonWearEdgeCorrection = params_cleaning[["nonWearEdgeCorrection"]]
  nsi = which(colnames(metalong) == "nonwearscore")
  csi = which(colnames(metalong) == "clippingscore")
  NLongEpochs = nrow(metalong)
  clipsig = as.numeric(as.matrix(metalong[,csi]))
  # clipsig is a value between 0 and 1 for each long window (default 15 minutes) 
  # and indicates the ratio of datapoints in the window where signal clips
  LC = length(which(clipsig > 0.05))
  turnoffclip = which(clipsig > 0.3)
  LC2 = length(turnoffclip)
  turnoffnonw = which(as.numeric(as.matrix(metalong[,nsi])) >= wearthreshold)
  # step 1: identify islands of invalid data
  r1 = r2 = r3 = matrix(0, nrow(metalong), 1)
  r1[turnoffnonw] = 1 # non-weartime
  r2[turnoffclip] = 1 # clipping
  
  #----------------------------------------------------------------
  # Optionally ignore short lasting nonwear during the night
  if (!is.null(params_cleaning[["nonwearFiltermaxHours"]])) {
    fnn = filterNonwearNight(r1, metalong, qwindowImp, desiredtz, params_cleaning, ws2)
    r1[, 1] = fnn$r1
    nonwearHoursFiltered = fnn$nonwearHoursFiltered
    nonwearEventsFiltered = fnn$nonwearEventsFiltered
  } else {
    nonwearEventsFiltered = nonwearHoursFiltered = 0
  }
  r1 = c(0, r1, 0)
  r2 = c(0, r2, 0)
  r3 = c(0, r3, 0)
  additionalNonWearDetection = function(s1, s3, ws2, EdgeCorrection = FALSE) {
    ch1 = which(diff(s1) == 1) + 1 # indices of start wear
    ch2 = which(diff(s1) == -1) + 1 # indiced of start non-wear)
    if (length(ch1) > 1) { # there are at least 2 non-wear periods
      wear = matrix(0, (length(ch1) - 1), 3) #characteristics of all wear period in between non-wear periods
      for (weari in 1:(length(ch1) - 1)) {
        wear[weari,1] = abs(ch1[weari + 1] - ch2[weari]) # length of wear
        wear[weari,2] = abs(ch2[weari + 1] - ch1[weari + 1]) # length of non-wear after it
        wear[weari,3] = abs(ch2[weari] - ch1[weari]) # length of non-wear before it  
        if (wear[weari, 1] < (6 / (ws2 / 3600)) &
            (wear[weari, 1] / (wear[weari, 2] + wear[weari, 3])) < 0.3) {
          # turning wear into non-wear if it is less then 0.3 of nearby non-wear
          # and wear lasts for less than 6 hours (24 blocks)
          s3[ch2[weari]:(ch1[weari + 1] - 1)] = 1
        }
        if (wear[weari, 1] < (3 / (ws2 / 3600)) &
            (wear[weari, 1] / (wear[weari, 2] + wear[weari, 3])) < 0.8) {
          # turning wear into non-wear if it is less then 0.8 of nearby non-wear
          # and wear lasts for less than 3 hours (12 blocks)
          s3[ch2[weari]:(ch1[weari + 1] - 1)] = 1
        }
        if (ch1[weari] > (NLongEpochs - (24 / (ws2 / 3600)))) {
          # for last 24 hours of the data
          if (wear[weari, 1] < (3 / (ws2 / 3600)) &
              wear[weari, 3] > (1 / (ws2 / 3600))) {
            # only if wear is shorter than 3 hours and non-wear before it > 1 hour
            s3[ch2[weari]:(ch1[weari + 1] - 1)] = 1
          }
        }
      }
    }
    if (nonWearEdgeCorrection == TRUE) {
      if (length(ch1) > 0) {
        if (ch1[1] < (3 / (ws2 / 3600)) &
            ch1[1] > 1) {
          # did first non-wear period started after the start of the measurement 
          # but before the start of the fourth measurement hour? If yes then 
          # consider entire first three hours as non-wear (probably monitor set
          # up and then put aside)
          s3[1:(ch1[1] - 1)] = 1
        }
        if (ch2[length(ch2)] > (length(s3) - (3 / (ws2 / 3600))) &
            ch2[length(ch2)] != length(s3)) {
          # did measurement not end with non-wear but was there a period of 
          # non-wear that ended in the last 3 hours of the measurement? If yes, 
          # then consider entire final 3 hours as nonwear (probably monitor download)
          s3[ch2[length(ch2)]:length(s3)] = 1
        }
      }
    }
    invisible(list(r1 = s1, r3 = s3))
  }
  NEWlabels = additionalNonWearDetection(s1 = r1, s3 = r3, ws2, EdgeCorrection = nonWearEdgeCorrection)
  r1 = NEWlabels$r1
  r3 = NEWlabels$r3
  r1 = r1[-c(1, length(r1))]
  r2 = r2[-c(1, length(r2))]
  r3 = r3[-c(1, length(r3))]
  #===========================================================================================================
  #===========================================================================================================
  for (i in 1:2) { #for 2 additional iterative phases
    # redo based on newly identified blocks and remove shorter periods of 
    # non-wear surround by low acceleration
    r1b = r3 + r1
    r1b[which(r1b > 1)] = 1
    r1b = c(0, r1b, 0)
    r3b = c(0, r3, 0)
    NEWlabels = additionalNonWearDetection(s1 = r1b, s3 = r3b, ws2, EdgeCorrection = FALSE)
    r1b = NEWlabels$r1
    r3b = NEWlabels$r3
    r3 = r3b[-c(1, length(r3b))]
  }
  invisible(list(
    r1 = r1,
    r2 = r2,
    r3 = r3,
    LC = LC,
    LC2 = LC2,
    nonwearHoursFiltered = nonwearHoursFiltered,
    nonwearEventsFiltered = nonwearEventsFiltered
  ))
}

datadir2fnames = function(datadir,filelist) {
  if (filelist == FALSE) {
    fnamesfull = c(dir(datadir, recursive = TRUE, ignore.case = TRUE, pattern = "[.]csv", full.names = TRUE),
                   dir(datadir, recursive = TRUE, ignore.case = TRUE, pattern = "[.]bin", full.names = TRUE),
                   dir(datadir, recursive = TRUE, ignore.case = TRUE, pattern = "[.]wav", full.names = TRUE),
                   dir(datadir, recursive = TRUE, ignore.case = TRUE, pattern = "[.]cwa", full.names = TRUE),
                   dir(datadir, recursive = TRUE, ignore.case = TRUE, pattern = "[.]gt3", full.names = TRUE))
    fnames = basename(fnamesfull)
    fnamesRD = dir(datadir,recursive = TRUE, pattern = "[.]RD", full.names = TRUE)
    if (length(fnames) == length(fnamesRD)) { #because filenames may have both .bin in the middle and .RData
      fnames = basename(fnamesRD)
      fnamesfull = fnamesRD
    }
  } else {
    if (ismovisens(datadir)) {
      fnamesfull = dir(datadir, recursive = TRUE, pattern = "acc.bin$", full.names = TRUE)
      foldersWithAccBin = dirname(fnamesfull)
      fnames = basename(foldersWithAccBin)

      allfolders = list.dirs(datadir, recursive=FALSE, full.names = TRUE)

      # are there any folders that don't contain acc.bin (directly, or in any subfolders)?
      noAccBin = c()
      for (fld in allfolders) {
        if (fld %in% foldersWithAccBin) next # folder contains acc.bin

        # do any subfolders contain acc.bin?
        if (!any(grepl(paste(fld, '/', sep = ''), foldersWithAccBin))) {
          noAccBin = c(noAccBin, fld)
        }
      }
      
      if (length(noAccBin) > 0) {
        warning(paste0("The following movisens data folders do not contain the ",
                       "acc.bin file with the accelerometer recording, and ",
                       "therefore cannot be processed in GGIR: ",
                       paste(noAccBin, collapse = ", ")), call. = FALSE)
      }
    } else {
      fnamesfull = datadir
      fnames = basename(fnamesfull)
    }
  }
  invisible(list(fnames = fnames, fnamesfull = fnamesfull))
}

checkMilestoneFolders = function(metadatadir, partNumber) {
  if (partNumber == 1) return()
  # This function checks whether expected output folder(s) exists
  # If not, it creates them
  # Further it check whether folders with expected content are empty.
  # If yesm it gives error.
  paths = c(paste0(metadatadir, "/meta/basic"),
            paste0(metadatadir, "/meta/ms2.out"),
            paste0(metadatadir, "/meta/ms3.out"),
            paste0(metadatadir, "/meta/ms4.out"),
            paste0(metadatadir, "/meta/ms5.out"))
  sleepqc = paste0(metadatadir, "/meta/sleep.qc")
  
  # Make sure folders exists
  for (i in 1:partNumber) {
    if (!dir.exists(paths[i])) {
      dir.create(file.path(paths[i]), recursive = TRUE)
    }
  }
  if (partNumber >= 3) {
    if (!dir.exists(sleepqc)) {
      dir.create(file.path(sleepqc), recursive = TRUE)
    }
  }
  
  # Make sure preceding parts have files
  if (partNumber == 2) {
    expectedParts = 1
  } else if (partNumber == 3) {
    expectedParts = 2
  } else if (partNumber == 4) {
    expectedParts = 3
  } else if (partNumber == 5) {
    expectedParts = c(2, 3, 4)
  }
   
  # Give error when no data was found in the expected parts
  warnAbout = NULL
  for (i in expectedParts) {
    N = length(dir(paths[i]))
    if (N == 0) {
      warnAbout = c(warnAbout, i)
    }
  }
  if (length(warnAbout) > 0) {
    stop(paste0("\nNo milestone data found for part(s) ", paste0(warnAbout, collapse = " and "),
                ". Run this/these first before running part ", partNumber, "."), call. = FALSE)
  }
}

library(GGIR)
context("checkMilestoneFolders")
test_that("checkMilestoneFolders creates expected folders and gives expected warnings", {
  test_folder = "test_milestone_folders"
  dir.create(test_folder)
  expect_error(checkMilestoneFolders(test_folder, partNumber = 5))
  expect_true(dir.exists(paste0(test_folder, "/meta/ms5.out")))
  if (dir.exists(test_folder)) unlink(test_folder, recursive = TRUE)
})

g.imputeTimegaps = function(x, sf, k=0.25, impute = TRUE, 
                            PreviousLastValue = c(0, 0, 1), PreviousLastTime = NULL, 
                            epochsize = NULL) {
  if (!is.null(epochsize)) {
    shortEpochSize = epochsize[1]
    longEpochSize = epochsize[2]
  }
  # dummy variables to control the process
  remove_time_at_end = FirstRowZeros = imputelast = FALSE
  # initialize numberofgaps and GapsLength
  NumberOfGaps = GapsLength = 0

  # add temporary timecolumn to enable timegap imputation where there are zeros
  if (!("time" %in% colnames(x))) { 
    x$time = seq(from = Sys.time(), by = 1/sf, length.out = nrow(x))
    remove_time_at_end = TRUE
  }
  
  xyzCol = which(colnames(x) %in% c("x", "y", "z"))

  # define function for imputation at raw level
  imputeRaw = function(x, sf) {
    # impute raw timestamps because timestamps still need to be meaningful when
    # resampling or when plotting
    gapp = which(x$gap != 1)
    if (length(gapp) > 0) {
      if (gapp[1] > 1) {
        newTime = x$time[1:(gapp[1] - 1)]
      } else {
        newTime = NULL
      }
      for (g in 1:length(gapp)) {
        newTime = c(newTime, x$time[gapp[g]] + seq(0, by = 1/sf, length.out = x$gap[gapp[g]]))
        if (g < length(gapp)) {
          newTime = c(newTime, x$time[(gapp[g] + 1):(gapp[g + 1] - 1)])
        }
      }
      newTime =  c(newTime, x$time[(gapp[g] + 1):length(x$time)])
    }
    x <- as.data.frame(lapply(x, rep, x$gap))
    
    if (length(gapp) > 0) {
      x$time = newTime[1:nrow(x)]
    }
    x = x[, which(colnames(x) != "gap")]
    return(x)
  }
  
  # find zeros and remove them from dataset
  zeros = which(x$x == 0 & x$y == 0 & x$z == 0)
  if (length(zeros) > 0) {
    # if first value is a zero, remember value from previous chunk to replicate
    # if chunk = 1, then it will use c(0, 0, 1)
    if (zeros[1] == 1) {
      zeros = zeros[-1]
      x[1, xyzCol] = PreviousLastValue
      FirstRowZeros = TRUE
    }
    # if last value is a zero, we should not remove it (to keep track of the time)
    # This last row of zeros will be imputed afterwards (i.e., imputelast = TRUE)
    if (length(zeros) > 0) { # safe check just in case that removing zeros[1] have made length(zeros) == 0
      if (zeros[length(zeros)] == nrow(x)) {
        zeros = zeros[-length(zeros)]
        imputelast = TRUE
      }
    }
    # Again, check that length(zeros) is still > 0 
    if (length(zeros) > 0) x = x[-zeros,]
  }
  # find missing timestamps (timegaps)
  if (impute == TRUE) { # this is default, in g.calibrate this is set to FALSE
    if (k < 2/sf) { # prevent trying to impute timegaps shorter than 2 samples
      k = 2/sf
    }
    deltatime = diff(x$time)
    if (!is.numeric(deltatime)) {  # in csv axivity, the time is directly read as numeric (seconds)
      units(deltatime) = "secs"
      deltatime = as.numeric(deltatime)
    }
    # refill if first value is not consecutive from last value in previous chunk
    if (!is.null(PreviousLastTime)) {
      if (!inherits(x = PreviousLastTime, what = "numeric") &&
          inherits(x = x$time[1], what = "numeric")) {
        PreviousLastTime = as.numeric(PreviousLastTime)
      }
      first_deltatime = diff(c(PreviousLastTime, x$time[1]))
      if (!is.numeric(first_deltatime)) {  # in csv axivity, the time is directly read as numeric (seconds)
        units(first_deltatime) = "secs"
        first_deltatime = as.numeric(first_deltatime)
      }
      if (first_deltatime >= k) { # don't impute a timegap shorter than the minimum requested
        x = rbind(x[1,], x)
        x$time[1] = PreviousLastTime
        x[1, xyzCol] = PreviousLastValue
        deltatime = c(first_deltatime, deltatime)
      }
    }
    # impute time gaps
    gapsi = which(deltatime >= k) # limit imputation to gaps longer than the minimum requested
    NumberOfGaps = length(gapsi)
    if (NumberOfGaps > 0) {
      x$gap = 1
      x$gap[gapsi] = round(deltatime[gapsi] * sf)   # as.integer was problematic many decimals close to wholenumbers (but not whole numbers) resulting in 1 row less than expected
      GapsLength = sum(x$gap[gapsi])
      #  normalisation to 1 G 
      normalise = which(x$gap > 1)
      en_lastknownvalue = sqrt(rowSums(x[normalise, xyzCol]^2))
      # only if it deviates more than 5 mg from 1 G
      good_to_normalise = which((abs(en_lastknownvalue) - 1) > 0.005)
      if (length(good_to_normalise) > 0) {
        x[normalise[good_to_normalise], xyzCol] = x[normalise[good_to_normalise], xyzCol] / en_lastknownvalue[good_to_normalise]
      }
      imputation_done = FALSE
      if (!is.null(epochsize)) {
        # identify gaps larger than 6 long epochs (defaults to 90 minutes) or
        # 90 minutes, where the highest of the two is the criterium
        GapLimit =  max(c(((longEpochSize / 60) * 6), 90)) * 60 * sf
        gap90 = ifelse(x$gap > GapLimit, x$gap, 1) # keep track of gaps > 90 min
        gap90i = which(gap90 > 1)
        if (length(gap90i) > 0) {
          # if gap > 90 min, then impute only to fill up the long epoch
          # and keep track of how many epochs to impute
          x$remaining_epochs = 1
          x$next_epoch_delay = 0
          longEpochDayCut = seq(0, 24 * 60^2, by = longEpochSize)
          x$imputation = 0; imp = 0 # keep track of the imputation to organize data later on
          Rversion_lt_420 = as.numeric(R.Version()$major) >= 4 && as.numeric(R.Version()$minor) > 2
          for (i in gap90i) { 
            imp = imp + 1
            if (Rversion_lt_420) {
              time_i = x$time[i]
              time_ip1 = x$time[i + 1]
            } else {
              # Further down data.table is used to extract hour and minute from time stamps
              # In R 4.2.0 and earlier numeric time cannot be provided to data.table
              # For those older version of R convert timestamp format
              # Tz is set to GMT because at this point we do not care about the specific time
              # as time is removed at the end of the function, we only use it
              # to identify time gaps and their size.
              time_i = as.POSIXct(x$time[i], origin = "1970-01-01", tz = "GMT")
              time_ip1 = as.POSIXct(x$time[i + 1], origin = "1970-01-01", tz = "GMT")
            }
            # short epochs to add to fill up to next long epoch cut
            seconds = data.table::hour(time_i) * 60^2 + data.table::minute(time_i) * 60 + data.table::second(time_i) + 1 
            seconds_from_prevCut = seconds - max(longEpochDayCut[which(longEpochDayCut <= seconds)])
            shortEpochs2add_1 = (longEpochSize - seconds_from_prevCut) / shortEpochSize
            # short epochs to add after time gap
            seconds = data.table::hour(time_ip1) * 60^2 + data.table::minute(time_ip1) * 60 + data.table::second(time_ip1) + 1 
            seconds_from_prevCut = seconds - max(longEpochDayCut[which(longEpochDayCut <= seconds)])
            shortEpochs2add_2 = (seconds_from_prevCut - 1) / shortEpochSize
            # short epochs to add - total
            shortEpochs2add = shortEpochs2add_1 + shortEpochs2add_2
            time2add = (shortEpochs2add * sf * shortEpochSize) + 1
            x$remaining_epochs[i] = ((x$gap[i] - time2add) / (sf * shortEpochSize)) + 1   # plus 1 to count current epoch
            x$gap[i] = time2add # redefine gap to only fill up to next epoch
            x$imputation[i] = imp
            # if remaining_epochs[i] has decimal places, there is part of the next 
            # epoch (after the time gap) that we also should impute now.
            decs = x$remaining_epochs[i] - floor(x$remaining_epochs[i])
            if (decs > 0) {
              x$next_epoch_delay[i] = (decs * (sf * shortEpochSize))
              x$gap[i] = x$gap[i] +  x$next_epoch_delay[i] # redefine gap to fill up the first epoch after the gap
              x$remaining_epochs[i] = x$remaining_epochs[i] - (x$next_epoch_delay[i] / (sf * shortEpochSize)) 
            } 
          }
          x$gap = round(x$gap) # to make sure that small decimals do not mess up the imputation in next line
          x$next_epoch_delay = round(x$next_epoch_delay)
          
          x = imputeRaw(x, sf)

          imputation_done = TRUE
          # when imputing, the track of remaining_epochs has been repeated through the data frame
          # let's keep only the last record for the imputation later on
          keep_remaining_epochs = data.frame(index = which(x$remaining_epochs > 1),
                                             delay = x$next_epoch_delay[which(x$remaining_epochs > 1)],
                                             imp = x$imputation[which(x$remaining_epochs > 1)])
          keep_remaining_epochs$index = keep_remaining_epochs$index - keep_remaining_epochs$delay
          points2keep = aggregate(index ~ imp, data = keep_remaining_epochs, FUN = max)
          points2keep = points2keep$index
          # when turn_to_one is 1, then the record in x should also be 1 (this way we only keep the last record)
          x$remaining_epochs[-c(points2keep)] = 1 
          # cleanup x, we only need remaining epochs
          x = x[, which(colnames(x) != "next_epoch_delay")]
          x = x[, which(colnames(x) != "imputation")]
        }
      }
      if (imputation_done == FALSE) {
        x = imputeRaw(x, sf)
      }
    }
  } else if (impute == FALSE) {
    if (FirstRowZeros == TRUE) x = x[-1,] # since zeros[1] was removed
    if (imputelast == TRUE) x = x[-nrow(x),] # since zeros[length(zeros)] was removed
  }
  # impute last value?
  if (imputelast) x[nrow(x), xyzCol] = x[nrow(x) - 1, xyzCol]

  # QClog
  start = as.numeric(as.POSIXct(x$time[1], origin = "1970-1-1"))
  end = start + nrow(x)
  imputed = NumberOfGaps > 0
  QClog = data.frame(imputed = imputed, 
                     start = start, end = end,
                     blockLengthSeconds = (end - start) / sf,
                     timegaps_n = NumberOfGaps, timegaps_min = GapsLength/sf/60)

  if (remove_time_at_end == TRUE) {
    x = x[, grep(pattern = "time", x = colnames(x), invert = TRUE)]
  }

  # return data and QClog
  return(list(x = x, QClog = QClog))
}

library(GGIR)
context("g.imputeTimegaps")
test_that("timegaps are correctly imputed", {
  N = 10000
  sf = 20
  x = data.frame(time = as.POSIXct(x = (1:N)/sf, tz = "", origin = "1970/1/1"),
                 x = 1:N, y = 1:N, z = 1:N)
  xyzCol = c("x", "y", "z")
  
  x_without_time = data.frame(x = 1:N, y = 1:N, z = 1:N)
  xyzCol = c("x", "y", "z")
  
  # Duration of each consecutive gap is equal to the distance netween 
  # the sample right before and the sample right after the zeros that got removed to create this gap. 
  # So the duration of each gap is equal to the number of zeros + 1.
  ngaps = 4
  zeros = c(5:200, 6000:6500, 7000:7500, 8000:9500)
  gaps_duration = length(zeros) + ngaps
  gaps_duration = gaps_duration/sf/60
  
  # TEST THAT SAME FILE WITH DIFFERENT FORMATS IS IMPUTED IN THE SAME WAY ----
  # Format 1: with timestamp & with timegaps (no zeroes, incomplete dataset)
  x1 = x[-zeros,]
  x1_imputed = g.imputeTimegaps(x1, sf = sf, k = 2/sf, impute = TRUE, PreviousLastValue = c(0,0,1))
  x1_imputed_QClog = x1_imputed$QClog; x1_imputed = x1_imputed$x
  x1_removed = g.imputeTimegaps(x1, sf = sf, k = 2/sf, impute = FALSE, PreviousLastValue = c(0,0,1))
  x1_removed_QClog = x1_removed$QClog; x1_removed = x1_removed$x

  # make sure the timestamps got imputed correctly
  # (here we are checking that the last imputed timestamp is correct relative to the first timestamp after the gap)
  expect_equal(as.numeric(x1_imputed$time[201] - x1_imputed$time[200]), 1/sf, tolerance = .01, scale = 1)

  # Format 2: with timestamp & with zeros (complete dataset)
  x2 = x
  x2[zeros, xyzCol] = 0
  x2_imputed = g.imputeTimegaps(x2, sf = sf, k = 2/sf, impute = TRUE, PreviousLastValue = c(0,0,1))
  x2_imputed_QClog = x2_imputed$QClog; x2_imputed = x2_imputed$x
  x2_removed = g.imputeTimegaps(x2, sf = sf, k = 2/sf, impute = FALSE, PreviousLastValue = c(0,0,1))
  x2_removed_QClog = x2_removed$QClog; x2_removed = x2_removed$x
  
  # Format 3: without timestamp & with zeros (complete dataset)
  x3 = x_without_time
  x3[zeros, xyzCol] = 0
  x3_imputed = g.imputeTimegaps(x3, sf = sf, k = 2/sf, impute = TRUE, PreviousLastValue = c(0,0,1))
  x3_imputed_QClog = x3_imputed$QClog; x3_imputed = x3_imputed$x
  x3_removed = g.imputeTimegaps(x3, sf = sf, k = 2/sf, impute = FALSE, PreviousLastValue = c(0,0,1))
  x3_removed_QClog = x3_removed$QClog; x3_removed = x3_removed$x
  
  # tests number of rows
  expect_equal(nrow(x1_imputed), N)
  expect_equal(nrow(x2_imputed), N)
  expect_equal(nrow(x3_imputed), N)
  
  expect_equal(nrow(x1_removed), N - length(zeros))
  expect_equal(nrow(x2_removed), N - length(zeros))
  expect_equal(nrow(x3_removed), N - length(zeros))
  
  # test imputations on different formats worked identically
  expect_equal(x1_imputed$X, x2_imputed$X)
  expect_equal(x1_imputed$X, x3_imputed$X)
  
  expect_equal(x1_removed$X, x2_removed$X)
  expect_equal(x1_removed$X, x3_removed$X)
  
  # test QClog
  expect_equal(x1_imputed_QClog$timegaps_n, 4)
  expect_equal(x2_imputed_QClog$timegaps_n, 4)
  expect_equal(x3_imputed_QClog$timegaps_n, 4)
  
  expect_equal(x1_imputed_QClog$timegaps_min, gaps_duration)
  expect_equal(x2_imputed_QClog$timegaps_min, gaps_duration)
  expect_equal(x3_imputed_QClog$timegaps_min, gaps_duration)
  
  # TEST IMPUTATION WHEN FIRST ROW IS NOT CONSECUTIVE TO PREVIOUS CHUNK ----
  # Format 4: with timestamp & with timegaps (no zeroes, incomplete dataset)
  x4 = x[-zeros,]
  PreviousLastTime = x[1,"time"] - 30 # dummy gap of 30 seconds between chunks
  suppressWarnings({ # warning arising from made up PreviousLastTime
    x4_imputed = g.imputeTimegaps(x4, sf = sf, k = 2/sf, impute = TRUE, 
                                  PreviousLastValue = c(0,0,1), PreviousLastTime = PreviousLastTime)
    x4_imputed_QClog = x4_imputed$QClog; x4_imputed = x4_imputed$x
    x4_removed = g.imputeTimegaps(x4, sf = sf, k = 2/sf, impute = FALSE, 
                                  PreviousLastValue = c(0,0,1), PreviousLastTime = PreviousLastTime)
    x4_removed_QClog = x4_removed$QClog; x4_removed = x4_removed$x
  })
  
  expect_equal(nrow(x4_imputed), N + sf*30)
  expect_equal(nrow(x4_removed), N - length(zeros))
  
  # TEST IMPUTATION WHEN FIRST AND LAST ROW CONTAIN ZEROS ----
  zeros = c(1:200, 6000:6500, 7000:7500, 8000:10000)
  # Format 5: with timestamp & with zeros (complete dataset)
  x5 = x
  x5[zeros, xyzCol] = 0
  x5_imputed = g.imputeTimegaps(x5, sf = sf, k = 2/sf, impute = TRUE, PreviousLastValue = c(0,0,1))
  x5_imputed_QClog = x5_imputed$QClog; x5_imputed = x5_imputed$x
  x5_removed = g.imputeTimegaps(x5, sf = sf, k = 2/sf, impute = FALSE, PreviousLastValue = c(0,0,1))
  x5_removed_QClog = x5_removed$QClog; x5_removed = x5_removed$x
  
  expect_equal(nrow(x5_imputed), N)
  expect_equal(nrow(x5_removed), N - length(zeros))
  
})
 

parametersVignette = function(params = "sleep") {
  
  if (!grepl("params_", params)) params = paste0("params_", params)
  arguments = names(load_params()[[params]])
  vignette = character()
  for (i in 1:length(arguments)) {
    vignette = paste0(vignette, "\n### ", arguments[i], " {-}\n",
                      ShellDoc2Vignette(arguments[i]), "\n\n ")
  }
  return(vignette)
}

g.convert.part2.long = function(daySUMMARY) {
  # local functions:
  extractlastpart = function(x) { 
    # function extract windowsegment specification which 
    # is at the end of the variable name
    tmp = unlist(strsplit(x,"_"))
    if (length(tmp) > 1) {
      tmp = tmp[length(tmp)]
      tmp2 = unlist(strsplit(tmp,"-"))
      if (length(tmp2) < 2) tmp = ""
    } else {
      tmp = ""
    }
    if (tmp != "") {
      out = c(unlist(strsplit(x, paste0(tmp,"_")))[1], tmp)
    } else {
      out = c("","")
    }
    return(out)
  }
  convert2minutes = function(x) {
    x = as.numeric(x)
    hour = floor(x)
    minutes = round((x - floor(x)) * 60)
    if (nchar(minutes) == 1) minutes = paste0("0", minutes)
    if (nchar(hour) == 1) minutes = paste0("0", hour)
    time = paste0(hour, ":", minutes)
    return(time)
  }
  
  convert2clocktime = function(x) {
    x2 = unlist(strsplit(x, "hr"))
    if (length(x2) > 0) {
      x3 = as.numeric(unlist(strsplit(x2[1], "-")))
      myfun = function(y) {
        for (i in 1:length(y)) {
          if (y[i] < 10) {
            y[i] = paste0(0, y[i])
          } else {
            y[i] = as.character(y[i])
          }
        }
        return(y)
      }
      x4 = paste0(paste0(myfun(x3), ":00"), collapse = "-")
    } else {
      x4 = ""
    }
    return(x4)
  }
  #------------------------------
  # main code
  # replace spaces by underscores, because melt function struggles with it later on
  colnames(daySUMMARY) = gsub(pattern = " ",
                              replacement = "_",
                              x = colnames(daySUMMARY))
  cn = names(daySUMMARY)
  # extract windowsegment specification from variable name
  tt = sapply(cn, FUN = extractlastpart)
  id.vars = colnames(tt[, which(tt[2,] == "")]) # identify other variables, including ID
  daySUMMARY = data.table::as.data.table(daySUMMARY)
  # turn into long format with melt, this will make it too long, so further
  # down we will move key variables to wide again
  daySUMMARY2 = data.table::melt(daySUMMARY, id.vars = id.vars, 
                                 measure.vars = colnames(tt)[which(tt[2, ] != "")])
  # extract variable names
  daySUMMARY2$variable2 = sub("_[^_]+$", "", daySUMMARY2$variable)
  # extract time segment names
  daySUMMARY2$timesegment2 = sub("^.*_", "", daySUMMARY2$variable)
  daySUMMARY2 = as.data.frame(daySUMMARY2)
  # tidy up
  rem = which(colnames(daySUMMARY2) == "variable")
  # It is a data.table object so column removal works different
  daySUMMARY2 = daySUMMARY2[, -rem]
  id.vars = c(id.vars, "timesegment2")
  cnt = 1
  # long format that is undersirable is now moved back to wide format
  for (i in unique(daySUMMARY2$variable2)) {
    if (cnt == 1) {
      df = daySUMMARY2[which(daySUMMARY2$variable2 == i),]
      colnames(df)[which(colnames(df) == "value")] = i
      rem = which(colnames(df) == "variable2")
      df = df[,-rem]
    } else {
      df = base::merge(x = df,
                       y = daySUMMARY2[which(daySUMMARY2$variable2 == i),],
                       by = id.vars, all.x = T, sort = F)
      colnames(df)[which(colnames(df) == "value")] = i
      rem = which(colnames(df) == "variable2")
      df = df[, -rem]
    }
    cnt = cnt + 1
  }
  # tidy up variable names
  Nvh.1 = which(colnames(df) == "N_valid_hours.1")
  Nh.1 = which(colnames(df) == "N_hours.1")
  if (length(Nvh.1) > 0) colnames(df)[Nvh.1] = "N_valid_hours_in_window"
  if (length(Nh.1) > 0) colnames(df)[Nh.1] = "N_hours_in_window"
  df$N_valid_hours_in_window[which(df$timesegment2 == "0-24hr")] = ""
  df$N_hours_in_window[which(df$timesegment2 == "0-24hr")] = ""
  # re-order columns to be more logical
  col2move = which(colnames(df) %in% c("N_valid_hours_in_window", "N_hours_in_window") == TRUE)
  col2NH = which(colnames(df) %in% c("N_valid_hours", "N_hours") == TRUE)
  df = df[,c(1:max(col2NH), col2move, (max(col2NH) + 1):(ncol(df) - 2))]
  # re-place underscore to hyphen
  skip.qwindow_name.conversion = FALSE
  if (all(df$qwindow_timestamps == df$qwindow_names)) { # using qwindow vector without diary
    df$qwindow_timestamps = sapply(X = df$timesegment2, FUN = convert2clocktime)
    skip.qwindow_name.conversion = TRUE
  }
  
  rplc = which(df$qwindow_timestamps == "00:00 24:00")
  if (length(rplc) > 0) {
    df$qwindow_timestamps[rplc] = "00:00-24:00"
  }
  # the segment names are not correct yet, rename them
  for (ji in unique(df$ID)) {
    for (hi in unique(df$calendar_date[which(df$ID == ji)])) {
      df_tmp = df[which(df$ID == ji & df$calendar_date == hi),
                  c("qwindow_timestamps", "qwindow_names", "timesegment2")]
      tms = unlist(strsplit(df_tmp$qwindow_timestamps[1],"_"))
      nms = unlist(strsplit(df_tmp$qwindow_names[1],"_"))
      namekey = matrix("",length(tms),2)
      for (gi in 1:(length(tms))) {
        if (gi == 1) {
          options(warn = -1)
          qwindownumeric = !is.na(as.numeric(tms[1]))
          options(warn = 0)
          if (qwindownumeric == FALSE) {
            namekey[1,] = c("00:00-24:00", "0-24hr")
          } else { # if qwindow is numeric then store as such to be consistent with other output
            namekey[1, ] = c("00:00-24:00", "0-24hr")
          }
        } else {
          if (qwindownumeric == FALSE) {
            namekey[gi, ] = c(paste0(tms[gi - 1], "-", tms[gi]),
                              paste0(nms[gi - 1], "-", nms[gi], "hr"))
          } else {
            namekey[gi, ] = c(paste0(convert2minutes(tms[gi - 1]),
                                     "-", convert2minutes(tms[gi])), 
                              paste0(nms[gi - 1], "-", nms[gi],"hr"))
          }
        }
        qwt_index = which(as.character(df_tmp$timesegment2) == namekey[gi,2])
        if (length(qwt_index) > 0) {
          df_tmp$qwindow_timestamps[qwt_index] = namekey[gi,1]
        } else {
          df_tmp$qwindow_timestamps[qwt_index] = "not_calculated"
        }
      }
      df[which(df$ID == ji & df$calendar_date == hi),
         c("qwindow_timestamps", "qwindow_names", "timesegment2")] = df_tmp
    }
  }
  df = df[, -which(colnames(df) %in% c("qwindow_names"))]
  
  first_feature_col = which(colnames(df) == "timesegment2") + 1
  if (first_feature_col <= ncol(df)) {
    emptystring = function(x) all(x == "")
    redundant_rows = which(df$N_valid_hours_in_window == "" &
                             df$N_hours_in_window  == "" & 
                             apply(df[first_feature_col:ncol(df)], 1, FUN = emptystring) |
                             df$qwindow_timestamps == "not_calculated")
    if (length(redundant_rows) > 0) df = df[-redundant_rows,]
  }
  colnames(df)[which(colnames(df) == "timesegment2")] = "qwindow_name"
  if (skip.qwindow_name.conversion == FALSE) {
    if (qwindownumeric == FALSE) {
      df$qwindow_name[which(df$qwindow_name == "0-24hr")] = "daystart-dayendhr"
    }
    # remove last two characters, because they are hr
    df$qwindow_name = substr(df$qwindow_name, 1, nchar(df$qwindow_name) - 2)
  }
  return(df)
}

library(GGIR)
context("convert.part2.long")
test_that("converts_part2_output_to_long_format", {
  skip_on_cran()
  varnames = c("ID", "filename", "calendar_date", "bodylocation", "N valid hours",
               "N hours", "weekday", "measurementday", "qwindow_timestamps", 
               "qwindow_names", "L2hr_ENMO_mg_0-24hr", "L2_ENMO_mg_0-24hr",
               "M2hr_ENMO_mg_0-24hr", "M2_ENMO_mg_0-24hr", "mean_ENMO_mg_1-6am", 
               "mean_ENMO_mg_0-24hr", 
               "p79.16667_ENMO_mg_0-24hr", "p75_ENMO_mg_0-24hr", "p70.83333_ENMO_mg_0-24hr",
               "p66.66667_ENMO_mg_0-24hr", "p91.66667_ENMO_mg_0-24hr", 
               "p95.83333_ENMO_mg_0-24hr", "p96.875_ENMO_mg_0-24hr", 
               "p97.91667_ENMO_mg_0-24hr", "p98.61111_ENMO_mg_0-24hr", 
               "p98.95833_ENMO_mg_0-24hr", 
               "p99.30556_ENMO_mg_0-24hr", "p99.65278_ENMO_mg_0-24hr", 
               "p99.86111_ENMO_mg_0-24hr", "p99.93056_ENMO_mg_0-24hr",
               "[0,50)_ENMO_mg_0-24hr", "[50,100)_ENMO_mg_0-24hr", 
               "[100,150)_ENMO_mg_0-24hr", "[150,200)_ENMO_mg_0-24hr",
               "ig_gradient_ENMO_0-24hr", "ig_intercept_ENMO_0-24hr", 
               "ig_rsquared_ENMO_0-24hr", "MVPA_E5S_T100_ENMO_0-24hr",
               "MVPA_E1M_T100_ENMO_0-24hr", "MVPA_E5M_T100_ENMO_0-24hr", 
               "MVPA_E5S_B1M80%_T100_ENMO_0-24hr", "MVPA_E5S_B5M80%_T100_ENMO_0-24hr",
               "MVPA_E5S_B10M80%_T100_ENMO_0-24hr", "MVPA_E5S_T400_ENMO_0-24hr",
               "MVPA_E1M_T400_ENMO_0-24hr", "MVPA_E5M_T400_ENMO_0-24hr", 
               "MVPA_E5S_B1M80%_T400_ENMO_0-24hr", "MVPA_E5S_B5M80%_T400_ENMO_0-24hr", 
               "MVPA_E5S_B10M80%_T400_ENMO_0-24hr", "L2hr_ENMO_mg_daystart-Starthr",
               "L2_ENMO_mg_daystart-Starthr", "M2hr_ENMO_mg_daystart-Starthr", 
               "M2_ENMO_mg_daystart-Starthr", "mean_ENMO_mg_daystart-Starthr", 
               "p79.16667_ENMO_mg_daystart-Starthr", "p75_ENMO_mg_daystart-Starthr",
               "p70.83333_ENMO_mg_daystart-Starthr", "p66.66667_ENMO_mg_daystart-Starthr", 
               "p91.66667_ENMO_mg_daystart-Starthr", "p95.83333_ENMO_mg_daystart-Starthr",
               "p96.875_ENMO_mg_daystart-Starthr", "p97.91667_ENMO_mg_daystart-Starthr",
               "p98.61111_ENMO_mg_daystart-Starthr", "p98.95833_ENMO_mg_daystart-Starthr",
               "p99.30556_ENMO_mg_daystart-Starthr", "p99.65278_ENMO_mg_daystart-Starthr", 
               "p99.86111_ENMO_mg_daystart-Starthr", "p99.93056_ENMO_mg_daystart-Starthr",
               "[0,50)_ENMO_mg_daystart-Starthr", "[50,100)_ENMO_mg_daystart-Starthr", 
               "[100,150)_ENMO_mg_daystart-Starthr", "[150,200)_ENMO_mg_daystart-Starthr", 
               "ig_gradient_ENMO_daystart-Starthr", "ig_intercept_ENMO_daystart-Starthr", 
               "ig_rsquared_ENMO_daystart-Starthr", "MVPA_E5S_T100_ENMO_daystart-Starthr", 
               "MVPA_E1M_T100_ENMO_daystart-Starthr", "MVPA_E5M_T100_ENMO_daystart-Starthr",
               "MVPA_E5S_B1M80%_T100_ENMO_daystart-Starthr", "MVPA_E5S_B5M80%_T100_ENMO_daystart-Starthr",
               "MVPA_E5S_B10M80%_T100_ENMO_daystart-Starthr", "MVPA_E5S_T400_ENMO_daystart-Starthr",
               "MVPA_E1M_T400_ENMO_daystart-Starthr", "MVPA_E5M_T400_ENMO_daystart-Starthr",
               "MVPA_E5S_B1M80%_T400_ENMO_daystart-Starthr", "MVPA_E5S_B5M80%_T400_ENMO_daystart-Starthr",
               "MVPA_E5S_B10M80%_T400_ENMO_daystart-Starthr", "L2hr_ENMO_mg_Start-Endhr",
               "L2_ENMO_mg_Start-Endhr", "M2hr_ENMO_mg_Start-Endhr", "M2_ENMO_mg_Start-Endhr",
               "mean_ENMO_mg_Start-Endhr", "p79.16667_ENMO_mg_Start-Endhr",
               "p75_ENMO_mg_Start-Endhr", "p70.83333_ENMO_mg_Start-Endhr", 
               "p66.66667_ENMO_mg_Start-Endhr", "p91.66667_ENMO_mg_Start-Endhr",
               "p95.83333_ENMO_mg_Start-Endhr", "p96.875_ENMO_mg_Start-Endhr", 
               "p97.91667_ENMO_mg_Start-Endhr", "p98.61111_ENMO_mg_Start-Endhr", 
               "p98.95833_ENMO_mg_Start-Endhr", "p99.30556_ENMO_mg_Start-Endhr",
               "p99.65278_ENMO_mg_Start-Endhr", "p99.86111_ENMO_mg_Start-Endhr", 
               "p99.93056_ENMO_mg_Start-Endhr", "[0,50)_ENMO_mg_Start-Endhr", 
               "[50,100)_ENMO_mg_Start-Endhr", "[100,150)_ENMO_mg_Start-Endhr",
               "[150,200)_ENMO_mg_Start-Endhr", "ig_gradient_ENMO_Start-Endhr",
               "ig_intercept_ENMO_Start-Endhr", "ig_rsquared_ENMO_Start-Endhr", 
               "MVPA_E5S_T100_ENMO_Start-Endhr", "MVPA_E1M_T100_ENMO_Start-Endhr", 
               "MVPA_E5M_T100_ENMO_Start-Endhr", "MVPA_E5S_B1M80%_T100_ENMO_Start-Endhr",
               "MVPA_E5S_B5M80%_T100_ENMO_Start-Endhr", "MVPA_E5S_B10M80%_T100_ENMO_Start-Endhr", 
               "MVPA_E5S_T400_ENMO_Start-Endhr", "MVPA_E1M_T400_ENMO_Start-Endhr", 
               "MVPA_E5M_T400_ENMO_Start-Endhr", "MVPA_E5S_B1M80%_T400_ENMO_Start-Endhr",
               "MVPA_E5S_B5M80%_T400_ENMO_Start-Endhr", "MVPA_E5S_B10M80%_T400_ENMO_Start-Endhr",
               "L2hr_ENMO_mg_End-dayendhr", "L2_ENMO_mg_End-dayendhr", 
               "M2hr_ENMO_mg_End-dayendhr", "M2_ENMO_mg_End-dayendhr", 
               "mean_ENMO_mg_End-dayendhr", "p79.16667_ENMO_mg_End-dayendhr", 
               "p75_ENMO_mg_End-dayendhr", "p70.83333_ENMO_mg_End-dayendhr",
               "p66.66667_ENMO_mg_End-dayendhr", "p91.66667_ENMO_mg_End-dayendhr", 
               "p95.83333_ENMO_mg_End-dayendhr", "p96.875_ENMO_mg_End-dayendhr", 
               "p97.91667_ENMO_mg_End-dayendhr", "p98.61111_ENMO_mg_End-dayendhr",
               "p98.95833_ENMO_mg_End-dayendhr", "p99.30556_ENMO_mg_End-dayendhr", 
               "p99.65278_ENMO_mg_End-dayendhr", "p99.86111_ENMO_mg_End-dayendhr", 
               "p99.93056_ENMO_mg_End-dayendhr", "[0,50)_ENMO_mg_End-dayendhr",
               "[50,100)_ENMO_mg_End-dayendhr", "[100,150)_ENMO_mg_End-dayendhr",
               "[150,200)_ENMO_mg_End-dayendhr","ig_gradient_ENMO_End-dayendhr",
               "ig_intercept_ENMO_End-dayendhr", "ig_rsquared_ENMO_End-dayendhr", 
               "MVPA_E5S_T100_ENMO_End-dayendhr", "MVPA_E1M_T100_ENMO_End-dayendhr", 
               "MVPA_E5M_T100_ENMO_End-dayendhr", "MVPA_E5S_B1M80%_T100_ENMO_End-dayendhr",
               "MVPA_E5S_B5M80%_T100_ENMO_End-dayendhr", "MVPA_E5S_B10M80%_T100_ENMO_End-dayendhr",
               "MVPA_E5S_T400_ENMO_End-dayendhr", "MVPA_E1M_T400_ENMO_End-dayendhr",
               "MVPA_E5M_T400_ENMO_End-dayendhr", "MVPA_E5S_B1M80%_T400_ENMO_End-dayendhr", 
               "MVPA_E5S_B5M80%_T400_ENMO_End-dayendhr", "MVPA_E5S_B10M80%_T400_ENMO_End-dayendhr",
               "N_valid_hours_daystart-Starthr", "N_hours_daystart-Starthr",
               "N_valid_hours_Start-Endhr", "N_hours_Start-Endhr", "N_valid_hours_End-dayendhr",
               "N_hours_End-dayendhr")

  daySUMMARY = as.data.frame(matrix(1,15, length(varnames)))
  colnames(daySUMMARY) = varnames
  daySUMMARY$ID = rep(c(1:3), each=5)
  daySUMMARY$calendar_date = rep(c(1:5),times=3)
  daySUMMARY$qwindow_timestamps = rep(c("00:00_08:00_14:00_24:00"), times=15)
  daySUMMARY$qwindow_names = rep(c("daystart_Start_End_dayend"), times=15)
  df = g.convert.part2.long(daySUMMARY)
  # 3 individuals x 5 days x 4 segments (daystart-Start, Start-End, End-dayend, daystart-dayend)
  expect_that(nrow(df),equals(60)) 
  expect_that(ncol(df),equals(50)) # 50 unique variable are in daySUMMARY (above)
  expect_that(length(unique(df$ID)),equals(3))
})

extractID = function(hvars, idloc, fname) {
  
  ID = hvars$ID
  iID = hvars$iID
  IDd = hvars$IDd
  
  # This part is specific to how data Pelotas cohort was stored.
  # It is legacy code from 2012, for later projects I tried to avoid
  # ending up with dataset specific solutions
  ID2 = ID
  iID2 = iID
  if (idloc == 3) { #remove hyphen in id-name for Pelotas id-numbers
    get_char_before_hyphen = function(x) {
      x2 = c()
      for (j in 1:length(x)) {
        temp = unlist(strsplit(x,"-"))
        if (length(temp) == 2) {
          x2[j] = as.character(temp[1])
        } else {
          x2[j] = as.character(x[j])
        }
      }
      return(x2)
    }
    ID2 = get_char_before_hyphen(ID)
    iID2 = get_char_before_hyphen(iID)
  }
  ID_NAs = which(ID == "NA")
  ID2_NAs = which(ID == "NA")
  if (length(ID_NAs) > 0) ID[ID_NAs] = iID[ID_NAs]
  if (length(ID2_NAs) > 0) ID2[ID2_NAs] = iID2[ID2_NAs]
  if (idloc == 2) { # default is idloc=1, where ID just stays ID
    ID = unlist(strsplit(fname,"_"))[1]
  } else if (idloc == 3) {
    ID = ID2
  } else if (idloc == 4) {
    ID = IDd
  } else if (idloc == 5) {
    ID = unlist(strsplit(fname," "))[1]
  } else if (idloc == 6) {
    ID = unlist(strsplit(fname,"[.]"))[1]
  } else if (idloc == 7) {
    ID = unlist(strsplit(fname,"-"))[1]
  }
  if (length(ID) == 0) { # If ID could not be extracted
    ID = basename(fname)
    warning(paste0("\nUnable to extract ID from, ", fname, ". Using filname instead. ",
                   " You may want to check argument idloc, which is currently set to ", idloc))
  } else {
    ID = gsub(pattern = " ", replacement = "", ID)
  }
  return(ID)
}

appendRecords = function(metadatadir, desiredtz = "", idloc = 1, maxRecordingInterval = NULL) {
  
  # Declare local functions:
  mergePair = function(M1, M2, overlap, tz) {
    # overlap in epochs
    if (overlap < 0) { 
      # there is a gap
      fillgaps = function(df, overlap, epochSize, pattern) {
        if (overlap < 0) {
          N = abs(overlap) - ifelse(pattern == "long", yes = 0, no = 1)
        } else {
          N = overlap - 1
        }
        if (N > 0) {
          NEWSHORT = df[nrow(df), ]
          NEWSHORT[1:N, ] = NEWSHORT
          T0 = as.POSIXct(x = NEWSHORT$timestamp[1], format = "%Y-%m-%dT%H:%M:%S%z", tz = tz) + epochSize
          NEWSHORT$timestamp = POSIXtime2iso8601(seq(T0, T0 + (N * epochSize) - 1, by =  epochSize), tz)
          if (pattern != "long") {
            NEWSHORT[, grep(pattern = pattern, x = colnames(NEWSHORT), invert = TRUE, ignore.case = TRUE)] = 0
          } else {
            enCol = grep("en", colnames(NEWSHORT), ignore.case = TRUE)
            NEWSHORT[, enCol] = 1
            NEWSHORT$clippingscore = 0
            NEWSHORT$nonwearscore = 3
            if (any(colnames(NEWSHORT) %in% c("lightmean", "lightpeak"))) {
              NEWSHORT$lightmean = 0
              NEWSHORT$lightpeak = 0
            }
            if ("temperaturemean" %in% colnames(NEWSHORT)) {
              NEWSHORT$temperaturemean = 0
            }
          }
          df = rbind(df, NEWSHORT)
        }
        return(df)
      }
      # insert missing values in metashort
      M1$metashort = fillgaps(df = M1$metashort, overlap = round(overlap / M1$windows[1]),
                              epochSize = M1$windows[1], pattern = "time|angle")
      # insert missing values in metalong
      M1$metalong = fillgaps(df = M1$metalong,
                             overlap = round(overlap / M1$windows[2]),
                             epochSize = M1$windows[2], pattern = "long")
      if (length(which(duplicated(M1$metashort$timestamp) == TRUE) > 0) |
          length(which(duplicated(M1$metalong$timestamp) == TRUE) > 0)) {
        stop("\nDuplicated timestamps found, something may have gone wrong with appending the timeseries")
      }
      
    }
    
    # Append data.frames (yes, we may expect duplicates here, which are going to be dealt with)
    M1$metashort = rbind(M1$metashort, M2$metashort)
    M1$metalong = rbind(M1$metalong, M2$metalong)
    # Omit oldest duplicated timestamps
    M1$metashort = M1$metashort[rev(!duplicated(rev(M1$metashort$timestamp))),]
    M1$metalong = M1$metalong[rev(!duplicated(rev(M1$metalong$timestamp))),]
    return(M1)
  }
  #==================================================
  # Main code:
  filefoldername = filename_dir = tail_expansion_log = NULL
  # Create overview of all recordings ID, start time, end time, and filename
  fns = dir(paste0(metadatadir, "/meta/basic"), full.names = TRUE)
  
  S = do.call("rbind", lapply(X = fns, FUN = getPart1BasicInfo, idloc = idloc, tz = desiredtz)) 
  if (length(S) > 0 & !is.null(maxRecordingInterval)) {
    S = S[!is.na(S$ID),]
    S = S[order(S$ID, S$start), ]
    S$overlap = NA
    # Identify recordings that can be appended (ID and brand need to match)
    doubleID = unique(S$ID[duplicated(S[, c("ID", "brand")])])
    if (length(doubleID) > 0) {
      for (j in 1:length(doubleID)) {
        Mlist = Ilist = Clist = list()
        # Does time interval overlap?
        # Label the recordings that are groups, this can be more than 2 recordings
        rowsOfInterest = which(S$ID == doubleID[j])
        Nrec = length(which(S$ID == doubleID[j]))
        for (k in Nrec:2) {
          overlap = as.numeric(difftime(S$end[rowsOfInterest[k - 1]], S$start[rowsOfInterest[k]], units = "secs"))
          # positive overlap is true overlap, negative overlap is a gap
          # Add load number
          S$overlap[rowsOfInterest[k]] = overlap 
          load(S$filename[rowsOfInterest[k]])
          Mlist[[k]] = M
          Ilist[[k]] = I
          Clist[[k]] = C
        }
        # First recording
        load(S$filename[rowsOfInterest[1]])
        Mlist[[1]] = M
        Ilist[[1]] = I
        Clist[[1]] = C
        
        # Append all recordings if they are a sequence of nearby recordings
        cnt = 0
        for (k in (Nrec:1)) {
          if (!is.na(S$overlap[rowsOfInterest[k]])) {
            interval = S$overlap[rowsOfInterest[k]]
          } else {
            interval = -2e10
          }
          if (interval > -abs(maxRecordingInterval) * 3600 & k > 1) {
            # Merge this pair
            Mlist[[k - 1]] = mergePair(M1 = Mlist[[k - 1]],
                                       M2 =  Mlist[[k]],
                                       overlap = S$overlap[rowsOfInterest[k]],
                                       tz = desiredtz)
            interval = round(interval / 3600, digits = 3)
            if ("interval" %in% names(Ilist[[k]])) {
              Ilist[[k - 1]]$interval = c(Ilist[[k]]$interval, interval)
            } else {
              Ilist[[k - 1]]$interval = interval
            }
            # Remove duplicate files
            unlink(S$filename[rowsOfInterest[k]], recursive = TRUE)
            cnt = cnt + 1 # keep track of how many files were appended
          } else {
            if (cnt > 0) {
              # we need to have at least two recording appended to make it worth saving the data
              M = Mlist[[k]]
              C = Clist[[k]]
              I = Ilist[[k]]
              # save as new milestone data for part 1
              # Note that we also keep copy of I and C for each of the original recordings
              # this to retain this information for reporting in part 2
              Nappended = cnt
              C_list = Clist[c(k:(cnt + 1))]
              I_list = Ilist[c(k:(cnt + 1))]
              save(M, C, I, C_list, I_list,
                   filefoldername, filename_dir, tail_expansion_log, Nappended,
                   file = S$filename[rowsOfInterest[k]])
            }
            cnt = 0 # start again
          }
        }
      }
    }
  }
  # THIS FUNCTION DOES NOT PRODUCE OUTPUT IT ONLY APPENDS FILES
  return(NULL)
}

library(GGIR)
context("appendRecords")
#==========================================
# Create dummy data: 4 files from which the first 3 can be appended
timestamps = matrix("", 10, 4)
# Timstamps for file 1
timestamps[, 1] = c("2020-05-13T14:00:00+0200", "2020-05-13T14:01:00+0200",
                    "2020-05-13T14:02:00+0200", "2020-05-13T14:03:00+0200",
                    "2020-05-13T14:04:00+0200", "2020-05-13T14:05:00+0200",
                    "2020-05-13T14:06:00+0200", "2020-05-13T14:07:00+0200",
                    "2020-05-13T14:08:00+0200", "2020-05-13T14:09:00+0200")
# Timstamps for file 2: overlapping time
timestamps[, 2] = c("2020-05-13T14:05:00+0200", "2020-05-13T14:06:00+0200",
                    "2020-05-13T14:07:00+0200", "2020-05-13T14:08:00+0200",
                    "2020-05-13T14:09:00+0200", "2020-05-13T14:10:00+0200",
                    "2020-05-13T14:11:00+0200", "2020-05-13T14:12:00+0200",
                    "2020-05-13T14:13:00+0200", "2020-05-13T14:14:00+0200")
# Timstamps for file 3: non-overlapping time but small gap
timestamps[, 3] = c("2020-05-13T14:30:00+0200", "2020-05-13T14:31:00+0200",
                    "2020-05-13T14:32:00+0200", "2020-05-13T14:33:00+0200",
                    "2020-05-13T14:34:00+0200", "2020-05-13T14:35:00+0200",
                    "2020-05-13T14:36:00+0200", "2020-05-13T14:37:00+0200",
                    "2020-05-13T14:38:00+0200", "2020-05-13T14:39:00+0200")
# Timstamps for file 4: non-overlapping time but long gap
timestamps[, 4] = c("2020-05-25T14:30:00+0200", "2020-05-25T14:31:00+0200",
                    "2020-05-25T14:32:00+0200", "2020-05-25T14:33:00+0200",
                    "2020-05-25T14:34:00+0200", "2020-05-25T14:35:00+0200",
                    "2020-05-25T14:36:00+0200", "2020-05-25T14:37:00+0200",
                    "2020-05-25T14:38:00+0200", "2020-05-25T14:39:00+0200")
# Calibration, same for all files:
C = list(cal.error.end = 0, cal.error.start = 0)
C$scale = c(1, 1, 1)
C$tempoffset = C$offset = c(0, 0, 0)
C$QCmessage = "Autocalibration not done"
C$npoints = C$nhoursused = 0
C$use.temp = TRUE
# Template M object
M1 = M2 = M3 = M4 = list(filecorrupt = FALSE, filetooshort = FALSE,
                         metalong = data.frame(
                           timestamp = rep("", 2), 
                           nonwearscore = rep(0, 2),
                           clippingscore = rep(0, 2),
                           lightmean = rep(0, 2),
                           lightpeak = rep(0, 2),
                           temperaturemean = rep(0, 2),
                           EN = rep(0, 2)
                         ),
                         metashort = data.frame(timestamp = rep("", 10),
                                                accmetric = rep(0, 10)), 
                         wday = 0,
                         wdayname = "Sunday",
                         windowsizes = c(60, 300, 300),
                         bsc_qc = data.frame(A = 1:3,
                                             B = 1:3))
# File 1, use value 1 as a way to mark the time series
M1$metashort$timestamp = timestamps[, 1]
M1$metashort$accmetric = 1
M1$metalong$timestamp = timestamps[c(1, 6), 1]
M1$metalong$lightmean = 1
# File 2, use value 2 as a way to mark the time series
M2$metashort$timestamp = timestamps[, 2]
M2$metashort$accmetric = 2
M2$metalong$timestamp = timestamps[c(1, 6), 2]
M2$metalong$lightmean = 2
# File 3, use value 3 as a way to mark the time series
M3$metashort$timestamp = timestamps[, 3]
M3$metashort$accmetric = 3
M3$metalong$timestamp = timestamps[c(1, 6), 3]
M3$metalong$lightmean = 3
# File 4, use value 4 as a way to mark the time series
M4$metashort$timestamp = timestamps[, 4]
M4$metashort$accmetric = 4
M4$metalong$timestamp = timestamps[c(1, 6), 4]
M4$metalong$lightmean = 4

dummyheader = data.frame(uniqueSerialCode = 0, 
                         frequency = 100,
                         Subject_Code = 1,
                         start = "1980-01-01 19:08:00",
                         device = "name",
                         firmwareVersion = "unknown",
                         block = 0)
dummyheader = t(dummyheader)
colnames(dummyheader) = "value"
I1 = I2 = I3 = I4 = list(
  header = dummyheader,
  monc = 2,
  monn = "geneactive",
  dformc = 1,
  dformn = "bin",
  sf = 100,
  filename = "",
  deviceSerialNumber = "12345")
# Three file names with same ID
I1$filename = "1_T1test.bin"
I2$filename = "1_T2test.bin"
I3$filename = "1_T3test.bin"
I4$filename = "1_T4test.bin"

tail_expansion_log = NULL
filefoldername = filename_dir = NA
# Save objects to RData files to imitate a real study situation
dn = "./testfolder/meta/basic"
if (!dir.exists(dn)) {
  dir.create(dn, recursive = TRUE)
}
I = I1
M = M1
save(I, M, C, filefoldername, filename_dir, tail_expansion_log, file = paste0(dn, "/meta_1_bin.RData"))
I = I2
M = M2
save(I, M, C, filefoldername, filename_dir, tail_expansion_log, file = paste0(dn, "/meta_2_bin.RData"))
I = I3
M = M3
save(I, M, C, filefoldername, filename_dir, tail_expansion_log, file = paste0(dn, "/meta_3_bin.RData"))
I = I4
M = M4
save(I, M, C, filefoldername, filename_dir, tail_expansion_log, file = paste0(dn, "/meta_4_bin.RData"))

test_that("Neighbouring recordings are correctly appended", {
  
  appendRecords(metadatadir = "./testfolder",
                desiredtz = "Europe/Amsterdam",
                idloc = 2,
                maxRecordingInterval = 120) # extract ID from filename
  expect_true(dir.exists("./testfolder/meta/basic"))
  expect_equal(length(dir("./testfolder/meta/basic")), 2)
  expect_true(file.exists("./testfolder/meta/basic/meta_1_bin.RData"))
  expect_true(file.exists("./testfolder/meta/basic/meta_4_bin.RData"))
  
  # File one, should now be merged
  load("./testfolder/meta/basic/meta_1_bin.RData")
  expect_equal(nrow(M$metashort), 40)
  expect_equal(sum(M$metashort$accmetric), 55)
  expect_equal(nrow(M$metalong), 8)
  expect_equal(sum(M$metalong$lightmean), 11)
  expect_equal(length(which(duplicated(M$metalong$timestamp) == TRUE)), 0)
  expect_equal(length(which(duplicated(M$metashort$timestamp) == TRUE)), 0)
  
  expect_equal(length(I_list), 3)
  expect_equal(I_list[[1]]$interval, c(-0.267, 0.067))
  expect_equal(I_list[[2]]$interval, -0.267)
  expect_false("interval" %in% names(I_list[[3]]))

  expect_equal(M$metalong$timestamp, c("2020-05-13T14:00:00+0200",
                                       "2020-05-13T14:05:00+0200",
                                       "2020-05-13T14:10:00+0200",
                                       "2020-05-13T14:15:00+0200",
                                       "2020-05-13T14:20:00+0200",
                                       "2020-05-13T14:25:00+0200",
                                       "2020-05-13T14:30:00+0200",
                                       "2020-05-13T14:35:00+0200"))
  
  expect_equal(M$metashort$timestamp[c(1, 12, 24, 36)], c( "2020-05-13T14:00:00+0200",
                                                           "2020-05-13T14:11:00+0200",
                                                           "2020-05-13T14:23:00+0200",
                                                           "2020-05-13T14:35:00+0200"))

  expect_equal(unique(M$metashort$accmetric), c(1, 2, 0, 3))
  
  # File four, should be unmerged
  rm(I_list)
  load("./testfolder/meta/basic/meta_4_bin.RData")
  expect_false(exists("I_list"))
  expect_equal(nrow(M$metashort), 10)
  expect_equal(nrow(M$metalong), 2)
  expect_equal(unique(M$metashort$accmetric), 4)
  if (dir.exists("./testfolder"))  unlink("./testfolder", recursive = TRUE)
})

check_myfun = function(myfun, windowsizes) { # Function to check myfun object
  status = 0
  # check that myfun is a list:
  if (is.list(myfun) == F) {
    status = 1
    stop("Object myfun is not a list.", call. = FALSE)
  }
  # check that there are no foreign object:
  foreignElements = which(names(myfun) %in% c("FUN", "parameters", "expected_sample_rate",
                                              "expected_unit", "colnames",
                                              "minlength", "outputres",
                                              "outputtype", "aggfunction",
                                              "timestamp","reporttype",
                                              "clevels", "ilevels", "qlevels",
                                              "ebout.dur", "ebout.th.cad", "ebout.th.acc",
                                              "ebout.criter", "ebout.condition", "name") == FALSE)
  if (length(foreignElements) != 0) {
    status = 1
    stop("Object myfun has unexpected elements.", call. = FALSE)
  }
  # check that essential objects are included:
  expectedElements = c("FUN", "parameters", "expected_sample_rate", "expected_unit",
                       "colnames", "minlength", "outputres")
  missingElements = which(expectedElements %in% names(myfun) == FALSE)
  if (length(missingElements) != 0) {
    status = 1
    stop(paste0("Object myfun misses the following elements: ",
                paste(expectedElements[missingElements],collapse = ", "), "."), call. = FALSE)
  }
  # check that FUN is a function:
  if (is.function(myfun$FUN) ==  FALSE) {
    status = 1
     stop("Element FUN in myfun is not a function.", call. = FALSE)
  }
  # check that expected_sample_rate is numeric
  if (is.numeric(myfun$expected_sample_rate) == F) {
    status = 1
    stop("Element expected_sample_rate in myfun is not numeric.", call. = FALSE)
  }
  # check that unit is specified:
  if (length(which(myfun$expected_unit %in%  c("mg", "g", "ms2") == TRUE)) != 1) {
    status = 1
    stop("Object myfun lacks a clear specification of the expected_unit.", call. = FALSE)
  }
  # check that colnames has at least one character value:
  if (length(myfun$colnames) == 0) {
    status = 1
    stop("Element colnames in myfun does not have a value.", call. = FALSE)
  }
  # Check that colnames is a character:
  if (is.character(myfun$colnames) == F) {
    status = 1
    stop("Element colnames in myfun does not hold a character value.", call. = FALSE)
  }
  # check that minlegnth has one value:
  if (length(myfun$minlength) != 1) {
    status = 1
    stop("Element minlength in myfun does not have one value.", call. = FALSE)
  }
  # Check that minlength is a number
  if (is.numeric(myfun$minlength) == F) {
    status = 1
    stop("Element minlength in myfun is not numeric.", call. = FALSE)
  }
  # check that outputres has one value:
  if (length(myfun$outputres) != 1) {
    status = 1
    stop("Element outputres in myfun does not have one value.", call. = FALSE)
  }
  # Check that outputres is a number:
  if (is.numeric(myfun$outputres) == F) {
    status = 1
    stop("Element outputres in myfun is not numeric.", call. = FALSE)
  }
  # check that it is a round number can add up to 900 seconds:
  if (myfun$outputres != round(myfun$outputres)) {
    status = 1
    stop("Element outputres in myfun should be a round number.", call. = FALSE)
  }
  # check that outputres is either a multitude of the epochs size or vica versa:
  if (myfun$outputres/windowsizes[1] != round(myfun$outputres/windowsizes[1]) &
      windowsizes[1]/myfun$outputres != round(windowsizes[1]/myfun$outputres)) {
    status = 1
    stop(paste0("Element outputres and the epochsize used in GGIR",
                " (first element of windowsizes) are not a multitude",
                " of each other.", call. = FALSE))
  }
  if ("outputtype" %in% names(myfun)) { # If outputtype is available:
    if (is.character(myfun$outputtype) == F) {
      status = 1
      stop(paste0("Error in check_myfun.R: Element outputtype is expected to be a",
                  " character specifying the ouput type"), call. = FALSE)
    }
  }
  if ("aggfunction" %in% names(myfun)) { # If aggfunction is available:
    if (is.function(myfun$aggfunction) == F) {
      status = 1
      stop("Error in check_myfun.R: Element aggfunction is not a function object.", call. = FALSE)
    }
  }

  # if ("timestamp" %in% names(myfun)) { # If timestamp is available:
  #   if (is.logical(myfun$timestamp) == F) {
  #     status = 1
  #     stop("Element timestamp is not of type logical.")
  #   }
  # }
  return(status)
}

library(GGIR)
context("applyExtFunction")
test_that("Function is applied to acceleration data results in expected output", {
  skip_on_cran()
  sf = 40
  S = rep(c(1,5,7,9), each = sf)
  data = as.matrix(data.frame(x = S, y = S, z = S))
  ws3 = 1
  # External function
  exampleExtFunction = function(data=c(), parameters=c()) {
    output = data * parameters[1]
    data = data.frame(data, agglevel = rep(1:4,each = 30))
    output = aggregate(data,by = list(data$agglevel),FUN = mean)
    output = output[,-c(1,ncol(output))]
    return(output)
  }
  
  myfun =  list(FUN = exampleExtFunction,
                parameters = 1.1,
                expected_sample_rate =  30, # resample data to 30 Hertz before applying function
                expected_unit = "mg",
                minlength = 1,
                outputres = 1,
                colnames = c("A","B","C"),
                outputtype = "numeric", #"numeric" (averaging is possible), "category" (majority vote)
                aggfunction = mean,
                timestamp = as.numeric(Sys.time())) # for unit test only
  
  # Note: If function applyExtFunction is used directly then object myfun 
  # cannnot carry a logical value because the timestamp can only be added 
  # in the g.getmeta function from which applyExtFunction is called. However,
  # you can provide a numeric value to indicate start time in seconds since
  # 1-1-1970 If the user nonetheless still tries to supply a logical value
  # then applyExtFunction overwrites this.
  
  output = applyExtFunction(data, myfun, sf, ws3)
  
  expect_that(ncol(output),equals(4))
  expect_that(nrow(output),equals(4))
  expect_that(sum(output[,2:4]),equals(66000))
  expect_that(output[3,2],equals(7000))
  
  # test with different settings ----
  ws3 = 2 # to force aggregation in output
  myfun =  list(FUN = exampleExtFunction,
                parameters = 1.1,
                expected_sample_rate = 30, # resample data to 30 Hertz before applying function
                expected_unit = "ms2",
                minlength = 1,
                outputres = 1,
                colnames = c("A","B","C"),
                outputtype = "numeric", #"numeric" (averaging is possible), "category" (majority vote)
                aggfunction = mean,
                timestamp = TRUE) # for unit test only
  
  expect_warning({ # warning from setting timestamp = TRUE
    output = applyExtFunction(data, myfun, sf, ws3)
  })
  
  expect_that(ncol(output),equals(3))
  expect_that(nrow(output),equals(2)) # 2 rows because they were aggregated to match ws3
  expect_that(sum(output[,2:3]),equals(215.82))
  
  # test with different settings ----
  myfun =  list(FUN = exampleExtFunction,
                parameters = 1.1,
                expected_sample_rate = 30, # resample data to 30 Hertz before applying function
                expected_unit = "ms2",
                minlength = 1,
                outputres = 4,
                colnames = c("A","B","C"),
                outputtype = "category", #"numeric" (averaging is possible), "category" (majority vote)
                aggfunction = mean,
                timestamp = TRUE) # for unit test only
  
  expect_warning({ # warning from setting timestamp = TRUE
    output = applyExtFunction(data, myfun, sf, ws3)
  })
  
  expect_that(ncol(output),equals(3))
  expect_that(nrow(output),equals(4))
  expect_that(sum(output[,2:3]),equals(431.64))
  
  # test check_myfun warnings and errors
  expect_error(check_myfun(myfun = 4, ws3),
               regexp = "not a list")
  expect_error(check_myfun(myfun = list(FUN = mean, 
                                        unexpected = "element"), ws3),
               regexp = "unexpected elements")
  expect_error(check_myfun(myfun = list(FUN = mean, 
                                        parameters = 4), ws3),
               regexp = "misses the following elements")
  expect_error(check_myfun(myfun = list(FUN = "not_a_function",
                                        parameters = 1.1,
                                        expected_sample_rate = 30, # resample data to 30 Hertz before applying function
                                        expected_unit = "ms2",
                                        minlength = 1,
                                        outputres = 4,
                                        colnames = c("A","B","C"),
                                        outputtype = "category", #"numeric" (averaging is possible), "category" (majority vote)
                                        aggfunction = mean,
                                        timestamp = Sys.time()), ws3),
               regexp = "FUN in myfun is not a function")
  expect_error(check_myfun(myfun = list(FUN = exampleExtFunction,
                                        parameters = 1.1,
                                        expected_sample_rate = "Sixty", # resample data to 30 Hertz before applying function
                                        expected_unit = "ms2",
                                        minlength = 1,
                                        outputres = 4,
                                        colnames = c("A","B","C"),
                                        outputtype = "category", #"numeric" (averaging is possible), "category" (majority vote)
                                        aggfunction = mean,
                                        timestamp = Sys.time()), ws3),
               regexp = "expected_sample_rate in myfun is not numeric")
  expect_error(check_myfun(myfun = list(FUN = exampleExtFunction,
                                        parameters = 1.1,
                                        expected_sample_rate = 30, # resample data to 30 Hertz before applying function
                                        expected_unit = "meters",
                                        minlength = 1,
                                        outputres = 4,
                                        colnames = c("A","B","C"),
                                        outputtype = "category", #"numeric" (averaging is possible), "category" (majority vote)
                                        aggfunction = mean,
                                        timestamp = Sys.time()), ws3),
               regexp = "lacks a clear specification of the expected_unit")
  expect_error(check_myfun(myfun = list(FUN = exampleExtFunction,
                                        parameters = 1.1,
                                        expected_sample_rate = 30, # resample data to 30 Hertz before applying function
                                        expected_unit = "ms2",
                                        minlength = 1,
                                        outputres = 4,
                                        colnames = c(),
                                        outputtype = "category", #"numeric" (averaging is possible), "category" (majority vote)
                                        aggfunction = mean,
                                        timestamp = Sys.time()), ws3),
               regexp = "Element colnames in myfun does not have a value")
  expect_error(check_myfun(myfun = list(FUN = exampleExtFunction,
                                        parameters = 1.1,
                                        expected_sample_rate = 30, # resample data to 30 Hertz before applying function
                                        expected_unit = "ms2",
                                        minlength = 1,
                                        outputres = 4,
                                        colnames = 1:3,
                                        outputtype = "category", #"numeric" (averaging is possible), "category" (majority vote)
                                        aggfunction = mean,
                                        timestamp = Sys.time()), ws3),
               regexp = "colnames in myfun does not hold a character value")
  expect_error(check_myfun(myfun = list(FUN = exampleExtFunction,
                                        parameters = 1.1,
                                        expected_sample_rate = 30, # resample data to 30 Hertz before applying function
                                        expected_unit = "ms2",
                                        minlength = 1:4,
                                        outputres = 4,
                                        colnames = c("X","Y","Z"),
                                        outputtype = "category", #"numeric" (averaging is possible), "category" (majority vote)
                                        aggfunction = mean,
                                        timestamp = Sys.time()), ws3),
               regexp = "minlength in myfun does not have one value")
  expect_error(check_myfun(myfun = list(FUN = exampleExtFunction,
                                        parameters = 1.1,
                                        expected_sample_rate = 30, # resample data to 30 Hertz before applying function
                                        expected_unit = "ms2",
                                        minlength = "one",
                                        outputres = 4,
                                        colnames = c("X","Y","Z"),
                                        outputtype = "category", #"numeric" (averaging is possible), "category" (majority vote)
                                        aggfunction = mean,
                                        timestamp = Sys.time()), ws3),
               regexp = "minlength in myfun is not numeric")
  expect_error(check_myfun(myfun = list(FUN = exampleExtFunction,
                                        parameters = 1.1,
                                        expected_sample_rate = 30, # resample data to 30 Hertz before applying function
                                        expected_unit = "ms2",
                                        minlength = 1,
                                        outputres = 4:10,
                                        colnames = c("X","Y","Z"),
                                        outputtype = "category", #"numeric" (averaging is possible), "category" (majority vote)
                                        aggfunction = mean,
                                        timestamp = Sys.time()), ws3),
               regexp = "outputres in myfun does not have one value")
  expect_error(check_myfun(myfun = list(FUN = exampleExtFunction,
                                        parameters = 1.1,
                                        expected_sample_rate = 30, # resample data to 30 Hertz before applying function
                                        expected_unit = "ms2",
                                        minlength = 1,
                                        outputres = "four",
                                        colnames = c("X","Y","Z"),
                                        outputtype = "category", #"numeric" (averaging is possible), "category" (majority vote)
                                        aggfunction = mean,
                                        timestamp = Sys.time()), ws3),
               regexp = "outputres in myfun is not numeric")
  expect_error(check_myfun(myfun = list(FUN = exampleExtFunction,
                                        parameters = 1.1,
                                        expected_sample_rate = 30, # resample data to 30 Hertz before applying function
                                        expected_unit = "ms2",
                                        minlength = 1,
                                        outputres = 4.4,
                                        colnames = c("X","Y","Z"),
                                        outputtype = "category", #"numeric" (averaging is possible), "category" (majority vote)
                                        aggfunction = mean,
                                        timestamp = Sys.time()), ws3),
               regexp = "outputres in myfun should be a round number")
  expect_error(check_myfun(myfun = list(FUN = exampleExtFunction,
                                        parameters = 1.1,
                                        expected_sample_rate = 30, # resample data to 30 Hertz before applying function
                                        expected_unit = "ms2",
                                        minlength = 1,
                                        outputres = 5,
                                        colnames = c("X","Y","Z"),
                                        outputtype = "numeric", #"numeric" (averaging is possible), "category" (majority vote)
                                        aggfunction = mean,
                                        timestamp = Sys.time()), ws3),
               regexp = "are not a multitude of each other")
  expect_error(check_myfun(myfun = list(FUN = exampleExtFunction,
                                        parameters = 1.1,
                                        expected_sample_rate = 30, # resample data to 30 Hertz before applying function
                                        expected_unit = "ms2",
                                        minlength = 1,
                                        outputres = 4,
                                        colnames = c("X","Y","Z"),
                                        outputtype = 2, #"numeric" (averaging is possible), "category" (majority vote)
                                        aggfunction = mean,
                                        timestamp = Sys.time()), ws3),
               regexp = "outputtype is expected to be a character")
  expect_error(check_myfun(myfun = list(FUN = exampleExtFunction,
                                        parameters = 1.1,
                                        expected_sample_rate = 30, # resample data to 30 Hertz before applying function
                                        expected_unit = "ms2",
                                        minlength = 1,
                                        outputres = 4,
                                        colnames = c("X","Y","Z"),
                                        outputtype = "numeric", #"numeric" (averaging is possible), "category" (majority vote)
                                        aggfunction = "mean",
                                        timestamp = Sys.time()), ws3),
               regexp = "Element aggfunction is not a function object")
  
  
})

g.analyse.avday = function(doquan, averageday, M, IMP, t_TWDI, quantiletype,
                           ws3, doiglevels, firstmidnighti, ws2, midnightsi, params_247 = c(), 
                           qcheck = c(), acc.metric = c(), params_phyact = NULL, ...) {
  #get input variables
  input = list(...)
  if (length(input) > 0 || length(params_247) == 0) {
    # Extract and check parameters if user provides more arguments than just the parameter arguments,
    # or if params_[...] aren't specified (so need to be filled with defaults).
    # So, inside GGIR this will not be used, but it is used when g.analyse is used on its own
    # as if it was still the old g.analyse function
    params = extract_params(params_247 = params_247,
                            input = input) # load default parameters
    params_247 = params$params_247
  }
  if (length(acc.metric) == 0) {
    acc.metric = colnames(IMP$metashort)[which(colnames(IMP$metashort) %in% 
                                                 c("anglex", "angley", "anglez", 
                                                   "ExtSleep", "timestamp", "marker") == FALSE)[1]]
  }

  getUnitRescale = function(varName) {
    gUnitMetric = length(grep(x = varName,
                              pattern = "BrondCount|ZCX|ZCY|ZCZ|marker|NeishabouriCount|ExtAct|ExtHeartRate", invert = TRUE)) > 0
    UnitReScale = ifelse(test = gUnitMetric, yes = 1000, no = 1)
    return(UnitReScale)
  }
  
  if (doquan == TRUE) {
    QLN = rep(" ",length(params_247[["qlevels"]]))
    for (QLNi in 1:length(params_247[["qlevels"]])) {
      QLN[QLNi] = paste0("p", (round((params_247[["qlevels"]][QLNi]) * 10000))/100)
    }
  }
  QUAN = qlevels_names = c()
  ML5AD = ML5AD_names = c()
  one2sixname = ""
  if (is.data.frame(t_TWDI) == TRUE) { # If an activity log was used then use for average day only 24 hour window.
    t_TWDI = c(0, 24)
  }
  if (doquan == TRUE) {
    for (quani in 1:ncol(averageday)) { # these columns of averageday correspond to the metrics from part 1
      if (colnames(M$metashort)[(quani + 1)] %in% c("anglex", "angley", "anglez",
                                                    "ExtSleep", "marker") == FALSE) {
        UnitReScale = getUnitRescale(colnames(M$metashort)[(quani + 1)])
        #--------------------------------------
        # quantiles
        QUANtmp =  quantile(averageday[((t_TWDI[1] * (3600 / ws3)) + 1):(t_TWDI[length(t_TWDI)] * (3600 / ws3)), quani],
                            probs = params_247[["qlevels"]], na.rm = TRUE, type = quantiletype)
        QUAN = c(QUAN, QUANtmp)
        qlevels_namestmp = rep(" ", length(params_247[["qlevels"]]))
        for (QLNi in 1:length(params_247[["qlevels"]])) {
          qlevels_namestmp[QLNi] = paste0(QLN[QLNi], "_", colnames(M$metashort)[(quani + 1)], "_mg_",
                                          t_TWDI[1], "-", t_TWDI[length(t_TWDI)], "h")
        }
        qlevels_names = c(qlevels_names, qlevels_namestmp)
        rm(QUANtmp)
        #--------------------------------------
        #M5L5 - (Note: averageday is relative to starttime of measurement, but so is firstmidnighti (index of midnights), so M5L5 are correct)
        #loop through winhr
        for (winhr_value in params_247[["winhr"]]) {
          # Time window for L5 & M5 analysis
          t0_LFMF = params_247[["L5M5window"]][1] #start in 24 hour clock hours
          t1_LFMF = params_247[["L5M5window"]][2] + (winhr_value - (params_247[["M5L5res"]] / 60)) #end in 24 hour clock hours (if a value higher than 24 is chosen, it will take early hours of previous day to complete the 5 hour window
          avday = averageday[, quani]
          avday = c(avday[(firstmidnighti * (ws2 / ws3)):length(avday)], avday[1:((firstmidnighti * (ws2 / ws3)) - 1)])
          # Note that t_TWDI[length(t_TWDI)] in the next line makes that we only calculate ML5 over the full day
          ML5ADtmp = g.getM5L5(avday,ws3,t0_LFMF = t_TWDI[1], t1_LFMF = t_TWDI[length(t_TWDI)],
                               M5L5res = params_247[["M5L5res"]],
                               winhr_value, qM5L5 = params_247[["qM5L5"]],
                               MX.ig.min.dur = params_247[["MX.ig.min.dur"]],
                               UnitReScale = UnitReScale)
          ML5AD = as.data.frame(c(ML5AD, ML5ADtmp), stringsAsFactors = TRUE)
        }
        ML5N = names(ML5AD)
        for (ML5ADi in 1:length(ML5N)) {
          if (length(grep(pattern = "1to6", ML5N[ML5ADi])) == 0) {
            ML5AD_names[ML5ADi] = paste0(ML5N[ML5ADi], "_", colnames(M$metashort)[(quani + 1)],"_mg_",
                                         params_247[["L5M5window"]][1], "-", params_247[["L5M5window"]][2], "h")
          }
        }
        # Acceleration 1-6am
        one2sixname = paste0("1to6am_", colnames(M$metashort)[(quani + 1)], "_mg")
        ML5AD_names = c(ML5AD_names, one2sixname)
        ML5AD$one2six = mean(avday[((1 * 60 * (60 / ws3)) + 1):(6 * 60 * (60 / ws3))]) * UnitReScale
        colnames(ML5AD)[which(colnames(ML5AD) == "one2six")] = one2sixname
      }	
    }	
  }
  igfullr_names = igfullr = c()
  if (doiglevels == TRUE) { 
    # intensity gradient (as described by Alex Rowlands 2018)
    # applied to the averageday per metric (except from angle metrics)
    igfullr = igfullr_names = c()
    for (igi in 1:ncol(averageday)) {
      if (colnames(M$metashort)[(igi + 1)] %in% c("anglex", "angley", "anglez", "marker") == FALSE) {
        UnitReScale = getUnitRescale(colnames(M$metashort)[(igi + 1)])
        y_ig = c()
        avday = averageday[, igi] 
        avday = c(avday[(firstmidnighti * (ws2 / ws3)):length(avday)], avday[1:((firstmidnighti * (ws2 / ws3)) - 1)])
        # we are not taking the segment of the day now (too much output)
        q60 = cut((avday * UnitReScale), breaks = params_247[["iglevels"]], right = FALSE)
        y_ig  = (as.numeric(table(q60)) * ws3) / 60 #converting to minutes
        x_ig = zoo::rollmean(params_247[["iglevels"]], k = 2)
        igout = g.intensitygradient(x_ig, y_ig)
        if (length(avday) > 0) {
          igfullr = c(igfullr, as.vector(unlist(igout)))
        } else {
          igfullr = c(igfullr, rep("", 3))
        }
        igfullr_names = c(igfullr_names, paste0(c("ig_gradient", "ig_intercept", "ig_rsquared"),
                                                paste0("_", colnames(IMP$metashort)[igi + 1], "_0-24hr")))
      }
    }
  }
  #----------------------------------
  # Cosinor analysis, (Extended) Cosinor analysis, including IV, IS, and phi
  # based on time series where invalid data points are set to NA
  if (params_247[["part2CR"]] == TRUE) {
    cosinor_coef = apply_cosinor_IS_IV_Analyses(ts = IMP$metashort[, c("timestamp", acc.metric)],
                                                qcheck = qcheck,
                                                midnightsi, epochsizes = c(ws3, ws2),
                                                threshold = params_phyact[["threshold.lig"]][1]) # only use one threshold
  } else {
    cosinor_coef = c()
  }
  invisible(list(igfullr_names = igfullr_names,
                 igfullr = igfullr, QUAN = QUAN,
                 qlevels_names = qlevels_names,
                 ML5AD = ML5AD,
                 ML5AD_names = ML5AD_names,
                 cosinor_coef = cosinor_coef))
}

g.analyse.perday = function(ndays, firstmidnighti, time, nfeatures,
                            midnightsi, metashort, averageday,
                            doiglevels, nfulldays,lastmidnight, ws3, ws2, qcheck,
                            fname, sensor.location, wdayname, tooshort, includedaycrit,
                            doquan, quantiletype, doilevels, domvpa,
                            mvpanames, wdaycode, ID,
                            deviceSerialNumber, ExtFunColsi, myfun, 
                            params_247 = c(), params_phyact = c(),
                            params_general = c(),
                            ...) {
  #get input variables
  input = list(...)
  if (length(input) > 0 ||
      length(params_247) == 0 || length(params_phyact) == 0) {
    # Extract and check parameters if user provides more arguments than just the parameter arguments,
    # or if params_[...] aren't specified (so need to be filled with defaults).

    # So, inside GGIR this will not be used, but it is used when g.analyse is used on its own
    # as if it was still the old g.analyse function
    params = extract_params(params_247 = params_247,
                            params_phyact = params_phyact,
                            input = input,
                            params2check = c("247", "phyact")) # load default parameters
    params_247 = params$params_247
    params_phyact = params$params_phyact
    rm(params)
  }
  
  startatmidnight = endatmidnight = 0
  if (nfulldays >= 1) {
    if (firstmidnighti == 1) {  #if measurement starts at midnight
      ndays = ndays - 1
      startatmidnight =  1
    }
    if (lastmidnight == time[length(time)] & nrow(metashort) < ((60/ws3) * 1440)) {	#if measurement ends at midnight
      ndays = ndays - 1
      endatmidnight = 1
    }
  } else {
    # Handle edge case for short recordings without a full day (e.g., a few hours) 
    # that both start and end at midnight or it does not include any midnight 
    if (firstmidnighti == 1 & 
        lastmidnight == time[length(time)] & 
        nrow(metashort) < ((60/ws3) * 1440)) {
      ndays = ndays - 1
      endatmidnight = 1
    }
  }
  
  daysummary = matrix("",ceiling(ndays),nfeatures)
  ds_names = rep("",nfeatures)
  windowsummary = ws_names = c()
  #===============================================
  # Features per day (based on on single variables)
  qwindow_names = qwindowbackup = params_247[["qwindow"]]
  qwindow_actlog = FALSE
  if (is.data.frame(params_247[["qwindow"]]) == TRUE) {
    qwindow_actlog = TRUE
  }
  unique_dates_recording = unique(as.Date(iso8601chartime2POSIX(time[c(seq(1, length(time),
                                                                           by = (3600/ws2) * 12),
                                                                       length(time))], tz = params_general[["desiredtz"]])))
  ExtFunColsi = ExtFunColsi - 1 # subtract 1 because code ignores timestamp
  ExtFunColsi_backup = ExtFunColsi
  for (di in 1:ndays) { #run through days
    params_247[["qwindow"]] = qwindowbackup
    if (is.data.frame(params_247[["qwindow"]]) == TRUE) {
      currentdate = which(params_247[["qwindow"]]$date == unique_dates_recording[di])
      if (length(currentdate) > 0) {
        qwindow_times = as.character(unlist(params_247[["qwindow"]]$qwindow_times[currentdate]))
        qwindow_names = as.character(unlist(params_247[["qwindow"]]$qwindow_names[currentdate]))
        params_247[["qwindow"]] = qwindow_values_backup = unlist(params_247[["qwindow"]]$qwindow_values[currentdate])
        qwindow_names = qwindow_names[!is.na(params_247[["qwindow"]])]
        qwindow_times = qwindow_times[!is.na(params_247[["qwindow"]])]
      }
      params_247[["qwindow"]] = params_247[["qwindow"]][!is.na(params_247[["qwindow"]])]
      if (length(params_247[["qwindow"]]) == 1) params_247[["qwindow"]] = c()
      if (length(params_247[["qwindow"]]) == 0 | length(currentdate) == 0)  {
        qwindow_times = c("00:00","24:00")
        qwindow_names = c("daystart","dayend")
        params_247[["qwindow"]] = qwindow_values_backup = c(0, 24)
      }
    } else {
      if (length(params_247[["qwindow"]]) == 0) {
        params_247[["qwindow"]] = c(0, 24)
      }
      qwindow_values_backup = params_247[["qwindow"]]
      qwindow_times = as.character(params_247[["qwindow"]])
      qwindow_names = as.character(params_247[["qwindow"]])
    }
    # order qwindow timestamps in current date
    qwindow_order = order(params_247[["qwindow"]])
    params_247[["qwindow"]] = qwindow_values_backup = params_247[["qwindow"]][qwindow_order]
    qwindow_times = qwindow_times[qwindow_order]
    qwindow_names = qwindow_names[qwindow_order]
    # extract day from matrix D and qcheck
    if (startatmidnight == 1 & endatmidnight == 1) {
      qqq1 = midnightsi[di] * (ws2/ws3)	#a day starts at 00:00
      qqq2 = (midnightsi[(di + 1)]*(ws2/ws3)) - 1
    } else if (startatmidnight == 1 & endatmidnight == 0) {
      if (di < floor(ndays)) { #applying floor because when day there is day saving time then ndays is not an integer
        qqq1 = (midnightsi[di] - 1) * (ws2 / ws3) + 1
        qqq2 = ((midnightsi[(di + 1)] - 1) * (ws2 / ws3))  #remove +1 because I got some extra seconds per day
      } else if (di == floor(ndays)) {
        qqq1 = (midnightsi[di] - 1) * (ws2 / ws3) + 1
        qqq2 = nrow(metashort)
      }
    } else if (startatmidnight == 0 & endatmidnight == 0) {
      if (di == 1) {
        qqq1 = 1
        qqq2 = ((midnightsi[di] - 1) * (ws2 / ws3))
      } else if (di > 1 & di < floor(ndays)) {
        qqq1 = (midnightsi[(di - 1)] - 1) * (ws2 / ws3) + 1 # a day starts at 00:00
        qqq2 = ((midnightsi[di] - 1) * (ws2 / ws3))
      } else if (di == floor(ndays)) {
        qqq1 = (midnightsi[(di - 1)] - 1)*(ws2 / ws3) + 1 # a day starts at 00:00
        qqq2 = nrow(metashort)
      }
    } else if (startatmidnight == 0 & endatmidnight == 1) {
      if (di == 1) {
        qqq1 = 1
        qqq2 = (midnightsi[di] * (ws2 / ws3)) - 1
      } else if (di > 1 & di <= floor(ndays)) {
        qqq1 = midnightsi[(di - 1)] * (ws2 / ws3) # a day starts at 00:00
        qqq2 = (midnightsi[di] * (ws2 / ws3)) - 1
      }
    }
    if (qqq2 > nrow(metashort)) qqq2 = nrow(metashort)
    vari = metashort[qqq1:qqq2,]
    val = qcheck[qqq1:qqq2]
    nvalidhours_qwindow = rep(0, length(params_247[["qwindow"]]) - 1)
    nhours_qwindow = rep(0, length(params_247[["qwindow"]]) - 1)
    # Ignore qwindow values that are not possible for this day
    LENVAL_hours = length(val) / (60 * (60 / ws3)) #11.2
    if (length(which(round(LENVAL_hours) %in% 23:25 == TRUE)) == 0) {
      if (di == 1) {
        # Following 8 lines were turned off on June 5 2018 because first and last day are imputed,
        # but turned on again on 14 March 2019 because when the first day is half the relative start
        # of the windows must be different. Yes, the half days are imputed, but we are still only
        # interested in the non-imputed part. This is probably were the confusion came from.
        hours2delta = 24 - LENVAL_hours
        qw_select = which(params_247[["qwindow"]] > hours2delta)
        # if (qw_select[1] > 1) qw_select = c(qw_select[1] - 1, qw_select)
        # params_247[["qwindow"]] = params_247[["qwindow"]][qw_select]
        # # qwindow_names = qwindow_names[qw_select]
        qwindowindices = params_247[["qwindow"]] - hours2delta # - LENVAL_hours # because 1 is now different
        if (length(which(qwindowindices < 0)) > 0) qwindowindices[which(qwindowindices < 0)] = 0
      } else if (di == ndays) {
        qwindowindices = params_247[["qwindow"]]
      }
    } else {
      hours2delta = 0
      qwindowindices = params_247[["qwindow"]]
    }
    deltaLengthQwindow = 0
    if (length(params_247[["qwindow"]]) < 2) params_247[["qwindow"]] = c()
    if (length(params_247[["qwindow"]]) > 0) {
      if (length(qwindow_names) == 1) {
        warning("Argument to qwindow is invalid, requires a vector of at least length 2")
      }
      if (length(qwindow_names) == 2) {
        if (params_247[["qwindow"]][1] != 0 | params_247[["qwindow"]][2] != 24) {
          if((qwindowindices[2] * 60 * (60 / ws3)) <= length(val)) {
            valq = val[((qwindowindices[1] * 60 * (60 / ws3)) + 1):(qwindowindices[2] * 60 * (60 / ws3))]
          } else {
            valq = val[((qwindowindices[1] * 60 * (60 / ws3)) + 1):length(val)]
          }
          nvalidhours_qwindow = length(which(valq == 0)) / (3600 / ws3)
          nhours_qwindow = length(valq) / (3600 / ws3) #valid hours per day (or half a day)
        }
      } else if (length(qwindow_names) > 2) {
        deltaLengthQwindow = length(qwindow_names) - length(qwindowindices)
        
        for (qwi in 1:(length(qwindowindices) - 1)) {
          startindex = qwindowindices[qwi] * 60 * (60/ws3)
          endindex = qwindowindices[qwi + 1] * 60 * (60/ws3)
          if (startindex == endindex) {
            # inexistent qwindow (may occure in the first day if recording started later than first qwindow)
            valq = c()
          } else if (startindex <= length(val) & endindex <= length(val)) {
            valq = val[(startindex + 1):endindex]
          } else if (startindex <= length(val) & endindex >= length(val)) {
            valq = val[(startindex + 1):length(val)]
          } else if (startindex >= length(val) & endindex >= length(val)) {
            valq = c()
          }
          if (length(valq) > 0) {
            nvalidhours_qwindow[qwi + deltaLengthQwindow] = length(which(valq == 0)) / (3600 / ws3)
            nhours_qwindow[qwi + deltaLengthQwindow] = length(valq) / (3600 / ws3) #valid hours per day (or half a day)
          } else {
            nvalidhours_qwindow[qwi + deltaLengthQwindow] = ""
            nhours_qwindow[qwi + deltaLengthQwindow] = "" #valid hours per day (or half a day)
          }
        }
      }
    }
    val = as.numeric(val)
    nvalidhours = length(which(val == 0)) / (3600 / ws3) #valid hours per day (or half a day)
    nhours = length(val) / (3600 / ws3) #valid hours per day (or half a day)
    #start collecting information about the day
    fi = 1
    
    
    check_daysummary_size = function(daysummary, ds_names, fi) {
      if (fi > ncol(daysummary) - 1000) {
        expand = fi - (ncol(daysummary) - 1000)
        daysummary = cbind(daysummary, matrix("", nrow(daysummary), expand +  1000))
        ds_names = c(ds_names, rep("", expand + 1000))
      }
      invisible(list(daysummary = daysummary, ds_names = ds_names))
    }
    daysummary[di,fi] = ID
    idremember = daysummary[di,fi]
    ds_names[fi] = "ID";      fi = fi + 1
    if (!is.null(params_general[["recording_split_times"]])) {
      filename = fname
      splitnames = getSplitNames(filename)
      segment_names = splitnames$segment_names
      filename = splitnames$filename
    } else {
      segment_names = NULL
      filename = fname
    }
    daysummary[di,fi] = filename
    ds_names[fi] = "filename";  fi = fi + 1
    if (!is.null(params_general[["recording_split_times"]]) && !is.null(segment_names)) {
      daysummary[di,fi] = segment_names[2]
      daysummary[di,fi + 1] = segment_names[3]
      ds_names[fi:(fi + 1)] = c("split1_name","split2_name")
      fi = fi + 2
    }
    calendardate = unlist(strsplit(as.character(vari[1,1])," "))[1]
    daysummary[di,fi] = calendardate
    daysummary[di,(fi + 1)] = sensor.location
    daysummary[di,(fi + 2)] = nvalidhours
    daysummary[di,(fi + 3)] = nhours
    ds_names[fi:(fi + 3)] = c("calendar_date","bodylocation","N valid hours","N hours")
    fi = fi + 4
    vari = vari[,2:ncol(vari), drop = FALSE] # remove timestamp now it is no longer needed, drop = FALSE ensures it is not coerced to numeric if only 1 metric is calculated
    colnames(averageday) = colnames(vari)
    
    if (length(params_247[["qwindow"]]) > 0) {
      if (length(params_247[["qwindow"]]) > 2 | params_247[["qwindow"]][1] != 0 | params_247[["qwindow"]][2] != 24) {
        for (qwi in 1:(length(qwindow_names) - 1)) {
          tmp_name = c(paste0("N_valid_hours_", qwindow_names[qwi], "-", qwindow_names[qwi + 1], "hr"),
                       paste0("N_hours_", qwindow_names[qwi], "-", qwindow_names[qwi + 1], "hr"))
          matchi1 = which(ds_names %in% tmp_name[1] == TRUE)
          matchi2 = which(ds_names %in% tmp_name[2] == TRUE)
          if (length(matchi1) == 1 & length(matchi2) == 1) {
            ds_names[matchi1] == tmp_name[1]
            ds_names[matchi2] == tmp_name[2]
            if (length(which(qwindow_values_backup[qwi] %in% params_247[["qwindow"]] == TRUE)) > 0) {
              daysummary[di,matchi1] = nvalidhours_qwindow[qwi]
              daysummary[di,matchi2] = nhours_qwindow[qwi]
            } else {
              daysummary[di,matchi1] = "" # note that we only consider qwindows when there is data for the entire window, otherwise duration is 0
              daysummary[di,matchi2] = ""
            }
            fi = fi + 2
          } else {  # variable does not exists, so add it,
            ds_names = c(ds_names,tmp_name)
            daysummary = cbind(daysummary, matrix("", nrow(daysummary), 2))
            newncol = ncol(daysummary)
            if (length(which(qwindow_values_backup[qwi] %in% params_247[["qwindow"]] == TRUE)) > 0) {
              daysummary[di, (newncol - 1):newncol] = c(nvalidhours_qwindow[qwi], nhours_qwindow[qwi])
            } else {
              daysummary[di, (newncol - 1):newncol] = c("", "")
            }
          }
        }
      }
    }
    if (di > 1 & length(which(ds_names == "weekday")) > 0) { #qwindow_actlog == TRUE &
      fi = which(ds_names == "weekday")
    }
    #--------------------------------------
    weekdays = c("Sunday","Monday","Tuesday","Wednesday","Thursday","Friday","Saturday")
    weekdays = rep(weekdays, 104) # hardcoded maximum number of weeks of 104, not ideal
    if (di == 1) {
      daysummary[di,fi] = wdayname
    } else {
      daysummary[di,fi] = weekdays[wdaycode + (di - 1)]
    }
    daysummary[di, (fi + 1)] = di #day number relative to start of measurement
    ds_names[fi:(fi + 1)] = c("weekday","measurementday")
    fi = fi + 2
    if (qwindow_actlog == TRUE | length(params_247[["qwindow"]]) > 2) { # add column with all timestamps in single character
      ds_names[fi] =  "qwindow_timestamps"
      daysummary[di, fi] = paste0(qwindow_times, collapse = "_")
      fi = fi + 1
      ds_names[fi] =  "qwindow_names"
      daysummary[di,fi] = paste0(qwindow_names, collapse = "_")
      fi = fi + 1
    }
    correct_fi = function(di, ds_names, fi, varname) {
      # function to reset variable index to allign with previous day
      if (di > 1) {
        varexists = which(ds_names == varname)
        if (length(varexists) > 0) {
          fi = varexists
        } else {
          fi = which(ds_names == "")[1]
        }
      }
      return(fi)
    }
    if (tooshort == 0) {
      if (nvalidhours >= includedaycrit[1]) {
        #--------------------------------------
        # Features per day and per segment of the day
        # guided by qwindow, which is a vector of indices to slice up the day
        keepindex_46 = keepindex_48 = matrix(NA, length(2:ncol(metashort)), 2)
        # generate objects to help with selecting the slices and to give names to the respective variables
        anwi_t0 = 1 # analysis window time 0
        anwi_t1 = nrow(vari) # analysis window time 1
        anwi_nameindices = "_0-24hr"
        anwi_index = 1
        if (length(params_247[["qwindow"]]) > 0) {
          if (length(params_247[["qwindow"]]) == 2 & params_247[["qwindow"]][1] == 0 & params_247[["qwindow"]][2] == 24) {
          } else {
            oddqwindow = 1:(length(qwindowindices) - 1)
            evenqwindow = 2:length(qwindowindices)
            anwi_t0 = c(anwi_t0, ((qwindowindices[oddqwindow] * 60 * (60 / ws3)) + 1))
            anwi_t1 = c(anwi_t1, (qwindowindices[evenqwindow] * 60 * (60 / ws3)))
            for (iin in 1:length(oddqwindow)) {
              anwi_nameindices = c(anwi_nameindices,paste0("_", qwindow_names[oddqwindow[iin]], "-",
                                                           qwindow_names[evenqwindow[iin]], "hr"))
            }
          }
        } else {
          anwi_t0 = c(anwi_t0, ((0 * 60 * (60 / ws3)) + 1))
          anwi_t1 = c(anwi_t1,(24 * 60 * (60 / ws3)))
          anwi_nameindices = c(anwi_nameindices, "")
        }
        fi_remember = fi
        vari_bu = vari
        for (anwi_index in 1:length(anwi_t0)) { # Loop over starts of the windows
          if (anwi_index != 1 & di == 1) {
            # increase value of fi to leave enough space for the variables to be calculated in second day of measurement
            shift = (deltaLengthQwindow * (fi - fi_remember))
            fi = fi + shift
            fi_remember = fi
          }
          vari = vari_bu
          new = check_daysummary_size(daysummary, ds_names, fi)
          daysummary = new$daysummary
          ds_names = new$ds_names
          
          L5M5window_name = anwi_nameindices[anwi_index]
          anwindices = anwi_t0[anwi_index]:anwi_t1[anwi_index] # indices of varnum corresponding to a segment
          if (length(anwindices) > 0 & all(diff(anwindices) > 0)) { # negative diff(anwindices) may occur in the first day if a qwindow is not within the recorded hours
            minames = colnames(vari)
            
            #=====================================
            # Note: vari equals the imputed time series (metashort) data from one day
            # Also impute the unrecorded period before and after the recording
            NRV = nrow(vari)
            deltaLength = NRV - nrow(averageday)
            if (deltaLength < 0) {
              # Less than 24 hours: Append data from averageday
              if (di == 1) {
                # On first day of recording append the average day to the start
                vari = rbind(averageday[1:abs(deltaLength), ], vari)
                # readjust anwi indices in case that varnum has been imputed
                if (max(anwi_t1) < nrow(vari)) { # since GGIR always calculates full window, max(anwi_t1) should always equals length(varnum)
                  anwi_t0 = anwi_t0 + abs(deltaLength)
                  anwi_t1 = anwi_t1 + abs(deltaLength)
                  # then, we reset the minimum anwi_t0 to 1 to consider the imputed varnum
                  # anwi_t0[which(anwi_t0 == min(anwi_t0))] = 1
                  anwi_t0[1] = 1
                }
              } else {
                # When it is not the first day of recording
                if (NRV == 23 * 60 * (60/ws3)) { # day has 23 hours (assuming DST)
                  # Append data after 2nd hour
                  startMissingHour = 2 * 60 * (60/ws3) + 1
                  enMissingHour = 3 * 60 * (60/ws3)
                  vari = rbind(vari[1:(startMissingHour - 1), , drop = FALSE], averageday[startMissingHour:enMissingHour, , drop = FALSE],
                               vari[startMissingHour:nrow(vari), , drop = FALSE])
                } else { # day has less than 24 hours for another reason
                  # Append the average day to the end
                  a56 = nrow(averageday) - abs(deltaLength) + 1
                  a57 = nrow(averageday)
                  vari = rbind(vari, averageday[a56:a57, , drop = FALSE])
                }
              }
            } else if (deltaLength > 0) { # 25 hour days, assuming DST
              # If there is more than 24 hours in a day then clock must
              # have moved backward, remove the double hour.
              startDoubleHour = 2 * 60 * (60/ws3) + 1
              endDoubleHour = 3 * 60 * (60/ws3)
              if (nrow(vari) > endDoubleHour) {
                vari = vari[-c(startDoubleHour:endDoubleHour)]
              }
            }
            if (anwi_index != 1) {
              if (length(anwindices) > 0) {
                if (max(anwindices) > length(vari)) {
                  anwindices = anwindices[which(anwindices <= nrow(vari))]
                }
                vari = vari[anwindices, ] #cut short varnum to match day segment of interest
              } else {
                vari = c()
              }
            }
            #===

            for (mi in 1:ncol(vari)) { #run through metrics (for features based on single metrics)
              #=======================================
              # Motivation on the code below:
              # Standardise the number of hours in a day as an attempt to create a
              # fair comparison between days in terms of day length. For example, to
              # compare time spent in intensity levels or MVPA if days are not of
              # equal length, such as when a recording starts in
              # the middle of the day.
              # To do this we impute the missing part based on the average day
              # (literally called averageday in the code).
              # Note: We only do this here in part 2 and not in part 5, and it has
              # been this way since the early days of GGIR.
              # In part 2, GGIR aims to use as much data as possible to provide
              # estimates of behaviour on each recording day, even for the
              # incomplete first and last recording day. As a result, it is
              # important to account for imbalance in day length, which we do below.
              # In part 5, however, GGIR forces the user to only work with complete
              # days and by that the day length is less of a problem and not accounted for.
              
              NRV = length(which(is.na(as.numeric(as.matrix(vari[,mi]))) == FALSE))
              # Note: vari equals the imputed time series (metashort) data from one day
              varnum = as.numeric(as.matrix(vari[,mi])) # Note: varnum is one column of vari
              if (length(ExtFunColsi_backup) > 0) { # If an external function was used
                if (mi %in% ExtFunColsi_backup == TRUE) {
                  rti = which(ExtFunColsi_backup == mi)
                  ExtFunColsi = ExtFunColsi_backup[rti] # column index where to find the corresponding info
                }
              }
              #==============================
              # varnum_step
              isAccMetric = minames[mi] %in% c("ENMO","LFENMO", "BFEN", "EN", "HFEN", "HFENplus", "MAD", "ENMOa",
                                 "ZCX", "ZCY", "ZCZ", "BrondCount_x", "BrondCount_y",
                                 "BrondCount_z", "NeishabouriCount_x", "NeishabouriCount_y",
                                 "NeishabouriCount_z", "NeishabouriCount_vm", "ExtAct",
                                 "ExtHeartRate", myfun$colnames)
              
              if (isAccMetric == TRUE & length(ExtFunColsi) > 0) {
                # Then also extract count metric
                
                varnum_event = as.numeric(as.matrix(vari[,ExtFunColsi]))
                Ndatapoints = nrow(averageday)
                if (NRV < Ndatapoints) {
                  if (di == 1) {
                    varnum_event = c(averageday[1:abs(deltaLength), ExtFunColsi], varnum_event)
                  } else {
                    a56 = Ndatapoints - abs(deltaLength)
                    a57 = Ndatapoints
                    varnum_event = c(varnum_event, averageday[a56:a57, ExtFunColsi])
                  }
                }
                if (anwi_index != 1) {
                  if (length(anwindices) > 0) {
                    varnum_event = as.numeric(varnum_event[anwindices]) #cut short varnum_event to match day segment of interest
                  } else {
                    varnum_event = c()
                  }
                }
              }
              #=====
              gUnitMetric = length(grep(x = colnames(vari)[mi], pattern = "BrondCount|ZCX|ZCY|ZCZ|marker|NeishabouriCount|ExtAct|ExtHeartRate", invert = TRUE)) > 0
              UnitReScale = ifelse(test = gUnitMetric, yes = 1000, no = 1)
              # Starting filling output matrix daysummary with variables per day segment and full day.
              if (minames[mi] %in% c("ZCX", "ZCY", "ZCZ", "BrondCount_x", "BrondCount_y",
                                     "BrondCount_z", "NeishabouriCount_x", "NeishabouriCount_y", 
                                     "NeishabouriCount_z", "NeishabouriCount_vm", "ExtAct", "ExtHeartRate",
                                     myfun$colnames) == FALSE) {
                unit = "_mg"
              } else if (minames[mi] == "ExtAct") {
                unit = ""
              } else if (minames[mi] == "ExtHeartRate") {
                unit = "_bpm"
              } else {
                unit = "_cnt"
              }
              if (isAccMetric == TRUE) {
                collectfi = c()
                for (winhr_value in params_247[["winhr"]]) { # Variable (column) names
                  # We are first defining location of variable names, before calculating
                  # variables
                  ML5colna = c(paste0("L",winhr_value,"hr"), paste0("L",winhr_value),
                               paste0("M",winhr_value,"hr"), paste0("M",winhr_value))
                  if (length(params_247[["iglevels"]]) > 0 & length(params_247[["MX.ig.min.dur"]]) == 1) { # intensity gradient (as described by Alex Rowlands 2018)
                    if (winhr_value >= params_247[["MX.ig.min.dur"]]) {
                      for (li in 1:2) { # do twice, once for LX and once for MX
                        varnameig = paste0(paste0(ifelse(li == 1, yes = "L", no = "M"),winhr_value,"_"), c("ig_gradient","ig_intercept","ig_rsquared"))
                        ML5colna = c(ML5colna, varnameig)
                      }
                    }
                  }
                  if (length(params_247[["qM5L5"]]) > 0) {
                    ML5colna = c(ML5colna,
                                 paste0("L", winhr_value, "_q", round(params_247[["qM5L5"]] * 100)),
                                 paste0("M", winhr_value, "_q", round(params_247[["qM5L5"]] * 100)))
                  }
                  # add metric name and timewindow name
                  fi = correct_fi(di, ds_names, fi, varname = paste0(ML5colna[1],"_", colnames(vari)[mi], unit,
                                                                     L5M5window_name))
                  ds_names[fi:(fi - 1 + length(ML5colna))] = paste0(ML5colna, "_", colnames(vari)[mi], unit, 
                                                                    L5M5window_name)
                  collectfi = c(collectfi, fi)
                  fi = fi + length(ML5colna)
                }
                if (length(varnum) > ((60 / ws3) * 60 * min(params_247[["winhr"]]) * 1.2)) { # Calculate values
                  for (winhr_value in params_247[["winhr"]]) {
                    exfi = collectfi[which(params_247[["winhr"]] == winhr_value)]
                    if (length(varnum) > (60 / ws3) * 60 * winhr_value * 1.2) { # Calculate values
                      # Time window for L5 & M5 analysis
                      t0_LFMF =  0 #start
                      # L5M5window[2] #end in 24 hour clock hours (if a value higher than 24 is chosen, it will take early hours of previous day to complete the 5 hour window
                      t1_LFMF = length(varnum) / (60 * (60 / ws3)) + (winhr_value - (params_247[["M5L5res"]] / 60))
                      ML5 = g.getM5L5(varnum, epochSize = ws3, t0_LFMF, t1_LFMF,
                                      params_247[["M5L5res"]], winhr_value,
                                      qM5L5 = params_247[["qM5L5"]],
                                      iglevels = params_247[["iglevels"]],
                                      MX.ig.min.dur = params_247[["MX.ig.min.dur"]],
                                      UnitReScale = UnitReScale)
                      ML5colna = colnames(ML5)
                      ML5 = as.numeric(ML5)
                      if (anwi_index > 1) {
                        L5M5shift = qwindow_values_backup[anwi_index - 1] #qwindow
                      } else {
                        L5M5shift = 0
                      }
                      if (length(ML5) > 3) {
                        ML5 = as.numeric(ML5)
                        ML5[c(1, 3)] = ML5[c(1, 3)] + L5M5shift
                        daysummary[di, (exfi):(exfi + length(ML5) - 1)] = ML5
                      } else {
                        daysummary[di, (exfi):(exfi + length(ML5) - 1)] = ""
                      }
                    } else {
                      daysummary[di,(exfi):(exfi + (4 * length(params_247[["winhr"]])) - 1 + (length(params_247[["qM5L5"]]) * 2))] = ""
                    }
                    # (Below) 4 is the length of ML5 and then 2 extra variables for every qM5L5 value
                    # winhr is not considered because we are in the winhr loop:
                    exfi = exfi + 4 + (length(params_247[["qM5L5"]]) * 2)
                  }
                } else {
                  daysummary[di,collectfi] = ""
                }
                if (anwindices[1] == 1 & length(anwindices) > 6 * 60 * (60/ws3)) {
                  # only derive if 1-6am falls within window
                  fi = correct_fi(di, ds_names, fi, varname = paste0("mean_", colnames(vari)[mi], unit, "_1-6am"))
                  daysummary[di,fi] = mean(varnum[((1 * 60 * (60 / ws3)) + 1):(6 * 60 * (60 / ws3))]) * UnitReScale #from 1am to 6am
                  ds_names[fi] = paste0("mean_",colnames(vari)[mi], unit, "_1-6am"); fi = fi + 1
                }
                if (anwi_nameindices[anwi_index] == "") { # for consistency with previous GGIR version
                  anwi_nameindices[anwi_index] = "_24hr"
                }
                cn_metashort = colnames(vari)
                fi = correct_fi(di, ds_names, fi, varname = paste0("mean_", cn_metashort[mi], unit,
                                                                   anwi_nameindices[anwi_index]))
                if (length(varnum) > 0) {
                  daysummary[di,fi] = mean(varnum) * UnitReScale
                } else {
                  daysummary[di,fi] = ""
                }
                ds_names[fi] = paste0("mean_", cn_metashort[mi], unit, anwi_nameindices[anwi_index]); fi = fi + 1
                if (anwi_nameindices[anwi_index] == "_24hr") {
                  anwi_nameindices[anwi_index] = ""
                }
                if (doquan == TRUE) {
                  q46 = c()
                  q46 = quantile(varnum, probs = params_247[["qlevels"]], na.rm = T, type = quantiletype) * UnitReScale
                  keepindex_46[mi - 1,] = c(fi,(fi + (length(params_247[["qlevels"]]) - 1)))
                  namesq46 = rep(0,length(rownames(as.matrix(q46))))
                  for (rq46i in 1:length(rownames(as.matrix(q46)))) {
                    tmp1 = rownames(as.matrix(q46))[rq46i]
                    tmp2 = as.character(unlist(strsplit(tmp1, "%")))
                    namesq46[rq46i] = paste0("p", tmp2, "_", cn_metashort[mi], unit,
                                             anwi_nameindices[anwi_index])
                  }
                  fi = correct_fi(di, ds_names, fi, varname = namesq46[1])
                  if (length(varnum) > 0) {
                    daysummary[di, fi:(fi + (length(params_247[["qlevels"]]) - 1))] = q46
                  } else {
                    daysummary[di, fi:(fi + (length(params_247[["qlevels"]]) - 1))] = ""
                  }
                  ds_names[fi:(fi + (length(params_247[["qlevels"]]) - 1))] = namesq46
                  fi = fi + length(params_247[["qlevels"]])
                }
                if (doilevels == TRUE) {
                  q48 = c()
                  #times 1000 to convert to mg only if it g-unit metric
                  q47 = cut((varnum * UnitReScale), breaks = params_247[["ilevels"]], right = FALSE)
                  q47 = table(q47)
                  q48  = (as.numeric(q47) * ws3) / 60 #converting to minutes
                  keepindex_48[mi - 1,] = c(fi, (fi + (length(q48) - 1)))
                  namesq47 = rep(0, length(rownames(q47)))
                  for (rq47i in 1:length(rownames(q47))) {
                    namesq47[rq47i] = paste0(rownames(q47)[rq47i], "_", cn_metashort[mi], unit,
                                             anwi_nameindices[anwi_index])
                  }
                  fi = correct_fi(di, ds_names, fi, varname = namesq47[1])
                  if (length(varnum) > 0) {
                    daysummary[di,fi:(fi + (length(q48) - 1))] = q48
                  } else {
                    daysummary[di, fi:(fi + (length(q48) - 1))] = ""
                  }
                  ds_names[fi:(fi + (length(q48) - 1))] = namesq47
                  fi = fi + length(q48)
                }
                if (doiglevels == TRUE) { # intensity gradient (as described by Alex Rowlands 2018)
                  q49 = c()
                  q50 = cut((varnum * UnitReScale), breaks = params_247[["iglevels"]], right = FALSE)
                  q50 = table(q50)
                  q49  = (as.numeric(q50) * ws3) / 60 #converting to minutes
                  x_ig = zoo::rollmean(params_247[["iglevels"]], k = 2)
                  y_ig = q49
                  igout = g.intensitygradient(x_ig, y_ig)
                  varnameig = paste0(c("ig_gradient", "ig_intercept", "ig_rsquared"),
                                     paste0("_", cn_metashort[mi], anwi_nameindices[anwi_index]))
                  fi = correct_fi(di, ds_names, fi, varname = varnameig[1])
                  if (length(varnum) > 0) {
                    daysummary[di, fi:(fi + 2)] = as.vector(unlist(igout))
                  } else {
                    daysummary[di, fi:(fi + 2)] = rep("", 3)
                  }
                  ds_names[fi:(fi + 2)] = varnameig
                  fi = fi + 3
                }
                #=========================================
                if (domvpa == TRUE) {
                  for (mvpai in 1:length(params_phyact[["mvpathreshold"]])) {
                    mvpa = rep(0,6)
                    if (length(varnum) < 100) {
                      mvpa[1:6] = 0
                    } else {
                      # METHOD 1: time spent above threhold based on 5 sec epoch
                      mvpa[1] = length(which(varnum * UnitReScale >= params_phyact[["mvpathreshold"]][mvpai])) / (60/ws3) #time spent MVPA in minutes
                      # METHOD 2: time spent above threshold based on 1minute epoch
                      varnum2 = cumsum(c(0, varnum))
                      select = seq(1, length(varnum2), by = 60/ws3)
                      varnum3 = diff(varnum2[round(select)]) / abs(diff(round(select)))
                      mvpa[2] = length(which(varnum3 * UnitReScale >= params_phyact[["mvpathreshold"]][mvpai])) #time spent MVPA in minutes
                      # METHOD 3: time spent above threshold based on 5minute epoch
                      select = seq(1, length(varnum2), by = 300/ws3)
                      varnum3 = diff(varnum2[round(select)]) / abs(diff(round(select)))
                      mvpa[3] = length(which(varnum3 * UnitReScale >= params_phyact[["mvpathreshold"]][mvpai])) * 5 #time spent MVPA in minutes
                      # METHOD 4: time spent above threshold
                      boutduration = params_phyact[["mvpadur"]][1] * (60/ws3) # per minute
                      rr1 = matrix(0, length(varnum), 1)
                      p = which(varnum * UnitReScale >= params_phyact[["mvpathreshold"]][mvpai]); rr1[p] = 1
                      getboutout = g.getbout(x = rr1, boutduration = boutduration,
                                             boutcriter = params_phyact[["boutcriter"]], ws3 = ws3)
                      mvpa[4] = length(which(getboutout == 1)) / (60/ws3) #time spent MVPA in minutes
                      
                      # METHOD 5: time spent above threshold 5 minutes
                      boutduration = params_phyact[["mvpadur"]][2] * (60/ws3) #per five minutes
                      rr1 = matrix(0, length(varnum), 1)
                      p = which(varnum * UnitReScale >= params_phyact[["mvpathreshold"]][mvpai]); rr1[p] = 1
                      getboutout = g.getbout(x = rr1, boutduration = boutduration,
                                             boutcriter = params_phyact[["boutcriter"]], ws3 = ws3)
                      mvpa[5] = length(which(getboutout == 1))   / (60/ws3) #time spent MVPA in minutes
                      # METHOD 6: time spent above threshold 10 minutes
                      boutduration = params_phyact[["mvpadur"]][3] * (60/ws3) # per ten minutes
                      rr1 = matrix(0,length(varnum),1)
                      p = which(varnum * UnitReScale >= params_phyact[["mvpathreshold"]][mvpai]); rr1[p] = 1
                      getboutout = g.getbout(x = rr1, boutduration = boutduration,
                                             boutcriter = params_phyact[["boutcriter"]], ws3 = ws3)
                      mvpa[6] = length(which(getboutout == 1)) / (60/ws3) #time spent MVPA in minutes
                    }
                    if (length(which(is.nan(mvpa) == TRUE)) > 0) mvpa[which(is.nan(mvpa) == TRUE)] = 0
                    mvpanames[,mvpai] = c( paste0("MVPA_E", ws3, "S_T", params_phyact[["mvpathreshold"]][mvpai]),
                                           paste0("MVPA_E1M_T", params_phyact[["mvpathreshold"]][mvpai]),
                                           paste0("MVPA_E5M_T", params_phyact[["mvpathreshold"]][mvpai]),
                                           paste0("MVPA_E", ws3, "S_B", params_phyact[["mvpadur"]][1], "M", (params_phyact[["boutcriter"]] * 100), "%_T", params_phyact[["mvpathreshold"]][mvpai]),
                                           paste0("MVPA_E", ws3, "S_B", params_phyact[["mvpadur"]][2], "M", (params_phyact[["boutcriter"]] * 100), "%_T", params_phyact[["mvpathreshold"]][mvpai]),
                                           paste0("MVPA_E", ws3, "S_B", params_phyact[["mvpadur"]][3], "M", (params_phyact[["boutcriter"]] * 100), "%_T", params_phyact[["mvpathreshold"]][mvpai]))
                    for (fillds in 1:6) {
                      mvpavarname = paste0(mvpanames[fillds,mvpai], "_", cn_metashort[mi], anwi_nameindices[anwi_index])
                      fi = correct_fi(di, ds_names, fi, varname = mvpavarname)
                      daysummary[di,fi] = mvpa[fillds]
                      ds_names[fi] = mvpavarname; fi = fi + 1
                    }
                  }
                }
                
                if (length(ExtFunColsi_backup) > 0) { # If events are detected with external function
                  # bout detection depends on access to non-wear bout variables
                  # so, skip detection of event bouts when working with external function output
                  if (mi %in% ExtFunColsi == FALSE) { # INSERT HERE VARIABLES DERIVED WITH EXTERNAL FUNCTION
                    # if (length(rti) == 1 && myfun$reporttype[rti] == "event") {
                    if (length(varnum_event) > 0) {
                      # Step bout detection
                      eventBouts = detectEventBouts(myfun, varnum_event = varnum_event,
                                                    varnum = varnum,
                                                    UnitReScale = UnitReScale,
                                                    daysummary = daysummary,
                                                    ds_names = ds_names,
                                                    di = di, fi = fi,
                                                    ws3 = ws3,
                                                    boutnameEnding = paste0(cn_metashort[mi],
                                                                            anwi_nameindices[anwi_index]))
                      
                      daysummary = eventBouts$daysummary
                      ds_names = eventBouts$ds_names
                      di = eventBouts$di
                      fi = eventBouts$fi
                    }
                  }
                }
              }
              if (mi %in% ExtFunColsi == TRUE) { # INSERT HERE VARIABLES DERIVED WITH EXTERNAL FUNCTION
                if (myfun$reporttype[rti] == "event") { # For the event report type we take the sum
                  segmentInfo = list(anwi_nameindices = anwi_nameindices,
                                     anwi_index = anwi_index,
                                     anwi_t0 = anwi_t0,
                                     anwi_t1 = anwi_t1)
                  eventAgg = aggregateEvent(metric_name = cn_metashort[mi],
                                            epochsize = ws3, 
                                            daysummary = daysummary,
                                            ds_names = ds_names,
                                            fi = fi, di = di,
                                            vari = vari,
                                            segmentInfo, myfun, params_247)
                  daysummary = eventAgg$daysummary
                  ds_names = eventAgg$ds_names
                  fi = eventAgg$fi
                } else if (myfun$reporttype[rti] == "scalar") { # For the scalar report type we take the mean
                  # Scalars are already summarised as mean
                  # Specific summary statistics for non-acceleration scalar
                  # such as heart rate have not been implemented yet
                } else if (myfun$reporttype[rti] == "type") { # For type we calculate time spent in each class
                  # Not implemented yet
                }
              }
            }
          }
        }
      }
      rm(val); rm(vari)
    }
  }
  invisible(list(daysummary = daysummary, ds_names = ds_names,
                 windowsummary = windowsummary, ws_names = ws_names))
}

splitRecords = function(metadatadir, params_general = NULL) {
  
  desiredtz = params_general[["desiredtz"]]
  idloc = params_general[["idloc"]]
  windowsizes = params_general[["windowsizes"]]
  
  filefoldername = filename_dir = tail_expansion_log = NULL
  
  # Create overview of all recordings ID, start time, end time, and filename
  fns = dir(paste0(metadatadir, "/meta/basic"), full.names = TRUE)
  fns = fns[grep(pattern = "_split", x = basename(fns), invert = TRUE)]
  if (length(fns) == 0) return()
  S = do.call("rbind", lapply(X = fns, FUN = getPart1BasicInfo, idloc = idloc, tz = desiredtz)) 
  #------------------------------------
  # Load recording split times file
  splitTime = data.table::fread(params_general[["recording_split_times"]], data.table = FALSE, stringsAsFactors = FALSE, colClasses = "character")
  
  # Identify ID columns
  IDcol = grep(pattern = "ID", x = colnames(splitTime))
  if (length(IDcol) == 0) {
    stop(paste0("\nFile specified by recording_split_times does not have a ",
                "column with ID in it, please fix."), call. = FALSE)
  }
  # Make sure timestamps have date and time
  countSpaces = function(x) {
    return(length(unlist(strsplit(x, " "))) - 1)
  }
  if (length(grep(pattern = "%H:%M:%S",
                  x = params_general[["recording_split_timeformat"]])) > 0) {
    defaultTime = "00:00:00"
  } else {
    defaultTime = "00:00"
  }
  for (j in (IDcol + 1):ncol(splitTime)) {
    space_count = unlist(lapply(X = splitTime[, j], FUN = countSpaces))
    no_space = which(space_count == 0)
    if (length(no_space) > 0) {
      splitTime[no_space, j] = paste0(splitTime[no_space, j], " ", defaultTime)
    }
  }
  if (length(S) > 0) S = S[!is.na(S$ID),]
  if (length(S) > 0 & length(splitTime) > 0) {
    # Identify recordings that need to be split
    buffer = params_general[["recording_split_overlap"]] * 3600
    for (j in 1:nrow(S)) {
      # matching ID(s) between milestone data and splitTime
      thisID = which(splitTime[, IDcol] == S$ID[j])
      if (length(thisID) > 0) {
        for (i in 1:length(thisID)) {
          splitTime_tmp = as.character(splitTime[thisID[i], (IDcol + 1):ncol(splitTime)])
          splitTime_tmp = as.POSIXct(splitTime_tmp, tz = desiredtz, format = params_general[["recording_split_timeformat"]])
          
          splitTime_tmp = as.POSIXct(round(as.numeric(splitTime_tmp) / windowsizes[2]) * windowsizes[2], tz = desiredtz)
          split_names = colnames(splitTime)[(IDcol + 1):ncol(splitTime)]
          # Only consider timestamps that overlap with recording
          within_time_range = which(splitTime_tmp >= S$start[j] &
                                      splitTime_tmp <= S$end[j])
          # If first split time was skipped but overlaps for more than 12 hours
          # with the accelerometer recording include it
          
          # minutes betweens start recording and first split that overlaps
          mins_start_firstoverlap = as.numeric(difftime(splitTime_tmp[within_time_range[1]], S$start[j], units = "mins"))
          if (within_time_range[1] != 1 &&
              mins_start_firstoverlap > 12 * 60) {
            within_time_range = c(within_time_range[1] - 1, within_time_range)
          }
          # If last split time was skipped but overlaps for more than 12 hours
          # with the accelerometer recording include it
          
          # minutes between last split that overlaps and end recording
          mins_lastoverlap_end = as.numeric(difftime(S$end[j], splitTime_tmp[within_time_range[length(within_time_range)]], units = "mins"))
          if (splitTime_tmp[max(within_time_range)] < splitTime_tmp[length(splitTime_tmp)] &
              mins_lastoverlap_end > 12 * 3600) {
            within_time_range = c(within_time_range, within_time_range[length(within_time_range)] + 1)
          }
          splitTime_tmp = splitTime_tmp[within_time_range]
          
          if (length(splitTime_tmp) == 0) next
          split_names = split_names[within_time_range]
          # tidy up split_names
          split_names = tolower(gsub(pattern = " |[.]|-|TO", replacement = "", x = split_names))
          split_names = substr(split_names, start = 1, stop = 10) # consider max 8 characters
          split_names = make.unique(split_names, sep = "") # make names unique
          if (all(is.na(splitTime_tmp))) {
            stop(paste0("Timestamp format ", splitTime[thisID[i], (IDcol + 1):ncol(splitTime)],
                        " not recognised. You may want to check parameter ",
                        "recording_split_timeformat", call. = FALSE))
          }
          # If yes, use those splits to split the recording
          load(file = S$filename[j])
          timestamp_short = iso8601chartime2POSIX(x = M$metashort$timestamp, tz = desiredtz)
          timestamp_long = iso8601chartime2POSIX(x = M$metalong$timestamp, tz = desiredtz)
          Mbu = M
          segment_names = segment_starts = segment_ends = NULL
          # Define segments
          if (splitTime_tmp[1] > timestamp_short[1] &&
              params_general[["recording_split_ignore_edges"]] == FALSE) {
            segment_starts = timestamp_short[1]
            segment_ends = splitTime_tmp[1]
            segment_names = paste0("startrecTO", split_names[1])
          }
          for (segment_index in 1:length(splitTime_tmp)) {
            if (segment_index < length(splitTime_tmp)) {
              segment_starts = c(segment_starts, splitTime_tmp[segment_index])
              segment_ends = c(segment_ends, splitTime_tmp[segment_index + 1])
              segment_names = c(segment_names, paste0(split_names[segment_index], "TO",
                                                      split_names[segment_index + 1]))
            } else {
              if (params_general[["recording_split_ignore_edges"]] == FALSE) {
                segment_starts = c(segment_starts, splitTime_tmp[segment_index])
                segment_ends = c(segment_ends, timestamp_short[length(timestamp_short)])
                segment_names = c(segment_names, paste0(split_names[segment_index],
                                                        "TOendrec"))
              }
            }
          }
          # Store each part separately
          if (length(segment_starts) > 0) {
            file_was_split = FALSE
            # round to resolution that matches long epoch
            # segment_starts = as.POSIXct(round(as.numeric(segment_starts) / windowsizes[2]) * windowsizes[2], tz = desiredtz)
            buffer = round((buffer / 2) / windowsizes[2]) * windowsizes[2]
            for (g in 1:length(segment_starts)) {
              # Take subsection
              segment_short = which(timestamp_short >= segment_starts[g] - buffer &
                                      timestamp_short < segment_ends[g] + buffer)
              segment_long = which(timestamp_long >= segment_starts[g] - buffer &
                                     timestamp_long <= segment_ends[g] + buffer)
              # Make long a multitude of short
              Nshort = length(segment_short)
              Nlong = length(segment_long)
              segment_short = segment_short[1:pmin(Nshort, Nlong * (windowsizes[2] / windowsizes[1]))]
              Nshort = length(segment_short)
              segment_long = segment_long[1:pmin(Nlong, floor(Nshort / (windowsizes[2] / windowsizes[1])))]
              
              Nhours_long = length(segment_long) /  (3600 / windowsizes[2])
              # Only save if there are at least 12 hours of data
              if (Nhours_long > 2) {
                M$metashort = Mbu$metashort[segment_short, ]
                M$metalong = Mbu$metalong[segment_long, ]
                # Save the split
                
                # Take RData filename and split it based on dot
                fname_tmp = unlist(strsplit(basename(S$filename[j]), "[.]"))
                # Last two elements are the RData + Original file extension (e.g. .AWD, .csv)
                newfilebase = paste0(fname_tmp[1:pmax(1, length(fname_tmp) - 2)], collapse = ".")
                # Remember original file extension
                extension = fname_tmp[length(fname_tmp) - 1]
                # Reassamble file name but use character used for locating the ID
                # to ensure that the new part never becomes part of the extracted ID
                # In other words activity and sleep diary (part 2, 4 and 5) can
                # keep using the general participant ID and do not need to work with
                # segment specific IDs
                if (idloc == 6) {
                  id_separator = "._"
                } else {
                  id_separator = "_"
                }
                newFileName = paste0(newfilebase, id_separator, "split", g, "_", segment_names[g], ".", extension)
                # add back to I$filename because part 2 uses this in the csv report
                filename_dir = gsub(pattern = "meta_", replacement = "", x = newFileName)
                I$filename = filename_dir
                newRDataFileName = paste0(dirname(S$filename[j]), "/", newFileName, ".RData")
                file_was_split = TRUE
                
                # update weekday code and name
                wday = as.POSIXlt(x = timestamp_short[1], tz = desiredtz)$wday + 1
                weekdays = c("Sunday","Monday","Tuesday","Wednesday","Thursday","Friday","Saturday")
                M$wday = wday
                M$wdayname = weekdays[wday]
                # save
                save(M, C, I,
                     filefoldername, filename_dir, tail_expansion_log,
                     file = newRDataFileName)
              }
            }
            if (file_was_split == TRUE) {
              # Delete original RData file
              unlink(S$filename[j], recursive = TRUE)
            }
          }
        }
      }
    }
  }
  # THIS FUNCTION DOES NOT PRODUCE OUTPUT IT ONLY SPLITS FILES
  return()
}

getfolderstructure = function(datadir = c(),
                              referencefnames = c()) {
  
  filelist = isfilelist(datadir)
  if (filelist == FALSE) {
    fnamesfull = dir(datadir, recursive = TRUE, full.names = TRUE, pattern = "[.](csv|bin|Rda|wav|cw|gt3)")
  } else {
    fnamesfull = datadir
  }
  fullfilenames = foldername = rep("",length(referencefnames))
  if (length(fnamesfull) > 0) {
    fnamesshort = apply(
      X = as.matrix(fnamesfull),
      MARGIN = 1,
      FUN = function(X)
        basename(X)
    )
    foldername_new = apply(
      X = as.matrix(fnamesfull),
      MARGIN = 1,
      FUN = function(X)
        basename(dirname(X))
    )
    for (i in 1:length(referencefnames)) {
      index_match = which(fnamesshort == referencefnames[i])
      if (length(index_match) > 0 & referencefnames[i] %in% c("", " ", NA) == FALSE) {
        fullfilenames[i] = fnamesfull[index_match]
        foldername[i] = foldername_new[index_match]
      }
    }
  }
  invisible(list(fullfilenames = fullfilenames, foldername = foldername))
}

library(GGIR)
context("getfolderstructure")
test_that("getfolderstructure", {
  dn = "A" # name of the datadir folder
  if (file.exists(dn)) unlink(dn, recursive = TRUE)
  dummyfolderA = "A"
  dummyfolderB = "B"
  dummyfolderC = "C"
  dir.create(file.path("A"))
  dir.create(file.path("A/B"))
  dir.create(file.path("A/C"))
  
  write.csv(matrix(0, 4, 4), file = "A/B/testB.csv")
  write.csv(matrix(0, 4, 4), file = "A/C/testC.csv")
  
  folderstructure = getfolderstructure(datadir = dn,
                                       referencefnames = c("testB.csv", "testC.csv"))
  expect_that(  folderstructure$foldername[1],equals("B"))
  expect_that(  folderstructure$foldername[2],equals("C"))
  if (file.exists(dn))
    unlink(dn, recursive = TRUE
    )
})

cosinor_IS_IV_Analyses = function(Xi, epochsize = 60, timeOffsetHours = 0, threshold = NULL) {
  if (length(threshold) > 1) {
    threshold = threshold[1]
    warning("Multiple threshold values supplied to cosinor analysis, only first value used.")
  }
  # Apply Cosinor function from ActRC
  N = 1440 * (60 / epochsize) # Number of epochs per day
  Xi = Xi[1:(N * floor(length(Xi) / N))] # ActCR expects integer number of days
  
  # omit all days at the end with no data
  end2 = length(Xi)
  while (end2 >= N) {
    if (all(is.na(Xi[(end2 - N + 1):end2]))) {
      end2 = end2 - N
      Xi = Xi[1:end2]
    } else {
      end2 = -1
    }
  }
  # transform data to millig if data is stored in g-units
  notna = !is.na(Xi)
  if (max(Xi, na.rm = TRUE) < 13 && mean(Xi, na.rm = TRUE) < 1) {
    # 13 because a typical 8g accelerometer could in theory measure 7.5 in each axis without
    # being considered clipping, which results in a vector of 13
    # as soon as the time series has values above 13 then it is most likely that
    # it is either expressed in counts or in mg.
    Xi[notna] = Xi[notna] * 1000
  }
  # log transform data for ActCosinor, IV IS further down will use the non-transformed signal
  Xi_log = Xi
  Xi_log[notna] = log(Xi[notna] + 1)
  coef = ActCR::ActCosinor(x = Xi_log, window = 1440 / N)

  # Apply Extended Cosinor function from ActRC (now temporarily turned of to apply my own version)
  coefext = ActCR::ActExtendCosinor(x = Xi_log, window = 1440 / N, export_ts = TRUE)
  # Correct time estimates by offset in start of recording
  add24ifneg = function(x) {
    if (x < 0) x = x + 24
    return(x)
  }
  coef$params$acrotime = add24ifneg(coef$params$acrotime - timeOffsetHours)
  coefext$params$UpMesor = add24ifneg(coefext$params$UpMesor - timeOffsetHours)
  coefext$params$DownMesor = add24ifneg(coefext$params$DownMesor - timeOffsetHours)
  coefext$params$acrotime = add24ifneg(coefext$params$acrotime - timeOffsetHours)
  # do same for acrophase in radians (24 hours: 2 * pi)
  # take absolute value of acrophase, because it seems ActCR provides negative value in radians,
  # which is inverse correlated with acrotime
  coef$params$acr = abs(coef$params$acr) - ((timeOffsetHours / 24) * 2 * pi)
  k = ceiling(abs(coef$params$acr) / (pi * 2))
  if (coef$params$acr < 0) coef$params$acr = coef$params$acr + (k * 2 * pi)
  # Perform IVIS on the same input signal to allow for direct comparison
  IVIS = g.IVIS(Xi = Xi,
                epochSize = epochsize, 
                threshold = threshold) # take log, because Xi is logtransformed with offset of 1
  
  coefext$params$R2 = cor(coefext$cosinor_ts$original, coefext$cosinor_ts$fittedYext)^2
  coef$params$R2 = cor(coefext$cosinor_ts$original, coefext$cosinor_ts$fittedY)^2
  
 # this should equal: https://en.wikipedia.org/wiki/Coefficient_of_determination

  invisible(list(coef = coef, coefext = coefext, IVIS = IVIS))
}

library(GGIR)
context("cosinor_IS_IV_Analyses")

test_that("cosinorAnalyses provides expected output", {
  ActCosDummy = function(epochSizeSeconds, missingdata = FALSE, timeOffsetHours = 0) {
    N = 1440 * (60 / epochSizeSeconds) # Number of epochs per day
    if (timeOffsetHours == 0) {
      time = seq(1/N, 7, by = 1 / N)
    } else {
      time = seq(-timeOffsetHours/24 + (1/N), -(timeOffsetHours/24) + 7, by = 1 / N)
    }
    counts = sin(time * 2 * pi) + 1
    if (missingdata == TRUE) {
      is.na(counts[N:floor(N * 2)]) = TRUE
    }
    # Plot code below commented out, but useful for debugging to 
    # understand how dummy time series looks like.
    # x11()
    # plot(log((counts * 1000) + 1), type = "l")
    mod = cosinor_IS_IV_Analyses(Xi = counts * 1000, epochsize = epochSizeSeconds, timeOffsetHours = timeOffsetHours)
    mod$coefext = mod$coefext[which(names(mod$coefext) != "cosinor_ts")]
    return(mod)
  }
  # no missing data
  coef60 = ActCosDummy(epochSizeSeconds = 60)
  coef300 = ActCosDummy(epochSizeSeconds = 300)
  coef60_Offset9 = ActCosDummy(epochSizeSeconds = 60, timeOffsetHours = 9)
  coef60_Offset17 = ActCosDummy(epochSizeSeconds = 60, timeOffsetHours = 17)
  
  expect_equal(coef60_Offset9$coef, coef60$coef, tolerance = 0.001)
  expect_equal(coef60_Offset17$coef, coef60$coef, tolerance = 0.001)
  expect_equal(coef60_Offset9$coefext, coef60$coefext, tolerance = 0.001)
  expect_equal(coef60_Offset17$coefext, coef60$coefext, tolerance = 0.001)
  
  # standard cosinor analyses
  expect_equal(coef60$coef$params$mes, 6.259326, tolerance = 0.01)
  expect_equal(coef60$coef$params$amp, 1.912535, tolerance = 0.01)
  expect_equal(coef60$coef$params$acr, pi * 1/2, tolerance = 0.01)
  expect_equal(coef60$coef$params$acrotime, 6, tolerance = 0.01)
  expect_equal(coef60$coef$params$ndays, 7)
  expect_equal(coef60$coef, coef60$coef, tolerance = 0.01)
  expect_equal(coef60$coef, coef300$coef, tolerance = 0.01)
  expect_equal(coef60$IVIS, coef300$IVIS, tolerance = 0.01)
  
  # extended cosinor analyses
  expect_equal(coef60$coefext$params$minimum, 0)
  expect_equal(coef60$coefext$params$amp, 7.245069, tolerance = 0.01)
  expect_equal(coef60$coefext$params$alpha, -0.921499, tolerance = 0.01)
  expect_equal(coef60$coefext$params$beta, 5.939461, tolerance = 0.01)
  expect_equal(coef60$coefext$params$UpMesor, 19.52358, tolerance = 0.01)
  expect_equal(coef60$coefext$params$DownMesor, 16.47642, tolerance  = 0.01)
  expect_equal(coef60$coefext$params$MESOR, 3.622534, tolerance  = 0.01)
  expect_equal(coef60$coefext$params$R2, 0.8976208, tolerance  = 0.01)
  
  # IV IS
  expect_equal(coef60$IVIS$InterdailyStability, 1, tolerance  = 0.01)
  expect_equal(coef60$IVIS$IntradailyVariability, 0.06774028, tolerance  = 0.01)
  
  fields_to_compare = c("minimum", "amp", "alpha", "beta", "acrotime", "DownMesor", "MESOR")
  expect_equal(coef60$coefext[fields_to_compare], coef300$coefext[fields_to_compare], tolerance = 0.1)
  
  # handling of missing data
  coef60_missingdata = ActCosDummy(epochSizeSeconds = 60, missingdata = TRUE)
  expect_equal(coef60$coef, coef60_missingdata$coef, tolerance = 0.1)
  expect_equal(coef60$coefext[fields_to_compare], coef60_missingdata$coefext[fields_to_compare], tolerance = 0.5)
})

g.IVIS = function(Xi, epochSize = 60, threshold = NULL) {
  if (!is.null(threshold)) {
    Xi = ifelse(Xi > threshold, 0, 1)  
  }
  if (epochSize < 3600) {
    # Always aggregate Xi to 1 hour resolution if it is not provided at 1 hour resolution
    df = data.frame(Xi = Xi, time = numeric(length(Xi)))
    time = seq(0, length(Xi) * epochSize, by = epochSize)
    df$time = time[1:nrow(df)]
    df = aggregate(x = df, by = list(floor(df$time / 3600) * 3600), 
                   FUN = mean, na.action = na.pass, na.rm = TRUE)
    Xi = df$Xi
  }
  # ts hourly time series, including NA values
  hour = (1:ceiling(length(Xi))) - 1
  ts = data.frame(Xi = Xi, hour = hour, stringsAsFactors = TRUE)
  IS = IV = phi = NA
  if (nrow(ts) > 1) {
    ts$day = floor(ts$hour/24) + 1
    ts$hour = ts$hour - (floor(ts$hour / 24) * 24) # 24 hour in a day
    if (nrow(ts) > 1) {
      # Average day
      aveDay = aggregate(. ~ hour, data = ts, mean, na.action = na.omit)
      Xh = aveDay$Xi
      Xm = suppressWarnings(mean(Xh,na.rm = TRUE)) # Average acceleration per day
      deltaXi = diff(Xi)^2
      N = length(Xi[!is.na(Xi)])
      
      # phi
      model = arima(Xi[!is.na(Xi)], order = c(1, 0, 0))
      phi = model$coef[[1]]

      # IS: lower is less synchronized with the 24 hour zeitgeber
      ISnum = sum((Xh - Xm)^2, na.rm = TRUE) * N
      ISdenom = 24 * sum((Xi - Xm)^2, na.rm = TRUE)
      IS = ISnum / ISdenom
      
      #IV: higher is more variability within days (fragmentation)
      IVnum = sum(deltaXi, na.rm = TRUE) * N
      IVdenom = (N - 1) * sum((Xi - Xm)^2, na.rm = TRUE)
      IV = IVnum / IVdenom 
    }
  }
  invisible(list(InterdailyStability = IS, IntradailyVariability = IV, phi = phi))
}

library(GGIR)
context("g.IVIS")
test_that("g.IVIS returns expected output value without missing data", {
  set.seed(1234)
  xx = rnorm(n = 1440) * 1000
  Xi1 = rep(xx, 7)
  Xi2 = rep(0:7, each = 1440)
  T1 = g.IVIS(Xi1, epochSize = 60)
  T2 = g.IVIS(Xi2, epochSize = 60)
  expect_equal(T1$InterdailyStability, 1, tolerance = 0.01) # every day is the same, so we expect IS = 1
  expect_equal(T2$InterdailyStability, 0, tolerance = 0.01) # every day is entirely different, so IS = 0
  expect_equal(T1$IntradailyVariability, 2.244335, tolerance = 0.05) #within the day there is variation, so IV>0
  expect_equal(T2$IntradailyVariability, 0, tolerance = 0.01) #within the day there is no variation, so IV=0
  expect_equal(T1$phi, -0.1453851, tolerance = 0.001)
  expect_equal(T2$phi, 0.9982517, tolerance = 0.001)
})

test_that("g.IVIS returns expected output value with missing data", {
  set.seed(1234)
  xx = abs(rnorm(n = 1440))
  Xi1 = rep(xx, 7)
  Xi2 = rep(0:7, each = 1440)
  is.na(Xi1[1441:2880]) = TRUE
  is.na(Xi2[1441:2880]) = TRUE
  T3 = g.IVIS(Xi1, epochSize = 60)
  T4 = g.IVIS(Xi2, epochSize = 60)
  expect_equal(T3$InterdailyStability, 1, tolerance = 0.01) # every day is the same, so we expect IS = 1
  expect_equal(T4$InterdailyStability, 0, tolerance = 0.01) # every day is entirely different, so IS = 0
  expect_equal(T3$IntradailyVariability, 2.51238, tolerance = 0.05) #within the day there is variation, so IV>0
  expect_equal(T4$IntradailyVariability, 0, tolerance = 0.01) #within the day there is no variation, so IV=0
  expect_equal(T3$phi, -0.2577584, tolerance = 0.001)
  expect_equal(T4$phi, 0.9972666, tolerance = 0.001)
})

g.extractheadervars = function(I) {
  header = I$header
  mon = I$monn
  hnames = rownames(header)
  hvalues = as.character(as.matrix(header))
  pp = which(hvalues == "")
  if (length(pp) > 0) hvalues[pp] = c("not stored in header")
  # set defaults
  ID = I$filename # recording ID
  SX = "not available" # sex
  iID = "not extracted" #investigator ID
  HN = "not extracted" #handedness
  sensor.location = "not extracted" #body location
  deviceSerialNumber = "not extracted"
  # attempt to extract from hvalues and hnames
  if (mon == "genea") {
    IDd = hvalues[which(hnames == "Volunteer_Number")]    	
    ID = as.character(unlist(IDd))
    iIDd = hvalues[which(hnames == "Investigator_Id")]			
    iID = as.character(unlist(iIDd))
    sensor.location = hvalues[which(hnames == "Body_Location")]
    deviceSerialNumber = hvalues[which(hnames == "Serial_Number")] #serial number
  } else if (mon == "geneactive") {
    check_GENEAread = which(hnames == "Subject_Code")
    if (length(check_GENEAread) > 0) { 
      # This if-statement can be deprecated once GENEAread is deprecated as a dependency
      ID = hvalues[which(hnames == "Subject_Code")] #; temp2 = unlist(strsplit(as.character(temp)," "))
      iID = hvalues[which(hnames == "Investigator_ID")] #investigator ID
      HN = hvalues[which(hnames == "Handedness_Code")] #handedness
      sensor.location = as.character(as.matrix(hvalues[which(hnames == "Device_Location_Code")])) #body location
      SX = hvalues[which(hnames == "Sex")] #gender
      deviceSerialNumber = hvalues[which(hnames == "Device_Unique_Serial_Code")] #serial number
    } else {
      ID = hvalues[which(hnames == "RecordingID")]
      HN = hvalues[which(hnames == "Handedness")]
      sensor.location = hvalues[which(hnames == "DeviceLocation")]
      deviceSerialNumber = hvalues[which(hnames == "serial_number")] #serial number
    }
  } else if (mon == "actigraph" | mon == 'verisense') {
    if (I$dformn == "gt3x") {
      header = I$header
      deviceSerialNumber = as.character(header["Serial Number",])
      firmwareversion = as.character(header["Firmware",])
    } else { # .csv format
      deviceSerialNumber = as.character(I$header$value[which(row.names(I$header) == "Serial Number:")])
      if (length(deviceSerialNumber) == 0) deviceSerialNumber = "not extracted" #serial number
      # try to extract firmware version and append it to serial number,
      # at a later point in time this could become a separate variable in the reports
      firmwareversion = unlist(strsplit(unlist(strsplit(as.character(I$header$value[1]),"Firmware[.]"))[2],"[.]date"))[1]
    }
    if (length(firmwareversion) == 1) deviceSerialNumber = paste0(deviceSerialNumber,"_firmware_",firmwareversion)
  } else if (mon == "axivity" | mon == "movisens") {
    if (mon  == "actigraph") {
      deviceSerialNumber = as.character(I$header$value[grep(pattern = "serial number", x = row.names(I$header), ignore.case = TRUE, value = FALSE)])
    }
    if (mon == "axivity") {
      seriali = which(hnames %in% c("uniqueSerialCode", "IART2Id"))
      if (length(seriali) > 0) deviceSerialNumber = hvalues[seriali[1]] #serial number			
    }
    if (mon == "movisens") {
      deviceSerialNumber = as.character(I$header$value[which(row.names(I$header) == "serialnumber")])
      ID = as.character(I$header$value[which(row.names(I$header) == "ID")])
    }
  } else if (mon == "unknown") {
    if (length(which(hnames == "recordingID")) > 0) {
      ID = hvalues[which(hnames == "recordingID")]
    }
    if (length(which(hnames == "device_serial_number")) > 0) {
      deviceSerialNumber = hvalues[which(hnames == "device_serial_number")]			 #serial number
    }
  }
  invisible(list(ID = ID, iID = iID, HN = HN, sensor.location = sensor.location,
                 SX = SX, deviceSerialNumber = deviceSerialNumber)) #  wdayname=wdayname,wdaycode=wdaycode,wday=wday,ws3=ws3,ws2=ws2
}

g.report.part6 = function(metadatadir = c(), f0 = c(), f1 = c(),
                          params_cleaning = NULL, params_output,
                          verbose = TRUE) {
  # description: function to load milestone data generated by g.part6 and to merge these into a spreadsheet
  # here no additional information of analyses are added. This function therefore is primary to wrap up the up the output
  # from parallel processed accelerometer files
  ms6.out = "/meta/ms6.out"
  if (file.exists(paste(metadatadir, ms6.out, sep = ""))) {
    if (length(dir(paste(metadatadir, ms6.out, sep = ""))) == 0) {
      try.generate.report = FALSE #do not run this function if there is no milestone data from g.part5
    } else {
      try.generate.report = TRUE
    }
  } else {
    try.generate.report = FALSE #do not run this function if there is no milestone data from g.part5
    warning(paste0('Cannot generate part6 report because no part 6 milestone data",
                   " is available. First run part 6 with argument mode=6.'))
  }
  if (try.generate.report == TRUE) {
    #======================================================================
    # loop through meta-files
    fnames.ms6 = list.files(paste0(metadatadir, ms6.out), full.names = TRUE)
    if (f1 > length(fnames.ms6)) {
      f1 = length(fnames.ms6)
    }
    if (verbose == TRUE) {
      cat(paste0(" loading all the milestone data from part 6 this can",
                 " take a few minutes\n"))
    }
    myfun = function(x, expectedCols = c()) {
      output_part6 = NULL
      load(file = x)
      cut = which(output_part6[, 1] == "")
      if (length(cut) > 0 & length(cut) < nrow(output_part6)) {
        output_part6 = output_part6[-cut, which(colnames(output_part6) != "")]
      }
      out = as.matrix(output_part6)
      if (length(expectedCols) > 0) {
        tmp = as.data.frame(matrix(0, 0, length(expectedCols)))
        colnames(tmp) = expectedCols
        out = base::merge(tmp, out, all = TRUE)
      }
      if (any(c("cosinorIS", "cosinorIV", "GGIRversion") %in% colnames(out) == FALSE)) {
        out = as.data.frame(out)
        out$cosinorIS = NA
        out$cosinorIV = NA
        out$GGIRversion = ""
      }
      return(out)
    }
    # Use large milestone file as reference point, assuming that this has the largest number of columns
    all_files <- fnames.ms6[f0:f1]
    file_info <- file.info(all_files)
    file_sizes <- file_info$size
    # Find the file with the largest size. Since the .RData files in ms6.out are processed, the largest file is likely to have the most column names.
    largest_file <- all_files[which.max(file_sizes)]
    # Extract column names from this file
    out_try <- myfun(largest_file)
    expectedCols = colnames(out_try)
    outputfinal = as.data.frame(do.call(rbind,
                                        lapply(fnames.ms6[f0:f1], myfun, expectedCols)),
                                stringsAsFactors = FALSE)
    outputfinal = outputfinal[,which(colnames(outputfinal) %in% c("cosinorIS", "cosinorIV") == FALSE)]
    # Find columns filled with missing values
    cut = which(sapply(outputfinal, function(x) all(x == "")) == TRUE)
    if (length(cut) > 0) {
      outputfinal = outputfinal[,-cut]
    }
    # order data.frame
    outputfinal = outputfinal[order(outputfinal$filename, outputfinal$ID),]
    #-------------------------------------------------------------
    # store all summaries in csv files
    outputfinal_clean = tidyup_df(outputfinal)
    outputfinal_clean = addSplitNames(outputfinal_clean) # If recording was split
    data.table::fwrite(outputfinal_clean, paste0(metadatadir, "/results/part6_summary.csv"),
                       row.names = FALSE, na = "", sep = params_output[["sep_reports"]],
                       dec = params_output[["dec_reports"]])
    rm(outputfinal)
  }
}

g.report.part5_dictionary = function(metadatadir, params_output) {
  # internal function to get class ------------------------------------------
  getclass = function(x) {
    if (length(x) == 0) return(NULL)
    class = NULL
    if (grepl("2", x)) x = unlist(strsplit(x, "2"))
    for (i in 1:length(x)) {
      if (x[i] == "IN") class[i] = "inactivity"
      if (x[i] == "LIG" | x[i] == "LIPA") class[i] = "light physical activity"
      if (x[i] == "MOD") class[i] = "moderate physical activity"
      if (x[i] == "VIG") class[i] = "vigorous physical activity"
      if (x[i] == "MVPA") class[i] = "moderate-to-vigorous physical activity"
      if (x[i] == "PA") class[i] = "physical activity"
      if (x[i] == "sleep") class[i] = "sleep"
      if (x[i] == "wake") class[i] = "awake"
      if (x[i] == "nonwear") class[i] = "non-wear time"
      if (x[i] == "mostactive60min") class = "the most active 60 minutes of the day are accumulated"
      if (x[i] == "mostactive30min") class = "the most active 30 minutes of the day are accumulated"
    }
    if (length(class) == 2) class = paste(class[1], "to", class[2])
    return(class)
  }
  # -------------------------------------------------------------------------
  # main script -------------------------------------------------------------
  # -------------------------------------------------------------------------
  # identify individual reports
  reports = dir(file.path(metadatadir, "results"), full.names = TRUE, 
                pattern = "^part5.*\\.csv$")
  # Select one daysummary, personsummary, and Segment report as variables are the same across configurations
  ds = grep("^part5_daysummary", basename(reports))[1]
  ps = grep("^part5_personsummary", basename(reports))[1]
  reports = reports[c(ds, ps)]
  if (length(reports) == 0 || all(is.na(reports))) {
    # No cleaned part 5 report probably because no valid windows, so try full report instead
    reports = dir(file.path(metadatadir, "results", "QC"), full.names = TRUE, 
                    pattern = "^part5.*\\.csv$")
    if (length(reports) == 0 || all(is.na( reports))) return()
  }
  reports = reports[!is.na(reports)]
  if (!exists("baseDictionary")) return()
  # read col names of each report and derive definitions
  for (ri in 1:length(reports)) {
    # read report
    R = data.table::fread(reports[ri], verbose = FALSE, nrows = 2)
    cnames = colnames(R)
    # initialise dictionary
    dictionary = data.frame(Variable = cnames,
                            Definition = NA)
    # get definitions
    for (coli in 1:length(cnames)) {
      # definitions will be a combination of what, window, when (only for LUX segments), class, unit
      what = window = class = unit = NULL
      # get variable names from baseDictionary
      nam = gsub("_pla|_wei|_WD|_WE", "", cnames[coli])
      if (grepl("Nvalid", cnames[coli])) nam = cnames[coli]
      if (nam %in% names(baseDictionary)) {
        what = baseDictionary[[nam]]
      }
      # if the variable is not in baseDictionary, extract definition:
      if (is.null(what)) {
        # split up variable name 
        elements = unlist(strsplit(cnames[coli], "[.]|_"))
        # WHAT -------------------------------------------------------------
        if ("dur" %in% elements | "nonwear" %in% elements) {
          what = "Time accumulated"
        } else if ("ACC" %in% elements) {
          what = "Mean acceleration"
          if ("median" %in% elements) {
            what = "Median acceleration"
          } else if ("stdev" %in% elements) {
            what = "Standard deviation of acceleration"
          }
        } else if ("Nbouts" %in% elements) {
          what = "Number of bouts"
        } else if ("Nblocks" %in% elements) {
          what = "Number of blocks (defined as consecutive series of epochs with the same behavioural class)"
        } else if ("quantile" %in% elements) {
          what = "Acceleration above which (percentile)"
        } else if ("LUX" %in% elements) {
          if ("min" %in% elements | "above1000" %in% elements) {
            # time in lux ranges
            what = "Time accumulated with LUX"
            if ("above1000" %in% elements) {
              numbers = 1000
            } else {
              numbers = suppressWarnings(as.numeric(elements))
              numbers = numbers[!is.na(numbers)]
            }
            if (any(numbers == Inf) | length(numbers) == 1) {
              thresholds = paste("above", numbers[1])
            } else if (length(numbers) == 2) {
              thresholds = paste("between", numbers[1], "and", numbers[2])
            } else {
              thresholds = ""
            }
            what = paste(what, thresholds, "luxes")
            unit = "(minutes)"
          } else if ("mean" %in% elements) {
            what = "Mean LUX value"
            if ("mvpa" %in% elements) what = paste(what, "in moderate-to-vigorous physical activity time")
            unit = "(luxes)"
          } else if ("max" %in% elements) {
            what = "Maximum LUX value"
            unit = "(luxes)"
          } else if ("timeawake" %in% elements) {
            what = "Time classified as awake"
            unit = "(minutes)"
          } else if ("imputed" %in% elements) {
            what = "Time in which the LUX has been imputed"
            unit = "(minutes)"
          } else if ("ignored" %in% elements) {
            what = "Time in which the LUX has been ignored"
            unit = "(minutes)"
          }
          # lux segments
          if (any(grepl("hr", elements))) {
            t1 = elements[grep("hr", elements)]; t1 = gsub("hr", ":00", t1)
            t0 = elements[grep("hr", elements) - 1]
            if (nchar(t0) == 1) t0 = paste0(t0, ":00")
            what = paste(what, "in the segment from", t0, "to", t1)
          }
        }
        # fragmentation metrics
        if ("FRAG" %in% elements) {
          what_bu = tolower(what)
          what = "Fragmentation analysis:"
          if ("TP" %in% elements) {
            what = paste(what, "transition probability (%)")
          } else if ("Nfrag" %in% elements) {
            what = paste(what, "number of fragments")
          } else if ("NFragPM" %in% elements) {
            what = paste(what, "number of fragments per minute")
          } else if ("mean" %in% elements) {
            what = paste(what, "mean duration in the")
          } else if ("Gini" %in% elements) {
            what = paste(what, "Gini inequality index as calculated in the ineq R package")
          } else if ("CoV" %in% elements) {
            what = paste(what, "Coefficient of variance as proposed in https://shorturl.at/nsDU9")
          } else if ("alpha" %in% elements) {
            what = paste(what, "Alpha power law exponent metric as proposed in https://shorturl.at/gwzB8")
          } else if ("x0" %in% elements) {
            what = paste(what, "x0.5 power law exponent metric as proposed in https://shorturl.at/gwzB8")
          } else if ("W0" %in% elements) {
            what = paste(what, "W0.5 power law exponent metric as proposed in https://shorturl.at/gwzB8")
          } else if ("SD" %in% elements) {
            what = paste(what, "standard deviation in the")
          }
          what = paste(what, what_bu)
        }
        # WINDOW -----------------------------------------------------------
        if ("day" %in% elements & "spt" %in% elements) {
          window = "during the waking time and sleep period time (i.e., full window)"
        } else if ("day" %in% elements) {
          window = "during the waking time"
        } else if ("spt" %in% elements) {
          window = "during the sleep period time"
        }
        # CLASS ------------------------------------------------------------
        classes = c("IN", "LIG", "LIPA", "MOD", "VIG", "MVPA", "PA", 
                    "sleep", "nonwear", "mostactive60min", "mostactive30min",
                    "IN2PA", "PA2IN", "IN2LIPA", "IN2MVPA", "sleep2wake", "wake2sleep")
        x = elements[which(elements %in% classes)]
        class = getclass(x)
        # sleep efficiency (overwrite prev classification as sleep)
        if ("sleep" %in% elements & "efficiency" %in% elements) {
          class = "Sleep efficiency after onset"
          unit = "(%)"
        }
        # wakefulness after sleep onset (overwrite prev classification of intensity)
        if ("wake" %in% elements) class = paste("awake", class)
        # bouts
        if ("bts" %in% elements & !("Nbouts" %in% elements)) {
          numbers = suppressWarnings(as.numeric(elements))
          boutdur = numbers[!is.na(numbers)]
          class = paste("bouts of", paste(boutdur, collapse = "-"), "min", class)
        }
        # unbouted time
        if ("unbt" %in% elements & !("Nbouts" %in% elements)) {
          intensity = elements[which(elements %in% c("IN", "LIG", "MOD", "VIG"))]
          if (intensity %in% c("MOD", "VIG")) intensity = "MVPA"
          look4 = paste("dur_day", intensity, "bts", sep = "_")
          boutVars = grep(look4, cnames, value = TRUE)
          boutdurs = c()
          for (i in 1:length(boutVars)) {
            x = unlist(strsplit(boutVars[i], split = "_"))
            numbers = suppressWarnings(as.numeric(x))
            boutdurs = c(boutdurs, numbers[!is.na(numbers)])
          }
          minboutdur = min(boutdurs)
          class = paste("unbouted", class, paste0("(",0,"-",minboutdur, " min)"))
        }
        # intensity gradient
        if ("ig" %in% elements) {
          if ("gradient" %in% elements) class = "Intensity gradient"
          if ("rsquared" %in% elements) class = paste("R-squared from the log-log time to intensity regression to calculate the intensity gradient")
          if ("intercept" %in% elements) class = paste("Intercept from the log-log time to intensity regression to calculate the intensity gradient")
        }
        # connector (in)
        if (!is.null(class)) {
          if (class != "Sleep efficiency after onset" 
              & substr(class, 1, 8) != "the most"
              & !("ig" %in% elements)) class = paste("in", class)
        }
        # UNITS -------------------------------------------------------------------
        if ("min" %in% elements) {
          unit = "(minutes)"
        } else if ("mg" %in% elements) {
          unit = "(mili-gravity units)"
        } else if ("perc" %in% elements) {
          unit = "(%)"
        }
        # ------------------------------------------------------------------
        # Rest of column names...
        if (is.null(what)) {
          if ("sleeponset" %in% elements | "wakeup" %in% elements) {
            if ("sleeponset" %in% elements) what = "Sleep onset time"
            if ("wakeup" %in% elements) what = "Wake up time"
            if ("ts" %in% elements) {
              unit = "(hh:mm:ss)"
            } else {
              unit = "(hours from previous midnight)"
            }
          } else if ((substr(elements[1], 1, 1) == "L" | substr(elements[1], 1, 1) == "M")
                     & elements[1] != "LUX") {
            # LX and MX metrics
            # what
            if (grepl("TIME", elements[1])) {
              what = "Starting time"
              unit = "(timestamp)"
              if ("num" %in% elements) {
                unit = "(timestamp)"
              }
            } else if (grepl("VALUE", elements[1])) {
              what = "Mean acceleration"
              unit = "(mili-gravity units)"
            } else if ("peakLUX" %in% elements) {
              if ("mean" %in% elements) {
                what = "Mean peak Lux"
              } else if ("max" %in% elements) {
                what = "Max peak Lux"
              }
            }
            # class
            X = as.numeric(gsub("\\D", "", elements))[1]
            if (substr(elements[1], 1, 1) == "L") {
              class = paste("during the", X, "consecutive hours with the lowest acceleration")
            } else if (substr(elements[1], 1, 1) == "M") {
              class = paste("during the", X, "consecutive hours with the highest acceleration")
            }
          } else if ("daytype" %in% elements) {
            what = "WD = weekday; WE = weekend day"
          } else {
            what = ""
          } 
        } 
      }
      # build up definition
      def = paste(what, class, window, unit)
      # only for personsummary - aggregation method
      if (grepl("personsummary", reports[ri])) {
        agg = NULL
        if (!grepl("Nvalid", cnames[coli])) {
          if ("pla" %in% elements) agg = "- plain average"
          if ("wei" %in% elements) agg = "- weighted average"
          if ("WD" %in% elements) agg = "- weekdays average"
          if ("WE" %in% elements) agg = "- weekend days average"
        } else {
          # adapt definition of valid days to part 5 reports
          def = gsub("part 2, 4, and 5", "part 5", def)
        }
        def = paste(def, agg)
      }
      def = gsub("\\s+", " ", def) # this removes extra spaces
      def = gsub("^ ", "", def) # this removes leading spaces
      
      # if window and class have not been identified, then return blank definition
      if (is.null(what) & is.null(class)) def = NA
      
      # store definition
      dictionary[coli, "Definition"] = def
    }
    # write csv reports with data dictionaries
    directory = file.path(metadatadir, "results", "variableDictionary/")
    if (!dir.exists(directory)) dir.create(directory)
    fn = gsub("part5_", "part5_dictionary_", basename(reports[ri]))
    fn = unlist(strsplit(fn, "_MM|_WW|_OO"))[1]
    fn = paste0(fn, ".csv")
    data.table::fwrite(dictionary, file = file.path(directory, fn), 
                       row.names = FALSE, na = "", sep = params_output[["sep_reports"]],
                       dec = params_output[["dec_reports"]])
  }
}

g.part6 = function(datadir = c(), metadatadir = c(), f0 = c(), f1 = c(),
                   params_general = c(), params_phyact = c(), params_247 = c(),
                   params_cleaning = c(),
                   verbose = TRUE, ...) {
  
  # This function called by function GGIR
  # and aims to facilitate time-pattern analysis building on the labelled time
  # series derived in GGIR part 5
  #----------------------------------------------------------
  # Extract and check parameters
  input = list(...)
  params = extract_params(params_general = params_general,
                          params_phyact = params_phyact,
                          params_247 = params_247,
                          input = input, params2check = c("general", "phyact", "247"))
  params_general = params$params_general
  params_phyact = params$params_phyact
  params_247 = params$params_247
  
  #======================================================================
  # create new folder (if not existent) for storing milestone data
  ms6.out = "/meta/ms6.out"
  if (!file.exists(paste(metadatadir, ms6.out, sep = ""))) {
    dir.create(file.path(metadatadir, ms6.out))
  }
  
  #======================================================================
  # compile lists of milestone data filenames
  
  # Identify correct subfolder
  expected_ts_path = paste0(metadatadir, "/meta/ms5.outraw/", params_phyact[["part6_threshold_combi"]])
  expected_ms5raw_path = paste0(metadatadir, "/meta/ms5.outraw")
  if (!dir.exists(expected_ts_path)) {
    if (!dir.exists(expected_ms5raw_path)) {
      stop("\nPath ", expected_ms5raw_path, " does not exist", call. = FALSE)
    } else {
      subDirs = list.dirs(expected_ms5raw_path)
      if (length(subDirs) > 0) {
        expected_ts_path = paste0(expected_ms5raw_path, "/", subDirs[1])
        warning(paste0("\nThreshold combi ", params_phyact[["part6_threshold_combi"]],
                       " in time series ouput. Instead now using", subDirs[1]), call. = FALSE)
      } else {
        stop(paste0("\nNo subfolders found inside ", expected_ms5raw_path), call. = FALSE)
      }
    }
  }
  fnames.ms5raw = dir(expected_ts_path)
  # It can be that the use stored rdata and csv files
  # only use rdata in that case
  getExt = function(x) {
    tmp = unlist(strsplit(basename(x), "\\."))
    return(tmp[length(tmp)])
  }
  EXT = unlist(lapply(fnames.ms5raw, FUN = getExt))
  uniqueEXT = unique(EXT)
  if (length(uniqueEXT) == 2) {
    fnames.ms5raw = fnames.ms5raw[which(EXT == "RData")]
    EXT = EXT[which(EXT == "RData")]
  }
  EXT = EXT[1]
  # Identify existing part 6 milestone data
  fnames.ms6 = dir(paste(metadatadir, "/meta/ms6.out", sep = ""))
  ffdone = fnames.ms6
  #------------------------------------------------
  # specify parameters
  N5 = length(fnames.ms5raw)
  if (length(f0) == 0) f0 = 1
  if (length(f1) == 0) f1 = N5
  
  if (f0 > N5) {
    stop("Argument f0 is larger than the number of files in meta/ms5.outraw.", call. = FALSE)
  }
  if (f1 > N5 | f1 == 0) f1 = N5
  if (f0 == 0) f0 = 1
  
  
  resultsdir = paste0(metadatadir , "/results")
  if (!dir.exists(resultsdir)) dir.create(resultsdir)
  
  #=========================================================
  # Recording-group (e.g. household members) level co-analysis,
  # which at the end of this g.part6 is either
  if (params_247[["part6HCA"]] == TRUE) {
    part6AlignIndividuals(GGIR_ts_dir = expected_ts_path,
                          outputdir = resultsdir,
                          path_ggirms = paste0(metadatadir , "/meta"),
                          desiredtz = params_general[["desiredtz"]],
                          verbose = verbose)
    
    part6PairwiseAggregation(outputdir = paste0(metadatadir, "/results"),
                             desiredtz = params_general[["desiredtz"]],
                             verbose = verbose)
  }
  
  #=========================================================
  # Declare recording level functionality, which at the end of this g.part6 is either
  # applied to the file in parallel with foreach or serially with a loop
  main_part6_recordlevel = function(i, metadatadir = c(), f0 = c(), f1 = c(),
                                    fnames.ms5raw, ffdone, EXT, verbose,
                                    params_general = c(), params_phyact = c(), params_247 = c(),
                                    params_cleaning = c()) {
    if (length(ffdone) > 0) {
      if (length(which(ffdone == fnames.ms5raw[i])) > 0) {
        skip = 1 #skip this file because it was analysed before")
      } else {
        skip = 0 #do not skip this file
      }
    } else {
      skip = 0
    }
    if (params_general[["overwrite"]] == TRUE) skip = 0
    Lnames = cosinor_coef = mdat = desiredtz_part1 = NULL
    if (skip == 0) {
      # Load time series:
      if (EXT == "RData") {
        load(file = paste0(metadatadir, "/meta/ms5.outraw/",
                           params_phyact[["part6_threshold_combi"]], "/", fnames.ms5raw[i]))
        if (!is.null(desiredtz_part1)) {
          params_general[["desiredtz"]] = desiredtz_part1
        }
        if (is.null(Lnames)) stop("Part 5 was processed with an older version of GGIR, reprocess part 5")
        mdat$time = mdat$timestamp # duplicate column because cosinor function expect columntime
      } else {
        # use behavioural codes files to derive Lnames object (names of behavioural classes)
        legendfile = dir(path = paste0(metadatadir, "/meta/ms5.outraw"), pattern = "behavioralcodes", full.names = TRUE)
        if (length(legendfile) > 1) {
          df <- file.info(legendfile)
          legendfile = rownames(df)[which.max(df$mtime)]
        } else if (length(legendfile) == 0) {
          stop(paste0("behaviouralcodes file could not be found in ",
                      paste0(metadatadir, "/meta/ms5.outraw")))
        }
        Lnames = data.table::fread(file = legendfile)$class_name
        mdat = data.table::fread(file = paste0(metadatadir, "/meta/ms5.outraw/", 
                                               params_phyact[["part6_threshold_combi"]],  "/", fnames.ms5raw[i]), data.table = FALSE)
        filename = fnames.ms5raw[i]
      }
      nfeatures = 200
      summary = matrix(NA, nfeatures, 1)
      s_names = rep("", nfeatures)
      fi = 1
      deltatime = diff(mdat$timenum)
      epochSize = min(deltatime[which(deltatime != 0)])
      if (any(deltatime > epochSize)) {
        imputeTimeGaps = function(mdat, epochSize) {
          # impute timegaps
          mdat$gap = 1
          mdat$gap[1:(nrow(mdat) - 1)] = diff(mdat$timenum) / epochSize
          gapp = which(mdat$gap != 1)
          if (length(gapp) > 0) {
            if (gapp[1] > 1) {
              newTime = mdat$timenum[1:(gapp[1] - 1)]
            } else {
              newTime = NULL
            }
            for (g in 1:length(gapp)) {
              newTime = c(newTime, mdat$timenum[gapp[g]] + seq(0, by = epochSize, length.out = mdat$gap[gapp[g]]))
              if (g < length(gapp)) {
                newTime = c(newTime, mdat$timenum[(gapp[g] + 1):(gapp[g + 1] - 1)])
              }
            }
            newTime =  c(newTime, mdat$timenum[(gapp[g] + 1):length(mdat$timenum)])
          }
          mdat <- as.data.frame(lapply(mdat, rep, mdat$gap))
          invalid = duplicated(mdat)
          if (length(gapp) > 0) {
            mdat$timenum = newTime[1:nrow(mdat)]
          }
          mdat = mdat[, which(colnames(mdat) != "gap")]
          mdat$ACC[invalid] = NA
          mdat$class_id[invalid] = NA
          mdat$guider[invalid] = NA
          mdat$invalidepoch[invalid] = 1
          mdat$window[invalid] = 9999
          mdat$invalid_sleepperiod[invalid] = 100
          mdat$invalid_wakinghours[invalid] = 100
          mdat$time = mdat$timestamp = as.POSIXct(mdat$timenum,
                                                  tz =  params_general[["desiredtz"]],
                                                  origin = "1970-01-01")
          return(mdat)
        }
        mdat = imputeTimeGaps(mdat, epochSize)
      }
      # Select relevant section of the time series
      wakeuptimes = which(diff(c(1, mdat$SleepPeriodTime, 0)) == -1)
      onsettimes = which(diff(c(0, mdat$SleepPeriodTime, 1)) == 1) 
      Nmdat = nrow(mdat)
      if (wakeuptimes[length(wakeuptimes)] > Nmdat) wakeuptimes[length(wakeuptimes)] = Nmdat
      if (onsettimes[length(onsettimes)] > Nmdat) onsettimes[length(onsettimes)] = Nmdat
      getIndex = function(x, ts, wakeuptimes, onsettimes, epochSize) {
        if (x == "start") {
          y = 1 # simply the start of the time series
        } else if (x == "end") {
          y = nrow(ts) # simply the end of the time series
        } else if (substr(x, start = 1, stop = 1) == "W") {
          ni = as.numeric(substr(x, start = 2, stop = 5))
          if (ni > 0) { # nth Wakeup after the start
            y = wakeuptimes[ni]
          } else if (ni < 0) { # nth Wakeup before the end
            y = wakeuptimes[length(wakeuptimes) + ni + 1] # + because it already is -
          }
        } else if (substr(x, start = 1, stop = 1) == "O") {
          ni = as.numeric(substr(x, start = 2, stop = 5))
          if (ni > 0) { # nth Onset after the start
            y = onsettimes[ni]
          } else if (ni < 0) { # nth Onset before the end
            y = onsettimes[length(onsettimes) + ni + 1] # + because it already is -
          }
        } else if (substr(x, start = 1, stop = 1) == "H") {
          ni = as.numeric(substr(x, start = 2, stop = 5))
          if (ni > 0) { # nth Hour after the start
            y = round((ni * 3600) / epochSize)
          } else if (ni < 0) { # nth Hour before the end
            y = nrow(ts) - round((ni * 3600) / epochSize)
          }
        }
        return(y)
      }
      t0 = getIndex(params_247[["part6Window"]][1], ts = mdat, wakeuptimes, onsettimes, epochSize)
      t1 = getIndex(params_247[["part6Window"]][2], ts = mdat, wakeuptimes, onsettimes, epochSize)
      if (length(t1) == 1 && length(t0) == 1 && !is.na(t1) && !is.na(t0) && t1 > t0) {
        do.cr = TRUE
      } else {
        do.cr = FALSE
      }
      
      if (do.cr == TRUE) {
        ts = mdat[t0:t1, ]
        rm(mdat)
        ts = ts[which(ts$window != 0), ]
      } else {
        ts = mdat[1,]
      }
      
      # Set windows to missing that do not have enough valid data in either spt or day.
      # Note that columns invalid_wakinghours and invalid_sleepperiod here
      # are constants per window, we use this to identify the entire window as valid/invalid
      invalidWindows = which(ts$invalid_wakinghours > (1 - params_cleaning[["includecrit.part6"]][1]) * 100 | 
                               ts$invalid_sleepperiod > (1 - params_cleaning[["includecrit.part6"]][2]) * 100)
      if (length(invalidWindows) > 0) {
        ts[invalidWindows,c("class_id", "ACC")] = NA
        ts$invalidepoch[invalidWindows] = 1
      }
      # Include basic information in the summary
      summary[fi] = extractID(hvars = list(ID = "NA", iID = "NA", IDd = "NA"),
                              idloc = params_general[["idloc"]],
                              fname = fnames.ms5raw[i])
      s_names[fi] = "ID"
      fi = fi + 1
      starttime = as.POSIXlt(ts$time[1], tz = params_general[["desiredtz"]],
                             origin = "1970-01-01")
      summary[fi] = format(starttime)
      s_names[fi] = "starttime"
      fi = fi + 1
      summary[fi] = gsub(pattern = "[.]RData$|[.]csv$", replacement = "", x = filename)
      s_names[fi] = "filename"
      fi = fi + 1
      summary[fi] = ifelse(test = nrow(ts) == 1,
                           yes = 0,
                           no = nrow(ts) / ((3600 * 24) / epochSize))
      s_names[fi] = "N_days"
      fi = fi + 1
      N_valid_days = ifelse(test = nrow(ts) == 1,
                            yes = 0,
                            no = length(which(ts$invalidepoch == 0)) / ((3600 * 24) / epochSize))
      summary[fi] = N_valid_days
      s_names[fi] = "N_valid_days"
      fi = fi + 1
      CountWeekDays = table(weekdays(ts$time[which(ts$invalidepoch == 0)]))
      CountWeekDays = CountWeekDays / (1440 * (60 / epochSize))
      CountWeekDays = CountWeekDays[which(CountWeekDays > 0.66)]
      if (length(CountWeekDays) > 0) {
        Nweekendays = length(which(names(CountWeekDays) %in% c("Saturday", "Sunday")))
      } else {
        Nweekendays = 0
      }
      summary[fi] = Nweekendays
      s_names[fi] = "N_valid_weekend_days"
      fi = fi + 1
      #=================================================================
      # Circadian rhythm analysis
      if (do.cr == TRUE & N_valid_days > 2) {
        #===============================================
        # Cosinor analysis which comes with IV IS estimtes
        # Note: applyCosinorAnalyses below uses column invalidepoch to turn
        # imputed values to NA.
        colnames(ts)[which(colnames(ts) == "timenum")] = "time"
        acc4cos = ts[, c("time", "ACC")]
        qcheck = ts$invalidepoch

        # Add missing value to complete an interger number of full days
        Na4c = nrow(acc4cos)
        NepochsPerDay = 24 * (3600 / epochSize)
        NepochsNeeded = ((ceiling(Na4c / NepochsPerDay) + 4) * NepochsPerDay - Na4c) + 1
        if (NepochsNeeded > 0)  {
          acc4cos[(Na4c + 1):(Na4c + NepochsNeeded), ] = NA
          acc4cos$time[Na4c:(Na4c + NepochsNeeded)] = seq(from = acc4cos$time[Na4c], by = epochSize, length.out = NepochsNeeded + 1)
          qcheck = c(qcheck, rep(1, NepochsNeeded))
        }

        threshold = as.numeric(unlist(strsplit( params_phyact[["part6_threshold_combi"]], "_"))[1])
        
        # extract nightsi again
        tempp = unclass(as.POSIXlt(acc4cos$time, tz = params_general[["desiredtz"]], origin = "1970-01-01"))
        sec = tempp$sec
        min = tempp$min
        hour = tempp$hour
        if (params_general[["dayborder"]] == 0) {
          nightsi = which(sec == 0 & min == 0 & hour == 0)
        } else {
          nightsi = which(sec == 0 & min == (params_general[["dayborder"]] - floor(params_general[["dayborder"]])) * 60 & hour == floor(params_general[["dayborder"]])) #shift the definition of midnight if required
        }
        cosinor_coef = apply_cosinor_IS_IV_Analyses(ts = acc4cos,
                                                    qcheck = qcheck,
                                                    midnightsi = nightsi,
                                                    epochsizes = rep(epochSize, 2),
                                                    threshold = threshold)
        rm(acc4cos)
        try(expr = {summary[fi:(fi + 6)] = c(cosinor_coef$timeOffsetHours,
                                             cosinor_coef$coef$params$mes,
                                             cosinor_coef$coef$params$amp,
                                             cosinor_coef$coef$params$acr,
                                             cosinor_coef$coef$params$acrotime,
                                             cosinor_coef$coef$params$ndays,
                                             cosinor_coef$coef$params$R2)},
            silent = TRUE)
        
        s_names[fi:(fi + 6)] = c("cosinor_timeOffsetHours", "cosinor_mes", "cosinor_amp", "cosinor_acrophase",
                                 "cosinor_acrotime", "cosinor_ndays", "cosinor_R2")
        fi = fi + 7
        try(expr = {summary[fi:(fi + 10)] = c(cosinor_coef$coefext$params$minimum,
                                              cosinor_coef$coefext$params$amp,
                                              cosinor_coef$coefext$params$alpha,
                                              cosinor_coef$coefext$params$beta,
                                              cosinor_coef$coefext$params$acrotime,
                                              cosinor_coef$coefext$params$UpMesor,
                                              cosinor_coef$coefext$params$DownMesor,
                                              cosinor_coef$coefext$params$MESOR,
                                              cosinor_coef$coefext$params$ndays,
                                              cosinor_coef$coefext$params$F_pseudo,
                                              cosinor_coef$coefext$params$R2)},
            silent = TRUE)
        s_names[fi:(fi + 10)] = c("cosinorExt_minimum", "cosinorExt_amp", "cosinorExt_alpha",
                                  "cosinorExt_beta", "cosinorExt_acrotime", "cosinorExt_UpMesor",
                                  "cosinorExt_DownMesor", "cosinorExt_MESOR",
                                  "cosinorExt_ndays", "cosinorExt_F_pseudo", "cosinorExt_R2")
        fi = fi + 11
        try(expr = {summary[fi:(fi + 2)] = c(cosinor_coef$IVIS$InterdailyStability,
                                             cosinor_coef$IVIS$IntradailyVariability,
                                             cosinor_coef$IVIS$phi)},
            silent = TRUE)
        s_names[fi:(fi + 2)] = c("IS", "IV", "phi")
        fi = fi + 3
        #===============================================
        # Transition probabilities
        for (fragmode in c("day", "spt")) {
          ts_temp = ts
          # turn other half of data to NA
          ts_temp$class_id[which(ts$SleepPeriodTime == ifelse(fragmode == "spt", yes = 0, no = 1))] = NA
          frag.out = g.fragmentation(frag.metrics = params_phyact[["frag.metrics"]],
                                     LEVELS = ts_temp$class_id,
                                     Lnames = Lnames, xmin = 60/epochSize, mode = fragmode)
          # fragmentation values can come with a lot of decimal places
          summary[fi:(fi + (length(frag.out) - 1))] = round(as.numeric(frag.out), digits = 6)
          s_names[fi:(fi + (length(frag.out) - 1))] = paste0("FRAG_", names(frag.out), "_", fragmode)
          fi = fi + length(frag.out)
        }
        
        #===============================================
        # LXMX: code copied from g.part 5
        # To be refactored as a central generic function once
        # new MXLX function is merged
        window_number = unique(ts$window)
        dsummary = matrix(NA, 100, length(params_247[["winhr"]]) * 80)
        ds_names = rep("", ncol(dsummary))
        lightpeak_available = "lightpeak" %in% names(ts)
        si = 1
        for (wi in window_number) {
          gi = 1
          sse = which(ts$window == wi)
          if (is.na(ts$ACC[sse[2]])) {
            next
          }
          ts$time_num = ts$time
          ts$time = ts$timestamp
          WLH = length(sse)/((60/epochSize) * 60)
          if (WLH <= 1) WLH = 1.001
          for (wini in params_247[["winhr"]]) {
            reso = params_247[["M5L5res"]] #resolution at 5 minutes
            endd = floor((WLH * 10) / 10) # rounding needed for non-integer window lengths
            nwindow_f = (endd - wini) #number of windows for L5M5 analyses
            ignore = FALSE
            if (endd <= wini | nwindow_f < 1) ignore = TRUE # day is shorter then time window, so ignore this
            nwindow_f = nwindow_f * (60/reso)
            if (ignore == FALSE) {
              # Calculate running window variables
              ACCrunwin = matrix(0, nwindow_f, 1)
              TIMErunwin = matrix("", nwindow_f, 1)
              for (hri in 0:floor((((endd - wini) * (60/reso)) - 1))) {
                e1 = (hri * reso * (60/epochSize)) + 1
                e2 = (hri + (wini * (60/reso))) * reso * (60/epochSize)
                if (e2 > length(sse)) e2 = length(sse)
                ACCrunwin[(hri + 1), 1] = mean(ts$ACC[sse[e1:e2]])
                TIMErunwin[(hri + 1), 1] = format(ts$time[sse[e1]])
              }
              ACCrunwin = ACCrunwin[is.na(ACCrunwin) == F]
              TIMErunwin = TIMErunwin[is.na(ACCrunwin) == F]
              if (length(ACCrunwin) > 0 & length(TIMErunwin) > 0) {
                # Derive day level variables
                L5VALUE = min(ACCrunwin)
                L5HOUR = TIMErunwin[which(ACCrunwin == L5VALUE)[1]]
                M5VALUE = max(ACCrunwin)
                M5HOUR = TIMErunwin[which(ACCrunwin == M5VALUE)[1]]
                if (lightpeak_available == TRUE) {
                  if (length(unlist(strsplit(M5HOUR, " |T"))) == 1) M5HOUR = paste0(M5HOUR, " 00:00:00")
                  startM5 = which(format(ts$time) == M5HOUR)
                  M5_mean_peakLUX = round(mean(ts$lightpeak[startM5[1]:(startM5[1] + (wini*60*(60/epochSize)))], na.rm = TRUE), digits = 1)
                  M5_max_peakLUX = round(max(ts$lightpeak[startM5[1]:(startM5[1] + (wini*60*(60/epochSize)))], na.rm = TRUE), digits = 1)
                }
              } else {
                L5HOUR = M5HOUR = "not detected"
                L5VALUE = M5VALUE = ""
                if (lightpeak_available == TRUE) {
                  M5_mean_peakLUX = M5_max_peakLUX = ""
                }
              }
              # Add variables calculated above to the output matrix
              dsummary[si, gi:(gi + 1)] = c(L5VALUE, M5VALUE)
            }
            ds_names[gi:(gi + 1)] = c(paste0("L", wini, "VALUE"),
                                      paste0("M", wini, "VALUE"))
            gi = gi + 2
            if ("lightpeak" %in% colnames(ts)) {
              if (ignore == FALSE) {
                dsummary[si,gi] = M5_mean_peakLUX
                dsummary[si,gi + 1] = M5_max_peakLUX
              }
              ds_names[gi] = paste("M", wini, "_mean_peakLUX", sep = "")
              ds_names[gi + 1] = paste("M", wini,"_max_peakLUX", sep = "")
              gi = gi + 2
            }
            if (ignore == FALSE) {
              # Add also numeric time
              if (is.ISO8601(L5HOUR)) { # only do this for ISO8601 format
                L5HOUR = format(iso8601chartime2POSIX(L5HOUR, tz = params_general[["desiredtz"]]))
                M5HOUR = format(iso8601chartime2POSIX(M5HOUR, tz = params_general[["desiredtz"]]))
              }
              if (length(unlist(strsplit(L5HOUR," "))) == 1) L5HOUR = paste0(L5HOUR," 00:00:00") #added because on some OS timestamps are deleted for midnight
              if (length(unlist(strsplit(M5HOUR," "))) == 1) M5HOUR = paste0(M5HOUR," 00:00:00")
              if (L5HOUR != "not detected") {
                time_num = sum(as.numeric(unlist(strsplit(unlist(strsplit(L5HOUR," "))[2], ":"))) * c(3600, 60, 1)) / 3600
                if (time_num <= 12) {
                  time_num = time_num + 24
                }
                dsummary[si,gi] = time_num
              } else {
                dsummary[si,gi] = NA
              }
            }
            ds_names[gi] = paste("L", wini, "TIME_num", sep = "");      gi = gi + 1
            if (ignore == FALSE) {
              if (M5HOUR != "not detected") {
                time_num = sum(as.numeric(unlist(strsplit(unlist(strsplit(M5HOUR," "))[2], ":"))) * c(3600, 60, 1)) / 3600
                dsummary[si, gi] = time_num
              } else {
                dsummary[si, gi] = NA
              }
            }
            ds_names[gi] = paste("M", wini, "TIME_num", sep = "");      gi = gi + 1
          }
          si = si + 1
          if (si > nrow(dsummary)) {
            dsummary[nrow(dsummary) + 1, ] = NA
            ds_names = c(ds_names, "")
          }
        }
        gi = which(ds_names == "")[1]
        dsummary = dsummary[1:(si - 1), 1:(gi - 1), drop = FALSE]
        ds_names = ds_names[1:(gi - 1)]
        dsummary = as.data.frame(x = dsummary)
        colnames(dsummary) = ds_names
        # aggregate across recording
        MXLXsummary = colMeans(x = dsummary, na.rm = TRUE)
        # Add to summary
        summary[fi:(fi + length(MXLXsummary) - 1)] = as.numeric(MXLXsummary)
        s_names[fi:(fi + length(MXLXsummary) - 1)] = names(MXLXsummary)
        fi = fi + length(MXLXsummary)
        # Translate times to clock time and add to summary
        cols2convert = grep(pattern = "TIME_num", x = names(MXLXsummary), value = FALSE)
        if (length(cols2convert) > 0) {
          for (cvi in cols2convert) {
            s_names[fi] = gsub(pattern = "_num", replacement = "_clock", x = names(MXLXsummary)[cvi])
            hr = as.numeric(floor(MXLXsummary[cvi]))
            min = as.numeric(floor((MXLXsummary[cvi] - hr) * 60))
            sec = as.numeric(floor((MXLXsummary[cvi] - hr - (min / 60)) * 3600))
            if (!is.na(hr) && !is.na(min) && !is.na(sec)) {
              if (hr >= 24) hr = hr - 24
              clock_time = paste0(hr, ":", min, ":", sec)
            } else {
              clock_time = NA
            }
            summary[fi] = clock_time 
            fi = fi + 1
          }
        }
        #=======================================================================
        # DFA analyses is used independent of do.cr because it is time consuming
        if (params_247[["part6DFA"]] == TRUE) {
          ssp = SSP(ts$ACC[!is.na(ts$ACC)])
          abi = ABI(ssp)
          summary[fi:(fi + 1)] = c(ssp, abi)
          s_names[fi:(fi + 1)] = c("SSP", "ABI")
          fi = fi + 2
        }
        #------------------------------------------------------------
        # Sleep Regularity Index
        if (!is.null(params_247[["SRI2_WASOmin"]])) {
          sleepnap_classid = grep(pattern = "spt_sleep|_nap", x = Lnames) - 1
          ts$sleepnap = 0
          ts$sleepnap[which(ts$class_id %in% sleepnap_classid)] = 1
          # remove short lasting 'WASO' based on WASOmin parameters
          waso_classid = grep(pattern = "spt_wake", x = Lnames) - 1
          waso = rep(0, nrow(ts))
          waso[which(ts$class_id %in% waso_classid)] = 1
          rlew = rle(waso)
          shortwaso = which(rlew$values == 1 & (rlew$lengths / (60/epochSize)) <= params_247[["SRI2_WASOmin"]])
          if (length(shortwaso) > 0) {
            rlew$values[shortwaso] = 2
            waso = rep(rlew$values, rlew$lengths)
            if (2 %in% rlew$values) {
              ts$sleepnap[which(waso == 2)] = 1
            }
          }
          SRI = CalcSleepRegularityIndex(data = ts,
                                         epochsize = epochSize,
                                         desiredtz = params_general[["desiredtz"]])
          
          SRI = SRI[which(SRI$frac_valid > params_cleaning[["includecrit.part6"]][1]), ]
          if (nrow(SRI) > 0) {
            summary[fi] = weighted.mean(x = SRI$SleepRegularityIndex, w = SRI$frac_valid)
            summary[fi + 1] = nrow(SRI)
          } else {
            summary[fi:(fi + 1)] = NA
          }
          s_names[fi:(fi + 1)] = c("SleepRegularityIndex2", "SleepRegularityIndex2_Ndaypairs")
          fi = fi + 2
          rm(rlew, shortwaso, waso)
        }
      }
      #=============================================
      # Store results in milestone data
      summary = summary[1:(fi - 1),]
      s_names = s_names[1:(fi - 1)]
      summary = summary[which(s_names != "")]
      s_names = s_names[which(s_names != "")]
      output_part6 = data.frame(t(summary), stringsAsFactors = FALSE)
      names(output_part6) = s_names
      nonnumeric = grep(pattern = "TIME_clock", x = s_names, invert = TRUE)
      nonnumeric = nonnumeric[nonnumeric %in% 1:3 == FALSE]
      output_part6[, nonnumeric] = as.numeric(output_part6[, nonnumeric])
      if (length(output_part6) > 0) {
        if (length(cosinor_coef) > 0) {
          cosinor_ts = cosinor_coef$coefext$cosinor_ts
        } else {
          cosinor_ts = c()
        }
        GGIRversion = "GGIR not used"
        if (is.element('GGIR', installed.packages()[,1])) {
          GGIRversion = as.character(utils::packageVersion("GGIR"))
          if (length(GGIRversion) != 1) GGIRversion = sessionInfo()$otherPkgs$GGIR$Version
        }
        output_part6$GGIRversion = GGIRversion
        save(output_part6, cosinor_ts, GGIRversion, file = paste0(metadatadir,
                                                                  ms6.out, "/", gsub(pattern = "[.]csv|[.]RData",
                                                                                     replacement = "",
                                                                                     x = fnames.ms5raw[i]), ".RData"))
      }
      rm(output_part6, summary)
    }
    # Function has no output_part6 because ideally all relevant output_part6
    # is stored in milestone data by now
  }
  
  if (params_247[["part6CR"]] == TRUE) {
    #======================================================================
    # loop through milestone data-files or filenames stored in output of g.part5
    # setup parallel backend to use many processors
    if (params_general[["do.parallel"]] == TRUE) {
      cores = parallel::detectCores()
      Ncores = cores[1]
      if (Ncores > 3) {
        if (length(params_general[["maxNcores"]]) == 0) params_general[["maxNcores"]] = Ncores
        Ncores2use = min(c(Ncores - 1, params_general[["maxNcores"]], (f1 - f0) + 1))
        if (Ncores2use > 1) {
          cl <- parallel::makeCluster(Ncores2use) # not to overload your computer
          parallel::clusterExport(cl = cl, 
                                  varlist = c(unclass(lsf.str(envir = asNamespace("GGIR"), all = T)),
                                              "MONITOR", "FORMAT"),
                                  envir = as.environment(asNamespace("GGIR"))
          )
          parallel::clusterEvalQ(cl, Sys.setlocale("LC_TIME", "C"))
          doParallel::registerDoParallel(cl)
        } else {
          # Don't process in parallel if only one core
          params_general[["do.parallel"]] = FALSE
        }
      } else {
        if (verbose == TRUE) cat(paste0("\nparallel processing not possible because number of available cores (",Ncores,") < 4"))
        params_general[["do.parallel"]] = FALSE
      }
    }
    if (params_general[["do.parallel"]] == TRUE) {
      if (verbose == TRUE) cat(paste0('\n Busy processing ... see ', metadatadir, ms6.out, ' for progress\n'))
      # check whether we are in development mode:
      GGIRinstalled = is.element('GGIR', installed.packages()[,1])
      packages2passon = functions2passon = NULL
      GGIRloaded = "GGIR" %in% .packages()
      if (GGIRloaded) { #pass on package
        packages2passon = 'GGIR'
        errhand = 'pass'
      } else {
        # pass on functions
        functions2passon = c("cosinor_IS_IV_Analyses", "apply_cosinor_IS_IV_Analyses", "g.IVIS")
        errhand = 'stop'
      }
      i = 0 # declare i because foreach uses it, without declaring it
      `%myinfix%` = foreach::`%dopar%`
      output_list = foreach::foreach(i = f0:f1,  .packages = packages2passon,
                                     .export = functions2passon, .errorhandling = errhand) %myinfix% {
                                       tryCatchResult = tryCatch({
                                         main_part6_recordlevel(i, metadatadir, f0, f1,
                                                                fnames.ms5raw, ffdone, EXT, verbose,
                                                                params_general, params_phyact, params_247,
                                                                params_cleaning)
                                       })
                                       return(tryCatchResult)
                                     }
      on.exit(parallel::stopCluster(cl))
      for (oli in 1:length(output_list)) { # logged error and warning messages
        if (is.null(unlist(output_list[oli])) == FALSE) {
          if (verbose == TRUE) cat(paste0("\nErrors and warnings for ", fnames.ms5raw[oli]))
          print(unlist(output_list[oli])) # print any error and warnings observed
        }
      }
    } else {
      errors = list()
      for (i in f0:f1) {
        if (verbose == TRUE) cat(paste0(i, " "))
        function_to_evaluate = expression(
          main_part6_recordlevel(i, metadatadir, f0, f1,
                                 fnames.ms5raw, ffdone, EXT, verbose,
                                 params_general, params_phyact, params_247,
                                 params_cleaning)
        )
        if (params_general[["use_trycatch_serial"]] == TRUE) {
          tryCatch(
            eval(function_to_evaluate),
            error = function(e) {
              err_msg = conditionMessage(e)
              errors[[as.character(fnames.ms5raw[i])]] <<- err_msg
            }
          )
        } else {
          eval(function_to_evaluate)
        }
      }
      # show logged errors after the loop:
      if (params_general[["use_trycatch_serial"]] == TRUE && verbose == TRUE) {
        if (length(errors) > 0) {
          cat(paste0("\n\nErrors in part 6... for:"))
          cat(paste0("\n-", names(errors), ": ", unlist(errors), collapse = ""))
        }
      }
    }
  }
}

g.loadlog = function(loglocation = c(), coln1 = c(), colid = c(), 
                     sleeplogsep = ",", meta.sleep.folder = c(), desiredtz="", sleepwindowType = c()) {
  # declare local functions:
  getIDstartdate = function(x) {
    rec_starttime = ID = c()
    load(x)
    invisible(list(ID = ID, rec_starttime = rec_starttime))
  }
  remove_empty_rows_cols = function(logmatrix, name) {
    logmatrix = as.data.frame(logmatrix[which((rowSums(logmatrix != "") != 0) == TRUE),
                                        which((colSums(logmatrix != "") != 0) == TRUE)])
    logmatrix = as.data.frame(logmatrix)
    if (length(name) > 0 & nrow(logmatrix) > 0) {
      newnames = c("ID", "date", rep(paste0(name, 1:ncol(logmatrix)), each = 2))
      colnames(logmatrix) = newnames[1:ncol(logmatrix)]
    }
    return(logmatrix)
  }
  removeEmptyCells = function(x) {
    if (nrow(x) == 1) {
      if (sum(x[, 2:ncol(x)] == "") == ncol(x) - 1) {
        emptyrows = 1
      } else {
        emptyrows = NULL
      }
    } else {
      emptyrows = which(rowSums(x[, 2:ncol(x)] == "") == ncol(x) - 1)
    }
    if (length(emptyrows)) {
      x = as.matrix(x[-emptyrows,])
    }
    if (length(x) != 0) {
      if (ncol(x) == 1 & nrow(x) > 1) {
        x = t(x)
      }
    }
    emptycols = which(colSums(x == "") == nrow(x))
    colp = ncol(x)
    twocols = c(colp - 1, colp)
    while (min(twocols) > 0) {
      if (all(twocols %in% emptycols)) {
        x = as.matrix(x[, -twocols])
        if (ncol(x) == 1) x = t(x)
        twocols = twocols - 2
      } else {
        break
      }
    }
    if (ncol(x) == 1) x = NULL
    return(x)
  }
  adjustLogFormat = function(S, nnights, mode = "sleeplog", coln1) {
    # function to adjust the sleeplog or bedlog format
    log = matrix(0,(nrow(S)*nnights),3)
    log_times = matrix(" ", (nrow(S) * nnights), 2)
    cnt = 1
    sli = coln1
    wki = sli + 1
    night = 1
    while (wki <= ncol(S)) { #loop through nights
      SL = as.character(S[,sli])
      WK = as.character(S[,wki])
      # Check whether any correction need to be made to the sleep log:
      for (j in 1:length(SL)) { #loop through participant
        # idtmp = S[j,colid]
        if (is.na(WK[j]) == FALSE & is.na(SL[j]) == FALSE & WK[j] != "" & SL[j] != "") {
          SLN = as.numeric(unlist(strsplit(SL[j],":")))
          WKN = as.numeric(unlist(strsplit(WK[j],":")))
          if (length(SLN) == 2) SLN = c(SLN,0) #add seconds when they are not stored
          if (length(WKN) == 2) WKN = c(WKN,0) #add seconds when they are not stored
          SL[j] = paste0(SLN[1], ":", SLN[2], ":", SLN[3])
          WK[j] = paste0(WKN[1], ":", WKN[2], ":", WKN[3])
          SLN2 = SLN[1] * 3600 + SLN[2] * 60 + SLN[3]
          WKN2 = WKN[1] * 3600 + WKN[2] * 60  + WKN[3]
          if (is.na(WKN2)) stop(paste0(WKN[1]," as found in the ", mode, " is not a valid timestamp"), call. = FALSE)
          if (is.na(SLN2)) stop(paste0(SLN[1]," as found in the ", mode, " is not a valid timestamp"), call. = FALSE)
          if (WKN2 > SLN2) { #e.g. 01:00 - 07:00
            dur = WKN2 - SLN2
          } else if (WKN2 < SLN2) { #e.g. 22:00 - 07:00
            dur = ((24*3600) - SLN2) + WKN2
          }
          dur = dur / 3600
        } else {
          cnt_time_notrecognise = cnt_time_notrecognise + 1
          dur = 0
          is.na(dur) =  TRUE
        }
        # add extra row if needed
        if (nrow(log) < cnt) {
          log = rbind(log, matrix(0, 1, 3))
          log_times = rbind(log_times, matrix(0, 1, 2))
        }
        # store information in log
        log[cnt,1] = as.character(S[j,colid])
        log[cnt,2] = night #ifelse(deltadate > 0, yes = i, no = i + abs(deltadate))
        log[cnt,3] = dur
        log_times[cnt,1] = SL[j]
        log_times[cnt,2] = WK[j]
        cnt = cnt + 1
      }
      sli = sli + 2
      wki = wki + 2
      night = night + 1
    }
    # delete id-numbers that are unrecognisable
    empty_rows = which(as.character(log[,1]) == "0")
    if (length(empty_rows) > 0) {
      log = log[-empty_rows, , drop = FALSE]
      log_times = log_times[-empty_rows, , drop = FALSE]
    }
    log = as.data.frame(log, stringsAsFactors = FALSE)
    names(log) = c("ID","night","duration")
    if (mode == "sleeplog") {
      log$sleeponset = log_times[,1]
      log$sleepwake = log_times[,2]
    } else if (mode == "bedlog") {
      log$bedstart = log_times[,1]
      log$bedend = log_times[,2]
    }
    # keep only the non-empty rows as they can only lead to confusion later on
    log = log[which(is.na(log$duration) == FALSE),]
    return(log)
  }
  
  #==========================================================================
  # Main code starts here
  #==========================================================================
  dateformat_correct = "%Y-%m-%d" # set default value
  deltadate = 0
  #===============================
  # Load sleep log data...
  S = data.table::fread(file = loglocation, stringsAsFactors = FALSE, data.table = FALSE,
                        check.names = TRUE, colClasses = "character")
  # Check for duplicated ID
  if (any(duplicated(S[, colid]))) {
    duplicatedIDs = paste0(S[duplicated(S[, colid]), colid], collapse = " ")
    duplicatedIDs = gsub(pattern = " ", replacement = "", x = duplicatedIDs)
    if (duplicatedIDs != "") {
      stop(paste0("Sleeplog has duplicated entries (rows) for ID(s) ",
                  duplicatedIDs, 
                  ", please fix. GGIR expects one sleeplog row per unique ID. "), call. = FALSE)
    } else {
      # Duplicated rows without an ID, ignore these
      S = S[!duplicated(S[, colid]),]
    }
  }
  
  if (colnames(S)[1] == "V1" && any(S[1, ] == "")) {
    stop(paste0("Sleeplog column found with empty header, please fix. This can also happen if ",
                "there are empty columns at the end, delete those columns if applicable."), call. = FALSE)
  }
  cnt_time_notrecognise = 0
  advanced_sleeplog = length(grep(pattern = "date", x = colnames(S), ignore.case = TRUE)) > 0
  if (advanced_sleeplog ==  TRUE) {
    if (length(meta.sleep.folder) > 0) {
      
      startdates = lapply(X = dir(meta.sleep.folder, full.names = T), FUN = getIDstartdate)
      startdates = as.data.frame(data.table::rbindlist(startdates, fill = TRUE))
      
      startdates$startAtMidnight = FALSE
      # If recording starts at midnight or before 4am that first half night
      # is still counted in part 3, which means that the start date of the recording
      # does not equal the start date of the nights. Via startdates$startAtMidnight
      # we correct for this
      starthour = as.numeric(format(as.POSIXct(x = startdates$rec_starttime, format = "%Y-%m-%dT%H:%M:%S%z", tz = desiredtz), format = "%H"))
      startdates$startAtMidnight[which(starthour <= 4)] = TRUE
      colnames(startdates)[1:2] = c("ID", "startdate")
      startdates$startdate = as.Date(iso8601chartime2POSIX(startdates$startdate, tz = desiredtz), tz = desiredtz)
    } else {
      warning("\nArgument meta.sleep.folder has not been specified")
    }
  }
  if (length(S) == 0) {
    warning(paste0("Could not read sleeplog file, check that file path is correct.",
                   "Tip: Try to aply function g.loadlog to your sleeplog file first",
                   " to verify that sleeplog is correctly processed."), call. = FALSE)
  } else {
    if (nrow(S) == 0 | ncol(S) <= 2) {
      warning(paste0("Could not read sleeplog file. Does it have at least 3 columns",
                     " and comma seperated values?",
                     " Tip: Try to aply function g.loadlog to your sleeplog file ",
                     "first to verify that sleeplog is correctly processed."), call. = FALSE)
    }
  }
  count = 1 # to keep track of row in new sleeplog matrix
  naplog = nonwearlog = newsleeplog = imputecodelog = c()
  if (advanced_sleeplog ==  TRUE) {
    # assumptions:
    # if date occurs in column names we assume it is an advanced sleeplogreport
    # columnames with onset|inbed|tobed|lightsout represent start of the main sleep/inbed window in a day
    # columnames with wakeup represent end of the main sleep/inbed window in a day
    # columnames with nap represent nap start or end-times
    # columnames with nonwear represent nonwear start or end-times
    # dates are expressed as YYYY-mm-dd
    datecols = grep(pattern = "date", x = colnames(S), value = FALSE, ignore.case = TRUE)
    nnights = length(datecols)
    # if date occurs in column names at least twice we assume it is an advanced sleeplogreport
    if (length(datecols) > 1) { # if yes, do:
      bedstartcols = grep(pattern = "lightsout|inbed|tobed|bedstart",x = colnames(S), value = FALSE, ignore.case = TRUE)
      bedendcols = grep(pattern = "lightson|outbed|bedend",x = colnames(S), value = FALSE, ignore.case = TRUE)
      wakecols = grep(pattern = "wakeup",x = colnames(S), value = FALSE, ignore.case = TRUE)
      onsetcols = grep(pattern = "onset",x = colnames(S), value = FALSE, ignore.case = TRUE)
      napcols = grep(pattern = "nap",x = colnames(S), value = FALSE, ignore.case = TRUE)
      nonwearcols = grep(pattern = "nonwear",x = colnames(S), value = FALSE, ignore.case = TRUE)
      imputecols = grep(pattern = "impute",x = colnames(S), value = FALSE, ignore.case = TRUE)
      # Create new sleeplog consisting of:
      # - original ID column
      # - empty columns if relevant to make sleeplog match accelerometer recording, make sure coln1 argument is used
      # - onset and wakup times of sleeplog, for this extract dates from sleeplog to check for missing days
      newsleeplog = matrix("", nrow(S), max(c(nnights*2, 100)) + 1)
      newbedlog = matrix("", nrow(S), max(c(nnights*2, 100)) + 1)
      naplog = matrix("", nrow(S)*nnights * 5, 50) #ID date start end
      nonwearlog = matrix("", nrow(S)*nnights * 5, 50) #ID date start end
      if (length(imputecols) > 0) {
        imputecodelog = matrix("", nrow(S)*nnights * 2, 50) #ID date start end
      } else {
        imputecodelog = NULL
      }
      napcnt = nwcnt = iccnt = 1
      IDcouldNotBeMatched = TRUE
      dateformat_found = FALSE
      dateformats_to_consider = c("%Y-%m-%d", "%d-%m-%Y", "%m-%d-%Y", "%Y-%d-%m",
                                  "%y-%m-%d", "%d-%m-%y", "%m-%d-%y", "%y-%d-%m",
                                  "%Y/%m/%d", "%d/%m/%Y", "%m/%d/%Y", "%Y/%d/%m",
                                  "%y/%m/%d", "%d/%m/%y", "%m/%d/%y", "%y/%d/%m")
      for (i in 1:nrow(S)) { # loop through rows in sleeplog
        ID = S[i,colid]
        if (ID %in% startdates$ID == TRUE) { # matching ID in acc data, if not ignore ID
          IDcouldNotBeMatched = FALSE
          matchingID = which(startdates$ID == ID)
          if (length(matchingID) > 1) {
            warning(paste0("There is more than 1 accelerometer recording for ID ",
                           ID, ", assuming that the first corresponds to the ",
                           "sleeplog entry for this ID"), call. = FALSE)
            matchingID  = matchingID[1]
          }
          startdate_acc = as.Date(startdates$startdate[matchingID], tz = desiredtz)
          startdate_sleeplog = as.character(S[i, datecols[1:pmin(length(datecols), 5)]])
          Sdates_correct = c()
          if (dateformat_found == TRUE && dateformats_to_consider[1] != dateformat_correct) {
            # If found then first try that before trying anything else
            dateformats_to_consider = unique(c(dateformat_correct, dateformats_to_consider))
          }
          # Detect date format in sleeplog:
          for (dateformat in dateformats_to_consider) {
            startdate_sleeplog_tmp = as.Date(startdate_sleeplog[which(startdate_sleeplog != "")], format = dateformat, tz = desiredtz)
            if (is.null(startdate_sleeplog_tmp)) next
            Sdates = as.Date(as.character(S[i,datecols]), format = dateformat, tz = desiredtz)
            if (length(which(diff(which(is.na(Sdates))) > 1)) > 0) {
              stop(paste0("\nSleeplog for ID: ", ID, " has missing date(s)"), call. = FALSE)
            }
            if (all(is.na(startdate_sleeplog_tmp) == FALSE)) {
              deltadate = as.numeric(startdate_sleeplog_tmp - startdate_acc)
              if (all(is.na(deltadate) == FALSE)) {
                if (all(abs(deltadate) < 30)) {
                  startdate_sleeplog = startdate_sleeplog_tmp[1]
                  Sdates_correct = Sdates
                  dateformat_correct = dateformat
                  deltadate = deltadate[1]
                  dateformat_found = TRUE
                  break
                }
              }
            }
          }
          if (is.null(Sdates_correct)) {
            # skip this row because it is empty
            warning(paste0("\nSkipping sleeplog row for ID ", ID,
                           " because it has no date(s) or",
                           " date range sleeplog does not overlap with",
                           " date range accelerometer. Please check",
                           " diary dates are correct."), call. = FALSE)
            next
          }
          if (deltadate > 300) {
            warning(paste0("For ID ", ID, " the sleeplog start date is more than 300 days separated ",
                           "from the dates in the accelerometer recording, this may indicate a ",
                           "problem with date formats or their recognition, please check."), call. = FALSE)
          }
          if (startdates$startAtMidnight[matchingID] == TRUE) {
            # If the first day in the advanced sleeplog is 28/11 
            # and the recording starts at midnight 27/11 00:00:00
            # then that means that we miss the first 2 nights.
            # However, the code above only sees a
            # difference of 1 day (deltadate) between 27/11 and 28/11.
            # If the recording starts at 27/11 00:00:05 this is correct 
            # because 27/11 is not counted as a night in g.part3 and g.part4.
            # This is why we need to do + 1 if the recording starts at midnight.
            deltadate = deltadate + 1
          }
          
          if (length(Sdates_correct) == 0 | is.na(startdate_sleeplog) == TRUE) {
            warning(paste0("\nSleeplog for ID: ",ID," not used because first date",
                           " not within 30 days of first date in accerometer recording"), call. = FALSE)
          } else {
            # handle missing dates
            ndates = as.numeric(diff(range(Sdates_correct[!is.na(Sdates_correct)]))) + 1
            if (ndates > 300) warning(paste0("For ID ", ID, " the sleeplog has has ",
                                             "more than 300 missing dates, this may ",
                                             "indicate a problem with date format ",
                                             "recognition. Please check."), call. = FALSE)
            if (ndates > nnights) {
              extraColumns = matrix("", nrow(newsleeplog), max(c((ndates - nnights)*2, 100)) + 1)
              newsleeplog = cbind(newsleeplog, extraColumns)
              newbedlog = cbind(newbedlog, extraColumns)
              extraColumns = matrix("", nrow(naplog), max(c((ndates - nnights)*2, 100)) + 1)
              naplog = cbind(naplog, extraColumns)
              nonwearlog = cbind(nonwearlog, extraColumns)
              nnights = ndates
            }
            if (startdate_sleeplog - deltadate > startdate_sleeplog + nnights) {
              warning(paste0("Accelerometer recording for ID ",
                             ID, " does not overlap with sleeplog date",
                             " range"), call. = FALSE)
              next
            } 
            # only attempt to use sleeplog if start date could be recognised
            # Add row to newsleeplog if somehow there are not enough rows
            if (count > nrow(newsleeplog)) {
              newsleeplog = rbind(newsleeplog, matrix(NA, 1, ncol(newsleeplog)))
            }
            if (count > nrow(newbedlog)) {
              newbedlog = rbind(newbedlog, matrix(NA, 1, ncol(newbedlog)))
            }
            newsleeplog[count ,1] = ID
            newbedlog[count ,1] = ID
            newsleeplog_times = newbedlog_times = rep("time", 100)
            newCounter = 1
            expected_dates = seq(startdate_sleeplog - deltadate, startdate_sleeplog + nnights, by = 1)
            # loop over expect dates giving start date of sleeplog
            for (ni in 1:(length(expected_dates) - 1)) {
              # checking whether date exists in sleeplog
              ind = which(Sdates_correct == as.Date(expected_dates[ni], tz = desiredtz))
              if (length(ind) > 0) {
                if (length(ind) > 1) {
                  duplicatedDate = unique(as.character(S[i, datecols[ind]]))
                  stop(paste0("\n", ID, " has duplicate dates in the diary, please fix ", duplicatedDate), call. = FALSE)
                }
                curdatecol = datecols[ind]
                nextdatecol =  datecols[which(datecols > curdatecol)[1]]
                doublenextdatecol =  datecols[which(datecols > nextdatecol)[1]]
                lastday = FALSE
                if (is.na(nextdatecol)) {
                  nextdatecol = ncol(S) + 1
                  lastday = TRUE
                }
                if (is.na(doublenextdatecol)) {
                  doublenextdatecol = ncol(S) + 1
                }
                # Handle mixed reporting of time in bed and SPT"
                if (length(bedendcols) == 0 & length(bedstartcols) != 0 &
                    length(onsetcols) == 0 & length(wakecols) != 0) {
                  # bedstart and wakeup are present, but bedend and sleeponset not,
                  # treat wake as bedend such that this can be treated as time in bed period
                  bedendcols = wakecols
                  wakecols = NULL
                }
                if (length(bedendcols) != 0 & length(bedstartcols) == 0 &
                    length(onsetcols) != 0 & length(wakecols) == 0) {
                  # bedend and onsetpresent, but bedstart and wakeup not,
                  # treat bedend as wakeup such that this can be treat as SPT
                  wakecols = bedendcols
                  bedendcols = NULL
                }
                # Sleeplog:
                onseti = onsetcols[which(onsetcols > curdatecol & onsetcols < nextdatecol)]
                if (lastday == FALSE) {
                  wakeupi = wakecols[which(wakecols > nextdatecol & wakecols < doublenextdatecol)[1]]
                  wakeuptime = S[i,wakeupi]
                } else {
                  wakeuptime = ""
                }
                if (length(onseti) == 1 & length(wakeupi) == 1) {
                  newsleeplog_times[newCounter:(newCounter + 1)] = c(S[i,onseti], wakeuptime)
                } else {
                  newsleeplog_times[newCounter:(newCounter + 1)] = c("", "")
                }
                # time in bed
                bedstarti = bedstartcols[which(bedstartcols > curdatecol & bedstartcols < nextdatecol)]
                if (lastday == FALSE) {
                  bedendi = bedendcols[which(bedendcols > nextdatecol & bedendcols < doublenextdatecol)[1]]
                  bedendtime = S[i,bedendi]
                } else {
                  bedendtime = ""
                }
                if (length(bedstarti) == 1 & length(bedendi) == 1) {
                  newbedlog_times[newCounter:(newCounter + 1)] = c(S[i,bedstarti], bedendtime)
                } else {
                  newbedlog_times[newCounter:(newCounter + 1)] = c("", "")
                }
                # Also grap nap, non-wear, and imputation code info and put those in separate matrices:
                naps = napcols[which(napcols  > curdatecol & napcols < nextdatecol)]
                nonwears = nonwearcols[which(nonwearcols  > curdatecol & nonwearcols < nextdatecol)]
                imputecodes = imputecols[which(imputecols  > nextdatecol & imputecols < doublenextdatecol)[1]]
                if (is.na(imputecodes)) imputecodes = NULL
                if (length(naps) > 0) {
                  naplog[napcnt, 1] = ID
                  naplog[napcnt, 2] = S[i, curdatecol]
                  naplog[napcnt, 3:(2 + length(naps))] = as.character(S[i, naps])
                  napcnt = napcnt + 1
                }
                if (length(nonwears) > 0) {
                  nonwearlog[nwcnt, 1] = ID
                  nonwearlog[nwcnt, 2] = S[i, curdatecol]
                  nonwearlog[nwcnt, 3:(2 + length(nonwears))] = as.character(S[i, nonwears ])
                  nwcnt = nwcnt + 1
                }
                if (length(imputecodes) > 0 && imputecodes <= ncol(S)) {
                  imputecodelog[iccnt, 1] = ID
                  imputecodelog[iccnt, 2] = S[i, curdatecol]
                  imputecodelog[iccnt, 3:(2 + length(imputecodes))] = gsub("[^0-9.-]", "", as.character(S[i, imputecodes]))
                  iccnt = iccnt + 1
                }
              } else {
                newsleeplog_times[newCounter:(newCounter + 1)] = c("", "")
                newbedlog_times[newCounter:(newCounter + 1)] = c("", "")
              }
              newCounter = newCounter + 2
              if (newCounter > length(newbedlog_times) - 5) {
                newbedlog_times = c(newbedlog_times, rep("time", 100))
                newsleeplog_times = c(newsleeplog_times, rep("time", 100))
              }
            }
            newsleeplog_times = newsleeplog_times[which(newsleeplog_times != "time")]
            newbedlog_times = newbedlog_times[which(newbedlog_times != "time")]
            # add columns to sleeplog
            extracols = (length(newsleeplog_times) + 2) - ncol(newsleeplog)
            if (extracols > 0) {
              newsleeplog = cbind(newsleeplog, matrix(NA, nrow(newsleeplog), extracols))
            }
            newsleeplog[count, 2:(length(newsleeplog_times) + 1)] = newsleeplog_times
            
            # add columns to bedlog
            extracols = (length(newbedlog_times) + 2) - ncol(newbedlog)
            if (extracols > 0) {
              newbedlog = cbind(newbedlog, matrix(NA, nrow(newbedlog), extracols))
            }
            newbedlog[count, 2:(length(newbedlog_times) + 1)] = newbedlog_times
            
            count  = count + 1
          }
        }
      }
      if (IDcouldNotBeMatched == TRUE) {
        warning(paste0("\nNone of the IDs in the accelerometer data could be matched with",
                       " the ID numbers in the sleeplog. You may want to check that the ID",
                       " format in your sleeplog is consistent with the ID column in the GGIR part2 csv-report,",
                       " and that argument coldid is correctly set."), call. = FALSE)
      }
      # remove empty rows and columns:
      
      if (length(naplog) > 0) {
        naplog = remove_empty_rows_cols(naplog, name = "nap")
      }
      if (length(nonwearlog) > 0) {
        nonwearlog = remove_empty_rows_cols(nonwearlog, name = "nonwear")
      }
      if (length(imputecodelog) > 0) {
        imputecodelog = remove_empty_rows_cols(imputecodelog, name = "imputecode")
        colnames(imputecodelog)[3] = "imputecode"
        imputecodelog$date = as.Date(imputecodelog$date, dateformat_correct)
      }
      
      if (length(newsleeplog) > 0) {
        newsleeplog = removeEmptyCells(newsleeplog)
        if (!is.null(newsleeplog)) {
          S = as.data.frame(newsleeplog)
        }
        coln1 = 2
        colid = 1
      }
      if (length(newbedlog) > 0) {
        newbedlog = removeEmptyCells(newbedlog)
        if (!is.null(newbedlog)) {
          B = as.data.frame(newbedlog)
        }
        coln1 = 2
        colid = 1
      }
    }
  } else {
    nnights = (ncol(S) - coln1 + 1) / 2
  }
  # test whether number of columns with night information in sleeplog is odd
  # this would provide nnights %% 2 == 0.5
  if (nnights %% 2 == 0.5) { 
    warning(paste0("\nWe see an odd number of timestamp columns",
                   " in the sleeplog. The last column will be ignored. If this is incorrect,",
                   " please check that argument coln1 is correctly specified if you use a basic sleeplog format and",
                   " that all days have a date column if you use an advanced sleeplog format."), call. = FALSE)
    nnights = floor(nnights)
  }
  nnights = nnights + deltadate + 1 # to account for the possibility of extra night at the beginning of recording
  if (!is.null(sleepwindowType) && advanced_sleeplog == FALSE) {
    if (sleepwindowType == "TimeInBed") {
      B = S
      rm(S)
    }
  }
  # # From here we continue with original code focused on sleeplog only
  if (exists("S") && ncol(S) > 0 && nnights > 0) {
    sleeplog = adjustLogFormat(S, nnights, mode = "sleeplog", coln1)
  } else {
    sleeplog = NULL
  }
  if (exists("B") && ncol(B) > 0 && nnights > 0) {
    bedlog = adjustLogFormat(B, nnights, mode = "bedlog", coln1)
  } else {
    bedlog = NULL
  }
  invisible(list(sleeplog = sleeplog, nonwearlog = nonwearlog, naplog = naplog, bedlog = bedlog,
                 imputecodelog = imputecodelog,
                 dateformat = dateformat_correct))
}

g.analyse =  function(I, C, M, IMP, params_247 = c(), params_phyact = c(),
                      params_general = c(), params_cleaning = c(),
                      quantiletype = 7, myfun = c(), ID, ...) {
  
  #get input variables
  input = list(...)
  if (length(input) > 0 ||
      length(params_247) == 0 || length(params_phyact) == 0 ||
      length(params_general) == 0 || length(params_cleaning) == 0) {
    # Extract and check parameters if user provides more arguments than just the parameter arguments,
    # or if params_[...] aren't specified (so need to be filled with defaults).
    # So, inside GGIR this will not be used, but it is used when g.analyse is used on its own
    # as if it was still the old g.analyse function
    params = extract_params(params_247 = params_247,
                            params_phyact = params_phyact,
                            params_general = params_general,
                            params_cleaning = params_cleaning,
                            input = input,
                            params2check = c("247", "phyact", "general", "cleaning")) # load default parameters
    params_247 = params$params_247
    params_phyact = params$params_phyact
    params_general = params$params_general
    params_cleaning = params$params_cleaning
    rm(params)
  }
  params_247[["L5M5window"]] = c(0,24) # as of version 1.6-0 this is hardcoded because argument qwindow now
  # specifies the window over which L5M5 analysis is done. So, L5M5window is a depricated
  # argument and this is also clarified in the documentation
  
  desiredtz = params_general[["desiredtz"]]
  idloc = params_general[["idloc"]]
  includedaycrit = params_cleaning[["includedaycrit"]][1]
  acc.metric = params_general[["acc.metric"]]
  
  fname = I$filename
  averageday = IMP$averageday
  if (is.null(IMP$data_masking_strategy)) {
    data_masking_strategy = IMP$strategy
  } else {
    data_masking_strategy = IMP$data_masking_strategy
  }
  hrs.del.start = IMP$hrs.del.start
  hrs.del.end = IMP$hrs.del.end
  maxdur = IMP$maxdur
  windowsizes = M$windowsizes
  metalong = M$metalong
  metashort = IMP$metashort
  rout = IMP$rout
  nonwearHoursFiltered = IMP$nonwearHoursFiltered
  nonwearEventsFiltered = IMP$nonwearEventsFiltered
  wdaycode = M$wday
  wdayname = M$wdayname
  if (length(params_phyact[["mvpadur"]]) > 0) params_phyact[["mvpadur"]] = sort(params_phyact[["mvpadur"]])
  LC2 = IMP$LC2
  LC = IMP$LC
  dcomplscore = IMP$dcomplscore
  r4 = as.numeric(as.matrix(rout[,4]))
  r5 = as.numeric(as.matrix(rout[,5]))
  ws3 = windowsizes[1]
  ws2 = windowsizes[2]
  keepindex_46 = keepindex_48 = c()
  # Extracting basic information about the file
  hvars = g.extractheadervars(I)
  # ID = hvars$ID;              iID = hvars$iID; IDd = hvars$IDd
  HN = hvars$HN;              sensor.location = hvars$sensor.location
  SX = hvars$SX;                deviceSerialNumber = hvars$deviceSerialNumber
  n_ws2_perday = (1440*60) / ws2
  n_ws3_perday = (1440*60) / ws3
  if (((nrow(metalong)/((1440*60)/ws2)*10) - (nrow(metashort)/((60/ws3)*1440)) * 10) > 1) {
    stop("Matrices 'metalong' and 'metashort' are not compatible")
  }
  #----------------------
  # Extract ID centrally
  # ID = extractID(hvars = hvars, idloc = idloc, fname = I$filename)
  
  #--------------------------------------------------------------
  # Extract qwindow if an activity log is provided:
  qwindow_actlog = FALSE
  if (is.data.frame(params_247[["qwindow"]]) == TRUE) {
    qwindow_actlog = TRUE
    IDacc = gsub(pattern = " ", replacement = "", x = as.character(ID))
    IDlog = gsub(pattern = " ", replacement = "", x = as.character(params_247[["qwindow"]]$ID))
    params_247[["qwindow"]] = params_247[["qwindow"]][which(IDlog == IDacc),]
  }
  # # Time window for L5 & M5 analysis (commented out because this is now defined further down)
  # t0_LFMF = L5M5window[1] #start in 24 hour clock hours
  # t1_LFMF = L5M5window[2]+(winhr-(M5L5res/60)) #end in 24 hour clock hours (if a value higher than 24 is chosen, it will take early hours of previous day to complete the 5 hour window
  # Time window for distribution analysis
  t_TWDI = params_247[["qwindow"]] #start and of 24 hour clock hours
  if (length(params_247[["qwindow"]]) == 0) {
    t_TWDI = c(0,24)
    if ((length(params_247[["qlevels"]]) > 0 | length(params_247[["ilevels"]]) > 0)) params_247[["qwindow"]] = c(0,24)
    qwindow_actlog = FALSE # ignore qwdinow_actlog if it does not produce actual qwindow values
  }
  if (length(params_247[["qwindow"]]) > 0 & qwindow_actlog == FALSE) {
    if (params_247[["qwindow"]][1] != 0) params_247[["qwindow"]] = c(0,params_247[["qwindow"]])
    if (params_247[["qwindow"]][length(params_247[["qwindow"]])] != 24) params_247[["qwindow"]] = c(params_247[["qwindow"]],24)
  }
  #==========================================================================================
  # Setting paramters (NO USER INPUT NEEDED FROM HERE ONWARDS)
  domvpa = doilevels = doiglevels = doquan = FALSE
  if (length(params_247[["qlevels"]]) > 0) doquan = TRUE
  if (length(params_247[["ilevels"]]) > 0) doilevels = TRUE
  if (length(params_247[["iglevels"]]) > 0) {
    if (length(params_247[["iglevels"]]) == 1) {
      params_247[["iglevels"]] = c(seq(0, 4000, by = 25), 8000) # to introduce option to just say TRUE
    }
    doiglevels = TRUE
  }
  if (length(params_phyact[["mvpathreshold"]]) > 0) domvpa = TRUE
  doperday = TRUE
  #------------------------------------------------------
  NVARS = (length(colnames(metashort)) - 1)
  if (NVARS < 1) NVARS = 1
  if (length(params_247[["qwindow"]]) > 0 | qwindow_actlog == TRUE) {
    NVARS = NVARS + 2 # for qwindow non-wear time
  }
  nfeatures = 50 + NVARS * (21 + length(params_247[["qlevels"]]) + length(params_247[["ilevels"]]))    #levels changed into qlevels
  if (length(params_247[["qwindow"]]) > 0 | qwindow_actlog == TRUE) {
    nfeatures = 50 + NVARS*(length(params_247[["qwindow"]]) * 
                              (21 + (length(params_247[["qlevels"]]) + length(params_247[["ilevels"]]))))
  }
  i = 1
  #---------------
  if (domvpa) { #create dummy data
    mvpanames = matrix(0, 6, length(params_phyact[["mvpathreshold"]]))
    mvpanames[, 1:length(params_phyact[["mvpathreshold"]])] = c("MVPA1","MVPA2","MVPA3","MVPA4","MVPA5","MVPA6")
  }
  # What is the minimum number of accelerometer axis needed to meet the criteria for nonwear in order for the data to be detected as nonwear?
  wearthreshold = 2 #needs to be 0, 1 or 2 (hard coded to avoid inconsistency in literature)
  #---------------------
  # detect first and last midnight and all midnights
  tsi = which(colnames(metalong) == "timestamp")
  time = as.character(metalong[,tsi])
  startt = as.character(metalong[1, tsi])
  # basic file characteristics
  LD = nrow(metalong) * (ws2/60) #length data in minutes
  ND = nrow(metalong)/n_ws2_perday #number of days
  #  time variable
  timeline = seq(0, ceiling(nrow(metalong)/n_ws2_perday), by = 1/n_ws2_perday)
  timeline = timeline[1:nrow(metalong)]
  tooshort = 0
  dmidn = g.detecmidnight(time,
                          desiredtz =  params_general[["desiredtz"]],
                          dayborder = params_general[["dayborder"]]) #ND,
  firstmidnight = dmidn$firstmidnight;  firstmidnighti = dmidn$firstmidnighti
  lastmidnight = dmidn$lastmidnight;    lastmidnighti = dmidn$lastmidnighti
  midnights = dmidn$midnights;          midnightsi = dmidn$midnightsi
  starttimei = 1
  endtimei = nrow(M$metalong)
  if (data_masking_strategy == 2) {
    starttimei = firstmidnighti
    endtimei = lastmidnighti - 1
  }
  # get r5long - this is vector with the 5 second scores of what data needs to be imputed
  r5long = matrix(0,length(r5),(ws2/ws3))
  r5long = replace(r5long,1:length(r5long),r5)
  r5long = t(r5long)
  dim(r5long) = c((length(r5)*(ws2/ws3)),1)
  if (length(myfun) > 0) {
    ExtFunColsi = which(colnames(M$metashort) %in% myfun$colnames ==  TRUE)
  } else {
    ExtFunColsi = c()
  }
  #===============================================
  # Extract features from the imputed data
  qcheck = r5long
  # LW = length(which(as.numeric(qcheck) != 1)) / (60/ws3) #number of minutes wear time between first and last midnights
  nfulldays = (lastmidnighti - firstmidnighti) / ((3600/ws2)*24)
  ndays = length(midnights) + 1 #ceiling(nfulldays + 2) # ceiling to cope with days with 23 hours
  #-------------------------------------
  # Detect orientation (at the moment only designed for hip with accelerometer):
  # assess which angle per axis is most strongly 24 hour correlated:
  # for hip worn devices this will be the vertical axis
  longitudinal_axis_id = ""
  epochday = 24 * 60 * (60/ws3)
  Ndays = floor(nrow(IMP$metashort)/epochday)
  if (length(which(c("anglex","angley","anglez") %in% colnames(IMP$metashort) == FALSE)) == 0 &
      Ndays >= 2) {
    Nhalfdays = Ndays - 1
    CorrA = rep(0,3)
    cnt = 1
    for (anglename in  c("anglex","angley","anglez") ) {
      if (sd(IMP$metashort[,anglename]) > 0) {
        CorrA[cnt] = stats::cor(IMP$metashort[1:(Nhalfdays*epochday), anglename],
                                IMP$metashort[(((Ndays - Nhalfdays) *
                                                  epochday) + 1):(Ndays * epochday), anglename])
      } else {
        CorrA[cnt] = NA
      }
      cnt = cnt + 1
    }
    if (length(which(is.na(CorrA) == FALSE)) > 0) {
      longitudinal_axis_id = which.max(CorrA)
    } else {
      longitudinal_axis_id = ""
    }
  }
  #--------------------------------------
  # Analysis of the average day
  # Derivation of distribution characteristics of the average day: quantiles (percentiles) and L5M5 method
  # Note that this is done here before all the other analyses because it only relies on the average day
  # The values and variablenames are, however, stored in the filesummary matrix towards the end (not here
  # in function g.analyse.avday).
  output_avday = g.analyse.avday(doquan = doquan, averageday = averageday,
                                 M = M, IMP = IMP, t_TWDI = t_TWDI,
                                 quantiletype = quantiletype, ws3 = ws3,
                                 doiglevels = doiglevels, firstmidnighti = firstmidnighti, ws2 = ws2,
                                 midnightsi = midnightsi, params_247 = params_247, qcheck = qcheck,
                                 acc.metric = acc.metric, params_phyact = params_phyact)
  cosinor_coef = output_avday$cosinor_coef
  
  #--------------------------------------------------------------
  # Analysis per day
  if (doperday == TRUE) {
    output_perday = g.analyse.perday(ndays = ndays,
                                     firstmidnighti = firstmidnighti, time = time,
                                     nfeatures = nfeatures, midnightsi = midnightsi,
                                     metashort = metashort, averageday = averageday,
                                     doiglevels = doiglevels, nfulldays = nfulldays,
                                     lastmidnight = lastmidnight,
                                     ws3 = ws3, ws2 = ws2, qcheck = qcheck, fname = fname,
                                     sensor.location = sensor.location,
                                     wdayname = wdayname,
                                     tooshort = tooshort, includedaycrit = includedaycrit,
                                     quantiletype = quantiletype, doilevels = doilevels, 
                                     domvpa = domvpa,
                                     mvpanames = mvpanames, wdaycode = wdaycode, ID = ID, 
                                     deviceSerialNumber = deviceSerialNumber,
                                     doquan = doquan,  ExtFunColsi = ExtFunColsi,
                                     myfun = myfun,
                                     params_247 = params_247, params_phyact = params_phyact,
                                     params_general = params_general)
  }
  #metashort is shortened from midgnight to midnight if requested (data_masking_strategy 2)
  if (data_masking_strategy == 2) {
    if (starttimei == 1) {
      metashort = as.matrix(metashort[starttimei:(endtimei*(ws2/ws3)),])
    } else {
      metashort = as.matrix(metashort[(starttimei*(ws2/ws3)):(endtimei*(ws2/ws3)),])
    }
  }
  ND = nrow(metashort) / n_ws3_perday #update this because of reduction in datapoints (added on 20-11-2012)
  # LW = length(which(r5 < 1)) * (ws2/60) #length wear in minutes (for entire signal)
  LWp = length(which(r5[which(r4 == 0)] < 1)) * (ws2/60) #length wear in minutes (for protocol)
  LMp = length(which(r4 == 0)) * (ws2/60) #length protocol
  #====================================================================
  # Analysis per recording (entire file), and merge in average day analysis results
  #====================================================================
  # Extract the average 24 hr but ignore angle metrics and externally extracted sleep
  lookattmp = which(colnames(metashort) %in% c("angle","anglex", "angley", "anglez",
                                               "ExtSleep", "marker") ==  FALSE)
  lookat = lookattmp[which(lookattmp > 1)] #]c(2:ncol(metashort[,lookattmp]))
  colnames_to_lookat = colnames(metashort)[lookat]
  AveAccAve24hr = matrix(NA,length(lookat),1)
  if (length(which(r5 == 0)) > 0) { #to catch data_masking_strategy 2 with only 1 midnight and other cases where there is no valid data
    for (h in 1:length(lookat)) {
      average24h = matrix(0,n_ws3_perday,1)
      average24hc = matrix(0,n_ws3_perday,1)
      if (floor(ND) != 0) {
        for (j in 1:floor(ND)) {
          dataOneDay = as.numeric(as.matrix(metashort[(((j - 1) * n_ws3_perday) + 1):(j * n_ws3_perday),lookat[h]]))
          val = which(is.na(dataOneDay) == F)
          average24h[val,1] = average24h[val,1] + dataOneDay[val] #mean acceleration
          average24hc[val,1] = average24hc[val,1] + 1
        }
      }
      if (floor(ND) < ND) {
        if (floor(ND) == 0) {
          dataOneDay = as.numeric(as.matrix(metashort[,lookat[h]]))
        } else {
          dataOneDay = as.numeric(as.matrix(metashort[((floor(ND) * n_ws3_perday) + 1):nrow(metashort), lookat[h]]))
        }
        val = which(is.na(dataOneDay) == F)
        average24h[val,1] = average24h[val,1] + dataOneDay[val]  #mean acceleration
        average24hc[val,1] = average24hc[val,1] + 1
      }
      average24h = average24h / average24hc
      AveAccAve24hr[h] = mean(average24h) #average acceleration in an average 24 hour cycle
    }
  }
  rm(metalong); rm(metashort)
  dataqual_summary = data.frame(clipping_score = LC2  / ((LD/1440)*96),
                                meas_dur_dys =  LD/1440,
                                dcomplscore = dcomplscore,
                                meas_dur_def_proto_day = LMp / 1440,
                                wear_dur_def_proto_day = LWp / 1440,
                                nonwearHoursFiltered = nonwearHoursFiltered,
                                nonwearEventsFiltered = nonwearEventsFiltered)
  file_summary = data.frame(wdayname = wdayname,
                            deviceSerialNumber = deviceSerialNumber,
                            sensor.location = sensor.location,
                            ID = ID,
                            fname = fname,
                            startt = startt)
  
  if (!is.null(M$QClog)) {
    # Summarise the QC log (currently only expected from cwa Axivity, actigraph, and csv files)
    QCsummarise = function(QClog, wx) {
      x = ifelse(test = length(wx) > 0,
                 yes = sum(QClog$end[wx] - QClog$start[wx]) / 60,
                 no = 0)
      return(x)
    }
    # total imputation
    if ("imputed" %in% colnames(M$QClog)) {
      impdone = which(M$QClog$imputed == TRUE)
      if (any(colnames(M$QClog) == "timegaps_min")) {
        file_summary$Dur_imputed = sum(M$QClog$timegaps_min)
        file_summary$Nblocks_imputed = sum(M$QClog$timegaps_n)
      } else {
        file_summary$Dur_imputed = QCsummarise(M$QClog, impdone)
        file_summary$Nblocks_imputed = length(impdone)
      }
    }
    
    # checksum
    if ("checksum_pass" %in% colnames(M$QClog)) {
      chsum_failed = which(M$QClog$checksum_pass == FALSE)
      file_summary$Dur_chsum_failed = QCsummarise(M$QClog, chsum_failed)
      file_summary$Nblocks_chsum_failed = length(chsum_failed)
      
    }
    
    # nonincremental block ID
    if ("blockID_current" %in% colnames(M$QClog)) {
      nonincremental = which(M$QClog$blockID_current - M$QClog$blockID_next != 1)
      file_summary$Dur_nonincremental = QCsummarise(M$QClog, nonincremental)
      file_summary$Nblocks_nonincremental = length(nonincremental)
    }
    
    # sampling frequency issues
    if ("frequency_blockheader" %in% colnames(M$QClog)) {
      freqBlockHead = M$QClog$frequency_blockheader
      frequency_bias = abs(M$QClog$frequency_observed - freqBlockHead) / freqBlockHead
    }
    if ("frequency_bias" %in% colnames(M$QClog)) {
      freqissue = which(frequency_bias >= 0.05 & frequency_bias < 0.1)
      file_summary$Dur_freqissue_5_10 = QCsummarise(M$QClog, freqissue)
      file_summary$Nblock_freqissue_5_10 = length(freqissue)
      
      freqissue = which(frequency_bias >= 0.1 & frequency_bias < 0.2)
      file_summary$Dur_freqissue_10_20 = QCsummarise(M$QClog, freqissue)
      file_summary$Nblock_freqissue_10_20 = length(freqissue)
      
      freqissue = which(frequency_bias >= 0.2 & frequency_bias < 0.3)
      file_summary$Dur_freqissue_20_30 = QCsummarise(M$QClog, freqissue)
      file_summary$Nblock_freqissue_20_30 = length(freqissue)
      
      freqissue = which(frequency_bias >= 0.3)
      file_summary$Dur_freqissue_30 = QCsummarise(M$QClog, freqissue)
      file_summary$Nblock_freqissue_30 = length(freqissue)
    }
    if ("filehealth_ExtSleep_y_ExtAct_missing" %in% colnames(M$QClog)) {
      # Fitbit data
      file_summary = cbind(file_summary, as.data.frame(M$QClog))
    }
  }
  
  metrics_nav = list(lookat = lookat,
                     colnames_to_lookat = colnames_to_lookat,
                     longitudinal_axis_id = longitudinal_axis_id)
  output_perfile = g.analyse.perfile(I, C, metrics_nav,
                                     AveAccAve24hr, 
                                     doquan, doiglevels, tooshort, 
                                     params_247 = params_247, 
                                     params_cleaning = params_cleaning,
                                     params_general = params_general,
                                     output_avday = output_avday,
                                     output_perday = output_perday,
                                     dataqual_summary = dataqual_summary,
                                     file_summary = file_summary)
  
  filesummary = output_perfile$filesummary
  daysummary = output_perfile$daysummary
  
  if (length(cosinor_coef) > 0) {
    cosinor_ts = cosinor_coef$coefext$cosinor_ts
  } else {
    cosinor_ts = c()
  }
  invisible(list(summary = filesummary, daysummary = daysummary, cosinor_ts = cosinor_ts))
}

CalcSleepRegularityIndex = function(data = c(), epochsize = c(), desiredtz= c(),
                                    SRI1_smoothing_wsize_hrs = NULL,
                                    SRI1_smoothing_frac = NULL) {
  if (inherits(data$time[1], "character") || inherits(data$time[1], "factor")) {
    data$time = iso8601chartime2POSIX(data$time, tz = desiredtz)
  }
  # ignore columns that this function does not need
  data = data[, grep("temperature|selfreported|angle|invalid_|guider|ACC|step_count", colnames(data), invert = TRUE)]
  sleepcol = grep("sleepnap", colnames(data))
  if (length(sleepcol) != 1) {
    # Only run when used in GGIR part 3
    # because there we use sustained inactivity bouts which optionally need
    # to be smoothed
    sleepcol = which(colnames(data) %in% c("time", "invalid", "night") == FALSE)[1]
    # Optionally smooth sleep/wake classification
    if (!is.null(SRI1_smoothing_wsize_hrs) && !is.null(SRI1_smoothing_frac)) {
      data[,sleepcol] = zoo::rollmean(x = data[,sleepcol], k = (SRI1_smoothing_wsize_hrs * 3600) / epochsize, fill = 0)
      data[, sleepcol] = ifelse(data[,sleepcol] >= SRI1_smoothing_frac, yes = 1, no = 0)
    }
    invalid_epochs = which(data$invalid == 1)
  } else {
    # Only run when used in GGIR part 6
    invalid_epochs = which(data$invalidepoch == 1)
  }
  # Ignore invalid epochs and set them to NA
  if (length(invalid_epochs) > 0) { 
    is.na(data[invalid_epochs, sleepcol]) = TRUE
  }
  # Aggregate to 30-second epoch if epochsize is less than 30 seconds to be in line
  # with original paper by Phillips et al. 2017
  if (epochsize < 30 & (30/epochsize) == round(30/epochsize)) {
    data$time_num = floor(as.numeric(data$time) / 30) * 30
    # aggregate by mean and round off value in the next step, to make sure we still
    # have a simple 1 (sleep) or 0 (wakefullness) for the resulting epoch.
    data = aggregate(data[,sleepcol],
                     by = list(data$time_num), FUN = mean)
    colnames(data) = c("time_num",  "sleepstate")
    data$sleepstate = round(data$sleepstate)
    M = 24 * 60 * 2
  } else { # if larger epoch size, keep as it is
    data$time_num = as.numeric(data$time)
    colnames(data)[sleepcol] = "sleepstate"
    M = (24*3600) / epochsize
  }
  epochsize = max(c(epochsize, 30))
  # reinsert timestamps
  data$time = as.POSIXlt(data$time_num, tz = desiredtz, origin = "1970-01-01")
  data$date = as.Date(data$time)
  # add sec(ond) in the day
  convert2SecInDay = function(x) {
    tmp1 = format(x, format = "%H %M %S")
    tmp2 = unlist(strsplit(tmp1," "))
    SecInDay = sum(as.numeric(tmp2) * c(3600, 60, 1))
  }
  data$SecInDay = sapply(data$time, FUN = convert2SecInDay)
  # Expand data frame with empty values such that data.frame
  # has full days
  uniqueDates = unique(data$date)
  Ndays = length(uniqueDates)
  MaxSecInDay = max(data$SecInDay) # maximum number of seconds in a day
  SecSequence = seq(0, MaxSecInDay, by = epochsize) # Typical sequences of seconds in the day
  SecInDayTmp = rep(SecSequence, times = Ndays)
  DatesTmp = rep(uniqueDates, each = length(SecSequence))
  temp_df = data.frame(SecInDay = SecInDayTmp, date = DatesTmp, stringsAsFactors = FALSE)
  data = merge(x = data, y = temp_df, by.all = c("SecInDay", "date"), all = TRUE)
  
  # Calculations of Sleep Regularity Index per day pair
  NR = Ndays - 1
  if (NR > 1) {
    SleepRegularityIndex = data.frame(day = 1:NR, SleepRegularityIndex = numeric(NR),
                                      weekday = character(NR), frac_valid = numeric(NR), 
                                      date = character(NR), stringsAsFactors = FALSE)
    for (i in 1:NR) { # Loop over all days
      thisday = which(data$date == uniqueDates[i] & data$SecInDay < (24*3600))
      nextday = which(data$date == uniqueDates[i + 1] & data$SecInDay < (24*3600))
      ne25 = ((3600*25)/epochsize)
      ne24 = ((3600*24)/epochsize)
      ne23 = ((3600*23)/epochsize)
      if (length(thisday) == ne25) {
        thisday = thisday[1:ne24]
      }
      if (length(nextday) == ne25) {
        nextday = nextday[1:ne24]
      }
      if (length(thisday) == ne23) {
        nextday = nextday[1:ne23]
      }
      if (length(nextday) == ne23) {
        thisday = thisday[1:ne23]
      }
      EqualState = c(data$sleepstate[thisday] == data$sleepstate[nextday])
      testNA = c(is.na(data$sleepstate[thisday]) | is.na(data$sleepstate[nextday]))
      SummedValue = sum(ifelse(test = EqualState[which(testNA == FALSE)] == TRUE, yes = 1, no = 0))
      NValuesSkipped = length(which(testNA == TRUE))
      SleepRegularityIndex$frac_valid[i] = round((M - NValuesSkipped) / M, digits = 4)
      if (M != NValuesSkipped) {
        SleepRegularityIndex$SleepRegularityIndex[i] = round(-100 + (200/(M - NValuesSkipped)) * SummedValue, digits = 3)
      }
      SleepRegularityIndex$weekday[i] = weekdays(abbreviate = FALSE, x =  uniqueDates[i])
      SleepRegularityIndex$date[i] = format(as.Date(uniqueDates[i], 
                                                          origin = "1970-01-01"), format = "%d/%m/%Y")
    }
  } else {
    SleepRegularityIndex = NA
  }
  # SleepRegularityIndex is now a data.frame with SleepRegularityIndex per calendar date
  return(SleepRegularityIndex)
}

library(GGIR)
context("CalcSleepRegularityIndex")
test_that("CalcSleepRegularityIndex gives output as expected", {
  # Create dummy data:
  tz = "Europe/Amsterdam"
  time = seq(from = as.POSIXlt(x = "2020-01-01 00:01:10", tz = tz), 
             to = as.POSIXlt(x = "2020-01-06 00:01:10", tz = tz), by = 5)
  time = POSIXtime2iso8601(x = time, tz = tz)
  NID =  12*1440
  sleep = rep(0, length(time))
  R1 = round(NID*(4/10))
  R2 = round(NID*(5/10))
  R3 = round(NID*(8/10))
  R4 = round(NID*(9/10))
  sleep[R1:R3] = 1
  sleep[(NID + R2):(NID + R3)] = 1
  sleep[(NID*2 + R2):(NID*2 + R4)] = 1
  sleep[(NID*3 + R1):(NID*3 + R4)] = 1
  sleep[(NID*4 + R1):(NID*4 + R3)] = 1
  sleep[(NID*5 + R2):(NID*5 + R3)] = 1
  sleep = sleep[1:length(time)]
  testdata = data.frame(time = time, invalid = rep(0, length(time)), sleep = sleep)
  # Use dummy data in function call:
  SRI = CalcSleepRegularityIndex(data = testdata, epochsize = 5, desiredtz = tz)
  expect_equal(nrow(SRI), 5)
  expect_equal(ncol(SRI), 5)
  expect_equal(SRI$SleepRegularityIndex[2], 80)
  expect_equal(SRI$SleepRegularityIndex[3], 80)
  # Scenario 2, perfect regularity
  testdata2 = data.frame(time = time, sleep = rep(0,length(time)))
  SRI2 = CalcSleepRegularityIndex(data = testdata2, epochsize = 5, desiredtz = tz)
  expect_equal(SRI2$SleepRegularityIndex[3], 100)
  expect_equal(SRI2$SleepRegularityIndex[4], 100)
  # Scenario 3, perfect irregularity
  testdata3 = data.frame(time = time, sleep = 1:length(time))
  SRI3 = CalcSleepRegularityIndex(data = testdata3, epochsize = 5, desiredtz = tz)
  expect_equal(SRI3$SleepRegularityIndex[2], -100)
  expect_equal(SRI3$SleepRegularityIndex[3], -100)
})

detect_nonwear_clipping = function(data = c(), windowsizes = c(5, 900, 3600), sf = 100,
                                   clipthres = 7.5, sdcriter = 0.013, racriter = 0.05,
                                   nonwear_approach = "2013",
                                   params_rawdata = c()) {
  MediumEpochSize = windowsizes[2] * sf # Medium window size in samples
  LongEpochSize = windowsizes[3] * sf #Long window size in samples
  NMediumEpochs = floor(nrow(data)/(MediumEpochSize)) # Number of non-overlapping Medium Epochs that fit in data
  # Initialise matrices to collect clipping and nonwear info per axis
  ClipLog = NonwearLog = matrix(0,NMediumEpochs,3)
  ClipLogCollapsed = NonwearLogCollapsed = rep(0, NMediumEpochs) # averages
  # Define minimum number of Medium epochs that need to be present in Long Epoch
  minimumEpochCount = ((LongEpochSize/MediumEpochSize)/2) + 1
  
  if (nonwear_approach %in% c("2013", "2023")) {
    for (h in 1:NMediumEpochs) {
      # define start and end index for clipping detection (clipstart and clipend)
      clipstart = (((h - 1) * MediumEpochSize) + MediumEpochSize * 0.5 ) - MediumEpochSize * 0.5
      clipend = (((h - 1) * MediumEpochSize) + MediumEpochSize * 0.5 ) + MediumEpochSize * 0.5
      # define start and end index for nonwear detection (nwstart and nwend)
      if (nonwear_approach == "2013") {
        NonwearLogflag = h
        if (h <= minimumEpochCount) {
          nwstart = 1
          nwend = LongEpochSize
        } else if (h >= (NMediumEpochs - minimumEpochCount)) {
          nwstart = (NMediumEpochs - minimumEpochCount) * MediumEpochSize
          nwend = NMediumEpochs * MediumEpochSize
        } else if (h > minimumEpochCount & h < (NMediumEpochs - minimumEpochCount)) {
          nwstart = (((h - 1) * MediumEpochSize) + MediumEpochSize * 0.5 ) - LongEpochSize * 0.5
          nwend = (((h - 1) * MediumEpochSize) + MediumEpochSize * 0.5 ) + LongEpochSize * 0.5
        }
      } else if (nonwear_approach == "2023") {
        NonwearLogflag = h:(h + LongEpochSize/MediumEpochSize - 1)
        if (NonwearLogflag[length(NonwearLogflag)] > NMediumEpochs) NonwearLogflag = NonwearLogflag[-which(NonwearLogflag > NMediumEpochs)]
        # LongEpochSize to check (not aggregated values)
        nwstart = h * MediumEpochSize - MediumEpochSize
        nwend = nwstart + LongEpochSize
        if (nwend > nrow(data)) {
          nwend = nrow(data)
        }
      }
      # ---
      xyzCol = which(colnames(data) %in% c("x", "y", "z"))
      for (jj in seq(3)) {
        # Detect clipping
        aboveThreshold = which(abs(data[(1 + clipstart):clipend, xyzCol[jj]]) > clipthres)
        ClipLog[h, jj] = length(aboveThreshold)
        if (length(aboveThreshold) > 0) {
          if (length(which(abs(data[c((1 + clipstart):clipend)[aboveThreshold],  xyzCol[jj]]) > clipthres * 1.5)) > 0) {
            ClipLog[h, jj] = MediumEpochSize # If there is a a value that is more than 150% the dynamic range then ignore entire block.
          }
        }
        # Detect nonwear
        if (nonwear_approach == "2013") {
          indices = (1 + nwstart):nwend
        } else if (nonwear_approach == "2023") {
          indices = seq((1 + nwstart), nwend, by = ceiling(sf / 5))
        }
        maxwacc = max(data[indices, xyzCol[jj]], na.rm = TRUE)
        minwacc = min(data[indices, xyzCol[jj]], na.rm = TRUE)
        absrange = abs(maxwacc - minwacc)
        if (absrange < racriter) {
          sdwacc = sd(data[indices, xyzCol[jj]], na.rm = TRUE)
          if (sdwacc < sdcriter) {
            NonwearLog[NonwearLogflag,jj] = 1
          }
        }
      }
      # Summarise clipping
      ClipLog = ClipLog / (MediumEpochSize) # Express as fraction of epoch length
      ClipLogCollapsed[h] = max(c(ClipLog[h, 1], ClipLog[h, 2], ClipLog[h, 3])) #indicator of clipping
      
      # Summarise nonwear
      if ("wear" %in% colnames(data)) {
        wearTable = table(data[(1 + nwstart):nwend, "wear"], useNA = "no")
        NonwearLogCollapsed[h] = as.numeric(tail(names(sort(wearTable)), 1)) * 3 # times 3 to simulate heuristic approach
      }
      if (!("wear" %in% colnames(data))) {
        NonwearLogCollapsed[h] = (NonwearLog[h,1] + NonwearLog[h,2] + NonwearLog[h,3]) #indicator of non-wear
      }
      
    }
  }
  # In NonwearLogCollapsed: single 1's surrounded by 2's or 3's --> 2 (so it is considered nonwear)
  ones = which(NonwearLogCollapsed == 1)
  if (length(ones) > 0) {
    for (one_i in ones) {
      if (one_i - 1 < 1) next # skip if we are in the very beginning
      if (one_i + 1 > length(NonwearLogCollapsed)) next # skip if we are in the very end
      if (NonwearLogCollapsed[one_i - 1] > 1 & NonwearLogCollapsed[one_i + 1] > 1) NonwearLogCollapsed[one_i] = 2
    }
  }
  return(list(NWav = NonwearLogCollapsed, CWav = ClipLogCollapsed, nmin = NMediumEpochs))
}

g.getstarttime = function(datafile, data, mon, dformat, desiredtz, configtz = NULL) {
  starttime = NULL
  if (is.null(configtz)) {
    configtz = desiredtz
  }
  if ("time" %in% colnames(data)) {
    starttime = as.POSIXlt(data[1, "time"], tz = desiredtz, origin = "1970-01-01")
  } else if (mon == MONITOR$MOVISENS) {
      starttime = unisensR::readUnisensStartTime(dirname(datafile))

      # movisens timestamp is stored in unisens.xml file as a string "%Y-%m-%dT%H:%M:%S",
      # without a timezone.
      # unisensR::readUnisensStartTime() converts this string into a POSIXct object,
      # but since no timezone information was stored, this conversion happens using
      # the system timezone.
      # If configtz is different from the system timezone (""), we need to force the
      # correct timezone (force, while keeping the same hh:mm:ss time, not just convert to congigtz)
      if (configtz != "") {
        starttime = lubridate::force_tz(starttime, configtz)
      }
      starttime = as.POSIXlt(starttime, tz = desiredtz)

  } else if (dformat == FORMAT$CSV && (mon == MONITOR$ACTIGRAPH || mon == MONITOR$VERISENSE)) {
    starttime = startdate = NULL
    header_rows = 8
    tmph = read.csv(datafile, nrow = header_rows, skip = 1)

    tmphi = 1
    while (tmphi < header_rows) {
      tmp = unlist(strsplit(format(tmph[tmphi,1]),"Start Time"))
      if (length(tmp) > 1) {
        starttime = tmp[2]
        break
      }
      tmphi = tmphi + 1
    }
    if (is.null(starttime)) {
      stop(paste0("Start Time not found in the header of ", datafile))
    }
    #-------------------------------
    tmphi = 1
    while (tmphi < header_rows) {
      tmp = unlist(strsplit(format(tmph[tmphi,1]),"Start Date"))
      if (length(tmp) > 1) {
        startdate = tmp[2]
        break
      }
      tmphi = tmphi + 1
    }
    if (is.null(startdate)) {
      stop(paste0("Start Date not found in the header of ", datafile))
    }

    #-----------------------------------------
    # trim any spaces
    startdate = gsub(" ", "", startdate, fixed = TRUE)
    starttime = gsub(" ", "", starttime, fixed = TRUE)

    #-----------------------------------------
    # flexible four date/time formats
    starttime = paste0(startdate," ",starttime)
    getOption("digits.secs")
    options(digits.secs = 3)

    options(warn = -1)
    topline = as.matrix(colnames(as.matrix(read.csv(datafile, nrow = 1, skip = 0))))
    topline = topline[1]  #To avoid dots
    options(warn = 0)
    # Extraction of date format.
    # all formats to consider following R date formatting symbols:
    # Y year including century, y year excluding century, b month as character, m month as a number
    # Note: year in the middle of a date is assumed to not occur.
    allformats = c("mdY", "mdy","bdY", "bdy",
                   "dmY", "dmy", "dbY", "dby",
                   "Ymd", "ymd", "Ybd", "ybd",
                   "Ydm", "ydm", "Ydb", "ydb")
    fc = data.frame(matrix(0,1,length(allformats))) # data.frame to keep track of formats that have been checked
    names(fc) = allformats
    # Now systematically go through all possible formats to identify which one it is
    fc$mdY = length(grep("MM[.]dd[.]yyyy|M[.]d[.]yyyy|M[.]dd[.]yyyy|MM[.]d[.]yyyy",topline))
    fc$mdy = length(grep("MM[.]dd[.]yy|M[.]d[.]yy|M[.]dd[.]yy|MM[.]d[.]yy",topline))
    fc$bdY = length(grep("MMM[.]dd[.]yyyy|MMM[.]d[.]yyyy",topline))
    fc$bdy = length(grep("MMM[.]dd[.]yy|MMM[.]d[.]yy",topline))
    if (fc$mdY == 1 & fc$mdy == 1) fc$mdy = 0 # not yy when yyyy is also detected
    if (fc$bdY == 1 & fc$bdy == 1) fc$bdy = 0 # not yy when yyyy is also detected
    if (fc$mdY == 1 & fc$bdY == 1) fc$mdY = 0 # not M(M) when MMM is also detected for yyyy
    if (fc$mdy == 1 & fc$bdy == 1) fc$mdy = 0 # not M(M) when MMM is also detected for yy
    fc$dmY = length(grep("d[.]M[.]yyyy|d[.]MM[.]yyyy",topline))
    fc$dmy = length(grep("d[.]M[.]yy|d[.]MM[.]yy",topline))
    fc$dbY = length(grep("d[.]MMM[.]yyyy",topline))
    fc$dby = length(grep("d[.]MMM[.]yy",topline))
    if (fc$dmY == 1 & fc$dmy == 1) fc$dmy = 0 # not yy when yyyy is also detected
    if (fc$dbY == 1 & fc$dby == 1) fc$dby = 0 # not yy when yyyy is also detected
    if (fc$dmY == 1 & fc$dbY == 1) fc$dmY = 0 # not M(M) when MMM is also detected for yyyy
    if (fc$dmy == 1 & fc$dby == 1) fc$dmy = 0 # not M(M) when MMM is also detected for yy
    fc$Ymd = length(grep("yyyy[.]M[.]d|yyyy[.]MM[.]d",topline))
    fc$ymd = length(grep("yy[.]M[.]d|yy[.]MM[.]d",topline))
    fc$Ybd = length(grep("yyyy[.]MMM[.]d",topline))
    fc$ybd = length(grep("yy[.]MMM[.]d",topline))
    if (fc$Ymd == 1 & fc$ymd == 1) fc$ymd = 0 # not yy when yyyy is also detected
    if (fc$Ybd == 1 & fc$ybd == 1) fc$ybd = 0 # not yy when yyyy is also detected
    if (fc$Ymd == 1 & fc$Ybd == 1) fc$Ymd = 0 # not M(M) when MMM is also detected for yyyy
    if (fc$ymd == 1 & fc$ybd == 1) fc$ymd = 0 # not M(M) when MMM is also detected for yy
    fc$Ydm = length(grep("yyyy[.]dd[.]MM|yyyy[.]d[.]M|yyyy[.]d[.]MM|yyyy[.]dd[.]M",topline))
    fc$ydm = length(grep("yy[.]dd[.]MM|yy[.]d[.]M|yy[.]d[.]MM|yy[.]dd[.]M",topline))
    fc$Ydb = length(grep("yyyy[.]dd[.]MMM|yyyy[.]d[.]MMM",topline))
    fc$ydb = length(grep("yy[.]dd[.]MMM|yy[.]d[.]MMM",topline))
    if (fc$Ydm == 1 & fc$ydm == 1) fc$ydm = 0 # not yy when yyyy is also detected
    if (fc$Ydb == 1 & fc$ydb == 1) fc$ydb = 0 # not yy when yyyy is also detected
    if (fc$Ydm == 1 & fc$Ydb == 1) fc$Ydm = 0 # not M(M) when MMM is also detected for yyyy
    if (fc$ydm == 1 & fc$ydb == 1) fc$ydm = 0 # not M(M) when MMM is also detected for yy
    # Now we can identify which format it is:
    theformat = names(fc)[which(fc == 1)[1]]
    if (is.na(theformat)) warning("date format not recognised")
    splitformat = unlist(strsplit(theformat,""))
    # identify separater:
    if (length(grep("/",starttime)) > 0) {
      sepa = "/"
    } else {
      if (length(grep("-",starttime)) > 0) {
        sepa = "-"
      } else {
        if (length(grep(".",starttime)) > 0) {
          sepa = "."
        } else {
          warning("separator character for dates not identified")
        }
      }
    }
    expectedformat = paste0('%',splitformat[1],sepa,'%',splitformat[2],sepa,'%',splitformat[3],' %H:%M:%S')
    starttime = as.POSIXct(starttime, format = expectedformat, tz = configtz)
    starttime = as.POSIXlt(starttime, tz = desiredtz)
  } else {
    stop(paste0("Timestamps not found for monitor type ", mon, " and file format type ", dformat,
                "\nThis should not happen."))
  }

  return(starttime)
}

g.getbout = function(x, boutduration, boutcriter = 0.8, ws3 = 5) {
  p = which(x == 1)
  x[is.na(x)] = 0 # ignore NA values in the unlikely event that there are any
  xt = x
  #look for breaks larger than 1 minute
  # Note: we do + 1 to make sure we look for breaks larger than but not equal to a minute,
  # this is critical when working with 1 minute epoch data
  # we need to append 0's at the beginning and end so that first and last epochs in timeseries do not get ignored
  zeroes = rep(0, ceiling(60/ws3/2))
  xtmp = c(zeroes, x, zeroes)
  lookforbreaks = zoo::rollmean(x = xtmp, k = (60/ws3) + 1, align = "center", fill = rep(0,3))
  # remove appended zeroes
  keep = (length(zeroes) + 1):(length(lookforbreaks) - length(zeroes))
  #insert negative numbers to prevent these minutes to be counted in bouts
  #in this way there will not be bouts breaks lasting longer than 1 minute
  xtmp[lookforbreaks == 0] = -boutduration
  xt = xtmp[keep]
  # append 0's to define boutcriter
  append = rep(-boutduration, ceiling(boutduration/2))
  xtmp = c(append, xt, append)
  RM = zoo::rollmean(x = xtmp, k = boutduration, align = "center", fill = rep(0,3))
  keep = (length(append) + 1):(length(RM) - length(append))
  RM = RM[keep]
  p = which(RM >= boutcriter)
  half1 = floor(boutduration/2)
  half2 = boutduration - half1
  
  # now mark all epochs that are covered by the detected bout/s
  detected_bouts = split(p, cumsum(c(1, diff(p) != 1)))
  if (length(detected_bouts) == 1 & length(detected_bouts[[1]]) == 0) {
    # if no bouts are detected
    x[which(xt != 2)] = 0
    x[which(xt == 2)] = 1
    boutcount = x # distinction not made anymore, but object kept to preserve output structure
  } else {
    # if bouts are detected
    for (bout_i in 1:length(detected_bouts)) {
      bout = detected_bouts[[bout_i]]
      # find start of bout
      start_found = FALSE
      adjust = 0
      while (start_found == FALSE) {
        start = min(bout) - half1 + adjust
        if (start < 1) {
          adjust = adjust + 1
          next
        }
        if (xt[start] != 1) {
          # a bout cannot start without meetin threshold crit
          adjust = adjust + 1
          next
        }
        start_period = start:max(bout)
        look4start = split(xt[start_period], cumsum(c(1, diff(xt[start_period]) != 0)))
        max_gap = 60/ws3
        zeros = c()
        for (i in 1:length(look4start)) {
          zeros = c(zeros, sum(look4start[[i]] == 0))
        }
        if (all(zeros <= max_gap)) start_found = TRUE
        adjust = adjust + 1
      }
      # find end of bout
      end_found = FALSE
      adjust = 0
      while (end_found == FALSE) {
        end = max(bout) + half1 - adjust
        if (end > length(xt)) {
          adjust = adjust + 1
          next
        }
        if (xt[end] != 1) {
          # a bout cannot start without meetin threshold crit
          adjust = adjust + 1
          next
        }
        end_period = min(bout):end
        look4end = split(xt[end_period], cumsum(c(1, diff(xt[end_period]) != 0)))
        max_gap = 60/ws3
        zeros = c()
        for (i in 1:length(look4start)) {
          zeros = c(zeros, sum(look4start[[i]] == 0))
        }
        if (all(zeros <= max_gap)) end_found = TRUE
        adjust = adjust + 1
      }
      xt[start:end] = 2
    }
    x[which(xt != 2)] = 0
    x[which(xt == 2)] = 1
  }
  return(x)
}

library(GGIR)
context("ggetbout")
test_that("g.getbout produces expected output", {
  
  for (ws3 in c(5, 15, 30, 60)) { # epoch size should not influence the result, so we can turn it into a loop
    cat(paste0("\nepochsize ", ws3))
    boutcriter = 1
    N = 60 / ws3 # N is number of epochs in minute
    boutduration = 10 * N
    # Metric 7 - 2 bouts out of which 1 is long enough
    xtest = c(rep(0, 15 * N), rep(1, 6 * N), rep(0, 1 * N), rep(1, 10 * N), rep(0, 20 * N))
    expected_bout_start = ((15 + 6 + 1) * N) + 1
    expected_bout_end = ((15 + 6 + 1 + 10) * N)
    bouts1 = g.getbout(x = xtest, boutduration = boutduration, boutcriter = boutcriter, ws3 = ws3)
    expect_equal(sum(bouts1), 10 * N)
    expect_equal(bouts1, c(rep(0, 22 * N), rep(1, 10 * N), rep(0, 20 * N)))
    expect_equal(bouts1, c(rep(0, 22 * N), rep(1, 10 * N), rep(0, 20 * N)))
    expect_equal(min(which(bouts1 == 1)), expected_bout_start)
    expect_equal(max(which(bouts1 == 1)), expected_bout_end)
    
    # Metric 7 - ignore gaps when boutcriter is 1
    xtest = x = c(rep(1, 4 * N), rep(0, 3 * N), rep(1, 6 * N),rep(0, 3 * N), rep(1, 10 * N), rep(0, 1 * N) , rep(1, 9 * N))
    expected_bout_start = ((4 + 3 + 6 + 3) * N) + 1
    expected_bout_end = ((4 + 3 + 6 + 3 + 10) * N)
    bouts2 = g.getbout(x = xtest, boutduration = boutduration, boutcriter = boutcriter, ws3 = ws3)
    expect_equal(sum(bouts2), 10 * N)
    expect_equal(sum(bouts2), 10 * N)
    expect_equal(min(which(bouts2 == 1)), expected_bout_start)
    expect_equal(max(which(bouts2 == 1)), expected_bout_end)
    
    # Metric 7 - ability to detect odd length bouts
    boutduration = 9 * N
    xtest = x = c(rep(0, 20 * N), rep(1, 9 * N), rep(0, 20 * N))
    expected_bout_start = (20 * N) + 1
    expected_bout_end = ((20 + 9) * N)
    bouts3 = g.getbout(x = xtest, boutduration = boutduration, boutcriter = boutcriter, ws3 = ws3)
    expect_equal(sum(bouts3), 9 * N)
    expect_equal(min(which(bouts3 == 1)), expected_bout_start)
    expect_equal(max(which(bouts3 == 1)), expected_bout_end)
    
    # Metric 7 - ability to detect bouts with grace periods
    xtest = c(rep(0, 15 * N), rep(1,6 * N), rep(0, 1 * N), rep(1,10 * N), rep(0,20 * N))
    expected_bout_start = (15 * N) + 1
    expected_bout_end = ((15 + 6 + 1 + 10) * N)
    boutcriter = 0.9
    boutduration = 10 * N
    bouts4 = g.getbout(x = xtest, boutduration = boutduration, boutcriter = boutcriter, ws3 = ws3)
    expect_equal(sum(bouts4), 17 * N)
    expect_equal(min(which(bouts4 == 1)), expected_bout_start)
    expect_equal(max(which(bouts4 == 1)), expected_bout_end)
    
    # Metric 7 - ability to detect bouts with grace periods, but also ignore too many grace periods
    xtest = c(rep(0, 15 * N), rep(1, 6 * N), rep(0, 1 * N), rep(1,3 * N), rep(0, 1 * N), rep(1, 4 * N), rep(0, 20 * N))
    expected_bout_start = (15 * N) + 1
    expected_bout_end = ((15 + 6 + 1 + 3) * N)
    bouts5 = g.getbout(x = xtest, boutduration = boutduration, boutcriter = boutcriter, ws3 = ws3)
    expect_equal(sum(bouts5), 10 * N)
    expect_equal(min(which(bouts5 == 1)), expected_bout_start)
    expect_equal(max(which(bouts5 == 1)), expected_bout_end)
    
    # Metric 7 - bout length equals epoch length
    boutduration = 1
    xtest = c(rep(0, 10 * N), 1, rep(0, 10 * N))
    bouts6 = g.getbout(x = xtest, boutduration = boutduration, boutcriter = boutcriter, ws3 = ws3)
    expect_equal(sum(bouts6), 1)
    expect_equal(bouts6, xtest)
    expect_equal(bouts6, xtest)
  }
  
  
  
  
})

g.dotorcomma = function(inputfile, dformat, mon, ...) {
  #get input variables (relevant when read.myacc.csv is used)
  input = list(...)
  decn = getOption("OutDec") # extract system decimal separator
  if (length(decn) == 0) decn = "." # assume . if not retrieved
  if (length(input) > 0) {
    for (i in 1:length(names(input))) {
      txt = paste0(names(input)[i], "=", input[i])
      if (is(unlist(input[i]), "character")) {
        txt = paste0(names(input)[i], "='", unlist(input[i]), "'")
      }
      eval(parse(text = txt))
    }
  }
  #------------------------------------------------------------
  if (exists("rmc.firstrow.acc") == FALSE) rmc.firstrow.acc = c()
  if (exists("rmc.dec") == FALSE) rmc.dec = decn
  if (length(rmc.firstrow.acc) == 1) {
    dformat = FORMAT$AD_HOC_CSV
    mon = MONITOR$AD_HOC
    decn = rmc.dec
  }
  if (dformat == FORMAT$CSV) {
    skiprows = 100
    # Note: I have added the below lines because some ActiGraph files start with a
    # lot of zeros, which makes it impossible to detect decimal separator
    # "." will then be the default, which is not correct for "," systems.
    while (skiprows < 1000000) { #foundnonzero == FALSE &
      tmp = try(expr = {as.matrix(read.csv(inputfile, skip = skiprows, nrow = 10))}, silent = TRUE)
      if (inherits(tmp, "try-error")) break  # nothing left in the file to read
      deci = tmp

      skiprows = skiprows + 10000
      if (length(unlist(strsplit(as.character(deci[2,2]), ","))) > 1) {
        decn = ","
        break
      }
      numtemp = as.numeric(deci[2,2])
      if (is.na(numtemp) == FALSE && numtemp != 0) break
    }
    if (!exists("deci")) stop("Problem with reading .csv file in GGIR function dotorcomma")
    if (is.na(suppressWarnings(as.numeric(deci[2,2]))) == T & decn == ".") decn = ","
  } else if (dformat == FORMAT$BIN) {
    if (mon == MONITOR$GENEACTIV ) {
      try(expr = {deci = GGIRread::readGENEActiv(filename = inputfile,
                                                 start = 1, end = 3)}, silent = TRUE)
      if (!exists("deci")) stop("Problem with reading .bin file in GGIR function dotorcomma")
      if (is.na(as.numeric(deci$data.out[2, 2])) == T & decn == ".") decn = ","
    } else if (mon == MONITOR$PARMAY_MTX) {
      decn = "." 
    }
  } else if (dformat == FORMAT$CWA) {
    try(expr = {deci = GGIRread::readAxivity(filename = inputfile, start = 1, end = 10,
                                             interpolationType = 1)$data}, silent = TRUE)
    if (!exists("deci")) stop("Problem with reading .cwa file in GGIR function dotorcomma")
    if (is.na(suppressWarnings(as.numeric(deci[2,2]))) == T & decn == ".") decn = ","
  } else if (dformat == FORMAT$GT3X) {
    if (length(grep(pattern = "[.]GT", x = inputfile)) > 0 & file.exists(inputfile) == FALSE) {
      inputfile = gsub(pattern = "[.]GT3X", replacement = "[.]gt3x", x = inputfile)
    }
    try(expr = {deci = as.data.frame(read.gt3x::read.gt3x(path = inputfile,
                                                       batch_begin = 1, batch_end = 10,
                                                       asDataFrame = TRUE))}, silent = TRUE)
    if (!exists("deci")) stop("Problem with reading .gt3x file in GGIR function dotorcomma")
    if (is.na(suppressWarnings(as.numeric(deci[2,2]))) == T & decn == ".") decn = ","
  }
  dotorcomma = decn
}

ismovisens = function(data){
        # ---------------------------
        # Is data a directory? Then use the first file in this directory 
        # to test if recorded with movisens
        first_file = dir(data, recursive = T, full.names = T)[1]
        isdir = !is.na(first_file)
        if (isdir) data = first_file
        # ---------------------------
        # At this point, data is guaranteed to be a path to a file.
        # Now, focus on the directory containing this file.
        data_tmp = strsplit(data, "/")

        data_ln = length(data_tmp[[1]]) - 1
        data = paste(data_tmp[[1]][1:data_ln], collapse = "/")
        # ---------------------------
        unisensXML = paste(data, "unisens.xml", sep = "/")
        is.mv = file.exists(unisensXML)
        return(is.mv)
}

markerButtonForRest = function(sibreport, params_sleep, ts) {
  #-----------------------------------------------------------
  # This function looks for marker buttons near the edges of sustained
  # inactivity bouts (sib) to guide the nap detection.
  # For the time being marker buttons are only extracted from Actiwatch-like
  # devices (Philips Health Band, MotionWatch).
  
  # nap_markerbutton_method is explained in the documentation (see GGIR.Rd)
  
  # Explanation objects as used below
  # nap_copy_timing_mb:
  # Boolean to decide whether to rely on marker for final timing of the nap
  # if TRUE, only consider marker buttons  before midtime of sib for start
  # and only consider marker buttons after midtime of sib for end
  # if FALSE, then consider any mark button, this uses the marker solely as 
  # a confirmation that nap nearby
  
  # nap_require_mb:
  # Boolean to decide whether to require availability of marker to consider 
  # detection of a nap
  # if TRUE, only keep sibs that had marker button nearby
  # if FALSE, also keep sibs that may not have marker button but meet 
  # criteria for another reason
  # 
  sibreport$ignore = FALSE # if no marker button than keep sibs
  if  (params_sleep[["nap_markerbutton_method"]] > 0) {
    if (params_sleep[["nap_markerbutton_method"]] > 1) {
      sibreport$ignore = TRUE # if no marker button then ignore all sibs
    }
    # maximum gap allowed between sib edge and marker
    nap_mb_max_dist = params_sleep[["nap_markerbutton_max_distance"]]
    
    # If marker button data available
    if ("marker" %in% colnames(ts)) {
      if (1 %in% ts$marker) {
        if (params_sleep[["nap_markerbutton_method"]] == 1) {
          nap_require_mb = FALSE
          nap_copy_timing_mb = TRUE
        } else if (params_sleep[["nap_markerbutton_method"]] == 2) {
          nap_require_mb = TRUE
          nap_copy_timing_mb = FALSE
        } else if (params_sleep[["nap_markerbutton_method"]] == 3) {
          nap_require_mb = TRUE
          nap_copy_timing_mb = TRUE
        }
        # Calculate time in minutes to nearest button press
        # that starts before and after midway the sib, for start and end respectively
        sibreport$midtime = sibreport$start + (sibreport$end - sibreport$start) / 2
        marker_times = ts$time[which(ts$marker == 1)]
        sibreport$start_to_marker = NA
        sibreport$end_to_marker = NA
        
        for (hi in 1:nrow(sibreport)) {
          # assess duration marker buttons relative to sib midtime
          delta_times = as.numeric(sibreport$midtime[hi]) - as.numeric(marker_times)
          if (nap_copy_timing_mb == TRUE) {
            # Plan is to use marker button timing, so only consider
            # for start the marker buttons that start before midtime
            end_delta_times = delta_times[which(delta_times < 0)]
            # Plan is to use marker button timing, so only consider
            # for end the marker buttons that start after midtime
            start_delta_times = delta_times[which(delta_times > 0)]
          } else {
            # Plan is NOT to use marker button timing, so any nearby
            # marker button is sufficient
            start_delta_times = end_delta_times = delta_times
          }
          nap_start_found = nap_end_found = FALSE
          half_sib_dur = sibreport$duration[hi] / 2
          tmp_start = tmp_end = NULL
          # find marker for start nap
          if (length(start_delta_times) > 0) {
            sibreport$start_to_marker[hi] = min(abs(start_delta_times)) / 60 - half_sib_dur
            if (any(sibreport$start_to_marker[hi] < nap_mb_max_dist)) {
              # marker button nearby
              if (nap_copy_timing_mb == TRUE) {
                # use nearby marker button to overwrite sib start
                tmp_start = sibreport$midtime[hi] - start_delta_times[which.min(abs(start_delta_times))]
              }
              nap_start_found = TRUE
            }
          }
          # find marker for end nap
          if (length(end_delta_times) > 0) {
            sibreport$end_to_marker[hi] = min(abs(end_delta_times)) / 60 - half_sib_dur
            if (any(sibreport$end_to_marker[hi] < nap_mb_max_dist)) {
              # marker button nearby
              if (nap_copy_timing_mb == TRUE) {
                # use nearby marker button to overwrite sib end
                tmp_end = sibreport$midtime[hi] + abs(end_delta_times[which.min(abs(end_delta_times))])
              }
              nap_end_found = TRUE
            }
          }
          if (nap_require_mb == TRUE && (nap_start_found == FALSE || nap_end_found == FALSE)) { 
            # if nearby marker button is required but none was available
            # ignore this sib
            sibreport$ignore[hi] = TRUE
          } else {
            # enable nap detection further down becaue marker button was found
            sibreport$ignore[hi] = FALSE
            # overwrite all other sibs that overlap such that they form duplicated rows
            if (length(tmp_start) > 0) {
              sibreport$start[hi] = tmp_start
              sibreport$start[which(sibreport$start < sibreport$midtime[hi] & sibreport$start >= tmp_start)] = sibreport$start[hi]
            }
            if (length(tmp_end) > 0) {
              sibreport$end[hi] = tmp_end
              sibreport$end[which(sibreport$end < sibreport$midtime[hi] & sibreport$end >= tmp_end)] = sibreport$end[hi]
            }
            # recalculate midtime
            sibreport$midtime = sibreport$start + (sibreport$end - sibreport$start) / 2
          }
        }
      }
    }
  }
  # remove duplicate rows
  sibreport = sibreport[!duplicated(sibreport[, c("start", "end")]), ]
  return(sibreport)
}

g.part5.definedays = function(nightsi, wi, indjump, epochSize, qqq_backup = c(), ts, timewindowi, 
                              Nwindows, qwindow, ID = NULL,
                              dayborder = 0) {
  Nts = nrow(ts)
  lastDay = FALSE
  # define local functions ----
  qwindow2timestamp = function(qwindow, epochSize) {
    hour = floor(qwindow)
    minute = floor((qwindow - hour) * 60)
    second = floor((qwindow - hour - minute/60) * 60 * 60)
    expected_seconds = seq(0, 60, by = epochSize)
    seconds_tobe_revised = which(!second %in% expected_seconds)
    if (length(seconds_tobe_revised) > 0) {
      for (si in seconds_tobe_revised) {
        second[si] = expected_seconds[which.min(abs(expected_seconds - second[si]))]
        if (second[si] == 60) { # shift minute if second == 60
          minute[si] = minute[si] + 1
          second[si] = 0
        }
      }
    }
    hour = as.character(hour); minute = as.character(minute); second = as.character(second)
    hour = ifelse(nchar(hour) == 1, paste0("0", hour), hour)
    minute = ifelse(nchar(minute) == 1, paste0("0", minute), minute)
    second = ifelse(nchar(second) == 1, paste0("0", second), second)
    HMS = paste(hour, minute, second, sep = ":")
    if (HMS[1] != "00:00:00") HMS = c("00:00:00", HMS)
    return(HMS)
  }
  subtractEpochFromTimeName = function(x, epochSize) {
    hms = strptime(x, format = "%H:%M:%S")
    hms = hms - epochSize
    hms = format(hms, format = "%H:%M:%S")
    return(hms)
  }
  # main script -----
  # Check that this is a meaningful day (that is all the qqq variable is used for),
  # before storing it.
  qqq = rep(0,2); segments = segments_names = c()
  # Check that it is possible to find both windows (WW and MM)
  # in the data for this day.
  if (timewindowi == "MM") {
    # include first and last partial days in MM
    if (nightsi[1] > 1 && nightsi[1] < 25 * 3600 / epochSize) {
      nightsi = c(1, nightsi)
    }
    if (nightsi[length(nightsi)] < nrow(ts) && 
        nrow(ts) - nightsi[length(nightsi)] < 25 * 3600 / epochSize) {
      nightsi = c(nightsi, nrow(ts))
    }
    # define window
    qqq[1] = nightsi[wi]
    if (length(nightsi) >= wi + 1) {
      qqq[2] = nightsi[wi + 1] - 1
    } else {
      qqq[2] = Nts
      lastDay = TRUE
    }
    # is this the last day?
    if (wi == length(nightsi) - 1) {
      lastDay = TRUE
    }
    if (qqq[2] >= Nts - 1) {
      qqq[2] = Nts
      lastDay = TRUE
    }
    qqq_backup = qqq
    # in MM, also define segments of the day based on qwindow
    if (!is.na(qqq[1]) & !is.na(qqq[2])) {
      segments_timing = NULL
      if (qqq[2] > Nts) qqq[2] = Nts
      fullQqq = qqq[1]:qqq[2]
      firstepoch = format(ts$time[qqq[1]],  "%H:%M:%S")
      lastepoch = format(ts$time[qqq[2]],  "%H:%M:%S")
      qnames = NULL
      if (is.data.frame(qwindow)) {
        date_of_interest = substr(ts$time[qqq[1]], 1, 10)
        qdate = which(qwindow$ID == ID & qwindow$date == date_of_interest)
        if (length(qdate) == 1) { # if ID/date matched with activity log
          qnames = unlist(qwindow$qwindow_names[qdate])
          qwindow = unlist(qwindow$qwindow_values[qdate])
          qwindow_order = order(qwindow)
          qwindow = qwindow[qwindow_order]
          qnames = qnames[qwindow_order]
        } else { # if ID/date not correctly matched with activity log
          qwindow = c(0, 24)
        }
      } else {
        qwindow = sort(qwindow)
        if (qwindow[1] != 0) qwindow = c(0, qwindow)
        if (qwindow[length(qwindow)] != 24) qwindow = c(qwindow, 24)
      }
      # define segments timing in H:M:S format
      breaks = qwindow2timestamp(qwindow, epochSize)
      startOfSegments = breaks[-length(breaks)]
      endOfSegments = subtractEpochFromTimeName(breaks[-1], epochSize)
      if (length(startOfSegments) > 1) { # when qwindow segments are defined, add fullwindow at the beginning
        startOfSegments = c(firstepoch, startOfSegments)
        endOfSegments = c(lastepoch, endOfSegments)
      } 
      segments_timing = paste(startOfSegments, endOfSegments, sep = "-")
      # define segment names based on qnames or segmentX
      if (is.null(qnames)) {
        segments_names = paste0("segment", 0:(length(segments_timing) - 1))
        segments_names = gsub("segment0", "MM", segments_names)
      } else {
        segments_names = c("MM", paste(qnames[-length(qnames)], qnames[-1], sep = "-"))
      }
      # Get indices in ts for segments start and end limits
      hms = format(ts$time[fullQqq], format = "%H:%M:%S")
      segments = vector("list", length = length(segments_timing))
      names(segments) = segments_timing
      for (si in 1:length(segments_timing)) {
        s0s1 = unlist(strsplit(segments_timing[si], split = "[-]"))
        s0s1 = format(s0s1, format = "%H:%M:%S")
        # tryCatch is needed in the case that the segment is not available in ts,
        # then a no non-missing values warning would be triggered by the which function
        segments[[si]] = tryCatch(range(fullQqq[which(hms >= s0s1[1] & hms <= s0s1[2])]), #segStart and segEnd
                                  warning = function(w) rep(NA, 2))
      }
    }
  } else if (timewindowi == "WW" || timewindowi == "OO") {
    windowEdge = ifelse(timewindowi == "WW", yes = -1, no = 1)
    if (wi <= (Nwindows - 1)) { # all full windows
      qqq[1] = which(diff(ts$diur) == windowEdge)[wi] + 1
      qqq[2] = which(diff(ts$diur) == windowEdge)[wi + 1]
    } else {
      # time after last reliable waking up or onset (this can be more than 24 hours)
      # ignore this window, because if the night was ignored for sleep analysis
      # then the description of the day in part 5 including that night is
      # not informative.
      qqq = c(NA, NA)
    }
    # build up segments
    if (!is.na(qqq[1]) & !is.na(qqq[2])) {
      segments = list(qqq)
      start = format(ts$time[qqq[1]], "%H:%M:%S")
      end = format(ts$time[qqq[2]], "%H:%M:%S")
      names(segments) = paste(start, end, sep = "-")
      segments_names = timewindowi
    }
    if (wi >= Nwindows) lastDay = TRUE
  }
  return(invisible(list(qqq = qqq, qqq_backup = qqq_backup, lastDay = lastDay,
                        segments = segments, segments_names = segments_names)))
}

library(GGIR)
context("g.part5.definedays")
test_that("g.part5.definedays considers qwindow to generate segments", {
  skip_on_cran()
  
  # set windows
  timestamp = seq.POSIXt(from = as.POSIXct("2022-01-01 00:00:00"), 
                         to = as.POSIXct("2022-01-04 23:59:00"), by = 60)
  ts = data.frame(timestamp = timestamp,
                  ACC = 0)
  nightsi = grep("00:00:00", format(ts$timestamp))
  qwindow = c(7, 18)
  
  definedays = g.part5.definedays(nightsi = nightsi, wi = 1, indjump = 1, 
                                  epochSize = 60, qqq_backup = c(), 
                                  ts = ts, 
                                  timewindowi = "MM", Nwindows = 1, 
                                  qwindow = qwindow, ID = NULL)
  
  expect_false(is.null(definedays$segments))
  expect_true(is.list(definedays$segments))
  expect_equal(length(definedays$segments), 4)
  expect_equal(names(definedays$segments)[1], "00:00:00-23:59:00")
  expect_equal(definedays$segments[[1]], c(1, 1440))
  expect_equal(names(definedays$segments)[2], "00:00:00-06:59:00")
  expect_equal(definedays$segments[[2]], c(1, 420))
  expect_equal(names(definedays$segments)[3], "07:00:00-17:59:00")
  expect_equal(definedays$segments[[3]], c(421, 1080))
  expect_equal(names(definedays$segments)[4], "18:00:00-23:59:00")
  expect_equal(definedays$segments[[4]], c(1081, 1440))
  expect_equal(definedays$segments_names, c("MM", "segment1", "segment2", "segment3"))
  
  # With slash as separator
  actlog = data.frame(id = c("1RAW"),
                      date = c("01/01/2022"),
                      work = c("7:45:00"),
                      travelhome = c("17:00:00"),
                      home = c("17:30:00"))
  fn = "testactlog.csv"
  write.csv(x = actlog, file = fn, row.names = FALSE)
  LOG = g.conv.actlog(qwindow = fn, qwindow_dateformat = "%d/%m/%Y")
  
  # definedays
  definedays = g.part5.definedays(nightsi = nightsi, wi = 1, indjump = 1, 
                                  epochSize = 60, qqq_backup = c(), 
                                  ts = ts, timewindowi = "MM", Nwindows = 1, 
                                  qwindow = LOG, ID = "1RAW")
  
  expect_false(is.null(definedays$segments))
  expect_true(is.list(definedays$segments))
  expect_equal(length(definedays$segments), 5)
  expect_equal(names(definedays$segments)[1], "00:00:00-23:59:00")
  expect_equal(definedays$segments[[1]], c(1, 1440))
  expect_equal(names(definedays$segments)[2], "00:00:00-07:44:00")
  expect_equal(definedays$segments[[2]], c(1, 465))
  expect_equal(names(definedays$segments)[3], "07:45:00-16:59:00")
  expect_equal(definedays$segments[[3]], c(466, 1020))
  expect_equal(names(definedays$segments)[4], "17:00:00-17:29:00")
  expect_equal(definedays$segments[[4]], c(1021, 1050))
  expect_equal(names(definedays$segments)[5], "17:30:00-23:59:00")
  expect_equal(definedays$segments[[5]], c(1051, 1440))
  expect_equal(definedays$segments_names, c("MM", "daystart-work", 
                                            "work-travelhome", 
                                            "travelhome-home", "home-dayend"))
  
  # With times not multiple of epoch size and date format %Y-%m-%d
  actlog = data.frame(id = c("1RAW"),
                      date = c("2022-01-01"),
                      work = c("08:09:59"),
                      travelhome = c("16:30:43"),
                      home = c("17:30:00"))
  fn = "testactlog.csv"
  write.csv(x = actlog, file = fn, row.names = FALSE)
  LOG = g.conv.actlog(qwindow = fn, qwindow_dateformat = "%Y-%m-%d")
  
  # definedays
  definedays = g.part5.definedays(nightsi = nightsi, wi = 1, indjump = 1, 
                                  epochSize = 60, qqq_backup = c(), 
                                  ts = ts, timewindowi = "MM", Nwindows = 1, 
                                  qwindow = LOG, ID = "1RAW")
  
  expect_false(is.null(definedays$segments))
  expect_true(is.list(definedays$segments))
  expect_equal(length(definedays$segments), 5)
  expect_equal(names(definedays$segments)[1], "00:00:00-23:59:00")
  expect_equal(definedays$segments[[1]], c(1, 1440))
  expect_equal(names(definedays$segments)[2], "00:00:00-08:09:00")
  expect_equal(definedays$segments[[2]], c(1, 490))
  expect_equal(names(definedays$segments)[3], "08:10:00-16:30:00")
  expect_equal(definedays$segments[[3]], c(491, 991))
  expect_equal(names(definedays$segments)[4], "16:31:00-17:29:00")
  expect_equal(definedays$segments[[4]], c(992, 1050))
  expect_equal(names(definedays$segments)[5], "17:30:00-23:59:00")
  expect_equal(definedays$segments[[5]], c(1051, 1440))
  expect_equal(definedays$segments_names, c("MM", "daystart-work", 
                                            "work-travelhome", 
                                            "travelhome-home", "home-dayend"))
  
  expect_true(file.exists(fn))
  if (file.exists(fn)) file.remove(fn)
  
  
  # With times in not chronological order
  actlog = data.frame(id = c("1RAW"),
                      date = c("2022-01-01"),
                      travelhome = c("16:30:43"),
                      work = c("08:09:59"),
                      home = c("17:30:00"))
  fn = "testactlog.csv"
  write.csv(x = actlog, file = fn, row.names = FALSE)
  LOG = g.conv.actlog(qwindow = fn, qwindow_dateformat = "%Y-%m-%d")
  
  # definedays
  definedays = g.part5.definedays(nightsi = nightsi, wi = 1, indjump = 1, 
                                  epochSize = 60, 
                                  qqq_backup = c(), 
                                  ts = ts, timewindowi = "MM", Nwindows = 1, 
                                  qwindow = LOG, ID = "1RAW")
  
  expect_false(is.null(definedays$segments))
  expect_true(is.list(definedays$segments))
  expect_equal(length(definedays$segments), 5)
  expect_equal(names(definedays$segments)[1], "00:00:00-23:59:00")
  expect_equal(definedays$segments[[1]], c(1, 1440))
  expect_equal(names(definedays$segments)[2], "00:00:00-08:09:00")
  expect_equal(definedays$segments[[2]], c(1, 490))
  expect_equal(names(definedays$segments)[3], "08:10:00-16:30:00")
  expect_equal(definedays$segments[[3]], c(491, 991))
  expect_equal(names(definedays$segments)[4], "16:31:00-17:29:00")
  expect_equal(definedays$segments[[4]], c(992, 1050))
  expect_equal(names(definedays$segments)[5], "17:30:00-23:59:00")
  expect_equal(definedays$segments[[5]], c(1051, 1440))
  expect_equal(definedays$segments_names, c("MM", "daystart-work", 
                                            "work-travelhome", 
                                            "travelhome-home", "home-dayend"))
  
  expect_true(file.exists(fn))
  if (file.exists(fn)) file.remove(fn)
  
})

g.part5.wakesleepwindows = function(ts, part4_output, desiredtz, nightsi,
                                    sleeplog, epochSize, ID,
                                    Nepochsinhour) {
  #========================================================
  # DIURNAL BINARY CLASSIFICATION INTO NIGHT (onset-wake) OR DAY (wake-onset) PERIOD 
  Nts = nrow(ts)
  findIndex = function(timeChar = NULL, wc = NULL) {
    wc_withzerotime = paste0(wc," 00:00:00")
    if (length(grep(pattern = " ", x = wc)) == 0) {
      wc = wc_withzerotime
    }
    index = which(timeChar == wc)[1]
    if (is.na(index) == TRUE) {
      index = which(timeChar == wc_withzerotime)[1]
    }
    return(index)
  }
  clock2numtime = function(x) { # function used for converting sleeplog times to hour times
    x2 = as.numeric(unlist(strsplit(x,":"))) / c(1,60,3600)
    return(sum(x2))
  }
  w0 = w1 = rep(0,length(part4_output$calendar_date))
  # Round seconds to integer number of epoch lengths (needed if cleaningcode = 5).
  round_seconds_to_epochSize = function(x, epochSize) {
    if (length(as.numeric(unlist(strsplit(x,":")))) == 3) {
      xPOSIX = as.POSIXct(x, format = "%H:%M:%S", origin = "1970-01-01")
      xPOSIX_rounded = as.POSIXct(round(as.numeric(xPOSIX) / epochSize) * epochSize,
                                  origin = "1970-01-01")
      x = format(xPOSIX_rounded, format = "%H:%M:%S")
    } else {
      x = ""
    }
    return(x)
  }
  # standardise time series to character without iso8601 respresentation
  timeChar = format(ts$time) 
  if (is.ISO8601(timeChar[1]) == TRUE) { # only do this for ISO8601 format
    timeChar = format(iso8601chartime2POSIX(timeChar, tz = desiredtz))
  }
  
  for (k in 1:length(part4_output$calendar_date)) { # loop through nights from part 4
    part4_output$wakeup_ts[k] = round_seconds_to_epochSize(part4_output$wakeup_ts[k], epochSize)
    part4_output$sleeponset_ts[k] = round_seconds_to_epochSize(part4_output$sleeponset_ts[k], epochSize)
    
    part4_output$sleeponset[k] = (round((part4_output$sleeponset[k]*3600) / epochSize) * epochSize) / 3600
    part4_output$wakeup[k] = (round((part4_output$wakeup[k]*3600) / epochSize) * epochSize) / 3600
    
    # Load sleep onset and waking time from part 4 and convert them into timestamps
    tt = unlist(strsplit(as.character(part4_output$calendar_date[k]),"/")) # calendar date
    # if sleep onset is not available in from acc and/or sleep then us the following default
    # in order to still have some beginning and end of the night, these days will be discared
    # anyway, because typically this coincides with a lot of non-wear time:
    if (is.na(part4_output$sleeponset[k]) == T) {
      defSO = 22 
      defSO_ts = "22:00:00"
    } else {
      defSO = part4_output$sleeponset[k]
      defSO_ts = part4_output$sleeponset_ts[k]
    }
    if (is.na(part4_output$wakeup[k]) == T) {
      defWA = 31
      defWA_ts = "07:00:00"
    } else {
      defWA = part4_output$wakeup[k]
      defWA_ts = part4_output$wakeup_ts[k]
    }
    w0[k] = paste(tt[3],"-", tt[2], "-", tt[1], " ", as.character(defSO_ts), sep = "")
    w1[k] = paste(tt[3],"-", tt[2], "-", tt[1], " ", as.character(defWA_ts), sep = "")
    # if time is beyond 24 then change the date
    if (defSO >= 24) { 
      tmp_w0 = as.POSIXlt(w0[k],tz = desiredtz, origin = "1970-01-01")
      w0[k] = as.character(tmp_w0 + (24*3600))
    }
    if (defWA >= 24 |
        (part4_output$daysleeper[k] == 1 & defWA < 18)) {
      w1[k] = format(as.POSIXlt(w1[k], tz = desiredtz) + (24*3600))
    }
    s0 = findIndex(timeChar, wc = format(as.POSIXlt(w0[k], tz = desiredtz)))
    s1 = findIndex(timeChar, wc = format(as.POSIXlt(w1[k], tz = desiredtz)))
    
    if (is.na(s0) == TRUE) {
      if (format(as.POSIXlt(w0[k], tz = desiredtz)) < timeChar[1]) {
        # safe check, only turn s0 to 1 when the part4 wake up is before the first
        # timestamp. Otherwise, if wake up is after the last timestamp, then the 
        # whole time series is set to sleep period time (only occurs in nights with
        # cleaningcode = 5 in part 4 (no accelerometer data available, but spt extracted from sleeplog))
        s0 = 1
      }
    }
    if (is.na(s1) == TRUE) {
      # might still be NA if the timestamps is not in ts (expanded time from expand_tail)
      # if so, we assume the participant is sleeping at the end of the recording, this night will be disregarded later on
      if (format(as.POSIXlt(w1[k], tz = desiredtz)) > timeChar[length(timeChar)]) {
        # safe check, only turn s1 to nrow(ts) when the part4 wake up is after the last
        # timestamp. Otherwise, if sleep onset is before the first timestamp, then the 
        # whole time series is set to sleep period time (only occurs in nights with
        # cleaningcode = 5 in part 4 (no accelerometer data available, but spt extracted from sleeplog))
        s1 = nrow(ts)
      }
    }
    
    if (length(s1) != 0 & length(s0) != 0 & is.na(s0) == FALSE & is.na(s1) == FALSE && length(nightsi) > 0) {
      distance2midnight = abs(nightsi - s1) + abs(nightsi - s0)
      closestmidnighti = which.min(distance2midnight)
      closestmidnight = nightsi[closestmidnighti]
      noon0 = closestmidnight - (12 * (60/epochSize) * 60)
      noon1 = closestmidnight + (12 * (60/epochSize) * 60)
      if (noon0 < 1) noon0 = 1
      if (noon1 > Nts) noon1 = Nts
      nonwearpercentage = mean(ts$nonwear[noon0:noon1])
      if ((length(sleeplog) > 0 & (nonwearpercentage > 0.33) | part4_output$sleeponset_ts[k] == "")) { # added condition to detect nights that are not detected in part 4
        # If non-wear is high for this day and if sleeplog is available
        sleeplogonset = sleeplog$sleeponset[which(sleeplog$ID == ID & sleeplog$night == part4_output$night[k])]
        sleeplogwake = sleeplog$sleepwake[which(sleeplog$ID == ID & sleeplog$night == part4_output$night[k])]
        if (length(sleeplogonset) != 0 & length(sleeplogwake) != 0) {
          if (!is.na(sleeplogonset) &  !is.na(sleeplogwake)) {
            # ... and if there is sleeplog data for the relevant night
            # rely on sleeplog for defining the start and end of the night
            sleeplogonset_hr = clock2numtime(sleeplogonset)
            sleeplogwake_hr = clock2numtime(sleeplogwake)
            # express hour relative to midnight within the noon-noon:
            if (sleeplogonset_hr > 12) {
              sleeplogonset_hr = sleeplogonset_hr - 24
            }
            if (sleeplogwake_hr > 18 & part4_output$daysleeper[k] == 1) {
              sleeplogwake_hr = sleeplogwake_hr - 24 # 18 because daysleepers can wake up after 12
            } else if (sleeplogwake_hr > 12 & part4_output$daysleeper[k] == 0) {
              sleeplogwake_hr = sleeplogwake_hr - 24
            }
            if (sleeplogwake_hr > 36 &  sleeplogonset_hr > 36) {
              sleeplogwake_hr = sleeplogwake_hr - 24
              sleeplogonset_hr = sleeplogonset_hr - 24
            }
            s0 = closestmidnight + round(sleeplogonset_hr * Nepochsinhour)
            if (s0 < 1) {
              warning("Impossible index for first night, consider setting excludefirst.part4=TRUE")
              s0 = 1
            }
            s1 = closestmidnight + round(sleeplogwake_hr * Nepochsinhour)
            # if sleeplog indicates a time after the ending of the recording, 
            # then set the nrow(ts) + 1, so that the next line will set as 
            # SPT all the time from sleep onset until the end of the recording.
            if (s1 > nrow(ts) + 1) s1 = nrow(ts) + 1
          }
        }
      }
      # it might also be that both the sleeplog reported sleeponset and
      # wakeup goes beyond recording length, in that case, s1 < s0,
      # so only assign SPT if s0 < s1
      if (s0 < s1) {
        ts$diur[s0:(s1 - 1)] = 1
      }
    }
  }
  return(ts)
}

g.part5.handle_lux_extremes = function(lux) {
  # detect sequence of extreme lux values
  detect.pattern = function(patrn, x) {
    # credits to Berend Hasselman for following three lines of code
    # https://r.789695.n4.nabble.com/matching-a-sequence-in-a-vector-tp4389523p4389909.html
    patrn.rev = rev(patrn)
    w = embed(x,length(patrn))
    w.pos <- which(apply(w,1,function(r) all(r == patrn.rev)))
    if (length(w.pos) > 0) {
      result = c()
      for (w.pos_i in 1:length(w.pos)) {
        result = c(result, w.pos[w.pos_i]:(w.pos[w.pos_i] + length(patrn) - 1))
      }
    } else {
      result = c()
    }
    return(result)
  }
  extremes_lux_values = which(lux > 120000)
  correction_log = rep(0, length(lux))
  correction_log[extremes_lux_values] = 1
  # remove sequences lasting 3 long epochs or longer (default 45 minutes)
  extreme_sequence = detect.pattern(patrn = rep(1, 3), x = correction_log)
  if (length(extreme_sequence) > 0) { 
    lux[extreme_sequence] = NA
    correction_log[extreme_sequence] = 2
  }
  # for remaining extremes impute by average of neighbours
  extremes = which(correction_log == 1)
  if (length(extremes) > 0) {
    lux[extremes] = NA
    for (eb in 1:length(extremes)) {
      if (extremes[eb] == 1) {
        impluxi = 1:2
      } else if (extremes[eb] > 1 & extremes[eb] < length(lux)) {
        impluxi = (extremes[eb]-1):(extremes[eb]+1)
      } else if (extremes[eb] == length(lux)) {
        impluxi = (length(lux)-1):length(lux)
      }
      implux = mean(lux[impluxi], na.rm = TRUE)
      if (length(implux) > 0) lux[extremes[eb]] = implux
    }
  }
  invisible(list(lux=lux, correction_log=correction_log))
}

library(GGIR)
context("g.part5.handle_lux_extremes")
test_that("is able to handle lux extremes", {
  skip_on_cran()
  lux = c(130000, 1, 1, 1, 130000, 1, 1, 1, 130000, 130000, 1, 1, 1, 130000, 130000, 130000, 1, 1, 130000)
  out = g.part5.handle_lux_extremes(lux)
  expect_equal(length(which(out$correction_log == 2)), 3)
  expect_equal(length(which(out$correction_log == 1)), 5)
  expect_equal(mean(out$lux, na.rm=TRUE), 1)
  
  # test also absense of extreme values
  lux = rep(1, 12)
  out = g.part5.handle_lux_extremes(lux)
  expect_equal(sum(out$lux), 12)
  expect_equal(sum(out$correction_log), 0)
})

tidyup_df = function(df = c(), digits = 3) {
  
  df = df[rowSums(!is.na(df)) > 0, ]
  if (ncol(df) > 1) {
    # Round columns that can be coerced to numeric
    # But round fragmentation metrics to 6 decimal places instead of default 3
    # because some fragmentation metrics can reveal relevant info
    # at 4 or 5 decimal places
    for (fragCols in c(TRUE, FALSE)) {
      if (fragCols == TRUE) {
        colsID = grep(pattern = "FRAG_|ABI|SSP", x = colnames(df), invert = FALSE)
        digitsRound = 6
      } else {
        colsID = grep(pattern = "FRAG_|ABI|SSP", x = colnames(df), invert = TRUE)
        digitsRound = digits
      }
      myRoundFun = function(x) {
        tryCatch(round(as.numeric(as.character(x)), digits = digitsRound), 
                 error = function(cond) return(x),
                 warning = function(cond) return(x))
      }
      if (length(colsID) > 0) {
        df[, colsID] = lapply(df[, colsID], FUN = myRoundFun)
      }
    }
    # list to data frame
    df = as.data.frame(df)
  }
  return(df)
}

library(GGIR)
context("tidyup_df")
test_that("tidyup_df returns a data frame with the numeric columns rounded", {
  # POSIXtime2iso8601
  df = data.frame(letters = c("a", "b"), nums = as.character(c(1.543218, 8.216856483)), 
                  factor = factor(c("f1", "f2")))
  df_converted = tidyup_df(df = df, digits = 3)
  expect_equal(df_converted$letters, df$letters)
  expect_equal(df_converted$nums, c(1.543, 8.217))
  expect_equal(df_converted$factor, df$factor)
})

read.myacc.csv = function(rmc.file=c(), rmc.nrow=Inf, rmc.skip=c(), rmc.dec=".",
                          rmc.firstrow.acc = c(), rmc.firstrow.header=c(),
                          rmc.header.length = c(),
                          rmc.col.acc = 1:3, rmc.col.temp = c(), rmc.col.time=c(),
                          rmc.unit.acc = "g", rmc.unit.temp = "C", 
                          rmc.unit.time = "POSIX",
                          rmc.format.time = "%Y-%m-%d %H:%M:%OS",
                          rmc.bitrate = c(), rmc.dynamic_range = c(),
                          rmc.unsignedbit = TRUE,
                          rmc.origin = "1970-01-01",
                          rmc.desiredtz = NULL,
                          rmc.configtz = NULL,
                          rmc.sf = c(),
                          rmc.headername.sf = c(),
                          rmc.headername.sn = c(),
                          rmc.headername.recordingid = c(),
                          rmc.header.structure = c(),
                          rmc.check4timegaps = FALSE,
                          rmc.col.wear = c(),
                          rmc.doresample=FALSE,
                          rmc.scalefactor.acc = 1,
                          interpolationType = 1,
                          PreviousLastValue = c(0, 0, 1),
                          PreviousLastTime = NULL,
                          desiredtz = NULL,
                          configtz = NULL,
                          header = NULL) {

  if (length(rmc.col.time) > 0 && !(rmc.unit.time %in% c("POSIX", "character", "UNIXsec", "UNIXmsec", "ActivPAL"))) {
    stop(paste0("\nUnrecognized rmc.col.time value. The only accepted values are \"POSIX\", ",
                "\"character\", \"UNIXsec\", \"UNIXmsec\", and \"ActivPAL\"."), call. = FALSE)
  }

  if (!is.null(rmc.desiredtz) || !is.null(rmc.configtz)) {
    generalWarning = paste0("Argument rmc.desiredtz and rmc.configtz are scheduled to be deprecated",
                   " and will be replaced by the existing arguments desiredtz and configtz, respectively.")

    # Check if both types of tz are provided:
    if (!is.null(desiredtz) && desiredtz != "" && !is.null(rmc.desiredtz)) {
      if (rmc.desiredtz != desiredtz) { # if different --> error (don't know which one to use)
        stop(paste0("\n", generalWarning, "Please, specify only desiredtz and set ",
             "rmc.desiredtz to NULL to ensure it is no longer used."))
      }
    }
    if (!is.null(configtz) && !is.null(rmc.configtz)) { # then both provided 
      if (rmc.configtz != configtz) { # if different --> error (don't know which one to use)
        stop(paste0("\n", generalWarning, "Please, specify only configtz and set ",
             "rmc.configtz to NULL to ensure it is no longer used."))
      }
    }
    warning(paste0("\n", generalWarning))

    # Until deprecation still allow rmc. to be used, 
    # so use it to overwrite normal tz in this function:
    if (is.null(desiredtz)) desiredtz = rmc.desiredtz 
    if (desiredtz == "" && !is.null(rmc.desiredtz)) desiredtz = rmc.desiredtz
    if (is.null(configtz)) configtz = rmc.configtz
   
  }
  # check if none of desiredtz and rmc.desiredtz are provided
  if (is.null(desiredtz) && is.null(rmc.desiredtz)) {
    stop(paste0("Timezone not specified, please provide at least desiredtz",
                " and consider specifying configtz."))
  }
  
  if (is.null(rmc.firstrow.acc) || rmc.firstrow.acc < 1) {
    stop(paste0("\nParameter rmc.firstrow.acc always need to be specified ",
                "when working with ad-hoc csv format data"))
  }
  skip = rmc.firstrow.acc - 1
  if (!is.null(rmc.skip) && length(rmc.skip) > 0) {
    skip = skip + rmc.skip
  }

  # only extract the header if it hasn't been extracted for this file before
  if (is.null(header)) {
    # bitrate should be either header item name as character, or the actual numeric bit rate.
    # unit.temp can take C(elsius), F(ahrenheit), and K(elvin) and converts it into Celsius
    # Note all argument names start with rmc (read myacc csv) to avoid name clashes when passed on throughout GGIR
    if (length(rmc.firstrow.header) == 0) { # no header block
      sf = rmc.sf
      header = "no header"
    } else {
      # extract header information:
      if (length(rmc.header.length) == 0) {
        rmc.header.length = rmc.firstrow.acc - 1
      }
      
      options(warn = -1) # fread complains about quote in first row for some file types
      header_tmp = data.table::fread(file = rmc.file,
                                     nrows = rmc.header.length, 
                                     skip = rmc.firstrow.header - 1,
                                     dec = rmc.dec, showProgress = FALSE, header = FALSE,
                                     blank.lines.skip = TRUE,
                                     data.table=FALSE, stringsAsFactors=FALSE)
      validrows = which(is.na(header_tmp[,1]) == FALSE & header_tmp[,1] != "")
      header_tmp = header_tmp[validrows,1:2]
      
      options(warn = 0)
      if (length(rmc.header.structure) != 0) { # header is stored in 1 column, with strings that need to be split
        if (length(header_tmp) == 1) { # one header item
          header_tmp = as.matrix(unlist(strsplit(as.character(header_tmp[,1]), rmc.header.structure)))
        } else { # multiple header items
          mysplit = function(x){
            tmp = strsplit(as.character(x), rmc.header.structure)
            tmp = unlist(tmp)
            return(tmp)
          }
          header_tmp0 = header_tmp
          header_tmp = unlist(lapply(header_tmp[,1], FUN = mysplit))
          if (length(header_tmp) > 2) {
            header_tmp = data.frame(matrix(unlist(header_tmp), nrow = nrow(header_tmp0), byrow = T), stringsAsFactors = FALSE)
            colnames(header_tmp) = NULL
          } else {
            header_tmp = data.frame(matrix(unlist(header_tmp), nrow = 1, byrow = T), stringsAsFactors = FALSE)
            colnames(header_tmp) = NULL
          }
        }
        if (ncol(header_tmp) == 1) header_tmp = t(header_tmp)
        header_tmp2 = as.data.frame(as.character(unlist(header_tmp[,2])), stringsAsFactors = FALSE)
        row.names(header_tmp2) = header_tmp[,1] 
        colnames(header_tmp2) = NULL
        header = header_tmp2
      } else { # column 1 is header name, column 2 is header value
        colnames(header_tmp) = NULL
        header_tmp2 = as.data.frame(header_tmp[,2], stringsAsFactors = FALSE)
        row.names(header_tmp2) = header_tmp[,1]
        colnames(header_tmp2) = NULL
        header = header_tmp2
      }
      # assess whether accelerometer data conversion is needed
      if (length(rmc.bitrate) > 0 && length(rmc.dynamic_range) > 0 && rmc.unit.acc == "bit") {
        if (is.character(rmc.bitrate[1]) == TRUE) { # extract bitrate if it is in the header
          rmc.bitrate = as.numeric(header[which(row.names(header) == rmc.bitrate[1]),1])
        }
        if (is.character(rmc.dynamic_range[1]) ==  TRUE) { # extract dynamic range if it is in the header
          rmc.dynamic_range = as.numeric(header[which(row.names(header) == rmc.dynamic_range[1]),1])
        } 
      }
      # extract sample frequency:
      sf = as.numeric(header[which(row.names(header) == rmc.headername.sf[1]),1])
      
      if (is.na(sf)) { # sf not retrieved from header
        # first see if maybe sf *is* in the header, just not under the rmc.headername.sf name
        sf = as.numeric(header[which(row.names(header) == "sample_rate"),1])
        # if sf isn't in the header under the default name either, then use the default value
        if (is.na(sf)) {
          sf = rmc.sf # this could be null, that's fine. At least we can only end up with a null, not either null or NA
          if (!is.null(sf)) {
            header = rbind(header, sf) # also add it to the header
            row.names(header)[nrow(header)] = "sample_rate"
          }
        }
      }

      # standardise key header names to ease use elsewhere in GGIR:
      if (length(rmc.headername.sf) > 0) {
        row.names(header)[which(row.names(header) == rmc.headername.sf[1])] = "sample_rate"
      }
      if (length(rmc.headername.sn) > 0) {
        row.names(header)[which(row.names(header) == rmc.headername.sn[1])] = "device_serial_number"
      }
      if (length(rmc.headername.recordingid) > 0) {
        row.names(header)[which(row.names(header) == rmc.headername.recordingid[1])] = "recordingID"
      }
    }
  }
  # read data from file
  P = data.table::fread(rmc.file, nrows = rmc.nrow, skip = skip,
                        dec = rmc.dec, showProgress = FALSE, header = "auto",
                        data.table=FALSE, stringsAsFactors=FALSE)
  
  if (length(configtz) == 0) {
    configtz = desiredtz
  }
  if (length(rmc.col.wear) > 0) {
    wearIndicator = P[, rmc.col.wear] # keep wear channel seperately and reinsert at the end
  }
  # select relevant columns, add standard column names
  P = P[,c(rmc.col.time, rmc.col.acc, rmc.col.temp)]
  if (length(rmc.col.time) > 0 && length(rmc.col.temp) > 0) {
    colnames(P) = c("time","x","y","z","temperature")
  } else if (length(rmc.col.time) > 0 && length(rmc.col.temp) == 0) {
    colnames(P) = c("time","x","y","z")
  } else if (length(rmc.col.time) == 0 && length(rmc.col.temp) > 0) {
    colnames(P) = c("x","y","z","temperature")
  } else if (length(rmc.col.time) == 0 && length(rmc.col.temp) == 0) {
    colnames(P) = c("x","y","z")
  }
  # acceleration and temperature as numeric
  P$x = as.numeric(P$x)
  P$y = as.numeric(P$y)
  P$z = as.numeric(P$z)
  if (length(rmc.col.temp) > 0) P$temperature = as.numeric(P$temperature) 
  # Convert timestamps
  if (length(rmc.col.time) > 0) {
    if (rmc.unit.time == "POSIX") {
      P$time = as.POSIXct(format(P$time), origin = rmc.origin, tz = configtz, format = rmc.format.time)
      checkdec = function(x) {
        # function to check whether timestamp has decimal places
        return(length(unlist(strsplit(as.character(x), "[.]|[,]"))) == 1)
      }
      first_chunk_time = P$time[1:pmin(nrow(P), 1000)]
      checkMissingDecPlaces = unlist(lapply(first_chunk_time, FUN = checkdec))
      if (all(checkMissingDecPlaces) &&
          !is.null(sf) && sf != 0 &&
          length(which(duplicated(first_chunk_time) == TRUE)) > 0) {
        # decimal places are not present and there are duplicated timestamps,
        # so insert decimal places
        #-----
        # dummy data, to test the following code:
        # ttt = as.POSIXlt("2022-11-02 14:46:50", tz = "Europe/Amsterdam")
        # sf = 10
        # P = data.frame(timestamps = c(rep(ttt - 1, 3), rep(ttt, 10), rep(ttt + 1, 9), rep(ttt + 2, 10), rep(ttt + 3, 4)))
        #------
        trans = unique(c(1, which(diff(P$time) > 0), nrow(P)))
        sf_tmp = diff(trans)
        timeIncrement = seq(from = 0, length.out = sf, by = 1/sf) # expected time increment per second
        
        # All seconds with exactly the sample frequency
        trans_1 = trans[which(sf_tmp == sf)]
        indices_1 = sort(unlist(lapply(trans_1, FUN = function(x){x + (1:sf)})))
        P$time[indices_1] =  P$time[indices_1] + rep(timeIncrement, length(trans_1))
        # First second
        if (sf_tmp[1] != sf) {
          indices_2 = 1:trans[2]
          P$time[indices_2] = P$time[indices_2] + seq(1 - (trans[2]/sf), 1 - 1/sf, by = 1/sf)
        }
        # Last second
        if (sf_tmp[length(sf_tmp)] != sf) {
          indices_3 = (trans[length(trans)-1] + 1):trans[length(trans)]
          P$time[indices_3] = P$time[indices_3] + timeIncrement[1:length(indices_3)]
        }
        # All seconds with other sample frequency and not the first or last second
        # This code now assumes that most samples in the second were sampled
        # at the correct rate but that some samples were dropped or doubled
        # as a result of which the sample rate appears different
        # It seems that this may be a safer assumption than the assumption
        # that the entire second had a different sample rate
        if (length(trans) > 4) {
          trans_cut = trans[2:(length(trans)-1)]
          sf_tmp_cut = sf_tmp[2:(length(sf_tmp)-1)]
          sf_tmp_odd = unique(sf_tmp_cut[which(sf_tmp_cut != sf)])
          if (length(sf_tmp_odd) > 0) {
            for (ji in 1:length(sf_tmp_odd)) {
              sf2 = sf_tmp_odd[ji]
              trans_4 = trans_cut[which(sf_tmp_cut == sf2)]
              indices_4 = sort(unlist(lapply(trans_4, FUN = function(x){x + (1:sf2)})))
              if (length(timeIncrement) > sf2) {
                timeIncrement2 = timeIncrement[1:sf2]
              } else if (length(timeIncrement) < sf2) {
                timeIncrement2 = c(timeIncrement, rep(timeIncrement[sf], sf2 - sf))
              }
              P$time[indices_4] =  P$time[indices_4] + rep(timeIncrement2, length(trans_4))
            }
          }
        }
      }
    } else if (rmc.unit.time == "character") {
      P$time = as.POSIXct(P$time, format = rmc.format.time, origin = rmc.origin, tz = configtz)
    } else if (rmc.unit.time == "UNIXsec" || rmc.unit.time == "UNIXmsec") {
      if (rmc.unit.time == "UNIXmsec") {
        P$time = P$time / 1000
      }
      if (rmc.origin != "1970-01-01") {
        P$time = as.POSIXct(P$time, origin = rmc.origin, tz = desiredtz)
      }
    } else if (rmc.unit.time == "ActivPAL") {
      # origin should be specified as: "1899-12-30"
      P$time = lubridate::force_tz(as.POSIXct(P$time * 86400, origin = "1899-12-30", tz = "UTC"), tz = desiredtz)
    }
    if (length(which(is.na(P$time) == FALSE)) == 0) {
      stop("\nExtraction of timestamps unsuccesful, check timestamp format arguments")
    }
    if (!is.numeric(P$time)) { # we'll return Unix timestamps
      P$time = as.numeric(P$time)
    }
  }
  
  # If acceleration is stored in mg units then convert to gravitational units
  if (rmc.unit.acc == "mg") {
    P$x = P$x / 1000
    P$y = P$y / 1000
    P$z = P$z / 1000
  }
  if (rmc.scalefactor.acc != 1) {
    P$x = P$x * rmc.scalefactor.acc
    P$y = P$y * rmc.scalefactor.acc
    P$z = P$z * rmc.scalefactor.acc
  }
  # If acceleration is stored in bit values then convert to gravitational unit
  if (length(rmc.bitrate) > 0 && length(rmc.dynamic_range) > 0 && rmc.unit.acc == "bit") {
    if (rmc.unsignedbit == TRUE) {
      P$x = ((P$x / (2^rmc.bitrate)) - 0.5) * 2 * rmc.dynamic_range
      P$y = ((P$y / (2^rmc.bitrate)) - 0.5) * 2 * rmc.dynamic_range
      P$z = ((P$z / (2^rmc.bitrate)) - 0.5) * 2 * rmc.dynamic_range
    } else if (rmc.unsignedbit == FALSE) { # signed bit
      P$x = (P$x / ((2^rmc.bitrate)/2)) * rmc.dynamic_range
      P$y = (P$y / ((2^rmc.bitrate)/2)) * rmc.dynamic_range
      P$z = (P$z / ((2^rmc.bitrate)/2)) * rmc.dynamic_range
    }
  }
  # Convert temperature units
  if (rmc.unit.temp == "K") {
    P$temperature = P$temperature + 272.15 # From Kelvin to Celsius
  } else if (rmc.unit.temp == "F") {
    P$temperature = (P$temperature - 32) * (5/9) # From Fahrenheit to Celsius
  }
  if (length(rmc.col.wear) > 0) { # reinsert the nonwear channel
    P$wear = wearIndicator
  }
  # check for jumps in time and impute
  if (rmc.check4timegaps == TRUE && ("time" %in% colnames(P))) {
    sfBackup = sf
    if (is.null(sf) || sf == 0) { # estimate sample frequency if not given in header
      deltatime = abs(diff(as.numeric(P$time)))
      gapsi = which(deltatime > 0.25)
      sf = (P$time[gapsi[1]] - P$time[1]) / (gapsi[1] - 1)
    }
    P = g.imputeTimegaps(P, sf = sf, k = 0.25, 
                         PreviousLastValue = PreviousLastValue,
                         PreviousLastTime = PreviousLastTime, epochsize = NULL)
    sf = sfBackup
    P = P$x
    PreviousLastValue = P[nrow(P), c("x", "y", "z")]
    PreviousLastTime = as.POSIXct(P[nrow(P), "time"], origin = "1970-01-01")
  }
  if (rmc.doresample == TRUE && ("time" %in% colnames(P)) && !is.null(sf) && sf != 0) { # resample
    rawTime = P$time
    rawAccel = as.matrix(P[,-c(which(colnames(P) == "time"))])
    timeRes = seq(from = rawTime[1], to = rawTime[length(rawTime)], by = 1/sf)
    accelRes = GGIRread::resample(rawAccel, rawTime, timeRes, nrow(rawAccel), interpolationType) # this is now the resampled acceleration data
    colnamesP = colnames(P)[-which(colnames(P) == "time")]
    P = as.data.frame(accelRes, stringsAsFactors = FALSE)
    colnames(P) = colnamesP
    P$time = timeRes
  }
  return(list(data = P, header = header, 
              PreviousLastValue = PreviousLastValue,
              PreviousLastTime = PreviousLastTime))
}

library(GGIR)
context("read.myacc.csv")
test_that("read.myacc.csv can handle files without header, no decimal places in timestamps, temperature", {
  skip_on_cran()
  old_options = options()
  # create test files
  N = 30
  sf = 30
  options(digits.secs = 4)
  
  # NOTE THIS UNIT TEST STILL USES POSIXlt A LOT
  # BECAUSE OLDER R RELEASE CANNOT HANDLE POSIXct for this purpose
  t0 = as.POSIXlt(x = "2022-11-02 14:01:16.00", tz = "Europe/Amsterdam") 
  timeseq = t0 + ((0:(N - 1))/sf)
  time = as.POSIXlt(timeseq, origin = "1970-1-1", tz = "Europe/London")
  testfile = matrix("", 4, 1)
  set.seed(100)
  accx = rnorm(N)
  set.seed(200)
  accy = rnorm(N)
  set.seed(300)
  accz = rnorm(N)
  set.seed(400)
  temp = rnorm(N)
  wear = c(rep(TRUE,N/3),rep(FALSE,N/6),rep(TRUE,N/3),rep(TRUE,N/6))
  
  # No header, but otherwise normal
  S1 = data.frame(x = accx, time = time, 
                  y = accy, z = accz, temp = temp + 20, 
                  stringsAsFactors = TRUE)
  
  testfile[1] = "testcsv1.csv"
  write.csv(S1, file = testfile[1], row.names = FALSE)
  
  # With decimal places in seconds removed
  testfile[2] = "testcsv2.csv"
  S1$time = round(S1$time)
  write.csv(S1, file = testfile[2], row.names = FALSE)
  
  # Without temperature
  S2 = data.frame(x = accx, time = time, y = accy, z = accz, stringsAsFactors = TRUE)
  testfile[3] = "testcsv3.csv"
  write.csv(S2, file = testfile[3], row.names = FALSE)
  
  # Without temperature, without time, with a wear channel
  S3 = data.frame(x = accx, y = accy, z = accz, wear = wear, stringsAsFactors = TRUE)
  testfile[4] = "testcsv4.csv"
  write.csv(S3, file = testfile[4], row.names = FALSE)
  
  # UNIXsec
  testfile[5] = "testcsv5.csv"
  S5 = S1
  S5$time = as.numeric(S5$time)
  write.csv(S5, file = testfile[5], row.names = FALSE)
  
  # UNIXmsec
  testfile[6] = "testcsv6.csv"
  S6 = S5
  S6$time = S6$time * 1000
  write.csv(S6, file = testfile[6], row.names = FALSE)
  
  # attempt to load
  D1 = read.myacc.csv(rmc.file = testfile[1], rmc.nrow = 20, rmc.dec = ".",
                      rmc.firstrow.acc = 1, rmc.firstrow.header = c(),
                      rmc.col.acc = c(1,3,4), rmc.col.temp = 5, rmc.col.time = 2,
                      rmc.unit.acc = "g", rmc.unit.temp = "C",
                      rmc.format.time = "%Y-%m-%d %H:%M:%OS", rmc.origin = "1970-01-01",
                      desiredtz = "Europe/London", rmc.sf = sf,
                      rmc.headername.sf = "sample_frequency",
                      rmc.headername.sn = "serial_number",
                      rmc.headername.recordingid = "ID")
  
  expect_equal(mean(D1$data$x), 0.1078671, tol = 0.0001)
  expect_equal(mean(D1$data$y), -0.06675716, tol = 0.0001)
  
  # attempt to load as if values are in mg and scale by 100
  D1B = read.myacc.csv(rmc.file = testfile[1], rmc.nrow = 20, rmc.dec = ".",
                      rmc.firstrow.acc = 1, rmc.firstrow.header = c(),
                      rmc.col.acc = c(1,3,4), rmc.col.temp = 5, rmc.col.time = 2,
                      rmc.unit.acc = "mg", rmc.unit.temp = "C",
                      rmc.format.time = "%Y-%m-%d %H:%M:%OS", rmc.origin = "1970-01-01",
                      rmc.scalefactor.acc = 100,
                      desiredtz = "Europe/London", rmc.sf = sf,
                      rmc.headername.sf = "sample_frequency",
                      rmc.headername.sn = "serial_number",
                      rmc.headername.recordingid = "ID")
  
  expect_equal(mean(D1B$data$x), 0.01078671, tol = 0.0001)
  expect_equal(mean(D1B$data$y), -0.006675716, tol = 0.0001)
  
  #------------------------
  # Warnings and errors related to desiredtz and configtz
  # warning if desiredtz not provided but rmc.desiredtz provided:
  expect_warning(read.myacc.csv(rmc.file = testfile[1], rmc.nrow = 20, rmc.dec = ".",
                                rmc.firstrow.acc = 1, rmc.firstrow.header = c(),
                                rmc.col.acc = c(1,3,4), rmc.col.temp = 5, rmc.col.time = 2,
                                rmc.unit.acc = "g", rmc.unit.temp = "C", 
                                rmc.format.time = "%Y-%m-%d %H:%M:%OS",
                                rmc.origin = "1970-01-01",
                                rmc.desiredtz = "Europe/London", rmc.sf = sf,
                                rmc.headername.sf = "sample_frequency",
                                rmc.headername.sn = "serial_number",
                                rmc.headername.recordingid = "ID"))
  
  #  now also with configtz
  expect_warning(read.myacc.csv(rmc.file = testfile[1], rmc.nrow = 20, rmc.dec = ".",
                                rmc.firstrow.acc = 1, rmc.firstrow.header = c(),
                                rmc.col.acc = c(1,3,4), rmc.col.temp = 5, rmc.col.time = 2,
                                rmc.unit.acc = "g", rmc.unit.temp = "C", 
                                rmc.format.time = "%Y-%m-%d %H:%M:%OS",
                                rmc.origin = "1970-01-01",
                                desiredtz = "Europe/London", 
                                rmc.configtz = "Europe/Madrid",
                                rmc.sf = sf,
                                rmc.headername.sf = "sample_frequency",
                                rmc.headername.sn = "serial_number",
                                rmc.headername.recordingid = "ID"))
  
  # error if none of rmc.desiredtz and desiredtz are provided:
  expect_error(read.myacc.csv(rmc.file = testfile[1], rmc.nrow = 20, rmc.dec = ".",
                              rmc.firstrow.acc = 1, rmc.firstrow.header = c(),
                              rmc.col.acc = c(1,3,4), rmc.col.temp = 5, rmc.col.time = 2,
                              rmc.unit.acc = "g", rmc.unit.temp = "C", 
                              rmc.format.time = "%Y-%m-%d %H:%M:%OS",
                              rmc.origin = "1970-01-01",
                              rmc.sf = sf,
                              rmc.headername.sf = "sample_frequency",
                              rmc.headername.sn = "serial_number",
                              rmc.headername.recordingid = "ID"))
  
  # error if both rmc.desiredtz and desiredtz are provided:
  expect_error(read.myacc.csv(rmc.file = testfile[1], rmc.nrow = 20, rmc.dec = ".",
                              rmc.firstrow.acc = 1, rmc.firstrow.header = c(),
                              rmc.col.acc = c(1,3,4), rmc.col.temp = 5, rmc.col.time = 2,
                              rmc.unit.acc = "g", rmc.unit.temp = "C", 
                              rmc.format.time = "%Y-%m-%d %H:%M:%OS",
                              rmc.origin = "1970-01-01",
                              desiredtz = "Europe/London", rmc.desiredtz = "Europe/Madrid",
                              rmc.sf = sf,
                              rmc.headername.sf = "sample_frequency",
                              rmc.headername.sn = "serial_number",
                              rmc.headername.recordingid = "ID"))
  
  
  # Evaluate with decimal places in seconds
  expect_equal(nrow(D1$data), 20)
  expect_equal(ncol(D1$data), 5)
  expect_equal(strftime(as.POSIXct(D1$data$time[1:5], tz = "Europe/London", origin = "1970-01-01"),
                        format = '%Y-%m-%d %H:%M:%OS2', tz = "Europe/London"), 
               c("2022-11-02 13:01:16.00",
                 "2022-11-02 13:01:16.03",
                 "2022-11-02 13:01:16.06",
                 "2022-11-02 13:01:16.09",
                 "2022-11-02 13:01:16.13"))
  expect_that(D1$header,equals("no header"))
  
  # Evaluate without decimal places decimal places in seconds and
  # different config timezone
  D2 = read.myacc.csv(rmc.file = testfile[2], rmc.nrow = 20, rmc.dec = ".",
                      rmc.firstrow.acc = 1, rmc.firstrow.header = c(),
                      rmc.col.acc = c(1,3,4), rmc.col.temp = 5, rmc.col.time = 2,
                      rmc.unit.acc = "g", rmc.unit.temp = "C", rmc.format.time = "%Y-%m-%d %H:%M:%OS",
                      rmc.origin = "1970-01-01",
                      desiredtz = "Europe/London", rmc.sf = sf,
                      configtz = "America/Chicago",
                      rmc.headername.sf = "sample_frequency",
                      rmc.headername.sn = "serial_number",
                      rmc.headername.recordingid = "ID")
  expect_equal(nrow(D2$data), 20)
  expect_equal(ncol(D2$data), 5)
  expect_equal(strftime(as.POSIXct(D2$data$time[1:5], tz = "Europe/London", origin = "1970-01-01"), 
                        format = '%Y-%m-%d %H:%M:%OS2', tz = "Europe/London"),
               c("2022-11-02 18:01:16.50", "2022-11-02 18:01:16.53",
                 "2022-11-02 18:01:16.56", "2022-11-02 18:01:16.59",
                 "2022-11-02 18:01:16.63"))
  expect_that(D2$header,equals("no header"))
  
  D3 = read.myacc.csv(rmc.file = testfile[3], rmc.nrow = 20, rmc.dec = ".",
                      rmc.firstrow.acc = 1, rmc.firstrow.header = c(),
                      rmc.col.acc = c(1,3,4), rmc.col.temp = c(),
                      rmc.col.time = 2,
                      rmc.unit.acc = "g", rmc.unit.temp = "C",
                      rmc.format.time = "%Y-%m-%d %H:%M:%OS",
                      rmc.origin = "1970-01-01",
                      desiredtz = "Europe/London", 
                      rmc.sf = sf,
                      rmc.headername.sf = "sample_frequency",
                      rmc.headername.sn = "serial_number",
                      rmc.headername.recordingid = "ID")
  expect_that(nrow(D3$data),equals(20))
  expect_that(ncol(D3$data),equals(4))
  expect_that(D3$header,equals("no header"))
  
  D4 = read.myacc.csv(rmc.file = testfile[4], rmc.nrow = 20, 
                      rmc.dec = ".",
                      rmc.firstrow.acc = 1, rmc.firstrow.header = c(),
                      rmc.col.acc = 1:3, rmc.col.temp = c(), rmc.col.time = c(),
                      rmc.unit.acc = "g", rmc.unit.temp = "C", 
                      rmc.format.time = "%Y-%m-%d %H:%M:%OS",
                      rmc.origin = "1970-01-01",
                      desiredtz = "Europe/London", rmc.sf = sf,
                      rmc.headername.sf = "sample_frequency",
                      rmc.headername.sn = "serial_number",
                      rmc.headername.recordingid = "ID",
                      rmc.col.wear = 4)
  expect_that(nrow(D4$data), equals(20))
  expect_true(D4$data[1, 4])
  expect_that(ncol(D4$data), equals(4))
  expect_that(D4$header, equals("no header"))
  
  # Timestamps in UNIXsec
  D5 = read.myacc.csv(rmc.file = testfile[5], rmc.nrow = 20, rmc.dec = ".",
                      rmc.firstrow.acc = 1, rmc.firstrow.header = c(),
                      rmc.col.acc = c(1,3,4), rmc.col.temp = 5, rmc.col.time = 2,
                      rmc.unit.acc = "g", 
                      rmc.unit.time = "UNIXsec",
                      desiredtz = "Europe/London", rmc.sf = sf)
  
  expect_equal(mean(D5$data$x), 0.1078671, tol = 0.0001)
  expect_equal(mean(D5$data$y), -0.06675716, tol = 0.0001)
  expect_equal(D5$data$time[1], 1667394076, tol = 0.1)
  
  # Timestamps in UNIXmsec
  D6 = read.myacc.csv(rmc.file = testfile[6], rmc.nrow = 20, rmc.dec = ".",
                      rmc.firstrow.acc = 1, rmc.firstrow.header = c(),
                      rmc.col.acc = c(1,3,4), rmc.col.temp = 5, rmc.col.time = 2,
                      rmc.unit.acc = "g", 
                      rmc.unit.time = "UNIXmsec",
                      desiredtz = "Europe/London", rmc.sf = sf)
  expect_equal(mean(D6$data$x), 0.1078671, tol = 0.0001)
  expect_equal(mean(D6$data$y), -0.06675716, tol = 0.0001)
  expect_equal(D6$data$time[1], 1667394076, tol = 0.1)
  
  for (i in 1:length(testfile)) {
    expect_true(file.exists(testfile[i]))
    if (file.exists(testfile[i])) file.remove(testfile[i])
  }
  options(old_options)
})

test_that("read.myacc.csv can handle header and bit-value acceleration", {
  skip_on_cran()
  old_options = options()
  # create test files
  N = 30
  sf = 30
  options(digits.secs = 4)
  t0 = as.POSIXlt(x = "2022-11-02 14:01:16.00", tz = "Europe/Amsterdam")
  timeseq = t0 + ((0:(N - 1))/sf)
  time = as.POSIXlt(timeseq, origin = "1970-1-1", tz = "Europe/London")
  testfile = matrix("", 3, 1)
  set.seed(100)
  accx = rnorm(N)
  set.seed(200)
  accy = rnorm(N)
  set.seed(300)
  accz = rnorm(N)
  set.seed(400)
  temp = rnorm(N)
  wear = c(rep(TRUE, N / 3), rep(FALSE, N / 6),
           rep(TRUE, N / 3), rep(TRUE, N / 6))

  # 1: With 2-column header, with temperature, with time
  S1 = as.matrix(data.frame(x = accx, time = time, y = accy,
                            z = accz, temp = temp + 20, stringsAsFactors = TRUE))
  hd_NR = 10
  hd = matrix("", hd_NR + 1, ncol(S1))
  hd[1, 1:2] = c("ID","12345")
  hd[2, 1:2] = c("sample_rate","30")
  hd[3, 1:2] = c("serial_number","30")
  hd[4, 1:2] = c("bit","8")
  hd[5, 1:2] = c("dynamic_range","6")
  S1 = rbind(hd, S1)
  S1[hd_NR + 1,] = colnames(S1)
  colnames(S1) = NULL
  testfile[1] = "testcsv1.csv"
  write.table(S1, file = testfile[1], col.names = FALSE, row.names = FALSE)

  # 2: With 2-column header, with temperature, with time, bit-value acceleration unit
  bits = 8
  set.seed(100)
  xb = sample(x = 1:(2^bits),size = N,replace = TRUE)
  set.seed(200)
  yb = sample(x = 1:(2^bits),size = N,replace = TRUE)
  set.seed(300)
  zb = sample(x = 1:(2^bits),size = N,replace = TRUE)
  set.seed(400)
  temp3 = rnorm(N)
  S3 = S2 = as.matrix(data.frame(x = xb, time = time, y = yb,
                            z = zb, temp = temp3 + 20, stringsAsFactors = TRUE))
  S2 = rbind(hd, S2)
  S2[hd_NR + 1, ] = colnames(S2)
  colnames(S2) = NULL
  testfile[2] = "testcsv2.csv"
  write.table(S2, file = testfile[2], col.names = FALSE, row.names = FALSE)

  # 2. A header in 1 column:
  hd2 = matrix("", hd_NR + 1, ncol(S3))
  hd2[1, 1:2] = c("ID: 12345", "")
  hd2[2, 1:2] = c("sample_rate: 30", "")
  hd2[3, 1:2] = c("serial_number: 30", "")
  hd2[4, 1:2] = c("bit: 8", "")
  hd2[5, 1:2] = c("dynamic_range: 6", "")
  S3 = rbind(hd2, S3)
  S3[hd_NR + 1,] = colnames(S3)
  colnames(S3) = NULL
  testfile[3] = "testcsv3.csv"
  write.table(S3, file = testfile[3], col.names = FALSE, row.names = FALSE)



  #------------------------
  # Test 1 - 2 column header
  D1 = read.myacc.csv(rmc.file = testfile[1], rmc.nrow = 20, rmc.dec = ".",
                      rmc.firstrow.acc = 11, rmc.firstrow.header = 1,
                      rmc.col.acc = c(1, 3, 4), rmc.col.temp = 5, rmc.col.time = 2,
                      rmc.unit.acc = "g", rmc.unit.temp = "C",
                      rmc.format.time = "%Y-%m-%d %H:%M:%OS",
                      rmc.origin = "1970-01-01",
                      desiredtz = "Europe/London",
                      rmc.headername.sf = "sample_frequency",
                      rmc.headername.sn = "serial_number",
                      rmc.headername.recordingid = "ID")
  expect_that(nrow(D1$data), equals(20))
  expect_that(ncol(D1$data), equals(5))
  expect_that(nrow(D1$header), equals(5))
  expect_that(ncol(D1$header), equals(1))
  expect_equal(as.numeric(D1$header["sample_rate",1]), 30)
  
  # Test 2 - 2 column header, bit-valued acceleration
  D2 = read.myacc.csv(rmc.file = testfile[2], rmc.nrow = 20, rmc.dec = ".",
                      rmc.firstrow.acc = 11, rmc.firstrow.header = 1,
                      rmc.col.acc = c(1,3,4), rmc.col.temp = 5,
                      rmc.col.time = 2,
                      rmc.unit.acc = "bit", rmc.unit.temp = "C",
                      rmc.format.time = "%Y-%m-%d %H:%M:%OS",
                      rmc.origin = "1970-01-01",
                      desiredtz = "Europe/London", rmc.sf = sf,
                      rmc.headername.sf = "sample_rate",
                      rmc.headername.sn = "serial_number",
                      rmc.headername.recordingid = "ID", rmc.bitrate = "bit",
                      rmc.dynamic_range = "dynamic_range",
                      rmc.header.structure = c())
  expect_that(nrow(D2$data),equals(20))
  expect_that(ncol(D2$data),equals(5))
  expect_that(nrow(D2$header),equals(5))
  expect_that(ncol(D2$header),equals(1))
  
  # Test 3 - 1 column header, bit-valued acceleration
  D3 = read.myacc.csv(rmc.file = testfile[3], rmc.nrow = 20, rmc.dec = ".",
                      rmc.firstrow.acc = 11, rmc.firstrow.header = 1,
                      rmc.header.length = 5,
                      rmc.col.acc = c(1,3,4), rmc.col.temp = 5,
                      rmc.col.time = 2,
                      rmc.unit.acc = "bit", rmc.unit.temp = "C",
                      rmc.format.time = "%Y-%m-%d %H:%M:%OS",
                      rmc.origin = "1970-01-01",
                      desiredtz = "Europe/London", rmc.sf = sf,
                      rmc.headername.sf = "sample_rate",
                      rmc.headername.sn = "serial_number",
                      rmc.headername.recordingid = "ID",
                      rmc.bitrate = "bit", rmc.dynamic_range = "dynamic_range",
                      rmc.header.structure = ": ", rmc.check4timegaps = TRUE)
  # 
  expect_equal(mean(D3$data[,2]), 0.761, tolerance = 3)
  expect_equal(mean(D3$data[,3]), -0.62, tolerance = 2)
  expect_equal(mean(D3$data[,4]), -0.36, tolerance = 1)
  expect_equal(mean(D3$data[,5]), 20.1, tolerance = 1)
  expect_equal(D3$data[2,2], 0.9768219, tolerance  = 3)
  expect_equal(ncol(D3$data), 5)
  expect_equal(nrow(D3$header),5)
  expect_equal(ncol(D3$header), 1)
  expect_equal(as.numeric(D3$header[1,]), 12345)
  expect_equal(as.numeric(D3$header[2,]), 30)

  for (i in 1:length(testfile)) {
    expect_true(file.exists(testfile[i]))
    if (file.exists(testfile[i])) file.remove(testfile[i])
  }
   options(old_options)
})

test_that("read.myacc.csv can handle gaps in time and irregular sample rate", {
  skip_on_cran()
  old_options = options()
  # create test files
  N = 30
  sf = 30
  options(digits.secs = 4)
  t0 = as.POSIXlt(x = "2022-11-02 14:01:16.00", tz = "Europe/Amsterdam")
  timeseq = t0 + ((0:(N - 1))/sf)
  time = as.POSIXlt(timeseq, origin = "1970-1-1", tz = "Europe/London")
  testfile = matrix("", 1, 1)
  set.seed(100)
  accx = rnorm(N)
  set.seed(200)
  accy = rnorm(N)
  set.seed(300)
  accz = rnorm(N)
  set.seed(400)
  temp = rnorm(N)
  wear = c(rep(TRUE, N / 3), rep(FALSE, N / 6),
           rep(TRUE, N / 3), rep(TRUE, N / 6))
  
  # 1: With 2-column header, with temperature, with gap in time, variation in sample rate
  time_gap = time
  time_gap[10:length(time_gap)] = time_gap[10:length(time_gap)] + 5 # add gap of 5 seconds
  time_gap = time_gap + rnorm(n = length(time_gap), mean = 0, sd = 0.0001)
  S1 = as.matrix(data.frame(x = accx, time = time_gap, y = accy,
                            z = accz, temp = temp + 20, stringsAsFactors = TRUE))
  hd_NR = 10
  hd = matrix("", hd_NR + 1, ncol(S1))
  hd[1, 1:2] = c("ID","12345")
  hd[2, 1:2] = c("sample_rate","30")
  hd[3, 1:2] = c("serial_number","30")
  S1 = rbind(hd, S1)
  S1[hd_NR + 1,] = colnames(S1)
  colnames(S1) = NULL
  testfile[1] = "testcsv1.csv"
  write.table(S1, file = testfile[1], col.names = FALSE, row.names = FALSE)
  
  
  #------------------------
  # Test 1 - time gaps but otherwise regular sample rate
  D1 = read.myacc.csv(rmc.file = testfile[1], rmc.nrow = 20, rmc.dec = ".",
                      rmc.firstrow.acc = 11, rmc.firstrow.header = 1,
                      rmc.col.acc = c(1, 3, 4), rmc.col.temp = 5, rmc.col.time = 2,
                      rmc.unit.acc = "g", rmc.unit.temp = "C",
                      rmc.format.time = "%Y-%m-%d %H:%M:%OS",
                      rmc.origin = "1970-01-01",
                      desiredtz = "Europe/London",
                      rmc.sf = sf,
                      rmc.headername.sf = "sample_rate",
                      rmc.headername.sn = "serial_number",
                      rmc.headername.recordingid = "ID",
                      rmc.check4timegaps = TRUE, rmc.doresample = FALSE)
  expect_that(nrow(D1$data), equals(170)) # because data expands with 5 seconds that are now imputed
  expect_that(ncol(D1$data), equals(5))
  
  #------------------------
  # Test 2 - time gaps and irregular sample rate
  D2 = read.myacc.csv(rmc.file = testfile[1], rmc.nrow = 20, rmc.dec = ".",
                      rmc.firstrow.acc = 11, rmc.firstrow.header = 1,
                      rmc.col.acc = c(1, 3, 4), rmc.col.temp = 5, rmc.col.time = 2,
                      rmc.unit.acc = "g", rmc.unit.temp = "C",
                      rmc.format.time = "%Y-%m-%d %H:%M:%OS",
                      rmc.origin = "1970-01-01",
                      desiredtz = "Europe/London",
                      rmc.sf = sf,
                      rmc.headername.sf = "sample_rate",
                      rmc.headername.sn = "serial_number",
                      rmc.headername.recordingid = "ID",
                      rmc.check4timegaps = TRUE, rmc.doresample = TRUE)
  expect_that(nrow(D2$data), equals(170)) # because data expands with 5 seconds that are now imputed
  expect_that(ncol(D2$data), equals(5))
  
  
  for (i in 1:length(testfile)) {
    expect_true(file.exists(testfile[i]))
    if (file.exists(testfile[i])) file.remove(testfile[i])
  }
  options(old_options)
})

part6AlignIndividuals = function(GGIR_ts_dir = NULL, outputdir = NULL,
                                 path_ggirms = NULL, desiredtz = "", verbose = TRUE) {
  if (verbose == TRUE) {
    cat("\n  Align individuals:")
  }
  #======================================================================
  # identify houdesholds and IDs based on filenames and put these in table
  path_timeseries = GGIR_ts_dir
  path_results = paste0(outputdir, "/part6HouseholdCoAnalysis")
  if (!dir.exists(path_results)) dir.create(path = path_results, recursive = TRUE)
  
  path_results_ts = paste0(path_results, "/alignedTimeseries")
  if (!dir.exists(path_results_ts)) dir.create(path = path_results_ts, recursive = TRUE)
  
  fns = dir(path_timeseries, pattern = "RData")
  getIDs = function(x) {
    tmp = unlist(strsplit(x, "-|_"))
    return(data.frame(HID = tmp[2], MID = tmp[3], filename = x))
  }
  fileOverview = lapply(fns, getIDs)
  fileOverview = do.call(rbind,fileOverview)
  S = table(fileOverview$HID)
  fileOverview$houshold_size = NA
  for (i in 1:length(S)) {
    rowi = which(fileOverview$HID == names(S)[i])
    fileOverview$houshold_size[rowi] = S[i] 
  }
  
  # remove all households with less than two individuals:
  fileOverview = fileOverview[which(fileOverview$houshold_size > 1),]
  uHID = unique(fileOverview$HID)
  
  pdf(file = paste0(path_results, "/timeseriesPlot.pdf"))
  #======================================================================
  # Loop over households and IDs to check and merge the data
  
  mdat = NULL
  # Loop over houdeshold IDS
  for (h in 1:length(uHID)) {
    if (verbose == TRUE) cat(paste0("\n    Houdeshold ", uHID[h],": "))
    uMID = fileOverview$MID[which(fileOverview$HID == uHID[h])]
    uFilename = fileOverview$filename[which(fileOverview$HID == uHID[h])]
    out = NULL
    # Loop over Houdeshold members
    for (p in 1:length(uMID)) {
      if (verbose == TRUE) cat(paste0(uMID[p], " ")) #, uFilename[p]))
      # load data
      load(file = paste0(path_timeseries, "/", uFilename[p]))
      D = mdat
      if (nrow(D) > 0) { # check if file is not empty
        # Identify corresponding GGIR part 1 milestone data
        pat1 = unlist(strsplit(x = uFilename[p], split = "[.]csv"))[1]
        pat2 = unlist(strsplit(x = pat1, split = "_"))
        pat2 = paste0(pat2[1:(length(pat2) - 1)], collapse = "_")
        ptrn = paste0("meta_",pat2)
        msfile = dir(path = path_ggirms,  pattern = ptrn, full.names = TRUE, recursive = TRUE, ignore.case = TRUE)
        # merge temperature and light if they are available
        extravars = c()
        if (length(msfile) > 0) {
          load(msfile)
          if (all(c("lightmean", "lightpeak", "temperaturemean") %in% colnames(M$metalong))) {
            M$metalong$timestamp = as.POSIXct(M$metalong$timestamp,
                                              format = "%Y-%m-%dT%H:%M:%S%z", tz = desiredtz, origin = "1970-1-1")
            M$metalong$timenum = as.numeric(M$metalong$timestamp)
            
            epochSizeMetalong = mean(abs(diff(M$metalong$timenum[1:3])))
            epochSizeMS5 = mean(abs(diff(D$timenum[1:3])))
            if (epochSizeMetalong > epochSizeMS5) { # If long epoch size in part 1 was larger than 60 seconds 
                          M$metalong = M$metalong[rep(1:nrow(M$metalong),
                                                      each = epochSizeMetalong / epochSizeMS5), ]
              newTimenum = seq(M$metalong$timenum[1], M$metalong$timenum[nrow(M$metalong)] + epochSizeMetalong, by = epochSizeMS5)
              M$metalong$timenum = newTimenum[1:nrow(M$metalong)]
              M$metalong$timestamp = as.POSIXct(M$metalong$timenum,
                                                origin = "1970-1-1", tz = desiredtz)
            }
            if ("lightpeak" %in% colnames(D)) D = D[, -which(colnames(D) == "lightpeak")]
            D = merge(D, M$metalong[,c("lightmean", "lightpeak", "temperaturemean", "timenum")], by = "timenum")
            extravars = c("lightmean", "lightpeak", "temperaturemean")
          }
        } else {
          warning(paste0("\nGGIR milestone data not found for ", msfile))
        }
        invalid = which((D$invalid_fullwindow > 20 |
                           D$invalid_wakinghours > 20 | 
                           D$invalid_sleepperiod > 20) & D$window != 0)
        if (length(invalid) > 0) {
          is.na(D[invalid, 
                  c("ACC", "SleepPeriodTime", "invalidepoch", "guider", "window", "class_id", extravars)]) = TRUE
        }
        D = D[,c("timenum", "ACC", "SleepPeriodTime", "invalidepoch", "guider", "window", "class_id", extravars)]
        # add wakeup indicators
        D$onset = 0
        D$onset[which(diff(D$SleepPeriodTime) > 0)] = 1
        D$wakeup = 0
        D$wakeup[which(diff(D$SleepPeriodTime) < 0)] = 1
        
        if (length(which(is.na(D$ACC) == FALSE)) > 0) {
          D$MID = uMID[p]
          D$HID = uHID[h]
          if (length(out) == 0) {
            out = D
          } else {
            out = rbind(out, D)
          }
        }
      }
      rm(mdat, M)
    }
    
    out$validepoch = (-out$invalidepoch) + 1
    # only continue if there are at least 2 Houdeshold members left
    if (length(unique(out$MID)) > 1) {
      uMID = unique(out$MID)
      out = out[order(out$timenum),]
      
      # reshape to be in wide format to ease doing pairwise comparisons
      alignedTimeseries = reshape(out,
                     idvar = c("timenum", "HID"),
                     timevar = "MID",
                     direction = "wide")
      colnames = paste0("ACC.", uMID)
      
      alignedTimeseries$time_POSIX = as.POSIXct(alignedTimeseries$timenum,
                                                tz = "America/Halifax", origin = "1970-01-01")
      
      # Add indicator of valid number of family members per time point
      alignedTimeseries$N_valid_hhmembers = rowSums(alignedTimeseries[,grep(pattern = "validepoch",
                                                                            x = names(alignedTimeseries), value = FALSE)],
                                                    na.rm = TRUE)
      # Add indicator of valid data for a pair of family members per time point
      varn_remember = c()
      Nmembers = length(uMID)
      for (pa1 in 1:Nmembers) {
        for (pa2 in 1:Nmembers) {
          if (pa1 != pa2 & pa2 > pa1) {
            col1 = grep(pattern = paste0("validepoch.", uMID[pa1]),
                        x = names(alignedTimeseries), value = FALSE)
            col2 = grep(pattern = paste0("validepoch.", uMID[pa2]),
                        x = names(alignedTimeseries), value = FALSE)
            validpair = ifelse(test = rowSums(alignedTimeseries[, c(col1, col2)], na.rm = TRUE) == 2,
                               yes = TRUE, no = FALSE)
            alignedTimeseries$new = validpair
            varname = paste0("validpair_",uMID[pa1],"_", uMID[pa2])
            colnames(alignedTimeseries)[which(colnames(alignedTimeseries) == "new")] = varname
            varn_remember = c(varn_remember, varname)
          }
        }
      }
      Npairs = length(varn_remember)
      #======================================================================
      # Visualise data
      if (exists("currentColorPalette")) rm(currentColorPalette)
      Ncols = (Npairs + Nmembers)
      palette(rainbow(n = Ncols, alpha = 1)) 
      currentColorPalette = palette()
      colorsM = currentColorPalette[1:Nmembers]
      colorsP = currentColorPalette[(Nmembers + 1):(Nmembers + Npairs)]
      
      par(mfrow = c(2,1))
      plot(alignedTimeseries$time_POSIX, alignedTimeseries[, colnames[1]], type = "l", col = colorsM[1],
           main = paste0("Houshold ", uHID[h]), xlab = "", ylab = "ACC", bty = "l")
      for (jj in 2:length(colnames)) {
        lines(alignedTimeseries$timenum, alignedTimeseries[, colnames[jj]], type = "l", col = colorsM[jj])
      }
      legend("topright", legend = colnames,
             col = colorsM[1:length(colnames)], lty = 1, cex = 0.6, bg = "white", ncol = 3)
      LTY = rep(1, Npairs)
      
      for (vn in 1:Npairs) {
        if (vn == 1) {
          plot(alignedTimeseries$time_POSIX, alignedTimeseries[,varn_remember[vn]] - 0.05 + (vn/50), type = "l",
               col = colorsP[vn], ylim = c(0, 1.5),
               xlab = "", ylab = "Valid pair", bty = "l",
               axes = FALSE, lty = LTY[vn], lwd = 1.3)
        } else {
          lines(alignedTimeseries$time_POSIX, alignedTimeseries[,varn_remember[vn]] - 0.05 + (vn/50), type = "l",
                col = colorsP[vn], lty = LTY[vn], lwd = 1.3)
        }
      }
      axis(side = 2, at = c(0, 1), labels = c("FALSE", "TRUE"))
      legend("topright", legend = gsub(pattern = "valid", replacement = "", x = varn_remember),
             col = colorsP[1:Npairs], lty = LTY[1:Npairs], cex = 0.6, bg = "white", ncol = 3)
      for (i in 1:ncol(alignedTimeseries)) {
        NA_NaN = which(is.na(alignedTimeseries[, i]))
        alignedTimeseries[NA_NaN, i] = ""
      }
      rm(currentColorPalette)

      # Convert character to appropriate data format
      numericColumns = grep(pattern = "onset|wake|valid_hhmembers|epoch|temperature|light|class|SleepPeriod|ACC|window|guider", x = colnames(alignedTimeseries))
      for (jj in numericColumns) {
        alignedTimeseries[,jj] = as.numeric(alignedTimeseries[ ,jj])
      }
      boolColumns = grep(pattern = "validpair", x = colnames(alignedTimeseries))
      for (jj in boolColumns) {
        alignedTimeseries[,jj] = as.logical(alignedTimeseries[ ,jj])
      }
      # Store data
      save(alignedTimeseries,file = paste0(path_results, "/alignedTimeseries/timeseries_HID_", uHID[h], ".RData"))
      # write.csv(x = alignedTimeseries,
      #           file = paste0(path_results, "/alignedTimeseries/timeseries_HID_",uHID[h],".csv"),
      #           row.names = FALSE)
    }
  }
  dev.off()
}

HASIB = function(HASIB.algo = "vanHees2015", timethreshold = c(), anglethreshold = c(), 
                 time = c(), anglez = c(), ws3 = c(), 
                 zeroCrossingCount = c(), NeishabouriCount = c(),
                 activity = NULL, oakley_threshold = NULL) {
  epochsize = ws3 #epochsize in seconds
  sumPerWindow = function(x, epochsize, summingwindow = 60) {
    x2 = cumsum(c(0, x))
    select = seq(1, length(x2), by = summingwindow/epochsize)
    x3 = diff(x2[select])
  }
  create_rollfun_mat = function(x, Ncol) {
    Nz = length(x)
    # Generate matrix to ease applying rolling function
    x_matrix = matrix(0, Nz + (Ncol - 1), Ncol)
    for (jj in 1:Ncol) {
      x_matrix[jj:(Nz + jj - 1), jj] = x
      if (jj > 1) {
        x_matrix[1:(jj - 1), jj] = x[1]
      }
      if (jj < Ncol) {
        x_matrix[((Nz - ((Ncol - 1) - jj)):Nz) + (Ncol - 1), jj] = tail(x, 1)
      }
    }
    # Remove redundant rows at start and end
    Nremove = ceiling(Ncol / 2)
    x_matrix = x_matrix[Nremove:(nrow(x_matrix) - (Nremove - 1)),]
    return(x_matrix)
  }
  
  reformat_output = function(x, time, new_epochsize = 5, current_epochsize = 60) {
    # resample to original resolution
    x = rep(x, each = (current_epochsize / new_epochsize))
    # resize to match length of time
    if (length(x) < length(time)) {
      x = c(x,rep(0, length(time) - length(x)))
    } else if (length(x) > length(time)) {
      x = x[1:length(time)]
    }
    # format as expected matrix
    sib_classification = matrix(0,length(x), 1)
    sib_classification[,1] = as.matrix(x)
    return(sib_classification)
  }
  #===============================
  Nvalues = max(length(anglez), length(zeroCrossingCount), length(NeishabouriCount),
                length(activity))
  if (HASIB.algo == "vanHees2015") { # default
    cnt = 1
    Ndefs = length(timethreshold) * length(anglethreshold)
    sib_classification = matrix(0,Nvalues, Ndefs)
    for (i in timethreshold) {
      for (j in anglethreshold) {
        sdl1 = rep(0,length(time))
        postch = which(abs(diff(anglez)) > j) #posture change of at least j degrees
        # count posture changes that happen less than once per ten minutes
        q1 = c()
        if (length(postch) > 1) {
          q1 = which(diff(postch) > (i*(60/epochsize))) #less than once per i minutes
        }
        if (length(q1) > 0) {
          for (gi in 1:length(q1)) {
            sdl1[postch[q1[gi]]:postch[q1[gi] + 1]] = 1 #periods with no posture change
          }
        } else { #possibly a day without wearing
          if (length(postch) < 10) {  #possibly a day without wearing
            sdl1[1:length(sdl1)] = 1 #periods with no posture change
          } else {  #possibly a day with constantly posture changes
            sdl1[1:length(sdl1)] = 0 #periodsposture change
          }
        }
        sib_classification[,cnt] = sdl1
        cnt = cnt + 1
      }
    }
    cnt = 1
    sib_classification = as.data.frame(sib_classification, stringsAsFactors = TRUE)
    for (i in timethreshold) {
      for (j in anglethreshold) {
        colnames(sib_classification)[cnt] = paste("T", i, "A", j, sep = "")
        cnt = cnt + 1
      }
    }
  } else if (HASIB.algo == "Sadeh1994") {
    count_types = c()
    if (length(zeroCrossingCount) > 0) count_types = "zeroCrossingCount"
    if (length(NeishabouriCount) > 0) count_types = c(count_types, "NeishabouriCount")
    sib_classification = as.data.frame(matrix(0, Nvalues, length(count_types)))
    cti = 1
    for (count_type in count_types) {
      if (count_type == "zeroCrossingCount") {
        Countpermin = sumPerWindow(zeroCrossingCount, epochsize = epochsize, summingwindow = 60)
      } else if (count_type == "NeishabouriCount") {
        Countpermin = sumPerWindow(NeishabouriCount, epochsize = epochsize, summingwindow = 60)
      }
      if (count_type != "zeroCrossingCount") {
        # because this is what ActiLife does for their counts
        # https://actigraphcorp.my.site.com/support/s/article/Where-can-I-find-documentation-for-the-Sadeh-and-Cole-Kripke-algorithms#:~:text=The%20Sadeh%20algorithm%20uses%20the%20y%2Daxis%20epoch%20data.%20If%20any%20of%20the%20epoch%20counts%20are%20over%20300%2C%20it%20reduces%20them%20to%20300.
        Countpermin = ifelse(test = Countpermin > 300, yes = 300, no = Countpermin)
      }
      Countpermin_matrix = create_rollfun_mat(Countpermin, Ncol = 11)
      CalcSadehFT = function(x) {
        MeanW5 = mean(x, na.rm = TRUE)
        SDlast = sd(x[6:11]) #last five in this matrix means columns 6:11
        NAT = length(which(x > 50 & x < 100))
        LOGact = log(x[6] + 1)
        return(data.frame(MeanW5 = MeanW5, SDlast = SDlast,
                          NAT = NAT, LOGact = LOGact))
      }
      SadehFT1 = apply(X = Countpermin_matrix, MARGIN = 1, FUN = CalcSadehFT)
      rm(Countpermin_matrix)
      SadehFT = data.frame(matrix(unlist(SadehFT1), nrow = length(SadehFT1), byrow = TRUE))
      rm(SadehFT1)
      colnames(SadehFT) = c("MeanW5", "SDlast", "NAT", "LOGact")
      # apply Sadeh algorithm
      PS = 7.601 - (0.065 * SadehFT$MeanW5) - (1.08 * SadehFT$NAT) - (0.056 * SadehFT$SDlast) - (0.703 * SadehFT$LOGact)
      PS = c(rep(-2, 5), PS) # add five zeros because first 5 epochs are not classified by the algorithm
      
      PSscores = rep(0, length(PS))
      PSsibs = which(PS >= 0) # sleep
      if (length(PSsibs) > 0) {
        PSscores[PSsibs] = 1 # sleep
      }
      # resample to original resolution and ensure length matches length of time
      sib_classification[,cti] = reformat_output(x = PSscores, time, new_epochsize = epochsize,
                                                 current_epochsize = 60)
      if (count_type == "zeroCrossingCount") colnames(sib_classification)[cti] = paste0(HASIB.algo, "_ZC")
      if (count_type == "NeishabouriCount") colnames(sib_classification)[cti] = paste0(HASIB.algo, "_Neishabouri")
      cti = cti + 1
    }
  } else if (HASIB.algo == "ColeKripke1992") {
    count_types = c()
    if (length(zeroCrossingCount) > 0) count_types = "zeroCrossingCount"
    if (length(NeishabouriCount) > 0) count_types = c(count_types, "NeishabouriCount")
    sib_classification = as.data.frame(matrix(0, Nvalues, length(count_types)))
    
    if (epochsize <= 60) {
      aggwindow = 60
    } else if (epochsize > 60) {
      stop("Cole Kripke algorithm is not designed for epochs larger than 1 minute")
    }
    cti = 1
    for (count_type in count_types) {
      # We use the algorithms on page 6 of Cole, R. J., Kripke, D. F., http://doi.org/10.1093/sleep/15.5.461
      # which corresponds to non-overlapping 1 minute epochs of activity per minute.
      # The other algorithms proposed are less clearly described:
      # - How does aggration per epoch takes place, sum or average? Without knowing this it is hard
      # to oversee whether it matters whether we use 2 or 5 seconds as input. 
      # - Are maximum 10 seconds converted back to the unit of counts per minute?
      # For now we assume that the unit is counts per 10 seconds
      if (count_type == "zeroCrossingCount") {
        CountperWindow = sumPerWindow(zeroCrossingCount, epochsize = epochsize, summingwindow = aggwindow)
      } else if (count_type == "NeishabouriCount") {
        CountperWindow = sumPerWindow(NeishabouriCount, epochsize = epochsize, summingwindow = aggwindow)
      }
      # Convert unit to counts per minute if aggwindow is not 60
      if (aggwindow != 60) CountperWindow = CountperWindow * (60/aggwindow)
      # Convert back to Counts per 10 seconds as impression is that this provides more plausible output
      CountperWindow = CountperWindow / 6 
      # Prepare matrix to ease applying weights
      CountperWindow_matrix = create_rollfun_mat(CountperWindow, Ncol = 7)
      # Apply weights
      CKweights = c(67, 74, 230, 76, 58, 54, 106) # reversed to match order of matrix
      PS = 0.001 * (CountperWindow_matrix %*% CKweights)
      # Add 4 wake score because first 4 epochs are not classified by the algorithm
      PS = c(rep(2, 4), PS) 
      # Not applying rescoring as described by Cole 1992, because accuracy improvements
      # were reported to be marginal which makes the added complexity not justified
      PSscores = rep(0, length(PS))
      PSsibs = which(PS < 1) # sleep
      if (length(PSsibs) > 0) {
        PSscores[PSsibs] = 1 #sleep
      }
      # re sample to original resolution and ensure length matches length of time
      sib_classification[,cti] = reformat_output(x = PSscores, time, new_epochsize = epochsize,
                                                 current_epochsize = 60)
      if (count_type == "zeroCrossingCount") colnames(sib_classification)[cti] = paste0(HASIB.algo, "_ZC")
      if (count_type == "NeishabouriCount") colnames(sib_classification)[cti] = paste0(HASIB.algo, "_Neishabouri")
      cti = cti + 1
    }
  } else if (HASIB.algo == "Galland2012") {  
    count_types = c()
    if (length(zeroCrossingCount) > 0) count_types = "zeroCrossingCount"
    if (length(NeishabouriCount) > 0) count_types = c(count_types, "NeishabouriCount")
    sib_classification = as.data.frame(matrix(0, Nvalues, length(count_types)))
    cti = 1
    for (count_type in count_types) {
      # Aggregate per minute
      if (count_type == "zeroCrossingCount") {
        Countpermin = sumPerWindow(zeroCrossingCount, epochsize = epochsize, summingwindow = 60)
      } else if (count_type == "NeishabouriCount") {
        Countpermin = sumPerWindow(NeishabouriCount, epochsize = epochsize, summingwindow = 60)
      }
      mean_nonzero = mean(Countpermin[which(Countpermin != 0)])
      CountScaled = Countpermin / mean_nonzero
      CountScaled_matrix = create_rollfun_mat(CountScaled, Ncol = 7)
      # In next line I use intentionally a reversed order relative to Galland publication
      # because our matrix is reversed relative to paper
      WeightCounts = abs(CountScaled_matrix %*% c(1,3,5:1)) * 2.7 
      WeightCounts = WeightCounts[-c(1:2)] # remove first two time stamps, to align with 5th element (7-5=2)
      GallandScore = rep(0, length(WeightCounts))
      GallandScore[which(WeightCounts < 1)] = 1
      sib_classification[,cti] = reformat_output(x = GallandScore, time, new_epochsize = epochsize,
                                                 current_epochsize = 60)
      if (count_type == "zeroCrossingCount") colnames(sib_classification)[cti] = paste0(HASIB.algo, "_ZC")
      if (count_type == "NeishabouriCount") colnames(sib_classification)[cti] = paste0(HASIB.algo, "_Neishabouri")
      cti = cti + 1
    }
  } else if (HASIB.algo == "Oakley1997") {
    # Sleep Detection based on description in
    # Information bulletin no.3 sleep algorithms by Cambrige Neurotechnologies
    if (epochsize > 60) {
      stop("GGIR only facilitates the Oakley algorithm for epoch sizes up to 1 minute.", call. = FALSE)
    }
    count_types = c()
    if (length(zeroCrossingCount) > 0) count_types = "zeroCrossingCount"
    if (length(NeishabouriCount) > 0) count_types = c(count_types, "NeishabouriCount")
    sib_classification = as.data.frame(matrix(0, Nvalues, length(count_types)))
    cti = 1
    for (count_type in count_types) {
      if (epochsize %in% c(15, 30, 60) == FALSE) {
        # aggregate to 5, 30 or 60 seconds
        if (epochsize < 15 && 15 %% epochsize == 0) {
          aggwindow = 15
        } else if (epochsize < 30 && 30 %% epochsize == 0) {
          aggwindow = 30
        } else {
          aggwindow = 60
        }
        if (count_type == "zeroCrossingCount") {
          counts = sumPerWindow(zeroCrossingCount, epochsize = epochsize, summingwindow = aggwindow)
        } else if (count_type == "NeishabouriCount") {
          counts = sumPerWindow(NeishabouriCount, epochsize = epochsize, summingwindow = aggwindow)
        }
      } else {
        counts = zeroCrossingCount
        aggwindow = epochsize
      }
      # Oakley coefficients are epoch size specific
      if (aggwindow == 15) {
        oakley_coef = c(rep(0.04, 4), rep(0.20, 4), 4, rep(0.20, 4), rep(0.04, 4))
      } else if (aggwindow == 30) {
        oakley_coef = c(0.04, 0.04, 0.20, 0.20, 2.0, 0.2, 0.2, 0.04, 0.04)
      } else if (aggwindow == 60) {
        oakley_coef = c(0.04, 0.2, 1, 0.20, 0.04)
      }
      # Apply coefficients to estimate sleep
      Noak = length(oakley_coef)
      counts_matrix = create_rollfun_mat(counts, Ncol = Noak)
      PS = counts_matrix %*% oakley_coef
      rm(counts_matrix)
      PSscores = rep(0, length(PS))
      PSsibs = which(PS <= oakley_threshold)
      if (length(PSsibs) > 0) {
        PSscores[PSsibs] = 1 # sleep
      }
      # Resample to original resolution and ensure length matches length of time
      sib_classification[,cti] = reformat_output(x = PSscores, time,
                                                 new_epochsize = epochsize,
                                                 current_epochsize = aggwindow)
      if (count_type == "zeroCrossingCount") colnames(sib_classification)[cti] = paste0(HASIB.algo, "_ZC")
      if (count_type == "NeishabouriCount") colnames(sib_classification)[cti] = paste0(HASIB.algo, "_Neishabouri")
      cti = cti + 1
    }
  } else if (HASIB.algo == "NotWorn") {  
    # For the rare study protocols where sensor is not worn during the night
    # there is no point in looking at sleep. Nonetheless for the GGIR 
    # framework to work we need some estimate. This is not ideal but seems a pragmatic work
    # around to avoid having to completely redesign GGIR just for those studies.
    
    sib_classification = as.data.frame(matrix(0, Nvalues, 1))
    activity2 = zoo::rollmax(x = activity, k = 300 / epochsize, fill = 1)
    # ignore zeros because in ActiGraph with many zeros it skews the distribution
    nonzero = which(activity2 != 0)
    if (length(nonzero) > 0) {
      activityThreshold = sd(activity2[nonzero], na.rm = TRUE) * 0.05
      if (activityThreshold < min(activity)) {
        activityThreshold = quantile(activity2[nonzero], probs = 0.1)
      }
    } else {
      activityThreshold = 0
    }
    zeroMovement = which(activity2 <= activityThreshold)
    if (length(zeroMovement) > 0) {
      sib_classification[zeroMovement, 1] = 1
    }
    colnames(sib_classification) = "NotWorn"
  }
  return(sib_classification)
}

library(GGIR)
context("HASIB")
test_that("HASIB generate correct output", {
  skip_on_cran()
  time = seq(as.POSIXlt(x = "2021-3-3 15:00:00", tz = "Europe/Amsterdam"),
             as.POSIXlt(x = "2021-3-3 16:05:00", tz = "Europe/Amsterdam"), by = 5)
  tmp = c(rep(0, 10), 10, 20, 30, rep(30, 310), 20, 10, -5, -20, 0, -15, -30, rep(-40, 400))
  anglez = c(tmp, rep(0, length(time) - length(tmp)))
  set.seed(1245)
  zeroCrossingCount = round(abs(rnorm(n = length(anglez), mean = 0, sd = 50)))
  zeroCrossingCount[40:500] = 0
  vanHees2015 = HASIB(HASIB.algo = "vanHees2015", timethreshold = 5, anglethreshold = 5, 
                   time = time, anglez = anglez, ws3 = 5, zeroCrossingCount = c())
  Sadeh1994 = HASIB(HASIB.algo = "Sadeh1994", timethreshold = c(), anglethreshold = c(), 
                      time = time, anglez = c(), ws3 = 5, zeroCrossingCount = zeroCrossingCount)
  ColeKripke1992 = HASIB(HASIB.algo = "ColeKripke1992", timethreshold = c(), anglethreshold = c(), 
                    time = time, anglez = c(), ws3 = 5, zeroCrossingCount = zeroCrossingCount)
  Galland2012 = HASIB(HASIB.algo = "Galland2012", timethreshold = c(), anglethreshold = c(), 
                    time = time, anglez = c(), ws3 = 5, zeroCrossingCount = zeroCrossingCount)
  Oakley1997 = HASIB(HASIB.algo = "Oakley1997", timethreshold = c(), anglethreshold = c(), 
                     time = time, anglez = c(), ws3 = 5, zeroCrossingCount = zeroCrossingCount,
                     oakley_threshold = 20)
  
  expect_equal(nrow(vanHees2015), 781)
  expect_equal(nrow(Sadeh1994), 781)
  expect_equal(nrow(Galland2012), 781)
  expect_equal(nrow(ColeKripke1992), 781)
  expect_equal(nrow(Oakley1997), 781)
  expect_equal(length(which(vanHees2015[,1] == 1)), 713)
  expect_equal(length(which(Sadeh1994[,1] == 1)), 372)
  expect_equal(length(which(ColeKripke1992[,1] == 1)), 372)
  expect_equal(length(which(Galland2012[,1] == 1)), 396)
  expect_equal(length(which(Oakley1997[,1] == 1)), 435)
})

g.sib.det = function(M, IMP, I, twd = c(-12, 12),
                     acc.metric = "ENMO", desiredtz = "",
                     myfun=c(), sensor.location = "wrist",
                     params_sleep = c(), zc.scale = 1, ...) {
  #get input variables
  input = list(...)
  if (length(input) > 0 || length(params_sleep) == 0) {
    # Extract and check parameters if user provides more arguments than just the parameter arguments,
    # or if params_[...] aren't specified (so need to be filled with defaults).
    # So, inside GGIR this will not be used, but it is used when g.sleep is used on its own
    # as if it was still the old g.sleep function
    params = extract_params(params_sleep = params_sleep,
                            input = input,
                            params2check = "sleep") # load default parameters
    params_sleep = params$params_sleep
  }
  #==============================================================
  # Abbreviaton SPTE = Sleep Period Time Estimate, although in case of HorAngle it is the Time in Bed estimate
  # but then we would have to come up with yet another term to represent the main sleep and/or time in bed window of the day
  # So, out of convenience I keep the object name SPTE. 
  dstime_handling_check = function(tmpTIME = tmpTIME, spt_estimate = spt_estimate, tz = c(),
                                   calc_SPTE_end = c(), calc_SPTE_start = c()) {
    # Function to readjust estimated SPTE_start and SPTE_end in DST days
    t = tmpTIME[c(1, spt_estimate$SPTE_start, spt_estimate$SPTE_end)]
    timezone = format(GGIR::iso8601chartime2POSIX(t, tz = tz), "%z")
    if (length(unique(timezone)) == 1) {
      return(c(calc_SPTE_start, calc_SPTE_end))
    } else {
      sign = ifelse(substr(timezone, 1, 1) == "-", -1, 1)
      hours = as.numeric(substr(timezone, 2, 3))
      minutes = as.numeric(substr(timezone, 4, 5))
      offset = sign * (hours + minutes / 60)
      offset = offset[2:3] - offset[1]
      return(c(calc_SPTE_start + offset[1], calc_SPTE_end + offset[2]))
    }
  }
  decide_guider = function(HASPT.algo, nonwear_percentage) {
    # Function to decide the guider to use in the case that HASPT.algo is of length 2
    if (HASPT.algo[1] == "NotWorn" && nonwear_percentage < 25 && length(HASPT.algo) == 2) {
      return(2)
    } else {
      return(1)
    }
  }
  #==================================================
  # get variables
  nD = nrow(IMP$metashort)
  mon = I$monn
  ws3 = M$windowsizes[1] #short epoch size
  ws2 = M$windowsizes[2] # medium length epoch size used for non-wear detection
  n_ws2_perday = (1440*60) / ws2
  n_ws3_perday = (1440*60) / ws3
  #--------------------
  # get indicator of non-wear periods
  rout = IMP$rout
  r5 = as.numeric(as.matrix(rout[,5]))
  r5long = matrix(0, length(r5), (ws2/ws3))
  r5long = replace(r5long, 1:length(r5long), r5)
  r5long = t(r5long)
  dim(r5long) = c((length(r5) * (ws2/ws3)),1)
  if (nD < length(r5long)) {
    invalid = r5long[1:nD]
  } else {
    invalid = c(r5long, rep(0, (nD - length(r5long))))
  }
  rm(r5,rout)
  # Deriving file characteristics from 15 min summary files
  ND = nD/n_ws3_perday #number of days
  if (ND > 0.2) {
    #========================================================================
    # timestamps
    time = format(IMP$metashort[,1])
    fix_NA_invector = function(x){
      if (length(which(is.na(x) == TRUE)) > 0) {
        x[which(is.na(x) == T)] = 0
      }
      return(x)
    }
    anglez = as.numeric(as.matrix(IMP$metashort[,which(colnames(IMP$metashort) == "anglez")]))
    anglez = fix_NA_invector(anglez)
    anglex = angley = c()
    do.HASPT.hip = FALSE
    if (sensor.location == "hip" &
        "anglex" %in% colnames(IMP$metashort) &
        "angley" %in% colnames(IMP$metashort) &
        "anglez" %in% colnames(IMP$metashort)) {
      do.HASPT.hip = TRUE
      if (length(colnames(IMP$metashort) == "anglex") > 0) {
        anglex =  IMP$metashort[, which(colnames(IMP$metashort) == "anglex")]
        anglex = fix_NA_invector(anglex)
      }
      if (length(colnames(IMP$metashort) == "angley") > 0) {
        angley = IMP$metashort[,which(colnames(IMP$metashort) == "angley")]
        angley = fix_NA_invector(angley)
      }
    }
    if (acc.metric %in% colnames(IMP$metashort) == FALSE) {
      if ("ExtAct" %in% colnames(IMP$metashort) == TRUE) {
        acc.metric = "ExtAct"
      } else {
        stop("Argument acc.metric is set to ",acc.metric," but not found in GGIR part 1 output data")
      }
    }
    ACC = as.numeric(as.matrix(IMP$metashort[,which(colnames(IMP$metashort) == acc.metric)]))
    night = spt_crude_estimate = rep(0, length(ACC))
    
    if ("marker" %in% colnames(IMP$metashort)) {
      MARKER = as.numeric(as.matrix(IMP$metashort[,which(colnames(IMP$metashort) == "marker")]))
    } else {
      MARKER = NULL
    }
    sib_90s_algo_names = c("Sadeh1994", "Galland2012", "ColeKripke1992", "Oakley1997")
    if (any(params_sleep[["HASIB.algo"]] %in% sib_90s_algo_names)) { # extract zeroCrossingCount
      # note that for external derived metrics we refer to it as ZCY here
      # even though the exact calculation may have differed
      if (params_sleep[["Sadeh_axis"]] %in% c("X", "Y", "Z") == FALSE) params_sleep[["Sadeh_axis"]] = "Z"
      count_column_index = which(colnames(IMP$metashort) %in% c(paste0("ZC",
                                                                       params_sleep[["Sadeh_axis"]]),
                                                                "ExtAct"))[1]
      zeroCrossingCount =  IMP$metashort[, count_column_index]
      zeroCrossingCount = fix_NA_invector(zeroCrossingCount)
      zeroCrossingCount = zeroCrossingCount * zc.scale
      # optionally add NeishabouriCounts for comparison
      NeishabouriCount_colname = paste0("NeishabouriCount_", tolower(params_sleep[["Sadeh_axis"]]))
      if (NeishabouriCount_colname %in% colnames(IMP$metashort)) {
        NeishabouriCount =  IMP$metashort[, NeishabouriCount_colname]
        NeishabouriCount = fix_NA_invector(NeishabouriCount)
      } else {
        NeishabouriCount = c()
      }
    } else {
      zeroCrossingCount = c()
      NeishabouriCount = c()
    }
    #==================================================================
    # 'sleep' detection if sleep is not provided by external function.
    # Note that inside the code we call it sustained inactivity bouts
    # to emphasize that we know that this is not actually neurological sleep
    getSleepFromExternalFunction = FALSE
    if (length(myfun) != 0) {
      if ("wake_sleep" %in% myfun$colnames) {
        if (myfun$outputtype[which(myfun$colnames == "wake_sleep")] == "character") {
          getSleepFromExternalFunction = TRUE
          sleepColName = "wake_sleep"
          sleepColType = "character"
        }
      }
      if ("ExtSleep" %in% myfun$colnames) {
        if (params_sleep[["HASIB.algo"]][1] == "data") {
          getSleepFromExternalFunction = TRUE
          sleepColName = "ExtSleep"
          sleepColType = "numeric"
        } else {
          if (length(params_sleep[["HASIB.algo"]]) == 1 && params_sleep[["HASIB.algo"]] == "data") {
            stop(paste0("No sleep classification found in input data. ",
                        " please provide second sib algorithm to HASIB.algo"), call. = FALSE)
          } else if (length(params_sleep[["HASIB.algo"]]) == 2) {
            params_sleep[["HASIB.algo"]] = params_sleep[["HASIB.algo"]][2]
          }
        }
      }
    }
    if (getSleepFromExternalFunction == FALSE) {
      sleep = HASIB(HASIB.algo = params_sleep[["HASIB.algo"]], 
                    timethreshold = params_sleep[["timethreshold"]],
                    anglethreshold = params_sleep[["anglethreshold"]], 
                    time = time, anglez = anglez, ws3 = ws3,
                    zeroCrossingCount = zeroCrossingCount,
                    NeishabouriCount = NeishabouriCount, activity = ACC,
                    oakley_threshold = params_sleep[["oakley_threshold"]])
    } else { # getSleepFromExternalFunction == TRUE
      # Code now uses the sleep estimates from the external function
      # So, the assumption is that the external function provides a 
      # character "Sleep" when it detects sleep
      sleep = matrix(0, nD, 1)
      if (sleepColType == "character") {
        sleep[which(M$metashort[, sleepColName] == "Sleep")] = 1 
      } else {
        sleep[which(M$metashort[, sleepColName] == 1)] = 1 
      }
    }
    #-------------------------------------------------------------------
    # detect midnights
    detemout = g.detecmidnight(time, desiredtz, dayborder = 0) # ND, # for sleep dayborder is always 0, it is the summary of sleep that will be dayborder specific
    midnights = detemout$midnights
    midnightsi = detemout$midnightsi
    countmidn = length(midnightsi)
    tib.threshold = SPTE_end = SPTE_start = L5list = part3_guider = rep(NA,countmidn)
    if (countmidn != 0) {
      if (countmidn == 1) {
        tooshort = 1
        lastmidnight = midnights[length(midnights)]
        lastmidnighti = midnightsi[length(midnights)]
        firstmidnight = time[1]
        firstmidnighti = midnightsi[1]
      } else {
        cut = which(as.numeric(midnightsi) == 0)
        if (length(cut) > 0) {
          midnights = midnights[-cut]
          midnightsi = midnightsi[-cut]
        }
        lastmidnight = midnights[length(midnights)]
        lastmidnighti = midnightsi[length(midnights)]
        firstmidnight = midnights[1]
        firstmidnighti = midnightsi[1]
      }
      # if recording started before 4am, then also derive first awakening
      first4am = grep("04:00:00", time[1:pmin(nD, (n_ws3_perday + 1))])[1]
      # only do this if there is a 4am in the recording
      if (!is.na(first4am)) {
        if (first4am < firstmidnighti) { # this means recording started after midnight and before 4am
          midn_start = 0
        } else {
          midn_start = 1
        }
      } else {
        # since there's no 4am in the recording, then even if there is a midnight, skip this midnight. 
        # (basically treat this midnight like any other midnight without a preceding 4am)
        midn_start = countmidn
      }
      sptei = 0
      if (length(MARKER) > 0) {
        # derive typical marker button times
        # to fall back on when marker button is missing
        button_pressed = which(MARKER != 0)
        if (length(button_pressed) > 0) {
          if (params_sleep[["impute_marker_button"]] == TRUE) {
            newmarkers = button_pressed
            for (j in -countmidn:countmidn) {
              newmarkers = button_pressed + (j * 24 * (3600/ws3))
              newmarkers = newmarkers[which(newmarkers %in% button_pressed == FALSE &
                                              newmarkers > 0 &
                                              newmarkers <= length(MARKER))]
              if (length(newmarkers) > 0) {
                MARKER[newmarkers] = 0.9 / (abs(j) + 1)
              }
            }
          }
        }
      }
      for (j in midn_start:(countmidn)) { #Looping over the midnight
        if (j == 0) {
          qqq1 = 1 # preceding noon (not available in recording)
          qqq2 = midnightsi[1] + (twd[1] * (3600 / ws3))# first noon in recording
        } else {
          qqq1 = midnightsi[j] + (twd[1] * (3600 / ws3)) + 1 #preceding noon
          qqq2 = midnightsi[j] + (twd[2] * (3600 / ws3)) #next noon
        }
        # twd assumed 24 hour window, which is not the case for DST
        if (qqq2 < length(time) & qqq2 > 0) {
          qqq2_hour = as.numeric(format(iso8601chartime2POSIX(time[qqq2], tz = desiredtz), "%H"))
          if (qqq2_hour == 11) {
            qqq2 = qqq2 + (3600 / ws3)
          } else if (qqq2_hour == 13) {
            qqq2 = qqq2 - (3600 / ws3)
          }
        }
        sptei = sptei + 1
        if (qqq2 - qqq1 < 60) next # skip night if it has less than 60 epochs
        if (qqq2 > length(time))  qqq2 = length(time)
        if (qqq1 < 1)             qqq1 = 1
        if (qqq1 == 1 && qqq2 != 24 * 3600 / ws3) {
          partialFirstDay = TRUE
        } else {
          partialFirstDay = FALSE
        }
        tSegment = qqq1:qqq2
        night[tSegment] = sptei
        detection.failed = FALSE
        # Calculate nonwear percentage for this window
        nonwear_percentage = (length(which(invalid[tSegment] == 1)) /  (qqq2 - qqq1 + 1)) * 100
        guider_to_use = decide_guider(params_sleep[["HASPT.algo"]], nonwear_percentage)
        #------------------------------------------------------------------
        # calculate L5 because this is used as back-up approach
        tmpACC = ACC[tSegment]
        windowRL = round((3600/ws3) * 5)
        if ((windowRL / 2) == round(windowRL / 2)) windowRL = windowRL + 1
        if (length(tmpACC) <= windowRL) {
          L5 = 0
        } else {
          ZRM = zoo::rollmean(x = tmpACC, k = windowRL, fill = "extend", align = "center")
          L5 = which(ZRM == min(ZRM))[1]
          if (sd(ZRM) == 0) {
            L5 = 0
          } else {
            L5 = (L5  / (3600 / ws3)) + 12
          }
          if (length(L5) == 0) L5 = 0 #if there is no L5, because full day is zero
        }
        L5list[sptei] = L5
        # Estimate Sleep Period Time window, because this will be used by g.part4 if sleeplog is not available
        tmpANGLE = anglez[tSegment]
        tmpTIME = time[tSegment]
        daysleep_offset = 0
        if (do.HASPT.hip == TRUE & params_sleep[["HASPT.algo"]][guider_to_use] != "NotWorn") {
          if (params_sleep[["longitudinal_axis"]] == 1) {
            tmpANGLE = anglex[tSegment]
          } else if (params_sleep[["longitudinal_axis"]] == 2) {
            tmpANGLE = angley[tSegment]
          }
        }
        if (length(params_sleep[["def.noc.sleep"]]) == 1) {
          spt_estimate = HASPT(angle = tmpANGLE, ws3 = ws3,
                               params_sleep = params_sleep,
                               HASPT.algo = params_sleep[["HASPT.algo"]][guider_to_use],
                               invalid = invalid[tSegment],
                               activity = tmpACC,
                               marker = MARKER[tSegment],
                               sibs = sleep[tSegment, 1])
        } else {
          spt_estimate = list(SPTE_end = NULL, SPTE_start = NULL, tib.threshold = NULL, part3_guider = NULL)
        }
        tSegment_backup = tSegment
        if (length(spt_estimate$SPTE_end) != 0 & length(spt_estimate$SPTE_start) != 0) {
          if (spt_estimate$SPTE_end + qqq1 >= qqq2 - (1 * (3600 / ws3))) {
            # if estimated SPT ends within one hour of noon, re-run with larger window
            # to be able to detect sleep in daysleepers
            daysleep_offset = 6 # hours in which the window of data sent to SPTE is moved fwd from noon
            newqqq1 = qqq1 + (daysleep_offset * (3600 / ws3))
            newqqq2 = qqq2 + (daysleep_offset * (3600 / ws3))
            if (qqq1 == 1 && newqqq2 - newqqq1 < (24*3600) / ws3 && newqqq2 > (24*3600) / ws3) {
              newqqq1 = newqqq2 - (24 * 3600) / ws3
              partialFirstDay = FALSE
            }
            if (newqqq2 > length(anglez)) newqqq2 = length(anglez)
            # only try to extract SPT again if it is possible to extract a window of more than 23 hour
            if (newqqq2 < length(anglez) & (newqqq2 - newqqq1) > (23 * (3600/ws3)) ) {
              tSegment = newqqq1:newqqq2
              # Recalculate nonwear percentage for new window (6pm to 6pm)
              nonwear_percentage = (length(which(invalid[tSegment] == 1)) /  (newqqq2 - newqqq1 + 1)) * 100
              guider_to_use = decide_guider(params_sleep[["HASPT.algo"]], nonwear_percentage)
              # new TIME (6pm tp 6pm)
              tmpTIME = time[tSegment]
              if (params_sleep[["HASPT.algo"]][guider_to_use] != "NotWorn") {
                tmpANGLE = anglez[tSegment]
                if (do.HASPT.hip == TRUE) {
                  if (params_sleep[["longitudinal_axis"]] == 1) {
                    tmpANGLE = anglex[tSegment]
                  } else if (params_sleep[["longitudinal_axis"]] == 2) {
                    tmpANGLE = angley[tSegment]
                  }
                }
              }
              spt_estimate_tmp = HASPT(angle = tmpANGLE, ws3 = ws3,
                                       params_sleep = params_sleep,
                                       HASPT.algo = params_sleep[["HASPT.algo"]][guider_to_use],
                                       invalid = invalid[tSegment],
                                       activity = ACC[tSegment],
                                       sibs = sleep[tSegment, 1])
             
              if (length(spt_estimate_tmp$SPTE_start) > 0) {
                # If new SPTE_end is beyond noon (qqq2) then use the new SPTE_end
                if (spt_estimate_tmp$SPTE_end + newqqq1 >= qqq2) {
                  spt_estimate = spt_estimate_tmp
                } else {
                  daysleep_offset  = 0
                }
              } else {
                daysleep_offset  = 0
              }
            } else {
              daysleep_offset  = 0
            }
          }
          if (qqq1 == 1 && partialFirstDay == TRUE) {  
            # only use startTimeRecord if the start of the block send into SPTE was after noon
            startTimeRecord = unlist(iso8601chartime2POSIX(IMP$metashort$timestamp[1], tz = desiredtz))
            startTimeRecord = sum(as.numeric(startTimeRecord[c("hour", "min", "sec")]) / c(1, 60, 3600))
            daysleep_offset = daysleep_offset + startTimeRecord
          } else {
            daysleep_offset = daysleep_offset + 12
          }
          SPTE_end[sptei] = (spt_estimate$SPTE_end / (3600 / ws3)) + daysleep_offset
          SPTE_start[sptei] = (spt_estimate$SPTE_start / (3600 / ws3)) + daysleep_offset
          SPTE_dst = dstime_handling_check(tmpTIME = tmpTIME, spt_estimate = spt_estimate,
                                           tz = desiredtz, 
                                           calc_SPTE_end = SPTE_end[sptei],
                                           calc_SPTE_start = SPTE_start[sptei])
          SPTE_start[sptei] = SPTE_dst[1]
          SPTE_end[sptei] = SPTE_dst[2]
          tib.threshold[sptei] = spt_estimate$tib.threshold
          part3_guider[sptei] = spt_estimate$part3_guider
        }
        # log crude estimate as well:
        if (length(spt_estimate$spt_crude_estimate) == length(tSegment)) {
          spt_crude_estimate[tSegment] = spt_estimate$spt_crude_estimate
        } else if (length(spt_estimate$spt_crude_estimate) == length(tSegment_backup)) {
          spt_crude_estimate[tSegment_backup] = spt_estimate$spt_crude_estimate
        } else {
          if (!is.null(spt_estimate$spt_crude_estimate)) {
            warning("Crude estimate of sleep has unexpected length, please contact GGIR maintainer.")
          }
        }
      }
      detection.failed = FALSE
    } else {
      message("No midnights found in file")
      detection.failed = TRUE
    }
    metatmp = data.frame(time, invalid, night = night, sleep = sleep, stringsAsFactors = T)
    if (!is.null(spt_crude_estimate)) {
      metatmp$spt_crude_estimate = spt_crude_estimate
    }
  } else {
    metatmp = L5list = SPTE_end = SPTE_start = tib.threshold = part3_guider = NULL
    detection.failed = TRUE
  }
  invisible(list(output = metatmp, detection.failed = detection.failed, L5list = L5list,
                 SPTE_end = SPTE_end, SPTE_start = SPTE_start,
                 tib.threshold = tib.threshold, longitudinal_axis = params_sleep[["longitudinal_axis"]],
                 part3_guider = part3_guider))
}

g.getM5L5 = function(varnum, epochSize, t0_LFMF, t1_LFMF, M5L5res, winhr, 
                     qM5L5  = c(), iglevels = c(), MX.ig.min.dur = 10,
                     UnitReScale = 1000) {
  #diurnal pattern features extracted only meaningful if more than 16 hours
  meanVarnum = mean(varnum)
  do.M5L5 = meanVarnum > 0 # & length(varnum) > 1440*(60/ws3)
  nwindow_f = (t1_LFMF - winhr) - t0_LFMF #number of windows for L5M5 analyses
  if (length(do.M5L5) == 0 | is.na(do.M5L5) == TRUE | nwindow_f < 1) do.M5L5 = FALSE
  
  if (do.M5L5 == TRUE) { # only do the analysis if varnum has values other than zero
    reso = M5L5res #resolution
    nwindow_f = nwindow_f * (60/reso)
    DAYrunav5 = matrix(NA,nwindow_f,1)
    first_hri = (t0_LFMF * (60/reso))
    last_hri = min(nrow(DAYrunav5), floor(((t1_LFMF - winhr) * (60/reso)) - 1))
    for (hri in first_hri:last_hri) { #e.g.9am-9pm
      e1 = (hri * reso * (60/epochSize)) + 1 #e.g. 9am
      e2 = (hri + (winhr * (60/reso))) * reso * (60/epochSize) #e.g. 9am + 5 hrs
      if (e2 <= length(varnum)) {
        einclude = e1:e2
      } else { #allow for analyses beyond end of day 
        einclude = c(1:(e2 - length(varnum)), e1:length(varnum))
      }
      DAYrunav5[((hri - (t0_LFMF * (60/reso))) + 1),1] = mean(varnum[einclude])
    }
    valid = which(is.na(DAYrunav5) == F)
    DAYL5HOUR = ((which(DAYrunav5 == min(DAYrunav5[valid], na.rm = T) &
                          is.na(DAYrunav5) == F) - 1)/(60/reso)) + t0_LFMF #- 1
    DAYL5VALUE = min(DAYrunav5[valid]) * UnitReScale
    DAYM5HOUR = ((which(DAYrunav5 == max(DAYrunav5[valid], na.rm = T) & is.na(DAYrunav5) == F) - 1)/(60/reso)) + t0_LFMF #- 1
    DAYM5VALUE = max(DAYrunav5[valid]) * UnitReScale

    #-------------------------------------
    if (length(DAYL5VALUE) > 1) { DAYL5VALUE = sort(DAYL5VALUE)[ceiling(length(DAYL5VALUE)/2)] }
    if (length(DAYL5HOUR) > 1) { DAYL5HOUR = sort(DAYL5HOUR)[ceiling(length(DAYL5HOUR)/2)] }
    if (length(DAYM5VALUE) > 1) { DAYM5VALUE = sort(DAYM5VALUE)[ceiling(length(DAYM5VALUE)/2)] }
    if (length(DAYM5HOUR) > 1) { DAYM5HOUR = sort(DAYM5HOUR)[ceiling(length(DAYM5HOUR)/2)] }
    M5L5vars = data.frame(DAYL5HOUR = DAYL5HOUR[1],
                          DAYL5VALUE = DAYL5VALUE, DAYM5HOUR = DAYM5HOUR[1],
                          DAYM5VALUE = DAYM5VALUE, stringsAsFactors = TRUE)
  } else {
    M5L5vars = data.frame(DAYL5HOUR = NA, DAYL5VALUE = NA, 
                          DAYM5HOUR = NA, DAYM5VALUE = NA, stringsAsFactors = TRUE)
  }
  ML5N = c(paste0("L", winhr,"hr"), paste0("L", winhr),
           paste0("M", winhr, "hr"), paste0("M", winhr))
  names(M5L5vars) = ML5N
  if ((length(iglevels) > 0 || length(qM5L5 ) > 0) && do.M5L5 == TRUE) { #
    # get indices of window
    L5start = (DAYL5HOUR * (3600 / epochSize)) + 1 # (hour * (number of epochSize epochs in an hour)) + 1 because first epoch is one and not zero
    L5end = L5start + (winhr * (3600 / epochSize))
    M5start = (DAYM5HOUR * (3600 / epochSize)) + 1 # (hour * (number of epochSize epochs in an hour)) + 1 because first epoch is one and not zero
    M5end = M5start + (winhr * (3600 / epochSize))
  }
  # added 24 hours to create continuous scale to achieve more meaningful aggregation at person level
  if (do.M5L5 == TRUE) {
    if (!is.null(M5L5vars[1]) &&
        is.na(M5L5vars[1]) == FALSE &&
        M5L5vars[1] < 12) {
      M5L5vars[1] = M5L5vars[1] + 24 
    }
  }
  #-----------------------
  if (length(iglevels) > 0 && length(MX.ig.min.dur) == 1 && do.M5L5 == TRUE) { # intensity gradient (as described by Alex Rowlands 2018)
    if (winhr >= MX.ig.min.dur) {
      for (li in 1:2) { # do twice, once for LX and once for MX
        q49 = c()
        if (li == 1) {
          q50 = cut(varnum[L5start:L5end] * UnitReScale, breaks = iglevels, right = FALSE)
        } else {
          q50 = cut(varnum[M5start:M5end] * UnitReScale, breaks = iglevels, right = FALSE)
        }
        q50 = table(q50)
        q49  = (as.numeric(q50) * epochSize)/60 #converting to minutes
        x_ig = zoo::rollmean(iglevels, k = 2)
        y_ig = q49
        igout = g.intensitygradient(x_ig, y_ig) # produces three values
        varnameig = paste0(paste0(ifelse(li == 1, yes = "L", no = "M"), winhr, "_"),
                           c("ig_gradient", "ig_intercept", "ig_rsquared"))
        if (length(varnum) > 0) {
          M5L5vars = as.data.frame(c(M5L5vars, as.vector(unlist(igout))), stringsAsFactors = TRUE)
        } else {
          M5L5vars = as.data.frame(c(M5L5vars, rep("", 3)), stringsAsFactors = TRUE)
        }
        names(M5L5vars) = c(ML5N, varnameig)
      }
    }
  }
  #-----------------------
  if (length(qM5L5 ) > 0) {
    if (do.M5L5 == TRUE) {
      # calculate statistics on acceleration metrics
      L5q = quantile(varnum[L5start:L5end], probs = qM5L5 , na.rm = TRUE)
      M5q = quantile(varnum[M5start:M5end], probs = qM5L5 , na.rm = TRUE)
      M5L5varsExtra = as.numeric(c(L5q, M5q)) * UnitReScale
    } else {
      M5L5varsExtra = rep(NA, length(qM5L5) * 2)
    }
    M5L5varsExtraNames = c(paste0("L", winhr, "_q", round(qM5L5 * 100)),
                           paste0("M", winhr, "_q", round(qM5L5 * 100)))
    # add to M5L5vars data.frame
    M5L5vars = as.data.frame(c(M5L5vars, M5L5varsExtra), stringsAsFactors = TRUE)
    names(M5L5vars) = c(ML5N, M5L5varsExtraNames)
  }
  return(M5L5vars)
}

extract_params = function(params_sleep = c(), params_metrics = c(),
                          params_rawdata = c(), params_247 = c(),
                          params_phyact = c(), params_cleaning = c(),
                          params_output = c(), params_general = c(), input = c(), configfile_csv = c(),
                          params2check = c("sleep", "metrics", "rawdata", "247", "phyact",
                                           "cleaning", "output", "general")) {
  # Order of priority using parameters:
  # 1. Arguments provided as direct input by user
  # 2. Arguments provided via configuration file
  # 3. Default values in parameter objects
  #==================================================================================
  # Get default values in parameter objects
  if (length(params_sleep) == 0) {
    params = load_params(topic = "sleep")
    params_sleep = params$params_sleep
  }
  if (length(params_metrics) == 0) {
    params = load_params(topic = "metrics")
    params_metrics = params$params_metrics
  }
  if (length(params_rawdata) == 0) {
    params = load_params(topic = "rawdata")
    params_rawdata = params$params_rawdata
  }
  if (length(params_247) == 0) {
    params = load_params(topic = "247")
    params_247 = params$params_247
  }
  if (length(params_phyact) == 0) {
    params = load_params(topic = "phyact")
    params_phyact = params$params_phyact
  }
  if (length(params_cleaning) == 0) {
    params = load_params(topic = "cleaning")
    params_cleaning = params$params_cleaning
  }
  if (length(params_output) == 0) {
    params = load_params(topic = "output")
    params_output = params$params_output
  }
  if (length(params_general) == 0) {
    params = load_params(topic = "general")
    params_general = params$params_general
  }
  #==================================================================================
  # Overwrite them by arguments provided via configuration file
  if (length(configfile_csv) > 0) {
    config = data.table::fread(file = configfile_csv, stringsAsFactors = FALSE, data.table = FALSE)
    argNames = names(input)
    if (nrow(config) > 1) {
      for (ci in 1:nrow(config)) {
        varName = as.character(config[ci, 1])
        if (varName %in% c(argNames, "") == FALSE) {
          # only use config file values if argument is not provided as argument to g.shell.GGIR and if no empty
          
          # establish variable class
          conv2logical = conv2num = c()
          suppressWarnings(try(expr = {conv2num = as.numeric(config[ci,2])},silent = TRUE))
          suppressWarnings(try(expr = {conv2logical = as.logical(config[ci,2])},silent = TRUE))
          if (length(conv2num) > 0) {
            numi = is.na(conv2num) == FALSE
          } else {
            numi = FALSE
          }
          logi = FALSE
          if (numi == FALSE & is.na(conv2logical) == FALSE) logi = TRUE
          newValue = "notfound"
          if (logi == TRUE) {
            newValue = as.logical(config[ci,2])
          } else if (numi == TRUE) {
            newValue = as.numeric(config[ci,2])
          } else if (numi == FALSE & logi == FALSE) {
            if (length(config[ci,2]) > 0 & !is.na(config[ci,2])) {
              if (config[ci,2] == 'c()') {
                if (config[ci,1] == "def.noc.sleep") def.noc.sleep = c()
                if (config[ci,1] == "backup.cal.coef") backup.cal.coef = c()
                # Note that we do not re-assign the c(), because they are the default for most arguments that
                # can hold a c() anyway. def.noc.sleep is the only exception.
              } else if (as.character(config[ci,2]) == "NULL") {
                newValue = "c()"
              } else if (config[ci,2] != 'c()' & as.character(config[ci,2]) != "NULL") {
                if (grepl("c\\(", config[ci,2])) { # vector
                  tmp = c(gsub(pattern = "c|\\(|\\)", x = config[ci,2], replacement = ""))
                  tmp = unlist(strsplit(tmp, ","))
                  suppressWarnings(try(expr = {conv2logical = all(as.logical(tmp))},silent = TRUE))
                  suppressWarnings(try(expr = {conv2num = all(!is.na(as.numeric(tmp)))},silent = TRUE))
                  if (is.na(conv2logical)) conv2logical = FALSE
                  if (is.na(conv2num)) conv2num = FALSE
                  if (conv2logical == TRUE) {
                    newValue = as.logical(tmp)
                  } else if (conv2num == TRUE) {
                    newValue = as.numeric(tmp)
                  } else {
                    newValue = tmp
                  }
                } else {
                  newValue = config[ci,2]
                }
              }
            }
          }
          
          # Ignore arguments that are irrelevant or related to deprecated code
          # Note VvH 7 Dec 2022: I have added closedbout and boutmetric
          # because for the time being many groups may still have this in
          # their config.csv files. Eventuallythese can be removed here, which will
          # trigger an error for anyone who still uses config file with those arguments.
          ArgNames2Ignore = c("f0", "f1", "studyname", "datadir",
                              "outputdir", "do.report", "R_version",
                              "GGIR_version", "GGIRversion", "config_file", "mode",
                              "config_file_in_outputdir", "imputeTimegaps",
                              "argNames", "dupArgNames", "do.sgAccEN", "do.sgAnglex",
                              "do.sgAngley", "do.sgAnglez", "frag.classes.spt", "i",
                              "isna", "tmp", "vecchar", "dupi", "GGIRread_version",
                              "closedbout", "bout.metric", "sleeplogidnum", "LC_TIME_backup",
                              "constrain2range", "is_readxl_installed")
          # Find argument in the various parameter objects
          if (newValue[1] != "notfound" & varName %in% ArgNames2Ignore == FALSE) {
            if (varName %in% names(params_general)) {
              params_general[[varName]] = newValue
            } else {
              if (varName %in% names(params_rawdata)) {
                params_rawdata[[varName]] = newValue
              } else {
                if (varName %in% names(params_metrics)) {
                  params_metrics[[varName]] = newValue
                } else {
                  if (varName %in% names(params_sleep)) {
                    params_sleep[[varName]] = newValue
                  } else {
                    if (varName %in% names(params_phyact)) {
                      params_phyact[[varName]] = newValue
                    } else {
                      if (varName %in% names(params_output)) {
                        params_output[[varName]] = newValue
                      } else {
                        if (varName %in% names(params_247)) {
                          params_247[[varName]] = newValue
                        } else {
                          if (varName %in% names(params_cleaning)) {
                            params_cleaning[[varName]] = newValue
                          } else {
                            warning(paste0("\nIgnoring argument/parameter ", varName,
                                           " as stored in the configuration file",
                                           " as no (longer) used by GGIR. To avoid this",
                                           " warning remove the corresponding rows",
                                           " from the config.csv file."), call. = FALSE)
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  #==================================================================================
  # Overwrite them by arguments provided as direct input by user
  if (length(input) > 0) {
    argNames = names(input)
    if (exists("relyonsleeplog") == TRUE)  params_sleep[["relyonguider"]] = params_sleep[["relyonsleeplog"]]
    update_params = function(x, aN, input) {
      if (is.null(input[[aN]])) {
        x[aN] = list(NULL)
      } else {
        x[[aN]] = input[[aN]]
      }
      return(x)
    }
    expected_sleep_params = names(params_sleep)
    expected_metrics_params = names(params_metrics)
    expected_rawdata_params = names(params_rawdata)
    expected_247_params = names(params_247)
    expected_phyact_params = names(params_phyact)
    expected_cleaning_params = names(params_cleaning)
    expected_output_params = names(params_output)
    expected_general_params = names(params_general)
    for (aN in argNames) {
      if (aN %in% expected_sleep_params == TRUE) { # Sleep
        params_sleep = update_params(x = params_sleep, aN, input)
      } else if (aN %in% expected_metrics_params == TRUE) { # Metrics
        params_metrics = update_params(x = params_metrics, aN, input)
      } else if (aN %in% expected_rawdata_params == TRUE) { # Rawdata
        params_rawdata = update_params(x = params_rawdata, aN, input)
      } else if (aN %in% expected_247_params == TRUE) { # 247
        params_247 = update_params(x = params_247, aN, input)
      } else if (aN %in% expected_phyact_params == TRUE) { # phyact
        params_phyact = update_params(x = params_phyact, aN, input)
      } else if (aN %in% expected_cleaning_params == TRUE) { # cleaning
        if (aN == "data_masking_strategy") { # account for this parameter name change
          params_cleaning[["strategy"]] = input[[aN]]
        }
        params_cleaning = update_params(x = params_cleaning, aN, input)
      } else if (aN %in% expected_output_params == TRUE) { # output
        params_output = update_params(x = params_output, aN, input)
      } else if (aN %in% expected_general_params == TRUE) { # general
        params_general = update_params(x = params_general, aN, input)
      }
    }
  }
  #==================================================================================
  # Check that all parameter values have expect data class
  # and perform some checks on reasonable parameter combinations
  # prevent checking of parameters that are not used in GGIR part
  if (!"sleep" %in% params2check) params_sleep = c()
  if (!"metrics" %in% params2check) params_metrics = c()
  if (!"rawdata" %in% params2check) params_rawdata = c()
  if (!"247" %in% params2check) params_247 = c()
  if (!"phyact" %in% params2check) params_phyact = c()
  if (!"cleaning" %in% params2check) params_cleaning = c()
  if (!"output" %in% params2check) params_output = c()
  if (!"general" %in% params2check) params_general = c()
  
  params = check_params(params_sleep = params_sleep, params_metrics = params_metrics,
                        params_rawdata = params_rawdata, params_247 = params_247,
                        params_phyact = params_phyact, params_cleaning = params_cleaning,
                        params_output = params_output, params_general = params_general)
  return(params)
}

MXLX = function(Y = NULL, X = 5, epochSize = 1, tseg = c(0, 24), resolutionMin = 10) {
  # Y 24 hours of data
  # tseg is vector of length 2 with the start and end time as hour in the day
  # X is the window size in hours to look for
  
  do.MXLX = mean(Y, na.rm = TRUE) > 0
  Nwindows = (tseg[2] - X) - tseg[1]
  if (length(do.MXLX) == 0 ||
      is.na(do.MXLX) == TRUE ||
      Nwindows < 1 ||
      length(Y) < length(X) * (3600 / epochSize)) {
    do.MXLX = FALSE
  }
  
  nStepsPerHour = 60 / resolutionMin # number of steps per hour
  nEpochsPerStep = (resolutionMin * 60) / epochSize # number of epochs per step
  if (do.MXLX == TRUE) { # only do the analysis if Y has values other than zero
    Y = Y[((((tseg[1] - tseg[1]) * 3600)/epochSize) + 1):(((tseg[2] - tseg[1]) * 3600)/epochSize)]
    Nwindows = ceiling(Nwindows * nStepsPerHour)
    rollingMean = matrix(NA, Nwindows, 1)
    for (hri in 1:Nwindows) { #e.g.9am-9pm
      # start and end in terms of steps
      indexStart = hri - 1 #e.g. 9am
      indexEnd = indexStart + (X * nStepsPerHour) #e.g. 9am + 5 hrs
      # start and end in terms of Y index
      indexStart = (indexStart * nEpochsPerStep) + 1
      indexEnd = indexEnd * nEpochsPerStep
      einclude = indexStart:indexEnd
      if (indexEnd <= length(Y)) {
        rollingMean[hri,1] = mean(Y[einclude])
      }
    }
    valid = which(is.na(rollingMean) == F)
    LXvalue = min(rollingMean[valid], na.rm = T)
    MXvalue = max(rollingMean[valid], na.rm = T)
    LXhr = ((which(rollingMean == LXvalue & is.na(rollingMean) == F) - 1) / nStepsPerHour) + tseg[1]
    MXhr = ((which(rollingMean == MXvalue & is.na(rollingMean) == F) - 1) / nStepsPerHour) + tseg[1]
    
    #-------------------------------------
    if (length(LXvalue) > 1) { LXvalue = sort(LXvalue)[ceiling(length(LXvalue)/2)] }
    if (length(LXhr) > 1) { LXhr = sort(LXhr)[ceiling(length(LXhr)/2)] }
    if (length(MXvalue) > 1) { MXvalue = sort(MXvalue)[ceiling(length(MXvalue)/2)] }
    if (length(MXhr) > 1) { MXhr = sort(MXhr)[ceiling(length(MXhr)/2)] }
    
    #Note it is + 1 because first epoch is one and not zero:
    LXstart = ((LXhr - tseg[1]) * (3600 / epochSize)) + 1
    LXend = LXstart + (X * (3600 / epochSize)) - 1
    MXstart = ((MXhr - tseg[1]) * (3600 / epochSize)) + 1 
    MXend = MXstart + (X * (3600 / epochSize)) - 1
    
    MXLX = data.frame(LXhr = LXhr[1],
                      LXvalue = LXvalue,
                      LXindex0 = LXstart,
                      LXindex1 = LXend,
                      MXhr = MXhr[1],
                      MXvalue = MXvalue,
                      MXindex0 = MXstart,
                      MXindex1 = MXend, stringsAsFactors = TRUE)
  } else {
    MXLX = data.frame(LXhr = NA, LXvalue = NA, 
                      LXindex0 = NA,
                      LXindex1 = NA,
                      MXhr = NA, MXvalue = NA,
                      MXindex0 = NA,
                      MXindex1 = NA, stringsAsFactors = TRUE)
  }
  MXLXnames = c(paste0("L", X, "hr"), paste0("L", X), paste0("start_L", X),  paste0("end_L", X),
                paste0("M", X, "hr"), paste0("M", X),  paste0("start_M", X),  paste0("end_M", X))
  names(MXLX) = MXLXnames
  return(MXLX)
}

library(GGIR)
context("MXLX")

test_that("MXLX is able to identify correct mean and timing of MX and LX", {
  Y = rep(1, 1440)
  # An M5 at 1:00 with M5 == 2
  Y[61:360] = 2 
  # An L5 at 12:00 with L5 == 0
  Y[721:1020] = 0 
  # A second 'M5' at 18:00 with M5 == 1.5 which we expect to find when only
  # the second half of the data is passed on
  Y[1081:1380] = 1.5
  # Add some missing values as code should be able to handle this
  is.na(Y[20:30]) = TRUE 

  # Full 24 hours
  out1 = MXLX(Y = Y, X = 5, epochSize = 60, tseg = c(0, 24), resolutionMin = 10)
  expect_false(all(is.na(out1)))
  expect_equal(out1$L5hr, 12)
  expect_equal(out1$L5, 0)
  expect_equal(out1$start_L5, 721)
  expect_equal(out1$end_L5, 1020)
  expect_equal(out1$M5hr, 1)
  expect_equal(out1$M5, 2)
  expect_equal(out1$start_M5, 61)
  expect_equal(out1$end_M5, 360)
  
  # When selecting only the first 12 hours we still expect to find M5 at 1:00
  out3 = MXLX(Y = Y, X = 5, epochSize = 60, tseg = c(0, 12), resolutionMin = 10) 
  expect_equal(nrow(out3), 1)
  expect_equal(out3$M5hr, 1)
  expect_equal(out3$M5, 2)
  expect_equal(out3$start_M5, 61)
  expect_equal(out3$end_M5, 360)
  # and L5 is now the first 5 hour block after M5
  expect_equal(out3$L5hr, 6.3333, tolerance = 0.001)
  expect_equal(out3$L5, 1)
  expect_equal(out3$start_L5, 381)
  expect_equal(out3$end_L5, 680)
  
  # When only passing on the second 12 hours (note: Y is trimmed now)
  # we expect same L5 as 24 hours because L5 occurs in second half of the day
  # only differences are:
  # - indices are shifted to reflect new Y time series
  # - M5 is different because M5 occurs in first half of the day
  out4 = MXLX(Y = Y[721:1440], X = 5, epochSize = 60, tseg = c(12, 24), resolutionMin = 10) 
  expect_equal(nrow(out4), 1)
  expect_equal(out4$L5hr, out1$L5hr)
  expect_equal(out4$L5, out1$L5)
  expect_equal(out4$start_L5, out1$start_L5 - 720)
  expect_equal(out4$end_L5, out1$end_L5 - 720)
  expect_equal(out4$M5hr, 18)
  expect_equal(out4$M5, 1.5)
  expect_equal(out4$start_M5, 361)
  expect_equal(out4$end_M5, 660)

  # When recording is short
  Y = 1
  out = MXLX(Y = Y, X = 5, epochSize = 60, tseg = c(0, 24), resolutionMin = 10) 
  expect_equal(nrow(out), 1)
  expect_true(all(is.na(out)))
})

POSIXtime2iso8601 = function(x,tz){
  chartime2iso8601 = function(x, tz) {
    POStime = as.POSIXlt(as.numeric(as.POSIXlt(x, tz)), origin = "1970-1-1", tz)
    POStimeISO = strftime(POStime, format = "%Y-%m-%dT%H:%M:%S%z")
    return(POStimeISO)
  }
  POStime = as.POSIXlt(x,tz) #turn to right timezone
  POStime_z = chartime2iso8601(format(POStime),tz) #change format
  return(POStime_z)
}

library(GGIR)
context("POSIX")
test_that("g.impute returns a non-empty dataframe", {
  # POSIXtime2iso8601
  x ="2017-05-07 13:15:17 CEST"
  tz = "Europe/Amsterdam"
  x_converted = POSIXtime2iso8601(x,tz)
  expect_equal(as.character(x_converted),"2017-05-07T13:15:17+0200")
  
  # iso8601chartime2POSIX
  x ="2017-05-07T13:00:00+0200"
  tz = "Europe/Amsterdam"
  x_converted = iso8601chartime2POSIX(x,tz)
  expect_equal(as.character(x_converted),"2017-05-07 13:00:00")
})

g.applymetrics = function(data, sf, ws3, metrics2do,
                          n = 4, lb = 0.2, hb = 15,
                          zc.lb = 0.25, zc.hb = 3, zc.sb = 0.01,
                          zc.order = 2, 
                          actilife_LFE = FALSE){
  # data should be a 3 column matrix with the x, y, and z acceleration
  data = data[, c("x", "y", "z")]
  
  epochsize = ws3 #epochsize in seconds
  do.bfen = metrics2do$do.bfen
  do.enmo = metrics2do$do.enmo
  do.lfenmo = metrics2do$do.lfenmo
  do.en = metrics2do$do.en
  do.hfen = metrics2do$do.hfen
  do.hfenplus = metrics2do$do.hfenplus
  do.mad = metrics2do$do.mad
  do.anglex = metrics2do$do.anglex
  do.angley = metrics2do$do.angley
  do.anglez = metrics2do$do.anglez
  do.roll_med_acc_x = metrics2do$do.roll_med_acc_x
  do.roll_med_acc_y = metrics2do$do.roll_med_acc_y
  do.roll_med_acc_z = metrics2do$do.roll_med_acc_z
  do.dev_roll_med_acc_x = metrics2do$do.dev_roll_med_acc_x
  do.dev_roll_med_acc_y = metrics2do$do.dev_roll_med_acc_y
  do.dev_roll_med_acc_z = metrics2do$do.dev_roll_med_acc_z
  do.enmoa = metrics2do$do.enmoa
  do.lfen = metrics2do$do.lfen
  do.hfx = metrics2do$do.hfx
  do.hfy = metrics2do$do.hfy
  do.hfz = metrics2do$do.hfz
  do.lfx = metrics2do$do.lfx
  do.lfy = metrics2do$do.lfy
  do.lfz = metrics2do$do.lfz
  do.bfx = metrics2do$do.bfx
  do.bfy = metrics2do$do.bfy
  do.bfz = metrics2do$do.bfz
  do.zcx = metrics2do$do.zcx
  do.zcy = metrics2do$do.zcy
  do.zcz = metrics2do$do.zcz
  do.brondcounts = metrics2do$do.brondcounts
  do.neishabouricounts = metrics2do$do.neishabouricounts
  allmetrics = c()
  averagePerEpoch = function(x,sf,epochsize) {
    x2 = cumsum(c(0,x))
    select = seq(1,length(x2),by = sf*epochsize)
    x3 = diff(x2[round(select)]) / abs(diff(round(select)))
  }
  sumPerEpoch = function(x, sf, epochsize) {
    x2 = cumsum(c(0,x))
    select = seq(1,length(x2), by = sf*epochsize)
    x3 = diff(x2[round(select)])
  }
  
  if (sf <= (hb * 2)) { #avoid having a higher filter boundary higher than sf/2
    hb = round(sf/2) - 1
  }
  gravity = 1
  #================================================
  # Functions to aid metric extraction
  process_axes = function(data, filtertype, cut_point, n = 4, sf = c()) {
    if (length(sf) == 0) warning("sf not found")
    if (filtertype == "pass" | filtertype == "high" | filtertype == "low") {
      hf_lf_filter = function(bound, n, sf, filtertype) {
        return(signal::butter(n, c(bound/(sf/2)), type = filtertype))
      }
      bf_filter = function(lb, hb, n, sf) {
        Wc = matrix(0,2,1)
        Wc[1,1] = lb / (sf/2)
        Wc[2,1] = hb / (sf/2)
        if (sf/2 < hb | sf/2 < hb) {
          warning("\nSample frequency ", sf, " too low for calculating this metric.")
        }
        return(signal::butter(n, Wc, type = c("pass")))
      }
      if (filtertype == "pass") {
        coef = bf_filter(cut_point[1], cut_point[2], n, sf)
      } else {
        if (filtertype == "low") {
          bound = hb
        } else {
          bound = lb
        }
        coef = hf_lf_filter(bound, n, sf, filtertype)
      }
      data_processed = data
      for (i in 1:3) {
        data_processed[, i] = signal::filter(coef, data[,i])
      }
    } else if (filtertype == "rollmedian") {
      # rollmed = function(x, sf) {
      #   winsi = round(sf * 5)
      #   if (round(winsi/2) == (winsi/2)) winsi = winsi + 1
      #   xm = zoo::rollmedian(x, k = winsi, na.pad = TRUE)
      #   xm[which(is.na(xm[1:1000]) == TRUE)] = xm[which(is.na(xm[1:1000]) == FALSE)[1]]
      #   return(xm)
      # }
      rollmed = function(x, sf) {
        stepsize = max(c(floor(sf / 10), 1))
        newsf = ifelse(test = stepsize > 1, yes = 10, no = sf)
        winsi = round(newsf * 5)
        if (round(winsi/2) == (winsi/2)) winsi = winsi + 1
        if ((winsi %% 2) == 0) winsi = winsi + 1
        xm = zoo::rollmedian(x[seq(1, length(x), by = stepsize)],
                             k = winsi, fill = c(0, 0, 0), na.pad = FALSE)
        # replace zeros:
        S2check = max(c(sf * 60, 1000))
        # head
        xm[which(xm[1:S2check] == 0)] = xm[which(xm[1:S2check] != 0)[1]]
        # tail
        LN = length(xm)
        S2check = min(LN-1, S2check)
        xm_tail = xm[(LN - S2check):LN]
        lastvalue = xm_tail[which(xm_tail != 0)][1]
        if (length(lastvalue) == 0) lastvalue = xm[which(xm != 0)][1]
        xm_tail[which(xm_tail == 0)] = lastvalue
        xm[(LN - S2check):LN] = xm_tail
        # repeat values to get back to original sample rate
        xm = rep(xm, each = stepsize)
        LN = length(xm)
        LX = length(x)
        if (LN > LX) {
          xm = x[1:LX]
        } else if (LN < LX) {
          xm = c(xm, rep(xm[LN], LX - LN))
        }
        return(xm)
      }
      data_processed = data
      for (i in 1:3) {
        data_processed[,i] = rollmed(data[,i], sf)
      }
      if (length(which(is.na(data_processed[,1]) == TRUE |
                       is.na(data_processed[,2]) == TRUE |
                       is.na(data_processed[,3]) == TRUE)) > 0) {
        for (j in 1:3) {
          p1 = which(is.na(data_processed[,j]) == FALSE)
          data_processed[which(is.na( data_processed[,j]) == TRUE),j] =  data_processed[p1[length(p1)],j]
        }
      }
    }
    return(data_processed)
  }
  anglex = function(xyz) {
    return(atan(xyz[,1] / (sqrt(xyz[,2]^2 + xyz[,3]^2))) / (pi/180))
  }
  angley = function(xyz) {
    return(atan(xyz[,2] / (sqrt(xyz[,1]^2 + xyz[,3]^2))) / (pi/180))
  }
  anglez = function(xyz) {
    return(atan(xyz[,3] / (sqrt(xyz[,1]^2 + xyz[,2]^2))) / (pi/180))
  }
  EuclideanNorm = function(xyz) {
    return(sqrt((xyz[,1]^2) + (xyz[,2]^2) + (xyz[,3]^2)))
  }
  #==========================
  # Band-pass filtering related metrics
  if (do.bfen == TRUE | do.bfx == TRUE | do.bfy == TRUE | do.bfz == TRUE) {
    data_processed = abs(process_axes(data, filtertype = "pass", cut_point = c(lb,hb), n, sf))
    if (do.bfx == TRUE) {
      allmetrics$BFX = averagePerEpoch(x = data_processed[,1], sf, epochsize)
    }
    if (do.bfy == TRUE) {
      allmetrics$BFY = averagePerEpoch(x = data_processed[,2], sf, epochsize)
    }
    if (do.bfz == TRUE) {
      allmetrics$BFZ = averagePerEpoch(x = data_processed[,3], sf, epochsize)
    }
    if (do.bfen == TRUE) {
      allmetrics$BFEN = averagePerEpoch(x = EuclideanNorm(data_processed),sf,epochsize)
    }
  }
  if (do.zcx == TRUE | do.zcy == TRUE | do.zcz == TRUE) { # Zero crossing count
    
    # 1) apply band-pass frequency filter to mimic old-sensor
    # 0.25 - 3 Hertz to be in line with Ancoli Isreal's paper from 2003,
    # and online "Motionlogger Users Guide Version 2K1.1" from Ambulatory Monitoring,
    # Inc. Ardsley, New York 10502
    # Be aware that specific boundaries differ between Actigraph brands that copied the
    # Sadeh algorithm.
    # We use a second order filter because if it was an analog filter it was 
    # most likely not very steep filter.
    data_processed = process_axes(data, filtertype = "pass", cut_point = c(zc.lb, zc.hb), n = zc.order, sf)
    zil = c()
    
    # 2) Sadeh reported to have used the y-axis but did not specify the orientation of
    # the y-axis in their accelerometer. Therefore, we keep selection of axis flexible for the user
    if (do.zcx == TRUE) zil = 1
    if (do.zcy == TRUE) zil = c(zil, 2)
    if (do.zcz == TRUE) zil = c(zil, 3)
    Ndat = nrow(data_processed)
    for (zi in zil) { # loop over axes
      # 3) apply stop-band to minic sensitivity of 1980s and 1990s accelerometer
      # technology. Using a 0.01g threshold based on book by Tyron 
      # "Activity Measurementy in Psychology And Medicine"
      smallvalues = which(abs(data_processed[,zi]) < zc.sb)
      if (length(smallvalues) > 0) {
        data_processed[smallvalues, zi] = 0
      }
      rm(smallvalues)
      # output binary time series zeros and ones with 1 for zero-crossing
      data_processed[,zi] = ifelse(test = data_processed[,zi] >= 0,yes = 1, no = -1)
      # 4) detect zero-crossings
      # The exact algorithm for the original monitor not found, maybe it happened analog
      # we will use: http://rug.mnhn.fr/seewave/HTML/MAN/zcr.html
      zerocross = function(x, Ndat) {
        tmp = abs(sign(x[2:Ndat]) - sign(x[1:(Ndat - 1)])) * 0.5
        tmp = c(tmp[1], tmp) # add value to ensure length aligns
        return(tmp)
      }
      if (zi == 1) {
        allmetrics$ZCX = sumPerEpoch(zerocross(data_processed[,zi], Ndat), sf, epochsize)
      } else if (zi == 2) {
        allmetrics$ZCY = sumPerEpoch(zerocross(data_processed[,zi], Ndat), sf, epochsize)
      } else if (zi == 3) {
        allmetrics$ZCZ = sumPerEpoch(zerocross(data_processed[,zi], Ndat), sf, epochsize)
      }
    }
    
    # Note that this is per epoch, in GGIR part 3 we aggregate (sum) this per minute
    # to follow Sadeh. In Sadeh 1987 this resulted in values up to 280
    # 280 = 60 x 2 x frequency of movement which would mean near 2.33 Hertz average
    # movement frequencies, which may have reflected walking
  }
  #================================================
  # Low-pass filtering related metrics
  if (do.lfenmo == TRUE | do.lfx == TRUE | do.lfy == TRUE | do.lfz == TRUE | do.lfen == TRUE) {
    data_processed = abs(process_axes(data, filtertype = "low", cut_point = hb, n, sf))
    if (do.lfx == TRUE) {
      allmetrics$LFX = averagePerEpoch(x = data_processed[,1], sf, epochsize)
    }
    if (do.lfy == TRUE) {
      allmetrics$LFY = averagePerEpoch(x = data_processed[,2], sf, epochsize)
    }
    if (do.lfz == TRUE) {
      allmetrics$LFZ = averagePerEpoch(x = data_processed[,3], sf, epochsize)
    }
    if (do.lfen == TRUE) {
      allmetrics$LFEN = averagePerEpoch(x = EuclideanNorm(data_processed), sf, epochsize)
    }
    if (do.lfenmo == TRUE) {
      LFENMO = EuclideanNorm(data_processed) - gravity
      LFENMO[which(LFENMO < 0)] = 0
      allmetrics$LFENMO = averagePerEpoch(x = LFENMO,sf,epochsize)
    }
  }
  #================================================
  # High-pass filtering related metrics
  if (do.hfen == TRUE | do.hfx == TRUE | do.hfy == TRUE | do.hfz == TRUE) {
    data_processed = abs(process_axes(data, filtertype = "high", cut_point = c(lb), n, sf))
    if (do.hfx == TRUE) {
      allmetrics$HFX = averagePerEpoch(x = data_processed[,1], sf, epochsize)
    }
    if (do.hfy == TRUE) {
      allmetrics$HFY = averagePerEpoch(x = data_processed[,2], sf, epochsize)
    }
    if (do.hfz == TRUE) {
      allmetrics$HFZ = averagePerEpoch(x = data_processed[,3], sf, epochsize)
    }
    allmetrics$HFEN = averagePerEpoch(x = EuclideanNorm(data_processed), sf, epochsize)
  }
  #================================================
  # Combined low-pass and high-pass filtering metric:)
  if (do.hfenplus == TRUE) { 
    # Note that we are using intentionally the lower boundary for the low pass filter
    data_processed = process_axes(data, filtertype = "low", cut_point = lb, n, sf)
    GCP = EuclideanNorm(data_processed) - gravity
    # Note that we are using intentionally the lower boundary for the high pass filter
    data_processed = process_axes(data, filtertype = "high", cut_point = lb, n, sf)
    HFENplus = EuclideanNorm(data_processed) + GCP
    HFENplus[which(HFENplus < 0)] = 0
    allmetrics$HFENplus = averagePerEpoch(x = HFENplus,sf,epochsize)
  }
  #================================================
  # Rolling median filter related metrics:
  roll_median_done  = FALSE
  if (do.roll_med_acc_x == TRUE | do.roll_med_acc_y == TRUE | do.roll_med_acc_z == TRUE | 
      do.dev_roll_med_acc_x == TRUE | do.dev_roll_med_acc_y == TRUE | do.dev_roll_med_acc_z == TRUE) {
    data_processed = process_axes(data, filtertype = "rollmedian", cut_point = c(), n, sf)
    roll_median_done  = TRUE
    if (do.roll_med_acc_x == TRUE | do.roll_med_acc_y == TRUE | do.roll_med_acc_z == TRUE) { #rolling median of acceleration
      allmetrics$roll_med_acc_x = averagePerEpoch(x = data_processed[,1], sf, epochsize)
      allmetrics$roll_med_acc_y = averagePerEpoch(x = data_processed[,2], sf, epochsize)
      allmetrics$roll_med_acc_z = averagePerEpoch(x = data_processed[,3], sf, epochsize)
    }
    if (do.dev_roll_med_acc_x == TRUE | do.dev_roll_med_acc_y == TRUE | do.dev_roll_med_acc_z == TRUE) { #rolling median of acceleration
      allmetrics$dev_roll_med_acc_x = averagePerEpoch(x = abs(data[,1] - data_processed[,1]), sf, epochsize)
      allmetrics$dev_roll_med_acc_y = averagePerEpoch(x = abs(data[,2] - data_processed[, 2]), sf, epochsize)
      allmetrics$dev_roll_med_acc_z = averagePerEpoch(x = abs(data[,3] - data_processed[, 3]), sf, epochsize)
    }
  }
  if (do.anglex == TRUE | do.angley == TRUE | do.anglez == TRUE) {
    if (roll_median_done == FALSE) { 
      data_processed = process_axes(data, filtertype = "rollmedian", cut_point = c(), n, sf)
    }
    if (do.anglex == TRUE) {
      allmetrics$angle_x = averagePerEpoch(x = anglex(data_processed), sf, epochsize)
    }
    if (do.angley == TRUE) {
      allmetrics$angle_y = averagePerEpoch(x = angley(data_processed), sf, epochsize)
    }
    if (do.anglez == TRUE) {
      allmetrics$angle_z = averagePerEpoch(x = anglez(data_processed), sf, epochsize)
    }
  }
  #================================================
  # Filter-free Euclidean norm related metrics:
  EN = EuclideanNorm(data)
  if (do.enmo == TRUE) {
    ENMO = EN - 1
    ENMO[which(ENMO < 0)] = 0 #turning negative values into zero
    allmetrics$ENMO = averagePerEpoch(x = ENMO, sf, epochsize)
  }
  if (do.mad == TRUE) { # metric MAD (Mean Amplitude Deviation)
    MEANS = rep(averagePerEpoch(x = EN, sf, epochsize), each = sf * epochsize)
    MAD = abs(EN - MEANS)
    allmetrics$MAD = averagePerEpoch(x = MAD, sf, epochsize)
  }
  if (do.en == TRUE) {
    allmetrics$EN = averagePerEpoch(x = EN,sf,epochsize)
  }
  if (do.enmoa == TRUE) {
    ENMOa = abs(EN - gravity)
    allmetrics$ENMOa = averagePerEpoch(x = ENMOa, sf, epochsize)
  }
  #================================================
  # Brond Counts)
  if (do.brondcounts == TRUE) {
    stop(paste0("\nThe brondcounts option has been deprecated following issues with the ",
                "following issues with the activityCounts package. We will reinsert brondcounts ",
                "once the issues are resolved."), call. = FALSE)
    # if (ncol(data) > 3) data = data[,2:4]
    # mycounts = activityCounts::counts(data = data, hertz = sf, 
    #                                   x_axis = 1, y_axis = 2, z_axis = 3,
    #                                   start_time = Sys.time()) # ignoring timestamps, because GGIR has its own timestamps
    # if (sf < 30) {
    #   warning("\nNote: activityCounts not designed for handling sample frequencies below 30 Hertz")
    # }
    # # activityCount output is per second
    # # aggregate to our epoch size:
    # allmetrics$BrondCount_x = sumPerEpoch(mycounts[, 2], sf = 1, epochsize)
    # allmetrics$BrondCount_y = sumPerEpoch(mycounts[, 3], sf = 1, epochsize)
    # allmetrics$BrondCount_z = sumPerEpoch(mycounts[, 4], sf = 1, epochsize)
  }
  #================================================
  # Actilife Counts)
  if (do.neishabouricounts == TRUE) {
    # I do not include actilifecounts as dependency to avoid overwritten methods
    # message from package signal (this package has not well defined the source 
    # package of their functions and we get an inoffensive but unwanted message 
    # printed in the console).
    # I followed suggestion in: https://bit.ly/3SaP69u
    suppressMessages(requireNamespace("actilifecounts"))
    if (ncol(data) > 3) data = data[,2:4]
    mycounts = actilifecounts::get_counts(raw = data, sf = sf,
                                          epoch = epochsize, lfe_select = actilife_LFE,
                                          verbose = FALSE)
    if (sf < 30) {
      warning("\nNote: activityCounts not designed for handling sample frequencies below 30 Hertz")
    }
    # activityCount output is per second
    # aggregate to our epoch size:
    allmetrics$NeishabouriCount_x = mycounts[, 1]
    allmetrics$NeishabouriCount_y = mycounts[, 2]
    allmetrics$NeishabouriCount_z = mycounts[, 3]
    allmetrics$NeishabouriCount_vm = mycounts[, 4]
  }
  return(allmetrics)
} 

g.part5.addsib = function(ts, epochSize, part3_output, desiredtz, sibDefinition, nightsi) {
  #========================================================
  # SUSTAINED INACTIVITY BOUTS
  # These are stored in part 3 milestone data as start- and end-times
  # in the following code we convert the them into indices of the recording sequence
  Nts = nrow(ts)
  ts$sibdetection = 0# initialize output vector that will hold the sibs
  s0s1 = c()
  # pr0 and pr1 define the indices relative the start of the recording
  # and specifying a 6 day time window 
  # always relative to the most recently processed sustained inactivity bout
  # s0 and s1 are the indices within that time window 
  # that match the start and end of the next sustained inactivity bout 
  pr0 = 1
  pr1 = pr0 + ((60/epochSize)*1440*6)
  pr2 = Nts
  if (nrow(part3_output) > 0) {
    gik.ons = format(part3_output$sib.onset.time)
    gik.end = format(part3_output$sib.end.time)
    timeChar = format(ts$time)
    if (is.ISO8601(timeChar[1]) == FALSE) { # only do this for POSIX format
      timeChar = POSIXtime2iso8601(timeChar, tz = desiredtz)
    }
    if (timeChar[1] != as.character(timeChar[1])) {
      #not s0 because s0 does not exist yet if classes differ
      timeChar = as.character(timeChar)
    }
    for (g in 1:nrow(part3_output)) { # sustained inactivity bouts
      lastpr0 = pr0
      pr1 = pr0 + ((60/epochSize)*1440*6)
      if (pr1 > pr2) pr1 = pr2
      if (pr0 > pr1) pr0 = pr1
      #Coerce time into iso8601 format, so it is sensitive to daylight saving times when hours can be potentially repeated
      timeChar_tmp = timeChar[pr0:pr1]
      s0 = which(timeChar_tmp == gik.ons[g])[1]
      s1 = which(timeChar_tmp == gik.end[g])[1]
      if (is.na(s0) == TRUE) s0 = which(timeChar_tmp == paste0(gik.ons[g], " 00:00:00"))[1]
      if (is.na(s1) == TRUE) s1 = which(timeChar_tmp == paste0(gik.end[g], " 00:00:00"))[1]
      # add pr0 to make s0 and s1 be relative to the start of the recording
      s0 = s0 + pr0 - 1
      s1 = s1 + pr0 - 1
      pr0 = s1
      if (length(s1) != 0 & length(s0) != 0 & is.na(s0) == FALSE & is.na(s1) == FALSE) {
        s0s1 = c(s0s1,s0:s1)
      } else {
        pr0 = lastpr0 + ((60/epochSize)*1440*6)
      }
    }
  }
  ts$sibdetection[s0s1] = 1
  if (length(grep(pattern = "A", sibDefinition)) > 0 &&
      length(grep(pattern = "T", sibDefinition)) > 0 &&
      length(nightsi) > 0) {
    #=========
    # If excludefirstlast was set to TRUE in part 4 then 
    # SUSTAINED INACTIVITY BOUTS are not assessed for the time between
    # the first midnight and the first noon
    # because it is not used for sleep reports (first night is ignored in this case).
    # Therefore, expand the SIB sibdetection to include this time segment, for part 5 only.
    # This part of the code is only performed if using sib estimates based on vanHees2015
    #----------------------------------------------------------------------
    # Step 1: Identify time window that needs to be processed
    redo1 = nightsi[1] - ((60/epochSize)*60) # 1 hour before first midnight
    if (redo1 < 1) redo1 = 1
    redo2 = nightsi[1] + (14*(60/epochSize)*60) # 14 hours after first midngiht
    if (redo2 > Nts) redo2 = Nts
    # Specify defintion of sustained inactivity bout
    anglethreshold = as.numeric(unlist(strsplit(sibDefinition,"A"))[2])
    tempi = unlist(strsplit(unlist(strsplit(sibDefinition,"A"))[1],"T"))
    timethreshold = as.numeric(tempi[length(tempi)])
    # ts$angle[which(is.na(ts$angle[redo1:redo2]) == T)] = 0
    if (any(is.na(ts$angle[redo1:redo2]))) {
      ts$angle[which(is.na(ts$angle[redo1:redo2]) == T)] = 0
    }
    sdl1 = rep(0,length(ts$time[redo1:redo2]))
    postch = which(abs(diff(ts$angle[redo1:redo2])) > anglethreshold) #posture change of at least j degrees
    # count posture changes that happen less than once per ten minutes
    q1 = c()
    if (length(postch) > 1) {
      q1 = which(diff(postch) > (timethreshold * (60/epochSize))) #less than once per i minutes
    }
    if (length(q1) > 0) {
      for (gi in 1:length(q1)) {
        sdl1[postch[q1[gi]]:postch[q1[gi] + 1]] = 1 #periods with no posture change
      }
    } else { #possibly a day without wearing
      if (length(postch) < 5) {  #possibly a day without wearing
        sdl1[1:length(sdl1)] = 1 #periods with no posture change
      } else {  #possibly a day with constantly posture changes
        sdl1[1:length(sdl1)] = 0 #periodsposture change
      }
    }
    # update variable sibdetection
    if (redo2 > Nts) {
      delta = redo2 - Nts
      redo2 = Nts
      sdl1 = sdl1[1:(length(sdl1) - delta)]
    }
    if (redo1 > Nts) {
      redo1 = Nts
      sdl1 = sdl1[1]
    }
    ts$sibdetection[redo1:redo2] = sdl1
  }
  return(ts)
}

g.part5_analyseSegment = function(indexlog, timeList, levelList,
                                  segments,
                                  segments_names,
                                  dsummary, ds_names,
                                  params_general, params_output,
                                  params_sleep, params_247,
                                  params_phyact,
                                  sumSleep, sibDef,
                                  fullFilename,
                                  add_one_day_to_next_date,
                                  lightpeak_available,
                                  tail_expansion_log,
                                  foldernamei, sibreport = NULL) {
  # unpack list objects:
  # indexlog
  fileIndex = indexlog$fileIndex
  timewindowi = indexlog$winType
  wi = indexlog$winIndex
  qqq = indexlog$winStartEnd
  si = indexlog$segIndex1
  current_segment_i = indexlog$segIndex2
  segStart = indexlog$segStartEnd[1]
  segEnd = indexlog$segStartEnd[2]
  fi = indexlog$columnIndex
  
  # timeList
  ts = timeList$ts
  sec = timeList$sec
  min = timeList$min
  hour = timeList$hour
  time_POSIX = timeList$time_POSIX
  ws3new = timeList$epochSize
  
  # levelList
  TRLi = levelList$threshold[1]
  TRMi = levelList$threshold[2]
  TRVi = levelList$threshold[3]
  LEVELS = levelList$LEVELS
  Lnames = levelList$Lnames
  OLEVELS = levelList$OLEVELS
  bc.mvpa = levelList$bc.mvpa
  bc.in = levelList$bc.in
  bc.lig = levelList$bc.lig
  
  skiponset = skipwake = TRUE
  
  #==========================
  # The following is to avoid issue with merging sleep variables from part 4
  # Note that this means that for MM windows there can be multiple or no wake or onsets in window
  date = as.Date(ts$time[qqq[1]], tz = params_general[["desiredtz"]]) # changed segStart for qqq[1] in case segment is not available in this day
  if (add_one_day_to_next_date == TRUE & timewindowi %in% c("WW", "OO")) { # see below for explanation
    date = date + 1
    add_one_day_to_next_date = FALSE
  }
  weekday = weekdays(date, abbreviate = FALSE)
  dsummary[si,fi:(fi + 1)] = c(weekday, as.character(date))
  ds_names[fi:(fi + 1)] = c("weekday", "calendar_date"); fi = fi + 2
  # window and segment of the day
  segmentName = segments_names[current_segment_i]
  segmentTiming = names(segments)[current_segment_i]
  dsummary[si, fi:(fi + 2)] = c(wi, segmentName, segmentTiming)
  ds_names[fi:(fi + 2)] = c("window_number", "window", "start_end_window"); fi = fi + 3
  # Get onset and waking timing, both as timestamp and as index
  onsetwaketiming = g.part5.onsetwaketiming(qqq, ts, min, sec, hour, timewindowi)
  onset = onsetwaketiming$onset; wake = onsetwaketiming$wake
  onseti = onsetwaketiming$onseti; wakei = onsetwaketiming$wakei
  skiponset = onsetwaketiming$skiponset; skipwake = onsetwaketiming$skipwake
  if (wake < 24 & timewindowi == "WW") {
    # waking up before midnight means that next WW window
    # will start a day before the date we refer to when discussing it's SPT
    # So, for next window we have to do date = date + 1
    add_one_day_to_next_date = TRUE
  }
  if (onset > 24 & timewindowi == "OO") {
    # onset after midnight means that next OO window
    # will start a date after the date we refer to when discussing it;s SPT
    # So, for next window we have to do date = date + 1
    add_one_day_to_next_date = TRUE
  }
  # Add to dsummary output matrix
  if (skiponset == FALSE) {
    dsummary[si,fi] = onset
    dsummary[si,fi + 1] = as.character(strftime(format(time_POSIX[onseti]), tz = params_general[["desiredtz"]], format = "%H:%M:%S"))
  } else {
    dsummary[si,fi:(fi + 1)] = rep(NA, 2)
  }
  ds_names[fi:(fi + 1)] = c("sleeponset", "sleeponset_ts");      fi = fi + 2
  if (skipwake == FALSE) {
    dsummary[si,fi] = wake
    dsummary[si,fi + 1] = as.character(strftime(format(time_POSIX[wakei]), tz = params_general[["desiredtz"]], format = "%H:%M:%S"))
  } else {
    dsummary[si,fi:(fi + 1)] = rep(NA, 2)
  }
  ds_names[fi:(fi + 1)] = c("wakeup", "wakeup_ts");      fi = fi + 2
  # if skiponset and skipwake and "MM", then set full window as awake
  if (skiponset == TRUE & skipwake == TRUE & timewindowi == "MM") {
    ts$diur_bu = ts$diur
    ts$diur[qqq[1]:qqq[2]] = 0
  }
  # extract date and use this to retrieve corresponding part 4 information about the nights:
  options(encoding = "UTF-8")
  # look up matching part4 entry:
  recDates = as.Date(sumSleep$calendar_date, format = "%d/%m/%Y", origin = "1970-01-01")
  dsummary[si, fi] = sibDef
  ds_names[fi] = "sleepparam";      fi = fi + 1
  dayofinterest = which(recDates == date)
  if (length(dayofinterest) > 0) {
    dayofinterest = dayofinterest[1]
    dsummary[si,fi:(fi + 5)] = c(sumSleep$night[dayofinterest],
                                 sumSleep$daysleeper[dayofinterest],
                                 sumSleep$cleaningcode[dayofinterest],
                                 sumSleep$guider[dayofinterest],
                                 sumSleep$sleeplog_used[dayofinterest],
                                 sumSleep$acc_available[dayofinterest])
    ds_names[fi:(fi + 5)] = c("night_number", "daysleeper", "cleaningcode",
                              "guider", "sleeplog_used", "acc_available");      fi = fi + 6
    if (!is.na(segStart) & !is.na(segEnd)) {
      # segment available in time series
      ts$guider[segStart:segEnd] = sumSleep$guider[dayofinterest] # add guider also to timeseries
    }
  } else {
    dsummary[si,fi:(fi + 5)] = rep(NA, 6)
    ds_names[fi:(fi + 5)] = c("night_number",
                              "daysleeper","cleaningcode","guider",
                              "sleeplog_used","acc_available"); fi = fi + 6
  }
  #==========================
  # define time windows:
  # We will get onset and wakeup
  # regarding to the same day of measurement in the same row.
  # which differs between MM and WW
  # Also, it allows for the analysis of the first day for those studies 
  # in which the accelerometer is started during the morning and the first day is of interest.
  # qqq1 is the start of the day/segment
  # qqq2 is the end of the day/segment
  qqq1 = segStart
  qqq2 = segEnd
  # keep track of threshold value
  dsummary[si, fi:(fi + 2)] = c(TRLi, TRMi, TRVi)
  ds_names[fi:(fi + 2)] = c("TRLi", "TRMi", "TRVi")
  fi = fi + 3
  wlih = ((qqq2 - qqq1) + 1)/((60/ws3new) * 60)
  if (!is.na(qqq1)) {
    if (qqq1 > length(LEVELS)) qqq1 = length(LEVELS)
    sse = qqq1:qqq2
  } else {
    sse = NULL
  }
  doNext = FALSE
  if (length(sse) >= 1) { #next
    #============================================================
    # percentage of available data
    zt_hrs_nonwear = (length(which(ts$diur[sse] == 0 & ts$nonwear[sse] == 1)) * ws3new) / 3600 #day
    zt_hrs_total = (length(which(ts$diur[sse] == 0)) * ws3new) / 3600 #day
    dsummary[si,fi] = (zt_hrs_nonwear/zt_hrs_total)  * 10000 / 100
    ds_names[fi] = "nonwear_perc_day";      fi = fi + 1
    zt_hrs_nonwear = (length(which(ts$diur[sse] == 1 & ts$nonwear[sse] == 1)) * ws3new) / 3600 #night
    zt_hrs_total = (length(which(ts$diur[sse] == 1)) * ws3new) / 3600 #night
    dsummary[si,fi] =  (zt_hrs_nonwear/zt_hrs_total)  * 10000 / 100
    ds_names[fi] = "nonwear_perc_spt";      fi = fi + 1
    zt_hrs_nonwear = (length(which(ts$nonwear[sse] == 1)) * ws3new) / 3600
    zt_hrs_total = (length(ts$diur[sse]) * ws3new) / 3600 #night and day
    dsummary[si,fi] =  (zt_hrs_nonwear/zt_hrs_total)  * 10000 / 100
    ds_names[fi] = "nonwear_perc_day_spt";      fi = fi + 1
    #=======================================================
    # nap/sib/nonwear overlap analysis
    #=======================================================
    if (params_output[["do.sibreport"]]  == TRUE &&
        !is.null(params_sleep[["possible_nap_window"]]) &&
        !is.null(params_sleep[["possible_nap_dur"]])) {
      restAnalyses = g.part5.analyseRest(sibreport = sibreport, dsummary = dsummary,
                                         ds_names = ds_names, fi = fi, di = si,
                                         ts = ts[sse[ts$diur[sse] == 0], ],
                                         tz = params_general[["desiredtz"]],
                                         params_sleep = params_sleep)
      fi = restAnalyses$fi
      si = restAnalyses$di
      ts[sse[ts$diur[sse] == 0], ] = restAnalyses$ts
      dsummary = restAnalyses$dsummary
      ds_names = restAnalyses$ds_names
      
      # If naps detected add these to LEVELS
      detectedNaps = which(ts$sibdetection[sse] == 2)
      if (length(detectedNaps) > 0) {
        if ("day_nap" %in% Lnames) {
          LEVELS[sse[detectedNaps]] = length(Lnames) - 1
        } else {
          LEVELS[sse[detectedNaps]] = length(Lnames)
        }
      }
      if ("day_nap" %in% Lnames == FALSE) {
        Lnames = c(Lnames, "day_nap")
      }
    }
    #===============================================
    # TIME SPENT IN WINDOWS (window is either midnight-midnight or waking up-waking up)
    test_remember = c(si,fi)
    for (levelsc in 0:(length(Lnames) - 1)) {
      dsummary[si,fi] = (length(which(LEVELS[sse] == levelsc)) * ws3new) / 60
      ds_names[fi] = paste0("dur_", Lnames[levelsc + 1],"_min");      fi = fi + 1
    }
    for (g in 1:4) {
      dsummary[si, (fi + (g - 1))] = (length(which(OLEVELS[sse] == g)) * ws3new) / 60
    }
    ds_names[fi:(fi + 3)] = c("dur_day_total_IN_min",
                              "dur_day_total_LIG_min",
                              "dur_day_total_MOD_min",
                              "dur_day_total_VIG_min")
    fi = fi + 4
    dsummary[si, fi] = (length(which(ts$diur[sse] == 0)) * ws3new) / 60
    ds_names[fi] = "dur_day_min";      fi = fi + 1
    dsummary[si, fi] = (length(which(ts$diur[sse] == 1)) * ws3new) / 60
    ds_names[fi] = "dur_spt_min";      fi = fi + 1
    dsummary[si, fi] = (length(c(sse)) * ws3new) / 60
    ds_names[fi] = "dur_day_spt_min";      fi = fi + 1
    #============================================
    # Number of long wake periods (defined as > 5 minutes) during the night
    Nawake = length(which(abs(diff(which(LEVELS[sse] == 0))) > (300 / ws3new))) - 2
    if (Nawake < 0) Nawake = 0
    dsummary[si, fi] = Nawake
    ds_names[fi] = "N_atleast5minwakenight";      fi = fi + 1
    #=============================
    # sleep efficiency
    dsummary[si,fi] = length(which(ts$sibdetection[sse] == 1 &
                                     ts$diur[sse] == 1)) / length(which(ts$diur[sse] == 1))
    ds_names[fi] = "sleep_efficiency_after_onset";      fi = fi + 1
    #===============================================
    # NAPS (estimation)
    if (params_output[["do.sibreport"]] == TRUE & "nap1_nonwear2" %in% colnames(ts) &
        length(params_sleep[["nap_model"]]) > 0) {
      dsummary[si,fi] = length(which(diff(c(-1, which(ts$nap1_nonwear2[sse] == 1 & ts$diur[sse] == 0))) > 1))
      ds_names[fi] = "nap_count";      fi = fi + 1
      dsummary[si,fi] = round((sum(ts$nap1_nonwear2[sse[which(ts$nap1_nonwear2[sse] == 1 & ts$diur[sse] == 0)]]) * ws3new) / 60, digits = 2)
      ds_names[fi] = "nap_totalduration";      fi = fi + 1
    }
    if (length(tail_expansion_log) != 0) {
      # do not store sleep variables if data was expanded in GGIR part 1
      dsummary[si, fi] = (tail_expansion_log[["short"]] * ws3new) / 60
    } else {
      dsummary[si, fi] = 0
    }
    ds_names[fi] = "tail_expansion_minutes";      fi = fi + 1
    #===============================================
    # AVERAGE ACC PER WINDOW
    for (levelsc in 0:(length(Lnames) - 1)) {
      dsummary[si,fi] = mean(ts$ACC[sse[LEVELS[sse] == levelsc]], na.rm = TRUE)
      ds_names[fi] = paste("ACC_", Lnames[levelsc + 1], "_mg", sep = "");      fi = fi + 1
    }
    for (g in 1:4) {
      dsummary[si,(fi + (g - 1))] = mean(ts$ACC[sse[OLEVELS[sse] == g]], na.rm = TRUE)
    }
    ds_names[fi:(fi + 3)] = c("ACC_day_total_IN_mg", "ACC_day_total_LIG_mg",
                              "ACC_day_total_MOD_mg", "ACC_day_total_VIG_mg")
    fi = fi + 4
    dsummary[si, fi] = mean(ts$ACC[sse[ts$diur[sse] == 0]], na.rm = TRUE)
    ds_names[fi] = "ACC_day_mg";      fi = fi + 1
    dsummary[si, fi] = mean(ts$ACC[sse[ts$diur[sse] == 1]], na.rm = TRUE)
    ds_names[fi] = "ACC_spt_mg";      fi = fi + 1
    dsummary[si, fi] = median(ts$ACC[sse[ts$diur[sse] == 1]], na.rm = TRUE)
    ds_names[fi] = "ACC_spt_mg_median";      fi = fi + 1
    dsummary[si, fi] = sd(ts$ACC[sse[ts$diur[sse] == 1]], na.rm = TRUE)
    ds_names[fi] = "ACC_spt_mg_stdev";      fi = fi + 1
    dsummary[si, fi] = mean(ts$ACC[sse], na.rm = TRUE)
    ds_names[fi] = "ACC_day_spt_mg";      fi = fi + 1
    
    # #======================================================
    # # STEPS... (IF STEP_COUNT IS AVAILABLE)
    # if ("step_count" %in% names(ts)) {
    #   if (length(params_247[["clevels"]]) > 1) {
    #     for (windowhalf in c("day", "spt")) {
    #       step_subsegment = sse[ts$diur[sse] == ifelse(windowhalf == "day", yes = 0, no = 1)]
    #       cadence = ts$step_count[step_subsegment] / (60 / ws3new)
    #       step_count = ts$step_count[step_subsegment]
    #       for (cleveli in 1:(length(params_247[["clevels"]]) - 1)) {
    #         cadence_condition_met = which(cadence >= params_247[["clevels"]][cleveli] &
    #                                         cadence < params_247[["clevels"]][cleveli + 1])
    #         dsummary[si, fi] = floor(sum(step_count[cadence_condition_met]))
    #         tmp_ending_of_name = paste0("in_candence_", params_247[["clevels"]][cleveli],
    #                                     "_", params_247[["clevels"]][cleveli + 1], "_spm")
    #         ds_names[fi] = paste0("STEPS_", windowhalf, "_count_", tmp_ending_of_name)
    #         fi = fi + 1
    #         dsummary[si, fi] = length(step_count[cadence_condition_met]) * (ws3new / 60)
    #         ds_names[fi] = paste0("STEPS_", windowhalf, "_min_", tmp_ending_of_name)
    #         fi = fi + 1
    #       }
    #     }
    #   }
    # }
    #===============================================
    # QUANTILES...
    WLH = ((qqq2 - qqq1) + 1)/((60/ws3new) * 60)
    if (WLH <= 1) WLH = 1.001
    dsummary[si, fi] = quantile(ts$ACC[sse],probs = ((WLH - 1)/WLH), na.rm = TRUE)
    ds_names[fi] = paste("quantile_mostactive60min_mg", sep = "");      fi = fi + 1
    dsummary[si, fi] = quantile(ts$ACC[sse],probs = ((WLH - 0.5)/WLH), na.rm = TRUE)
    ds_names[fi] = paste("quantile_mostactive30min_mg", sep = "");      fi = fi + 1
    #===============================================
    NANS = which(is.nan(dsummary[si,]) == TRUE) #average of no values will results in NaN
    if (length(NANS) > 0) dsummary[si,NANS] = ""
    #===============================================
    # NUMBER OF BOUTS
    checkshape = function(boutcount) {
      if (is.matrix(boutcount) == FALSE) {# if there is only one bout setting
        boutcount = as.matrix(boutcount)
        if (nrow(boutcount) > ncol(boutcount)) boutcount = t(boutcount)
      }
      return(boutcount)
    }
    bc.mvpa = checkshape(bc.mvpa)
    for (bci in 1:nrow(bc.mvpa)) {
      RLE = rle(bc.mvpa[bci, sse])
      dsummary[si, fi + (bci - 1)] = length(which(RLE$values == 1))
      if (bci == 1) {
        ds_names[fi + (bci - 1)] = paste0("Nbouts_day_MVPA_bts_", params_phyact[["boutdur.mvpa"]][bci])
      } else {
        ds_names[fi + (bci - 1)] = paste0("Nbouts_day_MVPA_bts_", params_phyact[["boutdur.mvpa"]][bci], "_", params_phyact[["boutdur.mvpa"]][bci - 1])
      }
    }
    fi = fi + bci
    bc.in = checkshape(bc.in)
    for (bci in 1:nrow(bc.in)) {
      RLE = rle(bc.in[bci,sse])
      dsummary[si,fi + (bci - 1)] = length(which(RLE$values == 1))
      if (bci == 1) {
        ds_names[fi + (bci - 1)] = paste0("Nbouts_day_IN_bts_",params_phyact[["boutdur.in"]][bci])
      } else {
        ds_names[fi + (bci - 1)] = paste0("Nbouts_day_IN_bts_",
                                          params_phyact[["boutdur.in"]][bci], "_",
                                          params_phyact[["boutdur.in"]][bci - 1])
      }
    }
    fi = fi + bci
    bc.lig = checkshape(bc.lig)
    for (bci in 1:nrow(bc.lig)) {
      RLE = rle(bc.lig[bci,sse])
      dsummary[si,fi + (bci - 1)] = length(which(RLE$values == 1))
      if (bci == 1) {
        ds_names[fi + (bci - 1)] = paste0("Nbouts_day_LIG_bts_",
                                          params_phyact[["boutdur.lig"]][bci])
      } else {
        ds_names[fi + (bci - 1)] = paste0("Nbouts_day_LIG_bts_",
                                          params_phyact[["boutdur.lig"]][bci], "_",
                                          params_phyact[["boutdur.lig"]][bci - 1])
      }
    }
    fi = fi + bci
    #===============================================
    # NUMBER OF WINDOWS / BLOCKS
    RLE_LEVELS = rle(LEVELS[sse])
    RLE_OLEVELS = rle(OLEVELS[sse])
    # RunLengthEncoding
    for (levelsc in 0:(length(Lnames) - 1)) {
      dsummary[si, fi] = length(which(RLE_LEVELS$values == levelsc))
      ds_names[fi] = paste("Nblocks_", Lnames[levelsc + 1], sep = "");      fi = fi + 1
    }
    for (g in 1:4) {
      dsummary[si, (fi + (g - 1))] = length(which(RLE_OLEVELS$values == g))
    }
    ds_names[fi:(fi + 3)] = c("Nblocks_day_total_IN", "Nblocks_day_total_LIG",
                              "Nblocks_day_total_MOD", "Nblocks_day_total_VIG")
    fi = fi + 4
    dsummary[si, fi:(fi + 5)] = c(params_phyact[["boutcriter.in"]],
                                  params_phyact[["boutcriter.lig"]],
                                  params_phyact[["boutcriter.mvpa"]],
                                  paste(params_phyact[["boutdur.in"]], collapse = "_"),
                                  paste(params_phyact[["boutdur.lig"]], collapse = "_"),
                                  paste(params_phyact[["boutdur.mvpa"]], collapse = "_"))
    ds_names[fi:(fi + 5)] = c("boutcriter.in", "boutcriter.lig", "boutcriter.mvpa",
                              "boutdur.in",  "boutdur.lig", "boutdur.mvpa"); fi = fi + 6
    #===========================
    # Intensity gradient over waking hours
    if (length(params_247[["iglevels"]]) > 0) {
      q55 = cut(ts$ACC[sse[ts$diur[sse] == 0]], breaks = params_247[["iglevels"]], right = FALSE)
      x_ig = zoo::rollmean(params_247[["iglevels"]], k = 2)
      y_ig = (as.numeric(table(q55)) * ws3new)/60 #converting to minutes
      dsummary[si,fi:(fi + 2)] = as.numeric(g.intensitygradient(x_ig, y_ig))
      ds_names[fi:(fi + 2)] = c("ig_day_gradient", "ig_day_intercept", "ig_day_rsquared")
      fi = fi + 3
    }
    #===========================
    # Intensity gradient over the full window (waking + spt)
    if (length(params_247[["iglevels"]]) > 0) {
      q55 = cut(ts$ACC[sse], breaks = params_247[["iglevels"]], right = FALSE)
      x_ig = zoo::rollmean(params_247[["iglevels"]], k = 2)
      y_ig = (as.numeric(table(q55)) * ws3new)/60 #converting to minutes
      dsummary[si,fi:(fi + 2)] = as.numeric(g.intensitygradient(x_ig, y_ig))
      ds_names[fi:(fi + 2)] = c("ig_day_spt_gradient", "ig_day_spt_intercept", "ig_day_spt_rsquared")
      fi = fi + 3
    }
    #===============================================
    # FRAGMENTATION
    if (length(params_phyact[["frag.metrics"]]) > 0) {
      # Only do daytime fragmentation and not spt window because in part 5 the spt window does
      # not necessarily include valid data. spt window fragmentation will be performed in part 6
      fragmode = "day"
      frag.out = g.fragmentation(frag.metrics = params_phyact[["frag.metrics"]],
                                 LEVELS = LEVELS[sse[ts$diur[sse] == ifelse(fragmode == "day", 0, 1)]],
                                 Lnames = Lnames, xmin = 60/ws3new, mode = fragmode)
      # fragmentation values can come with a lot of decimal places
      dsummary[si, fi:(fi + (length(frag.out) - 1))] = round(as.numeric(frag.out), digits = 6)
      ds_names[fi:(fi + (length(frag.out) - 1))] = paste0("FRAG_", names(frag.out), "_", fragmode)
      fi = fi + length(frag.out)
    }
    #===============================================
    # LIGHT, IF AVAILABLE
    if ("lightpeak" %in% colnames(ts) & length(params_247[["LUX_day_segments"]]) > 0) {
      # mean LUX
      if (length(which(ts$diur[sse] == 0)) > 0 & length(which(ts$diur[sse] == 1)) > 0) {
        dsummary[si,fi] =  round(max(ts$lightpeak[sse[ts$diur[sse] == 0]], na.rm = TRUE), digits = 1)
        dsummary[si,fi + 1] = round(mean(ts$lightpeak[sse[ts$diur[sse] == 0]], na.rm = TRUE), digits = 1)
        dsummary[si,fi + 2] = round(mean(ts$lightpeak[sse[ts$diur[sse] == 1]], na.rm = TRUE), digits = 1)
        dsummary[si,fi + 3] = round(mean(ts$lightpeak[sse[ts$diur[sse] == 0 & ts$ACC[sse] > TRMi]], na.rm = TRUE),
                                    digits = 1)
      } else {
        dsummary[si,fi:(fi + 3)] = NA
      }
      ds_names[fi:(fi + 3)] = c("LUX_max_day", "LUX_mean_day", "LUX_mean_spt", "LUX_mean_day_mvpa"); fi = fi + 4
      # time in LUX ranges
      Nluxt = length(params_247[["LUXthresholds"]])
      for (lti in 1:Nluxt) {
        if (lti < Nluxt) {
          dsummary[si, fi + lti - 1] =  length(which(ts$lightpeak[sse[ts$diur[sse] == 0]] >= params_247[["LUXthresholds"]][lti] &
                                                       ts$lightpeak[sse[ts$diur[sse] == 0]] < params_247[["LUXthresholds"]][lti + 1])) / (60/ws3new)
          ds_names[fi + lti - 1] = paste0("LUX_min_", params_247[["LUXthresholds"]][lti], "_", params_247[["LUXthresholds"]][lti + 1], "_day")
        } else {
          dsummary[si, fi + lti - 1] =  length(which(ts$lightpeak[sse[ts$diur[sse] == 0]] >= params_247[["LUXthresholds"]][lti])) / (60/ws3new)
          ds_names[fi + lti - 1] = paste0("LUX_min_", params_247[["LUXthresholds"]][lti], "_inf_day")
        }
      }
      fi = fi + Nluxt
      # LUX per segment of the day
      luxperseg = g.part5.lux_persegment(ts, sse,
                                         LUX_day_segments = params_247[["LUX_day_segments"]],
                                         epochSize = ws3new,
                                         desiredtz = params_general[["desiredtz"]])
      dsummary[si, fi:(fi + (length(luxperseg$values) - 1))] = luxperseg$values
      ds_names[fi:(fi + (length(luxperseg$values) - 1))] = luxperseg$names
      fi = fi + length(luxperseg$values)
    }

    #===============================================
    # FOLDER STRUCTURE
    if (params_output[["storefolderstructure"]] == TRUE) {
      if ("filename_dir" %in% ds_names) fi = which( ds_names == "filename_dir")
      dsummary[si,fi] = fullFilename #full filename structure
      ds_names[fi] = "filename_dir"; fi = fi + 1
      dsummary[si,fi] = foldernamei #store the lowest foldername
      if ("foldername" %in% ds_names) fi = which(ds_names == "foldername")
      ds_names[fi] = "foldername"; fi = fi + 1
    }
  } else {
    doNext = TRUE
  }
  # if skiponset and skipwake and "MM", then reset ts$diur
  if (skiponset == TRUE & skipwake == TRUE & timewindowi == "MM") {
    ts$diur = ts$diur_bu
    ts = ts[, -which(colnames(ts) == "diur_bu")]
  }
  # group categories of object back into lists
  indexlog = list(fileIndex = fileIndex,
                  winStartEnd = qqq,
                  segIndex1 = si,
                  segIndex2 = current_segment_i,
                  segStartEnd = c(segStart, segEnd),
                  columnIndex = fi)
  timeList = list(ts = ts,
                  epochSize = ws3new,
                  LEVELS = LEVELS,
                  Lnames = Lnames)
  invisible(list(
    indexlog = indexlog,
    ds_names = ds_names,
    dsummary = dsummary,
    timeList = timeList,
    doNext = doNext,
    add_one_day_to_next_date = add_one_day_to_next_date
  ))
}

# NOTE ABOUT DOCUMENTATION:
# GGIR does not use Roxygen. The documentation below is not used. 
# All function documentation can be found in the man/*.Rd files.
# Please edit documentation there.
#
#- @title Detrended Fluctuation Analysis
#- 
#- @description ...
#- @usage DFA(data, scale = 2^(1/8), box_size = 4, m = 1)
#- @param data Univariate time series (must be a vector or data frame)
#- @param scale Specifies the ratio between successive box sizes (by default scale = 2^(1/8))
#- @param box_size Vector of box sizes (must be used in conjunction with scale = "F")
#- @param m An integer of the polynomial order for the detrending (by default m=1)
#- 
#- @details The DFA fluctuation can be computed in a geometric scale or for different choices of boxes sizes.
#- 
#- @return Estimated alpha is a real number between zero and two.
#- 
#- @note It is not possible estimating alpha for multiple time series at once.
#- 
#- @author Ian Meneghel Danilevicz  and Victor Mesquita
#- 
#- @references C.-K. Peng, S.V. Buldyrev, S. Havlin, M. Simons, H.E. Stanley, A.L. Goldberger Phys. Rev. E, 49 (1994), p. 1685
#- Mesquita, Victor & Filho, Florencio & Rodrigues, Paulo. (2020). Detection of crossover points in detrended fluctuation analysis: An application to EEG signals of patients with epilepsy. Bioinformatics. 10.1093/bioinformatics/btaa955.
#- 
#- @examples
#- # Estimate self-similarity of a very known time series available on R base: the sunspot.year.
#- # Then the spend time with each method is compared.
#- 

DFA = function(data, scale = 2^(1/8), box_size = 4, m = 1){
  if (inherits(x = data, "data.frame")) {
    data = data[, 1]
  }
  N = length(data)
  if (scale != "F") {
    box_size <- NULL
    n = 1
    n_aux <- 0
    box_size[1] <- 4
    for (n in 1:N) {
      while (n_aux < round(N/4)) {
        n_aux <- box_size[1]
        n = n + 1
        box_size[n] <- ceiling(scale * box_size[n - 1])
        n_aux <- box_size[n] + 4
      }
    }
  }
  ninbox2 <- NULL
  for (j in 1:length(box_size)) {
    ninbox2[j] <- N %/% box_size[j]
  }
  aux_seq = seq_along(ninbox2)  
  aux_length = aux_seq[length(aux_seq)]  
  y_k = cumsum(data) - mean(data)
  rm(data, n_aux, scale)
  aux_mat = matrix(nrow = aux_length, ncol = 2)
  for (j in seq_along(ninbox2)) {
    W = box_size[j] * ninbox2[j]
    aux_j = numeric(W)
    fit = y_k
    for (i in seq_len(W)) {
      if (i == 1) {
        aux_j[1] = box_size[j]
        tmp1 = i:aux_j[i]
      } else if (i >= 2) {
        aux_j[i] = aux_j[i - 1] + box_size[j]
        tmp1 = (aux_j[i - 1] + 1):(aux_j[i])
      }
      if (m == 1) {
        fit[tmp1] = lm.fit(x = cbind(1, tmp1), y = y_k[tmp1])$fitted.values
      } else {
        fit[tmp1] = lm(y_k[tmp1] ~ poly(tmp1, m, raw = TRUE))$fitted.values
      }
      if (i >= ninbox2[j]) {
        aux_j[i] <- 0
      }
    }
    aux_mat[j,] = c(round(box_size[j], digits = 0),
                    sqrt((1 / N) * sum((y_k[1:W] - fit[1:W]) ^ 2)))
  }
  colnames(aux_mat) <- c("box", "DFA")
  aux_list = aux_mat
  return(aux_list)
}

apply_cosinor_IS_IV_Analyses = function(ts, qcheck, midnightsi, epochsizes, threshold = NULL) {
  # qcheck - vector of length ts to indicate invalid values
  ws2 = epochsizes[2]
  ws3 = epochsizes[1]
  # Re-derive Xi but this time include entire time series
  # Here, we ignore duplicated values (when clock moves backward due to DST)
  handleDST = !duplicated(ts$time)
  qcheck = qcheck[handleDST]
  Xi = ts[handleDST, grep(pattern = "time", x = colnames(ts), invert = TRUE)]
  Nlong_epochs_day =  (1440 * 60) / ws2 # this is 96 by default
  dstgap = which(diff(midnightsi) != Nlong_epochs_day)
  if (length(dstgap) > 0) {
    # Time moved forward due to DST
    gaplocation = ((midnightsi[dstgap[1]] * ws2) / ws3) + (2 * (3600/ws3))
    # Insert NA values
    Xi = c(Xi[1:gaplocation], rep(NA, 3600/ws3), Xi[(gaplocation + 1):length(Xi)])
    qcheck = c(qcheck[1:gaplocation], rep(NA, 3600/ws3), qcheck[(gaplocation + 1):length(qcheck)])
  }
  
  # Xi = log((Xi * 1000) + 1)  # log transformed to be more robust against peaks in the data
  # set non-wear to missing values, because for Cosinor fit
  # it seems more logical to only fit with real data
  # this comes at the price of not being able to extract F_pseudo
  firstvalid = 1
  if (length(which(qcheck == 1)) > 0) {
    is.na(Xi[which(qcheck == 1)]) = TRUE
    # ignore invalid start of recording (if applicable)
    # such that 24 hour blocks start from first valid value
    firstvalid = which(qcheck == 0)[1]
    if (is.na(firstvalid) == FALSE) {
      if (firstvalid != 1) {
        Xi = Xi[firstvalid:length(Xi)]
      }
    }
  }
  if (length(which(is.na(Xi) == FALSE)) > (1440 * (60/ws3))) { # Only attempt cosinor analyses if there is more than 24 hours of data
    midnightsi_ws3 = (midnightsi - 1) * (ws2 / ws3)
    timeOffsetHours = (midnightsi_ws3[which(midnightsi_ws3 >= firstvalid - 1)[1]] - (firstvalid - 1)) / (3600 / ws3)
    if (ws3 < 60) {
      # If epochsize < 1 minute then aggregate to 1 minute by taking average value
      # but keep NA values
      XTtime = rep(1:length(Xi), each = 60 / ws3)
      XT = data.frame(Xi = Xi, time = XTtime[1:length(Xi)])
      custommean = function(x) {
        y = NA
        if (length(x) > 0) {
          if (length(which(is.na(x) == FALSE) ) > 0) {
            y = mean(x, na.rm = TRUE)
          }
        }
        return(y)
      }
      XT = aggregate(x = XT, by = list(XT$time), FUN = custommean)
      if (length(which(is.nan(XT$Xi) == TRUE)) > 0) {
        is.na(XT$Xi[which(is.nan(XT$Xi) == TRUE)]) = TRUE
      }
      Xi = XT$Xi
      epochsize = 60
    } else {
      epochsize = ws3
    }
    cosinor_coef = cosinor_IS_IV_Analyses(Xi = Xi, epochsize = epochsize, 
                                   timeOffsetHours = timeOffsetHours, threshold = threshold)
    cosinor_coef$timeOffsetHours = timeOffsetHours
  } else {
    cosinor_coef = c()
  }
  return(cosinor_coef)
}

g.sib.plot = function(SLE, M, I, plottitle, nightsperpage = 7, desiredtz = "") {
  A = SLE$output
  # invalid = A$invalid
  night = A$night
  sleep = A[,(which(colnames(A) == "night") + 1):ncol(A)]
  D = as.matrix(M$metashort) #IMP$
  nD = nrow(D)
  S = as.matrix(M$metalong)
  nS = nrow(S)
  space = ifelse(length(unlist(strsplit(format(A$time[1])," "))) > 1,TRUE,FALSE)
  temptime = format(unlist(A$time))
  if (space == FALSE) {
    time = as.POSIXlt(temptime,tz = desiredtz, format = "%Y-%m-%dT%H:%M:%S%z")
    timeL = iso8601chartime2POSIX(S[1:nS, 1], tz = desiredtz)
  } else {
    time = as.POSIXlt(temptime, tz = desiredtz)
    timeL = as.POSIXlt(as.character(S[1:nS,1]), tz = desiredtz)
  }
  ws3 = M$windowsizes[1]
  ws2 = M$windowsizes[2]
  n_ws2_perday = (1440*60) / ws2
  n_ws3_perday = (1440*60) / ws3
  #ENMO
  ENMO = as.numeric(as.matrix(D[1:nD,2])) * 1000 #indicator of acceleration
  # angle
  angle = as.numeric(as.matrix(D[1:nD,which(colnames(M$metashort) == "anglez")]))
  if (length(which(is.na(angle) == TRUE)) > 0) {
    if (which(is.na(angle) == TRUE)[1] == length(angle)) {
      angle[length(angle)] = angle[length(angle) - 1]
    }
  }
  # get temperature and lightpeak
  if (I$monn == "geneactive") {
    temperature = as.numeric(as.matrix(S[1:nS,6]))
    lightpeak = as.numeric(as.matrix(S[1:nS,5]))
  } else {
    lightpeak = temperature = rep(0,nS)
  }
  # transform temperature signal into something that is easy to plot on top of other data
  if (length(which(is.na(lightpeak) == TRUE)) > 0) {
    lightpeak[which(is.na(lightpeak) == TRUE)] = 0
  }
  if (length(which(is.na(temperature) == TRUE)) > 0) {
    temperature[which(is.na(temperature) == TRUE)] = mean(temperature, na.rm = TRUE)
  }
  if (min(temperature, na.rm = TRUE) < 0) {
    temperature = temperature + abs(min(temperature, na.rm = TRUE))
  }
  temperature = temperature - min(temperature)
  temperature = (scale(temperature) * 70) - 90
  # transform light signal into something that is easy to plot on top of other data
  light = lightpeak
  light = light + min(light)
  light = (scale(light) * 210) - 120
  un = unique(night)
  if (length(which(un == 0 | is.na(un) == TRUE)) > 0) {
    un = un[-c(which(un == 0 | is.na(un) == TRUE))]
  }
  ncolsleep = ncol(sleep)
  if (length(ncolsleep) == 0) {
    zeros = which(sleep == 0)
    if (length(zeros) > 0) sleep[zeros] = NA
  } else {
    for (j in 1:ncolsleep) {
      zeros = which(sleep[,j] == 0)
      if (length(zeros) > 0) sleep[zeros,j] = NA
    }
  }
  for (i in 1:length(un)) {
    if (i == 1 | (i-1)/9 == round((i-1)/9)) {
      par(mfrow=c(9,1),mar=c(1,5,3,1),oma=c(2,0,0,0)) 
    }
    qqq1 = which(night == i)[1]
    qqq2 = which(night == i)[length(which(night == i))]
    qqq3 = which(format(timeL) == format(time[qqq1]))
    qqq4 = which(format(timeL) == format(time[qqq2]))
    if (length(qqq3) == 0) qqq3 = round(qqq1 / (ws2/ws3))
    if (length(qqq4) == 0) qqq4 = round(qqq2 / (ws2/ws3))
    if (qqq3 == 0) qqq3 = 1
    
    
    #activity in the background, so go first
    ENMOscaled = scale(ENMO[qqq1:qqq2],scale = abs(diff(range(ENMO[qqq1:qqq2])))/180) - 204
    plot(time[qqq1:qqq2],ENMOscaled,type="l",xlab="",main=paste("Night: ",i," ",plottitle),col="grey",
          ylab="angle (black)",   ylim=c(-250,90),xaxt="n",yaxt="n",cex.lab=0.8,cex.main=0.6,bty="n",lwd=0.4)
    #temperature
    lines(timeL[qqq3:qqq4],light[qqq3:qqq4],type="l",col="yellow",lty=1,lwd=1)
    #light
    lines(timeL[qqq3:qqq4],temperature[qqq3:qqq4],type="l",col="lightblue",lty=1,lwd=1)
    
    lines(time[qqq1:qqq2],angle[qqq1:qqq2],type="l",lwd=0.4,col="black")
    axis(2, at = c(-90,0,90))
    r <- as.POSIXct(round(range(time[qqq1:qqq2]), "hours"))
    abline(h=0,lty=3,lwd=1.4,col="grey")
    abline(v=seq(r[1], r[2], by = "hour"),lty=3,lwd=1.4,col="grey")
    #time axis
    axis.POSIXct(1, at = seq(r[1], r[2], by = "hour"), format = "%H")
    
    if (length(ncol(sleep)) > 0) {
      ncolsleep = ncol(sleep)
      CL = rainbow(ncolsleep)
      for (j in 1:ncolsleep) {
        lines(time[qqq1:qqq2],(-(240)+(sleep[qqq1:qqq2,j]*((80/ncolsleep)*j))),type="l",col=CL[j],lwd=1)
      }
    } else {
      ncolsleep = 1
      CL = rainbow(ncolsleep)
      for (j in 1:ncolsleep) {
        lines(time[qqq1:qqq2],(-(240)+(sleep[qqq1:qqq2]*((80/ncolsleep)*j))),type="l",col=CL[j],lwd=1)
      }
    }
  }
}

addSplitNames = function(x) {
  if (nrow(x) != 0) {
    x$filename = as.character(x$filename)
    for (spi in 1:nrow(x)) {
      splitnames = getSplitNames(as.character(x$filename[spi]))
      segment_names = splitnames$segment_names
      filename = splitnames$filename
      if (!is.null(segment_names)) {
        x$filename[spi] = filename
        if ("split1_name" %in% colnames(x) == FALSE) {
          x$split1_name = NA
          x$split2_name = NA
          col_index_filename = which(colnames(x) == "filename")
          
          x = x[,c(1:col_index_filename, 
                   which(colnames(x) %in% c("split1_name", "split2_name")),
                   (col_index_filename + 1):(ncol(x) - 2))] 
        }
        x$split1_name[spi] = segment_names[2]
        x$split2_name[spi] = segment_names[3]
      }
    }
    x$filename = as.factor(x$filename)
  }
  return(x)
}

g.intensitygradient = function(x,y) {
  # x: numeric vector of bin mid-points
  # y: numeric vector of time spent in bins
  y = ifelse(test = y <=0,yes = NA,no = y)
  ly = log(y)
  lx = log(x)
  y_intercept = NA
  gradient = NA
  rsquared = NA
  if (length(which(is.na(lx) == FALSE)) > 1 & length(which(is.na(ly) == FALSE)) > 1) {
    if (sd(lx,na.rm = TRUE) != 0 & sd(ly,na.rm = TRUE) != 0) {
      fitsum = summary(stats::lm(ly ~ lx))
      y_intercept = stats::coef(fitsum)[1,1]
      gradient = stats::coef(fitsum)[2,1]
      rsquared = fitsum$r.squared
    }
  }
  invisible(list(gradient=gradient,y_intercept=y_intercept,rsquared=rsquared))
}

library(GGIR)
context("g.intensitygradient")
options(encoding = "UTF-8")
test_that("g.intensitygradient returns expected output value", {
  x = c(seq(12.5,4000,by=25),6000)
  y = (x * -1.5) + 10000
  IG = g.intensitygradient(x = x, y = y)
  expect_equal(round(IG$gradient,digits=2),-0.23)
  expect_equal(round(IG$y_intercept,digits=2),10.50)
  expect_equal(round(IG$rsquared,digits=2),0.59)
})

g.part5.savetimeseries = function(ts, LEVELS, desiredtz, rawlevels_fname,
                                  DaCleanFile = NULL,
                                  includedaycrit.part5 = 2/3,
                                  includenightcrit.part5 = 0,
                                  ID = NULL,
                                  params_output,
                                  params_247 = NULL,
                                  Lnames = NULL, timewindow = NULL,
                                  filename = "") {

  ms5rawlevels = data.frame(date_time = ts$time, class_id = LEVELS,
                            # class_name = rep("",Nts),
                            stringsAsFactors = FALSE)

  if (length(grep(pattern = " ", x = ts$time[1])) == 0) {
    ts$timestamp = as.POSIXct(ts$time, tz = desiredtz, format = "%Y-%m-%dT%H:%M:%S%z")
    ms5rawlevels$date_time = as.POSIXct(ms5rawlevels$date_time, 
                                        tz = desiredtz, format = "%Y-%m-%dT%H:%M:%S%z")
  } else {
    ts$timestamp = ts$time
  }
  # Create numeric time to faciltiate merging
  ts$timenum = as.numeric(ts$timestamp)
  epochSize = ts$timenum[2] - ts$timenum[1]
  ms5rawlevels$timenum = as.numeric(ms5rawlevels$date_time)
  mdat = merge(ts, ms5rawlevels, by = "timenum")
  rm(ts, ms5rawlevels)
  names(mdat)[which(names(mdat) == "nonwear")] = "invalidepoch"
  names(mdat)[which(names(mdat) == "diur")] = "SleepPeriodTime"
  mdat = mdat[,-which(names(mdat) == "date_time")]
  if ("require_complete_lastnight_part5" %in% names(params_output) &&
      params_output[["require_complete_lastnight_part5"]] == TRUE) {
    N_window0_at_end = which(rev(mdat$window) != 0)[1]
    N_hours_window0_at_end = ((N_window0_at_end * epochSize) / 3600)
    lastHour = as.numeric(format(mdat$timestamp[length(mdat$timestamp)], "%H"))
    # If last hour is less than 9 then we know recording ended between midnight and 9am
    # which means that sleep onset may not be reliably detected
    # If last window with a non-zero number ends after 6pm in the before last day
    # then that confirms that sleep was estimated fromt his incomplete night
    if ((timewindow == "MM" || timewindow == "OO") && lastHour < 9 &&
        N_hours_window0_at_end < 9 + 6) {
      mdat$window[which(mdat$window == max(mdat$window))] = 0 
    }
    # If last hour is less than 15 then we know recording ended between midnight and 3pm
    # which means that wakeup may not be reliably detected
    # If last window with a non-zero number ends after 6pm in the before last day
    # then that confirms that sleep was estimated from this incomplete night
    if (timewindow == "WW" && lastHour < 15 &&
        N_hours_window0_at_end < 15 + 6) {
      mdat$window[which(mdat$window == max(mdat$window))] = 0 
    }
  }
  
  # Add invalid day indicator
  mdat$invalid_wakinghours = mdat$invalid_sleepperiod =  mdat$invalid_fullwindow = 100
  window_starts = which(abs(diff(c(0, mdat$window, 0))) > 0) # first epoch of each window
  if (length(window_starts) > 0) {
    if (length(window_starts) > 1) {
      for (di in 1:(length(window_starts) - 1)) {
        window_indices = window_starts[di]:pmin((window_starts[di + 1] - 1), nrow(mdat))
        wake = which(mdat$SleepPeriodTime[window_indices] == 0)
        sleep = which(mdat$SleepPeriodTime[window_indices] == 1)
        mdat$invalid_wakinghours[window_indices] = round(mean(mdat$invalidepoch[window_indices[wake]]) * 100, digits = 2)
        mdat$invalid_sleepperiod[window_indices] = round(mean(mdat$invalidepoch[window_indices[sleep]]) * 100, digits = 2)
        mdat$invalid_fullwindow[window_indices] = round(mean(mdat$invalidepoch[window_indices]) * 100, digits = 2)
      }
    } else {
      wake = which(mdat$SleepPeriodTime == 0)
      sleep = which(mdat$SleepPeriodTime == 1)
      mdat$invalid_wakinghours = round(mean(mdat$invalidepoch[wake]) * 100, digits = 2)
      mdat$invalid_sleepperiod = round(mean(mdat$invalidepoch[sleep]) * 100, digits = 2)
      mdat$invalid_fullwindow = round(mean(mdat$invalidepoch) * 100, digits = 2)
    }
    # round acceleration values to 3 digits to reduce storage space
    mdat$ACC = round(mdat$ACC, digits = 3)
    # round some columns to reduce storage space
    if ("lightpeak" %in% names(mdat)) mdat$lightpeak = round(mdat$lightpeak)
    if ("temperature" %in% names(mdat)) mdat$temperature = round(mdat$temperature, digits = 2)
    if (params_output[["save_ms5raw_without_invalid"]] == TRUE) {
      # Remove days based on data_cleaning_file
      if (length(DaCleanFile) > 0) { 
        if (ID %in% DaCleanFile$ID) {
          days2exclude = DaCleanFile$day_part5[which(DaCleanFile$ID == ID)]
          if (length(days2exclude) > 0) {
            cut = which(mdat$window %in% days2exclude == TRUE)
            if (length(cut) > 0) mdat = mdat[-cut,] # remove days from which we already know that they are not going to be included (first and last day)       
          }
        }
      }
      # Reformat includedaycrit.part5 to maximum percentage non-wear during waking hours:
      if (includedaycrit.part5 >= 0 & includedaycrit.part5 <= 1) { # if includedaycrit.part5 is used as a ratio
        includedaycrit.part5 = includedaycrit.part5 * 100
      } else if (includedaycrit.part5 > 1 & includedaycrit.part5 <= 25) { # if includedaycrit.part5 is used like includedaycrit as a number of hours
        includedaycrit.part5 = (includedaycrit.part5 / 24) * 100
      }
      # Reformat includenightcrit.part5 to maximum percentage non-wear during waking hours:
      if (includenightcrit.part5 >= 0 & includenightcrit.part5 <= 1) { # if includenightcrit.part5 is used as a ratio
        includenightcrit.part5 = includenightcrit.part5 * 100
      } else if (includenightcrit.part5 > 1 & includenightcrit.part5 <= 25) { # if includenightcrit.part5 is used like includedaycrit as a number of hours
        includenightcrit.part5 = (includenightcrit.part5 / 24) * 100
      }
      maxpernwday = 100 - includedaycrit.part5
      maxpernwnight = 100 - includenightcrit.part5
      # Exclude days that have 100% nonwear over the full window or 100% over wakinghours
      cut = which(mdat$invalid_fullwindow == 100 |
                    mdat$invalid_wakinghours > maxpernwday |
                    mdat$invalid_sleepperiod > maxpernwnight |
                    mdat$window == 0)
      if (length(cut) > 0) mdat = mdat[-cut,] # remove days from which we already know that they are not going to be included (first and last day)
    }
    new_guider_number = ifelse(mdat$guider == 'sleeplog', yes = 1, # digitize guider to save storage space
                         no = ifelse(mdat$guider == 'HDCZA' | mdat$guider == 'HDCZA+invalid', yes = 2,
                                     no = ifelse(mdat$guider == 'HorAngle' | mdat$guider == 'HorAngle+invalid', yes = 5,
                                                 no = ifelse(mdat$guider == 'NotWorn' | mdat$guider == 'NotWorn+invalid', yes = 6,
                                                             no = 0))))
    # use loop for other guiders:
    guidernames = c('setwindow', 'L512', 'markerbutton', 'HLRB', 'MotionWare')
    guidernumbers = c(3, 4, 7, 8, 9)
    for (gi in 1:length(guidernames)) {
      guidername_instance = which(mdat$guider == guidernames[gi])
      if (length(guidername_instance) > 0) {
        new_guider_number[guidername_instance] = guidernumbers[gi]
      }
    }
    mdat$guider = new_guider_number
    mdat = mdat[,-which(names(mdat) %in% c("timestamp","time"))]
    # re-oder columns
    if ("csv" %in% params_output[["save_ms5raw_format"]]) {
      # save to csv file
      fname = rawlevels_fname[grep("*csv$", rawlevels_fname)]
      data.table::fwrite(mdat, fname, row.names = F, sep = params_output[["sep_reports"]],
                         dec = params_output[["dec_reports"]])
    }
    if ("RData" %in% params_output[["save_ms5raw_format"]] || params_247[["part6HCA"]] == TRUE || params_247[["part6CR"]] == TRUE) {
      # only doing this for RData output, because it would affect file size too much in csv,
      # remember that this function can create many files: sample sizes times all combinations of thresholds.
      mdat$timestamp = as.POSIXct(mdat$timenum, origin = "1970-01-01",tz = desiredtz)
      rawlevels_fname = gsub(pattern = "[.]csv", replacement = ".RData", x = rawlevels_fname)
      fname = unique(rawlevels_fname[grep("*RData$", rawlevels_fname)])
      save(mdat, filename, Lnames, desiredtz_part1 = desiredtz, file = fname)
    }
    #===============================
    rm(mdat)
  }
  return()
}

g.inspectfile = function(datafile, desiredtz = "", params_rawdata = c(),
                         configtz = c(), ...) {
  #get input variables
  input = list(...)
  if (length(input) > 0 || length(params_rawdata) == 0) {
    # or if params_[...] aren't specified (so need to be filled with defaults).
    # So, inside GGIR this will not be used, but it is used when g.inspectfile is used on its own
    # as if it was still the old g.inspectfile function
    params = extract_params(params_rawdata = params_rawdata,
                            input = input,
                            params2check = "rawdata") # load default parameters
    params_rawdata = params$params_rawdata
    rm(params)
  }
  
  # note that if the file is an RData file then this function will not be called
  # the output of this function for the original datafile is stored inside the RData file in the form of object I
  getbrand = function(filename = c(), datafile = c()) {
    sf = c(); isitageneactive = c(); mon = c(); dformat = c() #generating empty variables
    extension = unlist(strsplit(filename,"[.]"))
    if (tolower(extension[length(extension)]) == "gz") {
      extension = extension[length(extension) - 1]
    } else {
      extension = extension[length(extension)]
    }
    switch(extension,
           "bin" = ,
           "BIN" = { dformat = FORMAT$BIN },
           "cwa" = ,
           "CWA" = { mon = MONITOR$AXIVITY
           dformat = FORMAT$CWA
           },
           "gt3x" = { mon = MONITOR$ACTIGRAPH
           dformat = FORMAT$GT3X
           },
           "GT3X" = { mon = MONITOR$ACTIGRAPH
           dformat = FORMAT$GT3X
           if (file.access(datafile, 2) == 0) { # test for write access to file
             # rename file to be lower case gt3x extension
             file.rename(from = datafile, to = gsub(pattern = ".GT3X", replacement = ".gt3x", x = datafile))
             datafile = gsub(pattern = ".GT3X", replacement = ".gt3x", x = datafile)
             warning("\nWe have renamed the GT3X file to gt3x because GGIR dependency read.gt3x cannot handle uper case extension")
           } else {
             stop("\nGGIR needs to change the file extension from GT3X to gt3x, but it does not seem to have write permission to the file.")
           }
           },
           "csv" = { dformat = FORMAT$CSV
           testheader = read.csv(datafile, nrow = 1, skip = 0, header = FALSE)
           
           if (grepl("ActiGraph", testheader[1], fixed=TRUE)) {
             mon = MONITOR$ACTIGRAPH
           } else {
             testcsv = read.csv(datafile, nrow = 10, skip = 10)
             testcsvtopline = read.csv(datafile, nrow = 2,skip = 1)
             if (ncol(testcsv) == 2 && ncol(testcsvtopline) < 4) {
               mon = MONITOR$GENEACTIV
             } else if (ncol(testcsv) >= 4 && ncol(testcsvtopline) >= 4) {
               mon = MONITOR$AXIVITY
             } else {
               stop(paste0("\nError processing ", filename, ": unrecognised csv file format.\n"))
             }
           }
           },
           "wav" = { stop(paste0("\nError processing ", filename, ": Axivity .wav file format is no longer supported.\n")) },
           { stop(paste0("\nError processing ", filename, ": unrecognised file format.\n")) }
    )
    
    if (ismovisens(datafile)) {
      dformat = FORMAT$BIN
      mon = MONITOR$MOVISENS
      sf = 64
      header = "no header"
    } else if (dformat == FORMAT$BIN) { # .bin and not movisens, could be GENEActiv or Parmay Matrix
      mon = inspect_binFile_brand(filename = datafile)
      if (mon == MONITOR$PARMAY_MTX && utils::packageVersion("GGIRread") < "1.0.4") {
        stop("Please update R package GGIRread to version 1.0.4 or higher", call. = FALSE)
      }
      # try read the file as if it is a geneactiv and store output in variable 'isitageneactive'
      if (mon == MONITOR$GENEACTIV) {
        if (all(names(isitageneactive) %in% c("header", "data.out") == TRUE)) {
          isitageneactive = GGIRread::readGENEActiv(filename = datafile, start = 0, end = 1)
          tmp = unlist(strsplit(unlist(as.character(isitageneactive$header$SampleRate))," "))[1]
          # occasionally we'll get a decimal seperated by comma; if so, replace the comma with a dot
          tmp = sub(",", ".", tmp, fixed = TRUE)
          sf = as.numeric(tmp)
          if (sf != round(sf)) sf = round(sf) # Round GENEACtiv sampling rate. In g.readaccfile the raw data is resampled to this sampling rate
          #also try to read sf from first page header
          sf_r = sf
          csvr = c()
          suppressWarnings(expr = {
            try(expr = {csvr = as.matrix(read.csv(datafile, nrow = 10,
                                                  skip = 200, sep = ""))
            }, silent = TRUE)
          })
          if (length(csvr) > 1) {
            for (ii in 1:nrow(csvr)) {
              tmp3 = unlist(strsplit(as.character(csvr[ii,1]),"quency:")) #part of 'frequency'
              if (length(tmp3) > 1) {
                # occasionally we'll get a decimal seperated by comma; if so, replace the comma with a dot
                tmp3 = sub(",", ".", tmp3, fixed = TRUE)
                sf_r = as.numeric(tmp3)
              }
            }
            if (length(sf_r) > 0 && !is.na(sf_r)) {
              if (sf_r != sf && abs(sf_r - sf) > 5) { # use pageheader sample frequency if it is not the same as header sample frequency
                sf = sf_r
                warning(paste0("sample frequency used from page header: ", sf, " Hz"))
              }
            }
          }
        } else {
          stop(paste0("\nError processing ", filename, ": possibibly a corrupt GENEActive file"))
        }
      } else if (mon == MONITOR$PARMAY_MTX) {
        header = NULL
        sf = GGIRread::readParmayMatrix(datafile, output = "sf")
      } else {
        stop(paste0("\nError processing ", filename, ": unrecognised .bin file"))
      }
    } else if (dformat == FORMAT$CSV) { #no checks for corrupt file yet...maybe not needed for csv-format?
      if (mon == MONITOR$GENEACTIV) {
        tmp = read.csv(datafile, nrow = 50, skip = 0)
        tmp = as.character(tmp[which(as.character(tmp[,1]) == "Measurement Frequency"),2])
        tmp = as.numeric(unlist(strsplit(tmp," "))[1])
        # occasionally we'll get a decimal seperated by comma; if so, replace the comma with a dot
        tmp = sub(",", ".", tmp, fixed = TRUE)
        sf = as.numeric(tmp)
        
      } else if (mon == MONITOR$ACTIGRAPH) {
        tmp = read.csv(datafile, nrow = 9, skip = 0)
        tmp = colnames(tmp)
        tmp = as.character(unlist(strsplit(tmp,".Hz"))[1])
        # tmp3 = as.character(unlist(strsplit(tmp2,"yy.at."))[2])
        # following suggestion by XInyue on github https://github.com/wadpac/GGIR/issues/102 replaced by:
        tmp = as.character(unlist(strsplit(tmp, ".at.",fixed = T))[2])
        # occasionally we'll get a decimal seperated by comma; if so, replace the comma with a dot
        tmp = sub(",", ".", tmp, fixed = TRUE)
        sf = as.numeric(tmp)
      } else if (mon == MONITOR$AXIVITY) {
        # sample frequency is not stored
        tmp = read.csv(datafile, nrow = 100000, skip = 0)
        tmp = as.numeric(as.POSIXct(tmp[, 1], origin = "1970-01-01"))
        sf = length(tmp) / (tmp[length(tmp)] - tmp[1])
        sf = floor((sf) / 5 ) * 5 # round down to nearest integer of 5, we never want to assume that there is more frequency content in a signal than there truly is
      }
    } else if (dformat == FORMAT$CWA) {
      PP = GGIRread::readAxivity(datafile, start = 1, end = 10, desiredtz = desiredtz)
      H = PP$header
      sf = H$frequency
    } else if (dformat == FORMAT$GT3X) {
      info = try(expr = {read.gt3x::parse_gt3x_info(datafile, tz = desiredtz)},silent = TRUE)
      if (inherits(info, "try-error") == TRUE || is.null(info)) {
        warning(paste0("\nFile info could not be extracted from ", datafile), call. = FALSE)
        sf = NULL # set to NULL in order to tell other GGIR functions that file was corrupt
      } else {
        info = info[lengths(info) != 0] # remove odd NULL in the list
        sf = info[["Sample Rate"]]
      }
    }
    invisible(list(dformat = dformat, mon = mon, sf = sf, datafile = datafile))
  }
  #======================================================================
  # main script
  filename = unlist(strsplit(as.character(datafile),"/"))
  filename = filename[length(filename)]
  monnames = c("genea", "geneactive", "actigraph", "axivity", "movisens", "verisense", "parmay_mtx") #monitor names
  fornames = c("bin", "csv", "wav", "cwa", "csv", "gt3x", "BIN") #format names
  
  if (length(filename) == 0) {
    warning("no files to analyse", call. = FALSE)
  }
  
  if (length(params_rawdata[["rmc.firstrow.acc"]]) == 0) {
    INFI = getbrand(filename, datafile)
    mon = INFI$mon
    dformat = INFI$dformat
    sf = INFI$sf
    datafile = INFI$datafile
  } else {
    dformat = FORMAT$AD_HOC_CSV
    mon = MONITOR$AD_HOC
    Pusercsvformat = read.myacc.csv(rmc.file = datafile,
                                    rmc.nrow = 5,
                                    rmc.dec = params_rawdata[["rmc.dec"]],
                                    rmc.firstrow.acc = params_rawdata[["rmc.firstrow.acc"]],
                                    rmc.firstrow.header = params_rawdata[["rmc.firstrow.header"]],
                                    rmc.header.length = params_rawdata[["rmc.header.length"]],
                                    rmc.col.acc = params_rawdata[["rmc.col.acc"]],
                                    rmc.col.temp = params_rawdata[["rmc.col.temp"]],
                                    rmc.col.time = params_rawdata[["rmc.col.time"]],
                                    rmc.unit.acc = params_rawdata[["rmc.unit.acc"]],
                                    rmc.unit.temp = params_rawdata[["rmc.unit.temp"]],
                                    rmc.unit.time = params_rawdata[["rmc.unit.time"]],
                                    rmc.format.time = params_rawdata[["rmc.format.time"]],
                                    rmc.bitrate = params_rawdata[["rmc.bitrate"]],
                                    rmc.dynamic_range = params_rawdata[["rmc.dynamic_range"]],
                                    rmc.unsignedbit = params_rawdata[["rmc.unsignedbit"]],
                                    rmc.origin = params_rawdata[["rmc.origin"]],
                                    rmc.desiredtz = params_rawdata[["rmc.desiredtz"]],
                                    rmc.configtz = params_rawdata[["rmc.configtz"]],
                                    rmc.sf = params_rawdata[["rmc.sf"]],
                                    rmc.headername.sf = params_rawdata[["rmc.headername.sf"]],
                                    rmc.headername.sn = params_rawdata[["rmc.headername.sn"]],
                                    rmc.headername.recordingid = params_rawdata[["rmc.headername.sn"]],
                                    rmc.header.structure = params_rawdata[["rmc.header.structure"]],
                                    rmc.check4timegaps = params_rawdata[["rmc.check4timegaps"]],
                                    rmc.scalefactor.acc = params_rawdata[["rmc.scalefactor.acc"]],
                                    desiredtz = desiredtz,
                                    configtz = configtz)
    if (inherits(Pusercsvformat$header, "character") && Pusercsvformat$header == "no header") {
      sf = params_rawdata[["rmc.sf"]]
    } else {
      sf = as.numeric(Pusercsvformat$header["sample_rate",1])
    }
    if (is.null(sf) || is.na(sf)) {
      stop("\nFile header doesn't specify sample rate. Please provide rmc.sf value to process ", datafile)
    } else if (sf == 0) {
      stop("\nFile header doesn't specify sample rate. Please provide a non-zero rmc.sf value to process ", datafile)
    }
  }
  
  if (mon == MONITOR$GENEACTIV && dformat == FORMAT$CSV) {
    stop(paste0("The GENEActiv csv reading functionality is deprecated in",
                " GGIR from version 2.6-4 onwards. Please, use either",
                " the GENEActiv bin files or try to read the csv files with",
                " GGIR::read.myacc.csv"), call. = FALSE)
  }
  
  if (dformat == FORMAT$BIN) {
    if (mon == MONITOR$GENEACTIV) {
      H = GGIRread::readGENEActiv(filename = datafile, start = 0, end = 1)$header
    } else if (mon == MONITOR$MOVISENS) {
      H = "file does not have header" # these files have no header
      xmlfile = paste0(dirname(datafile), "/unisens.xml")
      if (file.exists(xmlfile)) {
        # as we are only interested in two character fields
        # try extract value by reading xml as text file to avoid having to add
        # software dependencies
        header = as.character(read.csv(xmlfile, nrow = 1))
        tmp1 = unlist(strsplit(header, "measurementId="))[2]
        ID = gsub(pattern = " ",replacement = "",  unlist(strsplit(tmp1, " timestampStart"))[1])
        
        header = paste0(read.csv(xmlfile, nrow = 10, skip = 2), collapse = " ")
        tmp1 = unlist(strsplit(header, "sensorSerialNumber value="))[2]
        SN = unlist(strsplit(tmp1, "/>"))[1]
        header = data.frame(serialnumber = SN, ID = ID)
        if (length(header) > 1) {
          H = t(header)
        }
        filename = unlist(strsplit(as.character(datafile),"/"))
        filename = filename[length(filename) - 1]
      }
    } else if (mon == MONITOR$PARMAY_MTX) {
      H = "file does not have header" # these files have no header
    }
  } else if (dformat == FORMAT$CSV) {
    if (mon == MONITOR$ACTIGRAPH) {
      H = read.csv(datafile, nrow = 9, skip = 0)
    } else if (mon == MONITOR$AXIVITY) {
      H = "file does not have header" # these files have no header
    }
  } else if (dformat == FORMAT$CWA) {
    PP = GGIRread::readAxivity(datafile, start = 1, end = 10, desiredtz = desiredtz)
    H = PP$header
    
  } else if (dformat == FORMAT$AD_HOC_CSV) { # csv data in a user-specified format
    header = Pusercsvformat$header
  } else if (dformat == FORMAT$GT3X) { # gt3x
    info = try(expr = {read.gt3x::parse_gt3x_info(datafile, tz = desiredtz)},silent = TRUE)
    if (inherits(info, "try-error") == TRUE || is.null(info)) {
      warning(paste0("\nFile info could not be extracted from ", datafile), call. = FALSE)
      sf = NULL # set to NULL in order to tell other GGIR functions that file was corrupt
      H = NULL
      header = NULL
    } else {
      info = info[lengths(info) != 0] # remove odd NULL in the list
      
      H = matrix("", length(info), 2)
      H[, 1] = names(info)
      for (ci in 1:length(info)) {
        if (inherits(info[[ci]], "POSIXct") == TRUE) {
          H[ci, 2] = format(info[[ci]])
        } else {
          H[ci, 2] = as.character(info[[ci]])
        }
      }
      sf = as.numeric(H[which(H[,1] == "Sample Rate"), 2])
    }
  }
  if (is.null(sf) || sf == 0) {
    warning(paste0("\nSample frequency not recognised in ", basename(datafile)), call. = FALSE)
  }
  
  if (dformat != FORMAT$AD_HOC_CSV && is.null(sf) == FALSE) {
    H = as.matrix(H)
    if (ncol(H) == 3 && dformat == FORMAT$CSV && mon == MONITOR$ACTIGRAPH) {
      if (length(which(is.na(H[,2]) == FALSE)) == 0) {
        H = as.matrix(H[,1])
      }
    }
    if (ncol(H) == 1 && dformat == FORMAT$CSV) {
      if (mon == MONITOR$ACTIGRAPH) {
        vnames = c("Number:","t Time","t Date",":ss)","d Time","d Date","Address:","Voltage:","Mode =")
        Hvalues = Hnames = rep(" ",length(H))
        firstline = colnames(H)
        for (run in 1:length(H)) {
          for (runb in 1:length(vnames)) {
            tmp = unlist(strsplit(H[run],vnames[runb]))
            if (length(tmp) > 1) {
              Hnames[run] = paste(tmp[1], vnames[runb], sep = "")
              Hvalues[run] = paste(tmp[2], sep = "")
            }
          }
        }
        H = cbind(Hnames,Hvalues)
        H = rbind(c("First line",firstline),H)
      } else {
        H = cbind(c(1:length(H)),H)
      }
    }
    if (dformat == FORMAT$CWA) {
      header = data.frame(value = H, row.names = rownames(H), stringsAsFactors = TRUE)
    } else {
      if ((mon == MONITOR$GENEACTIV && dformat == FORMAT$BIN) || (mon == MONITOR$MOVISENS && length(H) > 0)) {
        varname = rownames(as.matrix(H))
        H = data.frame(varname = varname,varvalue = as.character(H), stringsAsFactors = TRUE)
      } else if (dformat != FORMAT$AD_HOC_CSV) {
        if (length(H) > 1 && class(H)[1] == "matrix") H = data.frame(varname = H[,1],varvalue = H[,2], stringsAsFactors = TRUE)
      }
    }
    if (dformat != FORMAT$CWA && length(H) > 1 && (class(H)[1] == "matrix" || class(H)[1] == "data.frame")) {
      RowsWithData = which(is.na(H[,1]) == FALSE)
      header = data.frame(value = H[RowsWithData, 2], row.names = H[RowsWithData, 1], stringsAsFactors = TRUE)
    }
    if (H[1,1] == "file does not have header") { #no header
      header = "no header"
    }
    if (mon == MONITOR$ACTIGRAPH && dformat != FORMAT$GT3X) {
      verisense_check = substr(colnames(read.csv(datafile,nrow = 1)[1]), start = 36, stop = 44)
      if (identical('Verisense', toString(verisense_check))) {
        mon = MONITOR$VERISENSE
      }
    }
  }
  if (!is.null(sf)) {
    # detect dot or comma separator in the data file
    op <- options(warn = -1) # turn off warnings
    on.exit(options(op))
    suppressWarnings(expr = {decn = g.dotorcomma(datafile, dformat, mon,
                                                 rmc.dec = params_rawdata[["rmc.dec"]])})
    options(warn = 0) # turn on warnings
  } else {
    decn = "."
  }
  monc = mon
  monn = ifelse(mon > 0, monnames[mon], "unknown")
  dformc = dformat
  dformn = fornames[dformat]
  invisible(list(header = header, monc = monc, monn = monn,
                 dformc = dformc, dformn = dformn, sf = sf, decn = decn, filename = filename))
}

library(GGIR)
testthat::context("g.inspectfile CSV")

quick_gzip = function(filename) {
  fext = strsplit(filename, "[.]")[[1]]
  fext = fext[length(fext)]
  destnameT = tempfile(fileext = paste0(".", fext, ".gz"))
  x = readLines(filename, warn = FALSE)
  gzcon = gzfile(destnameT, "w");
  on.exit({
    try({close(gzcon)})
  })
  write(x, gzcon);
  return(destnameT)
}

testthat::test_that("g.inspectfile works with a CSV.gz", {
  testthat::skip_on_cran()

  Ax3CsvFile  = system.file("testfiles/ax3_testfile_unix_timestamps.csv", package = "GGIR")[1]
  Ax6CsvFile  = system.file("testfiles/ax6_testfile_formatted_timestamps.csv", package = "GGIR")[1]

  testthat::expect_no_error({
    ax3out = g.inspectfile(Ax3CsvFile)
  })
  testthat::expect_no_error({
    ax6out = g.inspectfile(Ax6CsvFile)
  })

  testthat::expect_no_error({
    ax3out_gz = g.inspectfile(quick_gzip(Ax3CsvFile))
  })
  testthat::expect_no_error({
    ax6out_gz = g.inspectfile(quick_gzip(Ax6CsvFile))
  })

  ax3out$filename = ax3out_gz$filename =  NULL
  ax6out$filename = ax6out_gz$filename =  NULL

  testthat::expect_equal(ax3out, ax3out_gz)
  testthat::expect_equal(ax6out, ax6out_gz)
})

get_starttime_weekday_truncdata = function(monc, dformat, data, 
                                           header, desiredtz, sf, datafile,
                                           ws2, configtz = NULL) {
  # ensures that first window starts at logical timepoint relative to its size
  # (15,30,45 or 60 minutes of each hour)
  start_meas = ws2/60 
  
  starttime = g.getstarttime( # this will return a POSIXlt object
    datafile = datafile,
    data = data,
    mon = monc,
    dformat = dformat,
    desiredtz = desiredtz,
    configtz = configtz
  )

  # assess weekday
  wday = starttime$wday #day of the week 0-6 and 0 is Sunday
  wday = wday + 1
  weekdays = c("Sunday","Monday","Tuesday","Wednesday","Thursday","Friday","Saturday")
  wdayname = weekdays[wday]

  # assess how much data to delete till next 15 minute period
  secshift = 60 - starttime$sec # shift in seconds needed
  if (secshift == 60) {
    secshift = 0 # if starttime is 00:00 then we do not want to remove data
  } else {
    starttime$min = starttime$min + 1 # shift in minutes needed (+1 one to account for seconds comp)
  }

  minshift = start_meas - (starttime$min %% start_meas)
  if (minshift == start_meas) {
    minshift = 0
  }

  sampleshift = (minshift * 60 * sf) + (secshift * sf) # derive sample shift
  sampleshift = floor(sampleshift)
  if (sampleshift > 1) {
    data = data[-c(1 : sampleshift),] # delete data accordingly
  }

  # recalculate the timestamp
  starttime$min = starttime$min + minshift # if the result is >= 60, hours (and possibly date) will adjust automatically
  starttime$sec = 0

  invisible(
    list(
      starttime = starttime,
      wday = wday,
      wdayname = wdayname,
      data = data
    )
  )
}

g.part1 = function(datadir = c(), metadatadir = c(), f0 = 1, f1 = c(), myfun = c(),
                   params_metrics = c(), params_rawdata = c(),
                   params_cleaning = c(), params_general = c(), verbose = TRUE, ...) {
  #----------------------------------------------------------
  # Extract and check parameters
  input = list(...)
  params = extract_params(params_metrics = params_metrics,
                          params_rawdata = params_rawdata,
                          params_cleaning = params_cleaning,
                          params_general = params_general,
                          input = input,
                          params2check = c("metrics", "rawdata",
                                           "cleaning", "general")) # load default parameters
  params_metrics = params$params_metrics
  params_rawdata = params$params_rawdata
  params_cleaning = params$params_cleaning
  params_general = params$params_general
  
  if (f1 == 0) warning("\nWarning: f1 = 0 is not a meaningful value")
  filelist = isfilelist(datadir)
  if (ismovisens(datadir)) filelist = TRUE

  # list all accelerometer files
  dir2fn = datadir2fnames(datadir, filelist)
  fnames = dir2fn$fnames
  fnamesfull = dir2fn$fnamesfull
  filesizes = file.size(fnamesfull) # in bytes
  bigEnough = which(filesizes/1e6 > params_rawdata[["minimumFileSizeMB"]])
  fnamesfull = fnamesfull[bigEnough]
  
  if (length(bigEnough) > 0) {
    fnames_toosmall = fnames[-bigEnough]
  } else {
    fnames_toosmall = fnames
  }
  
  fnames = fnames[bigEnough]

  if (verbose && length(fnames_toosmall) > 0) {
    warning(paste0("\nSkipping files that are too small for analysis: ", toString(fnames_toosmall),
                   " (configurable with parameter minimumFileSizeMB)."), call. = FALSE) 
  }

  if (length(fnamesfull) == 0) {
    stop(paste0("\nNo files to analyse. Check that there are accelerometer files ",
                "in the directory specified with argument datadir"), call. = FALSE)
  }

  # create output directory if it does not exist
  if (!file.exists(metadatadir)) {
    dir.create(metadatadir)
    dir.create(file.path(metadatadir, "meta"))
    dir.create(file.path(metadatadir, "meta", "basic"))
    dir.create(file.path(metadatadir, "results"))
    dir.create(file.path(metadatadir, "results", "QC"))
  }
  use.temp = TRUE
  daylimit = FALSE

  # check access permissions
  Nfile_without_readpermission = length(which(file.access(paste0(fnamesfull), mode = 4) == -1)) #datadir,"/",
  if (Nfile_without_readpermission > 0) {
    if (verbose == TRUE) cat("\nChecking that user has read access permission for all files in data directory: No")
    warning(paste0("\nThere are/is ", Nfile_without_readpermission,
                   " file(s) in directory specified with argument datadir for which the user does not have read access permission\n"))
  } else {
    if (verbose == TRUE) cat("\nChecking that user has read access permission for all files in data directory: Yes\n")
  }

  if (length(f0) ==  0) f0 = 1
  if (length(f1) ==  0) f1 = length(fnames)
  if (f0 > length(fnames)) f0 = 1
  if (f1 > length(fnames)) f1 = length(fnames)
  #========================================================================
  # check which files have already been processed, such that no double work is done
  # ffdone a matrix with all the binary filenames that have been processed
  ffdone = dir(file.path(metadatadir, "meta", "basic"))

  if (length(ffdone) > 0) {
    for (ij in 1:length(ffdone)) {
      tmp = unlist(strsplit(ffdone[ij],".RData"))
      tmp2 = unlist(strsplit(tmp[1],"meta_"))
      ffdone[ij] = tmp2[2]
    }
  }

  # Following lines turned off because does not work when using subfolders
  # It seems best to leave sorting efforts until the very end when reports are created
  # fnames = sort(fnames)
  # fnamesfull = sort(fnamesfull)


  #=========================================================
  # Declare core functionality, which at the end of this g.part1 is either
  # applied to the file in parallel with foreach or serially with a loop
  main_part1 = function(i, params_metrics, params_rawdata,
                        params_cleaning, params_general, datadir, fnames, fnamesfull,
                        myfun, filelist, ffdone,
                        use.temp, daylimit, metadatadir, verbose) {
    tail_expansion_log = NULL
    if (params_general[["print.filename"]] == TRUE & verbose == TRUE) {
      cat(paste0("\nFile name: ",fnames[i]))
    }
    # if (filelist == TRUE) {
    datafile = as.character(fnamesfull[i])
    # } else {
    #   # datafile = paste0(datadir,"/",fnames[i])
    #   datafile = fnamesfull[i] #paste0(datadir,"/",fnames[i])
    # }

    #=========================================================
    #check whether file has already been processed
    #by comparing filename to read with list of processed files
    if (params_general[["overwrite"]] == FALSE) {
      fname_without = unlist(strsplit(fnames[i], ".csv"))[1]
      fname_without = unlist(strsplit(fname_without,"[.]RD"))[1]

      ffdone_without = unlist(strsplit(ffdone, ".csv"))

      if (length(which(ffdone_without == fname_without)) > 0) {
        return() # skip this file because it was analysed before
      } 
    }
    #=============================================================
    # Inspect file (and store output later on)
    I = g.inspectfile(datafile, desiredtz = params_general[["desiredtz"]],
                        params_rawdata = params_rawdata,
                        configtz = params_general[["configtz"]])
    if (verbose == TRUE) cat(paste0("\nP1 file ",i))
    turn.do.cal.back.on = FALSE
    if (params_rawdata[["do.cal"]] == TRUE & I$dformc == FORMAT$WAV) { # do not do the auto-calibration for wav files (because already done in pre-processign)
      params_rawdata[["do.cal"]] = FALSE
      turn.do.cal.back.on = TRUE
    }
    data_quality_report_exists = file.exists(paste0(metadatadir, "/results/QC/data_quality_report.csv"))
    assigned.backup.cal.coef = FALSE
    if (length(params_rawdata[["backup.cal.coef"]]) > 0) {
      if (params_rawdata[["backup.cal.coef"]] == "retrieve") {
        if (data_quality_report_exists == TRUE) { # use the data_quality_report as backup for calibration coefficients
          params_rawdata[["backup.cal.coef"]] = paste0(metadatadir, "/results/QC/data_quality_report.csv")
          assigned.backup.cal.coef = TRUE
        }
      } else if (params_rawdata[["backup.cal.coef"]] == "redo") { #ignore the backup calibration coefficients, and derive them again
        params_rawdata[["backup.cal.coef"]] = c()
        assigned.backup.cal.coef = TRUE
      } else if (params_rawdata[["backup.cal.coef"]] != "redo" & params_rawdata[["backup.cal.coef"]] != "retrieve") {
        # Do nothing, backup.cal.coef is the path to the csv-file with calibration coefficients
        assigned.backup.cal.coef = TRUE
      }
    }
    #data_quality_report.csv does not exist and there is also no ot
    if (assigned.backup.cal.coef == FALSE) params_rawdata[["backup.cal.coef"]] = c()
    #--------------------------------------
    # Set default C object
    C = list(cal.error.end = 0, cal.error.start = 0)
    C$scale = c(1,1,1)
    C$offset = c(0,0,0)
    C$tempoffset =  c(0,0,0)
    C$QCmessage = "Autocalibration not done"
    C$npoints = 0
    C$nhoursused = 0
    C$use.temp = use.temp
    Cdefault = C
    if (params_rawdata[["do.cal"]] == TRUE & length(params_rawdata[["backup.cal.coef"]]) == 0 &
        is.null(I$sf) == FALSE) {
      # cat(paste0("\n",rep('-',options()$width),collapse=''))
      if (verbose == TRUE) cat("\n")
      if (verbose == TRUE) cat("\nInvestigate calibration of the sensors with function g.calibrate:\n")
      C = g.calibrate(datafile,
                      params_rawdata = params_rawdata,
                      params_general = params_general,
                      params_cleaning = params_cleaning,
                      inspectfileobject = I,
                      verbose = verbose)
    }
    if (turn.do.cal.back.on == TRUE) {
      params_rawdata[["do.cal"]] = TRUE
    }
    cal.error.end = C$cal.error.end
    cal.error.start = C$cal.error.start
    if (length(cal.error.start) == 0) {
      #file too shortcorrupt to even calculate basic calibration value
      cal.error.start = NA
    }
    check.backup.cal.coef = FALSE
    if (is.na(cal.error.start) == T | length(cal.error.end) == 0) {
      C$scale = c(1,1,1); C$offset = c(0,0,0); C$tempoffset = c(0,0,0)
      check.backup.cal.coef = TRUE
    } else {
      if (cal.error.start < cal.error.end) {
        C$scale = c(1,1,1); C$offset = c(0,0,0); C$tempoffset =  c(0,0,0)
        check.backup.cal.coef = TRUE
      }
    }
    if (length(params_rawdata[["backup.cal.coef"]]) > 0) check.backup.cal.coef = TRUE
    #if calibration fails then check whether calibration coefficients are provided in a separate csv-spreadsheet
    # this csv spreadhseet needs to be created by the end-user and should contain:
    # column with names of the accelerometer files
    # three columns respectively named scale.x, scale.y, and scale.z
    # three columns respectively named offset.x, offset.y, and offset.z
    # three columns respectively named temperature.offset.x, temperature.offset.y, and temperature.offset.z
    # the end-user can generate this document based on calibration analysis done with the same accelerometer device.
    if (length(params_rawdata[["backup.cal.coef"]]) > 0 & check.backup.cal.coef == TRUE) {
      bcc.data = data.table::fread(params_rawdata[["backup.cal.coef"]], data.table = FALSE)
      if (isTRUE(params_rawdata[["do.cal"]]) & verbose == TRUE) {
        cat("\nRetrieving previously derived calibration coefficients")
      }
      bcc.data$filename = as.character(bcc.data$filename)
      for (nri in 1:nrow(bcc.data)) {
        tmp = unlist(strsplit(as.character(bcc.data$filename[nri]),"meta_"))
        if (length(tmp) > 1) {
          bcc.data$filename[nri] = tmp[2]
          bcc.data$filename[nri] = unlist(strsplit(bcc.data$filename[nri],".RData"))[1]
        }
      }
      if (length(which(as.character(bcc.data$filename) == fnames[i])) > 0) {
        bcc.i = which(bcc.data$filename == fnames[i])
        bcc.cal.error.start = which(colnames(bcc.data) == "cal.error.start")
        bcc.cal.error.end = which(colnames(bcc.data) == "cal.error.end")
        bcc.scalei = which(colnames(bcc.data) == "scale.x" | colnames(bcc.data) == "scale.y" | colnames(bcc.data) == "scale.z")
        bcc.offseti = which(colnames(bcc.data) == "offset.x" | colnames(bcc.data) == "offset.y" | colnames(bcc.data) == "offset.z")
        bcc.temp.offseti = which(colnames(bcc.data) == "temperature.offset.x" | colnames(bcc.data) == "temperature.offset.y" | colnames(bcc.data) == "temperature.offset.z")
        bcc.meantempcali = which(colnames(bcc.data) == "meantempcal")
        bcc.QCmessagei = which(colnames(bcc.data) == "QCmessage")
        bcc.npointsi = which(colnames(bcc.data) == "n.10sec.windows")
        bcc.nhoursusedi = which(colnames(bcc.data) == "n.hours.considered")
        bcc.use.tempi = which(colnames(bcc.data) == "use.temperature")
        C$scale = as.numeric(bcc.data[bcc.i[1],bcc.scalei])
        C$offset = as.numeric(bcc.data[bcc.i[1],bcc.offseti])
        C$tempoffset =  as.numeric(bcc.data[bcc.i[1],bcc.temp.offseti])
        C$meantempcal = bcc.data[bcc.i[1], bcc.meantempcali]
        C$cal.error.start = as.numeric(bcc.data[bcc.i[1],bcc.cal.error.start])
        C$cal.error.end = as.numeric(bcc.data[bcc.i[1],bcc.cal.error.end])
        C$QCmessage = bcc.data[bcc.i[1],bcc.QCmessagei]
        C$npoints = bcc.data[bcc.i[1], bcc.npointsi]
        C$nhoursused = bcc.data[bcc.i[1], bcc.nhoursusedi]
        C$use.temp = bcc.data[bcc.i[1], bcc.use.tempi]
        if (verbose == TRUE) {
          cat(paste0("\nRetrieved Calibration error (g) before: ",as.numeric(bcc.data[bcc.i[1],bcc.cal.error.start])))
          cat(paste0("\nRetrieved Callibration error (g) after: ",as.numeric(bcc.data[bcc.i[1],bcc.cal.error.end])))
          cat(paste0("\nRetrieved offset correction ",c("x","y","z"),": ",C$offset))
          cat(paste0("\nRetrieved scale correction ",c("x","y","z"),": ",C$scale))
          cat(paste0("\nRetrieved tempoffset correction ",c("x","y","z"),": ",C$tempoffset))
          cat("\n")
        }
      } else {
        # cat("\nNo matching filename found in backup.cal.coef\n")
        # cat(paste0("\nCheck that filename ",fnames[i]," exists in the csv-file\n"))
        if (params_rawdata[["do.cal"]] == TRUE) { # If no matching filename could be found, then try to derive the calibration coeficients in the normal way
          if (verbose == TRUE) {
            cat("\n")
            cat("\nInvestigate calibration of the sensors with function g.calibrate:\n")
          }
          C = g.calibrate(datafile,
                          params_rawdata = params_rawdata,
                          params_general = params_general,
                          params_cleaning = params_cleaning,
                          inspectfileobject = I,
                          verbose = verbose)
          if (is.null(C)) C = Cdefault # in case gt3x file is corrupt then this value is needed
        }
      }
    }
    #------------------------------------------------
    if (verbose == TRUE) cat("\nExtract signal features (metrics) with the g.getmeta function:\n")
    M = g.getmeta(datafile,
                  params_metrics = params_metrics,
                  params_rawdata = params_rawdata,
                  params_general = params_general,
                  params_cleaning = params_cleaning,
                  daylimit = daylimit,
                  tempoffset = C$tempoffset, scale = C$scale, offset = C$offset,
                  meantempcal = C$meantempcal,
                  myfun = myfun,
                  inspectfileobject = I,
                  verbose = verbose)

    if (!is.null(params_general[["recordingEndSleepHour"]])) {
      # Identify gap between last timestamp and following midnight
      ws3 = M$windowsizes[1]
      ws2 = M$windowsizes[2]
      # Check whether gap is less then criteria
      last_ts = c(Sys.time(), Sys.time())
      secs_to_midnight = c(0, 0)
      lastTimeLong = as.character(tail(M$metalong$timestamp, n = 1))
      lastTimeShort = as.character(tail(M$metashort$timestamp, n = 1))
      last_ts[1] = iso8601chartime2POSIX(x = lastTimeLong, tz = params_general[["desiredtz"]])
      last_ts[2] = iso8601chartime2POSIX(x = lastTimeShort, tz = params_general[["desiredtz"]])
      refhour = 24 + params_general[["dayborder"]]
      for (wsi in 1:2) {
        secs_to_midnight[wsi] = (refhour * 3600) -
          (as.numeric(format(last_ts[wsi], format = "%H", tz = params_general[["desiredtz"]])) * 3600 +
             as.numeric(format(last_ts[wsi], format = "%M", tz = params_general[["desiredtz"]])) * 60  +
             as.numeric(format(last_ts[wsi], format = "%S", tz = params_general[["desiredtz"]])))
      }
      # only expand if recording ends at 19PM or later
      max_expand_time = (refhour - (params_general[["recordingEndSleepHour"]] - params_general[["dayborder"]])) * 3600
      if (secs_to_midnight[1] <= max_expand_time) {
        # If yes, expand data
        secs_to_midnight = secs_to_midnight + (8 * 3600) # also add 8 hour till the morning
        N_long_epochs_expand = ceiling(secs_to_midnight[1] / ws2) + 1
        N_short_epochs_expand = ceiling(secs_to_midnight[2] / ws3) + 1
        if (N_short_epochs_expand / (ws2 / ws3) < N_long_epochs_expand) {
          N_short_epochs_expand = N_long_epochs_expand * (ws2 / ws3)
        }
        # N_short_epochs_expand = ((N_long_epochs_expand) * (ws2/ws3))
        # Expand metashort
        NR = nrow(M$metashort)
        metashort_expand = M$metashort[NR,]
        metashort_expand[, grep(pattern = "timestamp|angle", x = names(metashort_expand), invert = TRUE, value = FALSE)] = 0
        if ("EN" %in% names(metashort_expand)) {
          metashort_expand$EN = 1
        }
        expand_indices = (NR + 1):(NR + N_short_epochs_expand)
        expand_tsPOSIX = seq(last_ts[2] + ws3, last_ts[2] + (N_short_epochs_expand * ws3), by = ws3)
        M$metashort[expand_indices,] = metashort_expand
        M$nonwear[expand_indices] = 0
        M$metashort$timestamp[expand_indices] = POSIXtime2iso8601(expand_tsPOSIX, tz = params_general[["desiredtz"]])
        anglecol = grep(pattern = "angle", x = names(metashort_expand), value = FALSE)
        if (length(anglecol) > 0) {
          M$metashort[expand_indices,anglecol] = round(sin((1:length(expand_indices)) / (ws2/ws3))) * 15
        }
        tail_expansion_log = list(short = length(expand_indices))
        # Expand metalong
        NR = nrow(M$metalong)
        metalong_expand = M$metalong[NR,]
        metalong_expand[, grep(pattern = "timestamp", x = names(metalong_expand), invert = TRUE, value = FALSE)] = 0
        metalong_expand[, "nonwearscore"] = -1
        metalong_expand$en = tail(M$metalong$en, n = 1)
        expand_indices = (NR + 1):(NR + N_long_epochs_expand)
        expand_tsPOSIX = seq(last_ts[1] + ws2, last_ts[1] + (N_long_epochs_expand * ws2), by = ws2)
        M$metalong[expand_indices,] = metalong_expand
        M$metalong$timestamp[expand_indices] = POSIXtime2iso8601(expand_tsPOSIX, tz = params_general[["desiredtz"]])
        # Keep log of data expansion
        tail_expansion_log[["long"]] = length(expand_indices)
      } else {
        tail_expansion_log = NULL
      }
    }
    #------------------------------------------------
    if (verbose == TRUE) cat("\nSave .RData-file with: calibration report, file inspection report and all signal features...\n")
    # remove directory in filename if present
    filename = unlist(strsplit(fnames[i],"/"))
    if (length(filename) > 0) {
      filename = filename[length(filename)]
    } else {
      filename = fnames[i]
    }
    filefoldername = filename_dir = fnames[i] # not used elsewhere in GGIR, but just keeping it for consistency
    if (length(unlist(strsplit(fnames[1], "[.]RD"))) == 1) { # to avoid getting .RData.RData
      filename = paste0(filename,".RData")
    }
    GGIRversion = utils::packageVersion("GGIR")
    desiredtz_part1 = params_general[["desiredtz"]]
    save(M, I, C, filename_dir, filefoldername, tail_expansion_log, GGIRversion,
         desiredtz_part1, file = paste0(metadatadir, "/meta/basic/meta_", filename))
    rm(M); rm(I); rm(C)
  } # end of main_part1

  #--------------------------------------------------------------------------------
  # Run the code either parallel or in serial (file index starting with f0 and ending with f1)
  if (params_general[["do.parallel"]] == TRUE) {
    cores = parallel::detectCores()
    Ncores = cores[1]
    if (Ncores > 3) {

      if (length(params_general[["maxNcores"]]) == 0) params_general[["maxNcores"]] = Ncores
      Ncores2use = min(c(Ncores - 1, params_general[["maxNcores"]], (f1 - f0) + 1))
      if (Ncores2use > 1) {
        cl <- parallel::makeCluster(Ncores2use) # not to overload your computer
        parallel::clusterExport(cl = cl, 
                                varlist = c(unclass(lsf.str(envir = asNamespace("GGIR"), all = T)),
                                            "MONITOR", "FORMAT"),
                                envir = as.environment(asNamespace("GGIR"))
        )
        parallel::clusterEvalQ(cl, Sys.setlocale("LC_TIME", "C"))
        doParallel::registerDoParallel(cl)
      } else {
        # Don't process in parallel if only one core
        params_general[["do.parallel"]] = FALSE
      }
    } else {
      if (verbose == TRUE) cat(paste0("\nparallel processing not possible because number of available cores (",Ncores,") < 4"))
      params_general[["do.parallel"]] = FALSE
    }
  }
  if (params_general[["do.parallel"]] == TRUE) {
    # Check whether we are in development mode (this never applies when the package is installed):
    GGIRinstalled = is.element('GGIR', installed.packages()[,1])
    packages2passon = functions2passon = NULL
    GGIRloaded = "GGIR" %in% .packages()
    if (GGIRloaded) { # not in devleopment mode, pass on package
      packages2passon = 'GGIR'
      errhand = 'pass'
    } else { # development mode, pass on individual function that were sourced by developer
      # pass on functions
      # packages2passon = 'Rcpp'
      functions2passon = c("g.inspectfile", "g.calibrate","g.getmeta", "g.dotorcomma",
                           "g.applymetrics",
                           "g.readaccfile", "updateBlocksize",
                           "g.getstarttime", "POSIXtime2iso8601",
                           "iso8601chartime2POSIX", "datadir2fnames", "read.myacc.csv",
                           "get_nw_clip_block_params",
                           "get_starttime_weekday_truncdata", "ismovisens",
                           "g.extractheadervars", "g.imputeTimegaps", "extract_params",
                           "load_params",
                           "check_params", "detect_nonwear_clipping", "inspect_binFile_brand")
      errhand = 'stop'
      # Note: This will not work for cwa files, because those also need Rcpp functions.
      # So, it is probably best to turn off parallel when debugging cwa data.
    }
    Nmetrics2calc = sum(unlist(params_metrics[c("do.anglex", "do.angley", "do.anglez",
                                                "do.zcx", "do.zcy", "do.zcz",
                                                "do.enmo", "do.lfenmo", "do.en", "do.mad", "do.enmoa",
                                                "do.roll_med_acc_x", "do.roll_med_acc_y", "do.roll_med_acc_z",
                                                "do.dev_roll_med_acc_x", "do.dev_roll_med_acc_y", "do.dev_roll_med_acc_z",
                                                "do.bfen", "do.hfen", "do.hfenplus", "do.lfen",
                                                "do.lfx", "do.lfy", "do.lfz", "do.hfx", "do.hfy", "do.hfz",
                                                "do.bfx", "do.bfy", "do.bfz", "do.brondcounts",
                                                "do.neishabouricounts")]))
    if (Nmetrics2calc > 4) { #Only give warning when user wants more than 4 metrics.
      warning(paste0("\nExtracting many metrics puts higher demands on memory. Please consider",
                     " reducing the value for argument chunksize or setting do.parallel to FALSE"))
    }
    if (params_rawdata[["chunksize"]] > 0.6 & Nmetrics2calc >= 3 & Nmetrics2calc < 6) { # if user wants to extract 3-5 metrics
      params_rawdata[["chunksize"]] = 0.5 # put limit to chunksize, because when processing in parallel memory is more limited
    } else if (params_rawdata[["chunksize"]] > 0.6 & Nmetrics2calc >= 6) { # if user wants to extract more than 5 metrics
      params_rawdata[["chunksize"]] = 0.4 # put limit to chunksize, because when processing in parallel memory is more limited
    }


    if (verbose == TRUE) cat(paste0('\n Busy processing ... see ', metadatadir,'/meta/basic', ' for progress\n'))
    `%myinfix%` = foreach::`%dopar%`
    output_list = foreach::foreach(i = f0:f1, .packages = packages2passon,
                                   .export = functions2passon, .errorhandling = errhand) %myinfix% {
                                     tryCatchResult = tryCatch({
                                       main_part1(i, params_metrics, params_rawdata,
                                                  params_cleaning, params_general, datadir, fnames, fnamesfull,
                                                  myfun, filelist, ffdone,
                                                  use.temp, daylimit, metadatadir, verbose)
                                     })
                                     return(tryCatchResult)
                                   }
    on.exit(parallel::stopCluster(cl))
    for (oli in 1:length(output_list)) { # logged error and warning messages
      if (is.null(unlist(output_list[oli])) == FALSE) {
        if (verbose == TRUE) cat(paste0("\nErrors and warnings for ",fnames[oli]))
        print(unlist(output_list[oli])) # print any error and warnings observed
      }
    }
  } else { # simple loop to process files serially file by file
    errors = list()
    for (i in f0:f1) {
      if (verbose == TRUE) cat(paste0(i, " "))
      function_to_evaluate = expression(
        main_part1(i, params_metrics, params_rawdata,
                   params_cleaning, params_general, datadir, fnames, fnamesfull,
                   myfun, filelist, ffdone,
                   use.temp, daylimit, metadatadir, verbose)
      )
      if (params_general[["use_trycatch_serial"]] == TRUE) {
        tryCatch(
          eval(function_to_evaluate),
          error = function(e) {
            err_msg = conditionMessage(e)
            errors[[as.character(fnames[i])]] <<- err_msg
          }
        )
      } else {
        eval(function_to_evaluate)
      }
    }
    # show logged errors after the loop:
    if (params_general[["use_trycatch_serial"]] == TRUE && verbose == TRUE) {
      if (length(errors) > 0) {
        cat(paste0("\n\nErrors in part 1... for:"))
        cat(paste0("\n-", names(errors), ": ", unlist(errors), collapse = ""))
      }
    }
  }
}

applyExtFunction = function(data, myfun, sf, ws3,interpolationType=1) {
  data = data[, c("x", "y", "z")] # Needed because code below is not able to handle time column yet
  # check myfun object
  check_myfun(myfun, windowsizes=ws3) 
  # unit correction
  unitcorrection = 1 # default is g
  if (myfun$expected_unit != "g") { 
    if (myfun$expected_unit == "mg") {
      unitcorrection = 1000
    } else if (myfun$expected_unit == "ms2") {
      unitcorrection = 9.81
    }
  }
  if (is.logical(myfun$timestamp) == TRUE) {
    myfun$timestamp = c() 
    warning(paste0("Note: If function applyExtFunction is used directly",
                   " then object myfun cannnot carry a logical value",
                   " because the timestamp can only be added in the g.getmeta function",
                   " from which applyExtFunction is called. However,",
                   " you can provide a numeric value to indicate",
                   " start time in seconds since 1-1-1970"), call. = FALSE)
  }
  # sample rate correction
  if (sf != myfun$expected_sample_rate) {
    resampleAcc = function(rawAccel, sf, myfun) {
      step_old = 1/sf
      step_new = 1/myfun$expected_sample_rate
      start = 0
      end = nrow(data)/sf
      rawTime = seq(start, end, step_old)
      timeRes = seq(start, end, step_new)
      rawTime = rawTime[2:length(rawTime)]
      timeRes = timeRes[2:length(timeRes)]
      nr = length(timeRes) - 1
      timeResNew = as.vector(timeRes[1:nr])
      accelRes = matrix(0,nrow = nr, ncol = 3, dimnames = list(NULL, c("x", "y", "z")))
      # at the moment the function is designed for reading the r3 acceleration channels only,
      # because that is the situation of the use-case we had.
      rawLast = nrow(rawAccel)
      accelRes = GGIRread::resample(rawAccel, rawTime, timeRes, rawLast, 
                          type=interpolationType) # this is now the resampled acceleration data
      return(accelRes)
    }
    if (length(myfun$timestamp) == 1) { #resample, but do not apply function yet, because timestamp also needs to be added
      data = resampleAcc(data, sf, myfun)
    } else { # resample and apply function, because timestamp is not needed
      OutputExternalFunction = myfun$FUN(resampleAcc(data, sf, myfun) * unitcorrection, myfun$parameters)
    }
  } 
  if (length(myfun$timestamp) == 1) { # add timestamp and apply function
    st_num = as.numeric(myfun$timestamp) #POSIX converted to numeric time but relative to the desiredtz
    # Note that sample rate is now the expected sample rate, not the original sample rate
    time2beAdded = seq(st_num, (st_num + round(nrow(data)/myfun$expected_sample_rate)),by=1/myfun$expected_sample_rate) 
    LEtim = length( time2beAdded)
    NRda = nrow(data)
    if (LEtim > NRda) {
      time2beAdded = time2beAdded[1:NRda]
    } else if (LEtim < NRda) {
      time2beAdded = c(time2beAdded,time2beAdded[LEtim]+(c(1:(NRda-LEtim))/myfun$expected_sample_rate))
    }
    data = cbind(time2beAdded, data)
    OutputExternalFunction = myfun$FUN(data * unitcorrection, myfun$parameters)
  } else if (length(myfun$timestamp) == 0  & sf == myfun$expected_sample_rate ){ # if no resampling and timestamp column was needed
    OutputExternalFunction = myfun$FUN(data * unitcorrection, myfun$parameters)
  }
  # output resolution correction (either aggregate or repeat)
  if (myfun$outputtype == "numeric" | myfun$outputtype == "character") { # aggregation is possible with averaging, possibly later also allow for character output
    if (is.null(dim(OutputExternalFunction))) { # if OutputExternalFunction is a simple vector then convert it to 1 column matrix
      OutputExternalFunction = as.matrix(OutputExternalFunction)
      if (ncol(OutputExternalFunction) != 1 & nrow(OutputExternalFunction) == 1) OutputExternalFunction = t(OutputExternalFunction)
    }
    
    if (myfun$outputres < ws3) { # if function produces higher resolution output (shorter epoch length) then aggregate rows
      agglevel = rep(1:nrow(OutputExternalFunction)+(3*(ws3/myfun$outputres)),each=ws3/myfun$outputres)
      agglevel = agglevel[1:nrow(OutputExternalFunction)]
      OEF = data.frame(OutputExternalFunction, agglevel=agglevel)
      OEFA = aggregate(OEF,by=list(OEF$agglevel),FUN=myfun$aggfunction)
      OutputExternalFunction = OEFA[,-c(1,ncol(OEFA))]
      OutputExternalFunction = as.matrix(OutputExternalFunction)
      # OutputExternalFunction is now aggregated to ws3 which will enable merging it with metashort
    } else if (myfun$outputres > ws3) { # if function produces longer epoch length then repeat rows
      indx = rep(seq_len(nrow(OutputExternalFunction)), each = myfun$outputres/ws3)
      OutputExternalFunction = as.matrix(OutputExternalFunction[indx, ])
    }
  }
  return(OutputExternalFunction)
}

is_this_a_dst_night = function(calendar_date=c(),tz="Europe/London") {
  # calendardata in dd/mm/yyyy format
  # this function investigates whether it is a dst night
  splitdate = unlist(strsplit(calendar_date,"/"))
  t0 = as.POSIXlt(paste0(splitdate[3],"-",splitdate[2],"-",splitdate[1]," 21:00:00"),tz=tz)
  t1 = as.POSIXlt(as.numeric(t0) + 3600*9,origin="1970-01-01",tz=tz)
  hoursinbetween = as.numeric(format(seq(t0,t1,by=3600),"%H"))
  t1 = as.numeric(format(t1,"%H")) + 24
  t0 = as.numeric(format(t0,"%H"))
  dsthour = c()
  if (t1 - t0 < 9) { 
    # if time has moved backward (clock gives the impression that 8 hours 
    # elapsed, while physically 9 hours elased
    dst_night_or_not = -1
    dsthour = hoursinbetween[duplicated(hoursinbetween)] #the double hour
  } else if (t1 - t0 > 9) {
    # if time has moved forwared (clock gives the impression that 10 hours
    # elapsed, while physically 9 hours elapsed
    dst_night_or_not = 1
    expectedhours = c(21:23,0:7)
    dsthour = expectedhours[which(expectedhours %in% hoursinbetween == FALSE)[1]] #themissinghour
    
  } else if (t1 - t0 == 9) {
    dst_night_or_not = 0
  }
  return(invisible(list(
    dst_night_or_not = dst_night_or_not,
    dsthour = dsthour)))
}

library(GGIR)
context("is_this_a_dst_night")
test_that("is_this_a_dst_night", {
  dst_uk = is_this_a_dst_night(calendar_date = "29/03/2014",tz="Europe/London")
  expect_that(dst_uk$dst_night_or_not,equals(1))
  expect_that(dst_uk$dsthour,equals(1))
  dst_nl = is_this_a_dst_night(calendar_date = "25/10/2014",tz="Europe/Amsterdam")
  expect_that(dst_nl$dst_night_or_not,equals(-1))
  expect_that(dst_nl$dsthour,equals(2))
})

g.plot5 = function(metadatadir = c(), dofirstpage = FALSE, viewingwindow = 1,
                   f0 = c(), f1 = c(), overwrite = FALSE,
                   metric = "ENMO", desiredtz = "", threshold.lig = 30,
                   threshold.mod = 100, threshold.vig = 400,
                   visualreport_without_invalid = TRUE,
                   includedaycrit = 0.66, includenightcrit = 0.66,
                   verbose = TRUE) {
  if (file.exists(paste0(metadatadir, "/results/file summary reports"))) {
    fnames.fsr = sort(dir(paste0(metadatadir, "/results/file summary reports")))
    ffdone = fnames.fsr #ffdone is now a vector of filenames that have already been processed by g.part5
  } else {
    dir.create(file.path(paste0(metadatadir, "/results"), "file summary reports"))
    ffdone = c()
  }
  N_milestone_data_p3 = length(dir(paste0(metadatadir,"/meta/ms3.out")))
  if (f1 > N_milestone_data_p3) f1 = N_milestone_data_p3 # this is intentionally ms3 and not ms4, do not change!
  if (f1 == 0) f1 = N_milestone_data_p3
  # directories
  tail_expansion_log = NULL
  ms1dir = paste0(metadatadir, "/meta/basic")
  ms2dir = paste0(metadatadir,"/meta/ms2.out")
  ms3dir = paste0(metadatadir, "/meta/ms3.out")
  ms4dir = paste0(metadatadir, "/meta/ms4.out")
  results = paste0(metadatadir, "/results")
  # names of files in those directories
  fname_ms1 = dir(ms1dir)
  fname_ms3 = dir(ms3dir)
  fname_ms2 = dir(ms2dir)
  fname_ms4 = dir(ms4dir)
  removeRDa = function(x) as.character(unlist(strsplit(x,".RDa")))[1]
  fname_ms1_stripped = apply(as.matrix(as.character(fname_ms1)), MARGIN = c(1), FUN = removeRDa)
  removeEta = function(x) as.character(unlist(strsplit(x, "eta_")))[2]
  fnames_ms1_stripped = apply(as.matrix(as.character(fname_ms1_stripped)), MARGIN = c(1), FUN = removeEta) # part 2 milestone data files
  fnames_ms3_withoutRDa = apply(as.matrix(as.character(fname_ms3)), MARGIN = c(1), FUN = removeRDa) # part 3 milestone data files
  # create list of day names
  wdaynames = c("Sunday", "Monday", "Tuesday", "Wednesday",
                "Thursday", "Friday", "Saturday")
  checkfiles = dir(results)
  ## load summary spreadsheets for this study (no longer used as we now use milestone data)
  #if (!file.exists(paste0(results, "/part2_daysummary.csv"))) {
  # stop("Warning: File part2_daysummary.csv not generated yet")
  #}
  #P2daysummary = read.csv(paste0(results, "/part2_daysummary.csv"))
  
  M = c()
  
  if (f1 - f0 > 50 & verbose == TRUE) {
    cat(paste0("\nGGIR is now creating a visualreport (pdf) for each of the ", f1 - f0,
               " recordings as a final step in the GGIR pipeline. This can take a while as",
               " it is done file-by-file. If you do not want the visualreports then you",
               " can kill the process (Ctrl-C or ESC) as that will not affect the",
               " rest of the analyses and reports. To avoid this next time,",
               " set parameter 'old_visualreport' to FALSE to only derive the new",
               " reports. Alternatively, set 'visualreport' to FALSE which", 
               " suppresses both report types."))
  }
  # loop through files
  for (i in f0:f1) {
    if (length(ffdone) > 0) {
      if (length(which(ffdone == paste0("Report_", fnames_ms3_withoutRDa[i], ".pdf"))) > 0) {
        skip = 1 #skip this file because it was analysed before")
      } else {
        skip = 0 #do not skip this file
      }
    } else {
      skip = 0
    }
    if (overwrite == TRUE) skip = 0
    if (skip == 0) {
      sel = which(fnames_ms1_stripped == fnames_ms3_withoutRDa[i])
      ms2_file_index = which(fname_ms2 == fname_ms2[i])
      ms2_filepath = paste(ms2dir, "/", fname_ms2[ms2_file_index], sep = "")
      P2daysummary_tmp = IMP = SUM = NULL
      load(ms2_filepath)
      P2daysummary_tmp = SUM$daysummary
      P2daysummary_tmp$measurementday = as.numeric(P2daysummary_tmp$measurementday)
      if (length(sel) > 0) {
        ms4_file_index = which(fname_ms4 == fname_ms3[i])
        if (length(ms4_file_index) == 1) {
          nightsummary = c()
          ms4_filepath = paste0(ms4dir, "/", fname_ms4[ms4_file_index])
          load(ms4_filepath) #to load summary sleep
          nightsummary$night = as.numeric(nightsummary$night)
          summarysleep_tmp = nightsummary
        } else {
          warning(
            paste0("\nVisual report not generated for ",
                   fnames_ms1_stripped[sel], " because part 4 output was not available."
            )
          )
          next
        }
       
        sib.cla.sum = c()
        load(paste0(ms3dir,"/",fname_ms3[i]))
        load(paste0(ms1dir,"/",fname_ms1[sel]))
        ws3 = M$windowsizes[1]
        ws2 = M$windowsizes[2]
        if (nrow(sib.cla.sum) == 0) next
        pdf(paste0(metadatadir, "/results/file summary reports/old_report_",
                   fnames_ms1_stripped[sel], ".pdf"), paper = "a4",
            width = 0, height = 0)
        
        # P2daysummary_tmp = P2daysummary[which(P2daysummary$filename == fnames_ms1_stripped[sel]),]
        # note that the reports are generated from the raw sleep classification (part4) and no attempt is made to clean it up,
        # by deleting nights for which no diary was available or not enough accelerometer data was available
        if (length(unique(summarysleep_tmp$acc_def)) > 1) {
          if (length(which(unique(summarysleep_tmp$acc_def) == "T5A5")) == 1) {
            della = which(summarysleep_tmp$acc_def == "T5A5")
          } else {
            della = which(summarysleep_tmp$acc_def == unique(summarysleep_tmp$acc_def)[1])
          }
          if (length(della) > 0) summarysleep_tmp = summarysleep_tmp[-della,]
        }
        threshold_hrs_of_data_per_day = 0.5
        if (visualreport_without_invalid == TRUE) {
          incrementNight = 0
          # do not include days with no meaningful data
          if (includedaycrit < 1) includedaycrit = includedaycrit * 24
          if (includenightcrit > 1) includenightcrit = includenightcrit / 24
          d2excludeb = d2exclude = which(P2daysummary_tmp$`N valid hours` < max(c(includedaycrit,
                                                                                  threshold_hrs_of_data_per_day)))
          n2excludeb = n2exclude = which(summarysleep_tmp$fraction_night_invalid > (1 - includenightcrit)
                                         | summarysleep_tmp$SptDuration == 0)
          
          if (length(d2exclude) > 0) {
            d2excludeb = P2daysummary_tmp$measurementday[d2exclude]
            P2daysummary_tmp = P2daysummary_tmp[-d2exclude,] #ignore days with non-wear
            d2exclude = d2excludeb
          }
          if (length(n2exclude) > 0) {
            n2excludeb = summarysleep_tmp$night[n2exclude]
            summarysleep_tmp = summarysleep_tmp[-n2exclude,]
            n2exclude = n2excludeb
          }
          # nights missing in sleep summary?
          if (length(summarysleep_tmp$night) > 0) {
            allnights = 1:max(summarysleep_tmp$night)
            missingNights = which(allnights %in% summarysleep_tmp$night == FALSE)
          } else {
            missingNights = NULL
          }
          if (length(missingNights) > 0) {
            n2exclude = sort(unique(c(n2exclude, missingNights)))
          }
          # Account for shift in nights relative to windows
          if (nrow(P2daysummary_tmp) > 0) {
            # nights missing in sleep summary?
            allnights = 1:max(summarysleep_tmp$night)
            missingNights = which(allnights %in% summarysleep_tmp$night == FALSE)
            if (length(missingNights) > 0) {
              n2exclude = sort(unique(c(n2exclude, missingNights)))
            }
            if (P2daysummary_tmp$`N hours`[1] < 24 & P2daysummary_tmp$`N hours`[1] > 12 & length(n2exclude) > 0) {
              # First calendar day is between 12 and 24 hours
              # this means that the first night viewindow (=2) may include some data but
              # is not reflected by the sleep reportswindow
              # so our night counts should be increased by one
              n2excludeb = n2exclude = n2exclude + 1
              incrementNight = 1
            }
          }
          
          if (length(tail_expansion_log) != 0) { # then keep timing of sleeponset to plot
            lastnight = max(summarysleep_tmp$night) + incrementNight
            if (lastnight %in% n2exclude) {
              n2exclude = n2exclude[-which(n2exclude == lastnight)]
            }
          }
        } else {
          n2exclude = d2exclude = NULL
        }
        # detect which column is mvpa
        n45 = names(P2daysummary_tmp)
        varsVPA = grep(pattern = "VPA",x = n45)
        vars02 = grep(pattern = "_0-2|_0.2|_24",x = n45)
        c45 = varsVPA[which(varsVPA %in% vars02 == TRUE)]
        c45 = c45[length(c45)]
        #######################
        # First page of the report
        MainMetric_1 = paste0("mean_", metric, "_mg_24hr")
        MainMetric_2 = paste0("mean_", metric, "_mg_0-24hr")
        MainMetric_3 = paste0("mean_", metric, "_mg_0.24hr")
        MainMetric = NULL
        if (length(which(colnames(P2daysummary_tmp) == MainMetric_1)) == 1) {
          MainMetric = MainMetric_1
        } else if (length(which(colnames(P2daysummary_tmp) == MainMetric_2)) == 1) {
          MainMetric = MainMetric_2
        } else if (length(which(colnames(P2daysummary_tmp) == MainMetric_3)) == 1) {
          MainMetric = MainMetric_3
        }
        if (is.null(MainMetric) == TRUE) {
          dofirstpage = FALSE
        } else {
          if (length(which(is.na(P2daysummary_tmp[,MainMetric]) == FALSE)) <= 1) {
            dofirstpage = FALSE
          }
        }
        if (dofirstpage == TRUE & length(which(is.na(P2daysummary_tmp[,c45]) == FALSE)) > 1
            & nrow(summarysleep_tmp) > 0) {
          # abbreviate names of days
          days1 = P2daysummary_tmp$weekday
          daynames = as.character(days1)
          days_PA = g.abr.day.names(daynames)
          days2 = summarysleep_tmp$weekday
          daynames = as.character(days2)
          days_SLEEP = g.abr.day.names(daynames)
          # extract variables
          lengthnight = summarysleep_tmp$SptDuration #including wake periods
          nocsleepdur = summarysleep_tmp$SleepDurationInSpt
          sleepefficiency = (nocsleepdur / lengthnight) * 100
          f01 = P2daysummary_tmp[,c45]
          f02 = P2daysummary_tmp[,MainMetric]
          f05 = matrix(NA, nrow = 2, ncol = length(summarysleep_tmp$SptDuration))
          f05[1,] = summarysleep_tmp$SleepDurationInSpt
          f05[2,] = summarysleep_tmp$SptDuration - summarysleep_tmp$SleepDurationInSpt
          f05_2 = summarysleep_tmp$SptDuration
          f06 = sleepefficiency
          # f07 = P2daysummary_tmp$N.valid.hours # Previously needed when reading csv-report Part 2
          f07 = P2daysummary_tmp$`N valid hours`
          # allocate colours
          CLS = c("white", "black")
          CLS_A = rep(CLS[1], length(days_PA))
          CLS_B = rep(CLS[1], length(days_SLEEP))
          CLS_A[which(days_PA == "SUN" | days_PA == "SAT")] = CLS[2]
          CLS_B[which(days_SLEEP == "SUN" | days_SLEEP == "SAT")] = CLS[2]
          # headers
          vars = c(paste0("Time spent in moderate or vigorous activity (average is ",
                          round(mean(f01, na.rm = TRUE)), " minutes per day)"),
                   paste0("Total physical activity (average per day is ",
                          round(mean(f02, na.rm = TRUE)), " mg)"),
                   paste0("Sleep period time (average is ",
                          round(mean(f05_2, na.rm = TRUE), digits = 1), " hours per night)"),
                   paste0("Sleep efficiency (average is ", round(mean(f06, na.rm = TRUE)), "% per night)"),
                   paste0("Duration monitor worn (hours per day)"))
          # plot data
          CEXN = 0.9
          par(mfrow = c(5, 1),
              omi = c(0, 0, 0.2, 0),
              mar = c(3, 2, 2, 2) + 0.1)
          #MVPA
          YXLIM = c(0, (max(f01, na.rm = TRUE) * 1.3))
          if (YXLIM[2] == 0) YXLIM[2] = 60
          B3 = barplot(as.matrix(f01), names.arg = days_PA, beside = TRUE,
                       ylim = YXLIM, cex.names = CEXN, las = 0, col = CLS_A, density = 20)
          abline(h = 30, lty = 2, lwd = 2)
          topp = mean(as.matrix(round(f01))) * 0.1
          text(
            y = as.matrix(round(f01)) + topp + 5,
            x = B3,
            labels = as.character(as.matrix(round(f01))), xpd = TRUE, cex = 1
          )
          text(
            x = 1,
            y = (max(YXLIM) * 0.95),
            labels = vars[1], pos = 4, font = 2, cex = 1.2
          )
          #Acceleration
          YXLIM = c(0, (max(f02, na.rm = TRUE) * 1.3))
          B6 = barplot(as.matrix(f02), names.arg = days_PA, beside = TRUE,
                       ylim = YXLIM, cex.names = CEXN, las = 0,
                       col = CLS_A, density = 20)
          abline(h = 25, lty = 2, lwd = 2)
          topp = mean(as.matrix(round(f02))) * 0.1
          text(
            y = as.matrix(round(f02)) + topp,
            x = B6,
            labels = as.character(as.matrix(round(f02))), xpd = TRUE, cex = 1
          )
          text(
            x = 1,
            y = (max(YXLIM) * 0.95),
            labels = vars[2], pos = 4, font = 2, cex = 1.2
          )
          #Monitor wear duration
          YXLIM = c(0, (max(f07, na.rm = TRUE) * 1.3))
          B8 =  barplot(
            as.matrix(f07),
            names.arg = days_PA,
            beside = TRUE,
            ylim = YXLIM,
            cex.names = CEXN,
            las = 0,
            col = CLS_A,
            density = 20
          )
          topp = mean(as.matrix(round(f07))) * 0.1
          text(
            y = as.matrix(round(f07)) + topp,
            x = B8,
            labels = as.character(as.matrix(round(f07))),
            xpd = TRUE, cex = 1.1
          )
          abline(h = 16, lty = 2, lwd = 2)
          text(
            x = 1,
            y = (max(YXLIM) * 0.95),
            labels = vars[5], pos = 4, font = 2, cex = 1.2
          )
          #Sleep duration
          night_sleep_col <- rgb(0.8, 0.8, 0.8, alpha = 0.3)
          night_wake_col <- rgb(0, 0.8, 0.6, alpha = 0.2)
          space_vec = rep(0, length(f05_2))
          YXLIM = c(0, (max(f05_2, na.rm = TRUE) * 1.3))
          B4 = barplot(as.matrix(f05), names.arg = days_SLEEP,
                       ylim = YXLIM, las = 0,
                       col = c(night_sleep_col, night_wake_col),
                       space = space_vec)
          legend(
            "bottomleft",
            c('Nocturnal Wake', 'Sleep'),
            fill = c(night_wake_col, night_sleep_col),
            bg = 'white'
          )
          abline(h = 6, lty = 2, lwd = 2)
          topp = mean(as.matrix(round(f05_2, digits = 1))) * 0.1
          text(
            y = as.matrix(round(f05_2, digits = 1)) + topp,
            x = B4,
            labels = as.character(as.matrix(round(f05_2, digits = 1))),
            xpd = TRUE, cex = 1
          )
          # not sure why I had to change x to 0 (from 1) in the following line:
          text(
            x = 0,
            y = (max(YXLIM) * 0.95),
            labels = vars[3],
            pos = 4, font = 2, cex = 1.2
          )
          #Sleep efficiency
          YXLIM = c(0, 120)
          B5 = barplot(as.matrix(f06), names.arg = days_SLEEP, beside = TRUE,
                       ylim = YXLIM, cex.names = CEXN, las = 0, col = CLS_B, density = 20)
          abline(h = 60, lty = 2, lwd = 2)
          abline(h = 100,lty = 3, lwd = 1)
          topp = mean(as.matrix(round(f06))) * 0.1
          text(
            y = as.matrix(round(f06)) + topp,
            x = B5,
            labels = as.character(as.matrix(round(f06))), xpd = TRUE, cex = 1
          )
          text(
            x = 1,
            y = (max(YXLIM) - 10),
            labels = vars[4], pos = 4, font = 2, cex = 1.2
          )
          #-----------------------------------------------------------------------------------
          mtext(
            paste0("Activity and sleep report: ", fnames_ms1_stripped[sel]),
            side = 3, line = 0.5, outer = TRUE, font = 2, cex = 0.7
          )
          mtext(text = paste0("Warning: This GGIR legacy report is not intended to be used for data quality assessment. ",
                              " To quality check your data visually, see file names starting with report_ "),
                side = 3, line = -0.5, outer = TRUE, font = 2, cex = 0.6, col = "red"
          )
        }
        LWDX = 2.5 #linewidth for coloured lines
        LWDA = 0.2
        BLX = 0.6
        #=================================================
        # Next pages with day specific graphs
        # get variables - activity:
        ACC = as.numeric(as.matrix(M$metashort[, metric])) * 1000
        nonwearscore = as.numeric(as.matrix(M$metalong[,"nonwearscore"]))
        time =  as.character(M$metashort[,1])
        nw_time = as.character(M$metalong[,1])
        if (length(unlist(strsplit(time[1],"T"))) > 1) { # ISO timestamp format
          time = format(iso8601chartime2POSIX(time, desiredtz))
          nw_time = format(iso8601chartime2POSIX(nw_time, desiredtz))
        }
        time_unclassed = unclass(as.POSIXlt(time,desiredtz))
        sec = time_unclassed$sec
        min_vec = time_unclassed$min
        hour = time_unclassed$hour
        NONWEAR = IMP$r5long
        INACT = LIGPA = MODPA = VIGPA = rep(NA, length(ACC))  # PA vectors for plotting
        
        # Find bouts of light-PA (LPA):
        boutdur2 = 10 * (60/ws3)    # 10min bout duration
        boutcriter = 0.8            # 80% bout criteria
        rr_lpa = matrix(0,length(ACC),1)
        p_lpa = which(ACC >= threshold.lig)
        rr_lpa[p_lpa] = 1
        rr1t = rr_lpa
        jlpa = 1
        while (jlpa <= length(p_lpa)) {
          endi = p_lpa[jlpa] + boutdur2
          if (endi <= length(rr_lpa)) { #does bout fall without measurement?
            lengthbout = sum(rr_lpa[p_lpa[jlpa]:endi])
            if (lengthbout > (boutdur2*boutcriter)) { #0.8 => 80% of the bout needs to meet the criteria
              rr1t[p_lpa[jlpa]:endi] = 2 #remember that this was a bout in r1t
            } else {
              rr_lpa[p_lpa[jlpa]] = 0
            }
          } else { #bout does not fall within measurement
            if (length(p_lpa) > 1 & jlpa > 2) {
              rr_lpa[p_lpa[jlpa]] = rr_lpa[p_lpa[jlpa - 1]]
            }
          }
          jlpa = jlpa + 1
        }
        rr_lpa[which(rr1t == 2)] = 1
        LIGPA[which(rr_lpa == 1)] = 1 #moderate as part of 10 minute bouts of lpa
        # Find bouts of MVPA
        rr = matrix(0,length(ACC),1)
        p = which(ACC >= threshold.mod)
        rr[p] = 1
        rr1t = rr
        jmvpa = 1
        while (jmvpa <= length(p)) {
          endi = p[jmvpa] + boutdur2
          if (endi <= length(rr)) { #does bout fall without measurement?
            lengthbout = sum(rr[p[jmvpa]:endi])
            if (lengthbout > (boutdur2 * boutcriter)) { #0.9 => 90% of the bout needs to meet the criteria
              rr1t[p[jmvpa]:endi] = 2 # remember that this was a bout in r1t
            } else {
              rr[p[jmvpa]] = 0
            }
          } else { # bout does not fall within measurement
            if (length(p) > 1 & jmvpa > 2) {
              rr[p[jmvpa]] = rr[p[jmvpa - 1]]
            }
          }
          jmvpa = jmvpa + 1
        }
        rr[which(rr1t == 2)] = 1
        # moderate as part of 10 minute bouts of mvpa
        MODPA[which(rr == 1 & ACC < threshold.vig)] = 1
        # vigorous as part of 10 minute bouts of mvpa
        VIGPA[which(ACC >= threshold.vig & rr == 1)] = 1
        # cancel LPA sections that are MVPA
        LIGPA[which(MODPA == 1 | VIGPA == 1)] <- NA
        # create inactive sections that are NA on LPA/MPA/VPA
        INACT[which(is.na(LIGPA) & is.na(MODPA) & is.na(VIGPA))] <- 1
        #get variables - sleep:
        angle = as.matrix(M$metashort[,which(colnames(M$metashort) == "anglez")])
        detection = rep(NA,length(time))
        S = sib.cla.sum
        S$sib.onset.time = iso8601chartime2POSIX(S$sib.onset.time, tz = desiredtz)
        S$sib.end.time = iso8601chartime2POSIX(S$sib.end.time, tz = desiredtz)
        def = unique(S$definition)[1]
        S = S[which(S$definition == def),] # simplify to one definition
        for (j in 1:max(unique(S$night))) {
          tmp = S[which(S$night == j),]
          for (h in 1:nrow(tmp)) { # sleep periods
            s0 = which(time == format(tmp$sib.onset.time[h]))[1]
            s1 = which(time == format(tmp$sib.end.time[h]))[1]
            if (length(s0) == 0 | is.na(s0) == TRUE) {
              s0 = which(time == paste0(format(tmp$sib.onset.time[h])," 00:00:00"))[1]
            }
            if (length(s1) == 0 | is.na(s1) == TRUE) {
              s1 = which(time == paste0(format(tmp$sib.end.time[h]), " 00:00:00"))[1]
            }
            if (is.na(s0) == FALSE & is.na(s1) == FALSE) {
              detection[s0:s1] = 1
            }
          }
        }
        night_wake_full <- rep(NA,length(detection)) # opposite of detection (sleep/inactivity detection)
        night_wake_full[is.na(detection)] <- 1 # for plotting wake during SPT window
        # prepare to search for sleeplog_onst / sleeplog_wake
        sleep_dates = as.Date(summarysleep_tmp$calendar_date,
                              format = '%d/%m/%Y',
                              origin = "1970-1-1")
        # detect midnights
        if (viewingwindow == 1) {
          nightsi = which(sec == 0 & min_vec == 0 & hour == 0)
          xaxislabels = c("midnight","2am", "4am", "6am", "8am", "10am", "noon",
                          "2pm","4pm","6pm","8pm","10pm","midnight")
        } else if (viewingwindow == 2) {
          nightsi = which(sec == 0 & min_vec == 0 & hour == 12)
          xaxislabels = c("noon","2pm", "4pm", "6pm", "8pm", "10pm", "midnight",
                          "2am", "4am", "6am", "8am", "10am", "noon")
        }
        if (length(nightsi) > 0 && nrow(summarysleep_tmp) > 0 && nrow(P2daysummary_tmp)) {
          # Do not attempt to create a plot when there is no midnight in the data,
          # because calculation of t1 will be complicated.
          nplots = length(nightsi) + 1
          if (visualreport_without_invalid == TRUE) {
            if (viewingwindow == 2) {
              nplots = min(c(nplots, max(summarysleep_tmp$night) + incrementNight))
            } else {
              nplots = min(c(nplots, max(P2daysummary_tmp$measurementday)))
            }
          }
          # plot
          npointsperday = (60/ws3)*1440
          abcis = 1:npointsperday
          abcisL = length(abcis)
          NGPP = 7 #number of graphs per page
          par(
            mfcol = c(NGPP, 1),
            mar = c(2, 0.5, 2, 0.5) + 0.1,
            omi = c(0, 0, 0.3, 0),
            mgp = c(2, 0.8, 0)
          )
          daycount = 1
          for (g in 1:nplots) {
            skip = FALSE
            if (g == 1) {
              t0 = 1
              t1 = nightsi[g] - 1
              if ((t1 - t0) < (threshold_hrs_of_data_per_day * (60 / ws3) * 60)) {
                skip = TRUE # regardless of inclusion criterion only plot windows with at least half an hour of data
              }
            } else if (g > 1 & g < nplots) {
              t0 = nightsi[g - 1]
              t1 = nightsi[g] - 1
            }  else if (g == nplots) {
              t0 = nightsi[g - 1]
              t1 = min(c(length(time), t0 + (24 * 60 * (60 / ws3)) - 1))
              if ((t1 - t0) < (threshold_hrs_of_data_per_day * (60 / ws3) * 60)) {
                skip = TRUE
              }
            }
            NhoursInThisDay = ((t1 - t0) + 1) / (60 * 60 / ws3)
            if (NhoursInThisDay == 25) { # day with 25 hours, pretend that 25th hour did not happen
              t1 = t1 - (3600/ws3)
            } else if (NhoursInThisDay == 23) { # day with 23 hours, extend timeline with 1 hour
              t1 = t1 + (3600/ws3)
            }
            # Initialize daily 'what we think you did' vectors:
            if (length(tail_expansion_log) != 0) {
              from = length(ACC) - tail_expansion_log$short + 1
              to = length(ACC)
              ACC[from:to] = NA; angle[from:to] = NA
            }
            acc = abs(ACC[t0:t1])
            ang = angle[t0:t1]
            # if only expanded data in this day, then no plot
            if (all(is.na(acc)) & all(is.na(ang))) next
            non_wear <- NONWEAR[t0:t1]
            annot_mat = matrix(NA, nrow = length(acc), ncol = 6)
            annot_mat[,1] <- detection[t0:t1]          # night sleep
            annot_mat[,2] <- night_wake_full[t0:t1]    # nocturnal wake
            annot_mat[,3] <- INACT[t0:t1]              # inactivity
            annot_mat[,4] <- LIGPA[t0:t1]              # light pa
            annot_mat[,5] <- MODPA[t0:t1]              # moderate pa
            annot_mat[,6] <- VIGPA[t0:t1]              # vigorous pa
            # check to see if there are any sleep onset or wake annotations on this day
            sleeponset_loc = 0
            wake_loc = 0
            if (viewingwindow == 1) {  # use different search coefficients for noon or midnight centered plots
              sw_coefs = c(0,24)
            } else if (viewingwindow == 2) {
              sw_coefs = c(12,36)
            }
            # check for sleeponset & wake time that is logged on this day before midnight
            curr_date = as.Date(substr(time[t0], start = 1, stop = 10),
                                format = '%Y-%m-%d',
                                origin = "1970-1-1")  # what day is it?
            if (viewingwindow == 2) {
              # check to see if it is the first day that has less than 24 and starts after midnight
              if ((t1 - t0) < ((60*60*12)/ws3)) { # if there is less than half a days worth of data
                curr_date = curr_date - 1
              }
            }
            check_date = match(curr_date, sleep_dates)
            if (is.na(check_date) == FALSE) {
              sleeponset_time = summarysleep_tmp$sleeponset[check_date]  # get the time of sleep_onset
              if (sleeponset_time >= sw_coefs[1] & sleeponset_time < sw_coefs[2]) {
                sleeponset_hour = trunc(sleeponset_time)
                if (sleeponset_hour == 24) sleeponset_hour = 0
                if (sleeponset_hour > 24) sleeponset_hour = sleeponset_hour - 24 # only with viewingwindow==2
                sleeponset_min = round((sleeponset_time - trunc(sleeponset_time)) * 60)
                if (sleeponset_min == 60) sleeponset_min = 0
                sleeponset_locations = which(hour[t0:t1] == sleeponset_hour & min_vec[t0:t1] == sleeponset_min)
                if (!is.na(sleeponset_locations[1])) {
                  sleeponset_loc = sleeponset_locations[1]
                }
              }
              wake_time = summarysleep_tmp$wakeup[check_date]
              if (wake_time >= sw_coefs[1] & wake_time < sw_coefs[2]) {
                wake_hour = trunc(wake_time)
                if (wake_hour == 24) wake_hour = 0
                if (wake_hour > 24) {
                  wake_hour = wake_hour - 24
                }
                wake_min = round((wake_time - trunc(wake_time)) * 60)
                if (wake_min == 60) wake_min = 0
                wake_locations = which(hour[t0:t1] == wake_hour & min_vec[t0:t1] == wake_min)
                if (!is.na(wake_locations[1])) {
                  wake_loc = wake_locations[1]
                }
              }
            }
            # check for sleeponset & wake time that is logged on the previous day after midnight
            prev_date = curr_date - 1
            check_date = match(prev_date,sleep_dates)
            if (is.na(check_date) == FALSE) {
              wake_time = summarysleep_tmp$wakeup[check_date]
              if (wake_time >= sw_coefs[2]) {
                wake_hour = trunc(wake_time) - 24
                wake_min = round((wake_time - trunc(wake_time)) * 60)
                if (wake_min == 60) wake_min = 0
                wake_locations = which(hour[t0:t1] == wake_hour & min_vec[t0:t1] == wake_min)
                if (wake_loc > 0) {
                  if (!is.na(wake_locations[1])) {
                    wake_loc[2] = wake_locations[1]
                  }
                } else if (wake_loc == 0) {
                  if (!is.na(wake_locations[1])) {
                    wake_loc = wake_locations[1]
                  }
                }
              }
              # new way
              sleeponset_time = summarysleep_tmp$sleeponset[check_date]
              if (sleeponset_time >= sw_coefs[2]) {
                sleeponset_hour = trunc(sleeponset_time) - 24
                sleeponset_min = round((sleeponset_time - trunc(sleeponset_time)) * 60)
                if (sleeponset_min == 60) sleeponset_min = 0
                sleeponset_locations = which(hour[t0:t1] == sleeponset_hour & min_vec[t0:t1] == sleeponset_min)
                if (sleeponset_loc > 0) {
                  if (!is.na(sleeponset_locations[1])) {
                    sleeponset_loc[2] = sleeponset_locations[1]
                  }
                } else if (sleeponset_loc == 0) {
                  if (!is.na(sleeponset_locations[1])) {
                    sleeponset_loc = sleeponset_locations[1]
                  }
                }
              }
            }
            # if midnight centered plots, then search next day as well
            if (viewingwindow == 2) {
              next_day = curr_date + 1
              check_date = match(next_day,sleep_dates)
            }
            # add extensions if <24hr of data
            first_day_adjust = 0 # hold adjustment amounts on first and last day plots
            last_day_adjust = 0
            if (((t1 - t0) + 1) != npointsperday & t0 == 1) {
              daySize = (t1 - t0) - (floor(((t1 - t0) / npointsperday)) * npointsperday)
              extension = rep(NA, (npointsperday - (daySize + 1)))
              first_day_adjust = length(extension)
              acc = c(extension,acc)
              ang = c(extension,ang)
              non_wear = c(extension, non_wear)
              t1 = length(acc) # this was missing and causing x.y coords errors in some tests
              if (length(acc) == (abcisL + 1)) {
                extension = extension[2:(length(extension))]
                acc = acc[2:(length(acc))]
                ang = ang[2:(length(ang))]
                non_wear = non_wear[2:(length(non_wear))]
              }
              extension_mat = matrix(NA, nrow = length(extension), ncol = 6)
              annot_mat = rbind(extension_mat,annot_mat)
              # adjust any sleeponset / wake annotations if they exist:
              if (sleeponset_loc[1] != 0) {
                for (i2 in 1:length(sleeponset_loc)) {
                  sleeponset_loc[i2] = sleeponset_loc[i2] + length(extension)
                }
              }
              if (wake_loc[1] != 0) {
                for (i3 in 1:length(wake_loc)) {
                  wake_loc[i3] = wake_loc[i3] + length(extension)
                }
              }
            } else if (((t1 - t0) + 1) != npointsperday & t1 == length(time)) {
              daySize = (t1 - t0) - (floor(((t1 - t0) / npointsperday)) * npointsperday)
              extension = rep(NA, (npointsperday - (daySize + 1)))
              last_day_adjust = length(acc)
              acc = c(acc,extension)
              ang = c(ang,extension)
              non_wear = c(non_wear,extension)
              if (length(acc) == (abcisL + 1)) {
                extension = extension[1:(length(extension) - 1)]
                acc = acc[1:(length(acc) - 1)]
                ang = ang[1:(length(ang) - 1)]
                non_wear = non_wear[1:(length(non_wear) - 1)]
              }
              extension_mat = matrix(NA, nrow = length(extension), ncol = 6)
              annot_mat = rbind(annot_mat,extension_mat)
            }
            acc = as.numeric(acc)
            acc[which(acc >= 900)] = 900
            acc = (acc/9) - 210
            annot_mat[non_wear == 1, ] = 0   # no annotations in non-wear periods:
            sleep_mat = annot_mat[, 1:2]  # split annotation matrix
            wake_mat = annot_mat[, 3:6]
            # how many sleeponset and wake annotations are there?
            if (sleeponset_loc[1] != 0 | wake_loc[1] != 0) {
              # combine sleep and wake annotations and sort them
              if (wake_loc[1] == 0) {
                # assume there is only one sleeponset loc
                sleeponset_mat = matrix(nrow = length(sleeponset_loc), ncol = 2)
                sleeponset_mat[,1] = sleeponset_loc
                sleeponset_mat[,2] = 0 # sleeponset = 0
                sleep_wake_mat = sleeponset_mat
              } else if (sleeponset_loc[1] == 0) {
                # assume there is only one wake loc
                wake_time_mat = matrix(nrow = length(wake_loc), ncol = 2)
                wake_time_mat[,1] = wake_loc
                wake_time_mat[,2] = 1 # wake = 1
                sleep_wake_mat = wake_time_mat
              } else {
                # there is both sleeponst and wake annotations
                sleeponset_mat = matrix(nrow = length(sleeponset_loc), ncol = 2)
                sleeponset_mat[,1] = sleeponset_loc
                sleeponset_mat[,2] = 0 # sleeponset = 0
                wake_time_mat = matrix(nrow = length(wake_loc), ncol = 2)
                wake_time_mat[,1] = wake_loc
                wake_time_mat[,2] = 1 # wake = 1
                sleep_wake_mat = rbind(sleeponset_mat,wake_time_mat)
                sleep_wake_mat = sleep_wake_mat[order(sleep_wake_mat[,1]),] # sort rows
              }
              # create an annotation on the final data-point for the day that is opposite to the previous point
              if (last_day_adjust != 0) {  # adjust end location if it is the final day of plotting and there is less than 24h
                end_value = last_day_adjust
              } else {
                end_value = abcisL
              }
              if (sleep_wake_mat[length(sleep_wake_mat)] == 1) {    # add final location at end of 24h period
                sleep_wake_mat = rbind(sleep_wake_mat,c(end_value,0))
              } else {
                sleep_wake_mat = rbind(sleep_wake_mat,c(end_value,1))
              }
              # loop through each annotation and update data frames accordingly
              if (first_day_adjust != 0) {
                prev_loc = first_day_adjust  # start at adjusted start point on day one
                if (first_day_adjust > sleep_wake_mat[1,1]) warning('ERROR: sleep/wake annotation is before the recording begins')
              } else {
                prev_loc = 1  # otherwise, start from first sample
              }
              for (i4 in 1:length(sleep_wake_mat[, 1])) {
                annot_type = sleep_wake_mat[i4, 2] # get if it is wake (1) or sleep (0)
                curr_loc = sleep_wake_mat[i4, 1]
                if (annot_type == 0) {
                  # sleeponset annotation, previous data is wake  # turn off sleep_mat (NA)
                  sleep_mat[prev_loc:curr_loc,] <- 0
                } else {
                  # wake annotation, previous data is sleep    # turn off wake_mat (NA)
                  wake_mat[prev_loc:curr_loc,] <- 0
                }
                prev_loc = curr_loc
              }
            } else {
              # there are no sleeponset or wake annotations on this day.
              # plot both sleep and wake annotations for now ?
              # ? other suggestions welcome ?
            }
            if (viewingwindow == 1) { #focus on day
              if (length(d2exclude) > 0) {
                if (length(which(d2exclude == g)) > 0) skip = TRUE
              }
            } else { #focus on night
              if (length(n2exclude) > 0) {
                if (length(which(n2exclude == g)) > 0) skip = TRUE
              }
            }
            #VvH I have moved unclass to one location, to avoid doing this computation several times:
            curr_date_unclassed = unclass(as.POSIXlt(curr_date,desiredtz))
            title = paste0("Day ",daycount,": ",
                           wdaynames[curr_date_unclassed$wday + 1],
                           " | ",
                           curr_date_unclassed$mday, " ",
                           month.abb[curr_date_unclassed$mon + 1], " ",
                           curr_date_unclassed$year + 1900)
            rm(curr_date_unclassed)
            if (skip == FALSE) {
              YXLIM = c(-230,300)
              LJ = 2
              # plot accelerometer data:
              if (abcisL == length(acc) & abcisL == length(ang)) {  # check to prevent x.y coords errors, show more useful error msg
                plot(abcis, acc, type = "l", lwd = LWDA,
                     bty = "l", axes = FALSE, ylim = YXLIM,
                     xlab = "", ylab = "", main = "",
                     cex.main = 0.9, lend = LJ) #,axes=FALSE,ylim=YXLIM,xlab="",ylab="",main="",cex=0.3
                # plot z-angle:
                lines(abcis,ang, type = "l", lwd = LWDA, bty = "l",
                      xlab = "", ylab = "", cex = 0.3, lend = LJ)
              } else {
                break()
                # Following warning turned off because we expect condition not
                # to be met when recording ends
                # with many non-wear days
                # warning('\nplot5 error: index, acc and ang vectors are different lengths')
              }
              # add sleeponset time annotation to plot:
              arrow_line_length = abcisL * 0.01736 # make arrow line length adaptable to differnt short epochs
              if (sleeponset_loc[1] != 0) {
                for (i5 in sleeponset_loc) { # allow for multiple sleeponset_loc
                  set_pos = 4 # position of text in relation to arrow
                  ar_start_idx <- i5
                  ar_end_idx <- i5 + arrow_line_length # make arrow go to the right of the annotation line
                  if (i > (0.8 * abcisL)) {  # check to see if text should be placed on the left side of the line
                    set_pos = 2
                    ar_end_idx <- i5 - arrow_line_length
                  }
                  # draw sleeponset annotation:
                  segments(i5, -230, i5, 210, col = 'black', lwd = 1.5)
                  arrows(ar_start_idx, 205, ar_end_idx, 205, length = 0.05,
                         angle = 20, code = 1, lwd = 0.5)
                  segments(ar_start_idx, 205, ar_end_idx, 205, col = "black", lwd = 0.5)
                  text(ar_end_idx, 205, labels = "Sleep-onset", pos = set_pos,
                       font = 1.8, cex = 0.8, col = "darkgrey")
                }
              }
              # add wake time annotation to plot:
              if (wake_loc[1] != 0) {
                for (i6 in wake_loc) {
                  set_pos <- 4
                  ar_start_idx <- i6
                  ar_end_idx <- i6 + arrow_line_length
                  if (i6 > (0.8 * abcisL)) {
                    # check to see if text should be placed on the left side of the line
                    set_pos = 2
                    ar_end_idx <- i6 - arrow_line_length
                  }
                  # draw wake annotation:
                  segments(i6,-230, i6, 210, col = 'black', lwd = 1.5)
                  arrows(ar_start_idx, 160, ar_end_idx, 160,
                         length = 0.05, angle = 20, code = 1, lwd = 0.5)
                  segments(ar_start_idx, 160, ar_end_idx, 160, col = "black", lwd = 0.5)
                  text(ar_end_idx, 160, labels = "Wake", pos = set_pos,
                       font = 1.8, cex = 0.8, col = "darkgrey")
                }
              }
              # rect for all annotations:
              # make everything 0 in case there are any NA's left:
              sleep_mat[is.na(sleep_mat)] <- 0
              wake_mat[is.na(wake_mat)] <- 0
              non_wear[is.na(non_wear)] <- 0
              # colours for annotations:
              nonwear_col <- rgb(0, 0.8, 0, alpha = 0.4)
              inactive_col <- rgb(0.3, 0.3, 1, alpha = 0.2)
              light_pa_col <- rgb(1, 1, 0, alpha = 0.2)
              mod_pa_col <- rgb(1, 0.5, 0, alpha = 0.2)
              vig_pa_col <- rgb(1, 0, 0, alpha = 0.2)
              night_sleep_col <- rgb(0.8, 0.8, 0.8, alpha = 0.3)
              night_wake_col <- rgb(0, 0.8, 0.6, alpha = 0.2)
              spt_heights <- c(-80, 110)
              pa_heights <- c(-220, -100)
              plot_rects <- function(vec, col_select, rect_heights) {
                # use the rect function to highlight sleep and activity annotations on the plot
                if (sum(vec) > 0) {
                  r_start_idx <- which(diff(c(0L, vec)) == 1L)
                  if (length(r_start_idx) != 0) {
                    r_end_idx <- which(diff(c(vec, 0L)) == -1L)
                    if (length(r_end_idx) + 1 == length(r_start_idx)) {
                      r_end_idx[length(r_start_idx)] <- length(vec)
                    }
                    for (idx in 1:length(r_start_idx)) {
                      rect(
                        r_start_idx[idx],
                        rect_heights[1],
                        r_end_idx[idx],
                        rect_heights[2],
                        col = col_select,
                        border = NA
                      )
                    }
                  }
                }
              }
              # daytime inactivity annotations
              plot_rects(vec = wake_mat[, 1],
                         col_select = inactive_col,
                         rect_heights = pa_heights)
              # light pa annotatoins
              plot_rects(vec = wake_mat[, 2],
                         col_select = light_pa_col,
                         rect_heights = pa_heights)
              # mod pa annotations
              plot_rects(vec = wake_mat[, 3],
                         col_select = mod_pa_col,
                         rect_heights = pa_heights)
              # vig pa annotations
              plot_rects(vec = wake_mat[, 4],
                         col_select = vig_pa_col,
                         rect_heights = pa_heights)
              # night_sleep annotations on graph
              plot_rects(vec = sleep_mat[, 1],
                         col_select = night_sleep_col,
                         rect_heights = spt_heights)
              # night_wake annotations on graph
              plot_rects(vec = sleep_mat[, 2],
                         col_select = night_wake_col,
                         rect_heights = spt_heights)
              # non_wear annotations on graph
              plot_rects(
                vec = non_wear,
                col_select = nonwear_col,
                rect_heights = c(-240, 130)
              )
              axis(side = 1, at = seq(1, (((60 / ws3) * 60 * 24) + 1),
                                      by = (2 * (60 / ws3) * 60)),
                   labels = xaxislabels, cex.axis = 0.7)
              abline(h = 0, untf = FALSE, lty = 3, lwd = 1, col = "grey")
              # x-axis coordinate for plotting text on the plot adaptable
              # to different short epoch lengths
              plot_loc = -abcisL * 0.05
              text(x = plot_loc, y = 285, labels = title, pos = 4, font = 2, cex = 1)
              text(x = plot_loc, y = -120, labels = "Arm movement:",
                   pos = 4, font = 1.8, cex = 0.9)
              text(x = plot_loc, y = 100,
                   labels = "Angle of sensor's z-axis relative to horizontal plane:",
                   pos = 4, font = 1.8, cex = 0.9)
              box("figure", col = "black")
              legend("topright", legend = c("SPT Window: Sleep",
                                            "SPT Window: Wake",
                                            "Inactivity", "Light PA",
                                            "Moderate PA", "Vigorous PA",
                                            "Non-Wear"),
                     lty = c(1, 1), col = c(
                       night_sleep_col, night_wake_col,
                       inactive_col, light_pa_col,
                       mod_pa_col, vig_pa_col,
                       nonwear_col
                     ),
                     lwd = c(LWDX, LWDX, LWDX), bg = "white", cex = 0.6, ncol = 7, box.lwd = BLX)
              if (daycount == 1 |
                  ((daycount - 1) / NGPP) == (round((daycount - 1) / NGPP))) {
                mtext(paste0("Filename: ", fnames_ms1_stripped[sel]),
                      side = 3, line = 1, outer = TRUE, font = 2, cex = 0.7)
                mtext(text = paste0("Warning: This GGIR legacy report is not intended to be used for data quality assessment. ",
                                    " To quality check your data visually, see file names starting with report_ "),
                      side = 3, line = 0, outer = TRUE, font = 2, cex = 0.6, col = "red"
                )
              }
            }
            daycount = daycount + 1
          }
          dev.off()
        }
      }
    }
  }
}

g.part5.addfirstwake = function(ts, summarysleep, nightsi, sleeplog, ID, 
                               Nepochsinhour, SPTE_end) {
  # Note related to if first and last night were ignored in part 4:
  # - diur lacks the first and last night at this point in the code.
  # - nightsi has all the midnights, so it is possible to check here
  # whether a wakeup time is missing on the first full day.
  # - if it is missing, then we will impute it in order for part5 to
  # the wake-to-wake analys on the second recording day.
  # Previously we accounted only for this later on in the code, which
  # did not benefit the exported timeseries.
  clock2numtime = function(x) { # function used for converting sleeplog times to hour times
    x2 = as.numeric(unlist(strsplit(x,":"))) / c(1,60,3600)
    return(sum(x2))
  }
  firstwake = which(diff(ts$diur) == -1)[1]
  firstonset = which(diff(ts$diur) == 1)[1]
  Nts = nrow(ts)
  epochSize = round(3600 / Nepochsinhour)
  if (is.na(SPTE_end[1]) == TRUE) {
    SPTE_end = SPTE_end[which(is.na(SPTE_end) == FALSE)]
  }
  # test whether wake for second day is missing
  # if the full sleep period happens before midnights
  if (length(nightsi) < 2) {
    return(ts)
  }
  guider = "unknown" # initialise guider name as unknown
  if (!is.na(firstwake) && firstwake > nightsi[2] ||
      (summarysleep$sleeponset[1] < 18 &&
       summarysleep$wakeup[1] < 18 &&
       firstwake < nightsi[2])) {
    wake_night1_index = c()
    if (length(sleeplog) > 0) {
      # use sleeplog for waking up after first night
      wake_night1 = sleeplog[which(sleeplog$ID == ID & sleeplog$night == 1),
                                       grep(pattern = "bedend|wake", x = colnames(sleeplog))]
      onset_night1 = sleeplog[which(sleeplog$ID == ID & sleeplog$night == 1),
                                         grep(pattern = "bedstart|sleeponset", x = colnames(sleeplog))]
      if (length(wake_night1) != 0 & length(onset_night1) != 0) {
        if (wake_night1 != "" & onset_night1 != "") {
          # express hour relative to midnight within the noon-noon:
          wake_night1_hour = clock2numtime(wake_night1)
          onset_night1_hour = clock2numtime(onset_night1)
          # If wake is in the afternoon or evening and onset hour is at noon
          # then this cannot be a daysleeper and wakie time need to be correct
          if (wake_night1_hour > 12 & (onset_night1_hour >= 12 & onset_night1_hour < 18)) {
            wake_night1_hour = wake_night1_hour - 24 # express hour relative to midnight
          }
          # onset_night1_hour = clock2numtime(onset_night1)
          wake_night1_index = nightsi[1] + round(wake_night1_hour * Nepochsinhour)
          # express hour relative to midnight within the noon-noon:
          if (wake_night1_index > Nts) wake_night1_index = Nts
          if (wake_night1_hour > 18) wake_night1_hour = wake_night1_hour - 24 # express hour relative to midnight
          if (wake_night1_index < 1) wake_night1_index = 1
          wake_night1_index = nightsi[1] + round(wake_night1_hour * Nepochsinhour)
          if (wake_night1_index > Nts) wake_night1_index = Nts
          if (wake_night1_index < 1) wake_night1_index = 1
          guider = "sleeplog"
        } else { # use SPTE algorithm as plan B
          wake_night1_index = nightsi[1] + round((SPTE_end[1] - 24) * Nepochsinhour)
          guider = "part3_estimate"
        }
      } else { # use SPTE algorithm as plan B
        wake_night1_index = nightsi[1] + round((SPTE_end[1] - 24) * Nepochsinhour)
        guider = "part3_estimate"
      }
    } else if (length(SPTE_end) > 0 & length(sleeplog) == 0) {
      # use SPTE algortihm for waking up after first night
      # if there was no sleep log
      if (is.na(SPTE_end[1]) == FALSE) {
        if (SPTE_end[1] != 0) {
          wake_night1_index = nightsi[1] + round((SPTE_end[1] - 24) * Nepochsinhour) 
          guider = "part3_estimate"
        }
      }
    }
    if (length(wake_night1_index) == 0) {
      # use waking up from next day and subtract 24 hours,
      # the final option if neither of the above routes works
      wake_night1_index = (firstwake - (24 * ((60 / epochSize) * 60))) + 1
    }
    if (is.na(wake_night1_index)) wake_night1_index = 0
    if (wake_night1_index < firstwake & wake_night1_index > 1 &
        (wake_night1_index - 1) > nightsi[1]) {
      newWakeIndex = c()
      firstSIBs = which(ts$sibdetection[1:(wake_night1_index - 1)] == 1)
      if (length(firstSIBs) > 0) newWakeIndex = max(firstSIBs)
      if (length(newWakeIndex) == 0) {
        newWakeIndex = wake_night1_index - 1
      }
      ts$diur[1:newWakeIndex] = 1
      ts$guider[1:newWakeIndex] = guider
    } else {
      # Person slept only during the afternoon on day 2
      # And there is no sleep data available for the first night
      # We will now add 5 minutes of dummy waking time before it this now, 
      # and label it as non-wear.
      # We do this to make sure that the day numbering
      # and merging of the sleep variables is still consistent with
      # the other recording.
      if (!is.na(firstonset)) {
        dummywake = max(firstonset - round(Nepochsinhour/12), nightsi[1] + round(Nepochsinhour * 6))
        ts$diur[1:dummywake] = 1 
        ts$nonwear[1:firstonset] = 1
        ts$guider[1:dummywake] = guider
      }
    }
  }
  return(ts)
}

g.sibreport = function(ts, ID, epochlength, logs_diaries=c(), desiredtz="") {
  dayind = which(ts$diur == 0)
  sib_starts = which(diff(c(0,ts$sibdetection[dayind],0)) == 1)
  sib_ends = which(diff(c(ts$sibdetection[dayind],0)) == -1)
  Nsibs = length(sib_starts)
  sibreport = c()
  if (Nsibs > 0) {
    sibreport = data.frame(ID = rep(ID, Nsibs),
                           type = rep("sib", Nsibs),
                           start = character(Nsibs),
                           end = character(Nsibs),
                           duration = numeric(Nsibs),
                           mean_acc_1min_before = numeric(Nsibs),
                           mean_acc_1min_after = numeric(Nsibs), stringsAsFactors = FALSE)
    for (sibi in 1:Nsibs) {
      sibreport$start[sibi]  = format(ts$time[dayind[sib_starts[sibi]]])
      sibreport$end[sibi] = format(ts$time[dayind[sib_ends[sibi]]])
      if (is.ISO8601(sibreport$start[sibi])) {
        sibreport$start[sibi] = format(iso8601chartime2POSIX(sibreport$start[sibi], tz = desiredtz))
        sibreport$end[sibi] = format(iso8601chartime2POSIX(sibreport$end[sibi], tz = desiredtz))
      }
      sibreport$duration[sibi] = ((sib_ends[sibi] - sib_starts[sibi]) + 1) / (60/epochlength)
      boutind = sib_starts[sibi]:sib_ends[sibi]
      minute_before = (sib_starts[sibi] - (60/epochlength)):(sib_starts[sibi] - 1)
      minute_after = (sib_ends[sibi] + 1):(sib_ends[sibi] + (60/epochlength))
      if (min(minute_before) > 1) {
        sibreport$mean_acc_1min_before[sibi]  = round(mean(ts$ACC[dayind[minute_before]]), digits = 3)
      }
      if (max(minute_after) < nrow(ts)) {
        sibreport$mean_acc_1min_after[sibi]  = round(mean(ts$ACC[dayind[minute_after]]), digits = 3)
      }
    }
  }
  if (length(logs_diaries) > 0) {
    # extract self-reported nonwear and naps
    nonwearlog = logs_diaries$nonwearlog
    naplog = logs_diaries$naplog
    sleeplog = logs_diaries$sleeplog
    bedlog = logs_diaries$bedlog
    imputecodelog = logs_diaries$imputecodelog
    dateformat = logs_diaries$dateformat

    firstDate = as.Date(ts$time[1], tz = desiredtz)
    # when recording starts after midnight and before 4am we count the previous date as the first night
    if (as.numeric(format(ts$time[1], "%H")) < 4) firstDate = firstDate - 1
    extract_logs = function(log, ID, logname, firstDate = NULL) {
      logreport = c()
      if (length(log) > 0) {
        idwi = g.part4_extractid(idloc = 1, # idloc is not relevant in place, we can take full ID
                                 fname = log$ID,
                                 dolog = TRUE,
                                 sleeplog = log, accid = ID)
        relevant_rows = idwi$matching_indices_sleeplog
        if (length(relevant_rows) > 0) {
          log = log[relevant_rows,] # extract ID
          if (!is.null(firstDate)) {
            # Add date if missing and remove unneeded columns
            log$date = firstDate + as.numeric(log$night) - 1
            log = log[,c("ID", "date", grep(pattern = "onset|wake|bed", x = names(log), value = TRUE))]
          }
          for (i in 1:nrow(log)) { # loop over lines (days)
            
            tmp = log[i,] # convert into timestamps
            # only attempt if there are at least 2 timestamps to process
            if (ncol(tmp) <= 2) next
            nonempty = which(tmp[3:ncol(tmp)] != "" & tmp[3:ncol(tmp)] != "NA")
            if (length(nonempty) > 1) {
              date = as.Date(tmp[1,2], format = dateformat, tz = desiredtz)
              times = as.character(unlist(tmp[1,3:ncol(tmp)]))
              times = grep(pattern = "NA", value = TRUE, invert = TRUE, x = times)
              times = gsub(pattern = " ", replacement = "", x = times)
              times = times[which(times %in% c("", "NA") == FALSE)]
              # ignore entries without start and/or end time
              t_to_remove = c()
              for (ji in 1:floor(length(times)/2)) {
                check = ((ji * 2) - 1):(ji * 2)
                if (length(which(times[check] == "")) > 0) {
                  t_to_remove = c(t_to_remove, check)
                }
              }
              if (length(t_to_remove) > 0) {
                times = times[-t_to_remove]
              }
              # put remaining timestamps in logreport
              if (length(times) > 1) {
                Nevents = floor(length(times) / 2)
                timestamps = as.POSIXlt(paste0(date, " ", times), tz = desiredtz)
                hour = as.numeric(format(timestamps, "%H"))
                if (!is.null(firstDate)) {
                  # date on which night start is default
                  # time before noon are likely to occur on the next day
                  AM = which(hour <= 12)
                  
                  # When wake is after noon and before 6pm
                  # then it is a daysleeper and the wake up time
                  # is also likely to be on the next day
                  if (hour[2] > 12 & hour[2] <= 18) { # daysleeper
                    AM = c(AM, 2)
                  }
                  # however, wake can never be after onset when wake is at AM
                  # 11am - 9am, simply means that person apparently reported to
                  # sleep from 11am to 9am next day. Most likely a mistake in the
                  # diary, but not to be interpretted as 9am-11am
                  if (hour[1] > hour[2] && 1 %in% AM) {
                    AM = AM[which(AM != 1)]
                  }
                  if (length(AM) > 0) {
                    timestamps[AM] = as.POSIXlt(paste0(date + 1, " ", times[AM]), tz = desiredtz)
                  }
                } else {
                  # self-reported nap or nonwear startign before midnight and ending after midnight
                  if (hour[1] >= 12 & hour[2] < 12) {
                    timestamps[2] = as.POSIXlt(paste0(date, " ", times[2]), tz = desiredtz)
                  }
                }
                timestamps = sort(timestamps)
                logreport_tmp = data.frame(ID = rep(ID, Nevents),
                                           type = rep(logname, Nevents),
                                           start = rep("", Nevents),
                                           end = rep("", Nevents),
                                           duration = rep(0, Nevents), stringsAsFactors = FALSE)
                for (bi in 1:Nevents) {
                  tt1 = as.POSIXct(timestamps[(bi * 2) - 1], tz = desiredtz)
                  tt2 = as.POSIXct(timestamps[(bi * 2)], tz = desiredtz)
                  logreport_tmp$start[bi]  = format(tt1)
                  logreport_tmp$end[bi] = format(tt2)
                  if (length(unlist(strsplit(logreport_tmp$start[bi], " "))) == 1) {
                    logreport_tmp$start[bi] = paste0(logreport_tmp$start[bi], " 00:00:00")
                  }
                  if (length(unlist(strsplit(logreport_tmp$end[bi], " "))) == 1) {
                    logreport_tmp$end[bi] = paste0(logreport_tmp$end[bi], " 00:00:00")
                  }
                  logreport_tmp$duration[bi] = abs(as.numeric(difftime(time1 = tt1, time2 = tt2, units = "mins")))
                }
              }
              if (length(logreport) == 0) {
                logreport = logreport_tmp
              } else {
                logreport = rbind(logreport, logreport_tmp)
              }
            }
          }
        }
      }
      return(logreport)
    }
    naplogreport = extract_logs(naplog, ID, logname = "nap")
    nonwearlogreport = extract_logs(nonwearlog, ID, logname = "nonwear")
    sleeplogreport = extract_logs(sleeplog, ID, logname = "sleeplog", firstDate = firstDate)
    bedlogreport = extract_logs(bedlog, ID, logname = "bedlog", firstDate = firstDate)
    # add imputecodelog
    if (length(imputecodelog) > 0) {
      addDate = function(x, tz = desiredtz) {
        start_time = as.POSIXct(x$start, tz = tz)
        x$date = as.Date(start_time, tz = tz)
        start_hour = as.numeric(format(start_time, "%H"))
        start_am = which(start_hour < 12)
        if (length(start_am) > 0) {
          x$date[start_am] = x$date[start_am] - 1
        }
        return(x)
      }
      if (length(sleeplogreport) > 0) sleeplogreport = addDate(sleeplogreport)
      if (length(bedlogreport) > 0) bedlogreport = addDate(bedlogreport)
      
      imputecodelog_tmp = imputecodelog[which(imputecodelog$ID == ID), ]
      if (length(sleeplogreport) > 0) {
        sleeplogreport = merge(sleeplogreport, imputecodelog_tmp, by = c("ID", "date"))
        sleeplogreport = sleeplogreport[, which(colnames(sleeplogreport) != "date")]
      }
      if (length(bedlogreport) > 0) {
        bedlogreport = merge(bedlogreport, imputecodelog_tmp, by = c("ID", "date"))
        bedlogreport = bedlogreport[, which(colnames(bedlogreport) != "date")]
      }
    }
    logreport = sibreport
    # append all together in one output data.frame
    if (length(logreport) > 0 & length(naplogreport) > 0) {
      logreport = merge(logreport, naplogreport, by = c("ID", "type", "start", "end", "duration"), all = TRUE)
    } else if (length(logreport) == 0 & length(naplogreport) > 0) {
      logreport = naplogreport
    }
    if (length(logreport) > 0 & length(nonwearlogreport) > 0) {
      logreport = merge(logreport, nonwearlogreport, by = c("ID", "type", "start", "end", "duration"), all = TRUE)
    } else if (length(logreport) == 0 & length(nonwearlogreport) > 0) {
      logreport = nonwearlogreport
    }
    if (length(logreport) > 0 & length(sleeplogreport) > 0) {
      logreport = merge(logreport, sleeplogreport, by = c("ID", "type", "start", "end", "duration"), all = TRUE)
    } else if (length(logreport) == 0 & length(sleeplogreport) > 0) {
      logreport = sleeplogreport
    }
    if (length(logreport) > 0 & length(bedlogreport) > 0) {
      if ("imputecode" %in% colnames(logreport) && "imputecode" %in% colnames(bedlogreport)) {
        include_imputecode = "imputecode"
      } else {
        include_imputecode = NULL
      }
      logreport = merge(logreport, bedlogreport,
                        by = c("ID", "type", "start", "end", "duration", include_imputecode), all = TRUE)
    } else if (length(logreport) == 0 & length(bedlogreport) > 0) {
      logreport = bedlogreport
    }
  } else {
    logreport = sibreport
  }
  # add midnight timetimes which got lost
  convert2POSIX = function(x) {
    if (length(unlist(strsplit(x, " "))) == 1) {
      x = paste0(x, " 00:00:00")
    }
    return(x)
  }
  logreport$start = as.POSIXct(unlist(lapply(X = logreport$start, FUN = convert2POSIX)), tz = desiredtz)
  logreport$end = as.POSIXct(unlist(lapply(X = logreport$end, FUN = convert2POSIX)), tz = desiredtz)
  return(logreport)
}

library(GGIR)
context("g.sibreportc")
test_that("g.sibreport creates expected object", {
  ts = data.frame(sibdetection = c(0, 0, 1, 1, 1, 0, 0),
                  diur = rep(0, 7),
                  time = seq(as.POSIXlt(x = "2021-3-3 15:00:00", tz = "Europe/Amsterdam"),
                             as.POSIXlt(x = "2021-3-3 15:06:00", tz = "Europe/Amsterdam"),
                             by = 60),
                  ACC = 1:7, angle = 7:1)
  ID = 12345
  epochlength  = 60
  nonwearlog = data.frame(ID = rep(ID, 5),
                          date = seq(as.Date("2021-3-3"), as.Date("2021-3-7"), by = 1),
                          nonwear1 = rep("15:00:00", 5),
                          nonwear2 = rep("15:30:00", 5))
  naplog = data.frame(ID = rep(ID, 2),
                      date = seq(as.Date("2021-3-3"),as.Date("2021-3-4"), by = 1),
                      nonwear1 = rep("13:00:00", 2),
                      nonwear2 = rep("13:30:00", 2))
  
  logs_diaries = list(sleeplog = c(), nonwearlog = nonwearlog, 
                      naplog = naplog, dateformat = "%Y-%m-%d")
  
  SIBREPORT = g.sibreport(ts, ID, epochlength, logs_diaries)
  
  expect_equal(nrow(SIBREPORT), 8)
  expect_equal(mean(SIBREPORT$duration, na.rm = TRUE), 26.625)
  expect_equal(mean(SIBREPORT$mean_acc_1min_before, na.rm = TRUE), 2)
  expect_equal(mean(SIBREPORT$mean_acc_1min_after, na.rm = TRUE), 6)
  expect_equal(format(SIBREPORT$start[8]), "2021-03-03 15:02:00")
  expect_equal(format(SIBREPORT$end[8]), "2021-03-03 15:04:00")
})

get_nw_clip_block_params = function(monc, dformat, deviceSerialNumber = "", sf,
                                    params_rawdata) {
  blocksize = round(14512 * (sf/50) * params_rawdata[["chunksize"]])
  if (monc == MONITOR$GENEA) blocksize = round(21467 * (sf/80)  * params_rawdata[["chunksize"]])
  if (monc == MONITOR$ACTIGRAPH && dformat == FORMAT$CSV) blocksize = round(blocksize)#round(blocksize/5)
  if (monc == MONITOR$ACTIGRAPH && dformat == FORMAT$GT3X) blocksize = (24 * 3600) * params_rawdata[["chunksize"]]
  if (monc == MONITOR$AXIVITY && dformat == FORMAT$CWA) {
    if (utils::packageVersion("GGIRread") >= "0.3.1") {
      # 24-hour block.
      # CWA data blocks can have 40, 80 or 120 samples each; we'll take 80 as the average number.
      blocksize = round(24 * 3600 * sf / 80 * params_rawdata[["chunksize"]])      
    } else {
      blocksize = round(blocksize * 1.0043)
    }
  }
  if (monc == MONITOR$AXIVITY && dformat == FORMAT$CSV) blocksize = round(blocksize)
  if (monc == MONITOR$MOVISENS) blocksize = sf * 60 * 1440
  if (monc == MONITOR$VERISENSE && dformat == FORMAT$CSV) blocksize = round(blocksize)
  if (monc == MONITOR$PARMAY_MTX) blocksize = round(1440 / 2 * params_rawdata[["chunksize"]])

  dynrange = params_rawdata[["dynrange"]]
  if (monc == MONITOR$ACTIGRAPH) {
    # If Actigraph then try to specify dynamic range based on Actigraph model
    if (length(grep(pattern = "CLE", x = deviceSerialNumber)) == 1) {
      dynrange = 6
    } else if (length(grep(pattern = "MOS", x = deviceSerialNumber)) == 1) {
      dynrange = 8
    } else if (length(grep(pattern = "NEO", x = deviceSerialNumber)) == 1) {
      dynrange = 6
    }
  }

  # Clipping threshold
  if (length(dynrange) > 0) {
    clipthres = dynrange - 0.5
  } else {
    clipthres = 7.5 # hard-coded assumption that dynamic range is 8g
    #if (monc == MONITOR$GENEA) clipthres = 5.5
    if (monc == MONITOR$MOVISENS) {
      clipthres = 15.5 # hard coded assumption that dynamic range is 16g
    } else if (monc == MONITOR$AD_HOC) {
      clipthres = params_rawdata[["rmc.dynamic_range"]]
    }
  }
  # Nonwear threshold: non-wear criteria are monitor-specific
  racriter = params_rawdata[["nonwear_range_threshold"]] / 1000
  sdcriter = 0.013
  #if (monc == MONITOR$GENEA) { sdcriter = 0.003;  racriter = 0.05 }
  if (monc == MONITOR$VERISENSE) {
    racriter = 0.20
  } else if (monc == MONITOR$AD_HOC) {
    if (length(params_rawdata[["rmc.noise"]]) == 0) {
      stop("Argument rmc.noise not specified, please specify expected noise level in g-units")
    }
    sdcriter = params_rawdata[["rmc.noise"]] * 1.2
  }
  invisible(list(clipthres=clipthres, blocksize=blocksize, sdcriter=sdcriter, racriter=racriter))
}

g.sib.sum = function(SLE,M,ignorenonwear=TRUE,desiredtz="") {
  A = as.data.frame(SLE$output, stringsAsFactors = TRUE)
  invalid = A$invalid
  if (ignorenonwear == TRUE) {
    if (length(which(A$invalid==1)) > 0) {
      A[which(A$invalid==1), which(colnames(A) %in% c("time", "night") == FALSE)] = 0
    }
  }
  space = ifelse(length(unlist(strsplit(format(A$time[1])," "))) > 1,TRUE,FALSE)
  time = format(unlist(A$time))
  #time stored as iso8601 what makes it sensitive to daylight saving days.
  if (space == TRUE) {
    time = POSIXtime2iso8601(as.POSIXlt(time, tz = desiredtz), tz = desiredtz)
  }
  night = A$night
  sleep = as.data.frame(as.matrix(A[,(which(colnames(A)=="night")+1):ncol(A)]), stringsAsFactors = TRUE)
  colnames(sleep) = colnames(A)[(which(colnames(A)=="night")+1):ncol(A)]
  ws3 = M$windowsizes[1]
  ws2 = M$windowsizes[2]
  sib.cla.sum = as.data.frame(matrix(-1, 1000,9))
  # label night after last night as -1 to differentiate it from night 0
  lastNightEnd = max(which(night == max(night)))
  if (length(night) > lastNightEnd) {
    night[(lastNightEnd + 1):length(night)] = -1
  }
  # unique nights
  un = unique(night) #unique nights
  missingnights = which(un == -1 | is.na(un) == TRUE)
  if (length(missingnights) > 0) {
    un = un[-missingnights]
  }
  # if recording starts after 4am, then also remove first night
  firstTS_posix = iso8601chartime2POSIX(M$metashort$timestamp[1], tz = desiredtz)
  firstday4am_posix = as.POSIXct(paste0(as.Date(firstTS_posix), " 04:00:00"), tz = desiredtz)
  if (firstday4am_posix < firstTS_posix) { # then first timestamp is later than 4am
    missingnights = which(un == 0)
    if (length(missingnights) > 0) {
      un = un[-missingnights]
    }
  }
  if (length(un) != 0) {
    if (is.numeric(max(un)) == TRUE) {
      cnt = 1
      for (i in un) { #nights #length(un)
        qqq1 = which(night == i)[1]
        qqq2 = which(night == i)[length(which(night == i))]
        if (length(qqq1) == 1 & length(qqq2) == 1) {
          time.t = time[qqq1:qqq2]
          sleep.t = as.data.frame(sleep[qqq1:qqq2,], stringsAsFactors = TRUE)
          colnames(sleep.t) = colnames(sleep)
          invalid.t = invalid[qqq1:qqq2]
          for (j in 1:ncol(sleep.t)) { #sleep definitions
            nsleepperiods = length(which(diff(c(0, sleep.t[,j])) == 1))
            if (nsleepperiods > 0) {
              start_sp = which(diff(c(0, sleep.t[,j])) == 1)
              end_sp = which(diff(c(sleep.t[,j], 0)) == -1)
              if(length(end_sp) == 0) end_sp = nrow(sleep.t) #if sleep period ends the next 'nightday'
              if (start_sp[1] > end_sp[1]) { #if period starts with sleep
                start_sp = c(1,start_sp)
              }
              if (start_sp[length(start_sp)] > end_sp[length(end_sp)]) { #if period ends with sleep
                end_sp = c(end_sp,nrow(sleep.t))
              }
              nsleepperiods = length(start_sp)
            }
            colnames(sib.cla.sum)[1:9] = c("night", "definition", "start.time.day",
                                           "nsib.periods", "tot.sib.dur.hrs",
                                           "fraction.night.invalid",
                                           "sib.period", "sib.onset.time",
                                           "sib.end.time")
            NepochsInDay = (60/ws3)*1440
            Nmissingvalues = max(c(0, (NepochsInDay - length(invalid.t))))
            fraction.night.invalid = (length(which(invalid.t != 0)) + Nmissingvalues) / NepochsInDay
            if (fraction.night.invalid > 1) fraction.night.invalid = 1 # day with 25 hours
            if (nsleepperiods == 0) {
              sib.cla.sum[cnt, 1] = i #night
              sib.cla.sum[cnt, 2] = colnames(sleep.t)[j] #definition
              sib.cla.sum[cnt, 3] = format(time.t[1])
              sib.cla.sum[cnt, 4] = nsleepperiods #number of sleep periods
              sib.cla.sum[cnt, 5] = 0 #total sleep duration
              sib.cla.sum[cnt, 6] = fraction.night.invalid #length(invalid.t)
              sib.cla.sum[cnt, 7] = 0
              sib.cla.sum[cnt, 8:9] = ""
              cnt = cnt + 1
            } else {
              for (spi in 1:nsleepperiods) {
                sib.cla.sum[cnt, 1] = i #night
                sib.cla.sum[cnt, 2] = colnames(sleep.t)[j] #definition
                sib.cla.sum[cnt, 3] = format(time.t[1])
                sib.cla.sum[cnt, 4] = nsleepperiods #number of sleep periods
                sleep_sp = sleep.t[start_sp[spi]:end_sp[spi],j]
                time_sp = time.t[start_sp[spi]:end_sp[spi]]
                sleep_dur = (round((length(which(sleep_sp == 1))/(60/ws3)) * 100)) / 100
                # general characteristics of the time window
                sib.cla.sum[cnt, 5] = sleep_dur/60 #total sleep duration
                sib.cla.sum[cnt, 6] = fraction.night.invalid #length(invalid.t)
                sib.cla.sum[cnt, 7] = spi
                sib.cla.sum[cnt, 8] = format(time_sp[which(sleep_sp == 1)[1]])
                sib.cla.sum[cnt, 9] = format(time_sp[length(sleep_sp)])
                cnt = cnt + 1
              }
            }
            if (cnt  > 800) {
              emptydf = as.data.frame(matrix(0, 1000, ncol(sib.cla.sum)))
              colnames(emptydf) = colnames(sib.cla.sum)
              sib.cla.sum = rbind(sib.cla.sum, emptydf)
            }
          }
        }
      }
    }
  }
  sib.cla.sum = sib.cla.sum[-which(sib.cla.sum$night == -1 | sib.cla.sum$sib.period == 0), ]
  return(sib.cla.sum)
}

g.impute = function(M, I, params_cleaning = c(), desiredtz = "",
                    dayborder = 0, TimeSegments2Zero = c(), acc.metric = "ENMO", 
                    ID, qwindowImp = c(), ...) {
  
  #get input variables
  input = list(...)
  if (length(input) > 0 || length(params_cleaning) == 0) {
    # Extract and check parameters if user provides more arguments than just the parameter arguments,
    # or if params_[...] aren't specified (so need to be filled with defaults).
    # So, inside GGIR this will not be used, but it is used when g.impute is used on its own
    # as if it was still the old g.impute function
    params = extract_params(params_cleaning = params_cleaning,
                            input = input,
                            params2check = c("cleaning")) # load default parameters
    params_cleaning = params$params_cleaning
    rm(params)
  }
  
  windowsizes = M$windowsizes #default: c(5,900,3600)
  metashort = M$metashort
  metalong = M$metalong
  shortEpoch = windowsizes[1]
  mediumEpoch = windowsizes[2]
  longEpoch = windowsizes[3]
  n_short_in_mediumEpoch = mediumEpoch/shortEpoch
  # Move some object from params_cleaning to local to reduce repeated retrieval
  # and for readability
  ndayswindow = params_cleaning[["ndayswindow"]] 
  data_masking_strategy = params_cleaning[["data_masking_strategy"]]
  hrs.del.start = params_cleaning[["hrs.del.start"]]
  hrs.del.end = params_cleaning[["hrs.del.end"]]
  maxdur = params_cleaning[["maxdur"]]
  max_calendar_days = params_cleaning[["max_calendar_days"]]
  rm(M)
  # What is the minimum number of accelerometer axis needed to meet the criteria for nonwear in order for the data to be detected as nonwear?
  wearthreshold = 2 #needs to be 0, 1 or 2
  # epochs per day (there are 86400 seconds in a day)
  n_longEpoch_perday = 86400 / longEpoch # note: 
  n_mediumEpoch_perday = 86400 / mediumEpoch
  n_shortEpoch_perday = 86400 / shortEpoch
  # epochs per minute
  n_shortEpoch_permin = 60 / shortEpoch
  # epochs per hour
  n_longEpoch_perhour = 3600 / longEpoch
  n_medium_perhour = 3600 / mediumEpoch
  n_shortEpoch_perhour = 3600 / shortEpoch
  
  #check that matrices match
  if ((nrow(metalong) / n_mediumEpoch_perday) - (nrow(metashort) / n_shortEpoch_perday) > 0.1) {
    warning("Matrices 'metalong' and 'metashort' are not compatible", call. = FALSE)
  }
  tmi = which(colnames(metalong) == "timestamp")
  time = as.character(as.matrix(metalong[,tmi]))
  startt = as.matrix(metalong[1, tmi])
  #====================================
  # Deriving file characteristics from 15 min summary files
  LD = nrow(metalong) * (mediumEpoch / 60) #length data in minutes
  ND = nrow(metalong) / n_mediumEpoch_perday #number of days
  #==============================================
  # Generating time variable
  timeline = seq(0, ceiling(nrow(metalong) / n_mediumEpoch_perday),
                 by = 1/n_mediumEpoch_perday)
  timeline = timeline[1:nrow(metalong)]
  
  #========================================
  # Extracting non-wear and clipping and make decision on which additional time needs to be considered non-wear
  out = g.weardec(metalong, wearthreshold, mediumEpoch,
                  params_cleaning = params_cleaning,
                  desiredtz = desiredtz,
                  qwindowImp = qwindowImp)
  r1 = out$r1 #non-wear
  r2 = out$r2 #clipping
  r3 = out$r3 #additional non-wear
  r4 = matrix(0,length(r3),1) #protocol based decisions on data removal
  LC = out$LC
  LC2 = out$LC2
  nonwearHoursFiltered = out$nonwearHoursFiltered
  nonwearEventsFiltered = out$nonwearEventsFiltered
  #========================================================
  # Check whether TimeSegments2Zero exist, because this means that the
  # user wants to ignore specific time windows. This feature is used
  # for example if the accelerometer was not worn during the night and the user wants
  # to include the nighttime acceleration in the analyses without imputation,
  # but wants to use imputation for the rest of the day.
  # So, those time windows should not be imputed.
  # and acceleration metrics should have value zero during these windows.
  if (length(TimeSegments2Zero) > 0) {
    r1long = matrix(0, length(r1), n_short_in_mediumEpoch)
    r1long = replace(r1long, 1:length(r1long), r1)
    r1long = t(r1long)
    dim(r1long) = c(length(r1) * n_short_in_mediumEpoch, 1)
    timelinePOSIX = iso8601chartime2POSIX(metashort$timestamp,tz = desiredtz)
    # Combine r1Long with TimeSegments2Zero
    for (i in 1:nrow(TimeSegments2Zero)) {
      startTurnZero = which(timelinePOSIX == TimeSegments2Zero$windowstart[i])
      endTurnZero = which(timelinePOSIX == TimeSegments2Zero$windowend[i])
      r1long[startTurnZero:endTurnZero] = 0
      # Force ENMO and other acceleration metrics to be zero for these intervals
      metashort[startTurnZero:endTurnZero, 
                which(colnames(metashort) %in% c("timestamp","anglex","angley","anglez") == FALSE)] = 0
    }
    # collaps r1long (short epochs) back to r1 (long epochs)
    r1longc = cumsum(c(0, r1long))
    select = seq(1, length(r1longc), by = n_short_in_mediumEpoch)
    r1 = diff(r1longc[round(select)]) / abs(diff(round(select)))
    r1 = round(r1)
  }
  
  #======================================
  # detect first and last midnight and all midnights
  tooshort = 0
  dmidn = g.detecmidnight(time, desiredtz, dayborder)
  firstmidnight = dmidn$firstmidnight;  firstmidnighti = dmidn$firstmidnighti
  lastmidnight = dmidn$lastmidnight;    lastmidnighti = dmidn$lastmidnighti
  midnights = dmidn$midnights;          midnightsi = dmidn$midnightsi
  #===================================================================
  # Trim data based on study_dates_file
  study_dates_log_used = c(FALSE, FALSE) # two boolean, one for start and one for end
  study_date_indices = NULL
  if (!is.null(params_cleaning[["study_dates_file"]])) {
    # Read content of study dates file 
    studyDates = data.table::fread(file = params_cleaning[["study_dates_file"]], data.table = FALSE)
    # Check ID and date formats
    studyDates = check_log(log = studyDates, 
                           dateformat = params_cleaning[["study_dates_dateformat"]], 
                           colid = 1, datecols = 2:3, 
                           logPath = params_cleaning[["study_dates_file"]],
                           logtype = "study dates log")
    # identify first and last study dates
    rowID = which(studyDates[, 1] == ID)
    if (length(rowID) > 0) {
      if (length(rowID) > 1) {
        # this is in the unlikely event that an ID appears twice in the log
        rowID = rowID[1]
        warning(paste0("The ID ", ID, " appears twice in the study dates log"), call. = FALSE)
      }
      # expected first and last midnight
      fmt = params_cleaning[["study_dates_dateformat"]]
      firstmidnight = as.Date(studyDates[rowID, 2], format = fmt)
      lastmidnight = as.Date(studyDates[rowID, 3], format = fmt)
      # find the dates in metalong
      if (!is.na(firstmidnight)) {
        firstmidnighti = midnightsi[grep(firstmidnight, midnights)][1]
      } else {
        firstmidnighti = NULL
      }
      if (!is.na(lastmidnight)) {
        lastmidnighti = midnightsi[grep(lastmidnight, midnights) + 1][1] # plus 1 to include the reported date in log
      } else {
        lastmidnighti = NULL
      }
      # trim start
      if (length(firstmidnighti) > 0) {
        r4[1:(firstmidnighti - 1)] = 1
        study_dates_log_used[1] = TRUE
      } else {
        # if midnight timestamp for the date is not available, 
        # do not trim the data and recover the firstmidnight value
        list2env(dmidn[c("lastmidnighti", "lastmidnight")], envir = environment())
        # warning(paste0("The start date provided in the study dates file for ID = ", 
        #                ID, "is not within the dates available in the recording. ",
        #                "The data was not trimmed at the beginning of the recording."), call. = FALSE)
      }
      # trim end
      if (length(lastmidnighti) > 0) {
        r4[lastmidnighti:nrow(r4)] = 1
        study_dates_log_used[2] = TRUE
      } else {
        # if midnight timestamp for the date is not available, 
        # do not trim the data and recover the lastmidnight value
        list2env(dmidn[c("lastmidnighti", "lastmidnight")], envir = environment())
        # warning(paste0("The end date provided in the study dates file for ID = ", 
        #                ID, "is not within the dates available in the recording. ",
        #                "The data was not trimmed at the end of the recording."), call. = FALSE)
      }
      # cut out r4 to apply strategies only on the trimmed portion of data
      # after application of strategies, r4 will be reset to the original length
      if (any(study_dates_log_used)) {
        r4_bu = r4 # backup copy of r4 before cutting it to be imputed later on
        study_date_indices = which(r4[, 1] == 0)
        r4 = as.matrix(r4[study_date_indices, , drop = FALSE])
      }
    } else if (length(rowID) == 0) {
      warning(paste0("The ID ", ID, " does not appear  in the study dates log ",
                     "the full recording has been considered within the study ",
                     "protocol selection"), call. = FALSE)
    }
  }
  #===================================================================
  # Select data based on data_masking_strategy
  if (data_masking_strategy == 1) { # only mask x hours from start and x hours from end
    if (hrs.del.start > 0) {
      r4[1:(hrs.del.start * n_medium_perhour)] = 1
    }
    if (hrs.del.end > 0) {
      if (length(r4) > hrs.del.end * n_medium_perhour) {
        r4[(length(r4) + 1 - (hrs.del.end * n_medium_perhour)):length(r4)] = 1
      } else {
        r4[1:length(r4)] = 1
      }
    }
    if (LD < 1440) {
      r4 = r4[1:floor(LD / (mediumEpoch / 60))]
    }
    starttimei = 1
    endtimei = length(r4)
  } else if (data_masking_strategy == 2) { # first to last midnight
    starttime = firstmidnight
    endtime = lastmidnight  
    # only apply data_masking_strategy 2 if study dates log is not used for trimming the data,
    # otherwise the data already start and finishes at a midnight
    if (!any(study_dates_log_used)) {
      starttimei = firstmidnighti
      endtimei = lastmidnighti
      if (firstmidnighti != 1) { #ignore everything before the first midnight
        r4[1:(firstmidnighti - 1)] = 1 #-1 because first midnight 00:00 itself contributes to the first full day
      }
      r4[(lastmidnighti):length(r4)] = 1  #ignore everything after the last midnight
    } else {
      starttimei = 1
      endtimei = length(r4)
    }
  } else if (data_masking_strategy %in% c(3, 5)) { #select X most active days
    # Prepare time series of activity needed for both strategies
    if (acc.metric %in% colnames(metashort)) {
      acc = as.numeric(as.matrix(metashort[, acc.metric]))
    } else {
      acc.metric = grep("timestamp|angle", colnames(metashort),
                        value = TRUE, invert = TRUE)[1]
      acc = as.numeric(as.matrix(metashort[, acc.metric]))
    }
    acc[which(rep(r2, each = n_short_in_mediumEpoch) == 1 |
                rep(r1, each = n_short_in_mediumEpoch) == 1)] = 0
    if (!is.null(study_date_indices)) {
      # If study_dates_file was used then r4 has possibly been trimmed
      # If this is the case then also trim acc to allow for direct comparisons
      if (study_dates_log_used[1] == TRUE) {
        tt1 = ((study_date_indices[1] - 1) * n_short_in_mediumEpoch) + 1
      } else {
        tt1 = 1
      }
      if (study_dates_log_used[2] == TRUE) {
        tt2 = max(study_date_indices) * n_short_in_mediumEpoch
      } else {
        tt2 = length(acc)
      }
      acc = acc[tt1:tt2]
    }
    if (data_masking_strategy == 3) {
      # Find the most active ndayswindow block via a rolling mean window
      NDAYS = length(acc) / n_shortEpoch_perday
      acc_roll_mean = zoo::rollmean(x = acc, k = ndayswindow * n_shortEpoch_perday, align = "left")
      start_ndayswindow_hour = floor(which(acc_roll_mean == max(acc_roll_mean))[1] / n_shortEpoch_perhour)
      hrs.del.start = start_ndayswindow_hour + hrs.del.start
      maxdur = ((start_ndayswindow_hour / 24) + ndayswindow) - (hrs.del.end/24)
      if (maxdur > NDAYS) maxdur = NDAYS
      # Update r4 (which is medianEpoch size)
      if (hrs.del.start > 0) {
        start_epoch = max((hrs.del.start * n_medium_perhour) - 1, 1)
        r4[1:start_epoch] = 1
      }
      ignore_from = round((maxdur * n_mediumEpoch_perday))
      if (maxdur > 0 && length(r4) > (ignore_from + 1)) {
        r4[ignore_from:length(r4)] = 1
      }
      if (LD < 1440) r4 = r4[1:floor(LD / (mediumEpoch / 60))]
    } else if (data_masking_strategy == 5) {
      # Select the most active calendar days
      acc_roll_mean = c()
      # readjust midnightsi if study dates log used for trimming the data
      if (study_dates_log_used[1] == TRUE) {
        midnightsi = midnightsi[midnightsi >= firstmidnighti & midnightsi <= lastmidnighti]
        firstmidnighti = firstmidnighti - (midnightsi[1] - 1)
        lastmidnighti = lastmidnighti - midnightsi[1]
        midnightsi = midnightsi - midnightsi[1] + 1
      }
      for (i in 1:length(midnightsi)) {
        p0 = ((midnightsi[i] - 1) * n_short_in_mediumEpoch) + 1
        if (i == length(midnightsi) && i + ndayswindow > length(midnightsi)) {
          p1 = length(acc)
        } else {
          p1 = (midnightsi[i + ndayswindow] - 1) * n_short_in_mediumEpoch
        }
        if (is.na(p1) || p1 > length(acc)) {
          break
        }
        acc_roll_mean[i] = mean(acc[p0:p1], na.rm = TRUE)
      }
      start_index_most_active_window = ifelse(length(acc_roll_mean) > 0, which.max(acc_roll_mean), 1)
      offset = ifelse(study_dates_log_used[1] == TRUE, firstmidnighti - 1, 0)
      # Ignore until
      ignore_until = midnightsi[start_index_most_active_window] + (hrs.del.start * n_medium_perhour) - 1 - offset
      if (ignore_until > 0) r4[1:pmin(ignore_until, length(r4))] = 1
      # Ignore from
      target_idx = start_index_most_active_window + ndayswindow
      if (target_idx > length(midnightsi)) target_idx = length(midnightsi)
      ignore_from = midnightsi[target_idx] - (hrs.del.end * n_medium_perhour) - offset
      if (!is.na(ignore_from) && ignore_from < length(r4)) {
        r4[pmax(1, ignore_from):length(r4)] = 1
      }
    }
    starttimei = 1
    endtimei = length(r4)
  } else if (data_masking_strategy == 4) { #from first midnight to end of recording
    starttime = firstmidnight
    endtime = lastmidnight  
    # only apply data_masking_strategy 4 if study dates log is not used for trimming the data,
    # otherwise the data already start and finishes at midnight
    if (!any(study_dates_log_used)) {
      starttimei = firstmidnighti
      endtimei = lastmidnighti
      if (firstmidnighti != 1) { #ignore everything before the first midnight
        r4[1:(firstmidnighti - 1)] = 1 # -1 because first midnight 00:00 itself contributes to the first full day
      }
    } else {
      starttimei = 1
      endtimei = length(r4)
    }
  }
  # Mask data based on maxdur
  if (maxdur > 0 &&
      (length(r4) > (maxdur * n_mediumEpoch_perday) + 1)) {
    r4[((maxdur * n_mediumEpoch_perday) + 1):length(r4)] = 1
  }
  # Mask data based on max_calendar_days
  if (max_calendar_days > 0) {
    # get dates that are part of the study protocol
    if (any(study_dates_log_used)) {
      datetime = metalong$timestamp[firstmidnighti:(lastmidnighti - 1)]
    } else {
      datetime = metalong$timestamp
    }
    dates = as.Date(iso8601chartime2POSIX(datetime, tz = desiredtz))
    if (max_calendar_days < length(unique(dates))) {
      lastDateToInclude = sort(unique(dates))[max_calendar_days]
      r4[which(dates > lastDateToInclude)] = 1
    }
  }
  #===================================================================
  # if data selected based on study dates log, recover the original length of r4
  if (any(study_dates_log_used)) {
    r4_bu[which(r4_bu == 0),] = r4
    r4 = r4_bu
  }
  #========================================================================================
  # Impute shortEpoch second data based on mediumEpoch minute estimates of non-wear time
  r5 = r1 + r2 + r3 + r4
  r5[which(r5 > 1) ] = 1
  r5[which(metalong$nonwearscore == -1) ] = -1 # expanded data with expand_tail_max_hours
  #r5long is the same as r5, but with more values per period of time
  r5long = matrix(0,length(r5), n_short_in_mediumEpoch) 
  r5long = replace(r5long, 1:length(r5long), r5)
  r5long = t(r5long)
  dim(r5long) = c(length(r5) * n_short_in_mediumEpoch,1)
  #------------------------------
  # detect which features have been calculated in part 1 and in what column they have ended up
  ENi = which(colnames(metashort) == "en")
  if (length(ENi) == 0) ENi = -1
  #==============================
  if (nrow(metashort) > length(r5long)) {
    metashort = metashort[1:length(r5long),]
  }
  wpd = 1440 * n_shortEpoch_permin #windows per day
  averageday = matrix(0, wpd, ncol(metashort) - 1)
  
  for (mi in 2:ncol(metashort)) {# generate 'average' day for each variable
    # The average day is used for imputation and defined relative to the starttime of the measurement
    # irrespective of dayborder as used in other parts of GGIR
    metr = as.numeric(as.matrix(metashort[, mi]))
    # turn all values of metr to NA if r5long is different to 0 (it now leaves the
    # expanded time with expand_tail_max out of the averageday calculation)
    is.na(metr[which(r5long != 0)]) = T 
    imp = matrix(NA, wpd, ceiling(length(metr) / wpd)) #matrix used for imputation of seconds
    ndays = ncol(imp) #number of days (rounded upwards)
    nvalidsec = matrix(0, wpd, 1)
    dcomplscore = length(which(r5 == 0)) / length(r5)
    if (ndays > 1 ) { # only do imputation if there is more than 1 day of data
      # all days except last one
      for (j in 1:(ndays - 1)) {
        imp[, j] = as.numeric(metr[(((j - 1) * wpd) + 1):(j * wpd)])
      }
      # last day
      lastday = metr[(((ndays - 1) * wpd) + 1):length(metr)]
      imp[1:length(lastday),ndays] = as.numeric(lastday)
      if (colnames(metashort)[mi] == "step_count") {
        # Median per row, which equals to median for one time point in the 'average' day
        # (median day would be a better term in this context)
        # chances are high that this will often be zero, because a person
        # would have to walk on a certain time point in the day for more than half of
        # each day in the to get a median above zero
        imp3 = apply(imp, 1, median, na.rm = TRUE)
      } else if (colnames(metashort)[mi] == "marker") {
        imp[is.na(imp)] <- 0 # to prevent imputation of marker data
        imp3 = apply(imp, 1, min, na.rm = TRUE)
      } else {
        # mean per row, which equals to mean or one time point in the 'average' day
        imp3 = rowMeans(imp, na.rm = TRUE)
      }
      dcomplscore = length(which(is.nan(imp3) == F | is.na(imp3) == F)) / length(imp3)
      
      if (length(imp3) < wpd)  {
        dcomplscore = dcomplscore * (length(imp3)/wpd)
      }
      if (ENi == mi) { #replace missing values for EN by 1
        imp3[which(is.nan(imp3) == T | is.na(imp3) == T)] = 1
      } else { #replace missing values for other metrics by 0
        # for those part of the data where there is no single data point for a 
        # certain part of the day (this is CRITICAL)
        imp3[which(is.nan(imp3) == T | is.na(imp3) == T)] = 0 
      }
      averageday[, (mi - 1)] = imp3
      for (j in 1:ndays) {
        missing = which(is.na(imp[,j]) == T)
        if (length(missing) > 0) {
          imp[missing,j] = imp3[missing]
        }
      }
      # imp is now the imputed time series
      dim(imp) = c(length(imp), 1)
      # but do not use imp for expanded time
      # do not impute the expanded time with expand_tail_max_hours
      toimpute = which(r5long != -1)
      #to cut off the latter part of the last day used as a dummy data
      metashort[toimpute, mi] = as.numeric(imp[toimpute]) 
    } else {
      dcomplscore = length(which(r5long == 0)) / wpd
    }
  }
  n_decimal_places = 4
  
  metashort[,2:ncol(metashort)] = round(metashort[,2:ncol(metashort)], digits = n_decimal_places)
  rout = data.frame(r1 = r1, r2 = r2, r3 = r3, r4 = r4, r5 = r5, stringsAsFactors = TRUE)
  invisible(list(metashort = metashort, rout = rout, r5long = r5long, dcomplscore = dcomplscore,
                 averageday = averageday, windowsizes = windowsizes, 
                 data_masking_strategy = data_masking_strategy,
                 LC = LC, LC2 = LC2, hrs.del.start = hrs.del.start, hrs.del.end = hrs.del.end,
                 maxdur = maxdur, nonwearHoursFiltered = nonwearHoursFiltered,
                 nonwearEventsFiltered = nonwearEventsFiltered))
}

library(GGIR)
context("g.impute")
test_that("g.impute returns a non-empty dataframe", {
  data(data.getmeta)
  data(data.inspectfile)
  data(data.calibrate)
  IMP = g.impute(M = data.getmeta, I = data.inspectfile)
  expect_equal(length(IMP), 14)
})

part6PairwiseAggregation = function(outputdir = NULL, desiredtz = "", verbose = TRUE) {
  #----------------------------------------------------------
  # declare internal functions only used in this function:
  remove_invalid = function(wx, N) {
    # wx are indices for wakeup times
    # N is length of time serieses
    valid = which(wx <= N - 15)
    if (length(valid) > 0) {
      wx = wx[valid]
    } else {
      wx = NULL
    }
    return(wx)
  }
  
  find_nearest_b_to_a = function(a, b) {
    # Example of how this works
    # a = c(1, 10, 20, 30)
    # b = c(2, 9, 17, 21, 28)
    # nearest_b_to_a = unlist(lapply(X = a,
    #                                FUN = find_nearest_b_to_a, b))
    # which should be: 2  9 21 28
    return(b[which.min(abs(a - b))])
  }
  
  
  calcSimilarity = function(x, y) {
    EqualState = c(x == y)
    testNA = c(is.na(x) | is.na(y))
    SummedValue = sum(ifelse(test = EqualState[which(testNA == FALSE)] == TRUE, yes = 1, no = 0))
    NValuesSkipped = length(which(testNA == TRUE))
    M = length(x)
    frac_valid = round((M - NValuesSkipped) / M, digits = 4)
    if (M != NValuesSkipped) {
      similarity = round(-100 + (200/(M - NValuesSkipped)) * SummedValue, digits = 3)
    } else {
      frac_valid = NA
      similarity = NA
    }
    invisible(list(similarity = similarity, frac_valid = frac_valid))
  }

  #----------------------------------------------------------
  
  if (verbose == TRUE) {
    cat("\n  Aggregate per pair:")
  }
  path_hh_ts = paste0(outputdir, "/part6HouseholdCoAnalysis/alignedTimeseries")
  if (!dir.exists(path_hh_ts)) stop("first process data with function align_individuals")
  
  compfolder =  paste0(outputdir, "/part6HouseholdCoAnalysis")

  fns = dir(path = path_hh_ts, full.names = TRUE)
  
  # loop over households
  pairsum_all_housholds = NULL
  if (length(fns) > 0) {
    for (i in 1:length(fns)) {
      # load entire data for household
      load(file = fns[i])
      uHID = unique(alignedTimeseries$HID)
      if (verbose == TRUE) cat(paste0("\n    Household: ", uHID))
      #unique member IDs
      uMID = gsub(pattern = "ACC.", replacement = "", x = grep(pattern = "ACC", x = colnames(alignedTimeseries), value = TRUE))
      #unique pair IDs
      uPID = gsub(pattern = "validpair_", replacement = "", x = grep(pattern = "validpair", x = colnames(alignedTimeseries), value = TRUE))
      alignedTimeseries$time_POSIX = as.POSIXct(alignedTimeseries$timenum, tz = desiredtz, origin = "1970-01-01")
      alignedTimeseries$date = as.character(as.Date(alignedTimeseries$time_POSIX))
      Npairs = length(uPID)
      pairsum_final = NULL
      if (length(uPID) > 0) {
        # Loop over pairs
        for (j in 1:length(uPID)) {
          # extrac MID for members in the pair
          pp = unlist(strsplit(uPID[j], "_"))
          m1 = pp[1]    
          m2 = pp[2]
          # simplify D to only columns of interest
          dataForPair = alignedTimeseries[, grep(pattern = paste0("time|date|HID|",m1,"|",m2), x = colnames(alignedTimeseries)), ] #guider|window|valid
          validpaircol = grep(pattern = uPID[j], x = colnames(dataForPair)) # column index where the indicator of valid/invalid can be found
          invalidepochs = which(dataForPair[,validpaircol] == FALSE)
          # Turn all data invalid
          if (length(invalidepochs) > 0) {
            is.na(dataForPair[invalidepochs, grep(pattern = "time|date|window|HID|guider|valid", x = colnames(dataForPair), invert = TRUE)]) = TRUE
          }
          # extract wake-up times in the entire recording
          index_wake1 = which(dataForPair[, paste0("wakeup.",m1)] == 1)
          index_wake2 = which(dataForPair[, paste0("wakeup.",m2)] == 1)
          index_onsetpair = rowSums(dataForPair[, paste0("onset.", c(m1, m2))])
          # ignore wake-up times that are  within 15 minutes before the end of the recording
          index_wake1 = remove_invalid(wx = index_wake1, N = length(dataForPair[,validpaircol]))
          index_wake2 = remove_invalid(wx = index_wake2, N = length(dataForPair[,validpaircol]))
          # pair nearest times
          N = length(index_wake1)
          do.pair.sum = FALSE
          if (N > 0 & length(index_wake2) > 0) {
            pw = data.frame(index_wake1 = index_wake1,
                            index_wake2 = unlist(lapply(X = index_wake1,
                                                        FUN = find_nearest_b_to_a, index_wake2)),
                            wakeup_time1 = character(N),
                            wakeup_time2 = character(N),
                            wakeup_firstMID = character(N))
            
            pairsum = pw[which(abs(pw[,2] - pw[, 1]) < 12 * 60), ]
            if (nrow(pairsum) > 0) do.pair.sum = TRUE
          } else {
            # No valid data, but still create empty row to 
            # communicate that there was no data
            pairsum = data.frame(index_wake1 = NA, index_wake2 = NA,
                                 wakeup_time1 = NA, wakeup_time2 = NA,
                                 wakeup_firstMID = NA)
          }
          if (do.pair.sum == TRUE) {
            pairsum$PID = uPID[j]
            pairsum$HID = uHID
            pairsum$Npairs = Npairs
            pairsum$event = 1:nrow(pairsum)
            pairsum$MID1 = m1
            pairsum$MID2 = m2
          } else {
            pairsum = data.frame(index_wake1 = NA, index_wake2 = NA,
                                 wakeup_time1 = NA, wakeup_time2 = NA,
                                 wakeup_firstMID = NA)
            pairsum[, c("PID", "HID",
                        "Npairs", "event",
                        "MID1", "MID2")] = NA
          }
          # Initialise wakeup variables
          pairsum[, c("wakeup_date1", "wakeup_date2",
                      "wakeup_acc_1_before_2_mg", "wakeup_acc_2_before_2_mg",
                      "wakeup_lux_1_before_2", "wakeup_lux_2_before_2",
                      "wakeup_secondMID", "wakeup_deltatime_min")] = NA
          if (do.pair.sum == TRUE) {
            pairsum$wakeup_date1 = as.Date(pairsum$wakeup_date1)
            pairsum$wakeup_date2 = as.Date(pairsum$wakeup_date2)
          }
          # initialise day variables
          pairsum[, c("day_r_acceleration", "day_ICC_active", "day_ICC_active_Fvalue",
                       "day_Kappa_active","day_Kappa_active_CI_lower", "day_Kappa_active_CI_upper",
                       "day_ActivitySimilarityIndex", "day_frac_valid", "day_duration_awake")] = NA
          
          # initialise night variables
          pairsum[, c("night_ICC_sleep", "night_ICC_sleep_Fvalue",
                      "night_Kappa_sleep", "night_Kappa_sleep_CI_lower",
                      "night_Kappa_sleep_CI_upper",
                      "night_SleepSimilarityIndex",
                      "night_frac_valid", "nigth_duration_night")] = NA
          
          # initialise sptPattern variables
          pairsum[, c("sptPattern_sameTime",
                      "sptPattern_M2thenM1", "sptPattern_M1thenM2",
                      "sptPattern_soloWakeM1", "sptPattern_soloWakeM2",
                      "sptPattern_sptduration", "sptPattern_fractionValid")] = 0
          if (do.pair.sum == TRUE) {
            # loop of wakeup pairs:
            for (h in 1:nrow(pairsum)) {
              # Assess who woke up first
              pairsum$wakeup_firstMID[h] = ifelse(pairsum$index_wake1[h] < pairsum$index_wake2[h], yes = m1,
                                                  no = ifelse(pairsum$index_wake1[h] > pairsum$index_wake2[h], yes = m2, no = "equal"))
              # Deduce who woke up second
              secondawake = ifelse(test = pairsum$wakeup_firstMID[h] == pairsum$MID1[h], yes = pairsum$MID2[h], no = pairsum$MID1[h])
              pairsum$wakeup_secondMID[h] = secondawake
              # Assess delta time of waking up
              pairsum$wakeup_deltatime_min[h] = difftime(time1 = dataForPair$time_POSIX[pairsum$index_wake2[h]], time2 = dataForPair$time_POSIX[pairsum$index_wake1[h]], units = "mins")
              # Extract dates of waking up
              pairsum$wakeup_date1[h] = as.Date(dataForPair$time_POSIX[pairsum$index_wake1[h]])
              pairsum$wakeup_date2[h] = as.Date(dataForPair$time_POSIX[pairsum$index_wake2[h]])
              
              pairsum$wakeup_time1[h] = format(dataForPair$time_POSIX[pairsum$index_wake1[h]], format = "%H:%M:%S")
              pairsum$wakeup_time2[h] = format(dataForPair$time_POSIX[pairsum$index_wake2[h]], format = "%H:%M:%S")
              
              index_last_wakeup = pairsum$index_wake1[h]
              if (pairsum$wakeup_firstMID[h] != "equal") {
                index_last_wakeup = max(pairsum[h,c("index_wake1", "index_wake2")])
                # Activity of person who first woke up during minute before second person wake up
                pairsum$wakeup_acc_1_before_2_mg[h] = mean(dataForPair[(index_last_wakeup - 3):index_last_wakeup,
                                                                    paste0("ACC.", pairsum$wakeup_firstMID[h])])
                # Activity of second person to wake up before they woke up
                pairsum$wakeup_acc_2_before_2_mg[h] = mean(dataForPair[(index_last_wakeup - 3):index_last_wakeup, 
                                                                    paste0("ACC.", secondawake)])
                # LUX of person who first woke up during minute before second person wake up
                pairsum$wakeup_lux_1_before_2[h] = max(dataForPair[(index_last_wakeup - 3):index_last_wakeup, 
                                                                   paste0("lightpeak.", pairsum$wakeup_firstMID[h])])
                # LUX of second person to wake up before they woke up
                pairsum$wakeup_lux_2_before_2[h] = max(dataForPair[(index_last_wakeup - 3):index_last_wakeup, 
                                                                   paste0("lightpeak.", secondawake)])
              } else if (pairsum$wakeup_firstMID[h] == "equal") {
                pairsum$wakeup_acc_1_before_2_mg[h] = NA
                pairsum$wakeup_acc_2_before_2_mg[h] = NA
                pairsum$wakeup_lux_1_before_2[h] = NA
                pairsum$wakeup_lux_2_before_2[h] = NA
              }
              #------------------------------------------------
              # Describe matching waking hours
              ind.wake = which(dataForPair$date == as.character(pairsum$wakeup_date1[h]) &
                                 dataForPair[, paste0("SleepPeriodTime.", m1)] == 0 &
                                 dataForPair[, paste0("SleepPeriodTime.", m2)] == 0)
              
              if (length(ind.wake) > 0) {
                ts_comp = dataForPair[ind.wake, 
                                      c(paste0("class_id.", c(m1, m2)),
                                        paste0("ACC.", c(m1, m2)))]
                if (length(which(rowSums(is.na(ts_comp[, 3:4])) == 0)) > 120) {
                  # Correlation between continuous acceleration values
                  pairsum$day_r_acceleration[h] = round(cor(x = ts_comp[, 3], y = ts_comp[,4]), digits = 3)
                  
                  # derive binary class of inactive / active
                  for (ja in 3:4) ts_comp[,ja] = ifelse(test = ts_comp[,ja] < 50, yes = 1, no = 0)
                  # ICC based on binary scores
                  IRR = irr::icc(ts_comp[, 3:4], model = "twoway", 
                                 type = "agreement", unit = "single")
                  pairsum$day_ICC_active[h] = round(IRR$value, digits = 3)
                  pairsum$day_ICC_active_Fvalue[h] = round(IRR$Fvalue, digits = 3)
                  # Cohen's Kappa based on binary scores
                  KP = psych::cohen.kappa(x = ts_comp[, 3:4])
                  KP$confid = round(KP$confid, digits = 3)
                  pairsum$day_Kappa_active[h] = round(KP$weighted.kappa, digits = 3)
                  pairsum$day_Kappa_active_CI_lower[h] = KP$confid[2,1]
                  pairsum$day_Kappa_active_CI_upper[h] = KP$confid[2,3]
                  # Similarity in binary scores
                  SIM = calcSimilarity(x = ts_comp[, 3], y = ts_comp[, 4])
                  pairsum$day_ActivitySimilarityIndex[h] = SIM$similarity
                  pairsum$day_frac_valid[h] = SIM$frac_valid
                  
                  pairsum$day_duration_awake = length(ind.wake)
                }
              }
              #------------------------------------------------
              # Describe noon-noon window of waking up in more detail
              ind.night = which(dataForPair$date == as.character(pairsum$wakeup_date1[h])) 
              # ignore first hour
              ind.night = ind.night - (12 * 60)
              ind.night = ind.night[which(ind.night > 0)]
              if (length(ind.night) > 0) {
                ts_comp = dataForPair[ind.night, c(paste0("class_id.", c(m1, m2)), paste0("ACC.", c(m1, m2)))]
                if (length(which(rowSums(is.na(ts_comp[,1:2])) == 0)) > 120) {
                  # derive binary class of sleep / wakefulness
                  for (ja in 1:2) ts_comp[,ja] = ifelse(test = ts_comp[,ja] < 1, yes = 1, no = 0)
                  # ICC based on binary scores
                  IRR = irr::icc(ts_comp[, 1:2], model = "twoway", 
                                 type = "agreement", unit = "single")
                  pairsum$night_ICC_sleep[h] = round(IRR$value, digits = 3)
                  pairsum$night_ICC_sleep_Fvalue[h] = round(IRR$Fvalue, digits = 3)
                  
                  # Cohen's Kappa based on binary scores
                  KP = psych::cohen.kappa(x = ts_comp[, 1:2])
                  KP$confid = round(KP$confid, digits = 3)
                  pairsum$night_Kappa_sleep[h] = round(KP$weighted.kappa, digits = 3)
                  pairsum$night_Kappa_sleep_CI_lower[h] = KP$confid[2,1]
                  pairsum$night_Kappa_sleep_CI_upper[h] = KP$confid[2,3]
                  # Similarity in binary scores
                  SIM = calcSimilarity(x = ts_comp[, 1], y = ts_comp[, 2])
                  pairsum$night_SleepSimilarityIndex[h] = SIM$similarity
                  pairsum$night_frac_valid[h] = SIM$frac_valid
                  pairsum$nigth_duration_night[h] = length(ind.night)
                }
              }
              # Describe wakefulness dynamics during the SPT prior to wakeup.
              onsetsi = which(index_onsetpair[1:pairsum$index_wake1[h]] > 0)
              if (length(onsetsi) > 0) {
                # indicies of the spt prior to the wake up
                ind.priorspt = (as.numeric(tail(onsetsi, n = 1)) + 1):pairsum$index_wake1[h]
                Npriorspt = length(ind.priorspt)
                # Fraction valid data
                fractionValid = sum(dataForPair[ind.priorspt,validpaircol], na.rm = TRUE) / Npriorspt
                pairsum$sptPattern_sptduration[h] = Npriorspt
                pairsum$sptPattern_fractionValid[h] = fractionValid
                Dtemp = dataForPair[ind.priorspt, ]
                # Identify wakeup times during the night
                sleep = Dtemp[, c(paste0("class_id.", m1), paste0("class_id.", m2))]
                sleep = ifelse(test = sleep >= 1, yes = 1, no = 0)
                for (nn in 1:(nrow(sleep) - 5)) {
                  # if at least one person wakes up
                  if (sum(sleep[nn, 1:2]) >= 1 & sum(sleep[nn - 1, 1:2]) == 0) {
                    if (sum(sleep[nn, 1:2]) == 2) {
                      # both wake up in the same minute
                      pairsum$sptPattern_sameTime[h] = pairsum$sptPattern_sameTime[h] + 1
                    } else {
                      next5min = rowSums(sleep[(nn + 1):(nn + 5), ])
                      if (any(next5min == 2)) {
                        # other person also wakes up
                        if (sleep[nn, 1] == 1) {
                          pairsum$sptPattern_M1thenM2[h] = pairsum$sptPattern_M1thenM2[h] + 1 # M1 first M2 second
                        } else {
                          pairsum$sptPattern_M2thenM1[h] = pairsum$sptPattern_M2thenM1[h] + 1 # M2 first M1 second
                        }
                      } else {
                        # other person does not wake up
                        if (sleep[nn, 1] == 1) {
                          pairsum$sptPattern_soloWakeM1[h] = pairsum$sptPattern_soloWakeM1[h] + 1 # M1 but not followed by M2
                        } else {
                          pairsum$sptPattern_soloWakeM2[h] = pairsum$sptPattern_soloWakeM2[h] + 1 # M2 but not followed by M1
                        }
                      }
                    }
                  }
                }
              }
            }
            # reorder columns
            colorder = c("HID", "Npairs", "PID", "MID1", "MID2", "event", "wakeup_time1", 
                         "wakeup_time2", "wakeup_date1", "wakeup_date2", "wakeup_firstMID", "wakeup_secondMID", 
                         "wakeup_deltatime_min", "wakeup_acc_1_before_2_mg", "wakeup_acc_2_before_2_mg",
                         "wakeup_lux_1_before_2", "wakeup_lux_2_before_2", grep(pattern = "sptPattern", x = names(pairsum), value = TRUE))
            pairsum = pairsum[, c(colorder, sort(colnames(pairsum)[which(colnames(pairsum) %in% colorder == FALSE)]))]
          } else {
            # no data for this pair, set the default zeros to missing
            is.na(pairsum[which(pairsum == 0)]) = TRUE
          }
          # remove columns we do not want in the ouput
          pairsum = pairsum[, which(colnames(pairsum) %in% c("index_wake1", "index_wake2") == FALSE)]
          # append to previous result
          pairsum_final = rbind(pairsum_final, pairsum)
          rm(pairsum)
        }
      }
      pairsum_all_housholds = rbind(pairsum_all_housholds, pairsum_final)
    }
    if (length(pairsum_all_housholds) > 0) {
      for (i in 1:ncol(pairsum_all_housholds)) {
        NA_NaN = which(is.na( pairsum_all_housholds[, i]))
        pairsum_all_housholds[NA_NaN, i] = ""
      }
      write.csv(x = pairsum_all_housholds, file = paste0(compfolder, "/pairwise_summary_all_housholds.csv"), row.names = FALSE)
    }
   }
}

g.calibrate = function(datafile, params_rawdata = c(),
                       params_general = c(),
                       params_cleaning = c(),
                       inspectfileobject = c(),
                       verbose = TRUE,
                       ...) {

  #get input variables
  input = list(...)
  if (length(input) > 0 ||
      length(params_rawdata) == 0 || length(params_general) == 0 || length(params_cleaning) == 0) {
    # Extract and check parameters if user provides more arguments than just the parameter arguments,
    # or if params_[...] aren't specified (so need to be filled with defaults).
    # So, inside GGIR this will not be used, but it is used when g.calibrate is used on its own
    # as if it was still the old g.calibrate function
    params = extract_params(params_rawdata = params_rawdata,
                            params_cleaning = params_cleaning,
                            params_general = params_general,
                            input = input,
                            params2check = c("rawdata", "cleaning", "general")) # load default parameters
    params_rawdata = params$params_rawdata
    params_cleaning = params$params_cleaning
    params_general = params$params_general
  }
  #-----------------
  # Define local functions:
  rollMean = function(x, fs, epochSize) {
    x = cumsum(c(0, x))
    select = seq(1, length(x), by = fs * epochSize)
    y = diff(x[round(select)]) / abs(diff(round(select)))
    return(y)
  }
  rollSD = function(x, sf, epochSize) {
    dim(x) = c(sf * epochSize, ceiling(length(x) / (sf * epochSize)))
    y = apply(x, 2, sd)
    return(y)
  }
  #-----------------
  use.temp = temp.available = TRUE
  
  filequality = data.frame(filetooshort = FALSE, filecorrupt = FALSE,
                           filedoesnotholdday = FALSE, stringsAsFactors = FALSE)
  calibEpochSize = 10 # epoch for recalibration as used in the 2014 paper
  blockResolution = 3600 # resolution at which raw data is loaded and processed (seconds)
  cal.error.start = cal.error.end = c()
  spheredata = c()
  tempoffset = c()
  npoints = c()
  PreviousEndPage = c() # needed for g.readaccfile
  scale = c(1,1,1)
  offset = c(0,0,0)
  bsc_qc = data.frame(time = c(), size = c(), stringsAsFactors = FALSE)
  #inspect file
  op <- options(stringsAsFactors = FALSE)
  on.exit(options(op))
  if (length(inspectfileobject) > 0) {
    INFI = inspectfileobject
  } else {
    stop("argument inspectfileobject not specified")
  }
  mon = INFI$monc
  if (mon == MONITOR$VERISENSE) mon = MONITOR$ACTIGRAPH
  dformat = INFI$dformc
  sf = INFI$sf
  
  if (is.null(sf)) {
    # If function g.inspectfile which produces the inspectfileobject
    # identifies a corrupt GT3X file then it sets the sf value to NULL
    # this is then used here to skip the calibration procedure
    return()
  }
  
  #creating matrices for storing output
  S = matrix(0,0,4) #dummy variable needed to cope with head-tailing succeeding blocks of data
  NR = ceiling((90*10^6) / (sf*calibEpochSize)) + 1000 #NR = number of rows to initialise features matrix with

  # setting size of blocks that are loaded (too low slows down the process)
  # the setting below loads blocks size of 12 hours (modify if causing memory problems)
  blocksize = round((14512 * (sf/50)) * (params_rawdata[["chunksize"]]*0.5))
  if (mon == MONITOR$MOVISENS) blocksize = (sf * 60 * 1440) / 2   #Around 12 hours of data for movisens
  if (mon == MONITOR$ACTIGRAPH && dformat == FORMAT$GT3X) blocksize = (12 * 3600) * params_rawdata[["chunksize"]]
  if (mon == MONITOR$AXIVITY && dformat == FORMAT$CWA) {
    if (utils::packageVersion("GGIRread") >= "0.3.1") {
      # CWA data blocks can have 40, 80 or 120 samples each; we'll take 80 as the average number
      blocksize = round(12 * 3600 * sf / 80 * params_rawdata[["chunksize"]])
    }
  }
  if (mon == MONITOR$PARMAY_MTX && dformat == FORMAT$BIN) {
    # Matrix data packets have ~2 minutes of data
    blocksize = round(60 * 12 / 2 * params_rawdata[["chunksize"]])
  }
  #===============================================
  # Read file
  i = 1 #counter to keep track of which binary block is being read
  count = 1 #counter to keep track of the number of seconds that have been read
  LD = 2 #dummy variable used to identify end of file and to make the process stop
  isLastBlock = FALSE # dummy variable part of "end of loop mechanism"
  header = NULL
  while (LD > 1) {
    if (verbose == TRUE) {
      if (i == 1) {
        cat(paste("\nLoading chunk: ",i,sep=""))
      } else {
        cat(paste(" ",i,sep=""))
      }
    }
    #try to read data blocks based on monitor type and data format
    options(warn=-1) #turn off warnings (code complains about unequal rowlengths
    accread = g.readaccfile(filename = datafile, blocksize = blocksize, blocknumber = i,
                            filequality = filequality, ws = blockResolution,
                            PreviousEndPage = PreviousEndPage, inspectfileobject = INFI,
                            params_rawdata = params_rawdata, params_general = params_general,
                            header = header)
    header = accread$header
    isLastBlock = accread$isLastBlock
    PreviousEndPage = accread$endpage

    if (i == 1) {
      use.temp = temp.available = ("temperature" %in% colnames(accread$P$data))
      if (use.temp) {
        features = matrix(99999, NR, 8)
      } else {
        features = matrix(99999, NR, 7)
      }
    }

    options(warn = 0) #turn on warnings
    #process data as read from binary file
    if (length(accread$P) > 0) { # would have been set to zero if file was corrupt or empty
      data = as.matrix(accread$P$data[,which(colnames(accread$P$data) %in% c("x", "y", "z", "time", "temperature"))]) # convert to matrix b/c rbind is much faster on matrices
      if (exists("accread")) {
        rm(accread)
      }
      # add leftover data from last time
      if (min(dim(S)) > 1) {
        data = rbind(S,data)
      }
      # remove 0s if ActiGraph csv (idle sleep mode) OR if similar imputation done in ad-hoc csv
      # current ActiGraph csv's are not with zeros but with last observation carried forward
      zeros = c()
      if (!(mon == MONITOR$ACTIGRAPH && dformat == FORMAT$CSV)) {
        zeros = which(data[, "x"] == 0 & data[, "y"] == 0 & data[, "z"] == 0)
      }
      if ((mon == MONITOR$ACTIGRAPH && dformat == FORMAT$CSV) || length(zeros) > 0) {
        data = g.imputeTimegaps(x = as.data.frame(data), sf = sf, impute = FALSE)
        data = as.matrix(data$x)
      }
      LD = nrow(data)
      #store data that could not be used for this block, but will be added to next block
      use = (floor(LD / (blockResolution*sf))) * (blockResolution*sf) #number of datapoint to use
      if (length(use) > 0) {
        if (use > 0) {
          if (use != LD) {
            S = data[(use + 1):LD,] # store leftover data

            # if S is only one row, it gets coersed to a vector, and what we need is a 1-row matrix
            if (is.vector(S)) {
              S = t(S)
            }
          }
          data = data[1:use,]
          LD = nrow(data) #redefine LD because there is less data

          Gx = data[, "x"]
          Gy = data[, "y"]
          Gz = data[, "z"]

          if (use.temp) {
            if (mean(data[1:10, "temperature"], na.rm = TRUE) > 120) {
              warning("\ntemperature ignored for auto-calibration because values are too high\n")
              use.temp = FALSE
            } else if (sd(data[, "temperature"], na.rm = TRUE) < 0.01) {
              warning("\ntemperature ignored for auto-calibration because no variance in values\n")
              use.temp = FALSE
            }
          }
          #=============================================
          #expand 'out' if it is expected to be too short
          if (count > (nrow(features) - (2.5 * (3600/calibEpochSize) * 24))) {
            extension = matrix(99999, ((3600/calibEpochSize) * 24), ncol(features))
            features = rbind(features, extension)
          }
          # mean acceleration for EN, x, y, and z
          EN = sqrt(Gx^2 + Gy^2 + Gz^2)
          EN2 = rollMean(EN, sf, calibEpochSize)
          endCount = count - 1 + length(EN2)
          features[count:endCount, 1] = EN2
          features[count:endCount, 2] = rollMean(Gx, sf, calibEpochSize)
          features[count:endCount, 3] = rollMean(Gy, sf, calibEpochSize)
          features[count:endCount, 4] = rollMean(Gz, sf, calibEpochSize)
          # sd acceleration
          features[count:endCount, 5] = rollSD(Gx, sf, calibEpochSize)
          features[count:endCount, 6] = rollSD(Gy, sf, calibEpochSize)
          features[count:endCount, 7] = rollSD(Gz, sf, calibEpochSize)
          if (use.temp == TRUE) {
            features[count:endCount, 8] = rollMean(data[, "temperature"], sf, calibEpochSize)
          }
          count = endCount + 1 # increasing count, the indicator of how many timesteps (calibEpochSize) have been read
          rm(Gx); rm(Gy); rm(Gz)
          # Update blocksize depending on available memory:
          BlocksizeNew = updateBlocksize(blocksize = blocksize, bsc_qc = bsc_qc)
          bsc_qc = BlocksizeNew$bsc_qc
          blocksize = BlocksizeNew$blocksize
        }
        #--------------------------------------------
      }
    } else {
      LD = 0 #once LD < 1 the analysis stops, so this is a trick to stop it
    }
    spherepopulated = 0
    if (isLastBlock) {
      LD = 0
    }
    features_temp = data.frame(V = features, stringsAsFactors = FALSE)
    # remove 999999, missing values
    cut = which(features_temp[,1] == 99999 |
                  is.na(features_temp[,4]) == TRUE | is.na(features_temp[,1]) == TRUE)
    if (length(cut) > 0) {
      features_temp = features_temp[-cut,]
    }
    # remove duplicates, e.g. caused by periods of nonwear
    if (nrow(features_temp) > 0) {
      cut = which(rowSums(features_temp[1:(nrow(features_temp) - 1), 2:7] == features_temp[2:nrow(features_temp), 2:7]) == 3)
      if (length(cut) > 0) {
        features_temp = features_temp[-cut,]
      }
    }
    nhoursused = 0
    if (nrow(features_temp) > (params_rawdata[["minloadcrit"]] - 21)) {  # enough data for the sphere?
      nhoursused = (nrow(features_temp) * 10) / 3600
      features_temp = features_temp[-1,]
      #select parts with no movement
      if (mon == MONITOR$AD_HOC) {
        if (length(params_rawdata[["rmc.noise"]]) == 0) {
          warning("Argument rmc.noise not specified, please specify expected noise level in g-units")
        }
        sdcriter = params_rawdata[["rmc.noise"]] * 1.2
        if (length(params_rawdata[["rmc.noise"]]) == 0) {
          stop(paste0("Please provide noise level for the acceleration sensors",
                      " in g-units with argument rmc.noise to aid non-wear detection"),
               call. = FALSE)
        }
      } else {
        sdcriter = 0.013
      }
      nomovement = which(features_temp[,5] < sdcriter & features_temp[,6] < sdcriter & features_temp[,7] < sdcriter &
                           abs(as.numeric(features_temp[,2])) < 2 & abs(as.numeric(features_temp[,3])) < 2 &
                           abs(as.numeric(features_temp[,4])) < 2) #the latter three are to reduce chance of including clipping periods
      if (length(nomovement) < 10) {
        # take only one row to trigger that autocalibration is skipped 
        # with the QCmessage that there is not enough data
        features_temp = features_temp[1, ] 
      } else {
        features_temp = features_temp[nomovement,]
      }
      if (min(dim(features_temp)) > 1) {
        npoints = nrow(features_temp)
        cal.error.start = sqrt(as.numeric(features_temp[,2])^2 + as.numeric(features_temp[,3])^2 + as.numeric(features_temp[,4])^2)
        cal.error.start = round(mean(abs(cal.error.start - 1)), digits = 5)
        #check whether sphere is well populated
        tel = 0
        for (axis in 2:4) {
          if ( min(features_temp[,axis]) < -params_rawdata[["spherecrit"]] & max(features_temp[,axis]) > params_rawdata[["spherecrit"]]) {
            tel = tel + 1
          }
        }
        if (tel == 3) {
          spherepopulated = 1
        } else {
          spherepopulated = 0
          QC = "recalibration not done because not enough points on all sides of the sphere"
        }
      } else {
        QC = "recalibration not done because no non-movement data available"
        features_temp = c()
      }
    } else {
      QC = "recalibration not done because not enough data in the file or because file is corrupt"
    }
    if (spherepopulated == 1) { #only try to improve calibration if there are enough datapoints around the sphere
      #---------------------------------------------------------------------------
      # START of Zhou Fang's code (slightly edited by vtv21 to use matrix features_temp from above
      # instead the similar matrix generated by Zhou Fang's original code. This to allow for
      # more data to be used as features_temp can now be based on 10 or more days of raw data
      input = features_temp[,2:4]
      if (use.temp == TRUE) {
        inputtemp = cbind(as.numeric(features_temp[,8]),as.numeric(features_temp[,8]),as.numeric(features_temp[,8])) #temperature
      } else {
        inputtemp = matrix(0, nrow(input), ncol(input)) #temperature, here used as a dummy variable
      }
      meantemp = mean(as.numeric(inputtemp[, 1]), na.rm = TRUE)
      inputtemp = inputtemp - meantemp
      offset = rep(0, ncol(input))
      scale = rep(1, ncol(input))
      tempoffset = rep(0, ncol(input))
      weights = rep(1, nrow(input))
      res = Inf
      maxiter = 1000
      tol = 1e-10
      for (iter in 1:maxiter) {
        curr = c()
        try(expr = {curr = scale(input, center = -offset, scale = 1/scale) +
          scale(inputtemp, center = F, scale = 1/tempoffset)}, silent = TRUE)
        if (length(curr) == 0) {
          # set coefficients to default, because it did not work.
          break
        }
        closestpoint = curr / sqrt(rowSums(curr^2))
        k = 1
        offsetch = rep(0, ncol(input))
        scalech = rep(1, ncol(input))
        toffch = rep(0, ncol(inputtemp))
        for (k in 1:ncol(input)) {
          #-----------------------------------------------------------------
          # Next few lines added 23 on april 2015 to deal with NaN values in
          # some of the sphere data for Actigraph monitor brand
          # Expanded on 17-Sep-2017 for the generic data format too
          if ((mon == MONITOR$ACTIGRAPH || mon == MONITOR$MOVISENS) && length(which(is.na(closestpoint[,k, drop = F]) == TRUE)) > 0 &
              length(which(is.na(closestpoint[,k, drop = F]) == FALSE)) > 10) { #needed for some Actigraph data
            invi = which(is.na(closestpoint[,k, drop = F]) == TRUE)
            closestpoint = closestpoint[-invi,]
            curr = curr[-invi,]
            inputtemp = inputtemp[-invi,]
            input = input[-invi,]
            weights = weights[-invi]
          }
          #------------------------------------------------
          fobj = lm.wfit(cbind(1, curr[,k],inputtemp[,k]) , closestpoint[,k, drop = F], w = weights)
          offsetch[k] = fobj$coef[1]
          scalech[k] = fobj$coef[2]
          if (use.temp == TRUE) {
            toffch[k] = fobj$coeff[3]
          }
          curr[,k] = fobj$fitted.values
        }
        offset = offset + offsetch / (scale  * scalech)
        if (use.temp == TRUE) {
          tempoffset = tempoffset * scalech + toffch
        }
        scale = scale * scalech
        res = c(res,  3 * mean(weights*(curr - closestpoint)^2 / sum(weights)))
        weights = pmin(1 / sqrt(rowSums((curr - closestpoint)^2)), 1 / 0.01)
        if (abs(res[iter + 1] - res[iter]) < tol)  break
      }
      if (use.temp == FALSE) {
        features_temp2 = scale(as.matrix(features_temp[,2:4]),center = -offset, scale = 1/scale)
      } else {
        yy = as.matrix(cbind(as.numeric(features_temp[,8]),as.numeric(features_temp[,8]),as.numeric(features_temp[,8])))
        features_temp2 = scale(as.matrix(features_temp[,2:4]),center = -offset, scale = 1/scale) +
          scale(yy, center = rep(meantemp,3), scale = 1/tempoffset)
      }     #equals: D2[,axis] = (D[,axis] + offset[axis]) / (1/scale[axis])
      # END of Zhou Fang's code
      #-------------------------------------------
      cal.error.end = sqrt(features_temp2[,1]^2 + features_temp2[,2]^2 + features_temp2[,3]^2)
      rm(features_temp2)
      cal.error.end = round(mean(abs(cal.error.end - 1)), digits = 5)
      # assess whether calibration error has sufficiently been improved
      if (cal.error.end < cal.error.start & cal.error.end < 0.01 & nhoursused > params_rawdata[["minloadcrit"]]) { #do not change scaling if there is no evidence that calibration improves
        if (use.temp == temp.available) {
          QC = "recalibration done, no problems detected"
        } else {
          QC = "recalibration done, but temperature values not used"
        }
        LD = 0 #stop loading
      } else {
        #continue loading data
        if (nhoursused > params_rawdata[["minloadcrit"]]) {
          if (verbose == TRUE) {
            cat(paste0("\nnew calibration error: ",cal.error.end, " g"))
            cat(paste0("\nnpoints around sphere: ", npoints))
          }
        }
        QC = "recalibration attempted with all available data, but possibly not good enough: Check calibration error variable to varify this"
      }
    }
    i = i + 1 #go to next block (12 hours-isch)
  }
  if (length(cal.error.end) > 0) {
    if (cal.error.end > cal.error.start) {
      QC = "recalibration not done because recalibration does not decrease error"
    }
  }
  if (!is.null(features_temp) && all(dim(features_temp)) != 0) {
    spheredata = features_temp
    if (use.temp == TRUE) {
      names(spheredata) = c("Euclidean Norm","meanx","meany","meanz","sdx","sdy","sdz","temperature")
    } else {
      names(spheredata) = c("Euclidean Norm","meanx","meany","meanz","sdx","sdy","sdz")
    }
  } else {
    spheredata = c()
  }
  rm(features_temp)
  QCmessage = QC
  if (params_rawdata[["printsummary"]] == TRUE & verbose == TRUE) {
    cat("\nSummary of autocalibration procedure:")
    cat("\n")
    cat(paste0("\nStatus: ",QCmessage))
    cat(paste0("\nCalibration error (g) before: ", cal.error.start))
    cat(paste0("\nCalibration error (g) after: ", cal.error.end))
    cat(paste0("\nOffset correction ",c("x","y","z"),": ", offset))
    cat(paste0("\nScale correction ",c("x","y","z"),": ", scale))
    cat(paste0("\nNumber of hours used: ",nhoursused))
    cat(paste0("\nNumber of 10 second windows around the sphere: ", npoints))
    cat(paste0("\nTemperature used (if available): ", use.temp))
    cat(paste0("\nTemperature offset (if temperature is available) ",
               c("x", "y", "z"),": ", tempoffset))
    cat("\n")
  }
  if (use.temp == TRUE && length(spheredata) > 0) {
      meantempcal = mean(spheredata[,8], na.rm = TRUE)
  } else {
    meantempcal = c()
  }
  invisible(list(scale = scale, offset = offset, tempoffset = tempoffset,
                 cal.error.start = cal.error.start, cal.error.end = cal.error.end,
                 spheredata = spheredata, npoints = npoints, nhoursused = nhoursused,
                 QCmessage = QCmessage, use.temp = use.temp, meantempcal = meantempcal, bsc_qc = bsc_qc))
}

check_log = function(log, dateformat, colid = 1, datecols = c(), 
                     logPath, logtype) {
  
  # at the moment this is used only for activity log and study dates log
  dateformat_param = ifelse(logtype == "activity log", "qwindow_dateformat", "study_dates_dateformat")
  
  # assume ID to be in first column
  log = log[which(log[, colid] != ""),] # ignore rows for which there is no id
  
  # replace cells with only a dot by empty character
  # the documentation clearly states that no dots should be used
  # to indicate missing value, but this is just in case users still forget about it
  for (jn in 1:ncol(log)) {
    dotcells = which(log[,jn] == ".")
    if (length(dotcells) > 0) {
      # is.na(actlog_vec[dotcells]) = TRUE
      log[dotcells, jn] = ""
    }
  }
  # Check dates
  # extract example date value
  if (is.null(datecols)) {
    datecols = grep(pattern = "date|Date|DATE",  x = colnames(log), value = FALSE)
  }
  if (length(datecols) > 0) {
    exampledates = unlist(log[,datecols])
    exampledates = exampledates[which(!is.na(exampledates))]
    
    e1 = strsplit(exampledates, split = "/|-")
    e1 = e1[lengths(e1) > 0L]
    date_consistency_check = t(as.data.frame(e1))
    numrange = apply(X = date_consistency_check, MARGIN = 2, FUN = function(x) diff(range(sapply(x, nchar))))
    if (any(numrange > 1)) {
      stop(paste0("\nDifferent date formats were encountered in your ", 
                  logtype, ": ", logPath,". ",
                  "Please note that this may not be visible when you open the ", logtype,
                  " in MS Excel, but it should be visible in a plain text editor. ",
                  "To resolve this update the format for all date cells in the ",
                  "log to be consistent and in line with the value of argument",
                  dateformat_param, ", which is currently: ", dateformat), call. = FALSE)
    }
  } else {
    exampledates = c()
  }
  
  Nhyphen = length(grep(pattern = "-", x = log[, datecols]))
  Ndash = length(grep(pattern = "/", x = log[, datecols]))
  if (Nhyphen > 0 & Ndash > 0) {
    warning(paste0("\nYou may be mixing slash and hyphen separated dates ",
                   "inside the same ", logtype , ", which could lead to errors. ",
                   "Please fix."),
            call. = FALSE)
  }
  # Replace / by - if that is what is expected dateformat
  if (Ndash > 0 & length(grep(pattern = "/", x = dateformat)) == 0) {
    for (rc in 1:ncol(log)) {
      log[, rc] = gsub(pattern = "/", replacement = "-", x = log[, rc])
    }
  }
  # Replace - by / if that is what is expected dateformat
  if (Nhyphen > 0 & length(grep(pattern = "-", x = dateformat)) == 0) {
    for (rc in 1:ncol(log)) {
      log[, rc] = gsub(pattern = "-", replacement = "/", x = log[, rc])
    }
  }
  # return
  return(log)
}

g.part3_alignIndexVectors = function(x, y, a, b, N) {
  # where N is length of time series
  # and x, y, a and b are index vectors
  # Assumptions:
  # - x and y are same length
  # - for all values y > x
  # This function ensures that:
  # - length of a and b match length of x and y
  # - for all values x < a < b < y
  #-------------------------------------
  # start of a is missing
  if (a[1] > y[1]) {
    a =  c(1, a)
  }
  # start of b is missing
  if (b[1] > y[1]) {
    b =  c(1, b)
  }
  #-------------------------------------
  # end of a is missing and end of y equals N
  if (a[length(a)] < x[length(x)] && y[length(y)] == N) {
    a =  c(a, N)
  }
  # end of b is missing and end of y equals N
  if (b[length(b)] < x[length(x)] && y[length(y)] == N) {
    b =  c(b, N)
  }
  #-------------------------------------
  # end of a is larger than end of y
  if (a[length(a)] > y[length(y)]) {
    a =  a[1:(length(a) - 1)]
  }
  # end of b is larger than end of y
  if (b[length(b)] > y[length(y)]) {
    b =  b[1:(length(b) - 1)]
  }
  #-------------------------------------
  # a is longer than y and x
  if (length(a) > length(y) &&
      length(a) > length(x)) {
    a =  a[1:(length(a) - 1)]
  }
  # b is longer than y and x
  if (length(b) > length(y) &&
      length(b) > length(x)) {
    b =  b[1:(length(b) - 1)]
  }
  # #-------------------------------------
  # # a is longer than y but the end of y 
  # if (length(a) > length(y) &&
  #     a[length(a)] == y[length(y)]) {
  #   a =  a[1:(length(a) - 1)]
  # }
  # if (length(b) > length(y) &&
  #     b[length(b)] == y[length(y)]) {
  #   b =  b[1:(length(b) - 1)]
  # }
  # 
  invisible(list(
    x = x,
    y = y,
    a = a,
    b = b
  ))
}

library(GGIR)
context("g.part3_correct_guider")
test_that("Align index vectors", {
  skip_on_cran()
  N = 280
  x = c(1, 100, 200)
  y = c(80, 180, 280)
  for (i in 1:11) {
    a = c(20, 120, 220)
    b = c(60, 160, 260)
    if (i == 1) {
      # no corrections expected
    } else if (i == 2) {
      # a one longer beyond N
      a = c(20, 120, 220, 320)
    } else if (i == 3) {
      # a one longer under N
      a = c(20, 120, 220, 275)
    } else if (i == 4) {
      # a and b one longer beyond N
      a = c(20, 120, 220, 320)
      b = c(60, 160, 260, 360)
    } else if (i == 5) {
      # b one longer equal to N
      b = c(60, 160, 260, N)
    } else if (i == 6) {
      # a one longer equal to N
      a = c(20, 120, 220, N)
    } else if (i == 7) {
      # a and b with missing start
      # a one longer equal to N
      a = c(120, 220, N)
      b = c(160, 260)
    } else if (i == 8) {
      # a and b with missing start
      a = c(120, 220)
      b = c(160, 260)
    } else if (i == 9) {
      # a with missing start
      a = c(120, 220)
    } else if (i == 10) {
      # a and b with missing end
      a = c(20, 120)
      b = c(60, 160)
    } else if (i == 11) {
      # b with missing start
      b = c(20, 120)
    }
    test = g.part3_alignIndexVectors(x, y, a, b, N)
    expect_equal(length(test$x), 3)
    expect_equal(length(test$a), 3)
    expect_equal(length(test$b), 3)
    expect_equal(length(test$y), 3)
    testthat::expect_true(all(test$x <= test$a))
    testthat::expect_true(all(test$a <= test$b))
    testthat::expect_true(all(test$b <= test$y))
  }
})

g.conv.actlog = function(qwindow, qwindow_dateformat = "%d-%m-%Y", epochSize = 5) {
  # Function to read activity log and convert it into data.frame
  # that has for each ID and date a different qwindow vector
  # local functions:
  time2numeric = function(x) {
    x = unlist(x)
    x = x[grep(pattern = "impute|imputa|uncertain", x = names(x), invert = TRUE)]
    c2t = function(x2) {
      tmp = as.numeric(unlist(strsplit(as.character(x2),":")))
      if (length(tmp) == 2) hourinday = tmp[1] + (tmp[2]/60)
      if (length(tmp) == 3) hourinday = tmp[1] + (tmp[2]/60) + (tmp[3]/60/60)
      return(hourinday)
    }
    out = as.numeric(sapply(x,c2t))
    return(out)
  }
  extract_names = function(x) {
    tmp = names(unlist(x))
    tmp2 = gsub(pattern = "[.].*|_|-",replacement = "",x = tmp)
    return(tmp2)
  }
  #-------------------------
  # main function code
  
  # Read content of activity diary file
  actlog = data.table::fread(file = qwindow, data.table = FALSE, colClasses = "character")
  
  # check ID and date cols in actlog
  actlog = check_log(actlog, dateformat = qwindow_dateformat,
                     colid = 1, datecols = NULL, 
                     logPath = qwindow, logtype = "activity log")
  
  # find dates
  actlog_vec = unlist(actlog) # turn into vector
  actlog_vec = sapply(actlog_vec, FUN = function(x) !all(is.na(as.Date(as.character(x),
                                                                 format = qwindow_dateformat))))
  exampledates = unlist(actlog)[which(actlog_vec == TRUE)]
  Ndates = length(which(actlog_vec == TRUE))
  dim(actlog_vec) = c(nrow(actlog), ncol(actlog))
  # create new qwindow object to archive all extracted information
  qwindow = data.frame(ID = rep(0, Ndates), date =  rep("", Ndates))
  qwindow$qwindow_times = qwindow$qwindow_values = qwindow$qwindow_names = c()
  cnt = 1
  processLogRow = function(actlog, qwindow_dateformat = qwindow_dateformat) {
    actlog_vec = unlist(actlog) # turn into vector
    actlog_vec = sapply(actlog_vec, FUN = function(x) !all(is.na(as.Date(as.character(x),
                                                                         format = qwindow_dateformat))))
    datei = which(actlog_vec == TRUE)
    Ndays = length(datei)
    out = data.frame(ID = character(Ndays), date = character(Ndays))
    out$qwindow_times = out$qwindow_values = qwindow$qwindow_names = NULL
    if (Ndays > 0) {
      out$ID[1:Ndays]  = rep(actlog[1], Ndays)
      out$date[1:Ndays]  = as.character(as.Date(as.character(actlog[datei]), format = qwindow_dateformat))
      # datei are the indices of the date columns
      # add number to be able to identify last class after last date
      datei = c(datei, max(which(actlog != "")) + 1)
      for (j in 1:(length(datei) - 1)) { # loop over days within the recording
        # Note: qindow is a data.frame and the timestamps for each day will be stored
        # as a list object in a single cell of qwindow. The same applies ot the corresponding
        # labels.
        k = cnt + j - 1
        if ((datei[j + 1] - datei[j]) >= 2) {
          actlog_tmp = actlog[(datei[j] + 1):(datei[j + 1] - 1)]
          actlog_tmp = actlog_tmp[which(is.na(actlog_tmp) == FALSE & actlog_tmp != "")]
          out$qwindow_times[k] = list(actlog_tmp) # list of times for that day
          # make sure that seconds are multiple of epochSize
          seconds = data.table::second(strptime(out$qwindow_times[[k]], format = "%H:%M:%S"))
          to_next_epoch = which(seconds %% epochSize != 0)
          if (length(to_next_epoch) > 0) {
            timechar = out$qwindow_times[[k]][to_next_epoch]
            time = strptime(timechar, format = "%H:%M:%S")
            seconds2sum = epochSize - seconds[to_next_epoch] %% epochSize
            time = time + seconds2sum
            timechar_new = as.character(time)
            time_new = unlist(lapply(timechar_new, FUN = function(x) strsplit(x, " ", fixed = T)[[1]][2]))
            out$qwindow_times[[k]][to_next_epoch] = time_new
          }
          out$qwindow_values[k] = list(time2numeric(out$qwindow_times[k]))
          out$qwindow_names[k] = list(extract_names(out$qwindow_times[k]))
          unlisted_qv = unlist(out$qwindow_values[k])
          if (length(which(is.na(unlisted_qv) == FALSE)) > 0) {
            unlisted_qn = unlist(out$qwindow_names[k])
            unlisted_qt = unlist(out$qwindow_times[k])
            if (min(unlisted_qv, na.rm = TRUE) > 0) {
              out$qwindow_values[k] = list(c(0, unlisted_qv))
              out$qwindow_times[k] = list(c("00:00", unlisted_qt))
              out$qwindow_names[k] = list(c("daystart", unlisted_qn))
            }
            if (max(unlisted_qv, na.rm = TRUE) < 24) {
              out$qwindow_values[k] = list(c(unlist(out$qwindow_values[k]), 24))
              out$qwindow_times[k] = list(c(unlist(out$qwindow_times[k]), "24:00:00"))
              out$qwindow_names[k] = list(c(unlist(out$qwindow_names[k]),"dayend"))
            }
          }
        } else {
          out$qwindow_values[k] = out$qwindow_names[k] = out$qwindow_times[k] = list("")
        }
      }
    }
    return(out)
  }
  test = apply(X = actlog, MARGIN = 1, FUN = processLogRow, qwindow_dateformat)
  qwindow = do.call(what = rbind, args = test)
  # When testing the code it seemed sometimes not to recognise the date.
  # The following lines should help to catch any errors people may encounter.
  if (is.na(as.Date(qwindow$date[1], format = "%y-%m-%d")) == FALSE) {
    qwindow$date =  as.Date(qwindow$date, format = "%y-%m-%d")
  } else {
    qwindow$date =  as.Date(qwindow$date)
  }
  if (is.na(qwindow$date[1]) == TRUE | !is(qwindow$date[1], "Date")) {
    if (length(exampledates) > 0) {
      warning(paste0("\n Date not recognised in activity diary. We expect format ", 
                     qwindow_dateformat, " because that is what you specified in ",
                     "argument qwindow_dateformat, but we see ",
                     paste0(head(exampledates), collapse = " "), 
                     ". You need to update the qwindow_dateformat argument, and check",
                     " that dates are in a consistent format."))
    } else {
      warning("\n Date not recognised in activity diary")
    }
  }
  return(qwindow)
}

library(GGIR)
context("Convert activity log")
test_that("g.conv.actlog loads activity log and puts it in data.frame", {
  
  # With slash as separator
  actlog = data.frame(id = c("1RAW", "2RAW"),
                      date = c("15/08/2022", "20/08/2022"),
                      work = c("7:45:30", "7:45:00"),
                      travelhome = c("17:00:00", "17:00:00"),
                      home = c("17:30:00", "18:00:00"),
                      date = c("16/08/2022", "21/08/2022"),
                      work = c("7:45:15", "7:45:00"),
                      travelhome = c("17:00:00", "17:00:00"),
                      home = c("17:30:00", "18:00:00"),
                      sport = c("", "19:00:00"),
                      home = c("", "20:00:00"))
  fn = "testactlog.csv"
  write.csv(x = actlog, file = fn, row.names = FALSE)
  LOG = g.conv.actlog(qwindow = fn, qwindow_dateformat = "%d/%m/%Y")
  
  expect_true(is.list(LOG$qwindow_times))
  expect_true(is.list(LOG$qwindow_values))
  expect_true(is.list(LOG$qwindow_names))
  expect_equal(length(LOG$qwindow_times[[1]]), 5)
  expect_equal(length(LOG$qwindow_times[[4]]), 7)
  expect_equal(length(LOG$qwindow_values[[1]]), 5)
  expect_equal(length(LOG$qwindow_values[[2]]), 5)
  expect_equal(length(LOG$qwindow_names[[3]]), 5)
  expect_equal(length(LOG$qwindow_names[[4]]), 7)
  expect_equal(LOG$qwindow_names[[3]], c("daystart", "work", "travelhome", "home", "dayend"))
  expect_equal(LOG$qwindow_values[[2]], c(0.00, 7.754167, 17.00, 17.50, 24.00), tolerance = 3)
  expect_equal(as.character(LOG$qwindow_times[[1]]), c("00:00", "7:45:30", "17:00:00", "17:30:00", "24:00:00"))
  
  # make sure it scales seconds when no compatible with epoch size
  actlog = data.frame(id = c("1RAW", "2RAW"),
                      date = c("15/08/2022", "20/08/2022"),
                      work = c("7:45:32", "7:45:14"),
                      travelhome = c("17:00:04", "17:00:44"),
                      home = c("17:30:31", "18:00:03"),
                      date = c("16/08/2022", "21/08/2022"),
                      work = c("7:45:14", "7:45:34"),
                      travelhome = c("17:00:12", "17:00:53"),
                      home = c("17:30:22", "18:00:34"),
                      sport = c("", "19:00:00"),
                      home = c("", "20:00:00"))
  fn = "testactlog.csv"
  write.csv(x = actlog, file = fn, row.names = FALSE)
  LOG = g.conv.actlog(qwindow = fn, qwindow_dateformat = "%d/%m/%Y", epochSize = 5)
  seconds = lapply(LOG$qwindow_times, FUN = function(x) substr(x, nchar(x), nchar(x)))
  seconds = as.numeric(unlist(seconds))
  expect_true(sum(seconds %in% c(0, 5)) == length(seconds))
  expect_equal(as.character(LOG$qwindow_times[[1]]), c("00:00", "07:45:35", "17:00:05", "17:30:35", "24:00:00"))
  
  expect_true(file.exists(fn))
  if (file.exists(fn)) file.remove(fn)
  
  # With hyphen as separator
  actlog = data.frame(id = c("1RAW", "2RAW"),
                      date = c("15-08-2022", "20-08-2022"),
                      work = c("7:45:00", "7:45:00"),
                      travelhome = c("17:00:00", "17:00:00"),
                      home = c("17:30:00", "18:00:00"),
                      date = c("16-08-2022", "21-08-2022"),
                      work = c("7:45:00", "7:45:00"),
                      travelhome = c("17:00:00", "17:00:00"),
                      home = c("17:30:00", "18:00:00"),
                      sport = c("", "19:00:00"),
                      home = c("", "20:00:00"))
  fn = "testactlog.csv"
  write.csv(x = actlog, file = fn, row.names = FALSE)
  LOG = g.conv.actlog(qwindow = fn, qwindow_dateformat = "%d-%m-%Y")
  
  expect_true(is.list(LOG$qwindow_times))
  expect_true(is.list(LOG$qwindow_values))
  expect_true(is.list(LOG$qwindow_names))
  expect_equal(length(LOG$qwindow_times[[1]]), 5)
  expect_equal(length(LOG$qwindow_times[[4]]), 7)
  expect_equal(length(LOG$qwindow_values[[1]]), 5)
  expect_equal(length(LOG$qwindow_values[[2]]), 5)
  expect_equal(length(LOG$qwindow_names[[3]]), 5)
  expect_equal(length(LOG$qwindow_names[[4]]), 7)
  expect_equal(LOG$qwindow_names[[3]], c("daystart", "work", "travelhome", "home", "dayend"))
  expect_equal(LOG$qwindow_values[[2]], c(0.00, 7.75, 17.00, 17.50, 24.00))
  expect_equal(as.character(LOG$qwindow_times[[1]]), c("00:00", "7:45:00", "17:00:00", "17:30:00", "24:00:00"))
  
  expect_true(file.exists(fn))
  if (file.exists(fn)) file.remove(fn)
})

# NOTE ABOUT DOCUMENTATION:
# GGIR does not use Roxygen. The documentation below is not used. 
# All function documentation can be found in the man/*.Rd files.
# Please edit documentation there.
#
#- @title Estimated self-similarity parameter
#- 
#- @description This function estimates the self-similarity parameter (SSP), also known as scaling exponent or alpha.  
#- @usage alpha_hat(data,scale = 2^(1/8),box_size = 4,m=1) 
#- @param data Univariate time series (must be a vector or data frame)
#- @param scale Specifies the ratio between successive box sizes (by default scale = 2^(1/8))
#- @param box_size Vector of box sizes (must be used in conjunction with scale = "F")
#- @param m An integer of the polynomial order for the detrending (by default m=1)
#-
#- @details The DFA fluctuation can be computed in a geometric scale or for different choices of boxes sizes.
#-
#- @return Estimated alpha is a real number between zero and two.
#- 
#- @note It is not possible estimating alpha for multiple time series at once. 
#-
#- @author Ian Meneghel Danilevicz and Victor Mesquita 
#- 
#- @references C.-K. Peng, S.V. Buldyrev, S. Havlin, M. Simons, H.E. Stanley, A.L. Goldberger Phys. Rev. E, 49 (1994), p. 1685
#- Mesquita, Victor & Filho, Florencio & Rodrigues, Paulo. (2020). Detection of crossover points in detrended fluctuation analysis: An application to EEG signals of patients with epilepsy. Bioinformatics. 10.1093/bioinformatics/btaa955. 
#-
#- @examples 
#- # Estimate self-similarity of a very known time series available on R base: the sunspot.year.
#- # Then the spend time with each method is compared.
#- 
#- SSP(sunspot.year, scale = 2)
#- SSP(sunspot.year, scale = 1.2)

SSP = function(data, scale = 2^(1/8), box_size = 4, m = 1){
  if (inherits(x = data, "data.frame")) {
    data = data[, 1]
  }
  if (length(data) <= box_size || any(is.na(data))) {
    alpha_hat = NA
  } else {
    dfa_hat = DFA(data, scale = scale, box_size = box_size, m = m)    
    est_ols = lm(log(dfa_hat[,2]) ~ log(dfa_hat[,1]))
    alpha_hat = est_ols$coefficients[[2]]    
  }
  return(alpha_hat)
} 

---
title: "Step and Cadence Analysis"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Step and Cadence Analysis}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

GGIR does not have a build-in step detection algorithm, but does facilitate the [embedding of an external function](https://wadpac.github.io/GGIR/articles/ExternalFunction.html) for event detection such as steps. Such algorithms typically count steps per epoch, where an epoch is usually a 5 second or larger time interval. Steps and cadence per epoch alone is not of direct value for research. Therefore, GGIR facilitates the extraction of summary statistics, which will be discussed below.


## Verisense algorithm

The only external algorithm for step detection currently available for GGIR is the Verisense algorithm, which was design for data collected on the wrist. The algorithm has been described and evaluated in studies by [Maylor 2022](https://doi.org/10.3390/s22249984) and [Rowlands 2022](https://doi.org/10.1080/02640414.2022.2147134).

The original [Verisense algortihm code](https://github.com/ShimmerResearch/Verisense-Toolbox/tree/master/Verisense_step_algorithm) is not actively maintained at the time of writing this documentation. An [improved copy of the Verisense algortihms code](https://github.com/wadpac/GGIR/blob/master/user-scripts/verisense_count_steps.R) with minor bug fixes is part of the GGIR GitHub repository.


### How to use the Verisense algorithm?

To use the algorithm, copy-paste the following code to the top of your R script and update the file path in the first line to the Verisense function file on your computer. A discussion of this code will follow further down.

```
source("C:/path_to_file/verisense_count_steps.R")
myfun = list(FUN = verisense_count_steps,
             parameters = c(4, 4, 20, -1.0, 4, 4, 0.01, 1.25),
             expected_sample_rate = 15,
             expected_unit = "g",
             colnames = "step_count",
             outputres = 1,
             minlength = 1,
             outputtype = "numeric",
             aggfunction = sum, # aggregate step by taking sum
             timestamp = FALSE,
             reporttype = "event",
             ilevels = c(0, 100, 250), # acceleration levels to be used
             clevels = c(0, 80, 100, 120), # cadence levels to be used
             qlevels = c(0.5, 0.9), # quantiles to be used
             ebout.dur = c(1, 5, 10), # event bout duration to extract
             ebout.th.cad = 30, # event bout threshold for cadence
             ebout.th.acc = 50, # event bout threshold for acceleration
             ebout.criter = 0.8, # event bout criteria (same as boutcriter)
             ebout.condition = "AND") # event bout logic (see below)
   
```

Note the that `parameters = c(4, 4, 20, -1.0, 4, 4, 0.01, 1.25)` is based on Rowlands et al. "Stepping up with GGIR" from 2022.

Next, use GGIR as you normally do, but include `myfun = myfun` as input argument. If we would rely on all GGIR default parameter values and segments days in 6 equal parts then this would look as follows:

```
GGIR(myfun = myfun,
     outputdir = "C:/myresults",
     datadir = "C:/mydata",
     qwindow = c(0, 4, 8, 12, 16, 20, 24))
```


## Time resolution for deriving statistics

### Per day or groups of days

All step and cadence statistics are derived as summary per recording day and per recording as a whole, and stored in **results/part2_dayeventsummary.csv** and **results/part2_eventsummary.csv**, respectively. The column name extensions **AD_**, **WD_**, **WE_**, **WWD_**, and **WWE_** inside the file **part2_eventsummary.csv** have the same meaning as elsewhere in the [GGIR part2 output](https://wadpac.github.io/GGIR/articles/chapter7_DescribingDataWithoutKnowingSleep.html?q=WWD#related-output), e.g. AD means aggregation across all recording days.

### Per day segment

When you use GGIR's [day segment analysis functionality](https://wadpac.github.io/GGIR/articles/TutorialDaySegmentAnalyses.html) the statistics are also derived per time segment within a day. The time segment for which each statistic is derived is clarified in the ending of a column name as **\_startHour-endHourhr**. For example, **\_0-14hr** indicates that it was extracted from data between midnight and 14:00.


## Step count and cadence summary statistics

Total step count is stored with **tot_step_count** in the column name. Cadence is derived in the unit 'steps per minute', even if the epoch size for step detection is not 1 minute. Cadence is abbreviated as **cad** in the column name. Mean cadence is indicated as **mn_cad**.

Note that [Chapter 6](https://wadpac.github.io/GGIR/articles/chapter6_DataImputation.html) discusses how acceleration metrics are imputed based on a mean of valid time points on other days of the recording. For steps and cadence this would be problematic as the average between walking and non-walking is not informative. Therefore, imputation of steps is done by the median instead and by that represents typical stepping behaviour at that time of the day.


### Stratified by acceleration and/or cadence level

#### Per acceleration level

As discussed in [Chapter 7](https://wadpac.github.io/GGIR/articles/chapter7_DescribingDataWithoutKnowingSleep.html) accelerometer data can be described in terms of acceleration levels. As such, we can also describe step count and cadence per acceleration level.

To control this aspect, include in your `myfun` object the item `ilevels` as a numeric vector with the acceleration values to define the acceleration levels, e.g. `ilevels= c(0, 50)`.

Output:

- Mean cadence for the acceleration range 0-50 m_g_ when working with acceleration metric `ENMO` will be stored with column name **mn_cad_acc0-50mg_ENMO**. Note: The step bout detection functionality automatically uses the acceleration metrics specified by the user. So, if you want to try it out with both ENMO, ENMOa, and MAD metric then that should work.

- Total step count for the acceleration range 0-50m_g_ when working with acceleration metric `ENMO` will be stored with column name **tot_step_count_acc0-50mg_ENMO**. Here **step_count** is the value as provided via `myfun` parameter `colnames`.


#### Per cadence level with absolute thresholds

Similarly, we can also describe the data based on cadence level and use absolute cadence thresholds. 

To control this aspect, include in your `myfun` object item `clevels` as a numeric vector with the cadence values that define the cadence levels, e.g. `clevels = c(0, 30, 120)`.

Output:

- Total step count for the cadence range 0-30spm will be stored with column name **tot_step_count_cad0-30spm**.
- Time spent in the cadence range 0-30spm will be stored with column name **dur_cad0-30spm**.
- Mean acceleration in the cadence range 0-30spm will be stored with column name **mn_ENMO_cad0-30spm**

#### Per cadence level with relative thresholds

Instead of setting absolute thresholds to define cadence levels we can use percentiles.

To control this aspect, include in your `myfun` object the item `qlevels` as a numeric vector with fractions of 1 to define the quantiles, e.g. `qlevels = c(0.5, 0.9)`.

Output:

- Cadence value for the 50th percentile of the cadence distribution, e.g. `cad_p50` for `qlevels = 0.5` and `cad_p90` for `qlevels = 0.9`.


#### Per most and least active consecutive cadence time window

The same statistics can also be derived from the consecutive X hours with the least and most cadence per (segment of a) day.

To control this aspect, include in your `myfun` object the item `winhr`, e.g. `winhr = c(5, 10)` which works the same as [`winhr` in GGIR itself](https://wadpac.github.io/GGIR/articles/GGIRParameters.html#winhr).

Output:

- The X hour window with the least amount of cadence will be stored with column name **LX_cad_**, e.g. `L10_cad_`.
- The X hour window with the most amount of cadence will be stored with column name **MX_cad_**, e.g. `M5_cad_`.

This is then followed by:

- **mean_Y** where Y reflects the metrics being average, which can be acceleration metrics, step count per epoch or the mean cadence. For example, "L10_cad_mean_ENMO_mg_0-24hr", "L10_cad_mean_step_count_mg_0-24hr", "L10_cad_meancad_0-24hr".
- **L10hr_** indicates the timing of the 10 hours with the least cadence.

### Walking bouts

Similar to the detection of physical activity bouts as discussed in [Chapter 11](https://wadpac.github.io/GGIR/articles/chapter11_DescribingDataCutPoints.html) we can detect bouts of stepping behaviour. In GGIR we define these based on duration, minimum cadence AND/OR minimum acceleration and the fraction of the bout for which these criteria need to be met.

The step bout detection functionality automatically uses all acceleration metrics specified by the user. So, if you want to try it out with both ENMO, ENMOa, and MAD metric then that should work.

To control the criteria for bout detection, include in your `myfun` object the items:

- `ebout.dur` a numeric vector of length 1 or larger with the minimum bout duration(s) of interest, e.g. `ebout.dur = c(1, 5, 10)`.
- `ebout.th.cad` a numeric vector of length 1 or larger being the minimum cadence value in steps per minute, e.g. `ebout.th.cad = 30`.
- `ebout.th.acc` a numeric vector of length 1 or larger being the minimum acceleration value in steps per minute, e.g. `ebout.th.acc = 50`.
- `ebout.criter` a single number as a fraction of 1 being the faction of a bout for which the inclusion criteria need to be met identical to `boutcriter` in the context of physical activity bouts, e.g. `ebout.criter = 0.8`.
- `ebout.condition` whether cadence and acceleration condition are both required (`ebout.condition = "AND"`) or either of the conditions is required (`ebout.condition = "OR"`). If both cadence and acc need to meet a thresholds then fill in `"AND"`. If it also acceptable if only one of the thresholds is met then fill in `"OR"`. If you do not want cadence or acceleration to be used in the equation then simply set `ebout.th.cad = 0` or `ebout.th.acc = 0`.

Output:

Columns names related to detected bouts start with the word **Bout_**, e.g. "Bout_meandur_E5S_B10M80%_cadT30_AND_accT50_ENMO", "Bout_number_E5S_B5M80%_cadT30_AND_accT50_ENMO" and "Bout_totdur_E5S_B5M80%_cadT30_AND_accT50_ENMO".

- Average bout duration is indicated with **Bout_meandur_** in the column name.
- The total duration is indicated with **Bout_totdur_** in the column name.
- The number of bouts is indicated with **Bout_number_** in the column name.
- Underlying epoch size is indicated with **E5S** in the column name where **5S** refers to a 5sec epochs.
- Minimum bout duration is indicated with **B10M** in the column name where **10M** would refer to a 10 minutes.
- The minimum fraction for which the bout conditions need to be met is indicated as a percentage  in the column name, e.g. **80%**.
- The minimum cadence and acceleration values are presented in the column name as **cadT30** and **accT50_ENMO** when using a 30 and 50 thresholds, respectively. 
- The setting of parameter **ebout.condition** is directly copied into the variable name as **AND** or **OR**.

---
title: "Contributing"
output:
   html_document:
    toc : true
    number_sections: true
    toc_depth: 3
    toc_float: true #by turning this on, the table of contents moves to the left of the page.
urlcolor: blue
vignette: >
  %\VignetteIndexEntry{Contributing}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
editor_options:
  markdown:
    wrap: 72
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
```


## If you have coding skills…

We welcome contributions in the development, maintenance, and documentation of GGIR. Please find the GGIR’s contributing guidelines [here]( https://github.com/wadpac/GGIR/blob/master/CONTRIBUTING.md).

## If you do not have coding skills….

You might not have the coding skills to contribute to the code base of GGIR, and still your contribution could be very important for us. For example: 

-	**Doing research on algorithms** is crucial to support the development of GGIR. We welcome contributions on any aspect in this regard.
-	**Reporting any issue you might have** in the google group dedicated to GGIR community. Our answers to your issues might also help other GGIR users. 
-	**Proofreading the GGIR documentation** and informing us if you miss any essential aspect or you found it difficult to follow any of the instructions provided.

---
title: "9. Circadian rhythms"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{9. Circadian rhythms}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

Work in progress...

## MXLX metrics (consecutive hours, instead of cumulative hours)

## Cosinor and extended cosinor analysis

## Intradaily variability and interdaily stability

## Key arguments

## Related output

---
title: "Embedding external functions in GGIR"
output:
   html_document:
    toc : true
    number_sections: true
    toc_depth: 3
    toc_float: true #by turning this on, the table of contents moves to the left of the page.
urlcolor: blue
vignette: >
  %\VignetteIndexEntry{Embedding external functions in GGIR}
  \usepackage[utf8]{inputenc}
  %\VignetteEngine{knitr::rmarkdown}
header-includes:
  - \usepackage{titling}
  - \pretitle{\begin{center}
    \includegraphics[]{GGIR-MASTERLOGO-RGB.png}\LARGE\\}
  - \posttitle{\end{center}}
---


```{r setup, include=FALSE}
knitr::opts_chunk$set(
  echo = TRUE,
  python.reticulate = FALSE
)
```

**NOTE: If you are viewing this page via CRAN note that the main GGIR documentation has been migrated to the [GGIR GitHub pages](https://wadpac.github.io/GGIR/).**

# Introduction

If you like GGIR but want to use algorithms for raw data not included in GGIR then the external function embedding feature can be the solution. For example, you may want to pilot a new machine learned classifiction algorithm but you do not want to write all the data cleaning and aggregation steps needed for analysis of real life 'out of the lab' acceleormeter data.

**How it works:**

Internally GGIR loads the raw accelerometer data in memory blocks of about 24 hours. When the data is in memory, corrected for calibration error, and resampled to the sample rate required by your function, GGIR applies its own default algorithms as well as the external function provided by you (Python or R). The external function is expected to take as input: A three-column matrix with the acceleration data corresponding to the three acceleration axes, and an optional parameters argument which can be of any R format (character, list, vector, data.frame, etc). As output your external function is expected to produce a matrix or data.frame with one or multiple columns corresponding to the output of your external function.

# Example with external R function

In this example we will apply the function counts() from R package [activityCounts](https://CRAN.R-project.org/package=activityCounts) to the raw data, which produces an estimate of Actigraph counts per second.

## Write external function

Create file **calculateCounts.R** and insert the following code:

```{R,eval=FALSE}
calculateCounts = function(data=c(), parameters=c()) {
  # data: 3 column matrix with acc data
  # parameters: the sample rate of data
  library("activityCounts")
  if (ncol(data) == 4) data = data[,2:4]
  mycounts = counts(data = data, hertz = parameters, 
                    x_axis = 1, y_axis = 2, z_axis = 3,
                    start_time = Sys.time())
  mycounts = mycounts[,2:4] #Note: do not provide timestamps to GGIR
  return(mycounts)
}
```

## Provide external function to GGIR

Create a new .R file for running the GGIR analysis, e.g. named `myscript.R`, and insert the following code.
Do not forget to update the filepath on the first line to point to your `calculateCounts.R` file.

```{R,eval=FALSE}
source("~/calculateCounts.R")
myfun =  list(FUN = calculateCounts,
              parameters = 30,
              expected_sample_rate = 30,
              expected_unit = "g",
              colnames = c("countsX","countsY","countsZ"),
              outputres = 1,
              minlength = 1,
              outputtype = "numeric",
              aggfunction = sum,
              timestamp = F,
              reporttype = c("scalar", "scalar", "scalar"))
```

The above code creates object `myfun` of type `list` which is expected to come with the following elements:

- `FUN` A character string specifying the location of the external function you want to apply.
- `parameters` The parameters used by the function, which can be stored in any format (vector, matrix, list, data.frame). The user should make sure that the external function can handle this object.
- `expected_sample_rate` Expected sample rate, if the inputdata has a difference sample rate, then the data will be resampled.
- `expected_unit` Expected unit of the acceleration by external function: "mg", "g" or "ms2". If input data is different it will be converted.
- `colnames` Character vector with the names of the columns produced by the external function.
- `outputres` The resolution (seconds) of the output produced by the external function. Note, that this needs to be equal to or a multitude of the short epoch size of the g.part1 output (5 seconds) or the short epoch size should be a multitude of this resolution. In this way GGIR can aggregate or repeat the external function output to be used inside GGIR.
- `minlength` The minimum length (seconds) of input data needed, typically the window per which output is provided.
- `outputtype` Character to indicate the type of external function output. Set to "numeric" if data is stored in numbers (any numeric format), or "character" if it is a character string.
- `aggfunction` If the data needs to be aggregated to match the short epoch size of the g.part1 output (5 seconds) then this element specifies what
function should be used for the aggregation, e.g. mean, sum, median.
- `timestamp` Boolean to indicated whether timestamps  (seconds since 1-1-1970) should be passed on to the external 
    function as first columm of the data matrix..
- `reporttype` Character to indicate the type of reporting by GGIR: "scalar" if it should be averaged per day, "event" if it should be summed per day, or "type" if it is categorical variable that can only be aggregated per day by tabulating it. For `reporttype` "event" GGIR will as of release 3.2-0 not summarise the events in part2_summary.csv and part2_daysummary.csv but in part2_eventsummary.csv and part2_dayeventsummary.csv.
- `name` Character to indicate the name of the entity being detected in the cases of event detection, this name is used in the output filename, e.g. part2_stepsummary.csv when name is "step". At the moment name is only use for event detection algorithms

Next, add a call to GGIR function GGIR with `myfun` provided as one of its arguments:

```{R,eval=FALSE}
library(GGIR)
GGIR(datadir = "~/myaccelerometerdata",
             outputdir = "~/myresults",
             mode = 1:2,
             epochvalues2csv = TRUE,
             do.report = 2,
             myfun = myfun) #<= this is where object myfun is provided to GGIR
```

Please see the [\underline{general GGIR vignette}](https://cran.r-project.org/package=GGIR/vignettes/GGIR.html) for more information about function GGIR.

# Example with external Python function

In this example we will use an external Python function to estimate the dominant signal frequency per acceleration axis. Note this can also be done in R, but it shows that even Python functions can be provided. 

## Write external function

Create **dominant_frequency.py** and insert the code shown below:

```{python, eval=FALSE}
import numpy

def dominant_frequency(x, sf):
  # x: vector with data values
  # sf: sample frequency
  fourier = numpy.fft.fft(x)
  frequencies = numpy.fft.fftfreq(len(x), 1/sf)
  magnitudes = abs(fourier[numpy.where(frequencies > 0)])
  peak_frequency = frequencies[numpy.argmax(magnitudes)]
  return peak_frequency
```

Create **dominant_frequency.R** that calls the python function and insert the following code:

```{R,eval=FALSE}
dominant_frequency = function(data=c(), parameters=c()) {
  # data: 3 column matrix with acc data
  # parameters: the sample rate of data
  source_python("dominant_frequency.py")
  sf=parameters
  N = nrow(data)
  ws = 5 # windowsize
  if (ncol(data) == 4) data= data[,2:4]
  data = data.frame(t = floor(seq(0,(N - 1)/sf, by = 1/sf)/ws),
                    x = data[,1], y = data[,2], z = data[,3])
  df = aggregate(data, by = list(data$t), 
                 FUN=function(x) {return(dominant_frequency(x, sf))})
  df = df[, -c(1:2)]
  return(df)
}
}
```

## Provide external function to GGIR

Create a new .R file for running the GGIR analysis, e.g. named myscript.R, and insert the following blocks of code.

Specification of Python environment to use, this can also be a conda environment or docker container (see documentation R package [\underline{reticulate}](https://rstudio.github.io/reticulate/) for further details). Make sure that that this Python environment has all the required dependencies for the external function, here we will only need `numpy`.

```{R,eval=FALSE}
  library("reticulate")
  use_virtualenv("~/myvenv", required = TRUE) # Local Python environment
  py_install("numpy", pip = TRUE)

```

Specify a `myfun` object as explained in the R example. Do not forget to update the filepath to the `"~/dominant_frequency.R"` file.

```{R,eval=FALSE}
source("~/dominant_frequency.R")
myfun =  list(FUN = dominant_frequency,
              parameters = 30,
              expected_sample_rate = 30,
              expected_unit = "g",
              colnames = c("domfreqX", "domfreqY", "domfreqZ"),
              minlength = 5,
              outputres = 5,
              outputtype = "numeric",
              aggfunction = median
              timestamp = F,
              reporttype = c("scalar", "scalar", "scalar"))
```

Add a call to function GGIR where `myfun` is provided as argument.
Note that, do.parallel is set to FALSE. Unfortunately Python embedding with R package reticulate and multi-threading with R package foreach as used in GGIR do not combine well.

```{R,eval=FALSE}
library(GGIR)
GGIR(datadir = "~/myaccelerometerdata",
             outputdir = "~/myresults",
             mode = 1:2,
             epochvalues2csv = TRUE,
             do.report = 2,
             myfun = myfun,
             do.parallel = FALSE)
```


# Integration in GGIR output

## Part 1
The external function output is included in the time series produced by function GGIR function g.part1 and stored in an RData-file in `/output_nameofstudy/meta/basic`.
The resolution of these output in GGIR is set by GGIR argument `windowsizes`, which is `c(5,900,3600)` by default. Here, the first element `5` specifies the short epoch size in seconds.
If the output of the external function is less then this resolution it will be aggregated with the function as specificied by aggfunction in the `myfun` object. In the count example we used the sum for this and for the dominant frequency example we used the median.

## Part 2
Next, in part2 GGIR aims to detect non-wear periods and imputes those. The impute time series can be found in the part 2 milestone data in folder: `/output_nameofstudy/meta/ms2.out`. If you want these to be directly stored in a csv file then set argument `epochvalues2csv = TRUE`.



# External functions released by GGIR collaborators:

- Wrist-based step detection algorithm: https://github.com/ShimmerResearch/Verisense-Toolbox/tree/master/Verisense_step_algorithm

which is not actively maintained, for the latest version of the function see:
https://github.com/wadpac/GGIR/blob/master/user-scripts/verisense_count_steps.R

- Wrist-based sleep classification as described by Sundararajan et al. 2021 [link to paper](https://www.nature.com/articles/s41598-020-79217-x), and corresponding code is here: https://github.com/wadpac/Sundararajan-SleepClassification-2021/tree/master/ggir_ext

---
title: "Published cut-points and how to use them in GGIR"
output:
   html_document:
    toc : true
    number_sections: true
    toc_depth: 3
    toc_float: true #by turning this on, the table of contents moves to the left of the page.
urlcolor: blue
vignette: >
  %\VignetteIndexEntry{Published cut-points and how to use them in GGIR}
  \usepackage[utf8]{inputenc}
  %\VignetteEngine{knitr::rmarkdown}
---

**NOTE: If you are viewing this page via CRAN note that the main GGIR documentation has been migrated to the [GGIR GitHub pages](https://wadpac.github.io/GGIR/).**

# Considerations

The physical activity research field has used so called cut-points to segment
accelerometer time series based on level of intensity. In this vignette we have 
compiled a list of published cut-points with instructions on how to use them with GGIR.
Please note that GGIR refers to cut-points as thresholds, but we are referring to the
same thing: A value or a set of values to help split levels of movement intensity.
As newer cut-points are frequently published the list below may not be up to date.
**Please let us know if you are aware of any published cut-points that we missed!**

## Cut-points expressed in gravitational units (this vignette)

This vignette focuses on cut-points for metrics that attempt to quantify average 
acceleration per epoch in gravitational units. The strength of these metrics is 
that their values are not affected by sampling rate and epoch length improving 
comparability across studies.

## Cut-points NOT expressed in gravitational units (not in this vignette)

However, GGIR also facilitates some metrics whose values are not expressed in 
gravitational units that were historically used. For example, the metric as 
described by Neishabouri (see GGIR argument `do.neishabouricounts`) which reflects
the indicator of accumulated body movement over time, referred to as counts, 
calculated by the ActiLife software from the ActiGraph accelerometer brand.
Cut-points for counts corresponding to the ActiGraph brand have been recurrently 
proposed in the literature, for example, see this [systematic review](https://doi.org/10.1007%2Fs40279-017-0716-0)
with a stratification by age group. Note that cut-points for ActiGraph counts proposed before
the introduction of multiday raw data collection are most likely
hardware-based calculations which may not perfectly align with ActiGraph
software-based (Actilife) calculations of counts that Neishabouri described.
As a result, older cut-points may need to be used with caution.

The cut-points you find in the literature for ActiGraph counts cannot be applied 
to Neishabouri counts directly because both are epoch length specific. The cut-points 
from the literature need to be corrected by a conversion factor. The conversion 
factor is calculated as the epoch length in the new study (e.g. 5 seconds) divided 
by the epoch length in the original study (e.g. 60 seconds). Note that no correction 
for differences in sampling rate is needed because Neishabouri counts already account 
for this via down-sampling.

If we would want to use cut-point "100 counts per minute" from the literature on 5
second epoch data, the GGIR function call would look like this:

```
GGIR([...],
     mode = 1:5,
     windowsizes = c(5, 900, 3600),
     do.neishabouricounts = TRUE,
     acc.metric = "NeishabouriCount_y",
     threshold.in = 100 * (5/60),
     [...])
```

As a side note - Although counts are epoch length specific it is possible to convert them into gravitational units by dividing the count value by the sampling rate as used by the count algorithm (30 Hertz for ActiGraph counts) multiplied by the number of seconds in the time window for which the count value was derived. For example, 100 counts per minute equates to an average acceleration per minute of 100 / (30 x 60) = 0.055 _g_, and 1950 counts per minute equals 1.0833 _g_. Note that the absolute acceleration values from different acceleration metrics are not comparable. However, expressing data in _g_-units can at least ease evaluating the validity of metrics relative to  criterion acceleration and removes the non-informative connection to sampling rate (van Hees et al. 2013 https://doi.org/10.1371/journal.pone.0061691). As this calculation shows, count-units and _g_-unit metrics only differ by a constant, by which there is no practical difference in their ability to explain variance in human behaviour. Similarly, if you would like to convert an average daily acceleration of 0.05 _g_ back to counts per day then you would have to multiply by (30 Hertz x 60 seconds x 1440 minutes), which gives 129600 counts per day.

# Relevant arguments to use cut-points in GGIR

The argument `mvpathreshold` is used in **part 2** to quantify the time
accumulated over a user-specified threshold over which the
moderate-to-vigorous intensity is expected to occur. The `mvpathreshold`
is applied over all the metrics extracted in part 1 with the arguments
*do.metric* (e.g., `do.enmo`, `do.mad`, `do.neishabouricounts`).

In **part 5**, `threshold.lig`, `threshold.mod`, and `threshold.vig` are
used to indicate the thresholds to separate inactivity from light, light
from moderate, and moderate from vigorous, respectively.These thresholds
are applied over the metric defined with `acc.metric` (default =
\"ENMO\"). Here a summary table for the parameters definition to calculate
some of the acceleration metrics that has been previously used for the
calibration of cut-points and how to define them to be used in the physical 
activity intensity classification with cut-points.

+--------------+-------------------------------+-------------------------------------+
| Metric       | To derive metric              | Define metric for cut-points        |
+==============+===============================+=====================================+
| ENMO         | `do.enmo = TRUE`              | `acc.metric = "ENMO"`               |
+--------------+-------------------------------+-------------------------------------+
| ENMOa        | `do.enmoa = TRUE`             | `acc.metric = "ENMOa"`              |
+--------------+-------------------------------+-------------------------------------+
| LFENMO       | `do.lfenmo = TRUE`            | `acc.metric = "LFENMO"`             |
+--------------+-------------------------------+-------------------------------------+
| MAD          | `do.mad = TRUE`               | `acc.metric = "MAD"`                |
+--------------+-------------------------------+-------------------------------------+
| Neishabouri\ | `do.neishabouricounts = TRUE` | `acc.metric = "NeishabouriCount_x"` |
| counts       |                               | `acc.metric = "NeishabouriCount_y"` |
|              |                               | `acc.metric = "NeishabouriCount_z"` |
|              |                               | `acc.metric = "NeishabouriCount_vm"`|
+--------------+-------------------------------+-------------------------------------+


# Summary of published cut-points

## Cut-points for preschoolers

+---------------+--------------------+-------------+------------------------+------------------+
| Cut-points    | Device\            | Age         | Relevant arguments     | thresholds       |
|               | Attachment site    |             |                        |                  |
+===============+:===================+:============+:=======================+:=================+
| Roscoe 2017\* | GENEActiv\         | 4-5 yr      | `do.enmoa = TRUE`\     | Light: 61.8\     |
|               | Non-dominant wrist |             | `do.enmo = FALSE`\     | Moderate: 100.4\ |
|               |                    |             | `acc.metric = "ENMOa"` | Vigorous: N/A    |
+---------------+--------------------+-------------+------------------------+------------------+
| Roscoe 2017\* | GENEActiv\         | 4-5 yr      | `do.enmoa = TRUE`\     | Light: 94.5\     |
|               | Dominant wrist     |             | `do.enmo = FALSE`\     | Moderate: 108.5\ |
|               |                    |             | `acc.metric = "ENMOa"` | Vigorous: N/A    |
+---------------+--------------------+-------------+------------------------+------------------+

\*These publications used acceleration metrics that sum their values per epoch rather than average them per epoch like GGIR does. So, to use their cut-point in GGIR, we provide a scaled version of the cut-points presented in the paper as: `(CutPointFromPaper_in_gsecs/85.7) * 1000`. Note that sample frequency of 87.5 as reported in the publication was incorrect and based on correspondence with authors we replaced this by 85.7.

## Cut-points for children/adolescents

+------------------+--------------------+----------+--------------------------------+------------------+
| Cut-points       | Device\            | Age      | Relevant arguments             | thresholds       |
|                  | Attachment site    |          |                                |                  |
+:=================+:===================+:=========+:===============================+:=================+
| Phillips 2013\*  | GENEA\             | 8-14 yr  | `do.enmoa = TRUE`\             | Light: 87.5\     |
|                  | Left wrist         |          | `do.enmo = FALSE`\             | Moderate: 250\   |
|                  |                    |          | `acc.metric = "ENMOa"`         | Vigorous: 750    |
+------------------+--------------------+----------+--------------------------------+------------------+
| Phillips 2013\*  | GENEA\             | 8-14 yr  | `do.enmoa = TRUE`\             | Light: 75\       |
|                  | Right wrist        |          | `do.enmo = FALSE`\             | Moderate: 275\   |
|                  |                    |          | `acc.metric = "ENMOa"`         | Vigorous: 700    |
+------------------+--------------------+----------+--------------------------------+------------------+
| Phillips 2013\*  | GENEA\             | 8-14 yr  | `do.enmoa = TRUE`\             | Light: 37.5\     |
|                  | Hip                |          | `do.enmo = FALSE`\             | Moderate: 212.5\ |
|                  |                    |          | `acc.metric = "ENMOa"`         | Vigorous: 637.5  |
+------------------+--------------------+----------+--------------------------------+------------------+
| Schaefer 2014\*  | GENEActiv\         | 6-11 yr  | `do.bfen = TRUE`\              | Light: 190\      |
|                  | Non-dominant wrist |          | `lb = 0.2`\                    | Moderate: 314\   |
|                  |                    |          | `hb = 15 `\                    | Vigorous: 998    |
|                  |                    |          | `do.enmo = FALSE`\             |                  |
|                  |                    |          | `acc.metric = "BFEN"`          |                  |
+------------------+--------------------+----------+--------------------------------+------------------+
| Hildebrand 2014\ | ActiGraph\         | 7-11 yr  | **Default values\              | Light: 35.6\     |
| Hildebrand 2016  | Non-dominant wrist |          | **`do.enmo = TRUE`\            | Moderate: 201.4\ |
|                  |                    |          | `acc.metric = "ENMO"`          | Vigorous: 707.0  |
+------------------+--------------------+----------+--------------------------------+------------------+
| Hildebrand 2014\ | GENEActiv\         | 7-11 yr  | **Default values\              | Light: 56.3\     |
| Hildebrand 2016  | Non-dominant wrist |          | **`do.enmo = TRUE`\            | Moderate: 191.6\ |
|                  |                    |          | `acc.metric = "ENMO"`          | Vigorous: 695.8  |
+------------------+--------------------+----------+--------------------------------+------------------+
| Hildebrand 2014\ | ActiGraph\         | 7-11 yr  | **Default values\              | Light: 63.3\     |
| Hildebrand 2016  | Hip                |          | **`do.enmo = TRUE`\            | Moderate: 142.6\ |
|                  |                    |          | `acc.metric = "ENMO"`          | Vigorous: 464.6  |
+------------------+--------------------+----------+--------------------------------+------------------+
| Hildebrand 2014\ | GENEActiv\         | 7-11 yr  | **Default values\              | Light: 64.1\     |
| Hildebrand 2016  | Hip                |          | **`do.enmo = TRUE`\            | Moderate: 152.8\ |
|                  |                    |          | `acc.metric = "ENMO"`          | Vigorous: 514.3  |
+------------------+--------------------+----------+--------------------------------+------------------+
| Aittasalo 2015   | ActiGraph\         | 13-15 yr | **Default values\              | Light: 26.9\     |
|                  | Hip                |          | **`do.mad = TRUE`\             | Moderate: 332\   |
|                  |                    |          | `do.enmo = FALSE`\             | Vigorous: 558.3  |
|                  |                    |          | `acc.metric = "MAD"`           |                  |
+------------------+--------------------+----------+--------------------------------+------------------+
| Aittasalo 2015   | Hookie AM20\       | 13-15 yr | **Default values\              | Light: 28.7\     |
|                  | Hip                |          | **`do.mad = TRUE`\             | Moderate: 338\   |
|                  |                    |          | `do.enmo = FALSE`\             | Vigorous: 558.3  |
|                  |                    |          | `acc.metric = "MAD"`           |                  |
+------------------+--------------------+----------+--------------------------------+------------------+

\*These publications used acceleration metrics that sum their values per epoch rather than average them per epoch like GGIR does. So, to use their cut-point in GGIR, we provide a scaled version of the cut-points presented in the paper as: `(CutPointFromPaper_in_gmins/(sampleRateFromPaper * EpochLengthInSecondsPaper)) * 1000`
\*\* This  publication used acceleration metrics that expressed their cut-points in _g_ units. So, to use their cut-point in GGIR, we provide a cut-point  multiplied by 1000.

## Cut-points for adults

+-------------------+--------------------+---------------+-------------------------+------------------+
| Cut-points        | Device\            | Age           | Relevant arguments      | thresholds       |
|                   | Attachment site    |               |                         |                  |
+:==================+:===================+:==============+:========================+:=================+
| Esliger 2011\*    | Left wrist         | 40-65 yr      | `do.enmoa = TRUE`\      | Light: 45\       |
|                   |                    |               | `do.enmo = FALSE`\      | Moderate: 134\   |
|                   |                    |               | `acc.metric = "ENMOa"`  | Vigorous: 377    |
+-------------------+--------------------+---------------+-------------------------+------------------+
| Esliger 2011\*    | Right wrist        | 40-65 yr      | `do.enmoa = TRUE`\      | Light: 80\       |
|                   |                    |               | `do.enmo = FALSE`\      | Moderate: 92\    |
|                   |                    |               | `acc.metric = "ENMOa"`  | Vigorous: 437    |
+-------------------+--------------------+---------------+-------------------------+------------------+
| Esliger 2011\*    | Waist              | 40-65 yr      | `do.enmoa = TRUE`\      | Light: 16\       |
|                   |                    |               | `do.enmo = FALSE`\      | Moderate: 46\    |
|                   |                    |               | `acc.metric = "ENMOa"`  | Vigorous: 428    |
+-------------------+--------------------+---------------+-------------------------+------------------+
| Hildebrand 2014\  | ActiGraph\         | 21-61 yr      | **Default values\       | Light: 44.8\     |
| Hildebrand 2016   | Non-dominant wrist |               | **`do.enmo = TRUE`\     | Moderate: 100.6\ |
|                   |                    |               | `acc.metric = "ENMO"`   | Vigorous: 428.8  |
+-------------------+--------------------+---------------+-------------------------+------------------+
| Hildebrand 2014\  | GENEActiv\         | 21-61 yr      | **Default values\       | Light: 45.8\     |
| Hildebrand 2016   | Non-dominant wrist |               | **`do.enmo = TRUE`\     | Moderate: 93.2\  |
|                   |                    |               | `acc.metric = "ENMO"`   | Vigorous: 418.3  |
+-------------------+--------------------+---------------+-------------------------+------------------+
| Hildebrand 2014\  | ActiGraph\         | 21-61 yr      | **Default values\       | Light: 47.4\     |
| Hildebrand 2016   | Hip                |               | **`do.enmo = TRUE`\     | Moderate: 69.1\  |
|                   |                    |               | `acc.metric = "ENMO"`   | Vigorous: 258.7  |
+-------------------+--------------------+---------------+-------------------------+------------------+
| Hildebrand 2014\  | GENEActiv\         | 21-61 yr      | **Default values\       | Light: 46.9\     |
| Hildebrand 2016   | Hip                |               | **`do.enmo = TRUE`\     | Moderate: 68.7\  |
|                   |                    |               | `acc.metric = "ENMO"`   | Vigorous: 266.8  |
+-------------------+--------------------+---------------+-------------------------+------------------+
| Mielke 2023       | GENEActiv\         | 35 (SD=11) yr | **Default values\       | Light: 36\       |
|                   | Non-dominant wrist |               | **`do.enmo = TRUE`\     | Moderate: 92\    |
|                   |                    |               | `acc.metric = "ENMO"`   | Vigorous: 283    |
+-------------------+--------------------+---------------+-------------------------+------------------+
| Mielke 2023       | ActiGraph\         | 35 (SD=11) yr | **Default values\       | Light: 25\       |
|                   | Non-dominant wrist |               | **`do.enmo = TRUE`\     | Moderate: 78\    |
|                   |                    |               | `acc.metric = "ENMO"`   | Vigorous: 249    |
+-------------------+--------------------+---------------+-------------------------+------------------+
| Mielke 2023       | GENEActiv\         | 35 (SD=11) yr | **Default values\       | Light: 30\       |
|                   | Dominant wrist     |               | **`do.enmo = TRUE`\     | Moderate: 85\    |
|                   |                    |               | `acc.metric = "ENMO"`   | Vigorous: 270    |
+-------------------+--------------------+---------------+-------------------------+------------------+
| Mielke 2023       | ActiGraph\         | 35 (SD=11) yr | **Default values\       | Light: 40\       |
|                   | Dominant waist     |               | **`do.enmo = TRUE`\     | Moderate: 65\    |
|                   |                    |               | `acc.metric = "ENMO"`   | Vigorous: 190    |
+-------------------+--------------------+---------------+-------------------------+------------------+
| Vähä-Ypyä 2015    | Hookie AM20\       | 35 (SD=11) yr | `do.mad = TRUE`\        | Light: N/A\      |
|                   | Hip                |               | `do.enmo = FALSE`\      | Moderate: 91\    |
|                   |                    |               | `acc.metric = "MAD"`    | Vigorous: 414    |
+-------------------+--------------------+---------------+-------------------------+------------------+
| Dillon 2016\*^,†^ | GENEActiv\         | 50-69 yr      | `do.enmoa = TRUE`\      | Light: 105.6\    |
|                   | Non-dominant wrist |               | `do.enmo = FALSE`\      | Moderate: 174.2\ |
|                   |                    |               | `acc.metric = "ENMOa"`  | Vigorous: 330    |
+-------------------+--------------------+---------------+-------------------------+------------------+
| Dillon 2016\*^,†^ | GENEActiv\         | 50-69 yr      | `do.enmoa = TRUE`\      | Light: 127.8\    |
|                   | Dominant wrist     |               | `do.enmo = FALSE`\      | Moderate: 187.6\ |
|                   |                    |               | `acc.metric = "ENMOa"`  | Vigorous: 396.4  |
+-------------------+--------------------+---------------+-------------------------+------------------+
| Buchan 2023\*^,†^ | activPAL\          | 23 (SD=4) yr  | **Default values\       | Light: 26.4\     |
|                   | Right thigh        |               | `do.enmo = TRUE`\       | Moderate: N/A\   |
|                   |                    |               | `acc.metric = "ENMO"`   | Vigorous: N/A    |
+-------------------+--------------------+---------------+-------------------------+------------------+
| Buchan 2023\*^,†^ | activPAL\          | 23 (SD=4) yr  | `do.mad = TRUE`\        | Light: 30.1\     |
|                   | Right thigh        |               | `do.enmo = FALSE`\      | Moderate: N/A\   |
|                   |                    |               | `acc.metric = "MAD"`    | Vigorous: N/A    |
+-------------------+--------------------+---------------+-------------------------+------------------+

\*These publications used acceleration metrics that sum their values per epoch rather than average them per epoch like GGIR does. So, to use their cut-point in GGIR, we provide a scaled version of the cut-points presented in the paper as: `(CutPointFromPaper_in_gmins/(sampleRateFromPaper * EpochLengthInSecondsPaper)) * 1000`
^†^ In this [publication](https://doi.org/10.1371%2Fjournal.pone.0109913), there are cut-point based on data sampled at 30 Hz and 100 Hz. When scaling the cut-points as specified in (\*), the resulting thresholds are virtually the same (the ones presented in this table).

## Cut-points for older adults

+------------------+--------------------+-------------------+------------------------+-----------------+
| Cut-points       | Device\            | Age               | Relevant arguments     | thresholds      |
|                  | Attachment site    |                   |                        |                 |
+:=================+:===================+:==================+:=======================+:================+
| Sanders 2019\*   | GENEActiv\         | 60-86 yr          | **Default values\      | Light: 20\      |
|                  | Non-dominant wrist |                   | **`do.enmo = TRUE`\    | Moderate: 32\   |
|                  |                    |                   | `acc.metric = "ENMO"`  | Vigorous: N/A   |
+------------------+--------------------+-------------------+------------------------+-----------------+
| Sanders 2019\*\* | GENEActiv\         | 60-86 yr          | **Default values\      | Light: 57\      |
|                  | Non-dominant wrist |                   | **`do.enmo = TRUE`\    | Moderate: 104\  |
|                  |                    |                   | `acc.metric = "ENMO"`  | Vigorous: N/A   |
+------------------+--------------------+-------------------+------------------------+-----------------+
| Sanders 2019\*   | ActiGraph\         | 60-86 yr          | **Default values\      | Light: 6\       |
|                  | Hip                |                   | **`do.enmo = TRUE`\    | Moderate: 19\   |
|                  |                    |                   | `acc.metric = "ENMO"`  | Vigorous: N/A   |
+------------------+--------------------+-------------------+------------------------+-----------------+
| Sanders 2019\*\* | ActiGraph\         | 60-86 yr          | **Default values\      | Light: 15\      |
|                  | Hip                |                   | **`do.enmo = TRUE`\    | Moderate: 69\   |
|                  |                    |                   | `acc.metric = "ENMO"`  | Vigorous: N/A   |
+------------------+--------------------+-------------------+------------------------+-----------------+
| Migueles 2021    | ActiGraph\         | ≥70 yr \          | **Default values\      | Light: 18\      |
|                  | Non-dominant wrist | (mean: 78.7 yr)   | **`do.enmo = TRUE`\    | Moderate: 60\   |
|                  |                    |                   | `acc.metric = "ENMO"`  | Vigorous: N/A   |
+------------------+--------------------+-------------------+------------------------+-----------------+
| Migueles 2021    | ActiGraph\         | ≥70 yr \          | **Default values\      | Light: 22\      |
|                  | Dominant wrist     | (mean: 78.7 yr)   | **`do.enmo = TRUE`\    | Moderate: 64\   |
|                  |                    |                   | `acc.metric = "ENMO"`  | Vigorous: N/A   |
+------------------+--------------------+-------------------+------------------------+-----------------+
| Migueles 2021    | ActiGraph\         | ≥70 yr \          | **Default values\      | Light: 7\       |
|                  | Hip                | (mean: 78.7 yr)   | **`do.enmo = TRUE`\    | Moderate: 14\   |
|                  |                    |                   | `acc.metric = "ENMO"`  | Vigorous: N/A   |
+------------------+--------------------+-------------------+------------------------+-----------------+
| Bammann 2021     | ActiGraph\         | 62.9 (SD=3.6) yr  | **Default values\      | Moderate: 94\   |
|                  | Hip                |                   | **`do.enmo = TRUE`\    | Vigorous: 230   |
|                  |                    |                   | `acc.metric = "ENMO"`  |                 |
+------------------+--------------------+-------------------+------------------------+-----------------+
| Bammann 2021     | ActiGraph\         | 62.9 (SD=3.6) yr  | **Default values\      | Moderate: 122\  |
|                  | Dominant Wrist     |                   | **`do.enmo = TRUE`\    | Vigorous: 234   |
|                  |                    |                   | `acc.metric = "ENMO"`  |                 |
+------------------+--------------------+-------------------+------------------------+-----------------+
| Bammann 2021     | ActiGraph\         | 62.9 (SD=3.6) yr  | **Default values\      | Moderate: 100\  |
|                  | Non-dominant Wrist |                   | **`do.enmo = TRUE`\    | Vigorous: 245   |
|                  |                    |                   | `acc.metric = "ENMO"`  |                 |
+------------------+--------------------+-------------------+------------------------+-----------------+
| Bammann 2021     | ActiGraph\         | 62.9 (SD=3.6) yr  | **Default values\      | Moderate: 342   |
|                  | Dominant Ankle     |                   | **`do.enmo = TRUE`\    |                 |
|                  |                    |                   | `acc.metric = "ENMO"`  |                 |
+------------------+--------------------+-------------------+------------------------+-----------------+
| Bammann 2021     | ActiGraph\         | 62.9 (SD=3.6) yr  | **Default values\      | Moderate: 331   |
|                  | Non-dominant Ankle |                   | **`do.enmo = TRUE`\    |                 |
|                  |                    |                   | `acc.metric = "ENMO"`  |                 |
+------------------+--------------------+-------------------+------------------------+-----------------+
| Fraysse 2020^†^  | GENEActiv\         | ≥70 yr \          | `do.enmoa = TRUE`\     | Light: 42.5\    |
|                  | Non-dominant wrist | (mean: 77 yr)     | `do.enmo = FALSE`\     | Moderate: 98\   |
|                  |                    |                   | `acc.metric = "ENMOa"` | Vigorous: N/A   |
+------------------+--------------------+-------------------+------------------------+-----------------+
| Fraysse 2020^†^  | GENEActiv\         | ≥70 yr \          | `do.enmoa = TRUE`\     | Light: 62.5\    |
|                  | Dominant wrist     | (mean: 77 yr)     | `do.enmo = FALSE`\     | Moderate: 92.5\ |
|                  |                    |                   | `acc.metric = "ENMOa"` | Vigorous: N/A   |
+------------------+--------------------+-------------------+------------------------+-----------------+
| Dibben 2020^‡^   | GENEActiv\         | 70.7 (SD=14.1) yr | `do.enmoa = TRUE`\     | Light: 18.6\    |
|                  | Right wrist        |                   | `do.enmo = FALSE`\     | Moderate: 45.5\ |
|                  |                    |                   | `acc.metric = "ENMOa"` | Vigorous: N/A   |
+------------------+--------------------+-------------------+------------------------+-----------------+
| Dibben 2020^‡^   | GENEActiv\         | 70.7 (SD=14.1) yr | `do.mad = TRUE`\       | Light: 18.3\    |
|                  | Right wrist        |                   | `do.enmo = FALSE`\     | Moderate: 26.2\ |
|                  |                    |                   | `acc.metric = "MAD"`   | Vigorous: N/A   |
+------------------+--------------------+-------------------+------------------------+-----------------+
| Dibben 2020^‡^   | GENEActiv\         | 70.7 (SD=14.1) yr | `do.enmoa = TRUE`\     | Light: 16.7\    |
|                  | Left wrist         |                   | `do.enmo = FALSE`\     | Moderate: 43.6\ |
|                  |                    |                   | `acc.metric = "ENMOa"` | Vigorous: N/A   |
+------------------+--------------------+-------------------+------------------------+-----------------+
| Dibben 2020^‡^   | GENEActiv\         | 70.7 (SD=14.1) yr | `do.mad = TRUE`\       | Light: 18.7\    |
|                  | Left wrist         |                   | `do.enmo = FALSE`\     | Moderate: 22.8\ |
|                  |                    |                   | `acc.metric = "MAD"`   | Vigorous: N/A   |
+------------------+--------------------+-------------------+------------------------+-----------------+
| Dibben 2020^‡^   | GENEActiv\         | 70.7 (SD=14.1) yr | `do.enmoa = TRUE`\     | Light: 7.6\     |
|                  | Hip                |                   | `do.enmo = FALSE`\     | Moderate: 40.6\ |
|                  |                    |                   | `acc.metric = "ENMOa"` | Vigorous: N/A   |
+------------------+--------------------+-------------------+------------------------+-----------------+
| Dibben 2020^‡^   | GENEActiv\         | 70.7 (SD=14.1) yr | `do.mad = TRUE`\       | Light: 1\       |
|                  | Hip                |                   | `do.enmo = FALSE`\     | Moderate: 2.4\  |
|                  |                    |                   | `acc.metric = "MAD"`   | Vigorous: N/A   |
+------------------+--------------------+-------------------+------------------------+-----------------+


\*Cut-points derived from applying the Youden index on ROC curves.\
\*\* Cut-points derived from increasing Sensitivity over Specificity for light and vice versa for moderate on ROC curves (see [paper](https://doi.org/10.1080/02640414.2018.1555904) for more details).\
^†^ These publications used acceleration metrics that sum their values per epoch rather than average them per epoch like GGIR does. So, to use their cut-point in GGIR, we provide a scaled version of the cut-points presented in the paper as: `(CutPointFromPaper_in_gmins/(sampleRateFromPaper * EpochLengthInSecondsPaper)) * 1000`
^‡^ More cut-points excluding data on aided walking and washing up activities can be found in the [publication](https://doi.org/10.1186/s13102-020-00196-7).

# Notes on cut-point validity

**Sensor calibration**

In all of the studies above, excluding Hildebrand et al. 2016, no effort
was made to calibrate the acceleration sensors relative to gravitational
acceleration prior to cut-point development. Theoretically this can be
expected to cause a bias in the cut-point estimates proportional to the
calibration error in each device, especially for cut-points based on
acceleration metrics which rely on the assumption of accurate
calibration such as metrics: ENMO, EN, ENMOa, and by that also metric
SVMgs used by studies such as Esliger 2011, Phillips 2013, and Dibben
2020.

**Idle sleep mode and ActiGraph**

As discussed in the [main package vignette](https://cran.r-project.org/package=GGIR/vignettes/GGIR.html),
studies using the ActiGraph sensor often forget to clarify whether idle sleep
mode was used and if so, how it was accounted for in the data processing.

**How about all the criticism towards cut-point methods?**

For a more elaborate reflection on the limitations of cut-points and a motivation
why cut-points still have value in GGIR see:
<https://www.accelting.com/why-does-ggir-facilitate-cut-points/>


# References

-   Aittasalo 2015: <https://doi.org/10.1186/s13102-015-0010-0>
-   Bammann 2021: <https://doi.org/10.1371/journal.pone.0252615>
-   Dibben 2020: <https://doi.org/10.1186/s13102-020-00196-7>
-   Dillon 2016: <https://doi.org/10.1371%2Fjournal.pone.0109913>
-   Esliger 2011: <https://doi.org/10.1249/mss.0b013e31820513be>
-   Fraysse 2020: <https://doi.org/10.3389%2Ffspor.2020.579278>
-   Hildebrand 2014: <https://doi.org/10.1249/mss.0000000000000289>
-   Hildebrand 2016: <https://doi.org/10.1111/sms.12795>
-   Mielke 2023: <https://doi.org/10.1111/sms.14416>
-   Migueles 2021: <https://doi.org/10.3390%2Fs21103326>
-   Phillips 2013: <https://doi.org/10.1016/j.jsams.2012.05.013>
-   Sanders 2018: <https://doi.org/10.1080/02640414.2018.1555904>
-   Schaefer 2014: <https://doi.org/10.1249%2FMSS.0000000000000150>
-   Roscoe 2017: <https://doi.org/10.1007/s00431-017-2948-2>
-   Vähä-Ypyä 2015: <https://doi.org/10.1371/journal.pone.0134813>

---
title: "Accelerometer data processing with GGIR"
output:
   html_document:
    toc : true
    number_sections: true
    toc_depth: 3
    toc_float: true #by turning this on, the table of contents moves to the left of the page.
urlcolor: blue
vignette: >
  %\VignetteIndexEntry{Accelerometer data processing with GGIR}
  \usepackage[utf8]{inputenc}
  %\VignetteEngine{knitr::rmarkdown}
editor_options:
  markdown:
    wrap: 72
---

```{r, echo=FALSE, out.width = "60%", out.extra='style="border: 0; padding:20px"', fig.alt="GGIR logo"}
knitr::include_graphics("GGIR-MASTERLOGO-RGB.png")
```

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
```

**NOTE: With the growing amount of functionality in GGIR we have decided to migrate the narrative documentation to the [GitHub pages of GGIR](https://wadpac.github.io/GGIR/). This to ease maintenance and accessibility. Therefore, many of the sections in this vignette have been replaced by a link to their new location.**

# Introduction

## What is GGIR?

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/chapter1_WhatIsGGIR.html) in the GGIR github-pages, which is now the main documentation resource for GGIR.


**Cite GGIR:**

This section has been migrated to this [section](https://wadpac.github.io/GGIR/authors.html#citation) in the GGIR github-pages, which is now the main documentation resource for GGIR.

## Contributing, Support, and Keeping up to date

# Setting up your work environment

## Install R and RStudio

## Prepare folder structure

## GGIR shell function

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/chapter2_Pipeline.html#the-ggir-function) in the GGIR github-pages, which is now the main documentation resource for GGIR.

### Key general arguments

You will probably never need to think about most of the arguments listed
above, because a lot of arguments are only included to facilitate
methodological studies where researchers want to have control over every
little detail. See previous paragraph for links to the documentation and
how to find the default value of each parameter.

The bare minimum input needed for `GGIR` is:

```{R,eval=FALSE}
library(GGIR)
GGIR(datadir="C:/mystudy/mydata",
 outputdir="D:/myresults")
```

Argument `datadir` allows you to specify where you have stored your
accelerometer data and `outputdir` allows you to specify where you would
like the output of the analyses to be stored. This cannot be equal to
`datadir`. If you copy paste the above code to a new R script (file
ending with .R) and Source it in R(Studio) then the dataset will be
processed and the output will be stored in the specified output
directory.

Below we have highlighted the key arguments you may want to be aware of.
We are not giving a detailed explanation, please see the package manual
for that.

-   `mode` - which part of GGIR to run, GGIR is constructed in five
    parts with a sixth part under development.
-   `overwrite` - whether to overwrite previously produced milestone
    output. Between each GGIR part, GGIR stores milestone output to ease
    re-running parts of the pipeline.
-   `idloc` - tells GGIR where to find the participant ID (default:
    inside file header)
-   `data_masking_strategy` - informs GGIR how to consider the design of the
    experiment.
    -   If `data_masking_strategy` is set to value 1, then check out arguments
        `hrs.del.start` and `hrs.del.end`.
    -   If `data_masking_strategy` is set to value 3 or 5, then check out arguments
        `ndayswindow`, `hrs.del.start` and `hrs.del.end`.
-   `maxdur` - maximum number of days you expect in a data file based on
    the study protocol.
-   `desiredtz` - time zone of the experiment.
-   `chunksize` - a way to tell GGIR to use less memory, which can be
    useful on machines with limited memory.
-   `includedaycrit` - tell GGIR how many hours of valid data per day
    (midnight-midnight) is acceptable.
-   `includenightcrit` - tell GGIR how many hours of a valid night
    (noon-noon) is acceptable.
-   `qwindow` - argument to tell GGIR whether and how to segment the day
    for day-segment specific analysis.
-   `mvpathreshold` and `boutcriter` - acceleration threshold and bout
    criteria used for calculating time spent in MVPA (only used in GGIR
    part2).
-   `epochvalues2csv` - to export epoch level magnitude of acceleration
    to a csv files (in addition to already being stored as RData file)
-   `dayborder` - to decide whether the edge of a day should be other
    than midnight.
-   `iglevels` - argument related to intensity gradient method proposed
    by A. Rowlands.
-   `do.report` - specify reports that need to be generated.
-   `viewingwindow` and `visualreport` - to create a visual report, this
    only works when all five parts of GGIR have successfully run.
    Note that the visual report was initially developed to provide something to show to
    study participants and not for data quality checking purposes. Over time we have improved
    the visual report to also be useful for QC-ing the data. however, some of
    the scorings as shown in the visual report are created for the visual report only and
    may not reflect the scorings in the main GGIR analyses as reported in the
    quantitative csv-reports. Most of our effort in the past 10 years has gone 
    into making sure that the csv-report are correct, while the visualreport has
    mostly been a side project. This is unfortunate and we hope to find funding
    in the future to design a new report specifically for the purpose of
    QC-ing the anlayses done by GGIR.
-   `maxRecordingInterval` - if specified controls whether neighboring or
    overlapping recordings with the same participant ID and brand are appended
    at epoch level. This can be useful when the intention is to monitor behaviour
    over larger periods of time but accelerometers only allow for a few weeks of 
    data collection. GGIR will never append or alter the raw input file, this operation
    is preformed on the derived data.
-   `study_dates_file` - if specified trims the recorded data to the first and 
    last date in which the study took place. This is relevant for studies that started
    the recording several days before the accelerometers were actually worn by
    participants. This is used on the top of data_masking_strategy, so that it may be combined
    with the strategies in GGIR.

### Key arguments related to sleep analysis {#key-arguments-related-to-sleep-analysis}

For an explanation on how sleep is detected and the specific role of the
various function arguments see section [Sleep
analysis](#Sleep_analysis).

-   Arguments related to configuring the sleep detection algorithm:
    `anglethreshold`, `timethreshold`, `HASPT.algo`, `HASIB.algo`,
    `Sadeh_axis`, and `HASPT.ignore.invalid`.
-   `ignorenonwear` if set to TRUE then ignore detected monitor non-wear
    periods in the detection of sustained inactivity bouts to avoid
    confusion between monitor non-wear time.
-   If you want to create a visualisation of how sleep period time and
    sustained inactivity bouts match throughout a day then consider
    arguments `do.visual`, `outliers.only`, and `criterror`.
-   If you want to exclude the first and last night from the sleep
    analysis then used `excludefirstlast`.
-   `def.noc.sleep` specifies how the sleep period time window should be
    estimated if no sleeplog is used.
-   `includenightcrit` Minimum number of valid hours per night (24 hour
    window between noon and noon or 6pm-6pm).
-   `data_cleaning_file` to ginore specific nights for specific
    individuals, see also section [Data cleaning
    file](#Data_cleaning_file).
-   If you want the sleep analysis to be guided by a sleeplog (diary)
    then check out arguments `loglocation` which specifies the location
    of the spreadsheet (csv) with sleep log information. Further, use
    arguments `colid` and `coln1` to specify
    the details of your sleep log structure.

GGIR facilitates two possible sleeplog file structures:

#### Basic sleep log

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/chapter9_SleepFundamentalsGuiders.html#basic-sleep-log) in the GGIR github-pages, which is now the main documentation resource for GGIR.

#### Advanced sleep log

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/chapter9_SleepFundamentalsGuiders.html#advanced-sleep-log) in the GGIR github-pages, which is now the main documentation resource for GGIR.

### Key arguments related to time use analysis

### Published cut-points and how to use them

This section has been rewritten and moved. Please, visit the vignette
[Published cut-points and how to use them in GGIR](https://cran.r-project.org/package=GGIR)
for more details on the cut-points available, how to use them, and some
additional reflections on the use of cut-points in GGIR.

### Example call

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/chapter2_Pipeline.html) in the GGIR github-pages, which is now the main documentation resource for GGIR.

### Configuration file

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/chapter2_Pipeline.html#configuration-file-) in the GGIR github-pages, which is now the main documentation resource for GGIR.

# Time for action: How to run your analysis?

## From the R console on your own desktop/laptop

Create an R-script and put the GGIR call in it. Next, you can source the R-script
with the `source` function in R:

`source("pathtoscript/myshellscript.R")`

or use the Source button in RStudio if you use RStudio.

## In a cluster

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/chapter2_Pipeline.html#in-a-cluster) in the GGIR github-pages, which is now the main documentation resource for GGIR.

## Processing time

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/chapter2_Pipeline.html#processing-time) in the GGIR github-pages, which is now the main documentation resource for GGIR.

# Inspecting the results

GGIR generates the following types of output. 

- csv-spreadsheets with all the variables you need for physical activity, sleep and circadian
rhythm research
- Pdfs with on each page a low resolution plot of the data per file and quality indicators
- R objects with milestone data
- Pdfs with a visual summary of the physical activity and sleep patterns as identified (see example below)

```{r, out.width = "700px",echo=FALSE, fig.alt="Example visualreport"}
knitr::include_graphics("reportexample.jpg")
```

## Output part 2 {.tabset}

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/GGIRoutput.html) in the GGIR github-pages, which is now the main documentation resource for GGIR.

### Person level summary

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/GGIRoutput.html) in the GGIR github-pages, which is now the main documentation resource for GGIR.

### Day level summary

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/GGIRoutput.html) in the GGIR github-pages, which is now the main documentation resource for GGIR.

### Data_quality_report

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/GGIRoutput.html) in the GGIR github-pages, which is now the main documentation resource for GGIR.

## Output part 4 {.tabset}

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/GGIRoutput.html) in the GGIR github-pages, which is now the main documentation resource for GGIR.

### Night level summaries

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/GGIRoutput.html) in the GGIR github-pages, which is now the main documentation resource for GGIR.

#### Non-default variables in part 4 csv report

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/GGIRoutput.html) in the GGIR github-pages, which is now the main documentation resource for GGIR.

### visualisation_sleep.pdf

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/GGIRoutput.html) in the GGIR github-pages, which is now the main documentation resource for GGIR.

## Output part 5 {.tabset}

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/GGIRoutput.html) in the GGIR github-pages, which is now the main documentation resource for GGIR.

### Day level summary {#output5}

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/GGIRoutput.html) in the GGIR github-pages, which is now the main documentation resource for GGIR.

### Person level summary

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/GGIRoutput.html) in the GGIR github-pages, which is now the main documentation resource for GGIR.


## Output part 6 {.tabset}

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/GGIRoutput.html#ggir-part-6) in the GGIR github-pages, which is now the main documentation resource for GGIR.

# Motivation and clarification

In this chapter we will try to collect motivations and clarification
behind GGIR which may not have been clear from the existing
publications.

## Reproducibilty of GGIR analyses

Some tips to increase reproducibility of your findings:

1.  When you publish your findings, please remember to add the GGIR
    package version number. All of GGIR are archived by CRAN and
    available from the archive section on the package
    [website](https://CRAN.R-project.org/package=GGIR). GGIR has evolved
    over the years. To get a better understanding of how versions differ
    you should check the NEWS sections from the package
    [website](https://CRAN.R-project.org/package=GGIR)
2.  Report how you configured the accelerometer
3.  Report the study protocol and wear instructions given to the
    participants
4.  Report GGIR version
5.  Report how GGIR was used: Share the config.csv file or your R script
6.  Report how you post-processed / cleaned GGIR output
7.  Report how reported outcomes relate to the specific variable names
    in GGIR

## Auto-calibration

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/chapter3_QualityAssessment.html#auto-calibration-algorithm) in the GGIR github-pages, which is now the main documentation resource for GGIR.

## Non-wear detection

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/chapter3_QualityAssessment.html#non-wear-detection)  in the GGIR github-pages, which is now the main documentation resource for GGIR.

## Clipping score

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/chapter3_QualityAssessment.html#clipping-detection)  in the GGIR github-pages, which is now the main documentation resource for GGIR.

## Why collapse information to epoch level?

Although many data points are collected we decide to only work with
aggregated values (e.g. 1 or 5 second epochs) for the following reasons:

1.  Accelerometers are often used to describe patterns in metabolic
    energy expenditure. Metabolic energy expenditure is typically
    defined per breath or per minute (indirect calorimetry), per day
    (room calorimeter), or per multiple days (doubly labelled water
    method). In order to validate our methods against these reference
    standards we need to work with a similar time resolution.

2.  Collapsing the data to epoch summary measures helps to standardise
    for differences in sample frequency between studies.

3.  There is little evidence that the raw data is an accurate
    representation of body acceleration. All scientific evidence on the
    validity of accelerometer data has so far been based on epoch
    averages.

4.  Collapsing the data to epoch summary measures may help to average
    out different noise levels and make sensor brands more comparable.

### Why does the first epoch not allign with the original start of the recording

GGIR uses short (default 5 seconds) and long epochs (default 15
minutes). The epochs are aligned to the hour in the day, and to each
other. For example, if a recording starts at 9:52:00 then the GGIR will
work with epochs derived from 10:00:00 onward. If the recording starts
at 10:12 then GGIR will work with epochs derived from 10:15:00 onward.

Motivation:

-   This allows us to have a standardised time grid across recordings to
    describe behaviour.
-   This allows us to calculate behaviour exactly per day or per
    specified time interval in a day.

If the first 15 minute epochs would start at 9:52 then the next one
would start at 10:07, which makes it impossible to make statement about
behaviour between 10:00 and 13:00.

## Sleep analysis {#Sleep_analysis}

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/chapter8_SleepFundamentalsSibs.html) in the GGIR github-pages, which is now the main documentation resource for GGIR.

### Notes on sleep classification algorithms designed for count data


#### Replication of the movement counts needed

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/chapter4_AccMetrics.html#notes-on-implementation-of-zero-crossing-counts) in the GGIR github-pages, which is now the main documentation resource for GGIR.


#### Missing information for replicating movement counts

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/chapter4_AccMetrics.html#notes-on-implementation-of-zero-crossing-counts) in the GGIR github-pages, which is now the main documentation resource for GGIR.

#### An educated guess and how you can to help optimise the implementation

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/chapter4_AccMetrics.html#notes-on-implementation-of-zero-crossing-counts) in the GGIR github-pages, which is now the main documentation resource for GGIR.

### Guiders

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/chapter9_SleepFundamentalsGuiders.html) in the GGIR github-pages, which is now the main documentation resource for GGIR.

### Daysleepers (nights workers)

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/chapter9_SleepFundamentalsGuiders.html#time-window-used-for-sleep-analyses) in the GGIR github-pages, which is now the main documentation resource for GGIR.

### Cleaningcode {#Cleaningcode}

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/chapter10_SleepAnalysis.html#cleaningcode) in the GGIR github-pages, which is now the main documentation resource for GGIR.

### Difference between cleaned and full output

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/chapter10_SleepAnalysis.html#related-output) in the GGIR github-pages, which is now the main documentation resource for GGIR.

### Data cleaning file {#Data_cleaning_file}

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/chapter10_SleepAnalysis.html#data-cleaning-file) in the GGIR github-pages, which is now the main documentation resource for GGIR.

## Waking-waking or 24 hour time-use analysis {#Waking-waking_or_24_hour}

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/chapter12_TimeUseAnalysis.html) in the GGIR github-pages, which is now the main documentation resource for GGIR.

### Time series output files

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/chapter12_TimeUseAnalysis.html#exporting-time-series) in the GGIR github-pages, which is now the main documentation resource for GGIR.

### Day inclusion criteria

The full part 5 output is stored in the `results/QC` folder. The default
inclusion criteria for days in the cleaned output from part 5 (stored in
the `results` folder) are:

-   For both MM and WW defined days: The valid (sensor worn) time
    fraction of the day needs to be above the fraction specified with
    argument `includedaycrit.part5` (default 2/3).
-   For MM defined days only: The length of the day needs to be at least
    the number of hours as specified by `minimum_MM_length.part5`
    (default 23). Note that if your experiment started and ended in the
    middle of the day then this default setting will exclude those
    incomplete first and last days. If you think including these days is
    still meaningful for your work then adjust the argument
    `minimum_MM_length.part5`.

**Important notes:**

-   No criteria is set for the amount of valid data during the SPT
    window, because all we are interested in part 5 is knowing the
    borders of the night and we trust that this was sufficiently
    estimated by part 4. If you disagree then please notice that all the
    days are included in the full report available in `results/QC`
    folder.
-   This means that argument `includenightcrit` as used for part 4 is
    not used in part 5.

The `data_cleaning_file` argument discussed in
[Data_cleaning_file](#Data_cleaning_file) also allows you to tell GGIR
which person(s) and day(s) should be omitted in part 5. The the day
numbers to be excluded should be listed in a column `day_part5` as
header.

### Fragmentation metrics

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/chapter11_DescribingDataCutPoints.html#behavioural-fragmentation) in the GGIR github-pages, which is now the main documentation resource for GGIR.

**Difference between fragments and blocks:**

Elsewhere in the part5 we use the term `block`. A `block` is a sequence
of epochs that belong to the same behavioural class. This may sound
similar to the definition of a fragment, but for blocks we distinguish
every behavioural class, which includes the subcategories such as bouted
and unbouted behaviour. This means that variables `Nblock_day_total_IN`
and `Nblock_day_total_LIG` are identical to `Nfrag_IN_day` and
`Nfrag_LIPA_day`, respectively. In contrast, for fragments we may group LIPA and MVPA
together when refering to the fragmentation of PA.

**Differences with R package ActFrag:**

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/chapter11_DescribingDataCutPoints.html#differences-with-r-package-actfrag) in the GGIR github-pages, which is now the main documentation resource for GGIR.


## Why use data metric ENMO as default?

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/chapter4_AccMetrics.html#why-ggir-uses-enmo-as-a-default-) in the GGIR github-pages, which is now the main documentation resource for GGIR.

## What does GGIR stand for?

I wanted a short name and not to spend too much time finding it. The abbreviation has lost its functional meaning, which is why we now only use GGIR as the name.

## Circadian Rhythm analyses

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/chapter13_CircadianRhythm.html) in the GGIR github-pages, which is now the main documentation resource for GGIR.

## ActiGraph's idle sleep mode

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/GGIR.html?q=idle%20slee#actigraphs-idle-sleep-mode) in the GGIR github-pages, which is now the main documentation resource for GGIR.

### Time gap imputation

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/chapter3_QualityAssessment.html#time-gaps-in-actigraph-gt3x-and-ad-hoc-csv-files
) in the GGIR github-pages, which is now the main documentation resource for GGIR.

### The importance of reporting idle.sleep.mode usage

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/chapter3_QualityAssessment.html#time-gaps-in-actigraph-gt3x-and-ad-hoc-csv-files
) in the GGIR github-pages, which is now the main documentation resource for GGIR.

## MX metrics (minimum intensity of most active X minutes)


This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/chapter7_DescribingDataWithoutKnowingSleep.html#sets-of-quantiles-mx-metrics-by-rowlands-et-al-) in the GGIR github-pages, which is now the main documentation resource for GGIR.

## Minimum recording duration {#Minimum_recording_duration}

This section has been migrated to this [section](https://wadpac.github.io/GGIR/articles/GGIR.html?q=minimum%20recording#Minimum_recording_duration) in the GGIR github-pages, which is now the main documentation resource for GGIR.

## LUX sensor data processing

Although GGIR focuses on accelerometer data a few brands come with LUX data.

In part 1 GGIR calculates the peak lux per long epoch at a default resolution of 15 minutes,
which can be modified with argument windowsizes. Peak light offers a more reliable 
estimate of light exposure per time window compared with taking the average. Further,
LUX is used in the auto-calibration.

In GGIR part 2 we visualise the LUX values in the qc plot. In part 3 and 4 LUX is 
not used for sleep classification because relation between light exposure and sleep is weak.

In part 5 we calculate the mean and maximum of the peak LUX per epoch across all
waking hours of the day. Here, the mean (peak per epoch) LUX would then indicate
average light exposure per time segment, while max peak would indicate the maximum
light exposure per day. Further, we calculate the max and mean peak LUX per most 
active consecutive X hour of the day. This is intended to offer an alternative to LUX
exposure during waking hours which relies on correct sleep classification. LUX
exposure during M10 may be seen as an alternative if you are unsure whether you 
can trust the sleep classification in your data set.

# Other Resources

-   The [GGIR package manual](https://CRAN.R-project.org/package=GGIR)
    provides documentation on individual functions.
-   For general questions about how to use GGIR see the [GGIR discussion page](https://github.com/wadpac/GGIR/discussions).
-   For bug reports please post them
    [here](https://github.com/wadpac/GGIR/issues).

# Citing GGIR {#CitingGGIR}

A correct citation of research software is important to make your
research reproducible and to acknowledge the effort that goes into the
development of open-source software.

This section has been replaced by an updated citation instruction which you
can find via R command `citation(package = "GGIR")` and by citing the scientific
literature corresponding to the specific functionality you use as discussed in the GGIR [documentation](https://wadpac.github.io/GGIR/).

## Copyright for GGIR logo

The copyright of the GGIR logo lies with Accelting (Almere, The
Netherlands), please contact
[v.vanhees\@acceleting.com](mailto:v.vanhees@acceleting.com){.email} to
ask for permission to use this logo.

```{r, echo=FALSE, out.width = "60%", out.extra='style="border: 0; padding:20px"', fig.alt="GGIR logo"}
knitr::include_graphics("GGIR-MASTERLOGO-RGB.png")
```

---
title: "7. Describing the Data Without Knowing Sleep"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{7. Describing the Data Without Knowing Sleep}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

With the metric values imputed in the previous chapter, GGIR part 2 offers a first descriptive analysis of the data.
Although we will have to wait till later GGIR parts (chapters) to see the segmentation of days in waking and sleep hours, there is already enough that can be quantified at this point.
In this chapter we focus on descriptives of the data that are informative even without the knowledge on when the participant was sleeping or awake.

## Data quality indicators

GGIR part 2 summarises all the data quality checks done in the previous four chapters, ranging from a report on the successfulness of the auto-calibration procedure to the number of valid days.
In this way, GGIR part 2 is an ideal place to start for data quality assurance.

## Basic descriptives

Descriptive variables are calculated and reported for valid days only, where the criteria for valid day is defined by parameter `includedaycrit`.

### Average acceleration

Average acceleration is known to be correlated with the activity-related energy expenditure.
GGIR part 2 provide two types of average acceleration:

-   Average per day, only stored when the day was considered valid.
    Note that this descriptive and most other descriptives are also stored by GGIR as averages across all days, weekend days or weekday, which we will discuss in more detail.

-   Weighted average of all valid data points in the recording, weighted by timing in the day of all valid epochs, regardless of whether they come from days that as a whole are classified as valid or not.

### Acceleration distribution

To get a more detailed description of the data looking at acceleration value distribution can be informative, which we GGIR facilitates in two ways :

1.  By specifying the quantiles of the distribution with parameter `qlevels`, which are fed to build-in R function `quantile` and give us the metric values corresponding to such quantiles (a quantile multiple by 100 is the same as a percentile).

2.  By describing time spent in acceleration ranges, which are defined by parameter `ilevels` .

In the physical activity literature the distribution of acceleration values is often referred to as intensity distribution.

## Derived descriptives

### Sets of quantiles (MX metrics by [Rowlands et al.](https://pubmed.ncbi.nlm.nih.gov/29360664/))

The quantiles as discussed above can be used to describe the accelerations that participants exceed in their most active “X” accumulated minutes in a day.
In the specific approach as proposed by [Rowlands et al.](https://pubmed.ncbi.nlm.nih.gov/31808014/) these quantiles are referred to as the **MX metrics**.
The MX metrics should not be confused with the most active continuous X hours, e.g.
M10, as used in circadian rhythm research that also can be derived with GGIR (see parameter `winhr`).

To use the MX metrics as proposed by Rowlands et al., specify the durations of the 24h day that you wish to filter out the accelerations for.
For example, to generate the minimum acceleration value for the most active 30 minutes, you can call `qlevels = (1410/1440)`, which will filter out the lowest 1410 minutes of the day.
This can also be used as a nested term to generate multiple metrics, for example to call M60, M30 and M10, you can use the following parameter:

`qlevels = c(c(1380/1440), c(1410/1440), c(1430/1440))`.

Note: if parameter `qwindow` has values less than 24 h, e.g. the school day (e.g. [Fairclough et al 2020](https://pubmed.ncbi.nlm.nih.gov/31593604/)), the denominator should be changed from 1440 (24h) to the new `qwindow` window length, e.g. if an 8-hour window is used then the denominator would be 480 rather than 1440 and the numerator would be 480 minus for example 60, 30, 10 or 5.

The output in the part 2 summary files will refer to this as a percentile of the day.
Thus, for a 24-h day, M30 will appear as “p97.91666_ENMO_mg_0.24hr”.
To create the radar plots of these MX metrics as first described by [Rowlands et al.](https://pubmed.ncbi.nlm.nih.gov/31808014/), this [GitHub repository](https://github.com/Maylor8/RadarPlotGenerator) provides the R code and detailed instructions on how to make the radar plots using your own data.

### Intensity gradient

If we plot the time spent in equally spaced acceleration ranges we would end up with an asymptotic-shaped curve, indicating little time spent at high intensities (acceleration levels) and much time spent at low intensities.
The shape of the distribution may be informative but is hard to quantify with single number in its standard form.
Therefore, a new concept called the intensity gradient was proposed by [Rowlands and colleagues](https://pubmed.ncbi.nlm.nih.gov/29360664/).

The intensity gradient defines the slope of the log-transformed axes in the distribution.
More specifically, we calculate the time accumulated in incremental acceleration bins (bin size = 25 m*g*) but also keep track of the mid-point of each intensity bin, e.g. 62.5 m*g* for the bin ranging from 50 to 75 mg.
Both the mid-point acceleration of a bin expressed in m*g* and the time spent in a bin expressed in minutes are then log-transformed.
The log-transformation is expected to change the asymptotic-shaped curve into a straight line.
Subsequently, a linear regression is fitted through these data points.
The slope of this regression line represents the intensity gradient.
Further, we calculate the correlation coefficient for the data points to help verify the degree to which they form a straight line.

The intensity gradient is not calculated by default.
To include this metric in the part 2 output, set `iglevels = TRUE`, which will ensure that the algorithm is used.

Further, if you want to do more methodological research on this, you can use this parameter to define alternative acceleration bins , e.g. for using bins of 20 instead of 25 mg `iglevels = c(seq(0, 4000, by = 20), 8000)`.

## Key arguments

-   `includedaycrit`

-   `ilevels`

-   `qlevels`

-   `iglevels`

-   `qwindow`

-   `do.report`

## Related output

GGIR part 2 generates three csv reports: part2_daysummary.csv, part2_summary.csv, and data_quality_report.csv.
As data_quality_report.csv was discussed in chapter 3 we will focus here only on the first two reports.

The variables in part2_summary.csv are the recording level aggregates of the variables in part2_daysummary.csv.
Here, variable names starting with the “AD\_” refer to average across all days, “WD” refers to the average across weekdays, “WE” refers to the average across weekend days, “WWE” refers to weighted weekend days to ensure both weekend days contribute equally, “WWD” refers to weighted weekdays to ensure all weekdays contribute equally.

Further, GGIR part 2 generates a report named part2_daysummary_longformat.csv, which is generated when GGIR is used for day segment analysis, see documentation for parameter `qwindow`.
This report contains the exact same information as part2_daysummary.csv, yet in long format.
In part2_daysummary_longformat.csv, each row represents one segment from one day in one recording, while in part2_daysummary.csv, each row contains one day in one recording and the segments of the day are organized in different columns.

### Descriptive variables

| (Part of) variable name         | Description                                                                                                                                                                                                          | Report(s) |
|--------------------|----------------------------------|------------------|
| mean_ENMO_mg_0-24hr             | average acceleration defined by metric ENMO for the full day                                                                                                                                                         | `b`       |
| ENMO_fullRecordingMean          | Weighted average over acceleration across recording                                                                                                                                                                  | `r`       |
| p20_ENMO_mg_0-24h               | 20th percentile based on the full day                                                                                                                                                                                | `b`       |
| p20_ENMO_mg_0-24h_fullRecording | 20th percentile based on the weighted average day from the full recording                                                                                                                                            | `r`       |
| [0,50)\_ENMO_mg_0-24hr          | time spent in minutes between 0 and 50 m_g\_ acceleration defined by metric ENMO for the full day. Square bracket means that the value itself is included, round bracket indicated that value itself is not included | `b`       |
| ig_gradient_ENMO_0-24hr         | Gradient for the intensity gradient                                                                                                                                                                                  | `b`       |
| ig_intercept_ENMO_0-24hr        | Intercept for the intensity gradient                                                                                                                                                                                 | `b`       |
| ig_rsquared_ENMO_0-24hr         | R-squared for the intensity gradient                                                                                                                                                                                 | `b`       |

: To clarify that `b` refers to both part2_summary.csv and part2_daysummary.csv, `r` refers to part2_summary.csv only.

---
title: "Reading csv files with raw data in GGIR"
output:
   html_document:
    toc : true
    number_sections: true
    toc_depth: 3
    toc_float: true 
urlcolor: blue
vignette: >
  %\VignetteIndexEntry{Reading csv files with raw data in GGIR}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
```

**NOTE: If you are viewing this page via CRAN note that the main GGIR documentation has been migrated to the [GGIR GitHub pages](https://wadpac.github.io/GGIR/).**

# Introduction

GGIR can automatically read data from the most-frequently used accelerometer brands in the field:

- [GENEActiv](https://activinsights.com/) .bin
- [Axivity](https://axivity.com/) AX3 and AX6 .wav, .csv and .cwa
- [ActiGraph](https://ametris.com/) .csv and .gt3x (.gt3x only the newer format generated with firmware versions  above 2.5.0). Note for Actigraph users: If you want to work with .csv exports via the ActiLife then note that you have the option to export data with timestamps. Please do not do this as this causes memory issues for GGIR. To cope with the absence of timestamps GGIR will re-caculate timestamps from the sample frequency and the start time and date as presented in the file header
- [Movisens](https://www.movisens.com/en/) with data stored in folders
- Genea (an accelerometer that is not commercially available anymore, but which was used for some studies between 2007 and 2012) .bin and .csv

However, the accelerometer manufacturers are proliferating with an increasing number of brands in the market. For such reason, GGIR includes the `read.myacc.csv` function, which is able to read accelerometer raw triaxial data stored in csv files, independently of the brand. This vignette provides a general introduction on how to use GGIR to read accelerometer raw data stored in csv files.

**How it works:**

Internally GGIR loads csv files with accelerometer data and standardises the output format to make the data compatible with other GGIR functions. Data format standardisation includes unit of measurement, timestamp, file header format, and column locations.

# The read.myacc.csv function

As for the rest of GGIR functions, `read.myacc.csv` is intended to be used within function `GGIR`. All the arguments of the `read.myacc.csv` can be easily recognized as they all start by "rmc". GGIR checks whether argument `rmc.firstrow.acc` is provided by the user; in such case, GGIR will attempt to read the data with function `read.myacc.csv`. In other words you always need to specify `rmc.firstrow.acc` to use `read.myacc.csv`. Further, we recommend that you always first test the function argument settings by first trying to use function `read.myacc.csv` on its own. When that works copy the arguments to you GGIR call.

## Input arguments

As the `read.myacc.csv` function tries to read csv files with a wide variety of formats, the key arguments to specify depend on the characteristics of the csv file to process. Overall, if an argument is not relevant, it should be left in default setting (e.g., if the csv file does not contain temperature data, the arguments related to temperature settings should be left in default values).

Below we present a summary of the available input arguments. Please see the [parameters vignette](https://cran.r-project.org/package=GGIR/vignettes/GGIRParameters.html) for a more elaborate description of these input arguments. Further, the arguments are also covered by the function documentation for the `read.myacc.csv` function. 

### General arguments

- `rmc.file` - Filename of file to be read if it is in the working directory, or full path to the file otherwise.
- `rmc.nrow` - Number of rows to read, same as nrow argument in \link[utils]{read.csv} and nrows in \link[data.table]{fread}. The whole file is read by default (i.e., rmc.nrow = Inf).
- `rmc.skip` - Number of rows to skip, same as skip argument in \link[utils]{read.csv} and in \link[data.table]{fread}.
- `rmc.dec` - Decimal separator used for numbers, same as dec argument in \link[utils]{read.csv} and in data.table::\link[data.table]{fread}. If not "." (default) then usually ",".
- `rmc.firstrow.acc` - First row (number) of the acceleration data.
- `rmc.unit.acc` - Character with unit of acceleration values: "g", "mg", or "bit".
- `desiredtz` - Timezone in which device was worn.
- `confgitz` - Timezone in which device was configured.
- `rmc.sf` - Sample rate in Hertz, if this is stored in the file header then that will be used instead.

### Arguments for files containing a header {#header}

- `rmc.firstrow.header` - First row (number) of the file header. Leave blank (default) if the file does not have a file header. Not to be confused with a one row column header, a file header typically takes up several rows and has one or two columns. If the header has two columns, the first column is assumed to be the header item names and the second column is assumed to have the header item values. If the header has one column then it is assumed that each value contains both the name and the value of the item.
- `rmc.header.length` - If file has header, specify header length (numeric).
- `rmc.headername.sf` - If file has a header, row name (character) under which the sample frequency can be found, e.g. "sample_rate".
- `rmc.headername.sn` - If file has a header, row name (character) under which the serial number can be found, e.g. "serial_number".
- `rmc.headername.recordingid` -  If file has a header, row name (character) under which the recording ID can be found, e.g. "ID".
- `rmc.header.structure` - Character used to split the header name from the header value, e.g. ":" if a header value would look like "ID: 123" or " " if the ehader value would be like "ID 123".

### Arguments for files including timestamps

- `rmc.col.time` - Scalar with column (number) in which the timestamps are stored. Leave in default setting if timestamps are not stored.
- `rmc.unit.time` - Character with unit of timestamps: "POSIX", "UNIXsec" (seconds since origin, see argument rmc.origin), "UNIXmsec" (milliseconds since origin, see argument rmc.origin), "character", or "ActivPAL" (exotic timestamp format only used in the ActivPAL activity monitor).
- `rmc.format.time` - Character string giving a date-time format as used by \link[base]{strptime}. Only used for rmc.unit.time: character and POSIX.
- `rmc.origin` - Origin of time when unit of time is UNIXsec, e.g. 1970-1-1.

### Arguments for files with acceleration stored in bits

- `rmc.bitrate` - Numeric: If unit of acceleration is a bit then provide bit rate, e.g. 12 bit.
- `rmc.dynamic_range` - Numeric, if unit of acceleration is a bit then provide dynamic range deviation in g from zero, e.g. +/-6g would mean this argument needs to be 6. If you give this argument a character value the code will search the file header for elements with a name equal to the character value and use the corresponding numeric value next to it as dynamic range.
- `rmc.unsignedbit` - Boolean, if unsignedbit = TRUE means that bits are only positive numbers. If unsignedbit = FALSE then bits are both positive and negative.

### Arguments for files including temperature

- `rmc.col.temp` - Scalar with column (number) in which the temperature is stored. Leave in default setting if no temperature is avaible. The temperature will be used by \link{g.calibrate}.
- `rmc.unit.temp` - Character with unit of temperature values: (K)elvin, (C)elsius, or (F)ahrenheit.

### Arguments for files including wear time information

- `rmc.col.wear` - If external wear detection outcome is stored as part of the data then this can be used by GGIR. This argument specifies the column in which the wear detection (Boolean) is stored.

### Arguments to find time gaps and resampling

- `rmc.check4timegaps` - Boolean to indicate whether gaps in time should be imputed with most recent value normalised to 1 _g_.
- `rmc.doresample` - Boolean to indicate whether to resample the data based on the available timestamps and extracted sample rate from the file header
- `interpolationType` - Integer to indicate type of interpolation to be used when resampling time series (mainly relevant for Axivity sensors), 1=linear, 2=nearest neighbour.

# Usage of the read.myacc.csv function

This section shows an example real case in which the read.myacc.csv function can be used. The csv file to be read has the following structure:

dateTime | acc_x | acc_y | acc_z | ambient_temp
---|---|---|---|---
1/1/2022 16:48:26.000|-0.42041016|0.41114536|-0.76733398|24
1/1/2022 16:48:26.009|-0.41674805|0.40919218|-0.76611328|24
1/1/2022 16:48:26.019|-0.42407227|0.40845973|-0.77539062|24
1/1/2022 16:48:26.029|-0.41894531|0.41163366|-0.77246094|24
1/1/2022 16:48:26.039|-0.4206543|0.41749321|-0.76806641|24
1/1/2022 16:48:26.049|-0.42163086|0.42091128|-0.76757812|24
1/1/2022 16:48:26.059|-0.42236328|0.41627247|-0.76391602|24
1/1/2022 16:48:26.069|-0.42431641|0.4189581|-0.76171875|24
1/1/2022 16:48:26.079|-0.42138672|0.41993469|-0.76513672|24

This file contains timestamps in the column 1 (formatted as "%d/%m/%Y %H:%M:%OS"), the acceleration signals (in _g_'s) for the x, y, and z axis in the columns 2, 3, and 4, respectively, and temperature information in Celsius in the column 5. Also, this file has no file header.

Before we can use this with GGIR, we first test read this file using the `read.myacc.csv` function directly.

```{R,eval=FALSE}
library(GGIR)
data = read.myacc.csv(rmc.file = "C:/mystudy/mydata/datafile.csv",
               rmc.nrow = Inf,
               rmc.skip = 0,
               rmc.dec = ".",
               rmc.firstrow.acc = 2,
               rmc.col.acc = 2:4,
               rmc.col.temp = 5,
               rmc.col.time=1,
               rmc.unit.acc = "g",
               rmc.unit.temp = "C",
               rmc.unit.time = "POSIX",
               rmc.format.time = "%d/%m/%Y %H:%M:%OS",
               desiredtz = "Europe/London",
               rmc.sf = 100)
```

The object data is a list with a data.frame name data and a header.
The time column in the data.frame represents timestamps expressed in seconds since 1-1-1970.

## Example using the shell function

If the `rmc.firstrow.acc` argument is defined within the `GGIR` function, then the data will be read through `read.myacc.csv`. GGIR needs the user to specify in which row starts the accelerometer data within the csv, so this argument must be always explicitly specified by the user. Thus, a call to the `GGIR` including the rmc arguments would look as follows (note that the `rmc.file`, `rmc.nrow`, and `rmc.skip` arguments will not be used by `GGIR` as these arguments are already defined by `datadir`, `strategy`, and [header](#header) arguments, respectively).

```{R,eval=FALSE}
library(GGIR)
GGIR(
             mode=c(1,2,3,4,5),
             datadir="C:/mystudy/mydata/datafile.csv",
             outputdir="D:/myresults",
             do.report=c(2,4,5),
             #=====================
             # read.myacc.csv arguments
             #=====================
             rmc.nrow = Inf, 
             rmc.dec = ".",
             rmc.firstrow.acc = 2, 
             rmc.col.acc = 2:4, 
             rmc.col.temp = 5, 
             rmc.col.time=1,
             rmc.unit.acc = "g", 
             rmc.unit.temp = "C", 
             rmc.unit.time = "POSIX",
             rmc.format.time = "%d/%m/%Y %H:%M:%OS",
             desiredtz = "Europe/London",
             rmc.sf = 100,
             rmc.noise = 0.013
)
```





```{r, echo=FALSE, out.width = "60%", out.extra='style="border: 0; padding:20px"', fig.alt="GGIR logo"}
knitr::include_graphics("GGIR-MASTERLOGO-RGB.png")
```

---
title: "Household Co-Analysis"
output:
   html_document:
    toc: true
    number_sections: true
    toc_depth: 3
    toc_float: true
urlcolor: blue
vignette: >
  %\VignetteIndexEntry{Household Co-Analysis}
  \usepackage[utf8]{inputenc}
  %\VignetteEngine{knitr::rmarkdown}
---


```{r, echo=FALSE, out.width = "60%", out.extra='style="border: 0; padding:20px"'}
knitr::include_graphics("GGIR-MASTERLOGO-RGB.png")
```

See also complementary vignettes on:
[General introduction to GGIR](https://cran.r-project.org/package=GGIR/vignettes/GGIR.html), [Day segment analyses](https://CRAN.R-project.org/package=GGIR/vignettes/TutorialDaySegmentAnalyses.html), [GGIR parameters](https://CRAN.R-project.org/package=GGIR/vignettes/GGIRParameters.html), [Embedding external functions (pdf)](https://CRAN.R-project.org/package=GGIR/vignettes/ExternalFunction.pdf), and [Reading ad-hoc csv file formats](https://CRAN.R-project.org/package=GGIR/vignettes/readmyacccsv.html).


Household co-analysis facilitates the analysis of how sleep and physical activity relate
between members of the same household.

# Accelerometer file names

The household co-analysis requires that households and family member can be recognised.
To do this we assume the following logic in the file names:

`StudyNumber-HouseholdID-MemberID_anyotherinformation.bin`

This example is for a .bin file, but the same applies to .cwa or .csv files.

For example the files:

- `001-002-001_12345-2023.bin`
- `001-002-002_23456-2023.bin`
- `001-002-M_23456-2023.bin`

will be recognised as household ID `002` with member IDs `001`, `002`, and `M`.

# Configuring GGIR

The following input arguments are needed to run the Household co-analysis:

- The Household co-analysis is integrated in GGIR as part 6, so we need to run GGIR with
part 1 to 6, with `mode = 1:6`.
- `windowsizes = c(5, 60, 3600)`. Setting the second value to 60 ensures that light and 
temperature, if available, are aggregated per minute.
- `part5_agg2_60seconds = TRUE`. This ensures that GGIR part 5 stores
the time series at 1 minute resolution.
- `part6HCA = TRUE` to tell GGIR to perform Household Co-Analysis.
- `part6_threshold_combi = "30_100_400"` where 30, 100 and 400 need to correspond
to the accelerometer threshold combination that is used in part 5 that you want to use
in part 6. GGIR part 5 facilitates multiple threshold combinations but in part 6 you
need to select one.

Other GGIR arguments can be set according to your own needs. For an example:

```
datadir = "C:/projects/studyZ/binfiles"
outputdir = "C:/projects/studyZ"
library(GGIR)
GGIR(mode = 1:5, 
     datadir = datadir,
     idloc = 2, 
     outputdir = outputdir,
     do.report = c(2, 4, 5),
     do.parallel = TRUE,
     overwrite = FALSE,
     printsummary = TRUE,
     desiredtz = "America/Halifax",
     windowsizes = c(5, 60, 3600),
     threshold.lig = 30,
     threshold.mod = 100,
     threshold.vig = 400,
     part6_threshold_combi = "30_100_400",
     boutcriter.in = 1, boutcriter.lig = 1,  boutcriter.mvpa = 0.9,
     boutdur.in = 30,
     boutdur.lig = 10,
     boutdur.mvpa = 5,
     part6HCA = TRUE,
     save_ms5rawlevels = TRUE, # Not necessary because GGIR will set this to TRUE when part6HCA is TRUE.
     save_ms5raw_without_invalid = FALSE, # <= Needed for household co-analysis
     part5_agg2_60seconds = TRUE,
     visualreport = FALSE) 
     
```

# Household co-analysis

In GGIR part 1, 2, 3, 4, and 5 each recording is processed individually without
considering relations between recordings. Next, part 6 is subdivided in alligning
the time series produced by part 1 and 5 per household, and the pairwise analysis of the data.

## Align individuals

Household members with only one member are ignored.
Next, per household and per household member the code loads and merges the time
series produced by GGIR part 1 and part 5. Days, defined from waking-up to
waking-up, are removed if they have less than 20% valid data during waking hours,
the sleep period time window, or the day as a whole. Next, time series are completed
with indicates of valid household member pairs for each time points.

Finally, we store:

- The aligned time series per household in separate csv files in the GGIR
output directory (`.../results/part6HouseholdCoAnalysis/alignedTimeseries`). The columns
for this file are documented further down.
- A pdf file names timeseriesPlot.pdf with plots of the aligned time series to 
facilitate visual inspection of the data completeness per household.


## Pairwise analysis

Per household we identify all possible member pairs and loop over these pairs.

Per member pair the code identify wake-up time pairs. Here, wake-up times that occur 
within the last 15 minutes of the time series are ignored as we need at least some 
recording time to quantify behaviour after waking up.

Per wake-up pair we assess who woke up first and second, the time difference, and
the corresponding calendar dates of waking up.

Next, the code quantifies:

- Activity of person who first woke up during minute before second person woke up
- Activity of second person to wake up before they woke up
- LUX of person who first woke up during tminute before second person woke up
- LUX of second person to wake up before they woke up.

Describe matching waking hours between pairs:

- Correlation between continuous acceleration values (ENMO metric)
- Derive binary class of inactivity/active (ENMO metric, and threshold < 50)
- ICC based on binary scores (irr package, model=twoway, type=agreement, unit=single)
- Cohen’s Kappa (psych package)
- Similarity in binary scores (calculation in line with Sleep Regularity Index)

Describe noon-noon window with a stronger focus on sleep:

-	Describe binary class of sleep/wakefulness (note: does not attempt to classify daytime naps) 
-	ICC based on binary scores (irr package, model=twoway, type=agreement, unit=single)
-	Cohen’s Kappa (psych package)
-	Similarity in binary scores (calculation in line with Sleep Regularity Index)

Describe wakefulness dynamics during the SPT prior to wakeup:

-	Look up indices of spt prior to wakeup where both individuals where in SPT.
-	Assess fraction of data valid
-	Identify wake up times during the night

For each wake-up time:

-	Assess whether both persons woke up at the same time, other person wake up within 5 minutes, or other person does not wake up within 5 minutes.
-	Store output to csv one row per unique household pair, with columns to clarify who the household members are in the pair and from which household they are.


# Output variables

## alignedTimesieres

In the GGIR output folder `.../results/part6HouseholdCoAnalysis/alignedTimeseries`
you will find csv files with the time series per household. 

The data dictionary below shows the column names you would get for a household with
two members: X and Y. Most of the columns are copied from the time series output files, which
are documented [here](https://cran.r-project.org/package=GGIR/vignettes/GGIR.html#571_Time_series_output_files).
Therefore, those column are not documented here.

Variable name | Description
---- | ----------
timenum | Timestamp for tSeconds since since 1-1-1970
HID | Household ID
ACC.X | Average magnitude of acceleration for household member X
SleepPeriodTime.X | -
invalidepoch.X | -
guider.X | -
window.X | -
class_id.X | -
lightmean.X | Mean light sensor value for member X
lightpeak.X | Peak light sensor value for member X
temperaturemean.X | Mean temperature value for member X
onset.X | 1 to indicate sleep onset time for member X
wakeup.X | 1 to indicate waking up for member X
validepoch.X | 1 if epoch was valid
ACC.Y | Average magnitude of acceleration for household member X
SleepPeriodTime.Y | -
invalidepoch.Y | -
guider.Y | -
window.Y | -
class_id.Y | -
lightmean.Y | See lightmean.X but now for member Y
lightpeak.Y | See lightpeak.X but now for member Y
temperaturemean.Y | -
onset.Y | -
wakeup.Y | -
validepoch.Y |	 -
time_POSIX | Timestamp in character format
N_valid_hhmembers | Number of valid household member pairs
validpair_X_Y | TRUE or FALSE to indicate whether the pair had valid data for this timestamp


## Pairwise summary report

As stored inside `pairwise_summary_all_housholds.csv`

Variable name | Description
---- | ----------
HID	| Household ID
Npairs	| Number of unique pairs in a household
PID	| Pair ID
MID1	| Household member ID for the first person in the pair
MID2	| Household member ID for the second person in the pair
event	| Number to count the number of events, for now these are days with valid wake-up estimates in both individuals
time1	| Time person 1 wakes up
time2	| Time person 2 wakes up
date1	| Date person 1 wakes up
date2	| Date person 2 wakes up
index_wake1	| Index in the time series when person 1 woke up (probably not relevant for end-user)
index_wake2	| Index in the time series when person 2 woke up (probably not relevant for end-user)
firstawake | Who of the two individuals woke up first? If equal then value will be "equal"
wakeup_acc_wake1_before_wake2_mg | Mean acceleration (m*g*) of the person who wakes up first during the four minutes before the second person wakes up
wakeup_acc_wake2_before_wake2_mg | Mean acceleration (m*g*) of second person to wake up before they woke up
wakeup_lux_wake1_before_wake2 | Peak LUX of the person who wakes up first during the four minutes before the second person wakes up
wakeup_lux_wake2_before_wake2 | Peak LUX of second person to wake up before they woke up
wakeup_deltatime_min	| Difference in wake up time in minutes
r_acceleration	| Pearson correlation in continuous acceleration values
ICC_sleep	| ICC (from R package irr) calculated based on binary scores of sleep (excluding WASO)
ICC_sleep_Fvalue	| Fvalue for ICC_sleep
ICC_active | ICC (from R package irr) calculated based on binary scores of activity (metric ENMO < 50 chosen for now)
ICC_active_Fvalue | Favlue for ICC_active
frac_valid | Fraction of pair epochs between individuals for a certain day that were valid
SleepSimilarityIndex | Similarity in binary sleep/wake patterns between pairs per calendar day
ActivitySimilarityIndex	| Similarity in binary active/inactive patterns between pairs per calendar day
Kappa_sleep	| Weighted Cohen's Kappa coefficient calculated based on binary scores of sleep (excluding WASO)
Kappa_sleep_CI_lower | Lower confidence interval of Kappa_sleep
Kappa_sleep_CI_lupper | Upper confidence interval of Kappa_sleep
Kappa_active | Weighted Cohen's Kappa coefficient calculated based on binary scores of activity (metric ENMO < 50 chosen for now)
Kappa_active_CI_lower | Lower confidence interval of Kappa_active
Kappa_active_CI_upper | Upper confidence interval of Kappa_active

---
title: "Day segment analyses with GGIR"
output:
   html_document:
    toc : true
    number_sections: true
    toc_depth: 3
    toc_float: true #by turning this on, the table of contents moves to the left of the page.
urlcolor: blue
vignette: >
  %\VignetteIndexEntry{Day segment analyses with GGIR}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
```

**NOTE: If you are viewing this page via CRAN note that the main GGIR documentation has been migrated to the [GGIR GitHub pages](https://wadpac.github.io/GGIR/).**

# Introduction

Is this specific person more active in the morning or in the afternoon? Are children 
more active during their work hours or their leisure time? How much inactivity occurs
at work in office workers? Questions like these can be answered with GGIR but you 
first have to specify a few parameters.

The main input argument to be specified is `qwindow`, which can be used the following ways:

- To specify the clock hours in the day based on which the segmented day analyses should take place.
- To specify an activity log (diary) that should be used to guide the segmentation per individual and per day of the recording.

In the following sections I will discuss both scenarios.

# Clock hour-based segmentation

To perform clock hour segmentation, you will need to provide function GGIR with
argument `qwindow` and assign it a numeric vector with the hours for the segmentation.
If the start and end of the day, are not explicitly provided in the 
vector GGIR will add them.  Please find below some example values for `qwindow`.
The number of values used by `qwindow` is unlimited, but be aware that some of the analyses are such as MX-metrics are
impossible for very small windows and will produce empty results.

qwindow value | Resulting segment(s) to be analysed
-------------------| -------------------
c(0,24) | midnight to following midnight (24 hours), the full day is the only segment.
c(8,24) | midnight-8:00 (8 hour segment), 8:00-midnight (16 hour segment), and midnight-midnight (24 hour segment).
c(6,11, 13, 17) | midnight-6:00 (6 hour segment), 6:00-11:00 (5 hour segment), 11:00-13:00 (2 hour segment), 13:00-17:00 (4 hour segment), 17:00-midnight (7 hour segment), and midnight-midnight (24 hour segment).
c(0:24) | 25 segments: 24 segments of 1 hour corresponding to each hour of the day, and midnight-midnight (24 hour segment).    

Day Saving Time (DST) is taken into account when identifying the start of the day, but
not when identifying the day segments. In other words, a 23 hour days is processed as the 24 hours after the first midnight.
This to ensure that segment length is identical across days of the week, which is needed to ease comparison of outcome variables across days.


# Segmentation guided by activity log

To perform activity-log based segmentation, you will need to provide function `GGIR` with
argument `qwindow` and assign it the full path to your activity log in `.csv format`, e.g. `qwindow="C:/myactivitylog.csv"`.

The activity log is expected to be a .csv-file with the following structure:

ID | date | work | travelhome | home | date | work | travelhome | home 
-----| -------- | ------- | --------- | ----- | -------- | ------- | -------- | --------
1234 | 04-11-20 | 7:45:00 |  17:00:00 | 17:30:00 | 05-11-20 |  |  | 17:30:00
4567 | 24-11-20 | 7:45:00 |  17:00:00 | 17:30:00 | 25-11-20 | 7:45:00 |  17:00:00 | 17:30:00

**Rows:**
First row represents the column headers after which each row represents one accelerometer recording.

**ID-column:**
The first column is expected to hold the recording ID, which needs to match with the ID GGIR extracts from the accelerometer file. If unsure how to format the ID values, apply GGIR to a sample of your accelerometer files using the default argument settings. The ID column in the generated part 2 .csv reports will show how the participant ID is extracted by GGIR. If no ID is extracted, see documentation for argument `idloc`, which helps you to specify the location of the participant in the file name or file header. If ID extraction fails the accelerometer files cannot be matched with the corresponding activity log entries.

**Date-column:**
The ID column is followed by a date column for the first log day. To ensure GGIR recognises this date correctly, specify argument `qwindow_dateformat`. The default format is `"\%d-\%m-\%Y"` as in 23-2-2021 to indicate the 23rd of February 2021. If your date is formatted as 2-23-21 then specify`"\%m-\%d-\%y"`. The column name of the date column needs to include the character combination "date" or "Date" or "DATE". Use the same date format consistently throughout your activity diary.

**Start-times:**
The date column is followed by one or multiple columns with start times for the activity types in that day format in hours:minutes:seconds. Do not provide dates in these cells. The header of the column will be used as label for each activity type. Insert a new date column before continuing with activity types for next day. Leave missing values empty.

**Missing values:**
If values are missing the preceding and succeeding time point will be used as the edges of the segment.
In the example below this means that we will define a segment from `A-C` for ID `1234`, while for ID `6789` we only defined segments `A-B` and `B-C`, segment `A-C` is not derived here.

ID | date | A | B | C 
-----| -------- | ------- | --------- | ----- 
1234 | 04-11-20 | 7:45:00 |           | 17:30:00 
6789 | 24-11-20 | 7:45:00 |  13:00:00         | 17:30:00 

**Notes:**
- If an activity log was collected for some individuals then those will be processed with qwindow value c(0,24).
- Dates with no activity log data can be skipped, no need to have a column with the date followed by a column with the next date.
- The end time of one activity is assumed to be the start time of the next activity. We currently do not facilitate overlapping time segments.

# Implementation in GGIR

Both approaches are implemented in GGIR part 2 and part 5. Therefore the specific
output variables that are calculated both in part 2 and 5 are available per day,
per person, and per segment of the day based on the argument `qwindow` Note that 
`qwindow` is only used in **part 5** when `timewindow` includes `"MM"` (see specific
documentation for `timewindow`} in the [parameters vignette](https://CRAN.R-project.org/package=GGIR/vignettes/GGIRParameters.html))

At the moment, specifying the argument `qwindow` triggers the calculation of the
`qwindow` segments both in part 2 and part 5, which may result in a longer time
to finish the analysis. If only interested in the segments in either part 2 or part 5,
an option might be to run GGIR parts 1:2 with the argument `qwindow` of interest,
and then set `qwindow = NULL` and run GGIR parts 3:5 (or vice versa: `qwindow = NULL`
for GGIR parts 1:2, and then the desired `qwindow` segments when running GGIR parts 3:5).

For more information about the output variables calculated in each part
of the pipeline, see the main [GGIR vignette](https://CRAN.R-project.org/package=GGIR/vignettes/GGIR.html).


# Examples

For more information about how to use the GGIR function call
see explanation in the main [GGIR vignette](https://CRAN.R-project.org/package=GGIR/vignettes/GGIR.html).

## Clock-hour based segmentation:

```{R,eval=FALSE}
library("GGIR")
GGIR(datadir = "/your/data/directory",
             outputdir = "/your/output/directory",
             mode = 1:5, # <= run GGIR parts 1 to 5
             do.report = c(2, 5), # <= generate csv-report for GGIR part 2 and part 5
             qwindow = c(0, 6, 12, 18, 24),
             timewindow = "MM")
```


## Activity log based segmentation:

```{R,eval=FALSE}
library("GGIR")
GGIR(datadir = "/your/data/directory",
             outputdir = "/your/output/directory",
             mode = 1:5, # <= run GGIR parts 1 to 5
             do.report = c(2, 5), # <= generate csv-report for GGIR part 2 and part 5
             qwindow = "/path/to/your/activity/log.csv",
             timewindow = "MM")
```


After running this code GGIR creates an output folder in the output directory as 
specified with argument outputdir. In the subfolder `results` you will then find 
csv files with the reports generated in part 2 and part 5 of the pipeline:

**Part 2**

- `part2_summary.csv` the recording level summary, with 1 row per recording and in recording level aggregates of day segments in columns. 
- `part2_daysummary.csv` the day level summary, with 1 row per day and day segment specific outcomes in columns.
- `part2_daysummary_longformat.csv` the day level summary in long format, such that each row represents one segment from one day in one recording.

In both `part2_summary.csv` and `part2_daysummary.csv` the column names tell you the day segment they correspond to. For example, column names ending with `_18-24hr` refer to the time segment 18:00-24:00. In `part2_daysummary_longformat.csv` the time segment is clarified via columns qwindow_timestamps and qwindow_name.

**Part 5**

In part 5, information about the segments of the days are exported in different csv
reports than the person-level and day-level summaries. These files include the 
word "Segments" in the filename and are provided in the long format and 
aggregated per day and per person:

- `part5_daysummary_Segments[...].csv` the day level summary in long format, such that each row represents one segment from one day in one recording.
- `part5_personsummary_Segments[...].csv` the recording level summary in long format, such that each row represents the average for each 
outcome in one specific segments across all days in which that segment is available per participant.


# Cleaning parameters for day segments (in part 5):

In part 5, the analyses performed per segment of the day come with the possibility to 
clean the reports based on the information available in the segments. The users
can select to include only those segments with a given amount of wear time
during the segment (`segmentWEARcrit.part5`), as well as with a given 
awake time or sleep period time in the segment (`segmentDAYSPTcrit.part5`). 

These arguments are likely to be critical for a meaningful analysis of the data. 
The presence of sleep in a segment with physical activity will bias physical inactivity 
estimates and the presence of physical activity in a segment with sleep will bias 
sleep estimates. It will then become impossible to quantify whether it was the lack
of one or the presence of the other behaviour that drives the association with
for example a health outcome.


# Analyses performed per day segment

The analyses that GGIR per segment of the day, include:

**Acceleration distribution (in part 2):**
Derived if argument `ilevels` is specified. You will find these under the variable names such as `[0,36)_ENMO_mg` which means time spent between 0 and 36 mg defined by acceleration metric ENMO.

**Number of valid hours of data (in part 2):**
You will recognise these as `N_valid_hours_in_window` which tells you the number of valid hours per time window, and `N_valid_hours` which is the number of valid hours per day.

**Non-wear time percentage (in part 5):**
`nonwear_day_perc`, `nonwear_spt_perc`, and `nonwear_day_spt_perc` tell you the 
proportion of the segment classified as non-wear during  awake time (day) and during
sleep period time (spt).

**LXMX analysis (in part 2 and part5):**
LXMX analysis, which stands for least and most active X hours of the segment. You will recognise these variable names
like `L5hr_ENMO_mg` which is the start time of the least active five hours defined by metric ENMO, and `L5_ENMO_mg` which is the average acceleration for those hours.

**Intensity gradient analysis (in part 2 and part 5):**
You will find these as variables that start with `ig_gradient_` See [description](https://cran.r-project.org/package=GGIR/vignettes/GGIR.html#41_Output_part_2) of GGIR part 2 output in the main GGIR vignette for further details.

**Time spent in Moderate or Vigorous Physical Activity (MVPA) (in part 2 and part 5):**
You will find these as variables such as `MVPA_E5S_T201_ENMO` or `MVPA_E5S_B1M80%_T201_ENMO`. See [description](https://cran.r-project.org/package=GGIR/vignettes/GGIR.html#41_Output_part_2) of GGIR part 2 output in the main GGIR vignette for further details.

**Time spent in sleeping, in inactivity and physical activity intensities (part 5):**
You will find these variables in the part 5 reports, in their bouted, unbouted, and
total time version of the variables. See [description](https://cran.r-project.org/package=GGIR/vignettes/GGIR.html#41_Output_part_5) of GGIR part 5 output in the main GGIR vignette for further details.


```{r, echo=FALSE, out.width = "60%", out.extra='style="border: 0; padding:20px"', fig.alt="GGIR logo"}
knitr::include_graphics("GGIR-MASTERLOGO-RGB.png")
```

---
title: "5. Accounting for Study Protocol"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{5. Accounting for Study Protocol}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  markdown: 
    wrap: 72
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

When you use an accelerometer in a study, you are likely to give your
study participants specific instructions on when they should start
wearing the accelerometer, for how many days, and whether or not they
are expected to take the accelerometer off during specific activity
types or parts of the day. Further, it may be that you turned on the
accelerometers hours or even days before you gave it to the participant
or stopped it hours or days after you received it back. Your knowledge
about all these aspects of your study protocol can be used by GGIR to
mask certain periods of time in the recording. This is important because
this information is not necessarily obvious from the recorded data. For
instance, when a recording is started and dispatched to the participant
via mail, the time during which the devices are in transit and not worn
may be impossible to distinguish from a participant wearing the
accelerometer and commuting.

## Selecting/Masking the data

It is important that GGIR analyses masks all data outside the time
window for which the participant was instructed to wear the
accelererometer. Study protocols differ in duration and expected wear
period, which is why GGIR offers a variety of ways to account for study
protocol.

The main parameter to do this is `data_masking_strategy`, and it
requires a numeric value indicating one of the following strategies:

-   **data_masking_strategy = 1** to indicate the a specific number of
    hours should be masked from the start and/or the end of the
    recording, specified with parameters `hrs.del.start` and
    `hrs.del.end`, respectively.

-   **data_masking_strategy = 2** to indicate that only the data between
    the first and the last midnight in the recording should be
    considered.

-   **data_masking_strategy = 3** to indicate that only the most active
    X 24-h blocks starting any time in the day should be used, where X
    is specified by parameter `ndayswindow`. Note this can be combined
    with beforementioned parameters `hrs.del.start` and `hrs.del.end`,
    which will trim this window at the start and end.

-   **data_masking_strategy = 4** to indicate that only the data after
    the first midnight should be considered.

-   **data_masking_strategy = 5** is similar to data_masking_strategy =
    3, yet it selects X complete calendar days where X is specified by
    parameter `ndayswindow`.

Additionally, the maximum duration after the start of the recording that
the accelerometer is expected to be worn can also be set with parameter
`maxdur` for number of 24 hour blocks or parameter `max_calendar_days`
for number of calendar days.

## Key parameters

-   `data_masking_strategy`

-   `hrs.del.start`

-   `hrs.del.end`

-   `ndayswindow`

-   `maxdur`

-   `max_calendar_days`

## Related output

| (Part of) variable name                               | Description                                                                                                                                                                                     | Report(s)         |
|-------------------------------------------------------|-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|-------------------|
| data exclusion strategy                               | A log of the decision made when calling g.impute: value=1 mean ignore specific hours; value=2 mean ignore all data before the first midnight and after the last midnight                        | part2_summary.csv |
| n hours ignored at start of meas                      | Number of hours ignored at the end of the measurement (if data_masking_strategy = 1) or at the end of the ndayswindow (if data_masking_strategy = 3 or 5). A log of decision made in part2.R    | part2_summary.csv |
| n hours ignored at end of meas                        | Number of hours ignored at the start of the measurement (if data_masking_strategy = 1) or at the start of the ndayswindow (if data_masking_strategy = 3 or 5) A log of decision made in part2.R | part2_summary.csv |
| n days of measurement after which all data is ignored | Number of days of measurement after which all data is ignored (if data_masking_strategy = 1, 3 or 5) A log of decision made in part2.R                                                          | part2_summary.csv |

---
title: "GGIR configuration parameters"
output:
   html_document:
    toc : true
    number_sections: true
    toc_depth: 3
    toc_float: true #by turning this on, the table of contents moves to the left of the page.
urlcolor: blue
vignette: >
  %\VignetteIndexEntry{GGIR configuration parameters}
  \usepackage[utf8]{inputenc}
  %\VignetteEngine{knitr::rmarkdown}
---

**NOTE: If you are viewing this page via CRAN note that the main GGIR documentation has been migrated to the [GGIR GitHub pages](https://wadpac.github.io/GGIR/).**

The GGIR
[shell function](https://cran.r-project.org/package=GGIR/vignettes/GGIR.html#23_GGIR_shell_function)
takes the input arguments and groups them into parameter objects. 
The first section below displays all optional GGIR input argument names, the 
GGIR part (1, 2, 3, 4 and/or 5) they are used in, and the parameter object 
they are stored in. As you will see, a few parameters are not part of any parameter 
object because they are direct arguments of the GGIR shell function.

In the [second section of this vignette](#default-argument-values) you will 
find a description and default value for all the arguments.

# Input arguments/parameters overview

| Argument (parameter)        | Used in GGIR part | Parameter object         |
|-----------------------------|-------------------|--------------------------|
| datadir                     | 1, 2, 4, 5        | not in parameter objects |
| f0                          | 1, 2, 3, 4, 5     | not in parameter objects |
| f1                          | 1, 2, 3, 4, 5     | not in parameter objects |
| windowsizes                 | 1, 5              | params_general           |
| desiredtz                   | 1, 2, 3, 4, 5     | params_general           |
| overwrite                   | 1, 2, 3, 4, 5     | params_general           |
| do.parallel                 | 1, 2, 3, 5        | params_general           |
| maxNcores                   | 1, 2, 3, 5        | params_general           |
| use_trycatch_serial         | 1, 2, 3, 5, 6     | params_general           |
| myfun                       | 1, 2, 3           | not in parameter objects |
| outputdir                   | 1                 | not in parameter objects |
| studyname                   | 1                 | not in parameter objects |
| chunksize                   | 1                 | params_rawdata           |
| do.enmo                     | 1                 | params_metrics           |
| do.lfenmo                   | 1                 | params_metrics           |
| do.en                       | 1                 | params_metrics           |
| do.bfen                     | 1                 | params_metrics           |
| do.hfen                     | 1                 | params_metrics           |
| do.hfenplus                 | 1                 | params_metrics           |
| do.mad                      | 1                 | params_metrics           |
| do.anglex                   | 1                 | params_metrics           |
| do.angley                   | 1                 | params_metrics           |
| do.angle                    | 1                 | params_metrics           |
| do.enmoa                    | 1                 | params_metrics           |
| do.roll_med_acc_x           | 1                 | params_metrics           |
| do.roll_med_acc_y           | 1                 | params_metrics           |
| do.roll_med_acc_z           | 1                 | params_metrics           |
| do.dev_roll_med_acc_x       | 1                 | params_metrics           |
| do.dev_roll_med_acc_y       | 1                 | params_metrics           |
| do.dev_roll_med_acc_z       | 1                 | params_metrics           |
| do.lfen                     | 1                 | params_metrics           |
| do.lfx                      | 1                 | params_metrics           |
| do.lfy                      | 1                 | params_metrics           |
| do.lfz                      | 1                 | params_metrics           |
| do.hfx                      | 1                 | params_metrics           |
| do.hfy                      | 1                 | params_metrics           |
| do.hfz                      | 1                 | params_metrics           |
| do.bfx                      | 1                 | params_metrics           |
| do.bfy                      | 1                 | params_metrics           |
| do.bfz                      | 1                 | params_metrics           |
| do.zcx                      | 1                 | params_metrics           |
| do.zcy                      | 1                 | params_metrics           |
| do.zcz                      | 1                 | params_metrics           |
| do.neishabouricounts        | 1                 | params_metrics           |
| actilife_LFE                | 1                 | params_metrics           |
| lb                          | 1                 | params_metrics           |
| hb                          | 1                 | params_metrics           |
| n                           | 1                 | params_metrics           |
| do.cal                      | 1                 | params_rawdata           |
| spherecrit                  | 1                 | params_rawdata           |
| minloadcrit                 | 1                 | params_rawdata           |
| printsummary                | 1                 | params_rawdata           |
| print.filename              | 1                 | params_general           |
| backup.cal.coef             | 1                 | params_rawdata           |
| rmc.noise                   | 1                 | params_rawdata           |
| rmc.dec                     | 1                 | params_rawdata           |
| rmc.firstrow.acc            | 1                 | params_rawdata           |
| rmc.firstrow.header         | 1                 | params_rawdata           |
| rmc.col.acc                 | 1                 | params_rawdata           |
| rmc.col.temp                | 1                 | params_rawdata           |
| rmc.col.time                | 1                 | params_rawdata           |
| rmc.unit.acc                | 1                 | params_rawdata           |
| rmc.unit.temp               | 1                 | params_rawdata           |
| rmc.origin                  | 1                 | params_rawdata           |
| rmc.header.length           | 1                 | params_rawdata           |
| rmc.format.time              | 1                | params_rawdata           |
| rmc.bitrate                 | 1                 | params_rawdata           |
| rmc.dynamic_range           | 1                 | params_rawdata           |
| rmc.unsignedbit             | 1                 | params_rawdata           |
| rmc.desiredtz               | 1                 | params_rawdata           |
| rmc.sf                      | 1                 | params_rawdata           |
| rmc.headername.sf           | 1                 | params_rawdata           |
| rmc.headername.sn           | 1                 | params_rawdata           |
| rmc.headername.recordingid  | 1                 | params_rawdata           |
| rmc.header.structure        | 1                 | params_rawdata           |
| rmc.check4timegaps          | 1                 | params_rawdata           |
| rmc.col.wear                | 1                 | params_rawdata           |
| rmc.doresample              | 1                 | params_rawdata           |
| imputeTimegaps              | 1                 | params_rawdata           |
| dayborder                   | 1, 2, 5           | params_general           |
| dynrange                    | 1                 | params_rawdata           |
| nonwear_range_threshold     | 1                 | params_rawdata           |
| configtz                    | 1                 | params_general           |
| minimumFileSizeMB           | 1                 | params_rawdata           |
| interpolationType           | 1                 | params_rawdata           |
| expand_tail_max_hours       | deprecated        | params_general           |
| recordingEndSleepHour       | 1                 | params_general           |
| maxRecordingInterval        | 1                 | params_general           |
| recording_split_times       | 1                 | params_general           |
| recording_split_overlap     | 1                 | params_general           |
| recording_split_ignore_edges | 1                | params_general           |
| recording_split_timeformat  | 1                 | params_general           |
| nonwear_approach            | 1                 | params_general           |
| dataFormat                  | 1                 | params_general           |
| extEpochData_timeformat     | 1                 | params_general           |
| metadatadir                 | 2, 3, 4, 5        | not in parameter objects |
| minimum_MM_length.part5     | 5                 | params_cleaning          |
| strategy                    | 2, 5              | params_cleaning          |
| hrs.del.start               | 2, 5              | params_cleaning          |
| hrs.del.end                 | 2, 5              | params_cleaning          |
| maxdur                      | 2, 5              | params_cleaning          |
| max_calendar_days           | 2                 | params_cleaning          |
| includedaycrit              | 2, 5              | params_cleaning          |
| nonWearEdgeCorrection       | 2                 | params_cleaning          |
| nonwearFiltermaxHours       | 2                 | params_cleaning          |
| nonwearFilterWindow         | 2                 | params_cleaning          |
| L5M5window                  | 2                 | params_247               |
| M5L5res                     | 2, 5              | params_247               |
| winhr                       | 2, 5              | params_247               |
| qwindow                     | 2                 | params_247               |
| qlevels                     | 2                 | params_247               |
| ilevels                     | 2                 | params_247               |
| mvpathreshold               | 2                 | params_phyact            |
| boutcriter                  | 2                 | params_phyact            |
| ndayswindow                 | 2                 | params_cleaning          |
| idloc                       | 2, 4              | params_general           |
| do.imp                      | 2                 | params_cleaning          |
| storefolderstructure        | 2, 4, 5           | params_output            |
| epochvalues2csv             | 2                 | params_output            |
| do.part2.pdf                | 2                 | params_output            |
| sep_reports                 | 2, 4, 5           | params_output            |
| dec_reports                 | 2, 4, 5           | params_output            |
| sep_config                  | 1, 2, 3, 4, 5     | params_output            |
| dec_config                  | 1, 2, 3, 4, 5     | params_output            |
| mvpadur                     | 2                 | params_phyact            |
| bout.metric                 | 2, 5              | params_phyact            |
| closedbout                  | 2                 | params_phyact            |
| IVIS_windowsize_minutes     | 2                 | params_247               |
| IVIS_epochsize_seconds      | 2                 | params_247               |
| IVIS.activity.metric        | 2                 | params_247               |
| iglevels                    | 2, 5              | params_247               |
| TimeSegments2ZeroFile       | 2                 | params_cleaning          |
| qM5L5                       | 2                 | params_247               |
| MX.ig.min.dur               | 2                 | params_247               |
| qwindow_dateformat          | 2                 | params_247               |
| part2CR                     | 2                 | params_247               |
| anglethreshold              | 3                 | params_sleep             |
| timethreshold               | 3                 | params_sleep             |
| ignorenonwear               | 3                 | params_sleep             |
| HDCZA_threshold             | 3                 | params_sleep             |
| acc.metric                  | 3, 5              | params_general           |
| do.part3.pdf                | 3                 | params_output            |
| sensor.location             | 3, 4              | params_general           |
| HASPT.algo                  | 3                 | params_sleep             |
| HASIB.algo                  | 3                 | params_sleep             |
| Sadeh_axis                  | 3                 | params_sleep             |
| longitudinal_axis           | 3                 | params_sleep             |
| HASPT.ignore.invalid        | 3                 | params_sleep             |
| oakley_threshold            | 3                 | params_sleep             |
| consider_marker_button      | 3                 | params_sleep             |
| impute_marker_button        | 3                 | params_sleep             |
| SRI1_smoothing_wsize_hrs    | 3                 | params_sleep             |
| SRI1_smoothing_frac         | 3                 | params_sleep             |
| spt_min_block_dur           | 3                 | params_sleep             |
| spt_max_gap_dur             | 3                 | params_sleep             |
| spt_max_gap_ratio           | 3                 | params_sleep             |
| HorAngle_threshold          | 3                 | params_sleep             |
| guider_cor_maxgap_hrs       | 3                 | params_sleep             |
| guider_cor_min_frac_sib     | 3                 | params_sleep             |
| guider_cor_min_hrs          | 3                 | params_sleep             |
| guider_cor_meme_frac_out    | 3                 | params_sleep             |
| guider_cor_meme_frac_in     | 3                 | params_sleep             |
| guider_cor_meme_min_hrs     | 3                 | params_sleep             |
| guider_cor_do               | 3                 | params_sleep             |
| guider_cor_meme_min_dys     | 3                 | params_sleep             |
| loglocation                 | 4, 5              | params_sleep             |
| colid                       | 4                 | params_sleep             |
| coln1                       | 4                 | params_sleep             |
| do.visual                   | 4                 | params_output            |
| outliers.only               | 4                 | params_output            |
| excludefirstlast            | 4                 | params_cleaning          |
| criterror                   | 4                 | params_output            |
| includenightcrit            | 4                 | params_cleaning          |
| relyonguider                | 4                 | params_sleep             |
| relyonsleeplog              | 4                 | deprecated               |
| sleepefficiency.metric      | 4                 | params_sleep             |
| def.noc.sleep               | 4                 | params_sleep             |
| sib_must_fully_overlap_with_TimeInBed  | 4      | params_sleep             |
| data_cleaning_file          | 4, 5              | params_cleaning          |
| excludefirst.part4          | 4                 | params_cleaning          |
| excludelast.part4           | 4                 | params_cleaning          |
| sleepwindowType             | 4                 | params_cleaning          |
| excludefirstlast.part5      | 5                 | params_cleaning          |
| boutcriter.mvpa             | 5                 | params_phyact            |
| boutcriter.in               | 5                 | params_phyact            |
| boutcriter.lig              | 5                 | params_phyact            |
| threshold.lig               | 5                 | params_phyact            |
| threshold.mod               | 5                 | params_phyact            |
| threshold.vig               | 5                 | params_phyact            |
| boutdur.mvpa                | 5                 | params_phyact            |
| boutdur.in                  | 5                 | params_phyact            |
| boutdur.lig                 | 5                 | params_phyact            |
| save_ms5rawlevels           | 5                 | params_output            |
| part5_agg2_60seconds        | 5                 | params_general           |
| includedaycrit.part5        | 5                 | params_cleaning          |
| includenightcrit.part5      | 5                 | params_cleaning          |
| frag.metrics                | 5                 | params_phyact            |
| LUXthresholds               | 5                 | params_247               |
| LUX_cal_constant            | 5                 | params_247               |
| LUX_cal_exponent            | 5                 | params_247               |
| LUX_day_segments            | 5                 | params_247               |
| SRI2_WASOmin                | 5                 | params_247               |
| timewindow                  | 5                 | params_output            |
| save_ms5raw_format          | 5                 | params_output            |
| save_ms5raw_without_invalid | 5                 | params_output            |
| do.sibreport                | 5                 | params_output            |
| possible_nap_window         | 5                 | params_sleep             |
| possible_nap_dur            | 5                 | params_sleep             |
| possible_nap_gap            | 5                 | params_sleep             |
| nap_markerbutton_method     | 5                 | params_sleep             |
| nap_markerbutton_max_distance | 5               | params_sleep             |
| method_research_vars        | 5                 | params_sleep             |
| includecrit.part6           | 6                 | params_cleaning          |
| part6_threshold_combi       | 6                 | params_phyact            |
| part6CR                     | 6                 | params_247               |
| part6HCA                    | 6                 | params_247               |
| part6Window                 | 6                 | params_247               |
| part6DFA                    | 6                 | params_247               |
| require_complete_lastnight_part5 | 5            | params_output            |
| visualreport_without_invalid| visualreport      | params_output            |
| visualreport_hrsPerRow      | visualreport      | params_output            |
| visualreport_focus          | visualreport      | params_output            |
| visualreport_validcrit      | visualreport      | params_output            |
| dofirstpage                 | visualreport      | params_output            |
| visualreport                | visualreport      | params_output            |
| viewingwindow               | visualreport      | params_output            |


# Arguments/parameters description {#default-argument-values}

All information as shown below has been auto-generated and
is identical to the information provided in the [GGIR package pdf manual](https://CRAN.R-project.org/package=GGIR/GGIR.pdf).

## GGIR function input arguments

### mode {-}
`r format(GGIR::ShellDoc2Vignette("mode"))`

### datadir {-}
`r format(GGIR::ShellDoc2Vignette("datadir"))`

### outputdir {-}
`r format(GGIR::ShellDoc2Vignette("outputdir"))`

### studyname {-}
`r format(GGIR::ShellDoc2Vignette("studyname"))`

### f0 {-}
`r format(GGIR::ShellDoc2Vignette("f0"))`

### f1 {-}
`r format(GGIR::ShellDoc2Vignette("f1"))`

### do.report {-}
`r format(GGIR::ShellDoc2Vignette("do.report"))`

### configfile {-}
`r format(GGIR::ShellDoc2Vignette("configfile"))`

### myfun {-}
`r format(GGIR::ShellDoc2Vignette("myfun"))`

## General Parameters

`r format(GGIR::parametersVignette("general"))`

## Raw Data Parameters

`r format(GGIR::parametersVignette("rawdata"))`

## Metrics Parameters

`r format(GGIR::parametersVignette("metrics"))`

## Cleaning Parameters

`r format(GGIR::parametersVignette("cleaning"))`

## Sleep Parameters

`r format(GGIR::parametersVignette("sleep"))`


## Physical activity Parameters

`r format(GGIR::parametersVignette("phyact"))`

## 24/7 Parameters

`r format(GGIR::parametersVignette("247"))`

## Output Parameters

`r format(GGIR::parametersVignette("output"))`


```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
```


```{r, echo=FALSE, out.width = "60%", out.extra='style="border: 0; padding:20px"', fig.alt="GGIR logo"}
knitr::include_graphics("GGIR-MASTERLOGO-RGB.png")
```

---
title: "Installation of the GGIR R Package"
output:
   html_document:
    toc : true
    number_sections: true
    toc_depth: 3
    toc_float: true #by turning this on, the table of contents moves to the left of the page.
urlcolor: blue
vignette: >
  %\VignetteIndexEntry{Installation of the GGIR R Package}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
editor_options:
  markdown:
    wrap: 72
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
```

## Install R and RStudio

-   [Download and install R](https://cran.r-project.org/)

-   [Download and install
    RStudio](https://posit.co/products/open-source/rstudio/)

## Install GGIR

Install the latest released version of GGIR with its dependencies from
CRAN. You can do this with one command from the console command line:

```{R,eval=FALSE}
install.packages("GGIR", dependencies = TRUE)
library(GGIR)
```

Alternatively, you can install the latest development version, which
might include additional bug fixes and functionalities. To get the
development version, please use:

```{R,eval=FALSE}
install.packages("remotes", dependencies = TRUE)
remotes::install_github("wadpac/GGIR", dependencies = TRUE)
library(GGIR)
```

## Other packages you may need

Additionally, in some use-cases you will need to install one or multiple
additional packages. Note these packages are not installed by default,
so please follow the instructions if:

-   You want to derive Neishabouricounts (with
    `do.neishabouricounts = TRUE`), install the actilifecounts package
    with `install.packages("actilifecounts")`
-   You want to process Sensewear xlsx files (with
    `dataFormat = "sensewear"`), install the readxl package
    with `install.packages("readxl")`

---
title: "How can I get service and support?"
output:
   html_document:
    toc : true
    number_sections: true
    toc_depth: 3
    toc_float: true #by turning this on, the table of contents moves to the left of the page.
urlcolor: blue
vignette: >
  %\VignetteIndexEntry{How can I get service and support?}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
editor_options:
  markdown:
    wrap: 72
---

GGIR documentation is openly available for users in several places:

-   **This [dedicated website](https://wadpac.github.io/GGIR/)**
    containing instructions to install, get started, and specific
    documentation on the GGIR pipeline and functionalities.

-   **Vignettes available on
    [CRAN](https://cran.r-project.org/package=GGIR/index.html)**,
    containing an overview of GGIR and dedicated vignettes for some
    specific functionalities.

-   **The [GGIR configuration parameters
    vignette](https://cran.r-project.org/package=GGIR/vignettes/GGIRParameters.html)**
    explains each of the parameters that can be used in GGIR and how to
    use them.

-   The **GGIR function documentation**, to which you can get access to
    directly from R:

``` r
library(GGIR)
?GGIR
```

## Question and problems

For general questions and issues please report in the [GGIR google
group](https://groups.google.com/forum/#!forum/RpackageGGIR)[.]{.underline}

However, if you are familiar with GitHub you are also welcome to report
your issue via the [GitHub issue
tracker](https://github.com/wadpac/GGIR/issues).

Note that support in these places is based on voluntary efforts and we
encourage you to try and help other users with questions.

Questions are valuable as they help us understand the challenges you run
into and they occasionally help us to identify a bug in the code.
However, to make this practical it would be great if you doNOT send us
your questions via personal messages. Posting on public platforms
enables others to learn from the conversations, it reduces the chance
that we have to answer the same question multiple times, it increases
the chance that someone else can answer your question when we are not
available, and it gives recognition for the voluntary time people put
into answering questions.

## Commercial training services

Accelting provides online [training
options](https://www.accelting.com/ggir-training/) for using GGIR,
please check out the website. Check the dedicated
[website](https://www.accelting.com/ggir-training/) with all details? Do
not hesitate to reach out via:
[training\@accelting.com](mailto:training@accelting.com).

At the moment, the available training services are:

-   The [**GGIR standard
    training**](https://www.accelting.com/ggir-standard-training/),
    aimed to train you in both the hands-on use and understanding of
    GGIR's key functionalities and algorithms for common use-cases.

-   The [**R essentials for
    GGIR**](https://www.accelting.com/r-essentials-for-ggir/), designed
    to train you in the basic R skills that you will need for using
    GGIR. Also of interest may be the brief [free R introduction
    tutorial](https://www.accelting.com/updates/short-r-tutorial/).

-   The [**Private training in
    GGIR**](https://www.accelting.com/ggir-private-training-info/),
    dedicated to teach you how to use GGIR with a strong focus on the
    aspects relevant to you and to your research projects.

-   The [**Extra GGIR support
    session**](https://www.accelting.com/ggir-training/#:~:text=GGIR%20private%20training%22/%3E-,Extra%20GGIR%20support%20session,-Aim%3A%20Single%20session),
    a single session aimed to answer specific questions or address the
    challenges you may run into.

## Dedicated support

If you need dedicated support with the use of GGIR, want GGIR to be modified or enhanced in line to your needs, then please contact [Vincent van Hees](https://www.accelting.com/). 

---
title: "7. The 24-hour time use analysis"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{7. The 24-hour time use analysis}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

Work in progress...

## Waking-to-waking or 24-hour time-use analysis

## Merging the sleep and physical activity information

## Key arguments

## Related output

---
title: "2. The GGIR pipeline"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{2. The GGIR pipeline}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```


## The GGIR function

GGIR comes with a large number of functions and optional settings (function arguments also referred to as parameters) that together form a processing pipeline. To ease interacting with GGIR, it has one function to act as interface to all functionality, and this function is also named `GGIR`. You only need to learn how to work with this central function, but it is important to understand that this function interacts with all other functions in GGIR.

Further, it is important to understand that the `GGIR package` is structured in two complementary ways:

-   **Parts** to reflect computational order.
-   **Parameter themes** to reflect functional components.

### Parts of the pipeline

GGIR, has a computational structure of six parts which are applied sequentially to the data:

-   **Part 1:** Loads the data and stores derived signal features or also known as metrics as needed for the following parts.
-   **Part 2:** Basic data quality assessment and description per day and per file.
-   **Part 3:** Estimation of sustained inactivity bouts (rest periods) and sleep, needed for input to Part 4.
-   **Part 4:** Labels the sustained inactive periods derived in Part 3 as sleep, or daytime sustained inactivity, per night and per file.
-   **Part 5:** Compiles time series with classification of sleep and physical activity intensity categories by re-using information derived in part 2, 3 and 4. Next, Part 5 generates a describe summary such as total time in intensity categories, the number of bouts, time spent in bouts and average acceleration per category, and fragmentation.
-   **Part 6:** Facilitates analyses that rely on direct access to the full recording such as household co-analysis and circadian rhythm analysis. The reason why it split up in parts is that it avoids having the re-run all preceding parts if you only want to make a small change in the more downstream parts. The specific order and content of the parts has grown for historical with Part 1 and 2 created in 2011-2013, Part 3 and 4 created in 2013-2015, Part 5 created in 2017-2020, and Part 6 created in 2023-2024.

#### Milestone data

Each part when run stores its intermediate data, which we refer to as the milestone data, in an `.RData` file. This `.RData` file is then read by the next part. Therefore, there is no direct interaction between the parts, all exchange of information works via the `.RData` files. The advantage of this design is that it offers internal modularity to ease re-processing and code development.

The `.RData` files are stored in sub-folders of the output folder:

-   Part 1: `meta/basic`
-   Part 2: `meta/ms2.out`
-   Part 3: `meta/ms3.out`
-   Part 4: `meta/ms4.out`
-   Part 5: `meta/ms5.out`
-   Part 6: `meta/ms6.out`

As user, you are unlikely to ever need to interact with these milestone data because all relevant output for you is stored in csv and pdf files in the output folder `results`. However, it can be useful to know where the milestone data are stored for the following reasons:

-   By copying these `.RData` files to a new computer, you may continue your analyses without access to the original data files. This can for example be helpful if you process subsets of the same study on different computers, pooling the resulting milestone data allows you to finalise your analysis on a single computer.
-   When you run into a problem, these `.RData` files may allow you to share the problem as a reproducible example of the problem without having to share the original data file.

### Parameter themes

The arguments of all GGIR package functions can be used as input argument to the GGIR function. We will refer to these arguments as input parameters in this documentation. The parameters are structured thematically, independently of the five parts they are used in:

-   **params_rawdata:** parameters related to handling the raw data such as resampling or calibrating.
-   **params_metrics:** parameters related to aggregating the raw data to epoch level summary metrics.
-   **params_sleep:** parameters related to sleep detection.
-   **params_physact:** parameters related to physical activity.
-   **params_247:** parameters related to 24/7 behaviours that do not fall into the typical sleep or physical activity research category.
-   **params_output:** parameters relating to how and whether output is stored.
-   **params_general:** general parameters not covered by any of the above categories

There are a couple of ways to inspect the parameters in each parameter category and their default values:

#### GGIR function documentation:

```{R,eval=FALSE}
?GGIR
```

Or visit the .pdf manual and search for function `GGIR`.

#### Parameters vignette:

Documentation for all parameters in the parameter objects can be found the vignette: [GGIR configuration parameters](https://cran.r-project.org/package=GGIR).

#### From R command line:

```{R,eval=FALSE}
library(GGIR)
print(load_params())
```

If you are only interested in one specific category like sleep:

```{R,eval=FALSE}
library(GGIR)
print(load_params()$params_sleep)
```

If you are only interested in parameter `HASIB.algo` from the

`sleep_params` object:

```{R,eval=FALSE}
library(GGIR)
print(load_params()$params_sleep[["HASIB.algo"]])
```

All of these parameters are accepted as parameter to function `GGIR`, because `GGIR` is a shell around all GGIR functionality. However, the `params_` objects themselves cannot be provided as input to `GGIR`.

#### Configuration file.

GGIR stores all the parameter values as used in a csv-file named config.csv. The file is stored after each run of GGIR in the root of the output folder overwriting any existing config.csv file. So, if you would like to add annotations in the file, e.g. in the fourth column, then you will need to store it somewhere outside the output folder and explicitly specify the path with parameter `configfile`. Further, the config.csv file itself is accepted as input to `GGIR` with parameter `configfile` to replace the specification of all the parameters, except `datadir` and `outputdir`, see example below.

```{R,eval=FALSE}
library(GGIR)
GGIR(datadir = "C:/mystudy/mydata",
     outputdir = "D:/myresults", 
     configfile = "D:/myconfigfiles/config.csv")
```

The practical value of this is that it eases the replication of analysis, because instead of having to share you R script with colleagues, sharing your config.csv file will be sufficient.

#### Parameter extraction order

If parameters are provided in the GGIR call, then it always uses those. If parameters are not provided in the function call, GGIR checks whether there is a config.csv file either in the output folder or specified via parameter `configfile` and loads those values. If a parameter is neither specified in the GGIR function call and not available in the config.csv file then GGIR uses its default values which can be inspected as discussed in the section above. Here, it is important to realise that a consequence of this logic is that GGIR will not revert to its default parameter values in a repeated analysis unless you remove the parameter from the function call and delete the config.csv file.

To ensure this is clear here are some example:

-   If GGIR is used for the first time without specifying parameter `mvpathreshold` then it will use the default value, which is 100.
-   If you specify `mvpathreshold = 120` then GGIR will use this instead and store it in the config.csv file.
-   If you run GGIR again but this time delete `mvpathreshold = 120` from your GGIR call then GGIR will fall back on the value 120 as now stored in the config.csv file.
-   If you delete the config.csv file and run GGIR again then the value 100 will be used again.

## Input files

### Raw data

1.  GGIR currently works with the following accelerometer brands and formats:

-   [GENEActiv](https://activinsights.com/) .bin
-   [Axivity](https://axivity.com/) AX3 and AX6 .cwa
-   [ActiGraph](https://theactigraph.com) .csv and .gt3x (.gt3x only the newer format generated with firmware versions above 2.5.0. Serial numbers that start with “NEO” or “MRA” and have firmware version of 2.5.0 or earlier use an older format of the .gt3x file). If you want to work with .csv exports via the commercial ActiLife software then note that you have the option to export data with timestamps, which should be turned off. To cope with the absence of timestamps GGIR will calculate timestamps from the sample frequency, the start time and start date as presented in the file header.
-   [Movisens](https://www.movisens.com/en/) with data stored in folders.
-   Any other accelerometer brand that generates csv output, see documentation for functions `read.myacc.csv` and parameter `rmc.noise` in the vignette [Reading csv files with raw data in GGIR](https://cran.r-project.org/package=GGIR/vignettes/readmyacccsv.html).

### Externally derived epoch-level data

By default GGIR assumes that the data is raw as discussed at the start of this chapter. However, for some studies raw data is not available and all we have is an epoch level aggregate. For example, done by external software or done inside the accelerometer device. Although this can introduce some severe limitations to the transparency and flexibility of our analysis, GGIR makes an attempt to facilitate the analysis of these aggregations of the raw data. Please find below an overview of the file format currently facilitated:

| Sensor brand | Data format                                                 | Combine with parameter                                                       |
|------------------------|-----------------------|-------------------------|
| Actiwatch    | .csv                                                        | `dataFormat = "actiwatch_csv"` `extEpochData_timeformat = "%m/%d/%Y %H:%M:%S"` |
| Actiwatch    | .awd                                                        | `dataFormat = "actiwatch_awd"` `extEpochData_timeformat = "%m/%d/%Y %H:%M:%S"` |
| ActiGraph    | .csv (count files, not be confused with csv raw data files) | `dataFormat = "actigraph_csv"` `extEpochData_timeformat = "%m/%d/%Y %H:%M:%S"` |
| Axivity      | UKBiobank csv export of epoch level data                    | `dataFormat = "ukbiobank_csv"`                                                 |

: Note: For Actiwatch and ActiGraph, physical activity description and sleep classification needs to be tailored to count-specific algorithms: `do.neishabouricounts = TRUE`, `acc.metric = "NeishabouriCount_x"`, `HASPT.algo = "NotWorn"` `HASIB.algo = "NotWorn"`; Note 2: For UK Biobank csv epoch data GGIR does not facilitate sleep analysis as arm angle is not exported.

To process these files GGIR loads their content and saves it as GGIR part 1 milestone data, essentially fooling the rest of GGIR to think that GGIR part 1 created the data. As will be discussed in chapter 2, GGIR does non-wear detection in two steps: The first step is done in part 1 and the second step is done in part 2. In relation to externally derived epoch data non-wear is detected by looking for consecutive zeros of one hour (Actiwatch, ActiGraph) or derived from the file (UK Biobank csv).

2.  All accelerometer data that need to be analysed should be stored in one folder or subfolders of that folder. Make sure the folder does not contain any files that are not accelerometer data.
3.  Choose an appropriate name for the folder, preferable with a reference to the study or project it is related to rather than just 'data', because the name of this folder will be used as an identifier of the dataset and integrated in the name of the output folder GGIR creates.

## How to run your analysis?

The bare minimum input needed for `GGIR` is:

```{R,eval=FALSE}
library(GGIR)
GGIR(datadir = "C:/mystudy/mydata",
     outputdir = "D:/myresults")
```

Argument `datadir` allows you to specify where you have stored your accelerometer data and `outputdir` allows you to specify where you would like the output of the analyses to be stored. This cannot be equal to `datadir`. If you copy paste the above code to a new R script (file ending with .R) and Source it in R(Studio) then the dataset will be processed and the output will be stored in the specified output directory.

Next, we can add argument `mode` to tell GGIR which part(s) to run, e.g. `mode = 1:5` tells GGIR to run all five parts and is the default. With argument `overwrite` we can tell GGIR whether to overwrite previously produced milestone data.

Further, argument `idloc` tells GGIR where to find the participant ID. The default setting most likely does not work for most data formats, by which it is important that you tailor the value of this argument to your study setting.

GGIR stores some of its output csv file format with by default comma as column separator. However, this can be modified by argument `sep_reports`.

Each chapter in this digital book will highlight the key arguments related to specific topics.

### From the R console on your own desktop/laptop

Create an R-script and put the GGIR call in it. Next, you can source the R-script with the `source` button in RStudio: `source("pathtoscript/myshellscript.R")`

GGIR by default support multi-thread processing which is used to process one input file per process, which can be turned off by setting parameter `do.parallel = FALSE`.

### In a cluster

If processing data with GGIR on your desktop/laptop is not fast enough then we advise using a GGIR on a computing cluster. The way we did it on a Sun Grid Engine cluster is shown below, please note that some of these commands are specific to the computing cluster you are working on. Please consult your local cluster specialist to explore how to run GGIR on your cluster. Here, we only share how we did it. We had three files for the SGE setting:

**submit.sh**

```{bash,eval=FALSE}
for i in {1..707}; do
    n=1
    s=$(($(($n * $[$i-1]))+1))
    e=$(($i * $n))
    qsub /home/nvhv/WORKING_DATA/bashscripts/run-mainscript.sh $s $e
done
```

**run-mainscript.sh**

```{bash,eval=FALSE}
#! /bin/bash
#$ -cwd -V
#$ -l h_vmem=12G
/usr/bin/R --vanilla --args f0=$1 f1=$2 < /home/nvhv/WORKING_DATA/test/myshellscript.R
```

**myshellscript.R**

```{R,eval=FALSE}
options(echo=TRUE)
args = commandArgs(TRUE)
if(length(args) > 0) {
  for (i in 1:length(args)) {
    eval(parse(text = args[[i]]))
  }
}
GGIR(f0=f0,f1=f1,...)
```

You will need to update the `...` in the last line with the parameters you used for `GGIR`. Note that `f0=f0,f1=f1` is essential for this to work. The values of `f0` and `f1` are passed on from the bash script. Once this is all setup you will need to call `bash submit.sh` from the command line. With the help of computing clusters GGIR has successfully been run on some of the world’s largest accelerometer data sets such as UK Biobank and German NAKO study.

## Processing time

The time to process a typical seven-day recording should be anywhere in between 3 and 10 minutes depending on the sample frequency of the recording, the sensor brand, data format, the exact configuration of GGIR, and the specifications of your computer. If you are observing processing times of 20 minutes or longer for a seven-day recording then probably you are slowed down by other factors.

Some tips on how you may be able to address this:

-   Make sure the data you process is on the same machine as where GGIR is run. Processing data located somewhere else on a computer network can substantially slow down software.
-   Make sure your machine has 8GB or more RAM memory, using GGIR on old machines with only 4GB is known to be slow. However, total memory is not the only bottle neck, also consider the number of processes (threads) your CPU can run relative to the amount of memory. Ending up with 2GB per process seems a good target. It can be helpful to turn off parallel processing with `do.paralllel = FALSE`.
-   Avoid doing other computational activities with your machine while running GGIR. For example, if you use DropBox or OneDrive make sure they do not sync while you are running GGIR. When using GGIR to process large datasets it is probably best not to use the machine. Make sure the machine is configured not to automatically turn off after X hours as that would terminate GGIR. Further, you may like to configure the machine to not fall asleep as that would pause GGIR.
-   Reduce the amount of data GGIR loads in memory with parameter `chunksize`, which can be useful on machines with limited memory or when processing many files in parallel. A `chunksize` value of 0.2 (minimum) will make GGIR load the data in chunks of only 20% of the size relative to the chunks it loads by default which is approximately 12 hours of data for the auto-calibration routine and 24 hours of data for the calculation of signal metrics.

## GGIR output

GGIR always creates an output folder in the location as specified with parameter `outputdir`. The output folder name is constructed as `output_` followed by the name of the dataset which is derived from the most distal folder name of the data directory as specified with `datadir`. We recommend this approach because it ensures that output folder name and data directory have matching names. However, it is possible to use `datadir` to specify a vector with paths to individual files, which may helpful if you want to process a set of files but are not in a position to move them to a new folder. In that scenario you will need to set parameter `studyname` to tell GGIR what the dataset name is.

Inside the output folder GGIR will create two subfolders: `meta` and `results` as discussed earlier in this chapter. Inside `results` you will find folder named `QC` (Quality Checks). The name QC (Quality Checks) is possibly somewhat confusing. Data quality checks are best started with the files stored in the results folder, where the files in the QC subfolder offer complementary information to help with the quality check.

GGIR generates reports in the parts 2, 4, 5, and 6 of the pipeline. With parameter  `do.report` you can specify for which of these parts you want the reports to be generated. For example, `do.report = c(2, 5)` will only generated the report for parts 2 and 5. For a full GGIR analyses we expect at least the following output files:

***Output files in results subfolder:***

| File path                                                | Resolution                                                       | GGIR part |
|--------------------------------------|-----------------|-----------------|
| part2_summary.csv                                | one row per recording                                            | 2         |
| part2_daysummary.csv                             | one row per day                                                  | 2         |
| part4_summary_sleep_cleaned.csv                  | one row per recording                                            | 4         |
| part4_nightsummary_sleep_cleaned.csv             | one row per night                                                | 4         |
| visualisation_sleep.pdf                          | 40 recordings per page                                           | 4         |
| part5_daysummary_WW_L40M100V400_T5A5.csv         | one row per day                                                  | 5         |
| part5_personsummary_WW_L40M100V400_T5A5.csv      | one row per recording                                            | 5         |

: A detailed discussion of each output can be found in the other chapters.


***Output files in results/QC subfolder:***

| File path                                                | Resolution                                                       | GGIR part |
|--------------------------------------|-----------------|-----------------|
| data_quality_report.csv                       | one row per recording                                            | 5         |
| part4_summary_sleep_full.csv                  | one row per recording                                            | 4         |
| part4_nightsummary_sleep_full.csv             | one row per night                                                | 4         |
| part5_daysummary_full_WW_L40M100V400_T5A5.csv | one row per day                                                  | 5         |
| plots_to_check_data_quality_1.pdf             | one recording per page                                           | 2         |

: A detailed discussion of each output can be found in the other chapters.


***Output files in meta/ms5.outraw subfolder:***

| File path                                                | Resolution                                                       | GGIR part |
|--------------------------------------|-----------------|-----------------|
| 40_100_400/                               | one file with time series per recording                          |           |
| behavioralcodes2023-11-06.csv           | dictionary for behavioural codes as used in the time series file |           |

: A detailed discussion of each output can be found in the other chapters.

---
title: "Get started: the GGIR R package"
output:
   html_document:
    toc : true
    number_sections: true
    toc_depth: 3
    toc_float: true #by turning this on, the table of contents moves to the left of the page.
urlcolor: blue
vignette: >
  %\VignetteIndexEntry{Get started: the GGIR R package}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
editor_options:
  markdown:
    wrap: 72
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
```

This section will guide you through the process of configuring your work
environment to initiate the use of GGIR. For this purpose, you will need
at least one accelerometer file to be processed by GGIR, please feel
free to download [this .gt3x file](https://www.dropbox.com/s/covfyj31sarfaw0/GGIR01.gt3x?dl=0) or [this .cwa file](https://biobank.ndph.ox.ac.uk/ukb/refer.cgi?id=131620)
and follow the instructions below.

## Run GGIR for first time

1.  First, you will need to place your file(s) in a folder in your
    computer. Make sure that this folder does only contain accelerometer
    files.

2.  Use the following command to run GGIR over your data using all the
    default options.

    -   datadir refers to the directory where you have located the
        accelerometer files.

    -   outputdir refers to the directory where you want to store the
        GGIR’s output.

``` r
library(GGIR) 
GGIR(datadir="C:/mystudy/mydata", outputdir="D:/myresults")
```

After some minutes, you should be able to see how your output directory
gets populated with files, reports, and visualizations.

## Related links

-   [Install R and GGIR](https://wadpac.github.io/GGIR/articles/chapter0_Installation.html).
-   [Get support](https://wadpac.github.io/GGIR/articles/chapter0_Support.html)
-   [Suitable file formats for GGIR](https://wadpac.github.io/GGIR/articles/chapter2_Pipeline.html#input-files)

---
title: "8. Physical behaviours analysis"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{8. Physical behaviours analysis}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

Work in progress...

## Acceleration distribution

### Percentiles or MX metrics (Rowlands)

### Intensity levels

### Intensity gradient

### Key arguments

### Related output

## Physical activity intensities (or inactivity)

### Quantification of physical activity and inactivity in GGIR

### Using cut-points 

### Bouts of activity/inactivity

### Key arguments

### Related output

## Fragmentation metrics

### What is a fragment

### When and how are they calculated in GGIR

### Fragmentation metrics available

### Key arguments

## Related output

---
title: "Nap Detection"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Nap Detection}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

The nap detection functionality was recently introduced and has so far only been evaluated in data from pre-schoolers for hip- or wrist-worn accelerometers. These findings can hopefully be shared via journal publication in the upcoming year. Further, we are actively exploring how to best summarise the detect naps in the GGIR output. The description below reflects the current situation. 


### Algorithm description

The detection of both sleep and naps starts with the identification of sustained inactivity bouts as described in [Chapter 8](https://wadpac.github.io/GGIR/articles/chapter8_SleepFundamentalsSibs.html). Next, we continue with the identification of the main sleep period time window in the day as discussed in [Chapter 9](https://wadpac.github.io/GGIR/articles/chapter9_SleepFundamentalsGuiders.html) and [Chapter 10](https://wadpac.github.io/GGIR/articles/chapter10_SleepAnalysis.html). The sustained inactivity bouts that do not overlap with the detected sleep period time window are detected as nap if both of the following conditions are met:

1. Duration in minutes falls in range specified by parameter [possible_nap_dur](https://wadpac.github.io/GGIR/articles/GGIRParameters.html#possible_nap_dur). For pre-schoolers we used `c(30, 240)`.
2. Timing in the day falls in range specified by  parameter [possible_nap_window](https://wadpac.github.io/GGIR/articles/GGIRParameters.html#possible_nap_window), use `c(0, 24)` for full day. For pre-schoolers we used `c(6, 18)`.

Note that default settings are `possible_nap_dur=NULL` and `possible_nap_window=NULL` meaning that nap detection will not be performed.

## Output related to nap detection

Time spent in naps is stored as column "dur_day_nap_min" in the GGIR part 5 csv report where it is treated as one of several mutually exclusive behavioural categories. Meaning, time spent in naps never count towards time spent in inactivity (sedentary behaviour) and visa versa.

## Disclaimer

1.	Make sure that detected sleep period time (SPT) window is plausible. Without plausible SPT detection, nap detection is expected to struggle. This is particularly important for scenario where the accelerometer is not worn consistently or with highly irregular sleeping patterns.

2.	Consider using GGIR’s visualreport functionality to visually inspect the classifications before trusting the numeric output in the csv reports.

3.	Keep in mind that GGIR's nonwear detection is sensitive to non-wear lasting at least 60 minutes. Shorter episodes of non-wear are likely to be missed and my be detected as nap.

---
title: "GGIR output"
output:
   html_document:
    toc: true
    number_sections: true
    toc_depth: 3
    toc_float: true
urlcolor: blue
vignette: >
  %\VignetteIndexEntry{GGIR output}
  \usepackage[utf8]{inputenc}
  %\VignetteEngine{knitr::rmarkdown}
---

**NOTE: If you are viewing this page via CRAN note that the main GGIR documentation has been migrated to the [GGIR GitHub pages](https://wadpac.github.io/GGIR/).**

GGIR generates the following types of output:

-   csv-spreadsheets with all the variables you need for physical activity, sleep and circadian rhythm research
-   .png images with on each image a low time resolution plot of the data per file and quality indicators
-   .RData files with milestone data
-   Pdfs with a visual summary of the physical activity and sleep patterns as identified (see example below)

# GGIR Part 1

GGIR part 1, only outputs RData files which are used by GGIR part 2. The RData files are not intended for direct use by the GGIR user.

# GGIR Part 2

Part 2 generates the following output:

-   part2_summary.csv: Person level summary (see below)
-   part2_daysummary.csv: Day level summary (see below)
-   QC/data_quality_report.csv: Overview of calibration results and whether or not a file was corrupt or too short to be processed,
-   QC/plots_part2: A folder with .png files holding a visualisation of the acceleration time series in 15 minute resolution and with invalid data segments highlighted in colours (yellow: non-wear based on standard deviation threshold, brown: non-wear after extra filtering step (introduced in 2013), and purple: clipping)

## Person level summary (csv)

| (Part of) variable name | Description |
|--------------------------|----------------------------------------------|
| ID | Participant id |
| device_sn | Device serial number |
| bodylocation | Body location extracted from file header |
| filename | Name of the data file |
| start_time | Timestamp when recording started |
| startday | Day of the week on which recording started |
| samplefreq | Sample frequency (Hz) |
| device | Accelerometer brand, e.g. GENEACtiv |
| clipping_score | The Clipping score: Fraction of 15 minute windows per file for which the acceleration in one of the three axis was close to the maximum for at least 80% of the time. This should be 0. |
| meas_dur_dys | Measurement duration (days) |
| complete_24hcycle | Completeness score: Fraction of 15 minute windows per 24 hours for which no valid data is available at any day of the measurement. |
| meas_dur_def_proto_day | measurement duration according to protocol (days): Measurement duration (days) minus the hours that are ignored at the beginning and end of the measurement motivated by protocol design |
| wear_dur_def_proto_day | wear duration duration according to protocol (days): So, if the protocol was seven days of measurement, then wearing the accelerometer for 8 days and recording data for 8 days will still make that the wear duration is 7 days |
| nonwear_hours_filtered | Total number of nonwear hours filtered, only stored when using parameter `nonwearFiltermaxHours` |
| nonwear_events_filtered   | Total number of nonwear events filtered, only stored when using parameter `nonwearFiltermaxHours` |
| calib_err | Calibration error (static estimate) Estimated based on all 'non-movement' periods in the measurement after applying the autocalibration. |
| calib_status | Calibration status: Summary statement about the status of the calibration error minimisation |
| n hours ignored at start of meas (if data_masking_strategy=1, 3 or 5) | number of hours ignored at the start of the measurement (if data_masking_strategy = 1) or at the start of the `ndayswindow` (if data_masking_strategy = 3 or 5) A log of decision made in part2.R |
| n hours ignored at end of meas (if data_masking_strategy=1, 3 or 5) | number of hours ignored at the end of the measurement (if data_masking_strategy = 1) or at the end of the `ndayswindow` (if data_masking_strategy = 3 or 5). A log of decision made in part2.R |
| n days of measurement after which all data is ignored (if data_masking_strategy=1) | number of days of measurement after which all data is ignored (if data_masking_strategy = 1, 3 or 5) A log of decision made in part2.R |
| epoch size to which acceleration was averaged (seconds) | A log of decision made in part1.R |
| data_masking_strategy | A log of the decision made when calling g.impute: value=1 mean ignore specific hours; value=2 mean ignore all data before the first midnight and after the last midnight |

### Letter codes to indicate aggregation type

| (Part of) variable name | Description |
|--------------------------|----------------------------------------------|
| `AD_` | All days (plain average of all available days, no weighting). The variable was calculated per day and then averaged over all the available days |
| `WE_` | Weekend days (plain average of all available days, no weighting). The variable was calculated per day and then averaged over weekend days only |
| `WD_` | Week days (plain average of all available days, no weighting). The variable was calculated per day and then averaged over week days only |
| `WWE_` | Weekend days (weighted average) The variable was calculated per day and then averaged over weekend days. Double weekend days are averaged. This is only relevant for experiments that last for more than seven days. |
| `WWD_` | Week days (weighted average) The variable was calculated per day and then averaged over week days. Double week days were averaged. This is only relevant for experiments that last for more than seven days) |

### Distribution of acceleration

| (Part of) variable name | Description |
|--------------------------|----------------------------------------------|
| ENMO_fullRecordingMean | ENMO is the main summary measure of acceleration. The value presented is the average ENMO over all the available data normalised per 24-hour cycles (diurnal balanced), with invalid data imputed by the average at similar time points on different days of the week. In addition to ENMO it is possible to extract other acceleration metrics (i.e. BFEN, HFEN, HFENplus). We emphasize that it is calculated over the full recording because the alternative is that a variable is only calculated overmeasurement days with sufficient valid hours of data. |
| ENMO | (only available if set to true in part1.R) ENMO is the main summary measure of acceleration. The value presented is the average ENMO over all the available data normalised per 24 hour cycles, with invalid data imputed by the average at similar timepoints on different days of the week. In addition to ENMO it is possible to extract other acceleration metrics in part1.R (i.e. BFEN, HFEN, HFENplus) See also [van Hees PLoSONE April 2013](https://journals.plos.org/plosone/article?id=10.1371/journal.pone.0061691) for a detailed description and comparison of these techniques. |
| pX_A_mg_0-24h_fullRecording | This variable represents the Xth percentile in the distribution of short epoch metric value A of the average day. The average day may not be ideal for describing the distribution. Therefore, the code also extracts the following variable. |
| AD_pX_A_mg_0-24h | This variable represents the Xth percentile in the distribution of short epoch metric value A per day averaged across all days. |
| WWD_MVPA_E5S_T100_ENMO | Time spent in moderate-to-vigorous based on 5 second epoch size and an ENMO metric threshold of 100 |
| `WWE_MVPA_E5S_B1M80%_T100_ENMO` | Time spent in moderate-to-vigorous based on a bout criteria of at least 1 minute where 80% or more of the 5 second epochs are expected to meet the threshold criteria of of 100 mg based on acceleration metric ENMO (threshold is specified with parameter `mvpathreshold`) |
| `WE_[100, 150)_mg_0-24h_ENMO` | Time spent between (and including) 100 mg and 150 (excluding 150 itself) between 0 and 24 hours (the full day) using metric ENMO data exclusion data_masking_strategy (value=1, ignore specific hours; value=2, ignore all data before the first midnight and after the last midnight) |
| `_MVPA_E5S_B1M80_T100` | MVPA calculated based on 5 second epoch setting bout duration 1 Minute and inclusion criterion of more than 80 percent. |
| `_ENMO_mg` | ENMO or other metric was first calculated per day and then average according to AD, WD, WWE, WWD |

### Circadian rhythm: MXLX, IV, and IS

| (Part of) variable name | Description |
|--------------------------|----------------------------------------------|
| L5_A_mg_0-24 | Average of metric A during the least active five\* hours in the day that is the lowest rolling average value of metric A. (\* window size is modifiable by argument `winhr`) |
| M5_A_mg_0-24 | Average of metric A during the most active five\* hours in the day that is the lowest rolling average value of metric A. (\* window size is modifiable by argument `winhr`) |
| L5hr_A_mg_0-24 | Starting time in hours and fractions of hours of L5_A_mg_0-24, where hours below 12 are incremented with 24 to create a continuous scale throughout the night (e.g. 36 = 6am) in line with numeric timeing of sleep variables in GGIR part 4 output. |
| M5hr_A_mg_0-24 | Starting time in hours and fractions of hours of M5_A_mg_0-24 |
| ig_gradient_ENMO_0 -24hr_fullRecording | Intensity gradient calculated over the full recording. |
| 1to6am_ENMO_mg | Average metric value ENMO between 1am and 6am |
| N valid weekdays (WD) | Number of valid weekend days |
| N valid weekend days (WE) | Number of valid week days |
| IS | Interdaily stability as dicussed in [chapter 13](https://wadpac.github.io/GGIR/articles/chapter13_CircadianRhythm.html) |
| IV | Intradaily variability as dicussed in [chapter 13](https://wadpac.github.io/GGIR/articles/chapter13_CircadianRhythm.html) |
| phi | Indicator of auto-correlation in acceleration time series over multiple days, see [chapter 13](https://wadpac.github.io/GGIR/articles/chapter13_CircadianRhythm.html) for details. |
| SSP | Method for describing time series, see [chapter 13](https://wadpac.github.io/GGIR/articles/chapter13_CircadianRhythm.html) for details. |
| ABI | ABI measures how the activity over the observed period is balanced, see [chapter 13](https://wadpac.github.io/GGIR/articles/chapter13_CircadianRhythm.html) for details. |

### Circadian rhythm: Cosinor

The following variables in the GGIR output are derived by the ActCR package, the documentation has been copied from ActCR documentation.

| Variable name | Description |
|--------------------------|----------------------------------------------|
| cosinor_mes |  MESOR which is short for midline statistics of rhythm, which is a rhythm adjusted mean. This represents mean activity level. |
| cosinor_amp | amplitude, a measure of half the extend of predictable variation within a cycle. This represents the highest activity one can achieve |
| cosinor_acrophase | acrophase, a measure of the time of the overall high values recurring in each cycle. Here it has a unit of radian. This represents time to reach the peak. |
| cosinor_acrotime | acrotime represents time to reach the peak. |
| cosinor_ndays | Number of days modeled |
| cosinorExt_minimum | Minimum value of the of the function. |
| cosinorExt_amp | amplitude, a measure of half the extend of predictable variation within a cycle. This represents the highest activity one can achieve. |
| cosinorExt_alpha | It determines whether the peaks of the curve are wider than the troughs: when alpha is small, the troughs are narrow and the peaks are wide; when alpha is large, the troughs are wide and the peaks are narrow. |
| cosinorExt_beta | It dertermines whether the transformed function rises and falls more steeply than the cosine curve: large values of beta produce curves that are nearly square waves. |
| cosinorExt_acrotime | acrophase is the time of day of the peak in the unit of the time (hours) |
| cosinorExt_UpMesor | Time of day of switch from low to high activity. Represents the timing of the rest- activity rhythm. Lower (earlier) values indicate increase in activity earlier in the day and suggest a more advanced circadian phase. |
| cosinorExt_DownMesor | Time of day of switch from high to low activity. Represents the timing of the rest-activity rhythm. Lower (earlier) values indicate decline in activity earlier in the day, suggesting a more advanced circadian phase. |
| cosinorExt_MESOR | A measure analogous to the MESOR of the cosine model (or half the deflection of the curve) can be obtained from mes=min+amp/2. However, it goes through the middle of the peak, and is therefore not equal to the MESOR of the cosine model, which is the mean of the data.
| cosinorExt_ndays | Number of days modeled. |
| cosinorExt_F_pseudo | Measure the improvement of the fit obtained by the non-linear estimation of the transformed cosine model |


**Additionally, GGIR derives:**

| Variable name | Description |
|--------------------------|----------------------------------------------|
| cosinor_timeOffsetHours | (Only used for code development purposes) Time offset in hours between midnight and the start of the time series used for cosinor analysis |
| cosinor_R2 | Squared correlated coefficient between fitted and log-transformed original time series. |
| cosinorExt_R2 | Squared correlated coefficient between fitted and log-transformed original time series. |
| `cosinorIV` | Cosinor analysis compatible estimate of the Intradaily Variability (IV) |
| `cosinorIS` | Cosinor analysis compatible estimate of Interdaily Stability (IS) |

## Day level summary (csv)

This is a non-exhaustive list, because most concepts have been explained in summary.csv

| (Part of) variable name | Description |
|--------------------------|----------------------------------------------|
| ID | Participant id |
| filename | Name of the data file |
| calender_date | Timestamp and date on which measurement started |
| bodylocation | Location of the accelerometer as extracted from file header |
| N valid hours | Number of hours with valid data in the day |
| N hours | Number of hours of measurement in a day, which typically is 24, unless it is a day on which the clock changes (DST) resulting in 23 or 25 hours. The value can be less than 23 if the measurement started or ended this day |
| weekday | Name of weekday |
| measurement | Day of measurement Day number relative to start of the measurement |
| L5hr_ENMO_mg_0-24h | Hour on which L5 starts for these 24 hours (defined with metric ENMO), on a continuous scale relative to the preceding midnight such that 2am is 26 and 6am is 30 |
| L5_ENMO_mg_0-24h | Average acceleration for L5 (defined with metric ENMO) |
| `[A,B)_mg_0-24h_ENMO` | Time spent in minutes between (and including) acceleration value A in mg and (excluding) acceleration value B in mg based on metric ENMO |
| ig_gradient_ENMO_0-24hr | Gradient from intensity gradient analysis proposed by [Rowlands et al. 2018](https://doi.org/10.1249/mss.0000000000001561) based on metric ENMO for the time segment 0 to 24 hours |
| ig_intercept_ENMO_0-24hr | Intercept from intensity gradient analysis proposed by [Rowlands et al. 2018](https://doi.org/10.1249/mss.0000000000001561) based on metric ENMO for the time segment 0 to 24 hours |
| ig_rsquared_ENMO_0-24hr | r squared from intensity gradient analysis proposed by [Rowlands et al. 2018](https://doi.org/10.1249/mss.0000000000001561) based on metric ENMO for the time segment 0 to 24 hours |

## Data_quality_report (csv)

The data_quality_report.csv is stored in subfolder folder results/QC.

| (Part of) variable name | Description |
|--------------------------|----------------------------------------------|
| filename | file name |
| file.corrupt | Is file corrupt? TRUE or FALSE (mainly tested for GENEActiv bin files) |
| file.too.short | File too short for processing? ([definition](#Minimum_recording_duration)) TRUE or FALSE |
| use.temperature | Temperature used for auto-calibration? TRUE or FALSE |
| scale.x | Auto-calibration scaling coefficient for x-axis (same for y and z axis, not shown here) |
| offset.x | Auto-calibration offset coefficient for x-axis (same for y and z axis, not shown here) |
| temperature.offset.x | Auto-calibration temperature offset coefficient for x-axis (same for y and z axis, not shown here) |
| cal.error.start | Calibration error prior to auto-calibration |
| cal.error.end | Calibration error after auto-calibration |
| n.10sec.windows | Number of 10 second epochs used as sphere data in auto-calibration |
| n.hours.considered | Number of hours of data considered for auto-calibration |
| QCmessage | Character QC message at the end of the auto-calibration |
| mean.temp | Mean temperature in sphere data |
| device.serial.number | Device serial number |
| NFilePagesSkipped | (Only for Axivity .cwa format) Number of raw data blocks skipped |
| filehealth_totimp_min | (Only for Axivity .cwa, ActiGraph gt3x, and ad-hoc csv format) Total number of minutes of raw data imputed |
| filehealth_checksumfail_min | (Only for Axivity .cwa format) Total number of minutes of raw data where the checksum failed |
| filehealth_niblockid_min | (Only for Axivity .cwa format) Total number of minutes of raw data with non-incremental block ids |
| filehealth_fbias0510_min | (Only for Axivity .cwa format) Total number of minutes with a sampling frequency bias between 5 and 10% |
| filehealth_fbias1020_min | (Only for Axivity .cwa format) Total number of minutes with a sampling frequency bias between 10 and 20% |
| filehealth_fbias2030_min | (Only for Axivity .cwa format) Total number of minutes with a sampling frequency bias between 20 and 30% |
| filehealth_fbias30_min | (Only for Axivity .cwa format) Total number of minutes with a sampling frequency bias higher than 30% |
| filehealth_totimp_N | (Only for Axivity .cwa, ActiGraph gt3x, and ad-hoc csv format) Total number of data blocks that were imputed |
| filehealth_checksumfail_N | (Only for Axivity .cwa format) Total number of blocks where the checksum failed |
| filehealth_niblockid_N | (Only for Axivity .cwa format) Total number of raw data blocks with non-incremental block ids |
| filehealth_fbias0510_N | (Only for Axivity .cwa format) Total number of raw data blocks with a sampling frequency bias between 5 and 10% |
| filehealth_fbias1020_N | (Only for Axivity .cwa format) Total number of raw data blocks with a sampling frequency bias between 10 and 20% |
| filehealth_fbias2030_N | (Only for Axivity .cwa format) Total number of raw data blocks with a sampling frequency bias between 20 and 30% |
| filehealth_fbias30_N | (Only for Axivity .cwa format) Total number of raw data blocks with a sampling frequency bias higher than 30% |

# GGIR Part 3

GGIR part 3, only outputs RData files which are used by GGIR part 4 and 5. The RData files are not intended for direct use by the GGIR user.

# GGIR Part 4

Part 4 generates the following output:

## Night level summaries (csv)

-   part4_nightsummary_sleep_cleaned.csv
-   QC/part4_nightsummary_sleep_full.csv

The latter with '\_full' in the name is intended to aid clarifying why some nights (if any) are excluded from the cleaned summary report. Although, nights where the accelerometer was not worn at all are excluded from this. So, if you have a 30 day recording where the accelerometer was not worn from day 7 onward then you will not find the last 22 nights in either csv-report.

The csv. files contain the variables as shown below.

| (Part of) variable name | Description |
|--------------------------|----------------------------------------------|
| ID | Participant ID extracted from file |
| night | Number of the night in the recording |
| sleeponset | Detected onset of sleep expressed as hours since the midnight of the previous night. |
| wakeup | Detected waking time (after sleep period) expressed as hours since the midnight of the previous night. |
| SptDuration | Difference between onset and waking time. |
| sleepparam | Definition of sustained inactivity by accelerometer. |
| guider | guider used as discussed in paragraph [Sleep analysis](#Sleep_analysis). |
| guider_onset | Start of Sleep Period Time window derived from the guider. |
| guider_wake | End of Sleep Period Time window derived guider. |
| guider_SptDuration | Time SPT duration derived from guider_wake and guider_onset. |
| error_onset | Difference between sleeponset and guider_onset |
| error_wake | Difference between wakeup and guider_wake |
| fraction_night_invalid | Fraction of the night (noon-noon or 6pm-6pm) for which the data was invalid, e.g. monitor not worn or no accelerometer measurement started/ended within the night. |
| SleepDurationInSpt | Total sleep duration, which equals the accumulated nocturnal sustained inactivity bouts within the Sleep Period Time. |
| duration_sib_wakinghours | Accumulated sustained inactivity bouts during the day. These are the periods we would label during the night as sleep, but during the day they form a subclass of inactivity, which may represent day time sleep or wakefulness while being motionless for a sustained period of time |
| number_sib_sleepperiod | Number of nocturnal sleep periods, with nocturnal referring to the Sleep Period Time window. |
| duration_sib_wakinghours_atleast15min | Same as duration_sib_wakinghours, but limited to SIBs that last at least 15 minutes. |
| num ber_sib_wakinghours | Number of sustained inactivity bouts during the day, with day referring to the time outside the Sleep Period Time window. |
| sleeponset_ts | sleeponset formatted as a timestamp |
| wakeup_ts | wakeup formatted as a timestamp |
| guider_onset_ts | guider_onset formatted as a timestamp |
| guider_wake_ts | guider_wake formatted as a timestamp |
| daysleeper | If 0 then the person is a nightsleeper (sleep period did not overlap with noon) if value=1 then the person is a daysleeper (sleep period did overlap with noon) |
| weekday | Day of the week on which the night started |
| calendardate | Calendar date on which the night started in day/month/year format. |
| filename | Name of the accelerometer file |
| cleaningcode | see paragraph [Cleaningcode](https://wadpac.github.io/GGIR/articles/chapter10_SleepAnalysis.html#cleaningcode) |
| sleeplog_used | Whether a sleep log was used (TRUE/FALSE) |
| acc_available | Whether accelerometer data was available (TRUE/FALSE). |
| WASO | Wake After Sleep Onset: SptDuration - SleepDurationInSpt |
| SptDuration | Sleep Period Time window duration: wakeup - sleeponset |
| error_onset | Difference between sleeponset and guider_onset (this variable is only available in the full report as stored in the QC folder) |
| error_wake | Difference between wakeup and guider_wake (this variable is only available in the full report as stored in the QC folder) |
| SleepRegularityIndex1 | The Sleep Regularity Index inspired by [Phillips et al. 2017](https://www.nature.com/articles/s41598-017-03171-4), but calculated per day-pair to enable user to study patterns across days, and calculated based on sustained inactivity bouts. See [Chapter 10](https://wadpac.github.io/GGIR/articles/chapter10_SleepAnalysis.html#sleep-regularity-index-sri) for details. |
| SriFractionValid | Fraction of the 24 hour period that was valid in both current as well as in matching timestamps for the next calendar day. See GGIR function manual for details |
| nonwear_perc_spt | Non-wear percentage during the spt hours of this day. This is a copy of the nonwear_perc_spt calculated in `part 5`, only included in part 4 reports if part 5 has been run with timewindow = WW |

**Non-default variables in part 4 csv report**

These additional are only stored if you used a sleeplog that captures time in bed, or when using guider HorAngle for hip-worn accelerometer data. If either of these applies set argument `sleepwindowType` to "TimeInBed".

| (Part of) variable name | Description |
|--------------------------|----------------------------------------------|
| guider_inbedStart | Time of getting in bed |
| guider_inbedEnd | Time of getting out of bed |
| guider_inbedDuration | Time in Bed: guider_inbedEnd - guider_inbedStart |
| sleepefficiency | Sleep efficiency, calculated by one of two metrics as controlled by argument `sleepefficiency.metric`: SleepDurationInSpt / guider_inbedDuration (default) or SleepDurationInSpt / (SptDuration + latency) |
| sleeplatency | Sleep latency, calculated as: sleeponset - guider_inbedStart |
| guider_corrected | When using the guider correction algortihm with parameter `guider_cor_do` this output variable can have the following values: **0** means that the guider was not correct; **1** indicates that only basic expansion with neighboring windows was applied as done by step 4 of the algorithm; **2** indicates that main guider window was changes as done by step 3 of the algorithm, and; **3** indicates that both the main guider window was changed and the window was expanded. |

## Person level summaries (csv)

-   part4_summary_sleep_cleaned.csv
-   QC/part4_summary_sleep_full.csv

In the person level report the variables are derived from the variables in the night level summary. Minor extensions to the variable names explain how variables are aggregated across the days. Please find below extra clarification on a few of the variable names for which the meaning may not be obvious:

| (Part of) variable name | Description |
|--------------------------|----------------------------------------------|
| `_mn` | mean across days |
| `_sd` | standard deviation across days |
| `_AD` | All days |
| `_WE` | Weekend days |
| `_WD` | Week days |
| sleeplog_used | Whether a sleeplog was available (TRUE) or not (FALSE) |
| sleep_efficiency_after_onset | Accelerometer derived sleep efficiency within the sleep period time calculated as the ratio between acc_SleepDurationInSpt and guider_SptDuration (denominator) or acc_SleepDurationInSpt and acc_SptDuration + latency (denominator), as defined with sleepefficiency.metric. Only available at person level, because at night level the user can calculate this from existing variables. |
| n_nights_acc | Number of nights of accelerometer data |
| n_nights_sleeplog | Number of nights of sleeplog data. |
| n_nights_markerbutton | Only stored when parameter `consider_marker_button` is set to TRUE, reflecting the number of valid nights with marker button data. |
| n_WE_nights_complete | Number of weekend nights complete which means both accelerometer and estimate from guider. |
| n_WD_nights_complete | Number of weekday nights complete which means both accelerometer and estimate from guider. |
| n_WEnights_daysleeper | Number of weekend nights on which the person slept until after noon. |
| n_WDnights_daysleeper | Number of weekday nights on which the person slept until after noon. |
| duration_sib_wakinghour | Total duration of sustained inactivity bouts during the waking hours. |
| number_sib_wakinghours | Number of sustained inactivity bouts during the waking hours. |
| average_dur_sib_wakinghours | Average duration of the sustained inactivity bouts during the day (outside the sleep period duration). Calculated as duration_sib_wakinghour divided by number_sib_wakinghours per day, after which the mean and standard deviation are calculated across days. |

## Visualisation (pdf)

Visualisation to support data quality checks: - visualisation_sleep.pdf (optional)

When input argument `do.visual` is set to TRUE GGIR can show the following visual comparison between the time window of being asleep (or in bed) according to the sleeplog and the detected sustained inactivity bouts according to the accelerometer data. This visualisation is stored in the results folder as `visualisation_sleep.pdf`.

Explanation of the image: Each line represents one night. Colours are used to distinguish definitions of sustained inactivity bouts (2 definitions in this case) and to indicate existence or absence of overlap with the sleeplog. When argument `outliers.only` is set to FALSE it will visualise all available nights in the dataset. If `outliers.only` is set to TRUE it will visualise only nights with a difference in onset or waking time between sleeplog and sustained inactivity bouts larger than the value of argument `criterror`.

This visualisation with outliers.only set to TRUE and critererror set to 4 was very powerful to identify entry errors in sleeplog data in van Hees et al PLoSONE 2015. We had over 25 thousand nights of data, and this visualisation allowed us to quickly zoom in on the most problematic nights to investigate possible mistakes in GGIR or mistakes in data entry.

```{r, out.width = "700px",echo=FALSE, fig.alt="Example visualisation GGIR part 4"}
knitr::include_graphics("example_dovisual.jpg")
```

# GGIR Part 5

The output of part 5 is dependent on the parameter configuration, it will generate as many output files as there are unique combination of the three thresholds provided.

For example, the following files will be generated if the threshold configuration was 30 for light activity, 100 for moderate and 400 for vigorous activity:

-   part5_daysummary_MM_L30M100V400_T5A5.csv
-   part5_daysummary_WW_L30M100V400_T5A5.csv
-   part5_personsummary_MM_L30M100V400_T5A5.csv
-   part5_personsummary_WW_L30M100V400_T5A5.csv
-   file summary reports/Report_nameofdatafile.pdf

## Day level summary (csv)

| (Term in) variable name | Description |
|--------------------------|----------------------------------------------|
| sleeponset  | onset of sleep expressed in hours since the midnight in the night preceding the night of interest, e.g. 26 is 2am.        |
| wakeup      | waking up time express in the same way as sleeponset.       |
| sleeponset_ts           | onset of sleep expressed as a timestamp hours:minutes:seconds           |
| daysleeper  | if 0 then the person woke up before noon, if 1 then the person woke up after noon    |
| cleaningcode| See paragraph [Cleaningcode](https://wadpac.github.io/GGIR/articles/chapter10_SleepAnalysis.html#cleaningcode).   |
| dur_day_spt_min         | Total length of daytime waking hours and spt combined (typically 24 hours for MM report).        |
| `dur_`      | duration of a behavioral class that will be specified int he rest of the variable name           |
| `ACC_`      | (average) acceleration according to default metric specific by acc.metric|
| `_spt_wake_`| Wakefulness within the Sleep period time window.|
| `_spt_sleep_`           | Sleep within the Sleep period time window.     |
| `_IN_`      | Inactivity. Note that we use the term inactivity instead of sedentary behaviour for the lowest intensity level of behaviour. The reason for this is that GGIR does not attempt to classifying the activity type sitting at the moment, by which we feel that using the term sedentary behaviour would fail to communicate that.|
| `_LIG_`     | Light activity        |
| `_MOD_`     | Moderate activity     |
| `_VIG_`     | Vigorous activity     |
| `_MVPA_`    | Moderate or Vigorous activity      |
| `_unbt_`    | Unbouted  |
| `_bts_`     | Bouts (also known as sojourns), which are segments that for which the acceleration is within a specified range for a specified fraction of the time.|
| `_bts_1_10_`| Bouts lasting at least 1 minute and less than 10 minutes (1 and 9.99 minutes are included, but 10 minutes is not).        |
| Nblock      | number of blocks of a certain behavioral class, not these are not bouts but a count of the number of times the behavioral class occurs without interruptions.   |
| WW          | in filename refers to analyses based on the timewindow from waking to waking up      |
| MM          | in filename refers to analyses done on windows between midnight and midnight         |
| calendar_date           | calendar date on which the window started in day/month/year format. So, for WW window this could mean that you have two windows starting on the same date.      |
| weekday     | weekday on which the window started. So, for WW window this could mean that you have two windows starting on the weekday. |
| `_total_IN` | total time spent in inactivity (no distinction between bouted or unbouted behavior, this is a simple count of the number of epochs that meet the threshold criteria.        |
| `_total_LIG`| total time spent in light activity.|
| nonwear_perc_day        | Non-wear percentage during the waking hours of this day.    |
| nonwear_perc_spt        | Non-wear percentage during the spt hours of this day.       |
| nonwear_perc_day_spt    | Non-wear percentage during the whole day, including waking and spt.     |
| dur_day_min | Duration of waking hours within this day window|
| dur_spt_min | Duration of Sleep Period Time within this day window.       |
| dur_day_spt_min         | Duration this day window, including both waking hours and SPT.          |
| sleep_efficiency_after_onset        | sleep_efficiency_after_onset in part 5 is not the same as in part 4, but calculated as the percentage of sleep within the sleep period time window. The conventional approach is the approach used in part 4.    |
| `ig_`       | All variables related to intensity gradient analysis |
| `_gradient`   | Gradient from intensity gradient analysis proposed by [Rowlands et al. 2018](https://doi.org/10.1249/mss.0000000000001561) for the waking hours window (`_day_`) and for the full window (`_day_spt_`)      |
| `_intercept`  | Intercept from intensity gradient analysis proposed by [Rowlands et al. 2018](https://doi.org/10.1249/mss.0000000000001561) for the waking hours window (`_day_`) and for the full window (`_day_spt_`)      |
| `_rsquared`   | r squared from intensity gradient analysis proposed by [Rowlands et al. 2018](https://doi.org/10.1249/mss.0000000000001561) for the waking hours window (`_day_`) and for the full window (`_day_spt_`)      |


### Fragmentation

| (Term in) variable name | Description |
|--------------------------|----------------------------------------------|
| `FRAG_`     | All variables related to behavioural fragmentation analysis |
| `TP_`       | Transition probability|
| PA2IN       | Physical activity fragments followed by inactivity fragments|
| IN2PA       | Physical inactivity fragments followed by activity fragments|
| Nfrag       | Number of fragments   |
| IN2LIPA     | Inactivity fragments followed by LIPA          |
| IN2MVPA     | Inactivity fragments followed by MVPA          |
| `mean_dur`  | mean duration of a fragment category           |
| `Gini_dur`  | Gini index|
| `CoV_dur`   | Coefficient of Variation           |
| alpha       | Power law exponent    |
| `x0.5`      | Derived from power law exponent alpha, see [Chastin et al. 201 0](https://doi.org/10.1016/j.gaitpost.2009.09.002)        |
| `W0.5`      | Derived from power law exponent alpha, see [Chastin et al. 201 0](https://doi.org/10.1016/j.gaitpost.2009.09.002)        |


**Special note if you are working on compositional data analysis:**

The duration of all `dur_` variables that have `_total_` in their name should add up to the total length of the waking hours in a day. Similarly, the duration of all other `dur_` variables excluding the variables `_total_` in their name and excluding the variable with `dur_day_min`, `dur_spt_min`, and `dur_day_spt_min` should also add up to the length of the full day.

**Motivation for default boutcriter.in = 0.9:**

The idea is that if you allow for bouts of 30 minutes it would not make sense to allow for breaks of 20 percent (6 minutes!) this is why I used a more stringent criteria for the highest category. Please note that you can change these criteria via arguments `boutcriter.mvpa`, `boutcriter.in`, and `boutcriter.lig`.

## Person level summary (csv)

Most variables in the person level summary are derived from the day level summary, but extended with `_pla` to indicate that the variable was calculated as the plain average across all valid days. Variables extended with `_wei` represent the weighted average of across all days where weekend days always weighted 2/5 relative to the contribution of week days.

| Variable name | Description |
|--------------------------|----------------------------------------------|
| Nvaliddays | Total number of valid days. |
| Nvaliddays_WD | Number of valid week days. |
| Nvaliddays_WE | Number of valid weekend days, where the days that start on Saturday or Sunday are considered weekend. |
| NcleaningcodeX | Number of days that had cleaning code X for the corresponding sleep analysis in part 4. In case of MM analysis this refers to the night at the end of the day. |
| Nvaliddays_AL10F_WD | Number of valid week days with at least 10 fragments (5 inactivity or 5 inactive) |
| Nvaliddays_AL10F_WE | Number of valid weekend days with at least 10 fragments (5 inactivity or 5 inactive) |
| `_wei` | weighted average of weekend and week days, using a 2/5 ratio, see above. |
| `_pla` | plain average of all days, see above |
| Nsleeplog_used | Number of valid windows with sleeplog used. |
| Nmarkerbutton_used | Only stored when parameter `consider_marker_button` is set to TRUE, reflecting the number of valid windows with marker button data for SPT estimation. |

# GGIR Part 6

Part 6 only stores a person level summary (csv). Most column names overlap with part 5, but are now derived based on the for the full time series only, whereas part 5 only presents these variables per day window which is then aggregated at person level.

### MXLX output variables

| Variable name | Description |
|--------------------------|----------------------------------------------|
| L5VALUE, M5VALUE     | Acceleration value for least (L5) and most (M5) active consecutive 5hrs per 24 hours     |
| L5TIME_clock, M5TIME_clock | Clock time of L5 and M5, e.g. 15:00 |
| L5TIME_num, M5TIME_num  | Timing of L5 and M5, expressed as hours relative to previous midnight, meaning that 2am would be 26. This is done to ease averaging acros days/persons. |
| M5_mean_peakLUX | Mean peak lux during M5 |
| M5_max_peakLUX | Max peak lux during M5 |

### Cosinor output variables

These variables names are consistent with the variable names [as discussed for GGIR part 2 output](https://wadpac.github.io/GGIR/articles/GGIRoutput.html#circadian-rhythm-cosinor).

### Other output variables in part 6

| Variable name | Description |
|--------------------------|----------------------------------------------|
| IS | Interdaily Stability |
| IV | Intradaily Variability |
| phi | See [section on phy in chapter 13](https://wadpac.github.io/GGIR/articles/chapter13_CircadianRhythm.html#phi) |
| `FRAG_` | Fragmentation variables, as also [discussed for part 5 output](https://wadpac.github.io/GGIR/articles/GGIRoutput.html#fragmentation). Only difference now is that fragmentation variables name ending with `_day` are specific to the waking hours of a day, while variable names ending iwht `_spt` are specific to the SPT window.  |
| SSP | See [section on SSP in chapter 13](https://wadpac.github.io/GGIR/articles/chapter13_CircadianRhythm.html#self-similarity-paramerter-ssp) |
| ABI | See [section on ABI in chapter 13](https://wadpac.github.io/GGIR/articles/chapter13_CircadianRhythm.html#activity-balance-index-abi) |
| SleepRegularityIndex2 | The Sleep Regularity Index inspired by [Phillips et al. 2017](https://www.nature.com/articles/s41598-017-03171-4), but calculated per day-pair to enable user to study patterns across days, and calculated based on classified naps and sleep. See [Chapter 10](https://wadpac.github.io/GGIR/articles/chapter10_SleepAnalysis.html#sleep-regularity-index-sri) for details. |
| SleepRegularityIndex2_Ndaypairs | Number of day pairs used for calculating SleepRegularityIndex2. |

\name{g.detecmidnight}
\alias{g.detecmidnight}
\title{
  Detect all midnights in a time series
}
\description{
  Detect all midnights in a time series
}
\usage{
  g.detecmidnight(time,desiredtz, dayborder)
}
\arguments{
  \item{time}{
    Vector of timestamps, either in iso8601 or in POSIX format
  }
  \item{desiredtz}{
    See \link{g.part2}
  }
    \item{dayborder}{
  see \link{g.analyse}
  }
  
}
\value{
  Output of the function is list containing the following objects:\cr
  \itemize{
  \item firstmidnight = timestamp of first midnight
  \item firstmidnighti = index of first midnight
  \item lastmidnight = timestamp of last midnight
  \item lastmidnighti = index of last midnight
  \item midnights = timestamps of midnights
  \item midnightsi = indeces of midnights
  }
}
\keyword{internal}
\author{
Vincent T van Hees <v.vanhees@accelting.com>
}

\name{is.ISO8601}
\alias{is.ISO8601}
\title{
Check whether character timestamp is in iso8601 format.
}
\description{
Checks whether timestamp stored in character format is in ISO8601 format or not}
\usage{
is.ISO8601(x)	
}
\arguments{
  \item{x}{
Timestamps in character format either in ISO8601 or as "yyyy-mm-dd hh:mm:ss".
}
}
\examples{
x ="1980-1-1 18:00:00"
is.ISO8601(x)
}

\name{CalcSleepRegularityIndex}
\alias{CalcSleepRegularityIndex}
\title{
  Calculates Sleep Regularity Index
}
\description{
  Calculates Sleep Regularity Index per day pair proposed by Phillips and 
  colleagues in 2017 expanded with day-pair level estimates.
}
\usage{
  CalcSleepRegularityIndex(data = c(), epochsize = c(), desiredtz= c(),
                            SRI1_smoothing_wsize_hrs = NULL,
                            SRI1_smoothing_frac = NULL)
}
\arguments{
  \item{data}{
    Data.frame produced by function \link{g.sib.det}.
  }
  \item{epochsize}{
    Numeric value of epoch size in seconds.
  }
  \item{desiredtz}{
    See \link{GGIR}.
  }
  \item{SRI1_smoothing_wsize_hrs}{
    See \link{GGIR}.
  }
  \item{SRI1_smoothing_frac}{
    See \link{GGIR}. 
  }
}
\value{
  Data.frame with columns: day (day number); Sleep Regularity Index, which by definition must lie in the 
  range -100 (reversed regularity), to 0 (random pattern), to 100 (perfect regularity);
  weekday (e.g. Wednesday); frac_valid, number between 0 and 1 indicating the fraction of the 24 hour
  period for which valid data was available in both the current and the next day, and; date.
}
\details{
  Calculates Sleep Regularity Index per day pair. Absense of missing data is not used 
  as a criteria for calculation. Instead the code asses the fraction of the time for which
  matching valid data points were found in both days. Later in g.part4 this fraction is used
  to include or exclude days based on the excludenightcrit criteria it also uses for the other
  sleep variables. In g.report.part4 these day-level SRI values are stored, but also
  aggregated across all recording days, all weekend days, and all weekend days, respectively. 
  Therefore, this function is broader in functionality than the algorithm proposed by Phillips and 
  colleagues in 2017.
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}
\references{
\itemize{
  \item Andrew J. K. Phillips, William M. Clerx, et al. Irregular sleep/wake patterns 
  are associated with poorer academic performance and delayed circadian and 
  sleep/wake timing. Scientific Reports. 2017 June 12
}
}

\name{getfolderstructure}
\alias{getfolderstructure}
\title{
  Extracts folderstructure based on data directory.
}
\description{
  Extracts folderstructure based on data directory. This is used when
  accelerometer files are stored in a hierarchical folder structure and
  the user likes to have a reference to the exact position in the folder
  tree, rather than just the filename.  Function not
  intended for direct use by package user.
}
\usage{
getfolderstructure(datadir=c(),referencefnames=c())	
}
\arguments{
  \item{datadir}{
    Argument datadir as used in various other functions in GGIR
  }
  \item{referencefnames}{
    vector with filename to filter on
  }
}
\value{
  List with items:
  \itemize{
    \item fullfilenames: vector with all full paths to the folders including the
    name of the file itself
    \item foldername: vector with only the names of the folder in which each file
    is stroed (so only the most distal folder in the folder tree).
  }
}
\keyword{internal}
\examples{
  \dontrun{
    folderstructure = getfolderstructure(datadir)
  }
}

\name{g.part5.addfirstwake}
\alias{g.part5.addfirstwake}
\title{
  Adds first wake if it is missing in part 4 output.
}
\description{
  Not intended for direct use by GGIR users.
  Adds first wake if it is missing in part 4 output
  as part of \link{g.part5}.
}
\usage{
  g.part5.addfirstwake(ts, summarysleep, nightsi, sleeplog,
  ID, Nepochsinhour, SPTE_end)
}
\arguments{
  \item{ts}{
    Data.frame object as passed on from \link{g.part5}
  }
  \item{summarysleep}{
    Data.frame object as passed on from \link{g.part5} with
    sleep summary information from \link{g.part4}.
  }
  \item{nightsi}{
    Vector with indices for the nights
  }
  \item{sleeplog}{
    Data.frame with all sleeplog information
  }
  \item{ID}{
    Participant ID
  }
  \item{Nepochsinhour}{
    Number of epochs in an hour
  }
  \item{SPTE_end}{
    Sleep period time end index
  }
}
\value{
  Data.frame ts updated with first wakeup time
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{part6PairwiseAggregation}
\alias{part6PairwiseAggregation}
\title{part6PairwiseAggregation}
\description{
  Pairwise aggregation of the time series of a group.
}
\usage{
  part6PairwiseAggregation(outputdir = NULL, desiredtz = "", verbose = TRUE)
}
\arguments{
  \item{outputdir}{Directory where you would like to store your results}
  \item{desiredtz}{Character, specifying the timezone database name 
    of the timezone the data was collected in}
  \item{verbose}{
    See details in \link{GGIR}.
  }
}
\value{
  No object is returned, only files are created in the output directory
}

\name{g.convert.part2.long}
\alias{g.convert.part2.long}
\title{
  Convert part 2 report to long format
}
\description{
  Not for direct access by used. This function is used 
  inside g.report.part2 and convert2 part 2 report to long 
  ormat if there
  are multiple segments per day
}
\usage{
g.convert.part2.long(daySUMMARY)
}
\arguments{
  \item{daySUMMARY}{
  Object available inside g.report.part2
  }
}
\value{
  Data.frame with long format version of daySUMMARY
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{addSplitNames}
\alias{addSplitNames}
\title{
  Add split names to output object in g.part4 and g.part5
}
\description{
  Internal function to add the recording split names to output the reports 
  for GGIR part 4 and 5.
}
\usage{
   addSplitNames(x)
}
\arguments{
  \item{x}{
    Output data.frame as generated inside \link{g.part4} and \link{g.part5}.
  }
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{ABI}
\alias{ABI}
\title{Activity balance index (ABI)}
\usage{
  ABI(x)
}
\arguments{
  \item{x}{the estimated self-similarity parameter (SSP)}
}
\value{
  The estimated Activity balance index (ABI) is a real number between zero and one.
}
\description{
  This function estimates the Activity balance index (ABI), which is a transformation of the self-similarity parameter (SSP), also known as scaling exponent or alpha.
}
\details{
  ABI = exp(-abs(SSP-1)/exp(-2))
}
\examples{
  # Estimate Activity balance index of a very known time series
  # available on R base: the sunspot.year.
  \dontrun{  
    ssp = SSP(sunspot.year)
    abi = ABI(ssp)
  }
}
\references{
  C.-K. Peng, S.V. Buldyrev, S. Havlin, M. Simons, H.E. Stanley, A.L. Goldberger Phys. Rev. E, 49 (1994), p. 1685
  Mesquita, Victor & Filho, Florencio & Rodrigues, Paulo. (2020). Detection of crossover points in detrended fluctuation analysis: An application to EEG signals of patients with epilepsy. Bioinformatics. 10.1093/bioinformatics/btaa955.
}
\author{
  Ian Meneghel Danilevicz <ian.meneghel-danilevicz@inserm.fr>
}

\name{g.report.part6}
\alias{g.report.part6}
\title{
  Generate report from milestone data produced by \link{g.part6}
}
\description{
  Creates report from milestone data produced by \link{g.part6}. Not intended
  for direct use by package user
}
\usage{
  g.report.part6(metadatadir = c(), f0 = c(), f1 = c(),
                          params_cleaning = NULL, params_output, 
                          verbose = TRUE)
}
\arguments{
  \item{metadatadir}{
    Directory that holds a folder 'meta' and inside this a folder 'basic'
    which contains the milestone data produced by \link{g.part1}. The folderstructure
    is normally created by \link{g.part1} and \link{GGIR} will recognise what
    the value of metadatadir is.
  }
  \item{f0}{
    File index to start with (default = 1). Index refers to the filenames sorted
    in alphabetical order
  }
  \item{f1}{
    File index to finish with (defaults to number of files available, i.e., f1 = 0)
  }
  \item{params_cleaning}{
    See details in \link{GGIR}.
  }
  \item{params_output}{
    Parameters object, see \link{GGIR}
  }
  \item{verbose}{
    See details in \link{GGIR}.
  }
}
\value{
  Function does not produce data, but only writes reports
  in csv format
  
  The following files are stored in the root of the results folder:
  part6_summary.csv
  
  See package vignette "HouseHoldCoanalysis".
}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{MXLX}
\alias{MXLX}
\title{
  MX LX analysis
}
\description{
  Identifies the most and least active X hours
}
\usage{
   MXLX(Y = NULL, X = 5, epochSize = 1, tseg = c(0, 24), resolutionMin = 10)
}
\arguments{
  \item{Y}{
    Vector with numeric time series
  }
  \item{X}{
    Single numeric value of X in hours
  }
  \item{epochSize}{
    Numeric epoch size in seconds of Y
  }
  \item{tseg}{
    Numeric vector of length two reflecting the time windownof Y in real clock hours.
    For example, if Y represent noon-midnight then
    set \code{tseg} to c(12, 24).
  }
  \item{resolutionMin}{
    Numeric value to indicate the resolution in minutes of the MX and LX search
  }
}
\value{
  A data.frame with the:
    \itemize{
      \item LX the average value of Y for LX
      \item LXhr the hour in the day at which LX starts
      \item start_LX the index of Y where LX starts
      \item end_LX the index of Y where LX ends
      \item MX the average value of Y for MX
      \item MXhr the hour in the day at which MX starts
      \item start_MX the index of Y where MX starts
      \item end_MX the index of Y where MX ends
    }
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.part1}
\alias{g.part1}
\title{
  function to load and pre-process acceleration files
}
\description{
  Calls function \link{g.getmeta} and \link{g.calibrate}, and converts the
  output to .RData-format which will be the input for \link{g.part2}. Here,
  the function generates a folder structure to keep track of various output files.
  The reason why these \link{g.part1} and \link{g.part2} are not merged as one
  generic shell function is because g.part1 takes much
  longer to and involves only minor decisions of interest to the movement scientist.
  Function g.part2 on the other hand is relatively fast and comes with all the
  decisions that directly impact on the variables that are of interest to the
  movement scientist. Therefore, the user may want to run g.part1 overnight
  or on a computing cluster, while g.part2 can then be the main playing ground
  for the movement scientist. Function \link{GGIR} provides the main shell
  that allows for operating g.part1 and g.part2.
}
\usage{
  g.part1(datadir = c(), metadatadir = c(), f0 = 1, f1 = c(),
          myfun = c(), params_metrics = c(), params_rawdata = c(),
          params_cleaning = c(), params_general = c(), verbose = TRUE, ...)
}

\arguments{
  \item{datadir}{
    Directory where the accelerometer files are stored, e.g. 
    "C:/mydata", or list of accelerometer filenames and directories, e.g. 
    c("C:/mydata/myfile1.bin", "C:/mydata/myfile2.bin").
  }
  \item{metadatadir}{
    Directory where the output needs to be stored. Note that this 
    function will attempt to create folders in this directory and uses
    those folder to keep output.
  }
  \item{f0}{
    File index to start with (default = 1). Index refers to the filenames sorted
    in alphabetical order
  }
  \item{f1}{
    File index to finish with (defaults to number of files available, i.e., f1 = 0)
  }
  \item{myfun}{
    External function object to be applied to raw data.
    See details \link{applyExtFunction}.
  }
  \item{params_metrics}{
   See details in \link{GGIR}.
  }
  \item{params_rawdata}{
   See details in \link{GGIR}.
  }
  \item{params_cleaning}{
   See details in \link{GGIR}.
  }
  \item{params_general}{
   See details in \link{GGIR}.
  }
  \item{verbose}{
    See details in \link{GGIR}.
  }
  \item{...}{
    If you are working with a non-standard csv formatted files, \link{g.part1} also takes any input
    arguments needed for function \link{read.myacc.csv} and argument rmc.noise from
    \link{get_nw_clip_block_params}. First test these argument with function \link{read.myacc.csv} directly.
    To ensure compatibility with R scripts written for older GGIR versions, the user can also
    provide parameters listed in the params_ objects as direct argument.
  }
}
\details{
  GGIR comes with many processing parameters, which have been thematically grouped in
  parameter objects (R list). By running print(load_params()) you can
  see the default values of all the parameter objects. When g.part 1 is used via \link{GGIR}
  you have the option to specifiy a configuration file, which will overrule the default
  parameter values. Further, as user you can set parameter values as input argument to both \link{g.part1}
  and \link{GGIR}. Directly specified argument overrule the configuration file and default values.
  
  See the GGIR package vignette or the details section in \link{GGIR} for a more elaborate overview of parameter objects and their usage across GGIR.
  
}
\value{
 The function provides no values, it only ensures that the output from other
 functions is stored in .RData(one file per accelerometer file) in folder structure
}
\examples{
  \dontrun{
    datafile = "C:/myfolder/mydata"
    outputdir = "C:/myresults"
    g.part1(datadir,outputdir)
  }
}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}
\references{
  \itemize{
    \item van Hees VT, Gorzelniak L, Dean Leon EC, Eder M, Pias M, et al. (2013) Separating
      Movement and Gravity Components in an Acceleration Signal and Implications for the
      Assessment of Human Daily Physical Activity. PLoS ONE 8(4): e61691.
      doi:10.1371/journal.pone.0061691
    \item van Hees VT, Fang Z, Langford J, Assah F, Mohammad A, da Silva IC, Trenell MI,
      White T, Wareham NJ, Brage S. Auto-calibration of accelerometer data for
      free-living physical activity assessment using local gravity and temperature:
      an evaluation on four continents. J Appl Physiol (1985). 2014 Aug 7
    \item Aittasalo M, Vaha-Ypya H, Vasankari T, Husu P, Jussila AM, and Sievanen H. Mean
      amplitude deviation calculated from raw acceleration data: a novel method for
      classifying the intensity of adolescents physical activity irrespective of accelerometer
      brand. BMC Sports Science, Medicine and Rehabilitation (2015).
  }
}

\name{get_starttime_weekday_truncdata }
\alias{get_starttime_weekday_truncdata}
\title{
  Get starttime (adjusted), weekday, and adjust data accordingly.
}
\description{
  Function not intended for direct use by user.
  Used inside \link{g.getmeta} as an intermediate step between
  loading the raw data and calibrating it. This step includes extracting
  the starttime and adjusting it to nearest integer number of long epoch window
  lengths in an hour, truncating the data accordingly, and extracting the 
  corresponding weekday.
}
\usage{
  get_starttime_weekday_truncdata(monc, 
  dformat, data, header, desiredtz, sf,
  datafile, ws2, configtz = NULL)
}
\arguments{
  \item{monc}{
    See \link{g.inspectfile}
  }
  \item{dformat}{
    See \link{g.dotorcomma}
  }
  \item{data}{
    Data part of \link{g.readaccfile} output
  }
  \item{header}{
    Header part of \link{g.readaccfile} output
  }
  \item{desiredtz}{
    See \link{g.getmeta}
  }
  \item{sf}{
    Numeric, sample frequency in Hertz
  }
  \item{datafile}{
    See \link{g.getmeta}
  }
  \item{ws2}{
    Long epoch length
  }
   \item{configtz}{
    See \link{g.getmeta}
  }
}
\keyword{internal}
\author{
Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.report.part4}
\alias{g.report.part4}
\title{
  Generate report from milestone data produced by \link{g.part4}
}
\description{
  Creates report from milestone data produced by \link{g.part4}. Not intended
  for direct use by package user
}
\usage{
g.report.part4(datadir = c(), metadatadir = c(), f0 = c(),
  f1 = c(), data_cleaning_file = c(),
  params_sleep, params_output, verbose = TRUE)
}
\arguments{
  \item{datadir}{
    Directory where the accelerometer files are stored, e.g. 
    "C:/mydata", or list of accelerometer filenames and directories, e.g. 
    c("C:/mydata/myfile1.bin", "C:/mydata/myfile2.bin").
  }
  \item{metadatadir}{
    Directory that holds a folder 'meta' and inside this a folder 'basic'
    which contains the milestone data produced by \link{g.part1}. The folderstructure
    is normally created by \link{g.part1} and \link{GGIR} will recognise what
    the value of metadatadir is.
  }
  \item{f0}{
    File index to start with (default = 1). Index refers to the filenames sorted
    in alphabetical order
  }
  \item{f1}{
    File index to finish with (defaults to number of files available, i.e., f1 = 0)
  }
  \item{data_cleaning_file}{
    see \link{GGIR}
  }
  \item{params_sleep}{
    Parameters object, see \link{GGIR}
  }
  \item{params_output}{
    Parameters object, see \link{GGIR}
  }
  \item{verbose}{
    See details in \link{GGIR}.
  }
}
\value{
  Function does not produce data, but only writes reports
  in csv format and a visual report in pdf.
  
  The following files are stored in the root of the results folder:
  part4_nightsummary_sleep_cleaned.csv
  part4_summary_sleep_cleaned.csv
  
  The following files are stored in the folder results/QC:
  part4_nightsummary_sleep_full.csv
  part4_summary_sleep_full.csv
  
  If a sleeplog is used *_full.csv as stored in the QC folder includes estimates 
  for all nights in the data, and *_cleaned.csv in the results folder includes
  estimates for all nights in the data excluding the nights that did not had a
  sleeplog entry or had no valid accelerometer data.
  
  If a sleep log is not used then * _cleaned.csv includes the nights that are
  in *_full.csv excluding the nights with insufficient data.
  
  If you have a study where the sleeplog was available for a subset of the
  participants, but you want to include all individuals in your analysis, 
  then use the *_full.csv output and clean the night level data yourself
  by excluding  rows with cleaningcode > 1 which are the cases where no
  or invalid accelerometer data was present.
  
  The above means that for studies with missing sleeplog entries for some
  individuals and some nights using the *_full.csv output and excluding
  rows (nights) with cleaningcode > 1 will lead to the same as
  * _cleaned.csv plus sleep estimates for the nights with missing
  sleeplog, providing that there was enough accelerometer data for 
  those nights.
  
  In other words, *_cleaned.csv is perfect if you only want to rely on 
  nights with a sleeplog or if you do not use a sleeplog at all. For all
  other scenarios We advise using the *_full.csv report and to clean it 
  yourself.
  
  See package vignette sections "Sleep analysis" and "Output part 4"
  for a more elaborative description of the sleep analysis and reporting.
}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{extract_params}
\alias{extract_params}
\title{
  Extract parameters from input and add them to params
}
\description{
  Extracts parameters separately provided by input and adds them to the params objects.
  Not intended for direct use by GGIR users.
}
\usage{
  extract_params(params_sleep = c(), params_metrics = c(),
                 params_rawdata = c(), params_247 = c(),
                 params_phyact = c(), params_cleaning = c(),
                 params_output = c(), params_general = c(), input = c(),
                 configfile_csv = c(), params2check = c("sleep", "metrics",
                 "rawdata", "247", "phyact",
                                           "cleaning", "output", "general"))
}
\arguments{
  \item{params_sleep}{
    List with sleep parameters
  }
  \item{params_metrics}{
    List with parameters related to metrics
  }
  \item{params_rawdata}{
    List with parameters related to raw data reading and processing
  }
  \item{params_247}{
    List with parameters related to 24/7 behavioural analysis, which includes anything
    that does not fit with physical activity or sleep research
  }
  \item{params_phyact}{
    List with parameters related to physical activity analysis
  }
  \item{params_cleaning}{
    List with parameters related to cleaning the time series, including masking and imputation
  }
  \item{params_output}{
    List with parameters related to how GGIR stores its output
  }
  \item{params_general}{
    List with parameters related to general topics
  }
  \item{input}{
    All objects provided by users
  }
  \item{configfile_csv}{
    Csv configuration file
  }
  \item{params2check}{
    Character vector to indicate which params objects need to be checked.
    This allows us to prevent the function from checking params objects that are not used
    in the context where function extract_params is used.
  }
}
\value{
  Lists of updated parameter objects 
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{get_nw_clip_block_params}
\alias{get_nw_clip_block_params}
\title{
  Set monitor brand specific parameters
}
\description{
  Set monitor brand specific thresholds for non-wear detection, clipping
  etection, and blocksizes to be loaded.
  Not designed for direct use by user.
}
\usage{
  get_nw_clip_block_params(monc, dformat, deviceSerialNumber = "", sf,
                            params_rawdata)
}
\arguments{
  \item{monc}{
    See \link{g.inspectfile}
  }
  \item{dformat}{
    See \link{g.dotorcomma}
  }
  \item{deviceSerialNumber}{
    As produced by \link{g.extractheadervars}
  }
  \item{sf}{
    Numeric, sample frequency in Hertz
  }
  \item{params_rawdata}{
    See \link{GGIR}
  }
}
\keyword{internal}
\author{
Vincent T van Hees <v.vanhees@accelting.com>
}

\name{HASPT}
\alias{HASPT}
\title{
  Heuristic Algorithms estimating SPT window.
}
\description{
  As used in function \link{g.sib.det}. Function is not intended for direct use
  by GGIR user.
}
\usage{
HASPT(angle, params_sleep = NULL, ws3 = 5, 
      HASPT.algo="HDCZA", invalid, activity = NULL,
      marker = NULL, sibs = NULL)
}

\arguments{
  \item{angle}{
    Vector of epoch level estimates of angle
  }
  \item{params_sleep}{
    See \link{GGIR}
  }
  \item{ws3}{
    Number representing epoch length in seconds
  }
  \item{HASPT.algo}{
    See \link{GGIR}
  }
  \item{invalid}{
    Integer vector with per epoch an indicator of valid(=0) or invalid(=1) epoch.
  }
  \item{activity}{
    Numeric vector with magnitude of acceleration per epoch,
    only used when HASPT.algo is set to NotWorn.
    Acceleration metric used is specified by argument \code{acc.metric} 
    elsewhere in GGIR.
  }
  \item{marker}{
    Optional, numeric vector with for each epoch an indication of whether the
    marker button was pressed (1) or not (0). Only used for Actiwatch-like devices
    such as MotionWare.
  }
  \item{sibs}{
    Numeric vector of classified sibs with \code{HASIB}, experimental not used at the moment.
  }
}
\value{
  List with start and end times of the SPT window and the threshold as used.
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{HASIB}
\alias{HASIB}
\title{
  Heuristic algorithms for sustiained inactivty bouts detection
}
\description{
  Apply heuristic algorithms for sustiained inactivty bouts detection.
  Function not intended for direct use by package user
}
\usage{
  HASIB(HASIB.algo = "vanHees2015", timethreshold = c(), anglethreshold = c(), 
                 time = c(), anglez = c(), ws3 = c(), zeroCrossingCount = c(),
                 NeishabouriCount = c(), activity = NULL,
                 oakley_threshold = NULL)
}
\arguments{
  \item{HASIB.algo}{
    Character to indicator which sib algorithm should be used.
    Default value: "vanHees2015". Other options: "Sadeh1994", "Galland2012", "ColeKripke1992"
  }
  \item{anglethreshold}{
    See \link{g.sib.det}
  }
  \item{timethreshold}{
    See \link{g.sib.det}
  }
  \item{time}{
    Vector with time per short epoch
  }
  \item{anglez}{
    Vector with z-angle per short epoch
  }
  \item{ws3}{
    See \link{g.getmeta}
  }
  \item{zeroCrossingCount}{
    Vector with zero crossing counts per epoch as required for count-based algorithms
  }
  \item{NeishabouriCount}{
    Vector with Neishabouri counts per epoch to be used by the count-based algorithms
  }
  \item{activity}{
    Magnitude of acceleration, only used when HASIB.algo is set to NotWorn or Oakley1997.
    Acceleration metric used is specified by argument \code{acc.metric} 
    elsewhere in GGIR.
  }
  \item{oakley_threshold}{
    See \link{GGIR}
  }
}
\value{
  Vector with binary indicator of sustained inactivity bout, 1 is yes, 0 is no.
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.readtemp_movisens}
\alias{g.readtemp_movisens}
\title{
Reads the temperature from movisens files.
}
\description{
  Reads the temperature from movisens files, resamples it and adds 
  it to the matrix where accelerations are stored
}
\usage{
g.readtemp_movisens(datafile, from = c(), to = c(), acc_sf, acc_length,
                    interpolationType=1)
}
\arguments{
  \item{datafile}{
    Full path to the folder where the movisens bin files are stored. Note that 
    movisens store a set of bin file in one folder per recording. GGIR will read 
    the pertinent bin file to access to the temperature data.
  }
  \item{from}{
    Origin point to derive the temperature from movisens files (automatically 
    calculated by GGIR)
  }
  \item{to}{
    End point to derive the temperature from movisens files (automatically 
    calculated by GGIR)
  }
  \item{acc_sf}{
    Sample frequency of acceleration data
  }
  \item{acc_length}{
    number of acceleration data samples
  }
  \item{interpolationType}{
    Integer to indicate type of interpolation to be used when resampling time series (mainly relevant for Axivity sensors), 1=linear, 2=nearest neighbour.
  }
}
\value{
  Data matrix with the temperature values resampled at 64 Hz.
}
\keyword{internal}
\examples{
\dontrun{
  P = g.readtemp_movisens(datafile, from = c(), to = c(), acc_sf = 64, acc_length = 3000)
}
}

\name{g.part5.definedays}
\alias{g.part5.definedays}
\title{
  Fix missing night in part 4 output
}
\description{
  Not intended for direct use by GGIR users.
  Defines when day windows start and end as part of \link{g.part5}.
}
\usage{
  g.part5.definedays(nightsi, wi, indjump,
                     epochSize, qqq_backup = c(), ts,
                     timewindowi, Nwindows, qwindow, ID = NULL,
                     dayborder = 0)
}
\arguments{
  \item{nightsi}{
    Vector with indices for the midnights
  }
  \item{wi}{
    Numeric to indicate window number
  }
  \item{indjump}{
    Number of indices to jump
  }
  \item{epochSize}{
    Numeric epoch size in seconds
  }
  \item{qqq_backup}{
    Backup of qqq object, which holds the start and end indices of a window
  }
  \item{ts}{
    Data.frame with time series
  }
  \item{timewindowi}{
    Timewindow definition either "MM" or "WW"
  }
  \item{Nwindows}{
    Number of windows in the data
  }
  \item{qwindow}{
    qwindow argument
  }
  \item{ID}{
    ID of participant
  }
  \item{dayborder}{
    dayborder argument
  }
}
\value{
  List of qqq and qqq_backup 
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{check_params}
\alias{check_params}
\title{
  Check default parameters
}
\description{
  Checks parameter objects for class and logical combinations.
  Called from \link{extract_params}. Not intended for direct use by GGIR users.
}
\usage{
  check_params(params_sleep = c(), params_metrics = c(),
                        params_rawdata = c(), params_247 = c(),
                        params_phyact = c(), params_cleaning = c(),
                         params_output = c(), params_general = c())
}
\arguments{
 \item{params_sleep}{
    List with sleep parameters
  }
  \item{params_metrics}{
    List with parameters related to metrics
  }
  \item{params_rawdata}{
    List with parameters related to raw data reading and processing
  }
  \item{params_247}{
    List with parameters related to 24/7 behavioural analysis, which includes anything
    that does not fit with physical activity or sleep research
  }
  \item{params_phyact}{
    List with parameters related to physical activity analysis
  }
  \item{params_cleaning}{
    List with parameters related to cleaning the time series, including masking and imputation
  }
  \item{params_output}{
    List with parameters related to how GGIR stores its output
  }
  \item{params_general}{
    List with parameters related to general topics
  }
}
\value{
  Lists of updated parameter objects 
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{getSplitNames}
\alias{getSplitNames}
\title{
  get split names from filename
}
\description{
  Internal function to get split names from filename if recording was split.
}
\usage{
   getSplitNames(filename)
}
\arguments{
  \item{filename}{
    Character of filename
  }
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{tidyup_df}
\alias{tidyup_df}
\title{
  Round numeric columns and replace NA/NaN values by blank
}
\description{
  Identifies columns that can be coerced to numeric in a data frame, 
  transforms these columns to numeric and round them to the specified digits. 
  It also replaces NA and NaNs values by blank.
}
\usage{
  tidyup_df(df = c(), digits = 3)
}
\arguments{
  \item{df}{
    Data frame
}
\item{digits}{
    Integer indicating the number of decimal places (round) or significant digits (signif) to be used
}
}
\value{
  Data frame with all possible columns as numeric and rounded to the specified number of digits
}

\author{
  Jairo H Migueles
}
\keyword{internal}
\examples{
  # Test data frame
  df = data.frame(a = c("a", "b"), b = as.character(c(1.543218, 8.216856483)))
  tidyup_df(df = df, digits = 3)
}

\name{checkMilestoneFolders}
\alias{checkMilestoneFolders}
\title{
  Checks for existence of folders to process
}
\description{
  Checks whether milestone folders exist, create them if needed,
  and check whether folders are not empty. Only done for part 1 to 5
  and not part 6, which is different and handled inside \link{g.part6}.
}
\usage{
  checkMilestoneFolders(metadatadir, partNumber)
}
\arguments{
  \item{metadatadir}{
    Character, path to root of outputfolder.
  }
  \item{partNumber}{
    Numeric, number from the set 2, 3, 4 or 5.
  }
}
\value{
  No value is produced
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{create_test_acc_csv}
\alias{create_test_acc_csv}
\title{
Creates csv data file for testing purposes
}
\description{
Creates file in the Actigraph csv data format with dummy data
that can be used for testing. The file includes accelerometer data 
with bouts of higher acceleration, variations non-movement periods
in a range of accelerometer positions to allow for testing the
auto-calibration functionality.}
\usage{
  create_test_acc_csv(sf=3,Nmin=2000,storagelocation=c(),
                      start_time = NULL, starts_at_midnight = FALSE)
}
\arguments{
  \item{sf}{
    Sample frequency in Hertz, the default here is low to minimize file size
  }
  \item{Nmin}{
    Number of minutes (minimum is 720)
  }
  \item{storagelocation}{
    Location where the test file named testfile.csv will be stored
    If no value is provided then the function uses the current 
    working directory
  }
  \item{start_time}{
    Start time of the recording, in the hh:mm:ss format.
  }
  \item{starts_at_midnight}{
    Boolean indicating whether the recording should start at 
    midnight. Ignored if start_time is specified.
  }
}
\value{
 The function does not produce any output values. Only the file is
 stored
}
  
\examples{
  \dontrun{
    create_test_acc_csv()
  }
}

\name{inspect_binFile_brand}
\alias{inspect_binFile_brand}
\title{Identify the Device Brand from a Binary File}
\description{
  This function identifies the brand of a device (e.g., GENEActiv or MATRIX) based on the contents 
  of a binary file. If the file is unrecognized, it returns a default value indicating the device is not recognized.
}
\usage{
inspect_binFile_brand(filename)
}
\arguments{
  \item{filename}{
    The path to the binary file to be inspected.
  }
}
\details{
  The function performs the following steps:
  \enumerate{
    \item Reads the file header to check for the presence of a \code{"Device Type"} field.
    \item If the device type contains the string \code{"GENEActiv"}, the device is identified as a GENEActiv device.
    \item If the \code{"Device Type"} field is not found, it checks for a MATRIX device by validating a specific header value (\code{"MDTC"}).
    \item If neither device type is detected, the function returns \code{"not_recognised"}.
  }
}
\value{
  A character string representing the detected device brand:
  \itemize{
    \item \code{2}: Indicates the device is a GENEActiv device.
    \item \code{7}: Indicates the device is a MATRIX device.
    \item \code{"not_recognised"}: Indicates the file's brand is not recognized.
  }
}
\author{
  Jairo H Migueles <jairo@jhmigueles.com>
}

\name{g.analyse}
\alias{g.analyse}
\title{
  Function to analsyse meta-data generated by \link{g.getmeta} and \link{g.impute}
}
\description{
  Analyses the output from other functions within the packages to generate a basic
  descriptive summary for each accelerometer data file. Analyses include: Average 
  acceleration per day, per measurement, L5M5 analyses (assessment of the five
  hours with lowest acceleration and with highest acceleration). Further, the 
  traditionally popular variable MVPA is automatically extracted in six variants:
  without bout criteria in combination with epoch = epoch length as defined in
  g.getmeta (first value of the input argument windowsizes), 1 minute,
  and 5 minutes, and for bout durations 1 minute, 5 minutes or 10 minutes in
  combination with the epoch length as defined in g.getmeta.
}
\usage{
  g.analyse(I, C, M, IMP, params_247 = c(), params_phyact = c(),
                      params_general = c(), params_cleaning = c(),
                      quantiletype = 7, myfun = c(), ID, ...)
}
\arguments{
  \item{I}{
    the output from function \link{g.inspectfile}
  }
  \item{C}{
    the output from function \link{g.calibrate}
  }
  \item{M}{
    the output from function \link{g.getmeta}
  }
  \item{IMP}{
    the output from function \link{g.impute}
  }
  \item{params_247}{
    See \link{GGIR}
  }
  \item{params_phyact}{
    See \link{GGIR}
  }
  \item{params_general}{
    See \link{GGIR}
  }
  \item{params_cleaning}{
    See \link{GGIR}
  }
  \item{quantiletype}{
    type of quantile function to use (default recommended). For details, see 
    quantile function in STATS package
  }
  \item{myfun}{
    External function object to be applied to raw data, see \link{g.getmeta}.
  }
  \item{ID}{
    ID extracted in g.part2.
  }
  \item{...}{
     Any argument used in the previous version of g.analyse, which will now
     be used to overrule the arguments specified with the parameter objects.
  }
}
\value{
  g.analyse generated two data,franeL
  \item{\code{summary}}{summary for the file that was analysed}
  \item{\code{daysummary}}{summary per day for the file that was analysed}
  These data.frames are used by function g.report.part2 to generate csv reports.
  An exaplantion of all the columns in the data.frame and subsequent csv reports
  can be found in the package vignette (Output part 2).
}
\examples{
  data(data.getmeta)
  data(data.inspectfile)
  data(data.calibrate)
  \dontrun{
    #inspect file:
    I = g.inspectfile(datafile)
    
    #autocalibration:
    C = g.calibrate(datafile) 
    
    #get meta-data:
    M = g.getmeta(datafile, desiredtz = "Europe/London", 
    windowsizes = c(5, 900, 3600),
    daylimit = FALSE, offset = c(0, 0, 0), 
    scale = c(1, 1, 1), tempoffset = c(0, 0, 0))
  }
  #impute meta-data:
  IMP = g.impute(M = data.getmeta, I = data.inspectfile, ID = "01wk0")
  
  #analyse and produce summary:
  A = g.analyse(I = data.inspectfile, C = data.calibrate,
  M = data.getmeta, IMP, ID = "01wk0")
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{splitRecords}
\alias{splitRecords}
\title{
  Append GGIR milestone data from neighbouring or overlapping recordings
}
\description{
  Splits GGIR part 1 milestone data format files based on 
  params_general parameters \code{recording_split_times}, 
  \code{recording_split_timeformat}, and \code{recording_split_times}.
}
\usage{
   splitRecords(metadatadir, params_general = NULL)
}
\arguments{
  \item{metadatadir}{
    See \link{g.part2}
  }
  \item{params_general}{
    See \link{GGIR}
  }
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{apply_cosinor_IS_IV_Analyses}
\alias{apply_cosinor_IS_IV_Analyses}
\title{
  Apply Cosinor Analyses to time series
}
\description{
  Wrapper function around \link{cosinor_IS_IV_Analyses} that first prepares the time series
  before applying the cosinorAnlayses
}
\usage{
  apply_cosinor_IS_IV_Analyses(ts, qcheck, midnightsi, epochsizes, threshold = NULL)
}
\arguments{
  \item{ts}{
    Data.frame with timestamps and acceleration metric.
  }
  \item{qcheck}{
    Vector of equal length as number of rows in ts with value 1 for invalid
    timestamps, 0 otherwise.
  }
  \item{midnightsi}{
    Indices for midnights in the time series
  }
  \item{epochsizes}{
    Epoch size for ts and qcheck respectively
  }
  \item{threshold}{
    See \link{cosinor_IS_IV_Analyses}
  }
}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.getM5L5}
\alias{g.getM5L5}
\title{
  Extract M5 and L5 from time series
}
\description{
  Extract M5 and L5 from time series, function used by \link{g.analyse} and
  not intended for direct use by package user. Please see \link{g.analyse}
  for further clarification on functionalities
}
\usage{
  g.getM5L5(varnum, epochSize, t0_LFMF, t1_LFMF, M5L5res, winhr, qM5L5 = c(), 
            iglevels = c(), MX.ig.min.dur = 10, UnitReScale = 1000)	
}
\arguments{
  \item{varnum}{
    Numeric vector of epoch values
  }
  \item{epochSize}{
    Small epoch size in seconds
  }
  \item{t0_LFMF}{
    Start hour of the day for the M5L5 analyses, e.g. 0 for midnight
  }
  \item{t1_LFMF}{
    End hour of the day for the M5L5 analyses, e.g. 24 for midnight
  }
  \item{M5L5res}{
    Resolution of hte M5L5 analyses in minutes
  }
  \item{winhr}{
    windowsize of M5L5 analyses, e.g. 5 hours
  }
  \item{qM5L5}{
    Percentiles (quantiles) to be calculated over L5 and M5 window.
  }
  \item{iglevels}{
    See  \link{g.analyse}. If provided then the intensity gradient will be calculated
    for all MX windows larger or equal than argument MX.ig.min.dur
  }
  \item{MX.ig.min.dur}{
    Minimum MX duration needed in order for intensity gradient to be calculated
  }
  \item{UnitReScale}{
    Numeric value with which acceleration values are multiple to rescale
  }
}
\value{
  \itemize{
  \item DAYL5HOUR = Starting time in hours of L5
  \item DAYL5VALUE = average acceleration during L5
  \item DAYM5HOUR = Starting time in hours of M5
  \item DAYM5VALUE = average acceleration during M5
  \item V5NIGHT = average acceleration between 1am and 6am
  }
}
\examples{
  \dontrun{
    data(data.getmeta)
    g.getM5L5(varnum=data.getmeta,ws3=5,t0_LFMF=0,t1_LFMF=24,M5L5res=10,winhr=5)
  }
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.applymetrics}
\alias{g.applymetrics}
\title{
  Extract metrics from acceleration signals
}
\description{
  Function to extract metrics from acceleration signal. Not intended
  for direct use by user
}
\usage{
  g.applymetrics(data, sf, ws3, metrics2do,
                          n = 4, lb = 0.2, hb = 15,
                          zc.lb = 0.25, zc.hb = 3, 
                          zc.sb = 0.01, zc.order = 2,
                          actilife_LFE = FALSE)
}
\arguments{
  \item{data}{
    Three column matrix with x, y, and z acceleration data
  }
  \item{n}{
    filter order, see \link{GGIR} for details
  }
  \item{sf}{
     sample frequency
  }
   \item{ws3}{
     Epoch size in seconds
  }
  \item{metrics2do}{
    Dataframe with Boolean indicator for all metrics whether
    they should be extracted or not. For instance,
    metrics2do$do.bfen = TRUE, indicates that the bfen metric
    should be extracted
  }
  \item{lb}{
    Lower boundery of cut-off frequencies, see \link{GGIR}.
  }
  \item{hb}{
    Higher boundery of cut-off frequencies, see \link{GGIR}.
  }
  \item{zc.lb}{
    See \link{GGIR}
  }
  \item{zc.hb}{
    See \link{GGIR}
  }
  \item{zc.sb}{
    See \link{GGIR}
  }
  \item{zc.order}{
    See \link{GGIR}
  }
  \item{actilife_LFE}{
    See \link{GGIR}
  }
}
\value{
  Dataframe with metric values in columns average per epoch (ws3)
}

\examples{
  Gx = runif(n=10000,min=0,max=2)
  Gy = runif(n=10000,min=1,max=3)
  Gz = runif(n=10000,min=0,max=2)
  data = cbind(Gx, Gy, Gz)
  colnames(data) = c("x", "y", "z")
  metrics2do = data.frame(do.bfen=TRUE,do.enmo=TRUE,do.lfenmo=FALSE,
  do.en=FALSE,do.hfen=FALSE,do.hfenplus=FALSE,do.mad=FALSE,do.anglex=FALSE,
  do.angley=FALSE,do.anglez=FALSE,do.roll_med_acc_x=FALSE,
  do.roll_med_acc_y=FALSE,do.roll_med_acc_z=FALSE,
  do.dev_roll_med_acc_x=FALSE,do.dev_roll_med_acc_y=FALSE,
  do.dev_roll_med_acc_z=FALSE,do.enmoa=FALSE,
  do.lfx=FALSE, do.lfy=FALSE, do.lfz=FALSE, 
  do.hfx=FALSE, do.hfy=FALSE, do.hfz=FALSE, 
  do.bfx=FALSE, do.bfy=FALSE, do.bfz=FALSE,
  do.zcx=FALSE, do.zcy=FALSE, do.zcz=FALSE, 
  do.brondcounts=FALSE, do.neishabouricounts=FALSE)
  
  extractedmetrics = g.applymetrics(data,n=4,sf=40,ws3=5,metrics2do)
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.dotorcomma}
\alias{g.dotorcomma}
\title{
  Assesses whether decimals in fileheader are stored with comma
  or dot separated decimals
}
\description{
  The function is used by \link{g.readaccfile} to assess how numeric data
  should be interpretted
}
\usage{
  g.dotorcomma(inputfile, dformat, mon, ...)
}
\arguments{
  \item{inputfile}{
    full path to inputfile
  }
  \item{dformat}{
    Data format code: 1=.bin, 2=.csv, 3=.wav, 4=.cwa, 5=.csv for ad-hoc monitor 
    brand
  }
  \item{mon}{
    Monitor code (accelorometer brand): 0=undefined, 1=GENEA, 2=GENEActiv,
    3=Actigraph, 4=Axivity, 5=Movisense, 6=Verisense
  }
  \item{...}{
    Any input arguments needed for function \link{read.myacc.csv} if you
    are working with a non-standard csv formatted files.
  }
  
}
\value{
  Character object showing how decimals are separated
}

\examples{
  \dontrun{
    decn = g.dotorcomma(inputfile="C:/myfile.bin",dformat=1,mon=2)
  }
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{datadir2fnames}
\alias{datadir2fnames}
\title{
  Generates vector of file names out of datadir input
  argument
}
\description{
  Uses input argument datadir from \link{g.part1} and
  the output from \link{isfilelist} to generate vector of filenames
}
\usage{
  datadir2fnames(datadir,filelist)
}
\arguments{
  \item{datadir}{
    See \link{g.part1}
  }
  \item{filelist}{
    Produced by \link{isfilelist}
  }
}
\value{
  Character vector of filenames
}
\examples{
  \dontrun{
  datadir2fnames(datadir = "C:/mydatafolder",filelist=TRUE)
  }
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.getbout}
\alias{g.getbout}
\title{
  function to calculate bouts from vector of binary classes
}
\description{
  To detect bouts of behaviour in time series. The function is used by \link{g.analyse}
}
\usage{
  g.getbout(x, boutduration, boutcriter = 0.8, ws3 = 5)
}
\arguments{
  \item{x}{vector of zeros and/or ones to be screened for bouts of ones
  }
  \item{boutduration}{duration of bout in epochs
  }
  \item{boutcriter}{ Minimum percentage of boutduration for which the epoch values
    are expected to meet the threshold criterium
  }
  \item{ws3}{epoch length in seconds, only needed for bout.metric =3, because
  it needs to measure how many epochs equal 1 minute breaks
}

}
\value{
  Vector with binary numbers indicator where bouts where detected
}
\examples{
  y = g.getbout(x=round(runif(1000, 0.4, 1)), boutduration = 120, boutcriter=0.9,
    ws3 = 5)
}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
  Jairo Hidalgo Migueles
}

\name{g.analyse.perday}
\alias{g.analyse.perday}
\title{
  Function supports \link{g.analyse}. Not intended for direct use by user.
}
\description{
  Generates day specific analyses and fills corresponding
  output matrix, \link{g.analyse}.
}
\usage{
g.analyse.perday(ndays, firstmidnighti, time, nfeatures, 
                midnightsi, metashort, averageday,
                doiglevels, nfulldays,lastmidnight, ws3, ws2, qcheck,
                fname, sensor.location, wdayname, tooshort, includedaycrit,
                doquan, quantiletype, doilevels, domvpa,
                mvpanames, wdaycode, ID,
                deviceSerialNumber, ExtFunColsi, myfun,
                params_247 = c(), params_phyact = c(),
                params_general = c(),
                ...)
}
\arguments{
  \item{ndays}{Number of days in file} 
  \item{firstmidnighti}{see \link{g.detecmidnight}}
  \item{time}{timestamp column from metalong converted to character}
  \item{nfeatures}{estimate of number of variables that need to be stored in
  the output matrix}
  \item{midnightsi}{see \link{g.detecmidnight}}
  \item{metashort}{see \link{g.impute}}
  \item{averageday}{As produced by \link{g.impute}}
  \item{doiglevels}{Boolean to indicate whether iglevels should be calculated}
  \item{nfulldays}{Number of days between the first and last midnight in the recording}
  \item{lastmidnight}{see \link{g.detecmidnight}}
  \item{ws3}{Epoch size in seconds}
  \item{ws2}{see \link{g.weardec}}
  \item{qcheck}{vector with zeros and ones for each epoch, respenting the
    quality check derived with g.impute}
  \item{fname}{RData filename produced by g.part1}
  \item{sensor.location}{as produced by \link{g.extractheadervars}} 
  \item{wdayname}{character with weekdayname}
  \item{tooshort}{0 (file not too short) or 1 (file too short)} 
  \item{includedaycrit}{see \link{g.analyse}} 
  \item{doquan}{Boolean whether quantile analysis should be done}
  \item{quantiletype}{see \link{g.analyse}} 
  \item{doilevels}{Boolean whether to generate ilevels, see \link{g.analyse}} 
  \item{domvpa}{Boolean whether to do mvpa analysis}
  \item{mvpanames}{Matrix with 6 columns and 1 row holding the names for the six 
    mvpa variables}
  \item{wdaycode}{Equal to M$wday as produced by \link{g.getmeta}}
  \item{ID}{Person Identification number, this can be numeric or character}
  \item{deviceSerialNumber}{As produced by \link{g.extractheadervars}}
  \item{ExtFunColsi}{column index of metashort where metric is stored}
  \item{myfun}{External function object to be applied to raw data, see \link{g.getmeta}.}
  \item{params_247}{
    See \link{GGIR}
  }
  \item{params_phyact}{
    See \link{GGIR}
  }
  \item{params_general}{
    See \link{GGIR}
  }
  \item{...}{
   Any argument used in the previous version of g.analyse.perday, which will now
   be used to overrule the arguments specified with the parameter objects.
  }
}
\value{
  \item{\code{daysummary}}{Summary per day for the file that was analysed}
  \item{\code{ds_names}}{Variable names in daysummary}
  \item{\code{windowsummary}}{Window summary, only used when
  selectdayfile is specified}
  \item{\code{ws_names}}{Variable names in windowsummary}
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.sibreport}
\alias{g.sibreport}
\title{
  Generate sustiained inactivty bouts report
}
\description{
  Generate sustained inactivity bout report. Function not intended
  for direct use by package user
}
\usage{
  g.sibreport(ts, ID, epochlength, logs_diaries=c(), desiredtz="")
}
\arguments{
  \item{ts}{
    Data frame with time series as created inside function \link{g.part5}
  }
  \item{ID}{
   Recording identifier (character or numeric)
  }
  \item{epochlength}{
    Numeric to indicate epoch length in seconds in the ts object
  }
  \item{logs_diaries}{
  Object produced by \link{g.loadlog} function
  }
  \item{desiredtz}{
    See \link{g.getmeta}
  }
}
\value{
  Dataframe with one row per sustained inactivity bout and corresponding 
  properties stored in the data.frame columns.
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.part6}
\alias{g.part6}
\title{
  Perform temporal pattern analyses
}
\description{
  This function aims to facilitate time-pattern analysis building on the labelled time
  series derived in GGIR part 5
}
\usage{
g.part6(datadir = c(), metadatadir = c(), f0 = c(), f1 = c(),
                   params_general = c(), params_phyact = c(), params_247 = c(),
                   params_cleaning = c(), verbose = TRUE, ...)
}
\arguments{
  \item{datadir}{
    Directory where the accelerometer files are stored, e.g. 
    "C:/mydata", or list of accelerometer filenames and directories, e.g. 
    c("C:/mydata/myfile1.bin", "C:/mydata/myfile2.bin").
  }
  \item{metadatadir}{
    Directory that holds a folder 'meta' and inside this a folder 'basic'
    which contains the milestone data produced by \link{g.part1}. The folderstructure
    is normally created by \link{g.part1} and \link{GGIR} will recognise what
    the value of metadatadir is.
  }
  \item{f0}{
    File index to start with (default = 1). Index refers to the filenames sorted
    in alphabetical order
  }
  \item{f1}{
    File index to finish with (defaults to number of files available, i.e., f1 = 0)
  }
  \item{params_general}{
   See details in \link{GGIR}.
  }
  \item{params_phyact}{
   See details in \link{GGIR}.
  }
  \item{params_247}{
   See details in \link{GGIR}.
  }
  \item{params_cleaning}{
   See details in \link{GGIR}.
  }
  \item{verbose}{
    See details in \link{GGIR}.
  }
  \item{...}{
    To ensure compatibility with R scripts written for older GGIR versions,
    the user can also provide parameters listed in the params_ objects as direct argument.
  }
}
\value{
  The function does not produce values but generates an RData file
  in the milestone subfolder ms6.out which incudes ... (TO BE COMPLETED).
  This dataframe is used in g.report.part6 to create
  reports. See package vignette
  paragraph (TO BE COMPLETED) for description of all the variables.
}
\examples{
  \dontrun{
    metadatadir = "C:/myfolder/meta"
    g.part6(metadatadir=metadatadir)
  }
}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{parametersVignette}
\alias{parametersVignette}
\title{
  Builds Section for Parameters Vignette
}
\description{
  Function extracts the documentation for a given GGIR argument as provided in the \link{GGIR} documentation
  and builds the structure for the Parameters Vignette.
  Function not designed for direct use by package user.
}
\usage{
  parametersVignette(params = "sleep")
}
\arguments{
  \item{params}{
    Character (default = "sleep"). 
    Name of the parameters object to build its corresponding section in the Parameters vignette.
  }
}
\value{
  Structure for the vignette subsection.
}
\keyword{internal}
\author{
  Jairo Hidalgo Migueles <jairo.hidalgo.migueles@gmail.com>
}

\name{check_log}
\alias{check_log}
\title{
  Function to revise format of user-provided logs
}
\description{
  Function to revise the format for missing values and dates in user-provided
  
}
\usage{
  check_log(log, dateformat, colid = 1, datecols = c(), 
            logPath, logtype)	
}
\arguments{
  \item{log}{
    Data frame with the log as read by \code{data.table::fread}.
  }
  \item{dateformat}{
    Character specifying the expected date format used in the log.
  }
  \item{colid}{
    Numeric with the column containing the file ID.
  }
  \item{datecols}{
    Numeric with the column/s containing the dates.
  }
  \item{logPath}{
    Character containing the full path to the activity log that is being checked.
  }
  \item{logtype}{
    Character which accepts "activity log" or "study dates log" at the moment.
    Only used for warning messages.
  }
}
\value{
  Data.frame containing the revised log.
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
  Jairo H Migueles <jairo@jhmigueles.com>
}

\name{g.part5_analyseSegment}
\alias{g.part5_analyseSegment}
\title{
  Analyses the time series per time segment for part 5
}
\description{
  Not intended for direct use by GGIR users, part of \link{g.part5}.
}
\usage{
  g.part5_analyseSegment(indexlog, timeList, levelList,
                                  segments,
                                  segments_names,
                                  dsummary, ds_names,
                                  params_general, params_output,
                                  params_sleep, params_247,
                                  params_phyact,
                                  sumSleep, sibDef,
                                  fullFilename,
                                  add_one_day_to_next_date,
                                  lightpeak_available,
                                  tail_expansion_log,
                                  foldernamei, sibreport = NULL)
}
\arguments{
  \item{indexlog}{
    List of objects related to indices of window, file, and segment that
    are passed on from \link{g.part5} to aid selecting time segments or keeping track
    of where in file the code is.
  }
  \item{timeList}{
    List of objects related to time series passed on from \link{g.part5}.
  }
  \item{levelList}{
    List of objects related to intensity levels passed on from \link{g.part5}. 
  }
  \item{segments}{
    List produced by \link{g.part5}
  }
  \item{segments_names}{
    Vector produced by \link{g.part5}
  }
  \item{dsummary}{
    Matrix to hold all daysummary (and segment summary)
  }
  \item{ds_names}{
    Character vector with column names of the dsummary matrix. The code collects
    these separately in this vector and assigns them at the end.
  }
  \item{params_general}{
    See \link{GGIR}
  }
  \item{params_output}{
    See \link{GGIR}
  }
  \item{params_sleep}{
    See \link{GGIR}
  }
  \item{params_247}{
    See \link{GGIR}
  }
  \item{params_phyact}{
    See \link{GGIR}
  }
  \item{sumSleep}{
    Section of data.frame produced by \link{g.part4} passed on from \link{g.part5}.
  }
  \item{sibDef}{
    Character to identify sib definition. 
  }
  \item{fullFilename}{
    Character with full filename being processed
  }
  \item{add_one_day_to_next_date}{
    Boolean to indicate whether one day should be added to next date
  }
  \item{lightpeak_available}{
    Boolean to indicate whether light peak is available
  }
  \item{tail_expansion_log}{
    Object generated in \link{g.part1} and passed on to \link{g.part5} when 
    argument \code{recordingEndSleepHour} is used.
  }
  \item{foldernamei}{
    Character with folder name in which the data file is stored.
  }
  \item{sibreport}{
    Sibreport object as passed on from \link{g.part5}
  }
}
\value{
  List with objects: indexlog, timeList, and the matrix with the prelimenary results
  and column names (dsummary and ds_names, see input arguments above)
}
\keyword{internal}

\name{g.fragmentation}
\alias{g.fragmentation}
\title{
Fragmentation metrics from time series.
}
\description{
The function is used by \link{g.part5} to derive time series 
fragmentation metrics. The function assumes that NA values and nonwear time is 
accounted for before the data enters the function.
}
\usage{
  g.fragmentation(frag.metrics = c("mean", "TP", "Gini", "power",
        "CoV", "NFragPM", "all"), LEVELS = c(), Lnames=c(), xmin=1, mode = "day")
}
\arguments{
   \item{frag.metrics}{
   Character with fragmentation metric to exract. Can be "mean", "TP", "Gini", 
   "power", or "CoV", "NFragPM", or all the above metrics with "all". See details.
  }
  \item{LEVELS}{
    Numeric vector of behavioural level classes derived with \link{identify_levels}
  }
  \item{Lnames}{
    Character vector with names of classes used in LEVELS, see details.
  }
  \item{xmin}{
    Numeric scalar to indicate the minimum recordable fragment length. In \link{g.part5}
    this is derived from the epoch length.
  }
  \item{mode}{
    Character to indicate whether input data is daytime ("day") or sleep period time ("spt").
  }
  
}
\value{
List with Character object showing how decimals are separated
  \item{TP_PA2IN}{Transition probability physical activity to inactivity}.
  \item{TP_IN2PA}{Transition probability physical inactivity to activity}
  \item{Nfrag_IN2LIPA}{Number of inacitivty fragments succeeded by LIPA
  (light physical activity)}
  \item{TP_IN2LIPA}{Transition probability physical inactivity to LIPA}
  \item{Nfrag_IN2MVPA}{Number of inacitivty fragments succeeded by MVPA
  (moderate or vigorous physical activity)}
  \item{TP_IN2MVPA}{Transition probability physical inactivity to MVPA}
  \item{Nfrag_MVPA}{Number of MVPA fragments}
  \item{Nfrag_LIPA}{Number of LIPA fragments}
  \item{mean_dur_MVPA}{mean MVPA fragment duration}
  \item{mean_dur_LIPA}{mean LIPA fragment duration}
  \item{Nfrag_IN}{Number of inactivity fragments}
  \item{Nfrag_PA}{Number of activity fragments}
  \item{mean_dur_IN}{mean duration inactivity fragments}
  \item{mean_dur_PA}{mean duration activity fragments}
  \item{Gini_dur_IN}{Gini index corresponding to inactivity fragment durations}
  \item{Gini_dur_PA}{Gini index corresponding to activity fragment durations}
  \item{CoV_dur_IN}{Coefficient of Variance corresponding to inactivity fragment durations}
  \item{CoV_dur_PA}{Coefficient of Variance corresponding to activity fragment durations}
  \item{alpha_dur_IN}{Alpha of the fitted power distribution through inactivity fragment durations}
  \item{alpha_dur_PA}{Alpha of the fitted power distribution through activity fragment durations}
  \item{x0.5_dur_IN}{x0.5 corresponding to alpha_dur_IN}
  \item{x0.5_dur_PA}{x0.5 corresponding to alpha_dur_PA}
  \item{W0.5_dur_IN}{W0.5 corresponding to alpha_dur_IN}
  \item{W0.5_dur_PA}{W0.5 corresponding to alpha_dur_PA}
  \item{NFragPM_IN}{Number of IN fragments per minutes in IN}
  \item{NFragPM_PA}{Number of PA fragments per minutes in PA}
  \item{SD_dur_IN}{Standard deviation in the duration of inactivity fragments}
  \item{SD_dur_PA}{Standard deviation in the duration of physical activity fragments}
}
\details{
  See package vignette for description of fragmentation metrics.
  In short, abbreviation "TP" refers to transition probality metrics,
  abbreviation "CoV" refers to Coefficient of Variance, and
  metric "NFragPM" refers to the Number of fragments per minute.
  
  Regarding the Lnames argument. The class names included in this are categorised
  as follows:
  \itemize{
    \item Inactive, if name includes the character strings "day_IN_unbt" or "day_IN_bts"
    \item LIPA, if name includes the character strings "day_LIG_unbt" or "day_LIG_bts"
    \item MVPA, if name includes the character strings "day_MOD_unbt", "day_VIG_unbt", 
    or "day_MVPA_bts"
  }
}
\examples{
\dontrun{
    x = c(6, 5, 6, 7, 6, 6, 7, 6, 6, 5, 6, 6, 6, 5, 7, 6, 6, 5, 5, 5, 6, 7, 6,
        6, 6, 6, 7, 6, 5, 5, 5, 5, 5, 6, 6, 6, 5, 6, 6, 6, 6, 6, 6, 6, 6, 6,
        7, 7, 6, 5, 6, 5, 6, 5, rep(12, 11), 5, 6, 6, 6, 5, 6, rep(9, 14), 6,
        5, 7, 7, 6, 7, 7, 7, 6, 6, 6, 5, 6, 5, 5, 5, 6, 5, 5, 5, 5, 5, 5, 5)
  Lnames = c("spt_sleep", "spt_wake_IN", "spt_wake_LIG", "spt_wake_MOD",
            "spt_wake_VIG", "day_IN_unbt", "day_LIG_unbt", "day_MOD_unbt",
            "day_VIG_unbt", "day_MVPA_bts_10", "day_IN_bts_30",
             "day_IN_bts_10_30", "day_LIG_bts_10")
  out = g.fragmentation(frag.metrics = "all",
                        LEVELS = x,
                        Lnames=Lnames)}
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.loadlog}
\alias{g.loadlog}
\title{
  Load and clean sleeplog information
}
\description{
  Loads sleeplog from a csv input file and applies sanity checks
  before storing the output in a dataframe
}
\usage{
  g.loadlog(loglocation = c(), coln1 = c(), colid = c(),
            sleeplogsep = ",", meta.sleep.folder = c(),
            desiredtz = "", sleepwindowType = c())
}
\arguments{
  \item{loglocation}{
    Location of the spreadsheet (csv) with sleep log information.
    See package vignette for explanation on expected format
  }
  \item{coln1}{
    Column number in the sleep log spreadsheet where the onset of the first
  night starts
  }
  \item{colid}{
      Column number in the sleep log spreadsheet in which the participant
  ID code is stored (default = 1)

  }
  \item{sleeplogsep}{
    Value used as sep argument for reading sleeplog csv file, usually "," or ";".
    This argument has been deprecated.
  }
  \item{meta.sleep.folder}{
    Path to part3 milestone data, only specify if sleeplog is in advanced format.
  }
  \item{desiredtz}{
    See \link{GGIR}
  }
  \item{sleepwindowType}{
    See \link{GGIR}
  }
}
\value{
  Data frame with sleeplog, which can be either in basic format or in advanced
  format. See GGIR package vignette for discussion of these two formats.
}
\examples{
\dontrun{
  sleeplog = g.loadlog(loglocation="C:/mysleeplog.csv",coln1=2,
  colid=1)
}
}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{updateBlocksize}
\alias{updateBlocksize}
\title{
  Update blocksize of data to be read depending on available memory.
}
\description{
  Function queries available memory to either lower or increase the blocksize
  used by function \link{g.readaccfile}
}
\usage{
updateBlocksize(blocksize, bsc_qc)
}
\arguments{
  \item{blocksize}{
    Number of filepages (binary data) or rows (other dataformats).
  }
  \item{bsc_qc}{
    Data.frame with columns time (timestamp from Sys.time) and size (memory size).
    This is used for housekeeping in \link{g.calibrate} and \link{g.getmeta}
  }
}
\keyword{internal}
\value{
  List with blocksize and bsc_qc, same format as input, although bsc_qc has one new
  row.
}

\name{g.part5.fixmissingnight}
\alias{g.part5.fixmissingnight}
\title{
  Fix missing night in part 4 output
}
\description{
  Not intended for direct use by GGIR users.
  If a night is missing in the part4 output then this function
  tries to fix as part of \link{g.part5}.
}
\usage{
  g.part5.fixmissingnight(summarysleep, sleeplog = c(), ID)
}
\arguments{
  \item{summarysleep}{
    Object produced by \link{g.part4}
  }
  \item{sleeplog}{
    Sleeplog object as passed on from \link{g.part5}
  }
  \item{ID}{
    ID of participant
  }
}
\value{
  Corrected summarysleep_tmp2 object.
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{markerButtonForRest}
\alias{markerButtonForRest}
\title{
  Improve the labelling of rest with marker button data
}
\description{
  Improve the labelling of rest with marker button data.
  Currently only functional for Actiwatch and Philips Healthband.
}
\usage{
  markerButtonForRest(sibreport, params_sleep, ts)
}
\arguments{
  \item{sibreport}{
    sibreport data.frame produced by \link{g.sibreport}
  }
  \item{params_sleep}{
    See \link{GGIR}
  }
  \item{ts}{
    Data.frame with time series object passed on from \link{g.part5_analyseSegment} and \link{g.part5}.
  }
}
\value{
  Updated sibreport
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.part2}
\alias{g.part2}
\title{
  function to analyse and summarize pre-processed output from \link{g.part1}
}
\description{
  Loads the output from \link{g.part1} and then applies \link{g.impute} and
  \link{g.analyse}, after which the output is converted to .RData-format
  which will be used by \link{GGIR} to generate reports.
  The variables in these reports are the same variables as described in
  \link{g.analyse}.
}
\usage{
  g.part2(datadir = c(), metadatadir = c(), f0 = c(), f1 = c(),
        myfun = c(), params_cleaning = c(), params_247 = c(),
        params_phyact = c(), params_output = c(), params_general = c(),
        verbose = TRUE, ...)
}
\arguments{
  \item{datadir}{
    Directory where the accelerometer files are stored, e.g.
    "C:/mydata", or list of accelerometer filenames and directories, e.g.
    c("C:/mydata/myfile1.bin", "C:/mydata/myfile2.bin").
  }
  \item{metadatadir}{
    Directory that holds a folder 'meta' and inside this a folder 'basic'
    which contains the milestone data produced by \link{g.part1}. The folderstructure
    is normally created by \link{g.part1} and \link{GGIR} will recognise what
    the value of metadatadir is.
  }
  \item{f0}{
    File index to start with (default = 1). Index refers to the filenames sorted
    in alphabetical order
  }
  \item{f1}{
    File index to finish with (defaults to number of files available, i.e., f1 = 0)
  }
  \item{myfun}{
    External function object to be applied to raw data.
    See details \link{applyExtFunction}.
  }
  \item{params_cleaning}{
   See details in \link{GGIR}.
  }
  \item{params_247}{
   See details in \link{GGIR}.
  }
  \item{params_phyact}{
   See details in \link{GGIR}.
 }
  \item{params_output}{
   See details in \link{GGIR}.
  }
  \item{params_general}{
   See details in \link{GGIR}.
  }
  \item{verbose}{
    See details in \link{GGIR}.
  }
  \item{...}{
    To ensure compatibility with R scripts written for older GGIR versions,
    the user can also provide parameters listed in the params_ objects as direct argument.
  }
}
\value{
  The function provides no values, it only ensures that other functions are called
  and that their output is stored in the folder structure as created with \link{g.part1}.
}
\details{
  GGIR comes with many processing parameters, which have been thematically grouped in
  parameter objects (R list). By running print(load_params()) you can
  see the default values of all the parameter objects. When g.part 2 is used via \link{GGIR}
  you have the option to specifiy a configuration file, which will overrule the default
  parameter values. Further, as user you can set parameter values as input argument to both g.part2
  and \link{GGIR}. Directly specified argument overrule the configuration file and default values.

   See the GGIR package vignette or the details section in \link{GGIR} for a more elaborate overview of parameter objects and their usage across GGIR.
}
\examples{
  \dontrun{
    metadatadir = "C:/myresults/output_mystudy"
    g.part2(metadatadir)
  }
}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}
\references{
  \itemize{
    \item van Hees VT, Gorzelniak L, Dean Leon EC, Eder M, Pias M, et al. (2013) Separating
    Movement and Gravity Components in an Acceleration Signal and Implications for the
    Assessment of Human Daily Physical Activity. PLoS ONE 8(4): e61691.
    doi:10.1371/journal.pone.0061691
    \item van Hees VT, Fang Z, Langford J, Assah F, Mohammad A, da Silva IC, Trenell MI,
    White T, Wareham NJ, Brage S. Auto-calibration of accelerometer data for
    free-living physical activity assessment using local gravity and temperature:
    an evaluation on four continents. J Appl Physiol (1985). 2014 Aug 7
  }
}

\name{g.part3}
\alias{g.part3}
\title{
  Detection of sustained inactivity periods as needed for sleep detection
  in g.part4.
}
\description{
  Function called by function GGIR. It estimates the sustained inactivity
  periods in each day, which are used as input for g.part4 which then labels
  them as nocturnal sleep or day time sustained inactivity periods.
  Typical users should work with function GGIR only.
}
\usage{
g.part3(metadatadir = c(), f0, f1, myfun = c(), 
  params_sleep = c(), params_metrics = c(), params_output = c(), 
  params_general = c(), verbose = TRUE,
  ...)
}
\arguments{
  \item{metadatadir}{
    Directory that holds a folder 'meta' and inside this a folder 'basic'
    which contains the milestone data produced by \link{g.part1}. The folderstructure
    is normally created by \link{g.part1} and \link{GGIR} will recognise what
    the value of metadatadir is.
  }
  \item{f0}{
    File index to start with (default = 1). Index refers to the filenames sorted
    in alphabetical order
  }
  \item{f1}{
    File index to finish with (defaults to number of files available, i.e., f1 = 0)
  }
  \item{myfun}{
    External function object to be applied to raw data.
    See details \link{applyExtFunction}.
  }
  \item{params_sleep}{
   See details in \link{GGIR}.
  }
  \item{params_metrics}{
   See details in \link{GGIR}.
  }
  \item{params_output}{
   See details in \link{GGIR}.
  }
  \item{params_general}{
   See details in \link{GGIR}.
  }
  \item{verbose}{
    See details in \link{GGIR}.
  }
  \item{...}{
    To ensure compatibility with R scripts written for older GGIR versions,
    the user can also provide parameters listed in the params_ objects as direct argument.
  }
}
\details{
  GGIR comes with many processing parameters, which have been thematically grouped in
  parameter objects (R list). By running print(load_params()) you can
  see the default values of all the parameter objects. When g.part 3 is used via \link{GGIR}
  you have the option to specifiy a configuration file, which will overrule the default
  parameter values. Further, as user you can set parameter values as input argument to both g.part3
  and \link{GGIR}. Directly specified argument overrule the configuration file and default values.
  
  See the GGIR package vignette or the details section in \link{GGIR} for a more elaborate overview of parameter objects and their usage across GGIR.
}
\value{
  The function provides no values, it only ensures that other functions
  are called and that their output is stored in .RData files.
  \cr
  \itemize{
    \item \code{night} nightnumber
    \item \code{definition} definition of sustained inactivity. For example,
    T10A5 refers to 10 minute window and a 5 degree angle (see paper for
    further explaination).
    \item \code{start.time.day} timestamp when the day started
    \item \code{nsib.periods} number of sustained inactivity bouts
    \item \code{tot.sib.dur.hrs} total duration of all sustained inactivity bouts
    \item \code{fraction.night.invalid} fraction of the night for which
    accelerometer data was invalid, e.g. monitor not worn
    \item \code{sib.period} number of sustained inactivity period
    \item \code{sib.onset.time} onset time of sustained inactivity period
    \item \code{sib.end.time} end time of sustained inactivity period
  }
}
\examples{
  \dontrun{
    metadatadir = "C:/myfolder/meta" # assumes that there is a subfolder in
    # metadatadir named 'basic' containing the output from g.part1
    g.part3(metadatadir=metadatadir, anglethreshold=5,
    timethreshold=5, overwrite=FALSE)
  }
}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}
\references{
  \itemize{
    \item van Hees VT, Sabia S, et al. (2015) A novel, open access method to
    assess sleep duration using a wrist-worn accelerometer, PLoS ONE, November 2015
    \item van Hees VT, Sabia S, et al. (2018) Estimating sleep parameters
    using an accelerometer without sleep diary. Scientific Reports.
  }
}

\name{g.sib.sum}
\alias{g.sib.sum}
\title{
  sustiained inactivty bouts detection
}
\description{
  Detects sustiained inactivty bouts. Function not intended
  for direct use by package user
}
\usage{
  g.sib.sum(SLE,M,ignorenonwear=TRUE,desiredtz="")
}
\arguments{
  \item{SLE}{
    Output from \link{g.sib.det}
  }
  \item{M}{
    Object produced by \link{g.getmeta}
  }
  \item{ignorenonwear}{
    If TRUE then ignore detected monitor non-wear periods to avoid
  confusion between monitor non-wear time and sustained inactivity
  (default = TRUE)
  }
  \item{desiredtz}{
    See \link{g.part3}
  }
}
\value{
  Dataframe with per night and per definition of sustained inactivity bouts
  the start and end time of each sustained inactivity bout
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.plot5}
\alias{g.plot5}
\title{
  Generate user-friendly visual report. The first part of the report summarizes important daily metrics in bar plot format.
  The second part of the report shows the raw data and annotations in 24-hr periods.
  Angle-z is shown with sleep annotations during the SPT (sleep period time) window.
  ENMO is shown with daytime inactivity and PA (physical activity) annotations in the lower
  section of each 24-hr plot. The PA annotations are based on a 10 minute bout metric and
  80% bout criteria. Moderate PA is a short window of time above threshold.mod that is part
  of a 10 minute bout of MVPA. Vigorous PA is a short window of time above threshold.vig that
  is part of a bout of MVPA. Light PA is a short window of time above threshold.lig that is
  part of a bout of light PA.
}
\description{
  Function called by \link{GGIR} to generate report. Not intended
  for direct use by user
}
\usage{
  g.plot5(metadatadir = c(), dofirstpage = FALSE, viewingwindow = 1,
  f0 = c(), f1 = c(), overwrite = FALSE, metric="ENMO",desiredtz = "",
  threshold.lig = 30, threshold.mod = 100, threshold.vig = 400, 
  visualreport_without_invalid = TRUE, includedaycrit = 0.66, includenightcrit = 0.66,
  verbose = TRUE)
}
\arguments{
  \item{metadatadir}{
    Directory that holds a folder 'meta' and inside this a folder 'basic'
    which contains the milestone data produced by \link{g.part1}. The folderstructure
    is normally created by \link{g.part1} and \link{GGIR} will recognise what
    the value of metadatadir is.
  }
  \item{dofirstpage}{
    Boolean to indicate whether a first page with historgrams summarizing the whole
    measurement should be added
  }
  \item{viewingwindow}{
    See \link{GGIR}
  }
  \item{f0}{
    File index to start with (default = 1). Index refers to the filenames sorted
    in alphabetical order
  }
  \item{f1}{
    File index to finish with (defaults to number of files available, i.e., f1 = 0)
  }
  \item{overwrite}{
    See \link{GGIR}
  }
  \item{metric}{
    Which one of the metrics do you want to consider to describe behaviour. The
    metric of interest need to be calculated in M (see \link{g.part1})
  }
  \item{desiredtz}{
    See \link{GGIR}
  }
  \item{threshold.lig}{
    See \link{GGIR}
  }
  \item{threshold.mod}{
    See \link{GGIR}
  }
  \item{threshold.vig}{
    See \link{GGIR}
  }
  \item{visualreport_without_invalid}{
    See \link{GGIR}
  }
  \item{includenightcrit}{
    See \link{GGIR}
  }
    \item{includedaycrit}{
    See \link{GGIR}
  }
  \item{verbose}{
    See \link{GGIR}
  }
}
\value{
  No values, this function only generates a plot
}

\examples{
  \dontrun{
    # generate plots for the first 10 files:
    g.plot5(metadatadir="C:/output_mystudy/meta/basic",dofirstpage=TRUE,
    viewingwindow = 1,f0=1,f1=10,overwrite=FALSE,desiredtz = "Europe/London",
    threshold.lig,threshold.mod,threshold.vig)
  }
}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
  Matthew R Patterson <mpatterson@shimmersensing.com>
}

\name{DFA}
\alias{DFA}
\title{Detrended Fluctuation Analysis}
\description{
  Detrended Fluctuation Analysis (DFA)
}
\usage{
  DFA(data, scale = 2^(1/8), box_size = 4, m = 1)
}
\arguments{
  \item{data}{
    Univariate time series (must be a vector or data frame)
  }
  \item{scale}{
    Specifies the ratio between successive box sizes (by default scale = 2^(1/8))
  }
  \item{box_size}{
    Vector of box sizes (must be used in conjunction with scale = "F")
  }
  \item{m}{
    An integer of the polynomial order for the detrending (by default m=1)
  }
}
\value{
  Estimated alpha is a real number between zero and two.
}
\details{
  The DFA fluctuation can be computed in a geometric scale or for different choices of boxes sizes.
}
\note{
  It is not possible estimating alpha for multiple time series at once.
}
\examples{
  # Estimate self-similarity of a very known time series available
  # on R base: the sunspot.year.
  # Then the spend time with each method is compared.
  \dontrun{
    dfa = DFA(sunspot.year)
  }
}
\references{
  C.-K. Peng, S.V. Buldyrev, S. Havlin, M. Simons, H.E. Stanley, A.L. Goldberger Phys. Rev. E, 49 (1994), p. 1685
  Mesquita, Victor & Filho, Florencio & Rodrigues, Paulo. (2020). Detection of crossover points in detrended fluctuation analysis: An application to EEG signals of patients with epilepsy. Bioinformatics. 10.1093/bioinformatics/btaa955.
}
\author{
  Ian Meneghel Danilevicz <ian.meneghel-danilevicz@inserm.fr>
  Victor Barreto Mesquita <victormesquita40@hotmail.com>
}

\name{g.readaccfile}
\alias{g.readaccfile}
\title{
  Generic functiont to read large blocks of accelerometer data
}
\description{
  The function is used by \link{g.getmeta} and \link{g.calibrate} to
  read large blocks of the accelerometer file, which are processed and
  then deleted from memory. This is needed for memory management.
}
\usage{
  g.readaccfile(filename, blocksize, blocknumber, filequality,
                           ws, PreviousEndPage = 1, inspectfileobject = c(),
                           PreviousLastValue = c(0,0,1), PreviousLastTime = NULL,
                           params_rawdata = c(), params_general = c(), header = NULL, ...)
}
\arguments{
  \item{filename}{
    filename
  }
  \item{blocksize}{
    Size of blocks (in file pages) to be read
  }
  \item{blocknumber}{
    Block number relative to start of file, starting with 1.
  }
  \item{filequality}{
    Single row dataframe with columns: filetooshort, filecorrupt,
    and filedoesnotholdday. All with the value TRUE or FALSE
  }
  \item{ws}{
    Larger windowsize for non-detection, see documentation \link{g.part2}
  }
  \item{PreviousEndPage}{
    Page number on which previous block ended (automatically assigned within
    g.getmeta and g.calibrate).
  }
  \item{inspectfileobject}{
    Output from the function \link{g.inspectfile}.
  }
  \item{PreviousLastValue}{
    Automatically identified last value in previous chunk of data read.
  }
    \item{PreviousLastTime}{
    Automatically identified last timestamp in previous chunk of data read.
  }
  \item{params_rawdata}{
    See \link{g.part1}
  }
  \item{params_general}{
    See \link{g.part1}
  }
  \item{header}{
    Header information that was extracted the previous time this file was read,
    to be re-used instead of being extracted again.
  }
  \item{...}{
    Any input arguments needed for function \link{read.myacc.csv} if you
    are working with a non-standard csv formatted files. Furter,
    any argument used in the previous version of g.readaccfile, which will now
    be used to overrule the arguments specified with the parameter objects.
  }
}
\value{
  \itemize{
    \item \code{P} Block object extracted from file with format specific to
    accelerometer brand
    \item \code{filequality} Same as in function arguments
    \item \code{isLastBlock} Boolean indicating whether this was the last block to read
    \item \code{endpage} Page number on which blocked ends, this will be
    used as input for argument PreviousEndPage when reading the next block.
  }
}
\examples{
  \dontrun{
    filequality = data.frame(filetooshort = FALSE, filecorrupt = FALSE,
    filedoesnotholdday = FALSE)
    output = g.readaccfile(filename = "C:/myfile.bin", 
    blocksize = 20000, blocknumber = 1,
    selectdaysfile = c(), filequality = filequality,
    dayborder = 0, PreviousEndPage = c()) 
  }
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.report.part5_dictionary}
\alias{g.report.part5_dictionary}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
  Generate data dictionary for reports from milestone data produced by \link{g.part5}
}
\description{
  Creates a data dictionary with the definitions of the outcomes exported in the
  reports from milestone data produced by \link{g.part5}. Not intended
  for direct use by package user.
}
\usage{
g.report.part5_dictionary(metadatadir, params_output)
}
\arguments{
  \item{metadatadir}{
    Directory that holds a folder 'meta' and inside this a folder 'basic'
    which contains the milestone data produced by \link{g.part1}. The folderstructure
    is normally created by \link{g.part1} and \link{GGIR} will recognise what
    the value of metadatadir is.
  }
  \item{params_output}{
    Parameters object, see \link{GGIR}
  }
}
\value{
  Function does not produce data, but only writes data dictionaries for the
  reports in csv format
  
  The following files are stored in the root of the results folder:
  part5_dictionary_daysummary_*
  part5_dictionary_personsummary_*
}

\author{
  Vincent T van Hees <v.vanhees@accelting.com>
  Jairo Hidalgo Migueles <jairo@jhmigueles.com>
}

\name{create_test_sleeplog_csv}
\alias{create_test_sleeplog_csv}
\title{
  Creates csv sleeplog file for testing purposes
}
\description{
  Creates sleeplog file in the format as expected by g.part4
with dummy data (23:00 onset, 07:00 waking time for every night).
}
\usage{
  create_test_sleeplog_csv(Nnights = 7, storagelocation = c(),
                            advanced = FALSE, sep = ",", begin_date = "2016/06/25",
                            type = "sleeplog")
}
\arguments{
  \item{Nnights}{
    Number of nights (minimum is 1)
  }
  \item{storagelocation}{
    Location where the test file named testfile.csv will be stored
    If no value is provided then the function uses the current 
    working directory
  }
  \item{advanced}{
    Boolean to indicate whether to create an advanced sleeplog that also includes
    logs of nap times and nonwear
  }
  \item{sep}{
    Character to indicate the column separator of the csv file.
  }
  \item{begin_date}{
    Character to indicate first date (in format "2016/06/25") to be used
    in the advanced sleeplog format. Ignored when generated basic sleeplog format.
  }
  \item{type}{
    Character to indicate which type of advanced sleeplog to create,
    either "sleeplog" with "wakeup" and "onset" columns, "bedlog" with 
    "bedstart" and "bedend" columns, or "both" with both.
  }
}
\value{
 The function does not produce any output values. Only the file is
 stored
}
  
\examples{
  \dontrun{
    create_test_sleeplog_csv()
  }
}

\name{POSIXtime2iso8601}
\alias{POSIXtime2iso8601}
\title{
Convert POSIX to iso8601 timestamp
}
\description{
To avoid ambiguities when sharing and comparing timestamps. All timestamps
are expressed in iso8601 format: https://en.wikipedia.org/wiki/ISO_8601
}
\usage{
POSIXtime2iso8601(x,tz)	
}
\arguments{
  \item{x}{
Vector of timestamps in POSIX format
}
  \item{tz}{
 Timezone of data collection, e.g. "Europe/London".
 See
 https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
 for full list
}
}
  
\examples{
\dontrun{
x ="2017-05-07 13:15:17 CEST"
tz = "Europe/Amsterdam"
x_converted = POSIXtime2iso8601(x,tz)
}
}
\author{
Vincent T van Hees <v.vanhees@accelting.com>
}

\name{filterNonwearNight}
\alias{filterNonwearNight}
\title{
  Filters short lasting nonwear during the night
}
\description{
  Filters short lasting nonwear during the night. Not intended for
  direct use by GGIR user.
}
\usage{
  filterNonwearNight(r1, metalong, qwindowImp, desiredtz, params_cleaning, ws2)
}
\arguments{
  \item{r1}{
    Object r1 as created inside function \link{g.weardec} which is a vector
    of zeros and ones derived from metalong where 1 indicates nonwear and 0 wear.
  }
  \item{metalong}{
    Object metalong created in \link{g.part1}
  }
  \item{qwindowImp}{
    See \link{g.impute}
  }
  \item{desiredtz}{
    See \link{GGIR}
  }
  \item{params_cleaning}{
    Parameters object with cleaning paramete, see \link{GGIR}.
  }
  \item{ws2}{
    Object ws2 which equals the second value of vector parameter \code{windowsizes}
  }
}
\value{
  r1 object with short lasting nonwear during the night removed
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{GGIR-package}
\alias{GGIR-package}
\docType{package}
\title{
  A package to process multi-day raw accelerometer data
}
\description{
  Disclaimer: If you are a new GGIR user then please see
  \href{https://wadpac.github.io/GGIR/}{the GGIR github-pages}
   for a narrative overview of GGIR.\cr
  \cr
  This document is primarily aimed at documenting the functions and their input arguments.\cr
  \cr
  Please note that there is google discussion group for this package (link below).\cr
  \cr
  You can thank us for sharing the code in this package and for developing
  it as a generic purpose tool by citing the package name and by
  citing the supporting publications (e.g. Migueles et al. 2019) in your publications.
}
\details{
  \tabular{ll}{
  Package: \tab GGIR\cr
  Type: \tab Package\cr
  Version: \tab 3.3-6\cr
  Date: \tab 2026-04-21\cr
  License: \tab Apache License (== 2.0)\cr
  Discussion group: \tab https://groups.google.com/forum/#!forum/rpackageggir\cr
  }
}
\examples{
  \dontrun{
    #inspect file:
    I = g.inspectfile(datafile)

    #autocalibration:
    C = g.calibrate(datafile)

    #get meta-data:
    M = g.getmeta(datafile)
  }
  data(data.getmeta)
  data(data.inspectfile)
  data(data.calibrate)

  #impute meta-data:
  IMP = g.impute(M = data.getmeta, I = data.inspectfile)
  #analyse and produce summary:
  A = g.analyse(I = data.inspectfile, C = data.calibrate, M = data.getmeta, IMP, ID = "01wk0")
  #plot data
  g.plot(IMP, M = data.getmeta, I = data.inspectfile, durplot=4)
}
\author{
  \itemize{
    \item Vincent T van Hees <v.vanhees@accelting.com> main creator and developer
    \item Zhou Fang developed calibration algorithm used in function \link{g.calibrate}
    \item Joe Heywood helped develop the functionality to process specific recording days
    \item Severine Sabia, Mathilde Chen, and Manasa Yerramalla extensively tested and provided feedback on various functions
    \item Joan Capdevila Pujol helped to improve various functions
    \item Jairo H Migueles <jairohm@ugr.es> helped to improve various functions
    \item Matthew R Patterson helped with enhancing the visual report.
    \item Lena Kushleyeva helped fix bug in sleep detection.
    \item Taren Sanders helped tidy up the parallel processing functionality
  }
}
\references{
  \itemize{
    \item Migueles JH, Rowlands AV, et al. GGIR: A Research Community-Driven Open Source
    R Package for Generating Physical Activity and Sleep Outcomes From Multi-Day Raw
    Accelerometer Data. Journal for the Measurement of Physical Behaviour. 2(3) 2019.
    doi:10.1123/jmpb.2018-0063.
    \item van Hees VT, Gorzelniak L, Dean Leon EC, Eder M, Pias M, et al. (2013) Separating
    Movement and Gravity Components in an Acceleration Signal and Implications for the
    Assessment of Human Daily Physical Activity. PLoS ONE 8(4): e61691.
    doi:10.1371/journal.pone.0061691
    \item van Hees VT, Fang Z, Langford J, Assah F, Mohammad A, da Silva IC, Trenell MI,
    White T, Wareham NJ, Brage S. Auto-calibration of accelerometer data for
    free-living physical activity assessment using local gravity and temperature:
    an evaluation on four continents. J Appl Physiol (1985). 2014 Aug 7
    \item van Hees VT, Sabia S, et al. (2015) A novel, open access method to
    assess sleep duration using a wrist-worn accelerometer, PLoS ONE, November 2015
  }
}

\name{visualReport}
\alias{visualReport}
\title{
  Generate visualisation of time series produced by part 5.
}
\description{
  Function called by \link{GGIR}. Not intended
  for direct use by user
}
\usage{
  visualReport(metadatadir = c(), 
              f0 = c(), f1 = c(), 
              verbose = TRUE,
              part6_threshold_combi = NULL, GGIRversion = NULL,
              params_sleep = NULL,
              params_output = NULL,
              params_general = NULL)
}
\arguments{
  \item{metadatadir}{
    Directory that holds a folder 'meta' and inside this a folder 'basic'
    which contains the milestone data produced by \link{g.part1}. The folderstructure
    is normally created by \link{g.part1} and \link{GGIR} will recognise what
    the value of metadatadir is.
  }
  \item{f0}{
    File index to start with (default = 1). Index refers to the filenames sorted
    in alphabetical order
  }
  \item{f1}{
    File index to finish with (defaults to number of files available, i.e., f1 = 0)
  }
  \item{verbose}{
    See \link{GGIR}
  }
  \item{part6_threshold_combi}{
    See \link{GGIR}
  }
  \item{GGIRversion}{
    Character with GGIR version number
  }
  \item{params_sleep}{
    See \link{GGIR}
  }
  \item{params_output}{
    See \link{GGIR}
  }
  \item{params_general}{
    See \link{GGIR}
  }
}
\value{
  No values, this function only generates a plot
}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{ismovisens}
\alias{ismovisens}
\title{
  Checks whether the files to process are collected with movisens accelerometers.
}
\description{
  Checks whether the files in the datadir folder are files collected
  with movisens accelerometers. Note that movisens data are stored in one folder per recording that includes multiple bin-files
  (instead of one file per recording as usual in other accelerometer brands). Therefore, datadir indicates
  the directory where all the recording folders are stored, then, GGIR reads the pertinent bin files from
  every folder.
}
\usage{
  ismovisens(data)	
}
\arguments{
  \item{data}{
    Full path to the recording folder (with the bin files inside) or the 
    datadir (where all the recording folders are stored).
  }
}
\value{
  Boolean whether it is a movisens file (TRUE) or not (FALSE)
}
\keyword{internal}
\examples{
  \dontrun{
    is.mv = ismovisens(data)
  }
}

\name{GGIR}
\alias{GGIR}
\title{
  Shell function for analysing an accelerometer dataset.
}
\description{
  This function is designed to help users operate all steps of the
  analysis. It helps to generate and structure milestone data,
  and produces user-friendly reports. The function acts as a shell with
  calls to \link{g.part1}, \link{g.part2}, \link{g.part3},
  \link{g.part4} and \link{g.part5}.
}
\usage{
GGIR(mode = 1:5,
     datadir = c(),
     outputdir = c(),
     studyname = c(),
     f0 = 1, f1 = 0,
     do.report = c(2, 4, 5, 6),
     configfile = c(),
     myfun = c(),
     verbose = TRUE, ...)
}
\arguments{
  \item{mode}{
    Numeric (default = 1:5).
    Specify which of the five parts need to be run, e.g., mode = 1 makes
    that \link{g.part1} is run; or mode = 1:5 makes
    that the whole GGIR pipeline is run, from \link{g.part1} to \link{g.part5}.
    Optionally mode can also include the number 6 to tell GGIR to run \link{g.part6}
    which is currently under development.
  }
  \item{datadir}{
    Character (default = c()).
    Directory where the accelerometer files are stored, e.g.,
    \"C:/mydata\", or list of accelerometer filenames and directories, e.g.
    c(\"C:/mydata/myfile1.bin\", \"C:/mydata/myfile2.bin\").
  }
  \item{outputdir}{
    Character (default = c()).
    Directory where the output needs to be stored. Note that this
    function will attempt to create folders in this directory and uses
    those folder to keep output.
  }
  \item{studyname}{
    Character (default = c()).
    If the datadir is a folder, then the study will be given the name of the
    data directory. If datadir is a list of filenames then the studyname as specified
    by this input argument will be used as name for the study.
  }
  \item{f0}{
    Numeric (default = 1).
    File index to start with (default = 1). Index refers to the filenames sorted
    in alphabetical order.
  }
  \item{f1}{
    Numeric (default = 0).
    File index to finish with (defaults to number of files available).
  }
  \item{do.report}{
    Numeric (default = c(2, 4, 5, 6)).
    For which parts to generate a summary spreadsheet: 2, 4, 5, and/or 6. Default is c(2, 4, 5, 6).
    A report will be generated based on the available milestone data. When creating
    milestone data with multiple machines it is advisable to turn the report
    generation off when generating the milestone data, value = c(),
    and then to merge the milestone data and turn report generation back
    on while setting overwrite to FALSE.
  }
  \item{configfile}{
    Character (default = c()).
    Configuration file previously generated by function GGIR.
    See also https://wadpac.github.io/GGIR/articles/chapter2_Pipeline.html#configuration-file-
  }
  \item{myfun}{
    List (default = c()).
    External function object to be applied to raw data. See package vignette for detailed tutorial with
    examples on how to use the function embedding:
    https://wadpac.github.io/GGIR/articles/ExternalFunction.html
  }
  \item{verbose}{
    Boolean (default = TRUE).
    to indicate whether console message should be printed. Note that warnings and error are
    always printed and can be suppressed with suppressWarning() or suppressMessages().
  }
  \item{...}{
    Any of the parameters used GGIR. Given the large number of parameters used in GGIR
    we have grouped them in objects that start with \"params_\". These are documented in the
    details section. You cannot provide these objects as argument to function GGIR, but
    you can provide the parameters inside them as input to function GGIR.
  }
}
\value{
  The function provides no values, it only ensures that other functions are called
  and that their output is stored. Further, a configuration file is stored containing
  all the argument values used to facilitate reproducibility.
}
\details{
  Once you have used function GGIR and the output directory (outputdir) will be filled
  with milestone data and results. Function GGIR stores all the explicitely
  entered argument values and default values for the argument that are not explicitely
  provided in a csv-file named config.csv stored in the root of the output folder.
  The config.csv file is accepted as input to GGIR with argument \code{configfile}
  to replace the specification of all the arguments, except \code{datadir} and \code{outputdir}.

  The practical value of this is that it eases the replication of analysis, because
  instead of having to share you R script, sharing your config.csv file will be
  sufficient. Further, the config.csv file contribute to the reproducibility
  of your data analysis.

  Note: When combining a configuration file with explicitely provided argument
  values, the explicitely provided argument values will overrule
  the argument values in the configuration file. If a parameter is neither provided
  via the configuration file nor as input then GGIR uses its default paramter values which
  can be inspected with command \code{print(load_params())}, and if you are specifically
  interested in a certain subgroup of parameters, e.g., physical activity, then you
  can do \code{print(load_params()$params_phyact)}. These defaults are part of the GGIR
  code and cannot be changed by the user.

  The parameters that can be used in GGIR are:

    \subsection{params_general}{
    A list of parameters used across all GGIR parts that do not fall in any of the other
    categories.
    \describe{

      \item{overwrite}{
        Boolean (default = FALSE).
        Do you want to overwrite analysis for which milestone
        data exists? If overwrite = FALSE, then milestone data from a previous analysis will
        be used if available and visual reports will not be created again.}

      \item{dayborder}{
        Numeric (default = 0).
        Hour at which days start and end (dayborder = 4 would mean 4 am).}

      \item{do.parallel}{
        Boolean (default = TRUE).
        Whether to use multi-core processing (only works if at least 4 CPU cores are available).}

      \item{maxNcores}{
        Numeric (default = NULL).
        Maximum number of cores to use when argument do.parallel is set to true.
        GGIR by default uses either the maximum number of available cores or the number of files to
        process (whichever is lower), but this argument allows you to set a lower maximum.}

      \item{acc.metric}{
        Character (default = \"ENMO\").
        Which one of the acceleration metrics do you want to use for all acceleration magnitude
        analyses in GGIR part 5 and the visual report? For example: \"ENMO\", \"LFENMO\", \"MAD\",
        \"NeishabouriCount_y\", or \"NeishabouriCount_vm\". Only one acceleration metric can be specified
        and the selected metric needs to have been calculated in part 1 (see \link{g.part1})
        via arguments such as \code{do.enmo = TRUE} or \code{do.mad = TRUE}.}

      \item{part5_agg2_60seconds}{
        Boolean (default = FALSE).
        Whether to use aggregate epochs to 60 seconds as part of the GGIR 
        \link{g.part5} analysis. Aggregation is doen by averaging. 
        Note that when working with count metrics such as Neishabouri counts this
        means that the threshold can stay the same as in part 2, because again the
        threshold is expressed relative to the original epoch size, even if averaged
        per minute. For example if we want to use a cut-point 100 count per minute
        then we specify \code{mvpathreshold = 100 * (5/60)} as well as
        `threshold.mod = \code{100 * (5/60)} regardless of whether we set 
        part5_agg2_60seconds to TRUE or FALSE.}

      \item{print.filename}{
        Boolean (default = FALSE).
        Whether to print the filename before analysing it (in case do.parallel = FALSE).
        Printing the filename can be useful to investigate
        problems (e.g., to verify that which file is being read).}

      \item{desiredtz}{
        Character (default = \"\", i.e., system timezone).
        Timezone in which device was configured and experiments took place.
        If experiments took place in a different timezone, then use this
        argument for the timezone in which the experiments took place and
        argument \code{configtz} to specify where the device was configured.
        Use the \"TZ identifier\" as specified at
        https://en.wikipedia.org/wiki/Zone.tab to set desiredtz,
        e.g., \"Europe/London\".
        As of GGIR 3.2-12, \code{desiredtz} is only configurable in GGIR 
        part 1 and passed on to the following parts via de milestone data files 
        stored by each part. This means that if you would run
        \code{GGIR(mode = 2, desiredtz = "Europe/Berlin")} GGIR would ignore the
        timezone specification and rely on the timezone specified when part 1 was 
        processed specific to each recording. As a result, it is now possible
        to process part 1 separately with different timezones, then pool the milestone
        data into one output folder, and finally process all files jointly with the 
        subsequent GGIR parts (2-6) which will automatically accounts for each 
        individual's file timezone.
      }

      \item{configtz}{
        Character (default = \"\", i.e., system timezone).
        At the moment only functional for GENEActiv .bin, AX3 cwa, ActiGraph .gt3x,
        and ad-hoc csv file format.
        Timezone in which the accelerometer was configured. Only use this argument
        if the timezone of configuration and timezone in which recording took
        place are different. Use the \"TZ identifier\" as specified at
        https://en.wikipedia.org/wiki/Zone.tab to set configtz, 
        e.g., \"Europe/London\".}

      \item{sensor.location}{
        Character (default = \"wrist\").
        To indicate sensor location, default is wrist. If it is hip, the HDCZA algorithm for sleep detection
        also requires longitudinal axis of sensor to be between -45 and +45 degrees.}

      \item{windowsizes}{
        Numeric vector, three values
        (default = c(5, 900, 3600)).
        To indicate the lengths of the windows as in c(window1, window2, window3):
        window1 is the short epoch length in seconds, by default 5, and this is the time 
        window over which acceleration and angle metrics are calculated;
        window2 is the long epoch length in seconds for which non-wear and signal clipping 
        are defined, default 900 (expected to be a multitude of 60 seconds);
        window3 is the window length of data used for non-wear detection and by default 3600 seconds.
        So, when window3 is larger than window2 we use overlapping windows,
        while if window2 equals window3 non-wear periods are assessed by non-overlapping windows.}

      \item{idloc}{
        Numeric (default = 1).
        If idloc = 1 the code assumes that ID number is stored in the obvious header field. Note that for ActiGraph data
        the ID is never stored in the file header.
        For value set to 2, 5, 6, and 7, GGIR looks at the filename and extracts the character string preceding the first
        occurance of a \"_\" (idloc = 2), \" \" (space, idloc = 5), \".\" (dot, idloc = 6),
        and \"-\" (idloc = 7), respectively.
        You may have noticed that idloc 3 and 4 are skipped, they were used for one study in 2012,
        and not actively maintained anymore, but because it is legacy code not omitted.}

      \item{expand_tail_max_hours}{
        Numeric (default = NULL).
        This parameter has been replaced by \code{recordingEndSleepHour}.}
        
      \item{recordingEndSleepHour}{
        Numeric (default = NULL).
        Time (in hours) at which the recording should end (or later) to expand the
        \link{g.part1} output with synthetic data to trigger sleep detection for last night.
        Using argument \code{recordingEndSleepHour} implies the assumption that the
        participant fell asleep at or before the end of the recording if the recording
        ended at or after \code{recordingEndSleepHour} hour of the last day.
        This assumption may not always hold true and should be used with caution.
        The synthetic data for metashort entails: timestamps continuing
        regularly, zeros for acceleration metrics other than EN, one for EN.
        Angle columns are created in a way that it triggers the sleep detection using
        the equation: \code{round(sin((1:length_expansion) / (900/epochsize))) * 15}.
        To keep track of the tail expansion \link{g.part1} stores the length of the expansion in
        the RData files, which is then passed via \link{g.part2}, \link{g.part3},
        and \link{g.part4} to \link{g.part5}. In \link{g.part5} the tail expansion
        size is included as an additional variable in the csv-reports.
        In the \link{g.part4} csv-report the last night is omitted, because we know
        that sleep estimates from the last night will not be trustworthy. Similarly,
        in the \link{g.part5} output columns related to the sleep assessment will
        be omitted for the last window to avoid biasing the averages. Further,
        the synthetic data are also ignored in the visualizations and time series
        output to avoid biased output.}

      \item{dataFormat}{
        Character (default = \"raw\").
        To indicate what the format is of the data in datadir.
        Alternatives: ukbiobank_csv, actiwatch_csv, actiwatch_awd, 
        actigraph_csv, sensewear_xls, phb_xlsx, and fitbit_json which correspond 
        to epoch level data files from, respecitively, UK Biobank in csv format, 
        Actiwatch in csv format, Actiwatch in awd format, ActiGraph csv format, 
        Sensewear in xls format (also works with xlsx), Philips Health Band in 
        xlsx format, and Fitbit in json format. Here, the assumed epoch size for
        UK Biobank csvdata is 5 seconds. 
        The epoch size for the other non-raw data formats is 
        flexible, but make sure that you set first value of argument 
        \code{windowsizes} accordingly. Also when working with 
        non-raw data formats specify argument \code{extEpochData_timeformat} as
        documented below. For ukbiobank_csv nonwear is a column in the data itself,
        for actiwatch_csv, actiwatch_awd, actigraph_csv, and sensewear_xls non-wear 
        is detected as 60 minute rolling zeros. The length of this window can be 
        modified with the third value of argument \code{windowsizes} expressed in
        seconds.}
        
       \item{maxRecordingInterval}{
        Numeric (default = NULL).
        To indicate the maximum gap in hours between repeated measurements with the same
        ID for the recordings to be appended. So, the assumption is that the
        ID can be matched, make sure argument \code{idloc} is set correctly.
        If argument \code{maxRecordingInterval} is set to NULL (default) recordings
        are not appended. If recordings overlap then GGIR will use the data from
        the latest recording. If recordings are separated then the timegap between
        the recordings is filled with data points that resemble monitor not worn.
        The maximum value of maxFile gap is 504 (21 days). Only recordings from the
        same accelerometer brand are appended. This functionality is applied
        after all other aspect of GGIR part 1 are completed for all input files.
        The part 2 csv report
        will show number of appended recordings, sampling rate for each, time overlap or gap 
        and the names of the filenames of the respective recording.}
      
      \item{recording_split_times}{
        Character (default = NULL).
        To indicate path to a csv file with a column that has ID in the column name
        that holds the participant IDs, followed by columns with either dates or
        full timestamps on which the recording should be split. Here, timestamp
        format is specified with parameter \code{recording_split_timeformat}.
        The names of these columns are used to refer to each resulting segment.
        To prevent extremely long file name GGIR ignores spaces and does not consider
        more than 10 characters which are forced to lower case. If the recording
        starts before or ends after the first and last split time then the start and/or
        end of the recording is also used as split time and referred to as
        startrec and endrec, respectively. Files are split after all other aspect of
        GGIR part 1 are completed for all input files.
      }
      
      \item{recording_split_timeformat}{
        Character (default = \"\%d/\%m/\%Y \%H:\%M\")
        To indicate timestamp format as used in file as specified with 
        \code{recording_split_times}.
        For guidance on how to specify time formats in R see:
        https://wadpac.github.io/GGIR/articles/DateTimeFormatsInR.html.
      }
      
      \item{recording_split_ignore_edges}{
        Boolean (default = FALSE)
        To indicate whether the recording time before and after the time range defined by
        \code{recording_split_times} should be considered as segments.
      }
      
      \item{recording_split_overlap}{
        Numeric (default = 0).
        Number of hours to use as buffer when splitting recording. A possitive number
        means that the split recordings overlap, a nevative number indicates a gap.
      }
      
      \item{extEpochData_timeformat}{
        Character (default = \"\%d-\%m-\%Y \%H:\%M:\%S\").
        To specify the time format used in the external epoch level data when
        argument \code{dataFormat} is set to \"actiwatch_csv\", \"actiwatch_awd\", 
        \"actigraph_csv\" or \"sensewear_xls\". For example \"\%Y-\%m-\%d \%I:\%M:\%S \%p\" for 
        \"2023-07-11 01:24:01 PM\" or \"\%m/\%d/\%Y \%H:\%M:\%S\" \"2023-07-11 13:24:01\".
        For guidance on how to specify time formats in R see:
        https://wadpac.github.io/GGIR/articles/DateTimeFormatsInR.html.
      }
      
      \item{use_trycatch_serial}{
        Boolean (default = FALSE).
        If set to TRUE and when using the non-default 
        \code{do.parallel = FALSE}, the 
        core processing function in each GGIR part is wrapped for each file in 
        a \code{tryCatch} block. This allows GGIR to continue processing remaining 
        files even if one or more files cause an error (e.g., due to file corruption). 
        The error messages for failed files are logged and printed in the console 
        at the end of the processing. Note: these error messages are less
        detailed compared to when processing data with \code{use_trycatch_serial == FALSE}.
        So, to investigate an error message it may be advisable to set 
        \code{use_trycatch_serial == FALSE}. Further,
        it should be noted that \code{tryCatch} blocks are always used when processing 
        \code{do.parallel = TRUE}. The functionally described here is envisioned
        to be primarily useful when it not possible to process accelerometer files
        in parallel and there are many files. For example, when GGIR is expected 
        to run only on a single thread in a computing cluster. In that scenario it
        would actually be better to investigate whether GGIR can be given more
        computing threads as GGIR is designed to coordinate the processing of files across 
        processing threads. A second scenario where \code{use_trycatch_serial == TRUE}
        is useful is
        when processing files that are too large to be processed even if parameter 
        \code{chunksize} is set to a minimal value. In that scenario you may want to 
        set \code{use_trycatch_serial == TRUE} and \code{do.parallel = FALSE}.
        Note that under normal conditions we expect GGIR to
        not generate errors. If an error occurs, always investigate
        and report the error to the community to help make GGIR better. GGIR is free
        software and depends on you as community.
      }
    }
    }

    \subsection{params_rawdata}{
    A list of parameters used to related to reading and pre-processing
    raw data, excluding parameters related to metrics as those are in
    the params_metrics object.
    \describe{

      \item{backup.cal.coef}{
        Character (default = \"retrieve\").
        Option to use backed-up calibration coefficient instead of
        deriving the calibration coefficients when analysing the same file twice.
        Argument backup.cal.coef has two usecase. Use case 1: If the auto-calibration
        fails then the user has the option to provide back-up
        calibration coefficients via this argument. The value of the argument needs to
        be the name and directory of a csv-spreadsheet with the following column names
        and subsequent values: \"filename\" with the names of accelerometer files on which
        the calibration coefficients need to be applied in case auto-calibration fails;
        \"scale.x\", \"scale.y\", and \"scale.z\" with the scaling coefficients; \"offset.x\",
        \"offset.y\", and \"offset.z\" with the offset coefficients, and;
        \"temperature.offset.x\", \"temperature.offset.y\", and \"temperature.offset.z\"
        with the temperature offset coefficients. This can be useful for analysing
        short lasting laboratory experiments with insufficient sphere data to perform
        the auto-calibration, but for which calibration coefficients can be derived
        in an alternative way.  It is the users responsibility to compile the
        csv-spreadsheet. Instead of building this file the user can also
        Use case 2: The user wants to avoid performing the auto-calibration repeatedly
        on the same file. If backup.cal.coef value is set to \"retrieve\" (default) then
        GGIR will look out for the \"data_quality_report.csv\" file in the outputfolder
        QC, which holds the previously generated calibration coefficients. If you
        do not want this happen, then deleted the data_quality_report.csv from the
        QC folder or set it to value \"redo\".}

      \item{minimumFileSizeMB}{
        Numeric (default = 2).
        Minimum File size in MB required to enter processing.
        This argument can help to avoid having short uninformative 
        files to enter the analyses. Given that a typical accelerometer 
        collects several MBs per hour, the default setting should only skip 
        the very tiny files.}

      \item{do.cal}{
        Boolean (default = TRUE).
        Whether to apply auto-calibration or not by \link{g.calibrate}.
        Recommended setting is TRUE.}

      \item{imputeTimegaps}{
        Boolean (default = TRUE).
        To indicate whether timegaps larger than 1 sample should be imputed.
        Currently only used for .gt3x data and ActiGraph .csv format, where timegaps
        can be expected as a result of Actigraph's idle sleep.mode configuration.}

      \item{spherecrit}{
        Numeric (default = 0.3).
        The minimum required acceleration value (in g) on both sides of 0 g
        for each axis. Used to judge whether the sphere is sufficiently populated}

      \item{minloadcrit}{
        Numeric (default = 168).
        The minimum number of hours the code needs to read for the
        autocalibration procedure to be effective (only sensitive to
        multitudes of 12 hrs, other values will be ceiled).
        After loading these hours only extra data is loaded if 
        calibration error has not been reduced to under 0.01 g.}

      \item{printsummary}{
        Boolean (default = FALSE).
        If TRUE will print a summary of the calibration procedure in
        the console when done.}

      \item{chunksize}{
        Numeric (default = 1).
        Value to specify the size of chunks to be
        loaded as a fraction of an approximately 12 hour period for auto-calibration
        procedure and as fraction of 24 hour period for the metric calculation, e.g., 
        0.5 equals 6 and 12 hour chunks, respectively. 
        For machines with less than 4Gb of RAM memory or with < 2GB memory per process
        when using \code{do.parallel = TRUE} a value below 1 is recommended.
        The value is constrained by GGIR to not be lower than 0.05. Please note that
        setting 0.05 will not produce output when 3rd value of parameter windowsizes
        is 3600.}

      \item{dynrange}{
        Numeric (default = NULL).
        Provide dynamic range of 8 gravity.}

      \item{interpolationType}{
        Integer (default = 1).
        To indicate type of interpolation to be used
        when resampling time series (mainly relevant for Axivity and Parmay Matrix sensors),
        1=linear, 2=nearest neighbour.}

      \item{rmc.file}{
        Character (default = NULL).
        Filename of file to be read if it is in the working directory,
        or full path to the file otherwise.
      }

      \item{rmc.nrow}{
        Numeric (default = NULL).
        Number of rows to read, same as nrow argument in \link[utils]{read.csv} and nrows in \link[data.table]{fread}.
        The whole file is read by default (i.e., rmc.nrow = Inf).}

      \item{rmc.skip}{
        Numeric (default = 0).
        Number of rows to skip, same as skip argument in \link[utils]{read.csv} and in \link[data.table]{fread}.}

      \item{rmc.dec}{
        Character (default = \".\").
        Decimal used for numbers, same as dec argument in \link[utils]{read.csv} and in \link[data.table]{fread}.}

      \item{rmc.firstrow.acc}{
        Numeric (default = NULL).
        First row (number) of the acceleration data.}

      \item{rmc.firstrow.header}{
        Numeric (default = NULL).
        First row (number) of the header. Leave blank if the file does not have a header.}

      \item{rmc.header.length}{
        Numeric (default = NULL).
        If file has header, specify header length (number of rows).}

      \item{rmc.col.acc}{
        Numeric, three values
        (default = c(1, 2, 3)).
        Vector with three column (numbers) in which the acceleration signals
        are stored.}

      \item{rmc.col.temp}{
        Numeric (default = NULL).
        Scalar with column (number) in which the temperature is stored.
        Leave in default setting if no temperature is available. The temperature
        will be used by \link{g.calibrate}.}

      \item{rmc.col.time}{
        Numeric (default = NULL).
        Scalar with column (number) in which the timestamps are stored.
        Leave in default setting if timestamps are not stored.}

      \item{rmc.unit.acc}{
        Character (default = \"g\").
        Character with unit of acceleration values: \"g\", \"mg\", or \"bit\".}

      \item{rmc.unit.temp}{
        Character (default = \"C\").
        Character with unit of temperature values: (K)elvin, (C)elsius, or (F)ahrenheit.}

      \item{rmc.unit.time}{
        Character (default = \"POSIX\").
        Character with unit of timestamps: \"POSIX\", \"UNIXsec\" (seconds since origin, 
        see argument \code{rmc.origin}), \"UNIXmsec\" (same as UNIXsec but in milliseconds),
        \"character\", or \"ActivPAL\" (exotic timestamp format only used in the ActivPAL
        activity monitor).}

      \item{rmc.format.time}{
        Character (default = \"\%Y-\%m-\%d \%H:\%M:\%OS\").
        Character  giving a date-time format as used by \link[base]{strptime}.
        Only used for rmc.unit.time: character and POSIX..
        For guidance on how to specify time formats in R see:
        https://wadpac.github.io/GGIR/articles/DateTimeFormatsInR.html
        }

      \item{rmc.bitrate}{
        Numeric (default = NULL).
        If unit of acceleration is a bit then provide bit rate, e.g., 12 bit.}

      \item{rmc.dynamic_range}{
        Numeric or character (default = NULL).
        If unit of acceleration is a bit then provide dynamic range deviation
        in g from zero, e.g., +/-6g would mean this argument needs to be 6. If you give this
        argument a character value the code will search the file header for elements with
        a name equal to the character value and use the corresponding numeric value
        next to it as dynamic range.}

      \item{rmc.unsignedbit}{
        Boolean (default = TRUE).
        If unsignedbit = TRUE means that bits are only positive numbers.
        if unsignedbit = FALSE then bits are both positive and negative.}

      \item{rmc.origin}{
        Character (default = \"1970-01-01\").
        Origin of time when unit of time is UNIXsec, e.g., 1970-1-1.}

      \item{rmc.desiredtz}{
        Character (default = NULL).
        Timezone in which experiments took place. This argument is scheduled to
        be deprecated and is now used to overwrite \code{desiredtz} if not provided.}

      \item{rmc.configtz}{
        Character (default = NULL).
        Timezone in which device was configured. This argument is scheduled to
        be deprecated and is now used to overwrite \code{configtz} if not provided.}

      \item{rmc.sf}{
        Numeric (default = NULL).
        Sample rate in Hertz, if this is stored in the file header then that will be used
        instead (see argument \code{rmc.headername.sf}).}

      \item{rmc.headername.sf}{
        Character (default = NULL).
        If file has a header: Row name under which the sample frequency can be found.}

      \item{rmc.headername.sn}{
        Character (default = NULL).
        If file has a header: Row name under which the serial number can be found.}

      \item{rmc.headername.recordingid}{
        Character (default = NULL).
        If file has a header: Row name under which the recording ID can be found.}

      \item{rmc.header.structure}{
        Character (default = NULL).
        Used to split the header name from the header value, e.g., \":\" or \" \".}

        \item{rmc.check4timegaps}{
        Boolean (default = FALSE).
        To indicate whether gaps in time should be imputed with zeros.
        Some sensing equipment provides accelerometer with gaps in time. The rest of
        GGIR is not designed for this, by setting this argument to TRUE the gaps
        in time will be filled with zeros.}

      \item{rmc.col.wear}{
        Numeric (default = NULL).
        If external wear detection outcome is stored as part of the data then this can be used by GGIR.
        This argument specifies the column in which the wear detection (Boolean) is stored.}

      \item{rmc.doresample}{
        Boolean (default = FALSE).
        To indicate whether to resample the data based on the available timestamps and extracted
        sample rate from the file header.}

      \item{rmc.noise}{
        Numeric (default = 13).
        Noise level of acceleration signal in m\emph{g}-units, used when working 
        ad-hoc .csv data formats
        using \link{read.myacc.csv}. The \link{read.myacc.csv} does not take rmc.noise as argument,
        but when interacting with \link{GGIR} or \link{g.part1} rmc.noise is used.}
        
      \item{rmc.scalefactor.acc}{
        Numeric value (default 1) to scale the acceleration signals via multiplication.
        For example, if data is provided in m/s2 then by setting this to 1/9.81
        we would derive gravitational units.
      }

      \item{frequency_tol}{
        Number (default = 0.1) as passed on to readAxivity from the GGIRread package.
        Represents the frequency tolerance as fraction between 0 and 1. When the relative bias
        per data block is larger than this fraction then the data block will be imputed
        by lack of movement with gravitational oriationed guessed from most recent
        valid data block. Only applicable to Axivity .cwa data.
      }
      \item{nonwear_range_threshold}{
        Numeric (default 150) used to define maximum value range per axis for non-wear
        detection, used in combination with brand specific standard deviation per
        axis.
      }
      
    }
  }

  \subsection{params_metrics}{
  A list of parameters used to specify the signal metrics that need to be extract in GGIR \link{g.part1}.
  \describe{

      \item{do.anglex}{
        Boolean (default = FALSE).
        If TRUE, calculates the angle of the X axis relative to the horizontal:
        \deqn{angleX = (\tan{^{-1}\frac{acc_{rollmedian(x)}}{(acc_{rollmedian(y)})^2 +
        (acc_{rollmedian(z)})^2}}) * 180/\pi}}

      \item{do.angley}{
        Boolean (default = FALSE).
        If TRUE, calculates the angle of the Y axis relative to the horizontal:
        \deqn{angleY = (\tan{^{-1}\frac{acc_{rollmedian(y)}}{(acc_{rollmedian(x)})^2 +
        (acc_{rollmedian(z)})^2}}) * 180/\pi}}

      \item{do.anglez}{
        Boolean (default = TRUE).
        If TRUE, calculates the angle of the Z axis relative to the horizontal:
        \deqn{angleZ = (\tan{^{-1}\frac{acc_{rollmedian(z)}}{(acc_{rollmedian(x)})^2 +
        (acc_{rollmedian(y)})^2}}) * 180/\pi}}

      \item{do.zcx}{
        Boolean (default = FALSE).
        If TRUE, calculates metric zero-crossing count for x-axis. For computation specifics
        see source code of function \link{g.applymetrics}}

      \item{do.zcy}{
        Boolean (default = FALSE).
        If TRUE, calculates metric zero-crossing count for y-axis. For computation specifics
        see source code of function \link{g.applymetrics}}

      \item{do.zcz}{
        Boolean (default = FALSE).
        If TRUE, calculates metric zero-crossing count for z-axis. For computation
        specifics see source code of function \link{g.applymetrics}}

      \item{do.enmo}{
        Boolean (default = TRUE).
        If TRUE, calculates the metric: \deqn{ENMO = \sqrt{acc_x^2 + acc_y^2 + acc_z^2} - 1}
        (if ENMO < 0, then ENMO = 0).}

      \item{do.lfenmo}{
        Boolean (default = FALSE).
        If TRUE, calculates the metric \code{ENMO} over the low-pass filtered accelerations
        (for computation specifics see source code of function \link{g.applymetrics}).
        The filter bound is defined by the parameter \code{hb}.}

      \item{do.en}{
        Boolean (default = FALSE).
        If TRUE, calculates the Euclidean Norm of the raw accelerations:
        \deqn{EN = \sqrt{acc_x^2 + acc_y^2 + acc_z^2}}}

      \item{do.mad}{
        Boolean (default = FALSE).
        If TRUE, calculates the Mean Amplitude Deviation:
        \deqn{MAD = \frac{1}{n}\Sigma|r_i - \overline{r}|}}

      \item{do.enmoa}{
        Boolean (default = FALSE).
        If TRUE, calculates the metric:
        \deqn{ENMOa = \sqrt{acc_x^2 + acc_y^2 + acc_z^2} - 1} (if ENMOa < 0, then ENMOa = |ENMOa|).}

      \item{do.roll_med_acc_x}{
        Boolean (default = FALSE).
        If TRUE, calculates the metric. For computation
        specifics see source code of function \link{g.applymetrics}.}

      \item{do.roll_med_acc_y}{
        Boolean (default = FALSE).
        If TRUE, calculates the metric. For computation
        specifics see source code of function \link{g.applymetrics}.}

      \item{do.roll_med_acc_z}{
        Boolean (default = FALSE).
        If TRUE, calculates the metric. For computation
        specifics see source code of function \link{g.applymetrics}.}

      \item{do.dev_roll_med_acc_x}{
        Boolean (default = FALSE).
        If TRUE, calculates the metric. For computation
        specifics see source code of function \link{g.applymetrics}.}

      \item{do.dev_roll_med_acc_y}{
        Boolean (default = FALSE).
        If TRUE, calculates the metric. For computation
        specifics see source code of function \link{g.applymetrics}.}

      \item{do.dev_roll_med_acc_z}{
        Boolean (default = FALSE).
        If TRUE, calculates the metric. For computation
        specifics see source code of function \link{g.applymetrics}.}

      \item{do.bfen}{
        Boolean (default = FALSE).
        If TRUE, calculates the metric. For computation
        specifics see source code of function \link{g.applymetrics}.}

      \item{do.hfen}{
        Boolean (default = FALSE).
        If TRUE, calculates the metric. For computation
        specifics see source code of function \link{g.applymetrics}.}

      \item{do.hfenplus}{
        Boolean (default = FALSE).
        If TRUE, calculates the metric. For computation
        specifics see source code of function \link{g.applymetrics}.}

      \item{do.lfen}{
        Boolean (default = FALSE).
        If TRUE, calculates the metric. For computation
        specifics see source code of function \link{g.applymetrics}.}

      \item{do.lfx}{
        Boolean (default = FALSE).
        If TRUE, calculates the metric. For computation
        specifics see source code of function \link{g.applymetrics}.}

      \item{do.lfy}{
        Boolean (default = FALSE).
        If TRUE, calculates the metric. For computation
        specifics see source code of function \link{g.applymetrics}.}

      \item{do.lfz}{
        Boolean (default = FALSE).
        If TRUE, calculates the metric. For computation
        specifics see source code of function \link{g.applymetrics}.}

      \item{do.hfx}{
        Boolean (default = FALSE).
        If TRUE, calculates the metric. For computation
        specifics see source code of function \link{g.applymetrics}.}

      \item{do.hfy}{
        Boolean (default = FALSE).
        If TRUE, calculates the metric. For computation
        specifics see source code of function \link{g.applymetrics}.}

      \item{do.hfz}{
        Boolean (default = FALSE).
        If TRUE, calculates the metric. For computation
        specifics see source code of function \link{g.applymetrics}.}

      \item{do.bfx}{
        Boolean (default = FALSE).
        If TRUE, calculates the metric. For computation
        specifics see source code of function \link{g.applymetrics}.}

      \item{do.bfy}{
        Boolean (default = FALSE).
        If TRUE, calculates the metric. For computation
        specifics see source code of function \link{g.applymetrics}.}

      \item{do.bfz}{
        Boolean (default = FALSE).
        If TRUE, calculates the metric. For computation
        specifics see source code of function \link{g.applymetrics}.}

      \item{do.brondcounts}{
        Boolean (default = FALSE).
        this option has been deprecated (October 2022) due to issues with the 
        activityCounts package that we used as a dependency.
        If TRUE, calculated the metric via R package activityCounts.
        We called them BrondCounts because there are large number of activity counts in
        the physical activity and sleep research field. By calling them _brondcounts_
        we clarify that these are the counts proposed by Jan Br&#248;nd and implemented
        in R by Ruben Brondeel. The _brondcounts_ are intended to be an imitation of
        the counts produced by one of the closed source ActiLife software by ActiGraph.}

      \item{do.neishabouricounts}{
        Boolean (default = FALSE).
        If TRUE, calculates the metric via R package actilifecounts, which is an 
        implementation of the algorithm used in the closed-source software ActiLife 
        by ActiGraph (methods published in doi: 10.1038/s41598-022-16003-x). We use 
        the name of the first author (instead of ActiLifeCounts) of the paper and 
        call them NeishabouriCount under the uncertainty that ActiLife will implement 
        this same algorithm over time. To use the Neishabouri counts for the physical 
        activity intensity classification in part 5 (i.e., metric over the threshold.lig, 
        threshold.mod, and threshold.vig would be applied), the \code{acc.metric} 
        argument needs to be set as one of the following: \"NeishabouriCount_x\",
        \"NeishabouriCount_y\", \"NeishabouriCount_z\", \"NeishabouriCount_vm\" to use the 
        counts in the x-, y-, z-axis or vector magnitude, respectively.}

      \item{lb}{
        Numeric (default = 0.2).
        Lower boundary of the frequency filter (in Hertz) as used in the filter-based metrics.}

      \item{hb}{
        Numeric (default = 15).
        Higher boundary of the frequency filter (in Hertz) as used in the filter-based metrics.}

      \item{n}{
        Numeric (default = n).
        Order of the frequency filter as used in the filter-based metrics.}

      \item{zc.lb}{
        Numeric (default = 0.25).
        Used for zero-crossing counts only. Lower boundary of cut-off frequency filter.}

      \item{zc.hb}{
        Numeric (default = 3).
        Used for zero-crossing counts only. Higher boundary of cut-off frequencies in filter.}

      \item{zc.sb}{
        Numeric (default = 0.01).
        Stop band used for calculation of zero crossing counts. Value is the acceleration threshold
        in g units below which  acceleration will be rounded to zero.}

      \item{zc.order}{
        Numeric (default = 2).
        Used for zero-crossing counts only. Order of frequency filter.}

      \item{zc.scale}{
        Numeric (default = 1)
        Used for zero-crossing counts only. Scaling factor to be applied after
        counts are calculated (GGIR part 3).}

      \item{actilife_LFE}{
        Boolean (default = FALSE).
        If TRUE, calculates the NeishabouriCount metric with the low-frequency extension filter
        as proposed in the closed source ActiLife software by ActiGraph. Only applicable to
        the metric NeishabouriCount.}
    }
  }

  \subsection{params_cleaning}{
    A list of parameters used across all GGIR parts releated to masking or
    imputing data, abbreviated as \"cleaning\".
    \describe{

      \item{do.imp}{
        Boolean (default = TRUE).
        Whether to impute missing values (e.g., suspected of monitor non-wear 
        or clippling) or not by \link{g.impute} in GGIR \link{g.part2}.
        Recommended setting is TRUE.}

      \item{TimeSegments2ZeroFile}{
        Character (default = NULL).
        Takes path to a csv file that has columns \"windowstart\" and \"windowend\"
        to refer to the start and end time of a time windows in format
        \"2024-10-12 20:00:00\", and \"filename\" of the GGIR
        milestone data file without the \"meta_\" segment of the name. GGIR part 2
        uses this to set all acceleration values to zero and the non-wear classification 
        to zero (meaning sensor worn). Motivation: When the 
        accelerometer is not worn during the night GGIR automatically labels them
        as invalid, while the user may like to treat them as zero movement.
        Disclaimer: This functionality was developed in 2019. With hindsight it
        is not generic enough and in need for revision. Please contact GGIR 
        maintainers if you would like us to invest time in improving this 
        functionality.}

      \item{data_cleaning_file}{
        Character (default = NULL).
        Optional path to a csv file you create that holds four columns:
        ID, day_part5, relyonguider_part4, and night_part4. ID should hold the
        participant ID. Columns day_part5 and night_part4 allow you to specify which
        day(s) and night(s) need to be excluded from \link{g.part5} and 
        \link{g.part4}, respectively. When including multiple day(s)/night(s) 
        create a new line for each day/night.
        So, this will be done regardless of whether the rest of GGIR thinks 
        those day(s)/night(s) are valid. Column relyonguider_part4 allows you 
        to specify for which nights \link{g.part4} should fully rely on the
        guider. See also package vignette.}

      \item{excludefirstlast.part5}{
        Boolean (default = FALSE).
        If TRUE then the first and last window (waking-waking,
        midnight-midnight, or sleep onset-onset) are ignored in \link{g.part5}.}

      \item{excludefirstlast}{
        Boolean (default = FALSE).
        If TRUE then the first and last night of the measurement are ignored for
        the sleep assessment in \link{g.part4}.}

      \item{excludefirst.part4}{
        Boolean (default = FALSE).
        If TRUE then the first night of the measurement are ignored for the sleep 
        assessment in \link{g.part4}.}

      \item{excludelast.part4}{
        Boolean (default = FALSE).
        If TRUE then the last night of the measurement are ignored for the sleep 
        assessment in \link{g.part4}.}

      \item{includenightcrit}{
        Numeric (default = 16).
        Minimum number of valid hours per night (24 hour window between noon and noon),
        used for sleep assessment in \link{g.part4}.}

      \item{minimum_MM_length.part5}{
        Numeric (default = 23).
        Minimum length in hours of a MM day to be included in the cleaned \link{g.part5} results.}
        
      \item{study_dates_file}{
        Character (default = c()).
        Full path to csv file containing the first and last date of the expected 
        wear period for every study participant (dates are provided per individual).
        Expected format of the activity diary is: First column headers followed 
        by one row per recording. There should be three columns: first column is 
        recording ID, which needs to match with the ID GGIR extracts from the 
        accelerometer file; second column should contain the first date of the 
        study; and third column the last date of the study. Date columns should be
        by default in format \"23-04-2017\", or in the date format  specified by 
        argument \code{study_dates_dateformat} (below). If not specified (default), 
        then GGIR would use the first and last day of the recording as beginning 
        and end of the study. Note that these dates are used on top of the 
        \code{data_masking_strategy} selected.}
        
      \item{study_dates_dateformat}{
        Character (default = \"\%d-\%m-\%Y\").
        To specify the date format used in the \code{study_dates_file} as used 
        by \link[base]{strptime}.
        For guidance on how to specify time formats in R see:
        https://wadpac.github.io/GGIR/articles/DateTimeFormatsInR.html
        }
        
      \item{strategy}{
        Deprecated and replaced by \code{data_masking_strategy}. If \code{strategy} 
        is specified then its value is passed on and used for \code{data_masking_strategy}.}
        
      \item{data_masking_strategy}{
        Numeric (default = 1).
        How to deal with knowledge about study protocol.
        data_masking_strategy = 1 means select data based on \code{hrs.del.start} and \code{hrs.del.end}.
        data_masking_strategy = 2 makes that only the data between the first
        midnight and the last midnight is used.
        data_masking_strategy = 3 selects the most active X days in the file where X is
        specified by argument \code{ndayswindow}, where the days are a series 
        of 24-h blocks starting any time in the day (X hours at the beginning and end
        of this period can be deleted with arguments \code{hrs.del.start} and \code{hrs.del.end})
        data_masking_strategy = 4 to only use the data after the first midnight.
        data_masking_strategy = 5 is similar to \code{data_masking_strategy = 3}, but it selects X complete
        calendar days where X is specified by argument \code{ndayswindow} 
        (X hours at the beginning and end of this period can be deleted with 
        arguments \code{hrs.del.start} and \code{hrs.del.end}).}

      \item{hrs.del.start}{
        Numeric (default = 0).
        How many HOURS after start of experiment did wearing
        of monitor start? Used in GGIR \link{g.part2} when \code{data_masking_strategy = 1}.}

      \item{hrs.del.end}{
        Numeric (default = 0).
        How many HOURS before the end of the experiment did
        wearing of monitor definitely end? Used in GGIR \link{g.part2} when \code{data_masking_strategy = 1}.}

      \item{maxdur}{
        Numeric (default = 0).
        How many DAYS after start of experiment did experiment
        definitely stop? (set to zero if unknown).}

      \item{ndayswindow}{
        Numeric (default = 7).
        If \code{data_masking_strategy} is set to 3 or 5, then this is the size of 
        the window as a
        number of days. For data_masking_strategy 3 value can be fractional, e.g. 7.5,
        while for data_masking_strategy 5 it needs to be an integer.}

      \item{includedaycrit.part5}{
        Numeric (default = 2/3).
        Inclusion criteria used in part 5 for number of valid hours during the 
        waking hours of a day,
        when value is smaller than or equal to 1 used as fraction of waking hours, 
        when value above 1 used as absolute number of valid hours required.
        Do not confuse this argument with argument \code{includedaycrit} which is
        only used in GGIR part 2 and applies to the entire day.}
        
      \item{includenightcrit.part5}{
        Numeric (default = 0).
        Inclusion criteria used in part 5 for number of valid hours during the 
        sleep period hours of a day (the night),
        when value is smaller than or equal to 1 used as fraction of sleep period hours, 
        when value above 1 used as absolute number of valid hours required.
        Do not confuse this argument with argument \code{includenightcrit} which is
        only used in GGIR part 4 and applies to the entire 24 hour window from 
        noon to noon or 6pm to 6pm.}

      \item{segmentWEARcrit.part5}{
        Numeric (default = 0.5).
        Fraction of \code{qwindow} segment expected to be valid in part 5, where 
        0.3 indicates that at least 30 percent of the time should be valid.}
        
      \item{segmentDAYSPTcrit.part5}{
        Numeric vector or length 2 (default = c(0.9, 0)).
        Inclusion criteria for the proportion of the segment that should be 
        classified as day (awake) and spt (sleep period time) to be considered 
        valid. If you are interested in comparing time spent in behaviour then it
        is better to set one of the two numbers to 0, and the other defines the 
        proportion of the segment that should be classified as day or spt, respectively.
        The default setting would focus on waking hour
        segments and includes all segments that overlap for at least 90 percent 
        with waking hours. In order to shift focus to the SPT you could use
        c(0, 0.9) which ensures that all segments that overlap for at least
        90 percent with the SPT are included.
        Setting both to zero would be problematic when comparing time spent in
        behaviours between days or individuals: A complete segment
        would be averaged with an incomplete segments (someone going to bed or waking up
        in the middle of a segment) by which it is no longer clear whether the person
        is less active or sleeps more during that segment. Similarly it is not
        clear whether the person has more wakefulness during SPT for a segment or
        woke up or went to bed during the segment.
      }
      \item{includedaycrit}{
        Numeric (default = 16).
        Minimum required number of valid hours in calendar day specific to analysis 
        in part 2. If you specify two values as in c(16, 16) then the first value 
        will be used in part 2 and the second value will be used in part 5 and 
        applied as a criterion on the full part 5 window. Note that this is then 
        applied in addition to parameter includedaycrit.part5 which only looks 
        at valid data during waking hours.}

      \item{max_calendar_days}{
        Numeric (default = 0).
        The maximum number of calendar days to include (set to zero if unknown).}

      \item{nonWearEdgeCorrection}{
        Boolean (default = TRUE).
        If TRUE then the non-wear detection around the edges of the recording (first
        and last 3 hours) are corrected following description in vanHees2013 as
        has been the default since then. This functionality is advisable when working 
        with sleep clinic or exercise lab data typically lasting less than a day.}

      \item{nonwear_approach}{
        Character (default = \"2023\").
        Whether to use the traditional version of the non-wear detection algorithm 
        (nonwear_approach = \"2013\") or the new version (nonwear_approach = \"2023\"). 
        The 2013 version would use the longsize window (windowsizes[3], one hour 
        as default) to check the conditions for nonwear identification and would 
        flag as nonwear the mediumsize window (windowsizes[2], 15 min as default) 
        in the middle. The 2023 version differs in which it would flag as nonwear 
        the full longsize window. For the 2013 method the longsize window is centered 
        in the centre of the mediumsize window, while in the 2023 method the longsizewindow 
        is aligned with its left edge to the left edge of the mediumsize window.
      }
      
      \item{includecrit.part6}{
        Numeric (default = c(2/3, 2/3))
        Vector of two with the minimum fraction of valid data required for day
        and spt time, respectively. This criteria is only used for circadian 
        rhythm analysis.}

      \item{nonwearFiltermaxHours}{
        Numeric (default = NULL).
        If not NULL, ignore detected nonwear periods that last shorter than
        nonwearFiltermaxHours during a window as defined by
        parameter \code{NonwearFilterWindow}. If NonwearFilterWindow is not 
        provided (set to NULL as is the default) check
        whether \code{qwindow} is defined as activity diary file and whether it has 
        timestamps that allow for defining a time in bed, SPT or lightsoff window.
        For this the code looks for column names in the diary with the terms inbed,
        sleeponset, or lightsout to define the start of the window.
        Further, to define the end of the window the code looks for columns outbed,
        wakeup, or lightsoff. If yes, it uses the largest of the windows that can
        be defined with these timestamps. If not, use midnight-6am
        as fall back setting. You may realise that these columns are the same as the
        advanced format for the sleelog accepted by GGIR part 4. So, if you already
        have an advanced format sleeplog for part 4 then you can provide this as value
        for the qwindow parameter.  If you want to
        use \code{qwindow} to guide the nonwear filtering as described in here but 
        do not want GGIR to use it for day segment analysis in part 2 or 5 then 
        make sure the diary filename includes the word \"onlyfilter\" or \"filteronly\"
        to tell GGIR to only use qwindow for nonwear filtering.
        The main purpose of this functionality is to offer
        the option to ignore short lasting nonwear episodes during the night
        when there is suspicion that these are falsely detect, e.g. in individuals
        with extended motionless sleep periods caused by medication.}
      
      \item{nonwearFilterWindow}{
        Numeric (default = NULL).
        Vector of length 2 to specify the start and end hour of the night to be used
        for the functionality as described above for NonwearFiltermaxHours.}
    }
  }

  \subsection{params_phyact}{
    A list of parameters releated to physical activity as used in GGIR \link{g.part2} and GGIR \link{g.part5}.
    \describe{

      \item{mvpathreshold}{
        Numeric (default = NULL).
        Legacy parameter, if not provided GGIR uses the value of \code{threshold.mod} 
        for this.
        Acceleration threshold for MVPA estimation in GGIR \link{g.part2}.
        This can be a single number or an vector of numbers,
        e.g., \code{mvpathreshold = c(100, 120)}.
        In the latter case the code will estimate MVPA separately for each threshold.
        If this variable is left blank, e.g., \code{mvpathreshold = c()}, then
        MVPA is not estimated.}

      \item{mvpadur}{
        Numeric (default = c(1, 5, 10)).
        The bout duration(s) for which MVPA will be calculated. Only used in GGIR \link{g.part2}.}

      \item{boutcriter}{
        Numeric (default = NULL).
        Legacy parameter, if not provided GGIR uses the value of \code{boutcriter.mvpa} 
        for this.
        A number between 0 and 1, it defines what fraction of a bout needs to be above the
        mvpathreshold, only used in GGIR \link{g.part2}.}

      \item{threshold.lig}{
        Numeric (default = 40).
        In \link{g.part5}: Threshold for light physical activity to
        separate inactivity from light. Value can be one number or an vector of multiple
        numbers, e.g., \code{threshold.lig =c(30,40)}. If multiple numbers are entered then
        analysis will be repeated for each combination of threshold values. Threshold is
        applied to the first metric in the milestone data, so if you have only specified
        \code{do.enmo = TRUE} then it will be applied to ENMO.}

      \item{threshold.mod}{
        Numeric (default = 100).
        In \link{g.part5}: Threshold for moderate physical activity
        to separate light from moderate. Value can be one number or an vector of
        multiple numbers, e.g., \code{threshold.mod = c(100, 120)}.
        If multiple numbers are entered then analysis will be repeated for each
        combination of threshold values. Threshold is applied to the first metric in the
        milestone data, so if you have only specified \code{do.enmo = TRUE}
        then it will be applied to ENMO.}

      \item{threshold.vig}{
        Numeric (default = 400).
        In \link{g.part5}: Threshold for vigorous physical activity
        to separate moderate from vigorous. Value can be one number or an vector of
        multiple numbers, e.g., \code{threshold.vig =c(400,500)}. If multiple numbers are
        entered then analysis will be repeated for each combination of threshold values.
        Threshold is applied to the first metric in the milestone data, so if you
        have only specified \code{do.enmo = TRUE} then it will be applied to ENMO.}

      \item{boutdur.mvpa}{
        Numeric (default = c(1, 5, 10)).
        Duration(s) of MVPA bouts in minutes to be extracted.
        It will start with the identification of the longest to the shortest duration.
        In the default setting, it will start with the 10 minute bouts, followed by 5 minute
        bouts in the rest of the data, and followed by 1 minute bouts in the rest of the data.}

      \item{boutdur.in}{
        Numeric (default = c(10, 20, 30)).
        Duration(s) of inactivity bouts in minutes to be extracted.
        Inactivity bouts are detected in the segments of the data which
        were not labelled as sleep or MVPA bouts.
        It will start with the identification of the longest to the shortest duration.
        In the default setting, it will start with the identification of 30 minute bouts,
        followed by 20 minute bouts in the rest of the data, and followed by 10 minute
        bouts in the rest of the data. Note that we use the term inactivity instead
        of sedentary behaviour for the lowest intensity level of behaviour. The reason
        for this is that GGIR does not attempt to classifying the activity type
        sitting at the moment, by which we feel that using the term sedentary
        behaviour would fail to communicate that.
      }

      \item{boutdur.lig}{
        Numeric (default = c(1, 5, 10)).
        Duration(s) of light activity bouts in minutes
        to be extracted. Light activity bouts are detected in the segments of the data
        which were not labelled as sleep, MVPA, or inactivity bouts.
        It will start with the identification of the longest to the shortest duration.
        In the default setting, this will start with the identification of
        10 minute bouts, followed by 5 minute bouts in the rest of the data, and followed
        by 1 minute bouts in the rest of the data.}

      \item{boutcriter.mvpa}{
        Numeric (default = 0.8).
        A number between 0 and 1, it defines what fraction of a bout needs to be above the \code{threshold.mod}.}

      \item{boutcriter.in}{
        Numeric (default = 0.9).
        A number between 0 and 1, it defines what fraction of a bout needs to be below the \code{threshold.lig}.}

      \item{boutcriter.lig}{
        Numeric (default = 0.8).
        A number between 0 and 1, it defines what fraction of a bout needs to be between
        the \code{threshold.lig} and the \code{threshold.mod}.}

      \item{frag.metrics}{
        Character (default = NULL).
        Fragmentation metric to extract. Can be \"mean\", \"TP\", \"Gini\",
        \"power\", or \"CoV\", \"NFragPM\", or all the above metrics with \"all\".
        See package vignette for description of fragmentation metrics.}

      \item{part6_threshold_combi}{
        Character (default = NULL) to indicate the threshold combination derived in
        part 5 to be used for part 6. For example, \"40_100_120\". If left in default value
        GGIR will use the first threshold value from parameters \code{threshold.lig},
        \code{threshold.mod}, and \code{threshold.vig}.
      }
      \item{bout.metric}{
        Deprecated parameter that was previously used to choose which
        bout detection metric (algorithm) GGIR should use. GGIR now uses only one 
        algorithm.
      }
    }
  }

  \subsection{params_sleep}{
    A list of parameters used to configure the sleep analysis as performend in
    GGIR \link{g.part3} and \link{g.part4}.
    \describe{

      \item{relyonguider}{
        Boolean (default = FALSE).
        Sustained inactivity bouts (sib) that overlap with the guider are 
        labelled as sleep. If \code{relyonguider = FALSE} and the sib overlaps only
        partially with the guider then it is the sib that defines the edge 
        of the SPT window and not the guider.
        If \code{relyonguider = TRUE} and the sib overlaps only partially with the 
        guider then it is the guider that defines the edge of the SPT window 
        and not the sib. If participants were instructed NOT to wear the 
        accelerometer during waking hours and \code{ignorenonware=FALSE} then 
        set to \code{relyonguider=TRUE}, in all other scenarios set to FALSE.}

      \item{relyonsleeplog}{
        Boolean (default = FALSE).
        Do not use, now replaced by argument relyonguider.
        Values provided to argument relyonsleeplog will be passed on to
        argument relyonguider to not preserve functionality of old R scripts.}

      \item{def.noc.sleep}{
        Numeric (default = 1).
        The time window during which sustained
        inactivity will be assumed to represent sleep, e.g., \code{def.noc.sleep = c(21, 9)}.
        This is only used if no sleep log entry is available. If
        left blank \code{def.noc.sleep = c()} then the 12 hour window centred
        at the least active 5 hours of the 24 hour period will be used
        instead. Here, L5 is hardcoded and will not change by changing
        argument winhr in function \link{g.part2}. If def.noc.sleep is filled
        with a single integer, e.g., \code{def.noc.sleep=c(1)} then the window
        will be detected with based on built in algorithms.
        See argument \code{HASPT.algo} from \link{HASPT} for specifying which of the
        algorithms to use.}

      \item{sleepwindowType}{
        Character (default = \"SPT\").
        To indicate type of information in the sleeplog, \"SPT\" for sleep period time.
        Set to \"TimeInBed\" if sleep log recorded time in bed to enable calculation
        of sleep latency and sleep efficiency.}

      \item{nnights}{
        Numeric (default = NULL).
        This argument has been deprecated.}

      \item{loglocation}{
        Character (default = NULL).
        Path to csv file with sleep log information.
        See package vignette for how to format this file.}

      \item{colid}{
        Numeric (default = 1).
        Column number in the sleep log spreadsheet in which the participant ID code is stored.}

      \item{coln1}{
        Numeric (default = 2).
        Column number in the sleep log spreadsheet where the onset of the first night starts.}

      \item{ignorenonwear}{
        Boolean (default = TRUE).
        If TRUE then ignore detected monitor non-wear periods to avoid
        confusion between monitor non-wear time and sustained inactivity.}

      \item{constrain2range}{
        Deprecated, used to be a Boolean (default = TRUE)
        Whether or not to constrain the range of
        threshold used in the diary free sleep period time window detection.}

      \item{HASPT.algo}{
        Character (default = \"HDCZA\").
        To indicate what algorithm should be used for the sleep period time detection.
        Default \"HDCZA\" is Heuristic algorithm looking at Distribution of Change in Z-Angle as
        described in van Hees et al. 2018. Other options included:
        \"HorAngle\", which is based on HDCZA but replaces non-movement detection of
        the HDCZA algorithm by looking for time segments where the angle of the
        longitudinal sensor axis has an angle relative to the horizontal plane
        between -45 and +45 degrees. And \"NotWorn\" which is also the same as HDCZA
        but looks for time segments when a rolling average of acceleration
        magnitude is below 5 per cent of its standard deviation, see 
        Cookbook vignette in the Annexes of https://wadpac.github.io/GGIR/ 
        for more detailed guidance on how to use \"NotWorn\".}
    
      \item{HDCZA_threshold}{
        Numeric (default = c())
        If \code{HASPT.algo} is set to \"HDCZA\" and HDCZA_threshold is NOT NULL,
        (e.g., HDCZA_threshold = 0.2), then that value will be used as threshold
        in the 6th step in the diagram of Figure 1 in van Hees
        et al. 2018 Scientific Report (doi: 10.1038/s41598-018-31266-z). However,
        doing so has not been supported by research yet and is only intended to
        facilitate methodological research, so we advise sticking with the default in
        line with the publication. Further, if HDCZA_threshold is set to a numeric vector of
        length 2, e.g. c(10, 15), that will be used as percentile and 
        multiplier for the above mentioned 6th step.
      }

      \item{HASPT.ignore.invalid}{
        Boolean (default = FALSE).
        To indicate whether invalid time segments should be ignored in the
        heuristic guiders. If \code{FALSE} (default), the imputed 
        angle or activity metric during the invalid time segments are used.
        If \code{TRUE}, invalid time segments are ignored (i.e., they cannot 
        contribute to the guider). If \code{NA}, then invalid time segments are 
        considered to be no movement segments and can contribute to the guider.
        Further, the guider name in the output will be shown with \"+invalid\" 
        its end, e.g. \"HDCZA+invalid\", to reflect the \code{NA} setting.
        When HASPT.algo is \"NotWorn\", HASPT.ignore.invalid is automatically set to
        NA. 
        }

      \item{HASIB.algo}{
        Character (default = \"vanHees2015\").
        To indicate which algorithm should be used to define the
        sustained inactivity bouts (i.e., likely sleep).
        Options: \"vanHees2015\", \"Sadeh1994\", \"Galland2012\", \"NotWorn\",
        \"Oakley1997\", \"data\". For details see vignette:
        https://wadpac.github.io/GGIR/articles/chapter8_SleepFundamentalsSibs.html
        }
        
      \item{Sadeh_axis}{
        Character (default = \"Y\").
        To indicate which axis to use for the Sadeh1994 algorithm, and  other algortihms
        that relied on count-based Actigraphy such as Galland2012.}

      \item{sleeplogsep}{
        Character (default = NULL).
        This argument is deprecated.}

      \item{nap_model}{
        Character (default = NULL).
        To specify classification model. Currently the only option is \"hip3yr\", which
        corresponds to a model trained with hip data in 3-3.5 olds trained with 
        parent diary data. This functionality is currently superseded by
        nap detection which is triggered by \code{possible_nap_window}
        and \code{possible_nap_dur}. Leave nap_model as NULL when using the new
        functionality. More documentation to follow in 2025 when development work 
        has completed.}

      \item{longitudinal_axis}{
        Integer (default = NULL).
        To indicate which axis is the longitudinal axis.
        If not provided, the function will estimate longitudinal axis as the axis
        with the highest 24 hour lagged autocorrelation. Only used when
        \code{sensor.location = \"hip\"} or \code{HASPT.algo = \"HorAngle\"}.}

      \item{anglethreshold}{
        Numeric (default = 5).
        Angle threshold (degrees) for sustained inactivity periods detection.
        The algorithm will look for periods of time (\code{timethreshold})
        in which the angle variability is lower than \code{anglethreshold}.
        This can be specified as multiple thresholds, each of which will be implemented, e.g.,
        \code{anglethreshold = c(5,10)}.}

      \item{timethreshold}{
        Numeric (default = 5).
        Time threshold (minutes) for sustained inactivity periods detection.
        The algorithm will look for periods of time (\code{timethreshold})
        in which the angle variability is lower than \code{anglethreshold}.
        This can be specified as multiple thresholds, each of which will be implemented, e.g.,
        \code{timethreshold = c(5,10)}.}

      \item{possible_nap_window}{
        Numeric (default = NULL).
        Numeric vector of length two with range in clock hours during which naps are
        assumed to take place, e.g., \code{possible_nap_window = c(9, 18)}. The nap 
        classification is only applied
        if both \code{possible_nap_window} and \code{possible_nap_dur} are specified.
        More documentation to follow in 2025 when development work has completed.}

      \item{possible_nap_dur}{
        Numeric (default = NULL).
        Numeric vector of length two with range in duration (minutes) of a nap,
        e.g., \code{possible_nap_dur = c(15, 240)}. The nap classification is only applied
        if both \code{possible_nap_window} and \code{possible_nap_dur} are both specified.
        More documentation to follow in 2025 when development work has completed.}
        
      \item{sleepefficiency.metric}{
        Numeric (default = 1).
        If 1 (default), sleep efficiency is calculated as detected sleep time during
        the SPT window divided by log-derived time in bed. If 2, sleep efficiency is
        calculated as detected sleep time during the SPT window divided by detected
        duration in sleep period time plus sleep latency (where sleep latency refers
        to the difference between time in bed and sleep onset). sleepefficiency.metric is only considered
        when parameter \code{sleepwindowType = \"TimeInBed\"}}
        
      \item{possible_nap_edge_acc}{
        Numeric (default = Inf).
        Maximum acceleration before or after the SIB for the nap to be considered.
        By default this will allow all possible naps.
      }

      \item{possible_nap_gap}{
        Numeric (default = 0).
        Time gap expressed in seconds that is allowed between the sustained 
        inactivity bouts that form the naps.
      }

      \item{sib_must_fully_overlap_with_TimeInBed}{
        Boolean (default = c(TRUE, TRUE)).
        To indicate whether sib must fully overlap with TimeInBed to be considered sleep
        for the start and end of time in bed, respectively.
        Only considered when parameter \code{sleepwindowType = \"TimeInBed\"}. Note
        that negative sleep latency, if any, will be reported in the night summary 
        report (part4_nightsummary_sleep csv files) but these negative sleep latency 
        and corresponding sleep efficiency values when calculating the person summary
        aggregate as stored in the person level report (part4_summary_sleep csv files).
      }
      
      \item{oakley_threshold}{
        Numeric (default = 20)
        Threshold as used by the Oakley algorithm. Original documentation
        suggestion to choose between 20, 40 or 80, while in GGIR this can be
        any absolute number.
      }
      
      \item{consider_marker_button}{
        Boolean (default = FALSE)
        Whether to consider the marker button as guider. Currently only functional 
        for Actiwatch-type count-accelerometers such as 
        Philips Health Band and MotionWatch 8. 
        For details see:
        https://wadpac.github.io/GGIR/articles/chapter9_SleepFundamentalsGuiders.html
      }
      
      \item{impute_marker_button}{
        Boolean (default = FALSE)
        Whether to impute marker buttons on other days of the recording in the context
        of using it as guider with \code{consider_marker_button}. For details see
        https://wadpac.github.io/GGIR/articles/chapter9_SleepFundamentalsGuiders.html
      }
      
      \item{nap_markerbutton_method}{
        Numeric (default = 0)
        Integer to indicate whether and how to use marker button for nap
        detection:
        0 = do not use marker button for nap detection (default);
        1 = if marker button is available use it for nap detection,
        if not rely on accelerometer.
        2 = nearby marker button is condition for nap detection 
        but rely on accelerometer to define exact nap timing, and;
        3 = nearby marker button is condition for nap detection and
        used instead of accelerometer classification.
      }
      
      \item{nap_markerbutton_max_distance}{
        Numeric (default = 30)
        When using \code{nap_markerbutton_method} with a value other than 0
        \code{nap_markerbutton_max_distance}  sets the maximum distance in 
        minutes between the nearest marker button
        and the edge of a sustained inactivity bout. If the marker button lies 
        inside the sustained inactivity bout but beyond the midpoint, no maximum distance
        is applied. For example, if a sustained activity bout lasts
        from 11:00:00 to 13:00:00 then its midpoint will be 12:00:00 and with the
        default \code{nap_markerbutton_max_distance} any marker button between 10:30:00
        and 12:00:00 will be considered valid for the start of the nap and 
        any marker button between 12:00:00 and 13:30:00 will be considered valid 
        for the end of the nap.
      }
      
      \item{SRI1_smoothing_wsize_hrs}{
        Numeric (default = NULL)
        As used for Sleep Regularity Calculation in \link{g.part3}.
        If \code{SRI1_smoothing_wsize_hrs} and \code{SRI1_smoothing_frac} are both
        specified, this is the windowsize for smoothing of detected sustained inactivity bouts
        before calculating SRI1. For example, if set to 1 and 0.8, we apply a rolling 
        1 hour window where each window with 80\% or more sleep is classificed as sleep.
      }
      
      \item{SRI1_smoothing_frac}{
        Numeric (default = NULL)
        As used for Sleep Regularity Calculation in \link{g.part3}.
        If \code{SRI1_smoothing_wsize_hrs} and \code{SRI1_smoothing_frac} are both
        specified, this is the fraction used to smooth detected sustained inactivity bouts
        before calculating SRI1. For example, if set to 1 and 0.8, we apply a rolling 
        1 hour window where each window with 80\% or more sleep is classificed as sleep.
      }
      
      \item{spt_min_block_dur}{
        Numeric (default = 30).
        Minimum size in minutes of a resting block to be considered part of SPT
        in the estimation of the guider window (relevant for HASPT.algo settings: 
        HDCZA, HorAngle and NotWorn).
      }
      
      \item{spt_max_gap_dur}{
        Numeric (default = 60).
        Maximum size in minutes of a gap between resting blocks to be considered part of SPT
        in the estimation of the guider window (relevant for HASPT.algo settings: 
        HDCZA, HorAngle and NotWorn).
      }
      
      \item{spt_max_gap_ratio}{
        Numeric (default = 1).
        Only considered when set to a value less than 1
        Maximum ratio between gap duration and either rest block next to it. If
        the ratio for both is less than spt_max_gap_ratio the gap is merged with
        the neighbouring rest blocks and together they are considered as estimation 
        of the SPT guider window 
        (relevant for HASPT.algo settings: HDCZA, HorAngle and NotWorn).
      }
      
      \item{HorAngle_threshold}{
        Numeric (default = 60).
        Angle threshold used for the HorAngle algorithm.
      }

      \item{guider_cor_maxgap_hrs}{
        Numeric (default = 2).
        Maximum gap duration in hours as used for the correction method.
      }

      \item{guider_cor_min_frac_sib}{
        Numeric (default = 0.5)
        Required fraction of guider-based rest period with sustained inactivity bouts (sib) to be considered as possible extension of guider.
      }
      
      \item{guider_cor_min_hrs}{
        Numeric (default = 2)
        Minimum guider-based rest duration (hours) to be considered as possible extension of guider.
      }
      
      \item{guider_cor_meme_frac_out}{
        Numeric (default = 0.9)
        Minimum required fraction of initial guider classification outside the 
        median start to median end (abbreviated as meme) of the guider across days
        to initiate search for secondary that falls within this window. 
      }
      
      \item{guider_cor_meme_frac_in}{
        Numeric (default = 0.4)
        Minimum required fraction of alternative guider classification inside the
        median start to median end (abbreviated as meme) of the guider across days 
        to be considered the new guider window.
      }
      
      \item{guider_cor_meme_min_hrs}{
        Numeric (default = 1)
        Minimum required duration of the alternative guider classification inside the
        median start to median end (abbreviated as meme) of the guider across days 
        to be considered the new guider window. 
      }
      
      \item{guider_cor_do}{
        Boolean (default = FALSE)
        Whether to apply the guider correction algorithm as discussed in Chapter 9 of the online
        documentation: https://wadpac.github.io/GGIR/articles/chapter9_SleepFundamentalsGuiders.html.
        In short, when sleep is fragmented, algorithms to guide the detection of SPT such
        as HDCZA may occassionally miss a part of the night or incorrectly consider a day 
        time nap as the main sleep window in a day. This optional algorithm aims
        to correct for such misclassifications.
      }
      
      \item{guider_cor_meme_min_dys}{
        Numeric (default = 3)
        Minimum number of days required to consider median-median based step in the 
        guider correction algorithm as controlled by parameter \code{guider_cor_meme_frac_out},
        \code{guider_cor_meme_frac_in}, and \code{guider_cor_meme_min_hrs}.
        This parameter is set to 3 by default as a median becomes only meaningful with
        at least three values.
      }
    }
  }

  \subsection{params_247}{
    A list of parameters releated to description of 24/7 behaviours that do not fall
    under conventional physical activity or sleep outcomes, these parameters are used
    in GGIR \link{g.part2} and GGIR \link{g.part5}:
    \describe{

      \item{qwindow}{
        Numeric or character (default = c(0, 24)).
        To specify windows over which all variables are calculated, e.g., acceleration 
        distribution, number of valid hours, LXMX analysis, MVPA.
        If numeric, qwindow should have length two, e.g., \code{qwindow = c(0, 24)},
        all variables will only be calculated over the full 24 hours in a day. If
        \code{qwindow = c(8, 24)} variables will be calculated over the window 0-8, 8-24 and 0-24.
        All days in the recording will be segmented based on these values.
        If you want to use a day specific segmentation in each day then you can set 
        qwindow to be the full path to activity diary file (character). Expected 
        format of the activity diary is: First column headers followed by one row 
        per recording, first column is recording ID, which needs to match with the 
        ID GGIR extracts from the accelerometer file. Followed by date column in 
        format \"23-04-2017\", where date format is specified by parameter
        \code{qwindow_dateformat} (below). Use the character combination date, 
        Date or DATE in the column name. This is followed by one or multiple columns 
        with start times for the activity types in that day format in hours:minutes:seconds. 
        The header of the column will be used as label for each activity type. 
        Insert a new date column before continuing with activity types for next day. 
        Leave missing values empty. If an activity log is used then individuals who 
        do not appear in the activity log will still be processed with value 
        \code{qwindow = c(0, 24)}. Dates with no activity log data can be skipped,
        no need to have a column with the date followed by a column with the next 
        date. If times in the activity diary are not multiple of the short window 
        size (epoch length), the next epoch is considered (e.g., with epoch of 5 
        seconds, 8:00:02 will be redefined as 8:00:05 in the activity log).
        When using the qwindow functionality in combination with GGIR part 5 then
        make sure to check that parameters \code{segmentWEARcrit.part5} and 
        \code{segmentDAYSPTcrit.part5} are specified to your research needs.
        When using an activity diary be aware that any column name including the words
        \"impute\" or \"uncertain\" will be ignored. This means you can, for you own 
        convenience, add columns to log which timestamps have been manually imputed 
        or are considered uncertain.
        }

      \item{qwindow_dateformat}{
        Character (default = \"\%d-\%m-\%Y\").
        To specify the date format used in the activity log as used by \link[base]{strptime}.
        For guidance on how to specify time formats in R see:
        https://wadpac.github.io/GGIR/articles/DateTimeFormatsInR.html
        }

      \item{M5L5res}{
        Numeric (default = 10).
        Resolution of L5 and M5 analysis in minutes.}

      \item{winhr}{
        Numeric (default = 5).
        Vector of window size(s) (unit: hours) of LX and MX analysis,
        where look for least and most active consecutive number of X hours.}

      \item{qlevels}{
        Numeric (default = NULL).
        Vector of percentiles for which value needs to be extracted. These need to 
        be expressed as a fraction of 1, e.g., c(0.1, 0.5, 0.75). There is no limit 
        to the number of percentiles. If left empty then percentiles will not be extracted. 
        Distribution will be derived from short epoch metric data. Parameter qlevels 
        can for example be used for the MX-metrics (e.g. Rowlands et al) as discussed in 
        https://wadpac.github.io/GGIR/articles/chapter7_DescribingDataWithoutKnowingSleep.html#sets-of-quantiles-mx-metrics-by-rowlands-et-al-
        }

      \item{ilevels}{
        Numeric (default = NULL).
        Levels for acceleration value frequency distribution in m\emph{g}, e.g., 
        \code{ilevels = c(0,100,200)}. There is no limit to the number of levels. 
        If left empty then the intensity levels will not be extracted. Distribution 
        will be derived from short epoch metric data.}

      \item{iglevels}{
        Numeric (default = NULL).
        Levels for acceleration value frequency distribution
        in m\emph{g} used for intensity gradient calculation (according to the method by
        Rowlands 2018). By default this is parameter is empty and the intensity gradient
        calculation is not done. The user can either provide a single value (any) to
        make the intensity gradient use the bins \code{iglevels = c(seq(0, 4000, by = 25), 8000)}
        or the user could specify their own distribution. There is no constriction to the
        number of levels.}

      \item{IVIS_windowsize_minutes}{
        This argument has been deprecated.}

      \item{IVIS_epochsize_seconds}{
        Numeric (default = NULL).
        This parameter has been deprecated.}

      \item{IVIS.activity.metric}{
        This argument has been deprecated.}

      \item{IVIS_acc_threshold}{
        This argument has been deprecated.}

      \item{qM5L5}{
        Numeric (default = NULL).
        Percentiles (quantiles) to be calculated over L5 and M5 window.}

      \item{MX.ig.min.dur}{
        Numeric (default = 10).
        Minimum MX duration needed in order for intensity gradient to be calculated.}

      \item{LUXthresholds}{
        Numeric (default = c(0, 100, 500, 1000, 3000, 5000, 10000)).
        Vector with numeric sequence corresponding to
        the thresholds used to calculate time spent in LUX ranges.}

      \item{LUX_cal_constant}{
        Numeric (default = NULL).
        If both LUX_cal_constant and LUX_cal_exponent are
        provided LUX values are converted based on formula y = constant * exp(x * exponent)}

      \item{LUX_cal_exponent}{
        Numeric (default = NULL).
        If both LUX_cal_constant and LUX_cal_exponent are provided LUX
        LUX values are converted based on formula y = constant * exp(x * exponent)}

      \item{LUX_day_segments}{
        Numeric (default = NULL).
        Vector with hours at which the day should be segmented for
        the LUX analysis.}

      \item{L5M5window}{
        Has been deprecated after version 1.5-24.
        This parameter used to define the start and end time, in 24 hour clock hours,
        over which L5M5 needs to be calculated. Now this is done with parameter qwindow.}

      \item{cosinor}{
        Boolean (default = FALSE). Whether to apply the circadian rhythm analysis in part 2.
        This parameter will be superseded by new parameter \code{part2CR} to better reflect
        that this controls all circadian rhythm parameters. If \code{part2CR} is specified then
        that setting overwrites \code{cosinor}.}
      
      \item{part2CR}{
        Boolean (default = FALSE). Whether circadian rhythm analysis should be run in
        part 2, including: cosinor analysis, extended cosinor analysis, IS, IV, and phi.}
      
      \item{part6CR}{
        Boolean (default = FALSE) to indicate whether circadian rhythm analysis should
        be run by part 6, this includes: cosinor analysis, extended cosinor analysis,
        IS, IV, and phi. Optionally this can be expanded with detrended fluctutation analysis
        which is controlled by parameter `part6DFA`.
      }
      \item{part6HCA}{
        Boolean (default = FALSE) to indicate whether Household Co Analysis should 
        be run by part 6.
      }
      \item{part6Window}{
        Character vector with length two (default = c(\"start\", \"end\")) to indicate
        the start and the end of the time series to be used for circadian rhythm analysis
        in part 6. In other words, this parameters is not used for Household co-analysis.
        Alternative values are: \"Wx\", \"Ox\", \"Hx\", where \"x\" is a number to indicat
        the xth wakeup, onset or hour of the recording. Negative values for \"x\"
        are also possible and will count relative to the end of the recording. For example,
        c(\"W1\", \"W-1\") goes from the first till the last wakeup, c(\"H5\", \"H-5\") 
        ignores the first and last 5 hours, and c(\"O2\", \"W10\") goes from the second
        onset till the 10th wakeup time.
      }

      \item{part6DFA}{
        Boolean (default = FALSE) to indicate whether to perform Detrended Fluctuation
        Analysis. Turned off by default because it can be time consuming.
      }

      \item{clevels}{
        Not fully operational yet, to be actived in 2025.
        Numeric vector (default = c(30, 150)) with length 2 to indicate cadence ranges used 
        when summarising step counts, if available. When set to NULL, step counts will not be
        summarised in part 5 output. 
      }

      \item{SRI2_WASOmin}{
        Numeric (default = 30)
        Minimum WASO duration in minutes as used for Sleep Regularity Calculation in \link{g.part6}.
      }
    }
  }

  \subsection{params_output}{
    A list of parameters used to specify whether and how GGIR stores its output at various stages of the
    process.
    \describe{

      \item{storefolderstructure}{
        Boolean (default = FALSE).
        Store folder structure of the accelerometer data.}

      \item{do.part2.pdf}{
        Boolean (default = TRUE).
        Now replaced by \code{do.part2.png}. If specified and 
        \code{do.part2.png} is not specified then value will be assigned 
        to \code{do.part2.png}. Used in \link{g.part2} to indicate whether
        to generate a pdf before, but this is now a png file.}
        
      \item{do.part2.png}{
        Boolean (default = TRUE).
        Used in \link{g.part2} to indicate whether to generate a png file for 
        each recording.}

      \item{do.part3.pdf}{
        Boolean (default = FALSE).
        In \link{g.part3}: Whether to generate a pdf for \link{g.part3}.}

      \item{timewindow}{
        Character (default = c(\"MM\", \"WW\")).
        In \link{g.part5}: Timewindow over which summary statistics are derived.
        Value can be \"MM\" (midnight to midnight), \"WW\" (waking time to waking time),
        \"OO\" (sleep onset to sleep onset), or any combination of them.}

      \item{save_ms5rawlevels}{
        Boolean (default = TRUE).
        In \link{g.part5}: Whether to save the time series classification (levels)
        as csv or RData files (as defined by \code{save_ms5raw_format}). Note that 
        time stamps will be stored in the column \code{timenum} in UTC format (i.e., 
        seconds from 1970-01-01). To convert timenum to time stamp format, you 
        need to specify your desired time zone, e.g., 
        \code{as.POSIXct(mdat$timenum, tz = \"Europe/London\")}. If you are not using
        GGIR part 6, are not interested in the visualreport generation, and not interested
        in time series then you may want to consider setting this parameter to FALSE.}

      \item{save_ms5raw_format}{
        Character (default = \"RData\").
        In \link{g.part5}: To specify how data should be stored: \"csv\", 
        \"RData\", or both via c(\"csv\", \"RData\"). Only used if \code{save_ms5rawlevels = TRUE}.}

      \item{save_ms5raw_without_invalid}{
        Boolean (default = TRUE).
        In \link{g.part5}: To indicate whether to remove invalid days from the 
        time series output files. Only used if \code{save_ms5rawlevels = TRUE}.}

      \item{epochvalues2csv}{
        Boolean (default = FALSE).
        In \link{g.part2}: If TRUE then epoch values are exported to a csv file.
        Here, non-wear time is imputed where possible.}

      \item{do.sibreport}{
        Boolean (default = TRUE).
        In \link{g.part4}: To indicate whether to generate report for the sustained 
        inactivity bouts (SIB). If set to TRUE and when an advanced sleep diary is
        available in part 4 then part 5 will use this to generate summary statistics
        on the overlap between self-reported nonwear and napping with SIB. Here,
        SIB can be filter based on parameter possible_nap_edge_acc and the first value
        of possible_nap_dur}

      \item{do.visual}{
        Boolean (default = TRUE).
        In \link{g.part4}: If TRUE, the function will generate a pdf with a visual
        representation of the overlap between the sleeplog entries and the accelerometer
        detections. This can be used to visually verify that the sleeplog entries do
        not come with obvious mistakes.}

      \item{outliers.only}{
        Boolean (default = FALSE).
        In \link{g.part4}: Only used if \code{do.visual = TRUE}. If FALSE,
        all available nights are included in the visual representation of the data and sleeplog.
        If TRUE, then only nights with a difference in onset or waking time
        larger than the variable of parameter \code{criterror} will be included.}

      \item{criterror}{
        Numeric (default = 3).
        In \link{g.part4}: Only used if \code{do.visual = TRUE} and \code{outliers.only = TRUE}.
        criterror specifies the number of minimum number of hours difference
        between sleep log and  accelerometer estimate for the night to be
        included in the visualisation.}

      \item{visualreport}{
        Boolean (default = TRUE).
        If TRUE, store a visualreport in outputfolder 'results/file summary reports'
        for each recording. Note that this report was introduced in release 3.1-8 
        and replaced an older visual report that had been in GGIR for many years.
        The older report was only intended as a quick attempt to have 
        something visual to show
        to study participants back in 2015. It was not designed for data quality 
        checking purposes. The new visual report is specifically aimed at 
        supporting data quality checks as it shows the time series data as generated 
        and used in GGIR part 5. }.

      \item{viewingwindow}{
        Numeric (default = 1).
        Centre the day as displayed around noon (\code{viewingwindow = 1}) or around 
        midnight (\code{viewingwindow = 2}) in the \"Report_\" visual report generated with 
        \code{visualreport = TRUE}.}

      \item{week_weekend_aggregate.part5}{
        Boolean (default = FALSE).
        In \link{g.part5}: To indicate whether week and weekend-days aggregates
        should be stored. This is turned off by default as it generates a
        large number of extra columns in the output report.}

      \item{dofirstpage}{
        Boolean (default = TRUE).
        To indicate whether a first page with histograms summarizing the whole
        measurement should be added in the file summary reports named \"old_report_\" 
        generated with \code{visualreport = TRUE}.}

      \item{sep_reports}{
        Character (default = \",\").
        Value used as sep parameter in \link[data.table]{fwrite} for writing csv reports.}
        
      \item{dec_reports}{
        Character (default = \".\").
        Value used as dec parameter in \link[data.table]{fwrite} for writing csv reports.}

      \item{sep_config}{
        Character (default = \",\").
        Value used as sep parameter in \link[data.table]{fwrite} for writing csv config file.}
        
      \item{dec_config}{
        Character (default = \".\").
        Value used as dec parameter in \link[data.table]{fwrite} for writing csv config file.}

      \item{visualreport_without_invalid}{
          Boolean (default = TRUE).
          If TRUE, then reports generated with \code{visualreport = TRUE} named \"old_report_\"
          only show the windows with sufficiently valid data according to \code{includedaycrit}
          when viewingwindow = 1 or \code{includenightcrit} when viewingwindow = 2}

      \item{require_complete_lastnight_part5}{    
          Boolean (default = FALSE).
          When set to TRUE: The last WW window is excluded if the recording ends 
          between midnight and 3pm, and starts on a date that is on or one day before the 
          recording end date; The last OO and MM window are excluded if recording 
          ends between midnight and 9am, and starts on a date that is on or one
          day before the recording end date. This to avoid risk that recording end 
          biases the sleep estimates for the last night.}

      \item{old_visualreport}{
        Boolean (default = FALSE).
        If TRUE, then generate old visual report in addition to new visualreport.
        The old visualreport will eventually be deprecated and is not suitable for
        data quality assessment, see documentation for \code{visualreport} above
        for details.}
      
      \item{visualreport_hrsPerRow}{
        Numeric (default = 36).
        Width of the plots in the new visualreport (named \"report_...\") 
        expressed in hours.
        Expected to be in the range 24-48. If more than 24 then the extra time overlaps
        with the beginning of the next plot.}
      
      \item{visualreport_focus}{
         Character (default = \"day\")
         Whether new visual report (names start with \"report...\") focuss on day
         or night. If set ot \"night\" the focus will be
         on the night.
      }
      
      \item{visualreport_validcrit}{
         Numeric (default = 0).
         Value between 0 and 1 representing the fraction of data in a plot 
         that is expected to be valid. If this criteria is not met the row is skipped.
         Only used in new visual report (names start with \"report...\").
      }
      
      \item{method_research_vars}{
        Character (default = NULL).
        Vector with names of methodological variable categories to store in the
        csv output files. Currently on \"nap\" is available which affects the part5
        output. The variables are intended for methodological research only and
        are by default turned off.
      }
    }
  }
}
\examples{
\dontrun{
  mode = c(1,2,3,4,5)
  datadir = \"C:/myfolder/mydata\"
  outputdir = \"C:/myresults\"
  studyname =\"test\"
  f0 = 1
  f1 = 2
  GGIR(#-------------------------------
       # General parameters
       #-------------------------------
       mode = mode,
       datadir = datadir,
       outputdir = outputdir,
       studyname = studyname,
       f0 = f0,
       f1 = f1,
       overwrite = FALSE,
       do.imp = TRUE,
       idloc = 1,
       print.filename = FALSE,
       storefolderstructure = FALSE,
       #-------------------------------
       # Part 1 parameters:
       #-------------------------------
       windowsizes = c(5,900,3600),
       do.cal = TRUE,
       do.enmo = TRUE,
       do.anglez = TRUE,
       chunksize = 1,
       printsummary = TRUE,
       #-------------------------------
       # Part 2 parameters:
       #-------------------------------
       data_masking_strategy = 1,
       ndayswindow = 7,
       hrs.del.start = 1,
       hrs.del.end = 1,
       maxdur = 9,
       includedaycrit = 16,
       L5M5window = c(0,24),
       M5L5res = 10,
       winhr = c(5,10),
       qlevels = c(c(1380/1440),c(1410/1440)),
       qwindow = c(0,24),
       ilevels = c(seq(0,400,by=50),8000),
       mvpathreshold = c(100,120),
       #-------------------------------
       # Part 3 parameters:
       #-------------------------------
       timethreshold = c(5,10),
       anglethreshold = 5,
       ignorenonwear = TRUE,
       #-------------------------------
       # Part 4 parameters:
       #-------------------------------
       excludefirstlast = FALSE,
       includenightcrit = 16,
       def.noc.sleep = 1,
       loglocation = \"D:/sleeplog.csv\",
       outliers.only = FALSE,
       criterror = 4,
       relyonguider = FALSE,
       colid = 1,
       coln1 = 2,
       do.visual = TRUE,
       #-------------------------------
       # Part 5 parameters:
       #-------------------------------
       # Key functions: Merging physical activity with sleep analyses
       threshold.lig = c(30,40,50),
       threshold.mod = c(100,120),
       threshold.vig = c(400,500),
       excludefirstlast = FALSE,
       boutcriter = 0.8,
       boutcriter.in = 0.9,
       boutcriter.lig = 0.8,
       boutcriter.mvpa = 0.8,
       boutdur.in = c(10,20,30),
       boutdur.lig = c(1,5,10),
       boutdur.mvpa = c(1,5,10),
       timewindow = c(\"WW\"),
       #-----------------------------------
       # Report generation
       #-------------------------------
       do.report = c(2,4,5))

       # For externally derived Actiwatch data in .AWD format:
       GGIR(datadir = \"/media/actiwatch_awd\", # folder with epoch level .AWD file
          outputdir = \"/media/myoutput\",
          dataFormat = \"actiwatch_awd\",
          extEpochData_timeformat = \"\%m/\%d/\%Y \%H:\%M:\%S\",
          mode = 1:5,
          do.report = c(2, 4, 5),
          windowsizes = c(60, 900, 3600), # 60 is the expected epoch length
          visualreport = FALSE,
          outliers.only = FALSE,
          overwrite = TRUE,
          HASIB.algo = \"Sadeh1994\",
          def.noc.sleep = c()) # <= because we cannot use HDCZA for ZCY

       # For externally derived Actiwatch data in .CSV format:
       GGIR(datadir = \"/media/actiwatch_csv\", # folder with epoch level .AWD file
          outputdir = \"/media/myoutput\",
          dataFormat = \"actiwatch_csv\",
          extEpochData_timeformat = \"\%m/\%d/\%Y \%H:\%M:\%S\",
          mode = 1:5,
          do.report = c(2, 4, 5),
          windowsizes = c(15, 900, 3600), # 15 is the expected epoch length
          visualreport = FALSE,
          outliers.only = FALSE,
          HASIB.algo = \"Sadeh1994\",
          def.noc.sleep = c()) # <= because we cannot use HDCZA for ZCY

       # For externally derived UK Biobank data in .CSV format:
       GGIR(datadir = \"/media/ukbiobank\",
           outputdir = \"/media/myoutput\",
           dataFormat = \"ukbiobank_csv\",
           extEpochData_timeformat = \"\%m/\%d/\%Y \%H:\%M:\%S\",
           mode = c(1:2),
           do.report = c(2),
           windowsizes = c(5, 900, 3600), # We know that data was stored in 5 second epoch
           desiredtz = \"Europe/London\", # We know that data was collected in the UK
           visualreport = FALSE,
           overwrite = TRUE)
        
       # For externally derived ActiGraph count data in .CSV format assuming
       # a study protocol where sensor was not worn during the night:
       GGIR(datadir = \"/examplefiles\",
           outputdir = \"\",
           dataFormat = \"actigraph_csv\",
           mode = 1:5,
           do.report = c(2, 4, 5),
           windowsizes = c(5, 900, 3600),
           threshold.in = round(100 * (5/60), digits = 2),
           threshold.mod = round(2500 * (5/60), digits = 2),
           threshold.vig = round(10000 * (5/60), digits = 2),
           extEpochData_timeformat = \"\%m/\%d/\%Y \%H:\%M:\%S\",
           do.neishabouricounts = TRUE,
           acc.metric = \"NeishabouriCount_x\",
           HASPT.algo = \"NotWorn\",
           HASIB.algo = \"NotWorn\",
           do.visual = TRUE,
           includedaycrit = 10,
           includenightcrit = 10,
           visualreport = FALSE,
           outliers.only = FALSE,
           save_ms5rawlevels = TRUE,
           ignorenonwear = FALSE,
           HASPT.ignore.invalid = FALSE,
           save_ms5raw_without_invalid = FALSE)
           
           
       # For externally derived Sensear data in .xls format:  
        GGIR(datadir = \"C:/yoursenseweardatafolder\",
            outputdir = \"D:/youroutputfolder\",
            mode = 1:5,
            windowsizes = c(60, 900, 3600),
            threshold.in = 1.5,
            threshold.mod = 3,
            threshold.vig = 6,
            dataFormat = \"sensewear_xls\",
            extEpochData_timeformat = \"\%d-\%b-\%Y \%H:\%M:\%S\",
            HASPT.algo = \"NotWorn\",
            desiredtz = \"America/New_York\",
            overwrite = TRUE,
            do.report = c(2, 4, 5),
            visualreport = FALSE)
           
  }
}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}
\references{
  \itemize{
    \item van Hees VT, Gorzelniak L, Dean Leon EC, Eder M, Pias M, et al. (2013) Separating
    Movement and Gravity Components in an Acceleration Signal and Implications for the
    Assessment of Human Daily Physical Activity. PLoS ONE 8(4): e61691.
    doi:10.1371/journal.pone.0061691
    \item van Hees VT, Fang Z, Langford J, Assah F, Mohammad A, da Silva IC, Trenell MI,
    White T, Wareham NJ, Brage S. Auto-calibration of accelerometer data for
    free-living physical activity assessment using local gravity and temperature:
    an evaluation on four continents. J Appl Physiol (1985). 2014 Aug 7
    \item van Hees VT, Sabia S, et al. (2015) A novel, open access method to
    assess sleep duration using a wrist-worn accelerometer, PLoS ONE, November 2015
  }
}

\name{g.getstarttime}
\alias{g.getstarttime}
\title{
  Extract start time of a measurement
}
\description{
  Extract start time of a measurement. GGIR calculates all timestamps by
  using the first timestamp and sample frequency. Not intended
  for direct use by package user
}
\usage{
  g.getstarttime(datafile, data, mon, dformat, desiredtz,
  configtz = NULL)
}
\arguments{
  \item{datafile}{
   Full path to data file
  }
  \item{data}{
    Data part of \link{g.readaccfile} output
  }
  \item{mon}{
    Same as in \link{g.dotorcomma}
  }
  \item{dformat}{
    Same as in \link{g.dotorcomma}
  }
  \item{desiredtz}{
    Same as in \link{g.dotorcomma}
  }
  \item{configtz}{
    Same as in \link{g.dotorcomma}
  }
}
\value{
  The starttime
}
\keyword{internal}
\author{
Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.sib.plot}
\alias{g.sib.plot}
\title{
  Create plot of sustained inactivity bouts
}
\description{
  Function create plot of sustained inactivity bouts for quality
  check purposes as part of \link{g.part3}. Not intended for direct use by package user
}
\usage{
  g.sib.plot(SLE, M, I, plottitle, nightsperpage=7, desiredtz="")
}
\arguments{
  \item{SLE}{
    Output from \link{g.sib.det}
  }
  \item{M}{
    Output from \link{g.getmeta}
  }
  \item{I}{
    Output from \link{g.inspectfile}
  }
  \item{plottitle}{
    Title to be used in the plot
  }
  \item{nightsperpage}{
     Number of nights to show per page
  }
  \item{desiredtz}{
    See \link{g.part3}
  }
}
\value{
  Function has no output other than the plot
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.weardec}
\alias{g.weardec}
\title{
  Detects whether accelerometer is worn
}
\description{
  Uses the object produced by \link{g.part1} to assess
  whether the accelerometer was worn
}
\usage{
  g.weardec(metalong, wearthreshold, ws2, params_cleaning = NULL,
            desiredtz = "", qwindowImp = c())
}
\arguments{
  \item{metalong}{
    Object produced by \link{g.getmeta}
  }
  \item{wearthreshold}{
    Number of axis that at least need to meet the non-wear criteria
  }
  \item{ws2}{
    Large windowsize used in seconds to apply non-wear detection
    Small window size not needed, because this is inherent to the object M
  }
  \item{params_cleaning}{
    Parameters object with cleaning paramete, see \link{GGIR}.
  }
  \item{desiredtz}{
    See \link{GGIR}
  }
  \item{qwindowImp}{
    See \link{g.impute}
  }
}
\value{
  \itemize{
    \item \code{r1} Participant id extracted from file
    \item \code{r2} Night number
    \item \code{r3} Detected onset of sleep expressed as hours 
    since the previous midnight
    \item \code{LC} fraction of 15 minute windows with more than 5 percent
    clipping
    \item \code{LC2} fraction of 15 minute windows with more than 80
    percent clipping
  }
}
\examples{
  data(data.getmeta)
  params_cleaning = load_params()$params_cleaning
  params_cleaning[["nonwearFilterWindow"]] = c(22, 7)
  output = g.weardec(metalong = data.getmeta$metalong, wearthreshold = 2, ws2 = 900,  
                      params_cleaning = params_cleaning,
                      desiredtz = "", qwindowImp = NULL)
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.part5.wakesleepwindows}
\alias{g.part5.wakesleepwindows}
\title{
  Label wake and sleepperiod window
}
\description{
  Not intended for direct use by GGIR users.
  Label wake and sleepperiod window as part of \link{g.part5}.
}
\usage{
  g.part5.wakesleepwindows(ts, part4_output, desiredtz,
                    nightsi, sleeplog, epochSize,ID, Nepochsinhour)
}
\arguments{
  \item{ts}{
    data.frame with time series
  }
  \item{part4_output}{
    cleaned output from part 4
  }
  \item{desiredtz}{
    \link{GGIR}
  }
  \item{nightsi}{
    vector with indices for the midnights
  }
  \item{sleeplog}{
    Data.frame with sleeplog information as loaded by \link{g.loadlog}
  }
  \item{epochSize}{
    Short epochsize in seconds
  }
  \item{ID}{
    ID of the participant
  }
  \item{Nepochsinhour}{
    Number of epochs in an hour
  }
}
\value{
  Object ts
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{data.inspectfile}
\alias{data.inspectfile}
\docType{data}
\title{
Example output from g.inspectfile
}
\description{
data.inspectfile is example output from \link{g.inspectfile}
}
\usage{data(data.inspectfile)}
\format{
  The format is:
 chr "data.inspectfile"
}
\source{
The data was collected on one individual for testing purposes
}
\examples{
data(data.inspectfile)
}
\keyword{datasets}

\name{g.calibrate}
\alias{g.calibrate}
\title{
  function to estimate calibration error and make recommendation for addressing it
}
\description{
  Function starts by identifying ten second windows of non-movement. Next, the
  average acceleration per axis per window is used to estimate calibration error 
  (offset and scaling) per axis. The function provides recommended correction factors
  to address the calibration error and a summary of the callibration procedure.
}
\usage{
  g.calibrate(datafile, params_rawdata = c(), params_general = c(),
              params_cleaning = c(), inspectfileobject = c(), verbose = TRUE, ...)
}
\arguments{
  \item{datafile}{
    Name of accelerometer file
  }
  \item{params_rawdata}{
    See \link{g.part1}
  }
  \item{params_general}{
    See \link{g.part1}
  }
  \item{params_cleaning}{
    See \link{g.part1}
  }
  \item{inspectfileobject}{
    Output from the function \link{g.inspectfile}.
  }
  \item{verbose}{
    Boolean (default = TRUE). to indicate whether console message should be printed. 
    Note that warnings and error are always printed and can be suppressed with 
    suppressWarning() or suppressMessages().
  }
  \item{...}{
     Any argument used in the previous version of g.calibrate, which will now
     be used to overrule the arguments specified with the parameter objects.
  }  
}
\value{
 \item{\code{scale}}{scaling correction values, e.g. c(1,1,1) }
  \item{\code{offset}}{offset correction values, e.g. c(0,0,0)}
  \item{\code{tempoffset}}{correction values related to temperature, e.g. c(0,0,0)}
  \item{\code{cal.error.start}}{absolute difference between Euclidean norm during all
  non-movement windows and 1 g before autocalibration}
  \item{\code{cal.error.end}}{absolute difference between Euclidean norm during all 
  non-movement windows and 1 g after autocalibration}
  \item{\code{spheredata}}{average, standard deviation, Euclidean norm and temperature
  (if available) for all ten second non-movement windows as used for the
  autocalibration procedure}
  \item{\code{npoints}}{number of 10 second no-movement windows used to populate
  the sphere}
  \item{\code{nhoursused}}{number of hours of measurement data scanned to find
    the ten second time windows with no movement}
  \item{\code{meantempcal}}{mean temperature corresponding to the data as used for 
    autocalibration. Only applies to data where temperate data is collected
    and available to GGIR, such as GENEActiv, Axivity, and in some instances
    ad-hoc .csv data.
  }
}
\examples{
  \dontrun{
    datafile = "C:/myfolder/testfile.bin"
    
    #Apply autocalibration:
    C = g.calibrate(datafile)
    print(C$scale)
    print(C$offset)
  }
}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
  Zhou Fang
}
\references{
  \itemize{
    \item van Hees VT, Fang Z, Langford J, Assah F, Mohammad A, da Silva IC, Trenell MI, 
    White T, Wareham NJ, Brage S. Auto-calibration of accelerometer data for
    free-living physical activity assessment using local gravity and temperature: 
    an evaluation on four continents. J Appl Physiol (1985). 2014 Aug 7
  }
}

\name{g.inspectfile}
\alias{g.inspectfile}
\title{
function to inspect accelerometer file for brand, sample frequency and header
}
\description{
Inspects accelerometer file for key information, including: monitor brand, sample frequency and file header
}
\usage{
g.inspectfile(datafile, desiredtz = "", params_rawdata = c(),
                         configtz = c(), ...)
}
\arguments{
  \item{datafile}{
  	name of data file
  }
  \item{desiredtz}{
  	Desired timezone, see documentation \link{g.getmeta}
  }
  \item{params_rawdata}{
    See \link{g.part1}
  }
  \item{configtz}{
    ...
  }
  \item{...}{
   Any argument used in the previous version of g.getmeta, which will now
    be used to overrule the arguments specified with the parameter objects.
  }
}
  
\value{
 \item{header}{fileheader}
  \item{monn}{monitor name (genea, geneactive)}
  \item{monc}{monitor brand code (0 - ad-hoc file format, 1 = genea (non-commercial),
      2 = GENEActive, 3 = actigraph, 4 = Axivity (AX3, AX6), 5 = Movisense, 6 = Verisense)}
   \item{dformn}{data format name, e.g bin, csv, cwa, gt3x}
  \item{dformc}{data format code (1 = .bin, 2 = .csv, 3 = .wav, 4 = .cwa, 5 = ad-hoc .csv, 6 = .gt3x)}
  \item{sf}{samplefrequency in Hertz}
  \item{filename}{filename}
}

\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.abr.day.names}
\alias{g.abr.day.names}
\title{
  Abbreviates daynames to numbers, needed for report generation in
  \link{g.plot5}
}
\description{
  Abbreviates daynames Monday becomes MON and Sunday becomes SUN
}
\usage{
  g.abr.day.names(daynames)	
}
\arguments{
  \item{daynames}{
    Vector of daynames in character format
  }
}
\examples{
  daynames = c("Monday","Friday")
  daynames_converted = g.abr.day.names(daynames)
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.part5.onsetwaketiming}
\alias{g.part5.onsetwaketiming}
\title{
  Identify wake and sleepperiod window timing
}
\description{
  Not intended for direct use by GGIR users.
  Labels timing of wakeing up and sleep onset as part of \link{g.part5}.
}
\usage{
  g.part5.onsetwaketiming(qqq, ts, min, sec, hour, timewindowi)
}
\arguments{
  \item{qqq}{
    Start and end index of window to analyses
  }
  \item{ts}{
    Data.frame with time series as created in \link{g.part5}
  }
  \item{min}{
    Numeric vector with minute values
  }
  \item{sec}{
    Numeric vector with second values
  }
  \item{hour}{
    Numeric vector with hour values
  }
  \item{timewindowi}{
    Character to indicate what timewindow definition is used either "MM" or "WW"
  }
}
\value{
  A list with objects: wake, onset, wakei, onseti, skiponset, and skipwake.
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{read.myacc.csv}
\alias{read.myacc.csv}
\title{
  Read custom csv files with accelerometer data
}
\description{
  Loads csv files with accelerometer data and standardises 
  the output format (incl. unit of measurement, timestamp format,
  header format, and column locations) to make the data compatible
  with other GGIR functions.
}
\usage{
  read.myacc.csv(rmc.file=c(), rmc.nrow=Inf, rmc.skip = c(), rmc.dec=".",
                          rmc.firstrow.acc = c(), rmc.firstrow.header=c(),
                          rmc.header.length = c(),
                          rmc.col.acc = 1:3, rmc.col.temp = c(), 
                          rmc.col.time=c(),
                          rmc.unit.acc = "g", rmc.unit.temp = "C", 
                          rmc.unit.time = "POSIX",
                          rmc.format.time = "\%Y-\%m-\%d \%H:\%M:\%OS",
                          rmc.bitrate = c(), rmc.dynamic_range = c(), 
                          rmc.unsignedbit = TRUE,
                          rmc.origin = "1970-01-01",
                          rmc.desiredtz = NULL,
                          rmc.configtz = NULL,
                          rmc.sf = c(),
                          rmc.headername.sf = c(),
                          rmc.headername.sn = c(),
                          rmc.headername.recordingid = c(),
                          rmc.header.structure = c(),
                          rmc.check4timegaps = FALSE,
                          rmc.col.wear = c(),
                          rmc.doresample = FALSE,
                          rmc.scalefactor.acc = 1,
                          interpolationType=1, 
                          PreviousLastValue = c(0, 0, 1),
                          PreviousLastTime = NULL,
                          desiredtz = NULL,
                          configtz = NULL,
                          header = NULL)	
}
\arguments{
  \item{rmc.file}{
    Filename of file to be read if it is in the working directory, or full path to the file otherwise.
  }
  \item{rmc.nrow}{
    Number of rows to read, same as nrow argument in \link[utils]{read.csv} and nrows in \link[data.table]{fread}. The whole file is read by default (i.e., rmc.nrow = Inf).
  }
  \item{rmc.skip}{
    Number of rows to skip, same as skip argument in \link[utils]{read.csv} and in \link[data.table]{fread}.
  }
  \item{rmc.dec}{
    Decimal used for numbers, same as skip argument in \link[utils]{read.csv} and in \link[data.table]{fread}.
  }
  \item{rmc.firstrow.acc}{
    First row (number) of the acceleration data.
  }
  \item{rmc.firstrow.header}{
    First row (number) of the header. Leave blank if the file does not have
    a header.
  }
  \item{rmc.header.length}{
    If file has header, specify header length (numeric).
  }
  \item{rmc.col.acc}{
    Vector with three column (numbers) in which the acceleration signals
    are stored
  }
  \item{rmc.col.temp}{
    Scalar with column (number) in which the temperature is stored.
    Leave in default setting if no temperature is avaible. The temperature
    will be used by \link{g.calibrate}.
  }
  \item{rmc.col.time}{
    Scalar with column (number) in which the timestamps are stored.
    Leave in default setting if timestamps are not stored. 
  }
  \item{rmc.unit.acc}{
    Character with unit of acceleration values: "g", "mg", or "bit"
  }
  \item{rmc.unit.temp}{
    Character with unit of temperature values: (K)elvin, (C)elsius, or (F)ahrenheit
  }
  \item{rmc.unit.time}{
    Character with unit of timestamps: "POSIX",
    "UNIXsec" (seconds since origin, see argument rmc.origin), "character", or
    "ActivPAL" (exotic timestamp format only used in the ActivPAL
    activity monitor).
    
  }
  \item{rmc.format.time}{
    Character string giving a date-time format as used by \link[base]{strptime}. Only used for rmc.unit.time: character and POSIX.
  }
  \item{rmc.bitrate}{
    Numeric: If unit of acceleration is a bit then provide bit rate, e.g. 12 bit.
  }
  \item{rmc.dynamic_range}{
    Numeric, if unit of acceleration is a bit then provide dynamic range deviation
    in g from zero, e.g. +/-6g would mean this argument needs to be 6. If you give this
    argument a character value the code will search the file header for elements with
    a name equal to the character value and use the corresponding numeric value
    next to it as dynamic range.
  }
  \item{rmc.unsignedbit}{
    Boolean, if unsignedbit = TRUE means that bits are only positive numbers.
   if unsignedbit = FALSE then bits are both positive and negative.
  }
  \item{rmc.origin}{
    Origin of time when unit of time is UNIXsec, e.g. 1970-1-1
  }
  \item{rmc.desiredtz}{
    Deprecated, please see \code{desiredtz}.
  }
  \item{rmc.configtz}{
    Deprecated, please see \code{configtz}.
  }
  \item{rmc.sf}{
    Sample rate in Hertz, if this is stored in the file header then that will be used
    instead.
  }
  \item{rmc.headername.sf}{
    If file has a header: Row name (character) under which the sample
    frequency can be found.
  }
  \item{rmc.headername.sn}{
    If file has a header: Row name (character) under which the
    serial number can be found.
  }
  \item{rmc.headername.recordingid}{
    If file has a header: Row name (character) under which the
    recording ID can be found.
  }
  
  \item{rmc.header.structure}{
    Character used to split the header name from the header
    value, e.g. ":" or " "
  }
  \item{rmc.check4timegaps}{
    Boolean to indicate whether gaps in time should be imputed with zeros.
    Some sensing equipment provides accelerometer with gaps in time. The rest of 
    GGIR is not designed for this, by setting this argument to TRUE the the gaps
    in time will be filled with zeros.
  }
  \item{rmc.col.wear}{
    If external wear detection outcome is stored as part of the data then this can be used by GGIR.
    This argument specifies the column in which the wear detection (Boolean) is stored.
  }
  \item{rmc.doresample}{
    Boolean to indicate whether to resample the data based on the available timestamps and extracted 
   sample rate from the file header
  }
  \item{rmc.scalefactor.acc}{
    Numeric value (default 1) to scale the acceleration signals via multiplication.
    For example, if data is provided in m/s2 then by setting this to 1/9.81
    we would derive gravitational units.
  }
  \item{interpolationType}{
    Integer to indicate type of interpolation to be used when resampling time series (mainly relevant for Axivity sensors), 1=linear, 2=nearest neighbour.
  }
  \item{PreviousLastValue}{
    Automatically identified last value in previous chunk of data read.
  }
    \item{PreviousLastTime}{
    Automatically identified last timestamp in previous chunk of data read.
  }
  \item{desiredtz}{
    Timezone in which device was worn.
  }
  \item{configtz}{
    Timezone in which device was configured. If equal to desiredtz you can leave
    this in its default value.
  }
  \item{header}{
    Header information that was extracted the previous time this file was read,
    to be re-used instead of being extracted again.
  }
}
\details{
  To use this function in the context of GGIR use all arguments from this function,
  except rmc.file, rmc.nrow, and  rmc.skip as input for function \link{GGIR} or \link{g.part1} and also specify argument rmc.noise, which is not part of this function but needed to tell GGIR what noise level to expect in the data. The rmc.noise is taken from the params_rawdata object if not explicitly specified by user.
}
\value{
  List with objects data holding the time series of acceleration with among others
  a column named "time" that holds the time expressed in seconds since 1-1-1970, and
  header if a header was present in the input file.
}
\examples{

  # create test files: No header, with temperature, with time
  N = 30
  sf = 30
  x = Sys.time()+((0:(N-1))/sf)
  timestamps = as.POSIXlt(x, origin="1970-1-1", tz = "Europe/London")
  mydata = data.frame(x = rnorm(N), time = timestamps, y = rnorm(N), z = rnorm(N),
            temp = rnorm(N) + 20)
  testfile = "testcsv1.csv"
  write.csv(mydata, file= testfile, row.names = FALSE)
  loadedData = read.myacc.csv(rmc.file=testfile, rmc.nrow=20, rmc.dec=".",
                      rmc.firstrow.acc = 1, rmc.firstrow.header=c(),
                      desiredtz = "",
                      rmc.col.acc = c(1,3,4), rmc.col.temp = 5, rmc.col.time=2,
                      rmc.unit.acc = "g", rmc.unit.temp = "C", rmc.origin = "1970-01-01")
  if (file.exists(testfile)) file.remove(testfile)
  
}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{appendRecords}
\alias{appendRecords}
\title{
  Append GGIR milestone data from neighbouring or overlapping recordings
}
\description{
  Append GGIR part 1 milestone data format from neighbouring or overlapping recording.
  When recordings overlap we use data from newest recordings. When time gap is larger
  than 2 days the recordings are not appended. Not intended for direct use by the user.
}
\usage{
   appendRecords(metadatadir, desiredtz = "", idloc = 1, maxRecordingInterval = NULL)
}
\arguments{
  \item{metadatadir}{
    See \link{g.part2}
  }
  \item{desiredtz}{
    See \link{GGIR}
  }
  \item{idloc}{
    See \link{GGIR}
  }
  \item{maxRecordingInterval}{
    See \link{GGIR}
  }
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.part5.savetimeseries}
\alias{g.part5.savetimeseries}
\title{
  Saves part 5 time series to csv files
}
\description{
  Not intended for direct use by GGIR users.
  Saves part 5 time series to csv files as part of \link{g.part5}.
}
\usage{
  g.part5.savetimeseries(ts, LEVELS, desiredtz, rawlevels_fname,
                            DaCleanFile = NULL,
                            includedaycrit.part5 = 2/3,
                            includenightcrit.part5 = 0,
                            ID = NULL,
                            params_output,
                            params_247 = NULL,
                            Lnames = NULL, timewindow = NULL,
                            filename = "")
}
\arguments{
  \item{ts}{
    Data.frame with time series
  }
  \item{LEVELS}{
    As produced as one of the objects in the output of \link{identify_levels}
  }
  \item{desiredtz}{
     See \link{GGIR}.
  }
  \item{rawlevels_fname}{
    Path to the file where the output (time series) should be stored
  }
  \item{DaCleanFile}{
    Content of data_cleaning_file as documented in \link{g.report.part5}.
    Only used in this function if save_ms5rawlevels is TRUE,  and it 
    only affects the time 
    series files stored.
  }
  \item{includedaycrit.part5}{
    See \link{GGIR}. Only used in this function if
    save_ms5rawlevels is TRUE,  and it only affects the time 
    series files stored.
  }
  \item{includenightcrit.part5}{
    See \link{GGIR}. Only used in this function if
    save_ms5rawlevels is TRUE, and it only affects the time 
    series files stored.
  }
  \item{ID}{
    If data_cleaning_file is used then this argument specifies
    which participant ID the data correspond with.
  }
  \item{params_output}{
    Parameters object, see \link{GGIR}
  }
  \item{params_247}{
    See \link{GGIR}
  }
  \item{Lnames}{
    Level names as passed on from \link{identify_levels}, these are the names
    corresponding the ID of the behavioural classes as stored in column class_id.
  }
  \item{timewindow}{
    See \link{GGIR}
  }
  \item{filename}{
    Character (default = "") indicating the name of the accelerometer data file 
    that was used as input. This name will be stored inside the time series output file.
  }
}
\value{
  Function does not provide output, it only prepare data for saving
  and saves it to a file. For documention on columns see main vignette.
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.part5}
\alias{g.part5}
\title{
  Merge output from physical activity and sleep analysis into one report
}
\description{
  Function to merge the output from \link{g.part2} and \link{g.part4} into one report enhanced with
  profiling of sleep and physical activity stratified across intensity levels and
  based on bouted periods as well as non-bouted periods.
}
\usage{
g.part5(datadir = c(), metadatadir = c(), f0 = c(), f1 = c(),
                   params_sleep = c(), params_metrics = c(),
                   params_247 = c(), params_phyact = c(), 
                   params_cleaning = c(), params_output = c(),
                   params_general = c(), verbose = TRUE, ...)
}
\arguments{
  \item{datadir}{
    Directory where the accelerometer files are stored, e.g. 
    "C:/mydata", or list of accelerometer filenames and directories, e.g. 
    c("C:/mydata/myfile1.bin", "C:/mydata/myfile2.bin").
  }
  \item{metadatadir}{
    Directory that holds a folder 'meta' and inside this a folder 'basic'
    which contains the milestone data produced by \link{g.part1}. The folderstructure
    is normally created by \link{g.part1} and \link{GGIR} will recognise what
    the value of metadatadir is.
  }
  \item{f0}{
    File index to start with (default = 1). Index refers to the filenames sorted
    in alphabetical order
  }
  \item{f1}{
    File index to finish with (defaults to number of files available, i.e., f1 = 0)
  }
  \item{params_sleep}{
   See details in \link{GGIR}.
  }
  \item{params_metrics}{
   See details in \link{GGIR}.
  }
  \item{params_247}{
   See details in \link{GGIR}.
  }
  \item{params_phyact}{
   See details in \link{GGIR}.
  }
  \item{params_cleaning}{
   See details in \link{GGIR}.
  }
  \item{params_output}{
   See details in \link{GGIR}.
  }
  \item{params_general}{
   See details in \link{GGIR}.
  }
  \item{verbose}{
    See details in \link{GGIR}.
  }
  \item{...}{
    To ensure compatibility with R scripts written for older GGIR versions,
    the user can also provide parameters listed in the params_ objects as direct argument.
  }
}
\value{
  The function does not produce values but generates an RData file
  in the milestone subfolder ms5.out which incudes a dataframe
  named \code{output}. This dataframe is used in g.report.part5 to create
  two reports one per day and one per person. See package vignette
  paragraph "Output part 5" for description of all the variables.
}
\examples{
  \dontrun{
    metadatadir = "C:/myfolder/meta"
    g.part5(metadatadir=metadatadir)
  }
}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.conv.actlog}
\alias{g.conv.actlog}
\title{
  Function to read activity log and make it useful for the rest of GGIR.
}
\description{
 Function to read activity log and convert it into data.frame
 that has for each ID and date a different qwindow vector.
}
\usage{
  g.conv.actlog(qwindow, qwindow_dateformat="\%d-\%m-\%Y", epochSize = 5)	
}
\arguments{
  \item{qwindow}{
    Path to csv file with activity log. Expected format of the activity diary is:
    First column headers followed by one row per recording, first column is recording ID,
    which needs to match with the ID GGIR extracts from the accelerometer file.
    Followed by date column in format "23-04-2017", where date format is specified by 
    argument qwindow_dateformat (below). Use the character combination date, Date or 
    DATE in the column name. This is followed by 
    one or multiple columns with start times for the activity types in that day format in 
    hours:minutes:seconds. The header of the column will be used as label for each activity
    type. Insert a new date column before continuing with activity types for next day.
    Leave missing values empty. If an activitylog is used then individuals who do 
    not appear in the activitylog will still be processed with value c(0,24).
    Dates with no activiy log data can be skipped, no need to have a column with the 
    date followed by a column with the next date. If times in the activitylog are not multiple 
    of the short window size (epoch length), the next epoch is considered (e.g., with epoch of 5 seconds, 
    8:00:02 will be redefined as 8:00:05 in the activity log).
  }
  \item{qwindow_dateformat}{
    Character specifying the date format used in the activity log.
  }
  \item{epochSize}{
    Short epoch size (first value of windowsizes in \link{g.getmeta}).
  }
}
\value{
  Data.frame with column ID, date and qwindow, where each
  qwindow value is a qwindow vector
}
\keyword{internal}
\author{
Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.getmeta}
\alias{g.getmeta}
\title{
  Function to extract meta-data (features) from data in accelerometer file
}
\description{
  Reads a accelerometer file in blocks, extracts various features and stores average  feature
  value per short or long epoch. Acceleration and angle metrics are stored at short
  epoch length. The non-wear indication score, the clipping score, temperature
  (if available), light (if available), and Euclidean norm are stored at long epoch
  length. The function has been designed and thoroughly tested with accelerometer files
  from GENEA and GENEActiv bin files. Further, the function should be able to cope with ActiGraph gt3x and
  csv files, Axivity cwa and csv files, Movisens bin files, and ad-hoc csv files read through the
  \link{read.myacc.csv} function.
}
\usage{
  g.getmeta(datafile, params_metrics = c(), params_rawdata = c(),
                     params_general = c(), params_cleaning = c(), daylimit = FALSE,
                     offset = c(0, 0, 0), scale = c(1, 1, 1), tempoffset = c(0, 0, 0),
                     meantempcal = c(), myfun = c(), inspectfileobject = c(), 
                     verbose = TRUE, ...)
}
\arguments{
  \item{datafile}{
    name of accelerometer file
  }
  \item{params_metrics}{
    See details in \link{GGIR}.
  }
  \item{params_rawdata}{
    See details in \link{GGIR}.
  }
  \item{params_general}{
    See details in \link{GGIR}.
  }
  \item{params_cleaning}{
    See details in \link{GGIR}.
  }
  \item{daylimit}{
    number of days to limit (roughly), if set to FALSE no daylimit
    will be applied
  }
  \item{offset}{
    offset correction value per axis, usage:
    value = scale(value,center = -offset, scale = 1/scale)
  }
  \item{scale}{
    scaling correction value per axis, usage:
    value = scale(value,center = -offset, scale = 1/scale)
  }
  \item{tempoffset}{
    temperature offset correction value per axis, usage:
    value = scale(value,center = -offset, scale = 1/scale)
    + scale(temperature, center = rep(averagetemperate,3), scale = 1/tempoffset)
  }
  \item{meantempcal}{
    mean temperature corresponding to the data as used for
    autocalibration. If autocalibration is not done or if temperature was not
    available then leave blank (default)
  }
  \item{myfun}{
    External function object to be applied to raw data.
    See details \link{applyExtFunction}.
  }
  \item{inspectfileobject}{
    Output from the function \link{g.inspectfile}.
  }
  \item{verbose}{
    Boolean (default = TRUE). to indicate whether console message should be printed.
    Note that warnings and error are always printed and can be suppressed with
    suppressWarning() or suppressMessages().
  }
  \item{...}{
    Any argument used in the previous version of g.getmeta, which will now
    be used to overrule the arguments specified with the parameter objects.
  }
}
\value{
  \item{metalong}{dataframe with long epoch meta-data: EN, non-wear score,
  clipping score, temperature}
  \item{metashort}{dataframe with short epoch meta-data: timestamp and metric}
  \item{tooshort}{indicator of whether file was too short for processing (TRUE or FALSE)}
  \item{corrupt}{indicator of whether file was considered corrupt (TRUE or FALSE)}
}
\examples{
  \dontrun{
    datafile = "C:/myfolder/testfile.bin"

    #Extract meta-data:
    M = g.getmeta(datafile)

    #Inspect first couple of rows of long epoch length meta data:
    print(M$metalong[1:5,])

    #Inspect first couple of rows of short epoch length meta data:
    print(M$metalong[1:5,])
  }
}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}
\references{
  \itemize{
    \item van Hees VT, Gorzelniak L, Dean Leon EC, Eder M, Pias M, et al. (2013) Separating
    Movement and Gravity Components in an Acceleration Signal and Implications for the
    Assessment of Human Daily Physical Activity. PLoS ONE 8(4): e61691.
    doi:10.1371/journal.pone.0061691
    \item Aittasalo M, Vaha-Ypya H, Vasankari T, Husu P, Jussila AM, and Sievanen H. Mean
    amplitude deviation calculated from raw acceleration data: a novel method for
    classifying the intensity of adolescents physical activity irrespective of accelerometer
    brand. BMC Sports Science, Medicine and Rehabilitation (2015).
  }
}

\name{createConfigFile}
\alias{createConfigFile}
\title{
  Creates Config File based on variables in function GGIR environment
}
\description{
  Only used inside \link{GGIR}. Not intended for direct use by user.
}
\usage{
  createConfigFile(config.parameters = c(), GGIRversion = "")	
}
\arguments{
  \item{config.parameters}{
   List with all arguments used in \link{GGIR}.
  }
  \item{GGIRversion}{
   GGIR version mumber to be incorported in the ConfigFile.
  }
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.create.sp.mat}
\alias{g.create.sp.mat}
\title{
Converts sleep period information. Not intended for direct use
}
\description{
Function to convert data into sleep period matrix part of g.part4.R.
Not intended for direct use by package user
}
\usage{
g.create.sp.mat(nsp,spo,sleepdet.t,daysleep=FALSE)

}
\arguments{
  \item{nsp}{
  Integer indicating the number of sleep periods
  } 
 \item{spo}{
  Empty matrix with overview of sleep periods, 5 columns and
  as along as nps
  }
 \item{sleepdet.t}{
   Part of detected sleep from g.sib.det for one night and one 
   sleep definition
  }
 \item{daysleep}{
  Boolean to indicator whether this person woke up
  after noon (daysleeper)
  }
 
}
\value{
  \itemize{
  \item spo matrix with start and end of each sleep period
  \item calendardate date corresponding to the day on which the night started
  \item item wdayname weekdayname
  }
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{getPart1BasicInfo}
\alias{getPart1BasicInfo}
\title{
  Get basic info form the part 1 milestone file
}
\description{
  Internal function to get basic info form the part 1 milestone file
}
\usage{
  getPart1BasicInfo(fn, idloc, tz)
}
\arguments{
  \item{fn}{
    Character with full file path to RData file as stored by \link{g.part1}
  }
  \item{idloc}{
    See \link{GGIR}
  }
  \item{tz}{
    As passed by \link{GGIR} and equal to parameter desiredtz as documented in \link{GGIR}
  }
}
\value{
  Data.frame with ID, start and end time of the recording,
  filename and brand name.
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.part5.addsib}
\alias{g.part5.addsib}
\title{
  Adds the sustained inactivity bout to the ts series.
}
\description{
  Not intended for direct use by GGIR users.
  Adds the sustained inactivity bout to the ts series
  as part of \link{g.part5}.
}
\usage{
  g.part5.addsib(ts, epochSize, part3_output, desiredtz,
              sibDefinition, nightsi)
}
\arguments{
  \item{ts}{
    Data.frame object as passed on from \link{g.part5}
  }
  \item{epochSize}{
    Short epoch size in seconds
  }
  \item{part3_output}{
    Segment of part 3 output relevant for current sleep definition
  }
  \item{desiredtz}{
    see \link{GGIR}
  }
  \item{sibDefinition}{
    Character to indicate definition of sib (sustained inactivity bout)
  }
  \item{nightsi}{
    Vector with indices for the midnights
  }
}
\value{
  Data.frame ts
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.IVIS}
\alias{g.IVIS}
\title{
  Calculates IV and IS
}
\description{
  To extract interdaily stability and interdaily variability as originally proposed by
  van Someren.
}
\usage{
  g.IVIS(Xi, epochSize = 60, threshold = NULL)
}
\arguments{
  \item{Xi}{
    Vector with acceleration values, e.g. ENMO metric.
  }
  \item{epochSize}{
    Epoch size of the values in Xi expressed in seconds.
  }
  \item{threshold}{
    Acceleration threshold to distinguish inactive from active
  }
}
\value{
  \item{InterdailyStability}{}
  \item{IntradailyVariability}{}
}
\examples{
  Xi = abs(rnorm(n = 10000,mean = 0.2))
  IVISvariables = g.IVIS(Xi=Xi)
}
\keyword{internal}
\author{
  Ian Meneghel Danilevicz <ian.meneghel-danilevicz@inserm.fr>
  Vincent T van Hees <v.vanhees@accelting.com>
}
\references{
  \itemize{
    \item Eus J. W. Van Someren, Dick F. Swaab, Christopher C. Colenda, 
    Wayne Cohen, W. Vaughn McCall & Peter B. Rosenquist. Bright Light Therapy: 
    Improved Sensitivity to Its Effects on Rest-Activity Rhythms in Alzheimer 
    Patients by Application of Nonparametric Methods  Chronobiology
    International. 1999. Volume 16, issue 4.
  }
}

\name{data.ts}
\alias{data.ts}
\docType{data}
\title{
  Time series data.frame stored by part 5
}
\description{
  data.ts is example of the data.frame stored \link{g.part5}
}
\usage{data(data.ts)}
\format{
  The format is:
 chr "data.ts"
}
\source{
  The data was collected on one individual for testing purposes and matches the data
  in object data.metalong
}
\examples{
  data(data.ts)
}
\keyword{datasets}

\name{g.extractheadervars}
\alias{g.extractheadervars}
\title{
Extracts header variables from header object
}
\description{
Function is not intended for direct interaction by package end user
}
\usage{
g.extractheadervars(I)
}
\arguments{
  \item{I}{
  Object produced by \link{g.inspectfile}
  }
}
\value{
\itemize{
\item ID = participant identifier
\item iid = investigator identifier
\item HN = handedness
\item BodyLocation = Attachement location of the sensor
\item SX = sex
\item deviceSerialNumber = serial number
}
}
\examples{
data(data.inspectfile)
headervars = g.extractheadervars(I=data.inspectfile)
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{ShellDoc2Vignette}
\alias{ShellDoc2Vignette}
\title{
  Extract Definition from Shell Documentation
}
\description{
  Function extracts the documentation for a given GGIR argument as provided in the \link{GGIR} documentation.
  Function not designed for direct use by package user.
}
\usage{
  ShellDoc2Vignette(argument = "mode")
}
\arguments{
  \item{argument}{
    Character (default = "mode"). 
    Name of the argument to extract the definition.
  }
}
\value{
  Character object with the definition of the argument.
}
\keyword{internal}
\author{
  Jairo Hidalgo Migueles <jairo.hidalgo.migueles@gmail.com>
}

\name{g.part5.lux_persegment}
\alias{g.part5.lux_persegment}
\title{
  Extract key lux variables per segment of the data.
}
\description{
  Extracts per segment of the day: mean lux, time above 1000 lux, time awake,
  and time LUX imputed. Function not intended
  for direct use by package user.
}
\usage{
  g.part5.lux_persegment(ts, sse, LUX_day_segments, epochSize, desiredtz = "")
}
\arguments{
  \item{ts}{
    Data.frame with time series
  }
  \item{sse}{
    Indices corresponding to the current time window (e.g. MM or WW)
  }
  \item{LUX_day_segments}{
    See \link{GGIR}
  }
  \item{epochSize}{
    Numeric epoch size in seconds
  }
  \item{desiredtz}{
    See \link{GGIR}
  }
}
\value{
 List with values (vector) of the derived variables and corresponding names (vector).
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{convertEpochData}
\alias{convertEpochData}
\title{
  convert external Epoch data to GGIR part 1 milestone data format
}
\description{
  convert externally created Epoch data to GGIR part 1 milestone data format. Function not
  intended for direct use by user. The aim is to allow for using GGIR on top of
  extrnally derived epoch data. See argument dataFormat in the \link{GGIR} for details on
  how to use this functionality.
}
\usage{
   convertEpochData(datadir = c(), metadatadir = c(),
                    params_general = c(), f0 = c(), f1 = c(), verbose = TRUE)
}
\arguments{
  \item{datadir}{
    See \link{GGIR}
  }
  \item{metadatadir}{
    See \link{GGIR}
  }
  \item{params_general}{
    Parameters object see \link{GGIR}
  }
  \item{f0}{
    File index to start with (default = 1). Index refers to the filenames sorted
    in alphabetical order
  }
  \item{f1}{
    File index to finish with (defaults to number of files available, i.e., f1 = 0)
  }
  \item{verbose}{
    See \link{GGIR}
  }
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{data.metalong}
\alias{data.metalong}
\docType{data}
\title{
  Metalong object as part of part 1 milestone data
}
\description{
  data.metalong is example of the metalong data.frame stored \link{g.part1}
}
\usage{data(data.metalong)}
\format{
  The format is:
 chr "data.metalong"
}
\source{
  The data was collected on one individual for testing purposes
}
\examples{
  data(data.metalong)
}
\keyword{datasets}

\name{g.analyse.perfile}
\alias{g.analyse.perfile}
\title{
  Function supports \link{g.analyse}. Not intended for direct use by user.
}
\description{
  Generates recording specific analyses and fills corresponding
  output matrix, \link{g.analyse}.
}
\usage{
g.analyse.perfile(I, C, metrics_nav,
                 AveAccAve24hr, doquan, doiglevels, tooshort,
                 params_247, params_cleaning, params_general,
                 output_avday, output_perday,
                 dataqual_summary, file_summary)
}

\arguments{
  \item{I}{output \link{g.inspectfile}}
  \item{C}{output \link{g.calibrate}}
  \item{metrics_nav}{List with three objects to help navigate the acceleration metrics}
  \item{AveAccAve24hr}{Average acceleration in an average 24 hour cycle}
  \item{doquan}{Boolean whether quantile analysis should be done}
  \item{doiglevels}{Boolean to indicate whether iglevels should be calculated}
  \item{tooshort}{0 (file not too short) or 1 (file too short)}
  \item{params_247}{see \link{GGIR}}
  \item{params_cleaning}{see \link{GGIR}}
  \item{params_general}{see \link{GGIR}}
  \item{output_avday}{Output from g.analyse.avday}
  \item{output_perday}{Output from g.analyse.perday}
  \item{dataqual_summary}{Data.frame with data quality summary indicators
  produced in \link{g.analyse}}
}
\value{
  \item{\code{filesummary}}{summary for the file that was analysed}
  \item{\code{daysummary}}{Summary per day for the file that was analysed}
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.part3_alignIndexVectors}
\alias{g.part3_alignIndexVectors}
\title{
  Aligns index vectors used inside g.part3_correct_guider
}
\description{
  Aligns index vectors used inside g.part3_correct_guider
}
\usage{
   g.part3_alignIndexVectors(x, y, a, b, N) 
}
\arguments{
  \item{x}{
    Index vector for minimum
  }
  \item{y}{
    Index vector for maximum
  }
  \item{a}{
    Index vector for lower medium
  }
  \item{b}{
    Index vector for upper medium
  }
  \item{N}{
    The maximum index, which equals the length of the time series
  }
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{load_params}
\alias{load_params}
\title{
  Load default parameters
}
\description{
  Loads default paramter values
  Not intended for direct use by GGIR users.
}
\usage{
  load_params(topic = c("sleep", "metrics", "rawdata", "247",
                        "phyact", "cleaning", "output", "general"))
}
\arguments{
  \item{topic}{
    Character vector with parameter groups to be loaded.
  }
}
\value{
  Lists of parameter objects 
}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.part5.handle_lux_extremes}
\alias{g.part5.handle_lux_extremes}
\title{
  Check lux values for extremes and imputes or removes them
}
\description{
  Extreme values are imputed by mean of neightbours if they occur isolated or
  in a sequence of two, and removed if they occure in a sequence of 3 or longer.
}
\usage{
  g.part5.handle_lux_extremes(lux)
}
\arguments{
  \item{lux}{
  Vector with lux values
  }
}
\value{
  List of imputed lux values and a vector with matching length named
  correction_log indicating which timestamps where imputed (value=1),
  replaced by NA (value=2) or untouched (value=0).
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{detect_nonwear_clipping}
\alias{detect_nonwear_clipping}
\title{
Detect non-wear and clipping time in the raw accelerometer data
}
\description{
Detects periods in which the accelerometer has not been worn or the
accelerometer signal was stuck very close to the dynamic range limit
of the accelerometer.
}
\usage{
detect_nonwear_clipping(data = c(), windowsizes = c(5, 900, 3600), sf = 100,
                        clipthres = 7.5, sdcriter = 0.013, racriter = 0.05,
                        nonwear_approach = "2013",
                        params_rawdata = c())
}
\arguments{
  \item{data}{
  Matrix with raw accelerometer data for X, Y, and Z axes.
}
 \item{windowsizes}{
  Numeric vector of length three, with short, long epoch and window size in seconds.
}
 \item{sf}{
  Sample frequency in Hertz.
}
 \item{clipthres}{
  Threschold to detect clipping in _g_ units. Usually 0.5 _g_ below the 
  dynamic range of the accelerometer.
}
 \item{sdcriter}{
  Criteria to define non-wear time, defined as the estimated noise
  measured in the raw accelerometer data.
}
 \item{racriter}{
  Absolute criteria below which the absolute range of the accelerations
  should be to define non-wear time.
}
 \item{nonwear_approach}{
  Whether to use the traditional version of the non-wear detection algorithm
  (nonwear_approach = "2013", default) or the new version (nonwear_approach = "2023"). 
  The 2013 version would use the longsize window (windowsizes[3], one hour as default)
  to check the conditions for nonwear identification and would flag as nonwear the
  mediumsize window (windowsizes[2], 15 min as default) in the middle. The 2023 
  version differs in which it would flag as nonwear the full longsize window. For the 2013
  method the longsize window is centered in the centre of the mediumsize window, while in 
  the 2023 method the longsizewindow is aligned with its left edge to the left edge of the 
  mediumsize window.
 }
 \item{params_rawdata}{
  See details in \link{GGIR}.
}
}
\value{
  List containing the next numeric vectors: 
  \itemize{
    \item NWav (non-wear score), from 1 to 3 indicating the number of axes that met the non wear criteria. 
    \item CWav (clipping score), binary, 0-1 indicating non-clipping and clipping, respectively.
    \item nmin is the minimum numebr of windows in this block of data.
  }
  Each number in these vectors represent a long epoch duration (i.e., ws2, 900 seconds by default).
}
\examples{
\dontrun{
    detect_nonwear_clipping(data = data, windowsizes = c(900, 3600), sf = sf,
                            clipthres = clipthres, sdcriter = sdcriter, 
                            racriter = racriter, nonwear_approach = "old")
  }
}
\author{
Vincent T van Hees <v.vanhees@accelting.com>
Jairo Hidalgo Migueles <j.h.migueles@accelting.com>
}
\keyword{internal}
\references{
\itemize{
    \item van Hees  et al. 2011, doi: 10.1371/journal.pone.0022922.
    \item van Hees et al. 2013, doi: 10.1371/journal.pone.0061691 (in supplementary material).
  }
}

\name{g.shell.GGIR}
\alias{g.shell.GGIR}
\title{
  Wrapper function around function GGIR
}
\description{
  This function used to be the central function in the package,
  but has been renamed GGIR. You can still use function call g.shell.GGIR
  but all arguments will be passed on to function GGIR. We have done this to preserve
  consistency with older use cases of the GGIR package. All documentation can
  now be found in \link{GGIR}.

}
\usage{
  g.shell.GGIR(...)
}
\arguments{
  \item{...}{
    Any of the parameters used by \link{GGIR}.
  }
}
\value{
  The function provides no values, it only ensures that other functions are called
  and that their output is stored. See \link{GGIR}.
}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.analyse.avy}
\alias{g.analyse.avday}
\title{
  Function supports \link{g.analyse}. Not intended for direct use by user.
}
\description{
  Generatess average day analyses and fills corresponding output
  matrix, \link{g.analyse}.
}
\usage{
  g.analyse.avday(doquan, averageday, M, IMP, t_TWDI, quantiletype,
                   ws3, doiglevels, firstmidnighti, ws2, midnightsi,
                   params_247 = c(), qcheck = c(), acc.metric = c(),
                   params_phyact = NULL, ...)
}
\arguments{
  \item{doquan}{Boolean whether quantile analysis should be done}
  \item{averageday}{ As produced by \link{g.impute}}
  \item{M}{ As produced by \link{g.getmeta}}
  \item{IMP}{ As produced by \link{g.impute}}
  \item{t_TWDI}{ Same as qwindow as described in \link{g.analyse}}
  \item{quantiletype}{see \link{g.analyse}}
  \item{ws3}{ Epoch size in seconds}
  \item{doiglevels}{Boolean to indicate whether iglevels should be calculated}
  \item{firstmidnighti}{see \link{g.detecmidnight}}
  \item{ws2}{see \link{g.weardec}}
  \item{midnightsi}{see \link{g.detecmidnight}}
  \item{params_247}{
    See \link{GGIR}
  }
  \item{qcheck}{
    Vector with indicators of when data is valid (value=0) or invalid (value=1).
  }
  \item{acc.metric}{
    Character, see \link{GGIR}. Here, it is used to decided which acceleration metric
    to use for IVIS and cosinor analyses.
  }
  \item{params_phyact}{
    See \link{GGIR}
  }
  \item{...}{
   Any argument used in the previous version of g.analyse.avday, which will now
   be used to overrule the arguments specified with the parameter objects.
  }
}
\value{
  \item{\code{igfullr_names}}{Intensity gradient variable names}
  \item{\code{igfullr}}{Intensity gradient values}
  \item{\code{QUAN}}{Quantiles}
  \item{\code{qlevels_names}}{Quantile level names}
  \item{\code{ML5AD}}{}
  \item{\code{ML5AD_names}}{}
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.sib.det}
\alias{g.sib.det}
\title{
  sustiained inactivty bouts detection
}
\description{
  Detects sustiained inactivty bouts. Function not intended
  for direct use by package user
}
\usage{
  g.sib.det(M, IMP, I, twd = c(-12, 12),
             acc.metric = "ENMO", desiredtz = "",
             myfun=c(), sensor.location = "wrist", params_sleep = c(), zc.scale = 1, ...)
}
\arguments{
  \item{M}{
    Object produced by \link{g.getmeta}
  }
  \item{IMP}{
    Object produced by \link{g.impute}
  }
  \item{I}{
    Object produced by \link{g.inspectfile}
  }
  \item{twd}{
    Vector of length 2, indicating the time window to consider
    as hours relative to midnight.
  }
    \item{acc.metric}{
    Which one of the metrics do you want to consider to analyze L5. 
    The metric of interest need to be calculated in
    M (see \link{g.part1})
  }
  \item{desiredtz}{
    See \link{g.part3}
  }
  \item{myfun}{
    External function object to be applied to raw data.
    See details \link{applyExtFunction}.
  }
  \item{sensor.location}{
    Character to indicate sensor location, default is wrist.
    If it is hip HDCZA algorithm also requires longitudinal axis of sensor to be
    between -45 and +45 degrees.
  }
   \item{params_sleep}{
    See \link{g.part3}
  }
  \item{zc.scale}{
    Used for zero-crossing counts only. Scaling factor to be applied after
    counts are calculated (GGIR part 3). See \link{GGIR}.
  }
  \item{...}{
     Any argument used in the previous version of g.sib.det, which will now
     be used to overrule the arguments specified with the parameter objects.
  }
}
\value{
  \itemize{
    \item output = Dataframe for every epoch a classification
    \item detection.failed = Boolean whether detection failed
    \item L5list = L5 for every day (defined from noon to noon)
  }
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{correctOlderMilestoneData}
\alias{correctOlderMilestoneData}
\title{
Corrects milestone data from g.part1 generated in older GGIR versions
}
\description{
Some older versions of GGIR stored the milestone data in part 1 as factor.
This function identifies when that occurs and convert the affected columns
to the appropriate class (e.g., numeric).}
\usage{
  correctOlderMilestoneData(x)
}
\arguments{
  \item{x}{
    Data frame with metashort or metalong data as generated in \link{g.part1}
  }
}
\value{
 Data frame with the class fixed in the appropriate columns (i.e., light and 
 temperature columns)
}
\keyword{internal}
\examples{
  \dontrun{
    correctOlderMilestoneData(x)
  }
}

\name{is_this_a_dst_night}
\alias{is_this_a_dst_night}
\title{
 Check whether the night starting on a calendar date has DST.
}
\description{
Tests whether the night that follows the input calendar date is a
night with day saving time (DST) and on what hour the time moved.
}
\usage{
  is_this_a_dst_night(calendar_date=c(),tz="Europe/London")
}
\arguments{
  \item{calendar_date}{
    Character in the format dd/mm/yyyy 
  }
  \item{tz}{
    Time zone in "Europe/London" format.
  }
}
\value{
  \item{dst_night_or_not}{If value=0 no DST, if value=1 time moved forward,
  if value=-1 time moved forward}
  \item{dsthour}{Either the double hour or the hour that was skipped, 
  this differs between countries}
}
\keyword{internal}
\examples{
  test4dst = is_this_a_dst_night("23/03/2014",tz="Europe/London")
}

\name{cosinor_IS_IV_Analyses}
\alias{cosinor_IS_IV_Analyses}
\title{
  Apply cosinor anlaysis and extended cosinor analysis
}
\description{
  Applies cosinor anlaysis from the ActCR package to the time series, as well
  as IV, IS and phi estimates.
}
\usage{
  cosinor_IS_IV_Analyses(Xi, epochsize = 60, timeOffsetHours = 0, threshold = NULL)
}
\arguments{
  \item{Xi}{
    Vector with time series of movement indicators if the maximum < 8 and mean < 1
    then input is assumed to be in g-units and is multiplied by 1000.
  }
  \item{epochsize}{
    Numeric epochsize in seconds
  }
  \item{timeOffsetHours}{
    Numeric time in hours relative to next midnight
  }
  \item{threshold}{
    Numeric value to use as threshold to distinguish inactivity from active behaviour
    for the IV and IS analysis. GGIR uses parameter threshold.lig to set this threshold.
  }
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{extractID}
\alias{extractID}
\title{
  Extract ID from file header object
}
\description{
  Extract ID from file header object.
}
\usage{
   extractID(hvars, idloc, fname)
}
\arguments{
  \item{hvars}{
    Object extracted with \link{g.inspectfile}
  }
  \item{idloc}{
    See \link{GGIR}
  }
  \item{fname}{
    File (base)name.
  }
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.imputeTimegaps}
\alias{g.imputeTimegaps}
\title{
  Impute gaps in three axis raw accelerometer data
}
\description{
  Removes all sample with a zero in each of the three axes, and then (as default)
  imputes time  gaps by the last recorded value per axis normalised to 1 _g_
}
\usage{
  g.imputeTimegaps(x, sf, k = 0.25, impute = TRUE, 
                   PreviousLastValue = c(0,0,1), 
                   PreviousLastTime = NULL, epochsize = NULL)
}
\arguments{
  \item{x}{
    Data.frame with raw accelerometer data, and a timestamp column with
    millisecond resolution.
  }
  \item{sf}{
    Sample frequency in Hertz
  }
   \item{k}{
    Minimum time gap length to be imputed
  }
   \item{impute}{
    Boolean to indicate whether the time gaps identified should be imputed
  }
  \item{PreviousLastValue}{
    Automatically identified last value in previous chunk of data read.
  }
    \item{PreviousLastTime}{
    Automatically identified last timestamp in previous chunk of data read.
  }
  \item{epochsize}{
    Numeric vector of length two, with short and long epoch sizes.
  }
}
\value{
  List including:
  - x, data.frame based on input x with timegaps imputed (as default) or with 
  recordings with 0 values in the three axes removed (if impute = FALSE)
  - QClog, data.frame with information on the number of time gaps found and the 
  total time imputed in minutes
}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{check_myfun}
\alias{check_myfun}
\title{
  Checks myfun object before it is passed to applyExtfunction
}
\description{
  Checks that object myfun is a list and check the elements of the
  list for: that element names are as expected, that value of each
  element is of the expected type and length.
}
\usage{
  check_myfun(myfun, windowsizes)
  
}
\arguments{
  \item{myfun}{
    See \link{applyExtFunction}
  }
  \item{windowsizes}{
    See \link{g.getmeta}).
  }
}
\value{
  0 if all checkes passed, 1 if one or more checks did not pass. Error
  message are printed to the console with feedback on which checks did 
  not pass.
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.part5_initialise_ts}
\alias{g.part5_initialise_ts}
\title{
  Initialise time series data from for part 5
}
\description{
  Initialise time series dataframe ts, part of \link{g.part5}.
}
\usage{
  g.part5_initialise_ts(IMP, M, params_247, params_general, longitudinal_axis = c())
}
\arguments{
  \item{IMP}{
    Object derived from \link{g.part2}
  }
  \item{M}{
    Object derived from \link{g.part1}.
  }
  \item{params_247}{
    See \link{GGIR}
  }
  \item{params_general}{
    See \link{GGIR} 
  }
  \item{longitudinal_axis}{
    As passed on from g.part5, which could be specified by user or estimated
    from hip data in part 2. 
  }
}
\value{
  Data.frame ts
}
\keyword{internal}

\name{applyExtFunction}
\alias{applyExtFunction}
\title{
  Apply external function to acceleration data.
}
\description{
  Applies external function to the raw acceleration data within GGIR.
  This makes it easier for new algorithms developed to be pilotted
  on accelerometer data while taking advantage of the existing comprehensive GGIR
  data management and analysis infrastructure.
  This function is not for direct interaction by user, please supply object 
  \code{myfun} to \link{GGIR} or \link{g.part1}. Object \code{myfun}
  is a list as detailed below.
}
\usage{
  applyExtFunction(data, myfun, sf, ws3, interpolationType=1)
  
}
\arguments{
  \item{data}{
    Data data.frame as present internally in \link{g.getmeta}. It has at least
    four columns of which the first is the timestamp followed by the x, y,
    and z acceleration.
  }
  \item{myfun}{
    See details, in short: myfun is a list object that holds the external function
    to be applied to the data and various parameters to aid in the process.
  }
  \item{sf}{
    Sample frequency (Hertz) of the data object
  }
  \item{ws3}{
    Short epoch size (first value of windowsizes in \link{g.getmeta}).
  }
  \item{interpolationType}{
    Integer to indicate type of interpolation to be used when resampling time series (mainly relevant for Axivity sensors), 1=linear, 2=nearest neighbour.
  }
}
\value{
  The output of the external algorithm aggregated or repeated to fit the
  short epoch length of GGIR. Therefore, the short epoch length of GGIR
  should be a multitude of the resolution of the external function output,
  or visa versa.
}
\details{
    See package vignette for detailed tutorial with examples
    on how to use the function embedding:
    https://cran.r-project.org/web/package=GGIR/vignettes/applyExtFunction.pdf
    Function applyExtFunction
    is typically not used by the GGIR user directly.
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{part6AlignIndividuals}
\alias{part6AlignIndividuals}
\title{part6AlignIndividuals}
\description{
  Align individual time series per household where households are identified
  by the character or number string between the first and second '-' in the filename.
}
\usage{
  part6AlignIndividuals(GGIR_ts_dir = NULL, outputdir = NULL,
                      path_ggirms = NULL, desiredtz = "", verbose = TRUE)
}
\arguments{
  \item{GGIR_ts_dir}{
    Character, path to time series directory in the GGIR output
  }
  \item{outputdir}{
    Directory where you would like to store the output
  }
  \item{path_ggirms}{
    path to GGIR created folder named meta, with the milestone data files
  }
  \item{desiredtz}{
    Character, specifying the timezone database name of the timezone the 
    data was collected in.
  }
  \item{verbose}{
    See details in \link{GGIR}.
  }
}
\value{
  no object is returned, only files are created in the output directory
}

\name{g.part4}
\alias{g.part4}
\title{
  Labels detected sustained inactivity periods by g.part3 as either
  part of the Sleep Period Time window or not
}
\description{
  Combines output from \link{g.part3} and guider information to estimate
  sleep variables. See vignette paragraph "Sleep and full day
  time-use analysis in GGIR" for an elaborate descript of the sleep detection.
}
\usage{
 g.part4(datadir = c(), metadatadir = c(), f0 = f0, f1 = f1, params_sleep = c(), 
    params_metrics = c(),  params_cleaning = c(), params_output = c(),
                   params_general = c(), verbose = TRUE, ...)
}
\arguments{
  \item{datadir}{
     Directory where the accelerometer files are stored, e.g. 
    "C:/mydata", or list of accelerometer filenames and directories, e.g. 
    c("C:/mydata/myfile1.bin", "C:/mydata/myfile2.bin").
  }
  \item{metadatadir}{
    Directory that holds a folder 'meta' and inside this a folder 'basic'
    which contains the milestone data produced by \link{g.part1}. The folderstructure
    is normally created by \link{g.part1} and \link{GGIR} will recognise what
    the value of metadatadir is.
  }
  \item{f0}{
    File index to start with (default = 1). Index refers to the filenames sorted
    in alphabetical order
  }
  \item{f1}{
    File index to finish with (defaults to number of files available, i.e., f1 = 0)
  }
  \item{params_sleep}{
    List of parameters used for sleep analysis (GGIR part 3, 4, and 5): 
    see documentation \link{g.part3}.
  }
  \item{params_metrics}{
    List of parameters used for metrics extraction (GGIR part 1): 
    see documentation \link{g.part1}.
  }
  \item{params_cleaning}{
   See details in \link{GGIR}.
  }
  \item{params_output}{
   See details in \link{GGIR}.
  }
  \item{params_general}{
   See details in \link{GGIR}.
  }
  \item{verbose}{
    See details in \link{GGIR}.
  }
  \item{...}{
    To ensure compatibility with R scripts written for older GGIR versions,
    the user can also provide parameters listed in the params_ objects as direct argument.
  }
}
\value{
  The function does not produce values but generates an RData file
  in the milestone subfolder ms4.out which incudes a dataframe
  named \code{nightsummary}. This dataframe is used in g.report.part4 to create
  two reports one per night and one per person. See package vignette 
  paragraph "Output part 4" for description of all the variables.
}

\examples{
  \dontrun{
    metadatadir = "C:/myfolder/meta" # assumes that there is a subfolder in
    # metadatadir named 'ms3.out' containing the output from g.part3
    g.part4(metadatadir=metadatadir)
  }
}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}
\references{
  \itemize{
    \item van Hees VT, Sabia S, et al. (2018) AEstimating sleep parameters
    using an accelerometer without sleep diary, Scientific Reports.
    \item van Hees VT, Sabia S, et al. (2015) A novel, open access method
    to assess sleep duration using a wrist-worn accelerometer, PLoS ONE.
  }
}

\name{data.calibrate}
\alias{data.calibrate}
\docType{data}
\title{
Example output from g.calibrate
}
\description{
data.calibrate is example output from \link{g.calibrate}
}
\usage{data(data.calibrate)}
\format{
  The format is:
 chr "data.calibrate"
}
\source{
The data was collected on one individual for testing purposes
}
\examples{
data(data.calibrate)
}
\keyword{datasets}

\name{g.part3_correct_guider}
\alias{g.part3_correct_guider}
\title{
  Correct guider estimates
}
\description{
  After guider has been derived for all nights in the recording
  this function is optionally applied to correct individual nights based
  on knowledge about other nights in the recording.
}
\usage{
   g.part3_correct_guider(SLE, desiredtz, epochSize,
                    params_sleep)
}
\arguments{
  \item{SLE}{
    Object extracted with \link{g.sib.det}
  }
  \item{desiredtz}{
    See \link{GGIR}
  }
  \item{epochSize}{
    First value of parameter \code{windowsizes} as discussed in See \link{GGIR}.
  }
  \item{params_sleep}{
    Complete params_sleep object, see \link{GGIR}
  }
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.impute}
\alias{g.impute}
\title{
  Function to identify invalid periods in the meta-data as generated by \link{g.getmeta}
  and to impute these invalid periods with the average of similar timepoints on other
  days of the measurement
}
\description{
  Functions takes the output from \link{g.getmeta} and information about the study
  protocol to label impute invalid time segments in the data.
}
\usage{
  g.impute(M, I, params_cleaning = c(), desiredtz = "", dayborder = 0,
            TimeSegments2Zero = c(), acc.metric = "ENMO", ID, 
            qwindowImp = c(), ...)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{M}{
    output from \link{g.getmeta}
  }
  \item{I}{
    output from \link{g.inspectfile}
  }
  \item{params_cleaning}{
    See \link{g.part1}
  }
  \item{desiredtz}{
    See \link{g.part1}
  }
  \item{dayborder}{
    See \link{g.part1}
  }
  \item{TimeSegments2Zero}{
    Optional data.frame to specify which time segments need to be ignored for the imputation,
    and acceleration metrics to be imputed by zeros. The data.frame is expected
    to contain two columns named windowstart and windowend, with the start- and end
    time of the time segment in POSIXlt class.
  }
  \item{acc.metric}{
    See \link{GGIR}
  }
  \item{ID}{
    ID extracted in g.part2.
  }
  \item{qwindowImp}{
    qwindow object as documented in See \link{GGIR}. If qwindow is based on a diary 
    then the qwindow object is assumed to only include the values for the current
    participant ID. 
  }
  \item{...}{
     Any argument used in the previous version of g.impute, which will now
     be used to overrule the arguments specified with the parameter objects.
  }
}
\value{
  \item{metashort}{imputed short epoch variables}
  \item{rout}{matrix to clarify when data was imputed for each long epoch time window
  and the reason for imputation. Value = 1 indicates imputation. 
  Columns 1 = monitor non wear, column 2 = clipping, column 3 = additional nonwear,
  column 4 = protocol based exclusion and column5 = sum of column 1,2,3 and 4. }
  \item{averageday}{matrix with n columns for n metrics values and m rows for
  m short epoch time windows in an average 24 hours period}
}
\examples{
  \dontrun{
    #inspect file:
    I = g.inspectfile(datafile)
    #autocalibration:
    C = g.calibrate(datafile) 
    #get meta-data:
    M = g.getmeta(datafile)
  }
  data(data.getmeta)
  data(data.inspectfile)
  #impute meta-data:
  IMP = g.impute(M=data.getmeta, I=data.inspectfile)
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{identify_levels}
\alias{identify_levels}
\title{
  Identifies levels of behaviour for g.part5 function.
}
\description{
  Identifies levels of behaviour from acceleratioon
  and sustained inactivity sibdetection (using angles). Function not
  intended for direct use by package user.
}
\usage{
  identify_levels(ts, TRLi,TRMi,TRVi,
                  ws3, params_phyact, ...)
}
\arguments{
  \item{ts}{
    Data.frame with time series genrated in .gpart5
  }
  \item{TRLi}{
    Numeric acceleration threshold light
  }
  \item{TRMi}{
    Numeric acceleration threshold moderate
  }
  \item{TRVi}{
    Numeric acceleration threshold vigorous
  }
  \item{ws3}{
    Numeric size of epoch in seconds
  }
  \item{params_phyact}{
    See \link{g.part2}
  }
  \item{...}{
     Any argument used in the previous version of identify_level, which will now
     be used to overrule the arguments specified with the parameter objects.
  }
}
\value{
  List with items:
  \itemize{
    \item LEVELS
    \item OLEVELS
    \item Lnames
    \item bc.mvpa
    \item bc.lig
    \item bc.in
    \item ts
  }
}
\examples{
  \dontrun{
    levels = identify_levels(TRLi,TRMi,TRVi,
                               boutdur.mvpa,boutcriter.mvpa,
                               boutdur.lig,boutcriter.lig,
                               boutdur.in,boutcriter.in,
                               ws3,bout.metric)
  }
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.intensitygradient}
\alias{g.intensitygradient}
\title{
Intensity gradient calculation
}
\description{
Calculates the intensity gradient based on Rowlands et al. 2018.
The function assumes that the user has already calculated the value
distribution.
}
\usage{
g.intensitygradient(x,y)
}
\arguments{
  \item{x}{
  Numeric vector of mid-points of the bins (mg)
  }
  \item{y}{
  Numeric vector of time spent in bins (minutes)
  }
}
\value{
 \item{y_intercept}{y-intercept of a linear regression line in log-log space}
 \item{gradient}{Beta coefficient of a linear regression line in log-log space}
 \item{rsquared}{R squared of x and y values in log-log space}
}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}
\keyword{internal}
\references{
  Rowlands A, Edwardson CL, et al. (2018) Beyond Cut Points: Accelerometer Metrics 
  that Capture the Physical Activity Profile. MSSE 50(6):1. 
  doi:10.1249/MSS.0000000000001561
}

\name{isfilelist}
\alias{isfilelist}
\title{
Checks whether datadir is a directory or a vector with
filenames
}
\description{
Checks whether argument datadir used in various other functions in 
GGIR is the name of a directory that includes data files or whether
it is a vector with the full paths to one or more data files}
\usage{
isfilelist(datadir)	
}
\arguments{
  \item{datadir}{
Argument datadir as used in various other functions in GGIR
  }
}
\keyword{internal}
\value{
Boolean whether it is a list of files (TRUE) or not (FALSE)
}
  
\examples{
\dontrun{
isitafilelist = isfilelist(datadir)
}
}

\name{aggregateEvent}
\alias{aggregateEvent}
\title{
  Aggregates event data as produced by external function
}
\description{
  Aggregates event data as produced by external function
}
\usage{
   aggregateEvent(metric_name, epochsize, daysummary, ds_names,
                          fi, di, vari, segmentInfo, myfun = NULL,
                          params_247)
}
\arguments{
  \item{metric_name}{
    Character with name of metric
  }
  \item{epochsize}{
    Numeric epochsize in seconds of the timeseries
  }
  \item{daysummary}{
    Matrix with prelimenary day summary
  }
  \item{ds_names}{
    Daysummary column names
  }
  \item{fi}{
    Column index of daysummary
  }
  \item{di}{
    Row index of daysummary representing the recording day
  }
  \item{vari}{
    Time series derived from metashort
  }
  \item{segmentInfo}{
    List passed on from \link{g.analyse.perday} with time series indices of 
    segment to analyse segment name and segment number.
  }
  \item{myfun}{
    See \link{GGIR}
  }
  \item{params_247}{
    See \link{GGIR}
  }
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.part4_extractid}
\alias{g.part4_extractid}
\title{
 Extracts ID from filename and finds matching rows in sleeplog
}
\description{
  Extracts ID from filename and finds matching rows in sleeplog. Function not
  designed for direct use by GGIR users.
}
\usage{
  g.part4_extractid(idloc, fname, dolog, sleeplog, accid = c())
}
\arguments{
  \item{idloc}{
    See \link{g.part4}
  }
  \item{fname}{
    Full patth to filename
  }
  \item{dolog}{
    Boolean to indicate whether to rely on a sleeplog
  }
  \item{sleeplog}{
    Sleeplog data.frame passed on from g.part4
  }
  \item{accid}{
    ID extracted from the acceleration file in GGIR part3. If not available leave blank.
  }
}
\value{
 List with accid the ID and matching_indices_sleeplog a vector with matching row
 indices in the sleeplog
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{g.plot}
\alias{g.plot}
\title{
function to generate a plot for quality check purposes
}
\description{
Function takes meta-data as generated by \link{g.getmeta} and \link{g.impute}
to create a visual representation of imputed time periods
}
\usage{
g.plot(IMP, M, I, durplot)
}
\arguments{
  \item{IMP}{
  output from \link{g.impute}
}
  \item{M}{
  output from \link{g.getmeta}
}
  \item{I}{
  output from \link{g.inspectfile}
}
  \item{durplot}{
number of days to plot
}
}
\value{
function only produces a plot, no values
}
\examples{
\dontrun{
  #inspect file:
  I = g.inspectfile(datafile)
  
  #autocalibration:
  C = g.calibrate(datafile) 
  
  #get meta-data:
  M = g.getmeta(datafile)
}
data(data.getmeta)
data(data.inspectfile)

#impute meta-data:
IMP = g.impute(M = data.getmeta, I = data.inspectfile, strategy = 1,
hrs.del.start = 0, hrs.del.end = 0, maxdur = 0)

#plot data
g.plot(IMP, M = data.getmeta, I = data.inspectfile, durplot=4)
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}

\name{iso8601chartime2POSIX}
\alias{iso8601chartime2POSIX}
\title{
Convert iso8601 timestamps to POSIX timestamp
}
\description{
To avoid ambiguities when sharing and comparing timestamps. All timestamps
are expressed in iso8601 format: https://en.wikipedia.org/wiki/ISO_8601
However, to generate plots in R we need to convert them back to POSIX
}
\usage{
iso8601chartime2POSIX(x,tz)
}
\arguments{
  \item{x}{
Vector of timestamps in iso8601 in character format
}
  \item{tz}{
 Timezone of data collection, e.g. "Europe/London". 
 See List_of_tz_database_time_zones on Wikipedia
 for full list.
}
}
\examples{
x ="2017-05-07T13:00:00+0200"
tz = "Europe/Amsterdam"
x_converted = iso8601chartime2POSIX(x,tz)
}

\name{g.part5.analyseRest}
\alias{g.part5.analyseRest}
\title{
  Analyse rest (internal function)
}
\description{
  Analyses overlap self-reported napping, non-wear and sib.
  Internal function not intended for direct use by GGIR end-user.
}
\usage{
  g.part5.analyseRest(sibreport = NULL, dsummary = NULL,
                                ds_names = NULL, fi = NULL,
                                di = NULL, ts = NULL, tz = NULL,
                                params_sleep = NULL)
}
\arguments{
  \item{sibreport}{
    sibreport data.frame produced by \link{g.sibreport}
  }
  \item{dsummary}{
    matrix created internally by \link{g.part5}
  }
  \item{ds_names}{
    character vector with variable names corresponding to dsummary
    created internally by \link{g.part5}
  }
  \item{fi}{
    Numeric scalar to indicate variable index, created internally by \link{g.part5}
  }
  \item{di}{
    Numeric scalar to indicate recording index, created internally by \link{g.part5}
  }
  \item{ts}{
    Data.frame with time series object passed on from \link{g.part5_analyseSegment} and \link{g.part5},
  }
  \item{tz}{
    Timezone database name
  }
  \item{params_sleep}{
    See \link{GGIR}
  }
}
\value{
  List with updated objects dsummary, ds_names, fi, and di
}
\keyword{internal}
\author{
  Vincent T van Hees <v.vanhees@accelting.com>
}
